/*
 * geode.c - geode graphics acceleration functions.
 * 
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 03/29/01  Damian Kowalewski
 */

/* include some needed fgl modules */

FGL_GLOBAL char *geode_reg_base;
FGL_GLOBAL char *geode_blt0_base;
FGL_GLOBAL char *geode_blt1_base;
FGL_GLOBAL char *geode_buf_base;

typedef struct _fgl_hardware_sprite_t {
        fgl_color_t    color1;
        fgl_color_t    color2;
        u8             surface[256];
} fgl_hardware_sprite_t;

#define GEODE_TIMEOUT   100000
#define GEODE_SKIP_TEST 30

#define GEODE_MEM_BASE  0x40000000
#define GEODE_MEM_SIZE  0x1000000

#define GEODE_REG_BASE  0x8100
#define GEODE_BLT0_BASE 0x0930
#define GEODE_BLT0_SIZE 0x07d0
#define GEODE_BLT1_BASE 0x0400
#define GEODE_BLT1_SIZE 0x0530
#define GEODE_BUF_BASE  0x800000

#define GP_DST_XY       0x000
#define GP_WIDTH_HEIGHT 0x004
#define GP_SRC_XY       0x008
#define GP_SRC_COLOR0   0x00c
#define GP_SRC_COLOR1   0x00e
#define GP_PAT_COLOR0   0x010
#define GP_PAT_COLOR1   0x012
#define GP_PAT_COLOR2   0x014
#define GP_PAT_COLOR3   0x016
#define GP_PAT_DATA0    0x020
#define GP_PAT_DATA1    0x024
#define GP_PAT_DATA2    0x028
#define GP_PAT_DATA3    0x02c
#define GP_RASTERMODE   0x100
#define GP_VECTORMODE   0x104
#define GP_BLTMODE      0x108
#define GP_BLTSTATUS    0x10c

#define WAIT_BLT_PENDING   0x0004
#define WAIT_PIPELINE_BUSY 0x0002
#define WAIT_BUFFER_BUSY   0x0001

/*
 * define the operation types for accelerated functions
 */

#define FGL_ACCEL_DRAWLINE_OP 0x54
#define FGL_ACCEL_MASKBLT_OP  0x76
#define FGL_ACCEL_MONOBLT_OP  0x76
#define FGL_ACCEL_FILLRECT_OP 0x54
#define FGL_ACCEL_MOVEAREA_OP 0x74

/*
 * redefine the fgl general operation type
 */
#undef FGL_FRAMEBUFFER_OP
#define FGL_FRAMEBUFFER_OP 0x01


#define FGL_ACCELFN_WAIT
static inline
int fgl_accel_wait( fgl_bitmap_t* bmp ) {

  int w;
  
  u32 test = WAIT_BLT_PENDING | WAIT_PIPELINE_BUSY | WAIT_BUFFER_BUSY;
  
  for(w=0; w < GEODE_TIMEOUT; w++ ) {
    if ( w % GEODE_SKIP_TEST == 0 ) {
      if ( !((*(unsigned long *)(geode_reg_base + GP_BLTSTATUS)) & test) ) {
	return 0;
      }
    }
  }
  return 1;
}

#define FGL_ACCELFN_PENDING
static inline
int fgl_accel_pending( fgl_bitmap_t* bmp ) {

  int w;
  
  u32 test = (bmp->lastOperation >> 4 ) & bmp->nextOperation;
  
  for(w=0; w < GEODE_TIMEOUT; w++ ) {
    if ( w % GEODE_SKIP_TEST == 0 ) {
      if ( !((*(unsigned long *)(geode_reg_base + GP_BLTSTATUS)) & test) ) {
	return 0;
      }
    }
  }
  return 1;
}

#define FGL_ACCELFN_BLT_PENDING
static inline
int fgl_accel_blt_pending() {

  int w;
  
  for(w=0; w < GEODE_TIMEOUT; w++ ) {
    if ( w % GEODE_SKIP_TEST == 0 ) {
      if ( !((*(unsigned long *)(geode_reg_base + GP_BLTSTATUS)) & WAIT_BLT_PENDING) ) {
	return 0;
      }
    }
  }
  return 1;
}

#define FGL_ACCELFN_PIPELINE_BUSY
static inline
int fgl_accel_pipeline_busy() {

  int w;
  
  for(w=0; w < GEODE_TIMEOUT; w++ ) {
    if ( w % GEODE_SKIP_TEST == 0 ) {
      if ( !((*(unsigned long *)(geode_reg_base + GP_BLTSTATUS)) & WAIT_PIPELINE_BUSY) ) {
	return 0;
      }
    }
  }
  return 1;
}

#define FGL_ACCELFN_BUFFER_BUSY
static inline
int fgl_accel_buffer_busy() {

  int w;
  
  for(w=0; w < GEODE_TIMEOUT; w++ ) {
    if ( w % GEODE_SKIP_TEST == 0 ) {
      if ( !((*(unsigned long *)(geode_reg_base + GP_BLTSTATUS)) & WAIT_BUFFER_BUSY) ) {
	return 0;
      }
    }
  }
  return 1;
}



#define FGL_ACCELFN_INIT
static inline
int fgl_accel_init() {

  int mem;
  char *temp;

  mem = open("/dev/mem",O_RDWR);

  temp = (char *)mmap((unsigned char *)GEODE_MEM_BASE, GEODE_MEM_SIZE,
		      PROT_READ |PROT_WRITE, MAP_SHARED,mem,GEODE_MEM_BASE);

  if (mem == -1) return 1;
  
  geode_reg_base  = temp + GEODE_REG_BASE;
  geode_blt0_base = temp + GEODE_BLT0_BASE;
  geode_blt1_base = temp + GEODE_BLT1_BASE;
  geode_buf_base  = temp + GEODE_BUF_BASE;

  close(mem);
  //  fprintf(stderr, "***** geode debug: fgl_accel_init()\n");
  return 0;
}

#define FGL_ACCELFN_FILLRECT
static inline
void fgl_accel_fillRect(fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1, fgl_color_t clr ) {

  int dx, dy;
  
  dx = x1 - x0 + 1;
  dy = y1 - y0 + 1;

  x0 += ctx->xoff;
  y0 += ctx->yoff;

  (*(unsigned long *)(geode_reg_base + GP_PAT_COLOR0))   = (int)clr;
  (*(unsigned long *)(geode_reg_base + GP_DST_XY))       = ((short)y0 << 16) + (short)x0;
  (*(unsigned long *)(geode_reg_base + GP_WIDTH_HEIGHT)) = ((short)dy << 16) + (short)dx;
  (*(unsigned long *)(geode_reg_base + GP_RASTERMODE))   = 0xf0;
  (*(unsigned long *)(geode_reg_base + GP_BLTMODE))      = 0x0;

  //  fprintf(stderr, "***** geode debug: fgl_accel_fillRect()\n");

}

#define FGL_ACCELFN_MOVEAREA
static inline
void fgl_accel_moveArea( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1, i32 dx, i32 dy ) {

  i32 u0, v0, w0 ,h0;
  int mode = 0x001;

  x0 += ctx->xoff;
  y0 += ctx->yoff;
  x1 += ctx->xoff;
  y1 += ctx->yoff;

  u0 = x0 + dx;
  v0 = y0 + dy;
  w0 = x1 - x0 + 1;
  h0 = y1 - y0 + 1;

  if (dy > 0 ) {
    mode |= 0x100;
    y0 += h0;
    v0 += h0;
  }
   
  fgl_accel_buffer_busy();
  (*(unsigned long *)(geode_reg_base + GP_DST_XY))       = ((short)v0 << 16) + (short)u0;
  (*(unsigned long *)(geode_reg_base + GP_SRC_XY))       = ((short)y0 << 16) + (short)x0;
  (*(unsigned long *)(geode_reg_base + GP_WIDTH_HEIGHT)) = ((short)h0 << 16) + (short)w0;
  (*(unsigned long *)(geode_reg_base + GP_RASTERMODE))   = 0xcc;
  (*(unsigned long *)(geode_reg_base + GP_BLTMODE))      = mode;

  //  fprintf(stderr, "***** geode debug: fgl_accel_moveArea()\n");
}

#define FGL_ACCELFN_DRAWLINE
static inline
void fgl_accel_drawLine( fgl_context_t *ctx,  i32 x0, i32 y0, i32 x1, i32 y1 ) {

  int di,d1,d2,dx,dy,temp,raster,status = 0;

  x0 += ctx->xoff;
  y0 += ctx->yoff;
  x1 += ctx->xoff;
  y1 += ctx->yoff;

  dx = ( x1 - x0 ); dy = ( y1 - y0 );
  
  if (dx*dx < dy*dy) {
    temp = dx;
    dx = dy;
    dy = temp;
    status |= 0x1;
  }

  if (dx >= 0 ) status |= 0x2;
  else dx *= -1;

  if (dy >= 0 ) status |= 0x4;  
  else dy *= -1;

  di = 2 * dy - dx;
  d1 = 2 * dy;
  d2 = 2 * ( dy - dx);

  if (ctx->mode == PAINT_MODE) raster = 0xf0;
  else {
    raster  = 0x5a;
    status |= 0x08;
  }

  (*(unsigned long *)(geode_reg_base + GP_PAT_COLOR0))   = (int)(ctx->clr);
  (*(unsigned long *)(geode_reg_base + GP_DST_XY))       = ((short)y0 << 16) + (short)x0; 
  (*(unsigned long *)(geode_reg_base + GP_SRC_XY))       = ((short)d2 << 16) + (short)d1;
  (*(unsigned long *)(geode_reg_base + GP_WIDTH_HEIGHT)) = ((short)di << 16) + (short)dx;
  (*(unsigned long *)(geode_reg_base + GP_RASTERMODE))   = raster;
  (*(unsigned long *)(geode_reg_base + GP_VECTORMODE))   = status;

  //  fprintf(stderr, "***** geode debug: fgl_accel_drawLine()\n");
 
}

#define FGL_ACCELFN_DRAWHLINE
static inline
void fgl_accel_drawHLine( fgl_context_t *ctx,  i32 x0, i32 y0, i32 x1 ) {

  int d2, dx, raster, status = 0;

  x0 += ctx->xoff;
  y0 += ctx->yoff;
  x1 += ctx->xoff;

  dx = ( x1 - x0 ); 

  status |= 0x2;

  d2 = 2 * -dx;

  if (ctx->mode == PAINT_MODE) raster = 0xf0;
  else {
    raster  = 0x5a;
    status |= 0x08;
  }

  (*(unsigned long *)(geode_reg_base + GP_PAT_COLOR0))   = (int)(ctx->clr);
  (*(unsigned long *)(geode_reg_base + GP_DST_XY))       = ((short)y0  << 16) + (short)x0; 
  (*(unsigned long *)(geode_reg_base + GP_SRC_XY))       = ((short)d2  << 16);
  (*(unsigned long *)(geode_reg_base + GP_WIDTH_HEIGHT)) = ((short)-dx << 16) + (short)dx;
  (*(unsigned long *)(geode_reg_base + GP_RASTERMODE))   = raster;
  (*(unsigned long *)(geode_reg_base + GP_VECTORMODE))   = status;

  //  fprintf(stderr, "***** geode debug: fgl_accel_drawHLine()\n");
 
}

#define FGL_ACCELFN_DRAWVLINE
static inline
void fgl_accel_drawVLine( fgl_context_t *ctx,  i32 x0, i32 y0, i32 y1 ) {

  int d2, dy, raster, status = 0;

  x0 += ctx->xoff;
  y0 += ctx->yoff;
  y1 += ctx->yoff;

  dy = ( y1 - y0 ); 

  status |= 0x3;

  d2 = 2 * -dy;

  if (ctx->mode == PAINT_MODE) raster = 0xf0;
  else {
    raster  = 0x5a;
    status |= 0x08;
  }

  (*(unsigned long *)(geode_reg_base + GP_PAT_COLOR0))   = (int)(ctx->clr);
  (*(unsigned long *)(geode_reg_base + GP_DST_XY))       = ((short)y0  << 16) + (short)x0; 
  (*(unsigned long *)(geode_reg_base + GP_SRC_XY))       = ((short)d2  << 16);
  (*(unsigned long *)(geode_reg_base + GP_WIDTH_HEIGHT)) = ((short)-dy << 16) + (short)dy;
  (*(unsigned long *)(geode_reg_base + GP_RASTERMODE))   = raster;
  (*(unsigned long *)(geode_reg_base + GP_VECTORMODE))   = status;

  //  fprintf(stderr, "***** geode debug: fgl_accel_drawVLine()\n");
 
}

#define FGL_ACCELFN_MONOBLT
static inline
void fgl_accel_monoBlt ( fgl_context_t* dst, i32 u0, i32 v0,
			 fgl_monoBitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, 
			 fgl_color_t clr) {
  
  int dx, dy, t0, off, nx, i, j;
  int nlines, nsize, ntimes, nlength;
  char *sp1, *bp1;

  u0 += dst->xoff;
  v0 += dst->yoff;

  dx  = x1 - x0 + 1;
  dy  = y1 - y0 + 1;
  off = x0 / 8;
  nlength = src->sLength - x0 / 8 - (src->width - x1) / 8;
  nx  = x0 % 8;
  t0  = v0;

  nlines = GEODE_BLT1_SIZE / nlength;
  ntimes = dy / nlines;
  nsize  = nlines * nlength;

  sp1 = src->surface + y0 * src->sLength + off;
  (*(unsigned long *)(geode_reg_base+GP_SRC_COLOR0))   = (int)clr << 16;
  (*(unsigned long *)(geode_reg_base+GP_RASTERMODE))   = 0x9cc;
  (*(unsigned long *)(geode_reg_base+GP_WIDTH_HEIGHT)) = (nlines << 16) + dx;
  (*(unsigned long *)(geode_reg_base+GP_SRC_XY))       = nx;

  for( i = 0 ; i < ntimes ; i++, t0 += nlines ) {
    bp1 = geode_blt0_base;
    fgl_accel_blt_pending();
    fgl_accel_pipeline_busy();
    for( j = 0 ; i < nlines ; i++, sp1 += src->sLength, bp1 += nlength) {
      memcpy(bp1,sp1,nlength);
    }
    (*(unsigned long *)(geode_reg_base+GP_DST_XY))       = ((short)t0 << 16) + u0;
    (*(unsigned long *)(geode_reg_base+GP_BLTMODE))      = 0x42;
  }

  if (t0 < v0 + dy) {
    nlines = v0 + dy - t0;
    bp1 = geode_blt0_base;
    fgl_accel_blt_pending();
    fgl_accel_pipeline_busy();
    for( j = 0 ; i < nlines ; i++, sp1 += src->sLength, bp1 += nlength ) {
      memcpy(bp1,sp1,nlength);
    }
    (*(unsigned long *)(geode_reg_base+GP_WIDTH_HEIGHT)) = (nlines << 16) + dx;
    (*(unsigned long *)(geode_reg_base+GP_DST_XY))       = ((short)t0 << 16) + u0;
    (*(unsigned long *)(geode_reg_base+GP_BLTMODE))      = 0x42;
  }

  // fprintf(stderr, "***** geode debug: fgl_accel_monoBlt()\n");

}

#define FGL_ACCELFN_MASKBLT
static inline
void fgl_accel_maskBlt ( fgl_context_t* dst, i32 u0, i32 v0,
			 fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1 ) {
     
  unsigned short *pd, *ps;
  int i, dx, dy;


  u0 += dst->xoff;
  v0 += dst->yoff;

  dx  = x1 - x0 + 1;
  dy  = y1 - y0 + 1;

  //  fprintf(stderr,"transp: %i\n",src->transp);

  for( i = 0 ; i < GEODE_BLT1_SIZE / 2 ; i++) 
    { 
      *((unsigned short *)(geode_blt1_base) + 1 ) = src->transp; 
    }
  pd = (unsigned short *)geode_buf_base + dst->bmp->sLength * 0x3c0;
  ps = (unsigned short *)src->surface + x0 + src->sLength * y0;

  for( i = 0 ; i < dy ; i++, pd += dst->bmp->sLength, ps += src->sLength) 
    {
      fgl_cpySlc( ps, x0, pd, u0, dx );
    }

  /* one time register setup */

  (*(unsigned long *)(geode_reg_base + GP_PAT_COLOR0))   = 0xffffffff;
  //  (*(unsigned long *)(geode_reg_base + GP_PAT_COLOR1))   = 0xffffffff;

  (*(unsigned long *)(geode_reg_base + GP_SRC_XY))       = 0x2d0;
  (*(unsigned long *)(geode_reg_base + GP_DST_XY))       = 0x2d0;
  (*(unsigned long *)(geode_reg_base + GP_WIDTH_HEIGHT)) = (1 << 16) + 1;

  /*
   * a first 1 pixel dummy blt operation has to be performed
   * to set correctly the registers before attempting the
   * actual transparent blt.
   */
  (*(unsigned long *)(geode_reg_base + GP_RASTERMODE))   = 0x10c6;
  (*(unsigned long *)(geode_reg_base + GP_BLTMODE))      = 0xd;

  /* 
   * unfortunately Geode chip updates the src register so we have
   * to write values once again. 
   */

  (*(unsigned long *)(geode_reg_base + GP_SRC_XY))       = (0x3c0 << 16);
  (*(unsigned long *)(geode_reg_base + GP_DST_XY))       = (v0 << 16) + u0;

  (*(unsigned long *)(geode_reg_base + GP_WIDTH_HEIGHT)) = (dy << 16) + dx;
  (*(unsigned long *)(geode_reg_base + GP_BLTMODE))      = 0x1;

}

static inline
fgl_color_t fgl_hardware_getColor16 ( u32 argb ) /*{{{*/
{
	return (((argb & 0xf80000) >> 8) |
			((argb & 0xfc00) >> 5) |
			((argb & 0xf8) >> 3));
}
/*}}}*/

static inline
void  fgl_hardware_createSprite ( fgl_spriteData_t *spd, fgl_sprite_t* sp, 
				  fgl_hardware_sprite_t *hsp )
{
  char cursor_and[4*32];
  char cursor_xor[4*32];
  char *cc,*ca,*cx;
  int i,x,y;
  int color1;
  int color2;

  /* set all the pixels in the cursor transparent */
  for (i=0 ;i< 4*32; i ++ ) cursor_and[i] = 0xff;
  for (i=0 ;i< 4*32; i ++ ) cursor_xor[i] = 0x00;

  /* figure out color1, color2 and transparent */
  color1 = color2 = -1;
  for ( i = 0; i < spd->nColors; i++ ) {
    if (i != spd->transp && color1 == -1) color1 = i;
    if (i != spd->transp && color1 != -1 && color2 == -1) color2 = i;
  }
  /* 
   * we do not have the fgl_getPixelValue so we have to
   * do it ourselves here.
   */

  hsp->color1 = fgl_hardware_getColor16( spd->cmap[color1] );
  hsp->color2 = fgl_hardware_getColor16( spd->cmap[color2] );

  for (y = 0; y < spd->height ; y++ ) {
    for (x = 0; x < spd->width ; x++ ) {

      /* test if it is not transparent */
      if (spd->data[spd->width*y+x] != spd->transp ) {
	/* assigned pixels to color 1 */
	if (spd->data[spd->width*y+x] == color1) {
	  cursor_and[y*4 + x/8] &= ~(0x1 << (7 - x%8));
	}
	if (spd->data[spd->width*y+x] == color2) {
	  cursor_and[y*4 + x/8] &= ~(0x1 << (7 - x%8)); 
	  cursor_xor[y*4 + x/8] |= 0x1 << (7 - x%8); 
	}
      }      
    }
  }

  /* write geode native data to extension */

  cc = hsp->surface;

  ca = cursor_and;
  cx = cursor_xor;

  for(i = 0; i < 32 ; i ++ ) {
    *cc++ = *(cx+1);
    *cc++ = *cx++;
    cx++;
    *cc++ = *(ca+1);
    *cc++ = *ca++;
    ca++;

    *cc++ = *(cx+1);
    *cc++ = *cx++;
    cx++;
    *cc++ = *(ca+1);
    *cc++ = *ca++;
    ca++;
  }
  /* set the extension in the sprite */
  sp->ext = (void *)hsp;

}


