/*
 * cbmp.c - fgl module to handle framebuffer-depth bitmaps
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: cbmp.c,v 1.32 2001/08/16 01:36:35 samc Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_CBMP_INCLUDED
#define FGL_CBMP_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* these are our fgl imports */
#include "util.c"
#include FGL_DRAW_MODULE
#include "mbmp.c"
#include "cntx.c"
#include FGL_CMAP_MODULE
#include FGL_HEAP_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif


#ifdef FGL_INLINE_CBMP
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif
/*}}}*/


static inline
fgl_monoBitmap_t* fgl_createBitmapMask ( fgl_bitmap_t* bmp ) /*{{{*/
{
	return fgl_createMonoBitmap( bmp->width, bmp->height, 0xff);
}
/*}}}*/

static inline
void fgl_setBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y, fgl_color_t clr ) /*{{{*/
{
	fgl_slc_t *p;

	if ( (x < bmp->width) && (y < bmp->height) ){
#ifdef ROT_90
	        rotate32( bmp, &x, &y);
#endif
		p = fgl_getSlc( bmp, x, y);
		fgl_setSlcPixel( p, x, clr);		
	}
}
/*}}}*/

static inline
void fgl_setBmpTransparency ( fgl_bitmap_t* bmp, fgl_color_t clr ) /*{{{*/
{
       bmp->transp = clr;
}
/*}}}*/

static inline
fgl_color_t fgl_getBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y ) /*{{{*/
{
	fgl_slc_t    *p;

	if ( (x < bmp->width) && (y < bmp->height) ){
#ifdef ROT_90
	        rotate32( bmp, &x, &y);
#endif
		p = fgl_getSlc( bmp, x, y);
		return fgl_getSlcPixel( p, x);
	}
	else {
		return 0;
	}
}
/*}}}*/

static inline
void fgl_clear ( fgl_context_t* ctx ) /*{{{*/
{
        fgl_beginDraw( ctx->bmp, 0, 0, ctx->bmp->width-1,
                       ctx->bmp->height-1, FGL_FRAMEBUFFER_OP );  

        fgl_initBmp( ctx->bmp, ctx->clr);

        fgl_endDraw( ctx->bmp, 0, 0, ctx->bmp->width-1,
                     ctx->bmp->height-1 );  
}
/*}}}*/

static inline
void fgl_setPixel ( fgl_context_t* ctx, u32 x, u32 y, fgl_color_t clr ) /*{{{*/
{
	fgl_slc_t *p;

	if ( (x >= ctx->x0Clip) && (x <= ctx->x1Clip) &&
		 (y >= ctx->y0Clip) && (y <= ctx->y1Clip) ){
#ifdef ROT_90
	        rotate32( ctx->bmp, &x, &y);
#endif
	        p = fgl_getSlc( ctx->bmp, x, y);
		fgl_setSlcPixel( p, x, clr);
	}
}
/*}}}*/

FUNC_ATTR
fgl_bitmap_t* fgl_createBitmap ( u32 width, u32 height, i32 clr ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_CBMP)
{
	u32 size = sizeof(fgl_bitmap_t) + (width * fgl_getSlcSize() * height);
	fgl_bitmap_t *bmp = (fgl_bitmap_t*) FGL_MALLOC( size);

	if ( !bmp ){
		return 0;
	}
	
	bmp->width  = width;
	bmp->height = height;
#ifdef ROT_90
	bmp->sLength = fgl_getScanlineLength ( height );
#else
	bmp->sLength = fgl_getScanlineLength ( width );
#endif

	bmp->surface = (fgl_slc_t*)((u8*)bmp + sizeof(fgl_bitmap_t));

	if ( clr != -1 ) {
		fgl_initBmp( bmp, clr);
	}

	bmp->lastOperation = FGL_FRAMEBUFFER_OP;
	bmp->nextOperation = FGL_FRAMEBUFFER_OP;

	return bmp;
}
#endif /* !FGL_PROTO */
; /*}}}*/

static inline
void fgl_disposeBitmap ( fgl_bitmap_t* bmp ) /*{{{*/
{
	FGL_FREE( bmp);
}
/*}}}*/

FUNC_ATTR
void fgl_moveArea ( fgl_context_t* ctx,
					i32 x0, i32 y0, i32 x1, i32 y1, i32 dx, i32 dy ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_CBMP)
{
	i32 u0, v0, u1, v1, w;
	fgl_slc_t *p, *q, *pe;

	if ( !(dx | dy) )
		return;

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	u0 = (i32)x0 + dx;
	v0 = (i32)y0 + dy;

	if ( !clipBlt( ctx, &u0, &v0, &u1, &v1,
				   ctx->bmp->width, ctx->bmp->height, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90
	rotate32( ctx->bmp, (u32*)&x0, (u32*)&y0);
	rotate32( ctx->bmp, (u32*)&x1, (u32*)&y1);
	rotate32( ctx->bmp, (u32*)&u0, (u32*)&v0);
	rotate32( ctx->bmp, (u32*)&u1, (u32*)&v1);
	swap32( &x0, &x1);
	swap32( &u0, &u1);
#endif

#ifdef FGL_ACCELFN_MOVEAREA
	if (ctx->bmp->state & IMAGE_ON_SCREEN) {
	  fgl_beginDraw( ctx->bmp, u0, v0, u1, v1, FGL_ACCEL_MOVEAREA_OP );
	  fgl_accel_moveArea( fgl_screenCtx, x0, y0, x1, y1, dx, dy );
	  fgl_endDraw( ctx->bmp, u0, v0, u1, v1);
	  return;
	}
#endif

	w = x1 - x0 + 1;

	fgl_beginDraw( ctx->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	if ( v1 > y1 ){
		/* move down, start with bottom scanline */
		p  = fgl_getSlc( ctx->bmp, x0, y1);
		pe = fgl_getSlc( ctx->bmp, x0, y0);
		q  = fgl_getSlc( ctx->bmp, u0, v1);

		while ( p >= pe ){
			fgl_cpySlc( p, x0, q, u0, w);

			fgl_subSlc( &p, x0, ctx->bmp->sLength);
			fgl_subSlc( &q, u0, ctx->bmp->sLength);
		}
	}
	else {
		/* start with top scanline */
		p  = fgl_getSlc( ctx->bmp, x0, y0);
		pe = fgl_getSlc( ctx->bmp, x0, y1);
		q  = fgl_getSlc( ctx->bmp, u0, v0);

		/* non-overlapping scanline copy */
		if ( (v1 < y1) || (x1 < u0) || (u1 < x0) ){
			while ( p <= pe ){
				fgl_cpySlc( p, x0, q, u0, w);

				fgl_addSlc( &p, x0, ctx->bmp->sLength);
				fgl_addSlc( &q, u0, ctx->bmp->sLength);
			}
		}
		else {
			while ( p <= pe ){
				fgl_moveSlc( p, x0, q, u0, w);

				fgl_addSlc( &p, x0, ctx->bmp->sLength);
				fgl_addSlc( &q, u0, ctx->bmp->sLength);
			}
		}
	}

	fgl_endDraw( ctx->bmp, u0, v0, u1, v1);
}
#endif /* !FGL_PROTO */
; /*}}}*/

FUNC_ATTR
void fgl_bitBlt ( fgl_context_t* dst, i32 u0, i32 v0,
				  fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_CBMP)
{
	i32 u, x, u1, v1, w;
	fgl_slc_t *ps, *pse, *pd;
	u32 ds, dd;
	fgl_color_t pix;

	if ( src == dst->bmp ){
		fgl_moveArea( dst, x0, y0, x1, y1, u0-x0, v0-y0);
		return;
	}

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90
	rotate32( src, (u32*)&x0, (u32*)&y0);
	rotate32( src, (u32*)&x1, (u32*)&y1);
	rotate32( dst->bmp, (u32*)&u0, (u32*)&v0);
	rotate32( dst->bmp, (u32*)&u1, (u32*)&v1);
	swap32( &x0, &x1);
	swap32( &u0, &u1);
#endif
	w = x1 - x0 + 1;

	ps  = fgl_getSlc( src, x0, y0);
	pse = fgl_getSlc( src, x0, y1);
	pd  = fgl_getSlc( dst->bmp, u0, v0);

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	if ( dst->alpha == 0xff ){
		/* this is a opaque blit, delegate as much as possible to the driver */
		while ( ps <= pse ){
			fgl_cpySlc( ps, x0, pd, u0, w);
			fgl_addSlc( &ps, x0, src->sLength);
			fgl_addSlc( &pd, u0, dst->bmp->sLength);
		}
	}
	else {
		/* use the context's alpha value to blend this in */
		ds  = src->sLength - x1 + x0 - 1;
		dd  = dst->bmp->sLength - u1 + u0 - 1; 

		while ( ps <= pse ){
			for ( u=u0, x=x0; u<=u1; u++, x++ ){
				pix = fgl_blendPixelValues( dst->alpha,
							    fgl_getSlcPixel( pd, u),
							    fgl_getSlcPixel( ps, x));
				
				fgl_setSlcPixel( pd, u, pix);

				fgl_incSlc( &pd, u);
				fgl_incSlc( &ps, x);
			}

			fgl_addSlc( &pd, u1, dd);
			fgl_addSlc( &ps, x1, ds);
		}
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}
#endif /* !FGL_PROTO */
; /*}}}*/


FUNC_ATTR
void fgl_maskBlt ( fgl_context_t* dst, i32 u0, i32 v0,
				   fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1,
				   fgl_monoBitmap_t* mask ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_CBMP)
{
	u8 *pm;
	fgl_slc_t *pd, *ps;
	fgl_color_t clr;
	i32 x, u, u1, v1, sinc, dinc, minc;
	i32 w, h;
	register u8 a, i;
	u8 i0, off;

	if ( (mask->width < src->width) || (mask->height < src->height) ){
		/* illegal! */
		return;
	}

	if ( x0 > x1 )  swap32( &x0, &x1);
	if ( y0 > y1 )  swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90
	rotate32( src, (u32*)&x0, (u32*)&y0);
	rotate32( src, (u32*)&x1, (u32*)&y1);
	rotate32( dst->bmp, (u32*)&u0, (u32*)&v0);
	rotate32( dst->bmp, (u32*)&u1, (u32*)&v1);
	swap32( &x0, &x1);
	swap32( &u0, &u1);
#endif


#ifdef FGL_ACCELFN_MASKBLT
	if ( dst->bmp->state & IMAGE_ON_SCREEN) {
	  fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_ACCEL_MASKBLT_OP );
	  fgl_accel_maskBlt( fgl_screenCtx, u0, v0, src, x0, y0, x1, y1 );
	  fgl_endDraw( dst->bmp, u0, v0, u1, v1);
	  return;
	}
#endif

	w = x1 - x0;
	h = y1 - y0;
	off = x0 % 8;

	pm = mask->surface + (y0 * mask->sLength) + (x0 / 8);
	pd = fgl_getSlc( dst->bmp, u0, v0);
	ps = fgl_getSlc( src, x0, y0);

	a = *pm;
	u = u0;
	x = x0;
	i = i0 = 8-off;

	dinc = (dst->bmp->sLength - u1) + u0;               /* in pixels */
	sinc = (src->sLength - x1) + x0;                    /* in pixels */
	minc = mask->sLength + (x0>>3) - (x1>>3);           /* in bytes */
	
	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( h-- >= 0 ){
		a <<= off;

		while ( 1 ){
			if ( a & 0x80 ){
				clr = fgl_getSlcPixel( ps, x);
				fgl_setSlcPixel( pd, u, clr);
			}

			if ( u == u1 )
				break;

			if ( --i == 0 ){
				/* next source byte */
				i = 8;
				a = *(++pm);
			}
			else {
				a <<= 1;
			}

			fgl_incSlc( &pd, u);
			u++;
			fgl_incSlc( &ps, x);
			x++;
		}

		/* next source and destination scanlines */
		fgl_addSlc( &pd, u, dinc);
		fgl_addSlc( &ps, x, sinc);

		pm += minc;
		a = *pm;
		u = u0;
		x = x0;
		i = i0;
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}
#endif /* FGL_PROTO */
; /*}}}*/

FUNC_ATTR
void fgl_alphaMaskBlt ( fgl_context_t* dst, i32 u0, i32 v0,
						fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1,
						fgl_a8Bitmap_t* mask) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_CBMP)
{
	i32 u1, v1, u, x, da, ds, dd;
	fgl_slc_t *ps, *pse, *pd;
	u8 *pa;
	fgl_color_t pix;

	if ( (mask->width < src->width) || (mask->height < src->height) ){
		/* illegal! */
		return;
	}

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90
	rotate32( src, (u32*)&x0, (u32*)&y0);
	rotate32( src, (u32*)&x1, (u32*)&y1);
	rotate32( dst->bmp, (u32*)&u0, (u32*)&v0);
	rotate32( dst->bmp, (u32*)&u1, (u32*)&v1);
	swap32( &x0, &x1);
	swap32( &u0, &u1);
#endif

	ps  = fgl_getSlc( src, x0, y0);
	pse = fgl_getSlc( src, x0, y1);
	ds  = src->sLength - x1 + x0 -1;

	pd  = fgl_getSlc( dst->bmp, u0, v0);
	dd  = dst->bmp->sLength - u1 + u0 - 1; 

	pa  = mask->surface;
	da  = mask->sLength - x1 - 1;

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( ps <= pse ){
		for ( u=u0, x=x0; u<=u1; u++, x++ ){
			pix = fgl_blendPixelValues( *pa,
						    fgl_getSlcPixel( pd, u),
						    fgl_getSlcPixel( ps, x));
			fgl_setSlcPixel( pd, u, pix);

			fgl_incSlc( &pd, u);
			fgl_incSlc( &ps, x);
			pa++;
		}

		pa += da;
		fgl_addSlc( &pd, u1, dd);
		fgl_addSlc( &ps, x1, ds);
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}
#endif /* !FGL_PROTO */
; /*}}}*/


/*****************************************************************************
 * convenience functions
 */

static inline
void fgl_bitBltBmp ( fgl_context_t* dst, i32 u0, i32 v0, fgl_bitmap_t* src ) /*{{{*/
{
	fgl_bitBlt( dst, u0, v0, src, 0, 0, src->width-1, src->height-1);
}
/*}}}*/

static inline
void fgl_maskBltBmp ( fgl_context_t* dst, i32 u0, i32 v0,
					  fgl_bitmap_t* src, fgl_monoBitmap_t* mask ) /*{{{*/
{
	fgl_maskBlt( dst, u0, v0, src, 0, 0, src->width-1, src->height-1, mask);
}
/*}}}*/

static inline
void fgl_alphaMaskBltBmp ( fgl_context_t* dst, i32 u0, i32 v0,
						   fgl_bitmap_t* src, fgl_a8Bitmap_t* mask ) /*{{{*/
{
        fgl_alphaMaskBlt( dst, u0, v0, src, 0, 0, src->width-1, src->height-1, mask);
}
/*}}}*/

static inline
void fgl_saveArea ( fgl_bitmap_t* dst, i32 u0, i32 v0,
                    fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
{
    fgl_context_t dstCtx;

    fgl_setBmp( &dstCtx, dst);
	fgl_setAlpha( &dstCtx, 0xff);

    fgl_bitBlt( &dstCtx, u0, v0, src, x0, y0, x1, y1);
}
/*}}}*/

/* fgl module epilog */ /*{{{*/
#undef FUNC_ATTR

#endif /* FGL_CBMP_INCLUDED */
/*}}}*/

