/* -*- mode: C; mode: fold -*- */
/*
 * feng_bin.c - fgl module to implement a binary font file (*.fnt)
 *              font engine
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: feng_bin.c,v 1.22 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/

#ifndef FGL_FENG_INCLUDED
#define FGL_FENG_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* initialize locale specific stuff */
#if defined(LOCALE_INT)
#include "locale.h"
#endif

/* these are our fgl imports */
#include "util.c"
#include "mbmp.c"

#include FGL_HEAP_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif
/*}}}*/

#define IS_FILEMSB   0x01;
#define IS_ANTIALIAS 0x02;
#define IS_ROTATED   0x04;

static char *fntDir;
static int delta[] = { 0, -1, 1, -2, 2 };


int fgl_initFontEngine () /*{{{*/
{
	if ( !(fntDir = getenv( "FGL_FONTS")) ){
#if defined(DEF_FONTPATH)
		fntDir = DEF_FONTPATH;
#else
		fntDir = ".";
#endif
	}

	return 1;
}
/*}}}*/

#ifdef ROT_90
int fgl_rotate90font (fgl_font_t *fnt) /*{{{*/
{
	u32 newSurfSize = 0;
	u32 *newSurfOffset;
	u8  *newSurface;
	int i;

	/* alocate new memory offsets */

	newSurfOffset = (u32 *) FGL_MALLOC( sizeof(u32) * fnt->nGlyphs );

	/* recalcuate required memory */
	for ( i = 0 ; i < fnt->nGlyphs ; i++ ) {
		newSurfOffset[i] = newSurfSize;
		newSurfSize += fnt->width[i] * MONO_SLENGTH( fnt->height[i] );
	}

	newSurface = (u8 *)FGL_MALLOC( sizeof(u8) * newSurfSize );
	for ( i = 0 ; i < fnt->nGlyphs ; i++ ) {
		fgl_rotate90MonoSurface( fnt->surface + fnt->surfOff[i], 
					newSurface + newSurfOffset[i], 
					fnt->width[i], fnt->height[i] );

	}
	fnt->surface = newSurface;
	fnt->surfOff = newSurfOffset;
	fnt->swappedBO  = 1;
	fnt->rotated    = 1;

	return newSurfSize;

}
/*}}}*/
#endif

#ifdef ROT_90
void fgl_rotate90fontExt (fgl_fontExt_t *extension) /*{{{*/
{
#if defined(LOCALE_INT)
	u32 newSurfSize = 0;
	u32 *newSurfOffset;
	u8  *newSurface;
	int i;

	newSurfOffset = (u32 *) FGL_MALLOC( sizeof(u32) * extension->nGlyphs );
	newSurfSize = 0;

	for ( i = 0 ; i < extension->nGlyphs ; i++ ) {
		newSurfOffset[i] = newSurfSize;
		newSurfSize += extension->width[i] * 
			MONO_SLENGTH( extension->height[i] );
	}
	
	newSurface = (u8 *)FGL_MALLOC( sizeof(u8) * newSurfSize );
	for ( i = 0 ; i < extension->nGlyphs ; i++ ) {
		fgl_rotate90MonoSurface( extension->surface + extension->surfOff[i], 
					 newSurface + newSurfOffset[i], 
					 extension->width[i], extension->height[i] );	
	}
	extension->surface = newSurface;
	extension->surfOff = newSurfOffset;
	extension->swappedBO  = 1;
	extension->rotated    = 1;
	
#endif
}
/*}}}*/
#endif

fgl_font_t* fgl_loadFont ( char* fname, u8 ptSize ) /*{{{*/
{
	char buf[128];
	int  fh;
	struct stat fs;
	u8   *p;
	fgl_font_t *fnt;
	u32  i;
	u8   isFileMsb;
	u8   isFileRotated;
	static u8   isHostMsb = (__BYTE_ORDER == __BIG_ENDIAN);


	/* if the requested size is not found, iterate a little */
	for ( i=0; i < sizeof(delta)/sizeof(delta[0]); i++ ){
		sprintf( buf, "%s/%s_%d.fnt", fntDir, fname, ptSize + delta[i]);
		fh = open( buf, O_RDONLY);
		
		if ( fh >= 0 ){
			if ( fstat( fh, &fs) == 0 )
				break;
			else
				fh = -1;
		}
	}

	/* if we still don't find a match, fall back to the default font */
	if ( fh < 0 ){
		sprintf( buf, "%s/default.fnt", fntDir);
		fh = open( buf, O_RDONLY);
   
		if ( (fh < 0) || (fstat( fh, &fs) != 0) ){
			/* we have lost, there's nothing else we can try */
			err( "no matching font file found: %s, %d\n", fname, ptSize);
			return 0;
		}		
	}

	p = (u8*) mmap( NULL, fs.st_size, PROT_READ, (MAP_FILE | MAP_SHARED), fh, 0);
	if ( p == (void*)-1 ){
		err( "mmap %s failed: %d\n", buf, errno);

		p = (u8*) FGL_MALLOC( fs.st_size);
		if ( read( fh, p, fs.st_size) != fs.st_size ){
			FGL_FREE( p);
			return 0;
		}

		fh = -1;
	}
	
	fnt = (fgl_font_t*) FGL_CALLOC( 1, sizeof(fgl_font_t));

	fnt->fh = fh;
	fnt->fSize = fs.st_size;
	fnt->fAdr = p;

	/*
	 * Read font-info from mmaped data, and adapt to different
	 * byte order. But watch out - the compiler might do fancy things
	 * with mis-aligned short/int addresses (strongARM, for instance)
	 */

	isFileMsb       = (*p) & IS_FILEMSB;         
	isFileRotated   = (*p) & IS_ROTATED;
	fnt->antiAlias  = (*p) & IS_ANTIALIAS; p++;

	fnt->nGlyphs    = ((u16)p[1]<<8) + (u16)p[0];   p += sizeof(fnt->nGlyphs);
	fnt->uniOffset  = ((u16)p[1]<<8) + (u16)p[0];   p += sizeof(fnt->uniOffset);

	if ( isFileMsb != isHostMsb ){
		fnt->nGlyphs   = swapU16Bytes( fnt->nGlyphs);
		fnt->uniOffset = swapU16Bytes( fnt->uniOffset);
	}
	
	fnt->fontHeight = *p;          p += sizeof(fnt->fontHeight);
	fnt->ascend     = *p;          p += sizeof(fnt->ascend);
	fnt->descend    = *p;          p += sizeof(fnt->descend);

	fnt->surfOff    = (u32*)p;     p += sizeof(fnt->surfOff[0]) * fnt->nGlyphs;

	if ( isFileMsb != isHostMsb ){
		u32* newSurf = (u32*) FGL_MALLOC( sizeof(u32) * fnt->nGlyphs);
		
		for ( i=0; i< fnt->nGlyphs; i++ ){
			newSurf[i] = swapU32Bytes( fnt->surfOff[i]);
		}
		
		fnt->surfOff = newSurf;
	}
	
	fnt->width      = (u8*)p;         p += sizeof(fnt->width[0]) * fnt->nGlyphs;
	fnt->height     = (u8*)p;        p += sizeof(fnt->height[0]) * fnt->nGlyphs;
	fnt->xBearing   = (i8*) (u8*)p;  p += sizeof(fnt->xBearing[0]) * fnt->nGlyphs;
	fnt->yBearing   = (i8*) (u8*)p;  p += sizeof(fnt->yBearing[0]) * fnt->nGlyphs;
	fnt->advance    = (i8*) (u8*)p;  p += sizeof(fnt->advance[0]) * fnt->nGlyphs;

	fnt->surface    = (u8*)p;

	fnt->fname      = strdup( fname);
	fnt->ptSize     = ptSize; 
	fnt->swappedBO  = (isFileMsb != isHostMsb);
	/* maxAdvance stays uninitialized */
	fnt->maxAdvance   = -1;

	fnt->rotated = 0;
#ifdef ROT_90
	if (!isFileRotated) {
	  fgl_rotate90font(fnt);
	}
#endif
	return fnt;
}
/*}}}*/

u8 fgl_getFontExtensionSize( char *extname, u8 ptSize ) /*{{{*/
{
#if defined(LOCALE_INT)

	char buf[256];
        int j;
	int         ext_fh;
	struct stat ext_fs;

	for ( j=1; j < sizeof(ext_delta)/sizeof(ext_delta[0]); j++ ){
		sprintf( buf, "%s/%s_%d.fnt", fntDir, extname, ptSize + ext_delta[j]);
		ext_fh = open( buf, O_RDONLY);
		if ( ext_fh >= 0 ){
			if ( fstat( ext_fh, &ext_fs) == 0 ) {
			     close (ext_fh);
			     return (ptSize + ext_delta[j]); 
			}
			close (ext_fh);
		}
	}
#endif /* LOCALE_INT */
	return 0;
}
/*}}}*/

fgl_fontExt_t* fgl_loadFontExt(fgl_font_t* fnt, char *extname, u8 ptSize ) /*{{{*/
{
#if defined(LOCALE_INT)

	int         ext_fh;
	struct stat ext_fs;
	u8          *ext_p;
	u32         j;
	u32         ext_height;
	fgl_fontExt_t *extension;
	char buf[128];
	u8   isFileMsb;
	u8   isFileRotated;
	static u8   isHostMsb = (__BYTE_ORDER == __BIG_ENDIAN);

	/* adjust font size */
	if (ptSize < ext_delta[0]) ptSize = ext_delta[0];

	/* load each extension */

	sprintf( buf, "%s/%s_%d.fnt", fntDir, extname, ptSize );
	ext_fh = open( buf, O_RDONLY);
	if ( ext_fh >= 0 ){
	         if ( fstat( ext_fh, &ext_fs) != 0 ) {
		   ext_fh = -1;
		 }
	}
	ext_p = NULL;
	
	if (ext_fh >= 0) {
		ext_p = (u8*) mmap( NULL, ext_fs.st_size, PROT_READ, (MAP_FILE | MAP_SHARED), ext_fh, 0);
		if ( ext_p == (void*)-1 ){
			err( "mmap %s f	ailed: %d\n", buf, errno);
			
			ext_p = (u8*) FGL_MALLOC( ext_fs.st_size);
			if ( read( ext_fh, ext_p, ext_fs.st_size) != ext_fs.st_size ){
				FGL_FREE( ext_p);
				return 0;
			}
			
			ext_fh = -1;
		}
	}

	
	if (ext_fh <= 0) return NULL;


	extension = (fgl_fontExt_t *)FGL_MALLOC(sizeof(fgl_fontExt_t));
		
	extension->fh    = ext_fh;
	extension->fSize = ext_fs.st_size;
	extension->fAdr  = ext_p;
	
	isFileMsb              = (*ext_p) & IS_FILEMSB;	
	isFileRotated          = (*ext_p) & IS_ROTATED;
	extension->antiAlias  = (*ext_p) & IS_ANTIALIAS; ext_p++;

	extension->nGlyphs    = ((u16)ext_p[1]<<8) + (u16)ext_p[0];   ext_p += sizeof(extension->nGlyphs);
	extension->uniOffset  = ((u16)ext_p[1]<<8) + (u16)ext_p[0];   ext_p += sizeof(extension->uniOffset);
		
	if ( isFileMsb != isHostMsb ){
		extension->nGlyphs   = swapU16Bytes( extension->nGlyphs);
		extension->uniOffset = swapU16Bytes( extension->uniOffset);
	}
	
	/* calculate the last character in set now */
	extension->uniLast = extension->uniOffset + extension->nGlyphs;
	
	ext_height             = *ext_p;          ext_p += sizeof(fnt->fontHeight);
		
	/* carefull! adjust height of entire font */
	if ( ext_height > fnt->fontHeight ) fnt->fontHeight = ext_height;
	
	extension->ascend     = *ext_p;          ext_p += sizeof(extension->ascend);
	extension->descend    = *ext_p;          ext_p += sizeof(extension->descend);
	
	extension->surfOff    = (u32*)ext_p;     ext_p += sizeof(extension->surfOff[0]) * extension->nGlyphs;
		
	if ( isFileMsb != isHostMsb ){
		u32* newSurf = (u32*) FGL_MALLOC( sizeof(u32) * extension->nGlyphs);
		
		for ( j=0; j < extension->nGlyphs; j++ ){
			newSurf[j] = swapU32Bytes( extension->surfOff[j]);
		}
		
		extension->surfOff = newSurf;
	}
	
	extension->width      = (u8*)ext_p;        ext_p += sizeof(extension->width[0]) * extension->nGlyphs;
	extension->height     = (u8*)ext_p;        ext_p += sizeof(extension->height[0]) * extension->nGlyphs;
	extension->xBearing   = (i8*) (u8*)ext_p;  ext_p += sizeof(extension->xBearing[0]) * extension->nGlyphs;
	extension->yBearing   = (i8*) (u8*)ext_p;  ext_p += sizeof(extension->yBearing[0]) * extension->nGlyphs;
	extension->advance    = (i8*) (u8*)ext_p;  ext_p += sizeof(extension->advance[0]) * extension->nGlyphs;
	
	extension->surface    = (u8*)ext_p;
	extension->swappedBO  = (isFileMsb != isHostMsb);
	extension->rotated = 0;

#ifdef ROT_90
	if (!isFileRotated) {
	  fgl_rotate90fontExt(extension);
	}
#endif
	return extension;
#else 
	return 0;
#endif /* LOCALE_INT */
}

void fgl_unloadFont ( fgl_font_t* fnt ) /*{{{*/
{
	if ( fnt->fh != -1 ){
		munmap( fnt->fAdr, fnt->fSize);
		close( fnt->fh);
	}
	else {
		FGL_FREE( fnt->fAdr);
	}

	if ( fnt->rotated ){
		FGL_FREE( fnt->surface);
	}

	if ( fnt->swappedBO ){
		FGL_FREE( fnt->surfOff);
	}

	FGL_FREE( fnt);
}
/*}}}*/
void fgl_unloadFontExt ( fgl_fontExt_t* extension ) /*{{{*/
{
#if defined(LOCALE_INT)
	if ( extension->fh != -1 ){
		munmap( extension->fAdr, extension->fSize);
		close( extension->fh);
	}
	else {
		FGL_FREE( extension->fAdr);
	}

	if ( extension->rotated ){
		FGL_FREE( extension->surface);
	}

	if ( extension->swappedBO ){
		FGL_FREE( extension->surfOff);
	}

	FGL_FREE( extension );
#endif /* LOCALE_INT */
}
/*}}}*/

/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_FENG_INCLUDED */
/*}}}*/

