/* -*- mode: C; mode: fold -*- */
/*
 * feng_ttf.c - fgl module to implement a truetype font file (*.ttf)
 *              font engine (using the freetype library)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: feng_ttf.c,v 1.12 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/

#ifndef FGL_FENG_INCLUDED
#define FGL_FENG_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* fgl imports */
#include "util.c"

#include FGL_HEAP_MODULE

#ifdef FGL_RESET_PROTO
#  undef FGL_PROTO
#  undef FGL_RESET_PROTO
#endif

#include <freetype/freetype.h>
/*}}}*/

/* globals and defines */ /*{{{*/
static char *fntDir;
TT_Engine engine;

int xres = 96;
int yres = 96;

#define MAX_NGLYPHS 200

#define TTLOAD_FLAGS (TTLOAD_SCALE_GLYPH | TTLOAD_HINT_GLYPH)
/*}}}*/


int fgl_initFontEngine () /*{{{*/
{
	TT_Error error;

	if ( (error = TT_Init_FreeType( &engine)) ){
		err( "TT_Init_FreeType failed: %d\n", error);
		return 0;
	}

	if ( !(fntDir = getenv( "FGL_FONTS")) ){
#if defined(DEF_FONTPATH)
		fntDir = DEF_FONTPATH;
#else
		fntDir = ".";
#endif
	}

	return 1;
}
/*}}}*/


inline
fgl_font_t* allocFont ( u16 nGlyphs ) /*{{{*/
{
	fgl_font_t *fnt = 0;
	u32        arr32Size = sizeof(u32) * nGlyphs;
	u32        arr8Size = sizeof(u8) * nGlyphs;

	fnt = FGL_CALLOC( 1, sizeof(fgl_font_t) + 5*arr8Size + arr32Size);

	fnt->surfOff  = (u32*)((u8*)fnt + sizeof(fgl_font_t));
	fnt->width    = (i8*) ((u8*)(fnt->surfOff)  + arr32Size); 
	fnt->height   = (i8*) ((u8*)(fnt->width)    + arr8Size); 
	fnt->xBearing = (i8*) ((u8*)(fnt->height)   + arr8Size);
	fnt->yBearing = (i8*) ((u8*)(fnt->xBearing) + arr8Size);
	fnt->advance  = (i8*) ((u8*)(fnt->yBearing) + arr8Size);

	fnt->nGlyphs = nGlyphs;

	return fnt;
}
/*}}}*/


fgl_font_t* fgl_loadFont ( char* fname, u8 ptSize ) /*{{{*/
{
	TT_Error             error;
	TT_Face              face;
	TT_Instance          instance;
	TT_Glyph             glyph;
	TT_CharMap           charMap;
	TT_Glyph_Metrics     metrics;
	TT_Face_Properties   properties;
	TT_Raster_Map        rasterMap;

	char                 buf[128];
	int                  i, j, k, w, h, m;
    u16                  platform, encoding, nGlyphs;
	u8                   *pSurf;
	fgl_font_t           *fnt = 0;
	int                  ptres, em72;
	u16                  *map;


	sprintf( buf, "%s/%s.ttf", fntDir, fname);
	/*printf( "loading font %s\n", buf);*/

	if ( (error = TT_Open_Face( engine, buf, &face)) ){
		err( "TT_Open_Face \"%s\" failed: %d\n", buf, error);
		return 0;
	}
	if ( (error = TT_New_Instance( face, &instance)) ){
		err( "TT_New_Instance failed: %d\n", error);
		return 0;
	}
    if ( (error = TT_Set_Instance_Resolutions( instance, xres, yres)) ){
		err( "TT_New_Instance failed: %d\n", error);
		return 0;
	}
    if ( (error = TT_Set_Instance_CharSize( instance, ptSize * 64)) ){
		err( "TT_Set_Instance_CharSize %d failed: %d\n", ptSize, error);
		return 0;		
	}
	if ( (error = TT_New_Glyph( face, &glyph)) ){
		err( "TT_New_Glyph failed: %d\n", error);
		return 0;
	}
	if ( (error = TT_Get_Face_Properties( face, &properties)) ){
		err( "TT_New_Glyph failed: %d\n", error);
		return 0;
	}

	/* try to get a unicode char map */
	for ( i=0; i < properties.num_CharMaps; i++ ){
		TT_Get_CharMap_ID( face, i, &platform, &encoding );
		if ( (platform == 3 && encoding == 1 )  ||
			 (platform == 0 && encoding == 0 ) ) {
			TT_Get_CharMap( face, i, &charMap );
			break;
		}
    }
    if ( i == properties.num_CharMaps ){
		err( "no unicode mapping found");
		return 0;		
	}

	/* there can be quite a number of glyphs in a ttf file, check if we have to cap */
	nGlyphs = (properties.num_Glyphs > MAX_NGLYPHS) ? MAX_NGLYPHS : properties.num_Glyphs;
	
	/* BAD! reverse mapping eats up memory, and it probably is used
	 * just for gaps (all referring to the same void glyph). But how
	 * can we detect if we already processed a certain glyph index,
	 * without walking the whole unicode array?
	 */
	map = FGL_MALLOC( properties.num_Glyphs * sizeof(u16));
	memset( map, 0, properties.num_Glyphs * sizeof(u16));	

	fnt = allocFont( nGlyphs);
	m = 0;

	/* first pass, fill in font data */
    for ( i = 0; i < nGlyphs; i++ ){
		k = TT_Char_Index( charMap, (short)i );

		if ( map[k] && (map[k] < nGlyphs) ){
			/* if we already have a map entry for k, simply copy the data (if it
			 * doesn't refer to a out-of-range glyph
			 */
			j = map[k];

			fnt->width[i]    = fnt->width[j];
			fnt->height[i]   = fnt->height[j];			
			fnt->xBearing[i] = fnt->xBearing[j];
			fnt->yBearing[i] = fnt->yBearing[j];
			fnt->advance[i]  = fnt->advance[j];			
		}
		else {
			/* Ok, we don't have this one in our glyph space yet, query it */
			
			if ( !(error = TT_Load_Glyph( instance, glyph, k, TTLOAD_FLAGS )) ){
				TT_Get_Glyph_Metrics( glyph, &metrics );
				
				w = (metrics.bbox.xMax - metrics.bbox.xMin) / 64;
				h = (metrics.bbox.yMax - metrics.bbox.yMin) / 64;

				//printf( "%d '%c' : %d,%d, %d\n", i, i, w, h, metrics.advance/64);

				m += MONO_SLENGTH(w) * h;

				fnt->width[i]    = w;
				fnt->height[i]   = h;
				fnt->xBearing[i] = metrics.bearingX / 64;
				fnt->yBearing[i] = metrics.bearingY / 64;
				fnt->advance[i]  = metrics.advance  / 64;

				map[k] = i;
			}
		}
	}

	if ( properties.horizontal ){
		em72 = properties.header->Units_Per_EM * 72;
		ptres = ptSize * yres;

		/* must be ceiled, or we might loose pixels */
		fnt->ascend = (properties.horizontal->Ascender * ptres + em72/2) / em72;
		fnt->descend = (-properties.horizontal->Descender * ptres + em72/2) / em72;

		fnt->fontHeight = fnt->ascend + fnt->descend +
			(properties.horizontal->Line_Gap * ptres / em72);
	}

	pSurf = fnt->surface = FGL_CALLOC(1, m);
	rasterMap.flow = TT_Flow_Down;

	/* second pass, render glyphs */
    for ( i = 0; i < nGlyphs; i++ ){
		k = TT_Char_Index( charMap, (short)i );

		if ( map[k] == i ) {
			/* this is a new one, render the glyph */
			rasterMap.width = fnt->width[i];

			if ( rasterMap.width ) {
				fnt->surfOff[i]  = (unsigned long)pSurf - (unsigned long)fnt->surface;

				rasterMap.bitmap = pSurf;
				rasterMap.cols   = MONO_SLENGTH(rasterMap.width);
				rasterMap.rows   = fnt->height[i];
				rasterMap.size   = rasterMap.cols * rasterMap.rows;

				if ( !(error = TT_Load_Glyph( instance, glyph, k, TTLOAD_FLAGS )) ){
					TT_Get_Glyph_Metrics( glyph, &metrics );
					TT_Get_Glyph_Bitmap( glyph, &rasterMap,
										-metrics.bbox.xMin, -metrics.bbox.yMin);
				}

				pSurf += (rasterMap.cols * rasterMap.rows);
			}
		}
		else {
			/* We already had this one, copy the offset. Note that map[k]
			 * can't be outside our glyph space because of the way we set it in pass 1
			 */
			fnt->surfOff[i] = fnt->surfOff[ map[k] ];
		}
	}

	TT_Close_Face( face);

	FGL_FREE( map);

	fnt->fname  = strdup( fname);
	fnt->ptSize = ptSize;

	return fnt;
}
/*}}}*/


void fgl_unloadFont ( fgl_font_t* fnt ) /*{{{*/
{
	FGL_FREE( fnt->surface);
	FGL_FREE( fnt);
}
/*}}}*/

/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_FENG_INCLUDED */
/*}}}*/

