/* -*- mode: C; mode: fold -*- */
/*
 * fgl.h - fgl master header (only included part in clients)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: fgl.h,v 1.36 2001/11/28 20:18:35 damian Exp $
 */

/* fgl header epilog */ /*{{{*/
#ifndef FGL_H_INCLUDED
#define FGL_H_INCLUDED

#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdio.h>

#include "fgl_config.h"
#include "locale.h"
/*}}}*/


/*************************************************************************
 * standard types section (from config)
 */
/*{{{*/

typedef struct {
	i32 x;
	i32 y;
} fgl_point_t;

static inline
void swap8 ( u8* a, u8* b )
{
	u8 tmp = *a;
	*a = *b;
	*b = tmp;
}

static inline
void swap32 ( i32* a, i32* b )
{
	u32 tmp = *a;
	*a = *b;
	*b = tmp;
}

static inline
int inInterval ( int a, int r1, int r2 )
{
	return ((a >= r1) && (a <= r2));
}

/*}}}*/


/*************************************************************************
 * fgl types
 */

/* this include s just typedefs from the other modules */
#include FGL_FGLD_MODULE


typedef struct _fgl_bitmap_t {  /*{{{*/
	fgl_slc_t    *surface;
	u32          width;
	u32          height;
	u32          sLength;  /* length of scanline in scanline cell units */
        fgl_color_t  transp;
        u32          state;
        u32          lastOperation;
        u32          nextOperation;
} fgl_bitmap_t;
/*}}}*/

typedef struct _fgl_a8Bitmap_t { /*{{{*/
	u8           *surface;
	u32          width;
	u32          height;
	u32          sLength; /* same as width, but we keep it for conformity */
        u32          state;
} fgl_a8Bitmap_t;
/*}}}*/

typedef struct _fgl_a2Bitmap_t { /*{{{*/
	u8           *surface;
	u32          width;
	u32          height;
	u32          sLength;  /* length of scanline in bytes */
        u32          state;
} fgl_a2Bitmap_t;
/*}}}*/

typedef struct _fgl_monoBitmap_t { /*{{{*/
	u8           *surface;
	u32          width;
	u32          height;
	u32          sLength;  /* length of scanline in bytes */
        u32          state;
} fgl_monoBitmap_t;
/*}}}*/

#define MONO_SLENGTH(_width) (((_width)+7) >> 3)
#define A2_SLENGTH(_width) (((_width)+3) >> 2)

typedef struct _fgl_spriteData_t { /*{{{*/
	u32          width, height;
	u16          xhot, yhot;
	u8           nColors;
	u8           transp;
	u32          *cmap;
	u8           *data; /* this is in fgl_color_t so that we can replace in situ */
        u32          state;
} fgl_spriteData_t;
/*}}}*/

typedef struct _fgl_sprite_t { /*{{{*/
	fgl_bitmap_t      *bmp;
	fgl_monoBitmap_t  *mask;
	fgl_bitmap_t      *sav;
	i16               xhot, yhot;

	i32               x0, y0, x1, y1;
	u8                isVisible;
  
        void              *ext; /* extension to hook-up native data */
} fgl_sprite_t;
/*}}}*/

typedef struct _fgl_cmap_t { /*{{{*/
	int len;
	u8 *r;
	u8 *g;
	u8 *b;
} fgl_cmap_t;
/*}}}*/

typedef struct _fgl_fontExt_t {
 	u16         nGlyphs;
	u16         uniOffset;
	u16         uniLast;

	/* glyph data itself */
	u8          *surface;

	/* per glyph data */
	u32         *surfOff;
	u8          *width;
	u8          *height;
	i8          *xBearing;
	i8          *yBearing;
	i8          *advance;

	/* per font info */
	u8          ascend;
	u8          descend;

	u32         nRefs;

	u32         fSize;
	void        *fAdr;
        int         fh;
	int         swappedBO;
	int         antiAlias;
        int         rotated;
} fgl_fontExt_t;

typedef struct _fgl_font_t { /*{{{*/
	u16         nGlyphs;
	u16         uniOffset;

	/* glyph data itself */
	u8          *surface;

	/* per glyph data */
	u32         *surfOff;
	u8          *width;
	u8          *height;
	i8          *xBearing;
	i8          *yBearing;
	i8          *advance;

	/* per font info */
	u8          fontHeight;
	u8          ascend;
	u8          descend;
        int         maxAdvance;

	char        *fname;
	u16         ptSize;
	u32         nRefs;

#if defined(USE_FENG_BIN)
	u32         fSize;
	void        *fAdr;
	int         fh;
	int         swappedBO;    /* did we swap byte order during load ? */  
        int         antiAlias;    /* is the font antialiased */
	int         rotated;
#if defined(LOCALE_INT)

        /* introduce extended sets */
        u8          ext_number;
	fgl_fontExt_t* ext[LOCALE_EXTENSION_NUM];

#endif /* LOCALE_INT */

#endif /* USE_FENG_BIN */

} fgl_font_t;
/*}}}*/

typedef struct _fgl_fontEntry_t { /*{{{*/
	fgl_font_t *fnt;

	char        *fname;
	u16         ptSize;

	struct _fgl_fontEntry_t *next;
} fgl_fontEntry_t;
/*}}}*/

typedef struct _fgl_fontExtEntry_t { /*{{{*/
	fgl_fontExt_t *fntExt;

	char        *fextname;
	u16         ptSize;

	struct _fgl_fontExtEntry_t *next;
} fgl_fontExtEntry_t;
/*}}}*/

typedef enum _fgl_mode_t { /*{{{*/
	PAINT_MODE = 0,
	XOR_MODE   = 1
} fgl_mode_t;
/*}}}*/

#define MODE_BRANCH(_ctx,_func,_arg) \
  if ( (_ctx)->mode == PAINT_MODE ) {_func _arg;} else {_func##Xor _arg;}

typedef struct _fgl_context_t { /*{{{*/
	fgl_bitmap_t *bmp;

	i32          x0Clip, y0Clip, x1Clip, y1Clip;

	u32          xoff, yoff;

	fgl_color_t  clr;
	fgl_color_t  xclr;
	u8           alpha;

	fgl_font_t   *fnt;

	fgl_mode_t   mode;

} fgl_context_t;
/*}}}*/


/*
 * Evans&Sutherland clip codes
 */
#define CC_LEFT    1
#define CC_RIGHT   2
#define CC_BOTTOM  4
#define CC_TOP     8

/*
 * clip rect relation
 */
#define CR_OUT     0
#define CR_IN      1
#define CR_PART    2

/*
 * Operation type for frame buffer direct operation
 */
#define FGL_FRAMEBUFFER_OP 0x0000

/*
 * macros to extract [0..255] channel values from 8888 argbs and vice versa
 */
#define A_8(_rgb)    (u8) (((_rgb) & 0xff000000) >> 24)
#define R_8(_rgb)    (u8) (((_rgb) & 0x00ff0000) >> 16)
#define G_8(_rgb)    (u8) (((_rgb) & 0x0000ff00) >> 8)
#define B_8(_rgb)    (u8) (((_rgb) & 0x000000ff))

#define RGB_888(_r,_g,_b) (u32) ((_r << 16) | (_g << 8) | (_b))

/*************************************************************************
 * rotation section
 */
static inline
void rotate32 ( fgl_bitmap_t* bmp, u32* x, u32* y )
{
	u32 tmp  = *x;
	*x = bmp->height - *y - 1;
	*y = tmp;
}

/*************************************************************************
 * screen config section
 */
typedef struct _fgl_config_t {
	u32 xoff;
	u32 yoff;
	u32 width;
	u32 height;
} fgl_config_t;


/*************************************************************************
 * global data section
 */
/*{{{*/

#if defined(FGL_MAIN)
#  define FGL_GLOBAL
#  define FGL_INIT(_a) = _a
#else
#  define FGL_GLOBAL extern
#  define FGL_INIT(_a)
#endif

FGL_GLOBAL fgl_bitmap_t               _fgl_screenBmpObject;
FGL_GLOBAL fgl_bitmap_t               *fgl_screenBmp FGL_INIT( &_fgl_screenBmpObject);

FGL_GLOBAL fgl_context_t              _fgl_screenCtxObject;
FGL_GLOBAL fgl_context_t              *fgl_screenCtx FGL_INIT( &_fgl_screenCtxObject);

FGL_GLOBAL fgl_sprite_t               *fgl_cursor;
FGL_GLOBAL u8                         _fgl_redrawCursor;
FGL_GLOBAL u8                         _fgl_inCursorOp;

FGL_GLOBAL fgl_fontEntry_t            *_fgl_fontCache;
FGL_GLOBAL fgl_fontExtEntry_t         *_fgl_fontExtCache;
FGL_GLOBAL int                        _fgl_fontInitialized;

FGL_GLOBAL fgl_font_t                 *fgl_defaultFont;

FGL_GLOBAL fgl_color_t fgl_clrBlack;
FGL_GLOBAL fgl_color_t fgl_clrWhite;
FGL_GLOBAL fgl_color_t fgl_clrRed;
FGL_GLOBAL fgl_color_t fgl_clrGreen;
FGL_GLOBAL fgl_color_t fgl_clrBlue;
FGL_GLOBAL fgl_color_t fgl_clrYellow;
FGL_GLOBAL fgl_color_t fgl_clrCyan;
FGL_GLOBAL fgl_color_t fgl_clrGray;
FGL_GLOBAL fgl_color_t fgl_clrLightGray;
FGL_GLOBAL fgl_color_t fgl_clrOrange;

/*}}}*/


/*************************************************************************
 * fgl module import section
 */
/*{{{*/

/* make sure we don't get in definitions (unless our includer requests them) */
#define FGL_PROTO

/*
 * include the macros from fgld (we already have the types)
 */
#include FGL_FGLD_MODULE

/*
 * include the acceleraion module
 */
#ifdef FGL_ACCEL_MODULE
#include FGL_ACCEL_MODULE
#endif

#define IMAGE_ON_SCREEN 0x00000001

/*
 * in case this is included from a external source, make sure we see all
 * fgl modules
 */
#ifndef FGL_MODULE
#  define FGL_PROTO
#  include FGL_HEAP_MODULE
#  include FGL_INIT_MODULE
#  include "util.c"
#  include FGL_CMAP_MODULE
#  include "cntx.c"
#  include FGL_DRAW_MODULE
#  include "mbmp.c"
#  include "a8bmp.c"
#  include "a2bmp.c"
#  include "cbmp.c"
#  include "font.c"
#  include "spri.c"
#  include "line.c"
#  include "rect.c"
#  include "poly.c"
#  include "oval.c"
#  include "screencfg.c"
#endif

/*}}}*/

#endif /* FGL_H_INCLUDED */
