/* -*- mode: C; mode: fold -*- */
/*
 * fgld_32 - fgl header to handle framebuffer access and scanline cells
 *           for 32-bpp visuals (read/write value, compute/alter scanline 
 *           cell addresses), generic part (might be partly overriden by
 *           platform specific optimizations)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 * 12/13/00  Damian Kowalewski
 *
 * $tvt: fgld_32.h,v 1.3 2001/07/04 00:47:55 samc Exp $
 */


/*****************************************************************************
 * linear framebuffer access for 8-8-8 truecolor visual
 */

#if !defined(FGL_PROTO)

typedef u32 fgl_slc_t;     /* scanline cell type */
typedef u32 fgl_color_t;   /* pixel color value */

#else /* defined FGL_PROTO */


/*****************************************************************************
 * internal helper types and functions
 */

typedef unsigned long long u64;

/* #if defined (TARGET_CPU_I386)*/  /* simple intel optimizations */
/* #include "fgld_32_i386.h"*/
/* #endif *//* defined TARGET_CPU_I386 */

/***************************************************************************
 * public functions
 */


static inline
int fgl_getSlcSize () /*{{{*/
{
	return 4;
}
/*}}}*/

static inline
int fgl_getBitsPerPixel () /*{{{*/
{
	return 32;
}
/*}}}*/

// return length of scanline in cell units
static inline
int fgl_getScanlineLength ( int length ) /*{{{*/
{
	return length;
}
/*}}}*/

static inline
fgl_slc_t* fgl_getSlc ( fgl_bitmap_t* bmp, u32 x, u32 y ) /*{{{*/
{
	return (bmp->surface + (y * bmp->sLength) + x);
}
/*}}}*/

static inline
void fgl_incSlc ( fgl_slc_t **p, u32 x ) /*{{{*/
{
	(*p)++;
}
/*}}}*/

static inline
void fgl_decSlc ( fgl_slc_t **p, u32 x ) /*{{{*/
{
	(*p)--;
}
/*}}}*/

static inline
void fgl_addSlc ( fgl_slc_t **p, u32 x, i32 pixels ) /*{{{*/
{
	(*p) += pixels;
}
/*}}}*/

static inline
void fgl_subSlc ( fgl_slc_t **p, u32 x, i32 pixels ) /*{{{*/
{
	(*p) -= pixels;
}
/*}}}*/

static inline
fgl_color_t fgl_getSlcPixel ( fgl_slc_t *p, u32 x )  /*{{{*/
{
	return (fgl_color_t) *p;
}
/*}}}*/

static inline
void fgl_setSlcPixel ( fgl_slc_t *p, u32 x, fgl_color_t clr )  /*{{{*/
{
	*p = clr;
}
/*}}}*/

static inline
void fgl_setSlcPixelXor ( fgl_slc_t *p, u32 x, fgl_color_t clr ) /*{{{*/
{
	*p ^= clr;
}
/*}}}*/

#ifndef HAVE_SETSLCCOLS
static inline
void fgl_setSlcCols ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	if ( len < 8 ){
		fgl_slc_t *pmax = p + len;

		while ( p < pmax  ){
			*p++ = clr;
		}
	}
	else {
		/* do a little loop unrolling */
		u32 *pp;
		u32 cc = clr;

		if ( x & 1 ){ 			/* align */
			*p++ = clr;
			len--;
		}

		pp = (u32*) p;
		while (len > 7){
 			*pp++ = cc;
			*pp++ = cc;
 			*pp++ = cc;
			*pp++ = cc;
 			*pp++ = cc;
			*pp++ = cc;
 			*pp++ = cc;
			*pp++ = cc;

			len -= 8;
		}

		if ( len ){
			p = ((fgl_slc_t*)pp);
			while ( len-- ){
				*p++ = clr;
			}
		}
	}
}
/*}}}*/
#endif /* HAVE_SETSLCCOLS */

static inline
void fgl_setSlcColsXor ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	fgl_slc_t *pmax = p + len;

	while ( p < pmax  ){
		*p++ ^= clr;
	}
}
/*}}}*/

static inline
void fgl_setSlcRows ( fgl_slc_t *p, i32 inc, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	do {
		*p = clr;
		p += inc;
	} while ( --len );
}
/*}}}*/

static inline
void fgl_setSlcRowsXor ( fgl_slc_t *p, i32 inc, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	do {
		*p ^= clr;
		p += inc;
	} while ( --len );
}
/*}}}*/

#ifndef HAVE_INITBMP
static inline
void fgl_initBmp ( fgl_bitmap_t* bmp, i32 clr ) /*{{{*/
{
	fgl_slc_t* p = bmp->surface;
	u32 len = bmp->sLength * bmp->height;

	u32 *pp;
	u32 cc = clr;

	pp = (u32*) p;
	while (len > 7){
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		
		len -= 8;
	}

	if ( len ){
		p = ((fgl_slc_t*)pp);
		while ( len-- ){
			*p++ = (fgl_slc_t)clr;
		}
	}
}
/*}}}*/
#endif /* HAVE_INITBMP */


#ifndef HAVE_CPYSLC
static inline
void fgl_cpySlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len ) /*{{{*/
{
	memcpy( dest, src, len*4); 
}
/*}}}*/
#endif /* HAVE_CPYSLC */

#ifndef HAVE_MOVESLC
static inline
void fgl_moveSlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len ) /*{{{*/
{
	memmove( dest, src, len*4);
}
/*}}}*/
#endif /* HAVE_MOVESLC */

#endif  /* defined FGL_PROTO */




