/* -*- mode: C; mode: fold -*- */
/*
 * fgld_4  - fgl header to handle framebuffer access and scanline cells
 *           for 4-bpp visuals (read/write value, compute/alter scanline 
 *           cell addresses), generic part (might be partly overriden by
 *           platform specific optimizations)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: fgld_4.h,v 1.6 2001/07/04 00:47:55 samc Exp $
 */


#if !defined(FGL_PROTO)

typedef u8 fgl_slc_t;     /* scanline cell type */
typedef u8 fgl_color_t;   /* pixel color value */

#else /* defined FGL_PROTO */


static inline
int fgl_getSlcSize ()
{
	return 1;
}

static inline
int fgl_getBitsPerPixel ()
{
	return 4;
}

// return length of scanline in cell units
static inline
int fgl_getScanlineLength ( int length )
{
  if (( length / 2) & 1) {
    return ( length / 2 + 1 );
  } else {
    return ( length / 2 );
  }
}

static inline
fgl_slc_t* fgl_getSlc ( fgl_bitmap_t* bmp, u32 x, u32 y )
{
	return (bmp->surface + ((y * bmp->sLength) + x) / 2);
}

static inline
void fgl_incSlc ( fgl_slc_t **p, u32 x )
{
	if ( x & 1 ){
		(*p)++;
	}
}

static inline
void fgl_decSlc ( fgl_slc_t **p, u32 x )
{
	if ( !(x & 1) ){
		(*p)--;
	}
}

static inline
void fgl_addSlc ( fgl_slc_t **p, u32 x, i32 pixels )
{
	(*p) += pixels/2;

	if ( pixels > 0 ){
		if ( (x & 1) && (pixels & 1) )
			(*p)++;
	}
	else {
		if ( !(x & 1) && (pixels & 1) )
			(*p)--;
	}
}

static inline
void fgl_subSlc ( fgl_slc_t **p, u32 x, i32 pixels )
{
	fgl_addSlc( p, x, -pixels);
}

static inline
fgl_color_t fgl_getSlcPixel ( fgl_slc_t *p, u32 x )
{
	if ( x & 1 ){
		return (fgl_color_t) (*p & 0xf);
	}
	else {
		return (fgl_color_t) (*p >> 4);
	}
}

static inline
void fgl_setSlcPixel ( fgl_slc_t *p, u32 x, fgl_color_t clr )
{
	if ( x & 1 ){
		*p = (*p & 0xf0) | (clr & 0x0f);
	}
	else {
		*p = (clr << 4) | (*p & 0x0f);
	}
}
static inline
void fgl_setSlcPixelXor ( fgl_slc_t *p, u32 x, fgl_color_t clr )
{
	if ( x & 1 ){
		*p = (*p & 0xf0) | ((clr ^ *p) & 0x0f);
	}
	else {
		*p = ((clr << 4) ^ (*p & 0xf0)) | (*p & 0x0f);
	}
}



static inline
void fgl_setSlcCols ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr )
{
	u8 c = (clr << 4) | (clr & 0x0f);

	if ( x & 1 ) {
		*p++ = (*p & 0xf0) | (clr & 0x0f);
		len--;
	}

	while ( len > 1 ){
		*p++ = c;
		len-=2;
	}

	if ( len ) {
		*p = (clr << 4) | (*p & 0x0f);
	}
}

static inline
void fgl_setSlcColsXor ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr )
{
	u8 c = (clr << 4) | (clr & 0x0f);

	if ( x & 1 ) {
		*p++ = (*p & 0xf0) | ((clr ^ *p) & 0x0f);
		len--;
	}

	while ( len > 1 ){
		*p++ = (*p ^ c);
		len-=2;
	}

	if ( len ) {
		*p = ((clr << 4) ^ (*p & 0xf0))  | (*p & 0x0f);
	}
}


static inline
void fgl_setSlcRows ( fgl_slc_t *p, i32 inc, u32 x, u32 len, fgl_color_t clr )
{
	i32 inc2 = inc / 2;

	if ( inc & 1 ) {
		while ( len-- ){
			if ( x & 1 ){
				*p = (*p & 0xf0) | (clr & 0x0f);
				if ( inc > 0 )
					p++;
			}
			else {
				*p = (clr << 4) + (*p & 0x0f);
				if ( inc < 0 )
					p--;
			}
			p += inc2;
			x++;
		}
	}
	else {
		if ( x & 1 ){
			do {
				*p = (*p & 0xf0) | (clr & 0x0f);
				p += inc2;
			} while ( --len );
		}
		else {
			do {
				*p = (clr << 4) + (*p & 0x0f);
				p += inc2;
			} while ( --len );		
		}
	}
}
static inline
void fgl_setSlcRowsXor ( fgl_slc_t *p, i32 inc, u32 x, u32 len, fgl_color_t clr )
{
	i32 inc2 = inc / 2;

	if ( inc & 1 ) {
		while ( len-- ){
			if ( x & 1 ){
				*p = (*p & 0xf0) | ((clr ^ *p) & 0x0f);
				if ( inc > 0 )
					p++;
			}
			else {
				*p = ((clr << 4) ^ (*p & 0xf0)) | (*p & 0x0f);
				if ( inc < 0 )
					p--;
			}
			p += inc2;
			x++;
		}
	}
	else {
		if ( x & 1 ){
			do {
				*p = (*p & 0xf0) | ((clr ^ *p) & 0x0f);
				p += inc2;
			} while ( --len );
		}
		else {
			do {
				*p = ((clr << 4) ^ (*p & 0xf0)) | (*p & 0x0f);
				p += inc2;
			} while ( --len );		
		}
	}
}

static inline
void fgl_initBmp ( fgl_bitmap_t* bmp, i32 clr )
{
	fgl_slc_t* p = bmp->surface;
	u32 len = bmp->sLength * bmp->height / 2;
	u8 c = (clr << 4) + clr;

	while ( len ){
		*p++ = c;
		len--;
	}
}


static inline
void fgl_cpySlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len )
{
	u8  b;
	u32 l;

	if ( (x & 1) == (u & 1) ){  /* equal src and dest alignment */
		if ( x & 1 ){           /* start not aligned */
			*dest = (*dest & 0xf0) | (*src & 0xf);
			dest++;
			src++;
			len--;
		}

		l = len >> 1;
		memmove( dest, src, l);

		if ( len & 1 ){         /* end not aligned */
			dest += l;
			src += l;
			*dest = (*src & 0xf0) | (*dest & 0x0f);
		}
	}
	else {               
	  /* nasty - different alignment */
		if ( (x & 1) == 0 ){    
		  /* src aligned, dest not */
			*dest = (*dest & 0xf0) | (*src >> 4);
			dest++;
			len--;
		}

		while ( len > 1 ){
		  b = (src[0] << 4) | (src[1] >> 4);
		  src++;
		  dest[0] = b;
		  dest++;
		  len -= 2;
		}

		if ( len ){
			*dest = (*src << 4) | (*dest & 0x0f);
		}
	}
}

static inline
void fgl_moveSlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len )
{

	fgl_cpySlc( src, x, dest, u, len);
}

#endif  /* defined FGL_PROTO */

