/* -*- mode: C; mode: fold -*- */
/*
 * font.c - fgl module for generic font handling (load/unload
 *          display, metrics)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: font.c,v 1.20 2001/08/31 07:29:48 peter Exp $
 */

#ifndef FGL_FONT_INCLUDED
#define FGL_FONT_INCLUDED

#include "fgl.h"

/* fgl imports */
#include "util.c"
#include "cntx.c"
#include "mbmp.c"
#include "a2bmp.c"
#include "cbmp.c"

/* initialize locale specific stuff */
#if defined(LOCALE_INT)
#include "locale.h"
#endif

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif


#ifdef FGL_INLINE_FONT
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif

extern int fgl_initFontEngine ();
extern fgl_font_t* fgl_loadFont ( char* fname, u8 ptSize );
extern void fgl_unloadFont ( fgl_font_t* fnt );

#if defined(LOCALE_INT)
extern fgl_fontExt_t* fgl_loadFontExt ( fgl_font_t* fnt, char* extname, u8 ptSize );
extern char *ext_names[LOCALE_EXTENSION_NUM];
extern void fgl_unloadFontExt ( fgl_fontExt_t* extension );
extern u8 fgl_getFontExtensionSize( char *extname, u8 ptSize );
#endif

inline static char* font_strdup(const char* str)  /*{{{*/
{
	/* just a helper to avoid direct 'malloc' usage (heap fragmentation) */
	char* s;
        if (str == NULL) {
                return (NULL);
        }
        s = (char *)FGL_MALLOC(strlen(str) + 1);
        if (s != 0) {
                strcpy(s, str);
        }
        return (s);
}
/*}}}*/

FUNC_ATTR
fgl_font_t* fgl_getFont ( char* fname, u8 ptSize ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
	fgl_fontEntry_t *fe;
	fgl_font_t *fnt;

#if defined(LOCALE_INT)
	fgl_fontExtEntry_t* fex;
	fgl_fontExt_t* extension;
	int i;
	int ext_found;
	u8 ext_size;
#endif /* LOCALE_INT */


	if ( !_fgl_fontInitialized && !(_fgl_fontInitialized = fgl_initFontEngine()) ){
		err( "font initialization failed\n");
		return 0;
	}

	if ( !fname ){
		fname = "Default";
	}

	for ( fe = _fgl_fontCache; fe; fe = fe->next ){
		if ( (*fname == *fe->fname) &&
			 !strcmp( fname, fe->fname) && (ptSize == fe->ptSize) ){
			fe->fnt->nRefs++;
			return fe->fnt;
		}
	}

	if ( (fnt = fgl_loadFont( fname, ptSize)) ){
		fnt->nRefs = 1;

		fe = (fgl_fontEntry_t*) FGL_MALLOC( sizeof(fgl_fontEntry_t));

		/* update font cache */
		fe->fnt    = fnt;
		fe->fname  = font_strdup( fname);
		fe->ptSize = ptSize;
		fe->next   = _fgl_fontCache;

		_fgl_fontCache = fe;

		/* check if the extensions are loaded in */
#if defined(LOCALE_INT)

		fnt->ext_number = 0;

		for ( i = 0 ; i < LOCALE_EXTENSION_NUM ; i++ ) {

		  if ((ext_size = fgl_getFontExtensionSize( ext_names[i], ptSize ))) {
		        ext_found = 0;
		        for ( fex = _fgl_fontExtCache; fex; fex = fex->next ){
				if ( (*(ext_names[i]) == *fex->fextname) &&
				     !strcmp( ext_names[i], fex->fextname) && (ext_size == fex->ptSize) ){
					fex->fntExt->nRefs++;
					fnt->ext[fnt->ext_number] = fex->fntExt;
					fnt->ext_number++;
					ext_found = 1;
					break;
				}
			}
		      
			if ( !ext_found && (extension = fgl_loadFontExt( fnt, ext_names[i], ext_size)) ){
				extension->nRefs = 1;
				
				fex = (fgl_fontExtEntry_t*) FGL_CALLOC( 1, sizeof(fgl_fontExtEntry_t));
				

				fex->fntExt    = extension;
				fex->fextname  = font_strdup( ext_names[i]);
				fex->ptSize    = ext_size;
				fex->next      = _fgl_fontExtCache;

				_fgl_fontExtCache = fex;

				fnt->ext[fnt->ext_number] = fex->fntExt;
			
				fnt->ext_number++;
			}
		  }
		}
	    
#endif /* LOCALE_INT */

		return fnt;
	}

	return 0;
}
#endif /* !FGL_PROTO */
;
/*}}}*/


static inline
void fgl_addFontReference ( fgl_font_t* fnt ) /*{{{*/
{
	if ( fnt ) {
		fnt->nRefs++;
	}
}
/*}}}*/

FUNC_ATTR
void fgl_disposeFont ( fgl_font_t* fnt ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
#if !defined(USE_FENG_HDR)
	fgl_fontEntry_t *fe, *fe1=0, *fe2;
#if defined(LOCALE_INT)
	fgl_fontExtEntry_t *fex, *fex1=0, *fex2;
	int i;
	
	for ( i = 0 ; i < fnt->ext_number ; i++ ) {

		if ( fnt->ext[i] && fnt->ext[i]->nRefs ){
			fnt->ext[i]->nRefs--;
			
			if ( fnt->ext[i]->nRefs == 0 ){
				/* free the extension and all it's entries */
				for ( fex = _fgl_fontExtCache; fex; ){
					if ( fex->fntExt ==  fnt->ext[i]){
						if ( fex == _fgl_fontExtCache ){
							_fgl_fontExtCache = fex->next;
						}
						else {
							fex1->next = fex->next;
						}
						
						fex2 = fex;
						fex = fex->next;
						
						FGL_FREE( fex2->fextname);
						FGL_FREE( fex2);
					}
					else {
						fex1 = fex;
						fex = fex->next;
					}
				}
				
				fgl_unloadFontExt( fnt->ext[i]);
			}
		}
	}
#endif
	if ( fnt && fnt->nRefs ){
		fnt->nRefs--;
		
		if ( fnt->nRefs == 0 ){
			/* free the font and all it's entries */
			for ( fe = _fgl_fontCache; fe; ){
				if ( fe->fnt == fnt ){
					if ( fe == _fgl_fontCache ){
						_fgl_fontCache = fe->next;
					}
					else {
						fe1->next = fe->next;
					}

					fe2 = fe;
					fe = fe->next;

					FGL_FREE( fe2->fname);
					FGL_FREE( fe2);
				}
				else {
					fe1 = fe;
					fe = fe->next;
				}
			}

			fgl_unloadFont( fnt);
		}
	}

#endif /* USE_FENG_HDR */
}
#endif /* !FGL_PROTO */
;
/*}}}*/

FUNC_ATTR
int fgl_getMaxAdvance ( fgl_font_t* fnt ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
	int i, a = 0;
#if defined(LOCALE_INT)
	int j;
#endif
	if (fnt->maxAdvance != -1) return fnt->maxAdvance;
	for ( i=0; i<fnt->nGlyphs; i++ ){
		if ( fnt->advance[i] > a ){
			a = fnt->advance[i];
		}
	}
#if defined(LOCALE_INT)
	for ( j = 0 ; j < fnt->ext_number ; j++ ) {
	        for ( i = 0; i < fnt->ext[j]->nGlyphs ; i++ ) {
	                if ( fnt->ext[j]->advance[i] > a ) {
	                        a = fnt->ext[j]->advance[i];
	                } 
	        }	
	}	
#endif	
	fnt->maxAdvance = a;
	return a;
}
#endif /* !FGL_PROTO */
;
/*}}}*/

FUNC_ATTR
void fgl_drawText ( fgl_context_t* ctx, i32 x, i32 y, char* text, i32 len ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
  	char* s;
	int  i;
	fgl_font_t *fnt = ctx->fnt;
	fgl_monoBitmap_t bmp1;
	fgl_a2Bitmap_t   bmp2;

	if ( !fnt || !text )
		return;

	for ( s=text; (u32)len && *s; s++, len-- ){
		i = (*s - fnt->uniOffset);

		if ( (bmp1.width = bmp2.width = fnt->width[i]) ){

#ifdef DEBUG_FONT
			{ u8 *p8; int j, k;
			printf( "%c : %d,%d, %d\n", i, bmp1.width, bmp1.height, bmp1.sLength);
			for ( j=0, p8=bmp1.surface; j<bmp1.height; j++ ){
				for ( k=0; k<bmp1.sLength; k++, p8++ )
					printU8( *p8);
				putchar( '\n');
			}
			}
#endif

			if (fnt->antiAlias) {
			        bmp2.surface = fnt->surface + fnt->surfOff[i];
				bmp2.height  = fnt->height[i];
				bmp2.sLength = A2_SLENGTH(bmp2.width);

				fgl_a2BltBmp( ctx,
					      x + fnt->xBearing[i],
					      y - fnt->yBearing[i],
					      &bmp2, ctx->clr);
			} else {
			        bmp1.surface = fnt->surface + fnt->surfOff[i];
				bmp1.height  = fnt->height[i];
#ifdef ROT_90
				bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
				bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
				fgl_monoBltBmp( ctx,
						x + fnt->xBearing[i],
						y - fnt->yBearing[i],
						&bmp1, ctx->clr);
			}
		}

		x += fnt->advance[i];
	}
}
#endif /* !FGL_PROTO */
;
/*}}}*/

FUNC_ATTR
void fgl_drawWText ( fgl_context_t* ctx, i32 x, i32 y, u16* text, u32 len ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
	u16  *s, i;
	fgl_font_t *fnt = ctx->fnt;
	fgl_monoBitmap_t bmp1;
	fgl_a2Bitmap_t   bmp2;
#if defined(LOCALE_INT)
	u16  j;
	u8   match;
#endif

	if ( !fnt || !text )
		return;
#if defined(LOCALE_INT)
	for ( s=(u16*)text; len && *s; s++, len-- ){
	  match = 0;
	  if (*s > fnt->uniOffset && *s <= fnt->uniOffset+fnt->nGlyphs) {
	    i = *s - fnt->uniOffset;
	    if ( (bmp1.width = bmp2.width = fnt->width[i]) ){

	      if (fnt->antiAlias) {
		bmp2.surface = fnt->surface + fnt->surfOff[i];
		bmp2.height  = fnt->height[i];
		bmp2.sLength = A2_SLENGTH(bmp2.width);

		fgl_a2BltBmp( ctx,
			      x + fnt->xBearing[i],
			      y - fnt->yBearing[i],
			      &bmp2, ctx->clr);
	      } else {
		bmp1.surface = fnt->surface + fnt->surfOff[i];
		bmp1.height  = fnt->height[i];
#ifdef ROT_90
		bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
		bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif

		
		fgl_monoBltBmp( ctx,
				x + fnt->xBearing[i],
				y - fnt->yBearing[i],
				&bmp1, ctx->clr);
	      }
	    }
	    x += fnt->advance[i];
	    match = 1;
	  } else {
	    for ( j = 0 ; j < fnt->ext_number; j ++) {
	      if (*s > fnt->ext[j]->uniOffset && *s <= fnt->ext[j]->uniLast) {
		i = *s - fnt->ext[j]->uniOffset;		
		if ( (bmp1.width = bmp2.width = fnt->ext[j]->width[i]) ){
		  if (fnt->ext[j]->antiAlias) {

		    bmp2.surface = fnt->ext[j]->surface + fnt->ext[j]->surfOff[i];
		    bmp2.height  = fnt->ext[j]->height[i];
		    bmp2.sLength = A2_SLENGTH(bmp2.width);

		    fgl_a2BltBmp( ctx,
				  x + fnt->ext[j]->xBearing[i],
				  y - fnt->ext[j]->yBearing[i],
				  &bmp2, ctx->clr);
		  } else {
		    bmp1.surface = fnt->ext[j]->surface + fnt->ext[j]->surfOff[i];
		    bmp1.height  = fnt->ext[j]->height[i];
#ifdef ROT_90
		bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
		bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
		    fgl_monoBltBmp( ctx,
				    x + fnt->ext[j]->xBearing[i],
				    y - fnt->ext[j]->yBearing[i],
				    &bmp1, ctx->clr);

		  }
		}
		x += fnt->ext[j]->advance[i];
		match = 1;
		break;
	      }
	    }
	  }
	  if (!match) {
	    i = 0;
	    if ( (bmp1.width = bmp2.width = fnt->width[i]) ){
	      if (fnt->antiAlias) {
		bmp2.surface = fnt->surface + fnt->surfOff[i];
		bmp2.height  = fnt->height[i];
		bmp2.sLength = A2_SLENGTH(bmp2.width);

		fgl_a2BltBmp( ctx,
			      x + fnt->xBearing[i],
			      y - fnt->yBearing[i],
			      &bmp2, ctx->clr);
	      } else {
		bmp1.surface = fnt->surface + fnt->surfOff[i];
		bmp1.height  = fnt->height[i];
#ifdef ROT_90
		bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
		bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
		fgl_monoBltBmp( ctx,
				x + fnt->xBearing[i],
				y - fnt->yBearing[i],
				&bmp1, ctx->clr);
	      }
	    } 
	    x += fnt->advance[0];
	  }
	}
#else /* !LOCALE_INT */
	for ( s=(u16*)text; len && *s; s++, len-- ){

		i = *s - fnt->uniOffset;
		if ( i >= fnt->nGlyphs ) i = 0;

		if ( (bmp1.width = bmp2.width = fnt->width[i]) ){
		  if (fnt->antiAlias) {
		    bmp2.surface = fnt->surface + fnt->surfOff[i];
		    bmp2.height  = fnt->height[i];
		    bmp2.sLength = A2_SLENGTH(bmp2.width);
		    
		    fgl_a2BltBmp( ctx,
				  x + fnt->xBearing[i],
				  y - fnt->yBearing[i],
				  &bmp2, ctx->clr);
		  } else {
		    bmp1.surface = fnt->surface + fnt->surfOff[i];
		    bmp1.height  = fnt->height[i];
#ifdef ROT_90
		    bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
		    bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
		    fgl_monoBltBmp( ctx,
				    x + fnt->xBearing[i],
				    y - fnt->yBearing[i],
				    &bmp1, ctx->clr);
		  }
		}

		x += fnt->advance[i];
	}
#endif /* LOCALE_INT */
}
#endif /* !FGL_PROTO */
;
/*}}}*/

FUNC_ATTR
void fgl_drawWVertText ( fgl_context_t* ctx, i32 x, i32 y, u16* text, u32 len ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
	u16  *s, i;
	fgl_font_t *fnt = ctx->fnt;
	fgl_monoBitmap_t bmp1;
	fgl_a2Bitmap_t   bmp2;
	int xoff;
	int xadvance;
#if defined(LOCALE_INT)
	u16  j;
	u8   match;
#endif
	y += fnt->fontHeight;
	xadvance = fgl_getMaxAdvance(fnt);

	if ( !fnt || !text )
		return;
#if defined(LOCALE_INT)
	for ( s=(u16*)text; len && *s; s++, len-- ){
	  match = 0;
	  if (*s > fnt->uniOffset && *s <= fnt->uniOffset+fnt->nGlyphs) {
	    i = *s - fnt->uniOffset;
	    if ( (bmp1.width = bmp2.width = fnt->width[i]) ){
	      xoff = ( xadvance - fnt->width[i] - fnt->xBearing[i])/2;
	      if (fnt->antiAlias) {
		bmp2.surface = fnt->surface + fnt->surfOff[i];
		bmp2.height  = fnt->height[i];
		bmp2.sLength = A2_SLENGTH(bmp2.width);

		fgl_a2BltBmp( ctx,
			      x + fnt->xBearing[i]+xoff,
			      y - fnt->yBearing[i],
			      &bmp2, ctx->clr);
	      } else {
		bmp1.surface = fnt->surface + fnt->surfOff[i];
		bmp1.height  = fnt->height[i];
#ifdef ROT_90
		bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
		bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
		fgl_monoBltBmp( ctx,
				x + fnt->xBearing[i]+xoff,
				y - fnt->yBearing[i],
				&bmp1, ctx->clr);
	      }
	    }
	    y += fnt->fontHeight;
	    match = 1;
	  } else {
	    for ( j = 0 ; j < fnt->ext_number; j ++) {
	      if (*s > fnt->ext[j]->uniOffset && *s <= fnt->ext[j]->uniLast) {
		i = *s - fnt->ext[j]->uniOffset;		
		if ( (bmp1.width = bmp2.width = fnt->ext[j]->width[i]) ){

		  xoff = ( xadvance - fnt->ext[j]->width[i] - fnt->ext[j]->xBearing[i])/2;

		  if (fnt->ext[j]->antiAlias) {

		    bmp2.surface = fnt->ext[j]->surface + fnt->ext[j]->surfOff[i];
		    bmp2.height  = fnt->ext[j]->height[i];
		    bmp2.sLength = A2_SLENGTH(bmp2.width);

		    fgl_a2BltBmp( ctx,
				  x + fnt->ext[j]->xBearing[i]+xoff,
				  y - fnt->ext[j]->yBearing[i],
				  &bmp2, ctx->clr);
		  } else {
		    bmp1.surface = fnt->ext[j]->surface + fnt->ext[j]->surfOff[i];
		    bmp1.height  = fnt->ext[j]->height[i];
#ifdef ROT_90
		    bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
		    bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
		    fgl_monoBltBmp( ctx,
				    x + fnt->ext[j]->xBearing[i]+xoff,
				    y - fnt->ext[j]->yBearing[i],
				    &bmp1, ctx->clr);

		  }

		}

		y += fnt->ext[j]->height[i]+2;
		match = 1;
		break;
	      }
	    }
	  }
	  if (!match) {

	    i = 0;
	    if ( (bmp1.width = bmp2.width = fnt->width[i]) ){
	      xoff = ( xadvance - fnt->width[i] - fnt->xBearing[i])/2;
	      if (fnt->antiAlias) {
		bmp2.surface = fnt->surface + fnt->surfOff[i];
		bmp2.height  = fnt->height[i];
		bmp2.sLength = A2_SLENGTH(bmp2.width);

		fgl_a2BltBmp( ctx,
			      x + fnt->xBearing[i]+xoff,
			      y - fnt->yBearing[i],
			      &bmp2, ctx->clr);
	      } else {
		bmp1.surface = fnt->surface + fnt->surfOff[i];
		bmp1.height  = fnt->height[i];
#ifdef ROT_90
		bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
		bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
		fgl_monoBltBmp( ctx,
				x + fnt->xBearing[i]+xoff,
				y - fnt->yBearing[i],
				&bmp1, ctx->clr);
	      }
	    }
	    y += fnt->height[i] + fnt->yBearing[i];
	  }
	}
#else /* !LOCALE_INT */
	for ( s=(u16*)text; len && *s; s++, len-- ){

	  i = *s - fnt->uniOffset;
	  if ( i >= fnt->nGlyphs ) i = 0;
	  if ( (bmp1.width = bmp2.width = fnt->width[i]) ){
	    xoff = ( xadvance - fnt->width[i] - fnt->xBearing[i])/2;
	    if (fnt->antiAlias) {
	      bmp2.surface = fnt->surface + fnt->surfOff[i];
	      bmp2.height  = fnt->height[i];
	      bmp2.sLength = A2_SLENGTH(bmp2.width);
	      
	      fgl_a2BltBmp( ctx,
			    x + fnt->xBearing[i]+xoff,
			    y - fnt->yBearing[i],
			    &bmp2, ctx->clr);
	    } else {
	      bmp1.surface = fnt->surface + fnt->surfOff[i];
	      bmp1.height  = fnt->height[i];
#ifdef ROT_90
	      bmp1.sLength = MONO_SLENGTH(bmp1.height);
#else
	      bmp1.sLength = MONO_SLENGTH(bmp1.width);
#endif
	      
	      fgl_monoBltBmp( ctx,
			      x + fnt->xBearing[i]+xoff,
			      y - fnt->yBearing[i],
			      &bmp1, ctx->clr);
	    }

	    
	  }
	  y += fnt->fontHeight;
	}
#endif /* LOCALE_INT */
}
#endif /* !FGL_PROTO */
;
/*}}}*/

FUNC_ATTR
int fgl_getTextLength ( fgl_font_t* fnt, char* text, u32 len ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
	int w = 0;
	u16 i;
	char *s;
	
	for ( s=text; len && *s; s++, len-- ) {
		i = *s - fnt->uniOffset;
		if ( i < fnt->nGlyphs ){
			w += fnt->advance[i];
		}
	}

	return w;
}
#endif /* !FGL_PROTO */
;
/*}}}*/

FUNC_ATTR
int fgl_getWTextLength ( fgl_font_t* fnt, u16* text, u32 len ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
	int w = 0;
	u16 *s, i;
#if defined(LOCALE_INT)
	int j;
	int match;

	for ( s=(u16*)text; len && *s; s++, len-- ){
	        match = 0;
	        if (*s > fnt->uniOffset && *s <= fnt->uniOffset+fnt->nGlyphs) {
	                i = *s - fnt->uniOffset;
	                w += fnt->advance[i];
			match = 1;
		}
		else {
		        for ( j = 0 ; j < fnt->ext_number; j ++) {
	                      if (*s > fnt->ext[j]->uniOffset && *s <= fnt->ext[j]->uniLast) {
				       i = *s - fnt->ext[j]->uniOffset;		
				       w += fnt->ext[j]->advance[i];
				       match = 1;
			      }
		        }
		}
		if (!match) w += fnt->advance[0];
	}
#else 
	for ( s=(u16*)text; len && *s; s++, len-- ) {
		i = *s - fnt->uniOffset;
		if ( i < fnt->nGlyphs ){
			w += fnt->advance[i];
		}
	}

#endif
	return w;
}
#endif /* !FGL_PROTO */
;
/*}}}*/

FUNC_ATTR
int fgl_getFixedWidth ( fgl_font_t* fnt ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_FONT)
{
	int i, a = fnt->advance[0];
#if defined(LOCALE_INT)
	int j;
#endif

	for ( i=1; i<fnt->nGlyphs; i++ ){
		if ( fnt->advance[i] != a ){
			return 0;
		}
	}
#if defined(LOCALE_INT)
	for ( j = 0 ; j < fnt->ext_number ; j++ ) {
	        for ( i = 0; i < fnt->ext[j]->nGlyphs ; i++ ) {
	                if ( fnt->ext[j]->advance[i] != a ) {
	                        return 0;
	                } 
	        }	
	}	
#endif

	return a;
}
#endif /* !FGL_PROTO */
;
/*}}}*/

#undef FUNC_ATTR

#endif /* FGL_FONT_INCLUDED */

