/* -*- mode: C; mode: fold -*- */
/*
 * init_fb.c - FGL_INIT_MODULE to initialize the framebuffer device and
 *             global fgl variables (Linux version)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: init_fb.c,v 1.47 2001/10/10 18:36:42 jsimmons Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_INIT_INCLUDED
#define FGL_INIT_INCLUDED

#include <asm/page.h>

#define FGL_MODULE
#define FGL_MAIN
#include "fgl.h"

#include FGL_HEAP_MODULE
#include "util.c"
#include "screencfg.c"
#include "rect.c"

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif
/*}}}*/


/* globals and platform specifics */ /*{{{*/
#include <linux/fb.h>
/* the non page boundary fb address fix */
#include <sys/user.h>
#if defined(USE_VT_CTRL)
#include <linux/kd.h>
#endif

FGL_GLOBAL int                        _fgl_fh  FGL_INIT( -1);
FGL_GLOBAL struct fb_var_screeninfo   _fgl_varInfo;
FGL_GLOBAL struct fb_fix_screeninfo   _fgl_fixInfo;

#ifndef FGL_PROTO

static void* scr;

static int fdvt = -1;

#endif
/*}}}*/

static inline
int fgl_getScreenWidth () /*{{{*/
{
	return fgl_screenCtx->bmp->width;
}
/*}}}*/

static inline
int fgl_getScreenHeight () /*{{{*/
{
	return fgl_screenCtx->bmp->height;
}
/*}}}*/

static inline
int fgl_getScreenXoff () /*{{{*/
{
	return fgl_screenCtx->xoff;
}
/*}}}*/

static inline
int fgl_getScreenYoff () /*{{{*/
{
	return fgl_screenCtx->yoff;
}
/*}}}*/


/*
 * fgl_terminate - clean up system resources
 * 
 * This is automatically called during process exit, so don't inline.
 * It also has to be able to detect multiple calls
 */
void fgl_terminate () /*{{{*/
#if !defined(FGL_PROTO)
{
	char str[20];

	if ( scr ){
		munmap( scr, _fgl_fixInfo.smem_len);
		scr = 0;
	}
	
	if ( _fgl_fh >= 0 ) {
		close( _fgl_fh);
		_fgl_fh = -1;
	}
	
	if (fdvt >= 0) {
#if defined(USE_VT_CTRL)
		if (ioctl(fdvt, KDSETMODE, KD_TEXT) < 0) {
			err( "failed to set KD_TEXT mode\n");
		}
#endif
		sprintf(str, "\033[?25h"); // Turn on cursor
		write(fdvt, str, 6); 
		sprintf(str, "\0338"); // Restore saved state
		write(fdvt, str, 2); 
		sprintf(str, "\033[9;15]"); // Set screen blank timeout
					    // to 15 minutes
		write(fdvt, str, 7); 
		close(fdvt);
		fdvt = -1;
	}
}
#endif /* FGL_PROTO */
;/*}}}*/


/*
 * This is our color mapping set up defined in the cmap_XX module. We
 * can't inline because cmap uses our colormap functions (which are
 * device dependent)
 */
extern int  fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld );
extern void fgl_fillRectClr ( fgl_context_t *ctx, i32 x0, i32 y0,
			      i32 x1, i32 y1, fgl_color_t clr );

/*
 * fgl_init - initialize system resources, screenBmp and color map
 * 
 *   nReq    : number of requested RGB cells (pseudocolor only)
 *   rgbReqs : 888 RGB values of requested colors (pseudocolor only)
 *   restore : 1: restore colormap when process exits (pseudocolor only)
 */
int fgl_init ( int nReq, u32* rgbReqs, int restore ) /*{{{*/
#if !defined(FGL_PROTO)
{
	char  *fbdev;
	fgl_config_t cfg;

	char  *vtdev;
	char str[20];
	

	vtdev = getenv( "VTCTRL");
	if ( !vtdev ){
		vtdev = DEFAULT_VT;
	}
	fdvt = open(vtdev, O_RDWR|O_NDELAY, 0);
	if (fdvt < 0) {
		err( "failed to open device: '%s'\n", vtdev);
		return 0;
	}
	sprintf(str, "\0337"); // Save state
	write(fdvt, str, 2); 
	sprintf(str, "\033[9;0]"); // Set screen blank timeout
				   // to 0 minutes (disable)
	write(fdvt, str, 6); 
	sprintf(str, "\033[?25l"); // Turn off cursor
	write(fdvt, str, 6); 
#if defined(USE_VT_CTRL)
	if (ioctl(fdvt, KDSETMODE, KD_GRAPHICS) < 0) {
		err( "failed to set KD_GRAPHICS mode\n");
		return 0;
	}
#endif

	fbdev = getenv( "FRAMEBUFFER");
	if ( !fbdev ){
		fbdev = DEFAULT_FB;
	}

	if ( (_fgl_fh = open( fbdev, O_RDWR)) < 0 ){
		err( "failed to open device: '%s'\n", fbdev);
		return 0;
	}
  
	/* get framebuffer info */
	if ( ioctl( _fgl_fh, FBIOGET_VSCREENINFO, &_fgl_varInfo) ){
		err( "FBIOGET_VSCREENINFO failed: %s\n", strerror(errno));
		return 0;
	}

	/* scroll to top of virtual framebuffer */
	_fgl_varInfo.xoffset=0;
	_fgl_varInfo.yoffset=0;
	if ( ioctl( _fgl_fh, FBIOPUT_VSCREENINFO, &_fgl_varInfo) ){
		err( "FBIOPUT_VSCREENINFO failed: %s\n", strerror(errno));
		return 0;
	}

	/* Turn panning off. This could fail on some cards but we don't care */
	_fgl_varInfo.xoffset=0;
	_fgl_varInfo.yoffset=0;
	ioctl( _fgl_fh, FBIOPAN_DISPLAY, &_fgl_varInfo);

	if ( ioctl( _fgl_fh, FBIOGET_FSCREENINFO, &_fgl_fixInfo) ){
		err( "FBIOGET_FSCREENINFO failed: %s\n", strerror(errno));
		return 0;
	}

	/* check if it's compatible with our build-settings */
	if ( _fgl_varInfo.bits_per_pixel != fgl_getBitsPerPixel() ){
		err( "wrong framebuffer driver %d (display is %d)\n",
			 fgl_getBitsPerPixel(), _fgl_varInfo.bits_per_pixel);
		return 0;
	}

	/* map framebuffer device into our memory */
	scr = mmap( NULL, _fgl_fixInfo.smem_len,
				PROT_READ | PROT_WRITE, MAP_SHARED,
				_fgl_fh, 0);
	if ( scr == MAP_FAILED ){
		puts( "framebuffer mmap failed");
		return 0;
	}

	if ( atexit( fgl_terminate) == -1 ){
		puts( "failed to install exit handler");
	}

	/*
	 * in case of MobilePro the actual memory address of the fb 
	 * is not aligned on the page boundary. Calling mmap therefore
	 * returns memory aligned address to which we have to use
	 * offset based on the actual memory location. The only case
	 * so far is MobilePro 780 but we will do it for the benefit
	 * of other new platforms that may exhibit same problem
	 */
	fgl_screenBmp->surface = (fgl_slc_t*) ((char *)scr + ((long)_fgl_fixInfo.smem_start % (int)PAGE_SIZE));
#ifdef ROT_90
	fgl_screenBmp->width   = _fgl_varInfo.yres; /* might be bigger than xres */
	fgl_screenBmp->height  = _fgl_varInfo.xres;
#else
	fgl_screenBmp->width   = _fgl_varInfo.xres; /* might be bigger than xres */
	fgl_screenBmp->height  = _fgl_varInfo.yres;
#endif
	fgl_screenBmp->state  |= IMAGE_ON_SCREEN;
	/*
	 * The actual length of the fb line and screen with may differ. The line
	 * length calcualtion should be based therefore on the fb linegth returned
	 * in fixScreenInfo structure. for multiple byte pixels sLnegth holds number
	 * of pixels, otherwise it holds numnber of bytes.
	 */

	fgl_screenBmp->sLength = _fgl_fixInfo.line_length * 8 / _fgl_varInfo.bits_per_pixel;

	fgl_screenCtx->bmp     = fgl_screenBmp;
	fgl_screenCtx->alpha   = 0xff;
	fgl_screenCtx->x1Clip  = fgl_screenBmp->width-1;
	fgl_screenCtx->y1Clip  = fgl_screenBmp->height-1;
	fgl_screenCtx->xoff    = 0;
	fgl_screenCtx->yoff    = 0;

	/* set up any color mapping we might need */
	fgl_initRgbCmap ( nReq, rgbReqs, restore );

	/* reset the screen dimensions if config file exists */
	if (fgl_readConfig( &cfg)) {
		if ( cfg.xoff != -1 && cfg.yoff != -1 &&
		     cfg.width != -1 && cfg.height != -1 ) {

                        if( cfg.xoff ) {
	                    fgl_fillRectClr( fgl_screenCtx, 0, 0, cfg.xoff - 1, fgl_screenBmp->height - 1, 0 );
                        }
                        if ( cfg.yoff )	{
			     fgl_fillRectClr( fgl_screenCtx, cfg.xoff, 0, fgl_screenBmp->width - 1, cfg.yoff - 1, 0);
			}
			if ( fgl_screenBmp->width - cfg.xoff - cfg.width > 0 ) {
			     fgl_fillRectClr( fgl_screenCtx, cfg.xoff + cfg.width, cfg.yoff, fgl_screenBmp->width - 1, fgl_screenBmp->height - 1, 0 );
			}
			if ( fgl_screenBmp->height - cfg.yoff - cfg.height > 0 ) {
			     fgl_fillRectClr( fgl_screenCtx, cfg.xoff, cfg.yoff + cfg.height, cfg.xoff + cfg.width - 1, fgl_screenBmp->height - 1, 0 );
			}
#ifdef ROT_90
			fgl_screenBmp->surface += cfg.yoff +
				cfg.xoff * fgl_screenBmp->sLength -
				cfg.height;
#else
			fgl_screenBmp->surface += cfg.xoff + 
				cfg.yoff * fgl_screenBmp->sLength;
#endif
			fgl_screenBmp->width   = cfg.width;
			fgl_screenBmp->height  = cfg.height;


			fgl_screenCtx->xoff    = cfg.xoff;
			fgl_screenCtx->yoff    = cfg.yoff;

			fgl_screenCtx->x1Clip  = fgl_screenBmp->width-1;
			fgl_screenCtx->y1Clip  = fgl_screenBmp->height-1;
		}
	}
	/* run accel init if any */
#ifdef FGL_ACCELFN_INIT
	fgl_accel_init( );
#endif
	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/


/*
 * fgl_initialize - convenience wrapper for fgl_init w/o RGB requests
 */
static inline
int fgl_initialize () /*{{{*/
{	
	return fgl_init( 0, 0, 1);
}
/*}}}*/

/*
 * fgl_initialize - convenience wrapper for fgl_init w/o RGB requests
 * specifying size
 */
static inline
int fgl_open (u32 x, u32 y, u32 w, u32 h) /*{{{*/
{	
        int r;

        r = fgl_init( 0, 0, 1);

	fgl_screenBmp->width   = w; 
	fgl_screenBmp->height  = h;

#ifdef ROT_90
	rotate32( fgl_screenBmp, &x, &y);
	fgl_screenCtx->x1Clip  = h-1;
	fgl_screenCtx->y1Clip  = w-1;
#else
	fgl_screenCtx->x1Clip  = w-1;
	fgl_screenCtx->y1Clip  = h-1;
#endif
	fgl_screenBmp->surface = fgl_getSlc( fgl_screenBmp, x, y);
	return r;
}
/*}}}*/

/****************************************************************************
 * These are the low level functions to manage device specific colormaps.
 * Note that we can't inline them, because they are used from the cmap_XX
 * module, which is included
 */

static inline
int fgl_hasCmap () /*{{{*/
{
#if defined (IS_CMAP_MODE)
	return 1;
#else
	return 0;
#endif
}/*}}}*/

fgl_cmap_t* fgl_allocCmap () /*{{{*/
#if !defined(FGL_PROTO)
{
	fgl_cmap_t* cm = 0;

#if defined (IS_CMAP_MODE)
	int i, n;

	if ( fgl_hasCmap() ){
		for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
		cm = FGL_CALLOC( 1, sizeof(fgl_cmap_t) + 3 * n);

		cm->r = (u8*)cm + sizeof(fgl_cmap_t);
		cm->g = cm->r + n;
		cm->b = cm->g + n;
	}
#endif /* IS_CMAP_MODE */

	return cm;
}
#endif /* FGL_PROTO */
;/*}}}*/

void fgl_freeCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
	FGL_FREE( cm);
}
#endif /* FGL_PROTO */
;/*}}}*/

int fgl_getHWCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i, n;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;

	for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
	cmap.len = n;

	buf = (u16*) FGL_MALLOC( n * 3 * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + n;
	cmap.blue  = cmap.green + n;
	cmap.transp = 0;

	if ( ioctl( _fgl_fh, FBIOGETCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}

	cm->len = cmap.len;
	for ( i=0; i<cmap.len; i++ ){
		cm->r[i] = cmap.red[i] >> 8;
		cm->g[i] = cmap.green[i] >> 8;
		cm->b[i] = cmap.blue[i] >> 8;
	}

	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/

int fgl_setHWCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;
	cmap.len = cm->len;

	buf = (u16*) FGL_MALLOC( 3 * cm->len * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + cmap.len;
	cmap.blue  = cmap.green + cmap.len;
	cmap.transp = 0;

	for ( i=0; i<cm->len; i++ ){
		cmap.red[i]   = (u16)(cm->r[i]) << 8;
		cmap.green[i] = (u16)(cm->g[i]) << 8;
		cmap.blue[i]  = (u16)(cm->b[i]) << 8;
	}

	if ( ioctl( _fgl_fh, FBIOPUTCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}
	
	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/

/* fgl module prolog */ /*{{{*/
#endif /* FGL_INIT_INCLUDED */
/*}}}*/
