/* -*- mode: C; mode: fold -*- */
/*
 * init_fbx.c - FGL_INIT_MODULE to initialize the framebuffer device and
 *             global fgl variables (Linux version)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: init_fbx.c,v 1.5 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_INIT_INCLUDED
#define FGL_INIT_INCLUDED

#include <asm/page.h>

#define FGL_MODULE
#define FGL_MAIN
#include "fgl.h"

#include FGL_HEAP_MODULE
#include "util.c"
#include "screencfg.c"

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif
/*}}}*/


/* globals and platform specifics */ /*{{{*/

#include <sys/shm.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* the non page boundary fb address fix */
#include <sys/user.h>
#if defined(USE_VT_CTRL)
#include <linux/kd.h>
#endif

int socketfd;


/*}}}*/

static inline
int fgl_getScreenWidth () /*{{{*/
{
	return fgl_screenCtx->bmp->width;
}
/*}}}*/

static inline
int fgl_getScreenHeight () /*{{{*/
{
	return fgl_screenCtx->bmp->height;
}
/*}}}*/

static inline
int fgl_getScreenXoff () /*{{{*/
{
	return fgl_screenCtx->xoff;
}
/*}}}*/

static inline
int fgl_getScreenYoff () /*{{{*/
{
	return fgl_screenCtx->yoff;
}
/*}}}*/


/*
 * fgl_terminate - clean up system resources
 * 
 * This is automatically called during process exit, so don't inline.
 * It also has to be able to detect multiple calls
 */
void fgl_terminate () /*{{{*/
#if !defined(FGL_PROTO)
{
	unsigned short packet = 2;
	write( socketfd, &packet, sizeof(packet));

	close(socketfd);
}
#endif /* FGL_PROTO */
;/*}}}*/

void fgl_updateScreen(int x0, int y0, int x1, int y1)
#if !defined(FGL_PROTO)
{
	unsigned short packet = 1;
	
	write( socketfd, &packet, sizeof(packet));
	write( socketfd, &x0, sizeof(int));
	write( socketfd, &y0, sizeof(int));
	write( socketfd, &x1, sizeof(int));
	write( socketfd, &y1, sizeof(int));

}
#endif /* FGL_PROTO */
;/*}}}*/


/*
 * This is our color mapping set up defined in the cmap_XX module. We
 * can't inline because cmap uses our colormap functions (which are
 * device dependent)
 */
extern int fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld );

/*
 * fgl_init - initialize system resources, screenBmp and color map
 * 
 *   nReq    : number of requested RGB cells (pseudocolor only)
 *   rgbReqs : 888 RGB values of requested colors (pseudocolor only)
 *   restore : 1: restore colormap when process exits (pseudocolor only)
 */
int fgl_init ( int nReq, u32* rgbReqs, int restore ) /*{{{*/
#if !defined(FGL_PROTO)
{

	int i;
	fgl_config_t cfg;
	int xoff = 100;
	int yoff = 100;
	int width = 640;
	int height = 480;
	struct sockaddr_in serveraddr;
	char tcpip[256];
	int shmid;
	void *shmaddr;

	if (fgl_readConfig( &cfg)) {
	  if ( cfg.xoff != -1 && cfg.yoff != -1 &&
	       cfg.width != -1 && cfg.height != -1 ) {
	    xoff   = cfg.xoff;
	    yoff   = cfg.yoff;
	    width  = cfg.width;
	    height = cfg.height;
	  }
	}

	socketfd = socket( AF_INET, SOCK_STREAM, 0);
	bzero( &serveraddr, sizeof(serveraddr) );
	serveraddr.sin_family = AF_INET;
	serveraddr.sin_port = htons(6032);
	inet_pton( AF_INET, "127.0.0.1", &(serveraddr.sin_addr));

	if (connect( socketfd, (const struct sockaddr *)&serveraddr, sizeof(serveraddr))){
		err ("failed to connect to server.\n");
		return 0;
	}

	write(socketfd,&width,sizeof(width));
	write(socketfd,&height,sizeof(height));

	i = 0;
	while (i < sizeof(shmid)) {
		i += read(socketfd, tcpip+i, sizeof(shmid) - i);
	}
	bcopy( tcpip, &shmid, sizeof( shmid ));
	i = 0;
	while (i < sizeof(shmaddr)) {
		i += read(socketfd, tcpip+i, sizeof(shmaddr) - i);
	}

	bcopy( tcpip, &shmaddr, sizeof( shmaddr ));

	fgl_screenBmp->surface = (fgl_slc_t*) shmat( shmid, shmaddr, 0);

	if (!fgl_screenBmp->surface) {
		err ("failed to map shared memory.\n");
		return 0;
	}	  

	for( i = 0; i< width * height; i++) {
		fgl_screenBmp->surface[i] = 0;
	}

	fgl_updateScreen(0,0,width-1,height-1);


	fgl_initRgbCmap ( nReq, rgbReqs, restore );

	fgl_screenBmp->width  = width;
	fgl_screenBmp->height = height;
	fgl_screenBmp->state  |= IMAGE_ON_SCREEN;
	fgl_screenBmp->sLength = width;
	fgl_screenCtx->bmp     = fgl_screenBmp;
	fgl_screenCtx->alpha   = 0xff;
	fgl_screenCtx->x1Clip  = fgl_screenBmp->width-1;
	fgl_screenCtx->y1Clip  = fgl_screenBmp->height-1;
	fgl_screenCtx->xoff    = 0;
	fgl_screenCtx->yoff    = 0;

	return 1;
}
#endif
;/*}}}*/


/*
 * fgl_initialize - convenience wrapper for fgl_init w/o RGB requests
 */
static inline
int fgl_initialize () /*{{{*/
{	
	return fgl_init( 0, 0, 1);
}
/*}}}*/

/*
 * fgl_initialize - convenience wrapper for fgl_init w/o RGB requests
 * specifying size
 */
static inline
int fgl_open (u32 x, u32 y, u32 w, u32 h) /*{{{*/
{	
        int r;

        r = fgl_init( 0, 0, 1);

	fgl_screenBmp->width   = w; 
	fgl_screenBmp->height  = h;

#ifdef ROT_90
	rotate32( fgl_screenBmp, &x, &y);
	fgl_screenCtx->x1Clip  = h-1;
	fgl_screenCtx->y1Clip  = w-1;
#else
	fgl_screenCtx->x1Clip  = w-1;
	fgl_screenCtx->y1Clip  = h-1;
#endif
	fgl_screenBmp->surface = fgl_getSlc( fgl_screenBmp, x, y);
	return r;
}
/*}}}*/

/****************************************************************************
 * These are the low level functions to manage device specific colormaps.
 * Note that we can't inline them, because they are used from the cmap_XX
 * module, which is included
 */

static inline
int fgl_hasCmap () /*{{{*/
{
#if defined (IS_CMAP_MODE)
	return 1;
#else
	return 0;
#endif
}/*}}}*/

fgl_cmap_t* fgl_allocCmap () /*{{{*/
#if !defined(FGL_PROTO)
{
	fgl_cmap_t* cm = 0;

#if defined (IS_CMAP_MODE)
	int i, n;

	if ( fgl_hasCmap() ){
		for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
		cm = FGL_CALLOC( 1, sizeof(fgl_cmap_t) + 3 * n);

		cm->r = (u8*)cm + sizeof(fgl_cmap_t);
		cm->g = cm->r + n;
		cm->b = cm->g + n;
	}
#endif /* IS_CMAP_MODE */

	return cm;
}
#endif /* FGL_PROTO */
;/*}}}*/

void fgl_freeCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
	FGL_FREE( cm);
}
#endif /* FGL_PROTO */
;/*}}}*/

int fgl_getHWCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i, n;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;

	for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
	cmap.len = n;

	buf = (u16*) FGL_MALLOC( n * 3 * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + n;
	cmap.blue  = cmap.green + n;
	cmap.transp = 0;

	if ( ioctl( _fgl_fh, FBIOGETCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}

	cm->len = cmap.len;
	for ( i=0; i<cmap.len; i++ ){
		cm->r[i] = cmap.red[i] >> 8;
		cm->g[i] = cmap.green[i] >> 8;
		cm->b[i] = cmap.blue[i] >> 8;
	}

	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/

int fgl_setHWCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;
	cmap.len = cm->len;

	buf = (u16*) FGL_MALLOC( 3 * cm->len * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + cmap.len;
	cmap.blue  = cmap.green + cmap.len;
	cmap.transp = 0;

	for ( i=0; i<cm->len; i++ ){
		cmap.red[i]   = (u16)(cm->r[i]) << 8;
		cmap.green[i] = (u16)(cm->g[i]) << 8;
		cmap.blue[i]  = (u16)(cm->b[i]) << 8;
	}

	if ( ioctl( _fgl_fh, FBIOPUTCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}
	
	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/

/* fgl module prolog */ /*{{{*/
#endif /* FGL_INIT_INCLUDED */
/*}}}*/
