/* -*- mode: C; mode: fold -*- */
/*
 * mbmp.c - fgl module to handle monochrome bitmaps (depth 1)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: mbmp.c,v 1.24 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/

#ifndef FGL_MBMP_INCLUDED
#define FGL_MBMP_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* fgl imports */
#include "util.c"
#include FGL_DRAW_MODULE
#include FGL_HEAP_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif


#ifdef FGL_INLINE_MBMP
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif

/*}}}*/


#if !defined(FGL_PROTO)
u8 _monoBits[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
#else
extern u8 _monoBits[8];
#endif


static inline
void fgl_setMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) /*{{{*/
{
	u8 *p;
#ifdef ROT_90
	rotate32( (fgl_bitmap_t *)bmp, &col, &row);
#endif
	p = bmp->surface + (row * bmp->sLength) + (col / 8);
	*p |= _monoBits[col % 8];
}
/*}}}*/

static inline
void fgl_unsetMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) /*{{{*/
{
	u8 *p;
#ifdef ROT_90
	rotate32( (fgl_bitmap_t *)bmp, &col, &row);
#endif
	p = bmp->surface + (row * bmp->sLength) + (col / 8);
	*p &= ~_monoBits[col % 8];
}
/*}}}*/

static inline
int fgl_isMonoBitSet ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) /*{{{*/
{
	u8 *p;
#ifdef ROT_90
	rotate32( (fgl_bitmap_t *)bmp, &col, &row);
#endif
	p = bmp->surface + (row * bmp->sLength) + (col / 8);
	return (*p & _monoBits[col % 8]);
}
/*}}}*/
#ifdef ROT_90
static inline
void fgl_rotate90MonoSurface ( u8* src, u8* dst, u32 width, u32 height ) /*{{{*/
{
	u32 x, y;
	u32 u, v;
	u8* p;
	u32 sLength1, sLength2;
	u8 pix;

	sLength1 = ( width  + 7 ) / 8;
	sLength2 = ( height + 7 ) / 8;

	for( y = 0 ; y < height ; y++ ) {
		for( x = 0 ; x < width ; x++ ) {

			u = height - y - 1;
			v = x;

			pix = *( src + ( y * sLength1 ) + ( x / 8 )) & _monoBits[x % 8];
			p = dst + ( v * sLength2 ) + ( u / 8 );

			if ( pix ) {
				*p |= _monoBits[u % 8];
			} else {
				*p &= ~_monoBits[u % 8];
			}
			
		}
	}
}
/*}}}*/

static inline
void fgl_rotate90MonoBitmap ( fgl_monoBitmap_t* src , int rot) /*{{{*/
{
	u8 *temp;
	u32 size = ( src->width + 7 ) * ( src->height + 7 ) / 8;
	u32 sLength = ( src->height + 7 ) / 8;

	temp = (u8 *) FGL_MALLOC ( sizeof(u8) * sLength * src->width);

	if ( rot )fgl_rotate90MonoSurface ( src->surface, temp, src->width, src->height);

	memcpy( src->surface, temp, size );
	
	src->sLength = sLength;

	FGL_FREE(temp);

}
/*}}}*/
#endif

FUNC_ATTR
fgl_monoBitmap_t* fgl_createMonoBitmap ( u32 width, u32 height, int init ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_MBMP)
{

	u32 sLength = (width + 7) / 8;
	u32 size =  ( width + 7 ) * ( height + 7 ) / 8;
	fgl_monoBitmap_t *bmp = (fgl_monoBitmap_t*) FGL_MALLOC( sizeof(fgl_monoBitmap_t) +size);
	
	if ( !bmp ){
		return 0;
	}
	
	bmp->sLength = sLength;
	bmp->width = width;
	bmp->height = height;

	bmp->surface = (u8*)bmp + sizeof(fgl_monoBitmap_t);
	
	if ( init != -1 ){
		memset( bmp->surface, init, size);
#ifdef ROT_90
		fgl_rotate90MonoBitmap( bmp, 1 );
#endif
	}

	return bmp;
}
#endif /* !FGL_PROTO */
; /*}}}*/


static inline
void fgl_disposeMonoBitmap ( fgl_monoBitmap_t* bmp ) /*{{{*/
{
	FGL_FREE( bmp);
}
/*}}}*/


FUNC_ATTR
void fgl_monoBlt ( fgl_context_t* dst, i32 u0, i32 v0,
				   fgl_monoBitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, 
				   fgl_color_t clr ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_MBMP)
{
	u8 *ps;
	fgl_slc_t *pd;
	i32 u, u1, v1, sinc, dinc;
	i32 w, h;
	register u8 a, i;
	u8 i0, off;

	if ( x0 > x1 )  swap32( &x0, &x1);
	if ( y0 > y1 )  swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90
	rotate32( ( fgl_bitmap_t *) src, (u32*)&x0, (u32*)&y0);
	rotate32( ( fgl_bitmap_t *) src, (u32*)&x1, (u32*)&y1);
	rotate32( dst->bmp, (u32*)&u0, (u32*)&v0);
	rotate32( dst->bmp, (u32*)&u1, (u32*)&v1);
	swap32( &x0, &x1);
	swap32( &u0, &u1);
#endif

#ifdef FGL_ACCELFN_MONOBLT
	if (dst->bmp->state & IMAGE_ON_SCREEN) {

	  fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_ACCEL_MONOBLT_OP );
	  fgl_accel_monoBlt( fgl_screenCtx, u0, v0, src, x0, y0, x1, y1, clr);
	  fgl_endDraw( dst->bmp, u0, v0, u1, v1);
	  return ;
	}
#endif

	w = x1 - x0;
	h = y1 - y0;
	off = x0 % 8;

	ps = src->surface + (y0 * src->sLength) + (x0 / 8);
	pd = fgl_getSlc( dst->bmp, u0, v0);

	a = *ps;
	u = u0;
	u1 = u0 + w;
	i = i0 = 8-off;

	dinc = (dst->bmp->sLength - u1) + u0;                 /* in pixels */
	sinc = src->sLength + (x0>>3) - (x1>>3);  /* in bytes */

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( h-- >= 0 ){
		a <<= off;

		while ( 1 ){
			if ( a & 0x80 ){
				fgl_setSlcPixel( pd, u, clr);
			}

#ifdef DEBUG_MBMP
			putchar( (a & 0x80) ? 'O' : '.');
			if ( u == u1 ) putchar( '\n');
#endif

			if ( u == u1 ){
				break;
			}
		   
			if ( --i == 0 ){
				/* next source byte */
				i = 8;
				a = *(++ps);
			}
			else {
				a <<= 1;
			}

			fgl_incSlc( &pd, u);
			u++;
		}

		/* next source and destination scanlines */
		fgl_addSlc( &pd, u, dinc);
		ps += sinc;
		a = *ps;
		u = u0;
		i = i0;
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}
#endif /* !FGL_PROTO */
;/*}}}*/


static inline
void fgl_monoBltBmp ( fgl_context_t* dst, i32 u0, i32 v0,
					  fgl_monoBitmap_t* src, fgl_color_t clr ) /*{{{*/
{
	fgl_monoBlt( dst, u0, v0, src, 0, 0, src->width-1, src->height-1, clr);
}
/*}}}*/


/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_MBMP_INCLUDED */
/*}}}*/
