/* -*- mode: C; mode: fold -*- */
/*
 * sprite.c - fgl module to handle sprites (transparent bitmaps with
 *            screen save/restore) and cursor (global sprite)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: spri.c,v 1.20 2001/11/28 20:18:35 damian Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_SPRI_INCLUDED
#define FGL_SPRI_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* fgl imports */
#include FGL_CMAP_MODULE
#include "cntx.c"
#include "mbmp.c"
#include "cbmp.c"

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif

#ifdef FGL_INLINE_SPRI
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif
/*}}}*/


static inline
void fgl_hideSprite ( fgl_context_t* ctx, fgl_sprite_t* sp ) /*{{{*/
{
	if ( sp->isVisible ){
		sp->isVisible = 0;
		fgl_waitReady( ctx->bmp );
		fgl_bitBltBmp( ctx, sp->x0, sp->y0, sp->sav);
	}
}
/*}}}*/

static inline
void fgl_showSprite ( fgl_context_t* ctx, fgl_sprite_t* sp ) /*{{{*/
{	
	if ( !sp->isVisible ){
		sp->isVisible = 1;

		fgl_waitReady( ctx->bmp );
		fgl_saveArea( sp->sav, 0, 0,
					  ctx->bmp, sp->x0, sp->y0, sp->x1, sp->y1);
		fgl_waitReady( ctx->bmp );
		fgl_maskBlt( ctx, sp->x0, sp->y0, sp->bmp, 0, 0,
					 sp->x1 - sp->x0, sp->y1 - sp->y0, sp->mask);
	}
}
/*}}}*/

FUNC_ATTR
fgl_sprite_t* fgl_createSprite ( fgl_spriteData_t *spd ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_SPRI)
{
	u32         i, x, y;
	fgl_slc_t   *p;
	u8          *q;
	fgl_color_t *cmap;
	fgl_color_t clr;
	fgl_monoBitmap_t *mask = fgl_createMonoBitmap( spd->width, spd->height, 0xff);
	fgl_bitmap_t *bmp = fgl_createBitmap( spd->width, spd->height, -1);
	fgl_sprite_t *sp  = (fgl_sprite_t*) FGL_MALLOC( sizeof(fgl_sprite_t));
#ifdef USE_HW_CURSOR
	fgl_hardware_sprite_t *hsp;
#endif	
	sp->bmp  = bmp;
	sp->mask = mask;
	sp->sav  = fgl_createBitmap( spd->width, spd->height, -1);

	cmap = (fgl_color_t*) alloca( sizeof(fgl_color_t) * spd->nColors);
	for ( i=0; i<spd->nColors; i++ ){
		cmap[i] = fgl_getPixelValue( spd->cmap[i]);
	}

	bmp->transp = cmap[spd->transp];

	p = bmp->surface;
	q = spd->data;

	for ( y=0; y<spd->height; y++ ){
		for ( x=0; x<spd->width; x++ ){
			clr = (*q < spd->nColors) ? cmap[*q] : fgl_clrBlack;
			fgl_setBmpPixel( bmp, x, y, clr);
			// fgl_incSlc( &p, x);

			if ( *q == spd->transp ){
				fgl_unsetMonoBit( mask, x, y);
			}
			q++;
		} 
	}

	sp->x0 = sp->y0 = 0;
	sp->x1 = bmp->width-1;
	sp->y1 = bmp->height-1;

	sp->xhot = spd->xhot;
	sp->yhot = spd->yhot;

	sp->isVisible = 0;
#ifdef USE_HW_CURSOR
	/* 
	 * hardware cursor create function is defined to
	 * to early to be aware of the FGL_MALLOC macro
	 * so we have to allocate the memory now and pass
	 * it to the function.
	 */
	hsp = FGL_MALLOC( sizeof(fgl_hardware_sprite_t));
	fgl_hardware_createSprite( spd, sp, hsp );
#endif
	return sp;
}
#endif /* !FGL_PROTO */
;/*}}}*/

static inline
void fgl_moveSprite ( fgl_context_t* ctx, fgl_sprite_t* sp, i32 x, i32 y ) /*{{{*/
{
	i32 u, v, xw, yh; 

	xw = sp->sav->width-1;
	yh = sp->sav->height-1;

	u = x - sp->xhot;
	v = y - sp->yhot;
	sp->x1 = u+xw;
	sp->y1 = v+yh;

 	/* restore old contents (if it's not the first time) */
	if ( sp->isVisible ){
		/* restore old area */
		fgl_bitBltBmp( ctx, sp->x0, sp->y0, sp->sav);

		/* save new area */
		fgl_saveArea( sp->sav, 0, 0, ctx->bmp, u, v, sp->x1, sp->y1);

		/* draw sprite */
		fgl_maskBltBmp( ctx, u, v, sp->bmp, sp->mask);
	}

	sp->x0 = u;
	sp->y0 = v;
}
/*}}}*/


/*
 * Here comes the cursor support, which is just a global sprite
 */

static inline
void fgl_moveCursor ( i32 x, i32 y ) /*{{{*/
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
	if ( x <= 0 ) x = 0;
	if ( y <= 0 ) y = 0;

	_fgl_inCursorOp = 1;
	fgl_moveSprite( fgl_screenCtx, fgl_cursor, x, y);
	_fgl_inCursorOp = 0;
#endif
}
/*}}}*/

void fgl_showCursor () /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)

	_fgl_inCursorOp = 1;
	fgl_showSprite( fgl_screenCtx, fgl_cursor);
	_fgl_inCursorOp = 0;
#endif
}
#endif
;/*}}}*/

void fgl_hideCursor () /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
	_fgl_inCursorOp = 1;
	fgl_hideSprite( fgl_screenCtx, fgl_cursor);
	_fgl_inCursorOp = 0;
#endif
}
#endif
;/*}}}*/

void fgl_setCursor ( fgl_sprite_t* sp ) /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
	int x, y;

	if ( fgl_cursor && fgl_cursor->isVisible ){
		x = fgl_cursor->x0 + fgl_cursor->xhot;
		y = fgl_cursor->y0 + fgl_cursor->yhot;
		
		sp->x0 = x - sp->xhot;
		sp->y0 = y - sp->yhot;
		sp->x1 = sp->x0 + sp->bmp->width;
		sp->y1 = sp->y0 + sp->bmp->height;

		fgl_hideCursor();
		fgl_cursor = sp;
		fgl_showCursor();
	}
	else {
		fgl_cursor = sp;
	}
#endif /* USE_xx_CURSOR */
}
#endif
;/*}}}*/

/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_SPRI_INCLUDED */
/*}}}*/





