#define IPL_MAIN

#include <stdlib.h>

#include "ipl.h"

static int intrFd = -1;
static int intwFd = -1;
static int maxFd;

static fd_set readSet;

static inline
void updateReadSet ( int fd )
{
	if ( fd >= 0 ) {
		if ( fd >= maxFd ) {
			maxFd = fd+1;
		}
		FD_SET( fd, &readSet);
	}
}

/* some kernel drivers require to be in a stable state before we
 * start to read packets from them
 */
void _ipl_flush ( int fd )
{
	struct timeval tv = { 0, 0 };
	fd_set rs;
	char c;

	FD_ZERO(&rs);
	FD_SET(fd,&rs);

    	while ( select( fd+1, &rs, 0, 0, &tv) > 0 ){
		read( fd, &c, 1);

		FD_ZERO(&rs);
		FD_SET(fd,&rs);
	}
}


/****************************************************************************
 * We need a way to unblock/wakeup a ipl_getEvent (which blocks on
 * file descriptors), and to inject artificial (i.e. not device generated)
 * events into the event system. That all done by the wakeup routines
 */

static int pendingWakeup;

int initWakeup ()
{
	int fd[2];

	/* create interrupt pipe to enable select wakeup */
	if ( pipe( fd) == 0 ){
		intrFd = fd[0];
		intwFd = fd[1];
		return 1;
	}
	return 0;
}

int processWakeup ( ipl_event_t* evt )
{
	unsigned char buf[sizeof(ipl_event_t)];

	if ( read( intrFd, buf, sizeof(ipl_event_t)) < sizeof(ipl_event_t) ){
		return 0;
	}

	memcpy( evt, buf, sizeof(ipl_event_t));

	return (evt->type != EVT_NONE);
}

static inline
int postEventWakeup ( ipl_event_t* evt )
{
	pendingWakeup++;

	if ( write( intwFd, evt, sizeof(ipl_event_t)) < sizeof(ipl_event_t) ){
		return 0;
	}
	else {
		return 1;
	}
}


static inline
int postInteruptWakup ()
{
	unsigned char buf[sizeof(ipl_event_t)];
	memset( buf, 0, sizeof(buf));

	if ( !pendingWakeup ){
		pendingWakeup++;

		if ( write( intwFd, buf, sizeof(buf)) < sizeof(buf) ){
			return 0;
		}
	}

	return 1;
}


static inline
int checkWakeup ( fd_set* readset )
{
	if ( FD_ISSET( intrFd, readset) ){
		pendingWakeup--;
		return 1;
	}
	else {
		return 0;
	}
}

static inline
void resetWakeup ()
{
	close( intrFd);
	close( intwFd);
}

/****************************************************************************
 * device registration
 */

static
int registerDriver( ipl_driver_t *driver) 
{
	ipl_device_t *dev, *d, *temp;
	int fd;

	/* 
	 * Grab all the devices that use this driver.
	 */
	if (!driver->initDevice(&dev)) {
		return 0;
	}
	
	while (dev) {
		/* First save next device passed in. */
                temp = dev->next;
		fd = _ipl_threadedFd(dev->fd);
		
		if (fd != -1) {
			_ipl_flush(fd);
                        updateReadSet(fd);
                        dev->fd = fd;

			/* insert new device into global device list */
			if (!ipl_devices || (ipl_devices->prio < dev->prio)) {
                		dev->next = ipl_devices;
                		ipl_devices = dev;
			} else { 
				for (d = ipl_devices; d->next; d = d->next ) {
                			if (d->next->prio < dev->prio ) {
						dev->next = d->next;
						d->next = dev;
						break;
					}
				}
				d->next = dev;
			}
        	}
		/* start over with next device from list passed in. */
                dev = temp;
	}
	return 1;
}

static
int registerDevices ()
{
#if defined (DEVICE_1)
	{
		extern ipl_driver_t DEVICE_1;
		registerDriver( &DEVICE_1);
	}
#endif

#if defined (DEVICE_2)
	{
		extern ipl_driver_t DEVICE_2;
		registerDriver( &DEVICE_2);
	}
#endif

#if defined (DEVICE_3)
	{
		extern ipl_driver_t DEVICE_3;
		registerDriver( &DEVICE_3);
	}
#endif


#if defined (DEVICE_4)
	{
		extern ipl_driver_t DEVICE_4;
		registerDriver( &DEVICE_4);
	}
#endif

#if defined (DEVICE_5)
	{
		extern ipl_driver_t DEVICE_5;
		registerDriver( &DEVICE_5);
	}
#endif
	return 1;
}

/****************************************************************************
 * These are our public APIs
 */

int ipl_init ( int width, int height )
{
	int result;

	/* store max pointer space in case some device has to know about it */
	ipl_width  = width;
	ipl_height = height;

	/* we always have a wakeup "device" */ 
	if ( !initWakeup() ) {
		fprintf(stderr, "ipl_init failed in initWakeup\n");
                return 0;
        }
	updateReadSet( intrFd);
	
	/* Now we register real "devices" */
	registerDevices();

	atexit(ipl_reset);
	return 1;
}


int ipl_getEvent( ipl_event_t* evt )
{
	ipl_device_t *dev;
	fd_set  rs = readSet;

	if ( !pendingWakeup ){
		for ( dev = ipl_devices; dev; dev = dev->next ){
			if ( dev->pending ){
				return dev->driver->getEvent( dev, evt);
			}
		}
	}

    	if ( _ipl_select( maxFd, &rs, 0, 0, 0) > 0 ){
		if ( checkWakeup ( &rs ) ){
			return processWakeup( evt);
		}

		for ( dev = ipl_devices; dev; dev = dev->next ){
			if ( dev->fd != -1 && FD_ISSET( dev->fd, &rs) ){
			   if ( dev->driver->getEvent( dev, evt) ) {
				return 1;
			   }
			}
		}
	}
	return 0;
}

/*
 * queue an event which isn't generated by one of the input drivers
 * (ensure propper serialization w/o resorting to a specific thread system)
 */
int ipl_queueEvent ( ipl_event_t* evt )
{
	return postEventWakeup( evt);
}

/*
 * make sure we wake up from a indefinite select wait
 * (ensure propper serialization w/o resorting to a specific thread system)
 */
int ipl_wakeUp ()
{
	return postInteruptWakup();
}

void ipl_reset ()
{
	ipl_device_t *dev;

	if ( maxFd ) {
		resetWakeup();

		for ( dev = ipl_devices; dev != 0; dev = dev->next ){
			dev->driver->resetDevice(dev);
		}

		maxFd = 0;
	}
}

void ipl_setScreenRotation (char rot)
{
	rotation = rot;
}
