/*
 * gen_geod_vbi.c - vbi device for geode chip
 *
 * This is for a particular /dev/vbi device that does not
 * conform to the V4L linux interface -- instead, it returns
 * two bytes on the VBI stream at a time, already processed.
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <sys/ioctl.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

//#include <linux/videodev.h>

#include "ipl.h"

//#define FLT_MASK 0xff
//#undef RAW_INPUT

#define FLT_MASK 0x7f
#define RAW_INPUT 1

static int vbiFd = -1;
static int atvef_on = 0;
static int dIdx = 0;
static unsigned char data[256];
static unsigned char vbi_data[2];

ipl_device_t vbigeodeDevice;

static int geode_vbi_init(ipl_device_t **dev)
{
  char  *vbidev = getenv( "VBI");
  //struct v4l2_format fmt;	

  if ( vbiFd != -1 ) {
    *dev = &vbigeodeDevice;
    return 1;
  }

  if ( !vbidev ){
    vbidev = "/dev/vbi";
  }

  if ( (vbiFd = open( vbidev, O_RDONLY | O_NONBLOCK)) == -1 ) {
    fprintf(stderr, "ipl: Error opening VBI device: %s", vbidev);
    *dev = NULL;
    return 0;
  }

  //fcntl( vbiFd, F_SETFL, O_NONBLOCK );
   
  /*	
	memset( &fmt, 0, sizeof(fmt) );
	fmt.fmt.vbi.sampling_rate = 60;
	fmt.fmt.vbi.offset = 9;             
	fmt.fmt.vbi.samples_per_line = 2;   
	fmt.fmt.vbi.sample_format = V4L2_VBI_SF_UBYTE;          
	fmt.fmt.vbi.start[0] = 20;
	fmt.fmt.vbi.start[1] = 283;
	fmt.fmt.vbi.count[0] = 1;
	fmt.fmt.vbi.count[1] = 1;
	fmt.fmt.vbi.flags = 0;
	
	if( ioctl(vbiFd,VIDIOC_S_FMT,&fmt) < 0 ) {
	}
  */	
  vbigeodeDevice.fd = vbiFd;
  *dev = &vbigeodeDevice;
  return 1;
}

static
void geode_vbi_reset (ipl_device_t *dev)
{
  if ( vbiFd != -1 ){
    close( vbiFd);
  }
}
      
static
unsigned short checkSum( unsigned short* val, int byte_count) {
  unsigned long sum = 0;
 
  while ( byte_count > 1 ) {
    sum += *(val++);
    byte_count -= 2;
  }
 
  if ( byte_count > 0 ) {
    sum += *(unsigned char*)val;
  }
 
  while ( sum >> 16 ) {
    sum = (sum & 0xFFFF) + (sum >> 16);
  }
 
  return ((unsigned short) ~sum);
}       

static
int isReady() {
  //check for trailing checksum [xxxx]
  int idx;
 
  if ( (dIdx < 5) || (data[dIdx-5] != '[')) {
    return 0;
  }
  for ( idx = 1; idx < 5; idx++ ) {
    if ( ! isxdigit( data[dIdx-idx] ) ) {
      return 0;
    }
  }

  return 1;
}

static
int checkReady( ipl_event_t* evt) {
  //check for trailing checksum [xxxx]
  char csb[5];
  unsigned short icheck, check;
 
  //swap checksum byte order
  csb[0] = data[dIdx-2];
  csb[1] = data[dIdx-1];
  csb[2] = data[dIdx-4];
  csb[3] = data[dIdx-3];
  csb[4] = 0;
 
  icheck = (unsigned short)strtol( csb, (char**)0, 16);
  check = checkSum( (unsigned short*)data, dIdx-5) ;
 
  printf( "trigger checksum: %s(%04X/%04X)\n", (check == icheck) ? "passed" : "failed", check, icheck);
  if ( icheck == check ) {
 
    data[++dIdx] = 0;                    

    evt->x = evt->y = 0;
    evt->buttons = 0;
    evt->mods = 0;
    evt->keyChar = evt->keyCode = 0;
 
    gettimeofday( &(evt->tv), 0);
    evt->type = EVT_GENERIC_FIRST;
    evt->data = data;
    evt->dlen = dIdx;                                         
    
    dIdx = 0;
 
    printf( "trigger ready: %s\n", data);
 
    return 1;                                                                           
  }
  else {
    //bad checksum, start with next trigger
    dIdx = 0;
    return 0;
  }
 
}  

static
int geode_vbi_getEvent ( ipl_device_t *dev, ipl_event_t* evt ){
   int  n, bIdx;

   
   n = read( vbiFd, vbi_data, 2);

   if( ( n == 2) && ( vbi_data[0] != 0x00 ) && ( vbi_data[1] != 0x00 ) ) {
#if defined (RAW_INPUT)
      if( (vbi_data[0] == 0x1c) && (vbi_data[1] == 0xab) ) {
	 atvef_on = 1;
      }
      else if( (vbi_data[0] == 0x1c) && (vbi_data[1] == 0xad) ) {
      }
      else if( vbi_data[0] == 0x94 ) {
	 atvef_on = 0;
      }
      else if( ((vbi_data[0] & FLT_MASK) >= 0x10) && ((vbi_data[0] & FLT_MASK) <= 0x1f) ) {
	 atvef_on = 0;
      }
      else if( atvef_on ) {
#else
      if ( 1) {
#endif
         unsigned char b0 = vbi_data[0] & FLT_MASK;
         unsigned char b1 = vbi_data[1] & FLT_MASK;
    
         if ( dIdx == 0 ) {
            if ( (b0 != '<') && (b1 != '<') ) {
	       return 0;
	    }
	    if ( b1 == '<' ) {
	       data[dIdx++] = '<';
	       return 0;
	    }
	 }
	 for ( bIdx = 0; bIdx < 2; bIdx++ ) {
	    if ( vbi_data[bIdx] != 0x80 ) {
	       data[dIdx] = vbi_data[bIdx] & FLT_MASK;               
	       if ( data[dIdx] == ']' ) {
		  if ( isReady() ) {
		     return checkReady( evt);
		  }
	       }   
	       dIdx++;                                                            
	    }
	 }
      } 
   }
      
   return 0;
}
ipl_driver_t vbigeodeGen = { geode_vbi_init, geode_vbi_getEvent, geode_vbi_reset };   
ipl_device_t vbigeodeDevice = { 10, -1, 0, NULL, &vbigeodeGen, 0 };
