/*
 * input.c - Linux /dev/input/eventX device support. 
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * This file uses the linux input api. With the same interface you can
 * work with touchscreens, joysticks, graphics tablets, mice, keyboards 
 * and anything else you can get your hands on. We can even make beep noises
 * with it. It is also a api across all linux platforms (alpha, ix86, arm etc). 
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include <linux/input.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "vk_constants.h"
#include "ipl.h"
#include "input.h"

int event2ipl[255] = {
	0, VK_ESCAPE, VK_1, VK_2, VK_3, VK_4, VK_5, VK_6, VK_7, VK_8, VK_9, 
	VK_0, VK_SUBTRACT, VK_EQUALS, VK_BACK_SPACE, VK_TAB, VK_Q, VK_W, 
	VK_E, VK_R, VK_T, VK_Y, VK_U, VK_I, VK_O, VK_P, VK_OPEN_BRACKET, 
	VK_CLOSE_BRACKET, VK_ENTER, VK_CONTROL, VK_A, VK_S, VK_D, VK_F, VK_G,
	VK_H, VK_J, VK_K, VK_L, VK_SEMICOLON, VK_QUOTE, VK_, 
	VK_SHIFT, VK_BACK_SLASH, VK_Z, VK_X, VK_C, VK_V, VK_B, VK_N, VK_M,
	VK_COMMA, VK_PERIOD, VK_SLASH, VK_SHIFT, KEY_KPASTERISK, VK_ALT, 
	VK_SPACE, VK_CAPS_LOCK, VK_F1, VK_F2, VK_F3, VK_F4, VK_F5, VK_F6, 
	VK_F7, VK_F8, VK_F9, VK_F10, VK_NUM_LOCK, VK_SCROLL_LOCK, VK_NUMPAD7,
	VK_NUMPAD8, VK_NUMPAD9, VK_SUBTRACT, VK_NUMPAD4, VK_NUMPAD5, VK_NUMPAD6,
	VK_ADD, VK_NUMPAD1, VK_NUMPAD2, VK_NUMPAD3, VK_NUMPAD0, VK_PERIOD, 
	VK_, VK_, VK_, VK_F11, VK_F12, VK_, VK_, 
	VK_, VK_, VK_, VK_, VK_, VK_ENTER, VK_CONTROL,
	VK_SLASH, VK_PRINTSCREEN, VK_ALT, KEY_LINEFEED, VK_HOME, VK_UP, 
	VK_PAGE_UP, VK_LEFT, VK_RIGHT, VK_END, VK_DOWN, VK_PAGE_DOWN, 
	VK_INSERT, VK_DELETE, VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_PAUSE, 
	VK_, VK_, VK_, VK_, VK_COMMA, VK_META, VK_META, VK_, VK_, VK_, VK_,
	VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_, VK_,
	VK_, VK_, VK_, VK_, VK_, VK_, VK_  
};

typedef struct {
        unsigned char keyChar;
        unsigned char keyCode;
	int event;
} shiftMap_t;

shiftMap_t charkeyTab[] = {
	{ '~', VK_ , KEY_GRAVE },
	{ '!', VK_ , KEY_1 },
	{ '@', VK_ , KEY_2 },
	{ '#', VK_ , KEY_3 },
	{ '$', VK_ , KEY_4 },
	{ '%', VK_ , KEY_5 },
	{ '^', VK_ , KEY_6 },
	{ '&', VK_ , KEY_7 },
	{ '*', VK_MULTIPLY , KEY_8 },
	{ '(', VK_OPEN_BRACKET , KEY_9 },
	{ ')', VK_CLOSE_BRACKET , KEY_0 },
	{ '_', VK_ , KEY_MINUS },
	{ '+', VK_ADD , KEY_EQUAL },
	{ '|', VK_ , KEY_BACKSLASH },
	{ '{', VK_ , KEY_LEFTBRACE },
	{ '}', VK_, KEY_RIGHTBRACE },
	{ ':', VK_ , KEY_SEMICOLON },
	{ '"', VK_QUOTE, KEY_APOSTROPHE },
	{ '<', VK_ , KEY_COMMA },
	{ '>', VK_ , KEY_DOT },
	{ '?', VK_ , KEY_SLASH }
};

#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define BITS_PER_LONG (sizeof(long) * 8)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define LONG(x) ((x)/BITS_PER_LONG)
#define test_bit(bit, array)    ((array[LONG(bit)] >> OFF(bit)) & 1)

static int xmax, ymax;

static int input_init(ipl_device_t **dev);
static void input_reset(ipl_device_t *dev);
static int input_getEvent(ipl_device_t *dev, ipl_event_t *evt); 

struct ev_data {
	int x;
	int y;
	unsigned int absmax[2];
	unsigned int absmin[2];
};

ipl_driver_t event_dev = {
	initDevice:     input_init,
        getEvent:       input_getEvent,
        resetDevice:    input_reset,
};

ipl_device_t *listdev = NULL;

static int input_init(ipl_device_t **dev)
{
	unsigned long bit[EV_MAX][NBITS(KEY_MAX)];
	char *name = "/dev/input/";
	unsigned int abs[5], fd;
	ipl_device_t *evdev;
	struct ev_data *private;
	struct dirent *dev_dirp;
	char temp[15];
	DIR *dp;

	if ((dp = opendir(name)) == NULL) {
		fprintf(stderr, "ipl: error opening %s: %s\n", name, strerror(errno));
		*dev = NULL;
		return 0;
	}		
		

	while ((dev_dirp = readdir(dp)) != NULL) {
		if (strncmp(dev_dirp->d_name, "event", 5) != 0)
			continue;

		strcpy(temp, name);
		strcat(temp, dev_dirp->d_name);

		if ((fd = open(temp, O_RDONLY)) >= 0) {
			private = (struct ev_data *) malloc(sizeof(struct ev_data));
			if (!private) {
				perror("ipl: allocation failed");	
				continue;
			}	

			memset(bit, 0, sizeof(bit));
       			ioctl(fd, EVIOCGBIT(0, EV_MAX), bit[0]);

			if (test_bit(EV_ABS, bit[0])) {
       				ioctl(fd, EVIOCGABS(ABS_X), abs);

				private->absmin[0] = abs[1];
				private->absmax[0] = abs[2];
             	
				ioctl(fd, EVIOCGABS(ABS_Y), abs);
             	
				private->absmin[1] = abs[1];
				private->absmax[1] = abs[2];
			}
			evdev=(ipl_device_t *) malloc(sizeof(ipl_device_t));
			evdev->data = (unsigned char *) private;
       			evdev->fd = fd;
       			evdev->prio = 12;
       			evdev->pending = 0;
       			evdev->driver = &event_dev;
			evdev->next = listdev;
			listdev = evdev;
		}
	}
	rewinddir(dp);
	if (closedir(dp) < 0)
          {
            perror("ipl: failed to close directory");
          }

	xmax = ipl_width;
	ymax = ipl_height;
	ipl_xPointer = ipl_width / 2;
       	ipl_yPointer = ipl_height / 2;
	*dev = listdev;
	return 1;
}

static void input_reset(ipl_device_t *dev)
{
	struct input_event reset;
	int rd, fd = dev->fd;

	reset.type = EV_RST;	

	rd = write(fd, &reset, sizeof(struct input_event));
}

static int key_handling(ipl_event_t *evt, struct input_event ev) 
{
	int i, j = 0;

	switch (ev.value) {
        	/* key release */
		case 0:
                	evt->type = EVT_KEY_UP;
                        break;
                /* key press */
                case 1:
                        evt->type = EVT_KEY_DOWN;
                        break;
                /* autorepeat */
                case 2:
                	/* We need to send the previous
		 	 * event over and over again 
		 	 */
			return 0;
	}
	/* Next we see if it is a modifier key.
	 * If it is we return immediately to the
	 * event loop but keep trace of which 
	 * modifier was pressed. 
	 */
	if ((ev.code == KEY_LEFTSHIFT)  ||
	    (ev.code == KEY_RIGHTSHIFT) ||
	    (ev.code == KEY_LEFTCTRL)   ||
	    (ev.code == KEY_RIGHTCTRL)  || 
	    (ev.code == KEY_LEFTMETA)	||
	    (ev.code == KEY_RIGHTMETA)  ||
	    (ev.code == KEY_LEFTALT)	||
	    (ev.code == KEY_RIGHTALT)) {	
		switch (ev.code) {
			case KEY_LEFTSHIFT:
			case KEY_RIGHTSHIFT:
				evt->mods ^= SHIFT;
				evt->keyCode = VK_SHIFT;
				break;
			case KEY_LEFTCTRL:
			case KEY_RIGHTCTRL:
				evt->mods ^= CTRL;
				evt->keyCode = VK_CONTROL;
				break;
			case KEY_LEFTMETA:
			case KEY_RIGHTMETA:
				evt->mods ^= META;
				evt->keyCode = VK_META;
				break;
			case KEY_LEFTALT:
			case KEY_RIGHTALT:
				evt->mods ^= ALT;
				evt->keyCode = VK_ALT;
				break;	
		}
		evt->keyChar = 0;
		evt->x = evt->y = evt->buttons = 0;
		return 1;
	}
	/* 
	 * The regular printable keys. Note isascii really
	 * sucks!!! I have to do some work arounds. 
	 */
	evt->keyCode = event2ipl[ev.code];
	if ((isascii(evt->keyChar)) || (ev.code=!KEY_BACKSPACE)
	    || !(ev.code >= KEY_F1 && ev.code <= KEY_F10)) { 
		
		evt->keyChar = evt->keyCode;
		if (evt->mods & SHIFT) {
			for (i = 0; i < 19; i++) {
				if (ev.code == charkeyTab[i].event) {
		  	    	       evt->keyChar = charkeyTab[i].keyChar;
			   	       evt->keyCode = charkeyTab[i].keyCode;
			    	       j = 1;
				}
		  	}	
		  	if (!j)
 		  		evt->keyChar = tolower(evt->keyCode);
			} else {
				if (ev.code == KEY_MINUS)
					evt->keyChar = '-';
			}
		} else {
		/*
	 	 * Okay. isascii is going to miss a few keys. 
		 * We grab them here.
 	 	 */
		if (evt->mods & SHIFT) {
			switch (ev.code) {
				case KEY_BACKSLASH:
				case KEY_DOT:
				case KEY_SLASH:
					for (i = 0;i < 19;i++) {
                                   		if (ev.code == charkeyTab[i].event) {
                                            		evt->keyChar = charkeyTab[i].keyChar;
                                            		evt->keyCode = charkeyTab[i].keyCode;
                                        	 }
					}
					break;
				default:
					evt->keyChar = 0;
			}
		} else {
			evt->keyChar = 0;
		}
	}
	evt->x = evt->y = evt->buttons = 0;	
	return 1;
}

/*
 * Java only acknowledges mice and keyboards :-( So fancier devices have to
 * treated as such. If it is not a keyboard we fake it as a mouse.
 */
static int input_getEvent(ipl_device_t *dev, ipl_event_t *evt) 
{
    struct ev_data *data = (struct ev_data *) dev->data;
    int rd, size, fd = dev->fd;	
    struct input_event ev;

    /* 
     * Which device sent the packet? More than one input device could be 
     * sending event packets. 
     */	
    rd = read(fd, &ev, sizeof(struct input_event));

    if (rd < sizeof(struct input_event)) {
	perror("ipl: error reading /dev/eventX");
	return 0;
    }

    evt->tv.tv_sec = ev.time.tv_sec;
    evt->tv.tv_usec = ev.time.tv_usec;

    switch (ev.type) {
	case EV_KEY:
		if (ev.code > 0x100) {
			/* Buttons instead of key presses. */
	       		switch (ev.code) {
				case BTN_LEFT:
				case BTN_TOUCH:
					evt->buttons ^= IPL_BTN_LEFT;
					break;
				case BTN_RIGHT:
					evt->buttons ^= IPL_BTN_RIGHT;
					break;
				case BTN_MIDDLE:
					evt->buttons ^= IPL_BTN_MIDDLE;
					break;
				default:
					return 0;
			}

			if (ev.value) 
				evt->type = EVT_BUTTON_DOWN;
			else
				evt->type = EVT_BUTTON_UP;			
		} else
			return key_handling(evt, ev);
		break;
	case EV_REL:
                evt->type = EVT_MOUSE_MOVE;
	
		switch (ev.code) {
		case REL_X:
			data->x += ev.value;
			if ( data->x > xmax )   data->x = xmax;
			else if ( data->x < 0 ) data->x = 0;
			break;	
		case REL_Y:
			data->y += ev.value;              
			if ( data->y > ymax )   data->y = ymax;
			else if ( data->y < 0 ) data->y = 0;
			break;
		default:
			return 0;
		}	
		ROTATE(evt->x, evt->y, data->x, data->y);
		break;
	case EV_ABS:
		evt->type = EVT_MOUSE_MOVE;
		
		switch (ev.code) {
		case ABS_X:
			size = data->absmax[0] - data->absmin[0];
			data->x = (ev.value*ipl_width)/size - data->absmin[0];
			break;
		case ABS_Y:
			size = data->absmax[1] - data->absmin[1];
			data->y = (ev.value*ipl_height)/size - data->absmin[1];
			break;
		default:
			return 0;
		}
		ROTATE(evt->x, evt->y, data->x, data->y);
		break;			
	default:
		return 0;
    }
    return 1;
}
