/*
 * kbd_tty.c - simple stdin based keyboard device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "ipl.h"

#include "vk_constants.h"
#include "ttykeytab-102-us.h"

#include <stdlib.h>
#include <termios.h>
#include <unistd.h>

#ifdef TTY_RAW_MODE
#include <linux/kd.h>
#endif

#define RESET  0
#define CBREAK 1
#define RAW    2

static int kbdFd = -1;
static struct termios tiosSave;
static int ttyState = RESET;

static unsigned int mods, lastMods;

ipl_driver_t ttyKeyboard;
ipl_device_t ttyDevice;   /* actually a local forward decl */

static
void tty_reset (ipl_device_t *dev)
{
#ifdef TTY_RAW_MODE
	struct kbentry kbentry;
#endif

	if ( ttyState == CBREAK || ttyState == RAW ){
		if ( tcsetattr( kbdFd, TCSAFLUSH, &tiosSave) < 0 ){
			return;
		}
		ttyState = RESET;

#ifdef TTY_RAW_MODE
		// Remap Shift-PGUP and Shift-PGDN back to kernel defaults
		kbentry.kb_table=1;
		kbentry.kb_index=104;     // Shift-PGUP
		kbentry.kb_value=0;
		kbentry.kb_value=0x020b;  // Scroll_Backward
		if ( ioctl( kbdFd, KDSKBENT, &kbentry ) < 0 ) {
			fprintf(stderr, "ipl: error setting kernel keymapping for Shift-PGUP");
			return;
		}

		kbentry.kb_table=1;
		kbentry.kb_index=109;	  // Shift-PGDN
		kbentry.kb_value=0;
		kbentry.kb_value=0x020a;  // Scroll_Forward
		if ( ioctl( kbdFd, KDSKBENT, &kbentry ) < 0 ) {
			fprintf(stderr, "ipl: error setting kernel keymapping for Shift-PGDN");
			return;
		}
#endif
	}
}

static int tty_init(ipl_device_t **dev)
{
#ifdef TTY_RAW_MODE
	struct kbentry kbentry;	
#endif	
	struct termios tios;

	if ( ttyState == RESET ){
		kbdFd = STDIN_FILENO;

		if ( tcgetattr( kbdFd, &tiosSave) < 0 ){
			fprintf(stderr, "ipl: unable to get terminal attributes for stdin\n");
			*dev = NULL;
			return 0;
		}

		memcpy( &tios, &tiosSave, sizeof(struct termios));

#ifdef TTY_RAW_MODE
		tios.c_lflag &= ~( ECHO | ICANON | IEXTEN | ISIG);
		tios.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
		tios.c_cflag &= ~(CSIZE | PARENB);
		tios.c_cflag |= CS8;
		tios.c_oflag &= ~(OPOST);

		tios.c_cc[VMIN]  = 1;
		tios.c_cc[VTIME] = 0;

		if ( tcsetattr( kbdFd, TCSAFLUSH, &tios) < 0 ){
			fprintf(stderr, "ipl: unable to set RAW mode on stdin\n");
			*dev = NULL;
			return 0;
		}

		ttyState = RAW;

		// Remap Shift-PGUP and Shift-PGDN so that they can be
		// recieved (without putting the kernel keyboard translation
		// into RAW mode) - map them to F19/F20 (same as Shift-F7/Shift-F8 on keyboard)
		kbentry.kb_table=1;
		kbentry.kb_index=104;     // Shift-PGUP
		kbentry.kb_value=0;
		kbentry.kb_value=0x0112;  // F19
		if ( ioctl( kbdFd, KDSKBENT, &kbentry ) < 0 ) {
			fprintf(stderr, "ipl: error setting kernel keymapping for Shift-PGUP");
			*dev = NULL;
			return 0;
		}

		kbentry.kb_table=1;
		kbentry.kb_index=109;     // Shift-PGDN
		kbentry.kb_value=0;
		kbentry.kb_value=0x0113;  // F20
		if ( ioctl( kbdFd, KDSKBENT, &kbentry ) < 0 ) {
			fprintf(stderr, "ipl: error setting kernel keymapping for Shift-PGDN");
			*dev = NULL;
			return 0;
		}

#else
		tios.c_lflag &= ~( ECHO | ICANON);

		tios.c_cc[VMIN]  = 1;
		tios.c_cc[VTIME] = 0;

		if ( tcsetattr( kbdFd, TCSAFLUSH, &tios) < 0 ){
			fprintf(stderr, "ipl: unable to set CBREAK mode on stdin\n");
			*dev = NULL;
			return 0;
		}

		ttyState = CBREAK;
#endif
	}
	ttyDevice.fd = kbdFd;	
	*dev = &ttyDevice;
	return 1;
}

static inline
int checkMod ( ipl_event_t* evt, int modBit, int vkConst ) {
	if ( (lastMods & modBit) != (mods & modBit) ){
		if ( lastMods & modBit ){  /* CTRL up */
			evt->type = EVT_KEY_UP;
			lastMods &= ~modBit;
		}
		else {                   /* CTRL down */
			evt->type = EVT_KEY_DOWN;
			lastMods |= modBit; 
		}
		evt->keyCode = vkConst;
		
		return 1;
	}

	return 0;
}


static
int tty_getEvent (ipl_device_t *dev, ipl_event_t* evt )
{
  static char buf[128];
  static int  i, n;
  static ipl_eventType_t state;
  static unsigned short keyChar, keyCode;

  unsigned char c;

  if ( !ttyDevice.pending ) {
      if ( (n = read( kbdFd, buf, sizeof(buf))) > 0 ){
          i=0;
      }
      else {
          return 0;
      }
  }

  if ( n > 0 ){
      evt->type = EVT_KEY_DOWN;
      state = EVT_KEY_DOWN;

      c = buf[i];
      n--; i++;

      if ( c != 27 ){                        /* single or control+key */
          keyChar = keyTab[c].keyChar;
          keyCode = keyTab[c].keyCode;
          mods    = keyTab[c].mods;
      }
      else {
          if ( (buf[i] == '[') && (n > 1) ){ /* some function key */
              keyChar = keyCode = 0;
              mods = 0;
              n--; i++;

              switch ( buf[i] ){
                case 'A': keyCode = VK_UP;    n--; i++; break;
                case 'B': keyCode = VK_DOWN;  n--; i++; break;
                case 'C': keyCode = VK_RIGHT; n--; i++; break;
                case 'D': keyCode = VK_LEFT;  n--; i++; break;
                case 'a': keyCode = VK_UP;    mods |= SHIFT; n--; i++; break;
                case 'b': keyCode = VK_DOWN;  mods |= SHIFT; n--; i++; break;
                case 'c': keyCode = VK_RIGHT; mods |= SHIFT; n--; i++; break;
                case 'd': keyCode = VK_LEFT;  mods |= SHIFT; n--; i++; break;
                case '1': n--; i++;
                          if ( (n > 1) & (buf[i+1] == '~') ){
                              switch ( buf[i] ){
                                case '7': keyCode = VK_F6; n-=2; i+=2; break;
                                case '8': keyCode = VK_F7; n-=2; i+=2; break;
                                case '9': keyCode = VK_F8; n-=2; i+=2; break;
                              }
                          }
                          if ( (keyCode == 0) && (buf[i] == '~') ){
                              keyCode = VK_HOME;
                              n--; i++;
                          }
                          break;
                case '2':  n--; i++;
                           if ( (n > 1) & (buf[i+1] == '~') ){
                               switch ( buf[i] ){
                                 case '0': keyCode = VK_F9;  n-=2; i+=2; break;
                                 case '1': keyCode = VK_F10; n-=2; i+=2; break;
                                 case '3': keyCode = VK_F11; n-=2; i+=2; break;
                                 case '4': keyCode = VK_F12; n-=2; i+=2; break;
                                 case '5': keyCode = VK_F1; mods |= SHIFT; n-=2; i+=2; break;
                                 case '6': keyCode = VK_F2; mods |= SHIFT; n-=2; i+=2; break;
                                 case '8': keyCode = VK_F3; mods |= SHIFT; n-=2; i+=2; break;
                                 case '9': keyCode = VK_F4; mods |= SHIFT; n-=2; i+=2; break;
                               }
                           }
                           if ( (keyCode == 0) && (buf[i] == '~') ){
                               keyCode = VK_INSERT;
                               n--; i++;
                           }
                           break;
                case '3':  n--; i++;
                           if ( (n > 1) & (buf[i+1] == '~') ){
                               switch ( buf[i] ){
                                 case '1': keyCode = VK_F5; mods |= SHIFT; n-=2; i+=2; break;
                                 case '2': keyCode = VK_F6; mods |= SHIFT; n-=2; i+=2; break;
                                 case '3': keyCode = VK_PAGE_UP; mods |= SHIFT; n-=2; i+=2; break;
                                 case '4': keyCode = VK_PAGE_DOWN; mods |= SHIFT; n-=2; i+=2; break;
                               }
                           }
                           if ( (keyCode == 0) && (buf[i] == '~') ){
                               keyCode = VK_DELETE;
                               n--; i++;
                           }
                           break;
                case '4': 
                           if ( buf[i+1] == '~' ){
                               keyCode = VK_END; n-=2; i+=2;
                           }
                           break;
                case '5':
                           if ( buf[i+1] == '~' ){
                               keyCode = VK_PAGE_UP; n-=2; i+=2;
                           }
                           break;
                case '6':
                           if ( buf[i+1] == '~' ){
                               keyCode = VK_PAGE_DOWN; n-=2; i+=2;
                           }
                           break;
                case '[':
                           if ( n > 0 ){
                               switch ( buf[i+1] ){
                                 case 'A': keyCode = VK_F1; n-=2; i+=2; break;
                                 case 'B': keyCode = VK_F2; n-=2; i+=2; break;
                                 case 'C': keyCode = VK_F3; n-=2; i+=2; break;
                                 case 'D': keyCode = VK_F4; n-=2; i+=2; break;
                                 case 'E': keyCode = VK_F5; n-=2; i+=2; break;
                               }
                           }
                           if ( keyCode == 0 ){
                               keyChar = '[';
                               keyCode = keyTab['['].keyCode;
                               n--; i++;
                           }
                           break;
                default:
                           break;
              }
          }
          else if ( n == 0 ){                /* single escape key */
              keyCode = VK_ESCAPE;
          }
          else {                             /* alt+key */
              c = buf[i];
              keyChar = keyTab[c].keyChar;
              keyCode = keyTab[c].keyCode;
              mods    = keyTab[c].mods | ALT;
              n--; i++;
          }
      }
  }

  evt->x = evt->y = evt->buttons = 0;
  gettimeofday( &(evt->tv), 0);

  if ( mods != lastMods ){
      /* produce artificial modifier key events */
      if ( checkMod( evt, SHIFT, VK_SHIFT) ||
          checkMod( evt, CTRL, VK_CONTROL) ||
          checkMod( evt, ALT, VK_ALT) ){
          evt->keyChar = 0;	
          ttyDevice.pending = 1;
          return 1;
      }
  }

  evt->keyChar = keyChar;
  evt->keyCode = keyCode;
  evt->type    = state;
  evt->mods    = mods;

  if ( state == EVT_KEY_DOWN ){
      state = EVT_KEY_UP;
      ttyDevice.pending = 1;
  }
  else if ( state == EVT_KEY_UP ){
      state = EVT_NONE;
      ttyDevice.pending = (n > 0);
  }

  return 1;
}

ipl_driver_t ttyKeyboard = { tty_init, tty_getEvent, tty_reset };
ipl_device_t ttyDevice = { 12, -1, 0, NULL, &ttyKeyboard, NULL };
