/*
 * mou_ps2.c - raw ps2 mouse device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#include "ipl.h"

static int mouFd = -1;
static int x, y;
static int xmax, ymax;
static int threshold = 4, accel = 2;

#define ACC_TIME 0
#define TRASH 100

#define ABS(_a) (((_a) > 0) ? (_a) : -(_a))

extern ipl_driver_t ps2Mouse;
extern ipl_device_t ps2dev;

static int ps2_init(ipl_device_t **dev)
{
	char  *env;
	char  *moudev = getenv( "MOUSE");
	
	if ( mouFd != -1 ) {
	    ps2dev.fd = mouFd;			 	
	    *dev = &ps2dev;
	    return 1;
	}

	if ( !moudev ){
		moudev = "/dev/psaux";
	}
	
	if ( (mouFd = open (moudev, O_RDONLY)) == -1 ) {
		fprintf(stderr, "ipl: error opening %s: %s\n", moudev, strerror(errno));
		*dev = NULL;
		return 0;
	}

	/* if we don't flush and there are already packets, the driver
	 * gets confused, select never resets
	 */
	ps2dev.fd = mouFd;

	ipl_xPointer = x = ipl_width / 2;
	ipl_yPointer = y = ipl_height / 2;

	xmax = ipl_width-1;
	ymax = ipl_height-1;

	if ( (env = getenv("IPL_MOUSE")) ){
		sscanf( env, "%d %d", &threshold, &accel);
	}
	*dev = &ps2dev;
	return 1;
}

static
void ps2_reset (ipl_device_t *dev)
{
	if ( mouFd != -1 ){
		close( mouFd);
	}
}

static
int ps2_getEvent(ipl_device_t *dev, ipl_event_t* evt )
{
	static int buttons, lastButtons;
	static int dx, dy;
	static unsigned char b[MOU_PACKET_LENGTH];
	
	int i, n, adx, ady;
	fd_set rs;
	struct timeval to;
	
	while ( 1 ) {
    	/* logically, PS/2 mouse packets come in 3 bytes, but lots of drivers
    	 * like to deliver them in pieces
    	 */
    	for ( i=0; i<MOU_PACKET_LENGTH; ){
    	    /* make sure we never block if we get to this point */
    	    if ( (n = read( mouFd, &(b[i]), MOU_PACKET_LENGTH - i)) <= 0 ){
    	        if ( i == 0 )
    	            return 0;
    	    }
    	    else {
    	        i += n;
    	    }
    	}
		
    	/* printf( "%d %d %d\n", b[0], b[1], b[2]); */
		
    	buttons = 0;
    	if ( b[0] & 1 ) buttons |= IPL_BTN_LEFT;
    	if ( b[0] & 2 ) buttons |= IPL_BTN_RIGHT;
    	if ( b[0] & 4 ) buttons |= IPL_BTN_MIDDLE;
	
    	if ( (b[1] == 0) && (b[2] == 0) ){ /* press/release */
    	    if ( buttons == lastButtons ){   /* HW jitter */
    	        return 0;
    	    }
            evt->type = ( buttons ) ? EVT_BUTTON_DOWN : EVT_BUTTON_UP;
    	    break;
 	    }
	    else {                             /* move/drag */
	        dx = b[1] ? ((b[0] & 0x10) ? (signed char) b[1] : b[1]) : 0;
	        dy = b[2] ? ((b[0] & 0x20) ? (signed char) b[2] : b[2]) : 0;

		    adx = ABS(dx);
		    ady = ABS(dy);
			
		    if ( (adx >= TRASH) || (ady >= TRASH) ){ /* a trash spike */
			    return 0;
		    }
			
		    /* a simple acceleration */
		    if ( adx >= threshold ){
			    dx *= accel;
		    }
		    if ( ady >= threshold ){
			    dy *= accel;
		    }
			
		    x += dx;
		    if ( x > xmax )			x = xmax;
		    else if ( x < 0 )		x = 0;
		
		    y -= dy;                /* a little pitfall, dy is decremented */
		    if ( y > ymax )			y = ymax;
		    else if ( y < 0 )		y = 0;

		    /* now do a some solicitation, to avoid ending up in
		     * floods of mouse move events for too sensitive devices
		     */
		    FD_ZERO( &rs);
		    FD_SET( mouFd, &rs);
		    to.tv_sec = 0;
		    to.tv_usec = ACC_TIME;
			
		    if ( (select( mouFd+1, &rs, 0,0, &to) <= 0) || !FD_ISSET(mouFd, &rs) ){
		        evt->type = EVT_MOUSE_MOVE;
		        break;
		    }
        }
	}
	
	lastButtons = buttons;
	
	ROTATE(evt->x, evt->y, x, y);

	evt->buttons = buttons;
	evt->mods = 0;
	evt->keyChar = evt->keyCode = 0;

	gettimeofday( &(evt->tv), 0);
	ipl_xPointer = evt->x;
	ipl_yPointer = evt->y;
	
	return 1;
}

ipl_driver_t ps2Mouse = { ps2_init, ps2_getEvent, ps2_reset };
ipl_device_t ps2dev = { 10, -1, 0, NULL, &ps2Mouse, NULL };
