/*
 * mou_ts.c - PAWS touchscreen pseudo-mouse device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 * 
 * 00/00/00 Peter Mehlitz
 */


#include "ipl.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>

#include <linux/ts.h>

#include "tpconfig.c"          /* includes the generic config file reader */

#define TSDEV "/dev/mouse"   /* not /dev/ts, as one might assume */
#define THRESH_X 5
#define THRESH_Y 5

/* static int pressure = 0;               unused */
static int state;
static int x, y;

static int mouFd = -1;

int rawCoordinates;


static inline
int abs ( int v )
{
	return (v > 0) ? v : -v;
}

static int
setTpConfig ()
{	
	struct cal_info  ci;

	/* put the thing into touchscreen mode (abs pos etc.) */
	if ( ioctl( mouFd, TS_TP_MODE) < 0 ){
		puts( "can't get into TP_MODE");
		return -1;
	}

	if ( rawCoordinates )
		return 1;

	if ( ioctl( mouFd, TS_GET_CALM, &ci) < 0 ){
		puts( "can't get CALM");
		return 0;
	}

	if ( !readTsConfig( &ipl_cfg) ){
		puts( "no touchscreen config file");
		return 0;
	}
	
	if ( ipl_cfg.xoff != TS_UNDEF )
	  ci.xoff = ipl_cfg.xoff;

	if ( ipl_cfg.xscale != TS_UNDEF )
	  ci.xscale = ipl_cfg.xscale;

	if ( ipl_cfg.yoff != TS_UNDEF )
	  ci.yoff = ipl_cfg.yoff;

	if ( ipl_cfg.yscale != TS_UNDEF )
	  ci.yscale = ipl_cfg.yscale;

	if ( ioctl( mouFd, TS_SET_CALM, &ci) < 0 ){
		puts( "can't set CALM");
		return 0;
	}

	if ( ipl_cfg.swap != TS_UNDEF ){
		if ( ioctl( mouFd, TS_SET_XY_SWAP, &ipl_cfg.swap) < 0 ){
			puts( "can't swap XY");
			return 0;
		}
	}

	if ( ipl_cfg.rate != TS_UNDEF ){
		if ( ioctl( mouFd, TS_SET_RATE, &ipl_cfg.rate) < 0 ){
			puts( "can't set rate");
			return 0;
		}
	} 

	return 1;
}


/***************************************************************************
 * exported funcs
 */

static int ts_init(ipl_device_t **dev)
{
    if ( mouFd == -1 ) {
	  if ( (mouFd = open( TSDEV, O_RDONLY)) == -1 ){
	    fprintf(stderr, "ipl: Error opening %s\n", TSDEV);
	    *dev = NULL;
	    return 0;
	  }

	  setTpConfig();
	}
	tsDevice.fd = mouFd;	
	*dev = &tsDevice;
	return 1;
}

static
void ts_reset (ipl_device_t *dev)
{
  if ( mouFd != -1 ){
		close( mouFd);
		mouFd = -1;
	}
}

static
int ts_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{
	struct ts_sample ts;
	int n;
	static int u=0, v=0;   /* it likes to loose coords on pen_up */

	fd_set  rs;
	struct timeval t;

	do {
		ts.pen_down = ts.xpos = ts.ypos = ts.pressure = 0;

		if ( (n=read( mouFd, &ts, sizeof(ts))) < sizeof(ts) ){
			printf( "read: not enough data %d/%d\n", n, sizeof(ts));
		}

		if ( ts.pen_down ){ /* otherwise xpos, ypos are zero */
			u = ts.xpos;
			v = ts.ypos; 
		}

		FD_ZERO( &rs);
		FD_SET( mouFd, &rs);
		t.tv_sec = 0;
		t.tv_usec = 0;

		/* we always want to get the first press */
	} while ( (ts.pen_down == state) && _ipl_select( mouFd+1, &rs, 0, 0, &t) );


	if ( (!ts.pen_down && (state != ts.pen_down)) ||
		 (abs( u - x) > THRESH_X) || (abs( v - y) > THRESH_Y) ){
		x = u;
		y = v;
		state = ts.pen_down;

		evt->x = x;
		evt->y = y;
		evt->type = (state) ? EVT_BUTTON_DOWN : EVT_BUTTON_UP;

		evt->buttons = IPL_BTN_LEFT;
		evt->mods = 0;
		evt->keyChar = evt->keyCode = 0;

		gettimeofday( &(evt->tv), 0);

		return 1;
		//printf( "pen: %d, x: %d, y: %d  pres: %d\n", ts.pen_down, x, y, ts.pressure);
	}

	return 0;
}

ipl_driver_t tsMouse = { ts_init, ts_getEvent, ts_reset };
ipl_device_t tsDriver = { 10, -1, 0, NULL, &tsMouse, 0 };
