/*
 * mou_vr41x.c - Cassiopeia touchscreen pseudo-mouse device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 * 
 * 07/20/00 Peter Mehlitz
 */


#include "ipl.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "tpconfig.c"          /* includes the generic config file reader */

#define TSDEV "/dev/tpanel"

ipl_device_t vr41xDevice;

typedef struct {
	unsigned short state   __attribute__ ((packed));
	unsigned short xplus   __attribute__ ((packed));
	unsigned short xminus  __attribute__ ((packed));
	unsigned short yplus   __attribute__ ((packed));
	unsigned short yminus  __attribute__ ((packed));
	short z                __attribute__ ((packed));
} ts_sample;

static int mouFd = -1;

static int callibrate;

static inline
int abs ( int v )
{
	return (v > 0) ? v : -v;
}

/***************************************************************************
 * exported funcs
 */

static int vr41x_init(ipl_device_t **dev)
{
    if ( mouFd == -1 ) {
		if ( (mouFd = open( TSDEV, O_RDONLY)) == -1 ){
			fprintf(stderr, "ipl: Error opening %s\n", TSDEV);
			*dev = NULL;
			return 0;
		}

		callibrate = 1;
		if ( ! readTsConfig( &ipl_cfg) ){
		        /* defaults for the Cassiopeia */
			ipl_cfg.swap   = 1;
			
			ipl_cfg.xoff   = 705;
			ipl_cfg.xscale = 177;
			
			ipl_cfg.yoff   = -760;
			ipl_cfg.yscale = -216;
		}
	}
	vr41xDevice.fd = mouFd;	
	*dev = &vr41xDevice;
	return 1;
}

static
void vr41x_reset (ipl_device_t *dev)
{
  if ( mouFd != -1 ){
		close( mouFd);
		mouFd = -1;
	}
}


#define MIN_PRESSURE 800
#define THRESH_X 150
#define THRESH_Y 150

#define UP           0
#define DOWN         1

static
int vr41x_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{
    ts_sample  ts;
	int        n, state;
	int        u=0, v=0;
	static int x=0, y=0;
	static int xlast, ylast;
	static int lastState = UP;
	static int um, vm, downs = 0;
	
	fd_set  rs;
	struct timeval t;

	do {
		if ( (n=read( mouFd, &ts, sizeof(ts))) < sizeof(ts) ){
			//printf( "read: not enough data %d/%d\n", n, sizeof(ts));
			return 0;
		}

		//printf( "state: %x,  x: %d - %d (=%d),  y: %d - %d (=%d),   z: %d\n", ts.state, ts.xplus, ts.xminus, (int)ts.xplus - ts.xminus, ts.yplus, ts.yminus, (int)ts.yplus - ts.yminus, ts.z);

		if ( (ts.state & 0xc000) == 0x4000 ){
			return 0;  /* down with no valid xyz yet */
		}
		
		state = (ts.state & 0x4000) ? DOWN : UP;

		if ( state  == DOWN ){ /* otherwise xyz are not valid */

			/* below that, it's going to be very unreliable */
			if ( ts.z < MIN_PRESSURE ) return 0;

			u = (int)ts.xplus - ts.xminus;
			v = (int)ts.yplus - ts.yminus; 

			/* filter out peak noise */
			if ( downs ) {
				if ( (abs(u-um) > THRESH_X) || (abs(v-vm) > THRESH_Y) ){
					return 0;
				}
			}

			if ( ++downs == 1  ) {
				um = u;
				vm = v;

				break;
			}
			else {
				um = (um + u) / 2;
				vm = (vm + v) / 2;

				/* make sure we don't sample up to the next up event */
				if ( downs % 5 ){
					break;
				}
			}
		}
		else {
			if ( lastState == UP ){ /* consecutive ups? makes no sense */
				return 0;
			}

			/* we stil have to select-loop, because the thing likes
			 * to produce spurious UPs in the middle of a move
			 */
		}

		FD_ZERO( &rs);
		FD_SET( mouFd, &rs);
		t.tv_sec = 0;
		t.tv_usec = 50000;
	} while ( select( mouFd+1, &rs, 0, 0, &t) );


	if ( state == DOWN ) {
		if ( callibrate ) {
			if ( ipl_cfg.swap ){
				x = vm;
				y = um;
			}
			else {
				x = um;
				y = vm;
			}

			if ( ipl_cfg.xoff != TS_UNDEF )   x += ipl_cfg.xoff;
			if ( ipl_cfg.xscale != TS_UNDEF ) x = (x * ipl_cfg.xscale) >> 10;  /* / 1024 */
			
			if ( ipl_cfg.yoff != TS_UNDEF )   y += ipl_cfg.yoff;
			if ( ipl_cfg.yscale != TS_UNDEF ) y = (y * ipl_cfg.yscale) >> 10;  /* / 1024 */
		}
		else {
			x = um;
			y = vm;
		}

		if ( downs > 1 ) {
			if ((abs(x - xlast) + abs(y - ylast)) < ((downs==2) ? 4 : 1)){
				downs = 1;
				return 0;
			}
		}

		xlast = x;
		ylast = y;

		evt->type = (state == lastState) ? EVT_MOUSE_MOVE : EVT_BUTTON_DOWN;
	}
	else {
		evt->type = EVT_BUTTON_UP;
		downs = 0;
	}
	lastState = state;
	
	ROTATE(evt->x, evt->y, x, y);

	evt->buttons = IPL_BTN_LEFT;
	evt->mods = 0;
	evt->keyChar = evt->keyCode = 0;

	gettimeofday( &(evt->tv), 0);

	return 1;
}

ipl_driver_t vr41xMouse = { vr41x_init, vr41x_getEvent, vr41x_reset };
ipl_device_t vr41xDevice = { 10, -1, 0, NULL, &vr41xMouse, 0 };
