#!/usr/bin/env python
"""
Compiler for CML2

by Eric S. Raymond, <esr@thyrsus.com>
"""
import sys

if sys.version[0] < '2':
    print "Python 2.0 or later is required for this program."
    sys.exit(0)

import string, os, getopt, shlex, cPickle, cml, cStringIO

# Globals
debug = 0; errors = 0; profile = 0
outfile = "rules.out"
requirements = []
dictionary = {}
bad_symbols = {}
bool_tests = []
help_dict = {}
depth = 0
start = prefix = banner = ""
trit_tie = help_tie = icon = None
warndepend = []
forward_private = []
explicit_ancestors = {}
expression_parser = None	# Just to make this a global

# Lexical analysis
_keywords = (
    'banner',		'choices',	'condition',	'debug',
    'default',		'dependent',	'derive',	'enum',
    'explanation',	'explanations',	'icon',		'menu',
    'menus',		'nohelp',	'on',		'prefix',
    'private',		'prohibit',	'range',	'require',
    'start',		'suppress',	'symbols',	'text',
    'trits',		'unless',	'warndepend'
    )
_ternaryops = ('?', ':')
_arithops = ('*', '+', '-')
_boolops = ('and', 'or', 'implies')
_relops = ('==', '!=', '<', '>', '>=', '<=') 
_termops = ('|', '&', '$')		# min, max, similarity
_operators = _termops + _relops + _boolops + _arithops + _ternaryops + ("(", ")")
_tritvals = ("n", "m", "y")
_atoms = ("trit", "string", "decimal", "hexadecimal")
_suffixes = ("&", "?", "!", "%", "@", "$")

class Token:
    "CML2's internal token type."
    def __init__(self, type, attr=None):
	self.type = type
	self.attr = attr
	if debug > 1: print "CML token: ", `self`
    def __repr__(self):
        if self.type == "EOF":
            return "EOF"
        elif self.attr is not None:
            return self.type + "=" + `self.attr`
        else:
            return self.type
    def __cmp__(self, o):
        return cmp(self.type, o)
    def __getitem__(self, i):
        raise IndexError

class lexwrapper(shlex.shlex):
    "Lexer subclass that returns Tokens with type-annotation information."
    def __init__(self, stream, file):
        # This is a speed hack....
        file_contents = cStringIO.StringIO(stream.read())
        stream.close()
	shlex.shlex.__init__(self, file_contents, file)

    def lex_token(self):
	# Get a (type, attr) token tuple, handling inclusion  
	raw = self.get_token()
	if type(raw) is not type(""):	# Pushed-back token
	    return raw
	elif not raw:
	    return Token("EOF")
	elif raw[0] in self.quotes:
	    return Token('string', raw[1:-1])
	elif raw in _tritvals:
	    return Token('trit', raw)
	elif len(raw) > 2 and \
		raw[0] == '0' and raw[1] == 'x' and raw[2] in string.hexdigits:
            return Token('hexadecimal', long(raw[2:], 16))
	elif raw[0] in string.digits:
	    return Token('decimal', int(raw))
	elif raw in ('!', '=', '<', '>'):	# Relational tests
	    next = self.get_token()
	    if next == '=':
		return Token(raw+next)
	    else:
		self.push_token(next)
		return Token('word', raw)
        elif raw == 'text':
            data = ""
            while 1:
                line = self.instream.readline()
                if line == "" or line == ".\n":	# Terminated by dot.
                    break
                if line[0] == '.':
                    line = line[1:]
                data = data + line
            return Token("text", data)
        elif raw == 'icon':
            data = ""
            while 1:
                line = self.instream.readline()
                if line == "" or line == "\n":	# Terminated by blank line
                    break
                data = data + line
            self.push_token(data)
            return Token(raw)
	elif raw in _keywords or raw in _operators:
	    return Token(raw)
	else:
            # Nasty hack alert.  If there is a declared prefix for the
            # rulebase, ignore it as a prefix of names.  This will
            # enable us to be backward-compatible with names like like
            # CONFIG_3C515 that have leading numerics when stripped.
            if prefix and raw[:len(prefix)] == prefix:
                raw = raw[len(prefix):]
	    return Token('word', raw)

    def complain(self, str):
	# Report non-fatal parse error; format like C compiler message.
        global errors
        if not debug and not errors:
            sys.stderr.write('\n')
	sys.stderr.write(self.error_leader() + " " + str + "\n")
	errors = errors + 1

    def croak(self, str):
	# Report a fatal parse error and die
	self.complain(str)
	sys.exit(1)

    def demand(self, type, attr=None):
	# Require a given token or token type, croak if we don't get it 
	tok = self.lex_token()
	if tok.type == "EOF":
	    self.croak("premature EOF")
	elif attr is not None and tok.attr != attr:
	    self.croak("syntax error, saw `%s' while expecting `%s'" % (tok, attr))
	elif tok.type != type:
	    self.croak("syntax error, expecting token of type `%s' (actually saw %s=%s)" % (type, tok.type, tok.attr))
	else:
	    return tok.attr

    def sourcehook(self, newfile):
	# Override the hook in the shlex class
	try:
	    if newfile[0] == '"':
		newfile = newfile[1:-1]
                # This implements cpp-like semantics for relative-path inclusion.
                if type(self.infile) is type("") and not os.path.isabs(newfile):
                    newfile = os.path.join(os.path.dirname(self.infile), newfile)
	    return (newfile, open(newfile, "r"))
	except IOError:
	    self.complain("I/O error while opening '%s'" % (newfile,))
	    sys.exit(1)

# Parsing

class ExpressionError:
    "Express a compile-time error."
    def __init__(self, explain):
        self.args = ("expression error " + explain,)

def parse_atom(input):
    if debug >= 2: print "entering parse_atom..."
    op = input.lex_token()
    if op.type in _atoms:
        if debug >= 2: print "parse_atom returns", op.attr
        return op
    elif op.type == '(':
        sub = parse_expr_inner(input)
        close = input.lex_token()
        if close != ')':
            raise ExpressionError, "while expecting a close paren"
        else:
            if debug >= 2: print "parse_atom returns singleton", sub
            return sub
    elif op.type in _keywords:
        raise ExpressionError, "keyword %s while expecting atom" % op.type
    elif op.type == 'word':
        if debug >= 2: print "parse_atom returns", op.attr
        return op

def parse_term(input):
    if debug >= 2: print "entering parse_term..."
    left = parse_atom(input)
    op = input.lex_token()
    if op.type not in _termops:
        input.push_token(op)
        if debug >= 2: print "parse_term returns singleton", left
        return left
    right = parse_term(input)
    expr = (op.type, left, right)
    if debug >= 2: print "parse_term returns", expr
    return expr

def parse_relational(input):
    if debug >= 2: print "entering parse_relational..."
    left = parse_term(input)
    op = input.lex_token()
    if op.type not in _relops:
        input.push_token(op)
        if debug >= 2: print "parse_relational returns singleton", left
        return left
    right = parse_term(input)
    expr = (op.type, left, right)
    if debug >= 2: print "parse_relational returns", expr
    return(op.type, left, right)

def parse_assertion(input):
    if debug >= 2: print "entering parse_assertion..."
    negate = input.lex_token()
    if negate.type == 'not':
        return ('not', parse_relational(input))
    input.push_token(negate)
    return parse_relational(input)

def parse_conjunct(input):
    if debug >= 2: print "entering parse_conjunct..."
    left = parse_assertion(input)
    op = input.lex_token()
    if op.type !=  'and': 
        input.push_token(op)
        if debug >= 2: print "parse_conjunct returns singleton", left
        return left
    else:
        expr = ('and', left, parse_conjunct(input))
        if debug >= 2: print "parse_conjunct returns", expr
        return expr

def parse_disjunct(input):
    if debug >= 2: print "entering parse_disjunct..."
    left = parse_conjunct(input)
    op = input.lex_token()
    if op.type != 'or':
        input.push_token(op)
        if debug >= 2: print "parse_disjunct returns singleton", left
        return left
    else:
        expr = ('or', left, parse_disjunct(input))
        if debug >= 2: print "parse_disjunct returns", expr
        return expr

def parse_factor(input):
    if debug >= 2:
        print "entering parse_factor..."
    left = parse_disjunct(input)
    op = input.lex_token()
    if op.type != 'implies':
        input.push_token(op)
        if debug >= 2: print "parse_factor returns singleton", left
        return left
    else:
        expr = ('implies', left, parse_disjunct(input))
        if debug >= 2: print "parse_factor returns", expr
        return expr

def parse_summand(input):
    if debug >= 2: print "entering parse_summand..."
    left = parse_factor(input)
    op = input.lex_token()
    if op.type != '*':
        input.push_token(op)
        if debug >= 2: print "parse_summand returns singleton", left
        return left
    else:
        expr = ('*', left, parse_expr_inner(input))
        if debug >= 2: print "parse_summand returns", expr
        return expr

def parse_ternary(input):
    if debug >= 2: print "entering parse_ternary..."
    guard = parse_summand(input)
    op = input.lex_token()
    if op.type != '?':
        input.push_token(op)
        if debug >= 2: print "parse_ternary returns singleton", guard
        return guard
    else:
        trueval = parse_summand(input)
        op = input.lex_token()
        if op.type != ':':
            raise ExpressionError("while expecting : in ternary")
        falseval = parse_summand(input)
        expr = ('?', guard, trueval, falseval)
        if debug >= 2: print "parse_ternary returns", expr
        return expr

def parse_expr_inner(input):
    if debug >= 2: print "entering parse_inner_expr..."
    left = parse_ternary(input)
    op = input.lex_token()
    if op.type not in ('+', '-'):
        input.push_token(op)
        if debug >= 2: print "parse_expr_inner returns singleton", left
        return left
    else:
        expr = (op.type, left, parse_expr_inner(input))
        if debug >= 2: print "parse_expr_inner returns", expr
        return expr

def parse_expr(input):
    "Parse an expression."
    try:
        exp = parse_expr_inner(input)
        return exp
    except ExpressionError, exp:
        input.croak(exp.args[0])
        return None

# Expression parsing ends here

def pure_conjunction(guard):
    if type(guard) is not type(()):
        return 1
    elif guard[0] == 'and' or guard[0] in _relops:
        return pure_conjunction(guard[1]) and pure_conjunction(guard[2])
    else:
        return 0

def make_dependent(guard, symbol):
    "Create a dependency lnk, indirecting properly through menus."
    if debug > 2:
        print "Making %s dependent on %s" % (symbol.name, guard.name)
    # If symbol is a menu, we'd really like to create a dependency link
    # for each of its children.  But they won't be defined at this point
    # if the reference is forward.
    symbol.ancestors.append(guard)
    guard.dependents.append(symbol)

def parse(input, baton):
    # Parse an entire CML program
    global debug, start, requirements, dictionary, bad_symbols, depth
    global help_dict, prefix, banner, icon, trit_tie, help_tie
    global warndepend, explicit_ancestors
    input.source = "source"
    if debug:
    	print "Calling parse()"
    	input.debug = 1
    while 1:
        if not debug and not errors:
            baton.twirl()
	leader = input.lex_token()
	if debug > 1: print "Parsing declaration beginning with %s..." % (leader,)
	if leader.type == "EOF":
	    break
	elif leader.type == "start":
	    start = input.lex_token().attr
	elif leader.type in ("symbols", "menus", "explanations"):
	    deduced_type = None
	    if leader.type == "menus":
		deduced_type = "menu"
	    elif leader.type == "explanations":
		deduced_type = "explanation"
	    while 1:
		tok = input.lex_token()
		if tok.type == "word":
		    name = tok.attr
		else:
		    input.push_token(tok)
		    break
		prompt = input.demand("string")

		if dictionary.has_key(name):
		    ref = dictionary[name]
		    input.complain('duplicate symbols %s (see "%s", line %d)'
				   % (name, ref.file, ref.lineno))
		else:
		    dictionary[name] = cml.ConfigSymbol(name,
							deduced_type,
                                                        None, prompt,
							input.infile,
							input.lineno)
                    explicit_ancestors[dictionary[name]] = []
                # These symbols may be followed by optional help text
                tok = input.lex_token()
                if tok.type == "text":
                    dictionary[name].helptext = tok.attr
                else:
                    input.push_token(tok)
	    if debug:
                print "%d symbols read" % (len(dictionary),)
	elif leader.type == "unless":
	    guard = parse_expr(input)
	    bool_tests.append((guard, input.infile, input.lineno))
	    input.demand("suppress")
	    dependent = input.lex_token()
	    make_dep = 0
	    if dependent.type == "dependent":
		make_dep = 1
                if not pure_conjunction(guard):
                    input.complain("illegal guard clause")
	    else:
		input.push_token(dependent)
	    list = parse_wordlist(input)
            list.reverse()
	    for symbol in list:
		if make_dep:
                    #if debug: print "Creating dependencies on %s" % (guard,)
                    explicit_ancestors[symbol].append(guard)
                else:
                    # Add it to ordinary visibility constraints
                    if symbol.visibility:
                        symbol.visibility = ('and', guard, symbol.visibility)
                    else:
                        symbol.visibility = guard
	elif leader.type == "menu":
	    name = input.demand("word")
	    list = parse_wordlist(input, 1)
	    if not dictionary.has_key(name):
		input.complain("%s is not a declared menu ID" % name)
            #elif dictionary[name].items:
            #    input.complain("more than one declaration of %s" % name)
	    else:
		entry = dictionary[name]
		entry.items += list
		for symbol in list:
		    if symbol.menu:
			input.complain("symbol %s in %s occurs in another menu (%s)"
				       % (symbol.name, name, symbol.menu.name))
		    else:
			symbol.menu = dictionary[name]
	elif leader.type == "choices":
	    name = input.demand("word")
	    list = parse_wordlist(input)
	    input.demand('default')
	    default = input.demand("word")
	    if not dictionary.has_key(name):
		input.complain("%s is not a declared menu ID" % (name,))
	    elif not dictionary.has_key(default):
		input.complain("default %s must be a known symbol"%(`default`,))
	    elif dictionary[default] not in list:
		input.complain("default %s must be in the menu" % (`default`,))
	    else:
		dictionary[name].default = dictionary[default]
                dictionary[name].items = list
                dictionary[name].type = "choices"
                for symbol in list:
                    if symbol.menu:
                        input.complain("symbol %s occurs in another menu (%s)"
                                       % (symbol.name, symbol.menu.name))
                    else:
                        symbol.menu = dictionary[name]
	elif leader.type == "derive":
	    name = input.demand("word")
	    input.demand("word", "from")
	    expr = parse_expr(input)
	    if dictionary.has_key(name):
		input.complain("symbol %s is already defined" % (name,))
	    else:
		dictionary[name] = cml.ConfigSymbol(name,
                                                    None, expr, None,
                                                    input.infile, input.lineno)
                if name in forward_private:
                    dictionary[name].private = 1
                    forward_private.remove(name)
	elif leader.type in ("require", "prohibit"):
            expr = parse_expr(input)
            if leader.type == "prohibit":
                expr = ('==', expr, cml.n)
            next = input.lex_token()
            if next.type != 'explanation':
                input.push_token(next)
                msg = None
            else:
                expl = input.lex_token()
                if expl.type != 'word':
                    input.complain("while expecting a word of explanation, I see %s" % (`expl`,))
                    continue
                if not dictionary.has_key(expl.attr):
                    input.complain("unknown symbol where explanation expected")
                elif dictionary[expl.attr].type != 'explanation':
                    input.complain("expecting an explanation symbol here")
                else:
                    msg = dictionary[expl.attr].prompt
	    requirements.append(cml.Requirement(expr, msg, input.infile, input.lineno))	    
	    bool_tests.append((expr, input.infile, input.lineno))
	elif leader.type == "default":
	    symbol = input.demand("word")
	    input.demand("word", "from")
	    expr = parse_expr(input)
	    if not dictionary.has_key(symbol):
		input.complain("%s is not a valid symbol to default"%(symbol,))
	    elif dictionary[symbol].default: 
		input.complain("%s already has a default" % (symbol,))
	    else:
                entry = dictionary[symbol]
		entry.default = expr
            next = input.lex_token()
            if next.type == "range":
                entry.range = []
                while 1:
                    low = input.lex_token()
                    if low.type in _keywords:
                        input.push_token(low)
                        break
                    elif low.type in ("decimal", "hexadecimal"):
                        low = low.attr
                    else:
                        input.complain("bad token %s where range literal expected" % (low.attr))
                    rangesep = input.lex_token()
                    if rangesep.type in _keywords:
                        entry.range.append(low)
                        input.push_token(rangesep)
                        break
                    elif rangesep.type in ("decimal", "hexadecimal"):
                        entry.range.append(low)
                        input.push_token(rangesep)
                        continue
                    elif rangesep.type == '-':
                        high = input.lex_token()
                        if high.type in ("decimal", "hexadecimal"):
                            high = high.attr
                            entry.range.append((low, high))
                            continue
                        else:
                            input.croak("malformed range")
                            break
            elif next.type == "enum":
                entry.range = []
                entry.enum = 1
                while 1:
                    name = input.lex_token()
                    if name.type in _keywords:
                        input.push_token(name)
                        break
                    elif name.type != 'word':
                        input.complain("bad token %s where enum name expected" % (name.attr))
                        continue
                    elif not dictionary.has_key(name.attr):
                        input.croak("unknown symbol in enum")
                    else:
                        name = dictionary[name.attr].prompt
                    input.demand('word', "=")
                    value = input.lex_token()
                    if value.type in ("decimal", "hexadecimal"):
                        value = value.attr
                        entry.range.append((name, value))
                        continue
                    else:
                        input.croak("malformed enum")
            else:
                input.push_token(next)
                continue
	elif leader.type == 'private':
            while 1:
                nxt = input.lex_token()
                if nxt.type != 'word':
                    input.push_token(nxt)
                    break
                if dictionary.has_key(nxt.attr):
                    dictionary[nxt.attr].private = 1
                else:
                    forward_private.append(nxt.attr)
	elif leader.type == 'debug':
            debug = input.lex_token().attr
	elif leader.type == 'prefix':
            prefix = input.lex_token().attr
	elif leader.type == 'banner':
            banner = dictionary[input.lex_token().attr].prompt
        elif leader.type == 'icon':
            if icon:
                input.complain("multiple icon declarations")
            icon = input.lex_token().attr
        elif leader.type == 'condition':
            flag  = input.lex_token()
            input.demand("on")
            switch = input.demand("word")
            if not dictionary.has_key(switch):
                input.complain("unknown symbol in condition starement")
                continue
            # Someday is today
            if flag == "trits":
                trit_tie = dictionary[switch]
                trit_tie.vital = 1
            elif flag == "nohelp":
                help_tie = dictionary[switch]
            else:
                input.complain("unknown flag %s in condition statement" % (flag,))
        elif leader.type == 'warndepend':
            iffy = parse_wordlist(input)
            for symbol in iffy:
                warndepend.append(symbol)
	else:
	    input.croak("syntax error, unknown statement %s" % (leader.attr,))

def parse_wordlist(input, do_nesting=None):
    # Get a nonempty list of config symbols and menu ids. 
    # Interpret the {} shorthand if second argument is nonempty
    global requirements
    condition_stack = []	# Stack of active conditions for {} shorthand
    list = []			# Result list
    entry = None		# Last symbol resolved
    while 1:
	symbol = input.lex_token()
	if not symbol.type in ("word",) + _suffixes:
	    #if debug > 1:
	    #	print "parse_wordlist() terminates returning %s and pushes keyword %s" % (list, symbol.type)
	    input.push_token(symbol)
	    break
	elif symbol.attr == '}':
	    if not do_nesting:
		input.complain("sublists aren't allowed outside menus")
	    elif entry:
		condition_stack = condition_stack[0:-1]
	    continue
	elif symbol.type == '?':	# This is also an operator
	    if not do_nesting:
		input.complain("? type suffix is not allowed outside menus")
	    elif entry:
		entry.type = 'trit'
	    continue
	elif symbol.attr == '!':
	    if not do_nesting:
		input.complain("! type suffix is not allowed outside menus")
	    elif entry:
		entry.type = 'trit'
		entry.vital = 1
	    continue
	elif symbol.attr == '%':
	    if not do_nesting:
		input.complain("% type suffix is not allowed outside menus")
	    elif entry:
		entry.type = 'decimal'
	    continue
	elif symbol.attr == '@':
	    if not do_nesting:
		input.complain("@ type suffix is not allowed outside menus")
	    elif entry:
		entry.type = 'hexadecimal'
	    continue
	elif symbol.type == '$':	# This is also an operator
	    if not do_nesting:
		input.complain("$ type suffix is not allowed outside menus")
	    elif entry:
		entry.type = 'string'
	    continue
	elif symbol.attr == '{':
	    if not do_nesting:
		input.complain("sublists aren't allowed outside menus")
	    elif entry:
                if entry.type == "string":
                    input.complain("string symbol is not a legal submenu guard")
                else:
                    condition_stack.append(entry)
	    continue
	elif dictionary.has_key(symbol.attr):
	    entry = dictionary[symbol.attr]
            if entry.type is None:
                entry.type = "bool"
	    list.append(entry)
            if do_nesting:
                entry.depth = len(condition_stack)
            if condition_stack:
                make_dependent(condition_stack[-1], entry)
	else:
	    input.complain("unknown symbol %s in wordlist" % (symbol,))
	    # Fall through here in case there is a following {.
	    # The wrong condition will end up getting set, but
	    # at least the parse can continue detecting errors.
            # input.croak("unknown symbol %s in wordlist" % (symbol,))
            #raise SystemExit, 1
    if not list:
	input.complain("syntax error, expected a nonempty word list")
	# Because word lists are always the trailing element of a declaration,
	# we don't need to resync on error here.  Returning an empty list 
        # will allow the parse to proceed, catching more errors.
    return list

# Functions for validating the parse tree

def simple_error(file, line, errmsg):
    global errors
    if not debug and not errors:
        sys.stderr.write('\n')
    sys.stderr.write(error_leader(file, line) + errmsg)
    errors = errors + 1

def validate_boolean(expr, file, line, ok=0):
    # Check for ambiguous boolean expr terms.
    if isinstance(expr, cml.ConfigSymbol):
        if expr.type in ("trit", "decimal", "hexadecimal") and not ok:
            simple_error(file, line, "test of %s is ambiguous\n" % (expr.name,))
    elif type(expr) is type(()):
        validate_boolean(expr[1], file, line, expr[0] in _relops)
        validate_boolean(expr[2], file, line, expr[0] in _relops)

def validate_expr(expr, file, line):
    # Check for bad type combinations in expressions
    global errors
    # Return a leaf node type, inaccurate but good enough for
    # consistency checking.
    if isinstance(expr, cml.ConfigSymbol):
        if expr.is_numeric():
            return "integer"
        elif expr.is_logical():
            return "trit"
        else:
            return expr.type
    elif isinstance(expr, cml.trit):
        return "trit"
    elif type(expr) in (type(0), type(0L)):
        return "integer"
    elif type(expr) == type(""):
        return "string"
    elif expr[0] == '?':
        left = validate_expr(expr[2], file, line)
        right = validate_expr(expr[3], file, line)
        if left != right:
            simple_error(file, line, "types %s and %s don't match in ternary expression\n" % (left, right))
        return left
    elif expr[0] in _arithops:
        left = validate_expr(expr[1], file, line)
        if left not in ("integer", "trit"):
            simple_error(file, line, "bad %s left operand for arithmetic operator %s\n" % (left, expr[0]))
        right = validate_expr(expr[2], file, line)
        if right not in ("integer", "trit"):
            simple_error(file, line, "bad %s right operand for arithmetic operator %s\n" % (right, expr[0]))
        return "integer"
    elif expr[0] in _boolops or expr[0] in _termops:
        left = validate_expr(expr[1], file, line)
        if left != "trit":
            simple_error(file, line, "bad %s left operand for trit operator %s\n" % (left, expr[0]))
        right = validate_expr(expr[2], file, line)
        if right != "trit":
            simple_error(file, line, "bad %s right operand for trit operator %s\n" % (right, expr[0]))
        return "trit"
    elif expr[0] in _relops:
        left = validate_expr(expr[1], file, line)
        right = validate_expr(expr[2], file, line)
        if left != right:
            simple_error(file, line, "types %s and %s don't match in %s expression\n" % (expr[0], left, right))
        return "trit"
    else:
        if not debug and not errors:
            sys.stderr.write('\n')
        sys.stderr.write(error_leader(file, line) + \
                         "internal error: unexpected node %s in expression\n" % expr[0])
        errors = errors + 1

def symbols_by_preorder(node):
    # Get a list of config symbols in natural traverse order
    if node.items:
       sublists = map(symbols_by_preorder, node.items)
       flattened = []
       for m in sublists:
	   flattened = flattened + m
       return flattened
    else:
       return [node.name]

def resolve(exp):
    # Replace symbols in an expr with resolved versions
    global bad_symbols, errors
    if type(exp) is type(()):
	if exp[0] == 'not':
	    return ('not', resolve(exp[1]))
	elif exp[0] == '?':
	    return ('?', resolve(exp[1]), resolve(exp[2]), resolve(exp[3]))
	else:
	    return (exp[0], resolve(exp[1]), resolve(exp[2]))
    elif isinstance(exp, cml.ConfigSymbol):	# Symbol, already resolved
	return exp
    elif isinstance(exp, cml.trit):		# Trit, already resolved
	return exp
    elif type(exp) in (type(0), type("")):	# Constant, already resolved
	return exp
    elif not hasattr(exp, "type"):
	sys.stderr.write("Symbol %s has no type.\n" % (exp,))
	errors = errors + 1
	return None
    elif exp.type == 'trit':
        if exp.attr == 'y':
            return cml.y
        elif exp.attr == 'm':
            return cml.m
        elif exp.attr == 'n':
            return cml.n
    elif exp.type in _atoms:
	return exp.attr
    elif dictionary.has_key(exp.attr):
	return dictionary[exp.attr]
    else:
	bad_symbols[exp.attr] = 1
	return None

def ancestry_check(symbol, counts):
    # Check for circular ancestry chains
    # print "Checking ancestry of %s: %s" % (symbol, symbol.ancestors)
    counts[symbol] = 1
    for ancestor in symbol.ancestors:
        if counts.has_key(ancestor.name):
            raise NameError, symbol.name + " through " + ancestor.name
        else:
            map(lambda symbol, counts=counts: ancestry_check(symbol, counts), symbol.ancestors)

def circularity_check(name, exp, counts):
    # Recursive circularity check...
    # print "Expression check of %s against %s" % (name, exp)
    if type(exp) is type(()):
        if exp[0] == '?':
	    circularity_check(name, exp[1], counts)
	    circularity_check(name, exp[2], counts)
	    circularity_check(name, exp[3], counts)
	else:
	    circularity_check(name, exp[1], counts)
	    circularity_check(name, exp[2], counts)
    elif isinstance(exp, cml.ConfigSymbol) and name == exp.name:
        raise NameError, name
    elif hasattr(exp, "default"):
	vars = cml.flatten_expr(exp.default)
	# print "Components of %s default: %s" % (exp.name vars) 
	for v in vars:
	    if v.name == name:
		raise NameError, name
	    elif counts.has_key(v.name):
		pass		# Already checked this branch
	    else:
		counts[v.name] = 1
		circularity_check(name, v.name, counts)

def transitive_depends(guard, entry):
    "Test transitive completion of dependency."
    # We don't also check visibility, because visibility guards can
    # have disjunctions and it would be wrong to propagate up both branches.
    if entry.menu:
        searchpath = entry.ancestors + entry.menu.ancestors
    else:
        searchpath = entry.ancestors
    if guard in searchpath:
        return 1
    for x in searchpath:
        if transitive_depends(x, guard):
            return 1
    return 0

def error_leader(file, line):
    global errors
    return '"%s", line %d:' % (file, line)

def postcomplain(msg):
    global errors
    if not debug and not errors:
        sys.stderr.write('\n')
    sys.stderr.write("cmlcompile: " + msg)
    errors += 1

def main(debug, outfile, arguments):
    "Sequence a compilation"
    global profile, expression_parser, requirements, bad_symbols, errors

    if not debug:
        baton = cml.Baton("Compiling rules, please wait")
    else:
        baton = None

    if profile:
        import time 
        now = zerotime = basetime = time.time();

    # Parse everything
    try:
	if not arguments:
	    parse(lexwrapper(sys.stdin, "stdin"), baton)
	else:
	    for file in arguments:
		parse(lexwrapper(open(file), file), baton)
    except IOError, details:
	sys.stderr.write("cmlcompile: I/O error, %s\n" % (details,))
	raise SystemExit, 1


    if profile:
        now = time.time();
        print "Rule parsing:", now - basetime
        basetime = now
    if not debug and not errors:
        baton.twirl()

    # Sanity and consistency checks:

    # We need a main menu declaration
    if not start:
	postcomplain("missing a start declaration.\n")
	raise SystemExit, 1
    elif not dictionary.has_key(start):
	postcomplain("declared start menu '%s' does not exist.\n"%(start,))
	raise SystemExit, 1
    if not debug and not errors:
        baton.twirl()

    # Check for unfulfilled forward declarations
    if forward_private:
        postcomplain("symbols declared private but never defined: %s\n" % forward_private)
        
    # Check that all symbols other than those on the right side of
    # derives are either known or themselves derived.
    for entry in dictionary.values():
        if entry.visibility:
            entry.visibility = resolve(entry.visibility)
	if entry.default:
	    entry.default = resolve(entry.default)
    requirements = map(lambda x: cml.Requirement(resolve(x.predicate), x.message, x.file, x.line), requirements)
    if bad_symbols:
	postcomplain("%d symbols could not be resolved:\n"%(len(bad_symbols),))
	sys.stderr.write(`bad_symbols.keys()` + "\n")
    if not debug and not errors:
        baton.twirl()

    # Now associate a type with all derived symbols.  Since such symbols
    # are never queried, the only place this is used is in formatting
    # the symbol's appearance in the final configuration file.
    #
    # (The outer loop forces this to keep spinning until it has done all
    # possible deductions, even in the presence of forward declarations.)
    while 1:
        deducecount = 0
        for entry in dictionary.values():
            if entry.is_derived() and not entry.type:
                derived_type = None
                if entry.default == cml.m:
                    derived_type = "trit"
                elif entry.default == cml.n or entry.default == cml.y:
                    derived_type = "bool"
                elif type(entry.default) is type(()):
                    if entry.default[0] in _boolops + _relops:
                        derived_type = "bool"
                    elif entry.default[0] in _termops:
                        derived_type = "trit"
                    elif entry.default[0] in _arithops:
                        derived_type = "decimal"
                    elif entry.default[0] == '?':
                        if isinstance(entry.default[2], cml.ConfigSymbol):
                            derived_type = entry.default[2].type
                        elif isinstance(entry.default[2], cml.trit):
                            derived_type = "trit"
                        elif type(entry.default[2]) in (type(0), type(0L)):
                            derived_type = "decimal"
                        elif type(entry.default[2]) is type(""):
                            derived_type = "string"
                elif type(entry.default) is type(0):
                    derived_type = "decimal"		# Could be hex
                elif isinstance(entry.default, cml.ConfigSymbol):
                    derived_type = entry.default.type
                if derived_type:
                    entry.type = derived_type
                    deducecount = 1
        if not deducecount:
            break

    for entry in dictionary.values():
        if entry.is_derived() and not entry.type:
            postcomplain(error_leader(entry.file, entry.lineno) + \
                             'can\'t deduce type for derived symbol %s\n' % (entry.name,))
    if not debug and not errors:
        baton.twirl()

    # Now run our ambiguity check on all unless expressions.
    for (guard, file, line) in bool_tests:
        validate_boolean(resolve(guard), file, line)
    if not debug and not errors:
        baton.twirl()

    # Handle explicit dependencies
    for symbol in explicit_ancestors.keys():
        for guard in explicit_ancestors[symbol]:
            for guardsymbol in cml.flatten_expr(resolve(guard)):
                make_dependent(guardsymbol, symbol)
    if not debug and not errors:
        baton.twirl()


    # Check that every symbol in the table is referenced from a menu
    # exactly once (except explanations).  We checked for multiple
    # inclusions at parse-tree generation time.  Now...
    bad_symbols = {}
    for entry in dictionary.values():
	if entry.prompt and entry.type!='explanation' and not entry.menu and \
           	entry.name != start and entry.prompt != banner:
	    bad_symbols[entry.name] = 1 
    if bad_symbols:
	postcomplain("%d symbols have no menus"%(len(bad_symbols),))
        sys.stderr.write("\n" +`bad_symbols.keys()` + "\n")
    if not debug and not errors:
        baton.twirl()

    # Check for forward references in visibility constraints.
    # Note: this is *not* a fatal error.
    preorder = symbols_by_preorder(dictionary[start])
    for i in range(len(preorder)):
	key = preorder[i]
	forwards = []
        for guards in cml.flatten_expr(dictionary[key].visibility):
            if guards.name in preorder[i+1:]:
                forwards.append(guards.name)
	if forwards:
	    sym = dictionary[key]
            postcomplain('"%s", line %d: %s requires %s forward\n' % (sym.file, sym.lineno, key, forwards))
            errors -= 1
    if not debug and not errors:
        baton.twirl()

    # Check for circularities in derives and defaults.
    try:
        for entry in dictionary.values():
            if entry.default:
                expr_counts = {}
                circularity_check(entry.name, entry.default, expr_counts)
            if entry.visibility:
                expr_counts = {}
                circularity_check(entry.name, entry.visibility, expr_counts)
            if entry.ancestors:
                ancestor_counts = {}
                ancestry_check(entry, ancestor_counts)
    except NameError:
	postcomplain("%s depends on itself\n"%(sys.exc_value,))
    if not debug and not errors:
        baton.twirl()

    # Various small hacks combined here to save traversal overhead.
    for entry in dictionary.values():
        # Validate the formulas for boolean derived symbols.
        if entry.is_derived():
            if entry.menu:
		postcomplain("menu %s contains derived symbol %s\n"%(entry.menu.name, `entry`))
            if entry.type == "bool":
                validate_boolean(entry.default, entry.file, entry.lineno)
            else:
                validate_expr(entry.default, entry.file, entry.lineno)
            continue
        #if not entry.default is None:
        #    validate_expr(entry.default, entry.file, entry.lineno)
        # Give childless menus the `message' type.  This will make
        # it easier for the front end to do special things with these objects.
	if entry.type == 'menu':
	    if not entry.items:
		entry.type = 'message'
            continue
        # Check for type messages between symbols and their defaults.
        if entry.is_symbol() and not entry.default is None:
            if type(entry.default) in (type(0L),type(0)) and not entry.is_numeric():
                postcomplain("%s is not of numeric type but has numeric constant default\n" % entry.name)
            elif type(entry.default) == type("") and not entry.type == "string":
                postcomplain("%s is not of string type but has string constant default\n" % entry.name)

        # Symbols with decimal/hexadecimal/string type must have a default.
	if entry.type in ("decimal", "hexadecimal", "string"):
	    if entry.default is None:
		postcomplain("%s needs a default\n"%(`entry`,))
        elif entry.range:
            # This member can be used by front ends to determine whether the
            # entry's value should be queried with a pulldown of its values.
            entry.discrete = not filter(lambda x: type(x) is type(()), entry.range)
	    # This member can be used by front ends to determine whether the
	    # entry's value should be queried with a pulldown of enums.
            entry.enum = type(entry.range[0]) is type(()) \
                         and type(entry.range[0][0]) is type("")
        # Now hack the prompts of anything dependent on a warndepend symbol
        for guard in warndepend:
            if entry.prompt and transitive_depends(guard, entry):
                entry.prompt = entry.prompt + " (" + guard.name + ")"
    if not debug and not errors:
        baton.twirl()

    # Check for constraint violations.  If the defaults set up by the
    # rule file are not consistent, it's not likely the user will make
    # a consistent one.  Don't try this if we've seen syntax errors, as
    # they tend to produce Nones in expressions that this will barf on.
    if not errors:
        for wff in requirements:
            if not cml.evaluate(wff.predicate, debug):
                postcomplain(error_leader(wff.file, wff.line) + " constraint violation: %s\n" % `wff`)
        if not debug and not errors:
            baton.twirl()

    # Now integrate the help references
    for key in dictionary.keys():
	if help_dict.has_key(key):
	    dictionary[key].helptext = help_dict[key]
	    del help_dict[key]
    if debug:
	missing = []
	for entry in dictionary.values():
	    if not entry.type in ("message", "menu", "choices", "explanation") and entry.prompt and not entry.help():
		missing.append(entry.name)
	if missing:
	    postcomplain("The following symbols lack help entries: %s\n" % missing)
	orphans = help_dict.keys()
        if orphans:
            postcomplain("The following help entries do not correspond to symbols: %s\n" % orphans)
    if not debug and not errors:
        baton.end("Done")

    if profile:
        now = time.time();
        print "Sanity checks:", now - basetime
        basetime = now

    # Package everything up for pickling
    if errors:
	postcomplain("rulebase write suppressed due to errors.\n")
	raise SystemExit, 1
    else:
	system = cml.CMLRulebase(dictionary[start], dictionary, requirements, prefix, banner, icon)
        system.trit_tie = trit_tie
        system.help_tie = help_tie

        # Precomputation to speed up the configurator's load time
        system.reduced = map(lambda x: x.predicate, system.constraints)
        system.optimize_constraint_access()
        if debug:
            cc = dc = tc = 0
            for symbol in system.dictionary.values():
                if not symbol.is_derived():
                    tc = tc + 1
                if symbol.dependents:
                    dc = dc + 1
                if symbol.constraints:
                    cc = cc + 1
            print "%d total symbols; %d symbols are involved in constraints; %d in dependencies." % (tc, cc, dc)

	try:
	    if debug: print "cmlcompile: output directed to %s" % (outfile)
	    out = open(outfile, "wb")
	    cPickle.dump(system, out, 1)
	    out.close()
	except:
	    postcomplain('couldn\'t open output file "%s"\n' % (outfile,))
	    sys.exit(1)

    if profile:
        now = time.time();
        print "Rulebase output:", now - basetime
        basetime = now
        print "Total time:", now - zerotime

if __name__ == '__main__': 
    (options, arguments) = getopt.getopt(sys.argv[1:], "o:Pv")
    for (switch, val) in options:
	if switch == '-o':
	    outfile = val
	elif switch == '-P':
	    profile = 1
	elif switch == '-v':
	    debug = debug + 1

    if profile:
        import profile
        profile.run("main(debug, outfile, arguments)")
    else:
        main(debug, outfile, arguments)

# That's all, folks!
