"""
cmlsystem.py -- CML2 configurator front-end support

by Eric S. Raymond, <esr@thyrsus.com>

Mode-independent code for front ends.  This supplies one big class that you
initialize by loading a compiled rulebase.  The class provides functions for
doing tests and manipulations on the rulebase, and for writing out
configuration files.
"""

import os, sys, re
import cml

_eng = {
    "ABOUT":"About to write %s=%s (type %s)",
    "BADEQUALS":"bad token `%s' while expecting '='.",
    "BADVERSION":"Compiler/configurator version mismatch, recompile the rulebase please.\n",
    "BADTOKEN":"bad token `%s' while expecting symbol.",
    "BADTRIT":"Boolean symbol %s cannot have value m.",
    "BINDING":"    Binding from constraint: %s (source %s)",
    "CHECKDONE":"    ...menu %s check done",
    "CHECKING":"    Checking menu %s...",
    "COMMIT":"    Committing new bindings.",
    "CONSTRAINT":"    Value %s for %s failed constraint %s",
    #"DERIVED":"Cannot set derived symbol %s",
    "EXCLUDED":"    %s value %s excluded by %s",
    "FAILREQ":"    Failed constraint %s",
    #"FORBIDDEN":"%s=%s would violate %s",
    "FROZEN":" (frozen)",
    "HELPFLAG":"Help-required flag is %s",
    "INCONST":"Symbol %s forced to n during recovery attempt.\n",
    "INVISANC":"    %s not visible, ancestor %s false",
    "INVISME":"    %s not visible, %s guard %s is false",
    "INVISSTART":"    is_visible(%s) called",
    "MODULESN":"All M-valued symbols will be forced to Y.",
    "MODULESM":"All tristate symbols will default to M.",
    "MODULESY":"Tristate symbols won't default to M.",
    "NOHELP":"    %s not visible, it has no help",
    "RADIOINVIS":"    Query of choices menu %s elided, button pressed",
    "READING":"Reading configuration from %s",
    "RECOVERY":"Attempting recovery from invalid configuration:",
    "RECOVEROK":"Recovery OK.",
    "REDUNDANT":"    Redundant assignment forced by %s", 
    "RENAME":"Attempt to rename %s to %s failed.",
    "ROLLBACK":"    Rolling back new bindings: ",
    "SAVEEND":"#\n# That's all, folks!\n",
    "SETTING":"%s=%s",
    "SHAUTOGEN":"#\n# Automatically generated, don't edit\n#\n",
    "SHDERIVED":"#\n# Derived symbols\n#\n",
    "SIDEEFFECT":" (deduced from %s)",
    "SUBINVIS":"    %s not visible, all subqueries invisible.",
    "TRIGGER":"    Set of %s = %s triggered by guard %s",
    "TRITFLAG":"Trit flag is now %s",
    "TRITSOFF":"    %s not visible, trits are suppressed",
    "TYPEUNKNOWN":"Node %s unknown value type: %s %s",
    "UNCHANGED":"    Symbol %s unchanged",
    "UNLOAD1":"File load violated these constraints:",
    "UNLOAD2":"Undoing file loads, recovery failed these constraints:",
    #"UNSATISFIABLE":"Ruleset found unsatisfiable while setting %s",
    "USERSETTING":"User action on %s.",
    "VALIDRANGE":"    Valid range of %s is %s",
    "VALUNKNOWN":"Node %s %s has unknown value: %s",
    "VISIBLE":"    Query of %s *not* elided",
}

class CMLSystem(cml.CMLRulebase):
    "A rulebase, from the point of view of a front end."
    def __init__(self, rulebase):
        import cPickle
        rules = cPickle.load(open(rulebase, "rb"))

        # Copy the symbol table, adding runtime value state.
        self.__dict__ = rules.__dict__
        for entry in self.dictionary.values():
            entry.frozen = 0		# True if it has been frozen
            if entry.type == "choices":
                entry.menuvalue = entry.default
            else:
                entry.menuvalue = None
            entry.inspected = 0		# Track menu inspections
            entry.setcount = 0		# Should this symbol be written?
            entry.included = 0		# Seen in an inclusion?
            entry.bindingstack = []	# Symbol's value stack
        self.oldbindings = {}
        self.newbindings = {}
        self.chilled = {}
        self.touched = []		# Does it have an uncommitted binding?

        # Enhance the ConfigSymbol methods to deal with the value state
        cml.ConfigSymbol.oldeval = cml.ConfigSymbol.eval
        def _neweval(symbol, debug=0, self=self):	# Not a method!
            value = self.__bindeval(symbol, debug=debug)
            if value != None:
                return value
            value = cml.ConfigSymbol.oldeval(symbol, debug)
            if value != None:
                return value
            elif symbol.type in ("decimal", "hexadecimal"):
                return 0
            elif symbol.type == "string":
                return ""
            elif symbol.type == "bool":
                return cml.n
            elif symbol.type == "trit":
                if self.module_default and not symbol.constraints:
                    return cml.m
                else:
                    return cml.n
            else:
                return None	# for menu, choices, and message-valued symbols
        cml.ConfigSymbol.eval = _neweval

        def _newstr(symbol):		# Not a method!
            res = "%s={" % (symbol.name)
            res = res + symbol.dump()
            value = symbol.eval()
            if value != None:
                res = res + " value " + str(value) + "," 
            if symbol.frozen:
                res = res + " frozen,"
            return res[:-1] + "}"
        cml.ConfigSymbol.__str__ = _newstr

        # Extra state for the configuration object
        self.debug = 0			# Initially, no status logging
        self.errout = sys.stderr	# Where to do status logging
        self.suppressions = 1		# Yes, do visibility checks
        self.inclusion = 0		# Are we doing a file inclusion?
        self.trits_enabled = 0		# Initially, support booleans only
        self.commits = 0		# Value sets since last save 
        self.lang = _eng		# Someday we'll support more languages.
        self.side_effects = []		# Track side effects
        self.module_suppress = 0	# Make trits invisible	
        self.module_default = 0		# Should we default to modular?	

        if rules.version != cml.version:
            sys.stderr.write(self.lang["BADVERSION"])
            raise SystemExit, 1

    # Utility code

    def debug_emit(self, threshold, msg):
        "Conditionally emit a debug message to the designated error stream."
        if self.debug >= threshold:
            self.errout.write(msg + "\n")

    # Handling of symbol bindings is encapsulated here.  The semantics we
    # want is for every side-effect to be associated with the symbol
    # whose user-specified change in value triggered it.  That way, if
    # and when the user changes the symbol's value again, we can back out'
    # all previous side-effects contingent on that symbol.
    #
    # Essentially what we're doing here is journalling all bindings.
    # It has to be this way, because different values of the same symbol
    # could trigger completely different side effects depending on
    # how the constraints are written.
    class __Binding:
        def __init__(self, symbol, value, source, link):
            self.symbol = symbol
            self.value = value
            self.source = source
            self.link = link
            self.visible = 1

    def __bindeval(self, symbol, debug=0):
        "Get the most recent visible value for symbol off the binding stack."
        #self.debug_emit(2, "    bindeval(%s)" % (symbol.name))
        if not hasattr(symbol, "bindingstack"):
            return None
        for binding in symbol.bindingstack:
            if binding.visible:
                return binding.value
    def __bindsymbol(self, symbol, value, source=None):
        "Bind symbol to a given value."
        self.debug_emit(2, "    bindsymbol(%s, %s, %s)" % (symbol.name, value, `source`))
        if not source:
            source = symbol
        # Here is the actual binding-stack hack 
        newbinding = self.__Binding(symbol,value, source,self.newbindings.get(source))
        self.newbindings[source] = newbinding
        symbol.bindingstack = [newbinding] + symbol.bindingstack
        # Side-effect tracking
        if source == None or source == symbol:
            side = ""
        else:
            side = self.lang["SIDEEFFECT"] % (`source`,)
        side = self.lang["SETTING"] % (symbol.name, value) + side
        if source and source != symbol:
            self.side_effects.append(side)
        if symbol == self.trit_tie:
            if value == cml.n:
                self.side_effects.append(self.lang["MODULESN"])
            elif value == cml.m:
                self.side_effects.append(self.lang["MODULESM"])
            elif value == cml.y:
                self.side_effects.append(self.lang["MODULESY"])
        # Debugging support
        if self.debug:
            self.debug_emit(1, "    " + side)
        # Mark it to be written
        menu = symbol
        while menu:
            menu.setcount += 1
            menu = menu.menu
    def __unbindsymbol(self, symbol, context):
        "Remove all bindings of a given symbol from a given context"
        listhead = context.get(symbol)
        if listhead:
            while listhead:
                listhead.symbol.bindingstack.remove(listhead)
                listhead = listhead.link
            del context[symbol]
            if symbol.menu.type == "choices":
                for sibling in symbol.menu.items:
                    if sibling.eval():
                        symbol.menu.menuvalue = sibling
                        break
    def __bindcommit(self):
        "Commit all new bindings."
        undo = []
        # When we undo side-effects from a choice symbol, all the side-effects
        # from setting its siblings have to be backed out too.
        for primary in self.newbindings.keys():
            if primary.menu.type == "choices":
                undo.extend(primary.menu.items)
            else:
                undo.append(primary)
        # This is the magic moment that undoes side-effects
        for symbol in undo:
            self.__unbindsymbol(symbol, self.oldbindings)
        self.oldbindings.update(self.newbindings)
        self.newbindings.clear()
        self.commits = self.commits + 1
    def __bindreveal(self, primary):
        "Make every old binding of symbol visible."
        bindings = self.oldbindings.get(primary)
        if bindings:
            listhead = bindings
            while listhead:
                listhead.visible = 1
                listhead = listhead.link
    def __bindconceal(self, primary):
        "Temporarily make old bindings hanging on a given primary invisible."
        bindings = self.oldbindings.get(primary)
        if bindings:
            listhead = bindings
            while listhead:
                listhead.visible = 0
                listhead = listhead.link
    def binddump(self, context=None):
        "Dump the state of the bindings stack."
        res = ""
        if context == None:
            context=self.oldbindings
        for (primary, bindings) in context.items():
            res = res + "# %s(%s, touched=%d): " % (primary.name,("inactive","active")[bindings.visible], primary in self.touched)
            while bindings:
                res = res + bindings.symbol.name + "=" + `bindings.value` + ", "
                bindings = bindings.link
            res = res[:-2] + "\n"
        return res

    #
    # Loading and saving
    #
    def load(self, file, freeze=0):
        "Load bindings from a defconfig-format configuration file."
        import shlex
        stream = shlex.shlex(open(file), file)
        self.debug_emit(1, self.lang["READING"] % (file,))
        changes = 0
        errors = ""
        self.inclusion = 1
        self.side_effects = []	# Not needed if we're using set_symbol below 
        while 1:
            dobind = 1
            symname = stream.get_token()
            if not symname:
                break
            if len(self.prefix) and symname[0:len(self.prefix)] == self.prefix:
                symname = symname[len(self.prefix):]
            if self.dictionary.has_key(symname):
                symbol = self.dictionary[symname]
            else:
                symbol = None
                errmsg = stream.error_leader()+self.lang["BADTOKEN"]%(symname)
                errors = errors + errmsg + "\n"
                dobind = 0

            sep = stream.get_token()
            if sep != '=':
                errmsg = stream.error_leader()+self.lang["BADEQUALS"]%(sep,)
                errors = errors + errmsg + "\n"
                dobind = 0
            value = stream.get_token()

            # Do this check early to avoid Python errors if the file
            # is malformed.
            if not dobind:
                continue
            
            if value[0] in stream.quotes:
                value = value[1:-1]

            if not symbol:
                continue
            # We can't permit these files to override derivations
            if symbol.is_derived():
                #errmsg = stream.error_leader()+self.lang["DERIVED"]%(symbol.name,)
                #errors = errors + errmsg + "\n"
                continue

            # Michael Chastain's case -- treat variable as new if
            # value is m but the type cas been changed to bool
            if value == 'm' and symbol.type == "bool":
                errmsg = stream.error_leader()+self.lang["BADTRIT"]%(symbol,)
                errors = errors + errmsg + "\n"
                continue

            # Note that we've seen this in an inclusion -- it's not new.
            symbol.included = 1

            # If we load a configuration with trit values,
            # force those to be enabled.
            if symbol.type == "trit" and value == "m":
                self.trits_enabled = 1

            # Don't count changes to variables that are already set at the
            # desired value.  Do them, though, so they can be frozen.
            oldval = symbol.eval()
            # Use this for consistency checking by file
            self.__set_symbol_internal(symbol,
                                       self.value_from_string(symbol,value))
            self.chilled.clear()
            if value != oldval:
                changes = changes + 1

        stream.instream.close()
        self.inclusion = 0
        violations = self.sanecheck()
        if not violations:
            self.__commit(freeze)
        else:
            errors += self.lang["UNLOAD1"]+"\n" + "\n".join(violations) + "\n"
            # This is an attempt to recover from inconsistent
            # configurations.  (Unusual case -- typically happens only
            # when a new constraint is added to a rulebase, not in the
            # more common case of new symbols.)  General recovery is
            # very hard, it involves constrained satisfaction problems
            # for which there are not just no good algorithms, there
            # are no clean definitions.  We settle for a simple,
            # stupid hack.  Force all the unfrozen symbols in the
            # violated constraints to n and see what happens.
            nukem = []
            self.debug_emit(1, self.lang["RECOVERY"])
            for i in range(len(self.constraints)):
                if not cml.evaluate(self.reduced[i], self.debug):
                    flattened = cml.flatten_expr(self.constraints[i].predicate)
                    for sym in flattened:
                        if not sym in nukem and not sym.frozen and sym.is_logical():
                            nukem.append(sym)
            if nukem:
                errors += self.lang["RECOVERY"] + "\n"
                for sym in nukem:
                    errors += self.lang["INCONST"] % (sym.name,)
                    self.__set_symbol_internal(sym, cml.n)
                    self.chilled.clear()
                violations = self.sanecheck()
                if not violations:
                    self.__commit(freeze)
                    errors += self.lang["RECOVEROK"]
                else:
                    self.__rollback()
                    errors += self.lang["UNLOAD2"]+"\n" + "\n".join(violations) + "\n"
        return (changes, errors)

    def saveable(self, node):
        "Should this symbol be visible in the configuration written out?"
        if node.setcount:
            return 1
        if node.visibility == cml.n:	# Result of variable freezes
            return 0
        elif node.ancestors and node.ancestors[-1].eval() == cml.n:
            return 0
        return 1

    def save(self, config, baton=None):
        "Save a configuration to the named output file."
        newbindings = None
        if self.newbindings:
            newbindings = self.newbindings
            self.newbindings = {}
        try:
            shelltemp = ".tmpconfig%d.sh" % os.getpid()
            config_sh = open(shelltemp, "w")
            config_sh.write(self.lang["SHAUTOGEN"])
            # Write mutable symbols, including defaulted modular symbols.
            self.module_suppress = 0
            self.__save_recurse(self.start, config_sh, baton=None)
            if self.trit_tie:
                self.module_suppress = (self.trit_tie.eval() == cml.m)
            # Write all derived symbols
            config_sh.write(self.lang["SHDERIVED"])
            for entry in self.dictionary.values():
                if entry.is_derived():
                    if baton:
                        baton.twirl()
                    self.__save_recurse(entry, config_sh)
            # Perhaps this is a crash dump from an inconsistent ruleset?
            if newbindings:
                self.newbindings = newbindings
                config_sh.write("#\n# Uncommitted bindings\n#\n")
                config_sh.write(self.binddump(self.newbindings))
            config_sh.write(self.lang["SAVEEND"])
            config_sh.close()
            try:
                os.rename(shelltemp, config)
            except OSError:
                reason  = self.lang["RENAME"] % (shelltemp, config,)
                raise IOError, reason
            self.commits = 0
            if baton:
                baton.end()
            return None
        except IOError, details:
            return details.args[0]

    def __save_recurse(self, node, shellstream, baton=None):
        if not self.saveable(node):
            return
        elif node.items:
            shellstream.write("\n#\n# %s\n#\n" % (node.prompt,))
            for child in node.items:
                self.__save_recurse(child, shellstream)
            shellstream.write("\n")
        elif node.type != 'message':
            symname = self.prefix + node.name
            value = node.eval(self.debug)
            self.debug_emit(2, self.lang["ABOUT"] %(symname,value,type(value)))
            try:
                if node.type == "decimal":
                    shellstream.write("%s=%d\n" % (symname, value))
                elif node.type == "hexadecimal":
                    shellstream.write("%s=0x%x\n" % (symname, value))
                elif node.type == "string":
                    shellstream.write("%s=\"%s\"\n" % (symname, value))
                elif node.type in ("bool", "trit"):
                    shellstream.write("%s=%s\n" % (symname, `value`))
                elif value == None and node.is_logical():
                    shellstream.write("%s=n\n" % (symname,))
                else:
                    raise ValueError, self.lang["VALUNKNOWN"] % (node,node.type,value)
            except:
                (errtype, errval, errtrace) = sys.exc_info()
                print "Internal error %s while writing %s." % (errtype, node)
                raise SystemExit, 1
        if baton:
            baton.twirl()

    # Symbol predicates.

    def is_mutable(self, symbol):
        "Is a symbol mutable (not frozen, not derived)?"
        return isinstance(symbol, cml.ConfigSymbol) and not symbol.is_derived() and not symbol.frozen

    def is_visible(self, query):
        "Should we ask this question?"
        self.debug_emit(2, self.lang["INVISSTART"] % (query.name,))
        # Maybe we're not doing elisions
        if not self.suppressions:
            return 1
        # Maybe we're suppressing trits entirely.
        if query.type == "trit":
            if self.module_suppress and not query.vital and not query.constraints and not query.setcount:
                self.debug_emit(2, self.lang["TRITSOFF"] % query.name)
                return 0
        # Maybe it has no help.
        if query.is_symbol() and not query.help() and self.help_tie and not self.help_tie.eval():
            self.debug_emit(2, self.lang["NOHELP"] % query.name)
            return 0
        # Check to see if the symbol or any menu in the chain above it
        # is suppressed by a visibility constraint.
        upward = query
        while upward != self.start:
            if upward.visibility != None and not cml.evaluate(upward.visibility):
                self.debug_emit(2, self.lang["INVISME"] % (query.name, upward.name, cml.display_expression(upward.visibility)))
                return 0
            upward = upward.menu
        # OK, now check that all ancestors are visible
        if not self.__dep_visible(query):
            return 0
        # Now check ancestors of the containing menu
        if query.menu and not self.__dep_visible(query.menu):
            return 0
        # Elide a menu if all subqueries are invisible, or a choices if one is.
        if query.items:
            setcount = subqueries = 0
            self.debug_emit(2, self.lang["CHECKING"] % (query.name,))
            for child in query.items:
                if child.type != "message":
                    subqueries = subqueries + 1
                    if child.visibility and not cml.evaluate(child.visibility):
                        setcount = setcount + 1
            self.debug_emit(2, self.lang["CHECKDONE"] % (query.name,))
            if query.type == 'choices' and setcount > 0:
                self.debug_emit(2, self.lang["RADIOINVIS"] % (query.name,))
                return 0
            elif query.type == 'menu' and setcount == subqueries:
                self.debug_emit(2, self.lang["SUBINVIS"] % (query.name,))
                return 0
        # All tests passed, it's visible
        self.debug_emit(2, self.lang["VISIBLE"] % (query.name))
        return 1

    #
    # All the properties of the dependency relationship are implemented here
    #
    def __dep_value_ok(self, symbol, value):
        "Do ancestry relationships allow given value of given symbol"
        for ancestor in symbol.ancestors:
            v = cml.evaluate(ancestor, self.debug)
            if (symbol.type == "trit" and value > v):
                break
            elif (symbol.type =="bool" and value > (v != cml.n)):
                break
        else:
            return 1	# Tricky use of for-else
        self.debug_emit(2, self.lang["EXCLUDED"] % (`symbol`, `cml.trit(value)`, `ancestor`))
        return 0

    def __dep_visible(self, symbol):
        "Do ancestry relations allow a symbol to be visible?"
        # Note: we don't need to recurse here, assuming dependencies
        # get propagated correctly.
        for super in symbol.ancestors:
            if not cml.evaluate(super):
                self.debug_emit(2,self.lang["INVISANC"]%(symbol.name,super.name))
                return 0
        return 1

    def __dep_force_ancestors(self, dependent, dependvalue, ancestor):
        "Force a symbol's ancestors based on the symbol's value."
        self.debug_emit(2, "    dep_force_ancestors(%s, %s, %s)" % (`dependent`, `dependvalue`, `ancestor`))
        if dependent.type in ("bool", "trit") and ancestor.is_logical():
            anctype = ancestor.type
            ancvalue = cml.evaluate(ancestor)
            if dependvalue > ancvalue:
                if dependent.type == anctype or not self.trits_enabled:
                    self.__set_symbol_internal(ancestor, dependvalue,dependent)
                elif dependent.type == "trit":
                    self.__set_symbol_internal(ancestor, dependvalue,dependent)
        # Recurse upwards, first through ancestors... 
        for upper in ancestor.ancestors:
            self.__dep_force_ancestors(dependent, dependvalue, upper)
        # ...and then through the containing menu.
        for upper in ancestor.menu.ancestors:
            self.__dep_force_ancestors(dependent, dependvalue, upper)

    def __dep_force_dependents(self, guard, guardvalue, dependent):
        self.debug_emit(2, "    dep_force_dependents(%s, %s, %s)" % (`guard`, `guardvalue`, `dependent`))
        if guard.type in ("bool", "trit") and dependent.is_logical():
            deptype = dependent.type
            depvalue = cml.evaluate(dependent)
            if guardvalue < depvalue:
                if guard.type == deptype or not self.trits_enabled:
                    self.__set_symbol_internal(dependent, guardvalue, guard)
                elif deptype == "trit":
                    self.__set_symbol_internal(dependent, guardvalue, guard)
                elif guardvalue == cml.n:
                    self.__set_symbol_internal(dependent, guardvalue, guard)
        # Recurse downwards...
        if dependent.items:
            for child in dependent.items:
                self.__dep_force_dependents(guard, guardvalue, child)
        else:
            for lower in dependent.dependents:
                self.__dep_force_dependents(guard, guardvalue, lower)

    #
    # The following methods handle variable bindings
    #
    def __rollback(self):
        "Roll back all new bindings."
        self.debug_emit(1, self.lang["ROLLBACK"] + `self.touched`)
        self.touched = []
        for symbol in self.newbindings.keys():
            self.__unbindsymbol(symbol, self.newbindings)
        self.side_effects = []

    def __commit(self, freeze=0, baton=None):
        "Commit all new bindings."
        if freeze:
            self.debug_emit(1, self.lang["COMMIT"] + self.lang["FROZEN"])
        else:
            self.debug_emit(1, self.lang["COMMIT"])
            if self.trit_tie and self.trit_tie in self.touched:
                self.trits_enabled = cml.evaluate(self.trit_tie)
                self.module_suppress = self.module_default = (self.trits_enabled == cml.m)
                self.debug_emit(1, self.lang["TRITFLAG"] % (`cml.trit(self.trits_enabled)`,))
            if self.help_tie and self.help_tie in self.touched:
                self.debug_emit(1, self.lang["HELPFLAG"] % self.help_tie.eval())
        for entry in self.touched:
            if freeze:
                entry.frozen = freeze
                if entry.menu.type == "choices":
                    entry.menu.frozen = 1
                    entry.menu.value = entry.name
                    for sibling in entry.menu.items:
                        sibling.frozen = freeze
        self.touched = []
        if baton:
            baton.twirl("#")
        if freeze:
            # Optimization hack -- undo this if variables can ever be unfrozen.
            # In the meantime, this greatly reduces the amount of expression
            # evaluation needed after variables have been frozen.
            for i in range(len(self.reduced)):
                simplified = self.__eval_frozen(self.reduced[i])
                if simplified != None:
                    self.reduced[i] = simplified
            self.optimize_constraint_access()
            if baton:
                baton.twirl("#")
        # Must do this *after* checking freezes
        self.__bindcommit()		# The magic moment
        if baton:
            baton.end()

    def sanecheck(self):
        "Sanity-check a configuration and report on its side effects."
        violations = []
        for i in range(len(self.constraints)):
            if not cml.evaluate(self.reduced[i], self.debug):
                violations.append(repr(self.constraints[i]));
                self.debug_emit(1, self.lang["FAILREQ"]%(self.constraints[i],))
        return violations

    def set_symbol(self, symbol, value, freeze=0):
        "Bind a symbol, tracking side effects."
        self.debug_emit(1, self.lang["USERSETTING"] % (symbol.name,))
        self.side_effects = []
        self.__set_symbol_internal(symbol, value)
        self.chilled.clear()
        violations = self.sanecheck()
        if not violations:
            self.__commit(freeze)
            return (1, self.side_effects, [])
        else:
            effects = self.side_effects
            self.__rollback()		# This will clear self.side_effects
            return (0, effects, violations)

    def __set_symbol_internal(self, symbol, value, source=None):
        "Recursively bind a symbol, with side effects."
        self.debug_emit(2, "    set_symbol_internal(%s, %s, %s)" % (symbol.name, value, `source`))
        if not source:
            source = symbol
        # The "touched" property marks this symbol changed for freeze purposes.
        # It has to stay on until the next commit.
        self.touched.append(symbol)
        # If it already has the desired value, we're done.
        if symbol.eval(self.debug) == value:
            self.debug_emit(1, self.lang["UNCHANGED"] % (symbol.name,))
            return
        # Barf on attempt to change the value of a changed value
        if self.chilled.has_key(symbol):
            self.__bindsymbol(symbol, value, source)	# record for debugging
            raise "UNSATISFIABLE"
        # Membership in chilled means we should treat the binding as frozen for
        # simplification purposes.  It has to be turned off when the current
        # call to set_symbol is done; otherwise side effects from inclusion
        # sequences would collide with each other.
        self.chilled[symbol] = 1
        # Here's where the value actually gets set
        self.__bindsymbol(symbol, value, source)
        # Make the side-effects of this symbol's previous bindings
        # temporarily invisible while computing side effects.  This
        # is necessary because things like dependent suppressions
        # need to be calculated according to the effects they would
        # have when the old bindings are removed, as they will be
        # if this change is committed (whew!)
        self.__bindconceal(symbol)
        # If this symbol was in a choices and is being set true,
        # then unset and chill all its siblings.
        if symbol.menu.type == "choices" and value:
            symbol.menu.menuvalue = symbol
            for sibling in symbol.menu.items:
                if sibling != symbol:
                    self.__set_symbol_internal(sibling, cml.n, source)
        # Other side effects...
        if self.trit_tie and symbol == self.trit_tie and value == cml.n:
            for entry in self.dictionary.values():
                if entry.type == "trit" and not entry.is_derived() and entry.eval() == cml.m:
                    self.__set_symbol_internal(entry, cml.y, source)
        # Now propagate the value change through ancestry chains
        if symbol.is_logical():
            if value > cml.n:
                for ancestor in symbol.ancestors + symbol.menu.ancestors:
                    self.__dep_force_ancestors(symbol, value, ancestor)
            elif value < cml.y:
                for dependent in symbol.dependents:
                    self.__dep_force_dependents(symbol, value, dependent)
        # Perhaps we can deduce other values through explicit constraints?
        # This is where we'd plug in a full SAT algorithm if we were going
        # to use one.
        if not self.inclusion and symbol.visibility:
            self.__bind_from_constraint(symbol.visibility, source)
        # Now loop through the constraints associated with this
        # symbol, simplifying out assigned variables and trying to
        # freeze more variables each time.  The outer loop guarantees
        # that as long as the constraints imply at least one more
        # tentative setting, we'll keep going.
        while 1:
            cc = 0
            for wff in symbol.constraints:
                cc += self.__bind_from_constraint(wff, symbol)
            if not cc:
                break;
        # OK, now make the old bindings of this symbol visible again
        # (the change we just made might get rolled back later).
        self.__bindreveal(symbol)

    def value_from_string(self, sym, val):
        "Set symbol from string according to the symbol type."
        if sym.is_logical():
            if val == "y":
                val = cml.y
            elif val == "m":
                val = cml.m
            elif val == "n":
                val = cml.n
        elif sym.type == "decimal":
            val = int(val)
        elif sym.type == "hexadecimal":
            val = long(val, 16)
        return val

    def __eval_frozen(self, wff):
        "Test whether a given expr is entirely constant, chilled or frozen."
        if isinstance(wff,cml.trit) or type(wff) in (type(0),type(0L),type("")):
            return wff
        elif isinstance(wff, cml.ConfigSymbol):
            if wff.frozen or self.chilled.has_key(wff):
                return wff.eval()
            elif wff.is_derived():
                return self.__eval_frozen(wff.default)
            else:
                return None
        elif wff[0] == 'not':
            below = self.__eval_frozen(wff[1])
            if below == None:
                return None
            else:
                return cml.trit(not below)
        elif wff[0] == '?':
            guard = self.__eval_frozen(wff[1])
            if guard == None:
                return None
            if guard:
                return self.__eval_frozen(wff[2])
            else:
                return self.__eval_frozen(wff[3])
        else:
            left = self.__eval_frozen(wff[1])
            right = self.__eval_frozen(wff[2])
            if left != None and right != None:
                return cml.evaluate((wff[0], left, right))
            elif left == None and right == None:
                return None
            # OK, now the grotty part starts
            elif wff[0] == 'and':
                if left == cml.n or right == cml.n:
                    return cml.n
                elif left in (cml.y, cml.m):
                    return right
                elif right in (cml.y, cml.m):
                    return left
                else:
                    return None
            elif wff[0] == 'or':
                if left in (cml.y, cml.m) or right in (cml.y, cml.m):
                    return cml.y
                elif left == cml.n:
                    return right
                elif right == cml.n:
                    return left
                else:
                    return None
            elif wff[0] == 'implies':
                if left in (cml.y, cml.m):
                    return right
                elif left == cml.n:
                    return cml.y
                elif right == cml.n:
                    return not left
                else:
                    return None
            else:
                return None

    def __bind_from_constraint(self, wff, source):
        "Set symbols based on asserted equalities or inequalities."
        self.debug_emit(2, self.lang["BINDING"] % (cml.display_expression(wff), `source`))
        if not wff or self.__eval_frozen(wff) != None:
            return 0
        elif isinstance(wff, cml.ConfigSymbol):
            return 0
        op = wff[0]
        left = wff[1]
        right = wff[2]
        if op == '?':
            guard = self.__eval_frozen(left)
            if guard == None:
                return 0
            elif guard:
                return self.__bind_from_constraint(right, source)
            else:
                return self.__bind_from_constraint(wff[3], source)
        elif op == 'implies':
            if self.__eval_frozen(left):
                return self.__bind_from_constraint(right, source)
            elif self.__eval_frozen(right) == cml.n:
                return self.__bind_from_constraint(('==',right,cml.n), source)
        elif op == 'and':
            return self.__bind_from_constraint(left, source) or self.__bind_from_constraint(right, source)
        elif op == '<':
            return self.__bind_from_constraint(('>=', right, left), source)
        elif op == '>':
            return self.__bind_from_constraint(('<=', right, left), source)
        else:
            # Before we can force a binding, we need exactly one operand
            # to be mutable...
            left_mutable = self.is_mutable(left) and not self.chilled.has_key(left)
            right_mutable = self.is_mutable(right) and not self.chilled.has_key(right)
            if left_mutable == right_mutable:
                self.debug_emit(3, "0 or 2 mutables in %s, %s" % (left, right)) 
                return 0
            # We need to keep the recursion from spinning forever.
            leftval = cml.evaluate(left)
            if isinstance(left, cml.ConfigSymbol):
                leftnonnull = left.eval()
            else:
                leftnonnull = leftval
            rightval = cml.evaluate(right)
            if isinstance(right, cml.ConfigSymbol):
                rightnonnull = right.eval()
            else:
                rightnonnull = rightval
            # Now we may have the conditions to force a binding
            if op == '==':
                if left_mutable:
                    if leftnonnull == rightval:
                        self.debug_emit(2, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(left, rightval, source)
                        return 1
                elif right_mutable:
                    if rightnonnull == leftval:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(right, leftval, source)
                        return 1
            elif op == '<=':
                if left_mutable and rightval == cml.n:
                    if leftnonnull == cml.n:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(left, cml.n, source)
                        return 1
                if right_mutable and leftval == cml.y:
                    if rightnonnull == cml.y:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(right, cml.y, source)
                        return 1
            elif op == '>=':
                if left_mutable and rightval == cml.y:
                    if leftnonnull == cml.y:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(left, cml.y, source)
                        return 1
                if right_mutable and leftval == cml.y:
                    if rightnonnull == cml.y:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(right, cml.y, source)
                        return 1
            elif op == '!=':
                if left_mutable and rightval == cml.y and (left.type == "bool" or not self.trits_enabled):
                    if leftnonnull == cml.n:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(left, cml.n, source)
                        return 1
                if right_mutable and left == cml.y and (right.type == "bool" or not self.trits_enabled):
                    if rightnonnull == cml.n:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(right, cml.n, source)
                        return 1
                if left_mutable and rightval == cml.n and (left.type == "bool" or not self.trits_enabled):
                    if leftnonnull == cml.y:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(left, cml.y, source)
                        return 1
                if right_mutable and leftval == cml.n and (right.type == "bool" or not self.trits_enabled):
                    if rightnonnull == cml.y:
                        self.debug_emit(1, self.lang["REDUNDANT"] % (wff,))
                        return 0
                    else:
                        self.__set_symbol_internal(right, cml.y, source)
                        return 1
        return 0

    #
    # Navigation helpers 
    #

    def visit(self, entry):
        "Register the fact that we've visited a menu."
        #self.debug_emit(1, "Visiting %s (%s) starts"%(entry.name,entry.type))
        if not hasattr(entry, "visits"):
            entry.visits = 1
        else:
            entry.visits = entry.visits + 1
        # Set choices defaults -- do it now for the side-effects
        # (If you do it sooner you can get weird constraint failures)
        if entry.visits == 1 and entry.type == "choices" and not entry.frozen:
            if entry.default:
                self.set_symbol(entry.default, cml.y)
        #self.debug_emit(1, "Visiting %s (%s) ends" % (entry.name,entry.type))

    def next_node(self, here):
        "Return the next menu or symbol in depth-first order."
        if here.type == 'menu' and not here.default:
            here = here.items[0];
        else:
            while here.menu:
                up = here.menu
                where = up.items.index(here)
                if where >= len(here.menu.items) - 1:
                    here = up
                else:
                    here = up.items[where+1]
                    break
        if here == None:
            here = self.start
        return here

    def previous_node(self, here):
        "Return the previous menu or symbol in depth-first order."
        if here.type == 'menu' and not here.default:
            here = here.items[0];
        else:
            while here.menu:
                up = here.menu
                where = up.items.index(here)
                if where == 0:
                    here = up
                else:
                    here = up.items[where-1]
                    break
        if here == None:
            here = self.start
        return here

    def search(self, pattern, hook):
        "Return a menu composed of symbols matching a given regexp."
        regexp = re.compile(pattern)
        hits = cml.ConfigSymbol("search", "menu")
        for entry in self.dictionary.values():
            if entry.prompt and entry.type != "message":
                text = hook(entry)
                if text == None:
                    continue
                elif regexp.search(text):
                    hits.items.append(entry)
        # Give the result menu a parent only if all members have same parent
        hits.menu = None
        hits.frozen = 0
        for symbol in hits.items:
            if hits.menu == None:
                hits.menu = symbol.menu
            elif symbol.menu != hits.menu:
                hits.menu = None
                break
        return hits

    def symbolsearch(self, pattern):
        "Return a menu composed of symbols matching a given regexp."
        return self.search(pattern, lambda x: x.name + x.prompt)

    def helpsearch(self, pattern):
        "Return a menu composed of symbols matching a given regexp."
        return self.search(pattern, lambda x: x.help())

    # Input validation

    def range_check(self, symbol, value):
        "Check whether a value is within a symbol's specified valid range."
        if not symbol.range:
            return 1
        elif symbol.enum:
            for (label, possibility) in symbol.range:
                if value == possibility:
                    return 1
        else:
            for span in symbol.range:
                if type(span) is type(0):
                    if value == span:
                        return 1
                elif value >= span[0] and value <= span[1]:
                    return 1
        return 0

# End
