/*
 * bag.c
 * Generic bag
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "bag.h"
#include "gc.h"

#define	FREE	((void*)0)
#define	DELETED	((void*)-1)

static
void
initmem(bag* bg)
{
	void* mem;

	mem = gc_malloc_fixed_typed((2 * sizeof(void*) + sizeof(int*)) * bg->size, bg->memtype);
	bg->keys = mem;
	bg->values = (void*)(((uintp)mem) + sizeof(void*) * bg->size);
	bg->refcounts = (void*)(((uintp)mem) + 2 * sizeof(void*) * bg->size);
}

void
bagStart(bag* bg, int size, int pressure, hashfunc hash, equalfunc equals, int memtype)
{
	bg->size = size;
	bg->count = 0;
	bg->pressure = pressure;
	bg->memtype = memtype;
	bg->hashcode = hash;
	bg->equals = equals;
	initmem(bg);
}

void
bagFinish(bag* bg)
{
	gc_free_fixed(bg->keys);
}

void
bagRehash(bag* bg)
{
	void** okeys;
	void** ovalues;
	int* orefs;
	int osize;
	void* key;
	void** ptr;
	void** end;
	int i;

	if (bg->count * 100 < bg->size * bg->pressure) {
		return;
	}

	okeys = bg->keys;
	ovalues = bg->values;
	orefs = bg->refcounts;
	osize = bg->size;

	bg->size *= 2;
	initmem(bg);

	for (i = 0; i < osize; i++) {
		key = okeys[i];
		if (key != FREE && key != DELETED) {
			end = &bg->keys[(*bg->hashcode)(key) % bg->size];
			ptr = end;
			for (;;) {
				if (*ptr == FREE) {
					*ptr = key;
					bg->values[ptr - bg->keys] = ovalues[i];
					bg->refcounts[ptr - bg->keys] = orefs[i];
					break;
				}
				ptr++;
				if (ptr == &bg->keys[bg->size]) {
					ptr = &bg->keys[0];
				}
				if (ptr == end) {
					assert("bag out of space" == 0);
					break;
				}
			}
		}
	}

	gc_free_fixed(okeys);
}

void*
bagPut(bag* bg, void* key, void* value)
{
	void** deleted;
	void** ptr;
	void** end;

	deleted = 0;
	end = &bg->keys[(*bg->hashcode)(key) % bg->size];
	ptr = end;
	for (;;) {
		if (*ptr == FREE) {
			if (deleted == 0) {
				deleted = ptr;
			}
			break;
		}
		else if (*ptr == DELETED) {
			if (deleted == 0) {
				deleted = ptr;
			}
		}
		else if ((*bg->equals)(*ptr, key) == 0) {
			bg->refcounts[ptr - bg->keys]++;
			return (bg->values[ptr - bg->keys]);
		}

		ptr++;
		if (ptr == &bg->keys[bg->size]) {
			ptr = &bg->keys[0];
		}
		if (ptr == end) {
			break;
		}
	}
	if (deleted != 0) {
		*deleted = key;
		bg->values[deleted - bg->keys] = value;
		bg->refcounts[deleted - bg->keys] = 1;
		bg->count++;
		return (value);
	}

	return (0);
}

void*
bagGet(bag* bg, void* key)
{
	void** ptr;
	void** end;

	end = &bg->keys[(*bg->hashcode)(key) % bg->size];
	ptr = end;
	for (;;) {
		if (*ptr == FREE) {
			break;
		}
		else if (*ptr == DELETED) {
			/* Ignore */
		}
		else if ((*bg->equals)(*ptr, key) == 0) {
			return (bg->values[ptr - bg->keys]);
		}

		ptr++;
		if (ptr == &bg->keys[bg->size]) {
			ptr = &bg->keys[0];
		}
		if (ptr == end) {
			break;
		}
	}
	return (0);
}

void
bagRemove(bag* bg, void* key)
{
	void** ptr;
	void** end;

	end = &bg->keys[(*bg->hashcode)(key) % bg->size];
	ptr = end;
	for (;;) {
		if (*ptr == FREE) {
			break;
		}
		else if (*ptr == DELETED) {
			/* Ignore */
		}
		else if ((*bg->equals)(*ptr, key) == 0) {
			if (--bg->refcounts[ptr - bg->keys] == 0) {
				*ptr = DELETED;
				bg->count--;
			}
			break;
		}

		ptr++;
		if (ptr == &bg->keys[bg->size]) {
			ptr = &bg->keys[0];
		}
		if (ptr == end) {
			break;
		}
	}
}
