/*
 * classMethod.c
 * Dictionary of classes, methods and fields.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	EDBG(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "md.h"
#include "slots.h"
#include "access.h"
#include "objects.h"
#include "classMethod.h"
#include "code.h"
#include "file.h"
#include "readClass.h"
#include "baseClasses.h"
#include "threads.h"
#include "itypes.h"
#include "exception.h"
#include "external.h"
#include "lookup.h"
#include "support.h"
#include "gc.h"
#include "locks.h"
#include "engine.h"
#include "soft.h"
#include "jni.h"
#include "jnidefs.h"
#include "strings.h"
#include "jdebug.h"
#include "precompile.h"
#include "patch.h"
#include "bag.h"

iLock* classHashLock;
extern iLock* translatorlock;

typedef Hjava_lang_Class* (*prelookup)(char* name);
static prelookup prelist[32];

Hjava_lang_Class* ArrayInterfaces[2];
Hjava_lang_Class* ClassLoaderClass;
Hjava_lang_Class* ReferenceClass;

/*
 * We need direct access to various dispatch tables.
 * These are defined here including their sizes.  These sizes may have to be
 * changed it the java classes change (this is asserted for).
 */
DEFINE_DISPATCH_TABLE(KaffeDispatchTable_CharArray, 5);
DEFINE_DISPATCH_TABLE(KaffeDispatchTable_java_lang_Class, 5);
DEFINE_DISPATCH_TABLE(KaffeDispatchTable_java_lang_String, 6);

extern jboolean verify2(Hjava_lang_Class*);
extern void verify3(Hjava_lang_Class*);
extern void Kaffe_ExceptionClear(JNIEnv*);
extern void Kaffe_CallStaticVoidMethodA(JNIEnv*, jclass, jmethodID, jvalue*);
extern void error_stub(void);

static void internalSetupClass(Hjava_lang_Class*, char*, int, int, Hjava_lang_ClassLoader*);

static void allocStaticFields(Hjava_lang_Class*);
static void resolveObjectFields(Hjava_lang_Class*);
static void resolveStaticFields(Hjava_lang_Class*);
static void buildDispatchTable(Hjava_lang_Class*);
static void buildInterfaces(Hjava_lang_Class*);
static Hjava_lang_Class* lookupPrecompiledClass(char*, Hjava_lang_ClassLoader*);

#if !defined(ALIGNMENT_OF_SIZE)
#define	ALIGNMENT_OF_SIZE(S)	(S)
#endif

int
processClass(Hjava_lang_Class* class, int tostate)
{
	int i;
	Method* meth;
	Hjava_lang_Class* nclass;
	static iLock* classLock;
	int ret;
	Hjava_lang_Thread* tid;
	int iLockRoot;

	assert(tostate == CSTATE_UNLOADED || tostate == CSTATE_COMPLETE || tostate == CSTATE_LINKED);

	/* If we're trying to complete this and it already failed, bail */
	if (CLASS_STATE(class) == CSTATE_FAILED) {
		goto errorNoDefNoLock;
	}
	/* If this class is initialised to the required point, quit now */
	if (CLASS_STATE(class) >= tostate) {
		return (1);
	}

	/* If we're not loading then we're doing nothing */
	if (tostate == CSTATE_UNLOADED) {
		return (1);
	}

	/* If we have no constant pool then we have no class - load it */
	if (class->constants.size == 0) {
		assert (class->loader == 0);
		if (findClass(class) == false) {
			lockStaticMutex(&classLock);
			goto errorNoDef;
		}
	}

	/* For the moment we only allow one thread to initialise any classes
	 * at once.  This is because we've got circular class dependencies
	 * we've got to work out.
	 */
	lockStaticMutex(&classLock);

	/* If class has no static data then we we've not initialized the
	 * class yet.
	 */
	if (class->sdata == 0) {
		/* We handle array initialisation a little specially */
		if (CLASS_IS_ARRAY(class)) {
			if (ObjectClass == 0) {
				ObjectClass = loadClass("java.lang.Object", 0, CSTATE_LINKED);
			}
			class->superclass = ObjectClass;
                        GC_WRITE_SAFE(class, ObjectClass);
			if (ArrayInterfaces[0] == 0) {
				if (SerialClass == 0) {
					SerialClass = loadClass("java.io.Serializable", 0, CSTATE_LINKED);
				}
				if (CloneClass == 0) {
					CloneClass = loadClass("java.lang.Cloneable", 0, CSTATE_LINKED);
				}
				ArrayInterfaces[0] = SerialClass;
				ArrayInterfaces[1] = CloneClass;
			}
			addInterfaces(class, 2, ArrayInterfaces);
			class->total_interface_len = 2;
			buildDispatchTable(class);
			CLASS_STATE(class) = CSTATE_COMPLETE;
			unlockStaticMutex(&classLock);
			return (1);
		}

		/* Load and link the super class */
		if (class->superclass != 0) {
			nclass = getClass((constIndex)(int)class->superclass, class, 0);
			if (nclass == 0) {
				goto errorNoDef;
			}
			ret = processClass(nclass, CSTATE_LINKED);
			if (ret == 0) {
				goto errorSuperNoDef;
			}
			class->superclass = nclass;
                        GC_WRITE_SAFE(class, nclass); /* class->superclass */
			CLASS_FSIZE(class) = CLASS_FSIZE(class->superclass);
		}

		/* Allocate any static space required by class and initialise
		 * the space with any constant values.  If the space is pre-allocated
		 * then we don't need to do this.
		 */
		allocStaticFields(class);

		buildInterfaces(class);

		resolveObjectFields(class);
		resolveStaticFields(class);

		/* Build dispatch table */
		buildDispatchTable(class);

		/* Second stage verification - check the class format is okay */
		if (verify2(class) == true) {
			CLASS_STATE(class) = CSTATE_LINKED;
			unlockStaticMutex(&classLock);
			postException("java.lang.ClassFormatError", CLASS_CNAME(class));
			return (0);
		}
		/* Third stage verification - check the bytecode is okay */
		verify3(class);

		/* Note any interesting features about this class */
                if (ThreadClass != 0 && instanceof_class(ThreadClass, class)) {
                        SET_GC_TYPE( class, GC_ALLOC_THREAD );
                }
                else if (StringClass != 0 && instanceof_class(StringClass, class)) {
                        SET_GC_TYPE( class, GC_ALLOC_STRING );
                }
                else if (ReferenceClass != 0 && instanceof_class(ReferenceClass, class)) {
                        SET_GC_TYPE( class, GC_ALLOC_JREFERENCE );
                }
                else if (ClassLoaderClass != 0 && instanceof_class(ClassLoaderClass, class) ) {
                        SET_GC_TYPE( class, GC_ALLOC_CLASSLOADER );
                }
                else{
                        SET_GC_TYPE( class, GC_ALLOC_NORMALOBJECT );
                }

		/* Cache finalizer method */
		class->finalizer = findMethod(class, final_name, void_sig);
		assert(class->finalizer != 0);
		if (ObjectClass != 0 && ObjectClass != class && class->finalizer == ObjectClass->finalizer) {
			class->finalizer = 0;
			if( CLASS_GC_TYPE( class ) == GC_ALLOC_NORMALOBJECT ){
				CLASS_GC_TYPE( class ) = GC_ALLOC_NORMALOBJECT_NOFINALIZE;
			}
		}

		/* And note that it's done */
		CLASS_STATE(class) = CSTATE_LINKED;

		/* If that's all then quit now */
		if (tostate == CSTATE_LINKED) {
			unlockStaticMutex(&classLock);
			return (1);
		}
	}

	else if (CLASS_IS_PRECOMPILED(class)) {
		if (CLASS_STATE(class) < CSTATE_LINKED) {
			/* Setup superclass */
			if (class->superclass != 0) {
				class->superclass = loadClass((char*)class->superclass, class->loader, CSTATE_LINKED);
			}

			buildInterfaces(class);

			/* Make sure the static fields are known to the GC */
			for (i = CLASS_NSFIELDS(class) - 1; i >= 0; i--) {
				void** ref = (void**)FIELD_ADDRESS(&CLASS_SFIELDS(class)[i]);
				gc_add_ref(ref, ref+1);
			}

			buildDispatchTable(class);

			/* Cache finalizer method */
			class->finalizer = findMethod(class, final_name, void_sig);
			assert(class->finalizer != 0);
			if (ObjectClass != 0 && ObjectClass != class && class->finalizer == ObjectClass->finalizer) {
				class->finalizer = 0;
                                if( CLASS_GC_TYPE( class ) == GC_ALLOC_NORMALOBJECT ){
                                        CLASS_GC_TYPE( class ) = GC_ALLOC_NORMALOBJECT_NOFINALIZE;
                                }
			}
			/* And note that it's done */
			CLASS_STATE(class) = CSTATE_LINKED;
		}

		/* If that's all then quit now */
		if (tostate == CSTATE_LINKED) {
			unlockStaticMutex(&classLock);
			return (1);
		}
	}

        /* If class initialization failed before then it fails again */
        if (CLASS_STATE(class) == CSTATE_FAILED) {
		goto errorNoDef;
        }

	/* If init is in progress return.  This must be the same
	 * thread because we lock the class when we come in here.
	 */
	if (CLASS_STATE(class) >= CSTATE_DOING_CONSTINIT) {
		if (class->owner != Kaffe_TI_currentJava()) {
			while(CLASS_STATE(class) < tostate) {
				waitStaticCond(&classLock, NOTIMEOUT);
			}
		}
		if (CLASS_STATE(class) == CSTATE_FAILED) {
			goto errorNoDef;
		}
		unlockStaticMutex(&classLock);
		return (1);
	}
	CLASS_STATE(class) = CSTATE_DOING_CONSTINIT;
	tid = Kaffe_TI_currentJava();
	class->owner = tid;
        
	/* We release the class lock now because it should be safe
	 * to finish the class init without it - and it we hold
	 * onto it we get problems with <clinit> methods.
	 */

	unlockStaticMutex(&classLock);

	nclass = class->superclass;
	if (nclass != 0) {
		ret = processClass(nclass, CSTATE_COMPLETE);
		if (ret == 0) {
			lockStaticMutex(&classLock);
			goto errorSuperNoDef;
		}
	}

	/* Register precompiled methods */
	if (CLASS_IS_PRECOMPILED(class)) {
		lockStaticMutex(&translatorlock);
		for (i = CLASS_NMETHODS(class), meth = CLASS_METHODS(class); i > 0; i--, meth++) {
			makeMethodActive(meth, class, meth->ncode, meth->ncodee);
		}
		unlockStaticMutex(&translatorlock);
	}

#if defined(MEM2)
        /* Get pointer to object freelist */
        if (CLASS_FSIZE(class) < MIN_BLOB_SIZE)
                class->freelist = (void*)&free_heap_obj[SZ2FREELIST(CLASS_FSIZE(class))];
#endif

	DBG( vm_class, ("Initialising %s static %d\n", class->name,
					CLASS_FSIZE(class)));

	meth = findMethodLocal(class, init_name, void_sig);
	if (meth != NULL) {
		clearPostedException();
EDBG(		kprintf("Calling <clinit>: %s\n", CLASS_CNAME(class)); )
		Kaffe_CallStaticVoidMethodA(0, (jclass)class, (jmethodID)meth, 0);
		if (unhand(tid)->pending != 0) {
			lockStaticMutex(&classLock);
			CLASS_STATE(class) = CSTATE_FAILED;
			broadcastStaticCond(&classLock);
			unlockStaticMutex(&classLock);
EDBG(			kprintf("Exception in <clinit>: %s (%s)\n", CLASS_CNAME(class), CLASS_CNAME(OBJECT_CLASS(&getPostedException()->base))); )
			postExceptionInException("java.lang.ExceptionInInitializerError");
			return (0);
		}
	}

	lockStaticMutex(&classLock);
	CLASS_STATE(class) = CSTATE_COMPLETE;
	broadcastStaticCond(&classLock);
	unlockStaticMutex(&classLock);

	return (1);

errorNoDef:;
	broadcastStaticCond(&classLock);
	unlockStaticMutex(&classLock);
errorNoDefNoLock:;
	postException("java.lang.NoClassDefFoundError", CLASS_CNAME(class));
	return (0);
errorSuperNoDef:;
	broadcastStaticCond(&classLock);
	unlockStaticMutex(&classLock);
	CLASS_STATE(class) = CSTATE_FAILED;
	postException("java.lang.NoClassDefFoundError", CLASS_CNAME(nclass));
	return (0);
}

Hjava_lang_Class*
setupClass(Hjava_lang_Class* cl, constIndex c, constIndex s, u2 flags, Hjava_lang_ClassLoader* loader)
{
	constants* pool;
	char* name;
	Hjava_lang_Class* cl2;

	pool = CLASS_CONSTANTS(cl);

	/* Find the name of the class */
	if (pool->tags[c] != CONSTANT_Class) {
		DBG( vm_class, ("setupClass: not a class.\n"));
		postException("java.lang.ClassFormatError", "Bad format");
		return (0);
	}

	UTF_REPLACE(name, WORD2UTF(pool->data[c]), '/', '.');
	cl->name = addUtfCopy(name, -1);
	cl->loader = loader;

	cl2 = addClass(cl);
	if (cl2 != cl) {
		char* mess;
		mess = alloca(20 + strlen(name));
		strcpy(mess, "Duplicate name: ");
		strcat(mess, name);
		postException("java.lang.ClassFormatError", addUtfCopy(mess, -1));
		return (0);
	}
	internalSetupClass(cl, cl->name, flags, s, loader);

	return (cl);
}

static
void
internalSetupClass(Hjava_lang_Class* cl, char* name, int flags, int su, Hjava_lang_ClassLoader* loader)
{
	CLASS_CNAME(cl) = name;
	cl->loader = loader;
        GC_WRITE_SAFE(cl, loader); /* class->loader */
	CLASS_METHODS(cl) = NULL;
	CLASS_NMETHODS(cl) = 0;
	cl->superclass = (Hjava_lang_Class*)(uintp)su;
        GC_WRITE_SAFE(cl, (void*)su); /* class->superuser */
	cl->msize = 0;
	CLASS_FIELDS(cl) = NULL;
	CLASS_FSIZE(cl) = 0;
	cl->accflags = flags;
	cl->dtable = NULL;
	cl->total_interface_len = 0;
	cl->interface_len = 0;
        cl->interfaces = NULL;
	/* cl->state = CSTATE_LOADED; */
	cl->sdata = NULL;
	/*
	 * Deal with any special class flags we need to set.
	 */
	if (strcmp(name, "java.lang.ClassLoader") == 0) {
		ClassLoaderClass = cl;
                CLASS_GC_TYPE( cl ) = GC_ALLOC_CLASSLOADER;
	}
	else if (strcmp(name, "java.lang.ref/Reference") == 0) {
		ReferenceClass = cl;
                CLASS_GC_TYPE( cl ) = GC_ALLOC_JREFERENCE;
	}
	else {
#if defined(MEM2)
		cl->gc_type = GC_ALLOC_NORMAL;
#else
		cl->flags = 0;
#endif
	}
#if defined(MEM1) && 0
        cl->refs = NULL;
        cl->nrefs = 0;
#elif defined(MEM2)
        cl->refs = NULL;
        cl->nrefs = 0;
        cl->freelist = NULL;
#endif
        cl->finalizer = NULL;
}

Method*
addMethod(Hjava_lang_Class* c, method_info* m)
{
	constIndex nc;
	constIndex sc;
	Method* mt;
	constants* pool;
	char* name;
	char* signature;

	pool = CLASS_CONSTANTS (c);

	nc = m->name_index;
	if (pool->tags[nc] != CONSTANT_Utf8) {
		DBG( vm_class, ("addMethod: no method name.\n"));
		return (0);
	}
	sc = m->signature_index;
	if (pool->tags[sc] != CONSTANT_Utf8) {
		DBG( vm_class, ("addMethod: no signature name.\n"));
		return (0);
	}
	name = WORD2UTF (pool->data[nc]);
	signature = WORD2UTF (pool->data[sc]);

	DBG( vm_class, ("Adding method %s:%s%s (%x)\n", c->name, WORD2UTF(pool->data[nc]), WORD2UTF(pool->data[sc]), m->access_flags));

	mt = &c->methods[c->nmethods++];
	mt->name = addUtf(name);
        GC_WRITE_SAFE(c, name); /* class->method->name */
	mt->signature = addUtf(signature);
	mt->accflags = m->access_flags;
	BCODE(mt) = 0;
	mt->exception_table = 0;
	mt->exception_declared = 0;
	mt->idx = -1;

	/* Mark constructors as such */
	if (strcmp (name, constructor_name) == 0) {
		mt->accflags |= ACC_CONSTRUCTOR;
	}

	return (mt);
}

Field*
addField(Hjava_lang_Class* c, field_info* f)
{
	constIndex nc;
	constIndex sc;
	Field* ft;
	constants* pool;
	int index;
	char* sig;

	pool = CLASS_CONSTANTS (c);

	nc = f->name_index;
	if (pool->tags[nc] != CONSTANT_Utf8) {
		DBG( vm_class, ("addField: no field name.\n"));
		return (0);
	}

	--CLASS_FSIZE(c);
	if (f->access_flags & ACC_STATIC) {
		index = CLASS_NSFIELDS(c)++;
	}
	else {
		index = CLASS_FSIZE(c) + CLASS_NSFIELDS(c);
	}
	ft = &CLASS_FIELDS(c)[index];

	DBG( vm_class, ("Adding field %s:%s\n", c->name, WORD2UTF(pool->data[nc])));

	sc = f->signature_index;
	if (pool->tags[sc] != CONSTANT_Utf8) {
		DBG( vm_class, ("addField: no signature name.\n"));
		return (0);
	}
	ft->name = addUtf(WORD2UTF(pool->data[nc]));
        GC_WRITE_SAFE(c, ft->name);
	ft->accflags = f->access_flags;

	sig = WORD2UTF(pool->data[sc]);
	ft->ftype = addUtf(sig);
	GC_WRITE_SAFE(c, ft->ftype);
	switch (sig[0]) {
	case 'L':
	case '[':
		FIELD_SIZE(ft) = PTR_TYPE_SIZE;
		break;
	case 'B':
	case 'Z':
		FIELD_SIZE(ft) = 1;
		break;
	case 'S':
	case 'C':
		FIELD_SIZE(ft) = 2;
		break;
	case 'I':
	case 'F':
		FIELD_SIZE(ft) = 4;
		break;
	case 'J':
	case 'D':
		FIELD_SIZE(ft) = 8;
		break;
	}

	return (ft);
}

/*
 * Add class's filename
 */
void
addSourceFilename(Hjava_lang_Class* cls, int idx)
{
#if !defined(OPTIMIZE_MEMORY) || defined(DEBUG)
	cls->filename = WORD2UTF(CLASS_CONSTANTS(cls)->data[idx]);
#endif
}

void
setFieldValue(Field* ft, u2 idx)
{
	/* Set value index */
	FIELD_CONSTIDX(ft) = idx;
	ft->accflags |= FIELD_CONSTANT_VALUE;
}

void
finishFields(Hjava_lang_Class *cl)
{
	Field tmp;
	Field* fld;
	int n;

	/* Reverse the instance fields, so they are in "proper" order. */
	fld = CLASS_IFIELDS(cl);
	n = CLASS_NIFIELDS(cl);
	while (n > 1) {
		tmp = fld[0];
		fld[0] = fld[n-1];
		fld[n-1] = tmp;
		fld++;
		n -= 2;
	}
}

void
addInterfaces(Hjava_lang_Class* c, int inr, Hjava_lang_Class** inf)
{
	assert(inr > 0);

        c->interfaces = inf;
	c->interface_len = inr;
	c->total_interface_len = 0;
#if defined(GC_INCREMENTAL)
        for (; --inr >= 0;) {
                GC_WRITE(c, c->interfaces[inr]); /* class->interfaces[i] */
        }
#endif
}

/*
 * Lookup a named class, loading it if necessary.
 * The name is like "java.lang.String".
 */
Hjava_lang_Class*
loadClass(char* name, Hjava_lang_ClassLoader* loader, int state)
{
	Hjava_lang_Class* clazz;
	Hjava_lang_Class* clazz2;
	clazz = lookupClass(name, loader);
	if (clazz == 0) {
		clazz = lookupPrecompiledClass(name, loader);
	}
	if (clazz == 0) {
		if (name[0] == '[') {
			clazz = lookupArrayProcess(getClassFromSignature(&name[1], loader), state);
		}
		else if (loader == 0) {
			clazz = newClass();
			if (clazz == 0) {
				return (0);
			}
			clazz->name = addUtfCopy(name, -1);
			clazz->loader = loader;
                        GC_WRITE( clazz, loader ); /* class->loader */
		}
		else {
			Hjava_lang_String* str;

			str = makeReplaceJavaStringFromUtf(name, -1, 0, 0);
			clazz = (Hjava_lang_Class*)Kaffe_CallObjectMethod(0, loader, loadClassMethod, str, true);
			if (clazz == 0 || strcmp(CLASS_CNAME(clazz), name) != 0) {
				/* If there's no posted exception then
				 * we post NotFound.
				 */
				if (getPostedException() == 0) {
					postException("java.lang.NoClassDefFoundError", name);
				}
				return (0);
			}
		}

		/* Add class to class pool, but it already added by someone
		 * else discard and use the older one.
		 */
		clazz2 = addClass(clazz);
		if (clazz2 != clazz) {
			clazz = clazz2;
		}
	}

	if (processClass(clazz, state) == 0) {
		return (0);
	}

	return (clazz);
}

/*
 * Lookup a method given class, name and signature.
 */
Method*
lookupClassMethod(Hjava_lang_Class* cls, char* name, char* sig)
{
	Method* meth;

	processClass(cls, CSTATE_COMPLETE);
	meth = findMethod(cls, name, sig);

	return (meth);
}

Hjava_lang_Class*
resolveFieldType(Field *fld, Hjava_lang_Class* this)
{
	char* name = fld->ftype;
	return (classFromSigProcess(&name, this->loader, CSTATE_UNLOADED));
}

static
void
resolveObjectFields(Hjava_lang_Class* class)
{
	int fsize;
	int align;
	Field* fld;
	int n;
	int offset;
	Hjava_lang_Class* super;
	unsigned int ix;
	int malign;

	super = class->superclass;
	if (super != 0) {
		malign = super->alignment;
	}
	else {
		malign = 1;
	}
        ix = 0;

	/* Find start of new fields in this object.  If start is zero, we must
	 * allow for the object headers.
	 */
	offset = CLASS_FSIZE(class);
	if (offset == 0) {
		offset = sizeof(Hjava_lang_Object);
		/* This must be true for alignment to work correctly */
		assert(offset == sizeof(jlong));
	}

	/* Walk the object fields and count up how many are references.
	 * We also work out the max alignment for this object.
	 */
	fld = CLASS_IFIELDS(class);
	for (n = CLASS_NIFIELDS(class); n > 0; n--, fld++) {
		if (FIELD_ISREF(fld)) {
			fld->accflags |= FIELD_REFERENCE;
#if defined(MEM1) && 0
			class->nrefs++;
#elif defined(MEM2)
			class->nrefs++;
#endif
		}
		align = ALIGNMENT_OF_SIZE(FIELD_SIZE(fld));
		if (align > malign) {
			malign = align;
		}
        }

	/* Save the maximum alignment */
	class->alignment = malign;

#if (defined(MEM1) && 0) || defined(MEM2)
        /* Allocate an array used to hold all offsets to references
         * within the object.
         */
        if (super != 0) {
                class->nrefs += super->nrefs;
		ix = super->nrefs;
	}
	if (class->nrefs > 0) {
		class->refs = gc_malloc_fixed(sizeof(int) * class->nrefs);
		if (ix > 0) {
			memcpy(class->refs, super->refs, sizeof(int) * ix);
		}
	}
#endif

	/* Move start of new class files so they're aligned */
	offset = ((offset + malign - 1) / malign) * malign;

	/* Now work out where to put each field */
	fld = CLASS_IFIELDS(class);
	n = CLASS_NIFIELDS(class);
	for (; --n >= 0; fld++) {
		fsize = FIELD_SIZE(fld);
		/* Align field */
		align = ALIGNMENT_OF_SIZE(fsize);
		offset = ((offset + align - 1) / align) * align;
		FIELD_BOFFSET(fld) = offset;
#if (defined(MEM1) && 0) || defined(MEM2)
                if (fld->accflags & FIELD_REFERENCE) {
                        class->refs[ix++] = offset;
		}
#endif
		offset += fsize;
	}

	CLASS_FSIZE(class) = offset;
}

/*
 * Allocate the space for the static class data.
 */
static
void
allocStaticFields(Hjava_lang_Class* class)
{
	int fsize;
	int align;
	int offset;
	int n;
	Field* fld;

	/* Calculate size and position of static data */
	offset = 0;

	n = CLASS_NSFIELDS(class);
	fld = CLASS_SFIELDS(class);
	for (; --n >= 0; fld++) {
		fsize = FIELD_SIZE(fld);
		/* Align field offset */
		align = ALIGNMENT_OF_SIZE(fsize);
		offset = ((offset + align - 1) / align) * align;
		FIELD_SIZE(fld) = offset;
		offset += fsize;
	}

	/* Allocate memory required */
	class->sdata = gc_malloc(offset, GC_ALLOC_STATICDATA);
        GC_WRITE_SAFE(class, class->sdata); /* class->sdata */

	/* Rewalk the fields, pointing them at the relevant memory and/or
	 * setting any constant values.
	 */
	fld = CLASS_SFIELDS(class);
	n = CLASS_NSFIELDS(class);
	for (; --n >= 0; fld++) {
		offset = FIELD_SIZE(fld);
		FIELD_SIZE(fld) = FIELD_CONSTIDX(fld);	/* Keep idx in size */
		FIELD_ADDRESS(fld) = &((uint8*)class->sdata)[offset];
	}
}

static
void
resolveStaticFields(Hjava_lang_Class* class)
{
	uint8* mem;
	constants* pool;
	Field* fld;
	int idx;
	int n;
	int iLockRoot;

	lockMutex(&class->head);

	pool = CLASS_CONSTANTS(class);
	fld = CLASS_SFIELDS(class);
	n = CLASS_NSFIELDS(class);
	for (; --n >= 0; fld++) {
		if ((fld->accflags & FIELD_CONSTANT_VALUE) != 0) {

			mem = FIELD_ADDRESS(fld);
			idx = FIELD_SIZE(fld);

			switch (CONST_TAG((uint32)idx, pool)) {
			case CONSTANT_Integer:
				if (FIELD_TYPE_CNAME(fld)[0] == 'B' || FIELD_TYPE_CNAME(fld)[0] == 'Z') {
					*(jbyte*)mem = (jbyte)CLASS_CONST_INT(class, idx);
					FIELD_SIZE(fld) = TYPE_PRIM_SIZE(&_Jv_byteClass);
				}
				else if (FIELD_TYPE_CNAME(fld)[0] == 'C' || FIELD_TYPE_CNAME(fld)[0] == 'S') {
					*(jshort*)mem = (jshort)CLASS_CONST_INT(class, idx);
					FIELD_SIZE(fld) = TYPE_PRIM_SIZE(&_Jv_shortClass);
				}
				else {
					*(jint*)mem = (jint)CLASS_CONST_INT(class, idx);
					FIELD_SIZE(fld) = TYPE_PRIM_SIZE(&_Jv_intClass);
				}
				break;

			case CONSTANT_Float:
				*(jint*)mem = (jint)CLASS_CONST_INT(class, idx);
				FIELD_SIZE(fld) = TYPE_PRIM_SIZE(&_Jv_floatClass);
				break;

			case CONSTANT_Long:
				*(jlong*)mem = (jlong)CLASS_CONST_LONG(class, idx);
				FIELD_SIZE(fld) = TYPE_PRIM_SIZE(&_Jv_longClass);
				break;

			case CONSTANT_Double:
				*(jlong*)mem = (jlong)CLASS_CONST_DOUBLE(class, idx);
				FIELD_SIZE(fld) = TYPE_PRIM_SIZE(&_Jv_doubleClass);
				break;

			case CONSTANT_String:
				pool->data[idx] = (ConstSlot)internJavaString(makeReplaceJavaStringFromUtf(WORD2UTF(pool->data[idx]), -1, 0, 0));
				pool->tags[idx] = CONSTANT_ResolvedString;
                                GC_WRITE(class, (void*)pool->data[idx]); /* class->constants.data[idx] */
				/* ... fall through ... */
			case CONSTANT_ResolvedString:
				*(jref*)mem = (jref)CLASS_CONST_DATA(class, idx);
				GC_WRITE(class->sdata, *(jref*)mem); /* sdata->data */
				FIELD_SIZE(fld) = PTR_TYPE_SIZE;
				break;
			}
		}
	}

	unlockMutex(&class->head);
}

static
void
buildInterfaces(Hjava_lang_Class* class)
{
	Hjava_lang_Class* nclass;
	int i;
	int j;
	int k;
	int m;

	/* Load all the implemented interfaces. */
	j = class->interface_len;
	nclass = class->superclass;
	if (nclass != 0 && strcmp("java.lang.Object", CLASS_CNAME(nclass)) != 0) {
		/* If class is an interface, include the superclass
		 * as well.
		 */
		if (CLASS_IS_INTERFACE(class)) {
			j += 1;
		}
		j += nclass->total_interface_len;
	}
	for (i = 0; i < class->interface_len; i++) {
		uintp iface = (uintp)class->interfaces[i];
		if (!CLASS_IS_PRECOMPILED(class)) {
			class->interfaces[i] = getClass((constIndex)iface, class, 0);
			/* We ignore any unknown interfaces for now */
			(void)processClass(class->interfaces[i], CSTATE_LINKED);
		}
		else {
			class->interfaces[i] = loadClass((char*)iface, class->loader, CSTATE_LINKED);
		}
                GC_WRITE(class, class->interfaces[i]); /* class->interfaces[i] */
		j += class->interfaces[i]->total_interface_len;
	}
	/* We build a list of *all* interfaces this class can use */
	if (class->interface_len != j) {
		Hjava_lang_Class** newifaces = (Hjava_lang_Class**)gc_malloc_fixed_typed(sizeof(Hjava_lang_Class**) * j, GC_ALLOC_INTERFACE);
		for (i = 0; i < class->interface_len; i++) {
			newifaces[i] = class->interfaces[i];
			GC_WRITE(class, newifaces[i]); /* class->interfaces[i] */
		}
		if (nclass != 0 && nclass != ObjectClass) {
			if (CLASS_IS_INTERFACE(class)) {
				newifaces[i] = nclass;
				GC_WRITE(class, newifaces[i]); /* class->interfaces[i] */
				i++;
			}
			for (m = 0; m < nclass->total_interface_len; m++, i++) {
				newifaces[i] = nclass->interfaces[m];
				GC_WRITE(class, newifaces[i]); /* class->interfaces[i] */
			}
		}
		for (k = 0; k < class->interface_len; k++) {
			nclass = class->interfaces[k];
			for (m = 0; m < nclass->total_interface_len; m++, i++) {
				newifaces[i] = nclass->interfaces[m];
				GC_WRITE(class, newifaces[i]); /* class->interfaces[i] */
			}
		}
		if (!CLASS_IS_PRECOMPILED(class)) {
			gc_free_fixed(class->interfaces);
		}
		class->interfaces = newifaces;
	}
	class->total_interface_len = j;
}

/**
 * Allocate a dispatch table.  We provide well known ones if necessary.
 */
static
void*
allocDispatchTable(Hjava_lang_Class* class, int sz)
{
	char* cname = CLASS_CNAME(class);

#define	KNOWN_DTABLE(C,D)						\
	if (strcmp(cname, C) == 0) {					\
		if (sz != sizeof(D##.method) / sizeof(void*)) {		\
			kprintf("Object %s: need %d have %d\n", C, sz,	\
				 sizeof(D##.method) / sizeof(void*));	\
			ABORT();					\
		}							\
		return (&D);						\
	}

	KNOWN_DTABLE("[C", KaffeDispatchTable_CharArray);
	KNOWN_DTABLE("java.lang.Class",  KaffeDispatchTable_java_lang_Class);
	KNOWN_DTABLE("java.lang.String", KaffeDispatchTable_java_lang_String);

#undef	KNOWN_DTABLE

	return (gc_malloc_fixed_typed(sizeof(dispatchTable) + (sz-1) * sizeof(void*), GC_ALLOC_DISPATCHTABLE));
}

void
buildDispatchTable(Hjava_lang_Class* class)
{
	Method* meth;
#ifdef DIRECT_CALL
	void** mtab;
#else
	void*** mtab;
#endif
	int i;
	Hjava_lang_Class* cc;
	void* tramp;
	short msize;
	short vsize;
	Hjava_lang_Class* super;
	int j;
	Method* lmeth;
	int idx;

	if (!CLASS_IS_PRECOMPILED(class)) {
		/* Set msize to the number of virtual methods in the inherited
		 * dispatch table.
		 */
		msize = 0;
		if (class->superclass != 0) {
			msize = class->superclass->msize;
		}

		meth = CLASS_METHODS(class);
		i = CLASS_NMETHODS(class);
		for (; --i >= 0; meth++) {
			idx = -1;
			/* If method is static or private we skip for now
			 * since we cannot inherit or overload these.
			 * Otherwise ...
			 */
			if (METHOD_IS_INHERITABLE(meth)) {
				/* Search superclasses for equivalent method
				 * name.  If found extract its index nr.
				 */
				for (super = class->superclass; super != 0; super = super->superclass) {
					int j = CLASS_NMETHODS(super);
					Method* mt = CLASS_METHODS(super);
					for (; --j >= 0;  ++mt) {
						if (METHOD_IS_INHERITABLE(mt) && strcmp(mt->name, meth->name) == 0 && strcmp(mt->signature, meth->signature) == 0) {
							idx = mt->idx;
							goto out;
						}
					}
				}
				out:;
				/* No match found
				 * If method is final or the whole class is
				 * final then this method cannot be overridden.
				 * Since it also isn't overriding a superclass
				 * method we don't need to put it in the
				 * dispatch table.
				 */
				if (idx == -1 && !METHOD_IS_FINAL(meth) && !CLASS_IS_FINAL(class)) {
					/* Allocate a new index number */
					idx = msize++;
				}
			}
			meth->idx = idx;
		}

		/* Set the number of virtual methods in out dispatch table */
		class->msize = msize;
	}
	else {
		msize = class->msize;
	}

	/* Now we add in the sizes of dispatch tables for any interfaces.
	 * Now we're a little over generous here since we add in the
	 * static <clinit> methods too.
	 */
	vsize = msize;
	if (!CLASS_IS_INTERFACE(class)) {
		for (i = 0; i < class->total_interface_len; i++) {
			msize += CLASS_NMETHODS(class->interfaces[i]);
		}
	}

	class->dtable = allocDispatchTable(class, msize);

	class->dtable->class = class;
	class->dtable->vsize = vsize;
	mtab = class->dtable->method;

	/* Create a trampoline for each method.  If method is already
	 * translate (precompiled), we just put it into the dtable
	 */
	meth = CLASS_METHODS(class);
	i = CLASS_NMETHODS(class);
	for (; --i >= 0; meth++) {
		if (!METHOD_TRANSLATED(meth)) {
			tramp = Kaffe_EngineInterface_setupTrampoline(meth, class);
			meth->ncode = tramp;
		}
		/* If method is virtual, fill in the dispatch table */
		if (meth->idx != -1) {
#ifdef DIRECT_CALL
                        addMethodPatch( meth, &mtab[meth->idx], PS_addr );
			mtab[meth->idx] = meth->ncode;
#else
			mtab[meth->idx] = &meth->ncode;
#endif
			GC_WRITE_SAFE(class->dtable, meth->ncode); /* dtable->methods[i] */
		}
	}

	/* Now get any inherited methods (virtual only of course) */
	for (cc = class->superclass; cc != 0; cc = cc->superclass) {
		meth = CLASS_METHODS(cc);
		i = CLASS_NMETHODS(cc);
		for (; --i >= 0; meth++) {
			if (meth->idx != -1 && mtab[meth->idx] == 0) {
				assert(meth->idx >= 0 && meth->idx < vsize);
#ifdef DIRECT_CALL
                                addMethodPatch( meth, &mtab[meth->idx], PS_addr );
				mtab[meth->idx] = meth->ncode;
#else
				mtab[meth->idx] = &meth->ncode;
#endif
			}
		}
	}

	/* Add in the interface indirections */
	if (!CLASS_IS_INTERFACE(class)) {
		msize = vsize;
		for (i = 0; i < class->total_interface_len; i++) {
			meth = CLASS_METHODS(class->interfaces[i]);
			for (j = CLASS_NMETHODS(class->interfaces[i]); j > 0; j--) {
				if (!METHOD_IS_STATIC(meth)) {
					for (cc = class;; cc = cc->superclass) {
						lmeth = findMethodLocal(cc, meth->name, meth->signature);
						if (lmeth != 0) {
							break;
						}
					}
#ifdef DIRECT_CALL
                                        addMethodPatch( lmeth, &mtab[msize], PS_addr );
					mtab[msize] = lmeth->ncode;
#else
					mtab[msize] = &lmeth->ncode;
#endif
				}
				msize++;
				meth++;
			}
		}
	}
}

static
unsigned int
hashClassName(void* key)
{
	char* name = (char*)key;
	int len = (strlen(name)+1) >> 1;  /* Number of shorts */
	uint16 *ptr = (uint16*)name;
	int hash = len;
	while (--len >= 0) {
		hash = (hash << 1) ^ *ptr++;
	}
	return (hash & 0x7FFFFFFF);
}

static
int
equalsClassName(void* bagdata, void* key)
{
	return (strcmp((char*)bagdata, (char*)key));
}

Hjava_lang_Class*
lookupClass(char* name, Hjava_lang_ClassLoader* loader)
{
	if (loader == 0) {
		loader = &systemClassLoader;
	}
	if (loader->classPool == 0) {
		loader->classPool = gc_malloc(sizeof(bag), GC_ALLOC_CLASSPOOL);
                GC_WRITE( loader, loader->classPool );
#if defined(MEM2)
                if( loader == &systemClassLoader ){
                        gc_add_ref( loader->classPool, loader->classPool+1 );
                }
#endif
		bagStart(loader->classPool, 64, 75, hashClassName, equalsClassName, GC_ALLOC_CLASSPOOL);
	}
	return ((Hjava_lang_Class*)bagGet(loader->classPool, name));
}

Hjava_lang_Class*
addClass(Hjava_lang_Class* entry)
{
	Hjava_lang_ClassLoader* loader;
	Hjava_lang_Class* centry;
	int iLockRoot;

	/* Lock the class table and insert entry into it */
	lockStaticMutex(&classHashLock);

	loader = entry->loader;
	if (loader == 0) {
		loader = &systemClassLoader;
	}
	if (loader->classPool == 0) {
		loader->classPool = gc_malloc(sizeof(bag), GC_ALLOC_CLASSPOOL);
                GC_WRITE( loader, loader->classPool );
#if defined(MEM2)
                if( loader == &systemClassLoader ){
                        gc_add_ref( loader->classPool, loader->classPool+1 );
                }
#endif
		bagStart(loader->classPool, 64, 75, hashClassName, equalsClassName, GC_ALLOC_CLASSPOOL);
	}

	centry = (Hjava_lang_Class*)bagGet(loader->classPool, CLASS_CNAME(entry));
	if (centry != 0) {
		entry = centry;
	}
	else {
		for (;;) {
			if (bagPut(loader->classPool, CLASS_CNAME(entry), entry) != 0) {
				break;
			}
			unlockStaticMutex(&classHashLock);
			bagRehash(loader->classPool);
			lockStaticMutex(&classHashLock);
		}
                GC_WRITE( loader->classPool, entry );
	}

	unlockStaticMutex(&classHashLock);

	return (entry);
}

/*
 * Lookup a named field.
 */
Field*
lookupClassField(Hjava_lang_Class* clp, char* name)
{
	Field* fptr;
	int n;

	for (; clp != 0; clp = clp->superclass) {
		/* Search down class for field name */
		fptr = CLASS_FIELDS(clp);
		n = CLASS_NFIELDS(clp);
		while (--n >= 0) {
			if (strcmp(name, fptr->name) == 0) {
				/* Resolve field if necessary */
				resolveFieldType(fptr, clp);
				return (fptr);
			}
			fptr++;
		}
	}
	return (0);
}

/*
 * Calculate size (in words) of a signature item.
 */
static
int
sizeofSigItem(char** strp, bool want_wide_refs)
{
  int count;
  char* str;

  count = 0;
  for (str = *strp; ; str++) {
	switch (*str) {
	case '(':
	  continue;
	case 0:
	case ')':
	  count = -1;
	  break;
	case 'V':
	  count = 0;
	  break;
	case 'I':
	case 'Z':
	case 'S':
	case 'B':
	case 'C':
	case 'F':
	  count = 1;
	  break;
	case 'D':
	case 'J':
	  count = 2;
	  break;
	case '[':
	  count = want_wide_refs ? sizeof(void*) / sizeof(int32) : 1;
	arrayofarray:
	  str++;
	  if (*str == 'L') {
		while (*str != ';') {
		  str++;
		}
	  }
	  else if (*str == '[') {
		goto arrayofarray;
	  }
	  break;
	case 'L':
	  count = want_wide_refs ? sizeof(void*) / sizeof(int32) : 1;
	  /* Skip to end of reference */
	  while (*str != ';') {
		str++;
	  }
	  break;
	default:
	  ABORT();
	}

	*strp = str + 1;
	return (count);
}
}

/*
 * Calculate size of data item based on signature.
 */
static
int
sizeofSig(char** strp, bool want_wide_refs)
{
  int count;
  int c;

  count = 0;
  for (;;) {
	c = sizeofSigItem(strp, want_wide_refs);
	if (c == -1) {
	  return (count);
	}
	count += c;
  }
}

/*
 * Determine the number of arguments and return values from the
 * method signature.
 */
void
countInsAndOuts(char* str, short* ins, short* outs, char* outtype)
{
	*ins = sizeofSig(&str, false);
	*outtype = str[0];
	*outs = sizeofSig(&str, false);
}

Hjava_lang_Class*
lookupArrayProcess(Hjava_lang_Class* c, int state)
{
	char *arr_name;
	char sig[CLASSMAXSIG];  /* FIXME! unchecked fixed buffer! */
	Hjava_lang_Class* arr_class;
	Hjava_lang_Class* arr_class2;

	/* Build signature for array type */
	sig[0] = '[';
	if (CLASS_IS_PRIMITIVE (c)) {
		arr_class = CLASS_ARRAY_CACHE(c);
		if (arr_class) {
			goto found;
		}
		sig[1] = (char)CLASS_PRIM_SIG(c);
		sig[2] = 0;
	}
	else {
		char* cname = CLASS_CNAME (c);
		if (cname[0] == '[') {
			sig[1] = 0;
		}
		else {
			sig[1] = 'L';
			sig[2] = 0;
		}
		strcat(sig, cname);
	}
	arr_name = addUtfCopy(sig, -1);
	arr_class = lookupClass(arr_name, c->loader);

	/* If we have it already, don't create a new one */
	if (arr_class != 0) {
		goto found;
	}

	arr_class = newClass();
	if (arr_class == 0) {
		return (0);
	}

	internalSetupClass(arr_class, arr_name, ACC_ABSTRACT|ACC_FINAL|ACC_PUBLIC, 0, c->loader);
	CLASS_ELEMENT_TYPE(arr_class) = c;
	arr_class->head.dtable = (dispatchTable*)&KaffeDispatchTable_java_lang_Class;
        GC_WRITE_SAFE(arr_class, &KaffeDispatchTable_java_lang_Class);
	arr_class->sdata = 0;
	arr_class->constants.size = 1; /* dummy */

	arr_class2 = addClass(arr_class);
	if (arr_class2 != arr_class) {
		/* Someone else added it before me - jus discard this one */
		arr_class = arr_class2;
	}
	if (CLASS_IS_PRIMITIVE(c)) {
		CLASS_ARRAY_CACHE(c) = arr_class;
	}

	found:;

	if (processClass(arr_class, state) == 0) {
		return (0);
	}
	else {
		return (arr_class);
	}
}

/*
 * Unload the given class.
 */
void
unloadClass(Hjava_lang_Class* class)
{
	Method* mptr;
	int i;

	/* If class has no dtable then it was never really loaded */
	if (class->dtable == 0) {
		return;
	}

	/* We cannot unload the system classes */
	assert(class->loader != 0);

	/* Unloading a class requires the following steps: */

	/* 1. Remove any active methods. */

	for (i = CLASS_NMETHODS(class), mptr = CLASS_METHODS(class); i > 0; i--, mptr++) {
		if (METHOD_TRANSLATED(mptr)) {
                        makeMethodInactive(mptr);
		}
#ifndef JIT4
		else {
			gc_free_fixed(BCODE(mptr));
		}
#endif
#if !defined(OPTIMIZE_MEMORY) || defined(DEBUG)
		gc_free_fixed(mptr->lines);
#endif
		gc_free_fixed(mptr->exception_table);
	}
	gc_free_fixed(class->constants.data);
	gc_free_fixed(class->methods);
	gc_free_fixed(class->fields);
	if (class->interfaces != ArrayInterfaces) {
		gc_free_fixed(class->interfaces);
	}
#if (defined(MEM1) && 0) || defined(MEM2)
	gc_free_fixed(class->refs);
#endif
}

/**
 * Unload a class loader and anything is references.
 */
void
unloadClassLoader(Hjava_lang_ClassLoader* loader)
{
	bagFinish(loader->classPool);
	gc_free_fixed(loader->classPool);
	unloadNativeLibraries(loader);
}

/*
 * Lookup a precompiled class
 */
static
Hjava_lang_Class*
lookupPrecompiledClass(char* name, Hjava_lang_ClassLoader* loader)
{
	int i;
	Hjava_lang_Class* cls;

	/* All precompiled classes are in the system classloader right now */
	if (loader != 0) {
		return (0);
	}

	for (i = 0; prelist[i] != 0; i++) {
		cls = (*prelist[i])(name);
		if (cls != 0) {
			return (cls);
		}
	}
	return (0);
}

void
registerPrecompiledClasses(void* func)
{
	int i;
	for (i = 0; i < sizeof(prelist) / sizeof(prelookup); i++) {
		if (prelist[i] == 0) {
			prelist[i] = (prelookup)func;
			break;
		}
	}
	initBaseClasses();
}
