/*
 * cms.c
 *
 * Copyright (c) 1996-1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "classMethod.h"
#include "callMethodInfo.h"
#include "access.h"
#include "lookup.h"
#include "engine.h"
#include "objects.h"
#include "soft.h"
#include "support.h"
#include "java_lang_Throwable.h"
#include "exception.h"
#include "locks.h"
#include "threads.h"
#include "cms.h"

/*
 * Create an instance of a class.
 */
static
jref
cms_bc_new(void)
{
	Hjava_lang_Object* obj;
	Method* meth;

	CMS_PROLOGUE();

	meth = findFrameMethod((uintp)cms_pc);
	assert(meth != UNKNOWN_METHOD);

	obj = soft_new(getClass(((CMS_PC(1) << 8) | CMS_PC(2)), meth->class));

	CMS_EPILOGUE();

	return (obj);
}

/*
 * Create an array of a basic type.
 */
static
jref
cms_bc_newarray(jint size)
{
	Hjava_lang_Object* obj;

	CMS_PROLOGUE();

	obj = soft_newarray(CMS_PC(1), size);

	CMS_EPILOGUE();

	return (obj);
}

/*
 * Create an array of reference to objects.
 */
static
jref
cms_bc_anewarray(jint size)
{
	Hjava_lang_Object* obj;
	Method* meth;

	CMS_PROLOGUE();

	meth = findFrameMethod((uintp)cms_pc);
	assert(meth != UNKNOWN_METHOD);

	obj = soft_anewarray(getClass(((CMS_PC(1) << 8) | CMS_PC(2)), meth->class), size);

	CMS_EPILOGUE();

	return (obj);
}

static
jref
cms_bc_multianewarray(jint arg, ...)
{
	Method* meth;
	va_list ap;
	jint* arraydims;
	jint len;
	int i;
	Hjava_lang_Object* obj;

	CMS_PROLOGUE();

	meth = findFrameMethod((uintp)cms_pc);
	assert(meth != UNKNOWN_METHOD);

	len = CMS_PC(3);
	arraydims = gc_calloc(len+1, sizeof(jint), GC_ALLOC_ENGINE);
	va_start(ap, arg);
	for (i = 0; i < len; i++) {
		if (arg < 0) {
			throwException(NegativeArraySizeException);
		}
		arraydims[i] = arg;
		arg = va_arg(ap, jint);
	}
	arraydims[i] = -1;

	va_end(ap);

	obj = newMultiArray(getClass(((CMS_PC(1) << 8) | CMS_PC(2)), meth->class), arraydims);

	gc_free(arraydims);

	CMS_EPILOGUE();

	return (obj);
}

static
jint
cms_bc_instanceof(jref obj)
{
	Method* meth;
	jint ret;

	CMS_PROLOGUE();

	meth = findFrameMethod((uintp)cms_pc);
	assert(meth != UNKNOWN_METHOD);

	ret = soft_instanceof(getClass(((CMS_PC(1) << 8) | CMS_PC(2)), meth->class), (Hjava_lang_Object*)obj);

	CMS_EPILOGUE();

	return (ret);
}

static
void
cms_bc_checkcast(jref obj)
{
	Method* meth;

	CMS_PROLOGUE();

	meth = findFrameMethod((uintp)cms_pc);
	assert(meth != UNKNOWN_METHOD);

	soft_checkcast(getClass(((CMS_PC(1) << 8) | CMS_PC(2)), meth->class), (Hjava_lang_Object*)obj);

	CMS_EPILOGUE();
}

static
void
cms_bc_athrow(jref obj)
{
	CMS_PROLOGUE();

	soft_athrow((Hjava_lang_Throwable*)obj);

	CMS_EPILOGUE();
}

static
void
cms_bc_monitorenter(jref obj)
{
	CMS_PROLOGUE();

	lockObject((Hjava_lang_Object*)obj);

	CMS_EPILOGUE();
}

static
void
cms_bc_monitorexit(jref obj)
{
	CMS_PROLOGUE();

	unlockObject((Hjava_lang_Object*)obj);

	CMS_EPILOGUE();
}

static
void
cms_ex_badarrayindex(void)
{
	CMS_PROLOGUE();

	soft_badarrayindex();

	CMS_EPILOGUE();
}

static
void
cms_ex_nullpointer(void)
{
	CMS_PROLOGUE();

	soft_nullpointer();

	CMS_EPILOGUE();
}

static
void
cms_ex_divzero(void)
{
	CMS_PROLOGUE();

	soft_divzero();

	CMS_EPILOGUE();
}

static
void
cms_ex_nosuchmethod(void)
{
	CMS_PROLOGUE();

	throwException(NoSuchMethodError("???"));

	CMS_EPILOGUE();
}

static
void
cms_ex_nosuchfield(void)
{
	CMS_PROLOGUE();

	throwException(NoSuchFieldError("???"));

	CMS_EPILOGUE();
}

static
void
cms_ex_stackoverflow(void)
{
	CMS_PROLOGUE();

	soft_stackoverflow();

	CMS_EPILOGUE();
}

static
void
cms_ot_initialize_class(void)
{
	Method* meth;
	jint idx;
	Hjava_lang_Class* clazz;
	callInfo call;
	fieldInfo field;

	CMS_PROLOGUE();

	meth = findFrameMethod((uintp)cms_pc);
	assert(meth != UNKNOWN_METHOD);

	idx = (CMS_PC(1) << 8) | CMS_PC(2);
	switch (CLASS_CONSTANTS(meth->class)->tags[idx]) {
	case CONSTANT_Methodref:
	case CONSTANT_InterfaceMethodref:
		getMethodSignatureClass(idx, meth->class, false, &call);
		clazz = call.class;
		break;
	case CONSTANT_Fieldref:
		getField(idx, meth->class, true, &field);
		clazz = field.class;
		break;
	default:
		clazz = 0;
		abort();
	}

	soft_initialise_class(clazz);

	CMS_EPILOGUE();
}

static
jref
cms_ot_lookupmethod(jref obj, jref name, jref sig)
{
	jref ret;

	CMS_PROLOGUE();

	ret = soft_lookupmethod((Hjava_lang_Object*)obj, (Utf8Const*)name, (methodSig*)sig);

	CMS_EPILOGUE();

	return (ret);
}

static
void
cms_ot_checkarraystore(jref array, jref obj)
{
	CMS_PROLOGUE();

	soft_checkarraystore((Hjava_lang_Object*)array, (Hjava_lang_Object*)obj);

	CMS_EPILOGUE();
}

/*
 * Define the engine interface.
 */
BYTECODE_STATE_s _engineInterface = {

	0,	/* version */
	0,	/* optop */
	0,	/* j_sc_bottom */
	0,	/* j_const_pool */
	0,	/* initialized */

	{ 0, 0 },				/* bc_exit */
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },	/* bc_ret */


	/* Functions for opcodes that cannot be executed directly */

	cms_bc_new,
	cms_bc_newarray,
	cms_bc_anewarray,
	cms_bc_multianewarray,
	cms_bc_instanceof,
	cms_bc_checkcast,
	cms_bc_athrow,
	cms_bc_monitorenter,
	cms_bc_monitorexit,

	/* Functions for exceptions */

	cms_ex_badarrayindex,
	cms_ex_nullpointer,
	cms_ex_divzero,
	cms_ex_nosuchmethod,
	cms_ex_nosuchfield,
	cms_ex_stackoverflow,

	/* Others - these don't fit in anywhere */

	cms_ot_initialize_class,
	cms_ot_lookupmethod,
	cms_ot_checkarraystore

};

BYTECODE_STATE engineInterface = &_engineInterface;


/* --------------------------------------------------------------------- */

/*
 * Below is the CMS engine interface.  This still assumes something which
 * looks like an interpreter in many places until we change the way we
 * handle, amongst other things, the stack traces for the CMS engine.
 * Some of this is real, some of this is fake.
 */
char* engine_name = "CMS";
char* engine_version = KAFFEVERSION;

extern struct JNIEnv_ Kaffe_JNIEnv;
Method* interpreterMethod;

/*
 * Call the CMS engine.
 */
static
void
callCMSEngine(callMethodInfo* call)
{
	extern jint cmsFakeEngineCall(jref, jint, ...);
	register callMethodInfo* scall asm("%edi");
	scall = call;

	asm volatile ("
1:
		cmpl $0,%0
		je 3f
		decl %0
		cmpb $0,(%2,%0)
		je 1b
		cmpb $1,(%2,%0)
		je 2f
		pushl 4(%1,%0,8)
2:
		pushl (%1,%0,8)
		jmp 1b
3:
		pushl %6
		pushl %3		# FAKE
		call cmsFakeEngineCall	# CALL
		movl %5,%%ebx
		movb %4,%%cl
		cmpb $0x46,%%cl
		jne 4f
		fstps (%%ebx)
		jmp 6f
4:
		cmpb $0x44,%%cl
		jne 5f
		fstpl (%%ebx)
		jmp 6f
5:
		movl %%eax,(%%ebx)
		cmpb $0x4a,%%cl
		jne 6f
		movl %%edx,4(%%ebx)
6:
	" :
	  : "r" ((scall)->nrargs),
	    "r" ((scall)->args),
	    "r" ((scall)->callsize),
	    "m" ((scall)->function),
	    "m" ((scall)->rettype),
	    "m" ((scall)->ret),
	    "m" ((scall)->nrargs)
	  : "ebx", "ecx", "cc", "memory");
	asm volatile ("
		subl %0,%%esp
	" : : "r" ((scall)->argsize * sizeof(jint)) : "cc");
}

static
void
Einit(void)
{
	/* Install the interpreter as a pseudo-method */
	interpreterMethod = gc_malloc_fixed(sizeof(Method));
}

/*
 * Call inter-language method
 */
static
void
EcallInterEngine(callMethodInfo* call)
{
	Method* meth;
	Hjava_lang_Object* sync;
	int i;
	int iLockRoot;

	meth = (Method*)call->function;
	if ((meth->accflags & ACC_NATIVE) == 0) {
		callCMSEngine(call);
	}
	else {
		if (!METHOD_TRANSLATED(meth)) {
                        native(meth);
                }
		call->function = METHOD_NATIVECODE(meth);

		if (meth->accflags & ACC_SYNCHRONISED) {
			if (meth->accflags & ACC_STATIC) {
				sync = &meth->class->head;
			}
			else {
				sync = (Hjava_lang_Object*)call->args[0].l;
			}
		}
		else {
			sync = 0;
		}

		/* For a JNI method we must insert the JNI environment and
		 * possibly the class too.
		 */
		if (meth->accflags & ACC_JNI) {
			i = 1;
			if (METHOD_IS_STATIC(meth)) {
				i++;
			}
			memmove(&call->callsize[i], &call->callsize[0], sizeof(char) * call->nrargs);
			memmove(&call->calltype[i], &call->calltype[0], sizeof(char) * call->nrargs);
			memmove(&call->args[i], &call->args[0], sizeof(jvalue) * call->nrargs);
			call->callsize[0] = PTR_TYPE_SIZE / SIZEOF_INT;
			call->calltype[0] = 'L';
			call->args[0].l = (void*)&Kaffe_JNIEnv;
			call->argsize += call->callsize[0];
			call->nrargs++;

			if (METHOD_IS_STATIC(meth)) {
				call->callsize[1] = PTR_TYPE_SIZE / SIZEOF_INT;
				call->calltype[1] = 'L';
				processClass(meth->class, CSTATE_LINKED);
				call->args[1].l = (void*)meth->class;
				call->argsize += call->callsize[1];
				call->nrargs++;
			}
		}

		if (sync != 0) {
			lockMutex(sync);
		}

		/* Make the call - system dependent */
		sysdepCallMethod(call);

		if (sync != 0) {
			unlockMutex(sync);
		}
	}
}

/*
 * Dispatch an exception.
 */
static
void
EcallExceptionHandler(Method* meth, uintp handler, stackTraceInfo* frame)
{
        vmException* fm;

        fm = (vmException*)frame->arg0;

        fm->pc = handler;
        longjmp(fm->jbuf, 1);
}

static
char*
EgetName(void)
{
        return (engine_name);
}

static
char*
EgetVersion(void)
{
        return (engine_version);
}

static
int
EgetNumber(void)
{
        return (0);
}

static
void*
EsetupTrampoline(Method* meth, void** where)
{
	return (meth);
}

static
void
EprocessExceptionMethod(stackTraceInfo* info)
{
	vmException* except;

	except = (vmException*)unhand(Kaffe_TI_currentJava())->exceptPtr;
	for (; except != 0; except = except->prev) {
		if (except->frame == info->rframe) {
			info->meth = except->meth;
			info->object = except->mobj;
			info->arg0 = (uintp)except;
			info->pc = except->pc;
			return;
		}
	}

	/* Not found */
	info->arg0 = 0;
	info->object = 0;
	if (info->meth == interpreterMethod) {
		info->meth = UNKNOWN_METHOD;
	}
}

static
void
EsetupJNIcall(Method* xmeth, void* func)
{
	SET_METHOD_NATIVECODE(xmeth, func);
	xmeth->accflags |= ACC_JNI;
}

/*
 * Define the engine interface.
 */
EngineInterface Kaffe_EngineInterface = {

	Einit,

	EgetName,
	EgetVersion,
	EgetNumber,

	EsetupTrampoline,
	EprocessExceptionMethod,
	EcallInterEngine,
	EcallExceptionHandler,
	EsetupJNIcall,

};
