/* code-analyse.c
 * Analyse a method's bytecodes.
 *
 * Copyright (c) 1996 - 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "bytecode.h"
#include "objects.h"
#include "constants.h"
#include "access.h"
#include "classMethod.h"
#include "code-analyse.h"
#include "lookup.h"
#include "exception.h"
#include "icode.h"
#include "itypes.h"
#include "locks.h"
#include "md.h"
#include "gc.h"
#include "support.h"
#if defined(TRANSLATOR)
#include "slots.h"
#include "registers.h"
#endif
#include "code.h"

#include "jdebug.h"
#include "jmeasure.h"

/*
 * Instruction lengths
 *  Note: some instructions are variable length or affected by 'wide'.
 */
const int8 insnLen[256] = {
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	2,  3,  2,  3,  3,  2,  2,  2,  2,  2,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  2,  2,  2,  2,  2,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  3,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  3,  3,  3,  3,  3,  3,  3,  
	3,  3,  3,  3,  3,  3,  3,  3,  3,  2,  0,  0,  1,  1,  1,  1,  
	1,  1,  3,  3,  3,  3,  3,  3,  3,  5,  1,  3,  2,  3,  1,  1,  
	3,  3,  1,  1,  1,  4,  3,  3,  5,  5,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
	1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  
};

/*
 * Stack adjustments
 *  Note: some adjustments are variable.
 */
const int8 insnStk[256] = {
	 0, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -1, -1, -1, -2, -2, 
	-1, -1, -1, -1, -2, -1, -2, -1, -2, -1, -1, -1, -1, -1, -2, -2, 
	-2, -2, -1, -1, -1, -1, -2, -2, -2, -2, -1, -1, -1, -1,  1,  0, 
	 1,  0,  1,  1,  1,  1,  1,  2,  1,  2,  1,  1,  1,  1,  1,  2, 
	 2,  2,  2,  1,  1,  1,  1,  2,  2,  2,  2,  1,  1,  1,  1,  3, 
	 4,  3,  4,  3,  3,  3,  3,  1,  2, -1, -1, -1, -2, -2, -2,  0, 
	 1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2,  1,  2, 
	 1,  2,  1,  2,  0,  0,  0,  0,  1,  1,  1,  1,  1,  1,  1,  2, 
	 1,  2,  1,  2,  0, -1,  0, -1,  1,  1,  0,  0, -1, -1,  1,  0, 
	 1,  0,  0,  0,  3,  1,  1,  3,  3,  1,  1,  1,  1,  1,  1,  2, 
	 2,  2,  2,  2,  2,  2,  2,  0,  0,  0,  1,  1,  1,  2,  1,  2, 
	 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  1, 
	 0,  0,  1,  1,  0,  0,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
	 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 
};

static bool verifyBasicBlock(codeinfo*, basicBlockInfo*, Method*, Hjava_lang_Class*, int32, frameElement*);
static basicBlockInfo* insertBB(codeinfo*, int);

#include "code-analyse.inc"

/*
 * The real code verifier
 */
int
verifyMethod(Method* meth, Hjava_lang_Class* class, codeinfo** pcodeInfo)
{
	int32 pc;
	int32 tabpc;
	int32 jpc;
	int32 idx;
	int32 sp;
	int32 lcl;
	int rerun;
	int lastrerun;
	bool wide;
        localUse* localuse;
        frameElement* activeFrame;
	codeinfo* codeInfo;
	basicBlockInfo* currBB;

        MSR_TIME(code_analyse);

	/* We don't need to do this twice */
	meth->accflags |= ACC_VERIFIED;

        MSR_START(code_analyse);

	activeFrame = 0;
	codeInfo = allocCodeInfo(BCODE(meth)->codelen);

	(*pcodeInfo) = codeInfo;
	if (codeInfo == 0) {
		return (0);
	}
	codeInfo->localuse = allocLocalUse(BCODE(meth)->localsz);

	if (codeInfo->localuse == 0) {
		goto error;
	}

	localuse = codeInfo->localuse;
	for (lcl = 0; lcl < BCODE(meth)->localsz; lcl++) {
		localuse[lcl].use = 0;
		localuse[lcl].first = 0x7FFFFFFF;
		localuse[lcl].last = -1;
		localuse[lcl].write = -1;
		localuse[lcl].type = 0;
	}

	DBG( vm_verify, ("codeInfo = 0x%x\n", codeInfo));
	DBG( vm_verify, ("%s.%s\n", CLASS_CNAME(class), meth->signature));

	/* Allocate code info. block */
	codeInfo->localsz = BCODE(meth)->localsz;
	codeInfo->stacksz = BCODE(meth)->stacksz;
	codeInfo->codelen = BCODE(meth)->codelen;

	/* Insert first and last blocks */
	INSERT_NEW_BLOCK(0);
	INSERT_NEW_BLOCK(codeInfo->codelen);
	INSERT_NEW_BLOCK(0xFFFF);

	/* Scan the code and mark the beginning of basic blocks */
	wide = false;
	for (pc = 0; pc < codeInfo->codelen;) {
		switch (INSN(pc)) {
		case IFEQ:	case IFNE:	case IFLT:
		case IFGE:	case IFGT:	case IFLE:
		case IF_ICMPEQ: case IF_ICMPNE:
		case IF_ICMPLT: case IF_ICMPGE:
		case IF_ICMPGT: case IF_ICMPLE:
		case IF_ACMPEQ: case IF_ACMPNE:
		case IFNULL:	case IFNONNULL:
			tabpc = pc + WORD(pc+1);
			SET_STARTOFBASICBLOCK(tabpc);
			pc = pc + INSNLEN(pc);
			SET_STARTOFBASICBLOCK(pc);
			break;
		case GOTO:
			tabpc = pc + WORD(pc+1);
			SET_STARTOFBASICBLOCK(tabpc);
			pc = pc + INSNLEN(pc);
			break;
		case GOTO_W:
			tabpc = pc + DWORD(pc+1);
			SET_STARTOFBASICBLOCK(tabpc);
			pc = pc + INSNLEN(pc);
			break;
		case JSR:
			tabpc = pc + WORD(pc+1);
			SET_STARTOFBASICBLOCK(tabpc);
			pc = pc + INSNLEN(pc);
			SET_STARTOFBASICBLOCK(pc);
			break;
		case JSR_W:
			tabpc = pc + DWORD(pc+1);
			SET_STARTOFBASICBLOCK(tabpc);
			pc = pc + INSNLEN(pc);
			SET_STARTOFBASICBLOCK(pc);
			break;
		case TABLESWITCH:
			tabpc = (pc + 4) & -4;
			idx = DWORD(tabpc+8)-DWORD(tabpc+4)+1;
			for (; idx > 0; idx--) {
                                jpc = pc+DWORD(tabpc+idx*4+8);
				SET_STARTOFBASICBLOCK(jpc);
			}
                        jpc = pc+DWORD(tabpc);
			SET_STARTOFBASICBLOCK(jpc);
			pc = tabpc + (DWORD(tabpc+8)-DWORD(tabpc+4)+1+3) * 4;
			break;
		case LOOKUPSWITCH:
			tabpc = (pc + 4) & -4;
			idx = DWORD(tabpc+4);
			for (; idx > 0; idx--) {
                                jpc = pc+DWORD(tabpc+idx*8+4);
				SET_STARTOFBASICBLOCK(jpc);
			}
                        jpc = pc+DWORD(tabpc);
			SET_STARTOFBASICBLOCK(jpc);
			pc = tabpc + (DWORD(tabpc+4)+1) * 8;
			break;
		case IRETURN:	case LRETURN:	case ARETURN:
		case FRETURN:	case DRETURN:	case RETURN:
		case ATHROW:	case RET:
			pc = pc + INSNLEN(pc);
			break;
		case WIDE:
			wide = true;
			pc = pc + INSNLEN(pc);
			break;
		case ILOAD:	case LLOAD:	case FLOAD:
		case DLOAD:	case ALOAD:
		case ISTORE:	case LSTORE:	case FSTORE:
		case DSTORE:	case ASTORE:
			pc = pc + INSNLEN(pc);
			if (wide == true) {
				wide = false;
				pc += 1;
			}
			break;
		case IINC:
			pc = pc + INSNLEN(pc);
			if (wide == true) {
				wide = false;
				pc += 2;
			}
			break;
		default:
			/* The default */
			pc = pc + INSNLEN(pc);
			break;
		}
	}

	/* Setup exception info. */
	sp = BCODE(meth)->localsz + BCODE(meth)->stacksz - 1;
	if (meth->exception_table != 0) {
		for (lcl = 0; meth->exception_table[lcl].start_pc != (uintp)-1; lcl++) {
			pc = meth->exception_table[lcl].handler_pc;
			SET_STARTOFBASICBLOCK(pc);
			SET_STARTOFEXCEPTION(pc);
			SET_STACKPOINTER(pc, sp);
		}
	}

	/* Mark the various starting states.  These include the main
	 * entry point plus all the exception entry points, their arguments
	 * and stack values.
	 */
	pc = 0;
	SET_STACKPOINTER(pc, BCODE(meth)->localsz + BCODE(meth)->stacksz);

	/* Scan out list of basic blocks.  Unfortunately they're not in
	 * precise order so we have to do this until they're all done.
	 */
	activeFrame = ALLOCFRAME();
	if (activeFrame == 0) {
		goto error;
	}
	rerun = 0;
	do {
		lastrerun = rerun;
		rerun = 0;
		for (currBB = codeInfo->bbinfo; currBB->start < codeInfo->codelen; currBB = currBB->next) {
			pc = currBB->start;
			if (IS_DONEVERIFY(pc)) {
				/* Done already */
			}
			else if (IS_NEEDVERIFY(pc)) {
				if (verifyBasicBlock(codeInfo, currBB, meth, class, pc, activeFrame) != 0) {
					goto error;
				}
			}
			/* Else we need to run again when the block can
			 * be verified.
			 */
			else {
				rerun++;
			}
		}
	} while (rerun != lastrerun);

	/* We do not unlock at the end but leave this to tidyVerifyMethod */
	MSR_STOP(code_analyse);

	if (activeFrame != 0) {
		freeFrame(activeFrame);
	}
	DBG( vm_verify, ("Completed %s.%s\n", CLASS_CNAME(class), meth->signature));
	return (1);

error:;
	if (activeFrame != 0) {
		freeFrame(activeFrame);
	}
	return (0);
}

static
bool
verifyBasicBlock(codeinfo* codeInfo, basicBlockInfo* currBB, Method* meth, Hjava_lang_Class* class, int32 pc, frameElement* activeFrame)
{
	int32 tabpc;
	int32 idx;
	int32 sp;
	int32 lcl;
	int32 lclw;
	int32 lclww;
	int32 opc;
	callInfo call;
	fieldInfo finfo;
	bool wide;
	bool failed;
	bool unreachable;
	localUse* l;
	int pcd;
	int spd;
	int insn;

	opc = pc;

	/* Get stack pointer */
	sp = currBB->stackPointer;
	assert(sp != 0xFFFF); /* Check it's set */

	/* Reset activeFrame and use it to hold type data */
        memset(activeFrame, 0, codeInfo->localsz * sizeof(frameElement));

	/* Process basic block until we get to the beginning of a new one */
	wide = false;
	failed = false;
	unreachable = false;
	do {

		if (sp < BCODE(meth)->localsz || sp > BCODE(meth)->localsz + BCODE(meth)->stacksz) {
			failed = true;
			DBG( vm_verify, ("sp out of range: %d <%d> %d\n", BCODE(meth)->localsz, sp, BCODE(meth)->localsz + BCODE(meth)->stacksz));
			break;
		}

		/* Retrieve a potential local slot */
		lcl = BYTE(pc+1);
		lclw = WORD(pc+1);
		lclww = DWORD(pc+1);

		DBG( vm_verify, ("%d: %d\n", pc, INSN(pc)));

		insn = INSN(pc);
		spd = insnStk[insn];
		pcd = insnLen[insn];

		switch (insn) {
		case ILOAD_0:
		case ILOAD_1:
		case ILOAD_2:
		case ILOAD_3:
			lcl = INSN(pc) - ILOAD_0;
			LCL(lcl).used = 1;
			break;

		case ILOAD:
		case FLOAD:
		case ALOAD:
			if (wide) {
				wide = false;
				lcl = lclw;
				INCPC(1);
			}
			LCL(lcl).used = 1;
			break;

		case LLOAD_0:
		case LLOAD_1:
		case LLOAD_2:
		case LLOAD_3:
			lcl = INSN(pc) - LLOAD_0;
			LCL(lcl).used = 1;
			LCL(lcl+1).used = 1;
			break;

		case LLOAD:
		case DLOAD:
			if (wide) {
				wide = false;
				lcl = lclw;
				INCPC(1);
			}
			LCL(lcl).used = 1;
			LCL(lcl+1).used = 1;
			break;

		case FLOAD_0:
		case FLOAD_1:
		case FLOAD_2:
		case FLOAD_3:
			lcl = INSN(pc) - FLOAD_0;
			LCL(lcl).used = 1;
			break;

		case DLOAD_0:
		case DLOAD_1:
		case DLOAD_2:
		case DLOAD_3:
			lcl = INSN(pc) - DLOAD_0;
			LCL(lcl).used = 1;
			LCL(lcl+1).used = 1;
			break;

		case ALOAD_0:
		case ALOAD_1:
		case ALOAD_2:
		case ALOAD_3:
			lcl = INSN(pc) - ALOAD_0;
			LCL(lcl).used = 1;
			break;

		case ISTORE_0:
		case ISTORE_1:
		case ISTORE_2:
		case ISTORE_3:
			lcl = INSN(pc) - ISTORE_0;
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TINT;
			break;

		case ISTORE:
			if (wide) {
				lcl = lclw;
				INCPC(1);
				wide = false;
			}
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TINT;
			break;

		case LSTORE_0:
		case LSTORE_1:
		case LSTORE_2:
		case LSTORE_3:
			lcl = INSN(pc) - LSTORE_0;
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TLONG;
			LCL(lcl+1).used = 1;
			LCL(lcl+1).modified = 1;
			LCL(lcl+1).type = TVOID;
			break;

		case LSTORE:
			if (wide) {
				lcl = lclw;
				INCPC(1);
				wide = false;
			}
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TLONG;
			LCL(lcl+1).used = 1;
			LCL(lcl+1).modified = 1;
			LCL(lcl+1).type = TVOID;
			break;

		case FSTORE_0:
		case FSTORE_1:
		case FSTORE_2:
		case FSTORE_3:
			lcl = INSN(pc) - FSTORE_0;
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TFLOAT;
			break;

		case FSTORE:
			if (wide) {
				lcl = lclw;
				INCPC(1);
				wide = false;
			}
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TFLOAT;
			break;

		case DSTORE_0:
		case DSTORE_1:
		case DSTORE_2:
		case DSTORE_3:
			lcl = INSN(pc) - DSTORE_0;
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TDOUBLE;
			LCL(lcl+1).used = 1;
			LCL(lcl+1).modified = 1;
			LCL(lcl+1).type = TVOID;
			break;

		case DSTORE:
			if (wide) {
				lcl = lclw;
				INCPC(1);
				wide = false;
			}
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TDOUBLE;
			LCL(lcl+1).used = 1;
			LCL(lcl+1).modified = 1;
			LCL(lcl+1).type = TVOID;
			break;

		case ASTORE_0:
		case ASTORE_1:
		case ASTORE_2:
		case ASTORE_3:
			lcl = INSN(pc) - ASTORE_0;
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TOBJ;
			break;

		case ASTORE:
			if (wide) {
				lcl = lclw;
				INCPC(1);
				wide = false;
			}
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			LCL(lcl).type = TOBJ;
			break;

		case IINC:
			if (wide) {
				lcl = lclw;
				INCPC(2);
				wide = false;
			}
			LCL(lcl).type = TINT;
			LCL(lcl).used = 1;
			LCL(lcl).modified = 1;
			break;

		case IFEQ:
		case IFNE:
		case IFLT:
		case IFGE:
		case IFGT:
		case IFLE:
		case IFNULL:
		case IFNONNULL:
			SET_STACKPOINTER(pc + lclw, sp + 1);
			SET_STACKPOINTER(pc + 3, sp + 1);
			break;

		case IF_ICMPEQ:
		case IF_ICMPNE:
		case IF_ICMPLT:
		case IF_ICMPGT:
		case IF_ICMPGE:
		case IF_ICMPLE:
		case IF_ACMPEQ:
		case IF_ACMPNE:
			SET_STACKPOINTER(pc + lclw, sp + 2);
			SET_STACKPOINTER(pc + 3, sp + 2);
			break;

		case GOTO:
			SET_STACKPOINTER(pc + lclw, sp);
			unreachable = true;
			break;

		case GOTO_W:
			SET_STACKPOINTER(pc + lclww, sp);
			unreachable = true;
			break;

		case JSR:
			SET_STACKPOINTER(pc + lclw, sp - 1);
			SET_STACKPOINTER(pc + 3, sp);
			break;

		case JSR_W:
			SET_STACKPOINTER(pc + lclww, sp - 1);
			SET_STACKPOINTER(pc + 5, sp);
			break;

		case RET:
			LCL(lcl).used = 1;
			unreachable = true;
			break;

		case LOOKUPSWITCH:
			tabpc = (pc + 4) & -4;
			idx = DWORD(tabpc+4);
			for (; idx > 0; idx--) {
				SET_STACKPOINTER(pc+DWORD(tabpc+idx*8+4), sp + 1);
			}
			SET_STACKPOINTER(pc+DWORD(tabpc), sp + 1);
			pc = tabpc + (DWORD(tabpc+4)+1) * 8;
			unreachable = true;
			break;

		case TABLESWITCH:
			tabpc = (pc + 4) & -4;
			idx = DWORD(tabpc+8)-DWORD(tabpc+4)+1;
			for (; idx > 0; idx--) {
				SET_STACKPOINTER(pc+DWORD(tabpc+idx*4+8), sp + 1);
			}
			SET_STACKPOINTER(pc+DWORD(tabpc), sp + 1);
			pc = tabpc + (DWORD(tabpc+8)-DWORD(tabpc+4)+1+3) * 4;
			unreachable = true;
			break;

		case IRETURN:
		case LRETURN:
		case FRETURN:
		case DRETURN:
		case ARETURN:
		case RETURN:
		case ATHROW:
			unreachable = true;
			break;

		case GETSTATIC:
			getField((constIndex)lclw, class, &finfo);
			if (finfo.field == 0) {
				STKPUSH(1);
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]) {
			case 'L':
			case '[':
			case 'I':
			case 'Z':
			case 'S':
			case 'B':
			case 'C':
				STKPUSH(1);
				break;
			case 'F':
				STKPUSH(1);
				break;
			case 'J':
				STKPUSH(2);
				break;
			case 'D':
				STKPUSH(2);
				break;
			default:
				ABORT();
				break;
			}
			break;

		case PUTSTATIC:
			getField((constIndex)lclw, class, &finfo);
			if (finfo.field == 0) {
				STKPOP(1);
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]) {
			case 'L':
			case '[':
			case 'I':
			case 'Z':
			case 'S':
			case 'B':
			case 'C':
				STKPOP(1);
				break;
			case 'F':
				STKPOP(1);
				break;
			case 'J':
				STKPOP(2);
				break;
			case 'D':
				STKPOP(2);
				break;
			default:
				ABORT();
				break;
			}
			break;

		case GETFIELD:
			getField((constIndex)lclw, class, &finfo);
			if (finfo.field == 0) {
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]){
			case 'L':
			case '[':
			case 'I':
			case 'Z':
			case 'S':
			case 'B':
			case 'C':
				break;
			case 'F':
				break;
			case 'J':
				STKPUSH(1);
				break;
			case 'D':
				STKPUSH(1);
				break;
			default:
				ABORT();
				break;
			}
			break;

		case PUTFIELD:
			getField((constIndex)lclw, class, &finfo);
			if (finfo.field == 0) {
				STKPOP(2);
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]){
			case 'L':
			case '[':
			case 'I':
			case 'Z':
			case 'S':
			case 'B':
			case 'C':
				STKPOP(2);
				break;
			case 'F':
				STKPOP(2);
				break;
			case 'J':
				STKPOP(3);
				break;
			case 'D':
				STKPOP(3);
				break;
			default:
				ABORT();
				break;
			}
			break;

		case INVOKEVIRTUAL:
		case INVOKESPECIAL:
			if (getMethodSignatureClass((constIndex)lclw, class, (INSN(pc) == INVOKEVIRTUAL ? 0 : 1), &call) == 0) {
				/* We can tolerate some errors here */
				if (!postedExceptionIs("java.lang.NoClassDefFoundError") && !postedExceptionIs("java.lang.NoSuchMethodError")) {
					goto error;
				}
				clearPostedException();
			}

			STKPOP(call.in+1);
			STKPUSH(call.out);
			break;

		case INVOKEINTERFACE:
			if (getMethodSignatureClass((constIndex)lclw, class, 2, &call) == 0) {
				/* We can tolerate some errors here */
				if (!postedExceptionIs("java.lang.NoClassDefFoundError") && !postedExceptionIs("java.lang.NoSuchMethodError")) {
					goto error;
				}
				clearPostedException();
			}

			STKPOP(call.in+1);
			STKPUSH(call.out);
			break;

		case INVOKESTATIC:
			if (getMethodSignatureClass((constIndex)lclw, class, 0, &call) == 0) {
				/* We can tolerate some errors here */
				if (!postedExceptionIs("java.lang.NoClassDefFoundError") && !postedExceptionIs("java.lang.NoSuchMethodError")) {
					goto error;
				}
				clearPostedException();
			}

			STKPOP(call.in);
			STKPUSH(call.out);
			break;

		case MULTIANEWARRAY:
			STKPOP(INSN(pc+3) - 1);
                        if (getClass((constIndex)lclw, class, 0) == 0) {
				clearPostedException();
			}
			break;

		case NEW:
		case ANEWARRAY:
		case CHECKCAST:
		case INSTANCEOF:
                        if (getClass((constIndex)lclw, class, 0) == 0) {
				clearPostedException();
			}
			break;

		case WIDE:
			wide = true;
			break;
		}

		/* Adjust stack pointer */
		sp += spd;
		pc += pcd;

	} while (!unreachable && pc < currBB->next->start);

	currBB->end = pc;
	if (!unreachable) {
		currBB->next->flags |= FLAG_NORMALFLOW;
		currBB->next->stackPointer = sp;
	}

	/* Update local information */
	for (idx = 0; idx < codeInfo->localsz; idx++) {
		if (activeFrame[idx].used == 0) {
			continue;
		}
		l = &codeInfo->localuse[idx];
		if (opc < l->first) {
			l->first = opc;
		}
		if (opc > l->last) {
			l->last = opc;
		}
		if (activeFrame[idx].modified != 0 && opc > l->write) {
			l->write = opc;
		}
		l->use++;
		l->type |= activeFrame[idx].type;
	}

	/* Block verified */
	currBB->flags |= FLAG_DONEVERIFY;

	/* If verification failed - post an VerifyError */
	if (failed) {
		postException("java.lang.VerifyError", CLASS_CNAME(class));
	}
	return (failed);
error:
	return (true);
}

/**
 * Insert a new basic block into the basic block list.  This list
 * is ordered for easy traversal later.
 */
static
basicBlockInfo*
insertBB(codeinfo* codeInfo, int start)
{
	basicBlockInfo* ptr;
	basicBlockInfo** pptr;
	basicBlockInfo* newBB;

	ptr = 0;
	for (pptr = &codeInfo->bbinfo; (ptr = *pptr) != 0; pptr = &ptr->next) {
		if (start < ptr->start) {
			break;
		}
		else if (start == ptr->start) {
			return (ptr);	/* Duplicate */
		}
	}
	newBB = gc_malloc_fixed_typed(sizeof(basicBlockInfo), GC_ALLOC_ANALYSE);
	newBB->next = ptr;
	*pptr = newBB;

	newBB->start = start;
	newBB->stackPointer = 0xFFFF;

	return (newBB);
}

/*
 * Tidy up after verfication data has been finished with.
 */
void
tidyVerifyMethod(codeinfo** pcodeInfo)
{
	codeinfo* codeInfo;
	basicBlockInfo* ptr;
	basicBlockInfo* nptr;

	codeInfo = *pcodeInfo;
	*pcodeInfo = 0;

	if (codeInfo == 0) {
		return;
	}
	if (codeInfo->localuse != 0) {
		freeLocalUse(codeInfo->localuse);
	}
	for (ptr = codeInfo->bbinfo; ptr != 0; ptr = nptr) {
		nptr = ptr->next;
		gc_free_fixed(ptr);
	}
	freeCodeInfo(codeInfo);
}
