/*
 * engine.c
 * Define the common engine components.
 *
 * Copyright (c) 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-signal.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "jtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "code.h"
#include "exception.h"
#include "baseClasses.h"
#include "lookup.h"
#include "threads.h"
#include "locks.h"
#include "itypes.h"
#include "external.h"
#include "soft.h"
#include "md.h"
#include "engine.h"
#include "support.h"
#include "jnidefs.h"

#include "jdebug.h"
#include <limits.h>

#if defined(INTERPRETER)
extern void interpreterStart(void);
extern void interpreterEnd(void);
extern Method* interpreterMethod;
#endif

extern void Kaffe_JNIStart(void);
extern void Kaffe_JNIEnd(void);
extern Method* jniMethod;

typedef struct ActiveMethod {
	Method*			meth;
	Hjava_lang_Class*	clazz;
	struct ActiveMethod*	next;
} ActiveMethod;
static ActiveMethod* activeMethods;
static ActiveMethod* freeMethods;

#define	ALLOCSIZE	1024

stackTraceInfo*
buildStackTrace(void)
{
	return (buildStackTraceFrom(0, 0));
}

stackTraceInfo*
buildStackTraceGC(void)
{
	return (buildStackTraceFrom(0, 1));
}

stackTraceInfo*
buildStackTraceFrom(void* start, int gcable)
{
	int cnt;
	stackFrame fm;
	stackTraceInfo* binfo;
	stackTraceInfo* info;

	memset(&fm, 0, sizeof(fm));

	/* Count the frames */
	if (start != 0) {
		fm.return_frame = start;
	}
	else {
		Kaffe_TI_nextFrame(&fm);
	}

	cnt = 0;
	do {
		cnt++;
		Kaffe_TI_nextFrame(&fm);
	} while (fm.return_frame != 0 && fm.return_pc != 0);

	/* Allocate space for trace info */
	if (gcable) {
		binfo = gc_malloc(sizeof(stackTraceInfo) * (cnt+1), GC_ALLOC_ESTACK);
	}
	else {
		binfo = gc_malloc_fixed(sizeof(stackTraceInfo) * (cnt+1));
	}
	if (binfo == 0) {
		return (0);
	}

	/* Determine trace information */
	info = binfo;

	memset(&fm, 0, sizeof(fm));

	if (start != 0) {
		fm.return_frame = start;
	}
	else {
		Kaffe_TI_nextFrame(&fm);
	}

	do {
		info->pc = (uintp)fm.return_pc;
		info->rframe = fm.return_frame;
		info++;
		Kaffe_TI_nextFrame(&fm);
	} while (fm.return_frame != 0 && fm.return_pc != 0);

	/* Terminate trace */
	info->meth = ENDOFTRACE;

	DBG_ACTION( vm_exception, {
	  int i;
	  DBG( vm_exception, ("Stack trace ---\n"));
	  for (i = 0; binfo[i].meth != ENDOFTRACE; i++) {
		DBG( vm_exception, (" pc 0x%08x meth 0x%08x obj 0x%08x arg0 0x%08x\n",
			   binfo[i].pc, binfo[i].meth, binfo[i].object, binfo[i].arg0));
	  }
	});

	return (binfo);
}

void
freeStackTrace(stackTraceInfo* trace)
{
	gc_free_fixed(trace);
}

/*
 * Determine which method this PC is within.
 */
bool
findFrameMethod(stackTraceInfo* frame)
{
	ActiveMethod* curr;
	uintp pc;

	pc = frame->pc;

	if (pc > (uintp)Kaffe_JNIStart && pc < (uintp)Kaffe_JNIEnd) {
		frame->meth = jniMethod;
		frame->clazz = 0;
		return (true);
	}

#if defined(INTERPRETER)
	if (pc > (uintp)interpreterStart && pc < (uintp)interpreterEnd) {
		frame->meth = interpreterMethod;
		frame->clazz = interpreterMethod->class;
                GC_WRITE( frame, frame->clazz );
		return (true);
	}
	for (curr = activeMethods; curr != 0; curr = curr->next) {
		bytecodeInfo* info = curr->meth->ncode;
		if ((uint8*)pc >= info->code && pc - info->code < info->codelen) {
			frame->meth = curr->meth;
			frame->clazz = curr->clazz;
                        GC_WRITE( frame, frame->clazz );
			return (true);
	}
#else /* TRANSLATOR */
	for (curr = activeMethods; curr != 0; curr = curr->next) {
	        if (pc >= (uintp)METHOD_NATIVECODE(curr->meth) && pc < (uintp)METHOD_NATIVECODE_END(curr->meth)) {
			frame->meth = curr->meth;
			frame->clazz = curr->clazz;
                        GC_WRITE( frame, frame->clazz );
			return (true);
		}
	}
#endif
	frame->meth = UNKNOWN_METHOD;
	return (false);
}

/*
 * Make a method active.  We can then find it when backtracing
 * during exceptions.
 */
int
makeMethodActive(Method* meth, Hjava_lang_Class* clazz, void* spc, void* epc)
{
	ActiveMethod* entry;

	entry = freeMethods;
	if (entry == 0) {
		int i;
		entry = gc_malloc_fixed(ALLOCSIZE * sizeof(ActiveMethod));
		if (entry == 0) {
			return (0);
		}
		for (i = ALLOCSIZE - 1; i > 0; i--) {
			entry[i].next = freeMethods;
			freeMethods = &entry[i];
		}
		entry = &entry[0];
	}
	else {
		freeMethods = entry->next;
	}
	entry->meth = meth;
	entry->clazz = clazz;
	entry->next = activeMethods;
	activeMethods = entry;

	return (1);
}

/*
 * Make a method inactive before the class is unloaded.
 */
void
makeMethodInactive(Method* meth)
{
	ActiveMethod** curr;
	ActiveMethod* m;

	for (curr = &activeMethods; *curr != 0; curr = &(*curr)->next) {
		if ((*curr)->meth == meth) {
			m = (*curr);
			(*curr) = m->next;
			m->next = freeMethods;
			freeMethods = m;
			break;
		}
	}
}

void
printStackTrace(struct Hjava_lang_Throwable* o, struct Hjava_lang_Object* p, int includeNative)
{
	int i;
	stackTraceInfo* info;
	Method* meth;
	uintp pc;
	int32 linenr;
	uintp linepc;
	char buf[200];
	int len;
	int32 j;
	Hjava_lang_Object* str;
	jchar* cptr;
	Hjava_lang_Class* clazz;
	int start;

	info = (stackTraceInfo*)unhand(o)->backtrace;
	if (info == 0) {
		return;
	}
	start = 0;
	for (i = 0; info[i].meth != ENDOFTRACE; i++) {
		if (info[i].meth == 0 && findFrameMethod(&info[i]) == true) {
			Kaffe_EngineInterface_processExceptionMethod(&info[i]);
		}
		if (info[i].clazz == OBJECT_CLASS(&o->base)) {
			start = i+1;
		}
	}
	for (i = start; info[i].meth != ENDOFTRACE; i++) {
		clazz = info[i].clazz;
		meth = info[i].meth; 
		pc = info[i].pc;
		if (meth == UNKNOWN_METHOD || meth == jniMethod) {
			if (includeNative == 0) {
				continue;
			}
			sprintf(buf, "\tat %.80s(pc %p)",
				(meth == jniMethod ? "jnicall" : "native"),
				(void*)pc);
			goto print;
		}
#if !defined(OPTIMIZE_MEMORY) || defined(DEBUG)
		linepc = 0;
		linenr = -1;
		if (meth->lines != 0) {
			uintp pc_offset = (uintp)METHOD_NATIVECODE(meth);
			for (j = 0; j < (int)meth->lines->length; j++) {
				pc_offset += meth->lines->entry[j].pc_offset;
				if (pc >= pc_offset && linepc < pc_offset) {
					linenr = meth->lines->entry[j].line_nr;
					linepc = pc_offset;
				}
			}
		}
#endif
#if !defined(HAVE_SPRINTF)
		strcpy(buf, "\tat ");
		strcat(buf, CLASS_CNAME(clazz));
		strcat(buf, ".");
		strcat(buf, meth->name->data);
#else
#if !defined(OPTIMIZE_MEMORY) || defined(DEBUG)
		if (linenr != -1) {
			char* fname = "?";
			if (clazz->filename != 0) {
				fname = strrchr(clazz->filename, '/');
				if (fname == 0) {
					fname = clazz->filename;
				}
				else {
					fname++;
				}
			}
			sprintf(buf, "\tat %.80s.%.80s(%s:%d)",
				CLASS_CNAME(clazz), meth->name,
				fname, linenr);
		}
		else
#endif
		if (CLASS_IS_PRECOMPILED(clazz)) {
			sprintf(buf, "\tat %.80s.%.80s(Compiled: pc %p)",
				CLASS_CNAME(clazz),
				meth->name, (void*)pc);
		}
		else {
			sprintf(buf, "\tat %.80s.%.80s(pc %p)",
				CLASS_CNAME(clazz),
				meth->name, (void*)pc);
		}
#endif
		print:;
		if (p != 0) {
			len = strlen(buf);
			str = newArray(&_Jv_charClass, len);
			cptr = (jchar*)ARRAY_DATA(HArrayOfChar, str);
			for (j = (int32)len;  --j >= 0; ) {
				cptr[j] = (unsigned char)buf[j];
			}
			if (printlnMethod == 0) {
				printlnMethod = lookupClassMethod(PrintWriterClass, "println", "([C)V");
			}
			Kaffe_CallVoidMethod(0, p, printlnMethod, str);
		}
		else {
			kprintf("%s\n", buf);
		}
	}
	if (p != 0) {
		if (flushMethod == 0) {
			flushMethod = lookupClassMethod(PrintWriterClass, "flush", "()V");
		}
		Kaffe_CallVoidMethod(0, p, flushMethod);
	}
}
