/*
 * exception.c
 * Handle exceptions for the interpreter or translator.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#define	CDBG(s)
#define	PDBG(s)

#include "config-std.h"
#include "config-signal.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "jtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "code.h"
#include "exception.h"
#include "baseClasses.h"
#include "lookup.h"
#include "threads.h"
#include "itypes.h"
#include "external.h"
#include "soft.h"
#include "md.h"
#include "locks.h"
#include "engine.h"
#include "threads.h"
#include "support.h"
#include "jnidefs.h"

#include "jdebug.h"

extern Method* jniMethod;
extern Hjava_lang_Thread* garbageman;
extern Hjava_lang_Thread* finalman;

/* Some systems need special setups for the exception handling */
#if !defined(EXCEPTIONSTART)
#define	EXCEPTIONSTART()
#endif
#if !defined(EXCEPTIONEND)
#define	EXCEPTIONEND()
#endif

void
dispatchException(void)
{
	char* cname;
	Hjava_lang_Class* class;
	stackTraceInfo* frame;
	Hjava_lang_Thread* tid;
	Hjava_lang_Throwable* eobj;
	stackTraceInfo* trace;

	tid = Kaffe_TI_currentJava();

	/* Get exception and clear pending flag */
	eobj = getPostedException();
	unhand(tid)->pending = 0;

	class = OBJECT_CLASS(&eobj->base);
	cname = CLASS_CNAME(class);

	DBG( vm_exception, ("dispatch %s\n", cname));
	trace = buildStackTrace();
        GC_WRITE(eobj, trace);

	/* Search down exception stack for a match */

#if defined(INTERPRETER)
	/* In the interpreter we must process the stack record now otherwise
	 * we loose some of the vital information.
	 */
	for (frame = trace; frame->meth != ENDOFTRACE ; frame++) {
		if (frame->meth == 0) {
			if (findFrameMethod(frame) == true) {
				Kaffe_EngineInterface_processExceptionMethod(frame);
			}
		}
	}
#endif

	for (frame = trace; frame->meth != ENDOFTRACE ; frame++) {
		Method* method;
		uintp handler;

		if (frame->meth == 0) {
			/* We must now submit the found method to it's
			 * execution engine incase there's any transformations
			 * to be done.  For example, we only have one method
			 * which represents the 'interpreter' yet this maps
			 * onto many Java methods.  It's also up to the
			 * execution engine to add any addition data that
			 * might be useful.
			 */
			if (findFrameMethod(frame) == false) {
				continue;
			}

			Kaffe_EngineInterface_processExceptionMethod(frame);
		}
		else if (frame->meth == UNKNOWN_METHOD) {
			continue;
		}
		method = frame->meth;
		DBG( vm_exception, ("in %s.%s\n", frame->clazz->name,
		                                  method->name));
		if (method == jniMethod) {
			Kaffe_JNIExceptionHandler(frame, eobj);
		}
		/* Find handler and, if found, dispatch exception */
		handler = findExceptionHandlerInMethod(frame->clazz, method, frame->pc, class);
		/* If we get an error in the handler, we switch to looking
		 * for the error and continue (maybe?).
		 */
		if (handler == (uintp)-1) {
			eobj = getPostedException();
			clearPostedException();
		}
		else if (handler != 0) {
			stackTraceInfo info = *frame;
			freeStackTrace(trace);
		        DBG( vm_exception, ("call handler m=%p h=%p (nc=%p) f=%p\n",
			    method, handler, METHOD_NATIVECODE(method), frame));

			Kaffe_EngineInterface_callExceptionHandler(method, handler, &info);
		}
	}

	freeStackTrace(trace);

	/* Report bad exception threads */
	if (Kaffe_TI_currentJava() == garbageman) {
		kprintf("FATAL ERROR: Garbage collector exception\n");
	}
	else if (Kaffe_TI_currentJava() == finalman) {
		kprintf("FATAL ERROR: Finalizer exception\n");
	}

	exitThread();
}

/*
 * Post an exception for the current thread - the exception will be thrown
 * later.
 */
void
postException(char* name, char* mess)
{
	Hjava_lang_Thread* tid;
	Hjava_lang_Throwable* trw;
	Hjava_lang_Throwable* otrw;

PDBG(	kprintf("POST: %s(%s)\n", name, mess);	)

	tid = Kaffe_TI_currentJava();
	if (tid == 0) {
		/* Too early for errors - we have no threads yet!!! */
		kprintf("FATAL ERROR IN INITIALIZATION\n\t%s: %s\n", name, (mess == 0 ? "" : mess));
		EXIT(1);
	}

CDBG(	if (strcmp(name, "java.lang.NoClassDefFoundError") == 0) {
		kprintf("Missing class %s\n", mess);
	})

	otrw = unhand(tid)->exceptObj;
	if (strcmp(name, "java.lang.StackOverflowError") == 0) {
		unhand(tid)->exceptObj = unhand(tid)->stackOverflowError;
                GC_WRITE(tid, unhand(tid)->exceptObj); /* tid->exceptObj */
	}
	else {
		unhand(tid)->exceptObj = 0;

		/* Save new exception name and message */
		unhand(tid)->ename = (struct Hkaffe_util_Ptr*)name;
                GC_WRITE(tid, name); /* tid->ename */
		if (mess != 0) {
			if (mess[0] == 0) {
				unhand(tid)->emess = 0;
			}
			else {
				unhand(tid)->emess = (struct Hkaffe_util_Ptr*)mess;
                                GC_WRITE(tid, mess); /* tid->emess */
			}
		}
		/* If we have no message but we do have an old message we
		 * must reuse it.
		 */
		else if (otrw != 0 && unhand(otrw)->detailMessage != 0) {
			trw = getPostedException();
			unhand(trw)->detailMessage = unhand(otrw)->detailMessage;
                        GC_WRITE(trw, unhand(trw)->detailMessage); /* trw->detailMessage */
		}
		else {
			unhand(tid)->emess = 0;
		}
	}

	unhand(tid)->pending = 1;
}

/*
 * Is the posted exception the given one.
 * NB. We just do a simple name comparison so we don't handle inheritence.
 */
int
postedExceptionIs(char* name)
{
	if (strcmp(CLASS_CNAME(OBJECT_CLASS(&getPostedException()->base)), name) != 0) {
		return (0);
	}
	else {
		return (1);
	}
}

/*
 * Post an out-of-memory exception - the exception will be thrown later.
 */
void
postOutOfMemory(void)
{
	postException("java.lang.OutOfMemoryError", "");
}

/*
 * Post an exception in exception.
 */
void
postExceptionInException(char* name)
{
	Hjava_lang_Thread* tid;
	Hjava_lang_Throwable* trw;
	jclass cls;

	tid = Kaffe_TI_currentJava();

	cls = Kaffe_FindClass(0, name);
	trw = (Hjava_lang_Throwable*)Kaffe_NewObject(0, cls, Kaffe_GetMethodID(0, cls, "<init>", "(Ljava/lang/Throwable;)V"), getPostedException());
	unhand(tid)->exceptObj = trw;
	GC_WRITE(tid, trw); /* tid->exceptObj */
	unhand(tid)->pending = 1;
}

/*
 * Clear any posted exception.
 */
void
clearPostedException(void)
{
	Hjava_lang_Thread* tid;

	tid = Kaffe_TI_currentJava();
	if (tid != 0) {
		unhand(tid)->pending = 0;
	}
}

/*
 * Throw any posted exception.
 */
void
throwPostedException(void)
{
	Hjava_lang_Thread* tid;
	Hjava_lang_Throwable* trw;
	stackTraceInfo* info;

	tid = Kaffe_TI_currentJava();

	if (unhand(tid)->pending != 0) {
		trw = getPostedException();
		DBG( vm_exception, ("throw posted exception %p", trw));
		if (unhand(trw)->backtrace == 0) {
			info = buildStackTraceGC();
			unhand(trw)->backtrace = (Hjava_lang_Object*)info;
			GC_WRITE(trw, unhand(trw)->backtrace); /* trw->backtrace */
		}
		dispatchException();
	}
}

void
postExternalException(Hjava_lang_Throwable* throwable)
{
	Hjava_lang_Thread* tid;

PDBG(	kprintf("POSTE: %s\n", CLASS_CNAME(OBJECT_CLASS(&throwable->base))); )

	tid = Kaffe_TI_currentJava();
	unhand(tid)->exceptObj = throwable;
	GC_WRITE(tid, throwable); /* tid->exceptObj */
	unhand(tid)->pending = 1;
}

/*
 * Throw an exception generated by the java code.
 */
void
postAndThrowExternalException(Hjava_lang_Throwable* throwable)
{
	postExternalException(throwable);
	throwPostedException();
}


/*
 * Throw an out-of-memory error immediately.
 */
void
postAndThrowOutOfMemory(void)
{
	postOutOfMemory();
	throwPostedException();
}

/*
 * Throw an exception NOW.
 */
void
postAndThrowException(char* name, char* mess)
{
	postException(name, mess);
	throwPostedException();
}

/*
 * Get the current exception.
 */
Hjava_lang_Throwable*
getPostedException(void)
{
	Hjava_lang_Thread* tid;
	Hjava_lang_Throwable* trw;
	jboolean pending;
	jclass ecls;

	tid = Kaffe_TI_currentJava();
	trw = unhand(tid)->exceptObj;
	if (trw == 0 && unhand(tid)->ename != 0) {
		pending = unhand(tid)->pending; /* Save this */
		unhand(tid)->pending = 0;
		if (strcmp ((char*)unhand(tid)->ename, "java.lang.OutOfMemoryError") != 0) {
			ecls =  Kaffe_FindClass(0, (char*)unhand(tid)->ename);
			trw = (Hjava_lang_Throwable*)Kaffe_NewObject(0, ecls, Kaffe_GetMethodID(0, ecls, "<init>", "()V"));
		}
		if (trw == 0) {
			trw = unhand(tid)->outOfMemoryError;
		}
		else if (unhand(tid)->emess != 0) {
			unhand(trw)->detailMessage = Kaffe_NewStringUTF(0, (char*)unhand(tid)->emess);
                        GC_WRITE(trw, unhand(trw)->detailMessage); /* trw->detailMessage */
		}
		unhand(tid)->exceptObj = trw;
		unhand(tid)->pending = pending;
		GC_WRITE(tid, trw); /* tid->exceptObj */
	}
	return (trw);
}

/*
 * Used to generate exception for unimplemented features.
 */
void
unimp(char* mess)
{
	postAndThrowException("kaffe.util.NotImplemented", mess);
}
