/*
 * external.c
 * Handle method calls to other languages.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	NDBG(s)
#define	LDBG(s)
#define	EDBG(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "config-loadlib.h"
#include "gtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "slots.h"
#include "external.h"
#include "exception.h"
#include "support.h"
#include "system.h"
#include "md.h"
#include "paths.h"
#include "threads.h"
#include "engine.h"
#include "callMethodInfo.h"
#include "baseClasses.h"
#include "jdebug.h"

#if !defined(_STDERR)
#define _STDERR stderr
#endif

typedef struct _NativeLibrary {
	struct _NativeLibrary*	next;
	char*			name;
	int			ref;
	LIBRARYHANDLE		desc;
} NativeLibrary;

/* User directory - accessed by various external libraries */
char* userDir;

/* User properties */
userProperty* userProperties;

/* Native function table */
typedef struct _nativeFunction {
	char*	name;
	void*	func;
} nativeFunction;

#include "../../libraries/clib/external_native.h"

static nativeFunction null_funcs[1];
static nativeFunction* native_funcs = null_funcs;

#define	NATIVE_FUNC_INCREMENT	256

/*
 * Error stub function.  Point unresolved link errors here to avoid
 * problems.
 */
void*
error_stub(void)
{
	postAndThrowException("java.lang.UnsatisfiedLinkError", "");
	return (0);
}

void
initNative(void)
{
	LIBRARYINIT();
	loadNativeLibrary(LIBRARYPREFIX "native" LIBRARYSUFFIX, 0);
}

char*
loadNativeLibrary(char* lib, Hjava_lang_ClassLoader* loader)
{
	char* name;
	NativeLibrary* hand;
	int i;

	/* Null loader - must be the system one then */
	if (loader == 0) {
		loader = &systemClassLoader;
	}

	/* Check builtin libraries */
	for (i = 0; default_libraries[i] != 0; i++) {
		if (strcmp(lib, default_libraries[i]) == 0) {
			return (0);
		}
	}

	/* Look for a match in the current loader */
	for (hand = loader->nativeLibs; hand != 0; hand = hand->next) {
		if (strcmp(hand->name, lib) == 0) {
			hand->ref++;
			return (0);
		}
	}

	hand = gc_malloc_fixed(sizeof(NativeLibrary));
	if (hand == 0) {
		postOutOfMemory();
		return ("out of memory");
	}

	/* Open the library */
DBG( vm_lib,("Loading library %s\n", lib));
        LIBRARYLOAD(hand->desc, lib);

	if (hand->desc == 0) {
	    
DBG( vm_lib,("Library load error %s\n", LIBRARYERROR()));
LDBG(		kprintf("Library load failed: %s\n", LIBRARYERROR());	)
		return (LIBRARYERROR());
	}

	name = gc_malloc_fixed(strlen(lib) + 1);
	if (name == 0) {
		postOutOfMemory();
		return ("out of memory");
	}
	hand->ref = 1;
	hand->name = name;
	strcpy(name, lib);

	hand->next = loader->nativeLibs;
	loader->nativeLibs = hand;

	return (0);
}

void
unloadNativeLibraries(Hjava_lang_ClassLoader* loader)
{
	NativeLibrary* hand;
	NativeLibrary* nhand;

	for (hand = loader->nativeLibs; hand != 0; hand = nhand) {
		nhand = hand->next;
		LIBRARYUNLOAD(hand->desc);
		gc_free_fixed(hand);
	}
}

/*
 * Get pointer to symbol from symbol name.
 */
static
void*
loadNativeLibrarySym(char* name, Hjava_lang_ClassLoader* loader)
{
	int i;
	void* func;

	if (loader == 0) {
		loader = &systemClassLoader;
	}

	/* Try registered */
	if (loader == &systemClassLoader) {
		for (i = 0; default_natives[i].name != 0; i++) {
			if (strcmp(default_natives[i].name, name) == 0) {
				return (default_natives[i].func);
			}
		}

		/* Try built-ins */
		for (i = 0; native_funcs[i].name != 0; i++) {
			if (strcmp(native_funcs[i].name, name) == 0) {
				DBG( vm_lib,("Loading syms for built in library %s\n", name));
				return (native_funcs[i].func);
			}   
		}
	}

	/* Try share libraries */
	DBG( vm_lib,("Loading syms for dynamic library %s\n", name));
	func = 0;
	LIBRARYFUNCTION(func, name, loader);
	if(func != NULL)
	  {
	    return func;
	  }
	
	// look in the system class loader as well
	if (loader != &systemClassLoader) {
	  return loadNativeLibrarySym(name, &systemClassLoader);
	}
	return 0;
}

/*
 * Register an user function statically linked in the binary.
 */
int
addNativeMethod(char* name, void* func)
{
	static int funcs_nr = 0;
	static int funcs_max = 0;

	DBG( vm_lib,("Registering an user function statically linked\n"));
	/* If we run out of space, reallocate */
	if (funcs_nr + 1 >= funcs_max) {
		funcs_max += NATIVE_FUNC_INCREMENT;
		if (native_funcs != null_funcs) {
			native_funcs = gc_realloc_fixed(native_funcs, funcs_max * sizeof(nativeFunction));
		}
		else {
			native_funcs = gc_malloc_fixed(NATIVE_FUNC_INCREMENT * sizeof(nativeFunction));
		}
		if (native_funcs == 0) {
			return (0);
		}
	}
	native_funcs[funcs_nr].name = gc_malloc_fixed(strlen(name) + 1);
	if (native_funcs[funcs_nr].name == 0) {
		return (0);
	}
	strcpy(native_funcs[funcs_nr].name, name);
	native_funcs[funcs_nr].func = func;
	funcs_nr++;
	native_funcs[funcs_nr].name = 0;
	native_funcs[funcs_nr].func = 0;
	return (1);
}

static
void
strcatJNI(char* to, char* from)
{
	char* ptr;

	ptr = &to[strlen(to)];
	for (; *from != 0; from++) {
		switch (*from) {
		case '(':
			/* Ignore */
			break;
		case ')':
			/* Terminate here */
			goto end;
		case '_':
			*ptr++ = '_';
			*ptr++ = '1';
			break;
		case ';':
			*ptr++ = '_';
			*ptr++ = '2';
			break;
		case '[':
			*ptr++ = '_';
			*ptr++ = '3';
			break;
		case '.':
		case '/':
			*ptr++ = '_';
			break;
		default:
			*ptr++ = *from;
			break;
		}
	}

	end:;
	*ptr = 0;
}

static
void*
getNativeMethod(Method* meth, Hjava_lang_Class* class)
{
	char name[1024];
	void* func;

	/* Build the simple JNI name for the method */
	strcpy(name, "_Java_");
	strcatJNI(name, CLASS_CNAME(class));
	strcat(name, "_");
	strcatJNI(name, meth->name);

	DBG( vm_lib,("getNativeMethod : Loading syms %s from a library\n", name));
	/* Lookup name without and with underscore */
	func = loadNativeLibrarySym(&name[1], class->loader);
	if (func == 0) {
		func = loadNativeLibrarySym(name, class->loader);
	}
	if (func == 0) {
		/* Try the long signatures */
		DBG( vm_lib,("getNativeMethod : Loading syms from library %s FAILED!!! Try long signature now\n", name));
		strcat(name, "__");
		strcatJNI(name, meth->signature);
		func = loadNativeLibrarySym(&name[1], class->loader);
		if (func == 0) {
			func = loadNativeLibrarySym(name, class->loader);
		}
	}
	return (func);
}

void
native(Method* meth, Hjava_lang_Class* class)
{
	void* func;

	func = getNativeMethod(meth, class);
	if (func == 0 || Kaffe_EngineInterface_setupJNIcall(meth, class, func) == 0) {
EDBG(		kprintf("Failed to locate native function:\n\t%s.%s%s\n", CLASS_CNAME(class), meth->name->data, meth->signature->data);	)
		SET_METHOD_NATIVECODE(meth, (void*)error_stub, 0);
		SET_METHOD_FLUSHED(meth);
	}
}

/*
 * Determine whether a method is a native call. We stash the function
 * itself in the usual place.  If we fail, we probably don't have the
 * shared library loaded yet so no problem.
 */
int
checkNativeCall(Method* meth, Hjava_lang_Class* class)
{
	void* func;

	/* Try to locate the nature function using the JNI interface */
	func = getNativeMethod(meth, class);
	if (func != 0) {
NDBG(		kprintf("Native JNI func = 0x%x\n", func);		)
		/* Fill it in */
		SET_METHOD_NATIVECODE(meth, func, 0);
		SET_METHOD_FLUSHED(meth);
		return (1);
	}
	else {
		/* Okay, we failed to find the function - fall back on default
		 * behaviour - we may not have loaded the shared library yet!
		 */
		return (0);
	}
}

/*
 * Print messages.
 */
void
kprintf(const char* mess, ...)
{
	va_list argptr;

	va_start(argptr, mess);
	if (Kaffe_JavaVMArgs[0].vfprintf != 0) {
		(*Kaffe_JavaVMArgs[0].vfprintf)(_STDERR, mess, argptr);
	}
	va_end(argptr);
}

/*
 * Generic routine to call a native or Java method (array style).
 */
void
callMethodA(Method* meth, void* func, void* obj, jvalue* args, jvalue* ret)
{
	char* sig;
	int i;
	int s;
	callMethodInfo call;
	jvalue in[MAXMARGS];
	jvalue tmp;
	int j;

	if (ret == 0) {
		ret = &tmp;
	}
	sig = meth->signature;
	i = 0;
	s = 0;

	/* If this method isn't static, we must insert the object as
	 * the first argument.  To do this we copy down the argument
	 * array and stick the object at the beginning.
 	 */
	if ((meth->accflags & ACC_STATIC) == 0) {
		call.callsize[0] = PTR_TYPE_SIZE / SIZEOF_INT;
		s += PTR_TYPE_SIZE / SIZEOF_INT;
		call.calltype[0] = 'L';
		in[0].l = obj;
		if (args != 0) {
			memcpy(&in[1], args, sizeof(jvalue) * (MAXMARGS-1));
		}
		args = in;
		i++;
	} else {
                /* Must copy since we may shift args later */
		if (args != 0) {
			memcpy(&in[0], args, sizeof(jvalue) * (MAXMARGS));
		}
		args = in;
        }

	/* Skip leading ( and terminate at the ) */
	for (j = 1; sig[j] != ')'; j++, i++) {
		switch (sig[j]) {
		case 'I':
		case 'Z':
		case 'S':
		case 'B':
		case 'C':
		case 'F':
			call.calltype[i] = sig[j];
#if defined(HAVE_NO_FLOATING_POINT)
			if (call.calltype[i] == 'F') {
				call.calltype[i] = 'I';
			}
#endif
			call.callsize[i] = 1;
			s += 1;
			break;

		case 'D':
		case 'J':
#if defined(HAVE_NO_FLOATING_POINT)
			call.calltype[i] = 'J';
#else
			call.calltype[i] = sig[j];
#endif
			call.callsize[i] = 2;
			i++;
			call.callsize[i] = 0;
			s += 2;
			break;
		case 'L':
			call.calltype[i] = 'L';	/* Looks like an object */
			call.callsize[i] = PTR_TYPE_SIZE / SIZEOF_INT;
			s += PTR_TYPE_SIZE / SIZEOF_INT;
			while (sig[++j] != ';')
				;
			break;
		case '[':
			call.calltype[i] = 'L';	/* Looks like an object */
			call.callsize[i] = PTR_TYPE_SIZE / SIZEOF_INT;
			s += PTR_TYPE_SIZE / SIZEOF_INT;
			while (sig[++j] == '[')
				;
			if (sig[j] == 'L') {
				while (sig[++j] != ';')
					;
			}
			break;
		default:
			ABORT();
		}
	}

	/* Put the stack limit as the last argument */
	call.calltype[i] = 'L';
	call.callsize[i] = PTR_TYPE_SIZE / SIZEOF_INT;
	in[i].l = getCurrentStackLimit();
	s += PTR_TYPE_SIZE / SIZEOF_INT;
	i++;

	/* Return info */
	call.rettype = sig[j+1];
#if defined(HAVE_NO_FLOATING_POINT)
	if (call.rettype == 'F') {
		call.rettype = 'I';
	}
	else if (call.rettype == 'D') {
		call.rettype = 'J';
	}
#endif
	if (call.rettype == 'L' || call.rettype == '[') {
		call.rettype = 'L';
		call.retsize = PTR_TYPE_SIZE / SIZEOF_INT;
	}
	else if (call.rettype == 'V') {
		call.retsize = 0;
	}
	else if (call.rettype == 'D' || call.rettype == 'J') {
		call.retsize = 2;
	}
	else {
		call.retsize = 1;
	}

	/* Call info and arguments */
	call.function = func;
	call.nrargs = i;
	call.argsize = s;
	call.args = args;
	call.ret = ret;

	/* Make the call */
	Kaffe_EngineInterface_callInterEngine(&call);
}

/*
 * Generic routine to call a native or Java method (varargs style).
 */
void
callMethodV(Method* meth, void* func, void* obj, va_list args, jvalue* ret)
{
	char* sig;
	int i;
	int s;
	callMethodInfo call;
	jvalue in[MAXMARGS];
	jvalue tmp;
	int j;

	if (ret == 0) {
		ret = &tmp;
	}
	i = 0;
	s = 0;

	sig = meth->signature;

	/* If this method isn't static, we must insert the object as
	 * the first argument and get the function code.
 	 */
	if ((meth->accflags & ACC_STATIC) == 0) {
		call.callsize[i] = PTR_TYPE_SIZE / SIZEOF_INT;
		s += PTR_TYPE_SIZE / SIZEOF_INT;
		call.calltype[i] = 'L';
		in[i].l = obj;
		i++;
	}

	/* Skip leading ( and terminate at the ) */
	for (j = 1; sig[j] != ')'; j++, i++) {
		call.calltype[i] = sig[j];
		switch (call.calltype[i]) {
#if defined(HAVE_NO_FLOATING_POINT)
		case 'F':
#endif
		case 'I':
			call.callsize[i] = 1;
			in[i].i = va_arg(args, jint);
			s += 1;
			break;
		case 'Z':
			call.callsize[i] = 1;
			/* C will promote the jboolean to jint */
			in[i].z = (jboolean)va_arg(args, jint);
			s += 1;
			break;
		case 'S':
			call.callsize[i] = 1;
			/* C will promote the jshort to jint */
			in[i].s = (jshort)va_arg(args, jint);
			s += 1;
			break;
		case 'B':
			call.callsize[i] = 1;
			/* C will promote the jbyte to jint */
			in[i].b = (jbyte)va_arg(args, jint);
			s += 1;
			break;
		case 'C':
			call.callsize[i] = 1;
			/* C will promote the jchar to jint */
			in[i].c = (jchar)va_arg(args, jint);
			s += 1;
			break;
#if !defined(HAVE_NO_FLOATING_POINT)
		case 'F':
			call.callsize[i] = 1;
			/* C will promote the jfloat to a jdouble */
			in[i].f = (jfloat)va_arg(args, jdouble);
			s += 1;
			break;
		case 'D':
			call.callsize[i] = 2;
			in[i].d = va_arg(args, jdouble);
			i++;
			call.callsize[i] = 0;
			s += 2;
			break;
#endif
#if defined(HAVE_NO_FLOATING_POINT)
		case 'D':
#endif
		case 'J':
			call.callsize[i] = 2;
			in[i].j = va_arg(args, jlong);
			i++;
			call.callsize[i] = 0;
			s += 2;
			break;
		case 'L':
			call.callsize[i] = PTR_TYPE_SIZE / SIZEOF_INT;
			call.calltype[i] = 'L';	/* Looks like an object */
			in[i].l = va_arg(args, jref);
			s += PTR_TYPE_SIZE / SIZEOF_INT;
			while (sig[++j] != ';')
				;
			break;
		case '[':
			call.callsize[i] = PTR_TYPE_SIZE / SIZEOF_INT;
			call.calltype[i] = 'L';	/* Looks like an object */
			in[i].l = va_arg(args, jref);
			s += PTR_TYPE_SIZE / SIZEOF_INT;
			while (sig[++j] == '[')
				;
			if (sig[j] == 'L') {
				while (sig[++j] != ';')
					;
			}
			break;
		default:
			ABORT();
		}
	}

	/* Put the stack limit as the last argument */
	call.calltype[i] = 'L';
	call.callsize[i] = PTR_TYPE_SIZE / SIZEOF_INT;
	in[i].l = getCurrentStackLimit();
	s += PTR_TYPE_SIZE / SIZEOF_INT;
	i++;

	/* Return info */
	call.rettype = sig[j+1];
#if defined(HAVE_NO_FLOATING_POINT)
	if (call.rettype == 'F') {
		call.rettype = 'I';
	}
	else if (call.rettype == 'D') {
		call.rettype = 'J';
	}
#endif
	if (call.rettype == 'V') {
		call.retsize = 0;
	}
	else if (call.rettype == 'D' || call.rettype == 'J') {
		call.retsize = 2;
	}
	else if (call.rettype == 'L' || call.rettype == '[') {
		call.rettype = 'L';
		call.retsize = PTR_TYPE_SIZE / SIZEOF_INT;
	}
	else {
		call.retsize = 1;
	}

	/* Call info and arguments */
	call.function = func;
	call.nrargs = i;
	call.argsize = s;
	call.args = in;
	call.ret = ret;

	/* Make the call */
	Kaffe_EngineInterface_callInterEngine(&call);
}
