/*
 * jar.c
 * Handle JAR input files.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */


#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "jar.h"
#include "support.h"
#if defined(HAVE_LIBZ) && defined(HAVE_ZLIB_H)
#include <zlib.h>
#endif

#if !defined(KAFFEH)
#include "gc.h"
#else
#undef	gc_malloc_fixed_typed
#undef	gc_free_fixed
#define	gc_malloc_fixed_typed(S,T)	calloc(S, 1)
#define	gc_free_fixed(M)		free(M)
#endif

#include "jdebug.h"

static jarFile* openJars;

static
inline
int
jarFileNameHash(char *name)
{
	int hash = 0;
	while(*name !=0) {
		hash = (hash * 37) + *name;
		name++;
	}
	return hash & JARFILE_HASHTABLE_MASK;
}


/*
 * Read in a central directory record and return the record size.
 */
static
int
readCentralDirRecordSize(jarFile* file)
{
	jarCentralDirectoryRecord head;

	head.signature = Kaffe_JarRawInterface_read32(file);
	head.versionMade = Kaffe_JarRawInterface_read16(file);
	head.versionExtract = Kaffe_JarRawInterface_read16(file);
	head.flags = Kaffe_JarRawInterface_read16(file);
	head.compressionMethod = Kaffe_JarRawInterface_read16(file);
	head.lastModifiedTime = Kaffe_JarRawInterface_read16(file);
	head.lastModifiedDate = Kaffe_JarRawInterface_read16(file);
	head.crc = Kaffe_JarRawInterface_read32(file);
	head.compressedSize = Kaffe_JarRawInterface_read32(file);
	head.uncompressedSize = Kaffe_JarRawInterface_read32(file);
	head.fileNameLength = Kaffe_JarRawInterface_read16(file);
	head.extraFieldLength = Kaffe_JarRawInterface_read16(file);
	head.fileCommentLength = Kaffe_JarRawInterface_read16(file);
	head.diskNumberStart = Kaffe_JarRawInterface_read16(file);
	head.internalFileAttribute = Kaffe_JarRawInterface_read16(file);
	head.externalFileAttribute = Kaffe_JarRawInterface_read32(file);
	head.relativeLocalHeaderOffset = Kaffe_JarRawInterface_read32(file);

	if (head.signature != CENTRALHEADERSIGNATURE) {
		DBG( vm_jar, ("Bad central record signature: %x vs wanted %x\n", head.signature,  CENTRALHEADERSIGNATURE));
		file->error = "Bad central record signature";
		return (0);
	}

	Kaffe_JarRawInterface_seekHere(file, head.fileNameLength + head.extraFieldLength + head.fileCommentLength);

	/* Return space needed rounded up to nearest 4 bytes */
	return ((sizeof(jarEntry) + (head.fileNameLength + 1) + 3) & -4);
}

/*
 * Read in a central directory record.
 */
static
jarEntry*
readCentralDirRecord(jarFile* file)
{
	jarCentralDirectoryRecord head;
	jarEntry* ret;
	int len;

	head.signature = Kaffe_JarRawInterface_read32(file);
	head.versionMade = Kaffe_JarRawInterface_read16(file);
	head.versionExtract = Kaffe_JarRawInterface_read16(file);
	head.flags = Kaffe_JarRawInterface_read16(file);
	head.compressionMethod = Kaffe_JarRawInterface_read16(file);
	head.lastModifiedTime = Kaffe_JarRawInterface_read16(file);
	head.lastModifiedDate = Kaffe_JarRawInterface_read16(file);
	head.crc = Kaffe_JarRawInterface_read32(file);
	head.compressedSize = Kaffe_JarRawInterface_read32(file);
	head.uncompressedSize = Kaffe_JarRawInterface_read32(file);
	head.fileNameLength = Kaffe_JarRawInterface_read16(file);
	head.extraFieldLength = Kaffe_JarRawInterface_read16(file);
	head.fileCommentLength = Kaffe_JarRawInterface_read16(file);
	head.diskNumberStart = Kaffe_JarRawInterface_read16(file);
	head.internalFileAttribute = Kaffe_JarRawInterface_read16(file);
	head.externalFileAttribute = Kaffe_JarRawInterface_read32(file);
	head.relativeLocalHeaderOffset = Kaffe_JarRawInterface_read32(file);

	if (head.signature != CENTRALHEADERSIGNATURE) {
		DBG( vm_jar, ("Bad central record signature: %x vs wanted %x\n", head.signature,  CENTRALHEADERSIGNATURE));
		file->error = "Bad central record signature";
		return (0);
	}

	len = sizeof(jarEntry) + (head.fileNameLength + 1);
	ret = (jarEntry*)file->mem;
	file->mem += (len + 3) & -4;

	ret->next = 0;
	ret->fileName = (char*)(((uintp)ret) + sizeof(jarEntry));
	ret->compressionMethod  = head.compressionMethod;
        ret->compressedSize = head.compressedSize;
        ret->uncompressedSize = head.uncompressedSize;

	Kaffe_JarRawInterface_readBytes(file, (int)head.fileNameLength, ret->fileName);
	Kaffe_JarRawInterface_seekHere(file, head.extraFieldLength + head.fileCommentLength);

/*DBG(	vm_jar, ("Central record filename: %s\n", ret->fileName));*/

	/* Would be nice if we could store the direct offset of the file data
	 * (which is supposed to be relOffset + SIZEOFLOCALHEADER +
	 * fileNameLength + extraFieldLength), but extraFieldLength differs
	 * in the cdir entry and the local header. Even worse, it shows a
	 * 4 byte difference in zip 2.1, and a 8 byte difference in zip 2.2
	 * generated archives. Better not take chances and read
	 * the whole local header (i.e. store just the header offset, here)
	 */
	ret->dataPos = head.relativeLocalHeaderOffset;

	return (ret);
}

/*
 * Find first central directory record and return nr of actual records.
 * NB. We assume there's not comments in JARs.
 */
static
int
findFirstCentralDirRecord(jarFile* file)
{
	uint32 signature;
	uint32 ign;
	uint16 sz;
	uint32 off;

	if (Kaffe_JarRawInterface_seekEnd(file, -SIZEOFCENTRALEND) < 0) {
		file->error = "Failed to seek into JAR file";
		return (0);
	}

	signature = Kaffe_JarRawInterface_read32(file);
	if (signature != CENTRALENDSIGNATURE) {
		file->error = "Failed to find end of JAR record";
		return (0);
	}

	ign = Kaffe_JarRawInterface_read16(file);	/* Nr of disk */
	ign = Kaffe_JarRawInterface_read16(file);	/* Nr of disk with central directory */
	ign = Kaffe_JarRawInterface_read16(file);	/* Nr of entries in central directory on this disk */
	sz = Kaffe_JarRawInterface_read16(file);	/* Nr of entries in central directory */
	ign = Kaffe_JarRawInterface_read32(file);	/* Size of central directory */
	off = Kaffe_JarRawInterface_read32(file);	/* Offset of central directory */

	Kaffe_JarRawInterface_seekStart(file, off);

	return (sz);
}

void
Kaffe_JarInterface_init(void)
{
	Kaffe_JarRawInterface_init();
}

jarFile*
Kaffe_JarInterface_open(char* name)
{
	jarFile* file;
	int i;
	int len;

	/* Look for this jarfile if it's already open */
	for (file = openJars; file != 0; file = file->next) {
		if (strcmp(file->name, name) == 0) {
			file->ref++;
			return (file);
		}
	}

	DBG( vm_jar, ("Opening JAR %s\n", name));

	i = strlen(name);
	file = gc_malloc_fixed_typed(sizeof(jarFile) + i, GC_ALLOC_JAR);

	Kaffe_JarRawInterface_open(file, name);
	if (file->fp == -1) {
		DBG( vm_jar, ("Opening failed\n"));
		return (0);
	}

	strncpy(file->name, name, i);

	DBG( vm_jar, ("Find first central dir record\n"));
	i = findFirstCentralDirRecord(file);

	DBG_ACTION(	vm_jar,
	{
		if (i <= 0) {
			DBG( vm_jar, ("Error: %s\n", file->error));
		}
	});

	file->count = i;

	len = 0;
	for (i = 0; i < file->count; i++) {
		int elen = readCentralDirRecordSize(file);
		if (elen == 0) {
			return (0);
		}
		len += elen;
	}

	file->memstart = gc_malloc_fixed_typed(len, GC_ALLOC_JAR);
	file->mem = file->memstart;

	findFirstCentralDirRecord(file);	/* Reset */
	for (i = 0; i < file->count; i++) {
		jarEntry* curr = readCentralDirRecord(file);
		if ( curr != NULL ) {
			int index = jarFileNameHash(curr -> fileName);
			curr->next = file -> head[index];
			file->head[index] = curr;
/*DBG( vm_jar,("Register %s at %d\n", curr -> fileName, index)); */
		}
		else {
			return 0;
		}
	}
	DBG( vm_jar, ("Done\n"));

	/* Attach */
	file->ref++;
	file->next = openJars;
	openJars = file;

	return (file);
}

jarEntry*
Kaffe_JarInterface_lookup(jarFile* file, char* entry)
{
	jarEntry* curr;
	int i = jarFileNameHash(entry);

	DBG( vm_jar, ("Looking for %s in %s\n", entry, file -> name));
	curr = file->head[i];
	while (curr != 0) {
		DBG( vm_jar, ("Checking %s in JAR\n", curr->fileName));
		if (strcmp(entry, curr->fileName) == 0) {
			DBG( vm_jar, ("Found\n"));
			return (curr);
		}
		curr = curr->next;
	}
	DBG( vm_jar, ("Not found\n"));
	return (0);
}

int
Kaffe_JarInterface_getData(jarFile* file, jarEntry* entry, uint8* nbuf)
{
	jarLocalHeader locHdr;
	uint32         locSig;
	int            skip;

	if (Kaffe_JarRawInterface_seekStart(file, entry->dataPos) < 0) {
		file->error = "Failed to seek into JAR file";
		return (0);
	}

	locSig = Kaffe_JarRawInterface_read32(file);
	if ( locSig != LOCALHEADERSIGNATURE ){
		file->error = "illegal local header in JAR file";
		return (0);
	}
	locHdr.versionExtract = Kaffe_JarRawInterface_read16(file);
	locHdr.flags = Kaffe_JarRawInterface_read16(file);
	locHdr.compressionMethod = Kaffe_JarRawInterface_read16(file);
	locHdr.lastModifiedDosDTG = Kaffe_JarRawInterface_read32(file);
	locHdr.crc = Kaffe_JarRawInterface_read32(file);
	locHdr.compressedSize = Kaffe_JarRawInterface_read32(file);
	locHdr.uncompressedSize = Kaffe_JarRawInterface_read32(file);
	locHdr.fileNameLength = Kaffe_JarRawInterface_read16(file);
	locHdr.extraFieldLength = Kaffe_JarRawInterface_read16(file);

	/* Note that extraFieldLength differs in the central dir and the
	 * local header (and the difference depends on the zip version used
	 * to build the archive).  Using the local header value seems to be
	 * safe (that's what unzip does, for instance). Skip over the filename
	 * and the extra field, and start to decompress
	 */
	skip = locHdr.fileNameLength + locHdr.extraFieldLength;
	if ( Kaffe_JarRawInterface_seekHere(file, skip) < 0 ){
		file->error = "Failed to seek into JAR file data";
		return (0);
	}

	/* Decompress data */
	switch (entry->compressionMethod) {
	case COMPRESSION_STORED:
	{
		/*
		 *
		 * If its just stored, allocate a buffer and read the data
		 * into it..
		 */
		int size = Kaffe_JarRawInterface_readBytes(file, entry->compressedSize, nbuf);
		if ( size != (int)entry->compressedSize) {
			file->error = "Failed to read from JAR file";
			return 0;
		}
		return (1);
	}

	case COMPRESSION_DEFLATED:
	{
#if defined(HAVE_LIBZ) && defined(HAVE_ZLIB_H)
		int r;
		uint8* buffer;
		z_stream in;

		buffer = gc_malloc_fixed_typed(entry->compressedSize, GC_ALLOC_JAR);
		r = Kaffe_JarRawInterface_readBytes(file, entry->compressedSize, buffer);
		if (r != entry->compressedSize) {
			gc_free_fixed(buffer);
			file->error = "Decompression failed";
			return (0);
		}

		in.zalloc = 0;
		in.zfree = 0;
		in.opaque = 0;

		r = inflateInit2(&in, -15);
		if (r != Z_OK) {
			file->error = "Decompression failed";
			return (0);
		}

		in.next_in = buffer;
		in.avail_in = entry->compressedSize;
		in.next_out = nbuf;
		in.avail_out = entry->uncompressedSize;

		r = inflate(&in, Z_FINISH);
		switch (r) {
		case Z_OK:
		case Z_STREAM_END:
		case Z_BUF_ERROR:
			break;
		default:
			gc_free_fixed(buffer);
			inflateEnd(&in);
			file->error = "Decompression failed";
			return (0);
		}

		gc_free_fixed(buffer);
		inflateEnd(&in);
		return (1);
#else
		file->error = "Decompression failed: Not available";
		return (0);
#endif
	}

	/* These are not supported yet */
	case COMPRESSION_SHRUNK:
	case COMPRESSION_REDUCED1:
	case COMPRESSION_REDUCED2:
	case COMPRESSION_REDUCED3:
	case COMPRESSION_REDUCED4:
	case COMPRESSION_IMPLODED:
	case COMPRESSION_TOKENIZED:
	default:
		file->error = "Unsupported compression in JAR file";
		return (0);
	}
}


void
Kaffe_JarInterface_close(jarFile* file)
{
	jarFile** ptr;

	file->ref--;
	if (file->ref > 0) {
		return;
	}
	for (ptr = &openJars; *ptr != 0; ptr = &(*ptr)->next) {
		if (*ptr == file) {
			*ptr = file->next;
			Kaffe_JarRawInterface_close(file);
			gc_free_fixed(file->memstart);
			gc_free_fixed(file);
			return;
		}
	}
}
