/*
 * jar.h
 * Handle JAR input files.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __jar_h
#define __jar_h

/*
 * JAR files are like ZIP files.
 * Format:
 *   { [ jarLocalHeader + filename + extraField + fileData + jarDataDescriptor
 *	   ..... } + jarCentralDirectory ]
 */

/*
 * Central directory header.
 */
typedef struct _jarCentralDirectoryRecord {

	uint32		signature;
	uint16		versionMade;
	uint16		versionExtract;
	uint16		flags;
	uint16		compressionMethod;
	uint16		lastModifiedTime;
	uint16		lastModifiedDate;
	uint32		crc;
	uint32		compressedSize;
	uint32		uncompressedSize;
	uint16		fileNameLength;
	uint16		extraFieldLength;
	uint16		fileCommentLength;
	uint16		diskNumberStart;
	uint16		internalFileAttribute;
	uint32		externalFileAttribute;
	uint32		relativeLocalHeaderOffset;

} jarCentralDirectoryRecord;


typedef struct _jarLocalHeader {
	uint16       versionExtract;
	uint16       flags;
	uint16       compressionMethod;
	uint32       lastModifiedDosDTG;
	uint32       crc;
	uint32       compressedSize;
	uint32       uncompressedSize;
	uint16       fileNameLength;
	uint16       extraFieldLength;
} jarLocalHeader;


/*
 * Central directory end record.
 */
typedef struct _jarCentralDirectoryEnd {

	uint32	signature;
	uint16	numberOfDisk;
	uint16	numberOfDiskWithDirectory;
	uint16	nrOfEntriesInThisDirectory;
	uint16	nrOfEntriesInDirectory;
	uint32	sizeOfDirectory;
	uint32	offsetOfDirectory;
	uint16	commentLength;

} jarCentralDirectoryEnd;

/*
 * JAR file record.
 */
typedef struct _jarEntry {
	struct _jarEntry*	next;
	char*			fileName;
	long			dataPos;
	uint16			compressionMethod;
	uint32			compressedSize;
	uint32			uncompressedSize;
} jarEntry;

#define JARFILE_HASHTABLE_BITS (8)
#define JARFILE_HASHTABLE_MASK (~(-1 << (JARFILE_HASHTABLE_BITS)))
#define JARFILE_HASHTABLE_SIZE (1 << JARFILE_HASHTABLE_BITS)

typedef struct _jarFile {
	struct _jarFile*	next;
	int			ref;
	int			fp;
	uint32			mmapLength;
	uint8*			mmapBase;
	uint8*			mmapPos;
	int			count;
	uint8*			memstart;
	uint8*			mem;
	jarEntry*		head[JARFILE_HASHTABLE_SIZE];
	char*			error;
	char			name[1];
} jarFile;

/*
 * Various signatures.
 */
#define	LOCALHEADERSIGNATURE	0x04034b50
#define	CENTRALHEADERSIGNATURE	0x02014b50
#define	CENTRALENDSIGNATURE	0x06054b50

/*
 * Sizes of infile headers.
 */
#define	SIZEOFLOCALHEADER  30      /* sizeof(JarLocalHeader) + localSig */
#define	SIZEOFCENTRALEND	  22

/*
 * Compression methods.
 */
#define COMPRESSION_STORED	0
#define COMPRESSION_SHRUNK	1
#define COMPRESSION_REDUCED1	2
#define COMPRESSION_REDUCED2	3
#define COMPRESSION_REDUCED3	4
#define COMPRESSION_REDUCED4	5
#define COMPRESSION_IMPLODED	6
#define COMPRESSION_TOKENIZED	7
#define COMPRESSION_DEFLATED	8

extern void Kaffe_JarInterface_init(void);
extern jarFile* Kaffe_JarInterface_open(char*);
extern jarEntry* Kaffe_JarInterface_lookup(jarFile*, char*);
extern int Kaffe_JarInterface_getData(jarFile*, jarEntry*, uint8*);
extern void Kaffe_JarInterface_close(jarFile*);

/*
 * Interface.  This consists of two parts, the first the interface
 *  to the JAR reader, the second the interface to the raw data.
 */
#define	initJarFile()		Kaffe_JarInterface_init()
#define	openJarFile(F)		Kaffe_JarInterface_open(F)
#define	closeJarFile(F)		Kaffe_JarInterface_close(F)
#define	lookupJarFile(F,N)	Kaffe_JarInterface_lookup((F),(N))
#define	getDataJarFile(F,E,B)	Kaffe_JarInterface_getData((F),(E),(B))

/* Inline the jar file access methods */
#if defined(HAVE_JAR_FILE)
#include "jarFile.c"
#endif
#if defined(HAVE_JAR_MMAP)
#include "jarMmap.c"
#endif

#endif
