/*
 * jdebug.h - runtime configurable debug support
 * 
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * (1) This header has to be included in all modules which want to use
 * the DBG or DBG_ACTION macros like:
 * 
 *   -------------- foo.c ----------------
 *   #include "jdebug.h"
 *     ..
 *     DBG( my_topic, ("this is a sample message: %d\", par));
 *     ..
 *     DBG_ACTION( my_topic, { doThis(); doThat(); });
 *     ..
 *   ------------------------------------
 *
 * 
 * (2) In addition, there has to be one main module (usually containing main)
 * that initializes the debug support:
 * 
 *   ------------- bar.c -----------------
 *   #define DEBUG_MAIN
 *   #include "jdebug.h"
 *     ..
 *     DBG_INIT();
 *     ..
 *   -------------------------------------
 * 
 * (3) All suppored debug topics have to be defined in a file called
 * jdebugdef.h (which is included from jdebug.h).
 * 
 *   ------------ jdebugdef.h ------------
 *   DBG_BEGIN_DEFINE_TOPICS
 *   DBG_DEFINE( my_topic);
 *   ..
 *   DBG_END_DEFINE_TOPICS
 *
 *   dbg_setup() {
 *     ..
 *   }
 *   -------------------------------------
 * 
 * (4) Build the system with -DDEBUG
 * 
 * (5) Before running the app, specify the debug topics you are interested in
 * with a environment variable DEBUG, containing a space separated list
 * of the topics:
 * 
 *   export DEBUG="my_topic my_second_topic .."
 * 
 * (6) go and get the bugs...
 * 
 */

#ifndef __jdebug_h
#define __jdebug_h

#if defined(DEBUG)

#define	USE_VFPRINTF_DEBUG	1

#if defined(USE_VFPRINTF_DEBUG)
#include "support.h"
#define	dbg_message	kprintf
#endif

/**********************************************************************
 * public macros - this is all the DEBUG_MAIN modules have to know
 */

#define DBG(_topic,_msg)               \
  if ( _dbg_##_topic ) {               \
    dbg_message ("%s: ", #_topic);     \
    dbg_message _msg;                  \
  }

#define DBGOR(_topic1,_topic2, _msg)               \
  if ( _dbg_##_topic1 || _dbg_##_topic2) {               \
    dbg_message ("%s: ", #_topic1);       \
    dbg_message _msg;                  \
  }


#define DBG_ACTION(_topic,_op)   if ( _dbg_##_topic ) _op

#define DBG_ACTION_OR(_topic1,_topic2,_op)   if ( _dbg_##_topic1 || _dbg_##_topic2) _op

#define DBG_ACTION_AND_ELSE(_topic,_op,_else)   if ( _dbg_##_topic ) _op else _else


/**********************************************************************
 * internal macros & funcs
 */

/* these are just for the second pass of the DEBUG_MAIN module */
#define DBG_BEGIN_DEFINE_TOPICS
#define DBG_END_DEFINE_TOPICS
#undef  DEBUG_SETUP

#if defined(DEBUG_MAIN)

/*
 * These are the macros for the jdebugdef.h config
 */
#define DBG_INIT() dbg_init()
#define DBG_DEFINE(_topic)       char _dbg_##_topic
#define DBG_CHECK(_topic)        _dbg_##_topic = dbg_check( #_topic)
#define	DBG_DEFINE_ON(_topic)	if (strcmp(#_topic, topic) == 0) return 1

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

char* _dbg_env;
int   _dbg_fd;
int   _dbg_fd_set;

#if !defined(USE_VFPRINTF_DEBUG)
void
dbg_message( char* fmt, ...)
{
  va_list ap;
  char buf[256];

  va_start( ap, fmt);

  /*
   * don't use vfprintf, because that involves buffered streams, which
   * might require locking mechanisms in the C lib we simply can't rely
   * on (depending on the threading system and debug topic)
   */
  vsnprintf( buf, sizeof(buf), fmt, ap);
  write( _dbg_fd, buf, strlen(buf));

  va_end(ap);
}
#endif

/*
 * get the requested debug topics out of the environment var DEBUG
 */
char
dbg_check ( char* topic )
{
  int len;
  char *tok, *s;

  if ( !_dbg_fd_set ) {
	_dbg_fd = 2;      /* stderr */
	_dbg_fd_set = 1;
  }

  if ( !_dbg_env ) {
	_dbg_env = getenv( "DEBUG");
  }

  if ( _dbg_env ) {
	len = strlen( topic);

	for ( s=_dbg_env; *s && (tok = strstr( s, topic)); s+=len ){
	  if ( (tok[len] == ' ') || (tok[len] == 0) ){
		if ( (tok == s) || (tok[-1] == ' ') ) {
		  dbg_message( "debug: %s\n", topic);
		  return 1;
		}
	  }
	}
  }

  return 0;
}

/* first pass - declare the debug channels as global vars */
#include "jdebugdef.h"

#undef  DBG_BEGIN_DEFINE_TOPICS
#define DBG_BEGIN_DEFINE_TOPICS \
void dbg_init () {

#undef  DBG_END_DEFINE_TOPICS
#define DBG_END_DEFINE_TOPICS                          \
  if ( atexit( dbg_finish ))                           \
    dbg_message( "can't register DBG exit handler\n"); \
  dbg_setup();                                         \
}

#undef DBG_DEFINE
#define DBG_DEFINE(_topic)       DBG_CHECK(_topic)

void dbg_setup(void);
void dbg_finish(void);

/* second pass - check which channels are active */
#define DEBUG_SETUP
#include "jdebugdef.h"

#else  /* !defined(DEBUG_MAIN) */

#if !defined(USE_VFPRINTF_DEBUG)
extern void dbg_message( char* fmt, ...);
#endif

/*
 * This is what we want to see from "jdebugdef" in case we are
 * not in the DEBUG_MAIN module
 */
#define DBG_INIT()
#define DBG_CHECK(_topic)

#define DBG_DEFINE(_topic)       extern char _dbg_##_topic

/* get us the required external declarations */
#include "jdebugdef.h"

#endif /* DEBUG_MAIN */

#else /***** !defined(DEBUG) *********************************************/

/*
 * If the program was compiled without -DDEBUG
 */

#define DBG_BEGIN_DEFINE_TOPICS
#define DBG_END_DEFINE_TOPICS

#define DBG_INIT()
#define DBG_DEFINE(_topic)
#define DBG_CHECK(_topic)

#define DBG(_topic,_msg)
#define DBG_ACTION(_topic,_op)

#define DBGOR(_topic1,_topic2, _msg)

#define DBG_ACTION_OR(_topic1,_topic2,_op)   

#define DBG_ACTION_AND_ELSE(_topic,_op,_else) _else


#endif /* DEBUG */


#endif /* __jdebug_h */
