/* machine.c
 * Translate the Kaffe instruction set to the native one.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "md.h"
#include "classMethod.h"
#include "callMethodInfo.h"
#include "bytecode.h"
#include "slots.h"
#include "registers.h"
#include "seq.h"
#include "gc.h"
#include "machine.h"
#include "basecode.h"
#include "icode.h"
#include "labels.h"
#include "constpool.h"
#include "codeproto.h"
#include "checks.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "baseClasses.h"
#include "code.h"
#include "access.h"
#include "lookup.h"
#include "exception.h"
#include "locks.h"
#include "code-analyse.h"
#include "external.h"
#include "soft.h"
#include "threads.h"
#include "engine.h"

#include "jdebug.h"
#include "jmeasure.h"

char* engine_name = "Just-in-time";
char* engine_version = KAFFEVERSION;

int stackno;
int maxStack;
int maxLocal;
int maxTemp;
int maxArgs;
int maxPush;
int isStatic;

int tmpslot;
int argcount = 0;		/* Function call argument count */
uint32 pc;
uint32 npc;

/* Various exception related things */
extern Hjava_lang_Class javaLangArrayIndexOutOfBoundsException;
extern struct JNIEnv_ Kaffe_JNIEnv;
extern Method* jniMethod;
extern void startJNIcall(void);
extern void finishJNIcall(void);
extern void addJNIref(jref);
extern void removeJNIref(jref);

codeinfo* codeInfo;
uint16 ciMaxPC;
uint16 ciMaxLocal;

jitflags willcatch;

/* jit3 specific prototypes from icode.c */
void check_null (int x, SlotInfo* obj, int y);
void check_div (int x, SlotInfo* obj, int y);
void check_div_long (int x, SlotInfo* obj, int y);
void softcall_fakecall (label* from,label* to, void* func);

/* Unit in which code block is increased when overrun */
#define	ALLOCCODEBLOCKSZ	8192
/* Codeblock redzone - allows for safe overrun when generating instructions */
#define	CODEBLOCKREDZONE	256

nativecode* codeblock;
int codeblock_size;
static int code_generated;
static int bytecode_processed;
static int codeperbytecode;
static iLock* translatorlock;

int CODEPC;

struct {
	int time_jit;
	int time_load;
	int time_flush;
} jitStats;

extern int enable_readonce;

void	initInsnSequence(int, int, int);
void	finishInsnSequence(nativeCodeInfo*);
static void generateInsnSequence(void);
static void installMethodCode(Method*, nativeCodeInfo*);
static void pushContext(void);
static void popContext(void);
static int  countArgs(Method *meth, int *isStatic);
static void flushCache(Method *m);

void	endBlock(sequence*);
void	startBlock(sequence*);
void	endSubBlock(sequence*);
void	startSubBlock(sequence*);
void	cancelNoWriteback(void);
jlong	currentTime(void);
Method*	findMethodFromPC(uintp);

/*
 * Translate a method into native code.
 */
void
translate(Method* meth)
{
	int i;

	jint low;
	jint high;
	jvalue tmpl;
	int idx;
	SlotInfo* tmp;
	SlotInfo* tmp2;
	SlotInfo* mtable;

	bytecode* base;
	uint32 len;
	callInfo cinfo;
	fieldInfo finfo;
	Hjava_lang_Class* crinfo;

	nativeCodeInfo ncode;

	int64 tms_jit = 0;
	int64 tme_jit;
	int64 tms_load = 0;
	int64 tme_load = 0;
	int iLockRoot;

	static bool reinvoke = false;

	MSR_START( jit_time);

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tms_load = currentTime();
	}

	/* Since we get here by calling the trampoline, we must be about
	 * to call this method - so make sure it's OK before generating
	 * code to avoid any extra initialization calls.
	 */
	processClass(meth->class, CSTATE_COMPLETE);

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tme_load = currentTime();
		tms_jit = tme_load;
	}

	/* Only one in the translator at once. Must check the translation
	 * hasn't been done by someone else once we get it.
	 */
	lockStaticMutex(&translatorlock);

	lockMutex(meth->class->centry);

	if (METHOD_TRANSLATED(meth)) {
		unlockMutex(meth->class->centry);
		unlockStaticMutex(&translatorlock);
		return;
	}

DBG( vm_jit_translate, ("callinfo = 0x%x\n", &cinfo));

	/* If this code block is native, then just set it up and return */
	if ((meth->accflags & ACC_NATIVE) != 0) {
		native(meth);
		KAFFEJIT_TO_NATIVE(meth); /* Note that this is a real function not a trampoline.  */
		unlockMutex(meth->class->centry);
		unlockStaticMutex(&translatorlock);
		return;
	}

	/* Save current JIT context */
	pushContext();

	/* Scan the code and determine the basic blocks */
	verifyMethod(meth, &codeInfo);

	assert(reinvoke == false);
	reinvoke = true;

	initInsnSequence(codeperbytecode * meth->bcode->codelen, meth->bcode->localsz, meth->bcode->stacksz);

	maxLocal = meth->bcode->localsz;
	maxStack = meth->bcode->stacksz;
	maxArgs = countArgs(meth, &isStatic);

	base = (bytecode*)meth->bcode->code;
	len = meth->bcode->codelen;

	willcatch.ANY = false;
	willcatch.BADARRAYINDEX = false;

	/* Deterimine various exception conditions */
	if (meth->exception_table != 0) {
		willCatch(ANY);
		for (i = 0; i < (int)meth->exception_table->length; i++) {
			Hjava_lang_Class* etype;
			etype = meth->exception_table->entry[i].catch_type;
			if (etype == 0) {
				willCatch(BADARRAYINDEX);
			}
			else {
				if (instanceof(&javaLangArrayIndexOutOfBoundsException, etype)) {
					willCatch(BADARRAYINDEX);
				}
			}
		}
	}

	/***************************************/
	/* Next reduce bytecode to native code */
	/***************************************/

	start_basic_block();
	start_function();
	check_stack_limit();
	if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) {
		softcall_trace(meth);
	}
	monitor_enter();
	if (IS_STARTOFBASICBLOCK(0)) {
		end_basic_block();
		start_basic_block();
	}

	for (pc = 0; pc < len; pc = npc) {

		assert(stackno <= maxStack+maxLocal);

		npc = pc + insnLen[base[pc]];

		start_instruction();

		/* Note start of exception handling blocks */
		if (IS_STARTOFEXCEPTION(pc)) {
			stackno = meth->bcode->localsz + meth->bcode->stacksz - 1;
			start_exception_block();
		}

		switch (base[pc]) {
		default:
			kprintf("Unknown bytecode %d\n", base[pc]);
			unlockMutex(meth->class->centry);
			unlockStaticMutex(&translatorlock);
			postAndThrowExternalException(VerifyError);
			break;
#include "kaffe.def"
		}

		/* Note maximum number of temp slots used and reset it */
		if (tmpslot > maxTemp) {
			maxTemp = tmpslot;
		}
		tmpslot = 0;

		cancelNoWriteback();

		if (IS_STARTOFBASICBLOCK(npc)) {
			end_basic_block();
			start_basic_block();
			stackno = STACKPOINTER(npc);
		}

		generateInsnSequence();
	}

	finish_function();

	assert(maxTemp < MAXTEMPS);

	finishInsnSequence(&ncode);
	installMethodCode(meth, &ncode);

	tidyVerifyMethod(&codeInfo);

	reinvoke = false;
	/* Restore current JIT context */
	popContext();

DBG( vm_jit_translate, ("Translating %s.%s%s (%s) %p\n",
						meth->class->name->data,
						meth->name->data, makeStrFromSignature(meth->sig),
						isStatic ? "static" : "normal", METHOD_NATIVECODE(meth)));

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		extern int timeInJar;
		extern int timeInJarUncompress;
		tme_jit = currentTime();
		jitStats.time_jit += (int)(tme_jit - tms_jit);
		jitStats.time_load += (int)(tme_load - tms_load);
		kprintf("<JIT: %s.%s%s time %dms/%dms (%dms/%dms/%dms), jar=(%dms/%dms) @ %p (%p)>\n",
			CLASS_CNAME(meth->class), meth->name->data,
			makeStrFromSignature(meth->sig),
			(int)(tme_jit - tms_jit),
			(int)(tme_load - tms_load),
			jitStats.time_jit,
			(int)jitStats.time_load,
			(int)jitStats.time_flush,
			(int)timeInJar,
			(int)timeInJarUncompress,
			METHOD_NATIVECODE(meth), meth);
	}

	unlockMutex(meth->class->centry);
	unlockStaticMutex(&translatorlock);

	MSR_STOP( jit_time);
}

/*
 * Generate the code.
 */
void
finishInsnSequence(nativeCodeInfo* code)
{
	uint32 constlen;
	nativecode* methblock;
	nativecode* codebase;
	nativeInfo* info;

	/* Emit pending instructions */
	generateInsnSequence();

	/* Okay, put this into malloc'ed memory */
	constlen = nConst * sizeof(union _constpoolval);
	methblock = gc_malloc_fixed_typed(constlen + CODEPC + sizeof(nativeInfo), GC_ALLOC_JITCODE);
	info = (nativeInfo*)(methblock + constlen);
	codebase = ((nativecode*)info) + sizeof(nativeInfo);
	memcpy(codebase, codeblock, CODEPC);
	gc_free(codeblock);

	/* Establish any code constants */
	establishConstants(methblock);

	/* Link it */
	linkLabels((uintp)codebase);

	/* Note info on the compiled code for later installation */
	code->mem = methblock;
	code->memlen = constlen + CODEPC + sizeof(nativeInfo);
	code->code = codebase;
	code->codelen = CODEPC;

	/* Fill in the info block */
	info->ncode_start = methblock;
	info->ncode_end = codebase + CODEPC;
}

/*
 * Get the instruction offset corresponding to the given PC.
 * If the PC doesn't point at the start of a valid instruction,
 * look forward until we find one.
 */
static
int
getInsnPC(int pc)
{
	int res;

	for (;;) {
		res = INSNPC(pc);
		if (res != -1) {
			return (res);
		}
		pc++;
	}
}

/*
 * Flush the code for this method out of the D-cache
 */
static
void 
flushCache(Method* m)
{
	if (!METHOD_FLUSHED(m)) {
#if defined(FLUSH_DCACHE)
		nativeInfo *info = GET_NATIVEINFO(m);
		/* Flush code out of cache */
		FLUSH_DCACHE(info->ncode_start, info->ncode_end);
#endif
		SET_METHOD_FLUSHED(m);
	}
}

/*
 * Install the compiled code in the method.
 */
static
void
installMethodCode(Method* meth, nativeCodeInfo* code)
{
	uint32 i;
	jexceptionEntry* e;

	/* Work out new estimate of code per bytecode */
	code_generated += code->memlen;
	bytecode_processed += meth->bcode->codelen;
	codeperbytecode = code_generated / bytecode_processed;

	GC_WRITE(meth, code->mem);
	SET_METHOD_NATIVECODE(meth, code->code);
        MSR_ADD(ncode_mem, code->codelen);

	/* Translate exception table and make it available */
	if (meth->exception_table != 0) {
		for (i = 0; i < meth->exception_table->length; i++) {
			e = &meth->exception_table->entry[i];
			e->start_pc = getInsnPC(e->start_pc) + (uintp)code->code;
			e->end_pc = getInsnPC(e->end_pc) + (uintp)code->code;
			e->handler_pc = getInsnPC(e->handler_pc) + (uintp)code->code;
		}
	}

#if !defined(OPTIMIZE_MEMORY)
	/* Translate line numbers table */
	if (meth->lines != 0) {
		for (i = 0; i < meth->lines->length; i++) {
			meth->lines->entry[i].start_pc = getInsnPC(meth->lines->entry[i].start_pc) + (uintp)code->code;
		}
	}
#endif

	/* Make method active (so we can find if for exceptions) */
	makeMethodActive(meth);
}

/*
 * Init instruction generation.
 */
void
initInsnSequence(int codesize, int localsz, int stacksz)
{
	/* Clear various counters */
	tmpslot = 0;
	maxTemp = 0;
	maxPush = 0;  
	stackno = localsz + stacksz;
	maxTemp = MAXTEMPS - 1; /* XXX */ 
	npc = 0;

	initSeq();
	initRegisters();
	initSlots(stackno);
	resetLabels();

	/* Before generating code, try to guess how much space we'll need. */
	codeblock_size = codesize;
	if (codeblock_size < ALLOCCODEBLOCKSZ) {
		codeblock_size = ALLOCCODEBLOCKSZ;
	}
	codeblock = gc_malloc_fixed_typed(codeblock_size + CODEBLOCKREDZONE, GC_ALLOC_ENGINE);
	CODEPC = 0;
}

static
void
generateInsnSequence(void)
{
	sequence* t;

	for (t = firstSeq; t != currSeq; t = t->next) {

		/* If we overrun the codeblock, reallocate and continue.  */
		if (CODEPC >= codeblock_size) {
			codeblock_size += ALLOCCODEBLOCKSZ;
			codeblock = gc_realloc(codeblock, codeblock_size + CODEBLOCKREDZONE, GC_ALLOC_ENGINE);
		}

		/* Generate sequences */
		(*(t->func))(t);
	}

	/* Reset */
	initSeq();
}

/*
 * Start a new instruction.
 */
void
startInsn(sequence* s)
{
	SET_INSNPC(const_int(2), CODEPC);
}

/*
 * Mark slot not to be written back.
 */
void
nowritebackSlot(sequence* s)
{
	seq_slot(s, 0)->modified |= rnowriteback;
}

/*
 * Start a new basic block.
 */
void
startBlock(sequence* s)
{
	startSubBlock(s);
}

/*
 * Start a new basic sub-block.
 */
void
startSubBlock(sequence* s)
{
	int i;

	/* Invalidate all slots - don't use clobberRegister which will
	 * flush them - we do not want to do that even if they are dirty.
	 */
	for (i = maxslot - 1; i >= 0; i--) {
		if (slotinfo[i].regno != NOREG) {
			register_invalidate(slotinfo[i].regno);
			slot_invalidate(&slotinfo[i]);
		}
	}
#if defined(STACK_LIMIT)
	if (stack_limit[0].regno != NOREG) {
		register_invalidate(stack_limit[0].regno);
		slot_invalidate(stack_limit);
	}
#endif
}

/*
 * Fixup after a function call.
 */
void
fixupFunctionCall(sequence* s)
{
	int i;

	/* Invalidate all slots - don't use clobberRegister which will
	 * flush them - we do not want to do that even if they are dirty.
	 */
	for (i = maxslot - 1; i >= 0; i--) {
		if (slotinfo[i].regno != NOREG && !calleeSave(slotinfo[i].regno)) {
			register_invalidate(slotinfo[i].regno);
			slot_invalidate(&slotinfo[i]);
		}
	}
#if defined(STACK_LIMIT)
	if (stack_limit[0].regno != NOREG) {
		register_invalidate(stack_limit[0].regno);
		slot_invalidate(stack_limit);
	}
#endif
}

/*
 * End a basic block.
 */
void
endBlock(sequence* s)
{
	endSubBlock(s);
}

/*
 * End a basic sub-block.
 */
void
endSubBlock(sequence* s)
{
	int stkno;
	int tmpslot;
	int i;

	/* Spill locals */
	for (i = 0; i < maxLocal; i++) {
		if ((localinfo[i].modified & rwrite) != 0 && localinfo[i].regno != NOREG) {
			if ((localinfo[i].modified & rnowriteback) == 0) {
				spill(&localinfo[i]);
				localinfo[i].modified = 0;
			}
			else {
				localinfo[i].modified &= ~rnowriteback;
			}
		}
	}

	/* Spill stack */
	stkno = const_int(1);
	for (i = stkno; i < maxStack+maxLocal; i++) {
		if ((localinfo[i].modified & rwrite) != 0 && localinfo[i].regno != NOREG) {
			if ((localinfo[i].modified & rnowriteback) == 0) {
				spill(&localinfo[i]);
				localinfo[i].modified = 0;
			}
			else {
				localinfo[i].modified &= ~rnowriteback;
			}
		}
	}

	/* Spill temps currently in use */
	tmpslot = const_int(2);
	for (i = 0; i < tmpslot; i++) {
		if ((tempinfo[i].modified & rwrite) != 0 && tempinfo[i].regno != NOREG) {
			if ((tempinfo[i].modified & rnowriteback) == 0) {
				spill(&tempinfo[i]);
				tempinfo[i].modified = 0;
			}
			else {
				tempinfo[i].modified &= ~rnowriteback;
			}
		}
	}

#if defined(STACK_LIMIT)
	if ((stack_limit[0].modified & rwrite) != 0 && stack_limit[0].regno != NOREG) {
		if ((stack_limit[0].modified & rnowriteback) == 0) {
			spill(stack_limit);
			stack_limit[0].modified = 0;
		}
		else {
			stack_limit[0].modified &= ~rnowriteback;
		}
	}
#endif
}

/*
 * Prepare register state for function call.
 */
void
prepareFunctionCall(sequence* s)
{
	int stkno;
	int tmpslot;
	int i;

	/* Spill locals */
	for (i = 0; i < maxLocal; i++) {
		/* Spill any locals which have been modified and are still
		 * in a register */
		if ((localinfo[i].modified & rwrite) != 0 && localinfo[i].regno != NOREG) {
			/* We don't need to spill callee-saved values except
			 * if we could need the value in an exception handler.
			 */
			if (canCatch(ANY) || !calleeSave(localinfo[i].regno)) {
				spill(&localinfo[i]);
				localinfo[i].modified = 0;
			}
		}
	}

	/* Spill stack */
	stkno = const_int(1);
	for (i = stkno; i < maxStack+maxLocal; i++) {
		if ((localinfo[i].modified & rwrite) != 0 && localinfo[i].regno != NOREG) {
			if (canCatch(ANY) || !calleeSave(localinfo[i].regno)) {
				spill(&localinfo[i]);
				localinfo[i].modified = 0;
			}
		}
	}

	/* Spill temps currently in use */
	tmpslot = const_int(2);
	for (i = 0; i < tmpslot; i++) {
		if ((tempinfo[i].modified & rwrite) != 0 && tempinfo[i].regno != NOREG) {
			if (canCatch(ANY) || !calleeSave(tempinfo[i].regno)) {
				spill(&tempinfo[i]);
				tempinfo[i].modified = 0;
			}
		}
	}

#if defined(STACK_LIMIT)
	if ((stack_limit[0].modified & rwrite) != 0 && stack_limit[0].regno != NOREG) {
		spill(stack_limit);
		stack_limit[0].modified = 0;
	}
#endif
}

/*
 * Sync register state back to memory (but don't clean anything).
 */
void
syncRegisters(sequence* s)
{
	int stkno;
	int tmpslot;
	int i;
	int old_ro;

	old_ro = enable_readonce;
	enable_readonce = 0;

	/* Spill locals */
	for (i = 0; i < maxLocal; i++) {
		if ((localinfo[i].modified & rwrite) != 0 && localinfo[i].regno != NOREG && !calleeSave(localinfo[i].regno)) {
			spill(&localinfo[i]);
		}
	}

	/* Spill stack */
	stkno = const_int(1);
	for (i = stkno; i < maxLocal+maxStack; i++) {
		if ((localinfo[i].modified & rwrite) != 0 && localinfo[i].regno != NOREG && !calleeSave(localinfo[i].regno)) {
			spill(&localinfo[i]);
		}
	}

	/* Spill temps currently in use */
	tmpslot = const_int(2);
	for (i = 0; i < tmpslot; i++) {
		if ((tempinfo[i].modified & rwrite) != 0 && tempinfo[i].regno != NOREG && !calleeSave(tempinfo[i].regno)) {
			spill(&tempinfo[i]);
		}
	}

#if defined(STACK_LIMIT)
	if ((stack_limit[0].modified & rwrite) != 0 && stack_limit[0].regno != NOREG) {
		spill(stack_limit);
	}
#endif

	enable_readonce = old_ro;
}

/*
 * Cancel any pending nowriteback flags.
 */
void
cancelNoWriteback(void)
{
	int i;

	for (i = stackno; i < maxLocal+maxStack; i++) {
		localinfo[i].modified &= ~rnowriteback;
	}
}

/*
 * Build multi-dimensional array.
 *  This is sufficiently different from the standard
 *  to require our own routine.
 */
void*
jit_soft_multianewarray(Hjava_lang_Class* class, jint dims, ...)
{
	int array[16];
	Hjava_lang_Object* obj;
	jint arg;
	int i;
	int* arraydims;
	va_list ap;

        if (dims < 16) {
		arraydims = array;
	}
	else {
		arraydims = gc_calloc(dims+1, sizeof(int), GC_ALLOC_ENGINE);
	}

	/* Extract the dimensions into an array */
	va_start(ap, dims);
	for (i = 0; i < dims; i++) {
		arg = va_arg(ap, jint);
		if (arg < 0) {
                        postAndThrowExternalException(NegativeArraySizeException);
		}
		arraydims[i] = arg;
	}
	arraydims[i] = -1;
	va_end(ap);

	/* Mmm, okay now build the array using the wonders of recursion */
        obj = newMultiArray(class, arraydims);

	if (arraydims != array) {
		gc_free(arraydims);
	}

	/* Return the base object */
	return (obj);
}

/*
 * Dispatch an exception.  These odd structures are necessary
 *  to emulate the old exception calling convention (we will
 *  deprecate at some point).
 */
static
void
EcallExceptionHandler(Method* meth, uintp handler, stackTraceInfo* frame, Hjava_lang_Throwable* eobj)
{
	CALL_KAFFE_EXCEPTION(frame->arg0, handler, eobj);
}

/* 
 * Trampolines come in here - do the translation and replace the trampoline.
 */
nativecode*
soft_fixup_trampoline(FIXUP_TRAMPOLINE_DECL)
{
	Method* meth;
	void** where;
	void* tramp;

	FIXUP_TRAMPOLINE_INIT;

	/* Find the trampoline so we can free it later if necessary */
	tramp = *where;

	/* If this class needs initializing, do it now.  */
	processClass(meth->class, CSTATE_COMPLETE);

	/* Generate code on demand.  */
	if (!METHOD_TRANSLATED(meth)) {
		translate(meth);
		DBG( vm_jit, ("%x : %s:%s%s\n", METHOD_NATIVECODE(meth),
					  meth->class->name->data, meth->name->data, makeStrFromSignature(meth->sig)));
	}
	if (!METHOD_FLUSHED(meth)) {
		flushCache(meth);
	}

	/* Update dispatch table - note that 'where' might not
	 * be the same as 'METHOD_NATIVECODE' since this may have already
	 * been updated.  In this case we don't do anything.
	 * Can't help but feel there's still a race condition here
	 *   - TIM Jul 22nd 1999
	 */
	if (*where != METHOD_NATIVECODE(meth)) {
		*where = METHOD_NATIVECODE(meth);
		/* Free the trampoline */
		gc_free(tramp);
	}

	return (*where);
}

/*
 * Count and return the number of arguments in a method
 * Includes `this' ptr if method is non-static.
 * *isStatic will be set by this method.
 */
static
int
countArgs(Method *meth, int *isStatic)
{
        int i;
	int maxArgs = 0;
        for (i = 0; i < meth->sig->len; i++) {
                switch (meth->sig->elem[i]->data[0]) {
                case 'D':
                case 'J':
                        maxArgs += 2;
                        break;
                default:
                        maxArgs += 1;
                }
        }

	if (meth->accflags & ACC_STATIC) {
		*isStatic = 1;
	}
	else {
		*isStatic = 0;
		maxArgs += 1;
	}
	return (maxArgs);
}

/*
 * Include the machine specific trampoline functions.
 */
#include "trampolines.c"

/*
 * Initialize the engine.
 */
static
void
Einit(void)
{
	/* Does nothing */
}

static
char*
EgetName(void)
{
	return ("Just-in-time");
}

static
char*
EgetVersion(void)
{
	return (KAFFEVERSION);
}

static
int
EgetNumber(void)
{
	return (0);
}

static
void*
EsetupTrampoline(Method* meth, void** where)
{
	int64 tms_flush = 0;
	methodTrampoline* tramp;

	if (METHOD_TRANSLATED(meth)) {
		return (METHOD_NATIVECODE(meth));
	}
	else {
		/* These are freed as methods are patched up */
		tramp = gc_malloc_fixed_typed(sizeof(methodTrampoline), GC_ALLOC_ENGINE);

		FILL_IN_TRAMPOLINE(tramp, meth, where);

		if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
			tms_flush = currentTime();
		}
#if defined(FLUSH_DCACHE)
		FLUSH_DCACHE(tramp, tramp+1);
#endif
		if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
			jitStats.time_flush += currentTime() - tms_flush;
		}

		return (tramp);
	}
}

static
void
EprocessExceptionMethod(stackTraceInfo* info)
{
	if (info->meth != jniMethod) {
		stackFrame* fm = (stackFrame*)info->rframe;
		/* Does this arch support a FRAMEOBJECT macro */
#if defined(FRAMEOBJECT)        
		info->object = (void*)FRAMEOBJECT(fm);
#else
		/* if not, do it the hard way... */
		int iLockRoot;
		/* Set up the necessary state for the SLOT2 macros to work ---
		 * see jit/machine.c
		 * Since we have the translator lock, we can clobber the 
		 * max* variables.
		 */
		/* XXX: I am not sure if this is sufficient in the
		 * custom edition.  -- gback
		 */
		lockStaticMutex(&translatorlock);
		maxLocal = info->meth->bcode->localsz;
		maxStack = info->meth->bcode->stacksz;
		maxArgs = countArgs(info->meth, &isStatic);
		maxTemp = MAXTEMPS - 1;		/* XXX */

		/* NB: we assume that the JIT will have
		 * spilled the 'this' object in the
		 * stack location for slot zero.
		 * For static methods, we set object to zero.
		 */
		info->object = isStatic ? 0 : ((void**)fm->return_frame)[SLOT2ARGOFFSET(0)/SLOTSIZE];
		unlockStaticMutex(&translatorlock);
#endif
		info->arg0 = (uintp)fm->return_frame;
	}
	else {
		info->object = 0;
		info->arg0 = (uintp)unhand(Kaffe_TI_currentJava())->exceptPtr;
	}
}

/* This function is here primarily to fake the compiler into thinking
 * EcallInterEngine isn't a leaf function (if it were then things would
 * break.
 */
static
void
engineDummy(void)
{
}

/*
 * Call inter-machine method
 */
void
EcallInterEngine(callMethodInfo* call)
{
	engineDummy();
	sysdepCallMethod(call);
}

/*
 * Build an interface between the JIT and a JNI method.
 */
static
int
EsetupJNIcall(Method* xmeth, void* func)
{
	char buf[100];
	int i;
	methodSig* sig;
	int count;
	nativeCodeInfo ncode;
	SlotInfo* tmp;
	int iLockRoot;
	int64 tms_flush;

	lockStaticMutex(&translatorlock);
	/* Convert the signature into a simple string of types, and
	 * count the size of the arguments too.
	 */
	sig = xmeth->sig;
	if (METHOD_IS_STATIC(xmeth)) {
		isStatic = 1;
		count = 0;
	}
	else {
		isStatic = 0;
		count = 1;
	}

	for (i = 0; i < sig->len; i++) {
		buf[i] = sig->elem[i]->data[0];
		switch (buf[i]) {
		case 'D':
		case 'J':
			count++;
			break;
		default:
			break;
		}
		count++;
	}

	/* Construct a wrapper to call the JNI method with the correct
	 * arguments.
	 */
	maxArgs = count;
	maxLocal = count;
	initInsnSequence(0, maxLocal + !isStatic, 0);
	start_basic_block();
	prologue(xmeth);
        if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) {
                softcall_trace(xmeth);
        }

	/* Start a JNI call */
#if defined(NEED_JNIREFS)
	prepare_function_call();
	call_soft(startJNIcall);
	fixup_function_call();

	/* Make the necesary JNI ref calls first */
	if (!METHOD_IS_STATIC(xmeth)) {
		pusharg_ref(local(0), 0);
		end_sub_block();
		call_soft(addJNIref);
		popargs();
	}
	j = i;
	jcount = count;
	while (j > 0) {
		j--;
		jcount--;
		if (buf[j] == '[' || buf[j] == 'L') {
			pusharg_ref(local(jcount), 0);
			end_sub_block();
			call_soft(addJNIref);
			popargs();
		}
		if (buf[j] == 'J' || buf[j] == 'D') {
			jcount--;
		}
	}
	start_sub_block();
#endif

	/* Add synchronisation if necessary */
	if (xmeth->accflags & ACC_SYNCHRONISED) {
		mon_enter(xmeth, local(0));
	}

#if defined(PUSHARG_FORWARDS)
       /* Push the JNI info */
       pusharg_ref_const((void*)&Kaffe_JNIEnv, 0);
 
       /* If static, push the class, else push the object */
       if (METHOD_IS_STATIC(xmeth)) {
               pusharg_ref_const(xmeth->class, 1);
       }
       else {
               pusharg_ref(local(0), 1);
       }
 
       i = 0;
       count = !isStatic;
       /* Push the specified arguments */
       while (i < sig->len) {
               switch (buf[i]) {
               case '[':
               case 'L':
                       pusharg_ref(local(count), count+1+isStatic);
                       break;
               case 'I':
               case 'Z':
               case 'S':
               case 'B':
               case 'C':
                       pusharg_int(local(count), count+1+isStatic);
                       break;
               case 'F':
                       pusharg_float(local(count), count+1+isStatic);
                       break;
               case 'J':
                       pusharg_long(local(count), count+1+isStatic);
                       count++;
                       break;
               case 'D':
                       pusharg_double(local(count), count+1+isStatic);
                       count++;
                       break;
               default:
                       ABORT();
               }
               i++;
               count++;
       }
 
#else
	/* Push the specified arguments */
	while (i > 0) {
		i--;
		count--;
		switch (buf[i]) {
		case '[':
		case 'L':
			pusharg_ref(local(count), count+1+isStatic);
			break;
		case 'I':
		case 'Z':
		case 'S':
		case 'B':
		case 'C':
			pusharg_int(local(count), count+1+isStatic);
			break;
		case 'F':
			pusharg_float(local(count), count+1+isStatic);
			break;
		case 'J':
			count--;
			pusharg_long(local(count), count+1+isStatic);
			break;
		case 'D':
			count--;
			pusharg_double(local(count), count+1+isStatic);
			break;
		default:
			ABORT();
		}
	}

	/* If static, push the class, else push the object */
	if (METHOD_IS_STATIC(xmeth)) {
		pusharg_ref_const(xmeth->class, 1);
	}
	else {
		pusharg_ref(local(0), 1);
	}

	/* Push the JNI info */
	pusharg_ref_const((void*)&Kaffe_JNIEnv, 0);
#endif

	/* Make the call */
	end_sub_block();
	call_soft(func);
	popargs();

	start_sub_block();

	/* Determine return type */
	switch (sig->ret->data[0]) {
	case 'I':
	case 'Z':
	case 'S':
	case 'B':
	case 'C':
		slot_alloctmp(tmp);
		return_int(tmp);
		/* Remove synchronisation if necessary */
		if (xmeth->accflags & ACC_SYNCHRONISED) {
			mon_exit(xmeth, local(0));
		}
		end_sub_block();
		call_soft(finishJNIcall);
		start_sub_block();
		returnarg_int(tmp);
		break;
	case 'F':
		slot_alloctmp(tmp);
		return_float(tmp);
		/* Remove synchronisation if necessary */
		if (xmeth->accflags & ACC_SYNCHRONISED) {
			mon_exit(xmeth, local(0));
		}
		end_sub_block();
		call_soft(finishJNIcall);
		start_sub_block();
		returnarg_float(tmp);
		break;
	case 'J':
		slot_alloc2tmp(tmp);
		return_long(tmp);
		/* Remove synchronisation if necessary */
		if (xmeth->accflags & ACC_SYNCHRONISED) {
		mon_exit(xmeth, local(0));
		}
		end_sub_block();
		call_soft(finishJNIcall);
		start_sub_block();
		returnarg_long(tmp);
		break;
	case 'D':
		slot_alloc2tmp(tmp);
		return_double(tmp);
		/* Remove synchronisation if necessary */
		if (xmeth->accflags & ACC_SYNCHRONISED) {
			mon_exit(xmeth, local(0));
		}
		end_sub_block();
		call_soft(finishJNIcall);
		start_sub_block();
		returnarg_double(tmp);
		break;
	case 'V':
		/* Remove synchronisation if necessary */
		if (xmeth->accflags & ACC_SYNCHRONISED) {
			mon_exit(xmeth, local(0));
		}
		call_soft(finishJNIcall);
		break;
	case 'L':
	case '[':
		slot_alloctmp(tmp);
		return_ref(tmp);
		/* Remove synchronisation if necessary */
		if (xmeth->accflags & ACC_SYNCHRONISED) {
			mon_exit(xmeth, local(0));
		}
		end_sub_block();
		call_soft(finishJNIcall);
		start_sub_block();
		returnarg_ref(tmp);
		break;
	default:
		ABORT();
	}

	epilogue(0);
	ret();

	/* Generate the code */
	if (tmpslot > maxTemp) {
		maxTemp = tmpslot;
	}
	finishInsnSequence(&ncode);

	/* Install it */
	SET_METHOD_NATIVECODE(xmeth, ncode.code);
	xmeth->accflags |= ACC_JNI;
        MSR_ADD(ncode_mem, ncode.codelen);

	/* Flush code out of cache */
	tms_flush = 0;
	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tms_flush = currentTime();
	}
#if defined(FLUSH_DCACHE)
	FLUSH_DCACHE(ncode.code, (void*)((uintp)ncode.code + ncode.codelen));
#endif
	SET_METHOD_FLUSHED(xmeth);
	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		jitStats.time_flush += currentTime() - tms_flush;
	}

	unlockStaticMutex(&translatorlock);

	return (1);
}

/* 
 * Define the engine interface.
 */
EngineInterface Kaffe_EngineInterface = {

	Einit,

	EgetName,
	EgetVersion,
	EgetNumber,

	EsetupTrampoline,
	EprocessExceptionMethod,
	EcallInterEngine,
	EcallExceptionHandler,
	EsetupJNIcall,
};

typedef struct JitContext {
	struct JitContext*	next;
	codeinfo*		codeInfo;
        uint16			maxPC;
        uint16			maxLocal;
} JitContext;

static JitContext* lastContext;
static JitContext* saveContext;

/*
 * Push JIT context.
 */
static
void
pushContext(void)
{
	JitContext* nc;

	if (saveContext != 0) {
		nc = saveContext;
		saveContext = nc->next;
	}
	else {
		nc = gc_malloc_fixed(sizeof(JitContext));
	}

	codeInfo = nc->codeInfo;
	ciMaxPC = nc->maxPC;
	ciMaxLocal = nc->maxLocal;

	nc->next = lastContext;
	lastContext = nc;
}

static
void
popContext(void)
{
	JitContext* nc;

	nc = lastContext;
	lastContext = nc->next;
	nc->next = saveContext;
	saveContext = nc;

	nc->codeInfo = codeInfo;
	nc->maxPC = ciMaxPC;
	nc->maxLocal = ciMaxLocal;
}
