/* constpool.c
 * Manage the constant pool.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "constpool.h"
#include "gc.h"
#include "md.h"

#include <stdarg.h>

#if !defined(ALIGNMENT_OF_SIZE)
#define	ALIGNMENT_OF_SIZE(S)	(S)
#endif

extern uint32 pc;

constpool* firstConst;
constpool* lastConst;
constpool* currConst;
constpool* availConst;
uint32 szConst;

/*
 * Allocate a new constant.
 */
constpool*
newConstant(int type, ...)
{
	constpool *c;
	union _constpoolval val;
	va_list arg;
	int sz;

	memset(&val, 0, sizeof(val));
	va_start(arg, type);
	switch (type) {
	case CPint:
		val.i = va_arg(arg, jint);
		sz = sizeof(jint);
		break;
	case CPlong:
		val.l = va_arg(arg, jlong);
		sz = sizeof(jlong);
		break;
	case CPref:
	case CProutine:
	case CPstring:
		val.r = va_arg(arg, void*);
		sz = sizeof(jref);
		break;
	case CPfloat:
		/* Remember ISO rules about argument widening.  */
		val.f = (float)va_arg(arg, double);
		sz = sizeof(jfloat);
		break;
	case CPdouble:
		val.d = va_arg(arg, double);
		sz = sizeof(jdouble);
		break;
	case CPlabel:
		sz = sizeof(jref);
		break;
	default:
		ABORT();
		return (0);
	}
	va_end(arg);

	if (availConst != 0 && type != CPlabel) {
		/* Find out if we've allocated this constant (or its bit
		   equivalent) before, and if so, reuse it.  */
		for (c = availConst; c != currConst; c = c->next) {
			if (memcmp(&c->val, &val, sz) == 0) {
				if (c->sz < sz) {
					szConst = szConst - c->sz + sz;
					c->sz = sz;
				}
				return (c);
			}
		}
	}
	else {
		c = currConst;
	}

	if (!c) {
		int i;

		/* Allocate chunk of label elements */
		c = gc_calloc_fixed(ALLOCCONSTNR, sizeof(constpool));

		if (availConst == 0) {
			availConst = c;
		}

		/* Attach to current chain */
		if (lastConst == 0) {
			firstConst = c;
			availConst = firstConst;
		}
		else {
			lastConst->next = c;
		}
		lastConst = &c[ALLOCCONSTNR-1];

		/* Link elements into list */
		for (i = 0; i < ALLOCCONSTNR-1; i++) {
			c[i].next = &c[i+1];
		}
		c[ALLOCCONSTNR-1].next = NULL;
	}

	c->type = type;
	c->val = val;
	c->sz = sz;
	currConst = c->next;
	szConst += sz;
	return (c);
}

void
establishConstants(uintp base)
{
	constpool *c;

	for (c = firstConst; c != currConst; c = c->next) {
		c->at += base;
		switch (c->sz) {
		case 4:
			*(jint*)c->at = c->val.i;
			break;
		case 8:
			*(jlong*)c->at = c->val.l;
			break;
		}
	}

	availConst = firstConst;
	currConst = firstConst;
	szConst = 0;
}

uintp
inlineConstants(uintp at)
{
	constpool *c;

	for (c = availConst; c != currConst; c = c->next) {
		int align = ALIGNMENT_OF_SIZE(c->sz);
		at = (at + align - 1) & -align;
		c->at = at;
		at += c->sz;
	}
	availConst = currConst;
	szConst = 0;
	return (at);
}
