/* machine.c
 * Translate the Kaffe instruction set to the native one.
 *
 * Copyright (c) 1996-1999
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define SCHK(s)
#define	SUSE(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "md.h"
#include "classMethod.h"
#include "callMethodInfo.h"
#include "bytecode.h"
#include "slots.h"
#include "registers.h"
#include "seq.h"
#include "gc.h"
#include "machine.h"
#include "basecode.h"
#include "icode.h"
#include "labels.h"
#include "constpool.h"
#include "codeproto.h"
#include "checks.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "baseClasses.h"
#include "code.h"
#include "access.h"
#include "lookup.h"
#include "exception.h"
#include "support.h"
#include "locks.h"
#include "code-analyse.h"
#include "external.h"
#include "soft.h"
#include "threads.h"
#include "engine.h"
#include "itypes.h"
#include "code.h"
#include "strings.h"
#include "jsyscall.h"
#include "patch.h"

#include "jdebug.h"
#include "jmeasure.h"

char* engine_name = "jit";
char* engine_version = KAFFEVERSION;

/*
 * If we're optimizing memory then we limit the number of
 * instructions we emit in a basic block so we don't use up
 * too much sequence memory.
 */
#if defined(OPTIMIZE_MEMORY)
#define	INSTRUCTION_LIMIT	500
#endif

/**
 * On some architectures (SH3) we need to know the framesize
 * at the beginning.  The only bit that changes is the temps
 * so we can force a minimum value.
 */
#if !defined(MINTEMPS)
#define	MINTEMPS	0
#endif

int stackno;
int maxStack;
int maxLocal;
int maxTemp;
int maxArgs;
int maxPush;
int isStatic;
codeinfo* codeInfo;

int tmpslot;
int argcount = 0;		/* Function call argument count */
uint32 pc;
uint32 npc;
uint32 ipc;
#if defined(OPTIMIZE_MEMORY)
uint32 lpc;
#endif
Method* cmeth;

/* Various exception related things */
extern const struct JNIEnv_ Kaffe_JNIEnv;
extern Method* jniMethod;
extern void startJNIcall(void);
extern void finishJNIcall(void);
extern void addJNIref(jref);
extern void removeJNIref(jref);

jitflags willcatch;
jmp_buf tjbuf;

/* jit3 specific prototypes from icode.c */
void check_null (int x, SlotInfo* obj, int y);
void force_check_null (int x, SlotInfo* obj, int y);
void check_div (int x, SlotInfo* obj, int y);
void check_div_long (int x, SlotInfo* obj, int y);
void softcall_fakecall (label* from,label* to, void* func);

/* Unit in which code block is increased when overrun */
#define	ALLOCCODEBLOCKSZ	8192
/* Codeblock redzone - allows for safe overrun when generating instructions */
#define	CODEBLOCKREDZONE	256

nativecode* codeblock;
int codeblock_size = -1;
static int code_generated;
static int bytecode_processed;
static int codeperbytecode;
iLock* translatorlock;

int CODEPC;

struct {
	int time;
} jitStats;

extern int enable_readonce;
extern int inFuncSetup;

int initInsnSequence(Method*, int, int, int);
int finishInsnSequence(nativeCodeInfo*);
static void generateInsnSequence(void);
static int installMethodCode(Method*, Hjava_lang_Class*, nativeCodeInfo*);
static void nullCall(void);
static void throwAbstractMethodError(void);
static int countArgs(Method*, int*);

Method*	findMethodFromPC(uintp);
label* newFakeCall(void*, uintp);
static void makeFakeCalls(void);
static void inline_constants(void);

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#if defined(JIT4)
#include "jit3stub.def"
#endif

/*
 * Translate a method into native code.
 */
static
int
translate(Method* meth, Hjava_lang_Class* clazz)
{
	jint low;
	jint high;
	jvalue tmpl;
	int idx;
	SlotInfo* tmp;
	SlotInfo* tmp2;
	SlotInfo* mtable;

	bytecode* base;
	uint32 len;
	callInfo cinfo;
	fieldInfo finfo;
	Hjava_lang_Class* crinfo;
	codeinfo* mycodeInfo;
	char* cname;
#if defined(JIT4)
        int bytecode_len;
#endif
	nativeCodeInfo ncode;

	basicBlockInfo* currBB;

	int64 tms = 0;
	int64 tme;

	int iLockRoot;

	static bool reinvoke = false;
	MSR_TIME( jit_time);
	MSR_START( jit_time);

	if (METHOD_TRANSLATED(meth)) {
		MSR_STOP( jit_time);
		return (1);
	}

	/* Abstract?  Point at an exception */
	else if ((meth->accflags & ACC_ABSTRACT) != 0) {
		SET_METHOD_NATIVECODE(meth, (void*)throwAbstractMethodError, 0);
		MSR_STOP( jit_time);
		return (1);
	}

	/* If this code block is native, then just set it up and return */
	else if ((meth->accflags & ACC_NATIVE) != 0) {
		native(meth, clazz);
		KAFFEJIT_TO_NATIVE(meth);
		MSR_STOP( jit_time);
		return (1);
	}

	/* Handle null calls specially */
	else if (BCODE(meth)->codelen == 1 && BCODE(meth)->code[0] == RETURN) {
#ifndef JIT4
		gc_free_fixed(BCODE(meth));
#endif
		BCODE(meth) = 0;
		SET_METHOD_NATIVECODE(meth, (nativecode*)nullCall, 0);
		MSR_STOP( jit_time);
		return (1);
	}
#ifndef JIT4
	/* Scan the code and determine the basic blocks */
	else if (verifyMethod(meth, clazz, &mycodeInfo) == 0) {
                MSR_STOP( jit_time);
                return (0);
	}

#endif

	/* Only one in the translator at once. Must check the translation
	 * hasn't been done by someone else once we get it.
	 */
	lockStaticMutex(&translatorlock);

#if defined(JIT4)
        if( invoke_new( meth, clazz ) ){
        	goto compiled;
        }
        bytecode_len = METHOD_BYTECODE_LEN(meth);

	/* Scan the code and determine the basic blocks */
	if (verifyMethod(meth, clazz, &mycodeInfo) == 0) {
                goto error;
	}
#endif

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tms = currtime();
	}

DBG( vm_jit_translate, ("callinfo = 0x%x\n", &cinfo));

	/* This is horrible, but it's the only way we can deal with
	 * out-of-memory conditions in various deeply nested routines.
	 */
	if (setjmp(tjbuf) != 0) {
		goto error;
	}

	assert(reinvoke == false);
	reinvoke = true;
	codeInfo = mycodeInfo;
	currBB = codeInfo->bbinfo;

	maxLocal = BCODE(meth)->localsz;
	maxStack = BCODE(meth)->stacksz;
	maxArgs = countArgs(meth, &isStatic);

SUSE(
        kprintf("Method: %s.%s%s\n", CLASS_CNAME(clazz), meth->name, makeStrFromSignature(meth->sig));
	for (i = 0; i < maxLocal; i++) {
		kprintf(" L%d: %2d", i, codeInfo->localuse[i].use);
	}
	kprintf("\n");
)

DBG( vm_jit_translate, ("Method: %s.%s%s\n", CLASS_CNAME(clazz), meth->name, meth->signature));

	base = (bytecode*)BCODE(meth)->code;
	len = BCODE(meth)->codelen;

	willcatch.ANY = false;
	willcatch.BADARRAYINDEX = false;

	/* Deterimine various exception conditions */
	if (meth->exception_table != 0 && (meth->exception_table[1].start_pc != (uintp)-1 || !METHOD_IS_SYNCHRONISED(meth))) {
		/* If we have any exceptions then note then */
		willCatch(ANY);
		willCatch(BADARRAYINDEX);
	}

	/*
	 * Initialise the translator.
	 */
	if (initInsnSequence(meth, codeperbytecode * BCODE(meth)->codelen, BCODE(meth)->localsz, BCODE(meth)->stacksz) == 0) {
		goto error;
	}

	/***************************************/
	/* Next reduce bytecode to native code */
	/***************************************/

	cmeth = meth;
	pc = 0;
	start_function();
	check_stack_limit();
	if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) {
		softcall_trace(meth, clazz);
	}
	monitor_enter();

	ipc = pc;
#if defined(OPTIMIZE_MEMORY)
	lpc = pc;
#endif
	for (; pc < len; pc = npc) {

		ADVANCE_BB(pc);

		/* Skip unreachable code */
		if (IS_UNREACHABLE(pc)) {
			npc = currBB->next->start;
#if defined(USE_JIT_INLINE_CONSTANT_POOL)
			generateInsnSequence();
			CODEPC = inlineConstants(CODEPC);
			ipc = pc;
#endif
			continue;
		}

		npc = pc + insnLen[base[pc]];

		/* If start of a new basic block, finish up the old
		 * one, generate the instructions, then start the new block.
		 */
		if (IS_STARTOFBASICBLOCK(pc)) {
			end_basic_block();
			generateInsnSequence();
#if defined(USE_JIT_INLINE_CONSTANT_POOL)
			/* If we're inlining constants and this is the
			 * beginning of a new basic block which we only reach
			 * by jumping to, we add the constants now.
			 */
			if (!IS_NORMALFLOW(pc)) {
				CODEPC = inlineConstants(CODEPC);
				ipc = pc;
			}
#endif

SCHK(			sanityCheck();				)
			stackno = STACKPOINTER(pc);
			start_basic_block();
		}

#if defined(OPTIMIZE_MEMORY)
		/* If we've emitted too many instructions without a
		 * break we flush them now to avoid using too much memory.
		 */
		if (pc - lpc > INSTRUCTION_LIMIT) {
			begin_flush();
			generateInsnSequence();
SCHK(			sanityCheck();				)
			end_flush();
			lpc = pc;
		}
#endif
#if defined(MAX_INLINE_CONSTANT_POOL_DISTANCE)
		/* If we can only load constants from so far we need to
		 * make sure we inline them every so often.
		 */
		if (MAX_INLINE_CONSTANT_POOL_DISTANCE((pc - ipc) * codeperbytecode, szConst)) {
DBG( vm_jit_translate, ("inlining from %d to %d for %d constants\n", ipc, pc, szConst));
			label* l = newLabel();
			l->type |= Linternal|Linstruction;
			begin_flush();
			branch_a(l);
			generateInsnSequence();
SCHK(			sanityCheck();				)
			CODEPC = inlineConstants(CODEPC);
			fix_label(l);
			end_flush();
			ipc = pc;
		}
#endif

		start_instruction();

		/* Note start of exception handling blocks */
		if (IS_STARTOFEXCEPTION(pc)) {
			stackno = BCODE(meth)->localsz + BCODE(meth)->stacksz - 1;
			start_exception_block();
		}

		assert(stackno <= maxStack+maxLocal);
	        assert(stackno >= 0);

		switch (base[pc]) {
		default:
			kprintf("Unknown bytecode %d\n", base[pc]);
			/* this probably does not work */
			postException("java.lang.VerifyError", CLASS_CNAME(clazz));
			goto error;
#include "kaffe.def"
		}

		/* Note maximum number of temp slots used and reset it */
		if (tmpslot > maxTemp) {
			maxTemp = tmpslot;
		}
		tmpslot = 0;

SCHK(		sanityCheck();					)
	}

	/* We must build a catch-all
	 * exception handler it this method is synchronous so we can unlock
	 * it during exception handling.  The exception will already have
	 * been prepared above.
	 */
	if (METHOD_IS_SYNCHRONISED(meth)) {
		mon_exit(meth, clazz, local(0));
		softcall_athrow(stack(0));
	}

	finish_function();
	makeFakeCalls();

	assert(maxTemp < MAXTEMPS);

	if (finishInsnSequence(&ncode) == 0) {
		goto error;
	}
	if (installMethodCode(meth, clazz, &ncode) == 0) {
		goto error;
	}

#if defined(JIT4)
        _DISASM_method( meth, clazz, &ncode, 0, bytecode_len ); 
 compiled:
#endif

DBG( vm_jit_translate, ("Translating %s.%s%s (%s) %p\n",
						CLASS_CNAME(clazz),
						meth->name, meth->signature, isStatic ? "static" : "normal", METHOD_NATIVECODE(meth)));

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tme = currtime();
		jitStats.time += (int)(tme - tms);
		kprintf("<JIT: %s.%s%s time %dms (%dms) @ %p (%p)>\n",
		       CLASS_CNAME(clazz),
		       meth->name, meth->signature,
		       (int)(tme - tms), jitStats.time,
		       METHOD_NATIVECODE(meth), meth);
	}

	reinvoke = false;
	tidyVerifyMethod(&codeInfo);
	unlockStaticMutex(&translatorlock);
	MSR_STOP(jit_time);
	return (1);

error:;
	reinvoke = false;
	tidyVerifyMethod(&codeInfo);
	unlockStaticMutex(&translatorlock);
	MSR_STOP(jit_time);
	return (0);
}

/*
 * Generate the code.
 */
int
finishInsnSequence(nativeCodeInfo* code)
{
	nativecode* methblock;
	nativecode* codebase;
	int len;
	int sz;

	/* Emit pending instructions */
	generateInsnSequence();

	/* Work out final codesize */
#if defined(USE_JIT_INLINE_CONSTANT_POOL)
	CODEPC = inlineConstants(CODEPC);
	sz = 0;
#else
	sz = inlineConstants(0);
#endif
	len = CODEPC + sz;

	/* Okay, put this into malloc'ed memory */
	methblock = gc_malloc_fixed_typed(len, GC_ALLOC_JITCODE);
	if (methblock == 0) {
		return (0);
	}
	codebase = (nativecode*)(methblock + sz);
	memcpy(codebase, codeblock, CODEPC);

	/* Establish any code constants */
#if defined(USE_JIT_INLINE_CONSTANT_POOL)
	establishConstants((uintp)codebase);
#else
	establishConstants((uintp)methblock);
#endif

	/* Link it */
	linkLabels((uintp)codebase);

	/* Note info on the compiled code for later installation */
	code->mem = methblock;
	code->memlen = len;
	code->code = codebase;
	code->codelen = CODEPC;

	return (1);
}

/*
 * Get the instruction offset corresponding to the given PC.
 * If the PC doesn't point at the start of a valid instruction,
 * look forward until we find one.
 */
static
int
getInsnPC(int pc)
{
	int res;

	for (;;) {
		res = INSNPC(pc);
		if (res != -1) {
			return (res);
		}
		pc++;
	}
}

/*
 * Install the compiled code in the method.
 */
static
int
installMethodCode(Method* meth, Hjava_lang_Class* clazz, nativeCodeInfo* code)
{
	uint32 i;
	jexceptionEntry* e;
	void* end;

	/* Work out new estimate of code per bytecode */
	code_generated += code->memlen;
	bytecode_processed += BCODE(meth)->codelen;
	codeperbytecode = code_generated / bytecode_processed;

	/* Free the bytecode */
#ifndef JIT4
	gc_free_fixed(BCODE(meth));
#endif
	BCODE(meth) = 0;
        MSR_ADD(ncode_mem, code->codelen);

	/* Install the new code */
	end = (void*)((uintp)code->code + code->codelen);
	SET_METHOD_NATIVECODE(meth, code->code, end);

	/* Flush code out of cache */
#if defined(FLUSH_DCACHE)
	FLUSH_DCACHE(code->code, end);
#endif

	/* Translate exception table and make it available */
	if (meth->exception_table != 0) {
		for (i = 0; meth->exception_table[i].start_pc != (uintp)-1; i++) {
			e = &meth->exception_table[i];
			e->start_pc = getInsnPC(e->start_pc) + (uintp)code->code;
			e->end_pc = getInsnPC(e->end_pc) + (uintp)code->code;
			e->handler_pc = getInsnPC(e->handler_pc) + (uintp)code->code;
		}
	}

#if !defined(OPTIMIZE_MEMORY) || defined(DEBUG)
	/* Translate line numbers table */
	if (meth->lines != 0) {
		uint32 last = 0;
		for (i = 0; i < meth->lines->length; i++) {
			int insnpc = getInsnPC(meth->lines->entry[i].pc_offset);
			meth->lines->entry[i].pc_offset = insnpc - last;
			last = insnpc;
		}
	}
#endif

	/* Make method active (so we can find if for exceptions) */
	if (makeMethodActive(meth, clazz, code->code, end) == 0) {
		return (0);
	}

#if defined(DEBUG) && !defined(WIN32)
/* pjo, 12 may 2000 */
	if (Kaffe_JavaVMArgs[0].kxFile) {
		char buffer[256] ;
		snprintf(buffer, sizeof(buffer),
			"J%#x %#x %s.%s%s\n",
		       (uintp)code->code, (uintp)code->code + code->codelen,
		       CLASS_CNAME(clazz), meth->name, meth->signature);
		write(Kaffe_JavaVMArgs[0].kxFile, buffer, strlen(buffer)) ;
		fsync(Kaffe_JavaVMArgs[0].kxFile) ;
	}
#endif

	return (1);
}

/*
 * Init instruction generation.
 */
int
initInsnSequence(Method* meth, int codesize, int localsz, int stacksz)
{
	/* Clear various counters */
	tmpslot = 0;
	maxTemp = MINTEMPS;
	maxPush = 0;  
	stackno = localsz + stacksz;
	npc = 0;

	/* Do any machine dependent JIT initialization */
#if defined(INIT_JIT_MD)
	INIT_JIT_MD(meth);
#endif

	initSeq();
	initRegisters();
	initSlots(stackno);
	resetLabels();

	CODEPC = 0;
	if (codesize > codeblock_size) {
		codeblock_size = codesize;
		if (codeblock_size < ALLOCCODEBLOCKSZ) {
			codeblock_size = ALLOCCODEBLOCKSZ;
		}
		codeblock = gc_realloc_fixed(codeblock, codeblock_size + CODEBLOCKREDZONE);
	}

	if (codeblock == 0) {
		return (0);
	}
	else {
		return (1);
	}
}

/*
 * Generate instructions from current set of sequences.
 */
static
void
generateInsnSequence(void)
{
	sequence* t;
	int i;
	int m;

	for (t = firstSeq; t != currSeq; t = t->next) {

		/* If we overrun the codeblock, reallocate and continue.  */
		if (CODEPC >= codeblock_size) {
			codeblock_size += ALLOCCODEBLOCKSZ;
			codeblock = gc_realloc_fixed(codeblock, codeblock_size + CODEBLOCKREDZONE);
		}

SCHK(		sanityCheck();					)

		/* Generate sequences */
		assert(t->func != 0);
		if (t->refed != 0) {
			(*(t->func))(t);
		}
		else {
			/* printf("discard instruction\n"); */
		}

		/* Handle dead slots */
		m = t->lastuse;
		if (m != 0) {
			for (i = 0; m != 0; m = m >> 1, i++) {
				if ((m & 1) != 0) {
					assert(!isGlobal(seq_slot(t, i)));
					slot_invalidate(seq_slot(t, i));
				}
			}
		}
	}

	/* Reset */
	initSeq();
}

/*
 * Start a new instruction.
 */
void
startInsn(sequence* s)
{
	SET_INSNPC(const_int(2), CODEPC);
}

void
doSpill(sequence* s)
{
	SlotData** mem;
	SlotData* sd;
	int type;
	int old_ro;

SCHK(	sanityCheck();						)

	type = s->u[2].value.i;

	old_ro = enable_readonce;
	if (type == SR_SYNC) {
		enable_readonce = 0;
	}

	/* Spill the registers */
	for (mem = s->u[1].smask; *mem != 0; mem++) {
		sd = *mem;

		/* Determine if we need to spill this slot at all */
		if ((sd->modified & rwrite) != 0 && sd->regno != NOREG) {
			switch (type) {
			case SR_BASIC:
			case SR_SUBBASIC:
				/* We only spill if it's not global */
				if (!isGlobal(sd)) {
					spillAndUpdate(sd, true);
				}
				break;

			case SR_SYNC:
				spillAndUpdate(sd, false);
				break;

			case SR_FUNCTION:
				if (calleeSave(sd->regno) == 0 || canCatch(ANY) != 0) {
					spillAndUpdate(sd, true);
				}
				break;

			default:
				ABORT();
			}
		}
	}

	/* If this isn't being done for a function call we can free the
	 * spill mask now.  Otherwise it will be freed by the function
	 * reloader.
	 */
	if (type != SR_FUNCTION) {
		gc_free_fixed(s->u[1].smask);
	}
	/* If this is being done for a function, make a note for the register
	 * allocator.
	 */
	else {
		inFuncSetup = 1;
	}

	enable_readonce = old_ro;

SCHK(	sanityCheck();						)
}

void
doReload(sequence* s)
{
	SlotData* sd;
	SlotData** mem;
	int type;

SCHK(	sanityCheck();						)

	type = s->u[2].value.i;

        // clear slot-register associations   
	for (mem = s->u[1].smask; *mem != 0; mem++) {
		sd = *mem;
		if (sd->regno != NOREG && !isGlobal(sd)) {
			switch (type) {
			case SR_BASIC:
			case SR_SUBBASIC:
				slot_invalidate(sd);
				break;

			case SR_FUNCTION:
				if (calleeSave(sd->regno) == 0) {
					slot_invalidate(sd);
				}
				break;

			case SR_START:
			case SR_EXCEPTION:
				break;

			default:
				ABORT();
			}
		}
	}

	/* Deal with global reloading */
	for (mem = s->u[1].smask; *mem != 0; mem++) {
		sd = *mem;
		if (isGlobal(sd)) {
			switch (type) {
			case SR_BASIC:
			case SR_SUBBASIC:
				if (!isGlobalReadonly(sd)) {
					sd->modified = rwrite;
				}
				break;

			case SR_FUNCTION:
				break;

			case SR_START:
				if (isGlobalPreload(sd)) {
					reload(sd);
				}
				break;

			case SR_EXCEPTION:
				reload(sd);
				break;

			default:
				ABORT();
			}
		}
	}

	gc_free_fixed(s->u[1].smask);

	/* If this is the end of a function, clear the flag */
	if (type == SR_FUNCTION) {
		inFuncSetup = 0;
	}

	/* Clear the register notes */
	clearRegisterNotes(type == SR_FUNCTION ? 0 : 1);

SCHK(	sanityCheck();						)
}

/*
 * Create the spill/reload mask.
 * This contains a list of slot/global pairs indicating which slots are
 * active and should be spilled or reloaded now.  The global flag indicates
 * that at this point the slot is global (ie. carried between basic blocks)
 * so should be handled with care.
 */
SlotData**
createSpillMask(void)
{
	SlotData** mem;
	SlotData* d;
	int i;
	int c;

	/* Count the number of slots we need to remember.  Note: we
	 * include the number of potentially active temps.
	 */
	i = maxLocal + maxStack + tmpslot;
	c = 0;
	for (i--; i >= 0; i--) {
		d = slotinfo[i].slot;
		if (d->rseq != 0 || d->wseq != 0 || isGlobal(d)) {
			c++;
		}
	}
#if defined(STACK_LIMIT)
	d = stack_limit->slot;
	if (d->rseq != 0 || d->wseq != 0) {
		c++;
	}
#endif

	c++; /* Add null slot on the end */
	mem = gc_malloc_fixed_typed(c * sizeof(SlotData*), GC_ALLOC_ENGINE);
	if (mem == 0) {
		longjmp(tjbuf, 1);
	//	return (0);
	}
	i = maxLocal + maxStack + tmpslot;
	c = 0;
	for (i--; i >= 0; i--) {
		d = slotinfo[i].slot;
		if (d->rseq != 0 || d->wseq != 0 || isGlobal(d)) {
			mem[c++] = d;
		}
	}
#if defined(STACK_LIMIT)
	d = stack_limit->slot;
	if (d->rseq != 0 || d->wseq != 0) {
		mem[c++] = d;
	}
#endif

	return (mem);
}

static
int
sortSlots(const void* s1, const void* s2)
{
	localUse* lcluse1;
	localUse* lcluse2;
	
	lcluse1 = &codeInfo->localuse[*((SlotInfo**)s1) - localinfo];
	lcluse2 = &codeInfo->localuse[*((SlotInfo**)s2) - localinfo];

	if (lcluse1->use == lcluse2->use) {
		return (0);
	}
	else if (lcluse1->use > lcluse2->use) {
		return (-1);
	}
	else {
		return (1);
	}
}

static
void
bindSlotToGlobal(int lcl, int r, int type)
{
	kregs* reg;
	SlotData* dslot;
	int gtype;

	reg = &reginfo[r];
	dslot = localinfo[lcl].slot;

	reg->ctype = type;
	reg->type |= Rglobal;
	reg->refs++;
	reg->slot = dslot;
	dslot->regno = r;

	gtype = GL_ISGLOBAL;

	/* If this is an argument then pre-load */
	if (lcl < maxArgs) {
		gtype |= GL_PRELOAD;
	}
	/* If this slot is never written note that the global is
	 * read only.
	 */
	if (codeInfo->localuse[lcl].write == -1) {
		gtype |= GL_RONLY;
	}
	setGlobal(dslot, gtype);
}

/*
 * Setup global registers
 */
int
setupGlobalRegisters(void)
{
#if defined(NR_GLOBALS)
	SlotInfo** slots;
	int j;
	int max;

	/* If we don't have any code info we can't do any global
	 * optimization
	 */
	if (codeInfo == 0 || maxLocal == 0) {
		return (1);
	}

	/* Allocate an array for the slot pointers and copy them in */
	slots = alloca(maxLocal * sizeof(SlotInfo*));
	for (j = 0; j < maxLocal; j++) {
		slots[j] = &localinfo[j];
	}

	/* Sort the slots to the most used is first */
	qsort(slots, maxLocal, sizeof(SlotInfo*), sortSlots);

	/* Allocate the slots to globals */
	max = NR_GLOBALS;
	for (j = 0; j < maxLocal && max > 0; j++) {
		int k;
		int i;
		localUse* lcl;

		i = slots[j] - localinfo;
		lcl = &codeInfo->localuse[i];

		for (k = 0; k < MAXREG; k++) {
			kregs* reg;

			reg = &reginfo[k];
			if ((reg->flags & Rglobal) == 0) {
				/* Not a global */
			}
			else if ((reg->type & Rglobal) != 0) {
				/* Already used */
			}
			else if (lcl->type == TINT && (reg->type & Rint) != 0) {
				bindSlotToGlobal(i, k, Rint);
				max--;
				break;
			}
			else if (lcl->type == TOBJ && (reg->type & Rref) != 0) {
				bindSlotToGlobal(i, k, Rref);
				max--;
				break;
			}
		}
	}
#endif
	return (1);
}

int
setupArgumentRegisters(void)
{
#if defined(NR_ARGUMENTS)
	int args;
	int i;
	static sequence argseq[1];

	args = maxArgs;
	if (args > NR_ARGUMENTS) {
		args = NR_ARGUMENTS;
	}

	for (i = 0; i < args; i++) {
		writeslot(argseq, i, &localinfo[i], 1);
		localinfo[i].slot->modified = rwrite;
	}
#if defined(STACK_LIMIT)
	if (args < NR_ARGUMENTS) {
		writeslot(argseq, i, stack_limit, 1);
		stack_limit->slot->modified = rwrite;
	}
#endif
#endif
	return (1);
}

/*
 * Build multi-dimensional array.
 *  This is sufficiently different from the standard
 *  to require our own routine.
 */
void*
jit_soft_multianewarray(Hjava_lang_Class* class, jint dims, ...)
{
	Hjava_lang_Object* obj;
	jint arg;
	int i;
	int* arraydims;
	va_list ap;

	arraydims = alloca((dims+1) * sizeof(int));

	/* Extract the dimensions into an array */
	va_start(ap, dims);
	for (i = 0; i < dims; i++) {
		arg = va_arg(ap, jint);
		if (arg < 0) {
			postAndThrowException("java.lang.NegativeArraySizeException", "");
		}
		arraydims[i] = arg;
	}
	arraydims[i] = -1;
	va_end(ap);

	/* Mmm, okay now build the array using the wonders of recursion */
        obj = newMultiArray(class, arraydims);
	if (obj == 0) {
		throwPostedException();
	}
	return (obj);
}

/*
 * Dispatch an exception.  These odd structures are necessary
 *  to emulate the old exception calling convention (we will
 *  deprecate at some point).
 */
void
Kaffe_EngineInterface_callExceptionHandler(Method* meth, uintp handler, stackTraceInfo* frame)
{
	CALL_KAFFE_EXCEPTION(frame->arg0, handler);
}

/* 
 * Trampolines come in here - do the translation and replace the trampoline.
 */
nativecode*
soft_fixup_trampoline(FIXUP_TRAMPOLINE_DECL)
{
	Method* meth;
	void* tramp;
	Hjava_lang_Class* clazz;
	int iLockRoot;

	FIXUP_TRAMPOLINE_INIT;

	/* We process the class before we start the trampoline code to
	 * handle any <clinit> methods if they're pending.  We must do
	 * this otherwise we get recursion of trampolines causing them
	 * to be freed multiple times (not good).
	 */

	processClass(clazz, CSTATE_COMPLETE);

	lockMutex(clazz);

	/* Generate code on demand.  */
	if (!METHOD_TRANSLATED(meth)) {

		tramp = METHOD_NATIVECODE(meth);

		if (translate(meth, clazz) == 0) {
			unlockMutex(clazz);
			throwPostedException();
			return (0);
		}
		DBG( vm_jit, ("%x : %s:%s%s\n", METHOD_NATIVECODE(meth), CLASS_CNAME(clazz), meth->name, meth->signature));
		assert(tramp != METHOD_NATIVECODE(meth));

		/* If method is native then we allocated a trampoline
		 * which was not part of the bytecode (which was already
		 * freed).  Better free it now.
		 */
#ifdef DIRECT_CALL
                finishMethodPatches( meth, clazz, METHOD_NATIVECODE(meth) );
#endif
#ifndef JIT4
		if (METHOD_IS_NATIVE(meth)) {
			gc_free_fixed(tramp);
		}
#endif
	}
	unlockMutex(clazz);

	return (METHOD_NATIVECODE(meth));
}

/*
 * Include the machine specific trampoline functions.
 */
#include "trampolines.c"

/*
 * Initialize the engine.
 */
void
Kaffe_EngineInterface_init(void)
{
	/* Does nothing */
}

char*
Kaffe_EngineInterface_getName(void)
{
	return ("Just-in-time");
}

char*
Kaffe_EngineInterface_getVersion(void)
{
	return (KAFFEVERSION);
}

void*
Kaffe_EngineInterface_setupTrampoline(Method* meth, Hjava_lang_Class* clazz)
{
	methodTrampoline* tramp;

	assert(!METHOD_TRANSLATED(meth));

	/* These are freed as methods are patched up */
	if (!METHOD_IS_NATIVE(meth)) {
		tramp = &BCODE(meth)->trampoline;
	}
	else {
		tramp = gc_malloc_fixed_typed(sizeof(methodTrampoline), GC_ALLOC_JITCODE);
	}

	if (tramp != 0) {
		FILL_IN_TRAMPOLINE(tramp, meth, clazz);
		meth->ncode = tramp;
#if defined(FLUSH_DCACHE)
		FLUSH_DCACHE(tramp, tramp+1);
#endif
	}

	return (tramp);
}

void
Kaffe_EngineInterface_processExceptionMethod(stackTraceInfo* info)
{
	info->object = (void*)FRAMEOBJECT(info->rframe);
	info->arg0 = (uintp)info->rframe;
}

static void engineDummy(void);

/*
 * Call inter-machine method
 */
void
Kaffe_EngineInterface_callInterEngine(callMethodInfo* call)
{
	engineDummy();
	sysdepCallMethod(call);
}

/* This function is here primarily to fake the compiler into thinking
 * EcallInterEngine isn't a leaf function (if it were then things would
 * break.
 */
static
void
engineDummy(void)
{
}

/*
 * Build an interface between the JIT and a JNI method.
 */
int
Kaffe_EngineInterface_setupJNIcall(Method* meth, Hjava_lang_Class* clazz, void* func)
{
	char* sig;
	int count;
	nativeCodeInfo ncode;
	SlotInfo* tmp;
	int j;
	int an;
#if !defined(PUSHARG_FORWARDS)
	char* args;
#endif
	char rettype;
	void* end;
	int iLockRoot;

	sig = 0;

	lockStaticMutex(&translatorlock);

	/* This is horrible, but it's the only way we can deal with
	 * out-of-memory conditions in various deeply nested routines.
	 */
	if (setjmp(tjbuf) != 0) {
		goto error;
	}

	/* Construct a wrapper to call the JNI method with the correct
	 * arguments.
	 */
	meth->accflags |= ACC_JNI;
	sig = meth->signature;
	maxArgs = countArgs(meth, &isStatic);
	maxLocal = maxArgs;
	maxStack = 0;
	if (initInsnSequence(meth, 0, maxLocal, maxStack) == 0) {
		goto error;
	}
	start_function();
	check_stack_limit();
        if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) {
                softcall_trace(meth, clazz);
        }

	/* Start a JNI call */
#if defined(NEED_JNIREFS)
					NEEDS FIXING!!!
	begin_func_sync();
	call_soft(startJNIcall);
	end_func_sync();

	/* Make the necesary JNI ref calls first */
	if (!METHOD_IS_STATIC(meth)) {
		pusharg_ref(local(0), 0);
		begin_func_sync();
		call_soft(addJNIref);
		popargs();
		end_func_sync();
	}
	count = maxArgs;
	for (j = sig->len - 1; j >= 0; j--) {
		count--;
		if (sig->elem[j]->data[0] == '[' || sig->elem[j]->data[0] == 'L') {
			pusharg_ref(local(count), 0);
			begin_func_sync();
			call_soft(addJNIref);
			popargs();
			end_func_sync();
		}
		else if (sig->elem[j]->data[0] == 'J' || sig->elem[j]->data[0] == 'D') {
			count--;
		}
	}
	start_sub_block();
#endif

	/* Add synchronisation if necessary */
	mon_enter(meth, clazz, local(0));

#if defined(PUSHARG_FORWARDS)
        /* Push the JNI info */
        pusharg_ref_const((void*)&Kaffe_JNIEnv, 0);
  
        /* If static, push the class, else push the object */
        if (isStatic) {
                pusharg_ref_const(clazz, 1);
		an = 0;
        }
        else {
                pusharg_ref(local(0), 1);
		an = 1;
        }
        count = 2;
  
        /* Push the specified arguments */
        for (j = 1; sig[j] != ')'; j++) {
		switch (sig[j]) {
                case '[':
                        pusharg_ref(local(an), count);
			while (sig[++j] == '[')
				;
			if (sig[j] == 'L') {
				while (sig[++j] != ';')
					;
			}
                        break;
                case 'L':
                        pusharg_ref(local(an), count);
			while (sig[++j] != ';')
				;
                        break;
                case 'I':
                case 'Z':
                case 'S':
                case 'B':
                case 'C':
                        pusharg_int(local(an), count);
                        break;
                case 'F':
                        pusharg_float(local(an), count);
                        break;
                case 'J':
                        pusharg_long(local(an), count);
                        count++;
			an++;
                        break;
                case 'D':
                        pusharg_double(local(an), count);
                        count++;
			an++;
                        break;
                default:
                        ABORT();
                }
                count++;
		an++;
        }
	rettype = sig[j+1];
#else
	args = alloca(strlen(sig));
	an = 0;
        for (j = 1; sig[j] != ')'; j++) {
		args[an++] = sig[j];
		switch (sig[j]) {
                case '[':
			while (sig[++j] == '[')
				;
			if (sig[j] == 'L') {
				while (sig[++j] != ';')
					;
			}
                        break;
                case 'L':
			while (sig[++j] != ';')
				;
                        break;
                case 'I':
                case 'Z':
                case 'S':
                case 'B':
                case 'C':
                case 'F':
                case 'J':
                case 'D':
                        break;
                default:
                        ABORT();
                }
        }
	rettype = sig[j+1];

	/* Push the specified arguments
	 * We must allow for the JNIEnv and the class (if the method is
	 * static).
	 */
	j = an - 1;
	count = maxArgs + 1;
	an = maxArgs;
	if (isStatic) {
		count++;
	}
	for (; j >= 0; j--) {
		count--;
		an--;
		switch (args[j]) {
		case '[':
		case 'L':
			pusharg_ref(local(an), count);
			break;
		case 'I':
		case 'Z':
		case 'S':
		case 'B':
		case 'C':
			pusharg_int(local(an), count);
			break;
		case 'F':
			pusharg_float(local(an), count);
			break;
		case 'J':
			an--;
			count--;
			pusharg_long(local(an), count);
			break;
		case 'D':
			an--;
			count--;
			pusharg_double(local(an), count);
			break;
		default:
			ABORT();
		}
	}

	/* If static, push the class, else push the object */
	if (METHOD_IS_STATIC(meth)) {
		pusharg_ref_const(clazz, 1);
	}
	else {
		pusharg_ref(local(0), 1);
	}

	/* Push the JNI info */
	pusharg_ref_const((void*)&Kaffe_JNIEnv, 0);
#endif

	/* Make the call */
	begin_func_sync();
	call_soft(func);
	popargs();
	end_func_sync();

	/* Determine return type */
	switch (rettype) {
	case 'I':
	case 'Z':
	case 'S':
	case 'B':
	case 'C':
		slot_alloctmp(tmp);
		return_int(tmp);
		/* Remove synchronisation if necessary */
		if (meth->accflags & ACC_SYNCHRONISED) {
			mon_exit(meth, clazz, local(0));
		}
		begin_func_sync();
		call_soft(finishJNIcall);
		end_func_sync();
		returnarg_int(tmp);
		slot_freetmp(tmp);
		break;
	case 'F':
		slot_alloctmp(tmp);
		return_float(tmp);
		/* Remove synchronisation if necessary */
		if (meth->accflags & ACC_SYNCHRONISED) {
			mon_exit(meth, clazz, local(0));
		}
		begin_func_sync();
		call_soft(finishJNIcall);
		end_func_sync();
		returnarg_float(tmp);
		slot_freetmp(tmp);
		break;
	case 'J':
		slot_alloc2tmp(tmp);
		return_long(tmp);
		/* Remove synchronisation if necessary */
		if (meth->accflags & ACC_SYNCHRONISED) {
			mon_exit(meth, clazz, local(0));
		}
		begin_func_sync();
		call_soft(finishJNIcall);
		end_func_sync();
		returnarg_long(tmp);
		slot_free2tmp(tmp);
		break;
	case 'D':
		slot_alloc2tmp(tmp);
		return_double(tmp);
		/* Remove synchronisation if necessary */
		if (meth->accflags & ACC_SYNCHRONISED) {
			mon_exit(meth, clazz, local(0));
		}
		begin_func_sync();
		call_soft(finishJNIcall);
		end_func_sync();
		returnarg_double(tmp);
		slot_free2tmp(tmp);
		break;
	case 'V':
		/* Remove synchronisation if necessary */
		if (meth->accflags & ACC_SYNCHRONISED) {
			mon_exit(meth, clazz, local(0));
		}
		call_soft(finishJNIcall);
		break;
	case 'L':
	case '[':
		slot_alloctmp(tmp);
		return_ref(tmp);
		/* Remove synchronisation if necessary */
		if (meth->accflags & ACC_SYNCHRONISED) {
			mon_exit(meth, clazz, local(0));
		}
		begin_func_sync();
		call_soft(finishJNIcall);
		end_func_sync();
		returnarg_ref(tmp);
		slot_freetmp(tmp);
		break;
	default:
		ABORT();
	}

	epilogue(0);
	ret();

	/* Generate the code */
	if (tmpslot > maxTemp) {
		maxTemp = tmpslot;
	}
	if (finishInsnSequence(&ncode) == 0) {
		goto error;
	}

	/* Install it */
	end = (void*)((uintp)ncode.code + ncode.codelen);
	SET_METHOD_NATIVECODE(meth, ncode.code, end);
        MSR_ADD(ncode_mem, ncode.codelen);

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		kprintf("<JIT: Trampoline %s.%s%s @ %p (%p)>\n", CLASS_CNAME(clazz), meth->name, meth->signature, METHOD_NATIVECODE(meth), meth);
	}

	/* Flush code out of cache */
#if defined(FLUSH_DCACHE)
	FLUSH_DCACHE(ncode.code, end);
#endif

	unlockStaticMutex(&translatorlock);
	return (1);
error:;
	unlockStaticMutex(&translatorlock);
	return (0);
}

static
void
nullCall(void)
{
}

/*
 * We point any abstract methods here in case they ever get executed.
 */
static
void
throwAbstractMethodError(void)
{
	postAndThrowException("java.lang.AbstractMethodError", "");
}

typedef struct _fakeCall {
	struct _fakeCall*	next;
	label*			from;
	label*			to;
	void*			func;
} fakeCall;

static fakeCall* firstFake;
static fakeCall* lastFake;
static fakeCall* currFake;

/*
 * Build a fake call to a function.
 *  A fake call has a return address which is different from where it's
 *  actually called from.
 */
label*
newFakeCall(void* func, uintp currpc)
{
	label* from;
	label* to;

	/* This is used to mark where I'm calling from */
	from = reference_code_label(currpc);

	/* This is where I'm calling to */
	to = newLabel();
	if (to == 0) {
		return (0);
	}
	to->type = Linternal;
	to->at = 0;
	to->to = 0;
	to->from = 0;

	if (currFake == 0) {
		currFake = gc_malloc_fixed_typed(sizeof(fakeCall), GC_ALLOC_ENGINE);
		if (currFake == 0) {
			longjmp(tjbuf, 1);
			return (0);
		}
		if (lastFake == 0) {
			firstFake = currFake;
		}
		else {
			lastFake->next = currFake;
		}
		lastFake = currFake;
	}

	currFake->from = from;
	currFake->to = to;
	currFake->func = func;

	currFake = currFake->next;

	return (to);
}

static
void
makeFakeCalls(void)
{
	fakeCall* c;

	for (c = firstFake; c != currFake; c = c->next) {
		softcall_fakecall(c->from, c->to, c->func);
	}

#if !defined(OPTIMIZE_MEMORY)
	currFake = firstFake;
#else
	while (firstFake != 0) {
		c = firstFake->next;
		gc_free_fixed(firstFake);
		firstFake = c;
	}
	lastFake = 0;
#endif
}

static
void
inline_constants(void)
{
#if defined(USE_JIT_INLINE_CONSTANT_POOL)
	generateInsnSequence();
	CODEPC = inlineConstants(CODEPC);
	ipc = pc;
#endif
}

/*
 * Count and return the number of arguments in a method
 * Includes `this' ptr if method is non-static.
 * *isStatic will be set by this method.
 */
static
int
countArgs(Method *meth, int *isStatic)
{
        int i;
	int maxArgs = 0;
	char* sig;

	sig = meth->signature;
        for (i = 1; sig[i] != ')'; i++) {
                switch (sig[i]) {
                case 'D':
                case 'J':
                        maxArgs += 2;
                        break;
		case '[':
                        maxArgs += 1;
			while (sig[++i] == '[')
				;
			if (sig[i] == 'L') {
				while (sig[++i] != ';')
					;
			}
			break;
		case 'L':
			while (sig[++i] != ';')
				;
                        maxArgs += 1;
			break;
                default:
                        maxArgs += 1;
                }
        }

	if (meth->accflags & ACC_STATIC) {
		*isStatic = 1;
	}
	else {
		*isStatic = 0;
		maxArgs += 1;
	}
	return (maxArgs);
}
