/* 
 * cfg.c
 * Manage the code labels and links.
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "code.h"
#include "access.h"
#include "support.h"
#include "env.h"
#include "bitvector.h"
#include "bytecode.h"

#ifndef OPTIMIZE_MEMORY
static instr_block * block_cache = 0;
#endif


static void
ARC_reassign_succ(instr_block* old, instr_block* new, int type)
{
        block_list *old_blks, *new_blks, *tgt_blks;
        int n, m;

        switch (type) {
        case ARC_ret_from_jsr:
                old_blks = &old->ret;
                new_blks = &new->succ;
                break;
        default:
                old_blks = &old->succ;
                new_blks = &new->succ;
                break;
        }
        bcopy(old_blks, new_blks, sizeof(block_list));
        bzero(old_blks, sizeof(block_list));

        for (n = new_blks->length; --n >= 0;) {
                switch (type) {
                case ARC_jsr_finally:
                default:
                        tgt_blks = &(new_blks->blocks[n]->pred);
                        break;
                }
                for (m = tgt_blks->length; --m >= 0;) {
                        if (tgt_blks->blocks[m] == old) {
                                tgt_blks->blocks[m] = new;
                                break;
                        }
                }
        }
}

void
ARC_add(instr_block* from, instr_block* to, int type)
{
        block_list *succ, *pred;

        switch (type) {
        case ARC_ret_from_jsr:
                succ = &from->ret;
                pred = &to->pred;
                break;
        case ARC_jsr_finally:
                succ = &from->succ;
                pred = &to->pred;
                if( !BL_contains( &to->cfg->finally, to ) ){
                        BL_add( &to->cfg->finally, to );
                }
                break;
        case ARC_try_catch:
                succ = &from->catch;
                pred = &to->try;
                break;
        case ARC_ebb:
                succ = &from->ebb_succ;
                pred = &to->ebb_pred;
                break;
        default:
                succ = &from->succ;
                pred = &to->pred;
                if (from->block.next == to) {
                        from->flags |= BLK_has_next;
                }
                break;
        }

        BL_add( succ, to );
        BL_add( pred, from );
}

void
ARC_del( instr_block* from, instr_block* to, int type ) 
{
        block_list *succ, *pred;

        switch (type) {
        case ARC_normal:
                succ = &from->succ;
                pred = &to->pred;
                break;
        case ARC_try_catch:
                succ = &from->catch;
                pred = &to->try;
                break;
        default:
                succ = pred = NULL;
                FAILED();
                break;
        }
        BL_del( pred, from );
        BL_del( succ, to );
}

instr_block *
BLK_new( cfg_info * cfg, Method * method, Hjava_lang_Class * class )
{
        int localsz, stacksz;
        instr_block * blk;
        method_env * m_env;
        boolean new_env;

        if( method == NULL ){
                new_env = false;
                method = cfg->env.method.method;
        }
        else{
                new_env = true;
        }
        stacksz = method->bcode->stacksz;
        localsz = ADJUST_LOCALSZ( method->bcode->localsz );

#ifdef OPTIMIZE_MEMORY
        blk = gc_malloc_fixed( sizeof(instr_block) );
#else
        if( block_cache != NULL ){
                blk = block_cache;
                block_cache = blk->block.next;
                bzero( blk, sizeof( instr_block) );
        }
        else {
                blk = gc_malloc_fixed( sizeof(instr_block) );
        }
#endif
        blk->bid = cfg->n_blocks++;
        blk->locals.in = gc_malloc_fixed( 2*(localsz+stacksz) * sizeof(expr*) );
        blk->stack.in = &blk->locals.in[localsz];
        blk->locals.out = &blk->locals.in[localsz+stacksz];
        blk->stack.out = &blk->locals.in[2*localsz+stacksz];
        blk->exprs_by_op = gc_malloc_fixed(MAX_OPS*sizeof(expr*));
        blk->stats.locals = STATS_new( localsz );
        blk->cfg = cfg;
        LAB_init( cfg );
        if( new_env ){
                m_env = gc_malloc_fixed( sizeof(method_env) );
                m_env->source = m_env;
                m_env->root = blk;
                m_env->method = method;
                m_env->class = class;
                m_env->next = cfg->env.global.methods;
                m_env->inline_depth = cfg->env.method.inline_depth;
                m_env->n_locals = localsz;
                m_env->n_slots = localsz + stacksz;
                m_env->cframe.map = gc_malloc_fixed( m_env->n_slots * sizeof(slot_info) );
                cfg->env.global.methods = m_env;
                blk->env = m_env;
        }
        else{
                blk->env = cfg->env.method.source;
        }
        return blk;
}

void 
BLK_free( instr_block * blk )
{
        int n;
        join_info * join_i;

        BLK_free_exprs( blk->exprs_head, blk->exprs_tail );
        if( blk == blk->ebb_root ){
                gc_free_fixed( blk->locals.in[0] );
#ifdef OPTIMIZE_GCSE
                gc_free_fixed( blk->globals.in[0] );
                n = GLOBAL_INIT_SIZE;
                for( ; n < blk->cfg->env.global.globals.max; n <<= 1 ){
                        gc_free_fixed( blk->globals.in[n] );                
                }
#endif
        }
        BV_free( blk->dom ); /* can free earlier */
        gc_free_fixed( blk->succ.blocks );
        gc_free_fixed( blk->pred.blocks );
        gc_free_fixed( blk->try.blocks );
        gc_free_fixed( blk->catch.blocks );
        gc_free_fixed( blk->ret.blocks );
        gc_free_fixed( blk->ebb.blocks );
        gc_free_fixed( blk->ebb_succ.blocks );
        gc_free_fixed( blk->ebb_pred.blocks );
        gc_free_fixed( blk->locals.in );
#ifdef OPTIMIZE_GCSE
        gc_free_fixed( blk->globals.in );
#endif
        gc_free_fixed( blk->exprs_by_op );
        join_i = blk->succ_join;
        if( join_i != NULL ){
                for( n = join_i->out.length; --n >= 0; ){
                        join_i->out.blocks[n]->succ_join = NULL;
                }
                gc_free_fixed( join_i->out.blocks );
                gc_free_fixed( join_i->in.blocks );
                gc_free_fixed( join_i->slots );
                gc_free_fixed( join_i->regs );
                gc_free_fixed( join_i );
        }
        STATS_field_free( &blk->stats.fields );
        STATS_free( blk->stats.locals );
        STATS_free( blk->handler_locals );
        /* FIXME: need to free properly! Watch out for early freeing of blocks */
        if( blk->flags & BLK_loop_entry ){
                STATS_field_free( &blk->loop->fields );
                gc_free_fixed( blk->loop->blocks.blocks );
                gc_free_fixed( blk->loop->exits.blocks );
                gc_free_fixed( blk->loop->locals );
                gc_free_fixed( blk->loop );
        }
#ifdef OPTIMIZE_MEMORY
        gc_free_fixed( blk );
#else
        blk->block.next = block_cache;
        block_cache = blk;
#endif
}

static instr_block * 
CFG_split_block( cfg_info * cfg, instr_block * blk, int pc, int npc )
{
        instr_block * s1_b, * s2_b;

LOG(    fprintf( cfg->fd, "CFG_split():\t%d - %d: ", blk->pc.in, blk->pc.out ); );
        s1_b = blk;
        s2_b = BLK_new( cfg, NULL, NULL );
        s2_b->pc.in = npc;
        s2_b->pc.out = s1_b->pc.out;
        s1_b->pc.out = pc;
        DLL_INSERT_AS_NEXT( s1_b, s2_b, instr_block *, block );
        ARC_reassign_succ( s1_b, s2_b, ARC_normal );
        if( s1_b->flags & BLK_has_next ){
                s2_b->flags |= BLK_has_next;
                s1_b->flags &= ~BLK_has_next;
        }
LOG(    fprintf( cfg->fd, "%d - %d, %d - %d\n", s1_b->pc.in, s1_b->pc.out, s2_b->pc.in, s2_b->pc.out ); );
        return s2_b;
}

#ifdef OPTIMIZE_INLINE

inline void
CFG_set_method_env( cfg_info * cfg, method_env * s_env )
{
        if( s_env != cfg->env.method.source ){
                bcopy( &cfg->env.method, cfg->env.method.source, sizeof(method_env) );
                bcopy( s_env, &cfg->env.method, sizeof(method_env) );
LOG(            fprintf( cfg->fd, "ENV_method: D%d, %s.%s%s\n", cfg->env.method.inline_depth,
                        cfg->env.method.class->name, 
                        cfg->env.method.method->name, 
                        cfg->env.method.method->signature ); );
        }
}

static boolean
CFG_inline_check( cfg_info * cfg, instr_block * blk, Method * method, Hjava_lang_Class * class ){
        if( strstr( class->name, "Exception" ) != NULL 
           || strstr( class->name, "Error" ) != NULL 
           ){
                return false;
        }
        if( print_dbg && ++cfg->env.global.count > inline_limit ){
                return false;
        }
        return true;
}

void
CFG_inline( expr_frame * frame, Method * method, Hjava_lang_Class * class, instr_block * blk, int op )
{
        method_env * m_env;
        instr_block * new_b;
        cfg_info * cfg;
        boolean do_inline, do_specialize;

        cfg = blk->cfg;
        do_inline = false;
        do_specialize = false;
        if( enable_inline
           && !( method->accflags & ACC_INTERFACE )
           && !( method->accflags & ACC_NATIVE )
           && method != cfg->env.method.method
           && CFG_inline_check( cfg, blk, method, class )
           && method->exception_table == NULL
           && method->bcode != NULL
           && method->bcode->codelen <= 20 ){
                if( method->accflags & (ACC_PRIVATE | ACC_STATIC | ACC_FINAL | ACC_CONSTRUCTOR) 
                   || class->accflags & ACC_FINAL
                   || op == INVOKESPECIAL
                   || op == INVOKESTATIC ){
                        if( op == INVOKESPECIAL 
                           && strcmp( method->name, "<init>" ) && ( !(method->accflags & (ACC_PRIVATE | ACC_STATIC | ACC_FINAL | ACC_CONSTRUCTOR) ) || !(class->accflags & ACC_FINAL) ) ){
TRACE(                          kprintf( "INVALID INLINE: %s.%s\n", class->name, method->name ); );
                        }
                        else{
                                do_inline = true;
                        }
                }
                else if( method->accflags & (ACC_PUBLIC | ACC_PROTECTED) ){
                        // do_specialize = true;
                }
        }
	if( !do_specialize && !do_inline ){
                return;
	}
        new_b = BLK_new( cfg, method, class );
        m_env = new_b->env;
        m_env->inline_depth++;
        m_env->inlined = true;
        m_env->caller = blk;
        m_env->cframe.offset.slots = blk->env->cframe.offset.slots + blk->env->n_slots;
        if( m_env->cframe.offset.slots > cfg->env.global.cframe.max_slots ){
                cfg->env.global.cframe.max_slots = m_env->cframe.offset.slots;
        }
        CFG_set_method_env( cfg, m_env );
        new_b->pc.in = 0;
        new_b->pc.out = method->bcode->codelen - 1;
        COUNTER_accumulate( cfg, expr );
        SCAN_bytecodes( cfg );
        CFG_set_method_env( cfg, blk->env );
        COUNTER_accumulate( cfg, scan );
        if( !(m_env->flags & METH_has_branches) ){
                m_env->frame = frame;
                frame->env = m_env;
                cfg->env.global.n_inline++;
                if( do_inline ){
                        frame->flags |= CALL_inline;
                        /*
                        kprintf( "INLINE: B%d.%d MF=0x%x CF=0x%x OP=%d %s.%s%s %db\n", blk->bid, cfg->env.method.inline_depth, 
                                method->accflags, class->accflags, op, class->name, method->name, 
                                method->signature, method->bcode->codelen );
                                */
LOG(		    	fprintf( cfg->fd, "CFG_inline(): B%d.%d: %s.%s%s\n", blk->bid, 
                                cfg->env.method.inline_depth, class->name, 
                                method->name, method->signature ); );
                }
                if( do_specialize ){
                        frame->flags |= CALL_specialized_inline;
/*                        kprintf( "SPECIALIZE B%d.%d F0x%x %s.%s%s\n", blk->bid, cfg->env.method.inline_depth,
                                method->accflags, class->name, 
                                method->name, method->signature );
*/
LOG( 		        fprintf( cfg->fd, "CFG_specialize(): B%d.%d %x %s.%s%s\n", blk->bid, cfg->env.method.inline_depth,
                                method->accflags, class->name, 
                                method->name, method->signature ); );
                }
        }
        else{
                if( do_inline ){
                        kprintf( "INLINE: B%d.%d MF=0x%x CF=0x%x OP=%d %s.%s%s %db\n", blk->bid, cfg->env.method.inline_depth, 
                                method->accflags, class->accflags, op, class->name, method->name, 
                                method->signature, method->bcode->codelen );
                }
        }
}
#endif

cfg_info *
CFG_new( Method * method, Hjava_lang_Class * class, const char * ext )
{
        cfg_info * cfg;
        char fname[500];
        int i, offset, type;
        instr_block * root_b;
        char * p, * tp;

        cfg = gc_malloc_fixed( sizeof(cfg_info) );
LOG(    bzero( fname, 500 );
        strcat( fname, ext );
        strcat( fname, class->name );
        strcat( fname, "." );
        strcat( fname, method->name );
        strcat( fname, method->signature );
        i = 5;
        while( fname[i] != 0 ){
                if( fname[i] < 0x21 || fname[i] > 0x7d ){
                        fname[i] = '_';
                }
                if( fname[i] == '/' ){
                        fname[i] = '.';
                }
                i++;
        }
        if( strlen( fname ) > 200 ){
                strcpy( fname, "tmp" );
        }
        if( (cfg->fd = fopen( fname, "w+" )) == NULL ){
                kprintf( "(%d): %s\n", errno, strerror(errno) );
        } );

        /* Architecture dependent */
        offset = 0;
        if( !METHOD_IS_STATIC( method ) ){
                offset++;
        }
        for( p = method->signature; (tp = sigItemType( &p )) != NULL; ){
                type = *tp;
                offset += type == 'J' || type == 'D' ? 2 : 1;
        }
LOG(    fprintf( cfg->fd, "Num Args In: %d\n", cfg->env.global.n_args ); );
#ifdef OPTIMIZE_GCSE
//        cfg->env.global.globals.max = 4;
#endif
        cfg->env.global.bytecode_len = method->bcode->codelen;
        cfg->env.global.constants.max = 0;
        cfg->env.global.n_args = ADJUST_LOCALSZ( offset );
        cfg->env.global.id = 1;
        root_b = BLK_new( cfg, method, class );
        bcopy( root_b->env, &cfg->env.method, sizeof(method_env) );
        cfg->env.global.root_env = root_b->env;
        cfg->env.global.arg_map = gc_malloc_fixed( cfg->env.method.n_locals * sizeof(uint16) );
        cfg->env.global.cframe.max_slots = root_b->env->n_slots;
        cfg->env.global.spill_sideways = true;
        SCH_init( cfg );
        root_b->pc.in = 0;
        root_b->pc.out = method->bcode->codelen - 1;
LOG(    fprintf( cfg->fd, "Locals: %d Stack: %d\n", cfg->env.method.n_locals, 
                cfg->env.method.n_slots - cfg->env.method.n_locals ); );
        cfg->flags |= CFG_is_leaf;
        return cfg;
}

void
CFG_exception( cfg_info * cfg, jexceptionEntry * exception )
{
        instr_block * from_b, * to_b, * new_b;
        int start_pc, end_pc, handler_pc;
        
        start_pc = exception->start_pc;
        end_pc = exception->end_pc;
        handler_pc = exception->handler_pc;
        
LOG( 	fprintf( cfg->fd, "CFG_exception(): %4d to %4d --> %4d\n", start_pc, end_pc, handler_pc); );

        to_b = CFG_get_block( cfg, handler_pc, true );
        if( to_b->pc.in != handler_pc ){
                to_b = CFG_split_block( cfg, to_b, handler_pc-1, handler_pc );
        }
        to_b->flags |= BLK_catch;

        from_b = cfg->env.method.root;
        while (from_b) {
                if (start_pc <= from_b->pc.out && from_b->pc.in < end_pc) {
                        if( start_pc > from_b->pc.in && start_pc <= from_b->pc.out ){
                                new_b = CFG_split_block( cfg, from_b, start_pc-1, start_pc );
                                ARC_add( from_b, new_b, ARC_normal);
                                BL_add( &cfg->env.method.rescan.blocks, from_b );
                                BL_add( &cfg->env.method.rescan.blocks, new_b );
                                from_b = new_b;
                        }
                        if( end_pc-1 < from_b->pc.out && end_pc-1 >= from_b->pc.in ){
                                new_b = CFG_split_block( cfg, from_b, end_pc-1, end_pc );
                                ARC_add( from_b, new_b, ARC_normal);
                                BL_add( &cfg->env.method.rescan.blocks, from_b );
                                BL_add( &cfg->env.method.rescan.blocks, new_b );
                        }
                        ARC_add(from_b, to_b, ARC_try_catch);
                        to_b->handler_type = (char*)exception->catch;
                }
                from_b = from_b->block.next;
        }
        BL_add( &cfg->catch, to_b );
}

instr_block *
CFG_set(cfg_info * cfg, int pc, int npc, int jpc, int type) 
{
        instr_block * cur_b;
        instr_block * from1_b = 0;
        instr_block * from2_b = 0;
        instr_block * to1_b = 0;
        instr_block * to2_b = 0;
        instr_block * to_b;
        boolean split_from = false;
        boolean split_to = false;
        int n;

        if( jpc < pc ){
                cfg->env.method.has_loops = true;
        }
        if( cfg->env.method.inline_depth != 0 && ( type == CFG_jsr || type == CFG_ret ) ){
LOG(            fprintf( cfg->fd, "CFG_build(): nested jsr/ret\n" ); );
        }
LOG( 	fprintf( cfg->fd, "CFG_set():\t%d --> %d / %d (%s)\n", pc, npc, jpc, CFG_print_type( type ) ); );
        cur_b = cfg->env.method.root;
        while (cur_b) {
                if (pc >= cur_b->pc.in && pc <= cur_b->pc.out) {
                        from1_b = cur_b;
                }
                if (jpc >= cur_b->pc.in && jpc <= cur_b->pc.out) {
                        to1_b = cur_b;
                }
                cur_b = cur_b->block.next;
        }

        if (from1_b->pc.out == pc) {
        }
        else if (from1_b->pc.out+1 == npc) {
                from1_b->pc.out = pc;
        }
        else {
                from2_b = CFG_split_block( cfg, from1_b, pc, npc );
                split_from = true;
                if (from1_b == to1_b && jpc > pc) {
                        to1_b = from2_b;
                }
        }

        if (jpc == 0 && ( type == CFG_branch || type == CFG_jump || type == CFG_switch ) ){
                if( cfg->env.method.root == to1_b && to1_b->pc.out != 0 ){
                        to2_b = CFG_split_block( cfg, to1_b, 0, 0 );
                        to2_b->stage = to1_b->stage;
                        to2_b->sp.out = to1_b->sp.out;
#ifdef OPTIMIZE_LOOPS
                        to2_b->stats.locals = to1_b->stats.locals;
                        to1_b->stats.locals = NULL;
                        LIST_copy( &to2_b->stats.fields, &to1_b->stats.fields );
                        LIST_zero( &to1_b->stats.fields );
#endif
                        to1_b->sp.out = 0;
                        to1_b->flags |= BLK_prologue;
                        ARC_add(to1_b, to2_b, ARC_normal);
                        if (from1_b == to1_b && jpc < pc) {
                                from1_b = to2_b;
                        }
                        to1_b = to1_b->block.next;
                }
        }
        else if (to1_b->pc.in != jpc) {
                to2_b = CFG_split_block( cfg, to1_b, jpc-1, jpc );
                to2_b->stage = to1_b->stage;
                split_to = true;
                if( jpc < pc ){
                        BL_add( &cfg->env.method.rescan.blocks, to1_b );
                        BL_add( &cfg->env.method.rescan.blocks, to2_b );
                }
                if (from1_b == to1_b && jpc < pc) {
                        from1_b = to2_b;
                }
        }
        to_b = split_to ? to2_b : to1_b;

        switch( type ){
        case CFG_ret:
                if( !(from1_b->flags & BLK_ret) ){
                        for( n = from1_b->succ.length; --n >= 0; ){
                                ARC_del( from1_b, from1_b->succ.blocks[n], ARC_normal );
                        }
                }
                from1_b->flags |= BLK_ret;
                break;
        case CFG_return:
                for( n = from1_b->succ.length; --n >= 0; ){
                        ARC_del( from1_b, from1_b->succ.blocks[n], ARC_normal );
                }
                bzero( &(from1_b->succ), sizeof(block_list) );
                break;
        case CFG_jsr:
                to_b->flags |= BLK_finally;
                ARC_add( from1_b, to_b, ARC_jsr_finally );
                if( split_from ){
                        ARC_add( to_b, from2_b, ARC_ret_from_jsr );
                }
                else {
                        ARC_add( to_b, from1_b->block.next, ARC_ret_from_jsr );
                }
                break;
        case CFG_switch:
                if( split_to ){
                        ARC_add( to1_b, to2_b, ARC_normal );
                }
                if( !BL_contains( &from1_b->succ, to_b ) ){
                        ARC_add( from1_b, to_b, ARC_normal );
                }
                break;
        case CFG_branch:
                if( split_to ){
                        ARC_add( to1_b, to2_b, ARC_normal );
                }
                ARC_add( from1_b, to_b, ARC_normal );
                
                if( split_from ){
                        ARC_add( from1_b, from2_b, ARC_normal );
                }
                if( !split_from && !(from1_b->flags & BLK_has_next) ){
                        ARC_add( from1_b, from1_b->block.next, ARC_normal );
                }
                break;
        case CFG_jump:
                if( split_to ){
                        ARC_add( to1_b, to2_b, ARC_normal );
                }
                ARC_add( from1_b, to_b, ARC_normal );

                if( from1_b->flags & BLK_has_next
                    && from1_b->block.next != to_b ){
                        ARC_del(from1_b, from1_b->block.next, ARC_normal );
                        from1_b->flags &= ~BLK_has_next;
                }
                break;
        }
        return from1_b;
}

static void
CFG_get_ret( block_list * ret_bl, instr_block * cur_b )
{
        int n;

        if( cur_b->flags & BLK_ret ){
                if( !BL_contains( ret_bl, cur_b ) ){
                        BL_add( ret_bl, cur_b );
                }
                return;
        } 
        for( n = cur_b->succ.length; --n >= 0; ){
                CFG_get_ret( ret_bl, cur_b->succ.blocks[n] );
        }
}

void
CFG_fixup_ret( cfg_info * cfg )
{
        instr_block * finally_b, * ret_b;
        int i,j;
        block_list ret_bl;

        for( i = cfg->finally.length; --i >= 0; ){
                bzero( &ret_bl, sizeof(block_list) );
                finally_b = cfg->finally.blocks[i];
                CFG_get_ret( &ret_bl, finally_b );
                for( j = ret_bl.length; --j >= 0; ){
                        ret_b = ret_bl.blocks[j];
                        ARC_reassign_succ( finally_b, ret_b, ARC_ret_from_jsr );
                }
                gc_free_fixed( ret_bl.blocks );
        }
}

/*
* TODO: probably don't need return_first_match or next_b match
*/
instr_block *
CFG_get_block( cfg_info * cfg, int pc, boolean return_first_match )
{
        instr_block * cur_b, * next_b;

        cur_b = cfg->env.method.root;
        while (cur_b) {
                next_b = cur_b->block.next;
                if( pc >= cur_b->pc.in && ( pc <= cur_b->pc.out || pc < next_b->pc.in ) ){
                        if( pc == 0 && !return_first_match 
                           && cur_b->flags & BLK_prologue ){
                        }
                        else{
                                return cur_b;
                        }
                }
                cur_b = next_b;
        }
LOG(    fprintf( cfg->fd, "Couldn't find block for pc=%d\n",pc); );
        FAILED();
        return NULL;
}

void
CFG_free( cfg_info * cfg )
{
        instr_block *cur_b, *next;
        method_env * m_env, * next_env;

        m_env = cfg->env.global.methods;
        while( m_env != NULL ){
                cur_b = m_env->root;
                while( cur_b ){
                        next = cur_b->block.next;
                        BLK_free( cur_b );
                        cur_b = next;
                }
                gc_free_fixed( m_env->cframe.map );
                next_env = m_env->next;
                gc_free_fixed( m_env );
                m_env = next_env;
        }
        REG_done( cfg );
        LAB_done( cfg );
        CONST_done( cfg );
        TRW_done( cfg );
        gc_free_fixed( cfg->env.method.rescan.exprs.exprs );
        gc_free_fixed( cfg->env.method.rescan.blocks.blocks );
        gc_free_fixed( cfg->env.global.schedule.roots.instrs );
        gc_free_fixed( cfg->env.global.schedule.regs );
        gc_free_fixed( cfg->env.global.schedule.base );
        gc_free_fixed( cfg->env.global.schedule.delay.instrs );
        gc_free_fixed( cfg->env.global.schedule.buffer.start );
        gc_free_fixed( cfg->env.global.arg_map );
        gc_free_fixed( cfg->env.global.globals.map );
LOG(    fflush( cfg->fd );;
        fclose( cfg->fd ); )
        gc_free_fixed( cfg );
}

static void
_readjust_join_slots( join_info * join_i, int n_slots )
{
        slot * t_s;

        t_s = gc_malloc_fixed( ( n_slots + join_i->n_globals ) * sizeof(slot) );
        bcopy( join_i->slots, t_s, join_i->n_slots);
        gc_free_fixed( join_i->slots );
        join_i->slots = t_s;
        join_i->n_slots = n_slots;
}

void
CFG_build_join( instr_block * start_b )
{
        int i, j, n_slots, n_locals;
        boolean change = true;
        instr_block * cur_b, * succ_b, * pred_b;
        FILE * fd;
        join_info * join_i;

        fd = start_b->cfg->fd;
LOG(    fprintf( fd, "CFG_build_join() @ B%d: ", start_b->bid ); );

        join_i = gc_malloc_fixed( sizeof(join_info) );
        join_i->n_globals = start_b->cfg->env.global.globals.length;
        n_locals = start_b->cfg->env.method.n_locals;
        n_slots = start_b->sp.out + n_locals;
        join_i->slots = gc_malloc_fixed( ( n_slots + join_i->n_globals ) * sizeof(slot) );
        join_i->regs = gc_malloc_fixed( NUM_REGS * sizeof(reg_join_info) );

        BL_add( &join_i->out, start_b );
LOG(    fprintf( fd, "pB%d ", join_i->out.blocks[0]->bid ); );
        while( change ){
                change = false;
                for( i = join_i->out.length; --i >= 0; ){
                        cur_b = join_i->out.blocks[i];
                        for( j = cur_b->succ.length; --j >= 0; ){
                                succ_b = cur_b->succ.blocks[j];
                                if( !BL_contains( &join_i->in, succ_b ) ){
                                        BL_add( &join_i->in, succ_b );
                                        if( succ_b->sp.in + n_locals > n_slots ){
                                                n_slots = succ_b->sp.in + n_locals;
                                                _readjust_join_slots( join_i, n_slots );
                                        }
LOG(                                    fprintf( fd, "sB%d ", succ_b->bid ); );
                                        change = true;
                                }
                        }
                }
                for( i = join_i->in.length; --i >= 0; ){
                        cur_b = join_i->in.blocks[i];
                        for( j = cur_b->pred.length; --j >= 0; ){
                                pred_b = cur_b->pred.blocks[j];
                                if( !BL_contains( &join_i->out, pred_b ) ){
                                        BL_add( &join_i->out, pred_b );
                                        if( pred_b->sp.out + n_locals > n_slots ){
                                                n_slots = pred_b->sp.out + n_locals;
                                                _readjust_join_slots( join_i, n_slots );
                                        }
LOG(                                    fprintf( fd, "pB%d ", pred_b->bid ); );
                                        change = true;
                                }
                        }
                }
        }
        join_i->n_slots = n_slots;
        for( i = join_i->out.length; --i >= 0; ){
                join_i->out.blocks[i]->succ_join = join_i;
        }
        for( i = join_i->in.length; --i >= 0; ){
                join_i->in.blocks[i]->pred_join = join_i;
        }
LOG(    fprintf( fd, "\n" ); );
}

static void
EBB_add( instr_block * root, instr_block * block )
{
        boolean changed;

        BL_add( &root->ebb, block );
        block->ebb_root = root;
        if( block->succ.length == 0 ){
                return;
        }
        do {
                int n;
                instr_block **blocks;

                changed = false;
                blocks = block->succ.blocks;
                for( n = block->succ.length; --n >= 0; ){
                        instr_block * cur = blocks[n];

                        if( cur->pred.length == 1 ){
                                if( !BL_contains( &(root->ebb), cur ) ){
                                        EBB_add( root, cur );
                                        changed = true;
                                }
                        }
                        else{
                                block->flags |= BLK_succ_not_in_ebb;
                                if( !(cur->flags & BLK_is_ebb) ){
                                        ARC_add( root, cur, ARC_ebb );
                                        cur->flags |= BLK_is_ebb;
                                        changed = true;
                                }
                        }
                }
        } while ( changed );
}

void
EBB_build( instr_block * block )
{
        instr_block ** blocks;
        int n;
        
        block->flags |= BLK_is_ebb;
        EBB_add( block, block );

        blocks = block->ebb_succ.blocks;
        for( n = block->ebb_succ.length; --n >= 0; ){
                EBB_build( blocks[n] );
        }
}

void
CFG_order_block_ids( cfg_info * cfg )
{
        int id, n;
        instr_block * blk, * succ_b, * next_b;

        blk = cfg->env.method.root;
        id = 0;
        while( blk != NULL ){
                next_b = blk->block.next;
                if( blk->pred.length == 0 && blk->try.length == 0 
                   && blk != cfg->env.method.root ){
                        for( n = blk->succ.length; --n >= 0; ){
                                succ_b = blk->succ.blocks[n];
                                ARC_del( blk, succ_b, ARC_normal );
                        }
                        for( n = blk->catch.length; --n >= 0; ){
                                succ_b = blk->catch.blocks[n];
                                ARC_del( blk, succ_b, ARC_try_catch );
                        }
                        DLL_REMOVE( blk, instr_block *, block );
                        BLK_free( blk );
                }
                else{
                        blk->bid = id++;
                }
                blk = next_b;
        }

}

/*
 * Compute dominators & immediate dominators using bitvectors
 * From Muchnick, Compiler Design & Implemenation, sec. 7.3
 */
void
CFG_dominators( cfg_info * cfg )
{
        uint ** dom, ** idom, * t, * d;
        instr_block * blk;
        block_list * bl;
#ifdef COMPUTE_IDOM
        int sn, tn;
#endif
        int n, bn, n_words, n_blocks;
        boolean change;
        FILE * fd;

        fd = cfg->fd;
        n_blocks = cfg->n_blocks;
        if( n_blocks <= 1 ) return;
        n_words = BV_words( n_blocks );

        dom = BV_new_array( n_blocks, n_words );
        idom = BV_new_array( n_blocks, n_words );
        t = BV_new( n_words );

        /* Compute dominators */
        for( bn = n_blocks; --bn >= 1; ){
                BV_set_all( dom[bn], n_blocks );
        }
        BV_set( dom[0], 0 );
        change = true;
        while( change ){
                change = false;
                blk = cfg->env.method.root->block.next;
                bn = 1;
                while( blk ){
                        BV_set_all( t, n_blocks );
                        bl = blk->flags & BLK_catch ? &blk->try : &blk->pred;
                        for( n = bl->length; --n >= 0; ){
                                BV_and( t, t, dom[bl->blocks[n]->bid], n_words );
                        }
                        d = t;
                        BV_set( d, bn );
                        if( !BV_equal( d, dom[bn], n_words ) ){
                                change = true;
                                BV_copy( dom[bn], d, n_words );
                        }
                        blk = blk->block.next;
                        bn++;
                }
        }
#ifdef COMPUTE_IDOM
        /* Compute immediate dominators from dominators */
        for( n = n_blocks; --n >= 0; ){
                BV_copy( idom[n], dom[n], n_words );
                BV_unset( idom[n], n );
        }
        for( n = n_blocks; --n >= 1; ){
                for( sn = n_blocks; --sn >= 0; ){
                        if( BV_is_set( idom[n], sn ) ){
                                BV_copy( t, idom[n], n_words );
                                BV_unset( t, sn );
                                for( tn = n_blocks; --tn >= 0; ){
                                        if( BV_is_set( t, tn ) && BV_is_set( idom[sn], tn )){
                                                BV_unset( idom[n], tn );
                                        }
                                }
                        }
                }
        }
#endif
        blk = cfg->env.method.root;
        for( n = 0; blk != NULL; n++, blk = blk->block.next ){
                blk->dom = dom[blk->bid];
                if( blk->pred.length == 0 && blk->try.length == 0 ){
                        BV_unset_all( blk->dom, n_blocks );
                }
        }
        BV_free_array( idom, n_blocks );
        BV_free( t );
}
