/*
 * cfg.h
 * Structures for analyzing control- and data-flow
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __cfg_h
#define __cfg_h

#define CFG_branch		0x1
#define CFG_jump		0x2
#define CFG_switch		0x3
#define CFG_return		0x4
#define CFG_jsr			0x5
#define CFG_ret			0x6

#define CFG_is_leaf		0x1
#define METH_has_branches	0x1
#define METH_built_positions 	0x2

#define ARC_normal		0x1
#define ARC_jsr_finally		0x2
#define ARC_try_catch		0x3
#define ARC_ebb			0x4
#define ARC_ret_from_jsr	0x5

#define STG_scanned		0x1
#define STG_scanned_loops	0x2
#define STG_generated_exprs	0x3
#define STG_linked_uses		0x4
#define STG_linked_loops	0x5
#define STG_generated_code	0x6

#define BLK_reserved		0x001 /* See LAB_is_label */
#define BLK_has_next		0x002
#define BLK_is_ebb		0x004

#define BLK_linked_in_slots	0x200
#define BLK_linked_out_slots	0x400
#define BLK_switched_branch	0x800

#define BLK_catch		0x1000
#define BLK_finally		0x2000
#define BLK_ret			0x4000

#define BLK_prologue		0x10000 /* Necessary since we can't distinguish from 1 bytecode method */
#define BLK_loop_header		0x40000
#define BLK_loop_entry		0x4000000
#define BLK_loop_do_while	0x8000

#define BLK_buffer_code		0x8000000

#define BLK_succ_not_in_ebb	0x200000

#define BLK_inline_epilogue	0x400000

#define BLK_filled_delay_slot	0x800000

#define BLK_has_java_call	0x1000000
#define BLK_inserted_header	0x2000000
#define LOOP_has_java_call	0x1
#define LOOP_do_while		0x2
#define LOOP_infinite		0x4

typedef struct _native_info {
	uint8 *			start;
	uint8 *			end;
} native_info;

typedef struct _exception_info {
	struct _exception_info *	next;
        struct _instr_block *		handler;
	native_info		native;
        void *			catch;
} exception_info;

typedef struct _slot_info {
        uint8			size;
        int16			position;
} slot_info;

typedef	struct _global_env {
        int			id;
        struct {
                struct _label_info **	table;
                struct _label_info *	list;
        } labels;
        struct {
	        struct _const_info *	list;
        	int		max;
        	int16		uses;
        } constants;
        struct {
                struct _throw_info *	list;
        } throw;
        struct {
                native_info	consttable;
                native_info	code;
                uint8 *		max;
                int		estimate;
                int		unused;
        } native;
        uint8 *			pbase;
	struct _instr_block *	last_block;
	struct _instr_block *	header_block;
        uint			add_flags;
	uint			max_call_args;
	uint			n_args;
	uint			n_inline;
        uint16 *		arg_map;
        struct _expr *		last_call;
        boolean			in_nested_block;
        boolean			no_free;
        boolean			no_reduce_base;
        boolean			spill_sideways;
	int16			max_temps;
        uint16			max_loop_depth;
        struct {
                int16		max_slots;
                int16		max_words;
	} cframe;
	int			bytecode_len;
	struct _method_env *	root_env;
	struct _method_env *	methods;
        struct {
                int16		length;
                int16		max;
                int16		words;
		slot_info *	map;
	} globals;
        struct {
		struct _exception_info *	handlers;
		int 		count;
	} exception;
        struct {
                struct _instr_info ** regs;
                instr_list 	roots;
                instr_list 	delay;
                struct _instr_info * last_branch;
                struct _label_info * call_label;
                struct _expr * branch_expr;
                native_info	code;
                native_info	buffer;
                struct _instr_info * base;
	} schedule;
	struct {
		long long	lookup;
		long long	scan;
		long long	expr;
		long long	code;
	} time;
	int 			count;
        char                    print_buf[20];
} global_env;

typedef struct _reg_context {
        struct _reg_info *	free;
        struct _reg_info *	live;
        struct _reg_info *	all;
} reg_context;

typedef struct _temp_context {
        int16			free;
        struct _temp_info *	table;
        uint16			max;
        uint16			limit;
} temp_context;

typedef struct _block_env {
	struct _block_env *	stack;
	struct _instr_block *	block;
        struct {
	        struct _expr *	backup;
        } exprs;
        reg_context		regs;
        temp_context		temps;
} block_env;

typedef struct _method_env {
	struct _method_env *	source;
	struct _method_env *	next;
	struct _expr *		expr;
	Method *		method;
	Hjava_lang_Class *	class;
	struct _instr_block *	caller;
	struct _instr_block *	root;
	struct _expr *		return_val;
        struct _expr_frame * 	frame;
	int			first_id;
	int16			flags;
	int16			n_slots;
	int16			n_locals;
	int16			inline_depth;
        struct {
	        struct {
	                int16	slots;
	                int16	words;
		} offset;
		slot_info *	map;
	} cframe;
        boolean			inlined;
        boolean			has_loops;
        struct {
                block_list	blocks;
                expr_list	exprs;
        } rescan;
} method_env;

typedef struct _cfg_info {
        uint			flags;
	uint			n_blocks;
        struct{
               global_env  	global;
               method_env	method;
               block_env	block;
        } env;
	block_list		catch;
	block_list		finally;
        FILE *			fd;
} cfg_info;

typedef struct _frame_info {
	struct _expr **		in;
	struct _expr **		out;
} frame_info;

typedef struct _reg_join_info {
	boolean		used;
	struct _expr *	expr;
} reg_join_info;

typedef struct _join_info {
	block_list		out; /* unnecessary */
	block_list		in;
        struct _slot *		slots;
	reg_join_info *		regs;
        int16			n_slots;
        int16			n_globals;
} join_info;

typedef struct _call_info {
        struct _expr *		last;
        struct _expr *		first;
} call_info;

typedef struct _stats_info {
        uint			words;
        uint *			item[6]; /* should use NUM_STATS */
        uint			bv[1];
} stats_info;

typedef struct _field_stats {
        Hjava_lang_Class *	class;
        stats_info *		stats;
} field_stats;

typedef struct _loop_info {
        uint16			flags;
        struct _instr_block *	header;
        struct _instr_block *	cond;
        struct _instr_block *	in;
	block_list		blocks;
	block_list		exits;
	field_list		fields;
        stats_info *		locals;
} loop_info;

typedef struct _instr_block {
	uint16			bid;
	uint8			stage;
	uint32			flags;
	struct {
	        uint16		in;
        	uint16		out;
	} pc;
	struct {
	        int16		in;
        	int16		out;
	} sp;
	frame_info		locals;
	frame_info		stack;
	frame_info		globals;
        linked_list		block;

        struct _cfg_info *	cfg;

	uint *			dom;

	block_list		pred;
	block_list		succ;
	struct _join_info *	pred_join; /* unnecessary */
	struct _join_info *	succ_join;
	struct _instr_block * 	ebb_root;
	block_list		ebb;	/* may be unnecessary except during initialization */
	block_list		ebb_pred; /* unnecessary */
	block_list		ebb_succ;
	block_list		try;
	block_list		catch;
	block_list		ret;
	native_info		native;

	struct _expr *		exprs_head;
	struct _expr *		exprs_tail;
	struct _expr **		exprs_by_op;

        call_info		call;

	expr *			set_cc;
	loop_info *		loop;
	struct {
                int		n_move;
                int		n_other;
                stats_info *	locals;
		field_list	fields;
	} stats;
        struct _method_env * 	env;
        stats_info *		handler_locals;
	exception_info *	handler;
	char *			handler_type;
        boolean			in_init;
} instr_block;

extern cfg_info * CFG_new( Method *, Hjava_lang_Class *, const char * );
extern instr_block * CFG_set( cfg_info *, int, int, int, int );
extern void CFG_exception( cfg_info *, jexceptionEntry * );
extern void CFG_print( cfg_info * );
extern void CFG_free( cfg_info * );
extern instr_block * CFG_get_block( cfg_info *, int, boolean );
extern void CFG_dominators( cfg_info * );
extern void CFG_build_join( instr_block * );
extern void CFG_fixup_ret( cfg_info * );
extern void CFG_inline( expr_frame *, Method *, Hjava_lang_Class *, instr_block *, int );
extern inline void CFG_set_method_env( cfg_info *, method_env * );
extern void CFG_order_block_ids( cfg_info * );
extern void ARC_add( instr_block *, instr_block *, int );
extern void ARC_del( instr_block *, instr_block *, int );
extern instr_block * BLK_new( cfg_info *, Method *, Hjava_lang_Class * );
extern void BLK_free( struct _instr_block * );

extern void EBB_build( instr_block * );
extern void DFG_build( cfg_info * );
extern void EXPR_update_use( expr *, instr_block * );
extern void EXPR_add( expr *, expr **, instr_block *, instr_block * );
#endif
