/*
 * expr.c
 * Instruction-level sequencing
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "env.h"
#include "lookup.h"
#include "code-analyse.h"
#include "locks.h"
#include "code.h"
#include "nop.h"
#include "bytecode.h"
#include "access.h"
#include "external.h"
#include "soft.h"
#include "objects.h"
#include "support.h"
#include "engine.h"
#include "reduce.h"
#include "itypes.h"

#ifdef OPTIMIZE_INLINE
static void EBB_generate_code( block_list *, cfg_info * );
static boolean BLK_link( instr_block *, int, boolean );
#endif
extern void BLK_generate_exprs( instr_block *, instr_block * );

void
EXPR_reduce_class( expr * e, Hjava_lang_Class * new )
{
        expr * base;

        if( new == NULL ){
                return;
        }
        base = EXPR_RESOLVE_ORIGINAL( e );
        if( base->class == new || base->op == OP_const_0 ){
                return;
        }
        if( base->class == NULL ){
                base->class = new;
        }
        else if( instanceof( base->class, new ) ){
                base->class = new;
        }
        else{
                if( !instanceof( new, base->class ) ){
                        kprintf( "UNEXPECTED: [%d] %s %s %s\n", e->eid, new->name, base->class->name );
                        FAILED();
                        return;
                }
        }
        e->class = base->class;
}

#define OP_cbranch_a		OP_2a
#define OP_cc_cbranch_a 	OP_1a

char OP_type[] = {
	0, /* nop */
	0, /* const_m1 */
	0, /* const_0 */
	0, /* const_1 */
	0, /* const */
	0, /* move */
	OP_1a, /* load */
	OP_2a, /* store */
	OP_2a, /* add */
	OP_2a, /* sub */
	OP_2a, /* mul */
	OP_2a, /* div */
	OP_2a, /* rem */
	OP_1a, /* neg */
	OP_2a, /* lsl */
	OP_2a, /* lsr */
	OP_2a, /* asr */
	OP_2a, /* and */
	OP_2a, /* or */
	OP_2a, /* xor */
	OP_1a, /* cvt */
	OP_2a, /* set_cc */
	cbranch_eq_func( OP_, _a ), /* cbranch_eq */
	cbranch_ne_func( OP_, _a ), /* cbranch_ne */
	cbranch_lt_func( OP_, _a ), /* cbranch_lt */
	cbranch_ge_func( OP_, _a ), /* cbranch_ge */
	cbranch_gt_func( OP_, _a ), /* cbranch_gt */
	cbranch_le_func( OP_, _a ), /* cbranch_le */
	cbranch_ult_func( OP_, _a ), /* cbranch_ult */
	0, /* branch */
	OP_1a, /* branch_indirect */
	OP_1a, /* returnarg */
	OP_2a, /* cmp */
	OP_frame, /* call */
	0, /* target */
	0, /* prologue */
	0, /* epilogue */
	0, /* eprologue */
        0, /* copy */
        0, /* inline_data */
        0, /* get_arg */
        OP_1a, /* preload */
        0, /* replace */
        0, /* specialize_sync */
        OP_2a, /* tableswitch */
        OP_2a, /* lookupswitch */
        OP_1a, /* native_instr_1a */
        OP_2a, /* native_instr_2a */
        OP_2a, /* cmov_eq */
        OP_2a, /* cmov_ne */
        OP_2a, /* cmov_lt */
        OP_2a, /* cmov_ge */
        OP_2a, /* cmov_gt */
        OP_2a /* cmov_le */
};

#ifndef OPTIMIZE_MEMORY
static expr * expr_cache = NULL;

expr *
EXPR_new() 
{
        expr * new_e;

        if( expr_cache != NULL ){
                new_e = expr_cache;
                expr_cache = new_e->expr.next;
                bzero( new_e, sizeof(expr) );
        }
        else {
                new_e = gc_malloc_fixed( sizeof(expr) );
        }
        return new_e;
}

void
EXPR_free( expr * e )
{
        e->expr.next = expr_cache;
        expr_cache = e;
}
#endif

void
BLK_free_exprs(expr* list_start, expr* list_end)
{
        if( list_start == NULL && list_end == NULL ){
                return;
        }
        assert( list_start != NULL && list_end != NULL );
        list_end->expr.next = expr_cache;
        expr_cache = list_start;
}

static void
EXPR_make_copy( expr * copy_e, expr * s_e, instr_block * blk, boolean set_id )
{
        copy_e->op = OP_copy;
        copy_e->a[0] = s_e;
        copy_e->a[1] = s_e->op == OP_copy || s_e->op == OP_move ? s_e->a[1] : s_e;
        copy_e->type = s_e->type;
        EXPR_assign_class( copy_e, s_e->class );
//        copy_e->flags = s_e->flags & (SE_global | SE_slot);
        copy_e->flags = s_e->flags & (SE_global | SE_null_checked);
        if( set_id ){
                copy_e->eid = ++blk->cfg->env.global.id;
        }
        copy_e->val.expr = blk->exprs_by_op[OP_copy];
        blk->exprs_by_op[OP_copy] = copy_e;
}

expr *
EXPR_new_copy( expr * s_e, instr_block * blk, boolean set_id )
{
        expr * copy_e;

        copy_e = EXPR_new();
        EXPR_make_copy( copy_e, s_e, blk, set_id );
        return copy_e;
}

static inline void
_EXPR_make_move( expr * move_e, expr * s_e, instr_block * blk )
{
        move_e->op = OP_move;
        move_e->a[0] = s_e;
        move_e->a[1] = s_e->op == OP_copy || s_e->op == OP_move ? s_e->a[1] : s_e;
        s_e->uses.block++;
        move_e->type = s_e->type;
        EXPR_assign_class( move_e, s_e->class );
        move_e->eid = ++blk->cfg->env.global.id;
}

void
EXPR_make_move( expr * move_e, expr * s_e, instr_block * blk )
{
	_EXPR_make_move( move_e, s_e, blk );
LOG(    fprintf( blk->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( blk, s_e ), s_e->uses.block ); );
}

expr *
EXPR_new_move( expr * s_e, instr_block * blk )
{
        expr * move_e;

        move_e = EXPR_new();
        _EXPR_make_move( move_e, s_e, blk );
        return move_e;
}

static inline void
_EXPR_make_move_to( expr * move_e, expr * s_e, instr_block * blk, int base_r, int pair_r )
{
        _EXPR_make_move( move_e, s_e, blk );
        move_e->flags = SE_in_reg;
        move_e->flags |= s_e->flags & SE_in_frame_slot;
        move_e->uses.block = 1;
        move_e->uses.other = 0;
        move_e->loc.reg.base = base_r;
        move_e->loc.reg.pair = pair_r;
        move_e->loc.slot.frame = s_e->loc.slot.frame;
}

void
EXPR_make_move_to( expr * move_e, expr * s_e, instr_block * blk, int base_r, int pair_r )
{
	_EXPR_make_move_to( move_e, s_e, blk, base_r, pair_r );
LOG(    fprintf( blk->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( blk, s_e ), s_e->uses.block ); );
}

expr *
EXPR_new_move_to( expr * s_e, instr_block * blk, int base_r, int pair_r )
{
        expr * move_e;

        move_e = EXPR_new();
        _EXPR_make_move_to( move_e, s_e, blk, base_r, pair_r );
        return move_e;
}

expr *
EXPR_new_reg_src( int reg, instr_block * blk )
{
        expr * e;

        e = EXPR_new();
        e->op = OP_nop;
        e->type = PRIM_int;
        EXPR_SET_SRC_REG( e, reg );
        e->eid = ++blk->cfg->env.global.id;
        BLK_ADD_EXPR( blk, e );
        return e;
}

void
EXPR_remove_from_exprs_by_op( instr_block * blk, expr * e )
{
        expr ** cur_e;

        for( cur_e = &blk->exprs_by_op[e->op]; *cur_e != NULL; cur_e = &(*cur_e)->exprs_by_op ){
                if( *cur_e == e ){
                        *cur_e = e->exprs_by_op;
                        return;
                }
        }
}

/*
 * A more complex search that returns prev instructions in presence 
 * of inlined and specialized methods
 */
expr *
CSE_prev_expr( instr_block * blk, expr * e, expr * call_e, Hjava_lang_Class * class )
{
        return NULL;
}

#ifdef OPTIMIZE_GCSE
static void
BLK_new_globals( instr_block * blk, int start, int cnt )
{
        int i, n;
        instr_block * pred_b, * cur_b, * ebb_root;
        expr * copy_e, * s_e;

        pred_b = blk->pred.blocks[0];
        for( i = start+cnt; --i >= start; ){
                s_e = pred_b->globals.out[i];
                copy_e = EXPR_new_copy( s_e, blk, false );
                blk->globals.out[i] = blk->globals.in[i] = copy_e;
                EXPR_LIST_APPEND( blk, blk->exprs_head, copy_e );
                copy_e->eid = blk->exprs_head->eid;
                copy_e->loc.slot.frame = s_e->loc.slot.frame;
        }
        ebb_root = blk->ebb_root;
        for( n = blk->succ.length; --n >= 0; ){
                cur_b = blk->succ.blocks[n];
                if( cur_b->ebb_root == ebb_root 
                   && cur_b->globals.in[start] == NULL ){
                        BLK_new_globals( cur_b, start, cnt );
                }
        }
}

static void
EBB_new_globals( instr_block * ebb_root, int start, int cnt )
{
        int n, i, id;
        instr_block * cur_b;
        slot * slots;

        assert( ebb_root->globals.in[start] == NULL );
        slots = gc_malloc_fixed( cnt * sizeof(slot) );
        id = 0;
        /* Following necessary? */
        if( ebb_root->globals.in[0] != NULL ){
                id = ebb_root->globals.in[0]->eid;
        }
        for( i = cnt; --i >= 0; ){
                ebb_root->globals.out[start+i] = ebb_root->globals.in[start+i] = (expr*)&slots[i];
//                slots[i].flags = SE_slot | SE_in_frame_slot | SE_global;
                slots[i].flags = SE_in_frame_slot;
                slots[i].loc.slot.frame = start+i;
                slots[i].eid = id;
        }
        for( n = ebb_root->succ.length; --n >= 0; ){
                cur_b = ebb_root->succ.blocks[n];
                if( cur_b->ebb_root == ebb_root 
                   && cur_b->globals.in[start] == NULL ){
                        BLK_new_globals( cur_b, start, cnt );
                }
        }
        for( n = ebb_root->ebb_succ.length; --n >= 0; ){
                cur_b = ebb_root->ebb_succ.blocks[n];
                if( cur_b->globals.in[start] == NULL ){
                        EBB_new_globals( cur_b, start, cnt );
                }
        }
}

void
METHOD_new_globals( method_env * m_env, int old_max, int new_max )
{
        int n;
        instr_block * blk;
        expr ** globals;

        blk = m_env->root;
        old_max = blk->globals.in == NULL ? 0 : old_max;
        while (blk != NULL) {
                globals = gc_malloc_fixed( 2 * new_max * sizeof(expr*) );
                if( old_max != 0 ){
                        bcopy( blk->globals.in, globals, old_max * sizeof(expr*) );
                        bcopy( blk->globals.out, &globals[new_max], old_max * sizeof(expr*) );
                        gc_free_fixed( blk->globals.in );
                }
                blk->globals.in = globals;
                blk->globals.out = &globals[new_max];
                if( m_env->caller != NULL ){
                        for( n = new_max; --n >= old_max; ){
                                blk->globals.out[n] = blk->globals.in[n] = m_env->caller->globals.out[n];
                        }
                }
                blk = blk->block.next;
        }
        if( m_env->caller == NULL ){
                EBB_new_globals( m_env->root, old_max, new_max - old_max );
        }
}

static void
DFG_expand_globals( cfg_info * cfg, int cnt )
{
        int new_max, old_max, n;
        method_env * m_env;
        slot_info * o_map;
        
        old_max = new_max = cfg->env.global.globals.max;
        if( cfg->env.global.globals.length + cnt >= old_max ){
                if( old_max == 0 ){
                        new_max = cfg->env.global.globals.max = GLOBAL_INIT_SIZE;
                        cfg->env.global.globals.map = gc_malloc_fixed( new_max * sizeof(slot_info) );
                }
                else{
                        new_max = cfg->env.global.globals.max <<= 1;
                        o_map = cfg->env.global.globals.map;
                        cfg->env.global.globals.map = gc_malloc_fixed( new_max * sizeof(slot_info) );
                        bcopy( o_map, cfg->env.global.globals.map, old_max * sizeof(slot_info) );
                        gc_free_fixed( o_map );
                }
#ifdef OPTIMIZE_INLINE
                for( m_env = cfg->env.global.methods; m_env != NULL; m_env = m_env->next ){
#endif
                        METHOD_new_globals( m_env, old_max, new_max );
#ifdef OPTIMIZE_INLINE
                }
#endif
                for( n = cfg->catch.length; --n >= 0; ){
                        EBB_new_globals( cfg->catch.blocks[n], old_max, new_max - old_max );
                }
        }
}

static void
GLOBAL_set_type( instr_block * blk, int ix, expr * global_e, instr_block * header, boolean ebb_only, block_list * defs )
{
        int n, i, j;
        expr * e;
        instr_block * succ_b, * pred_b;
        boolean defined;

        for( n = blk->succ.length; --n >= 0; ){
                succ_b = blk->succ.blocks[n];
                e = succ_b->globals.in[ix];
                /* FIXME: try to extend the global further beyond current loop */
                if( !(e->flags & SE_global) ){
                        if( !ebb_only && !BV_is_set( succ_b->dom, header->bid ) ){
                                if( defs == NULL ){
                                        continue;
                                }
                                defined = false;
                                for( i = succ_b->pred.length; --i >= 0; ){
                                        pred_b = succ_b->pred.blocks[i];
                                        defined = false;
                                        for( j = defs->length; --j >= 0; ){
                                                if( BV_is_set( pred_b->dom, defs->blocks[j]->bid ) ){
                                                        defined = true;
                                                        break;
                                                }
                                        }
                                        if( !defined ){
                                                break;
                                        }
                                }
                                if( defined ){
                                        header = succ_b->ebb_root;
                                }
                                else{
                                        continue;
                                }
                        }
                        e->type = global_e->type;
                        e->class = global_e->class;
                        e->flags |= SE_slot | SE_global;
                        e->flags |= global_e->flags & (SE_object_field | SE_static_field | SE_null_checked);
                        if( ebb_only ){
                                if( succ_b->ebb_root != blk->ebb_root ){
TRACE(                                  kprintf( "HEADER TO B%d ", succ_b->ebb_root->bid ); );
                                }
                                GLOBAL_set_type( succ_b, ix, global_e, succ_b->ebb_root, succ_b->ebb_root == blk->ebb_root, defs );
                        }
                        else{
                                GLOBAL_set_type( succ_b, ix, global_e, header, false, defs );
                        }
                }
        }
        for( n = blk->catch.length; --n >= 0; ){
                succ_b = blk->catch.blocks[n];
                e = succ_b->globals.in[ix];
                if( !(e->flags & SE_global) ){
                        if( !ebb_only && !BV_is_set( succ_b->dom, header->bid ) ){
                                if( defs == NULL ){
                                        continue;
                                }
                                defined = false;
                                for( i = succ_b->pred.length; --i >= 0; ){
                                        pred_b = succ_b->pred.blocks[i];
                                        defined = false;
                                        for( j = defs->length; --j >= 0; ){
                                                if( BV_is_set( pred_b->dom, defs->blocks[j]->bid ) ){
                                                        defined = true;
                                                        break;
                                                }
                                        }
                                        if( !defined ){
                                                break;
                                        }
                                }
                                if( defined ){
                                        header = succ_b->ebb_root;
                                }
                                else{
                                        continue;
                                }
                        }
                        e->type = global_e->type;
                        e->class = global_e->class;
                        e->flags |= SE_slot | SE_global;
                        e->flags |= global_e->flags & (SE_object_field | SE_static_field | SE_null_checked);
                        if( ebb_only ){
                                GLOBAL_set_type( succ_b, ix, global_e, succ_b->ebb_root, succ_b->ebb_root == blk->ebb_root, defs );
                        }
                        else{
                                GLOBAL_set_type( succ_b, ix, global_e, header, false, defs );
                        }
                }
        }
}

static void
BLK_expr_to_global( instr_block * blk, expr * new_e, int ix )
{
        int n;
        expr * base_e, * cmp_e, * copy_e;
        instr_block * succ_b;
        boolean found;

        base_e = EXPR_RESOLVE_ORIGINAL( new_e );
        for( n = blk->succ.length; --n >= 0; ){
                succ_b = blk->succ.blocks[n];
                if( succ_b->ebb_root == blk->ebb_root ){
                        found = false;
                        if( new_e->uses.other > 0 ){
                                cmp_e = succ_b->exprs_by_op[base_e->op];
                                while( cmp_e != NULL ){
                                        if( cmp_e->a[0] == new_e ){
                                                /* Free the old copy expr ? */
                                                if( succ_b->globals.out[ix] == succ_b->globals.in[ix] ){
                                                        succ_b->globals.out[ix] = cmp_e;
                                                }
                                                succ_b->globals.in[ix] = cmp_e;
                                                cmp_e->loc.slot.frame = new_e->loc.slot.frame;
                                                BLK_expr_to_global( succ_b, cmp_e, ix );
                                                found = true;
                                                break;
                                        }
                                        cmp_e = cmp_e->exprs_by_op;
                                }
                        }
                        if( !found ){
                                copy_e = succ_b->globals.in[ix];
                                copy_e->a[0] = new_e;
                                copy_e->a[1] = new_e->op == OP_copy || new_e->op == OP_move ? new_e->a[1] : new_e;
                                copy_e->type = new_e->type;
                        }
                }
        }
}

static int 
GLOBAL_assign_expr( instr_block * blk, expr * new_e, int ix, block_list * defs )
{
        cfg_info * cfg;
        expr * tmp_e;
        int n_globals;

        cfg = blk->cfg;
        if( ix == -1 ){
                n_globals = 1;
                assert( new_e->type != PRIM_long || new_e->type != PRIM_double ); /* verify dword global is being allocated */
//                n_globals = new_e->type == PRIM_long || new_e->type == PRIM_double ? 2 : 1;

                DFG_expand_globals( cfg, n_globals );
                ix = cfg->env.global.globals.length;
                cfg->env.global.globals.length += n_globals;
        }

        tmp_e = EXPR_RESOLVE_ORIGINAL( new_e );
        if( EXPR_IS_CONST( tmp_e ) ){
                new_e = EXPR_new_move( new_e, blk );
                GENERATED_BLK_ADD_EXPR( blk, new_e );
        }
#if 1
        if( new_e->op == OP_copy ){
                while( new_e->op == OP_copy ){
                        new_e = new_e->a[0];
                        blk = blk->pred.blocks[0];
                }
        }
#endif
        
        new_e->loc.slot.frame = ix;
        new_e->flags |= SE_slot | SE_global;
        if( new_e->eid == 0 ){
                new_e->eid = ++cfg->env.global.id;
                GENERATED_BLK_ADD_EXPR( blk, new_e );
        }
        cfg->env.global.globals.map[ix].size = PRIM_SIZE( new_e->type ) == 8 ? 2 : 1;
        blk->globals.out[ix] = new_e;
        BLK_expr_to_global( blk, new_e, ix );
        GLOBAL_set_type( blk, ix, new_e, blk, blk->ebb_root != blk, defs );
        if( new_e->op == OP_load ){
		new_e->a[0]->flags |= SE_null_checked;
        }
//        GLOBAL_set_type( blk, ix, new_e, blk, true, defs );
TRACE(  kprintf( "NEW G%d @ B%d, ", ix, blk->bid ); );
        return ix;
}

static void
GLOBAL_kill_unused( cfg_info * cfg )
{
        method_env * m_env;
        instr_block * blk;
        int n, used, total;
        expr * e;

        used = cfg->env.global.globals.length;
        total = cfg->env.global.globals.max;
#ifdef OPTIMIZE_INLINE
        for( m_env = cfg->env.global.methods; m_env != NULL; m_env = m_env->next ){
#endif
                blk = m_env->root;
                if( m_env->inlined ){
                        blk = blk->block.next;
                }
                while( blk != NULL ){
                        if( blk->ebb_root != blk && blk->stage > STG_generated_exprs ){
                                for( n = total; --n >= used; ){
                                        e = blk->globals.in[n];
                                        blk->globals.in[n] = blk->globals.out[n] = NULL;
                                        EXPR_REMOVE( blk, e );
                                }
                        }
                        blk = blk->block.next;
                }
#ifdef OPTIMIZE_INLINE
        }
#endif
        cfg->env.global.globals.max = used;
}
#endif

expr *
CSE_recurse( expr * new_e, instr_block * cur_b, instr_block * ebb_root, cseFunc cse, int op, boolean globalize )
{
        expr ** exprs_by_op, * cur_e;
        
        exprs_by_op = &cur_b->exprs_by_op[op];
        cur_e = (*cse)( new_e, exprs_by_op, cur_b, true );
        if( cur_e != NULL ){
                return cur_e;
        }
        if( cur_b == ebb_root || cur_b->pred.length != 1 ){
                return NULL;
        }
        /* FIXME: terminate on arg redefine? */
#ifdef OPTIMIZE_INLINE /* Don't CSE in inlined method yet */
        cur_e = CSE_recurse( new_e, cur_b->pred.blocks[0], ebb_root, cse, op, globalize );
#else
        cur_e = CSE_recurse( new_e, cur_b->pred.blocks[0], ebb_root, cse, op, globalize );
#endif
        if( cur_e == NULL || cur_e == (void*)CSE_TERMINATE ){
                return cur_e;
        }
        if( cur_e->flags & SE_slot ){
#ifdef OPTIMIZE_GCSE
                if( cur_e->flags & SE_global ){
                        cur_e = cur_b->globals.in[cur_e->loc.slot.frame];
                }
                else{
#endif
                        cur_e = cur_b->locals.in[cur_e->loc.slot.frame];
#ifdef OPTIMIZE_GCSE
                }
#endif
        }
        else{
#ifdef OPTIMIZE_GCSE
                if( globalize ){
                        /* assign at lowest point ? */
                        GLOBAL_assign_expr( cur_b->pred.blocks[0], cur_e, -1, NULL );
                        cur_e = cur_b->globals.in[cur_e->loc.slot.frame];
                }
                else{
#endif
                        cur_e = EXPR_new_copy( cur_e, cur_b, false );
                        // cur_e->a[0]->uses.block++; /* FIX??? */
                        // cur_e->a[0]->uses.other++; /* FIX??? */
                        exprs_by_op = &(cur_b->exprs_by_op[op]);
                        CSE_LIST_ADD( exprs_by_op, cur_e);
                        EXPR_LIST_APPEND(cur_b, cur_b->exprs_head, cur_e );
                        cur_e->eid = cur_b->exprs_head->eid;
#ifdef OPTIMIZE_GCSE
                }
#endif
        }
        return cur_e;
}

#define EBB_CSE 1

#ifdef OPTIMIZE_LOOPS
#ifdef OPTIMIZE_GCSE

static boolean
SLOT_redefined( expr * a, instr_block * blk )
{
        instr_block * ebb_root, * tmp_b;
        int ix;

        ebb_root = blk->ebb_root;
        if( a->flags & SE_slot ){
                ix = a->loc.slot.frame;
                if( a->flags & SE_global ){
                        if( !(ebb_root->globals.in[ix]->flags & SE_global) ){
                                return true;
                        }
                }
                else{
                        for( tmp_b = blk; tmp_b != ebb_root ; tmp_b = tmp_b->pred.blocks[0] ){
                                if( tmp_b->locals.out[ix] != tmp_b->locals.in[ix] ){
                                        return true;
                                }
                        }
                }
        }
        return false;
}

static boolean
BLK_slot_redefined( expr * a, instr_block * blk )
{
        int ix;

        if( a->flags & SE_slot ){
                ix = a->loc.slot.frame;
                if( a->flags & SE_global ){
                        if( !(blk->globals.in[ix]->flags & SE_global) ){
                                return true;
                        }
                }
                else{
                        if( blk->locals.out[ix] != blk->locals.in[ix] ){
                                return true;
                        }
                }
        }
        return false;
}

instr_block *
GCSE_recurse( expr * new_e, instr_block * cur_b, instr_block * ebb_root, cseFunc cse, int op )
{
        expr ** exprs_by_op, * cur_e, * a0_e, * a1_e;
        instr_block * blk;

TRACE(  kprintf( "B%d ", cur_b->bid ); );
        exprs_by_op = &cur_b->exprs_by_op[op];
        cur_e = (*cse)( new_e, exprs_by_op, cur_b, false );
        if( cur_e != NULL ){
TRACE(          if( cur_e != (void*)CSE_TERMINATE ){
                        kprintf( "MB%d ", cur_b->bid );
                } );
                return cur_e == (void*)CSE_TERMINATE ? (instr_block*)CSE_TERMINATE : cur_b;
        }
        if( cur_b == ebb_root || cur_b->pred.length != 1 ){
                return NULL;
        }
#define EXPR_RESOLVE( _e ) ( ( (_e) != NULL && (_e)->op == OP_copy ) ? (_e)->a[1] : (_e) )

        a1_e = EXPR_RESOLVE( new_e->a[1] );
        a0_e = EXPR_RESOLVE( new_e->a[0] );
        switch( OP_type[op] ){
        case OP_2a:
                if( BLK_slot_redefined( a1_e, cur_b ) ){
                        return (instr_block*)CSE_TERMINATE;
                }
                break;
        case OP_1a:
                if( BLK_slot_redefined( a0_e, cur_b ) ){
                        return (instr_block*)CSE_TERMINATE;
                }
                break;
        }
#ifdef OPTIMIZE_INLINE /* Don't CSE in inlined method yet */
        blk = GCSE_recurse( new_e, cur_b->pred.blocks[0], ebb_root, cse, op );
#else
        blk = GCSE_recurse( new_e, cur_b->pred.blocks[0], ebb_root, cse, op );
#endif
        return blk;
}

static boolean
EXPR_repoint_arg( expr * e, expr * a, int ix, instr_block * blk, boolean insert_const )
{
        expr * const_e;
        instr_block * ret_b;

        if( a->flags & SE_slot ){
                e->a[ix] = a->flags & SE_global ?
                        blk->globals.out[a->loc.slot.frame]
                                : blk->locals.out[a->loc.slot.frame];
TRACE(          kprintf( "a[%d]: %c%d [%d], ", ix, a->flags & SE_global ? 'G' : 'L', 
                        a->loc.slot.frame, e->a[ix]->eid ); );
        }
        else if( EXPR_IS_CONST( a ) ){
                const_e = EXPR_new();
                const_e->op = a->op;
                const_e->type = a->type;
                const_e->val.l_const = a->val.l_const;
TRACE(          kprintf( "a[%d]: C 0x%x (%d), ", ix, a->val.i_const, a->val.i_const ); );
                if( insert_const ){
                        EXPR_add( const_e, &e->a[ix], blk, blk->ebb_root );
                }
                else{
                        ret_b = GCSE_recurse( const_e, blk, blk->ebb_root, CSE_const, OP_const );
                        if( ret_b == NULL || ret_b == (void*)CSE_TERMINATE ){
                                EXPR_free( const_e );
TRACE(                          kprintf( "CONST DEAD " ); );
                                return false;
                        }
                        else{
                                EXPR_add( const_e, &e->a[ix], ret_b, ret_b->ebb_root );
                        }
                }
        }
        else{
                FAILED();
        }
        return true;
}

static expr *
GCSE_find_expr( expr * new_e, instr_block * start_b )
{
        block_list stack, defs;
        expr * a0_e, * a1_e, tmp_e, * cur_e, * copy_e, * backup_e, * ret_e;
        expr ** exprs_by_op;
        instr_block * pred_b, * match_b, * ebb_root, * cur_b;
        cseFunc cse;
        int n, op, ix;
        cfg_info * cfg;
        boolean valid_args, matched_slot;
        loopFunc loop;

        cfg = start_b->cfg;
        op = new_e->op;
        cse = CSE_funcs[op];
        if( cse == NULL 
           || cfg->env.global.in_nested_block
           || start_b->ebb_root->pred.length == 0
           || cse == CSE_call /* FIXME: implement */
           || ( EXPR_IS_CBRANCH( new_e ) && new_e->val.label->type == LAB_block_target ) ){
                return NULL;
        }
        /* verify args are locals/global/constant */
        a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        switch( OP_type[op] ){
        case OP_2a:
                if( ( !(a1_e->flags & SE_slot) && !EXPR_IS_CONST( a1_e ) ) 
                   || ( a1_e->flags & SE_slot && !(a1_e->flags & SE_global) 
                       && a1_e->loc.slot.frame >= cfg->env.method.n_locals ) ){
                        return NULL;
                }
                /* Fall through */
        case OP_1a:
                if( ( !(a0_e->flags & SE_slot) && !EXPR_IS_CONST( a0_e ) )
                   || ( a0_e->flags & SE_slot && !(a0_e->flags & SE_global) 
                       && a0_e->loc.slot.frame >= cfg->env.method.n_locals ) ){
                        return NULL;
                }
                break;
        default:
                return NULL;
                break;
        }
        LIST_zero( &stack );
        LIST_zero( &defs );
        BL_add( &stack, start_b->ebb_root );
TRACE(	kprintf( "GCSE START @ B%d -> B%d [%d]:\n", start_b->bid, start_b->ebb_root->bid, cfg->env.global.id ); );
	while( stack.length > 0 ){
                cur_b = stack.blocks[--stack.length];
TRACE(          kprintf( "@ B%d:\n", cur_b->bid ); );
                if( ( start_b->loop == cur_b->loop && cur_b->flags & BLK_loop_entry )
                   || cur_b->ebb_root->flags & BLK_finally 
                   || cur_b->pred.length == 0 ){
                        LIST_free( &stack );
                        LIST_free( &defs );
TRACE(                  kprintf( "\tTERMINATE: stop\n" ); );
                        return NULL;
                }
                if( start_b->loop != cur_b->loop && cur_b->flags & BLK_loop_entry 
                   && cur_b->loop != NULL ){
                        loop = LOOP_funcs[op];
                        if( loop != NULL && !(*loop)( cur_b, new_e, cur_b->loop ) ){
                                LIST_free( &stack );
                                LIST_free( &defs );
TRACE(                          kprintf( "\tTERMINATE: loop dependency\n" ); );
                                return NULL;
                        }
                }
                for( n = cur_b->pred.length; --n >= 0; ){
                        pred_b = cur_b->pred.blocks[n];
                        ebb_root = pred_b->ebb_root;
                        /* this block must dominate cur_b */
                        if( start_b->loop != cur_b->loop && cur_b->flags & BLK_loop_entry
                           && !(pred_b->flags & BLK_loop_header) ){
                                continue;
                        }
TRACE(                  kprintf( "\tPB%d: %s, ", pred_b->bid, op_names[op] );
                        if( op == OP_load ){
                                kprintf( "O %d ", new_e->val.i_const );
                        } );
                        bcopy( new_e, &tmp_e, sizeof(expr) );
                        valid_args = true;
                        switch( OP_type[op] ){
                        case OP_2a:
                                valid_args &= EXPR_repoint_arg( &tmp_e, a1_e, 1, pred_b, false );
                                /* Fall through */
                        case OP_1a:
                                valid_args &= EXPR_repoint_arg( &tmp_e, a0_e, 0, pred_b, false );
                                break;
                        }
TRACE(                  if( !valid_args ){
                                kprintf( "!VALID_ARGS " );
                        } );
                        match_b = valid_args ? GCSE_recurse( &tmp_e, pred_b, ebb_root, cse, op ) : NULL;
                        if( match_b == NULL ){
                                /* terminate if arg locals/global def occurs */
                                switch( OP_type[op] ){
                                case OP_2a:
                                        if( SLOT_redefined( a1_e, pred_b ) ){
                                                LIST_free( &stack );
                                                LIST_free( &defs );
TRACE(                                          kprintf( "TERMINATE: a1 redefined\n" ); );
                                                return NULL;
                                        }
                                        /* Fall through */
                                case OP_1a:
                                        if( SLOT_redefined( a0_e, pred_b ) ){
                                                LIST_free( &stack );
                                                LIST_free( &defs );
TRACE(                                          kprintf( "TERMINATE: a0 redefined\n" ); );
                                                return NULL;
                                        }
                                }
TRACE(                          kprintf( "PUSH ROOT: B%d\n", ebb_root->bid ); );
                                if( !BL_contains( &stack, ebb_root ) ){
                                        BL_add( &stack, ebb_root );
                                }
                        }
                        else{
                                if( match_b == (void*)CSE_TERMINATE ){
TRACE(                                  kprintf( "FAIL\n" ); );
                                        LIST_free( &stack );
                                        LIST_free( &defs );
                                        return NULL;
                                }
TRACE(                          kprintf( "MATCH\n" ); );
                                if( !BL_contains( &defs, match_b ) ){
                                        BL_add( &defs, match_b );
                                }
                        }
                }
        }
        if( defs.length == 0 ){
                LIST_free( &stack );
                LIST_free( &defs );
TRACE(          kprintf( "\tTERMINATE: no match\n" ); );
                return NULL;
        }
TRACE(  kprintf( "MATCHES: " );
        for( n = defs.length; --n >= 0; ){
                kprintf( "B%d, ", defs.blocks[n]->bid );
        } );
        backup_e = ret_e = NULL;
        ix = -1;
        matched_slot = false;
        for( n = defs.length; --n >= 0; ){
                match_b = defs.blocks[n];
                if( n > 0 ){
                        backup_e = EXPR_new();
                        bcopy( new_e, backup_e, sizeof(expr) );
                }
                switch( OP_type[op] ){
                case OP_2a:
//                        if( n == 0 ){
//                                new_e->a[1]->uses.block--;
//                        }
                        valid_args &= EXPR_repoint_arg( new_e, a1_e, 1, match_b, true );
//                        new_e->a[1]->uses.block++;
                        /* Fall through */
                case OP_1a:
//                        if( n == 0 ){
//                                new_e->a[0]->uses.block--;
//                        }
                        valid_args &= EXPR_repoint_arg( new_e, a0_e, 0, match_b, true );
//                        new_e->a[0]->uses.block++;
                        break;
                }
#ifdef EBB_CSE
                // cur_e = CSE_recurse( new_e, match_b, match_b->ebb_root, cse, op, true );
                cur_e = CSE_recurse( new_e, match_b, match_b->ebb_root, cse, op, false );
#else
                exprs_by_op = &header_b->exprs_by_op[op];
                // cur_e = (*cse)( new_e, exprs_by_op, match_b, true );
                cur_e = (*cse)( new_e, exprs_by_op, match_b, false );
#endif
                
                if( cur_e != NULL && cur_e != (void*)CSE_TERMINATE ){
                        assert( new_e != cur_e );
                        EXPR_free( new_e );
                        if( cur_e->flags & SE_slot ){
                                if( cur_e->flags & SE_global ){
                                        if( ix != -1 && ix != cur_e->loc.slot.frame ){
                                                cur_e = EXPR_new_move( cur_e, match_b );
                                                ix = GLOBAL_assign_expr( match_b, cur_e, ix, &defs );
                                                GENERATED_BLK_ADD_EXPR( match_b, cur_e );
TRACE(                                          kprintf( "MATCH +G%d @ B%d, ", cur_e->loc.slot.frame, match_b->bid ); );
                                                ret_e = start_b->globals.in[cur_e->loc.slot.frame];
                                                matched_slot = true;
                                                if( n > 0 ){
                                                        new_e = backup_e;
                                                }
                                                continue;
                                        }
TRACE(                                  kprintf( "MATCH G%d @ B%d, ", cur_e->loc.slot.frame, match_b->bid ); );
                                        ret_e = start_b->globals.in[cur_e->loc.slot.frame];
                                        ix = cur_e->loc.slot.frame;
                                        matched_slot = true;
                                        if( n > 0 ){
                                                new_e = backup_e;
                                        }
                                        continue;
                                }
                                else{
// FIXME:
//                                        if( cur_e->loc.slot.frame >= cfg->env.method.n_locals 
//                                           || SLOT_redefined( cur_e, match_b ) ){
                                        if( 1 ){
                                                cur_e = EXPR_new_move( cur_e, match_b );
                                                ix = GLOBAL_assign_expr( match_b, cur_e, ix, &defs );
                                                GENERATED_BLK_ADD_EXPR( match_b, cur_e );
TRACE(                                          kprintf( "MATCH +G%d @ B%d, ", cur_e->loc.slot.frame, match_b->bid ); );
                                                ret_e = start_b->globals.in[cur_e->loc.slot.frame];
                                                matched_slot = true;
                                                if( n > 0 ){
                                                        new_e = backup_e;
                                                }
                                                continue;
                                        }
TRACE(                                  kprintf( "MATCH L%d @ B%d\n", cur_e->loc.slot.frame, match_b->bid ); );
                                        ret_e = start_b->locals.in[cur_e->loc.slot.frame];
                                        matched_slot = true;
                                        if( n > 0 ){
                                                new_e = backup_e;
                                        }
                                        continue;
                                }
                        }
                        else{
                                /* FIXME: may not be necessary anymore */
                                if( cur_e->op == OP_copy ){
                                        copy_e = cur_e;
                                        while( copy_e != NULL && copy_e->op == OP_copy ){
                                                copy_e->a[0]->uses.other++; /* FIX ?? */
                                                copy_e = copy_e->a[0];
                                        }
                                }
                                /* FIXME: if a copy, force a move first? */
TRACE(                          kprintf( "MATCH [%d] @ B%d, ", cur_e->eid, match_b->bid ); );
                                new_e = cur_e;
                        }
                }
                else{
#ifdef EBB_CSE
                        exprs_by_op = &match_b->exprs_by_op[op];
#endif
                        CSE_LIST_ADD( exprs_by_op, new_e );
                }
                ix = GLOBAL_assign_expr( match_b, new_e, ix, &defs );
                EXPR_update_use( new_e, match_b );
                if( n > 0 ){
                        new_e = backup_e;
                }
                ret_e = start_b->globals.in[ix];
TRACE(          kprintf( "RETURN G%d @ B%d, ", ix, start_b->bid ); );
        }
TRACE(  kprintf( "\n" ); );
        LIST_free( &stack );
        LIST_free( &defs );
        return ret_e;
}

static expr *
LOOP_add_expr( expr * new_e, instr_block * cur_b, int depth, loop_info * loop_i, loopFunc loop )
{
        expr * ret_e, * a_e, * cur_e, * copy_e;
        expr ** exprs_by_op;
        cfg_info * cfg;
        cseFunc cse;
        int ix, op;
        instr_block * header_b;

        header_b = loop_i->header;
        if( header_b->loop != NULL && BL_contains( &header_b->loop->blocks, cur_b ) ){
                ret_e = new_e;
                ret_e = LOOP_add_expr( new_e, cur_b, depth+1, loop_i->header->loop, loop );
                if( ret_e != new_e ){
                        return ret_e;
                }
        }
        if( !(*loop)( cur_b, new_e, loop_i ) ){
                return new_e;
        }
TRACE(  kprintf( "LOPT START @ B%d:\n\t", cur_b->bid ); );
        cfg = cur_b->cfg;
        op = new_e->op;
        cfg->env.global.header_block = header_b;
        /* FIXME: move down? */
        switch( OP_type[op] ){
        case OP_2a:
//                new_e->a[1]->uses.block--;
                a_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
                EXPR_repoint_arg( new_e, a_e, 1, header_b, true );
//                new_e->a[1]->uses.block++;
                /* Fall through */
        case OP_1a:
//                new_e->a[0]->uses.block--;
                a_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
                EXPR_repoint_arg( new_e, a_e, 0, header_b, true );
//                new_e->a[0]->uses.block++;
                break;
        case 0:
                break;
        }

        cse = CSE_funcs[new_e->op];
        if( cse != NULL 
           && !(cfg->env.global.in_nested_block)
           && !( EXPR_IS_CBRANCH( new_e ) && new_e->val.label->type == LAB_block_target ) ){
#ifdef EBB_CSE
                cur_e = CSE_recurse( new_e, header_b, header_b->ebb_root, cse, op, true );
#else
                exprs_by_op = &header_b->exprs_by_op[op];
                cur_e = (*cse)( new_e, exprs_by_op, header_b, true );
#endif
                if( cur_e != NULL && cur_e != (void*)CSE_TERMINATE ){
                        EXPR_free( new_e );
                        if( cur_e->flags & SE_slot ){
                                if( cur_e->flags & SE_global ){
TRACE(                                  kprintf( "MATCH G%d @ B%d\n", cur_e->loc.slot.frame, header_b->bid ); );
                                        return cur_b->globals.in[cur_e->loc.slot.frame];
                                }
                                else{
                                        if( cur_e->loc.slot.frame >= cur_b->cfg->env.method.n_locals 
                                           || STATS_local_get( SO_store, loop_i->locals, cur_e->loc.slot.frame ) ){
                                                cur_e = EXPR_new_move( cur_e, header_b );
                                                GLOBAL_assign_expr( header_b, cur_e, -1, NULL );
                                                GENERATED_BLK_ADD_EXPR( header_b, cur_e );
TRACE(                                          kprintf( "MATCH G%d @ B%d\n", cur_e->loc.slot.frame, header_b->bid ); );
                                                return cur_b->globals.in[cur_e->loc.slot.frame];
                                        }
TRACE(                                  kprintf( "MATCH L%d @ B%d\n", cur_e->loc.slot.frame, header_b->bid ); );
                                        return cur_b->locals.in[cur_e->loc.slot.frame];
                                }
                        }
                        else{
                                /* FIXME: may not be necessary anymore */
#if 1
                                if( cur_e->op == OP_copy ){
                                        copy_e = cur_e;
                                        while( copy_e != NULL && copy_e->op == OP_copy ){
                                                copy_e->a[0]->uses.other++; /* FIX ?? */
                                                copy_e = copy_e->a[0];
                                        }
                                }
#endif
                                /* FIXME: if a copy, force a move first? */
TRACE(                          kprintf( "MATCH [%d] @ B%d\n", cur_e->eid, cur_e->loc.slot.frame, header_b->bid ); );
                                new_e = cur_e;
                        }
                }
                else{
#ifdef OPTIMIZE_GCSE
                        if( enable_gcse && header_b->ebb_root != NULL
                           && cfg->env.global.globals.length < globals_limit
                           && cur_e != (void*)CSE_TERMINATE ){
                                cur_e = GCSE_find_expr( new_e, header_b );
                                if( cur_e != NULL ){
                                        assert( cur_e->flags & SE_slot );
                                        if( cur_e->flags & SE_global ){
TRACE(                                          kprintf( "MATCH G%d @ B%d\n", cur_e->loc.slot.frame, header_b->bid ); );
                                                return cur_b->globals.in[cur_e->loc.slot.frame];
                                        }
                                        else{
TRACE(                                          kprintf( "MATCH L%d @ B%d\n", cur_e->loc.slot.frame, header_b->bid ); );
                                                return cur_b->locals.in[cur_e->loc.slot.frame];
                                        }
                                }
                        }
#endif
#ifdef EBB_CSE
                        exprs_by_op = &header_b->exprs_by_op[op];
#endif
                        CSE_LIST_ADD( exprs_by_op, new_e );
                }
        }
                
        ix = GLOBAL_assign_expr( header_b, new_e, -1, NULL );
        EXPR_update_use( new_e, header_b );
TRACE(  kprintf( "RETURN G%d @ B%d\n", ix, cur_b->bid ); );
        return cur_b->globals.in[ix];
}
#endif
#endif

/*
 * Add expression to block, try optimizing along the way
 */
void
EXPR_add( expr * new_e, expr ** dst, instr_block * blk, instr_block * ebb_root )
{
        /* Try to reduce the expression */
        cpFunc cp;
        expr * cur_e, * a0_e, * a1_e;
        cseFunc cse;
        int op;
        expr ** exprs_by_op;
        prepFunc prep;
#ifdef OPTIMIZE_LOOPS
        loopFunc loop;
#endif
        uint flags;

        flags = blk->cfg->env.global.add_flags;
#ifdef OPTIMIZE_CONSTANTS
        cp = CP_funcs[new_e->op];
        /* First try CP on current expr */
        if( cp != NULL && !(flags & ADD_disable_cp) ) {
                uint a0_op, a1_op;
                a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
                a0_op = a0_e->op;
                switch( OP_type[new_e->op] ){
                case OP_2a:
                        a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
                        a1_op = a1_e->op;
                        if( a0_op <= OP_const && a0_op != 0 && a1_op <= OP_const && a1_op != 0 ) {
                                if( a0_e->type != PRIM_label && a1_e->type != PRIM_label ){
                                        (*cp)( new_e );
                                }
                        } 
                        else if( ( ( a0_op < OP_const && a0_op != 0 ) 
                                     || ( a1_op < OP_const && a1_op != 0 ) )
                                   && (cp = PP_funcs[new_e->op]) != NULL ){
                                (*cp)(new_e);
                        }
                        break;
                case OP_1a:
                        if( a0_op <= OP_const && a0_op != 0 && a0_e->type != PRIM_label ){
                                (*cp)( new_e );
                        }
                        break;
                default:
                        break;
                }
        }
#endif
        op = new_e->op;
#ifdef OPTIMIZE_CONSTANTS
        if (op == OP_move) {
                cur_e = new_e->a[0];
//                cur_e->uses.block--;
                EXPR_free(new_e);
                if (dst != 0) {
                        (*dst) = cur_e;
                }
                return;
        }
        if (op == OP_nop) {
                EXPR_free(new_e);
                return;
        }
#endif
#ifdef OPTIMIZE_LOOPS
        loop = LOOP_funcs[op];
        if( enable_loopopt && blk->loop != 0 && loop != NULL
           && blk->cfg->env.global.globals.length < globals_limit
           && !blk->cfg->env.global.no_reduce_base 
#ifdef USE_CC
            && ( new_e->op != OP_set_cc || ( new_e->op == OP_set_cc && new_e->flags & SE_unsigned ) )
#endif
            ){
                cur_e = LOOP_add_expr( new_e, blk, 1, blk->loop, loop );
                if( cur_e != new_e ){
                        if( dst != NULL ){
                                (*dst) = cur_e;
                        }
                        return;
                }
        }
#endif
#ifdef OPTIMIZE_CSE
        /* Search backward through operation list and try CSE */
        cse = CSE_funcs[op];
        if( cse != NULL 
           && !(blk->cfg->env.global.in_nested_block) && !(flags & ADD_disable_cse)
           && !( EXPR_IS_CBRANCH( new_e ) && new_e->val.label->type == LAB_block_target ) ){
#ifdef EBB_CSE
                cur_e = CSE_recurse( new_e, blk, ebb_root, cse, op, false );
#else
                exprs_by_op = &blk->exprs_by_op[op];
                cur_e = (*cse)( new_e, exprs_by_op, blk, true );
#endif
                if( cur_e != NULL && cur_e != (void*)CSE_TERMINATE ){
                        EXPR_free( new_e );
                        if( dst != NULL ){
                                (*dst) = cur_e;
                        }
                        return;
                }
#ifdef OPTIMIZE_GCSE
                if( enable_gcse && blk->ebb_root != NULL
                    && blk->cfg->env.global.globals.length < globals_limit
                    && cur_e != (void*)CSE_TERMINATE
#ifdef USE_CC
                    && ( new_e->op != OP_set_cc || ( new_e->op == OP_set_cc && new_e->flags & SE_unsigned ) )
#endif
                    ){
                        cur_e = GCSE_find_expr( new_e, blk );
                        if( cur_e != NULL ){
                                if( dst != NULL ){
                                        (*dst) = cur_e;
                                }
                                return;
                        }
                }
#endif
#ifdef EBB_CSE
                exprs_by_op = &blk->exprs_by_op[op];
#endif
                CSE_LIST_ADD( exprs_by_op, new_e );
        }
#endif
        prep = PREP_funcs[new_e->op];
        if( prep != NULL){
                (*prep)( new_e, blk );
        }
        new_e->eid = ++blk->cfg->env.global.id;
        if( !EXPR_IS_BRANCH( new_e ) && !EXPR_IS_CONST( new_e ) ){
                blk->stats.n_other++;
        }
        if( blk->stage < STG_generated_exprs ){
                BLK_ADD_EXPR( blk, new_e );
                verify_check( new_e->expr.next == NULL, blk->cfg->fd, 
                             "New expression [%d] is not last\n", 
                             new_e->eid );
        }
        else{
TRACE(          kprintf( "ADDTO: [%d] @ B%d, ", new_e->eid, blk->bid ); );
                GENERATED_BLK_ADD_EXPR( blk, new_e );
        }
        if( dst != NULL ){
                (*dst) = new_e;
        }
}

#ifdef OPTIMIZE_INLINE
static void
EXPR_increment_call_frame( expr_frame * frame )
{
        int i;
        arg_info * arg_p;

        for( i = frame->sz; --i >= 0; ){
                arg_p = &frame->a[i];
                if( arg_p->expr != NULL ){
                        arg_p->expr->uses.block++;
                }
        }
}

static void
EXPR_decrement_call_frame( expr_frame * frame )
{
        int i;
        arg_info * arg_p;

        for( i = frame->sz; --i >= 0; ){
                arg_p = &frame->a[i];
                if( arg_p->expr != NULL ){
                        arg_p->expr->uses.block--;
                }
        }
}

expr_frame *
EXPR_copy_call_frame( expr_frame * orig_f )
{
        expr_frame * new_f;
        int sz;

        sz = sizeof(expr_frame) + orig_f->sz * sizeof(arg_info);
        new_f = gc_malloc_fixed( sz );
        bcopy( orig_f, new_f, sz );
        EXPR_increment_call_frame( new_f );
        return new_f;
}
#endif


static void
EXPR_update_call_use( expr * call_e, instr_block * b )
{
        expr_frame * frame;
        expr * a0_e;
        int sz;

        a0_e = call_e->a[0];
        if( a0_e != NULL ){
                a0_e->uses.last = call_e->eid;
                a0_e->uses.block++;
        }
        frame = call_e->val.frame;
        sz = frame->sz;
        if( !(frame->flags & CALL_vm) ){
                sz--;
        }
        for( ; --sz >= 0; ){
                frame->a[sz].expr->uses.last = call_e->eid;
                frame->a[sz].expr->uses.block++;
        }
}

void
EXPR_update_use( expr * e, instr_block * b )
{
        switch( OP_type[e->op] ){
        case OP_1a:
		e->a[0]->uses.last = e->eid;
                e->a[0]->uses.block++;
                break;
        case OP_2a:
		e->a[0]->uses.last = e->a[1]->uses.last = e->eid;
                e->a[0]->uses.block++;
                e->a[1]->uses.block++;
                break;
        case OP_frame:
		EXPR_update_call_use( e, b );
                break;
        }
}

#if 0
static void
EBB_validate_types( instr_block * ebb_root )
{
        int n, i, n_pred, type, new_type, localsz, stacksz;
        expr ** locals;
        instr_block * cur_b;

        localsz = ebb_root->cfg->env.method.n_locals;
        locals = ebb_root->locals.in;
        stacksz = ebb_root->sp.in;

        for( i = localsz + stacksz; --i >= 0; ){
                n_pred = ebb_root->pred.length;
                type = 0;
                if( n_pred <= 1 
                    || !(ebb_root->locals.in[i]->uses.block > 0 
                        || ebb_root->locals.in[i]->uses.other > 0) ){
                        continue;
                }
                for( n = n_pred; --n >= 0; ){
                        cur_b = ebb_root->pred.blocks[n];
                        if( cur_b->stage != STG_generated_exprs ){ /* Remove once we have completely built all expressions */
                                continue;
                        }
                        new_type = cur_b->locals.out[i]->type;
                        if( type != new_type ){
                                if( type == 0 ){
                                        type = new_type;
                                }
                                else{
                                        verify_check( new_type == 0 
                                                     || ( type == PRIM_int && new_type == PRIM_ref )
                                                     || ( type == PRIM_ref && new_type == PRIM_int ), 
                                                     ebb_root->cfg->fd, "Unmatched types: B%d, L%d: %x %x\n", ebb_root->id, i, type, new_type );
                                }
                        }
                }
        }
}
#endif

static void
BLK_link_uses_( int sz, expr ** out_p, expr ** in_p, boolean loop_link )
{
        expr * cur_e, * succ_e;
        int i;

        for( i = sz; --i >= 0; in_p--, out_p-- ) {
                if( loop_link ){
                        cur_e = *in_p;
                        succ_e = *out_p;
                        if( cur_e == NULL || succ_e == NULL ){
                                continue;
                        }
                        if( cur_e->uses.other > 0 ){
                                succ_e->uses.other++; 
                        }
                }
                else{
                        cur_e = *out_p;
                        succ_e = *in_p;
                        if( cur_e == NULL || succ_e == NULL ){
                                continue;
                        }
                        cur_e->uses.other += succ_e->uses.block + succ_e->uses.other;
                }
        }
}

/*
* Need to find way to make this breadth first (for better estimate)
*/
static void
BLK_link_uses( instr_block * blk, int localsz, boolean loop_link )
{
        int n, sz;
        instr_block * succ_b;
	expr * call_e;
        cfg_info * cfg;
#ifdef OPTIMIZE_INLINE
        expr * prev_call_e;
        method_env * m_env;
#endif
        FILE * fd;

        cfg = blk->cfg;
        fd = cfg->fd;
LOG(    fprintf( fd, "\tB%d.%d <-- ", blk->bid, blk->env->inline_depth ); );
        for( n = blk->succ.length; --n >= 0; ){
                succ_b = blk->succ.blocks[n];
LOG(    	fprintf( fd, "B%d.%d ", succ_b->bid, succ_b->env->inline_depth ); );
                sz = localsz + blk->sp.out;
                BLK_link_uses_( sz, &blk->locals.out[sz-1], &succ_b->locals.in[sz-1], loop_link );
#ifdef OPTIMIZE_GCSE
                sz = cfg->env.global.globals.length;
                BLK_link_uses_( sz, &blk->globals.out[sz-1], &succ_b->globals.in[sz-1], loop_link );
#endif
        }
LOG(    fprintf( fd, "\n" ); );

#ifdef OPTIMIZE_INLINE
        call_e = blk->exprs_by_op[OP_call];
        while( call_e != NULL 
              && ( call_e->op == OP_copy || call_e->val.frame->flags & CALL_throw ) ){
                call_e = call_e->exprs_by_op;
        }
        while( call_e != NULL ){
                prev_call_e = call_e->exprs_by_op;
                while( prev_call_e != NULL 
                      && ( prev_call_e->op == OP_copy || prev_call_e->val.frame->flags & CALL_throw ) ){
                        prev_call_e = prev_call_e->exprs_by_op;
                }
                if( call_e->val.frame->flags & (CALL_inline | CALL_specialized_failed) ){
                        m_env = call_e->val.frame->env;
                        BLK_link( m_env->root, m_env->n_locals, loop_link );
                        if( call_e->val.frame->flags & CALL_specialized_failed ){
                                prev_call_e = prev_call_e->exprs_by_op;                                
                        }
                }
                call_e = prev_call_e;
        }
#endif
        if( !loop_link ){
                for( n = blk->catch.length; --n >= 0; ){
                        succ_b = blk->catch.blocks[n];
LOG(                    fprintf( fd, "eB%d ", succ_b->bid ); );
                        BLK_link_uses_( localsz, &blk->locals.out[localsz-1], &succ_b->locals.in[localsz-1], loop_link );
#ifdef OPTIMIZE_GCSE
                        sz = cfg->env.global.globals.length;
                        BLK_link_uses_( sz, &blk->globals.out[sz-1], &succ_b->globals.in[sz-1], loop_link );
#endif
                }
        }
}

static void
DFG_clear_link_flag( cfg_info * cfg )
{
        instr_block * blk;
#ifdef OPTIMIZE_INLINE
        method_env * m_env;
#endif

#ifdef OPTIMIZE_INLINE
        for( m_env = cfg->env.global.methods; m_env != NULL; m_env = m_env->next ){
#endif
                blk = m_env->root;
                while (blk != NULL) {
                        if( blk->stage == STG_linked_loops ){
                                blk->stage = STG_linked_uses;
                        }
                        blk = blk->block.next;
                }
#ifdef OPTIMIZE_INLINE
        }
#endif
}

static boolean
BLK_link( instr_block * blk, int localsz, boolean loop_link )
{
        int n, stage;
        instr_block * cur_b;
        boolean done;

        stage = loop_link ? STG_linked_loops : STG_linked_uses;
        assert( blk->stage != stage );

        done = true;
        blk->stage = stage;
        for( n = blk->succ.length; --n >= 0; ){
                cur_b = blk->succ.blocks[n];
                if( cur_b->stage != stage ){
                        done &= BLK_link( cur_b, localsz, loop_link );
                }
        }
        BLK_link_uses( blk, localsz, loop_link );
/*
        if( blk->ebb_root == blk ){
                EBB_validate_types( blk );
        }
*/
        return done;
}

void
DFG_prologue( cfg_info * cfg )
{
        int type;
        expr ** local_p;
        char * tp, * p;
        Method * method;
        Hjava_lang_Class * class;
        FILE * fd;

        method = cfg->env.method.method;
        class = cfg->env.method.class;
        local_p = cfg->env.method.root->locals.in;
        fd = cfg->fd;

LOG(    fprintf( fd, "DFG_prologue(): " ); );
        if( !METHOD_IS_STATIC( method ) ){
LOG(            fprintf( fd, "L" ); );
                (*local_p)->type = PRIM_ref;
                EXPR_reduce_class( *local_p, class );
                local_p++;
        }
        for( p = method->signature; (tp = sigItemType( &p )) != NULL; ){
                type = *tp;
LOG(            fprintf( fd, "%c", type ); );
                switch( type ){
                case 'L': 
                case '[': (*local_p)->type = PRIM_ref; 
                        EXPR_reduce_class( *local_p, getClassFromSignature( tp, class->loader ) );
                        break;
                case 'I': (*local_p)->type = PRIM_int; break;
                case 'Z': 
                case 'B': (*local_p)->type = PRIM_byte; break;
                case 'C': (*local_p)->type = PRIM_char; break;
                case 'S': (*local_p)->type = PRIM_short; break;
                case 'F': (*local_p)->type = PRIM_float; break;
                case 'D': (*local_p)->type = PRIM_double; local_p++; break;
                case 'J': (*local_p)->type = PRIM_long; local_p++; break;
                }
                local_p++;
        }
#if defined( CHECK_STACK_LIMIT ) && !defined( FAKE_STACK_LIMIT )
        cfg->env.method.root->locals.in[method->bcode->localsz]->type = PRIM_int;
#endif
LOG(    fprintf( fd, "\n" ); );
}


void
EBB_link_types( instr_block * blk, block_list * iblist, expr ** slots, int n_slots, boolean is_global )
{
        int n, i, type, valid_stage;
        boolean generated_pred, null_checked;
        expr * cur_e;
        instr_block * cur_b;
        Hjava_lang_Class * class;

LOG(    fprintf( blk->cfg->fd, "EBB_link_types(): B%d:%d from:\n", blk->bid, blk->env->inline_depth ); );
        valid_stage = is_global ? STG_generated_code : STG_generated_exprs;
        for( i = n_slots; --i >= 0; ){
                generated_pred = false;
                type = PRIM_void;
                class = NULL;
                null_checked = true;
                for( n = iblist->length; --n >= 0; ){
                        cur_b = iblist->blocks[n];
                        if( cur_b->stage == valid_stage ){
                                cur_e = is_global ? cur_b->globals.out[i] : cur_b->locals.out[i];
                                if( cur_e == NULL ){
                                        continue;
                                }
                                null_checked &= cur_e->flags & SE_null_checked ? true : false;
                                generated_pred = true;
LOG(                            fprintf( blk->cfg->fd, "\t" ); );
                                if( cur_e->type != 0 ){
                                        type = cur_e->type;
LOG(                                    fprintf( blk->cfg->fd, "%c%d:B%d:%x ", is_global ? 'G' : 'L', i, cur_b->bid, type ); );
                                }
                                if( cur_e->class != NULL ){
                                        class = cur_e->class;
                                }
LOG(                            fprintf( blk->cfg->fd, "%s, ", cur_e->class == NULL ? "" : class->name ); );
                        }
                }
                slots[i]->type = type;
                slots[i]->class = class;
                if( generated_pred && null_checked ){
                        slots[i]->flags |= SE_null_checked;
                }
                if( generated_pred && type != PRIM_void ){
LOG(    		fprintf( blk->cfg->fd, "\n" ); );
		}
                verify_check( generated_pred, blk->cfg->fd, 
                             "B%d: No generated predecessors\n", blk->bid );
        }
LOG(    fprintf( blk->cfg->fd, "\n" ); );
}

void
BLK_process_rescan( instr_block * blk )
{
        expr_list * el;
        expr * e, * a0_e, * a1_e;

        el = &blk->cfg->env.method.rescan.exprs;
        while( el->length > 0 ){
               e = el->exprs[el->length-1];
               EL_del( el, e );
               if( e->uses.block == 0 && e->uses.other == 0 ){
LOG(                   fprintf( blk->cfg->fd, "BLK_process_rescan(): B%d: [%d]\n", blk->bid, e->eid ); );
                       if( OP_type[e->op] & (OP_1a | OP_2a) ){
                               a0_e = e->a[0];
                               a0_e->uses.block--;
                               if( a0_e->uses.block == 0 && a0_e->uses.other == 0
                                  && !EL_contains( el, a0_e ) 
                                  && !(a0_e->flags & SE_slot)
                                  && a0_e->op != OP_copy && a0_e->op != OP_nop ){
                                       EL_add( el, a0_e );
                               }
                       }
                       if( OP_type[e->op] & OP_2a ){
                               a1_e = e->a[1];
                               a1_e->uses.block--;
                               if( a1_e->uses.block == 0 && a1_e->uses.other == 0
                                  && !EL_contains( el, a1_e )
                                  && !(a1_e->flags & SE_slot)
                                  && a1_e->op != OP_copy && a1_e->op != OP_nop ){
                                       EL_add( el, a1_e );
                               } 
                       }
                       EXPR_remove_from_exprs_by_op( blk, e );
                       EXPR_REMOVE( blk, e );
                       EXPR_free( e );
               }
        }
}

void
EBB_generate_exprs( block_list * stack ) 
{
        int n, ix;
        boolean generated_pred;
        instr_block * cur_b, * blk;
        block_list * bl;

        blk = NULL;
        while( (ix = stack->length) > 0 ){
                generated_pred = false;
                while( !generated_pred && --ix >= 0 ){
                        blk = stack->blocks[ix];
                        generated_pred = true;
                        bl = blk->flags & BLK_catch ? &blk->try : &blk->pred;
                        for( n = bl->length; --n >= 0; ){
                                cur_b = bl->blocks[n];
                                if( cur_b->stage == STG_generated_exprs ){
                                        if( blk->flags & BLK_catch
                                           || ( cur_b->flags & BLK_loop_header && !(blk->flags & BLK_catch) ) ){
                                                generated_pred = true;
                                                break;
                                        }
                                }
                                else{
                                        generated_pred = false;
                                }
                        }
                }
                if( ix < 0 ){
                        break;
                }
                BL_del( stack, blk );
                for( n = blk->ebb_succ.length; --n >= 0; ){
                        cur_b = blk->ebb_succ.blocks[n];
                        if( cur_b->stage != STG_generated_exprs ){
                                BL_add( stack, cur_b );
                        }
                }
                if( blk->stage != STG_generated_exprs ){
                        BLK_generate_exprs( blk, blk );
                }
        }
}

static int
_build_frame_positions( cfg_info * cfg, slot_info * map, int start, int end )
{
        int pos, i;

        pos = 0;
        for( i = start; i < end; i++ ){
                switch( map[i].size ){
                case 2:
                        DWORD_ALIGN( pos, 2 );
                case 1:
                        map[i].position = pos;
                        pos += map[i].size;
                        break;
                default:
                        break;
                }
LOG(            fprintf( cfg->fd, "\t[%d]: sz=%d, offset=%d\n", i, map[i].size, map[i].position ) );
        }
        DWORD_ALIGN( pos, 2 );
        return pos;
}

static void
METH_build_frame_positions( cfg_info * cfg, method_env * m_env )
{
        int max_words;

        if( m_env->caller != NULL && !(m_env->caller->env->flags & METH_built_positions) ){
                METH_build_frame_positions( cfg, m_env->caller->env );
        }
LOG(    fprintf( cfg->fd, "build_frame_positions(): %s.%s %s\n", m_env->class->name, m_env->method->name, m_env->method->signature ) );
        max_words = _build_frame_positions( cfg, m_env->cframe.map, 
                                    m_env->caller == NULL ? cfg->env.global.n_args : 0, m_env->n_slots );
        if( m_env->caller != NULL ){
                m_env->cframe.offset.words = m_env->caller->env->cframe.offset.words + max_words;
        }
        else{
                m_env->cframe.offset.words = max_words;
        }
        if( cfg->env.method.source == m_env ){
                cfg->env.method.cframe.offset.words = m_env->cframe.offset.words;
        }
        if( m_env->cframe.offset.words > cfg->env.global.cframe.max_words ){
                cfg->env.global.cframe.max_words = m_env->cframe.offset.words;
        }
        m_env->flags |= METH_built_positions;
}

static void
DFG_build_frame_positions( cfg_info * cfg )
{
        method_env * m_env;
        int max_words;

        for( m_env = cfg->env.global.methods; m_env != NULL; m_env = m_env->next ){
                if( !(m_env->flags & METH_built_positions) ){
                        METH_build_frame_positions( cfg, m_env );
                }
        }
LOG(    fprintf( cfg->fd, "build_frame_positions(): globals\n" ); );
        max_words = _build_frame_positions( cfg, cfg->env.global.globals.map, 0, cfg->env.global.globals.length );
        cfg->env.global.globals.words = max_words;
}

#define LG_CHAR (is_global ? 'G' : 'L')

static void
BLK_get_out_slots( instr_block * cur_b, expr ** slots, reg_join_info * join_r, slot * join_s, int n_slots, boolean is_global )
{
        int i;
        expr * s_e, * join_e;
        
        for( i = n_slots; --i >= 0; ){
                s_e = slots[i];
                join_e = (expr *)&join_s[i];
                if( s_e == NULL ){
                        continue;
                }
                if( s_e->flags & SE_in_reg && !( join_e->flags & SE_in_reg )
                   && ( !join_r[s_e->loc.reg.base].used 
                       || ( join_r[s_e->loc.reg.base].used 
                           && join_r[s_e->loc.reg.base].expr == s_e ) ) ){
LOG(                    fprintf( cur_b->cfg->fd, "BLK_get_out_slots(): B%d: %c%d: %s [%d]\n", 
                                cur_b->bid, LG_CHAR, i, reg_names[s_e->loc.reg.base], s_e->eid ); );
                        join_r[s_e->loc.reg.base].used = true;
                        join_r[s_e->loc.reg.base].expr = s_e;
                        join_e->flags |= SE_in_reg;
                        join_e->loc.reg.base = s_e->loc.reg.base;
                        join_e->loc.reg.pair = s_e->loc.reg.pair;
                        if( s_e->flags & SE_in_frame_slot ){
LOG(   			        fprintf( cur_b->cfg->fd, "BLK_get_out_slots(): B%d: %c%d: in_frame_slot\n", 
                                        cur_b->bid, LG_CHAR, i ); );
                                join_e->flags |= SE_in_frame_slot;
                        }
                }
        }
        /* Add routinue to move arg / return slot to persistent reg */
}

static void
BLK_get_in_slots( instr_block * cur_b, expr ** slots, reg_join_info * join_r, slot * join_s, int n_slots, boolean is_global )
{
        expr * s_e, * join_e;
        int i;
        
        if( cur_b->ebb_root != cur_b ){
                return;
        }
        if( cur_b->stage == STG_generated_code || (cur_b->flags & BLK_linked_in_slots) ){
                return;
        }
        for( i = n_slots; --i >= 0; ){
                s_e = slots[i];
                join_e = (expr *)&join_s[i];
#if 1
                if( ( cur_b->flags & BLK_finally && i < cur_b->cfg->env.method.n_locals ) ){
LOG(                    fprintf( cur_b->cfg->fd, "BLK_get_in_slots(): B%d: %c%d: undo\n", 
                                cur_b->bid, LG_CHAR, i ); );
kprintf( "undo: B%d: %c%d\n", cur_b->bid, LG_CHAR, i );

                        join_e->flags &= ~( SE_in_reg | SE_in_frame_slot );
                }
#else
                /* Deprecated, keep around for debugging */
                if( ( join_e->flags & SE_in_reg 
                   && s_e->uses.block == 0 && s_e->uses.other == 0 )
                   || ( cur_b->flags & BLK_finally && i < cur_b->cfg->env.method.n_locals ) ){
LOG(                    fprintf( cur_b->cfg->fd, "BLK_get_in_slots(): B%d: %c%d: undo\n", 
                                cur_b->bid, LG_CHAR, i ); );
                        join_e->flags &= ~( SE_in_reg | SE_in_frame_slot );
                }
#endif
        }
}

void
EXPR_repoint( instr_block * blk, expr * new_e, int ix, boolean is_global )
{
        int n;
        expr * copy_e;
        instr_block * succ_b;

        for( n = blk->succ.length; --n >= 0; ){
                succ_b = blk->succ.blocks[n];
                if( succ_b->ebb_root == blk->ebb_root ){
                        copy_e = is_global ? succ_b->globals.in[ix] : succ_b->locals.in[ix];
                        copy_e->a[0] = new_e;
                        copy_e->a[1] = new_e->op == OP_copy || new_e->op == OP_move ? new_e->a[1] : new_e;
                        copy_e->type = new_e->type;
                }
        }
}

static void
BLK_set_out_slots( instr_block * cur_b, expr ** slots, instr_block * link_b, slot * join_s, int n_slots, boolean is_global )
{
        int i;
        expr * join_e, * d_e, * new_e;

        if( cur_b->stage == STG_generated_code ){
                if( cur_b != link_b ){
                        return;
                }
                for( i = n_slots; --i >= 0; ){
                        join_e = (expr *)&join_s[i];
                        d_e = slots[i];
                        if( d_e == NULL ){
                                continue;
                        }
                        if( join_e->flags & SE_in_reg ){
                                if( !(join_e->flags & SE_set_out_slot) ){
                                        continue;
                                }
                                if( d_e->flags & SE_in_reg ){
                                        if( d_e->loc.reg.base != join_e->loc.reg.base 
                                           || d_e->loc.reg.pair != join_e->loc.reg.pair ){
                                                /* Generate a move */
LOG(						fprintf( cur_b->cfg->fd, "BLK_set_out_slots(): B%d: %c%d: %s (v1)\n", 
                                                        cur_b->bid, LG_CHAR, i, reg_names[join_e->loc.reg.base] ); );
                                                new_e = EXPR_new_move_to( d_e, cur_b, join_e->loc.reg.base, join_e->loc.reg.pair );
                                                ASM_move( new_e, cur_b, false );
						// EXPR_repoint( cur_b, new_e, i, is_global );
						/* FIX: Free move */
                                        }
                                        if( !(d_e->flags & SE_in_frame_slot) && join_e->flags & SE_in_frame_slot ){
                                                /* Force a spill */
LOG(						fprintf( cur_b->cfg->fd, "BLK_set_out_slots(): spill 1\n" ); ); 
                                                REG_spill_expr( d_e, cur_b, false );
                                        }
                                }
                                else{
                                        /* Generate a load */
LOG(					fprintf( cur_b->cfg->fd, "BLK_set_out_slots(): B%d: %c%d: %s (v2)\n", 
                                                cur_b->bid, LG_CHAR, i, reg_names[join_e->loc.reg.base] ); );
                                        d_e->flags |= SE_preset_reg;
                                        d_e->uses.block++; /* To pass assertion in ASM_load_expr() */
LOG(    				fprintf( cur_b->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( cur_b, d_e ), d_e->uses.block ); );
                                        d_e->loc.reg.base = join_e->loc.reg.base;
                                        d_e->loc.reg.pair = join_e->loc.reg.pair;
                                        ASM_load_expr( cur_b, d_e );
                                }
                        }
                        else{
                                if( !(d_e->flags & SE_in_frame_slot) && d_e->flags & SE_in_reg ){
                                        /* Force a spill */
LOG(					fprintf( cur_b->cfg->fd, "BLK_set_out_slots(): spill 2\n" ); ); 
                                        REG_spill_expr( d_e, cur_b, false );
                                }
                        }
                }
        }
        else if( !(cur_b->flags & BLK_linked_out_slots) && cur_b->stage >= STG_generated_exprs ){
                for( i = n_slots; --i >= 0; ){
                        join_e = (expr *)&join_s[i];
                        if( !(join_e->flags & SE_set_out_slot) ){
                                continue;
                        }
                        d_e = slots[i];
                        if( join_e->flags & SE_in_reg ){
                                if( d_e->flags & SE_in_reg ){
LOG(					fprintf( cur_b->cfg->fd, "BLK_set_out_slots(): B%d: %c%d: %s (v3)\n", 
                                                cur_b->bid, LG_CHAR, i, reg_names[join_e->loc.reg.base] ); );
                                        if( d_e->loc.reg.base != join_e->loc.reg.base 
                                           || d_e->loc.reg.pair != join_e->loc.reg.pair ){
                                                /* Insert a move */
                                                new_e = EXPR_new_move_to( d_e, cur_b, join_e->loc.reg.base, join_e->loc.reg.pair );
                                                slots[i] = new_e;
                                                GENERATED_BLK_ADD_EXPR( cur_b, new_e );
                                                new_e->uses.block++;
LOG(   			 			fprintf( cur_b->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( cur_b, new_e ), new_e->uses.block ); );
                                        }
                                        else{
                                                d_e->uses.block++;
LOG(   			 			fprintf( cur_b->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( cur_b, d_e ), d_e->uses.block ); );
                                        }
                                }
                                else{
                                        /* Preset a destination reg */
LOG(					fprintf( cur_b->cfg->fd, "BLK_set_out_slots(): B%d: %c%d: %s (v4)\n", 
                                                cur_b->bid, LG_CHAR, i, reg_names[join_e->loc.reg.base] ); );
                                        d_e->uses.block++;
LOG(   			 		fprintf( cur_b->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( cur_b, d_e ), d_e->uses.block ); );
                                        d_e->flags |= SE_in_reg | SE_out_slot_reg;
                                        d_e->loc.reg.base = join_e->loc.reg.base;
                                        d_e->loc.reg.pair = join_e->loc.reg.pair;
                                }
                        }
                }
        }
}

static void
BLK_set_in_slots( instr_block * cur_b, expr ** slots, slot * join_s, int n_slots, boolean is_global )
{
        int i;
        expr * join_e, * d_e;
        
        if( cur_b->stage != STG_generated_code && !(cur_b->flags & BLK_linked_in_slots) ){
                for( i = n_slots; --i >= 0; ){
                        d_e = slots[i];
                        join_e = (expr *)&join_s[i];
                        if( join_e->flags & SE_in_reg ){
                                if( d_e->uses.block > 0 || d_e->uses.other > 0 ){
                                        assert( !(d_e->flags & SE_reg_assigned) );
LOG(					fprintf( cur_b->cfg->fd, "BLK_set_in_slots(): B%d: %c%d: %s\n", 
                                                cur_b->bid, LG_CHAR, i, reg_names[join_e->loc.reg.base] ); );
                                        d_e->flags |= SE_reg_assigned;
                                        join_e->flags |= SE_set_out_slot;
                                        d_e->loc.start.base = join_e->loc.reg.base;
                                        d_e->loc.start.pair = join_e->loc.reg.pair;
                                }
                                if( !(join_e->flags & SE_in_frame_slot) ){
                                        /* Make expression live only in reg */
LOG(				        fprintf( cur_b->cfg->fd, "BLK_set_in_slots(): B%d: %c%d: in_reg only\n", 
                                                 cur_b->bid, LG_CHAR, i ); );
                                        d_e->flags &= ~SE_in_frame_slot;
                                        if( d_e->uses.block > 0 || d_e->uses.other > 0 ){
                                                 assert( d_e->flags & SE_reg_assigned );
                                        }
                                }
                        }
                }
        }
}

static void
EBB_link_slots( instr_block * start_b )
{
        int i, j, n_slots, n_globals;
        block_list * in_bl;
        instr_block * cur_b;
        join_info * join_i;
        reg_join_info * regs;
        slot * slots;

        if( start_b->succ_join == NULL ){
                CFG_build_join( start_b );
                join_i = start_b->succ_join;
                slots = join_i->slots;
                n_slots = join_i->n_slots;
                n_globals = join_i->n_globals;
                regs = join_i->regs;
                BLK_get_out_slots( start_b, start_b->locals.out, regs, slots, n_slots, false );
#ifdef OPTIMIZE_GCSE
                BLK_get_out_slots( start_b, start_b->globals.out, regs, &slots[n_slots], n_globals, true );
#endif
                in_bl = &join_i->in;
                for( i = in_bl->length; --i >= 0; ){
                        BLK_get_in_slots( in_bl->blocks[i], in_bl->blocks[i]->locals.in, 
                                         regs, slots, n_slots, false );
#ifdef OPTIMIZE_GCSE
                        BLK_get_in_slots( in_bl->blocks[i], in_bl->blocks[i]->globals.in,
                                         regs, &slots[n_slots], n_globals, true );
#endif
                }
        }
        else{
                join_i = start_b->succ_join;
                slots = join_i->slots;
                n_slots = join_i->n_slots;
                n_globals = join_i->n_globals;
        }
LOG(    fprintf( start_b->cfg->fd, "EBB_link_slots(): %d slots in B%d\n", join_i->n_slots, start_b->bid ); );

        in_bl = &join_i->in;
        for( i = in_bl->length; --i >= 0; ){
                cur_b = in_bl->blocks[i];
                if( cur_b->ebb_root == cur_b ){
                        BLK_set_in_slots( cur_b, cur_b->locals.in, slots, n_slots, false );
#ifdef OPTIMIZE_GCSE
                        BLK_set_in_slots( cur_b, cur_b->globals.in, &slots[n_slots], n_globals, true );
#endif
                        if( cur_b->stage != STG_generated_code && !(cur_b->flags & BLK_linked_in_slots) ){
                                cur_b->flags |= BLK_linked_in_slots;
                        }
                        for( j = cur_b->pred.length; --j >= 0; ){
                                BLK_set_out_slots( cur_b->pred.blocks[j], cur_b->pred.blocks[j]->locals.out, start_b, 
                                                  slots, n_slots, false );
#ifdef OPTIMIZE_GCSE
                                BLK_set_out_slots( cur_b->pred.blocks[j], cur_b->pred.blocks[j]->globals.out, start_b, 
                                                  &slots[n_slots], n_globals, true );
#endif
                                if( cur_b->stage != STG_generated_code && !(cur_b->flags & BLK_linked_out_slots) && cur_b->stage >= STG_generated_exprs ){
                                        cur_b->flags |= BLK_linked_out_slots;
                                }
                        }
                }
        }
}

// inline void
void
EXPR_generate_code( expr * e, instr_block * blk )
{
        asmFunc assm;
        expr * last;

LOG(    fprintf( blk->cfg->fd, "[%d]:\n", e->eid ); );
#if 1
        if( e->uses.block == 0 && e->uses.other == 0
            && ( ( e->op > 0 && e->op <= OP_set_cc && e->op != OP_store )
                 || e->op == OP_cmp ) ){
LOG(            fprintf( blk->cfg->fd, "\tungenerated b/c uses=0\n" ); );
                return;
        }
#endif
        assm = ASM_funcs[e->op];
        if( assm != NULL && !(e->flags & SE_generated) ){
                (*assm)( e, blk, false );
                last = e->last_local;
                if( e->flags & SE_last_local
#ifdef OPTIMIZE_CMOV
                    && !EXPR_IS_CMOV( e )
#endif
                   && last != NULL 
                   && last->uses.block > 0
                   && !(last->flags & SE_in_reg) ){
                        REG_reload_expr( last, blk );
                        last->flags &= ~(SE_slot | SE_in_frame_slot);
LOG(    fprintf( blk->cfg->fd, "HACK %p\n", last ); );
                        kprintf( "**** HACKED FIX ****\n" );
                }
                if( blk->catch.length > 0 && e->flags & SE_slot && !(e->flags & SE_in_frame_slot) 
                    && (e->uses.block > 0 || e->uses.other > 0)
#ifdef OPTIMIZE_GCSE
                    && !(e->flags & SE_global)
#endif
                    ){
                        ASM_catch_spill_check( e, blk );
                }
        }
        /* We can be more conservative about syncing */
LOG(    fprintf( blk->cfg->fd, ".\n" ); );
        assert( blk->native.end < blk->cfg->env.global.native.max );
}

static void
BLK_fix_catch( instr_block * blk )
{
        int n;
        instr_block * catch_b;
                
        for( n = blk->catch.length; --n >= 0; ){
                catch_b = blk->catch.blocks[n];
                if( catch_b->handler != NULL ){
                        catch_b->handler->native.end = blk->native.end;
                }
        }
}

extern int enable_cmov;

#ifdef OPTIMIZE_CMOV
void
BLK_cvt_to_cmov( instr_block * cmov_b )
{
        int n;
        instr_block * pred_b, * succ_b, * blk, * ebb_root;
        expr * br_e, * mov_e, * cbr_e, * src_e, * new_e, * cmov_e, * t_e;
        loop_info * loop_i;

        /* make sure we don't try to sync this move */
        if( !enable_cmov ){
                return;
        }
        pred_b = cmov_b->pred.blocks[0];
        succ_b = cmov_b->succ.blocks[0];
        if( pred_b->succ.length > 2 
            || cmov_b->flags & BLK_loop_header 
            || pred_b->loop != cmov_b->loop 
            || pred_b == succ_b 
            || cmov_b->sp.out > pred_b->sp.out ){
                return;
        }
        assert( cmov_b->pred.length == 1 );
        assert( cmov_b->succ.length == 1 );
        if( EXPR_IS_BRANCH( cmov_b->exprs_tail ) ){
                br_e = cmov_b->exprs_tail;
                mov_e = br_e->expr.prev;
                src_e = mov_e->a[0];
                if( ( EXPR_IS_CONST( src_e ) && !(src_e->flags & SE_imm) )
                    || PRIM_TYPE( mov_e->type ) != PRIM_vm_int ){
                        return;
                }
                assert( mov_e->val.label == NULL );
                mov_e->val.label = br_e->val.label;
                if( !BL_contains( &pred_b->succ, succ_b ) && !BL_contains( &succ_b->pred, pred_b ) ){
                        ARC_add( pred_b, succ_b, ARC_normal );
                }
        }
        else{
                mov_e = cmov_b->exprs_tail;
                src_e = mov_e->a[0];
                if( ( EXPR_IS_CONST( src_e ) && !(src_e->flags & SE_imm) ) 
                    || PRIM_TYPE( mov_e->type ) != PRIM_vm_int ){
                        return;
                }
                assert( mov_e->val.label == NULL );
                mov_e->val.label = LAB_new_block_branch( cmov_b->cfg, mov_e, succ_b );
                if( !BL_contains( &pred_b->succ, succ_b ) && !BL_contains( &succ_b->pred, pred_b ) ){
                        ARC_add( pred_b, succ_b, ARC_normal );
                }
        }
        for( n = cmov_b->catch.length; --n >= 0; ){
                blk = cmov_b->catch.blocks[n];
                ARC_del( cmov_b, blk, ARC_try_catch );
        }
        ARC_del( pred_b, cmov_b, ARC_normal );
        ARC_del( cmov_b, succ_b, ARC_normal );
        if( cmov_b->sp.out > pred_b->sp.out ){
                pred_b->sp.out = cmov_b->sp.out;
        }
        if( EXPR_IS_CONST( src_e ) ){
                assert( PRIM_TYPE( mov_e->type ) == PRIM_vm_int );
                blk = pred_b;
                ebb_root = pred_b->ebb_root;
                move_int_const( new_e, src_e->val.i_const );
                mov_e->a[0] = new_e;
                new_e->uses.block++;
                if( new_e->eid > mov_e->eid ){
                        new_e->eid = mov_e->eid;
                }
        }
        else if( src_e->op == OP_copy ){
                mov_e->a[0]->uses.block--;
                mov_e->a[0] = mov_e->a[0]->a[0];
                mov_e->a[0]->uses.block++;
        }
        else{
                FAILED();
        }
        assert( mov_e->flags & SE_slot );

        for( t_e = pred_b->exprs_tail; t_e != NULL; t_e = t_e->expr.prev ){
                if( EXPR_IS_CBRANCH( t_e ) ){
                        cbr_e = t_e;
                        mov_e->a[1] = cbr_e->a[0];
                        mov_e->op = OP_cmov_eq + cbr_e->op - OP_cbranch_eq;
                        mov_e->flags |= SE_last_local;
                        mov_e->last_local = pred_b->locals.out[mov_e->loc.slot.frame];
                        mov_e->last_local->uses.other++;
                        pred_b->locals.out[mov_e->loc.slot.frame] = mov_e;
                        if( cbr_e->val.label->to.block != cmov_b ){
                                mov_e->op ^= 0x1;
                        }
                        EXPR_REMOVE( pred_b, cbr_e );
                        EXPR_free( cbr_e );
                        break;
                }
                else if( EXPR_IS_CMOV( t_e ) ){
                        cmov_e = t_e;
                        mov_e->a[1] = cmov_e->a[1];
                        mov_e->op = cmov_e->op ^ 0x1;
                        mov_e->flags |= SE_last_local;
                        mov_e->last_local = pred_b->locals.out[mov_e->loc.slot.frame];
                        mov_e->last_local->uses.other++;
                        pred_b->locals.out[mov_e->loc.slot.frame] = mov_e;
#if 0
                        if( mov_e->val.label != NULL ){
                                assert( mov_e->val.label->to.block == succ_b );
                                mov_e->val.label = NULL;
                        }
                        if( cmov_e->val.label != NULL ){
                                assert( cmov_e->val.label->to.block == succ_b );
                                cmov_e->val.label = NULL;
                        }
#endif
                        break;
                }
                else{
                        assert( EXPR_IS_CONST( t_e ) );
                }
        }
        assert( t_e != NULL );
        BL_del( &cmov_b->ebb_root->ebb, cmov_b );
        loop_i = cmov_b->loop;
        while( loop_i != NULL ){
                BL_del( &loop_i->blocks, cmov_b );
                loop_i = loop_i->header->loop;
        }
        EXPR_REMOVE( cmov_b, mov_e );
        BLK_ADD_EXPR( pred_b, mov_e );
        DLL_REMOVE( cmov_b, instr_block *, block );
        BLK_free( cmov_b );
kprintf( "SUCCESS!\n" );
}
#endif

static void
BLK_patch_branch( instr_block * cur_b, cfg_info * cfg )
{
        instr_block * last_b, * tgt_b;
        expr * e, * new_e;
        uint8 * code_p;

        last_b = cfg->env.global.last_block;
        if( cur_b != NULL && (cur_b->flags & BLK_finally) ){
                return;
        }
        if( cur_b != NULL && cur_b->flags & BLK_buffer_code ){
                cur_b = NULL;
        }
        if( last_b != NULL && !(last_b->flags & BLK_ret) ){
                switch( last_b->succ.length ){
                case 1:
                        if( last_b->succ.blocks[0] == cur_b ){
                                e = last_b->exprs_tail;
                                if( EXPR_IS_BRANCH( e ) ){
                                        if( last_b->flags & BLK_filled_delay_slot ){
                                                /* Move instruction out of delay slot */
                                                code_p = cfg->env.global.native.code.end - WORD_SZ;
                                                *(uint32 *)(code_p - WORD_SZ) = *(uint32 *)(code_p);
                                                last_b->native.end = cfg->env.global.native.code.end = code_p;
                                        }
                                        else{
#ifdef HAS_DELAYSLOT
                                                last_b->native.end = cfg->env.global.native.code.end -= 2 * WORD_SZ;
#else
                                                last_b->native.end = cfg->env.global.native.code.end -= WORD_SZ;
#endif
                                        }
                                        e->val.label->patch = PATCH_undef;
LOG(    				fprintf( cur_b->cfg->fd, "BLK_patch_branch(): B%d: DROP_BRANCH\n", last_b->bid ); );
                                        BLK_fix_catch( last_b );
                                }
                                else{
LOG(    				fprintf( cfg->fd, "BLK_patch_branch(): B%d: END\n", last_b->bid ); );
                                        SCH_end_block( last_b );
                                }
                        }
                        else{
                                e = last_b->exprs_tail;
                                if( !EXPR_IS_BRANCH( e ) ){
                                        new_e = EXPR_new();
                                        new_e->op = OP_branch;
                                        new_e->type = PRIM_void;
                                        new_e->val.label = LAB_new_block_branch( cfg, new_e, last_b->succ.blocks[0] );
                                        EXPR_LIST_APPEND( last_b, last_b->exprs_tail, new_e );
                                        EXPR_generate_code( new_e, last_b );
                                        cfg->env.global.native.code.end = last_b->native.end;
LOG(    				fprintf( cfg->fd, "BLK_patch_branch(): B%d: ADD_BRANCH1 \n", last_b->bid ); );
                                        BLK_fix_catch( last_b );
                                }
                                else{
LOG(    				fprintf( cfg->fd, "BLK_patch_branch(): B%d: END\n", last_b->bid ); );
                                        SCH_end_block( last_b );
                                }
                        }
                        break;
                case 2:
#ifdef OPTIMIZE_CMOV
                        if( EXPR_IS_CMOV( last_b->exprs_tail ) ){
                                break;
                        }
#endif
                        if( last_b->succ.blocks[1] == cur_b ){
                        }
                        else if( last_b->succ.blocks[0] == cur_b ){
                                if( !(last_b->flags & BLK_switched_branch) ){
                                        last_b->exprs_tail->val.label->to.block = last_b->succ.blocks[1];
#ifdef HAS_DELAYSLOT
                                        code_p = (uint8 *)(last_b->native.end - 2 * WORD_SZ);
#else
                                        code_p = (uint8 *)(last_b->native.end - WORD_SZ);
#endif
                                        ASM_switch_branch( code_p );
                                        last_b->flags |= BLK_switched_branch;
LOG(    				fprintf( cfg->fd, "BLK_patch_branch(): B%d: SWITCH_BRANCH \n", last_b->bid ); );
				}
                        }
                        else{
	                        new_e = EXPR_new();
        	                new_e->op = OP_branch;
                                new_e->type = PRIM_void;
                                tgt_b = last_b->flags & BLK_switched_branch ? last_b->succ.blocks[0]: last_b->succ.blocks[1];
                                new_e->val.label = LAB_new_block_branch( cfg, new_e, tgt_b );
                                EXPR_LIST_APPEND( last_b, last_b->exprs_tail, new_e );
                                EXPR_generate_code( new_e, last_b );
                                cfg->env.global.native.code.end = last_b->native.end;
LOG(    			fprintf( cfg->fd, "BLK_patch_branch(): B%d: ADD_BRANCH2 to B%d\n", last_b->bid, tgt_b->bid ); );
                                BLK_fix_catch( last_b );
                        }
                        break;
                }
        }
}

static void
BLK_push_block_env( instr_block * old_b, instr_block * new_b )
{
        cfg_info * cfg;
        block_env * b_env;
        
        cfg = old_b->cfg;
LOG(    fprintf( cfg->fd, "BLK_push_block_env()\n" ); );
        b_env = gc_malloc_fixed( sizeof(block_env) );
        bcopy( &cfg->env.block, b_env, sizeof(block_env) );
        b_env->block = old_b;
        cfg->env.block.block = new_b;
        cfg->env.block.stack = b_env;

        REG_push_block_env( cfg, b_env );
        TEMP_push_block_env( cfg, b_env );
        EXPR_push_block_env( cfg );
}

static void
BLK_pop_block_env( instr_block * blk, boolean do_reset, boolean is_specialized )
{
        cfg_info * cfg;
        block_env * b_env;
        
        cfg = blk->cfg;
LOG(    fprintf( cfg->fd, "BLK_pop_block_env()\n" ); );
        b_env = cfg->env.block.stack;

        EXPR_pop_block_env( cfg, cfg->env.block.exprs.backup, is_specialized );
        TEMP_pop_block_env( cfg, b_env );
        REG_pop_block_env( cfg, b_env );

        bcopy( b_env, &cfg->env.block, sizeof(block_env) );
        gc_free_fixed( b_env );

        if( do_reset ){
                TEMP_reset_context( blk );
                REG_reset_context( blk );
        }
}

static void
BLK_set_exception_handler( instr_block * blk, cfg_info * cfg )
{
        instr_block * catch_b;
        exception_info * handler_i;
        int n;

        for( n = blk->catch.length; --n >= 0; ){
                catch_b = blk->catch.blocks[n];
                if( catch_b->handler == NULL 
                   || catch_b->handler->native.end != blk->native.start ){
                        handler_i = catch_b->handler = gc_malloc_fixed( sizeof(exception_info) );
                        handler_i->native.start = blk->native.start;
                        handler_i->native.end = blk->native.end;
                        handler_i->handler = catch_b;
                        handler_i->catch = catch_b->handler_type;
                        handler_i->next = cfg->env.global.exception.handlers;
                        cfg->env.global.exception.handlers = handler_i;
                        cfg->env.global.exception.count++;
                }
                else{
                        catch_b->handler->native.end = blk->native.end;
                }
        }
}


static void
BLK_repoint( instr_block * blk, uint8 * old, uint8 * new )
{
        expr_frame * frame;
        expr * call_e, ** exprs_by_op;
        int sz, offset;
        instr_block * cur_b;

        offset = blk->native.start - old;
        sz = blk->native.end - blk->native.start;
        blk->native.start = blk->native.end = new + offset;
        blk->native.end += sz;
#ifdef OPTIMIZE_INLINE
        exprs_by_op = &blk->exprs_by_op[OP_call];
        for( call_e = *exprs_by_op; call_e != NULL; call_e = call_e->exprs_by_op ){
                frame = call_e->val.frame;
                if( frame->flags & CALL_inline ){
                        for( cur_b = frame->env->root; cur_b != NULL; cur_b = cur_b->block.next ){
                                BLK_repoint( cur_b, old, new );
                        }
                }
        }
#endif
}

static void
BLK_buffer( instr_block * blk, cfg_info * cfg )
{
        uint8 * buf;
        int sz;

LOG(    fprintf( cfg->fd, "BLK_buffer(): B%d: 0x%x - 0x%x\n", blk->bid, (uint)blk->native.start, (uint)blk->native.end ); );
        SCH_end_block( blk );
        sz = blk->native.end - blk->native.start;
        buf = gc_malloc_fixed( sz );
        bcopy( blk->native.start, buf, sz );
        cfg->env.global.native.code.end = blk->native.start;
LOG(    fprintf( cfg->fd, "BLK_buffer(): reset code_p=0x%x\n", (uint)blk->native.start ); );
        cfg->env.global.last_block = NULL;
        BLK_repoint( blk, blk->native.start, buf );
#if 0
        blk->native.start = buf;
        blk->native.end = buf + sz;
#endif
#ifdef OPTIMIZE_INLINE
        /* repoint inlined methods */
#endif
}

static void
BLK_generate_code( instr_block * ebb_root, instr_block * cur_b ) 
{
        expr * e, * end_e, ** locals, * new_e;
        expr tmp_e;
        int localsz, i, n;
        Method * method;
        cfg_info * cfg;
        instr_block * succ_b;
	block_list stack;
        loop_info * loop_i;
        boolean check_init;
#ifdef OPTIMIZE_INLINE
        expr_frame * frame;
        boolean is_call;
        reg_context * reg_c;
        temp_context * temp_c;
        expr * restore_e, * return_e, move_e;
#endif
#ifdef OPTIMIZE_GCSE
        int n_globals;
        expr ** globals;
#endif

        cfg = cur_b->cfg;
LOG(    fprintf( cfg->fd, "BLK_generate_code(): B%d.%d\n", cur_b->bid, cur_b->env->inline_depth ); );

#ifdef OPTIMIZE_INLINE
        reg_c = NULL;
        temp_c = NULL;
        restore_e = NULL;
        CFG_set_method_env( cfg, cur_b->env );
#endif
        method = cfg->env.method.method;
        localsz = cfg->env.method.n_locals;
#ifdef OPTIMIZE_GCSE
        n_globals = cfg->env.global.globals.length;
#endif

#ifdef OPTIMIZE_INLINE
        if( !cfg->env.method.inlined ){
#endif
                cfg->env.block.block = cur_b;
                BLK_patch_branch( cur_b, cfg );
#ifdef OPTIMIZE_INLINE
        }
#endif
        cur_b->native.start = cur_b->native.end = cfg->env.global.native.code.end;

        if( cur_b == ebb_root 
#ifdef OPTIMIZE_INLINE
           && !cfg->env.method.inlined
#endif
           ){
                REG_set_base_context( ebb_root );
                locals = cur_b->locals.in;
                for( i = localsz + cur_b->sp.in; --i >=0;  ){
                        locals[i]->flags |= SE_generated;
                }
#ifdef OPTIMIZE_GCSE
                if( cfg->env.method.root != cur_b ){
                        globals = cur_b->globals.in;
                        for( i = n_globals; --i >=0;  ){
                                globals[i]->flags |= SE_generated;
                        }
                }
#endif
        }

        cfg->env.global.last_call = NULL;

#ifdef OPTIMIZE_INLINE
        if( !cfg->env.method.inlined ){
#ifdef OPTIMIZE_SCHEDULE
                cfg->env.global.schedule.code.start = cur_b->native.start;
#endif
        }
#endif
        cur_b->in_init = true;
        check_init = !cfg->env.method.inlined && cur_b != cur_b->ebb_root;
        tmp_e.expr.next = (void*)cur_b->exprs_head;
        for( e = end_e = &tmp_e; e != NULL; ){
                if( e->expr.next == NULL || ( ((expr*)e->expr.next)->op != OP_branch_indirect
                   && e->expr.next != cur_b->exprs_tail ) ){
                }
                else{
                        if( EXPR_IS_BRANCH( (expr *)e->expr.next ) ){
                                break;
                        }
                        else if( cur_b->succ.length > 0 && cur_b->succ.blocks[0]->stage == STG_generated_code ){
                                new_e = EXPR_new();
                                new_e->op = OP_branch;
                                new_e->type = PRIM_void;
                                new_e->val.label = LAB_new_block_branch( cfg, new_e, cur_b->succ.blocks[0] );
                                EXPR_LIST_APPEND( cur_b, cur_b->exprs_tail, new_e );
                        }
                }
                e = e->expr.next;
                if( e == NULL ){
                        break;
                }
#if 0
                if( check_init && e->op != OP_copy ){
                        cur_b->in_init = false;
                        REG_backup_exprs( cfg );
                        check_init = false;
                }
#endif
#ifdef OPTIMIZE_INLINE
                frame = e->val.frame;
                is_call = e->op == OP_call ? true : false;
                if( is_call && frame->flags & CALL_specialized_inline ){
                        FAILED();
                        BLK_push_block_env( cur_b, 0 );
                        EXPR_decrement_call_frame( frame );
                }
                if( is_call && frame->flags & (CALL_inline | CALL_specialized_inline) ){
LOG( 	 		fprintf( cfg->fd, "BLK_generate_code(): inline [%d]\n", e->eid ); );
                        cfg->env.global.native.code.end = cur_b->native.end;
			LIST_zero( &stack );
		        BL_add( &stack, frame->env->root );
                        EBB_generate_code( &stack, cfg );
       			LIST_free( &stack );
                        cur_b->native.end = cfg->env.global.native.code.end;
LOG( 	 		fprintf( cfg->fd, "BLK_generate_code(): end inline [%d]\n", e->eid ); );
                        CFG_set_method_env( cfg, cur_b->env );
                }
                else{
#endif
                EXPR_generate_code( e, cur_b );
#ifdef OPTIMIZE_INLINE
                }
                if( is_call && frame->flags & (CALL_specialized_failed | CALL_specialized_inline) ){
                        if( frame->flags & CALL_specialized_inline ){
                                reg_c = REG_copy_context( &cfg->env.block.regs );
                                temp_c = TEMP_copy_context( &cfg->env.block.temps );
                                restore_e = cfg->env.block.exprs.backup;
                                BLK_pop_block_env( cur_b, true, true );
                                FAILED();
                                BLK_push_block_env( cur_b, 0 );
                                EXPR_increment_call_frame( frame );
                        }
                        if( frame->flags & CALL_specialized_failed ){
                                return_e = frame->env->return_val;
                                if( return_e != NULL ){
                                        e->uses.block++;
                                }

                                REG_merge_inline( reg_c, cur_b );
                                BLK_pop_block_env( cur_b, false, false );
                                EXPR_pop_block_env( cfg, restore_e, false );

                                gc_free_fixed( cfg->env.block.temps.table );
                                bcopy( temp_c, &cfg->env.block.temps, sizeof(temp_context) );
                                gc_free_fixed( temp_c );       

                                gc_free_fixed( cfg->env.block.regs.free );
                                bcopy( reg_c, &cfg->env.block.regs, sizeof(reg_context) );
                                gc_free_fixed( reg_c );

                                if( return_e != NULL && 
                                   ( return_e->uses.block > 0 || return_e->uses.other > 0 ) ){
                                        e->uses.block--;
                                        if( !( return_e->flags & SE_in_reg ) ){
                                                REG_reload_expr( return_e, cur_b );
                                        }
                                        if( !(e->flags & SE_in_reg) ){
                                                e->flags |= SE_preset_reg;
                                                e->loc.reg.base = return_e->loc.reg.base;
                                                e->loc.reg.pair = return_e->loc.reg.pair;
                                                REG_free_expr( return_e, cfg, true );
                                                REG_reload_expr( e, cur_b );
                                        }
                                        else if( return_e->loc.reg.base != e->loc.reg.base
                                                || return_e->loc.reg.pair != e->loc.reg.pair ){
                                                bzero( &move_e, sizeof(expr) );
                                                EXPR_make_move_to( &move_e, e, cur_b, 
                                                                  return_e->loc.reg.base, return_e->loc.reg.pair );
                                                REG_free_expr( return_e, cfg, true );
                                                ASM_move( &move_e, cur_b, false );
                                        }
                                        return_e->flags |= SE_in_reg | SE_generated;
                                        REG_assign_expr( return_e, cfg, REGOP_assign );
                                }
                        }
                }
#endif
                end_e = e;
        }

        /* Prepare locals for the loop */
        if( cur_b->flags & BLK_loop_header ){
LOG( 	 	fprintf( cfg->fd, "BLK_generate_code(): generate loop header\n" ); );
#if 0
#ifdef OPTIMIZE_GCSE
                globals = cur_b->globals.out;
                for( i = n_globals; --i >=0;  ){
                        e = globals[i];
                        if( e->flags & SE_in_reg && !(e->flags & SE_in_frame_slot) ){ 
                                REG_spill_expr( e, cur_b, false );
                        }
                }
#endif
#endif
                locals = cur_b->locals.out;
                loop_i = cur_b->succ.blocks[0]->loop;
                for( i = localsz + cur_b->sp.in; --i >=0;  ){
                        e = locals[i];
                        if( STATS_local_get( SO_load, loop_i->locals, i ) || STATS_local_get( SO_store, loop_i->locals, i ) ){
                                if( (e->uses.block > 0 || e->uses.other > 0) ){
LOG( 	 		               fprintf( cfg->fd, "BLK_generate_code(): load L%d\n", i ); );
                                       e->uses.block++;
                                       ASM_load_expr( cur_b, e );
                                       e->uses.block--;
                                }
                        }
                        else{
                                if( (e->uses.block > 0 || e->uses.other > 0) ){
                                        if( e->flags & SE_in_reg ){
LOG( 	 		                        fprintf( cfg->fd, "BLK_generate_code(): free L%d\n", i ); );
                                                if( !(e->flags & SE_in_frame_slot) ){ 
                                                        if( REG_spill_expr( e, cur_b, false ) ){
                                                                REG_free_expr( e, cfg, true );
                                                        }
                                                }
                                        }
                                }
                        }
                }
        }

        /* Generate values on the stack */
        if( cur_b->flags & BLK_succ_not_in_ebb ){
                for( i = cur_b->sp.out; --i >= 0; ){
                        e = cur_b->stack.out[i];
                        if( e->op != OP_nop || e->flags & SE_slot ){
LOG( 	 			fprintf( cfg->fd, "BLK_generate_code(): Load stack value S%d\n", i ); );
                                e->uses.block++; /* ERROR: Shouldn't need this, but do */
LOG(   				fprintf( cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( cur_b, e ), e->uses.block ); );
                                ASM_load_expr( cur_b, e );
                        }
                        if( e->type == PRIM_double || e->type == PRIM_long ){
                                i--;
                        }
                }
        }

        /* Set slot status of successors */
        if( cur_b->succ.length > 0 ){
                EBB_link_slots( cur_b );
                /* Link blocks within EBB */
        }
        /* FIXSLOT: Check block_out_info for required reg locations */

        e = end_e->expr.next;
        while( e != NULL ){
                EXPR_generate_code( e, cur_b );
                e = e->expr.next;
        }

        if( cur_b->flags & BLK_buffer_code ){
                BLK_buffer( cur_b, cfg );
        }
        else{
                cfg->env.global.native.code.end = cur_b->native.end;
                cfg->env.global.last_block = (cur_b->succ.length > 0 && cur_b->succ.length < 3) ? cur_b : NULL;
                BLK_set_exception_handler( cur_b, cfg );
        }

        for( n = cur_b->succ.length; --n >= 0; ){
                succ_b = cur_b->succ.blocks[n];
                if( succ_b->ebb_root == ebb_root 
                   && succ_b->stage != STG_generated_code ){
                        succ_b->stage = STG_generated_code;
                        BLK_push_block_env( cur_b, succ_b );
                        BLK_generate_code( ebb_root, succ_b );
                        BLK_pop_block_env( cur_b, true, false );
                }
        }
}

static void
EBB_set_buffer( instr_block * ebb_root )
{
        int n;
        instr_block * blk;

        for( n = ebb_root->ebb.length; --n >= 0; ){
                blk = ebb_root->ebb.blocks[n];
                if( blk->bid >= ebb_root->bid ){
                        blk->flags |= BLK_buffer_code;
                }
        }
}

static void
BLK_unbuffer( instr_block * blk, instr_block * ebb_root, cfg_info * cfg )
{
        int n, sz;
        instr_block * succ_b;

        if( blk->flags & BLK_buffer_code ){
                blk->flags &= ~BLK_buffer_code;
                BLK_patch_branch( blk, cfg );
                /* FIXME: assertion against code max */
                sz = blk->native.end - blk->native.start;
                bcopy( blk->native.start, cfg->env.global.native.code.end, sz );
                gc_free_fixed( blk->native.start );
                BLK_repoint( blk, blk->native.start, cfg->env.global.native.code.end );
                cfg->env.global.native.code.end += sz;
LOG(            fprintf( cfg->fd, "BLK_unbuffer(): B%d @ 0x%x - 0x%x\n", blk->bid, (uint)blk->native.start, (uint)blk->native.end ); );
LOG(            fflush( cfg->fd ); );
                cfg->env.global.last_block = (blk->succ.length > 0 && blk->succ.length < 3) ? blk : NULL;
                BLK_set_exception_handler( blk, cfg );
                SCH_start_block( blk );
        }
        for( n = blk->succ.length; --n >= 0; ){
                succ_b = blk->succ.blocks[n];
                if( succ_b->ebb_root == ebb_root 
                    && succ_b != ebb_root ){
                        BLK_unbuffer( succ_b, ebb_root, cfg );
                }
        }
}

#if 0
static void
EBB_generate_code( block_list * stack, cfg_info * cfg ) 
{
        int n, ix;
        boolean generated_pred;
        instr_block * cur_b, * blk;
        block_list * bl;

        blk = NULL;
        while( (ix = stack->length) > 0 ){
                generated_pred = false;
                while( !generated_pred && --ix >= 0 ){
                        blk = stack->blocks[ix];
                        generated_pred = true;
                        bl = blk->flags & BLK_catch ? &blk->try : &blk->pred;
                        for( n = bl->length; --n >= 0; ){
                                cur_b = bl->blocks[n];
                                if( cur_b->stage == STG_generated_code ){
                                        if( blk->flags & BLK_catch
                                           || ( cur_b->flags & BLK_loop_header && !(blk->flags & BLK_catch) ) ){
                                                generated_pred = true;
                                                break;
                                        }
                                }
                                else{
                                        generated_pred = false;
                                }
                        }
                }
                if( ix < 0 ){
                        break;
                }
                BL_del( stack, blk );
                if( blk->stage != STG_generated_code ){
                        blk->stage = STG_generated_code;
                        BLK_generate_code( blk, blk );
                }
                for( n = blk->ebb_succ.length; --n >= 0; ){
                        cur_b = blk->ebb_succ.blocks[n];
                        if( cur_b->stage != STG_generated_code ){
                                BL_add( stack, cur_b );
                        }
                }
        }
}
#else
static void
EBB_generate_code( block_list * stack, cfg_info * cfg ) 
{
        int n, ix;
        boolean generated_pred;
        instr_block * cur_b, * blk;
        block_list * bl;
        loop_info * l;

        blk = NULL;
        while( (ix = stack->length) > 0 ){
                generated_pred = false;
                while( !generated_pred && --ix >= 0 ){
                        blk = stack->blocks[ix];
                        bl = blk->flags & BLK_catch ? &blk->try : &blk->pred;
                        if( blk->flags & BLK_catch ){
                                generated_pred = false;
                                for( n = bl->length; --n >= 0; ){
                                        cur_b = bl->blocks[n];
                                        if( cur_b->stage == STG_generated_code ){
                                                generated_pred = true;
                                                break;
                                        }
                                }
                        }
                        else{
                                generated_pred = true;
                                for( n = bl->length; --n >= 0; ){
                                        cur_b = bl->blocks[n];
                                        if( cur_b->stage != STG_generated_code ){
                                                generated_pred = false;
                                                break;
                                        }
                                }
                        }
                        if( blk->flags & BLK_loop_entry ){
                                l = blk->loop;
                                if( !generated_pred ){
                                        if( blk->stage != STG_generated_code ){
                                                blk->stage = STG_generated_code;
TRACE(                                          kprintf( "code: B%d\n", blk->bid ); );

                                                if( !(blk->loop->flags & LOOP_do_while) ){
                                                        EBB_set_buffer( blk );
                                                }
                                                BLK_generate_code( blk, blk );
                                                for( n = blk->ebb_succ.length; --n >= 0; ){
                                                        cur_b = blk->ebb_succ.blocks[n];
                                                        if( cur_b->stage != STG_generated_code 
                                                            && BL_contains( &l->blocks, cur_b ) ){
                                                                BL_add( stack, cur_b );
TRACE( kprintf( "add: B%d\n", cur_b->bid ); );
                                                        }
                                                }
                                                ix = stack->length;
                                        }
                                }
                                else{
                                        BLK_unbuffer( blk, blk->ebb_root, cfg );
                                        BL_del( stack, blk );
                                        for( n = blk->ebb_succ.length; --n >= 0; ){
                                                cur_b = blk->ebb_succ.blocks[n];
                                                if( cur_b->stage != STG_generated_code 
                                                    && !BL_contains( &l->blocks, cur_b ) ){
                                                        BL_add( stack, cur_b );
TRACE( kprintf( "add: B%d\n", cur_b->bid ); );
                                                }
                                        }
                                        ix = stack->length;
                                }
                                generated_pred = false;
                                continue;
                        }
                }
                if( ix < 0 ){
                        break;
                }
                BL_del( stack, blk );
                if( blk->stage != STG_generated_code ){
                        blk->stage = STG_generated_code;
TRACE( kprintf( "code: B%d\n", blk->bid ); );
                        BLK_generate_code( blk, blk );
                }
                for( n = blk->ebb_succ.length; --n >= 0; ){
                        cur_b = blk->ebb_succ.blocks[n];
                        if( cur_b->stage != STG_generated_code ){
                                BL_add( stack, cur_b );
TRACE( kprintf( "add: B%d\n", cur_b->bid ); );
                        }
                }
        }
}
#endif

void
BLK_backup_expr( expr * e, block_env * b_env, FILE * fd )
{
        expr * t_e;

LOG(    fprintf( fd, "REG_spill_expr(): backup [%d]\n", e->eid ); );
        t_e = EXPR_new();
        bcopy( e, t_e, sizeof(expr) );
        t_e->expr.prev = e;
        t_e->expr.next = b_env->exprs.backup;
        b_env->exprs.backup = t_e;
}

inline void
EXPR_push_block_env( cfg_info * cfg )
{
        cfg->env.block.exprs.backup = NULL;
}
 
void
EXPR_pop_block_env( cfg_info * cfg, expr * cur_e, boolean do_swap )
{
        expr * tmp_e, * s_e, * next_e, * prev_e;
        expr local_e;

        while( cur_e != NULL ){
                tmp_e = cur_e->expr.next;
                s_e = cur_e->expr.prev;
                next_e = s_e->expr.next;
                prev_e = s_e->expr.prev;
                if( do_swap ){
                        bcopy( s_e, &local_e, sizeof(expr) );
                }
                bcopy( cur_e, s_e, sizeof(expr) );
                s_e->expr.next = next_e;
                s_e->expr.prev = prev_e;
                if( do_swap ){
LOG(                    fprintf( cfg->fd, "EXPR_pop_block_env(): swap [%d]\n", cur_e->eid ); );
                        bcopy( &local_e, cur_e, sizeof(expr) );
                        cur_e->expr.next = tmp_e;
                        cur_e->expr.prev = s_e;
                }
                else{
LOG(                    fprintf( cfg->fd, "EXPR_pop_block_env(): restore [%d]\n", cur_e->eid ); );
                        EXPR_free( cur_e );
                }
                cur_e = tmp_e;
        }
} 

static void
DFG_set_exception_handlers( cfg_info * cfg )
{
        exception_info * handler_i, * next_i;
	jexceptionEntry * e;
        Method * method;

LOG(    fprintf( cfg->fd, "DFG_exception_handlers():\n" ); );
        method = cfg->env.method.method;
        e = gc_malloc_fixed_typed( ( cfg->env.global.exception.count + 1 ) * sizeof(jexceptionEntry), GC_ALLOC_EXCEPTIONTABLE );
        gc_free_fixed( method->exception_table );
        method->exception_table = e;
        handler_i = cfg->env.global.exception.handlers;
        while( handler_i != NULL ){
                next_i = handler_i->next;
                if( handler_i->native.start != handler_i->native.end ){
LOG(                    fprintf( cfg->fd, "\t0x%x - 0x%x -> 0x%x (B%d)\n", 
                                 handler_i->native.start - cfg->env.global.pbase,
                                 handler_i->native.end - cfg->env.global.pbase,
                                 handler_i->handler->native.start - cfg->env.global.pbase,
                                 handler_i->handler->bid ); );
                        e->start_pc = (uintp)handler_i->native.start;
                        e->end_pc = (uintp)handler_i->native.end;
                        e->handler_pc = (uintp)handler_i->handler->native.start;
                        e->catch = handler_i->catch;
                }
                else{
                        e->start_pc = e->end_pc = 0;
                }
                gc_free_fixed( handler_i );
                handler_i = next_i;
                e++;
        }
        e->start_pc = -1;
}

void
DFG_install_code( cfg_info * cfg )
{
	uint32 i;
        Method * meth;
        instr_block * blk;

        meth = cfg->env.method.method;
LOG(    fprintf( cfg->fd, "DFG_install_code(): 0x%p\n", cfg->env.global.native.code.start ); );

	SET_METHOD_NATIVECODE( meth, cfg->env.global.native.code.start, cfg->env.global.native.code.end );

#ifdef FLUSH_DCACHE
	FLUSH_DCACHE( cfg->env.global.native.consttable.start, cfg->env.global.native.code.end );
#endif
	if (meth->exception_table != 0) {
                /* FIXME: Add handlers for exception/jsr blocks since they have been moved */
                DFG_set_exception_handlers( cfg );
	}

#ifndef OPTIMIZE_MEMORY
	/* Translate line numbers table */
	/* FIXME: Not sure this is accurate */
	if (meth->lines != 0) {
                uint32 last = (uint32)cfg->env.global.native.code.start;
		for (i = 0; i < meth->lines->length; i++) {
                        blk = CFG_get_block( cfg, meth->lines->entry[i].pc_offset, true );
			meth->lines->entry[i].pc_offset = (uintp)blk->native.start - last;
                        last = (uintp)blk->native.start;
		}
	}
#endif

	/* Make method active (so we can find if for exceptions) */
	if( makeMethodActive( meth, cfg->env.method.class, 
                             cfg->env.global.native.code.start, cfg->env.global.native.code.end ) == 0) {
                FAILED();
	}
}

int enable_check = 0;

/*
 * Convert bytecodes into an expression tree.
 */
void
DFG_build( cfg_info * cfg ) 
{
        Method * method;
        Hjava_lang_Class * class;
        instr_block * blk;
        int alloc_sz, i;
        block_list stack;

        method = cfg->env.method.method;
        class = cfg->env.method.class;

LOG(    fprintf( cfg->fd, "DFG_build(): %s.%s %s:\n", class->name, method->name, method->signature ); );

        if( cfg->finally.length > 0 ){
                CFG_fixup_ret( cfg );
        }
        EBB_build( cfg->env.method.root );
        for( i = cfg->catch.length ; --i >=0; ){
                EBB_build( cfg->catch.blocks[i] );
        }

        /* Initialize */
        LAB_init( cfg );
#ifdef OPTIMIZE_GCSE
        DFG_expand_globals( cfg, 1 );
#endif
        LIST_zero( &stack );
        BL_add( &stack, cfg->env.method.root );
        EBB_generate_exprs( &stack );
        if( cfg->catch.length > 0 || cfg->finally.length > 0 ){
                for( i = cfg->catch.length; --i >=0; ){
                        BL_add( &stack, cfg->catch.blocks[i] );
                }
                for( i = cfg->finally.length; --i >=0; ){
                        BL_add( &stack, cfg->finally.blocks[i] );
                }
                EBB_generate_exprs( &stack );
        }
        if( stack.length > 0 ){
                kprintf( "REMAINING: " );
                for( i = stack.length; --i >=0; ){
                        kprintf( "B%d ", stack.blocks[i]->bid );
                }
                kprintf( "\n" );
        }
        assert( stack.length == 0 );
        LIST_free( &stack );

        blk = cfg->env.method.root;
LOG(    fprintf( blk->cfg->fd, "BLK_link_uses():\n" ); );
        BLK_link( blk, cfg->env.method.n_locals, false );
        if( cfg->env.method.has_loops ){
LOG(            fprintf( blk->cfg->fd, "BLK_link_loop_uses(): %d iterations\n", cfg->env.global.max_loop_depth ); );
#if 1
                for( i = cfg->env.global.max_loop_depth; --i >= 0; ){
                        BLK_link( blk, cfg->env.method.n_locals, true );
                        DFG_clear_link_flag( cfg );
                }
#else
//                BLK_link( blk, cfg->env.method.n_locals, true );

                while( !BLK_link( blk, cfg->env.method.n_locals, true ) ){
                        DFG_clear_link_flag( cfg );
                }
#endif

        }
        for( i = 0; i < cfg->catch.length; i++ ){
                blk = cfg->catch.blocks[i];
                BLK_link( blk, blk->env->n_locals, false );
                /* FIXME: link loops here? */
        }
        COUNTER( cfg->env.global.time.expr += CNTR_end(); )
#ifdef OPTIMIZE_INLINE
        CFG_set_method_env( cfg, cfg->env.global.root_env );
#endif
#ifdef OPTIMIZE_GCSE
        GLOBAL_kill_unused( cfg );
#endif
LOG(    fflush( cfg->fd ); );
LOG(    CFG_print( cfg ); );

        COUNTER( CNTR_begin(); )
        /* Guess code size using expr IDs */
        DWORD_ALIGN( cfg->env.global.constants.max, 2 );
        cfg->env.global.native.estimate += cfg->env.global.id + cfg->env.global.constants.max;
        alloc_sz = WORD_SZ * ( 4 * cfg->env.global.native.estimate );
        cfg->env.global.native.consttable.start = gc_malloc_fixed( alloc_sz );
        cfg->env.global.native.consttable.end = cfg->env.global.native.consttable.start + WORD_SZ * cfg->env.global.constants.max;
        cfg->env.global.native.code.start = cfg->env.global.native.code.end = cfg->env.global.native.consttable.end;
        cfg->env.global.native.max = cfg->env.global.native.consttable.start + alloc_sz;
LOG(    fprintf( cfg->fd, "code size: %db (0x%x) @ %p - %p\n", alloc_sz, alloc_sz, cfg->env.global.native.code.start, cfg->env.global.native.max ); );


#ifdef TGT_SPARC
        REG_init( cfg, cfg->flags & CFG_is_leaf ? leaf_regfile : base_regfile );
#else
        REG_init( cfg, base_regfile );
#endif

#ifdef OPTIMIZE_SCHEDULE
        SCH_reset( cfg );
#endif
        DFG_build_frame_positions( cfg );
        LIST_zero( &stack );
        BL_add( &stack, cfg->env.method.root );
        EBB_generate_code( &stack, cfg );
        if( cfg->catch.length > 0 || cfg->finally.length > 0 ){
                for( i = cfg->catch.length; --i >=0; ){
                        BL_add( &stack, cfg->catch.blocks[i] );
                }
                for( i = cfg->finally.length; --i >=0; ){
                        BL_add( &stack, cfg->finally.blocks[i] );
                }
                EBB_generate_code( &stack, cfg );
        }
        assert( stack.length == 0 );
        LIST_free( &stack );

        BLK_patch_branch( NULL, cfg );
        TRW_generate_all( cfg );
        LAB_patch_all( cfg );
        CONST_generate_all( cfg );

#ifdef OPTIMIZE_INLINE /* if we can inline methods with handlers */
        CFG_set_method_env( cfg, cfg->env.global.root_env );
#endif
#ifdef INSTALL_CODE
        DFG_install_code( cfg );
#endif
LOG(    DISASM_method( cfg ); );
LOG(    fflush( cfg->fd ); );

        COUNTER( cfg->env.global.time.code = CNTR_end(); )
}
