/*
 * expr.h
 * Structures for instruction-level sequencing
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __expr_h
#define __expr_h

typedef struct _expr_context {
        struct _expr *	        expr;
	struct _expr_context *	next;
} expr_context;

typedef struct _uses_info {
	int16 			block;
	uint16 			other;
	int16 			first;
	int16 			last;
} uses_info;

typedef struct _location_info {
	union {
		int16		frame;
		int16		arg;
	} slot;
	struct {
		uint8		base;
		uint8		pair;
	} reg;
	struct {
		uint8		base;
		uint8		pair;
	} start;
} location_info;

typedef union _constant_info {
	jint			i_const;
	jlong			l_const;
	jfloat			f_const;
	jdouble			d_const;
	void *			ref;
	Method *		method;
	struct _label_info *	label;
	struct _expr_frame *	frame;
	struct _expr *		expr;
        struct {
		int32		offset;
		Hjava_lang_Class * class;
        } field;
	uint			c_flags;
} constant_info;

#define FCMP_l			0x1
#define FCMP_g			0x2

typedef struct _slot {
	/* common fields between expr & slot */
	uint8			op;
	uint8			type;
	int16			eid;	
	uint32			flags;
	uses_info		uses;
	location_info		loc;
	struct _expr*		a[2];
	Hjava_lang_Class * 	class;
} slot;

typedef struct _expr {
	/* common fields between expr & slot */
	uint8			op;
	uint8			type;
	int16			eid;	
	uint32			flags;
	uses_info		uses;
	location_info		loc;
	struct _expr*		a[2];
	Hjava_lang_Class * 	class;

	/* unique fields */
	struct _expr *		exprs_by_op;
	linked_list		expr;

	constant_info		val;
	struct _expr*		last_local;
} expr;

#define CALL_static			0x0001
#define CALL_no_side_effect		0x0002

#define CALL_vm				0x0004
#define CALL_throw			0x0008
#define CALL_nested			0x0010
#define CALL_jni			0x0200

#define CALL_inline			0x0040
#define CALL_specialized_inline		0x0080
#define CALL_specialized_failed		0x0100

typedef struct _arg_info {
	expr *			expr;
	uint32			flags;
	location_info		loc;
} arg_info;

typedef struct _expr_frame {
	struct _method_env *	env;
        Method *		method;
        Hjava_lang_Class *	class;
	Hjava_lang_Class *	ret;
	uint16			flags;
	uint16			sz;
	arg_info		a[1];
} expr_frame;

/* Definitions for expr->flags */

/* Slot role */
#define SE_slot				0x0001 /* Is local/stack position */
#define SE_global			0x0002

#define SE_generated			0x0004
#define SE_sync				0x0008

/* Location of value */
#define SE_in_reg			0x0010
#define SE_in_frame_slot		0x0020
#define SE_in_arg_slot			0x0040

#define SE_reg_assigned			0x0080
#define SE_preset_reg			0x0100 /* Short-lived */
#define SE_set_out_slot			0x0100 /* exists only in join_info */

#define SE_no_spill			0x0200
#define SE_invert_cmp			0x0400
#define SE_will_assign			0x0800 /* Short-lived */
#define SE_out_slot_reg			0x1000
#define SE_in_slot_reg			0x2000

#define SE_use_spans_call		0x4000
#define SE_last_local			0x8000 /* flag needed since slot struct doesn't have this field */

#define SE_unsigned			0x10000

/* Set if value is an object */
#define SE_object_field			0x20000
#define SE_static_field			0x40000
#define SE_array_object			0x80000

#define SE_no_generate			0x100000
#define SE_null_checked			0x200000

#define SE_inline_return		0x800000

#define SE_req_fp_reg			0x2000000

#define MAX_OPS				60

/* 
 * Modifications to these codes require updating:
 *   expr.c: OP_type
 *   asm-xxxx.def: ASM_funcs, PREP_funcs
 *   reduce.c: CP_funcs, PP_funcs, CSE_funcs, LOOP_funcs
 *   print.c: op_name, print_funcs
 */
#define OP_nop				0x0
#define OP_const_m1			0x01
#define OP_const_0			0x02
#define OP_const_1			0x03
#define OP_const			0x04
#define OP_move				0x05
#define OP_load				0x06
#define OP_store			0x07
#define OP_add				0x08
#define OP_sub				0x09
#define OP_mul				0x0a
#define OP_div				0x0b
#define OP_rem				0x0c
#define OP_neg				0x0d
#define OP_lsl				0x0e
#define OP_lsr				0x0f
#define OP_asr				0x10
#define OP_and				0x11
#define OP_or				0x12
#define OP_xor				0x13
#define OP_cvt				0x14
#define OP_set_cc			0x15
#define OP_cbranch_eq			0x16
#define OP_cbranch_ne			0x17
#define OP_cbranch_lt			0x18
#define OP_cbranch_ge			0x19
#define OP_cbranch_gt			0x1a
#define OP_cbranch_le			0x1b
#define OP_cbranch_ult			0x1c
#define OP_branch			0x1d
#define OP_branch_indirect		0x1e
#define OP_returnarg			0x1f
#define OP_cmp				0x20
#define OP_call				0x21
#define OP_target			0x22
#define OP_prologue			0x23
#define OP_epilogue			0x24
#define OP_eprologue			0x25
#define OP_copy				0x26
#define OP_inline_data			0x27
#define OP_get_arg			0x28
#define OP_preload			0x29 /* ?? */
#define OP_replace			0x2a
#define OP_specialize_sync		0x2b
#define OP_tableswitch			0x2c
#define OP_lookupswitch			0x2d
#define OP_native_instr_1a		0x2e
#define OP_native_instr_2a		0x2f
#define OP_cmov_eq			0x30
#define OP_cmov_ne			0x31
#define OP_cmov_lt			0x32
#define OP_cmov_ge			0x33
#define OP_cmov_gt			0x34
#define OP_cmov_le			0x35

/* Quick argument lookup for opcodes */
#define OP_1a		0x1
#define OP_2a		0x2
#define OP_frame	0x4

/*
 * These are carefully encoded to provide fast lookup of characteristics
 * (Modify at your own risk)
 */
#define PRIM_SIZE( _t )			( _t & PRIM_size )
#define PRIM_TYPE( _t )			( ( _t & PRIM_type ) >> 4 )
#define PRIM_FP( _t )			( _t & 0x40 )
#define PRIM_size			0x0f
#define PRIM_type			0x70
#define PRIM_vm_int			0x1
#define PRIM_vm_long			0x2
#define PRIM_vm_label			0x3
#define PRIM_vm_float			0x4
#define PRIM_vm_double			0x5

#define PRIM_void			0x00
#define PRIM_byte			0x11
#define PRIM_char			0x12
#define PRIM_short			0x92
#define PRIM_int			0x14
#define PRIM_ref			0x94
#define PRIM_long			0x28
#define PRIM_label			0x34 /* internal use only */
#define PRIM_float			0x44
#define PRIM_double			0x58

#define update_use_detail( _new, _d ) \
	if( (_new) == (_d) ){ \
		EXPR_update_use( _new, blk ); \
	}

#define check_div(_instr, _ref, _op)
#define check_div_long(_instr, _ref, _op)

#define SET_VALUE( _f, _v ) if( (void*)(_v) != NULL ){ (_f) = (_v); }

/* Constants */
#define move_const_b(_v, _zero, _one, _minus_one) \
		new_e = EXPR_new(); \
		new_e->op = ((_v) == (_zero)) ? OP_const_0 : ((_v) == (_one)) ? OP_const_1 : ((_v) == (_minus_one)) ? OP_const_m1 : OP_const
#define move_const_e(_d)		EXPR_add(new_e, &(_d), blk, ebb_root)
#define move_ref_const(_d, _v)	new_e = EXPR_new(); \
                new_e->op = ((void *)(_v) == (void *)0) ? OP_const_0 : OP_const; \
		new_e->type = PRIM_ref; new_e->val.ref = (void*)(_v); move_const_e(_d)
#define move_int_const(_d, _v)	move_const_b(_v, 0, 1, -1); \
		new_e->type = PRIM_int; new_e->val.i_const = (_v); move_const_e(_d)
#define move_long_const(_d, _v)	move_const_b(_v, 0L, 1L, -1L); \
		new_e->type = PRIM_long; new_e->val.l_const = (_v); move_const_e(_d)
#define move_float_const(_d, _v)	move_const_b(_v, 0.0F, 1.0F, -1.0F); \
		new_e->type = PRIM_float; new_e->val.f_const = (_v); move_const_e(_d)
#define move_double_const(_d, _v)	move_const_b(_v, 0.0, 1.0, -1.0); \
		new_e->type = PRIM_double; new_e->val.d_const = (_v); move_const_e(_d)
#define move_string_const(_d, _v)	move_ref_const(_d, _v)
#define move_label_const(_d, _l)	new_e = EXPR_new(); \
                new_e->op = ((void *)(_l) == (void *)0) ? OP_const_0 : OP_const; \
		new_e->type = PRIM_label; new_e->val.label = (void*)(_l); move_const_e(_d)

#define swap_any(_d, _s)		t_e = (_d); (_d) = (_s); (_s) = t_e

#define preload( _d, _s ) \
		new_e = EXPR_new(); \
		new_e->op = OP_preload; \
		new_e->type = (_s)->type; \
		new_e->a[0] = (_s); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )

#define specialize_sync() \
		new_e = EXPR_new(); \
		new_e->op = OP_specialize_sync; \
		new_e->val.frame = frame; \
		EXPR_add(new_e, 0, blk, ebb_root)

#define tableswitch( _d, _s, _high, _low ) \
		move_int_const(t_e, _high ); \
		new_e = EXPR_new(); \
		new_e->op = OP_tableswitch; \
		new_e->a[0] = (_s); \
		new_e->type = (_s)->type; \
		new_e->a[1] = t_e; \
		new_e->val.i_const = (_low); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )

#define lookupswitch( _d, _s, _table, _idx ) \
		new_e = EXPR_new(); \
		new_e->op = OP_lookupswitch; \
		new_e->a[0] = (_s); \
		new_e->type = (_s)->type; \
		new_e->a[1] = (_table); \
		new_e->val.i_const = (_idx); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )

/* Moves */
#define move_any(_d, _s)		(_d) = (_s)
#define move_ref(_d, _s)		move_any(_d, _s)
#define move_int(_d, _s)		move_any(_d, _s)
#define move_long(_d, _s)		move_any(_d, _s)
#define move_float(_d, _s)		move_any(_d, _s)
#define move_double(_d, _s)		move_any(_d, _s)

/* Loads */
#ifdef  USE_RISC_LDST 
#define load(_d, _base, _c_offset, _type, _flags, _field, _class) \
		new_e = EXPR_new(); \
		new_e->op = OP_load; \
		SET_VALUE( new_e->flags, _flags ); \
		new_e->a[0] = (_base); \
		new_e->val.i_const = (_c_offset); \
		new_e->type = (_type); \
		SET_VALUE( new_e->a[1], (void*)_field ); \
		SET_VALUE( new_e->class, _class ); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#else
#define load(_d, _addr, _type, _flags, _field, _class) \
		new_e = EXPR_new(); \
		new_e->op = OP_load; \
		SET_VALUE( new_e->flags, _flags ); \
		new_e->a[0] = (_addr); \
		new_e->type = (_type); \
		SET_VALUE( new_e->a[1], (void*)_field ); \
		SET_VALUE( new_e->class, _class ); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#endif 

#ifdef  USE_RISC_LDST 
#define load_code_ref(_d, _s)	load(_d, _s, 0, PRIM_ref, 0, NULL, NULL)
#define load_key(_d, _s)		load(_d, _s, 0, PRIM_ref, 0, NULL, NULL)
#define load_ref(_d, _s)		load(_d, _s, 0, PRIM_ref, 0, NULL, NULL)
#else
#define load_code_ref(_d, _s)	load(_d, _s, PRIM_ref, 0, NULL, NULL)
#define load_key(_d, _s)		load(_d, _s, PRIM_ref, 0, NULL, NULL)
#define load_ref(_d, _s)		load(_d, _s, PRIM_ref, 0, NULL, NULL)
#endif

#ifdef  USE_RISC_LDST 
#if 0
#define load_indexed(_d, _base, _offset, _type, _flags, _field, _class) \
		new_e = EXPR_new(); \
		new_e->op = OP_load; \
		SET_VALUE( new_e->flags, _flags | SE_ldst_indexed ); \
		new_e->a[0] = (_base); \
		new_e->val.expr = (_offset); \
		new_e->type = (_type); \
		SET_VALUE( new_e->a[1], (void*)_field ); \
		SET_VALUE( new_e->class, _class ); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define load_offset_scaled(_d, _base, _ix, _c_offset, _type, _class) \
		move_int_const(t2_e, _c_offset ); \
		add_ref(t2_e, _base, t2_e); \
		t2_e->class = (_base)->class; \
		move_int_const(t3_e, size2shift[PRIM_SIZE( _type )] ); \
		lshl_int(t3_e, _ix, t3_e); \
		load_indexed(_d, t2_e, t3_e, _type, SE_array_object, (_base)->class, _class)
#else
#define load_offset_scaled(_d, _base, _ix, _c_offset, _type, _class) \
		t_e = (_base); \
		move_int_const(t2_e, size2shift[PRIM_SIZE( _type )] ); \
		lshl_int(t2_e, _ix, t2_e); \
		add_ref(t2_e, _base, t2_e); \
		t2_e->class = (_base)->class; \
		load(_d, t2_e, _c_offset, _type, SE_array_object, (_base)->class, _class); \
		softcall_profile_arrayload( _d, t_e, _ix )
#endif
#else
#define load_offset_scaled(_d, _base, _ix, _c_offset, _type, _class) \
		move_int_const(t_e, _c_offset); \
		add_ref(t_e, t_e, _base); \
		move_int_const(t2_e, size2shift[PRIM_SIZE( _type )] ); \
		lshl_int(t2_e, _ix, t2_e); \
		add_ref(t_e, t_e, t2_e); \
		load(_d, t_e, _type, SE_array_object, (_base)->class, _class)
#endif
#define load_offset_scaled_ref(_d, _base, _ix, _c_offset) \
		if( (_base)->class != NULL ){ \
			load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_ref, getClassFromSignature(&(_base)->class->name[1], class->loader ) ); \
		} \
		else{ \
			load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_ref, NULL); \
		}
#define load_offset_scaled_int(_d, _base, _ix, _c_offset)	load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_int, NULL)
#define load_offset_scaled_long(_d, _base, _ix, _c_offset)	load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_long, NULL)
#define load_offset_scaled_float(_d, _base, _ix, _c_offset)	load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_float, NULL)
#define load_offset_scaled_double(_d, _base, _ix, _c_offset)	load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_double, NULL)
#define load_offset_scaled_byte(_d, _base, _ix, _c_offset)	load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_byte, NULL)
#define load_offset_scaled_char(_d, _base, _ix, _c_offset)	load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_char, NULL)
#define load_offset_scaled_short(_d, _base, _ix, _c_offset)	load_offset_scaled(_d, _base, _ix, _c_offset, PRIM_short, NULL)

#ifdef  USE_RISC_LDST 
#define load_addr(_d, _c_base, _c_offset, _type, _class) \
		move_ref_const(t_e, _c_base); \
		EXPR_reduce_class( t_e, field_class() ); \
		load(_d, t_e, _c_offset, _type, SE_static_field, field_field(), _class); \
		softcall_profile_get( _d, t_e )
#else
#define load_addr(_d, _c_base, _c_offset, _type, _class) \
		move_ref_const(t_e, _c_base); \
		EXPR_reduce_class( t_e, field_class() ); \
		move_ref_const(t2_e, _c_offset); \
		add_ref(t_e, t_e, t2_e); \
		load(_d, t_e, _type, SE_static_field, field_field(), _class)
#endif
#define load_addr_ref(_d, _c_base, _c_offset)		load_addr(_d, _c_base, _c_offset, PRIM_ref, getClassFromSignature( field_field()->ftype, class->loader ) )
#define load_addr_int(_d, _c_base, _c_offset)		load_addr(_d, _c_base, _c_offset, PRIM_int, NULL)
#define load_addr_long(_d, _c_base, _c_offset)	load_addr(_d, _c_base, _c_offset, PRIM_long, NULL)
#define load_addr_float(_d, _c_base, _c_offset)	load_addr(_d, _c_base, _c_offset, PRIM_float, NULL)
#define load_addr_double(_d, _c_base, _c_offset)	load_addr(_d, _c_base, _c_offset, PRIM_double, NULL)
#define load_addr_byte(_d, _c_base, _c_offset)	load_addr(_d, _c_base, _c_offset, PRIM_byte, NULL)
#define load_addr_char(_d, _c_base, _c_offset)	load_addr(_d, _c_base, _c_offset, PRIM_char, NULL)
#define load_addr_short(_d, _c_base, _c_offset)	load_addr(_d, _c_base, _c_offset, PRIM_short, NULL)

#ifdef  USE_RISC_LDST 
#define load_field(_d, _base, _c_offset, _type, _class) \
		t_e = (_base); \
		load(_d, _base, _c_offset, _type, SE_object_field, field_field(), _class); \
		softcall_profile_get( _d, t_e )
#else
#define load_field(_d, _base, _c_offset, _type, _class) \
		move_int_const(t_e, _c_offset); \
		add_ref(t_e, t_e, _base); \
		load(_d, t_e, _type, SE_object_field, field_field(), _class)
#endif 
#define load_field_ref(_d, _base, _c_offset) \
	load_field(_d, _base, _c_offset, PRIM_ref, getClassFromSignature( field_field()->ftype, class->loader ) )
#define load_field_routine(_d, _base, _c_offset)	load_field(_d, _base, _c_offset, PRIM_ref, NULL)
#define load_field_int(_d, _base, _c_offset)		load_field(_d, _base, _c_offset, PRIM_int, NULL)
#define load_field_long(_d, _base, _c_offset)		load_field(_d, _base, _c_offset, PRIM_long, NULL)
#define load_field_float(_d, _base, _c_offset)	load_field(_d, _base, _c_offset, PRIM_float, NULL)
#define load_field_double(_d, _base, _c_offset)	load_field(_d, _base, _c_offset, PRIM_double, NULL)
#define load_field_byte(_d, _base, _c_offset)		load_field(_d, _base, _c_offset, PRIM_byte, NULL)
#define load_field_char(_d, _base, _c_offset)		load_field(_d, _base, _c_offset, PRIM_char, NULL)
#define load_field_short(_d, _base, _c_offset)	load_field(_d, _base, _c_offset, PRIM_short, NULL)

#ifdef  USE_RISC_LDST 
#define load_offset(_d, _base, _c_offset, _type) \
		load(_d, _base, _c_offset, _type, 0, NULL, NULL)
#else
#define load_offset(_d, _base, _c_offset, _type) \
		move_int_const(t_e, _c_offset); \
		add_ref(t_e, t_e, _base); \
		load(_d, t_e, _type, 0, NULL, NULL)
#endif 
#define load_offset_ref(_d, _base, _c_offset)		load_offset(_d, _base, _c_offset, PRIM_ref)
#define load_offset_routine(_d, _base, _c_offset)	load_offset(_d, _base, _c_offset, PRIM_ref)
#define load_offset_int(_d, _base, _c_offset)		load_offset(_d, _base, _c_offset, PRIM_int)
#define load_offset_long(_d, _base, _c_offset)	load_offset(_d, _base, _c_offset, PRIM_long)
#define load_offset_float(_d, _base, _c_offset)	load_offset(_d, _base, _c_offset, PRIM_float)
#define load_offset_double(_d, _base, _c_offset)	load_offset(_d, _base, _c_offset, PRIM_double)
#define load_offset_byte(_d, _base, _c_offset)	load_offset(_d, _base, _c_offset, PRIM_byte)
#define load_offset_char(_d, _base, _c_offset)	load_offset(_d, _base, _c_offset, PRIM_char)
#define load_offset_short(_d, _base, _c_offset)	load_offset(_d, _base, _c_offset, PRIM_short)

/* Stores */
#ifdef  USE_RISC_LDST 
#define store(_base, _c_offset, _val, _type, _flags) \
		new_e = EXPR_new(); \
		new_e->op = OP_store; \
		SET_VALUE( new_e->flags, _flags ); \
		new_e->a[0] = (_base); \
		new_e->val.i_const = (_c_offset); \
		new_e->a[1] = (_val); \
		new_e->type = (_type); \
		EXPR_add(new_e, &d_e, blk, ebb_root); \
		update_use_detail( new_e, d_e )
#else
#define store(_addr, _val, _type, _flags) \
		new_e = EXPR_new(); \
		new_e->op = OP_store; \
		SET_VALUE( new_e->flags, _flags ); \
		new_e->a[0] = (_addr); \
		new_e->a[1] = (_val); \
		new_e->type = (_type); \
		EXPR_add(new_e, &d_e, blk, ebb_root); \
		update_use_detail( new_e, d_e )
#endif

#ifdef  USE_RISC_LDST 
#if 0
#define store_indexed(_base, _offset, _val, _type, _flags) \
		new_e = EXPR_new(); \
		new_e->op = OP_store; \
		SET_VALUE( new_e->flags, _flags | SE_ldst_indexed ); \
		new_e->a[0] = (_base); \
		new_e->val.expr = (_offset); \
		new_e->a[1] = (_val); \
		new_e->type = (_type); \
		EXPR_add(new_e, &d_e, blk, ebb_root); \
		update_use_detail( new_e, d_e )
#define store_offset_scaled(_base, _ix, _c_offset, _val, _type) \
		move_int_const(t_e, _c_offset ); \
		add_ref(t_e, _base, t_e); \
		t_e->class = (_base)->class; \
		move_int_const(t2_e, size2shift[PRIM_SIZE( _type )] ); \
		lshl_int(t2_e, _ix, t2_e); \
		store_indexed(t_e, t2_e, _val, _type, 0)
#else
#define store_offset_scaled(_base, _ix, _c_offset, _val, _type) \
		softcall_profile_arraystore( _base, _ix ); \
		move_int_const(t_e, size2shift[PRIM_SIZE( _type )] ); \
		lshl_int(t_e, _ix, t_e); \
		add_ref(t_e, _base, t_e); \
		t_e->class = (_base)->class; \
		store(t_e, _c_offset, _val, _type, 0)
#endif
#else
#define store_offset_scaled(_base, _ix, _c_offset, _val, _type) \
		move_int_const(t_e, _c_offset); \
		add_ref(t_e, t_e, _base); \
		move_int_const(t2_e, size2shift[PRIM_SIZE( _type )] ); \
		lshl_int(t2_e, _ix, t2_e); \
		add_ref(t_e, t_e, t2_e); \
		store(t_e, _val, _type, 0)
#endif
#define store_offset_scaled_ref(_base, _ix, _c_offset, _val) \
		if( (_base)->class != NULL ){ \
			EXPR_reduce_class( (_val), getClassFromSignature(&(_base)->class->name[1], class->loader ) ); \
		} \
		store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_ref)
#define store_offset_scaled_int(_base, _ix, _c_offset, _val)	store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_int)
#define store_offset_scaled_long(_base, _ix, _c_offset, _val)	store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_long)
#define store_offset_scaled_float(_base, _ix, _c_offset, _val)	store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_float)
#define store_offset_scaled_double(_base, _ix, _c_offset, _val)	store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_double)
#define store_offset_scaled_byte(_base, _ix, _c_offset, _val)	store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_byte)
#define store_offset_scaled_char(_base, _ix, _c_offset, _val)	store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_char)
#define store_offset_scaled_short(_base, _ix, _c_offset, _val)	store_offset_scaled(_base, _ix, _c_offset, _val, PRIM_short)

#ifdef  USE_RISC_LDST 
#define store_addr(_c_base, _c_offset, _val, _type) \
		move_ref_const(t_e, _c_base); \
		softcall_profile_put( t_e ); \
		EXPR_reduce_class( t_e, field_class() ); \
		store(t_e, _c_offset, _val, _type, 0)
#else
#define store_addr(_c_base, _c_offset, _val, _type) \
		move_ref_const(t_e, _c_base); \
		EXPR_reduce_class( t_e, field_class() ); \
		move_ref_const(t2_e, _c_offset); \
		add_ref(t_e, t_e, t2_e); \
		store(t_e, _val, _type, 0)
#endif
#define store_addr_ref(_base, _c_offset, _val) \
		EXPR_reduce_class( (_val), getClassFromSignature( field_field()->ftype, class->loader ) ); \
		store_addr(_base, _c_offset, _val, PRIM_ref)
#define store_addr_int(_base, _c_offset, _val)		store_addr(_base, _c_offset, _val, PRIM_int)
#define store_addr_long(_base, _c_offset, _val)		store_addr(_base, _c_offset, _val, PRIM_long)
#define store_addr_float(_base, _c_offset, _val)	store_addr(_base, _c_offset, _val, PRIM_float)
#define store_addr_double(_base, _c_offset, _val)	store_addr(_base, _c_offset, _val, PRIM_double)
#define store_addr_byte(_base, _c_offset, _val)		store_addr(_base, _c_offset, _val, PRIM_byte)
#define store_addr_char(_base, _c_offset, _val)		store_addr(_base, _c_offset, _val, PRIM_char)
#define store_addr_short(_base, _c_offset, _val)	store_addr(_base, _c_offset, _val, PRIM_short)

#ifdef  USE_RISC_LDST 
#define store_offset(_base, _c_offset, _val, _type) \
		softcall_profile_put( _base ); \
		store(_base, _c_offset, _val, _type, 0)
#else
#define store_offset(_base, _c_offset, _val, _type) \
		move_int_const(t_e, _c_offset); \
		add_ref(t_e, t_e, _base); \
		store(t_e, _val, _type, 0)
#endif
#define store_offset_ref(_base, _c_offset, _val) \
		EXPR_reduce_class( (_val), getClassFromSignature( field_field()->ftype, class->loader ) ); \
		store_offset(_base, _c_offset, _val, PRIM_ref)
#define store_offset_int(_base, _c_offset, _val)	store_offset(_base, _c_offset, _val, PRIM_int)
#define store_offset_long(_base, _c_offset, _val)	store_offset(_base, _c_offset, _val, PRIM_long)
#define store_offset_float(_base, _c_offset, _val)	store_offset(_base, _c_offset, _val, PRIM_float)
#define store_offset_double(_base, _c_offset, _val)	store_offset(_base, _c_offset, _val, PRIM_double)
#define store_offset_byte(_base, _c_offset, _val)	store_offset(_base, _c_offset, _val, PRIM_byte)
#define store_offset_char(_base, _c_offset, _val)	store_offset(_base, _c_offset, _val, PRIM_char)
#define store_offset_short(_base, _c_offset, _val)	store_offset(_base, _c_offset, _val, PRIM_short)

/* Math Ops */
#define binary_op(_d, _s1, _s2, _op, _type) \
		new_e = EXPR_new(); \
		new_e->op = (_op); \
		new_e->type = (_type); \
		new_e->a[0] = (_s1); \
		new_e->a[1] = (_s2); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define unary_op(_d, _s, _op, _type) \
		new_e = EXPR_new(); \
		new_e->op = (_op); \
		new_e->type = (_type); \
		new_e->a[0] = (_s); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )

#define replace_int_const( _d, _v ) \
		move_int_const( t_e, _v ); \
		binary_op( d_e, _d, t_e, OP_replace, (_d)->type )
#define add_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_add, PRIM_int)
#define add_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_add, PRIM_long)
#define add_ref(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_add, PRIM_ref)
#ifdef  USE_SOFT_FLOAT 
#define add_float(_d, _s1, _s2)	softcall_2a( _d, soft_fadd, _s1, _s2, PRIM_float, CALL_no_side_effect, NULL )
#define add_double(_d, _s1, _s2)	softcall_2a( _d, soft_faddl, _s1, _s2, PRIM_double, CALL_no_side_effect, NULL )
#else
#define add_float(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_add, PRIM_float)
#define add_double(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_add, PRIM_double)
#endif

#define add_int_const(_d, _s, _v)	move_int_const(t_e, _v); add_int(_d, _s, t_e)
#define add_ref_const(_d, _s, _v)	move_ref_const(t_e, _v); add_ref(_d, _s, t_e)

#define sub_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_sub, PRIM_int)
#define sub_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_sub, PRIM_long)
#ifdef  USE_SOFT_FLOAT 
#define sub_float(_d, _s1, _s2)	softcall_2a( _d, soft_fsub, _s1, _s2, PRIM_float, CALL_no_side_effect, NULL )
#define sub_double(_d, _s1, _s2)	softcall_2a( _d, soft_fsubl, _s1, _s2, PRIM_double, CALL_no_side_effect, NULL )
#else
#define sub_float(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_sub, PRIM_float)
#define sub_double(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_sub, PRIM_double)
#endif

#define sub_int_const(_d, _s, _v)	move_int_const(t_e, _v); sub_int(_d, _s, t_e)

#define mul_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_mul, PRIM_int)
#ifdef  USE_SOFT_LONG 
#define mul_long(_d, _s1, _s2)	softcall_2a( _d, soft_lmul, _s1, _s2, PRIM_long, CALL_no_side_effect, NULL )
#else
#define mul_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_mul, PRIM_long)
#endif
#ifdef  USE_SOFT_FLOAT 
#define mul_float(_d, _s1, _s2)	softcall_2a( _d, soft_fmul, _s1, _s2, PRIM_float, CALL_no_side_effect, NULL )
#define mul_double(_d, _s1, _s2)	softcall_2a( _d, soft_fmull, _s1, _s2, PRIM_double, CALL_no_side_effect, NULL )
#else
#define mul_float(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_mul, PRIM_float)
#define mul_double(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_mul, PRIM_double)
#endif

#ifdef  USE_SOFT_DIV
#define div_int(_d, _s1, _s2)	softcall_2a( _d, soft_div, _s1, _s2, PRIM_int, CALL_no_side_effect, NULL )
#else
#define div_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_div, PRIM_int)
#endif
#ifdef  USE_SOFT_LONG 
#define div_long(_d, _s1, _s2)	softcall_2a( _d, soft_ldiv, _s1, _s2, PRIM_long, CALL_no_side_effect, NULL )
#else
#define div_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_div, PRIM_long)
#endif
#if defined(USE_SOFT_FLOAT) || defined(USE_SOFT_DIV)
#define div_float(_d, _s1, _s2)	softcall_2a( _d, soft_fdiv, _s1, _s2, PRIM_float, CALL_no_side_effect, NULL )
#define div_double(_d, _s1, _s2)	softcall_2a( _d, soft_fdivl, _s1, _s2, PRIM_double, CALL_no_side_effect, NULL )
#else
#define div_float(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_div, PRIM_float)
#define div_double(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_div, PRIM_double)
#endif

#ifdef  USE_SOFT_REM
#define rem_int(_d, _s1, _s2) \
		if( EXPR_IS_CONST( _s2 ) && count_bits( (_s2)->val.i_const, (int*)&tmp ) == 1 ){ \
			binary_op(_d, _s1, _s2, OP_rem, PRIM_int); \
		} \
		else{ \
			softcall_2a( _d, soft_rem, _s1, _s2, PRIM_int, CALL_no_side_effect, NULL ); \
		}
#else
#define rem_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_rem, PRIM_int)
#endif
#ifdef  USE_SOFT_LONG 
#define rem_long(_d, _s1, _s2)	softcall_2a( _d, soft_lrem, _s1, _s2, PRIM_long, CALL_no_side_effect, NULL )
#else
#define rem_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_rem, PRIM_long)
#endif
#if defined(USE_SOFT_FLOAT) || defined(USE_SOFT_REM)
#define rem_float(_d, _s1, _s2)	softcall_2a( _d, soft_frem, _s1, _s2, PRIM_float, CALL_no_side_effect, NULL )
#define rem_double(_d, _s1, _s2)	softcall_2a( _d, soft_freml, _s1, _s2, PRIM_double, CALL_no_side_effect, NULL )
#else
#define rem_float(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_rem, PRIM_float)
#define rem_double(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_rem, PRIM_double)
#endif

#define neg_int(_d, _s)		unary_op(_d, _s, OP_neg, PRIM_int)
#define neg_long(_d, _s)		unary_op(_d, _s, OP_neg, PRIM_long)
#define neg_float(_d, _s)		unary_op(_d, _s, OP_neg, PRIM_float)
#define neg_double(_d, _s)		unary_op(_d, _s, OP_neg, PRIM_double)

/* Logical Ops */
#define lshl_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_lsl, PRIM_int)
#define lshl_int_const(_d, _s, _v)	move_int_const(t_e, _v); binary_op(_d, _s, t_e, OP_lsl, PRIM_int)
#ifdef  USE_SOFT_LONG 
#define lshl_long(_d, _s1, _s2)	softcall_2a( _d, soft_lshll, _s1, _s2, PRIM_long, CALL_no_side_effect, NULL )
#else
#define lshl_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_lsl, PRIM_long)
#endif

#define lshr_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_lsr, PRIM_int)
#define lshr_int_const(_d, _s, _v)	move_int_const(t_e, _v); binary_op(_d, _s, t_e, OP_lsr, PRIM_int)
#ifdef  USE_SOFT_LONG 
#define lshr_long(_d, _s1, _s2)	softcall_2a( _d, soft_lshrl, _s1, _s2, PRIM_long, CALL_no_side_effect, NULL )
#else
#define lshr_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_lsr, PRIM_long)
#endif

#define ashr_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_asr, PRIM_int)
#define ashr_int_const(_d, _s, _v)	move_int_const(t_e, _v); binary_op(_d, _s, t_e, OP_asr, PRIM_int)
#ifdef  USE_SOFT_LONG 
#define ashr_long(_d, _s1, _s2)	softcall_2a( _d, soft_ashrl, _s1, _s2, PRIM_long, CALL_no_side_effect, NULL )
#else
#define ashr_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_asr, PRIM_long)
#endif

#define and_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_and, PRIM_int)
#define and_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_and, PRIM_long)	

#define or_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_or, PRIM_int)
#define or_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_or, PRIM_long)

#define xor_int(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_xor, PRIM_int)
#define xor_long(_d, _s1, _s2)	binary_op(_d, _s1, _s2, OP_xor, PRIM_long)

#define cvt_int_long(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_long)
#define cvt_int_float(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_float)
#define cvt_int_double(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_double)
#define cvt_int_byte(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_byte)
#define cvt_int_char(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_char)
#define cvt_int_short(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_short)

#define cvt_long_int(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_int)
#ifdef USE_SOFT_CVT_LONG
#define cvt_long_float(_d, _s)	softcall_1a( _d, soft_cvtlf, _s, PRIM_float, CALL_no_side_effect, NULL )
#define cvt_long_double(_d, _s)	softcall_1a( _d, soft_cvtld, _s, PRIM_double, CALL_no_side_effect, NULL )
#else
#define cvt_long_float(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_float)
#define cvt_long_double(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_double)
#endif

#define cvt_float_int(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_int)
#ifdef USE_SOFT_CVT_LONG
#define cvt_float_long(_d, _s)	softcall_1a( _d, soft_cvtfl, _s, PRIM_long, CALL_no_side_effect, NULL )
#else
#define cvt_float_long(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_long)
#endif
#define cvt_float_double(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_double)

#define cvt_double_int(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_int)
#ifdef USE_SOFT_CVT_LONG
#define cvt_double_long(_d, _s)	softcall_1a( _d, soft_cvtdl, _s, PRIM_long, CALL_no_side_effect, NULL )
#else
#define cvt_double_long(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_long)
#endif
#define cvt_double_float(_d, _s)	unary_op(_d, _s, OP_cvt, PRIM_float)

#define set_cc( _d, _s1, _s2, _cc_flags ) \
		new_e = EXPR_new(); \
		new_e->op = OP_set_cc; \
		new_e->type = PRIM_int; \
		SET_VALUE( new_e->flags, _cc_flags ); \
		new_e->a[0] = (_s1); \
		new_e->a[1] = (_s2); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define cc_cbranch( _d, _s1, _s2, _cmp, _tgt, _cmp_flags ) \
		set_cc( t_e, _s1, _s2, _cmp_flags ); \
		new_e = EXPR_new(); \
		new_e->op = (_cmp); \
                new_e->type = PRIM_int; \
		new_e->a[0] = t_e; \
		SET_VALUE( new_e->val.label, _tgt ); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define cbranch( _d, _s1, _s2, _cmp, _tgt, _cmp_flags ) \
		new_e = EXPR_new(); \
		new_e->op = (_cmp); \
                new_e->type = PRIM_int; \
		SET_VALUE( new_e->flags, _cmp_flags ); \
		new_e->a[0] = (_s1); \
		new_e->a[1] = (_s2); \
		SET_VALUE( new_e->val.label, _tgt ); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define branch_a(_tgt) \
		new_e = EXPR_new(); \
		new_e->op = OP_branch; \
		new_e->type = PRIM_void; \
		SET_VALUE( new_e->val.label, _tgt ); \
		EXPR_add(new_e, 0, blk, ebb_root)
#define branch_indirect(_tgt) \
		new_e = EXPR_new(); \
		new_e->op = OP_branch_indirect; \
		new_e->a[0] = (_tgt); \
		EXPR_add(new_e, &d_e, blk, ebb_root); \
		update_use_detail( new_e, d_e )

#define cbranch_ref_const_eq(_s, _v, _tgt) \
		move_ref_const(t2_e, _v); \
		cbranch_ref_eq( _s, t2_e, _tgt )
#define cbranch_ref_const_ne(_s, _v, _tgt) \
		move_ref_const(t2_e, _v); \
		cbranch_ref_ne( _s, t2_e, _tgt )

#define cbranch_int_const_eq(_s, _v, _tgt) \
		move_int_const(t2_e, _v); \
		cbranch_int_eq( _s, t2_e, _tgt )
#define cbranch_int_const_ne(_s, _v, _tgt) \
		move_int_const(t2_e, _v); \
		cbranch_int_ne( _s, t2_e, _tgt )
#define cbranch_int_const_lt(_s, _v, _tgt) \
		move_int_const(t2_e, _v); \
		cbranch_int_lt( _s, t2_e, _tgt )
#define cbranch_int_const_ge(_s, _v, _tgt) \
		move_int_const(t2_e, _v); \
		cbranch_int_ge( _s, t2_e, _tgt )
#define cbranch_int_const_gt(_s, _v, _tgt) \
		move_int_const(t2_e, _v); \
		cbranch_int_gt( _s, t2_e, _tgt )
#define cbranch_int_const_le(_s, _v, _tgt) \
		move_int_const(t2_e, _v); \
		cbranch_int_le( _s, t2_e, _tgt )

#define cbranch_ref_eq(_s1, _s2, _tgt) \
		cbranch_int_eq( _s1, _s2, _tgt )
#define cbranch_ref_ne(_s1, _s2, _tgt) \
		cbranch_int_ne( _s1, _s2, _tgt )

#define cbranch_int_eq(_s1, _s2, _tgt) \
		cbranch_eq_func( , )( d_e, _s1, _s2, OP_cbranch_eq, _tgt, 0 )
#define cbranch_int_ne(_s1, _s2, _tgt) \
		cbranch_ne_func( , )( d_e, _s1, _s2, OP_cbranch_ne, _tgt, 0 )
#define cbranch_int_lt(_s1, _s2, _tgt) \
		cbranch_lt_func( , )( d_e, _s1, _s2, OP_cbranch_lt, _tgt, 0 )
#define cbranch_int_ge(_s1, _s2, _tgt) \
		cbranch_ge_func( , )( d_e, _s1, _s2, OP_cbranch_ge, _tgt, 0 )
#define cbranch_int_gt(_s1, _s2, _tgt) \
		cbranch_gt_func( , )( d_e, _s1, _s2, OP_cbranch_gt, _tgt, SE_invert_cmp )
#define cbranch_int_le(_s1, _s2, _tgt) \
		cbranch_le_func( , )( d_e, _s1, _s2, OP_cbranch_le, _tgt, SE_invert_cmp )
#define cbranch_int_ult( _d, _s1, _s2, _tgt) \
		cbranch_ult_func( , )( _d, _s1, _s2, OP_cbranch_ult, _tgt, SE_unsigned )

#define start_sub_block()		cfg->env.global.add_flags |= ADD_disable_cp
#define end_sub_block()			cfg->env.global.add_flags &= ~ADD_disable_cp
#define stored_code_label(_e)		(_e)
#define table_code_label(_e)		(_e)
#define reference_code_label(_pc)	LAB_new_block_branch( cfg, new_e, CFG_get_block( cfg, _pc, false ) )
#define reference_label(_op, _ix)	LAB_new_local_branch( cfg, _ix, new_e, false )
#define reference_table_label(_ix)	LAB_new_local_branch( cfg, _ix, new_e, false )
#define set_label(_op, _ix) \
		new_e = EXPR_new(); \
		new_e->op = OP_target; \
		new_e->type = PRIM_label; \
		new_e->val.label = LAB_new_local_branch( cfg, _ix, new_e, true ); \
		EXPR_add(new_e, 0, blk, ebb_root)
#define clear_label(_ix) \
		LAB_free_target( cfg, _ix )
#define word_const(_pos) 		(_pos)[0] * 0x01000000 + (_pos)[1] * 0x00010000 + (_pos)[2] * 0x00000100 + (_pos)[3] * 0x00000001 
#define build_code_ref(_pos, _pc) \
		new_e = EXPR_new(); \
		new_e->op = OP_inline_data; \
		new_e->type = PRIM_label; \
		new_e->val.label = LAB_new_block_branch( cfg, new_e, CFG_get_block( cfg, (_pc) + word_const(_pos), false ) ); \
		new_e->val.label->patch = PATCH_inline_data; \
		EXPR_add(new_e, 0, blk, ebb_root)
#define build_key(_pos) \
		new_e = EXPR_new(); \
		new_e->op = OP_inline_data; \
		new_e->type = PRIM_ref; \
		new_e->val.i_const = word_const( _pos ); \
		EXPR_add(new_e, 0, blk, ebb_root)

#ifdef OPTIMIZE_INLINE
#define INLINE_RETURN (blk->env->inlined)
#else
#define INLINE_RETURN false
#endif

#define returnarg(_s, _type, _jtype) \
		if( INLINE_RETURN ){ \
			(_s)->flags |= SE_inline_return; \
			cfg->env.method.return_val = (_s); \
			blk->flags |= BLK_inline_epilogue; \
		} \
		else{ \
	                if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) { \
	                        softcall_trace_return( _jtype, _s ); \
	                } \
			softcall_profile_return(); \
			new_e = EXPR_new(); \
			new_e->op = OP_returnarg; \
			new_e->type = (_type); \
			new_e->a[0] = (_s); \
			EXPR_add(new_e, &d_e, blk, ebb_root); \
			update_use_detail( new_e, d_e ); \
			new_e = EXPR_new(); \
			new_e->op = OP_epilogue; \
			EXPR_add(new_e, 0, blk, ebb_root); \
		}

#define returnarg_void() \
		if( INLINE_RETURN ){ \
			blk->flags |= BLK_inline_epilogue; \
		} \
		else{ \
	                if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) { \
				move_ref_const(t_e, 0); \
	                        softcall_trace_return( 'V', t_e); \
			} \
			softcall_profile_return(); \
			new_e = EXPR_new(); \
			new_e->op = OP_epilogue; \
			EXPR_add(new_e, 0, blk, ebb_root); \
		}

#define returnarg_int(_s)		returnarg(_s, PRIM_int, 'I')
#define returnarg_long(_s)		returnarg(_s, PRIM_long, 'J')
#define returnarg_float(_s)		returnarg(_s, PRIM_float, 'F')
#define returnarg_double(_s)		returnarg(_s, PRIM_double, 'D')
#define returnarg_ref(_s)		returnarg(_s, PRIM_ref, 'L')

#define end_function()	

#define softcall_0a(_d, _func, _type, _flags, _class ) \
		move_ref_const( call_e, _func ); \
		new_e = EXPR_new(); \
		new_e->op = OP_call; \
		new_e->a[0] = call_e; \
		new_e->type = (_type); \
		SET_VALUE( new_e->class, _class ); \
		new_e->val.frame = EXPR_build_call_frame_native(0, (_flags)); \
		generated_call = true; \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define softcall_1a(_d, _func, _a1, _type, _flags, _class ) \
		move_ref_const( call_e, _func ); \
		new_e = EXPR_new(); \
		new_e->op = OP_call; \
		new_e->a[0] = call_e; \
		new_e->type = (_type); \
		SET_VALUE( new_e->class, _class ); \
		new_e->val.frame = EXPR_build_call_frame_native(1, (_flags), (_a1)); \
		generated_call = true; \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define softcall_2a(_d, _func, _a1, _a2, _type, _flags, _class ) \
		move_ref_const( call_e, _func ); \
		new_e = EXPR_new(); \
		new_e->op = OP_call; \
		new_e->a[0] = call_e; \
		new_e->type = (_type); \
		SET_VALUE( new_e->class, _class ); \
		new_e->val.frame = EXPR_build_call_frame_native(2, (_flags), (_a1), (_a2)); \
		generated_call = true; \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define softcall_3a(_d, _func, _a1, _a2, _a3, _type, _flags, _class ) \
		move_ref_const( call_e, _func ); \
		new_e = EXPR_new(); \
		new_e->op = OP_call; \
		new_e->a[0] = call_e; \
		new_e->type = (_type); \
		SET_VALUE( new_e->class, _class ); \
		new_e->val.frame = EXPR_build_call_frame_native(3, (_flags), (_a1), (_a2), (_a3)); \
		generated_call = true; \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )

#define call_soft(_func) \
		softcall_0a( d_e, _func, PRIM_void, 0, NULL )
#define softcall_monitorenter(_obj) \
                add_int_const( t_e, _obj, object_lock_offset ); \
                t2_e = EXPR_new_reg_src( R_fp, blk ); \
		softcall_2a( d_e, _lockMutex, t_e, t2_e, PRIM_void, 0, NULL )
#define softcall_monitorexit(_obj) \
                add_int_const( t_e, _obj, object_lock_offset ); \
                t2_e = EXPR_new_reg_src( R_fp, blk ); \
		softcall_2a( d_e, _unlockMutex, t_e, t2_e, PRIM_void, 0, NULL )
#define softcall_new(_d, _obj) \
		move_ref_const(t2_e, _obj); \
		softcall_1a( _d, _Jv_AllocObject, t2_e, PRIM_ref, 0, class_object() ); \
		(_d)->flags |= SE_null_checked
#define softcall_newarray(_d, _sz, _prim) \
		move_int_const(t2_e, _prim); \
		softcall_2a( _d, _Jv_NewArray, t2_e, _sz, PRIM_ref, 0, lookupArrayProcess( classIx[(_prim)-4], CSTATE_LINKED ) ); \
		(_d)->flags |= SE_null_checked
#define softcall_anewarray(_d, _sz, _obj) \
		move_ref_const(t2_e, _obj); \
		softcall_2a( _d, _Jv_NewObjectArray, _sz, t2_e, PRIM_ref, 0, lookupArrayProcess( class_object(), CSTATE_LINKED ) ); \
		(_d)->flags |= SE_null_checked
#define softcall_multianewarray(_d, _sz, _stack_top, _obj) \
		move_ref_const(t_e, jit4_soft_multianewarray); \
		move_int_const(t2_e, _sz); \
		move_ref_const(t3_e, _obj); \
		new_e = EXPR_new(); \
		new_e->op = OP_call; \
		new_e->a[0] = t_e; \
		new_e->type = PRIM_ref; \
		new_e->class = class_object(); \
		new_e->val.frame = EXPR_build_call_frame_multianewarray( t2_e, t3_e, stack_p ); \
		generated_call = true; \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		(_d)->flags |= SE_null_checked; \
		update_use_detail( new_e, _d ); \
		pop( _sz-1 )
#define softcall_athrow(_obj) \
		softcall_1a( d_e, soft_athrow, _obj, PRIM_void, CALL_throw, NULL )
#define softcall_exception_info( _d ) \
		softcall_0a(  _d, soft_get_thrown_exception, PRIM_ref, CALL_no_side_effect, NULL )
#define softcall_checkcast(_res, _s, _obj) \
		(_s)->flags |= SE_null_checked; \
		move_ref_const(t2_e, _obj); \
		softcall_2a( _res, _Jv_CheckCast, t2_e, _s, PRIM_ref, CALL_no_side_effect, _obj )
#define softcall_instanceof(_res, _s, _obj) \
		move_ref_const(t2_e, _obj); \
		softcall_2a( _res, _Jv_IsInstanceOf, _s, t2_e, PRIM_ref, CALL_no_side_effect, NULL )
#define softcall_initialise_class( _c ) \
		if( (_c) != NULL && ( (_c)->sdata == NULL || ( CLASS_STATE(_c) != CSTATE_COMPLETE ) ) ){ \
			move_ref_const( t2_e, (_c) ); \
			softcall_1a( d_e, _Jv_InitClass, t2_e, PRIM_void, CALL_no_side_effect, NULL ); \
		}
#define softcall_nosuchfield(_class, _name) \
		move_ref_const(t2_e, _class); \
		move_ref_const(t3_e, _name); \
		softcall_2a( d_e, soft_nosuchfield, t2_e, t3_e, PRIM_void, CALL_throw, NULL )
#define softcall_nosuchmethod(_cname, _mname, _msig) \
		move_ref_const(t2_e, _cname); \
		move_ref_const(t3_e, _mname); \
		move_ref_const(t4_e, _msig); \
		softcall_3a( d_e, soft_nosuchmethod, t2_e, t3_e, t4_e, PRIM_void, CALL_throw, NULL )
#define softcall_nosuchclass(_cname) \
		move_ref_const(t2_e, _cname); \
		softcall_1a( d_e, soft_nosuchclass, t2_e, PRIM_void, CALL_throw, NULL )
#define softcall_lookupmethod(_d, _method, _class, _obj) \
		move_ref_const(t2_e, (_class) ); \
		move_int_const(t3_e, (_method) - CLASS_METHODS( (_class) ) ); \
		softcall_3a( _d, soft_lookupmethod, _obj, t2_e, t3_e, PRIM_ref, CALL_no_side_effect, NULL )
#define softcall_lazylookupmethod(_d, _name, _sig, _obj) \
		move_ref_const(t2_e, (_name) ); \
		move_ref_const(t3_e, (_sig) ); \
		softcall_3a( _d, soft_lazylookupmethod, _obj, t2_e, t3_e, PRIM_ref, CALL_no_side_effect, NULL )
#define softcall_trace_call( _m, _c ) \
		move_ref_const(t2_e, _m); \
		move_ref_const(t3_e, _c); \
		t4_e = EXPR_new_reg_src( 30, blk ); \
		if( !METHOD_IS_STATIC(_m) ){ \
			blk->locals.in[0]->uses.block++; /* force object ptr to be live */ \
		} \
		softcall_3a( d_e, soft_trace, t2_e, t3_e, t4_e, PRIM_void, CALL_no_side_effect, NULL )
#define softcall_trace_return( _jt, _rv ) \
		move_int_const(t2_e, _jt ); \
		softcall_2a( d_e, soft_end, t2_e, _rv, PRIM_void, CALL_no_side_effect, NULL )
#ifdef PROFILE
#define softcall_profile_get( _d, _o ) \
		if( (_d) == new_e && new_e->op != OP_copy ){ \
			t2_e = (_o); \
			di = DEP_new( class, method, pc, field_class(), field_field(), DEP_get ); \
			move_ref_const(t3_e, di ); \
			add_ref_const(t4_e, t2_e, field_field()->poffset ); \
			softcall_2a( d_e, PROF_get, t3_e, t4_e, PRIM_void, CALL_no_side_effect, NULL ); \
			(_o) = t2_e; \
		}
#define softcall_profile_put( _o ) \
		di = DEP_new( class, method, pc, field_class(), field_field(), DEP_put ); \
		t2_e = (_o); \
		move_ref_const(t3_e, di ); \
		add_ref_const(t4_e, t2_e, field_field()->poffset ); \
		softcall_2a( d_e, PROF_put, t3_e, t4_e, PRIM_void, CALL_no_side_effect, NULL ); \
		(_o) = t2_e
#define softcall_profile_arrayload( _d, _o, _ix ) \
		if( (_d) == new_e && new_e->op != OP_copy ){ \
			di = DEP_new( class, method, pc, (_o)->class, 0, DEP_get ); \
			move_ref_const(t2_e, di ); \
			softcall_3a( d_e, PROF_arrayload, t2_e, _o, _ix, PRIM_void, CALL_no_side_effect, NULL ); \
		}
#define softcall_profile_arraystore( _o, _ix ) \
		di = DEP_new( class, method, pc, (_o)->class, 0, DEP_put ); \
		move_ref_const(t2_e, di ); \
		softcall_3a( d_e, PROF_arraystore, t2_e, _o, _ix, PRIM_void, CALL_no_side_effect, NULL )
#define softcall_profile_call( ) \
		li = LOC_new( class, method, pc ); \
		move_ref_const( t_e, li ); \
		softcall_1a( d_e, PROF_call, t_e, PRIM_void, CALL_no_side_effect, NULL )
#define softcall_profile_return( ) \
		softcall_0a( d_e, PROF_return, PRIM_void, CALL_no_side_effect, NULL )
#define softcall_profile_loop_preheader( ) \
		li = LOC_new( class, method, pc ); \
		move_ref_const( t_e, li ); \
		softcall_1a( d_e, PROF_loop_preheader, t_e, PRIM_void, CALL_no_side_effect, NULL )
#define softcall_profile_loop_iteration( ) \
		softcall_0a( d_e, PROF_loop_iteration, PRIM_void, CALL_no_side_effect, NULL )
#else
#define softcall_profile_get( _d, _o )
#define softcall_profile_put( _o )
#define softcall_profile_arrayload( _d, _o, _ix )
#define softcall_profile_arraystore( _o, _ix )
#define softcall_profile_call( )
#define softcall_profile_return( )
#define softcall_profile_loop_preheader( )
#define softcall_profile_loop_iteration( )
#endif
#if 1
#define softcall_writeref(_f, _t) \
		softcall_2a( d_e, gcWriteRef, _f, _t, PRIM_void, CALL_no_side_effect, NULL )
#define softcall_writeref_static(_f, _t) \
		move_ref_const( t_e, _f ); \
		softcall_2a( d_e, gcWriteRef, t_e, _t, PRIM_void, CALL_no_side_effect, NULL )
#else
#define softcall_writeref(_f, _t) \
		move_ref_const( t_e, 0 ); \
		preload( t2_e, _f ); \
		cbranch( t_e, _t, t_e, OP_cbranch_eq, reference_label(1,1), PRIM_ref ); \
		cfg->env.global.in_nested_block = true; \
		if( new_e == t_e ){ \
			softcall_2a( d_e, gcWriteRef, t2_e, _t, PRIM_void, CALL_no_side_effect | CALL_nested, NULL ); \
			set_label( 1, 1 ); \
		} \
		else{ \
			clear_label( 1 ); \
		} \
		cfg->env.global.in_nested_block = false
#define softcall_writeref_static( _f, _t ); \
		move_ref_const( t_e, 0 ); \
		move_ref_const( t2_e, _f ); \
		cbranch( d_e, _t, t_e, OP_cbranch_eq, reference_label(1,1), PRIM_ref ); \
		cfg->env.global.in_nested_block = true; \
		if( new_e == d_e ){ \
			softcall_2a( d_e, gcWriteRef, t2_e, _t, PRIM_void, CALL_no_side_effect | CALL_nested, NULL ); \
			set_label( 1, 1 ); \
		} \
		else{ \
			clear_label( 1 ); \
		} \
		cfg->env.global.in_nested_block = false
#endif
#define GLOBAL_set_block( _e ) \
		if( (_e)->flags & SE_global ){ \
			tmp_b = blk; \
			blk = cfg->env.global.header_block; \
		}
#define GLOBAL_restore_block() \
		if( tmp_b != NULL ){ \
			blk = tmp_b; \
			tmp_b = NULL; \
		}
#define check_array_index(_ref, _idx) \
		(_ref)->flags |= SE_null_checked; \
		load_offset_int(t_e, _ref, object_array_length); \
		cbranch_int_ult( d_e, _idx, t_e, reference_label(1, 1) ); \
		cfg->env.global.in_nested_block = true; \
		if( new_e->op != OP_copy \
		    && ( new_e->flags & SE_global || new_e == d_e ) ){ \
			GLOBAL_set_block( new_e ); \
			softcall_0a( d_e, soft_badarrayindex, PRIM_void, CALL_throw, NULL ); \
			set_label( 1, 1 ); \
			GLOBAL_restore_block(); \
		} \
		else{ \
			clear_label( 1 ); \
		} \
		cfg->env.global.in_nested_block = false
#define check_stack_overflow( _lim ) \
                t_e = EXPR_new_reg_src( 29, blk ); \
		cbranch_int_ult( d_e, _lim, t_e, reference_label(1, 1) ); \
		cfg->env.global.in_nested_block = true; \
		if (new_e == d_e) { \
			softcall_0a( d_e, soft_stackoverflow, PRIM_void, CALL_throw, NULL ); \
			set_label( 1, 1 ); \
		} \
		else{ \
			clear_label( 1 ); \
		} \
		cfg->env.global.in_nested_block = false
#define _force_check_null(_instr, _ref, _op) \
		(_ref)->flags |= SE_null_checked; \
		move_ref_const(t_e, 0); \
		cbranch( d_e, _ref, t_e, OP_cbranch_ne, reference_label(1,1), PRIM_ref); \
		cfg->env.global.in_nested_block = true; \
		if( new_e == d_e && new_e->op != OP_copy ) { \
			softcall_0a( d_e, soft_nullpointer, PRIM_void, CALL_throw, NULL ); \
			set_label( 1, 1 ); \
		} \
		else{ \
			clear_label( 1 ); \
		} \
		cfg->env.global.in_nested_block = false

#define force_check_null(_instr, _ref, _op)
#if 0
#define check_null(_instr, _ref, _op) force_check_null( _instr, _ref, _op )
#else
#define check_null(_instr, _ref, _op) (_ref)->flags |= SE_null_checked

#endif

#define lcmp(_d, _s1, _s2) \
		softcall_2a( _d, soft_lcmp, _s1, _s2, PRIM_int, CALL_no_side_effect, NULL )
#ifdef USE_SOFT_FLOAT
#define cmpl_float(_d, _s1, _s2) \
		softcall_2a( _d, soft_fcmpl, _s1, _s2, PRIM_int, CALL_no_side_effect, NULL )
#define cmpg_float(_d, _s1, _s2) \
		softcall_2a( _d, soft_fcmpg, _s1, _s2, PRIM_int, CALL_no_side_effect, NULL )
#define cmpl_double(_d, _s1, _s2) \
		softcall_2a( _d, soft_dcmpl, _s1, _s2, PRIM_int, CALL_no_side_effect, NULL )
#define cmpg_double(_d, _s1, _s2) \
		softcall_2a( _d, soft_dcmpg, _s1, _s2, PRIM_int, CALL_no_side_effect, NULL )
#else
#define cmp( _d, _s1, _s2, _c_flags ) \
		new_e = EXPR_new(); \
		new_e->op = OP_cmp; \
		new_e->type = PRIM_int; \
		new_e->val.c_flags |= (_c_flags); \
		new_e->a[0] = (_s1); \
		new_e->a[1] = (_s2); \
		EXPR_add(new_e, &(_d), blk, ebb_root); \
		update_use_detail( new_e, _d )
#define cmpl_float(_d, _s1, _s2) \
		cmp( _d, _s1, _s2, FCMP_l )
#define cmpg_float(_d, _s1, _s2) \
		cmp( _d, _s1, _s2, FCMP_g )
#define cmpl_double(_d, _s1, _s2) \
		cmp( _d, _s1, _s2, FCMP_l )
#define cmpg_double(_d, _s1, _s2) \
		cmp( _d, _s1, _s2, FCMP_g )
#endif

#define check_array_store(_ref, _idx) /* Do nothing */

#define softcall_addreference(_from, _to)
#define softcall_addreference_static(_from, _to)
#define softcall_breakpoint()

#define clearPostedException()

#define popargs()	/* Do nothing */
#define pusharg()	/* Do nothing */
		
#ifdef OPTIMIZE_INLINE
#define INLINE_CALL (frame->flags & (CALL_inline | CALL_specialized_inline) )
#else
#define INLINE_CALL false
#endif

#define build_instruction_from_call( _rettype, _class , _method ) \
		new_e = EXPR_build_instruction_from_call( &stack_p, blk, _class, _method ); \
		new_e->type = sigToPrimType( _rettype ); \
		if( new_e->type == PRIM_void ){ \
			tgt = &d_e; \
		} \
		else{ \
			if( PRIM_SIZE( new_e->type ) == 8 ){ \
				push( 2 ); \
			} \
			else{ \
				push( 1 ); \
			} \
			tgt = &stack(0); \
		} \
		EXPR_add( new_e, tgt, blk, ebb_root ); \
		update_use_detail( new_e, *tgt )

#define build_call_frame(_sig, _obj, _nargs, _jni, _class )
#define jit4_build_call_frame(_sig, _obj, _nargs, _jni, _class ) \
		frame = EXPR_build_call_frame( meth_i.method, meth_i.class, _sig, _obj, _jni, stack_p, locals, blk, getopcode() )
#define call(_m) \
		new_e = EXPR_new(); \
		new_e->op = OP_call; \
		if( (_m) != NULL ){ \
			new_e->a[0] = (_m); \
		} \
		new_e->val.frame = frame; \
		EXPR_assign_class( new_e, frame->ret ); \
       	        was_call = 1
#ifdef DIRECT_CALL
#define call_indirect_method(_m) \
		if( INLINE_CALL ){ \
			t_e = NULL; \
		} \
		else{ \
			move_ref_const(t_e, METHOD_NATIVECODE(_m) ); \
		} \
		call(t_e)
#else
#define call_indirect_method(_m) \
		if( INLINE_CALL ){ \
			t_e = NULL; \
		} \
		else{ \
			if( METHOD_TRANSLATED(_m) ){ \
				move_ref_const(t_e, METHOD_NATIVECODE(_m) ); \
			} \
			else { \
				move_ref_const(t_e, &METHOD_NATIVECODE(_m) ); \
				load_offset_routine(t_e, t_e, 0); \
			} \
		} \
		call(t_e)
#endif
#define return_val(_d, _type)	\
		if (was_call) { \
			new_e->type = (_type); \
			EXPR_add( new_e, &(_d), blk, ebb_root ); \
			update_use_detail( new_e, _d ); \
			if( new_e == (_d) && INLINE_CALL ){ \
				block_list stack; \
        			LIST_zero( &stack ); \
			        BL_add( &stack, frame->env->root ); \
			        EBB_generate_exprs( &stack ); \
        			LIST_free( &stack ); \
			        CFG_set_method_env( cfg, blk->env ); \
	                        (_d) = frame->env->return_val; \
			} \
			else{ \
				generated_call = true; \
			} \
		} \
                was_call = 0

#define return_void()			return_val(t_e, PRIM_void)
#define return_ref(_d)		return_val(_d, PRIM_ref)
#define return_int(_d)		return_val(_d, PRIM_int)
#define return_long(_d)		return_val(_d, PRIM_long)
#define return_float(_d)		return_val(_d, PRIM_float)
#define return_double(_d)		return_val(_d, PRIM_double)

#define mon_enter( _method, _class, _obj ) \
		if( (_method)->accflags & ACC_SYNCHRONISED ){ \
			if( METHOD_IS_STATIC(_method) ){ \
				move_ref_const( t2_e, (_class) ); \
				softcall_monitorenter( t2_e ); \
			} \
			else{ \
                                softcall_monitorenter( _obj ); \
			} \
		}
#define mon_exit( _method, _class, _obj ) \
		if( (_method)->accflags & ACC_SYNCHRONISED ){ \
			if( METHOD_IS_STATIC(_method) ){ \
				move_ref_const(t2_e, (_class) ); \
				softcall_monitorexit( t2_e ); \
			} \
			else{ \
                                softcall_monitorexit( _obj ); \
			} \
		}
#define monitor_enter()	mon_enter( method, class, local(0) )
#define monitor_exit()	mon_exit( method, class, local(0) )

#define slot_alloctmp(_slt) 		/* Do nothing */
#define slot_freetmp(_slt)		/* Do nothing */

#ifdef OPTIMIZE_MEMORY
#define EXPR_new() 	gc_malloc_fixed( sizeof(expr) )
#define EXPR_free( _e ) gc_free_fixed( _e )
#else
extern void EXPR_free( expr * );
extern expr * EXPR_new();
#endif
extern void BLK_free_exprs( expr *, expr * );
extern void EXPR_reduce_class( expr *, Hjava_lang_Class * );

#define assign_local( _ix ) EXPR_assign_local( local(_ix), _ix, last_local, blk, cfg )

#define EXPR_REMOVE( _blk, _e ) \
	if( (_blk)->exprs_tail == _e ){ \
		(_blk)->exprs_tail = (_e)->expr.prev; \
	} \
	if( (_blk)->exprs_head == _e ){ \
		(_blk)->exprs_head = (_e)->expr.next; \
	} \
	DLL_REMOVE( _e, expr *, expr )

#define EXPR_LIST_APPEND( _blk, _list, _e ) \
        if( (_list) != NULL ){ \
		EXPR_APPEND( _blk, _list, _e ); \
        } \
        else { \
                (_blk)->exprs_head = (_blk)->exprs_tail = (_e); \
        }

#define EXPR_LIST_PREPEND( _blk, _list, _e ) \
        if( (_list) != NULL ){ \
		EXPR_PREPEND( _blk, _list, _e ); \
        } \
        else { \
                (_blk)->exprs_head = (_blk)->exprs_tail = (_e); \
        }

#define EXPR_APPEND( _blk, _cur, _new ) \
        DLL_INSERT_AS_NEXT( (_cur), (_new), expr *, expr ); \
        if( (_blk)->exprs_tail == (_cur) ){ \
                (_blk)->exprs_tail = (_new); \
        }

#define EXPR_PREPEND( _blk, _cur, _new ) \
	DLL_INSERT_AS_PREV( (_cur), (_new), expr *, expr ); \
        if( (_blk)->exprs_head == (_cur) ){ \
                (_blk)->exprs_head = (_new); \
        }

#define BLK_ADD_EXPR( _blk, _e ) \
        if( (_blk)->exprs_tail != 0 ){ \
                DLL_INSERT_AS_NEXT( (_blk)->exprs_tail, (_e), expr *, expr ); \
        } \
        else { \
                (_blk)->exprs_head = (_e); \
        } \
        (_blk)->exprs_tail = (_e)

#define GENERATED_BLK_ADD_EXPR( _blk, _e ) \
	if( EXPR_IS_BRANCH( (_blk)->exprs_tail ) && (_blk)->exprs_tail->val.label->type == LAB_block_target ){ \
		EXPR_LIST_PREPEND( (_blk), (_blk)->exprs_tail, (_e) ); \
	} \
	else{ \
		BLK_ADD_EXPR( (_blk), (_e) ); \
	}

#define CSE_LIST_ADD( _list, _e ) \
        (_e)->exprs_by_op = *(_list); \
        *(_list) = (_e)

#define EXPR_SET_SRC_REG( _e, _r ) \
	(_e)->flags |= (SE_in_reg | SE_generated); \
	(_e)->loc.reg.base = (_r)

#define EXPR_SET_SRC_SLOT( _e ) \
	(_e)->flags |= (SE_in_frame_slot | SE_generated)

#define EXPR_SET_DST_REG( _e, _r ) \
	(_e)->flags |= (SE_in_reg); \
	(_e)->loc.reg.base = (_r)
	
#define EXPR_SET_DST_SLOT( _e, _o ) \
	(_e)->flags |= (SE_in_arg_slot); \
	(_e)->loc.slot.arg =  (_o)

#define EXPR_copy_reg( _d, _s ) \
	(_d)->loc.reg.base = (_s)->loc.reg.base; \
	(_d)->loc.reg.pair = (_s)->loc.reg.pair
#define EXPR_assign_reg( _e, _base, _pair ) \
	(_e)->loc.reg.base = (_base); \
	(_e)->loc.reg.pair = (_pair)
#define EXPR_copy_slot( _d, _s ) \
	(_d)->loc.slot.frame = (_s)->loc.slot.frame
#define EXPR_copy_location( _d, _s ) \
	bcopy( &(_s)->loc, &(_d)->loc, sizeof(location_info) ) 

#define EXPR_RESOLVE_ORIGINAL( _e ) ( (_e) != NULL && ( (_e)->op == OP_copy || (_e)->op == OP_move ) ? (_e)->a[1] : (_e) )

#define ADD_disable_cp		0x1
#define ADD_disable_cse		0x2

#define EXPR_assign_class( _d, _c ) (_d)->class = (_c)

#endif
