/*
 * expr.c
 * Instruction-level sequencing
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "env.h"
#include "lookup.h"
#include "code-analyse.h"
#include "locks.h"
#include "code.h"
#include "nop.h"
#include "bytecode.h"
#include "access.h"
#include "external.h"
#include "soft.h"
#include "objects.h"
#include "support.h"
#include "engine.h"
#include "reduce.h"
#include "itypes.h"

static Hjava_lang_Class * classIx[8] = {
        &_Jv_booleanClass, &_Jv_charClass, &_Jv_floatClass, &_Jv_doubleClass,
       &_Jv_byteClass, &_Jv_shortClass, &_Jv_intClass, &_Jv_longClass
};

#ifdef USE_BYTECODES_DEF
static const uint8 prim2type [8] = { PRIM_int, PRIM_long, PRIM_float, PRIM_double, PRIM_ref, PRIM_byte, PRIM_char, PRIM_short };
static const char prim2sig [8] = { 'I', 'J', 'F', 'D', 'L', 'B', 'C', 'S' };
static unsigned char sig2type( char c, boolean strict )
{
        if( strict ){
                switch( c ){
                case 'V':
                        return PRIM_void;
                case 'L':
                case '[':
                        return PRIM_ref;
                case 'I':
                        return PRIM_int;
                case 'S':
                        return PRIM_short;
                case 'B':
                case 'Z':
                        return PRIM_byte;
                case 'C':
                        return PRIM_char;
                case 'F':
                        return PRIM_float;
                case 'J':
                        return PRIM_long;
                case 'D':
                        return PRIM_double;
                }
        }
        else{
                switch( c ){
                case 'V':
                        return PRIM_void;
                case 'L':
                case '[':
                        return PRIM_ref;
                case 'I':
                case 'S':
                case 'B':
                case 'Z':
                case 'C':
                        return PRIM_int;
                case 'F':
                        return PRIM_float;
                case 'J':
                        return PRIM_long;
                case 'D':
                        return PRIM_double;
                }
        }
        ABORT();
        return PRIM_void;
}
#endif
static const short size2shift [9] = { 0, 0, 1, 0, 2, 0, 0, 0, 3 };

extern Hjava_lang_Class * SoftcallClass;
extern void EBB_generate_exprs( block_list * );
extern expr_frame * EXPR_copy_call_frame( expr_frame * );
extern void DFG_prologue( cfg_info * cfg );
extern void EBB_link_types( instr_block *, block_list *, expr **, int, boolean );
extern void BLK_process_rescan( instr_block * );
extern void EXPR_repoint( instr_block *, expr *, int, boolean );
extern expr * EXPR_new_copy( expr *, instr_block *, boolean );
extern expr * EXPR_new_reg_src( int, instr_block * );

#define CHECK_TYPE( _e, _cond ) verify_check( _cond, cfg->fd, "[%d]: Unexpected type: %x != %c\n", (_e)->eid, (_e)->type, type )

static expr *
EXPR_build_instruction_from_call( expr *** stack, instr_block * blk, Hjava_lang_Class * class, Method * method )
{
        expr * new_e, * cur_e, * orig_e, ** stack_p;
        char type, * p, * tp;
        int ix;
        cfg_info * cfg;

        cfg = blk->cfg;
        new_e = EXPR_new();
        new_e->class = class;
        new_e->val.method = method;
        stack_p = *stack;
        stack_p -= sigItemCount( method->signature, true );
        *stack = stack_p;
        ix = 0;
        for( p = method->signature; (tp = sigItemType( &p )) != NULL; ){
                type = *tp;
                stack_p++;
                if( type == 'D' || type == 'J' ){
                        stack_p++;
                }
                cur_e = new_e->a[ix++] = *stack_p;
                orig_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                switch( type ){
                case 'D': CHECK_TYPE( cur_e, cur_e->type == PRIM_double ); break;
                case 'F': CHECK_TYPE( cur_e, cur_e->type == PRIM_float ); break;
                case 'J': CHECK_TYPE( cur_e, cur_e->type == PRIM_long ); break;
                case 'L': 
                case '[': CHECK_TYPE( cur_e, cur_e->type == PRIM_ref 
                                     || (cur_e->type == PRIM_int && orig_e->op == OP_const_0) ); 
                        EXPR_reduce_class( cur_e, getClassFromSignature( tp, class->loader ) );
                        break;
                case 'I': CHECK_TYPE( cur_e, cur_e->type > 0
                                      && cur_e->type <= PRIM_ref ); break;
                case 'S': CHECK_TYPE( cur_e, cur_e->type == PRIM_short
                                      || (cur_e->type == PRIM_ref && orig_e->op == OP_const_0)
                                      || cur_e->type == PRIM_int ); break;
                case 'C': CHECK_TYPE( cur_e, cur_e->type == PRIM_char
                                      || (cur_e->type == PRIM_ref && orig_e->op == OP_const_0)
                                      || cur_e->type == PRIM_int ); break;
                case 'Z': 
                case 'B': CHECK_TYPE( cur_e, cur_e->type == PRIM_byte
                                      || (cur_e->type == PRIM_ref && orig_e->op == OP_const_0)
                                      || cur_e->type == PRIM_int ); break;
                }
        }
        assert( ix <= 2 );
        new_e->op = ix == 2 ? OP_native_instr_2a : OP_native_instr_1a;
        return new_e;
}

static expr_frame *
EXPR_build_call_frame( Method * method, Hjava_lang_Class * class, char * signature, expr * obj, boolean native, expr ** stack_p, expr ** locals, instr_block * blk, int op )
{
        int sz, ix;
        expr_frame * frame = NULL;
        arg_info * arg_p;
        expr * cur_e, * tmp_e, * new_e, * orig_e;
        char type, * p, * tp;
        instr_block * ebb_root;
        cfg_info * cfg;
        FILE * fd;

        cfg = blk->cfg;
        fd = cfg->fd;
        sz = (obj == NULL) ? 0 : 1;
        if( native ){
                sz++;
                if( obj == NULL ){
                        sz++;
                }
        }
LOG(	fprintf( fd, "EXPR_build_call_frame(): %s.%s%s\n", 
                class->name, method->name, method->signature ); );
        sz += sigItemCount( signature, false );
        stack_p -= sigItemCount( signature, true );
LOG(	fprintf( fd, "sz=%d stack=%d\n", sigItemCount( signature, false ), sigItemCount( signature,true ) ) ); 
#ifdef  CHECK_STACK_LIMIT 
        frame = gc_malloc_fixed( sizeof(expr_frame) + sz * sizeof(arg_info) );
        frame->sz = sz + 1;
#else
        frame = gc_malloc_fixed( sizeof(expr_frame) + (sz - 1) * sizeof(arg_info) );
        frame->sz = sz;
#endif
        frame->method = method;
        frame->class = class;
        ix = 0;
        if( native ) {
                frame->flags |= CALL_jni;
                ebb_root = blk->ebb_root;
                move_ref_const( tmp_e, &Kaffe_JNIEnv);
LOG(            fprintf( fd, "(%x) ", tmp_e->type ); );
                frame->a[0].expr = tmp_e;
                ix++;
                if( obj == NULL ){
                        move_ref_const( tmp_e, class);
LOG(                    fprintf( fd, "(%x) ", tmp_e->type ); );
                        frame->a[1].expr = tmp_e;
                        ix++;
                }
        }
        if (obj != NULL) {
LOG(            fprintf( fd, "obj=[%d] %x ", obj->eid, obj->type ); );
                type = obj->type;
                orig_e = EXPR_RESOLVE_ORIGINAL( obj );
                CHECK_TYPE( obj, obj->type == PRIM_ref
                           || (obj->type == PRIM_int && orig_e->op == OP_const_0) );
                frame->a[ix].expr = obj;
                EXPR_reduce_class( obj, class );
                ix++;
        }
        else{
                frame->flags |= CALL_static;
        }
        arg_p = &frame->a[ix];
        for( p = method->signature; (tp = sigItemType( &p )) != NULL; ){
                type = *tp;
                stack_p++;
                if ((type == 'D') || (type == 'J')) {
                        stack_p++;
                }
                cur_e = arg_p->expr = *stack_p;
                orig_e = EXPR_RESOLVE_ORIGINAL( cur_e );
LOG(            fprintf( fd, "| %c [%d] %x ", type, cur_e->eid, cur_e->type ); );
                switch( type ){
                case 'D': CHECK_TYPE( cur_e, cur_e->type == PRIM_double ); break;
                case 'F': CHECK_TYPE( cur_e, cur_e->type == PRIM_float ); break;
                case 'J': CHECK_TYPE( cur_e, cur_e->type == PRIM_long ); break;
                case 'L': 
                case '[': CHECK_TYPE( cur_e, cur_e->type == PRIM_ref 
                                     || (cur_e->type == PRIM_int && orig_e->op == OP_const_0) ); 
                        EXPR_reduce_class( cur_e, getClassFromSignature( tp, class->loader ) );
                        break;
                case 'I': CHECK_TYPE( cur_e, cur_e->type > 0
                                      && cur_e->type <= PRIM_ref ); break;
                case 'S': CHECK_TYPE( cur_e, cur_e->type == PRIM_short
                                      || (cur_e->type == PRIM_ref && orig_e->op == OP_const_0)
                                      || cur_e->type == PRIM_int ); break;
                case 'C': CHECK_TYPE( cur_e, cur_e->type == PRIM_char
                                      || (cur_e->type == PRIM_ref && orig_e->op == OP_const_0)
                                      || cur_e->type == PRIM_int ); break;
                case 'Z': 
                case 'B': CHECK_TYPE( cur_e, cur_e->type == PRIM_byte
                                      || (cur_e->type == PRIM_ref && orig_e->op == OP_const_0)
                                      || cur_e->type == PRIM_int ); break;
                }
                arg_p++;
        }
        tp = sigReturnType( signature );
        type = *tp;
        if( type == 'L' || type == '[' ){
                frame->ret = getClassFromSignature( tp, class->loader );
        }
#if defined( CHECK_STACK_LIMIT ) && !defined( FAKE_STACK_LIMIT )
LOG(    fprintf( fd, "<%x> ", PRIM_int ); );
        cur_e = frame->a[frame->sz-1].expr = locals[cfg->env.method.method->bcode->localsz];
        CHECK_TYPE( cur_e->type > 0 && cur_e->type <= PRIM_int );
#endif
LOG(    fprintf( fd, ")\n" ); );
#ifdef OPTIMIZE_INLINE
        CFG_inline( frame, method, class, blk, op );
#endif
        return frame;
        /* FIXME: Free the frame */
}

static expr_frame *
EXPR_build_call_frame_native( int sz, int flags, ... )
{
	va_list va_p;
        expr_frame * frame;
        arg_info * arg_p;

        frame = gc_malloc_fixed( sizeof(expr_frame) + (sz - 1) * sizeof(arg_info) );
        frame->flags |= CALL_vm | flags;
        frame->sz = sz;
        va_start(va_p, flags);
        arg_p = frame->a;
        for (; --sz >= 0;) {
                arg_p->expr = va_arg(va_p, expr*);
                arg_p++;
        }
        va_end(va_p);
        return frame;
        /* FIXME: Free the frame */
}

static expr_frame *
EXPR_build_call_frame_multianewarray( expr * sz_e, expr * class_e, expr ** dims_p )
{
        expr_frame * frame;
        arg_info * arg_p;
        int sz;

        sz = sz_e->val.i_const;
        frame = gc_malloc_fixed( sizeof(expr_frame) + (sz + 1) * sizeof(arg_info) );
        frame->flags |= CALL_vm;
        frame->sz = sz + 2;
        frame->a[0].expr = class_e;
        frame->a[1].expr = sz_e;
        arg_p = &frame->a[sz+1];
        for( ; --sz >= 0; ){
                arg_p->expr = *dims_p;
                arg_p--;
                dims_p--;
        }
        /* Set sp */
        return frame;
        /* FIXME: Free the frame */
}

/*
 * Assign a expression to be stored in a stack location
 */
void
EXPR_assign_stack( expr * cur_e, int ix, instr_block * blk )
{
        cfg_info * cfg;
        int sz, six;

        cfg = blk->cfg;
        assert( ix >= 0 && ix < cfg->env.method.method->bcode->stacksz );
        six = ix + cfg->env.method.n_locals;
LOG(    fprintf( cfg->fd, "\t[%d] --> S%d\n", cur_e->eid, ix ); );

        assert( cfg->env.method.caller == NULL );
        sz = PRIM_SIZE( cur_e->type ) == 8 ? 2 : 1;
        if( sz > cfg->env.method.cframe.map[six].size ){
                cfg->env.method.cframe.map[six].size = sz;
        }
        if( cur_e->op == OP_copy 
           || EXPR_IS_CONST( cur_e )
           || cur_e->flags & (SE_slot | SE_inline_return) ){
                if( cur_e->flags & SE_slot
                   && cur_e->loc.slot.frame - cfg->env.method.n_locals != ix ){
                        return;
                }
                /* Reassign to prevent incorrect aliasing at EBB boundaries */
                cur_e = EXPR_new_move( cur_e, blk );
                assert( blk->exprs_head != NULL );
                GENERATED_BLK_ADD_EXPR( blk, cur_e );
                blk->stack.out[ix] = cur_e;
                blk->stats.n_move++;
        }
        cur_e->flags |= SE_slot;
        cur_e->loc.slot.frame = six;
        cur_e->flags &= ~SE_in_frame_slot;
        cur_e->uses.block++;
}

/*
 * Assign a expression to be stored in a given local
 */
expr *
EXPR_assign_local( expr * cur_e, int ix, expr * last_local, instr_block * blk, cfg_info * cfg )
{

        assert( ix >= 0 && ix < cfg->env.method.method->bcode->localsz );
LOG(    fprintf( cfg->fd, "\t[%d] --> L%d\n", cur_e->eid, ix ); );

#if defined( CHECK_STACK_LIMIT ) && !defined( FAKE_STACK_LIMIT )
        if( ix >= cfg->env.global.n_args - 1 ){
                ix++;
        }
#endif

        cfg->env.method.cframe.map[ix].size = PRIM_SIZE( cur_e->type ) == 8 ? 2 : 1;
        if( cur_e->op == OP_copy 
            || EXPR_IS_CONST( cur_e )
            || cur_e->flags & (SE_slot | SE_inline_return) 
            || last_local->eid > cur_e->eid ){
                if( cur_e->flags & SE_slot
                   && cur_e->loc.slot.frame == ix ){
                        return cur_e;
                }
                cur_e = EXPR_new_move( cur_e, blk );
                BLK_ADD_EXPR( blk, cur_e );
                blk->locals.out[ix] = cur_e;
                blk->stats.n_move++;
        }
        if( cfg->env.method.caller == NULL ){
                cur_e->loc.slot.frame = ix;
                cur_e->last_local = last_local;
                assert( cur_e->last_local != blk->locals.out[ix] );
                cur_e->flags |= SE_slot | SE_last_local;
                cur_e->flags &= ~SE_in_frame_slot;
        }
        return cur_e;
}

extern void BLK_cvt_to_cmov( instr_block * );

void
BLK_generate_exprs( instr_block * blk, instr_block * ebb_root ) 
{
        cfg_info * cfg;
        int stacksz, localsz;
        bytecode* code, opcode;
        Method * method;
        Hjava_lang_Class * class;
        boolean is_ebb_root;
        char * cname;

        expr ** stack, ** stack_p, ** locals, ** by_op_e, ** tgt;
        expr * cur_e, * new_e, * mtable, * call_e, * d_e, * orig_e, * copy_e;
        expr * t_e, * t2_e, * t3_e, * t4_e, * tmp, * tmp2, * last_local;
        slot * base_s;
        instr_block * cur_b, * succ0_b;
#ifdef OPTIMIZE_GCSE
        instr_block * tmp_b;
#endif
#ifdef OPTIMIZE_INLINE
        int sz;
        expr ** save_stack_p, ** local_p;
#endif
#ifdef OPTIMIZE_CMOV
        int length;
#endif
        instr_block * pred_b;
        uint32 pc, npc, pc_out, type;
        int idx, n, m;
	jint low, high;
        jvalue tmpl;
        expr_frame * frame;
        int first_expr_id;
        boolean was_call, match, generated_call;
        int sp_in;

	Hjava_lang_Class * cls_i;
	callInfo meth_i;
	fieldInfo fld_i;

        FILE * fd;
#ifdef PROFILE
        dep_info * di;
        loc_info * li;
#endif

        frame = 0;
        new_e = 0;
        generated_call = false;
        was_call = false;
        cfg = blk->cfg;
        fd = cfg->fd;

#ifdef OPTIMIZE_GCSE
        tmp_b = NULL;
#endif
#ifdef OPTIMIZE_INLINE
        CFG_set_method_env( cfg, blk->env );
#endif
LOG(    fprintf( fd, "BLK_generate_exprs(): B%d:%d pc = %d - %d\n", blk->bid, blk->env->inline_depth, blk->pc.in, blk->pc.out ); );
LOG(    fflush( fd ); );
        method = cfg->env.method.method;
        class = cfg->env.method.class;
        code = method->bcode->code;
        if( method->bcode->localsz == 0 ){
                assert( method->accflags & ACC_STATIC );
        }
        stacksz = method->bcode->stacksz;
        localsz = cfg->env.method.n_locals;

        locals = blk->locals.out;
        stack = blk->stack.out;
        sp_in = blk->sp.in;
        is_ebb_root = ( blk == ebb_root ) ? true : false;
        first_expr_id = ++cfg->env.global.id;

        if( is_ebb_root 
#ifdef OPTIMIZE_INLINE
           && !cfg->env.method.inlined
#endif
           ){
                base_s = gc_malloc_fixed( (localsz+stacksz) * sizeof(slot) );
                for (n = localsz; --n >= 0;) {
                        cur_e = locals[n] = (expr*)&base_s[n];
                        cur_e->flags = SE_slot;
                        if( blk != cfg->env.method.root ){
                                locals[n]->flags |= SE_in_frame_slot;
                        }
#if defined( CHECK_STACK_LIMIT ) && !defined( FAKE_STACK_LIMIT )
                        if( n < cfg->env.global.n_args - 1 ){
                                cur_e->loc.slot.frame = n;
                        }
                        else{
                                if( n == localsz - 1 ){
                                        cur_e->loc.slot.frame = cfg->env.global.n_args - 1;
                                }
                                else{
                                        cur_e->loc.slot.frame = n + 1;
                                }
                        }
#else
                        cur_e->loc.slot.frame = n;
#endif
                        cur_e->eid = first_expr_id;
                }
                for (n = sp_in; --n >= 0;) {
                        cur_e = stack[n] = (expr*)&base_s[localsz+n];
                        cur_e->flags = SE_slot | SE_in_frame_slot;
                        cur_e->eid = first_expr_id;
                        cur_e->loc.slot.frame = n + localsz;
                }
#ifdef OPTIMIZE_GCSE
                for( n = cfg->env.global.globals.max; --n >= 0; ){
                        assert( blk->globals.in[n] != NULL );
                        blk->globals.in[n]->eid = first_expr_id;
                }
#endif
        }
        else{
#ifdef OPTIMIZE_INLINE
                if( cfg->env.method.inlined && cfg->env.method.root == blk ){
                        cfg->env.method.first_id = first_expr_id;
                        frame = cfg->env.method.frame;
                        sz = frame->sz;
#ifdef CHECK_STACK_LIMIT 
                        sz--;
#endif
                        local_p = locals;
                        for( n = localsz; --n >= sz; ){
                                locals[n] = cur_e = EXPR_new();
                                cur_e->eid = first_expr_id;
                                cur_e->flags = SE_slot;
                                cur_e->loc.slot.frame = n;
                                BLK_ADD_EXPR( blk, cur_e );
                        }
                        for( n = 0; n < sz; n++ ){
                                cur_e = frame->a[n].expr;
                                cur_e->uses.block--;
                                // *local_p = EXPR_new_copy( cur_e, blk );
                                // BLK_ADD_EXPR( blk, locals[n] );
                                *local_p = cur_e;
                                if( (*local_p)->type == PRIM_long || (*local_p)->type == PRIM_double ){
                                        local_p++;
                                }
                                local_p++;
                        }
                        assert( sp_in == 0 );
#ifdef OPTIMIZE_GCSE
                        METHOD_new_globals( &cfg->env.method, 0, cfg->env.global.globals.max );
#endif
                }
                else{
#endif
#ifdef OPTIMIZE_GCSE
                for( n = cfg->env.global.globals.max; --n >= 0; ){
                        blk->globals.in[n]->eid = first_expr_id;
                }
#endif
                assert( blk->pred.length == 1 );
                pred_b = blk->pred.blocks[0];
                for( n = localsz; --n >= 0; ){
                        locals[n] = EXPR_new_copy( pred_b->locals.out[n], blk, true );
                        BLK_ADD_EXPR( blk, locals[n] );
                }
                for( n = sp_in; --n >= 0; ){
                        cur_e = pred_b->stack.out[n];
                        if( cur_e == NULL ){
                                continue;
                        }
                        orig_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                        tmp = EXPR_RESOLVE_ORIGINAL( locals[orig_e->loc.slot.frame] );
                        if( orig_e->flags & SE_slot 
                           && orig_e->loc.slot.frame < cfg->env.method.n_locals
                           && tmp == orig_e ){
                                stack[n] = locals[orig_e->loc.slot.frame];
                        }
                        else{
                                match = false;
                                for( m = sp_in; --m > n; ){
                                        if( cur_e == pred_b->stack.out[m] ){
                                                stack[n] = stack[m];
                                                match = true;
                                                break;
                                        }
                                }
                                if( !match ){
                                        stack[n] = EXPR_new_copy( cur_e, blk, true );
                                        BLK_ADD_EXPR( blk, stack[n] );
                                }
                        }
                }
        }
#ifdef OPTIMIZE_INLINE
        }
#endif
        bcopy( locals, blk->locals.in, (localsz+stacksz) * sizeof(expr*) );
        if( is_ebb_root 
#ifdef OPTIMIZE_INLINE
           && !cfg->env.method.inlined
#endif
           ){
                if( blk == cfg->env.method.root ){
                        DFG_prologue( cfg );
                        new_e = EXPR_new();
                        new_e->op = OP_prologue;
                        EXPR_add( new_e, 0, blk, ebb_root );
#if defined( CHECK_STACK_LIMIT ) && !defined( FAKE_STACK_LIMIT )
                        check_stack_overflow( locals[localsz-1] );
#endif
                        if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) {
                                softcall_trace_call( method, class );
                        }
                        softcall_profile_call( );
                }
                else{
                        if( blk->flags & BLK_catch ){
                                EBB_link_types( blk, &blk->try, blk->locals.in, localsz + sp_in - 1, false );
                                blk->stack.in[0]->type = PRIM_ref;
                                new_e = EXPR_new();
                                new_e->op = OP_eprologue;
                                EXPR_add( new_e, 0, blk, ebb_root );
                                softcall_exception_info( stack[0] );
                        }
                        else{
                                EBB_link_types( blk, &blk->pred, blk->locals.in, localsz + sp_in, false );
                        }
                }
        }
        stack_p = &stack[sp_in-1];
        
        /* Initialize incoming arguments */
        if (blk->pc.in == 0) {
                monitor_enter();
        }
        pc_out = blk->pc.out;
        if( blk->flags & BLK_loop_header ){
                softcall_profile_loop_preheader();
        }
        else if( blk->flags & BLK_loop_entry ){
                softcall_profile_loop_iteration();
        }
        if( ( blk->pc.out == 0 && blk->pc.in == 0 && blk->flags & BLK_prologue )
           || blk->flags & BLK_inserted_header ){
        }
        else{
                for (pc = blk->pc.in; pc <= pc_out; pc = npc) {
                        opcode = code[pc];
                        npc = pc + iLen[opcode];
                        switch (opcode) {
#ifdef USE_BYTECODES_DEF
#include "bytecodes.def"
#else
#include "kaffe.def"
#endif
                        default:
LOG(                    fprintf( fd, "Unknown bytecode 0x%x @ %d \n", opcode, pc); );
                        break;
                        }
                }
        }
        if( blk->try.length > 0 ){
                if( METHOD_IS_SYNCHRONISED( method ) && !( EXPR_IS_BRANCH( blk->exprs_tail ) ) ){
                        monitor_exit();
                        softcall_athrow( stack(0) );
                }
        }

        if( generated_call ){
                cfg->flags &= ~CFG_is_leaf;
        }
        BLK_process_rescan( blk );
        /* Clear out unused stack expressions ? */

        /* Make sure stack operands survive */
        if( blk->flags & BLK_succ_not_in_ebb ){
                for( n = blk->sp.out; --n >= 0; ){
                        EXPR_assign_stack( blk->stack.out[n], n, blk );
                        type = blk->stack.out[n]->type;
                        if( type == PRIM_double || type == PRIM_long ){
                                n--;
                        }
                }
        }
#ifdef OPTIMIZE_CMOV
        if( !is_ebb_root && blk->stats.n_other == 0 ){
                if( blk->stats.n_move == 1 ){
                        if( blk->exprs_tail->op == OP_branch ){
                                kprintf( "BR_CMOV: B%d\n", blk->bid );
                                BLK_cvt_to_cmov( blk );
                        }
                        else if( blk->exprs_tail->op == OP_move ){
                                kprintf( "CMOV: B%d\n", blk->bid );
                                BLK_cvt_to_cmov( blk );
                        }
                }
                else if( blk->stats.n_move > 1 ){
                //        kprintf( "CMOV?: B%d\n", blk->bid );
                }
        }
#endif
#ifdef OPTIMIZE_LOOPS
        if( blk->flags & BLK_inserted_header ){
                if( blk->block.next != blk->succ.blocks[0] ){
                        new_e = EXPR_new();
                        new_e->op = OP_branch;
                        new_e->val.label = LAB_new_block_branch( cfg, new_e, blk->succ.blocks[0] );
                        new_e->eid = ++cfg->env.global.id;
                        BLK_ADD_EXPR( blk, new_e );
                }
                for( n = blk->pred.length; --n >= 0; ){
                        pred_b = blk->pred.blocks[n];
                        cur_e = pred_b->exprs_tail;
                        if( EXPR_IS_BRANCH( cur_e ) 
                           && cur_e->val.label->to.block == blk->succ.blocks[0] ){
                                cur_e->val.label->to.block = blk;
                        }
                }
        }
#endif
        if( blk->exprs_head == NULL ){
                new_e = EXPR_new();
                new_e->eid = ++cfg->env.global.id;
                BLK_ADD_EXPR( blk, new_e );
        }

        blk->stage = STG_generated_exprs;
        if( blk->succ.length == 2 ){
                /* Force follow through to be 1st successor */
                succ0_b = blk->succ.blocks[0];
                if( succ0_b == blk->block.next ){
                        blk->succ.blocks[0] = blk->succ.blocks[1];
                        blk->succ.blocks[1] = succ0_b;
                }
        }
#ifdef OPTIMIZE_CMOV
        length = blk->succ.length;
#endif
//        for( n = blk->succ.length; --n >= 0; ){
        for( n = 0; n < blk->succ.length; n++ ){
                cur_b = blk->succ.blocks[n];
                if( cur_b->ebb_root == ebb_root
                    && cur_b->stage != STG_generated_exprs ){
                        BLK_generate_exprs( cur_b, ebb_root );
                }
#ifdef OPTIMIZE_CMOV
                if( blk->succ.length != length ){
                        length = blk->succ.length;
                        n = -1;
                }
#endif
        }

        /* Increment uses for OP_copy expressions actually used */
        cur_e = blk->exprs_by_op[OP_copy];
        while( cur_e != NULL ){
                copy_e = cur_e;
                if( copy_e->uses.block > 0 ){
                        while( copy_e != NULL && copy_e->op == OP_copy ){ /* FIXME: Recursion necessary ? */
                                copy_e->a[0]->uses.other++; /* FIX ?? */
                                // copy_e->a[0]->uses.block++; /* FIX ?? */
                                copy_e = copy_e->a[0];
                        }
                }
#if 1 /* FIXME: make more efficient */
                copy_e = cur_e;
                orig_e = EXPR_RESOLVE_ORIGINAL( copy_e );
                if( copy_e->uses.block == 0 && copy_e->uses.other == 0 && orig_e->op != OP_nop ){
                        by_op_e = &blk->exprs_by_op[orig_e->op];
                        copy_e = *by_op_e;
                        while( copy_e != NULL ){
                                if( copy_e->op == OP_copy && copy_e->uses.block == 0 && copy_e->uses.other == 0 ){
                                        EXPR_REMOVE( blk, copy_e );
                                        *by_op_e = copy_e->exprs_by_op;
                                        EXPR_free( copy_e );
                                }
                                else{
                                        by_op_e = &copy_e->exprs_by_op;
                                }
                                copy_e = *by_op_e;
                        }
                }
#endif
                cur_e = cur_e->val.expr;
        }
}
