/* 
 * print.c
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "code.h"
#include "support.h"
#include "env.h"
#include "locks.h"
#include "soft.h"
#include "access.h"

#define PRINT_EXPRS
#define MAX_BUFFER 1000

#ifdef  JIT_DBG
char* op_names[] = {
	"nop",
	"const_m1",
	"const_0",
	"const_1",
	"const",
	"move",
	"load",
	"store",
	"add",
	"sub",
	"mul",
	"div",
	"rem",
	"neg",
	"lsl",
	"lsr",
	"asr",
	"and",
	"or",
	"xor",
	"cvt",
	"set_cc",
	"cbranch_eq",
	"cbranch_ne",
	"cbranch_lt",
	"cbranch_ge",
	"cbranch_gt",
	"cbranch_le",
	"cbranch_ult",
	"branch",
	"branch_indirect",
	"returnarg",
	"cmp",
	"call",
	"target",
	"prologue",
	"epilogue",
	"eprologue",
        "copy",
        "inline_data",
        "get_arg",
        "preload",
        "replace",
        "specialize_sync",
        "tableswitch",
        "lookupswitch",
        "native_instr_1a",
        "native_instr_2a",
	"cmov_eq",
	"cmov_ne",
	"cmov_lt",
	"cmov_ge",
	"cmov_gt",
	"cmov_le",
	"cmov_ult"
};

#define soft_label( _f ) { soft_##_f, #_f }
static softname_info soft_names[] = {
	{ lockObject, "monitorenter" },
	{ unlockObject, "monitorexit" },
	{ _lockMutex, "monitorenter" },
	{ _unlockMutex, "monitorexit" },
	{ _Jv_AllocObject, "new" },
	{ _Jv_NewArray, "newarray" },
	{ _Jv_NewObjectArray, "anewarray" },
	{ jit4_soft_multianewarray, "multianewarray" },
        soft_label( athrow ),
	{ _Jv_CheckCast, "checkcast" },
	{ _Jv_IsInstanceOf, "instanceof" },
	{ _Jv_InitClass, "initialise_class" },
        soft_label( nosuchfield ),
        soft_label( nosuchmethod ),
        soft_label( nosuchclass ),
        soft_label( lookupmethod ),
        soft_label( lazylookupmethod ),
        soft_label( badarrayindex ),
        soft_label( get_thrown_exception ),
        soft_label( nullpointer ),
        soft_label( stackoverflow ),
        soft_label( lcmp ),
        soft_label( fcmpl ),
        soft_label( fcmpg ),
        soft_label( dcmpl ),
        soft_label( dcmpg ),
	{ gcWriteRef, "writeref" },
        soft_label( trace ),
        soft_label( end ),
        soft_label( fadd ),
        soft_label( faddl ),
        soft_label( fsub ),
        soft_label( fsubl ),
        soft_label( fdiv ),
        soft_label( fdivl ),
        soft_label( fmul ),
        soft_label( fmull ),
        soft_label( lmul ),
        soft_label( ldiv ),
        soft_label( lrem ),
        soft_label( rem ),
        soft_label( lshll ),
        soft_label( lshrl ),
        soft_label( ashrl ),
        soft_label( cvtfl ),
        soft_label( cvtdl ),
        soft_label( cvtlf ),
        soft_label( cvtld ),
	{ finishJNIcall, "finishJNIcall" },
#ifdef PROFILE
        { PROF_call, "profile_call" },
        { PROF_return, "profile_return" },
        { PROF_get, "profile_get" },
        { PROF_put, "profile_put" },
        { PROF_arrayload, "profile_arrayload" },
        { PROF_arraystore, "profile_arraystore" },
#endif
	{ 0, "null" }
};

#if defined( CHECK_STACK_LIMIT ) && !defined( FAKE_STACK_LIMIT )
static int
frame_to_local( int frame, cfg_info * cfg ){
        int n_args = cfg->env.global.n_args;
        int local = frame < n_args
                ? ( frame == n_args - 1 ? cfg->env.method.method->bcode->localsz : frame )
                        : frame - 1;
        return local;
}
#else
#define frame_to_local( _frame, _cfg ) (_frame)
#endif

const char *
CFG_print_type( int type )
{
#define CFG_print_case( _c ) case CFG_##_c: return #_c; break
        switch( type ){
                CFG_print_case( branch );
                CFG_print_case( jump );
                CFG_print_case( switch );
                CFG_print_case( return );
                CFG_print_case( jsr );
                CFG_print_case( ret );
        }
        return "???";
}

void
REG_print( FILE * fd, reg_info * r, int type, cfg_info * cfg )
{
        char * op;
        expr * e;

        e = r->expr;
        switch( type & 0x7 ){
        case REGOP_new: op = "new"; break;
        case REGOP_assign: op = "force assign"; break;
        case REGOP_free: op = "free"; break;
        default: op = "?"; break;
        }
        if( e != NULL ){
                if( e->flags & SE_slot 
#ifdef OPTIMIZE_INLINE
                && e->eid >= cfg->env.method.first_id
#endif
                   ){
#ifdef OPTIMIZE_GCSE
                        if( e->flags & SE_global ){
                                fprintf( fd, "\t[G%d]: %s %s ", e->loc.slot.frame, op, reg_names[r->num] );
                        } 
                        else
#endif
                        if( e->loc.slot.frame < cfg->env.method.n_locals ){
                                fprintf( fd, "\t[L%d]: %s %s ", frame_to_local( e->loc.slot.frame, cfg ), op, reg_names[r->num] );
                        }
                        else{
                                fprintf( fd, "\t[S%d]: %s %s ", e->loc.slot.frame - cfg->env.method.n_locals, op, reg_names[r->num] );
                        }
                }
                else {
                        fprintf( fd, "\t[%d]: %s %s ", e->eid, op, reg_names[r->num] );
                }
                if( type & REGOP_pair ){
                        assert( e->loc.reg.pair == r->num );
                        fprintf( fd, "(p)" );
                }
                else{
                        assert( e->loc.reg.base == r->num );
                }
                if( e->uses.other != 0 ) {
                        fprintf( fd, "{%d,%d} ", e->uses.block, e->uses.other );
                }
                else {
                        fprintf( fd, "{%d} ", e->uses.block );
                }
                if( !(type & REGOP_no_cr) ){
                        fprintf( fd, "\n" );
                }
        }
        else{
                        fprintf( fd, "\t[??]: %s %s ", op, reg_names[r->num] );
                        if( !(type & REGOP_no_cr) ){
                                fprintf( fd, "\n" );
                        }
        }
}

static void
print_type( char ** buf, expr * e )
{
#define print_type_case( _c, _s ) case PRIM_##_c: *buf += sprintf( *buf, #_s " " ); break

        switch (e->type) {
                print_type_case( byte, B );
                print_type_case( char, C );
                print_type_case( short, S );
                print_type_case( int, I );
                print_type_case( double, D );
                print_type_case( float, F );
                print_type_case( long, J );
                print_type_case( ref, R );
                print_type_case( void, V );
                print_type_case( label, L );
        default:
                *buf += sprintf(*buf, "? ");
                break;
        }
}

static void
print_const( char ** buf, instr_block * blk, expr * e ) 
{
        label_info * l;

        switch (e->type) {
        case PRIM_byte: case PRIM_char: case PRIM_short:
        case PRIM_int:
                *buf += sprintf(*buf, "%s %d ", op_names[e->op], 
                                e->val.i_const);
                break;
        case PRIM_double:
                *buf += sprintf(*buf, "%s %llf ", op_names[e->op], 
                                e->val.d_const);
                break;
        case PRIM_float:
                *buf += sprintf(*buf, "%s %f ", op_names[e->op], 
                                e->val.f_const);
                break;
        case PRIM_long:
                *buf += sprintf(*buf, "%s %lld ", op_names[e->op], 
                                e->val.l_const);
                break;
        case PRIM_ref:
                *buf += sprintf(*buf, "%s 0x%x ", op_names[e->op], 
                                (uint)e->val.ref);
                break;
        case PRIM_label:
                l = e->val.label;
                if( l->type == LAB_expr_target ){
                        *buf += sprintf(*buf, "%s addr @ [%d] ", op_names[e->op], 
                                        l->to.expr->eid );
                }
                else if( l->type == LAB_block_target ){
                        *buf += sprintf(*buf, "%s addr @ B%d ", op_names[e->op], 
                                        l->to.block->bid );
                }
                else{
                        *buf += sprintf(*buf, "%s addr @ ?? ", op_names[e->op] );
                }
                break;
        }
}

int
FLAG_print( char * buf, instr_block * blk, expr * e )
{
        char * bp = buf;

#define EXPR_print_flag( _f, _v ) \
        if( e->flags & SE_##_f ){ \
                bp += sprintf( bp, #_v ); \
        }


        if( e == NULL ){
                return 0;
        }
        bp += sprintf( bp, "(" );
        EXPR_print_flag( generated, g );
        EXPR_print_flag( sync, s );
        EXPR_print_flag( null_checked, v );
        EXPR_print_flag( req_fp_reg, f );
        if( e->flags & SE_in_reg ){
                bp += sprintf( bp, "l" );
                bp += sprintf( bp, " %s", reg_names[e->loc.reg.base] );
                if( PRIM_SIZE( e->type ) == 0x08 ){
                        bp += sprintf( bp, " %s", reg_names[e->loc.reg.pair] );
                }
        }
        if( e->flags & SE_in_frame_slot ){
#ifdef OPTIMIZE_GCSE
                if( e->flags & SE_global ){
                        bp += sprintf( bp, " G%d", e->loc.slot.frame );
                }
                else
#endif
                bp += sprintf( bp, " L%d", frame_to_local( e->loc.slot.frame, blk->cfg ) );
        }
        if( e->flags & SE_in_arg_slot ){
                bp += sprintf( bp, " A%d", e->loc.slot.frame );
        }
        assert( !( (e->flags & SE_in_frame_slot) && (e->flags & SE_in_arg_slot) ) );
        if( e->class != NULL ){
                bp += sprintf( bp, " %s", e->class->name );
        }
        bp += sprintf( bp, ") " );
        if( bp - buf == 3 ){
                buf[0] = 0;
                return 0;
        }
        else{
                return bp - buf;
        }
}

int
_SLOT_print( char * buf, instr_block * blk, expr * a )
{
        char * bp = buf;
        expr * o;

        o = EXPR_RESOLVE_ORIGINAL( a );
        if( a != NULL ){
                if( o->flags & SE_slot
#ifdef OPTIMIZE_INLINE
                && a->eid >= blk->env->first_id
#endif
                    ) {
                        bp += sprintf( buf, "[" );
                        if( a != o ){
                                bp += sprintf( bp, "%d|", a->eid );
                        }
#ifdef OPTIMIZE_GCSE
                        if( o->flags & SE_global ){
                                bp += sprintf( bp, "G%d] ", o->loc.slot.frame );
                        } 
                        else
#endif
                        if( o->loc.slot.frame < blk->cfg->env.method.n_locals ){
                                bp += sprintf( bp, "L%d] ", frame_to_local( o->loc.slot.frame, blk->cfg ) );
                        }
                        else{
                                bp += sprintf( bp, "S%d] ", o->loc.slot.frame - blk->cfg->env.method.n_locals );
                        }
                        return bp - buf;
                }
                else {
                        if( a != o ){
                                return sprintf( buf, "[%d|%d] ", a->eid, o->eid );
                        }
                        else{
                                return sprintf( buf, "[%d] ", a->eid);
                        }
                }
        }
        else{
                return sprintf( buf, "[null] ");
        }
}

inline char *
SLOT_print( instr_block * blk, expr * a )
{
        _SLOT_print( blk->cfg->env.global.print_buf, blk, a );
        return blk->cfg->env.global.print_buf;
}

PRINT_FUNCTION( load )
{
#ifdef  USE_RISC_LDST 
        *buf += sprintf(*buf, "%s %d+", op_names[e->op], e->val.i_const );
#else
        *buf += sprintf(*buf, "%s ", op_names[e->op] );
#endif
        *buf += _SLOT_print( *buf, blk, e->a[0] );
}

PRINT_FUNCTION( store )
{
#ifdef  USE_RISC_LDST 
        *buf += sprintf(*buf, "%s %d+", op_names[e->op], e->val.i_const );
#else
        *buf += sprintf(*buf, "%s ", op_names[e->op] );
#endif
        *buf += _SLOT_print( *buf, blk, e->a[0] );
        *buf += sprintf(*buf, "<- ");
        *buf += _SLOT_print( *buf, blk, e->a[1] );
}

PRINT_FUNCTION( unary )
{
        *buf += sprintf(*buf, "%s ", op_names[e->op]);
        *buf += _SLOT_print( *buf, blk, e->a[0] );
}

PRINT_FUNCTION( binary )
{
        *buf += sprintf(*buf, "%s ", op_names[e->op]);
        *buf += _SLOT_print( *buf, blk, e->a[0]);
        *buf += _SLOT_print( *buf, blk, e->a[1]);
}

PRINT_FUNCTION( branch )
{
        if( e->val.ref != NULL ){
                if( e->val.label->type == LAB_expr_target ){
                        *buf += sprintf(*buf, "%s --> [%d] ", op_names[e->op],
                                        e->val.label->to.expr->eid);
                }
                else{
                        *buf += sprintf(*buf, "%s --> B%d ", op_names[e->op],
                                        e->val.label->to.block->bid);
                }
        }
        else {
                *buf += sprintf(*buf, "%s --> ?? ", op_names[e->op]);
        }
}

PRINT_FUNCTION( cbranch )
{
        *buf += sprintf(*buf, "%s ", op_names[e->op]);
        *buf += _SLOT_print( *buf, blk, e->a[0]);
        *buf += _SLOT_print( *buf, blk, e->a[1]);

        if( e->val.ref != NULL ){
                if( e->val.label->type == LAB_expr_target ){
                        *buf += sprintf(*buf, "--> [%d] ", e->val.label->to.expr->eid);
                }
                else {
                        *buf += sprintf(*buf, "--> B%d ", e->val.label->to.block->bid);
                }
        } 
        else {
                *buf += sprintf(*buf, "--> ???");
        }
}

PRINT_FUNCTION( cc_cbranch )
{
        *buf += sprintf(*buf, "%s ", op_names[e->op]);
        *buf += _SLOT_print( *buf, blk, e->a[0]);

        if( e->val.ref != NULL ){
                if( e->val.label->type == LAB_expr_target ){
                        *buf += sprintf(*buf, "--> [%d] ", e->val.label->to.expr->eid);
                }
                else {
                        *buf += sprintf(*buf, "--> B%d ", e->val.label->to.block->bid);
                }
        } 
        else {
                *buf += sprintf(*buf, "--> ???");
        }
}

PRINT_FUNCTION( cmov )
{
        *buf += sprintf(*buf, "%s ", op_names[e->op]);
        *buf += _SLOT_print( *buf, blk, e->a[0]);
        *buf += _SLOT_print( *buf, blk, e->a[1]);

        if( e->val.ref != NULL ){
                *buf += sprintf(*buf, "--> B%d ", e->val.label->to.block->bid);
        } 
        else {
                *buf += sprintf(*buf, "--> ???");
        }
}

PRINT_FUNCTION( call )
{
        expr_frame* frame;
        expr * call_e;
        arg_info * arg_p;
        int n;
        Method * method;

        frame = e->val.frame;
        method = frame->method;
        if( frame->flags & CALL_vm ){
                for( n = 0; soft_names[n].func != NULL; n++ ){
                        call_e = e->a[0];
                        call_e = EXPR_RESOLVE_ORIGINAL( call_e );
                        if (call_e->val.ref == soft_names[n].func) {
                                *buf += sprintf(*buf, "%s ", soft_names[n].name);
                                break;
                        }
                }
                if( soft_names[n].func == NULL ){
                        *buf += sprintf(*buf, "vm_call " );
                }
        }
        else {
                if( method != NULL ){
                        *buf += sprintf(*buf, "%s.%s %s\n\t\t\t", frame->class->name, method->name, method->signature);
                }
                else{
                        *buf += sprintf(*buf, "%s ", op_names[e->op]);
                }
        }
        *buf += _SLOT_print( *buf, blk, e->a[0] );
        *buf += sprintf( *buf, "( " );
        for (n = 0; n < frame->sz; n++) {
                arg_p = &frame->a[n];
                *buf += _SLOT_print( *buf, blk, arg_p->expr );
                if( arg_p->flags != 0 ){
                        if( arg_p->flags & SE_in_reg ){
                                *buf += sprintf( *buf, "%s ", reg_names[arg_p->loc.reg.base] );
                                if( FAKE_STACK_LIMIT( arg_p->expr != NULL && ) PRIM_SIZE( arg_p->expr->type ) == 0x08 ){
                                        *buf += sprintf( *buf, "%s ", reg_names[arg_p->loc.reg.pair] );
                                }
                        }
                        if( arg_p->flags & SE_in_arg_slot ){
                                *buf += sprintf( *buf, "A%d ", arg_p->loc.slot.frame );
                        }
                }
        }
        *buf += sprintf(*buf, ") ");
}

static void
print_local( expr * e, char c, int ix, instr_block * blk )
{
        char buf[MAX_BUFFER];
        char * t;
        FILE * fd;
        int offset;

        fd = blk->cfg->fd;

        fprintf( fd, "\t\t%c%d   ", c, ix );
        bzero( buf, 30 );
        t = buf;
        print_type( &t, e);
        fprintf( fd, "%s", buf );
        if( e->uses.other != 0 ) {
                fprintf( fd, "{%d,%d} [%d] ", 
                        e->uses.block,
                        e->uses.other,
                        e->eid );
        }
        else {
                fprintf( fd, "{%d} [%d] ", 
                        e->uses.block,
                        e->eid );
        }
        bzero( buf, MAX_BUFFER );
        offset = FLAG_print( buf, blk, e);
        assert( offset < MAX_BUFFER );
        fprintf( fd, "%s\n", buf );
}

static void
print_locals( Method * method, instr_block * blk)
{
        int n;
        slot * cur_s;

        if( blk->locals.in[0] == NULL ){
                return;
        }
        fprintf( blk->cfg->fd, "\tincoming:\n" );
        for( n = blk->cfg->env.method.n_locals; --n >= 0; ){
                cur_s = (slot *)blk->locals.in[n];
                if( cur_s == NULL ){
                        continue;
                }
                print_local( (expr*)cur_s, 'L', n, blk );
        }
        for( n = blk->sp.in; --n >= 0; ){
                cur_s = (slot *)blk->stack.in[n];
                if( cur_s == NULL ){
                        continue;
                }
                print_local( (expr*)cur_s, 'S', n, blk );
        }
#ifdef OPTIMIZE_GCSE
        for( n = blk->cfg->env.global.globals.max; --n >= 0; ){
                cur_s = (slot *)blk->globals.in[n];
                print_local( (expr*)cur_s, 'G', n, blk );
        }
#endif
}

static void
BL_print( FILE * fd, block_list * pred, block_list * succ, const char * pred_str, const char * succ_str )
{
        int n, m;
        boolean print;

        n = pred->length;
        m = succ->length;
        print = (n > 0) || (m > 0);
        if( print ) fprintf( fd, "\t" );
        if( n > 0 ){
                fprintf( fd, "%s: ", pred_str );
                for( ;--n >= 0; ){
                        fprintf( fd, "B%d.%d ", pred->blocks[n]->bid, pred->blocks[n]->env->inline_depth );
                }
        }
        if( m > 0 ){
                fprintf( fd, "%s: ", succ_str );
                for( ;--m >= 0; ){
                        fprintf( fd, "B%d.%d ", succ->blocks[m]->bid, succ->blocks[m]->env->inline_depth );
                }
        }
        if( print ) fprintf( fd, "\n" );
}

static void
DOM_print( FILE * fd, uint * dom, int n_blocks )
{
        int n;

        fprintf( fd, "\tdom: " );
        for( n = n_blocks; --n >= 0; ){
                if( BV_is_set( dom, n ) ){
                        fprintf( fd, "B%d ", n );
                }
        }
        fprintf( fd, "\n" );
}

static void
CFG_print_node( instr_block * blk )
{
        block_list dummy;
        FILE * fd;
        dummy.length = 0;

        fd = blk->cfg->fd;
        BL_print( fd, &blk->pred, &blk->succ, "pred", "succ" );
        BL_print( fd, &blk->ebb, &dummy, "ebb", "" );
        BL_print( fd, &blk->ebb_pred, &blk->ebb_succ, "ex_pred", "ex_succ" );
        BL_print( fd, &blk->try, &blk->catch, "try", "catch" );
        BL_print( fd, &dummy, &blk->ret, "", "ret" );
        if( blk->dom != NULL ){
                DOM_print( fd, blk->dom, blk->cfg->n_blocks );
        }
}

static printFunc print_funcs[] = {
        0, /* nop */
	print_const, /* const_m1 */
	print_const, /* const_0 */
	print_const, /* const_1 */
	print_const, /* const */
	print_unary, /* move */
	print_load, /* load */
	print_store, /* store */
	print_binary, /* add */
	print_binary, /* sub */
	print_binary, /* mul */
	print_binary, /* div */
	print_binary, /* rem */
	print_unary, /* neg */
	print_binary, /* lsl */
	print_binary, /* lsr */
	print_binary, /* asr */
	print_binary, /* and */
	print_binary, /* or */
	print_binary, /* xor */
	print_unary, /* cvt */
	print_binary, /* set_cc */
	cbranch_eq_func( print_, ), /* cbranch_eq */
	cbranch_ne_func( print_, ), /* cbranch_ne */
	cbranch_lt_func( print_, ), /* cbranch_lt */
	cbranch_ge_func( print_, ), /* cbranch_ge */
	cbranch_gt_func( print_, ), /* cbranch_gt */
	cbranch_le_func( print_, ), /* cbranch_le */
	cbranch_ult_func( print_, ), /* cbranch_ult */
	print_branch, /* branch */
	print_unary, /* branch_indirect */
	print_unary, /* returnarg */
	print_binary, /* cmp */
	print_call, /* call */
	0, /* target */
	0, /* prologue */
	0, /* epilogue */
	0, /* eprologue */
        print_unary, /* copy */
        print_const, /* inline_data */
        print_unary, /* get_arg */
        print_unary, /* preload */
	print_binary, /* replace */
	0, /* specialize_sync */
	print_binary, /* tableswitch */
	print_binary, /* lookupswitch */
	print_unary, /* native_instr_1a */
	print_cmov, /* native_instr_2a */
        print_cmov, /* cmov_eq */
        print_cmov, /* cmov_ne */
        print_cmov, /* cmov_lt */
        print_cmov, /* cmov_ge */
        print_cmov, /* cmov_gt */
        print_cmov /* cmov_le */
};

void
EXPR_print(instr_block* blk, expr *e)
{
        printFunc print;
        char * buf_p, buffer[MAX_BUFFER];
        char * spans;

        print = print_funcs[e->op];
        buf_p = buffer;
        bzero( buffer, MAX_BUFFER );
        buf_p += sprintf(buf_p,"\t[%3d] ", e->eid);
#if 1
        buf_p += sprintf(buf_p,"\t[%3d] ", e->uses.last);
#endif
        print_type(&buf_p, e);
//        spans = blk->flags & SE_use_spans_call ? "C" : "";
        spans = "";
        if (e->uses.other != 0) {
                buf_p += sprintf( buf_p, "{%d%s,%d} ", e->uses.block, spans, e->uses.other);
        }
        else {
                buf_p += sprintf( buf_p, "{%d%s} ", e->uses.block, spans );
        }
        if( print != NULL ){
                (*print)( &buf_p, blk, e);
        }
        else {
                buf_p += sprintf( buf_p, "%s ", op_names[e->op]);
        }
        if( e->flags & SE_slot ){
#ifdef OPTIMIZE_GCSE
                if( e->flags & SE_global ){
                        buf_p += sprintf( buf_p, "-> [G%d] ", e->loc.slot.frame );
                } 
                else
#endif
                if( e->loc.slot.frame < blk->cfg->env.method.n_locals ){
                        buf_p += sprintf( buf_p, "-> [L%d] ", frame_to_local( e->loc.slot.frame, blk->cfg ) );
                }
                else{
                        buf_p += sprintf( buf_p, "-> [S%d] ", e->loc.slot.frame - blk->cfg->env.method.n_locals );
                }
        }

        buf_p += FLAG_print( buf_p, blk, e );
        assert( buf_p - buffer < MAX_BUFFER );
//        buf_p += sprintf( buf_p, "0x%p ", e );
        fprintf( blk->cfg->fd, "%s\n", buffer );
}

static void
LOOP_print( loop_info * loop_i )
{
        block_list dummy;
        int localsz, n, i;
        cfg_info * cfg;
        FILE * fd;
        field_stats * fs;
        stats_info * si;
        boolean is_static;
        char * buf_p, buf[200];

        cfg = loop_i->header->cfg;
        fd = cfg->fd;
        localsz = cfg->env.method.n_locals;
        dummy.length = 0;
        BL_print( fd, &dummy, &loop_i->blocks, "", "loop" );
        BL_print( fd, &dummy, &loop_i->exits, "", "exits" );
        fprintf( fd,         "\theader: B%d\n", loop_i->header->bid );
#define LOOP_print_flag( _f ) \
        if( loop_i->flags & LOOP_##_f ){ \
                buf_p += sprintf( buf_p, #_f " " ); \
        }
        buf_p = buf;
        LOOP_print_flag( has_java_call );
        LOOP_print_flag( do_while );
        LOOP_print_flag( infinite );
        if( buf != buf_p ){
                fprintf( fd, "\t(%s)\n", buf );
        }
        fprintf( fd,         "\t\t     ld st ml it\n" );
        for( n = localsz; --n >= 0; ){
                fprintf( fd, "\t\tL%d   %2c %2c %2c %2c\n", n, 
                         STATS_local_get( SO_load, loop_i->locals, n) ? 'T' : 'F',
                         STATS_local_get( SO_store, loop_i->locals, n) ? 'T' : 'F',
                         STATS_local_get( SO_loop_must_load, loop_i->locals, n) ? 'T' : 'F',
                         STATS_local_get( SO_iterator, loop_i->locals, n) ? 'T' : 'F'
                         );
        }
        for( n = loop_i->fields.length; --n >= 0; ){
                fs = loop_i->fields.classes[n];
                fprintf( fd, "\t\t%s ", fs->class->name );
                si = fs->stats;
                for( i = 32 * si->words; --i >= 0; ){
                        if( BV_is_set( si->item[SO_load], i ) ){
                                is_static = i < fs->class->nsfields ? true : false;
                                fprintf( fd, "%s get %d ", 
                                         is_static ? "static " : "",
                                         i - (!is_static ? fs->class->nsfields : 0) );
                        }
                        if( BV_is_set( si->item[SO_store], i ) ){
                                is_static = i < fs->class->nsfields ? true : false;
                                fprintf( fd, "%s put %d ", 
                                         is_static ? "static " : "",
                                         i - (!is_static ? fs->class->nsfields : 0) );
                        }
                }
                fprintf( fd, "\n" );
        }
}

void
BLK_print( instr_block * blk )
{
        expr * e;
        FILE * fd;
        char * buf_p, buf[200];

#define BLK_print_flag( _f ) \
        if( blk->flags & BLK_##_f ){ \
                buf_p += sprintf( buf_p, #_f " " ); \
        }

        e = blk->exprs_head;
        fd = blk->cfg->fd;
        fprintf( fd, "B%d.%d: %db - %db, sp: %d -> %d, in_loop = %c\n", blk->bid, blk->env->inline_depth, blk->pc.in, blk->pc.out, blk->sp.in, blk->sp.out, blk->loop != NULL ? 'Y' : 'N' );
        buf_p = buf;
        BLK_print_flag( catch );
        BLK_print_flag( finally );
        BLK_print_flag( ret );
        BLK_print_flag( loop_entry );
        BLK_print_flag( loop_header );
        BLK_print_flag( inserted_header );
        BLK_print_flag( succ_not_in_ebb );
        BLK_print_flag( inline_epilogue );
        BLK_print_flag( has_java_call );
        if( buf != buf_p ){
                fprintf( fd, "\t(%s)\n", buf );
        }
        if( blk->set_cc != NULL ){
                fprintf( fd, "\tset_cc: [%d]\n", blk->set_cc->eid );
        }
        if( blk->loop != NULL ){
                fprintf( fd, "\tloop: B%d\n", blk->loop->header->bid );
        }
        CFG_print_node( blk );
#if 0
        fprintf( fd, "\tblock_stats:\n" );
        for( n = blk->cfg->env.method.n_locals; --n >= 0; ){
                fprintf( fd, "\t\tL%d   %c %c\n", n, 
                         STATS_local_get( SO_null_checked, blk->stats.locals, n) ? 'T' : 'F',
                         STATS_local_get( SO_loop_must_load, blk->stats.locals, n) ? 'T' : 'F'
                         );
        }
#endif
#ifdef PRINT_EXPRS
        if( blk->loop != NULL && blk->loop->cond == blk ){
                LOOP_print( blk->loop );
        }
        print_locals(blk->cfg->env.method.method, blk);
        while (e != NULL) {
                EXPR_print( blk, e );
                e = e->expr.next;
        }
#endif
}

void
CFG_print( cfg_info * cfg )
{
        instr_block * blk;
        Method * method;
        FILE * fd;
        int n;
        localVariables * vars;
#ifdef OPTIMIZE_INLINE
        method_env * m_env;
#endif

        method = cfg->env.method.method;
        fd = cfg->fd;
        if( cfg->flags & CFG_is_leaf ){
                fprintf( fd, "leaf " );
        }
        if( method->accflags & ACC_STATIC ){
                fprintf( fd, "static " );
        }
#ifdef OPTIMIZE_INLINE
        for( m_env = cfg->env.global.methods; m_env != NULL; m_env = m_env->next ){
                CFG_set_method_env( cfg, m_env );
#endif
                method = cfg->env.method.method;
                fprintf( fd, "%s.%s %s:\n", cfg->env.method.class->name, method->name, method->signature );
#if 1
                vars = method->vars;
                if( vars != NULL ){
                        fprintf( fd, "LocalVariables:\n" );
                        for( n = vars->length; --n >= 0; ){
                                fprintf( fd, "\tL%d: %s %s\n", vars->entry[n].name, vars->entry[n].type );

                        }
                }
#endif
                blk = cfg->env.method.root;
                while (blk != NULL) {
                        BLK_print( blk );
                        blk = blk->block.next;
                }
#ifdef OPTIMIZE_INLINE
        }
        CFG_set_method_env( cfg, cfg->env.global.root_env );
#endif
}

#ifdef USE_GDB_DISASM
#define TRUE_FALSE_ALREADY_DEFINED 1
#include "dis-asm.h"
disassemble_info info;

void
DISASM_init( FILE * fd )
{
        info.fprintf_func = (fprintf_ftype)fprintf;
        info.stream = fd;
        info.flavour = bfd_target_unknown_flavour;
#if defined(TGT_SPARC)
        info.arch = bfd_mach_sparc_v9;
#elif defined(TGT_ARM)
        info.arch = bfd_mach_arm_4;
#endif
        info.read_memory_func = buffer_read_memory;
        info.memory_error_func = perror_memory;
        info.print_address_func = generic_print_address;
        info.symbol_at_address_func = generic_symbol_at_address;
}

#else
#include "disassembler.h"
#endif

void
DISASM_consttable( FILE * fd, uint8 * base_p, uint8 * end_p )
{
        uint8 * const_p;

        for( const_p = base_p; const_p < end_p; const_p += WORD_SZ ){
                fprintf( fd, "%#010x\t%-10s\t%x\n", (uint)const_p, ".word", *(uint*)const_p );
        }
}

void
DISASM_exceptiontable( FILE * fd, uint8 * base_p, Method * meth )
{
	jexceptionEntry* e;

	if (meth->exception_table != 0) {
		for( e = meth->exception_table; (int)e->start_pc != -1; e++ ){
                        fprintf( fd, "exception 0x%x: 0x%x - 0x%x -> 0x%x\n", (uint32)e->catch, (uint)e->start_pc, (uint)e->end_pc, (uint)e->handler_pc);
		}
	}
}

#define PRINT_RAW

void
DISASM_code( FILE * fd, uint8 * base_p, uint8 * start_p, uint8 * end_p )
{
#ifndef USE_GDB_DISASM
        char buffer[200];
        uint regmask, symbol, ls_reg;
#endif
        uint8 * code_p;

#ifdef USE_GDB_DISASM
        info.buffer = base_p;
        info.buffer_vma = (bfd_vma)base_p;
        info.buffer_length = (int)(end_p - base_p);
#endif
        for( code_p = start_p; code_p < end_p; code_p += WORD_SZ ){
#ifdef USE_GDB_DISASM
#ifdef PRINT_RAW
		fprintf( fd, "0x%x\t%08x\t", (uint)code_p, *(uint*)code_p );
#else
		fprintf( fd, "0x%x\t", (uint)code_p );
#endif
#if defined(TGT_SPARC)
                print_insn_sparc( (bfd_vma)code_p, &info );
#elif defined(TGT_ARM)
                print_insn_little_arm( (bfd_vma)code_p, &info );
#endif
                fprintf( fd, "\n" );
#else
                bzero( buffer, 200 );
                disasm( buffer, (uint)code_p, *(uint*)code_p, &regmask, &symbol, &ls_reg);
                fprintf( fd, "%s\n", buffer );
#endif
        }
}

void
DISASM_method( cfg_info * cfg )
{
        uint8 * base_p, * end_p, * start_p;
        FILE * fd;

        fd = cfg->fd;
        base_p = cfg->env.global.native.consttable.start;
        DISASM_consttable( fd, base_p, cfg->env.global.native.consttable.end );
        start_p = cfg->env.global.native.code.start;
        end_p = cfg->env.global.native.code.end;
        DISASM_code( fd, base_p, start_p, end_p );
#ifdef INSTALL_CODE
        DISASM_exceptiontable( fd, base_p, cfg->env.method.method );
#else
        fprintf( fd, "Exception tables not installed yet!\n" );
#endif
}

void
DISASM_instruction( FILE * fd, instr_block * blk, uint8 * code_p )
{
        uint8 * base_p;
#ifdef USE_GDB_DISASM

        base_p = blk->cfg->env.global.native.consttable.start;
        info.buffer = base_p;
        info.buffer_vma = (bfd_vma)base_p;
        info.buffer_length = (int)(code_p + 8);
#undef PRINT_RAW
#ifdef PRINT_RAW
	fprintf( fd, "\t\t0x%x\t%08x\t", (uint)code_p, *(uint*)code_p );
#else
	fprintf( fd, "\t\t0x%x\t", (uint)code_p );
#endif
#if defined(TGT_SPARC)
        print_insn_sparc( (bfd_vma)code_p, &info );
#elif defined(TGT_ARM)
        print_insn_little_arm( (bfd_vma)code_p, &info );
#endif
        fprintf( fd, "\n" );
#else
        char buffer[200];
        uint regmask, symbol, ls_reg;

        bzero( buffer, 200 );
        base_p = blk->cfg->env.global.native.consttable.start;
        disasm( buffer, (uint)code_p, *(uint*)code_p, &regmask, &symbol, &ls_reg);
        fprintf( fd, "%s\n", buffer );
#endif
}
#endif
