/*
 * reduce.c
 * Instruction-level sequencing
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "soft.h"
#include "env.h"
#include "reduce.h"
#include "access.h"

#ifdef OPTIMIZE_CONSTANTS
#define CP_FUNCTION( _name ) static void CP_##_name( expr * e )

#define CP_cc_lt			0x1
#define CP_cc_eq			0x2
#define CP_cc_gt			0x4

static void
CP_fixup_const( expr * e )
{
        if( PRIM_FP( e->type ) || e->type == PRIM_void ){
                return;
        }
        if( e->type == PRIM_long ){
                if( e->val.l_const == -1 ){
                        e->op = OP_const_m1;
                }
                else if( e->val.l_const == 0 ){
                        e->op = OP_const_0;
                }
                else if( e->val.l_const == 1 ){
                        e->op = OP_const_1;
                }
        }
        else{
                if( e->val.i_const == -1 ){
                        e->op = OP_const_m1;
                }
                else if( e->val.i_const == 0 ){
                        e->op = OP_const_0;
                }
                else if( e->val.i_const == 1 ){
                        e->op = OP_const_1;
                }
        }
}

CP_FUNCTION( arithmetic )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( e->a[1] );
        switch( PRIM_TYPE( e->type ) ){
        case PRIM_vm_int:
                switch (e->op) {
                case OP_add:
                        e->val.i_const = a0_e->val.i_const + a1_e->val.i_const;
                        break;
                case OP_sub:
                        e->val.i_const = a0_e->val.i_const - a1_e->val.i_const;
                        break;
                case OP_mul:
                        e->val.i_const = a0_e->val.i_const * a1_e->val.i_const;
                        break;
                case OP_div:
                        e->val.i_const = a0_e->val.i_const / a1_e->val.i_const;
                        break;
                case OP_rem:
                        e->val.i_const = a0_e->val.i_const % a1_e->val.i_const;
                        break;
                case OP_neg:
                        e->val.i_const = - a0_e->val.i_const;
                        break;
                default:
                        ABORT();
                }
                break;
        case PRIM_vm_long:
                switch (e->op) {
                case OP_add:
                        e->val.l_const = a0_e->val.l_const + a1_e->val.l_const;
                        break;
                case OP_sub:
                        e->val.l_const = a0_e->val.l_const - a1_e->val.l_const;
                        break;
                case OP_mul:
                        e->val.l_const = a0_e->val.l_const * a1_e->val.l_const;
                        break;
                case OP_div:
                        e->val.l_const = a0_e->val.l_const / a1_e->val.l_const;
                        break;
                case OP_rem:
                        e->val.l_const = a0_e->val.l_const % a1_e->val.l_const;
                        break;
                case OP_neg:
                        e->val.l_const = - a0_e->val.l_const;
                        break;
                default:
                        ABORT();
                }
                break;
        case PRIM_vm_float:
                switch (e->op) {
                case OP_add:
                        e->val.f_const = a0_e->val.f_const + a1_e->val.f_const;
                        break;
                case OP_sub:
                        e->val.f_const = a0_e->val.f_const - a1_e->val.f_const;
                        break;
                case OP_mul:
                        e->val.f_const = a0_e->val.f_const * a1_e->val.f_const;
                        break;
                case OP_div:
                        e->val.f_const = soft_fdiv(a0_e->val.f_const,a1_e->val.f_const);
                        break;
                case OP_rem:
                        e->val.f_const = javaRemainderf(a0_e->val.f_const,a1_e->val.f_const);
                        break;
                case OP_neg:
                        e->val.f_const = - a0_e->val.f_const;
                        break;
                default:
                        ABORT();
                }
                break;
        case PRIM_vm_double:
                switch (e->op) {
                case OP_add:
                        e->val.d_const = a0_e->val.d_const + a1_e->val.d_const;
                        break;
                case OP_sub:
                        e->val.d_const = a0_e->val.d_const - a1_e->val.d_const;
                        break;
                case OP_mul:
                        e->val.d_const = a0_e->val.d_const * a1_e->val.d_const;
                        break;
                case OP_div:
                        e->val.d_const = soft_fdivl(a0_e->val.d_const,a1_e->val.d_const);
                        break;
                case OP_rem:
                        e->val.d_const = javaRemainder(a0_e->val.d_const,a1_e->val.d_const);
                        break;
                case OP_neg:
                        e->val.d_const = - a0_e->val.d_const;
                        break;
                default:
                        ABORT();
                }
                break;
        default:
                ABORT();
        }
        e->op = OP_const;
        CP_fixup_const( e );
}

CP_FUNCTION( logical )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( e->a[1] );
        switch( PRIM_TYPE( e->type ) ){
        case PRIM_vm_int:
                switch (e->op) {
                case OP_lsl:
                        e->val.i_const = a0_e->val.i_const << a1_e->val.i_const; 
                        break;
                case OP_lsr:
                        e->val.i_const = (uint32)a0_e->val.i_const >> a1_e->val.i_const;
                        break;
                case OP_asr:
                        e->val.i_const = a0_e->val.i_const >> a1_e->val.i_const;
                        break;
                case OP_and:
                        e->val.i_const = a0_e->val.i_const & a1_e->val.i_const;
                        break;
                case OP_or:
                        e->val.i_const = a0_e->val.i_const | a1_e->val.i_const;
                        break;
                case OP_xor:
                        e->val.i_const = a0_e->val.i_const ^ a1_e->val.i_const;
                        break;
                default:
                        ABORT();
                }
                break;
        case PRIM_vm_long:
                switch (e->op) {
                case OP_lsl:
                        e->val.l_const = a0_e->val.l_const << a1_e->val.l_const;
                        break;
                case OP_lsr:
                        e->val.l_const = (uint64)a0_e->val.l_const >> a1_e->val.l_const;
                        break;
                case OP_asr:
                        e->val.l_const = a0_e->val.l_const >> a1_e->val.l_const;
                        break;
                case OP_and:
                        e->val.l_const = a0_e->val.l_const & a1_e->val.l_const;
                        break;
                case OP_or:
                        e->val.l_const = a0_e->val.l_const | a1_e->val.l_const;
                        break;
                case OP_xor:
                        e->val.l_const = a0_e->val.l_const ^ a1_e->val.l_const;
                        break;
                default:
                        ABORT();
                }
                break;
        default:
                ABORT();
        }
        e->op = OP_const;
        CP_fixup_const( e );
}

CP_FUNCTION( convert )
{
        expr * a0_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        switch (e->type) {
        case PRIM_byte: 
                if (a0_e->type == PRIM_int) {
                        e->val.i_const = (a0_e->val.i_const << 24) >> 24;
                }
                else {
                        ABORT();
                }
                break;
        case PRIM_short: 
                if (a0_e->type == PRIM_int) {
                        e->val.i_const = a0_e->val.i_const & 0xffff;
                }
                else {
                        ABORT();
                }
                break;
        case PRIM_char: 
                if (a0_e->type == PRIM_int) {
                        e->val.i_const = (a0_e->val.i_const << 16) >> 16;
                }
                else {
                        ABORT();
                }
                break;
        case PRIM_int:
                switch (a0_e->type) {
                case PRIM_long:
                        e->val.i_const = a0_e->val.l_const;
                        break;
                case PRIM_float:
                        e->val.i_const = a0_e->val.f_const;
                        break;
                case PRIM_double:
                        e->val.i_const = a0_e->val.d_const;
                        break;
                default:
                        ABORT();
                }
                break;
        case PRIM_long:
                switch (a0_e->type) {
                case PRIM_int:
                        e->val.l_const = a0_e->val.i_const;
                        break;
                case PRIM_float:
                        e->val.l_const = a0_e->val.f_const;
                        break;
                case PRIM_double:
                        e->val.l_const = a0_e->val.d_const;
                        break;
                default:
                        ABORT();
                }
                break;
        case PRIM_float:
                switch (a0_e->type) {
                case PRIM_int:
                        e->val.f_const = a0_e->val.i_const;
                        break;
                case PRIM_long:
                        e->val.f_const = a0_e->val.l_const;
                        break;
                case PRIM_double:
                        e->val.f_const = a0_e->val.d_const;
                        break;
                default:
                        ABORT();
                }
                break;
        case PRIM_double:
                switch (a0_e->type) {
                case PRIM_int:
                        e->val.d_const = a0_e->val.i_const;
                        break;
                case PRIM_long:
                        e->val.d_const = a0_e->val.l_const;
                        break;
                case PRIM_float:
                        e->val.d_const = a0_e->val.f_const;
                        break;
                default:
                        ABORT();
                }
                break;
        default:
                ABORT();
        }
        e->op = OP_const;
        CP_fixup_const( e );
}

CP_FUNCTION( set_cc )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( e->a[1] );
        assert( a0_e->type == PRIM_ref || a0_e->type == PRIM_int );
        assert( a1_e->type == PRIM_ref || a1_e->type == PRIM_int );
        if( a0_e->val.i_const < a1_e->val.i_const ){
                e->val.c_flags |= CP_cc_lt;
        }
        if( a0_e->val.i_const == a1_e->val.i_const ){
                e->val.c_flags |= CP_cc_eq;
        }
        if( a0_e->val.i_const > a1_e->val.i_const ){
                e->val.c_flags |= CP_cc_gt;
        }
        e->op = OP_const;
        CP_fixup_const( e );
}

CP_FUNCTION( cc_cbranch )
{
        expr * cmp_e;

        cmp_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        switch( e->op ){
        case OP_cbranch_eq:
                e->op = cmp_e->val.c_flags & CP_cc_eq ? OP_branch : OP_nop;
                break;
        case OP_cbranch_ne:
                e->op = !(cmp_e->val.c_flags & CP_cc_eq) ? OP_branch : OP_nop;
                break;
        case OP_cbranch_lt:
                e->op = cmp_e->val.c_flags & CP_cc_lt ? OP_branch : OP_nop;
                break;
        case OP_cbranch_ge:
                e->op = cmp_e->val.c_flags & CP_cc_eq && cmp_e->val.c_flags & CP_cc_gt 
                        ? OP_branch : OP_nop;
                break;
        case OP_cbranch_gt:
                e->op = cmp_e->val.c_flags & CP_cc_gt ? OP_branch : OP_nop;
                break;
        case OP_cbranch_le:
                e->op = cmp_e->val.c_flags & CP_cc_lt && cmp_e->val.c_flags & CP_cc_eq
                        ? OP_branch : OP_nop;
                break;
        default:
                ABORT();
                break;
        }
}

CP_FUNCTION( cbranch )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( e->a[1] );
        assert( a0_e->type == PRIM_ref || a0_e->type == PRIM_int );
        assert( a1_e->type == PRIM_ref || a1_e->type == PRIM_int );
        switch( e->op ){
        case OP_cbranch_eq:
                e->op = a0_e->val.i_const == a1_e->val.i_const ? OP_branch : OP_nop;
                break;
        case OP_cbranch_ne:
                e->op = a0_e->val.i_const != a1_e->val.i_const ? OP_branch : OP_nop;
                break;
        case OP_cbranch_lt:
                e->op = a0_e->val.i_const < a1_e->val.i_const ? OP_branch : OP_nop;
                break;
        case OP_cbranch_ge:
                e->op = a0_e->val.i_const >= a1_e->val.i_const ? OP_branch : OP_nop;
                break;
        case OP_cbranch_gt:
                e->op = a0_e->val.i_const > a1_e->val.i_const ? OP_branch : OP_nop;
                break;
        case OP_cbranch_le:
                e->op = a0_e->val.i_const <= a1_e->val.i_const ? OP_branch : OP_nop;
                break;
        default:
                ABORT();
                break;
        }
}

CP_FUNCTION( cmp )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( e->a[1] );
        switch( PRIM_TYPE( a0_e->type ) ){
        case PRIM_vm_float:
                assert( a1_e->type == PRIM_float );
                e->val.i_const = e->val.c_flags & FCMP_l
                        ? soft_fcmpl( a0_e->val.f_const, a0_e->val.f_const )
                                : soft_fcmpg( a0_e->val.f_const, a0_e->val.f_const );
                break;
        case PRIM_vm_double:
                assert( a1_e->type == PRIM_float );
                e->val.i_const = e->val.c_flags & FCMP_l
                        ? soft_dcmpl( a0_e->val.d_const, a0_e->val.d_const )
                                : soft_dcmpg( a0_e->val.d_const, a0_e->val.d_const );
                break;
        case PRIM_vm_long:
                assert( a1_e->type == PRIM_long );
                e->val.i_const = soft_lcmp( a0_e->val.l_const, a0_e->val.l_const );
                break;
        }
        e->op = OP_const;
        CP_fixup_const( e );
}

/* Constant propagation (CP) function table */
cpFunc CP_funcs[] = {
	0, /* nop */
	0, /* const_m1 */
	0, /* const_0 */
	0, /* const_1 */
	0, /* const */
	0, /* move */
	0, /* load */
	0, /* store */
	CP_arithmetic, /* add */
	CP_arithmetic, /* sub */
	CP_arithmetic, /* mul */
	CP_arithmetic, /* div */
	CP_arithmetic, /* rem */
	CP_arithmetic, /* neg */
	CP_logical, /* lsl */
	CP_logical, /* lsr */
	CP_logical, /* asr */
	CP_logical, /* and */
	CP_logical, /* or */
	CP_logical, /* xor */
	CP_convert, /* cvt */
	CP_set_cc, /* set_cc */
	cbranch_eq_func( CP_, ), /* cbranch_eq */
	cbranch_ne_func( CP_, ), /* cbranch_ne */
	cbranch_lt_func( CP_, ), /* cbranch_lt */
	cbranch_ge_func( CP_, ), /* cbranch_ge */
	cbranch_gt_func( CP_, ), /* cbranch_gt */
	cbranch_le_func( CP_, ), /* cbranch_le */
	cbranch_ult_func( CP_, ), /* cbranch_ult */
	0, /* branch */
	0, /* branch_indirect */
	0, /* returnarg */
	CP_cmp, /* cmp */
	0, /* call */
	0, /* target */
	0, /* prologue */
	0, /* epilogue */
	0, /* eprologue */
        0, /* copy */
        0, /* inline_data */
        0, /* get_arg */
        0, /* preload */
        0, /* replace */
        0, /* specialize_sync */
        0, /* tableswitch */
        0, /* lookupswitch */
        0, /* native_instr_1a */
        0, /* native_instr_2a */
        0, /* cmov_eq */
        0, /* cmov_ne */
        0, /* cmov_lt */
        0, /* cmov_ge */
        0, /* cmov_gt */
        0 /* cmov_le */
};

#define PP_FUNCTION( _name ) static void PP_##_name( expr * e )

PP_FUNCTION( arithmetic )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( e->a[1] );
        switch (e->op) {
        case OP_add:
                if (a0_e->op == OP_const_0) {
                        e->a[0] = e->a[1];
                        e->a[1] = NULL;
                        e->op = OP_move;
                }
                else if (a1_e->op == OP_const_0) {
                        e->a[1] = NULL;
                        e->op = OP_move;
                }
                break;
        case OP_sub:
                if (a0_e->op == OP_const_0) {
                        e->a[0] = e->a[1];
                        e->a[1] = NULL;
                        e->op = OP_neg;
                }
                else if (a1_e->op == OP_const_0) {
                        e->a[1] = NULL;
                        e->op = OP_move;
                }
                break;
        case OP_mul:
                if (a0_e->op == OP_const_1) {
                        e->a[0] = e->a[1];
                        e->a[1] = NULL;
                        e->op = OP_move;
                }
                else if (a1_e->op == OP_const_1) {
                        e->a[1] = NULL;
                        e->op = OP_move;
                }
                else if (a0_e->op == OP_const_m1) {
                        e->a[0] = e->a[1];
                        e->a[1] = NULL;
                        e->op = OP_neg;
                }
                else if (a1_e->op == OP_const_m1) {
                        e->a[1] = NULL;
                        e->op = OP_neg;
                }
                else if (a0_e->op == OP_const_0 ||
                         a1_e->op == OP_const_0) {
                        e->a[0] = NULL;
                        e->a[1] = NULL;
                        e->op = OP_const_0;
                }
                break;
        case OP_div:
                if (a1_e->op == OP_const_1) {
                        e->a[0] = e->a[1];
                        e->a[1] = NULL;
                        e->op = OP_move;
                }
                else if (a1_e->op == OP_const_m1) {
                        e->a[0] = e->a[1];
                        e->a[1] = NULL;
                        e->op = OP_neg;
                }
                else if (a0_e->op == OP_const_0) {
                        e->a[0] = NULL;
                        e->a[1] = NULL;
                        e->op = OP_const_0;
                        return;
                }
                break;
        case OP_rem:
                return;
                break;
        default:
                ABORT();
        }
}

PP_FUNCTION( logical )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( e->a[1] );
        switch( e->type ){
        case PRIM_int:
        case PRIM_long:
                switch (e->op) {
                case OP_lsl:
                case OP_lsr:
                case OP_asr:
                        if (a1_e->op == OP_const_0) {
                                e->a[1] = NULL;
                                e->op = OP_move;
                        }
                        else if (a0_e->op == OP_const_0) {
                                e->a[0] = NULL;
                                e->a[1] = NULL;
                                e->op = OP_const_0;
                        }
                        break;
                case OP_and:
                        if (a0_e->op == OP_const_0 || 
                            a1_e->op == OP_const_0) {
                                e->a[0] = NULL;
                                e->a[1] = NULL;
                                e->op = OP_const_0;
                        }
                        break;
                case OP_or:
                        if (a0_e->op == OP_const_0) {
                                e->a[0] = e->a[1];
                                e->a[1] = NULL;
                                e->op = OP_move;
                        }
                        else if (a1_e->op == OP_const_0) {
                                e->a[1] = NULL;
                                e->op = OP_move;
                        }
                        break;
                case OP_xor:
                        if (a0_e->op == OP_const_0) {
                                e->a[0] = e->a[1];
                                e->a[1] = NULL;
                                e->op = OP_move;
                        }
                        else if (a1_e->op == OP_const_0) {
                                e->a[1] = NULL;
                                e->op = OP_move;
                        }
                        break;
                }
                break;
        case PRIM_byte: case PRIM_short: case PRIM_char: 
        case PRIM_float: case PRIM_double:
        default:
                ABORT();
        }
}

/* Partial propagation (PP) of -1,1 & 0 function table */
cpFunc PP_funcs[] = {
	0, /* nop */
	0, /* const_m1 */
	0, /* const_0 */
	0, /* const_1 */
	0, /* const */
	0, /* move */
	0, /* load */
	0, /* store */
	PP_arithmetic, /* add */
	PP_arithmetic, /* sub */
	PP_arithmetic, /* mul */
	PP_arithmetic, /* div */
	PP_arithmetic, /* rem */
	0, /* neg */
	PP_logical, /* lsl */
	PP_logical, /* lsr */
	PP_logical, /* asr */
	PP_logical, /* and */
	PP_logical, /* or */
	PP_logical, /* xor */
	0, /* cvt */
	0, /* set_cc */
	0, /* cbranch_eq */
	0, /* cbranch_ne */
	0, /* cbranch_lt */
	0, /* cbranch_ge */
	0, /* cbranch_gt */
	0, /* cbranch_le */
	0, /* cbranch_ult */
	0, /* branch */
	0, /* branch_indirect */
	0, /* returnarg */
	0, /* cmp */
	0, /* call */
	0, /* target */
	0, /* prologue */
	0, /* epilogue */
	0, /* eprologue */
        0, /* copy */
        0, /* inline_data */
        0, /* get_arg */
        0, /* preload */
        0, /* replace */
        0, /* specialize_sync */
        0, /* tableswitch */
        0, /* lookupswitch */
        0, /* native_instr_1a */
        0, /* native_instr_2a */
        0, /* cmov_eq */
        0, /* cmov_ne */
        0, /* cmov_lt */
        0, /* cmov_ge */
        0, /* cmov_gt */
        0 /* cmov_le */
};

#endif /* OPTIMIZE_CONSTANTS */

#ifdef OPTIMIZE_CSE

#define CSE_FUNCTION( _name ) static EXTERN_CSE_FUNCTION( _name )
#define EXTERN_CSE_FUNCTION( _name ) expr * CSE_##_name( expr * new_e, expr ** exprs_by_op, instr_block * blk, boolean update ) 

static inline
int
equal_consts( expr * e1, expr * e2 ) 
{
        if (e1->type == e2->type) {
                switch (e1->type) {
                case PRIM_double: case PRIM_long:
                        return (e1->val.l_const == e2->val.l_const);
                default:
                        return (e1->val.i_const == e2->val.i_const);
                }
        } 
        else if ((e1->type == PRIM_int || e1->type == PRIM_ref) &&
                   (e2->type == PRIM_int || e2->type == PRIM_ref)) {
                return (e1->val.i_const == e2->val.i_const);
        } 
        else {
                return 0;
        }
}

/*
 * CSE: Constant value
 */
EXTERN_CSE_FUNCTION( const )
{
        expr * cur_e, * cmp_e;

        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                if( equal_consts(new_e, cmp_e) ){
                        return cur_e;
                }
        }
        return NULL;
}

static void
CSE_reduce_base( expr * e, instr_block * blk )
{
        expr * base_e, * a0_e, * a1_e;
        expr_list * el;

        if( blk->cfg->env.global.no_reduce_base ){
                return;
        }
        base_e = EXPR_RESOLVE_ORIGINAL( e->a[0] );
        if( base_e->op == OP_add ){
                el = &blk->cfg->env.method.rescan.exprs;
                a0_e = EXPR_RESOLVE_ORIGINAL( base_e->a[0] );
                a1_e = EXPR_RESOLVE_ORIGINAL( base_e->a[1] );
                if( EXPR_IS_CONST( a0_e ) && a0_e->type != PRIM_label 
                   && base_e->uses.block == 0 ){ 
                        e->val.i_const += a0_e->val.i_const;
                        e->a[0] = base_e->a[1];
                        if( !EL_contains( el, base_e ) ){
                                EL_add( el, base_e );
                        }
LOG(                    fprintf( blk->cfg->fd, "CSE_reduce_base(): B%d: [%d]\n", blk->bid, base_e->eid); );
                }
                else if( EXPR_IS_CONST( a1_e ) && a1_e->type != PRIM_label 
                        && base_e->uses.block == 0 ){
                        e->val.i_const += a1_e->val.i_const;
                        e->a[0] = base_e->a[0];
                        if( !EL_contains( el, base_e ) ){
                                EL_add( el, base_e );
                        }
LOG(                    fprintf( blk->cfg->fd, "CSE_reduce_base(): B%d: [%d]\n", blk->bid, base_e->eid); );
                }
        }
}

CSE_FUNCTION( store )
{
        if( update ){
                CSE_reduce_base( new_e, blk );
        }
        return (expr*)CSE_TERMINATE;
}

static expr *
CSE_match_store( expr * store_e, expr * new_e, boolean update )
{
        expr * store_a0_e, * new_a0_e;
#ifdef TGT_SPARC
        expr * store_ix_e, * new_ix_e;
        boolean indexed;
#endif

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        store_a0_e = EXPR_RESOLVE_ORIGINAL( store_e->a[0] );
#if 0
        indexed = new_e->flags & SE_ldst_indexed ? true : false;
        store_ix_e = indexed ? EXPR_RESOLVE_ORIGINAL( store_e->val.expr ) : NULL;
        new_ix_e = indexed ? EXPR_RESOLVE_ORIGINAL( new_e->val.expr ) : NULL;
#endif
        if( new_a0_e == store_a0_e
#ifdef USE_RISC_LDST
#if 0
            && new_e->flags & SE_ldst_indexed == store_e->flags & SE_ldst_indexed
            && ( indexed ? store_ix_e == new_ix_e : new_e->val.i_const == store_e->val.i_const )
#else
            && new_e->val.i_const == store_e->val.i_const
#endif
#endif
            && new_e->type == store_e->type ){
                return store_e->a[1];
        }
        return (expr*)CSE_TERMINATE;
}

/*
 * CSE: Load operation, do not allow loads to bypass stores
 */
CSE_FUNCTION( load )
{
        expr * cur_e, * cmp_e, * store_e, * cmp_a0_e, * new_a0_e, * call_e;
        Hjava_lang_Class * cmp_c, * base_c;
        boolean is_field;
        Field * f;
#ifdef TGT_SPARC
        expr * new_ix_e, * cmp_ix_e;
        boolean indexed;
#endif

        CSE_reduce_base( new_e, blk );
        is_field = new_e->flags & (SE_object_field | SE_static_field | SE_array_object) ? true : false;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        base_c = new_a0_e->class;

#if 0
        indexed = new_e->flags & SE_ldst_indexed ? true : false;
        new_ix_e = indexed ? EXPR_RESOLVE_ORIGINAL( store_e->val.expr ) : NULL;
#endif
        if( is_field ){
                for( store_e = blk->exprs_by_op[OP_store]; store_e != NULL; store_e = store_e->exprs_by_op ){
                        cmp_e = EXPR_RESOLVE_ORIGINAL( store_e );
#if 0
                        cmp_ix_e = indexed ? EXPR_RESOLVE_ORIGINAL( cmp_e->val.expr ) : NULL;
#endif
                        if( 
#if 0
                           new_e->flags & SE_ldst_indexed != cmp_e->flags & SE_ldst_indexed
                           || ( indexed ? cmp_ix_e != new_ix_e : new_e->val.i_const != cmp_e->val.i_const )
#else
                           new_e->val.i_const != cmp_e->val.i_const 
#endif
                           ){
                                continue;
                        }
                        cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                        cmp_c = cmp_a0_e->class;
                        if( cmp_c == NULL || base_c == NULL 
                            || instanceof( cmp_c, base_c ) || instanceof( base_c, cmp_c ) ){
                                store_e = cmp_e;
                                break;
                        }
                }
        }
        else{
                store_e = NULL;
        }

        call_e = blk->call.last;
        if( is_field ){
                f = (Field*)new_e->a[1];
                assert( f != NULL );
                if( f->accflags & ACC_FINAL ){
                        call_e = NULL;
                }
                else{
                        while( call_e != NULL ){
                                if( call_e->op != OP_copy
                                    && !(call_e->val.frame->flags & CALL_vm) ){
                                        break;
                                }
                                call_e = call_e->exprs_by_op;
                        }
                }
        }
        else{
                while( call_e != NULL ){
                        if( call_e->op != OP_copy
                           && call_e->val.frame->flags & CALL_vm
                           && !(call_e->val.frame->flags & (CALL_no_side_effect | CALL_throw) ) ){
                                break;
                        }
                        call_e = call_e->exprs_by_op;
                }
        }
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                if( ( call_e != NULL && cmp_e->eid < call_e->eid )
                   || ( store_e != NULL && cmp_e->eid < store_e->eid ) ){
                        break;
                }
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
#if 0
                cmp_ix_e = indexed ? EXPR_RESOLVE_ORIGINAL( cmp_e->val.expr ) : NULL;
#endif
                if (new_a0_e == cmp_a0_e
#ifdef USE_RISC_LDST
#if 0
                    && new_e->flags & SE_ldst_indexed == cmp_e->flags & SE_ldst_indexed
                    && ( indexed ? cmp_ix_e == new_ix_e : new_e->val.i_const == cmp_e->val.i_const )
#else
                    && new_e->val.i_const == cmp_e->val.i_const 
#endif
#endif
                    && new_e->type == cmp_e->type ){
                        if( update ){
                                EXPR_reduce_class( new_e, cmp_e->class );
                        }
                        return cur_e;
                }
        }
        if( call_e != NULL || store_e != NULL ){
                if( store_e == NULL
                   || ( call_e != NULL && store_e != NULL && call_e->eid > store_e->eid ) 
                   || !update ){ /* prevents incorrect global renaming */
                        return (expr*)CSE_TERMINATE;
                }
                return CSE_match_store( store_e, new_e, update );
        }
        return NULL;
}

/*
 * CSE: Unary operation
 */
CSE_FUNCTION( unary )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * new_a0_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );

                if( new_a0_e == cmp_a0_e
                    && new_e->type == cmp_e->type ){
                        return cur_e;
                }
        }
        return NULL;
}

/*
 * CSE: Binary operation where order of arguments matters
 */
CSE_FUNCTION( binary )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * cmp_a1_e;
        expr * new_a0_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                if( new_a0_e == cmp_a0_e
                     && new_a1_e == cmp_a1_e
                     && new_e->type == cmp_e->type ){
                        return cur_e;
                }
        }
        return NULL;
}

/*
 * CSE: Binary operation where order of arguments is irrelevant
 */
CSE_FUNCTION( binary_x )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * cmp_a1_e;
        expr * new_a0_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                if( (( new_a0_e == cmp_a0_e 
                        && new_a1_e == cmp_a1_e ) 
                      || ( new_a1_e == cmp_a0_e 
                           && new_a0_e == cmp_a1_e ))
                     && ( new_e->type == cmp_e->type ) ){
                        return cur_e;
                }
        }
        return NULL;
}

#ifdef USE_CC
/*
 * Tranditional CC setting
 */
CSE_FUNCTION( set_cc )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * cmp_a1_e;
        expr * new_a0_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        if( new_e->flags & SE_unsigned ){
                for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                        cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                        cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                        cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                        /* Expand this so it can match loop header comparisons */
                        if( new_a0_e == cmp_a0_e && new_a1_e == cmp_a1_e 
                            && cmp_e->flags & SE_unsigned ){
                                return cur_e;
                        }

                        if( new_a1_e == cmp_a1_e && cmp_e->flags & SE_unsigned 
                            && EXPR_IS_CONST( new_a0_e ) && EXPR_IS_CONST( cmp_a0_e ) ){
                                /* FIXME: Probably need to do method & store checks here for 
                                   correct exception behavior */
                                if( new_a0_e->val.i_const < cmp_a0_e->val.i_const ){
                                        assert( new_a0_e->val.i_const >= 0 );
                                        return cur_e;
                                }
                                else{
                                        assert( cmp_a0_e->val.i_const >= 0 );
                                        if( update ){
                                                instr_block * ebb_root;
                                                expr * const_e;

                                                ebb_root = blk->ebb_root;
                                                move_int_const( const_e, new_a0_e->val.i_const );
                                                if( const_e->eid > cur_e->eid ){
                                                        const_e->eid = cur_e->eid;
                                                        EXPR_REMOVE( blk, const_e );
                                                        EXPR_PREPEND( blk, cur_e, const_e );
                                                } 
                                                //cur_e->a[0]->uses.block--;
                                                cur_e->a[0] = const_e;
                                                cur_e->a[0]->uses.block++;
                                        }
                                        return cur_e;
                                }
                        }
                        else if( new_a1_e == cmp_a1_e && cmp_e->flags & SE_unsigned
                                 && (new_a0_e->op == OP_add || new_a0_e->op == OP_sub) ){
                                expr * add_a0_e, * add_a1_e;

                                add_a0_e = EXPR_RESOLVE_ORIGINAL( new_a0_e->a[0] );
                                add_a1_e = EXPR_RESOLVE_ORIGINAL( new_a0_e->a[1] );
                                if( add_a0_e == cmp_a0_e || add_a1_e == cmp_a0_e ){
//                                        kprintf( "---------------------------------------------------------------------------------\n" );
                                }
                        }
                        else if( new_a1_e == cmp_a1_e && cmp_e->flags & SE_unsigned
                                 && (cmp_a0_e->op == OP_add || cmp_a0_e->op == OP_sub) ){
                                expr * add_a0_e, * add_a1_e;

                                add_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_a0_e->a[0] );
                                add_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_a0_e->a[1] );
                                if( add_a0_e == new_a0_e || add_a0_e == new_a0_e ){
//                                        kprintf( "*********************************************************************************\n" );
                                }
                        }
                }
                return NULL;
        }
        else{
                cur_e = blk->set_cc;
                if( cur_e == NULL ){
                        return NULL;
                }
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );
                if( new_a0_e == cmp_a1_e && new_a1_e == cmp_a0_e ){
//                        kprintf( "----------------------------------------------------------------------------------------\n" );
                        kprintf( "[%d] <- [%d] [%d]\n", cmp_e->eid, cmp_a0_e->eid, cmp_a1_e->eid );
                        /* This means we need to invert branch.  Ugh! */
                }
                if( new_a0_e == cmp_a0_e && new_a1_e == cmp_a1_e ){
                        return cur_e;
                }
                return (expr*)CSE_TERMINATE;
        }
}
#else
/*
 * MIPS style set-less-than (slt)
 */
CSE_FUNCTION( set_cc )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * cmp_a1_e;
        expr * new_a0_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        if( new_a1_e->op == OP_const_0 ){
                return NULL;
        }
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                if( (new_e->flags & SE_unsigned) == (cmp_e->flags & SE_unsigned)
                   && ( ( new_a0_e == cmp_a0_e && new_a1_e == cmp_a1_e 
                         && (new_e->flags & SE_invert_cmp) == (cmp_e->flags & SE_invert_cmp) )
                       || ( new_a0_e == cmp_a1_e && new_a1_e == cmp_a0_e 
                           && (new_e->flags & SE_invert_cmp) != (cmp_e->flags & SE_invert_cmp) ) ) ){
                        if( update ){
                                if( cmp_e->flags & SE_no_generate ){ /* FIXME: correct fix for loading problem? */
                                        cmp_e->a[0]->uses.block++; 
                                        assert( cmp_a1_e->op == OP_const_0 );
                                }
                        }
                        return cur_e;
                }
        }
        return NULL;
}
#endif

/*
 * CSE: Try to reduce multiply into a shift
 */
CSE_FUNCTION( mul )
{
        expr * a0_e, * a1_e, * t_e, * t1_e, * t2_e, * swap_e, * a_e;
        expr * const_e, const_e_, sh_e;
        boolean do_swap, reduced;
        int b_cnt, b_pos, b_pos2;
        instr_block * ebb_root;

        ebb_root = blk->ebb_root;
        a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        do_swap = false;
        reduced = false;
        if( ( EXPR_IS_CONST( a1_e ) && PRIM_TYPE( a1_e->type ) == PRIM_vm_int )
            || (do_swap = EXPR_IS_CONST( a0_e ) && PRIM_TYPE( a0_e->type ) == PRIM_vm_int) ){
                EXPR_SWAP_IF( do_swap, a0_e, a1_e );
                a_e = do_swap ? new_e->a[1] : new_e->a[0];
                b_cnt = count_bits( a1_e->val.i_const, &b_pos );
                if( b_cnt == 1 ){
                        if( update ){
                                lshl_int_const( new_e, a_e, b_pos );
                                /* May need to add a neg sign too! */
                        }
                        else{
                                const_e = &const_e_;
                                bzero( const_e, sizeof(expr) );
                                const_e->op = OP_const;
                                const_e->val.i_const = b_pos;
                                const_e->type = PRIM_int;
                                CP_fixup_const( const_e );
                                const_e = CSE_recurse( const_e, blk, blk->ebb_root, CSE_const, OP_const, false );
                                if( const_e == NULL ){
                                        return NULL;
                                }
                                bzero( &sh_e, sizeof(expr) );
                                sh_e.op = OP_lsl;
                                sh_e.type = PRIM_int;
                                sh_e.a[0] = a_e;
                                sh_e.a[1] = const_e;
                                if( GCSE_recurse( const_e, blk, blk->ebb_root, CSE_binary, OP_lsl ) == NULL ){
                                        return NULL;
                                }
                        }
                        return new_e;
                }
                else if( b_cnt == 2 && a1_e->val.i_const & 0x1 && update ){
                        lshl_int_const( t1_e, a_e, b_pos );
                        add_int( new_e, t1_e, a_e );
                        return new_e;
                }
                else if( b_cnt == 2 && update ){
                        count_bits( a1_e->val.i_const & ~(1 << b_pos), &b_pos2 );
                        lshl_int_const( t1_e, a_e, b_pos );
                        lshl_int_const( t2_e, a_e, b_pos2 );
                        add_int( new_e, t1_e, t2_e );
                        return new_e;
                }
        }
        return CSE_binary_x( new_e, exprs_by_op, blk, update );
}

/*
 * CSE: Try to reduce divide into a shift
 */
CSE_FUNCTION( div )
{
        expr * a0_e, * a1_e, * a_e, * t_e;
        expr * const_e, const_e_, sh_e;
        int b_cnt, b_pos;
        instr_block * ebb_root;

        ebb_root = blk->ebb_root;
        a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        if( EXPR_IS_CONST( a1_e ) && PRIM_TYPE( a1_e->type ) == PRIM_vm_int ){
                b_cnt = count_bits( a1_e->val.i_const, &b_pos );
                if( b_cnt == 1 ){
                        a_e = new_e->a[0];
                        if( update ){
                                ashr_int_const( new_e, a_e, b_pos );
                                /* May need to add a neg sign too! */
                        }
                        else{
                                const_e = &const_e_;
                                bzero( const_e, sizeof(expr) );
                                const_e->op = OP_const;
                                const_e->val.i_const = b_pos;
                                const_e->type = PRIM_int;
                                CP_fixup_const( const_e );
                                const_e = CSE_recurse( const_e, blk, blk->ebb_root, CSE_const, OP_const, false );
                                if( const_e == NULL ){
                                        return NULL;
                                }
                                bzero( &sh_e, sizeof(expr) );
                                sh_e.op = OP_asr;
                                sh_e.type = PRIM_int;
                                sh_e.a[0] = a_e;
                                sh_e.a[1] = const_e;
                                if( GCSE_recurse( const_e, blk, blk->ebb_root, CSE_binary, OP_asr ) == NULL ){
                                        return NULL;
                                }
                        }
                        return new_e;
                }
        }
        return CSE_binary( new_e, exprs_by_op, blk, update );
}

/*
 * CSE: Try to reduce rem into shifts
 */
CSE_FUNCTION( rem )
{
        expr * a0_e, * a1_e, * a_e, * t_e, * t1_e;
        expr * const_e, const_e_, sh_e;
        int b_cnt, b_pos;
        instr_block * ebb_root;

        ebb_root = blk->ebb_root;
        a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        if( EXPR_IS_CONST( a1_e ) && PRIM_TYPE( a1_e->type ) == PRIM_vm_int ){
                b_cnt = count_bits( a1_e->val.i_const, &b_pos );
                if( b_cnt == 1 ){
                        a_e = new_e->a[0];
                        if( update ){
                                lshl_int_const( t1_e, a_e, 32 - b_pos );
                                lshr_int_const( t1_e, t1_e, 32 - b_pos );
                                /* May need to add a neg sign too! */
                        }
                        else{
                                FAILED();
                        }
                        return new_e;
                }
        }
        return CSE_binary( new_e, exprs_by_op, blk, update );
}

/*
 * CSE: Local branch where order of arguments matters
 */
CSE_FUNCTION( branch )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * new_a0_e;
        expr * cmp_a1_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                if( new_a0_e == cmp_a0_e
                    && new_a1_e == cmp_a1_e
                    && new_e->type == cmp_e->type ){
                        if( update ){
                                LAB_free( new_e->val.label, blk->cfg );
                        }
                        return cur_e;
                }
        }
        return NULL;
}

/*
 * CSE: Local branch where order of comparison matters
 */
CSE_FUNCTION( cbranch )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * new_a0_e;
        expr * cmp_a1_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                if( new_a0_e == cmp_a0_e
                    && new_a1_e == cmp_a1_e
                    && new_e->type == cmp_e->type ){
                        if( update ){
                                LAB_free( new_e->val.label, blk->cfg );
                        }
                        return cur_e;
                }
        }
        return NULL;
}

/*
 * CSE: Local branch where order of comparison is irrelevant
 */
CSE_FUNCTION( cbranch_x )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * new_a0_e;
        expr * cmp_a1_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                if( ( ( new_a0_e == cmp_a0_e && new_a1_e == cmp_a1_e ) ||
                     ( new_a1_e == cmp_a0_e && new_a0_e == cmp_a1_e ) )
                     && new_e->type == cmp_e->type ){
                        if( update ){
                                LAB_free( new_e->val.label, blk->cfg );
                        }
                        return cur_e;
                }
        }
        return NULL;
}

/*
 * CSE: Local branch where order of arguments matters, 
 * where branch uses results from a comparison
 */
CSE_FUNCTION( cc_cbranch )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * new_a0_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );

                if( new_a0_e == cmp_a0_e
                    && new_e->type == cmp_e->type ){
                        if( update ){
                                LAB_free( new_e->val.label, blk->cfg );
                        }
                        return cur_e;
                }
        }
        return NULL;
}

/*
 * CSE: Local branch where order of arguments doesn't matter, 
 * where branch uses results from a comparison
 */
CSE_FUNCTION( cc_cbranch_x )
{
        return CSE_cc_cbranch( new_e, exprs_by_op, blk, update );
}

/*
 * CSE: Binary operation where order of arguments matters
 */
CSE_FUNCTION( cmp )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * cmp_a1_e;
        expr * new_a0_e, * new_a1_e;

        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        new_a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                cmp_a1_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[1] );

                if( (new_e->val.c_flags & (FCMP_l | FCMP_g)) == (cmp_e->val.c_flags & (FCMP_l | FCMP_g))
                   && new_a0_e == cmp_a0_e && new_a1_e == cmp_a1_e ){
                        if( update ){
                                if( cmp_e->flags & SE_no_generate ){ /* FIXME: correct fix for loading problem? */
                                        cmp_e->a[0]->uses.block++; 
                                        cmp_e->a[1]->uses.block++;
                                }
                        }
                        return cur_e;
                }
        }
        return NULL;
}

EXTERN_CSE_FUNCTION( call )
{
        expr * cur_e, * cmp_e, * cmp_a0_e, * new_a0_e;
        expr * frame1_a_e, * frame2_a_e;
        expr_frame * frame1, * frame2;
        int n;
        boolean match, fail;
        
        match = false;
        if( !(new_e->val.frame->flags & CALL_no_side_effect) 
#ifdef OPTIMIZE_INLINE
           || new_e->val.frame->flags & (CALL_inline | CALL_specialized_inline | CALL_specialized_failed)
#endif
           ){
                return NULL;
        }
        new_a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        for (cur_e = *exprs_by_op; cur_e != NULL; cur_e = cur_e->exprs_by_op) {
                cmp_e = EXPR_RESOLVE_ORIGINAL( cur_e );
                cmp_a0_e = EXPR_RESOLVE_ORIGINAL( cmp_e->a[0] );
                if ( new_a0_e != cmp_a0_e || new_e->type != cmp_e->type ) {
                        continue;
                }
                frame1 = new_e->val.frame;
                frame2 = cmp_e->val.frame;
                if (frame1->sz != frame2->sz) {
                        continue;
                }
                if( frame1->sz == 0 ){
                        match = true;
                        break;
                }
                fail = false;
                for (n = frame1->sz; --n >= 0;) {
                        frame1_a_e = EXPR_RESOLVE_ORIGINAL( frame1->a[n].expr );
                        frame2_a_e = EXPR_RESOLVE_ORIGINAL( frame2->a[n].expr );
                        if (frame1_a_e != frame2_a_e) {
                                fail = true;
                                break;
                        }
                }
                if( !fail ){
                        match = true;
                        break;
                }
                
        }
        if (match) {
                frame1 = new_e->val.frame;
#ifdef OPTIMIZE_INLINE
                if( frame1->flags & CALL_inline ){
                        return frame1->env->return_val;
                }
#endif
                return cur_e;
        }
        else {
                return NULL;
        }
}

/* Common subexpression elimination (CSE) function table */
cseFunc CSE_funcs[] = {
	0, /* nop */
	CSE_const, /* const_m1 */
	CSE_const, /* const_0 */
	CSE_const, /* const_1 */
	CSE_const, /* const */
	0, /* move */
	CSE_load, /* load */
	CSE_store, /* store */
	CSE_binary_x, /* add */
	CSE_binary, /* sub */
	CSE_mul, /* mul */
	CSE_div, /* div */
	CSE_rem, /* rem */
	CSE_unary, /* neg */
	CSE_binary, /* lsl */
	CSE_binary, /* lsr */
	CSE_binary, /* asr */
	CSE_binary_x, /* and */
	CSE_binary_x, /* or */
	CSE_binary_x, /* xor */
	CSE_unary, /* cvt */
	CSE_set_cc, /* set_cc */
	cbranch_eq_func( CSE_, _x ), /* cbranch_eq */
	cbranch_ne_func( CSE_, _x ), /* cbranch_ne */
	cbranch_lt_func( CSE_, ), /* cbranch_lt */
	cbranch_ge_func( CSE_, ), /* cbranch_ge */
	cbranch_gt_func( CSE_, ), /* cbranch_gt */
	cbranch_le_func( CSE_, ), /* cbranch_le */
	cbranch_ult_func( CSE_, ), /* cbranch_ult */
	0, /* branch */
	0, /* branch_indirect */
	0, /* returnarg */
	CSE_cmp, /* cmp */
	CSE_call, /* call */
	0, /* target */
	0, /* prologue */
	0, /* epilogue */
	0, /* eprologue */
        0, /* copy */
        0, /* inline_data */
        0, /* get_arg */
        0, /* preload */
        0, /* replace */
        0, /* specialize_sync */
        0, /* tableswitch */
        0, /* lookupswitch */
        0, /* native_instr_1a */
        0, /* native_instr_2a */
        0, /* cmov_eq */
        0, /* cmov_ne */
        0, /* cmov_lt */
        0, /* cmov_ge */
        0, /* cmov_gt */
        0 /* cmov_le */
};
#endif /* OPTIMIZE_CSE */

#ifdef OPTIMIZE_LOOPS
#define LOOP_FUNCTION( _name ) static boolean LOOP_##_name( instr_block * blk, expr * new_e, loop_info * loop_i ) 

LOOP_FUNCTION( const )
{
        /* MD */
        assert( new_e->type != PRIM_void );
        if( new_e->type == PRIM_long ){
                if( new_e->val.l_const >= -32768 && new_e->val.l_const <= 32767 ){
                        return false;
                }
        }
        else if( new_e->type == PRIM_int || new_e->type == PRIM_ref ){
                if( new_e->val.i_const >= -32768 && new_e->val.i_const <= 32767 ){
                        return false;
                }
        }
        return false; /* Disable for now */
//        return true;
}

LOOP_FUNCTION( unary )
{
        expr * a0_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        if( !EXPR_IS_CONST( a0_e ) ){
                if( !(a0_e->flags & SE_slot) ){
                        return false;
                }
                if( a0_e->flags & SE_global ){
                        if( !(loop_i->header->globals.out[a0_e->loc.slot.frame]->flags & SE_global) ){
                                return false;
                        }
                }
                else{
                        if( a0_e->loc.slot.frame >= blk->cfg->env.method.n_locals
                           || STATS_local_get( SO_store, loop_i->locals, a0_e->loc.slot.frame ) ){
                                return false;
                        }
                }
        }
        return true;
}

static boolean
field_conflict( field_list * fl, Field * f, Hjava_lang_Class * c )
{
        int offset, n;
        Hjava_lang_Class * cmp_c;
        field_stats * fs;
        
        if( f == NULL ){
                return true;
        }
        offset = STATS_field_get_offset( f, c );
        for( n = fl->length; --n >= 0; ){
                fs = fl->classes[n];
                cmp_c = fs->class;
                if( c->nfields > cmp_c->nfields ){
                        if( instanceof( c, cmp_c )
                           && offset <= cmp_c->bfsize + cmp_c->nsfields
                           && STATS_field_get( SO_store, fs->stats, offset ) ){
                                return true;
                        }
                }
                else if( c->nfields < cmp_c->nfields ){
                        if( instanceof( cmp_c, c )
                           && STATS_field_get( SO_store, fs->stats, offset ) ){
                                return true;
                        }
                }
                else{
                        if( ( instanceof( cmp_c, c ) || instanceof( c, cmp_c ) )
                           && offset <= cmp_c->bfsize + cmp_c->nsfields
                           && STATS_field_get( SO_store, fs->stats, offset ) ){
                                return true;
                        }
                }
        }
        return false;
}

static boolean
array_conflict( field_list * fl, Hjava_lang_Class * c )
{
        field_stats * fs;
        int n;

TRACE(  kprintf( "CHECK %s\n", c->name ); );
        for( n = fl->length; --n >= 0; ){
                fs = fl->classes[n];
                if( c == fs->class && STATS_array_get( SO_load, fs->stats ) ){
TRACE(                  kprintf( "Conflict %s\n", c->name ); );
                        return true;
                }
        }
        return false;
}

LOOP_FUNCTION( load )
{
        expr * base_e, * cmp_e;
        Field * f;
        Hjava_lang_Class * c;
        boolean dominates;
        instr_block * check_b, * pred_b;
        int i, j;

        if( !LOOP_unary( blk, new_e, loop_i ) ){
                return false;
        }
        if( new_e->flags & (SE_object_field | SE_static_field | SE_array_object) ){
                if( loop_i->flags & LOOP_has_java_call ){
                        return false;
                }
                if( new_e->flags & SE_array_object ){
                        if( array_conflict( &loop_i->fields, (Hjava_lang_Class*)new_e->a[1] ) ){
                                return false;
                        }
                }
                else{
		        base_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
                        f = (Field*)new_e->a[1];
                        c = base_e->class;
                        if( c == NULL ){
kprintf( "WARN: %s.%s: NULL class: base=[%d] offset=%d\n", blk->cfg->env.method.class->name, blk->cfg->env.method.method->name, new_e->a[0]->eid, new_e->val.i_const );
LOG(                            fprintf( blk->cfg->fd, "WARN: NULL class: base=[%d] offset=%d\n", new_e->a[0]->eid, new_e->val.i_const ); );
				FAILED();
				return false;
                        }
                        if( field_conflict( &loop_i->fields, f, c ) ){
                                return false;
                        }
                }
        }
        base_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        if( base_e->class != NULL ){
                cmp_e = base_e->flags & SE_global 
                        ? loop_i->header->globals.out[base_e->loc.slot.frame] 
                                : loop_i->header->locals.out[base_e->loc.slot.frame];
                if( !(cmp_e->flags & SE_null_checked) ){
// #define OPTIMIZE_AGGRESSIVE
#ifdef OPTIMIZE_AGGRESSIVE
                        if( !(base_e->flags & SE_global) && STATS_local_get( SO_loop_must_load, loop_i->locals, base_e->loc.slot.frame ) ){
                                return true;
                        }
                        else{
                                dominates = true;
                                for( i = loop_i->exits.length; --i >= 0; ){
                                        check_b = loop_i->exits.blocks[i];
                                        if( check_b == loop_i->header->succ.blocks[0] ){
                                                for( j = check_b->pred.length; --j >= 0; ){
                                                        pred_b = check_b->pred.blocks[j];
                                                        if( BL_contains( &loop_i->blocks, pred_b ) ){
                                                                if( !BV_is_set( pred_b->dom, blk->bid ) ){
                                                                        dominates = false;
                                                                }
                                                        }
                                                }
                                        }
                                        else{
                                                if( !BV_is_set( check_b->dom, blk->bid ) ){
                                                        dominates = false;
                                                }
                                        }
                                }
                                if( dominates ){
                                        return true;
                                }
                        }
#endif
TRACE(                  kprintf( "FAILED NULL CHECK: B%d [%d:%d:%c]\n", blk->bid, cmp_e->eid, 
                                cmp_e->loc.slot.frame, cmp_e->flags & SE_global ? 'G' : 'L' ); );
                        return false;
                }
        }
        /* FIXME: fail if object is array & matching/subclass array class is written */
        return true;
}

LOOP_FUNCTION( binary )
{
        expr * a0_e, * a1_e;

        a0_e = EXPR_RESOLVE_ORIGINAL( new_e->a[0] );
        a1_e = EXPR_RESOLVE_ORIGINAL( new_e->a[1] );
        if( !EXPR_IS_CONST( a0_e ) ){
                if( !(a0_e->flags & SE_slot) ){
                        return false;
                }
                if( a0_e->flags & SE_global ){
                        if( !(loop_i->header->globals.out[a0_e->loc.slot.frame]->flags & SE_global) ){
                                return false;
                        }
                }
                else{
                        if( a0_e->loc.slot.frame >= blk->cfg->env.method.n_locals
                           || STATS_local_get( SO_store, loop_i->locals, a0_e->loc.slot.frame ) ){
                                return false;
                        }
                }
        }
        if( !EXPR_IS_CONST( a1_e ) ){
                if( !(a1_e->flags & SE_slot) ){
                        return false;
                }
                if( a1_e->flags & SE_global ){
                        if( !(loop_i->header->globals.out[a1_e->loc.slot.frame]->flags & SE_global) ){
                                return false;
                        }
                }
                else{
                        if( a1_e->loc.slot.frame >= blk->cfg->env.method.n_locals 
                           || STATS_local_get( SO_store, loop_i->locals, a1_e->loc.slot.frame ) ){
                                return false;
                        }
                }
        }
        return true;
}

LOOP_FUNCTION( unary_branch )
{
        if( new_e->val.label->type == LAB_block_target ){
                return false;
        }
        return LOOP_unary( blk, new_e, loop_i );
}

LOOP_FUNCTION( binary_branch )
{
        if( new_e->val.label->type == LAB_block_target ){
                return false;
        }
        return LOOP_binary( blk, new_e, loop_i );
}

/* Loop optimization (LOOP) function table */
/* FIXUP: May be simplified/rolled into argument table in expr.c */
loopFunc LOOP_funcs[] = {
	0, /* nop */
	LOOP_const, /* const_m1 */
	LOOP_const, /* const_0 */
	LOOP_const, /* const_1 */
	LOOP_const, /* const */
	0, /* move */
	LOOP_load, /* load */
	0, /* store */
	LOOP_binary, /* add */
	LOOP_binary, /* sub */
	LOOP_binary, /* mul */
	LOOP_binary, /* div */
	LOOP_binary, /* rem */
	LOOP_unary, /* neg */
	LOOP_binary, /* lsl */
	LOOP_binary, /* lsr */
	LOOP_binary, /* asr */
	LOOP_binary, /* and */
	LOOP_binary, /* or */
	LOOP_binary, /* xor */
	LOOP_unary, /* cvt */
	LOOP_binary, /* set_cc */
	LOOP_binary_branch, /* cbranch_eq */
	LOOP_binary_branch, /* cbranch_ne */
	LOOP_unary_branch, /* cbranch_lt */
	LOOP_unary_branch, /* cbranch_ge */
	LOOP_unary_branch, /* cbranch_gt */
	LOOP_unary_branch, /* cbranch_le */
	LOOP_unary_branch, /* cbranch_ult */
	0, /* branch */
	0, /* branch_indirect */
	0, /* returnarg */
	LOOP_binary, /* cmp */
	0, /* call */
	0, /* target */
	0, /* prologue */
	0, /* epilogue */
	0, /* eprologue */
        0, /* copy */
        0, /* inline_data */
        0, /* get_arg */
        0, /* preload */
        0, /* replace */
        0, /* specialize_sync */
        0, /* tableswitch */
        0, /* lookupswitch */
        0, /* native_instr_1a */
        0, /* native_instr_2a */
        0, /* cmov_eq */
        0, /* cmov_ne */
        0, /* cmov_lt */
        0, /* cmov_ge */
        0, /* cmov_gt */
        0 /* cmov_le */
};
#endif

