/*
 * regs.c
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "env.h"
#include "classMethod.h"
#include "code.h"
#include "support.h"

#if defined(TGT_MIPS)
#include "regs-mips.def"
#elif defined(TGT_SPARC)
#include "regs-sparc.def"
#elif defined(TGT_ARM)
#include "regs-arm.def"
#elif defined(TGT_I386)
#include "regs-i386.def"
#endif

#define TEMP_BLOCK 10

static void REG_free( int, cfg_info * );
static void REG_assign( expr *, int, cfg_info *, int );

#ifdef  JIT_DBG
static char * role_name[] = {
	"temp",
	"persist",
	"arg",
	"reserved"
};
#endif

#define REG_copy_regfile( _context, _base_regfile ) \
        bcopy( (_base_regfile), (_context)->free, ( 2 * NUM_REG_TYPES + NUM_REGS ) * sizeof(reg_info) )

#define REG_clone_regfile( _context, _base_regfile ) \
	(_context)->free = gc_malloc_fixed( ( 2 * NUM_REG_TYPES + NUM_REGS ) * sizeof(reg_info) ); \
	(_context)->live = &(_context)->free[NUM_REG_TYPES]; \
	(_context)->all = &(_context)->live[NUM_REG_TYPES]; \
        REG_copy_regfile( _context, _base_regfile )

#define REG_remove( _regfile, _node ) \
	if( (_node)->reg.prev != REG_NULL ){ \
		(_regfile)[(_node)->reg.prev].reg.next = (_node)->reg.next; \
	} \
	if( (_node)->reg.next != REG_NULL ){ \
        	(_regfile)[(_node)->reg.next].reg.prev = (_node)->reg.prev; \
	} \
	(_node)->reg.prev = REG_NULL; \
	(_node)->reg.next = REG_NULL

#define REG_insert_as_next( _regfile, _list, _node ) \
	(_node)->reg.prev = (_list)->num; \
	if( (_list)->reg.next != REG_NULL ){ \
		(_regfile)[(_list)->reg.next].reg.prev = (_node)->num; \
	} \
	(_node)->reg.next = (_list)->reg.next; \
	(_list)->reg.next = (_node)->num

#define TEMP_NULL -1

#define TEMP_remove( _all_temps, _free_list, _node ) \
	if( (_node)->temp.prev == TEMP_NULL ){ \
		(_free_list) = (_node)->temp.next; \
	} \
	else{ \
		(_all_temps)[(_node)->temp.prev].temp.next = (_node)->temp.next; \
	} \
	if( (_node)->temp.next != TEMP_NULL ){ \
        	(_all_temps)[(_node)->temp.next].temp.prev = (_node)->temp.prev; \
	} \
	(_node)->temp.prev = TEMP_NULL; \
	(_node)->temp.next = TEMP_NULL

#define TEMP_insert( _all_temps, _free_list, _node ) \
	(_node)->temp.prev = TEMP_NULL; \
	if( (_free_list) != TEMP_NULL ){ \
		(_all_temps)[(_free_list)].temp.prev = (_node)->offset; \
	} \
	(_node)->temp.next = (_free_list); \
	(_free_list) = (_node)->offset

#if 0
static void
TEMP_print_all( cfg_info * cfg )
{
        int i;
        temp_info * cur_t;
        FILE * fd;

        fd = cfg->fd;
        fprintf( fd, "free: %d", cfg->env.block.temps.free );
        for( i = cfg->env.block.temps.limit; --i >= 0; ) {
                cur_t = &cfg->env.block.temps.table[i];
                fprintf( fd, "[%d]: {%d,%d}\n", i, cur_t->temp.prev, cur_t->temp.next );
        }
}

static void
REG_p( FILE * fd, int num ){
        if( num == REG_NULL ){
                fprintf( fd, "[nil]" );
        }
        else{
                fprintf( fd, "[%d]", num );
        }
}

static void
REG_print_all( cfg_info * cfg )
{
        int i;
        reg_info * cur_r;
        FILE * fd;

        fd = cfg->fd;
        fprintf( fd, "live 0: " ); REG_p( fd, cfg->env.block.regs.live[0].reg.next); fprintf( fd, "\n" );
        fprintf( fd, "live 1: " ); REG_p( fd, cfg->env.block.regs.live[1].reg.next); fprintf( fd, "\n" );
        fprintf( fd, "live 2: " ); REG_p( fd, cfg->env.block.regs.live[2].reg.next); fprintf( fd, "\n" );
        fprintf( fd, "live 3: " ); REG_p( fd, cfg->env.block.regs.live[3].reg.next); fprintf( fd, "\n" );
        fprintf( fd, "free 0: " ); REG_p( fd, cfg->env.block.regs.free[0].reg.next); fprintf( fd, "\n" );
        fprintf( fd, "free 1: " ); REG_p( fd, cfg->env.block.regs.free[1].reg.next); fprintf( fd, "\n" );
        fprintf( fd, "free 2: " ); REG_p( fd, cfg->env.block.regs.free[2].reg.next); fprintf( fd, "\n" );
        fprintf( fd, "free 3: " ); REG_p( fd, cfg->env.block.regs.free[3].reg.next); fprintf( fd, "\n" );
        for( i = NUM_REGS; --i >= 0; ) {
                cur_r = &cfg->env.block.regs.all[i];
                fprintf( fd, "[%d]: [%p] ", cur_r->num, cur_r->expr ); REG_p( fd, cur_r->reg.prev ); REG_p( fd, cur_r->reg.next ); fprintf( fd, "\n" );
        }
}
#endif

int
TEMP_new( cfg_info * cfg, expr * e ) {
        temp_info * cur_t, * all_temps;
        int offset, i;

        /* Temp slots can handle doubleword temporaries */
        all_temps = cfg->env.block.temps.table;
        if( cfg->env.block.temps.free == TEMP_NULL ){
                all_temps = gc_malloc_fixed( (cfg->env.block.temps.limit + TEMP_BLOCK) * sizeof(temp_info) );
                if( cfg->env.block.temps.table != NULL ){
                        bcopy( cfg->env.block.temps.table, all_temps, cfg->env.block.temps.limit * sizeof(temp_info) );
                        gc_free_fixed( cfg->env.block.temps.table );
                }
                cfg->env.block.temps.table = all_temps;
                for( i = cfg->env.block.temps.limit + TEMP_BLOCK; --i >= cfg->env.block.temps.limit; ){
                        cur_t = &all_temps[i];
                        cur_t->offset = i;
                        TEMP_insert( all_temps, cfg->env.block.temps.free, cur_t );
                }
                cfg->env.block.temps.limit += TEMP_BLOCK;
        }
        cur_t = &all_temps[cfg->env.block.temps.free];
        TEMP_remove( all_temps, cfg->env.block.temps.free, cur_t );
        if( cur_t->offset + 1 > cfg->env.block.temps.max ){
                cfg->env.block.temps.max = 2*(cur_t->offset+1);
                if( cfg->env.block.temps.max > cfg->env.global.max_temps ){
                        cfg->env.global.max_temps = cfg->env.block.temps.max;
                }
        }
        cur_t->expr = e;
        offset = cur_t->offset + 1 + cfg->env.global.cframe.max_slots;
#ifdef OPTIMIZE_GCSE
        offset += cfg->env.global.globals.length;
#endif
LOG(    fprintf( cfg->fd, "TEMP_new():\t%d (%d)\n", offset, cur_t->offset ); );
        return offset;
}

void
TEMP_free( cfg_info * cfg, int offset ) {
        temp_info * cur_t, * all_temps;

LOG(    fprintf( cfg->fd, "TEMP_free():\t%d\n", offset ); );
#ifdef OPTIMIZE_GCSE
        offset = offset - cfg->env.global.cframe.max_slots - cfg->env.global.globals.length - 1;
#else
        offset = offset - cfg->env.global.cframe.max_slots - 1;
#endif
        all_temps = cfg->env.block.temps.table;
        cur_t = &all_temps[offset];
        if( verify_check( offset >= 0 && offset <= cfg->env.block.temps.max,
                     cfg->fd, "TEMP_free(): out of bounds: %d\n", offset ) ){
                cur_t->expr = NULL;
                TEMP_insert( all_temps, cfg->env.block.temps.free, cur_t );
        }
}

#ifdef OPTIMIZE_INLINE
temp_context *
TEMP_copy_context( temp_context * src_c )
{
        temp_context * new_c;
        new_c = gc_malloc_fixed( sizeof(temp_context) );
        bcopy( src_c, new_c, sizeof(temp_context) );
        new_c->table = gc_malloc_fixed( src_c->limit * sizeof(temp_info) );
        bcopy( src_c->table, new_c->table, src_c->limit * sizeof(temp_info) );
        return new_c;
}
#endif

inline void
TEMP_push_block_env( cfg_info * cfg, block_env * b_env )
{
        cfg->env.block.temps.table = gc_malloc_fixed( cfg->env.block.temps.limit * sizeof(temp_info) );
        bcopy( b_env->temps.table, cfg->env.block.temps.table, cfg->env.block.temps.limit * sizeof(temp_info) );
}

inline void
TEMP_pop_block_env( cfg_info * cfg, block_env * b_env )
{
        gc_free_fixed( cfg->env.block.temps.table );
}

void
TEMP_reset_context( instr_block * blk )
{
        int i;
        temp_info * cur_t, * all_temps;
        expr * cur_e;
        cfg_info * cfg;
        /* Reassign exprs to temp slots */

        cfg = blk->cfg;
        all_temps = cfg->env.block.temps.table;
        for( i = (cfg->env.block.temps.max - 1)/2; --i >= 0; ){
                cur_t = &all_temps[i];
                cur_e = cur_t->expr;
                if( cur_e != NULL && !(cur_e->flags & SE_in_frame_slot) ){
LOG(                    fprintf( cfg->fd, "TEMP_reset_context(): reassign temp %d to %s\n", i + cfg->env.method.n_locals, SLOT_print( blk, cur_e ) ); );
                        cur_e->flags |= SE_in_frame_slot;
                }
        }
}


void
REG_init( cfg_info * cfg, reg_info * rf )
{
        int i;

        /* Build queue of regs, with temps first, then arg regs, then globals */
	REG_clone_regfile( &cfg->env.block.regs, rf );
        for (i = NUM_REGS; --i >= 0; ) {
                REG_free( i, cfg );
        }

        /* TEMP_init() */
        cfg->env.block.temps.free = TEMP_NULL;
}

void
REG_done( cfg_info * cfg )
{
        gc_free_fixed( cfg->env.block.regs.free );
        /* FIX: Iterate through contexts and free */

        /* TEMP_done() */
        gc_free_fixed( cfg->env.block.temps.table );
        /* FIX: Iterate through contexts and free */
}

void
REG_free_expr( expr * cur_e, cfg_info * cfg, boolean update_expr )
{
        reg_info * regfile;

        assert( cur_e != NULL );
        regfile = cfg->env.block.regs.all;
        if( update_expr ){
                cur_e->flags &= ~SE_in_reg ;
                if( !(cur_e->flags & SE_in_frame_slot) 
                   || ( EXPR_IS_CONST( cur_e ) && !( cur_e->flags & SE_slot ) ) ){
                        cur_e->flags &= ~SE_generated;
                }
        }
        if( regfile[cur_e->loc.reg.base].expr == cur_e ){
LOG(            REG_print( cfg->fd, &regfile[cur_e->loc.reg.base], REGOP_free, cfg ); );
                REG_free( cur_e->loc.reg.base, cfg );
                if( PRIM_SIZE( cur_e->type ) == 0x08 ){
LOG(                    REG_print( cfg->fd, &regfile[cur_e->loc.reg.pair], REGOP_free | REGOP_pair, cfg ); );
                        REG_free( cur_e->loc.reg.pair, cfg );
                }
        }
        else if( regfile[cur_e->loc.reg.base].expr != NULL ){
LOG(            fprintf( cfg->fd, "REG_free_expr(): [%d] != [%d]\n", regfile[cur_e->loc.reg.base].expr->eid, cur_e->eid ); );
//kprintf( "FALSE FREE\n" );
        }
}

static void
REG_free( int reg, cfg_info * cfg )
{
        reg_info * free_r, * regfile, * free_list;

        assert( reg >= 0 && reg < NUM_REGS );
        regfile = cfg->env.block.regs.all;
        free_r = &regfile[reg];
        free_list = &cfg->env.block.regs.free[free_r->role];
        REG_remove( regfile, free_r );
        free_r->expr = NULL;
        free_r->status &= REG_was_used;
        REG_insert_as_next( regfile, free_list, free_r );
}

void
REG_assign_expr( expr * cur_e, cfg_info * cfg, int type )
{
        reg_info * regfile, * cur_r;
        expr * tmp_e;

        cur_e->flags |= SE_in_reg;
        if( !(cfg->env.global.no_free) && cur_e->flags & SE_slot ){
                if( cur_e->flags & SE_last_local && cur_e->last_local != NULL ){
                        regfile = cfg->env.block.regs.all;
                        tmp_e = cur_e->last_local;
                        if( tmp_e->flags & SE_slot && tmp_e->loc.slot.frame == cur_e->loc.slot.frame ){
                                cur_r = &regfile[tmp_e->loc.reg.base];
                                if( cur_r->status & REG_assigned && cur_r->expr == tmp_e
                                   && ( ( tmp_e->uses.block == 0 && tmp_e->uses.other == 0 ) || EXPR_IS_CONST( tmp_e ) ) ){
LOG(            			fprintf( cfg->fd, "REG_free_expr(): local free\n" ); );
                                        kprintf( "**** local free ****\n" );
                                        REG_free_expr( tmp_e, cfg, true );
                                }
                        }
                }
        }
        REG_assign( cur_e, cur_e->loc.reg.base, cfg, type );
        if( PRIM_SIZE( cur_e->type ) == 0x08 ){
                REG_assign( cur_e, cur_e->loc.reg.pair, cfg, type | REGOP_pair );
        }        
}

static void
REG_assign( expr * cur_e, int reg, cfg_info * cfg, int type )
{
        reg_info * cur_r, * regfile, * live_list;

        assert( reg >= 0 && reg < NUM_REGS );
        regfile = cfg->env.block.regs.all;
        cur_r = &regfile[reg];
        live_list = &cfg->env.block.regs.live[cur_r->role];
        REG_remove( regfile, cur_r );
        REG_insert_as_next( regfile, live_list, cur_r );
        cur_r->status = REG_assigned | REG_was_used;
        if( type & REGOP_pair ){
                cur_r->status |= REG_pair;
        }
        cur_r->expr = cur_e;
LOG(    REG_print( cfg->fd, cur_r, type, cfg ); );
}

void
REG_new( expr * cur_e, instr_block * blk, int cnt, ... )
{
        int reg_role, cur_n;
        reg_info * cur_r, * regfile, * pair_r;
        expr * s_e, * old_e, * pair_e;
	va_list va_p;
        FILE * fd;
        cfg_info * cfg;
        expr * next_call;
        boolean spill_sideways;

        cur_r = 0;
        cfg = blk->cfg;
        regfile = cfg->env.block.regs.all;
        fd = cfg->fd;
        spill_sideways = cfg->env.global.spill_sideways;
#if 1
        if( cur_e->uses.block > 1 || cur_e->uses.other > 0 ){
                if( cur_e->uses.other == 0 ){
                        if( cfg->env.global.last_call == NULL ){
                                next_call = blk->call.first != NULL ? blk->call.first : NULL;
                        }
                        else{
                                next_call = cfg->env.global.last_call->a[1];
                        }
                        if( next_call == NULL || next_call->eid >= cur_e->uses.last ){
                                reg_role = REG_temp;
                        }
                        else{
                                reg_role = REG_persist;
                        }
                }
                else{
                        reg_role = REG_persist;
                }
        }
        else{
                reg_role = REG_temp;
        }
#else
        reg_role = ((cur_e->uses.block > 1) || (cur_e->uses.other > 0))
                ? REG_persist : REG_temp;
#endif

        va_start( va_p, cnt );
        while( --cnt >= 0 ){
                s_e = va_arg(va_p, expr *);
                cur_r = &regfile[s_e->loc.reg.base];
                cur_r->status |= REG_lock;
        }
        va_end(va_p);

        if( cur_e->flags & SE_preset_reg ){
                cur_e->flags |= SE_in_reg;
                cur_e->flags &= ~SE_preset_reg;
        }
        if( cur_e->flags & SE_in_reg ){
                cur_r = &regfile[cur_e->loc.reg.base];
                if( cur_r->status & REG_assigned ){
                        old_e = cur_r->expr;
                        assert( !(old_e->flags & SE_no_spill) );
                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                REG_free_expr( old_e, cfg, true );
                        }
                }
                if( PRIM_SIZE( cur_e->type ) == 0x08 ){
                        cur_r = &regfile[cur_e->loc.reg.pair];
                        if( cur_r->status & REG_assigned ){
                                old_e = cur_r->expr;
                                assert( !(old_e->flags & SE_no_spill) );
                                if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                        REG_free_expr( old_e, cfg, true );
                                }
                        }
                }
                REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(            fprintf( fd, "(preassigned)\n" ); );
        }
        else{
                /* First try assignment to recently freed registers */
                va_start( va_p, cnt );
                while( --cnt >= 0 ){
                        s_e = va_arg(va_p, expr *);
                        if ( !(s_e->flags & SE_in_reg)
                            || s_e->flags & SE_no_spill
                            || s_e->uses.block != 0 || s_e->uses.other != 0 ){
                                continue;
                        }
                        switch( PRIM_TYPE( cur_e->type ) ){
                        case PRIM_vm_label:
                        case PRIM_vm_int:
                                cur_r = &regfile[s_e->loc.reg.base];
                                if( cur_r->type & REG_int && cur_r->role == reg_role ){
LOG(                                    REG_print( fd, cur_r, REGOP_free, cfg ); );
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(  				        fprintf( fd, "(preferred, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                break;
                        case PRIM_vm_float:
                                cur_r = &regfile[s_e->loc.reg.base];
                                if( cur_r->type & REG_float && cur_r->role == reg_role ){
LOG(                                    REG_print( fd, cur_r, REGOP_free, cfg ); );
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(  				        fprintf( fd, "(preferred, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                break;
                        case PRIM_vm_long:
                        case PRIM_vm_double:
                        default:
                                FAILED();
                                break;
                        }
                }
                va_end(va_p);
                /* Now try free lists and finally use a live register */
                assert( reg_role == REG_temp || reg_role == REG_persist );
                switch( PRIM_TYPE( cur_e->type ) ){
                case PRIM_vm_long:
                        cur_n = cfg->env.block.regs.free[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                pair_r = &regfile[cur_n+1];
                                if( cur_r->type & REG_int && cur_r->role == reg_role 
                                   && cur_r->num % 2 == 0 && !(pair_r->status & REG_assigned)
				    && !(pair_r->role == REG_reserved) ){
                                        cur_e->loc.reg.base = cur_r->num;
                                        cur_e->loc.reg.pair = cur_r->num+1;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(free list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.free[!reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                pair_r = &regfile[cur_n+1];
                                if( cur_r->type & REG_int && cur_r->role == !reg_role 
                                   && cur_r->num % 2 == 0 && !(pair_r->status & REG_assigned)
				    && !(pair_r->role == REG_reserved) ){
                                        cur_e->loc.reg.base = cur_r->num;
                                        cur_e->loc.reg.pair = cur_r->num+1;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(free list, %s)\n", role_name[!reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                pair_r = &regfile[cur_n+1];
                                old_e = cur_r->expr;
                                if( cur_r->type & REG_int && cur_r->role == reg_role 
                                   && cur_r->num % 2 == 0 && !(pair_r->status & REG_assigned)
                                   && !(old_e->flags & SE_no_spill) && old_e->uses.block == 0
				    && !(pair_r->role == REG_reserved) ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        cur_e->loc.reg.pair = cur_r->num+1;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                pair_r = &regfile[cur_n+1];
                                old_e = cur_r->expr;
                                pair_e = pair_r->expr;
                                if( cur_r->type & REG_int && cur_r->role == reg_role 
                                   && cur_r->num % 2 == 0 
                                   && !(old_e->flags & SE_no_spill) ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        if( pair_r->status & REG_assigned ){
                                                if( REG_spill_expr( pair_e, blk, spill_sideways ) ){
                                                        REG_free_expr( pair_e, cfg, true );
                                                }
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        cur_e->loc.reg.pair = cur_r->num+1;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        break;
                case PRIM_vm_label:
                case PRIM_vm_int:
                        cur_n = cfg->env.block.regs.free[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                if( cur_r->type & REG_int && cur_r->role == reg_role ){
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(free list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.free[!reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                if( cur_r->type & REG_int && cur_r->role == !reg_role ){
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(free list, %s)\n", role_name[!reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                old_e = cur_r->expr;
                                if( cur_r->type & REG_int && cur_r->role == reg_role 
                                   && !(old_e->flags & SE_no_spill) && old_e->uses.block == 0 ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                old_e = cur_r->expr;
                                if( cur_r->type & REG_int && cur_r->role == reg_role 
                                   && !(old_e->flags & SE_no_spill) ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        break;
                case PRIM_vm_float:
                        reg_role = REG_temp; /* No fp persist on MIPS */
                        cur_n = cfg->env.block.regs.free[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                if( cur_r->type & REG_float && cur_r->role == reg_role ){
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(free list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                old_e = cur_r->expr;
                                if( cur_r->type & REG_float && cur_r->role == reg_role
                                   && !(old_e->flags & SE_no_spill) && old_e->uses.block == 0 ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                old_e = cur_r->expr;
                                if( cur_r->type & REG_float && cur_r->role == reg_role
                                   && !(old_e->flags & SE_no_spill) ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        break;
                case PRIM_vm_double:
                        reg_role = REG_temp; /* No fp persist on MIPS */
                        cur_n = cfg->env.block.regs.free[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                pair_r = &regfile[cur_n+1];
                                if( cur_r->type & REG_double && cur_r->role == reg_role
                                   && !(pair_r->status & REG_assigned) ){ 
                                        cur_e->loc.reg.base = cur_r->num;
                                        cur_e->loc.reg.pair = cur_r->num+1;
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(free list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                pair_r = &regfile[cur_n+1];
                                old_e = cur_r->expr;
                                pair_e = pair_r->expr;
                                if( cur_r->type & REG_double && cur_r->role == reg_role
                                   && !(old_e->flags & SE_no_spill)
                                   && !(pair_e->flags & SE_no_spill) && old_e->uses.block == 0 ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        cur_e->loc.reg.pair = cur_r->num+1;
                                        if( pair_r->status & REG_assigned ){
                                                if( REG_spill_expr( pair_e, blk, spill_sideways ) ){
                                                        REG_free_expr( pair_e, cfg, true );
                                                }
                                        }
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        cur_n = cfg->env.block.regs.live[reg_role].reg.next;
                        while( cur_n != REG_NULL ){
                                cur_r = &regfile[cur_n];
                                pair_r = &regfile[cur_n+1];
                                old_e = cur_r->expr;
                                pair_e = pair_r->expr;
                                if( cur_r->type & REG_double && cur_r->role == reg_role
                                   && !(old_e->flags & SE_no_spill)
                                   && !(pair_e->flags & SE_no_spill) ){
                                        if( REG_spill_expr( old_e, blk, spill_sideways ) ){
                                                REG_free_expr( old_e, cfg, true );
                                        }
                                        cur_e->loc.reg.base = cur_r->num;
                                        cur_e->loc.reg.pair = cur_r->num+1;
                                        if( pair_r->status & REG_assigned ){
                                                if( REG_spill_expr( pair_e, blk, spill_sideways ) ){
                                                        REG_free_expr( pair_e, cfg, true );
                                                }
                                        }
                                        REG_assign_expr( cur_e, cfg, REGOP_new | REGOP_no_cr );
LOG(    			        fprintf( fd, "(live list, %s)\n", role_name[reg_role] ); );
                                        goto finish;
                                }
                                cur_n = cur_r->reg.next;
                        }
                        break;
                default:
LOG(    		fprintf( fd, "[%d]: unspecified type\n", cur_e->eid ); );
                        FAILED();
                        break;
                }
                FAILED();
        }
 finish:
        va_start( va_p, cnt );
        while( --cnt >= 0 ){
                s_e = va_arg(va_p, expr *);
                cur_r = &regfile[s_e->loc.reg.base];
                cur_r->status &= ~REG_lock;
        }
        va_end(va_p);

}

boolean
REG_spill_expr( expr * cur_e, instr_block * blk, boolean try_persistent_reg )
{
        expr * new_e, move_e;
        cfg_info * cfg;
        int tmp_n, flags;
        reg_info * tmp_r, * regfile;
        block_env * b_env;

        verify_strict( cur_e->flags & SE_in_reg, blk->cfg->fd, "[%d]: Not in register\n", cur_e->eid );

        if( !EXPR_IS_CONST( cur_e ) && ( cur_e->uses.block > 0 || cur_e->uses.other > 0 )
            && cur_e->flags & SE_in_reg && !(cur_e->flags & SE_in_frame_slot) 
            && !(cur_e->flags & SE_no_spill) && cur_e->flags & SE_generated ){

                /* make sure we operate on a local expr */
                cfg = blk->cfg;
                if( cfg->env.global.in_nested_block ){
                        return false;
                }
                if( cur_e->eid < blk->exprs_head->eid 
                   && ( !(cur_e->flags & SE_slot) 
                        || ( cur_e->flags & SE_in_reg && !(cur_e->flags & SE_in_frame_slot) ) ) ){
                        b_env = &cfg->env.block;
#ifdef OPTIMIZE_INLINE
#if 1
                        while( true ){
                                if( cur_e->eid >= b_env->block->env->first_id ){
                                        break;
                                }
                                b_env = b_env->stack;
                        }
#else
                        /* Yikes, this is broken */
                        if( cur_e->eid < blk->env->first_id && b_env->block != NULL ){
                                FAILED();
                                while( true ){
                                        if( cur_e->eid >= b_env->block->env->first_id ){
                                                break;
                                        }
                                        b_env = b_env->stack;
                                }
                        }
#endif
#endif
                        BLK_backup_expr( cur_e, b_env, blk->cfg->fd );
#if 1
                        if( cfg->env.method.caller == NULL && cur_e->eid < blk->exprs_head->eid && !(cur_e->flags & SE_slot) && !blk->in_init ){
LOG(    		        fprintf( blk->cfg->fd, "REG_spill_expr(): bail\n" ); );
                                return false;
                        }
#endif
                }
                regfile = blk->cfg->env.block.regs.all;
                if( try_persistent_reg
                   && cur_e->type != PRIM_float && cur_e->type != PRIM_double && cur_e->type != PRIM_long ){
                        tmp_n = cfg->env.block.regs.free[REG_persist].reg.next;
                        while( tmp_n != REG_NULL ){
                                tmp_r = &regfile[tmp_n];
                                if( tmp_r->type & REG_int && tmp_r->role == REG_persist 
                                   && !(tmp_r->status & REG_lock) ){
LOG(   			 		fprintf( blk->cfg->fd, "REG_spill_expr(): move to %d from %d\n", tmp_n, cur_e->loc.reg.base ); );
                                        bzero( &move_e, sizeof(expr) );
                                        EXPR_make_move_to( &move_e, cur_e, blk, tmp_n, 0 );
                                        flags = cur_e->flags;
                                        ASM_move( &move_e, blk, false );
                                        REG_free_expr( cur_e, blk->cfg, true );
                                        cur_e->loc.reg.base = move_e.loc.reg.base;
                                        cur_e->loc.reg.pair = move_e.loc.reg.pair;
                                        cur_e->flags = flags;
                                        REG_assign_expr( cur_e, blk->cfg, REGOP_assign );
                                        return false;
                                }
                                tmp_n = tmp_r->reg.next;
                        }
                }
                cur_e->flags |= SE_in_frame_slot;
                if( !(cur_e->flags & SE_slot) ){
                        cur_e->loc.slot.frame = TEMP_new( blk->cfg, cur_e );
                }
                ASM_spill( cur_e, blk );
        }
        return true;
}

static void
REG_fake_spill_expr( expr * cur_e, instr_block * blk )
{
        assert( cur_e->flags & SE_in_reg );
        if( !EXPR_IS_CONST( cur_e ) && ( cur_e->uses.block > 0 || cur_e->uses.other > 0 )
           && !(cur_e->flags & SE_in_frame_slot) && cur_e->flags & SE_generated ){
                if( !(cur_e->flags & SE_slot) ){
                        cur_e->loc.slot.frame = TEMP_new( blk->cfg, cur_e );
                }
                ASM_spill( cur_e, blk );
        }
}

static void
REG_spill_expr_before_throw( expr * cur_e, instr_block * blk )
{
        assert( cur_e->flags & SE_in_reg );
        if( !(cur_e->flags & SE_in_frame_slot) 
           && cur_e->flags & SE_slot && cur_e->loc.slot.frame < blk->cfg->env.method.n_locals ){
                ASM_spill( cur_e, blk );
        }
}

void
REG_reload_expr( expr * cur_e, instr_block * blk )
{
//        verify_strict( cur_e->flags & SE_in_frame_slot, blk->cfg->fd, "Not in frame slot" );
        REG_new( cur_e, blk, 0 );
        ASM_reload( cur_e, blk );
}

static void
REG_fake_reload_expr( expr * cur_e, instr_block * blk )
{
        assert( cur_e->flags & SE_in_reg );
        if( !EXPR_IS_CONST( cur_e ) && ( cur_e->uses.block > 0 || cur_e->uses.other > 0 )
           && !(cur_e->flags & SE_in_frame_slot) && cur_e->flags & SE_generated ){
                if( !(cur_e->flags & SE_slot) ){
                        TEMP_free( blk->cfg, cur_e->loc.slot.frame );
                }
                ASM_reload( cur_e, blk );
        }
}

void
REG_spill_before_call( instr_block * blk, boolean try_persistent_reg, boolean spill_all )
{
        reg_info * cur_r, * regfile;
        expr * cur_e;
        int i;

LOG(    fprintf( blk->cfg->fd, "REG_spill_before_call():\n" ); );
        regfile = blk->cfg->env.block.regs.all;
        for (i = NUM_REGS; --i >= 0; ) {
                cur_r = &regfile[i];
                if( cur_r->status & REG_assigned && !(cur_r->status & REG_pair) ){
                        cur_e = cur_r->expr;
                        if( cur_r->role != REG_persist || spill_all ){
LOG(    			fprintf( blk->cfg->fd, "REG_spill_expr(): check [%d] %s\n",  cur_e->eid, reg_names[cur_r->num] ); );
                                if( REG_spill_expr( cur_e, blk, try_persistent_reg ) ){
                                        REG_free_expr( cur_e, blk->cfg, true );
                                }
                        }
                }
        }
}

/*
 * Spill that doesn't update reg/expr state (for block-local calls)
 */
void
REG_fake_spill_before_call( instr_block * blk )
{
        reg_info * cur_r, * regfile;
        expr * cur_e;
        int i;

LOG(    fprintf( blk->cfg->fd, "REG_fake_spill_before_call():\n" ); );
        regfile = blk->cfg->env.block.regs.all;
        for (i = NUM_REGS; --i >= 0; ) {
                cur_r = &regfile[i];
                if( cur_r->status & REG_assigned && !(cur_r->status & REG_pair) ){
                        cur_e = cur_r->expr;
                        if( cur_r->role != REG_persist ){
                                REG_fake_spill_expr( cur_e, blk );
                        }
                }
        }
LOG(    fprintf( blk->cfg->fd, ".\n" ); );
}


/*
 * Reload that doesn't update reg/expr state (for block-local calls)
 */
void
REG_fake_reload_after_call( instr_block * blk )
{
        reg_info * cur_r, * regfile;
        expr * cur_e;
        int i;

LOG(    fprintf( blk->cfg->fd, "REG_fake_reload_after_call():\n" ); );
        regfile = blk->cfg->env.block.regs.all;
        for (i = NUM_REGS; --i >= 0; ) {
                cur_r = &regfile[i];
                if( cur_r->status & REG_assigned && !(cur_r->status & REG_pair) ){
                        cur_e = cur_r->expr;
                        if( cur_r->role != REG_persist ){
                                REG_fake_reload_expr( cur_e, blk );
                        }
                }
        }
}

/*
 * Spill that doesn't update reg/expr state (for throw)
 */
void
REG_spill_before_throw( instr_block * blk )
{
        reg_info * cur_r, * regfile;
        int i;

LOG(    fprintf( blk->cfg->fd, "REG_spill_before_throw():\n" ); );
        regfile = blk->cfg->env.block.regs.all;
        for (i = NUM_REGS; --i >= 0; ) {
                cur_r = &regfile[i];
                if( cur_r->status & REG_assigned && !(cur_r->status & REG_pair) ){
                        if( cur_r->expr->flags & SE_slot && !(cur_r->expr->flags & SE_in_frame_slot) ){
                                REG_spill_expr_before_throw( cur_r->expr, blk );
                        }
                }
        }
LOG(    fprintf( blk->cfg->fd, ".\n" ); );
}

inline void
REG_push_block_env( cfg_info * cfg, block_env * b_env )
{
	REG_clone_regfile( &cfg->env.block.regs, b_env->regs.free );
}

inline void
REG_pop_block_env( cfg_info * cfg, block_env * b_env )
{
        reg_info * base_regs, * new_regs;
        int i;

        base_regs = b_env->regs.all;
        new_regs = cfg->env.block.regs.all;
        for( i = NUM_REGS; --i >= 0; ) {
                base_regs[i].status |= new_regs[i].status & REG_was_used;
        }
        gc_free_fixed( cfg->env.block.regs.free );
}

#ifdef OPTIMIZE_INLINE
reg_context *
REG_copy_context( reg_context * src_c )
{
        reg_context * new_c;
        new_c = gc_malloc_fixed( sizeof(reg_context) );
	REG_clone_regfile( new_c, src_c->free );
        return new_c;
}

void
REG_backup_exprs( cfg_info * cfg )
{
        reg_info * rf;
        expr * e, * head_e;
        int i;
        instr_block * blk;

        rf = cfg->env.block.regs.all;
        blk = cfg->env.block.block;
        head_e = blk->exprs_head;
        if( head_e == NULL ){
                return;
        }
        for( i = NUM_REGS; --i >= 0; ) {
                if( rf[i].status & REG_assigned ){
                        e = rf[i].expr;
                        if( e->eid < head_e->eid ){
                                REG_free( i, cfg );
                        }
                }
        }
}

/*
 * After specialization, we have two reg contexts, of which we need to make
 * the exit state equalavent.  This performs the necessary reloads and moves.
 */
void
REG_merge_inline( reg_context * reg_c, instr_block * blk )
{
        reg_info * cur_r, * regfile;
        expr move_e, * cur_e;
        cfg_info * cfg;
        int i;

        cfg = blk->cfg;
        regfile = reg_c->all;
LOG(    fprintf( cfg->fd, "REG_merge_inline():\n" ); );
        for( i = NUM_REGS; --i >= 0; ){
                cur_r = &regfile[i];
                cur_e = cur_r->expr;
                if( !(cur_r->status & REG_assigned) 
                   || ( cur_r->status & REG_assigned
                       && cur_e->flags & SE_in_reg
                       && cur_e->loc.reg.base == cur_r->num
                       && cur_e->loc.reg.pair == cur_r->num+1 )
                   || cur_r->status & REG_pair ){
                        continue;
                }
                if( cur_e->flags & SE_in_reg
                   && cur_e->uses.block == 0
                   && cur_e->uses.other == 0 ){
                        REG_free_expr( cur_e, cfg, true );
                }
                else if( !(cur_e->flags & SE_in_reg) ){
                        cur_e->flags |= SE_preset_reg;
                        cur_e->loc.reg.base = cur_r->num; /* in case it was changed by a reg move */
                        cur_e->loc.reg.pair = cur_r->num+1;
                        REG_reload_expr( cur_e, blk );
                }                
                else if( cur_e->loc.reg.base != cur_r->num ){
                        bzero( &move_e, sizeof(expr) );
                        EXPR_make_move_to( &move_e, cur_e, blk, cur_r->num, cur_r->num+1 );
                        ASM_move( &move_e, blk, false );
                        /* below necessary? */
                        cur_e->loc.reg.base = cur_r->num;
                        cur_e->loc.reg.pair = cur_r->num+1;
                        REG_assign_expr( cur_e, cfg, REGOP_assign );
                        cur_e->flags |= SE_in_reg | SE_generated;
                }                
        }
}
#endif

void
REG_reset_context( instr_block * blk )
{
        reg_info * cur_r, * regfile;
        cfg_info * cfg;
        expr * cur_e;
        int i;

        cfg = blk->cfg;
        regfile = cfg->env.block.regs.all;
        for (i = NUM_REGS; --i >= 0; ) {
                cur_r = &regfile[i];
                cur_e = cur_r->expr;
                if( !(cur_r->status & REG_assigned) 
                   || cur_r->status & REG_pair ){
                        continue;
                }
                if( !(cur_e->flags & SE_in_reg) ){
                        cur_e->loc.reg.base = cur_r->num; /* in case it was changed by a reg move */
                        cur_e->loc.reg.pair = cur_r->num+1;
                        REG_assign_expr( cur_e, cfg, REGOP_assign );
                        cur_e->flags |= SE_in_reg | SE_generated;
LOG(                    fprintf( cfg->fd, "REG_reset_context(): reassign %s to %s\n", reg_names[cur_r->num], SLOT_print( blk, cur_e ) ); );
                }
        }
}

static void
REG_set_base_context_pass1( int n_slots, expr ** slots, expr ** temps, instr_block * blk )
{
        int i;
        expr * cur_e, * move_e;
        cfg_info * cfg;

        cfg = blk->cfg;
LOG(    fprintf( cfg->fd, "REG_set_base_context(): pass1\n" ); );
        for( i = n_slots; --i >= 0; ) {
                cur_e = slots[i];
                if( cur_e->flags & SE_in_reg && !(cur_e->flags & SE_out_slot_reg) ){
                        cur_e->flags |= SE_in_slot_reg;
                }
                if( cur_e->flags & SE_reg_assigned ){
                        if( cur_e->flags & SE_in_reg 
                           && ( cur_e->loc.reg.base != cur_e->loc.start.base
                               || cur_e->loc.reg.pair != cur_e->loc.start.pair ) ){
                                cur_e->flags |= SE_will_assign;
                                temps[i] = move_e = gc_malloc_fixed( sizeof(expr) );
                                EXPR_make_move_to( move_e, cur_e, blk, cur_e->loc.reg.base, cur_e->loc.reg.pair );
                                cur_e->loc.reg.base = cur_e->loc.start.base;
                                cur_e->loc.reg.pair = cur_e->loc.start.pair;
                                // cur_e->flags |= SE_generated;
                                cur_e->flags |= SE_generated | SE_in_slot_reg; /* removal generates error in java_cup */
                                REG_assign_expr( cur_e, cfg, REGOP_assign );
                        }
                        else{
                                cur_e->flags |= SE_in_reg | SE_generated;
                                cur_e->loc.reg.base = cur_e->loc.start.base;
                                cur_e->loc.reg.pair = cur_e->loc.start.pair;
                                REG_assign_expr( cur_e, cfg, REGOP_assign );
                        }
                }
                else{
                        if( cur_e->flags & SE_in_reg ){
                                cur_e->flags |= SE_will_assign;
                        }
                }
        }
}

static void
REG_set_base_context_pass2( int n_slots, expr ** slots, expr ** temps, instr_block * blk )
{
        int i;
        expr * cur_e, * move_e;
        cfg_info * cfg;

        cfg = blk->cfg;
LOG(    fprintf( cfg->fd, "REG_set_base_context(): pass2\n" ); );
        for( i = n_slots; --i >= 0; ) {
                cur_e = slots[i];
                if( cur_e->flags & SE_will_assign && !(cur_e->flags & SE_out_slot_reg) ){
                        cur_e->flags &= ~SE_will_assign;
                        if( cur_e->flags & SE_reg_assigned ){
                                /* A bit convolved, but correct */
                                move_e = temps[i];
                                if( cur_e->flags & SE_in_reg ){
                                        ASM_move( move_e, blk, false );
                                        REG_free_expr( cur_e, cfg, false );
                                        cur_e->loc.reg.base = move_e->loc.reg.base;
                                        cur_e->loc.reg.pair = move_e->loc.reg.pair;
                                        REG_assign_expr( cur_e, cfg, REGOP_assign );
                                }
                                else{
                                        cur_e->loc.reg.base = move_e->loc.reg.base;
                                        cur_e->loc.reg.pair = move_e->loc.reg.pair;
                                        cur_e->flags |= SE_in_reg;
                                        REG_new( cur_e, blk, 0 );
                                        ASM_reload( cur_e, blk );
                                }
                                gc_free_fixed( move_e );
                        }
                        else{
                                if( cur_e->uses.block == 0 ){ /* Hack */
                                        cur_e->uses.block++;
LOG(    			        fprintf( blk->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( blk, cur_e ), cur_e->uses.block ); );
                                }
                                REG_new( cur_e, blk, 0 );
                                ASM_reload( cur_e, blk );
                        }
                }
        }
}

static void
REG_set_base_context_pass3( int n_slots, expr ** slots, expr ** temps, instr_block * blk )
{
        int i;
        expr * cur_e, * move_e;
        cfg_info * cfg;
        reg_info * reg_i;
        reg_info * regfile;

        cfg = blk->cfg;
        regfile = cfg->env.block.regs.all;
LOG(    fprintf( cfg->fd, "REG_set_base_context(): pass3\n" ); );
        for( i = n_slots; --i >= 0; ) {
                cur_e = slots[i];
                if( cur_e->flags & SE_will_assign && cur_e->flags & SE_out_slot_reg ){
                        cur_e->flags &= ~SE_will_assign;
                        if( cur_e->flags & SE_reg_assigned ){
                                /* A bit convolved, but correct */
                                move_e = temps[i];
                                reg_i = &regfile[move_e->loc.reg.base];
                                if( !(reg_i->status & REG_assigned) 
                                   || ( reg_i->status & REG_assigned && !(reg_i->expr->flags & SE_in_slot_reg) ) ){
                                        if( cur_e->flags & SE_in_reg ){
                                                ASM_move( move_e, blk, false );
                                                REG_free_expr( cur_e, cfg, false );
                                                cur_e->loc.reg.base = move_e->loc.reg.base;
                                                cur_e->loc.reg.pair = move_e->loc.reg.pair;
                                                REG_assign_expr( cur_e, cfg, REGOP_assign );
                                        }
                                        else{
                                                cur_e->loc.reg.base = move_e->loc.reg.base;
                                                cur_e->loc.reg.pair = move_e->loc.reg.pair;
                                                cur_e->flags |= SE_in_reg;
                                                REG_new( cur_e, blk, 0 );
                                                ASM_reload( cur_e, blk );
                                        }
                                }
                                gc_free_fixed( move_e );
                        }
                        else{
                                reg_i = &regfile[cur_e->loc.reg.base];
                                if( !(reg_i->status & REG_assigned) 
                                   || ( reg_i->status & REG_assigned && !(reg_i->expr->flags & SE_in_slot_reg) ) ){
                                        if( cur_e->uses.block == 0 ){ /* Hack */
                                                cur_e->uses.block++;
LOG(    			        fprintf( blk->cfg->fd, "\t%s: uses++ {%d}\n", SLOT_print( blk, cur_e ), cur_e->uses.block ); );
                                        }
                                        REG_new( cur_e, blk, 0 );
                                        ASM_reload( cur_e, blk );
                                }
                                else{
                                        cur_e->flags &= ~SE_in_reg;
LOG(    			        fprintf( blk->cfg->fd, "\t%s: unassigned (v2)\n", SLOT_print( blk, cur_e ) ); );
                                }
                        }
                }
        }
}

void
REG_set_base_context( instr_block * blk )
{
        expr ** locals, ** tmp_l;
        reg_info * cur_r, * regfile;
        cfg_info * cfg;
        int n_slots, i;
        Method * method;
        int n_globals;
#ifdef OPTIMIZE_GCSE
        expr ** globals;
#endif

        cfg = blk->cfg;
        method = cfg->env.method.method;
        regfile = cfg->env.block.regs.all;
LOG(    fprintf( cfg->fd, "REG_set_base_context(): B%d\n", blk->bid ); );
        n_slots = blk->sp.in + blk->cfg->env.method.n_locals;
        locals = blk->locals.in;
#ifdef OPTIMIZE_GCSE
        n_globals = blk->cfg->env.global.globals.length;
        globals = blk->globals.in;
#else
        n_globals = 0;
#endif

        for (i = 2 * NUM_REG_TYPES; --i >= 0; ) {
                cur_r = &cfg->env.block.regs.free[i];
                cur_r->reg.prev = REG_NULL;
                cur_r->reg.next = REG_NULL;
        }
        for (i = NUM_REGS; --i >= 0; ) {
                cur_r = &regfile[i];
                cur_r->status &= REG_was_used;
                cur_r->reg.prev = REG_NULL;
                cur_r->reg.next = REG_NULL;
                cur_r->expr = NULL;
                REG_free( i, cfg );
        }
        if( blk == blk->cfg->env.method.root ){
                return;
        }
        tmp_l = gc_malloc_fixed( ( n_slots + n_globals ) * sizeof(expr*) );
        REG_set_base_context_pass1( n_slots, locals, tmp_l, blk );
#ifdef OPTIMIZE_GCSE
        REG_set_base_context_pass1( n_globals, globals, &tmp_l[n_slots], blk );
#endif
	/* Load register to be live on second pass, to prevent conflicts w/ incoming live regs */
        REG_set_base_context_pass2( n_slots, locals, tmp_l, blk );
#ifdef OPTIMIZE_GCSE
        REG_set_base_context_pass2( n_globals, globals, &tmp_l[n_slots], blk );
#endif
        REG_set_base_context_pass3( n_slots, locals, tmp_l, blk );
#ifdef OPTIMIZE_GCSE
        REG_set_base_context_pass3( n_globals, globals, &tmp_l[n_slots], blk );
#endif
LOG(    fprintf( cfg->fd, "REG_set_base_context(): done\n" ); );
        gc_free_fixed( tmp_l );
}


