/*
 * schedule.c
 * Machine instruction-level sequencing
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "env.h"

static int iid = 0;

#ifndef OPTIMIZE_MEMORY
static instr_info * instr_cache = NULL;

#define SCH_NUM_REGS 	(NUM_REGS + SCH_AUX_REGS)

instr_info *
SCH_new() 
{
        instr_info * new_i;

        if( instr_cache != NULL ){
                new_i = instr_cache;
                instr_cache = new_i->instr.next;
                bzero( new_i, sizeof(instr_info) );
        }
        else {
                new_i = gc_malloc_fixed( sizeof(instr_info) );
        }
        new_i->iid = iid++;
        return new_i;
}

void
SCH_free( instr_info * i )
{
        i->instr.next = instr_cache;
        instr_cache = i;
}
#endif

void
SCH_init_( cfg_info * cfg )
{
        int n;
        instr_info ** rf, * reg, * cur_i;

        rf = cfg->env.global.schedule.regs = gc_malloc_fixed( SCH_NUM_REGS * sizeof(instr_info*) );
        reg = cfg->env.global.schedule.base = gc_malloc_fixed( SCH_NUM_REGS * sizeof(instr_info) );
        for( n = SCH_NUM_REGS; --n >= 0; ){
                cur_i = rf[n] = &reg[n];
                cur_i->flags = SCH_base_reg | SCH_sorted | SCH_generated;
        }
}

void
SCH_reset( cfg_info * cfg )
{
        int n;
        instr_info ** rf, * reg, * cur_i;

        bzero( cfg->env.global.schedule.regs, SCH_NUM_REGS * sizeof(instr_info*) );
        bzero( cfg->env.global.schedule.delay.instrs, cfg->env.global.schedule.delay.limit * sizeof(instr_info*) );
        cfg->env.global.schedule.delay.length = 0;
        cfg->env.global.schedule.roots.length = 0;
        cfg->env.global.schedule.last_branch = NULL;
        cfg->env.global.schedule.branch_expr = NULL;
        cfg->env.global.schedule.call_label = NULL;
        iid = 0;

        rf = cfg->env.global.schedule.regs;
        reg = cfg->env.global.schedule.base;
        for( n = SCH_NUM_REGS; --n >= 0; ){
                cur_i = rf[n] = &reg[n];
                cur_i->flags = SCH_base_reg | SCH_sorted | SCH_generated;
                cur_i->uses = 0;
        }
}

void
SCH_print_instr_( FILE * fd, instr_info * i, instr_block * blk, boolean print_instr )
{
        instr_info * a_i;
        int n;
        cfg_info * cfg;

        cfg = blk->cfg;
        fprintf( fd, " i%-2d {%2d} f%02x :", i->iid, i->uses, i->flags );
        for( n = 0; n < N_SCH_ARGS; n++ ){
                a_i = i->a[n];
                if( a_i == NULL ){
                        fprintf( fd, "_   " );
                }
                else if( a_i->flags & SCH_base_reg ){
                        fprintf( fd, "B   " );
                }
                else{
                        fprintf( fd, "i%-2d ", a_i->iid );
                }
        }
        if( print_instr ){
                DISASM_instruction( fd, blk, i->addr );
        }
        else{
                fprintf( fd, "\n" );
        }
}

void
SCH_print_by_delay( instr_block * blk )
{
        instr_list * il;
        instr_info * cur_i, * last_branch;
        int n;
        cfg_info * cfg;

        cfg = blk->cfg;
        il = &cfg->env.global.schedule.delay;
        for( n = il->length+1; --n >= 0; ){
                cur_i = il->instrs[n];
                if( cur_i == NULL ){
                        continue;
                }
LOG(            fprintf( cfg->fd, "delay %d:\n", n ); );
                while( cur_i != NULL ){
LOG(                    SCH_print_instr_( cfg->fd, cur_i, blk, true ); );
                        cur_i = cur_i->instr.next;
                }
        }
        last_branch = cfg->env.global.schedule.last_branch;
        if( last_branch != NULL ){
LOG(            fprintf( cfg->fd, "branch:\n" );
                SCH_print_instr_( cfg->fd, last_branch, blk, true ); );
        }
}

uint8 *
SCH_get_buffer( cfg_info * cfg )
{
        int code_sz, buf_sz;

LOG(    fprintf( cfg->fd, "SCH_schedule(): 0x%x - 0x%x\n", 
                cfg->env.global.schedule.code.start - cfg->env.global.pbase, 
                cfg->env.global.schedule.code.end - cfg->env.global.pbase ); );
        code_sz = cfg->env.global.schedule.code.end - cfg->env.global.schedule.code.start;
        buf_sz = cfg->env.global.schedule.buffer.end - cfg->env.global.schedule.buffer.start;
        if( buf_sz < code_sz ){
                gc_free_fixed( cfg->env.global.schedule.buffer.start );
                cfg->env.global.schedule.buffer.start = gc_malloc_fixed( code_sz );
                cfg->env.global.schedule.buffer.end = cfg->env.global.schedule.buffer.start + code_sz;
        }
        return cfg->env.global.schedule.buffer.start;
}

void
SCH_schedule( instr_block * blk )
{
        int base_delay, cur_delay, tick, code_sz;
        instr_list * dl;
        instr_info * cur_i, * next_i, * last_branch, * branch_delay_slot;
        instr_info * a0_i, * a1_i, * a2_i;
#ifdef SCH_2DST
        instr_info * a3_i;
#endif
        uint8 * code_p;
        boolean scheduled, can_schedule_ds;
        expr * e;
        cfg_info * cfg;
        label_info * l;

        cfg = blk->cfg;
        dl = &cfg->env.global.schedule.delay;
        base_delay = dl->length;
        tick = 1;

        last_branch = cfg->env.global.schedule.last_branch;
        branch_delay_slot = NULL;
        if( last_branch != NULL ){
                cur_i = dl->instrs[0];
                while( cur_i != NULL ){
                        a0_i = cur_i->a[0];
                        a1_i = cur_i->a[1];
                        a2_i = cur_i->a[2];
#ifdef SCH_2DST
                        a3_i = cur_i->a[3];
                        if( a3_i == NULL
                            && !( ( a0_i != NULL && a0_i->flags & SCH_overwritten && a0_i != a2_i )
                              || ( a1_i != NULL && a1_i->flags & SCH_overwritten && a1_i != a2_i ) ) ){
#else
                        if( !( ( a0_i != NULL && a0_i->flags & SCH_overwritten && a0_i != a2_i )
                              || ( a1_i != NULL && a1_i->flags & SCH_overwritten && a1_i != a2_i ) ) ){
#endif
                                /* Put smallest delay here */
                                branch_delay_slot = cur_i;
                        }
                        cur_i = cur_i->instr.next;
                }
                if( branch_delay_slot != NULL ){
LOG(                    fprintf( cfg->fd, "delay_slot:\n" );
                        SCH_print_instr_( cfg->fd, branch_delay_slot, blk, true ); );
                        SCH_REMOVE( dl->instrs[0], branch_delay_slot );
                }
        }
        code_p = SCH_get_buffer( cfg );
        while( base_delay >= 0 ){
                cur_delay = base_delay;
                scheduled = false;
                while( dl->instrs[cur_delay] == NULL && cur_delay >= 0 ){
                        base_delay = --cur_delay;
                }
                can_schedule_ds = false;
                while( !scheduled && cur_delay >= 0 ){
                        cur_i = dl->instrs[cur_delay];
                        while( !scheduled && cur_i != NULL ){
                                next_i = cur_i->instr.next;
                                a0_i = cur_i->a[0];
                                a1_i = cur_i->a[1];
                                a2_i = cur_i->a[2];
#ifdef SCH_2DST
                                a3_i = cur_i->a[3];
#endif
                                if( ( a0_i == NULL 
                                     || ( a0_i->flags & SCH_generated 
                                         && a0_i->delay.total + a0_i->delay.node < tick ) )
                                   && ( a1_i == NULL 
                                       || ( a1_i->flags & SCH_generated 
                                           && a1_i->delay.total + a1_i->delay.node < tick ) )
                                   && ( a2_i == NULL
                                       || ( a2_i->flags & SCH_generated 
                                           && a2_i->delay.total + a2_i->delay.node < tick
                                           && ( a2_i->uses == 1 
                                               || ( a2_i->uses == 2 && (a2_i == a0_i || a2_i == a1_i) )
                                               || ( a2_i->uses == 3 && a2_i == a0_i && a2_i == a1_i )
                                               )
                                           ) )
#ifdef SCH_2DST
                                   && ( a3_i == NULL
                                       || ( a3_i->flags & SCH_generated 
                                           && a3_i->delay.total + a3_i->delay.node < tick
                                           ) )
#endif
                                   ){
                                        SCH_REMOVE( dl->instrs[cur_delay], cur_i );
                                        SCH_free( cur_i );
                                        if( a0_i != NULL ){
                                                a0_i->uses--;
                                        }
                                        if( a1_i != NULL ){
                                                a1_i->uses--;
                                        }
                                        if( a2_i != NULL ){
                                                a2_i->uses--;
                                        }
#ifdef SCH_2DST
                                        if( a3_i != NULL ){
                                                a3_i->uses--;
                                        }
#endif
LOG(                                    SCH_print_instr_( cfg->fd, cur_i, blk, true ); );
                                        (*(uint32*)code_p) = (*(uint32*)cur_i->addr);
                                        code_p += WORD_SZ;
                                        cur_i->flags |= SCH_generated;
                                        cur_i->delay.total = tick++;
                                        scheduled = true;
                                }
                                cur_i = next_i;
                        }
                        cur_delay--;
                }
                if( !scheduled ){
                        tick++;
                }
        }
        if( last_branch != NULL ){
LOG(            fprintf( cfg->fd, "B" );
                SCH_print_instr_( cfg->fd, last_branch, blk, true ); );
                (*(uint32*)code_p) = (*(uint32*)last_branch->addr);
                code_p += WORD_SZ;
                if( branch_delay_slot != NULL ){
LOG(                    fprintf( cfg->fd, "D" );
                        SCH_print_instr_( cfg->fd, branch_delay_slot, blk, true ); );
                        (*(uint32*)code_p) = (*(uint32*)branch_delay_slot->addr);
                        SCH_free( branch_delay_slot );
                }
                else{
                        (*(uint32*)code_p) = (uint32)I_NOP;
                }
                code_p += WORD_SZ;
        }
        if( enable_schedule ){
                code_sz = code_p - cfg->env.global.schedule.buffer.start;
                bcopy( cfg->env.global.schedule.buffer.start, 
                      cfg->env.global.schedule.code.start, code_sz );
                blk->native.end = cfg->env.global.schedule.code.start + code_sz;
                e = cfg->env.global.schedule.branch_expr;
                if( e != NULL && EXPR_IS_BRANCH( e ) ){
                        if( branch_delay_slot != NULL ){
                                blk->flags |= BLK_filled_delay_slot;
                        }
LOG(                    fprintf( cfg->fd, "SCH_schedule(): branch label change: 0x%x --> ", 
                                LAB_get_from_addr( e->val.label ) - cfg->env.global.pbase ); );
                        e->val.label->from.addr = (uint8*)(blk->native.end - blk->native.start - 2*WORD_SZ);
LOG(                    fprintf( cfg->fd, "0x%x\n", 
                                LAB_get_from_addr( e->val.label ) - cfg->env.global.pbase ); );
                }
                if( last_branch != NULL && last_branch->flags & SCH_last_call ){
                        l = cfg->env.global.schedule.call_label;
LOG(                    fprintf( cfg->fd, "SCH_schedule(): call label change: 0x%x --> ", 
                                LAB_get_from_addr( l ) ); );
                        l->from.addr = (uint8*)(blk->native.end - blk->native.start - 2*WORD_SZ);
LOG(                    fprintf( cfg->fd, "0x%x\n", 
                                LAB_get_from_addr( l ) ); );
                }
LOG(            fprintf( cfg->fd, "SCH_schedule(): now 0x%x - 0x%x\n", 
                        cfg->env.global.schedule.code.start - cfg->env.global.pbase, 
                        blk->native.end - cfg->env.global.pbase ); );
        }
}

static void
SCH_compute_delay( instr_list * dl, instr_info * cur_i )
{
        instr_info * a_i;
        int delay, n;

        for( n = N_SCH_ARGS; --n >= 0; ){
                a_i = cur_i->a[n];
                if( a_i == NULL || a_i->flags & SCH_base_reg ){
                        continue;
                }
                delay = cur_i->delay.total + a_i->delay.node + 1;
                if( delay > a_i->delay.total ){
                        if( a_i->flags & SCH_sorted ){
                                SCH_REMOVE( dl->instrs[a_i->delay.total], a_i );
                        }
                        if( delay >= dl->length ){
                                IL_set_limit( dl, delay );
                                dl->length = delay;
                        }
                        a_i->delay.total = delay;
                        a_i->flags |= SCH_sorted;
                        SCH_INSERT( dl->instrs[delay], a_i );
                        SCH_compute_delay( dl, a_i );
                }
        }
}

void
SCH_order( cfg_info * cfg )
{
        instr_info * cur_i, ** regs;
        instr_list * dl, * roots;
        int n;

        dl = &cfg->env.global.schedule.delay;
        IL_set_limit( dl, 50 );

        roots = &cfg->env.global.schedule.roots;
        for( n = roots->length; --n >= 0; ){
                cur_i = roots->instrs[n];
                if( cur_i->uses == 0 && !(cur_i->flags & SCH_base_reg) ){
                        /* all root instructions @ delay = 0 */
                        if( !(cur_i->flags & SCH_last_branch ) ){
                                SCH_INSERT( dl->instrs[cur_i->delay.total], cur_i );
                        }
                        SCH_compute_delay( dl, cur_i );
                }
        }
        roots->length = 0;
        regs = cfg->env.global.schedule.regs;
        for( n = SCH_NUM_REGS; --n >= 0; ){
                cur_i = regs[n];
                if( cur_i->uses == 0 && !(cur_i->flags & SCH_base_reg) ){
                        /* all root instructions @ delay = 0 */
                        if( !(cur_i->flags & SCH_last_branch ) && !(cur_i->flags & SCH_2d_inserted ) ){
#ifdef SCH_2DST
                                cur_i->flags |= SCH_2d_inserted;
#endif
                                SCH_INSERT( dl->instrs[cur_i->delay.total], cur_i );
                        }
                        SCH_compute_delay( dl, cur_i );
                }
        }
}

void
SCH_end_block_( instr_block * blk )
{
        cfg_info * cfg;

        cfg = blk->cfg;
        cfg->env.global.schedule.code.end = blk->native.end;
        if( iid > 0 && enable_schedule ){
                SCH_order( cfg );
                SCH_print_by_delay( blk );
                SCH_schedule( blk );
                SCH_reset( cfg );
        }
        SCH_start_block( blk );
LOG(    fflush( cfg->fd ); );
}
