/*
 * jmeasure.h - runtime configurable sense support
 * 
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * (1) This header has to be included in all modules which want to use
 * the measure macros like:
 * 
 *   -------------- foo.c ----------------
 *   #include "jmeasure.h"
 *     ..
 *     // counter measurement
 *     MSR_INC( my_channel);
 *     MSR_ADD( my_channel, number);
 *     MSR_SET( my_channel, number);
 *
 *     // conditional counters
 *     MSR_COND_INC( my_channel, (<expr>));
 *     MSR_COND_ADD( my_channel, number, (<expr>));
 *
 *     // state access
 *     MSR_SET_STATE( my_state, number);
 *     MSR_GET_STATE( my_state);
 *
 *     // local time measurement (thread safe)
 *     MSR_TIME( my_channel)
 *     MSR_START( my_channel);
 *      ..
 *     MSR_STOP( my_channel);
 *     ..
 *     // global time measurement (synchronize!)
 *     MSR_START_GLOBAL( my_channel);
 *     ..
 *     MSR_STOP_GLOBAL( my_channel);
 *
 *     // store log record
 *     MSR_LOG( my_channel, delta_threshold);
 *     ..
 *   ------------------------------------
 *
 * 
 * (2) In addition, there has to be one main module (usually containing main)
 * that initializes the measure support:
 * 
 *   ------------- bar.c -----------------
 *   #define MEASURE_MAIN
 *   #include "jmeasure.h"
 *     ..
 *   main () {    // doesn't have to be main
 *     MSR_INIT();
 *     ..
 *   }
 *   -------------------------------------
 * 
 * (3) All suppored measure channels have to be defined in a file called
 * jmeasuredef.h (which is included from jmeasure.h). This file defines
 * the (application specific) "measure channels":
 * 
 *   ------------ jmeasuredef.h ------------
 *   MSR_DEFINE_LOG( 1000);   // buffer can hold 1000 log records
 *
 *   MSR_BEGIN_DEFINE_CHANNELS
 *   MSR_DEFINE_TIMER( my_timer_channel, "a timer channel");
 *   MSR_DEFINE_COUNTER( my_counter_channel, "bytes", "a counter channel");
 *   ..
 *   MSR_END_DEFINE_CHANNELS;
 *   ..
 *   #if defined(MEASURE_SETUP)
 *   msr_setup() {
 *     MSR_START_GLOBAL( my_timer_channel);
 *     ..
 *   }
 *   msr_finish() {
 *     MSR_STOP_GLOBAL( my_timer_channel);
 *     MSR_FINISH();
 *   }
 *   #endif
 *   -------------------------------------
 * 
 * (4) Build the system with -DMEASURE
 * 
 * (5) Before running the app, specify the measure channels you are interested in
 * with a environment variable MEASURE, containing a space separated list
 * of the topics:
 * 
 *   export MEASURE="my_channel my_second_channel .."
 * 
 * (6) run the app, the report on all active channels will be printed automatically
 * when the app terminates (atExit)
 *
 * (7) get enlightened..
 * 
 */

#ifndef __jmeasure_h
#define __jmeasure_h


#if defined(MEASURE)


/**********************************************************************
 * public macros - this is all the non MEASURE_MAIN modules have to know
 */

/*
 * explicitly switches on/off measure channels (on by default)
 */
#define MSR_ON(_chan) \
  _msr_##_chan##.active = 1

#define MSR_OFF(_chan) \
  _msr_##_chan##.active = -1

/*
 * the measure macros for counters
 */
#define MSR_INC(_chan) \
  if ( _msr_##_chan##.active > 0 ) _msr_##_chan##.counter++

#define MSR_DEC(_chan) \
  if ( _msr_##_chan##.active > 0 ) _msr_##_chan##.counter--

#define MSR_ADD(_chan,_n)   \
  if ( _msr_##_chan##.active > 0 ) _msr_##_chan##.counter += _n

#define MSR_SET(_chan,_n) \
  if ( _msr_##_chan##.active > 0 ) _msr_##_chan##.counter = _n

/*
 * define start timer memory (we have to be thread safe)
 */
#define MSR_TIME(_chan) \
  unsigned long _msr_##_chan##tstart = 0;

/*
 * start time measurement
 */
#define MSR_START(_chan) \
  if ( _msr_##_chan##.active > 0 ) \
    MSR_GETTIME( _msr_##_chan##tstart);

/*
 * stop time measurement and store elapsed time (since start)
 */
#define MSR_STOP(_chan) \
  if ( _msr_##_chan##.active > 0 ) { \
    unsigned long te; \
    MSR_GETTIME( te); \
    _msr_##_chan##.counter++; \
    _msr_##_chan##.ttotal += (te - _msr_##_chan##tstart); \
  }


/*
 * these are the global timer variants - make sure they are used thread-safe
 */

#define MSR_START_GLOBAL(_chan) \
  if ( _msr_##_chan##.active > 0 ) \
    MSR_GETTIME( _msr_##_chan##.tstart);

#define MSR_STOP_GLOBAL(_chan) \
  if ( _msr_##_chan##.active > 0 ) { \
    unsigned long te; \
    MSR_GETTIME( te); \
    _msr_##_chan##.counter++; \
    _msr_##_chan##.ttotal += (te - _msr_##_chan##.tstart); \
  }



/*
 * conditional actions
 */
#define MSR_INC_STATE(_stat) \
  _msr_##_stat++;

#define MSR_DEC_STATE(_stat) \
  _msr_##_stat--;

#define MSR_SET_STATE(_stat,_val) \
  _msr_##_stat = _val

#define MSR_GET_STATE(_stat) \
  _msr_##_stat

#define MSR_COND_INC(_chan,_expr) \
  if ( _expr ) { MSR_INC(_chan); }

#define MSR_COND_DEC(_chan,_expr) \
  if ( _expr ) { MSR_DEC(_chan); }

#define MSR_COND_ADD(_chan,_n,_expr) \
  if ( _expr ) { MSR_ADD(_chan,_n); }



/*
 * log channel counter if abs difference against last logged value exceeds _delta
 */
#define MSR_LOG(_chan,_delta) \
  if ( (_msr_##_chan##.active > 0) && (_msr_tstart) ){ \
    MsrChannel *p = &_msr_##_chan; \
    unsigned d = (p->counter > p->clast) ? p->counter - p->clast : p->clast - p->counter; \
    if ( (_delta == 0) || (d > _delta) )  \
      MSR_RECORD( _chan); \
  }

/*
 * this sets the start time (all log records are stored with delta against this
 */
#define MSR_LOG_START()     MSR_GETTIME(_msr_tstart)

/*
 * forced log record write (used rarely, e.g. to set initial/terminal values)
 */
#define MSR_RECORD(_chan)  \
  if ( _msr_nlog < _msr_maxlog ) { \
    unsigned long t; \
    MSR_GETTIME( t); \
    _msr_plog->id   = _msr_##_chan##.id; \
    _msr_plog->time = (unsigned) (t - _msr_tstart); \
    _msr_plog->val  = _msr_##_chan##.clast = _msr_##_chan##.counter; \
    _msr_nlog++; \
    _msr_plog++; \
  }
    

/**********************************************************************
 * internal types & macros
 */

#include <sys/time.h>

typedef struct _channel {
  char           *name;
  char           *comment;
  char           *units;
  char           id;       /* for log output */
  signed char    active;
  unsigned long  counter;
  unsigned long  ttotal;

  unsigned long  tstart;   /* used for global timers */

  unsigned       cdelta;   /* for logging */
  unsigned long  clast;

  struct _channel *next;
} MsrChannel;

typedef struct _log_entry {
  char      id    __attribute__((packed));
  unsigned  time  __attribute__((packed));
  unsigned  val   __attribute__((packed));
} MsrLogEntry; 

/***
#define MSR_GETTIME(_var)     \
  do {                     \
    struct timeval tv;     \
    gettimeofday( &tv, 0); \
    _var = (tv.tv_sec * 1000) + (tv.tv_usec / 1000); \
  } while(0)
***/

/**** politically correct, but too slow to please Heisenberg ****/
#include <sys/resource.h>
#include <unistd.h>
#define MSR_GETTIME(_var)     \
  do {                     \
    struct rusage ru;     \
    getrusage( RUSAGE_SELF, &ru); \
    _var = ((ru.ru_utime.tv_sec + ru.ru_stime.tv_sec) * 1000) + \
             ((ru.ru_utime.tv_usec + ru.ru_stime.tv_usec) / 1000); \
  } while(0)
/***/

/*
 * return current counter value
 */
#define MSR_COUNTER_VAL(_v,_chan) \
_v = (_msr_##_chan##.active != 0) ? _msr_##_chan##.counter : 0


/*
 * these are just used in the 2nd pass of MEASURE_MAIN
 */
#define MSR_BEGIN_DEFINE_CHANNELS
#define MSR_END_DEFINE_CHANNELS


#if defined(MEASURE_MAIN)

/*
 * These are the macros for the jmeasuredef.h config
 */
#define MSR_INIT()   msr_init(); msr_atExit()
#define MSR_FINISH() msr_report()

#define MSR_DEFINE_STATE(_stat,_val) \
  int _msr_##_stat = _val

#define MSR_DEFINE_COUNTER(_chan,_units,_cmt) \
  MsrChannel _msr_##_chan = { #_chan, _cmt, _units, 0, 0, 0, 0, 0 }

#define MSR_DEFINE_TIMER(_chan,_cmt) \
  MsrChannel _msr_##_chan = { #_chan, _cmt, "ms", 0, 0, 0, 0, 0 }

#define MSR_CHECK(_chan)                              \
  if ( (_msr_##_chan.active = msr_check( #_chan)) ){  \
    if ( !_msr_first ) {                              \
	  _msr_first = _msr_last = &(_msr_##_chan);       \
    }                                                 \
    else {                                            \
      _msr_last->next = &(_msr_##_chan);              \
      _msr_last = &(_msr_##_chan);                    \
    }                                                 \
    _msr_##_chan.id = _msr_nchannel++;               \
  }

#define MSR_DEFINE_LOG(_const_size)      \
  MsrLogEntry _msr_log[_const_size];     \
  MsrLogEntry *_msr_plog = &_msr_log[0]; \
  unsigned    _msr_maxlog = _const_size; \
  unsigned long _msr_tstart;             \
  unsigned    _msr_nlog


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

char* _msr;
FILE* _msr_stream;
MsrChannel *_msr_first, *_msr_last;
unsigned _msr_timecal;               /* ms timechecks */
unsigned _msr_ntimecal;              /* number of rounds for _msr_timecal */
unsigned _msr_nchannel;


char msr_check ( char*); /* prototype for jmeasuredef.h */
void msr_report();

/*
 * first round: bring in the channel declarations from channel defines
 */
#define MEASURE_SETUP
#include "jmeasuredef.h"

/*
 * redefine declaration macros to turn them into initialization macros
 */
#undef MSR_BEGIN_DEFINE_CHANNELS
#undef MSR_END_DEFINE_CHANNELS
#undef MSR_DEFINE_STATE(_stat,_val)
#undef MSR_DEFINE_COUNTER(_chan,_units,_cmt)
#undef MSR_DEFINE_TIMER(_chan,_cmt)
#undef MSR_DEFINE_LOG(_n)

#define MSR_BEGIN_DEFINE_CHANNELS \
void msr_init() {
#define MSR_END_DEFINE_CHANNELS   \
  if ( msr_check( "log") ) MSR_LOG_START(); \
  msr_setup(); \
}
#define MSR_DEFINE_STATE(_stat,_val)
#define MSR_DEFINE_LOG(_n)
#define MSR_DEFINE_COUNTER(_chan,_units,_cmt)    MSR_CHECK(_chan)
#define MSR_DEFINE_TIMER(_chan,_cmt)             MSR_CHECK(_chan)

/*
 * second round: generate the msr_check function out of the channel defines
 */
#undef MEASURE_SETUP
#include "jmeasuredef.h"

unsigned
msr_calibrate ()
{
  unsigned long t1, t2;

  fprintf( _msr_stream, "measure: calibrating..");

  MSR_GETTIME(t1);
  do {
    _msr_ntimecal++;
    MSR_GETTIME( t2);
	_msr_timecal = t2 - t1;
  } while ( _msr_timecal < 30 );

  fprintf( _msr_stream, "%d (%d rounds)\n", _msr_timecal, _msr_ntimecal);
}

/*
 * get the requested measure channels out of the environment var MEASURE
 */
char
msr_check ( char* topic )
{
  int len;
  char key[32];
  char *tok, *s;

  if ( !_msr )
	_msr = getenv( "MEASURE");

  if ( !_msr_stream )
	_msr_stream = stdout;

  if ( _msr ) {
    if ( !_msr_timecal )
	  msr_calibrate();

	len = strlen( topic);

	for ( s=_msr; *s && (tok = strstr( s, topic)); s+=len ){
	  if ( (tok[len] == ' ') || (tok[len] == 0) ){
		if ( (tok == s) || (tok[-1] == ' ') ) {
		  /* fprintf( _msr_stream, "measure: %s\n", topic); */
		  return 1;
		}
	  }
	}
  }

  return 0;
}

void
msr_writelogs ()
{
  FILE *fp = 0;
  int i, j, k;
  unsigned long t, v, llt=0, over=0, t_min=0xffffffff, t_max=0, v_min=0xffffffff, v_max=0;
  char fn[128];
  MsrChannel *chan[_msr_nchannel];
  MsrChannel *pc;

  /* enable map from id to channel */
  for ( pc=_msr_first; pc; pc = pc->next )
	chan[pc->id] = pc;

  fp = fopen( "jmeasure.log", "w");
  fprintf( fp, "# jmeasure generated logfile (gnuplot format)\n\n");
  fprintf( _msr_stream, "writing log data file jmeasure.log\n");

  fprintf( fp, "#set terminal tgif\n");
  fprintf( fp, "#set output \"jmeasure-log.obj\"\n\n");

  fprintf( fp, "#set terminal postscript\n");
  fprintf( fp, "#set output \"jmeasure-log.ps\"\n\n");

  fprintf( fp, "set rmargin 5\n");
  fprintf( fp, "set grid\n\n");

  /* store each channel into a log file <channel>.dat */
  for ( i=0, k=0; i<_msr_nchannel; i++ ) {
	for ( j=0; j<_msr_nlog; j++ ) {
	  if ( _msr_log[j].id == i ) {
		t = _msr_log[j].time;
		v = _msr_log[j].val;

		if ( ++k == 1 ) {
		  fprintf( fp, "set title \"log of channel: %s\\n\"\n", chan[i]->name);
		  fprintf( fp, "set ylabel \"%s [%s]\"\n", chan[i]->comment, chan[i]->units);
          fprintf( fp, "set xlabel \"time [ms]\"\n\n");

          fprintf( fp, "plot '-' with lines\n");
          fprintf( fp, " 0  0\n");         /* we start everything in the oigin */

		  t_min = t;
		}
		fprintf( fp, "%8d  %d\n", t, v);
		if ( v > v_max )
		  v_max = v;
		if ( v < v_min )
		  v_min = v;
	  }
	}
	
	if ( k > 0 ) {
      if ( t < _msr_log[_msr_nlog-1].time )   /* force same x scaling */
        fprintf( fp, " %d %d\n", _msr_log[_msr_nlog-1].time, v);

      fprintf( fp, "e\n\n");
      fprintf( fp, "pause -1\n\n");

      t_max = t;
	  fprintf( fp, "#set label \"records: %d\\n"
                              "t_min: %d\\n"
                              "t_max: %d\\n"
                              "v_min: %d\\n"
                              "v_max: %d\" at %d,%d\n",
                k, t_min, t_max, v_min, v_max, (t_max - t_min)/2, (v_max - v_min)/2);
      k = 0;
    }
  }

  fclose( fp);
}

void
msr_report ()
{
  MsrChannel *pc;
  
  if ( _msr_first )
	fprintf( _msr_stream, "\nmeasure report ----------------------------\n");

  for ( pc=_msr_first; pc; pc = pc->next ) {
	if ( pc->ttotal ) {   /* timer channel */
	  fprintf( _msr_stream, "%+30s (t): %10d ms (%d periods, +%d ms check)\n",
			   pc->name, pc->ttotal, pc->counter,
			   (unsigned) ((pc->counter * 2) * _msr_timecal / (double) _msr_ntimecal));
	}
	else {
	  fprintf( _msr_stream, "%+30s (c): %10d %s\n", pc->name, pc->counter, pc->units);
	}
  }

  if ( _msr_nlog )
	msr_writelogs();
}

void
msr_atExit()
{
  if ( atexit( msr_finish) )
	fprintf( _msr_stream, "can't register exit handler\n");
}

#else  /* !defined(MEASURE_MAIN) */

/*
 * non MEASURE_MAIN modules see just the simple external declarations
 */
#define MSR_INIT()
#define MSR_FINISH()

#define MSR_DEFINE_STATE(_stat,_val)          extern int _msr_##_stat
#define MSR_DEFINE_COUNTER(_chan,_units,_cmt) extern MsrChannel _msr_##_chan
#define MSR_DEFINE_TIMER(_chan,_cmt)          extern MsrChannel _msr_##_chan
#define MSR_CHECK(_chan)

#define MSR_DEFINE_LOG(_const_size) \
  extern MsrLogEntry* _msr_plog;    \
  extern unsigned     _msr_maxlog;  \
  extern unsigned long _msr_tstart; \
  extern unsigned     _msr_nlog


#undef MEASURE_SETUP
#include "jmeasuredef.h"

#endif /* MEASURE_MAIN */


#else /***** !defined(MEASURE) *********************************************/

/*
 * If the program was compiled without -DMEASURE, all debug macros
 * expand to nothing
 */

#define MSR_INIT()
#define MSR_FINISH()
#define MSR_DEFINE_STATE(_stat,_val)
#define MSR_DEFINE_COUNTER(_chan,_units,_cmt)
#define MSR_DEFINE_TIMER(_chan,_cmt)
#define MSR_CHECK(_chan)

#define MSR_ON(_chan)
#define MSR_OFF(_chan)
#define MSR_INC(_chan)
#define MSR_DEC(_chan)
#define MSR_ADD(_chan,_n)
#define MSR_SET(_chan,_n)
#define MSR_SET_STATE(_stat,_val)
#define MSR_GET_STATE(_stat)
#define MSR_COND_INC(_chan,_expr)
#define MSR_COND_DEC(_chan,_expr)
#define MSR_COND_ADD(_chan,_n,_expr)
#define MSR_LOG(_chan,_thresh)
#define MSR_TIME(_chan)
#define MSR_START(_chan)
#define MSR_STOP(_chan)
#define MSR_START_GLOBAL(_chan)
#define MSR_STOP_GLOBAL(_chan)

#define MSR_TIMER_VAL(_v,_chan)
#define MSR_COUNTER_VAL(_v,_chan)

#endif /* DEBUG */

#endif /* __jmeasure_h */
