/*
 * jstats.h - runtime configurable statistics support
 * 
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * (1) This header has to be included in all modules which want to use
 * the STATS macros:
 * 
 *   -------------- foo.c ----------------
 *   #include "jstats.h"
 *     ..
 *     STATS_ACTION( my_topic, do_something_useful);
 *     ..
 *   ------------------------------------
 *
 * 
 * (2) In addition, there has to be one main module (usually containing main)
 * that initializes the statics support:
 * 
 *   ------------- bar.c -----------------
 *   #define STATS_MAIN
 *   #include "jstats.h"
 *     ..
 *     STATS_INIT();
 *     ..
 *   -------------------------------------
 * 
 * (3) All suppored statistics topics have to be defined in a file called
 * jstatsdef.h (which is included from jstats.h). This file defines
 * the (application specific) "statistics topics":
 * 
 *   ------------ jstatsdef.h ------------
 *   STATS_DEFINE( my_topic);
 *   ..
 *   stats_init() {
 *     STATS_CHECK( my_topic);
 *     ..
 *   }
 *   -------------------------------------
 * 
 * (4) Build the system with -DSTATS
 * 
 * (5) Before running the app, specify the topics you're interested in
 * with a environment variable STATS, containing a space separated list
 * of the topics:
 * 
 *   export STATS="my_topic my_second_topic .."
 * 
 * (6) go and get the bugs...
 * 
 */

#ifndef __jstats_h
#define __jstats_h

/* we use this as a default for now, but otherwise it should be a compiler option */
#define STATS 1

#if defined(STATS)

/**********************************************************************
 * public macros - this is all the STATS_MAIN modules have to know
 */

#define STATS_ACTION(_topic,_op) if ( _stats_##_topic ) _op
#define STATS_ENABLED(_topic) ( _stats_##_topic )

/**********************************************************************
 * internal macros & funcs
 */

#if defined(STATS_MAIN)

/*
 * These are the macros for the jstatsdef.h config
 */
#define STATS_INIT() stats_init()
#define STATS_DEFINE(_topic)       char _stats_##_topic
#define STATS_CHECK(_topic)        _stats_##_topic = stats_check( #_topic)

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>

char* _stats;
FILE* _stats_stream;

void
stats_message( char* fmt, ...)
{
  va_list ap;
  va_start( ap, fmt);
  
  if ( !_stats_stream )
	_stats_stream = stderr;

  vfprintf( _stats_stream, fmt, ap);
  va_end(ap);
}

/*
 * get the requested statistics topics out of the environment var STATS
 */
char
stats_check ( char* topic )
{
  int len;
  char key[32];
  char *tok, *s;

  if ( !_stats )
	_stats = getenv( "STATS");
  if ( !_stats_stream )
	_stats_stream = stderr;

  if ( _stats ) {
	len = strlen( topic);

	for ( s=_stats; *s && (tok = strstr( s, topic)); s+=len ){
	  if ( (tok[len] == ' ') || (tok[len] == 0) ){
		if ( (tok == s) || (tok[-1] == ' ') ) {
		  fprintf( _stats_stream, "stats: %s\n", topic);
		  return 1;
		}
	  }
	}
  }
  
  return 0;
}

#else  /* !defined(STATS_MAIN) */

void stats_message( char* fmt, ...);

/*
 * This is what we want to see from "jstatsdef" in case we are
 * not in the STATS_MAIN module
 */
#define STATS_INIT()
#define STATS_CHECK(_topic)

#define STATS_DEFINE(_topic)       extern char _stats_##_topic

#endif /* STATS_MAIN */


/*
 * Now bring in the project specific statistics sitches (this is the
 * variable part)
 */
#include "jstatsdef.h"


#else /***** !defined(STATS) *********************************************/

/*
 * If the program was compiled without -DSTATS, all statistics macros
 * expand to nothing
 */

#define STATS_INIT()
#define STATS_DEFINE(_topic)
#define STATS_CHECK(_topic)

#define STATS_ACTION(_topic,_op)

/* watch out, this is called with an expected return value */
#define STATS_ENABLED(_topic)     0

#endif /* STATS */

#endif /* __jstats_h */
