/*
 * locks.c
 * Manage locking system
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "objects.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "threads.h"
#include "locks.h"
#include "exception.h"
#include "md.h"

#include "jdebug.h"

/*
 * If we don't have an atomic compare and exchange defined then make one
 * out of a simple atmoc exchange (using the LOCKINPROGRESS value the place
 * holder).  If we don't have that, we'll just fake it.
 */
#if defined(COMPARE_AND_EXCHANGE)

/*
 * Do nothing - we have what we need.
 */

#elif defined(ATOMIC_EXCHANGE)

inline
static
int
COMPARE_AND_EXCHANGE(iLock** a, iLock* o, iLock* n)
{
	iLock* val = LOCKINPROGRESS;

	ATOMIC_EXCHANGE(a, val);

	/* Success - store the real new value and return success */
	if (val == o) {
		*a = n;
		return (1);
	}

	/* Failure - if retrieved value is LOCKINPROGRESS then someone
	 * got there first.  We don't do anything under these circumstances
	 * except report the failure.
	 */
	else if (val == LOCKINPROGRESS) {
		return (0);
	}

	/* If we got another value instead, we need to put it back and report
	 * the failure.
	 */
	else {
		*a = val;
		return (0);
	}
}

#elif defined(TEST_AND_SET)

inline
static
int
COMPARE_AND_EXCHANGE(iLock** a, iLock* o, iLock* n)
{
	static int taslock;
	int val;

	/* Claim global lock */
	while (TEST_AND_SET(&taslock) == 0)
		;

	/* Manipulate iLock */
	if (*a == o) {
		*a = n;
		val = 1;
	}
	else {
		val = 0;
	}

	/* Release global lock */
	taslock = 0;

	return (val);
}

#else

inline
static
int
COMPARE_AND_EXCHANGE(iLock** a, iLock* o, iLock* n)
{
	int val;
	Kaffe_TI_enterCritSect();
	if (*a == o) {
		*a = n;
		val = 1;
	}
	else {
		val = 0;
	}
	Kaffe_TI_exitCritSect();
	return (val);
}

#endif

/* We need to treat the gc locks special since they guard memory allocation
 * and cannot be acquired by the gc thread otherwise.
 * I'm not proud of this - it's a hack waiting for a better idea - TIM.
 */
iLock* gc_lock;
iLock* gcman1;
iLock* gcman2;
iLock* finman1;
iLock* finman2;
#if defined(UNIX_PTHREADS)
extern iLock* tLock;
#endif
static struct {
	iLock** key;
	iLock   lock;
} specialLocks[] = {
	{ &gc_lock,	{ 0, 0, 0 } },
	{ &gcman1,	{ 0, 0, 0 } },
	{ &gcman2,	{ 0, 0, 0 } },
	{ &finman1,	{ 0, 0, 0 } },
	{ &finman2,	{ 0, 0, 0 } },
#if defined(UNIX_PTHREADS)
	{ &tLock,	{ 0, 0, 0 } },
#endif
};

#define	NR_SPECIAL_LOCKS	(sizeof(specialLocks)/sizeof(specialLocks[0]))

jboolean usingPosixLocks = false;

static jboolean _SemGet(void*, jlong);
static void _SemPut(void*);

/*
 * Initialise the locking system.
 */
void
initLocking(void)
{              

	if (Kaffe_LockInterface.semget == 0) {
		Kaffe_LockInterface.semget = _SemGet;
		Kaffe_LockInterface.semput = _SemPut;
		usingPosixLocks = true;
	}
}

#if defined(DEBUG)
void
dumpObjectLocks(void)
{
}
#endif

/*
 * Get a heavy lock for the object and hold it.  If the object doesn't
 * have a heavy lock then we create one.
 */
static
iLock*
getHeavyLock(iLock** lkp)
{
	iLock* old;
	iLock* lk;
	Hjava_lang_Thread* tid;
	jlong timeout;
	int i;

	lk = 0;
	timeout = 1;
	for (;;) {
		/* Get the current lock and replace it with -1 to indicate that
		 * changes are in progress.
		 */
		old = *lkp;
		if (old == LOCKINPROGRESS || !COMPARE_AND_EXCHANGE(lkp, old, LOCKINPROGRESS)) {
		        /*another thread is getting it */
			tid = Kaffe_TI_currentJava();
			Kaffe_LI_semget(unhand(tid)->sem, timeout);
			/* Back off */ 
			timeout = (timeout << 1)|timeout;
			continue; /* been awoken does not mean we got it, must try again */
		}
		
		if ((((uintp)old) & 1) == 1) { /* if was assigned to struct (Object), not to stack adress (object has a heavy lock)*/
			lk = (iLock*)(((uintp)old) & (uintp)-2); /* make it as if assigned to 
			                                            stack adress (clears the lsb) */
		}
		else {
			if (lk == 0) { /* acquiring an internal (VM) lock ? */
				for (i = 0; i < NR_SPECIAL_LOCKS; i++) {
					if (specialLocks[i].key == lkp) {
						lk = &specialLocks[i].lock;
						break;
					}
				}
				if (lk == 0) {
					/* Release the lock before we go into malloc.
					 * We have to reclaim the lock afterwards (at beginning
					   of loop)
					 */
					*lkp = old;
					lk = (iLock*)gc_malloc_fixed_typed(sizeof(iLock), GC_ALLOC_LOCK);
					/* If this fails we're in trouble!! */
					assert(lk != 0);
					continue;
				}
			}
			lk->holder = (void*)old;
			lk->mux = 0;
			lk->cv = 0;
		}
		
		return (lk);
		
	} /* for */
}

/*
 * Release the lock - only the one who has claimed it can do this
 * so there's no need for locked instructions.
 */
static
void
putHeavyLock(iLock** lkp, iLock* lk)
{
	assert(*lkp == LOCKINPROGRESS);
	if (lk == LOCKFREE) {
		*lkp = LOCKFREE;
	}
	else {
		*lkp = (iLock*)(1|(uintp)lk);
	}
}

/*
 * Slowly lock a mutex.  We get the heavy lock and lock that instead.
 * If we can't lock it we suspend until we can.
 */
static
void
slowLockMutex(iLock** lkp, void* where)
{
	iLock* lk;
	Hjava_lang_Thread* tid;

	DBG( vm_lock, ("Slow Lock @ %x\n", lkp));

	for (;;) {
		lk = getHeavyLock(lkp);

		/* If I hold the heavy lock then just keep on going */
		if (onCurrentThreadStack(lk->holder)) {
			putHeavyLock(lkp, lk);
			return;
		}

		/* If no one holds the heavy lock then claim it */
		if (lk->holder == 0) {
			lk->holder = where;
			putHeavyLock(lkp, lk);
			return;
		}

		/* Otherwise wait for holder to release it */
		tid = Kaffe_TI_currentJava();
		unhand(tid)->nextlk = lk->mux; /* put myself on wait queue */
                GC_WRITE( tid, lk->mux ); /* unhand(tid)->nextlk */
		lk->mux = tid;
		putHeavyLock(lkp, lk);
		Kaffe_LI_semget(unhand(tid)->sem, NOTIMEOUT);
	}
}

/*
 * Slowly unlock a mutex.  If there's someone waiting then we wake them up
 * so they can claim the lock.  If no one is waiting we revert the lock to
 * a fast thin lock.
 */
static
void
slowUnlockMutex(iLock** lkp, void* where)
{
	iLock* lk;
	Hjava_lang_Thread* tid;
	int i;

	DBG( vm_lock, ("Slow Unlock @ %x\n", lkp));

	lk = getHeavyLock(lkp);

	/* Only the lock holder can be doing an unlock */
	if (!onCurrentThreadStack(lk->holder)) {
		putHeavyLock(lkp, lk);
		postAndThrowException("java.lang.IllegalMonitorStateException", "");
	}

	/* If holder isn't where we are now then this isn't the final unlock */
	if (lk->holder > where) {
		putHeavyLock(lkp, lk);
		return;
	}

	/* Final unlock - if someone is waiting for it now would be a good
	 * time to tell them.
	 */
	if (lk->mux != 0) {
		tid = lk->mux;
		lk->mux = unhand(tid)->nextlk;
		unhand(tid)->nextlk = 0;
		lk->holder = 0;
		putHeavyLock(lkp, lk);
		Kaffe_LI_semput(unhand(tid)->sem);
	}
	/* If someone's waiting to be signaled keep the heavy in place */
	else if (lk->cv != 0) {
		lk->holder = 0;
		putHeavyLock(lkp, lk);
	}
	else {
                for (i = 0; i < NR_SPECIAL_LOCKS; i++) {
                        if (specialLocks[i].key == lkp) {
                                lk = 0;
                                break;
                        }
                }
                if (lk != 0) {
                        gc_free_fixed(lk);
                }
		putHeavyLock(lkp, LOCKFREE);
	}
}

void*
_releaseLock(iLock** lkp)
{
	iLock* lk;
	void* holder;

	lk = getHeavyLock(lkp);
	holder = lk->holder;

	/* I must be holding the damn thing */
	if (!onCurrentThreadStack(holder)) {
		putHeavyLock(lkp, holder);
		postAndThrowException("java.lang.IllegalMonitorStateException", "");
	}

	putHeavyLock(lkp, lk);
	slowUnlockMutex(lkp, holder);

	return (holder);
}

void
_acquireLock(iLock** lkp, void* holder)
{
	slowLockMutex(lkp, holder);
}

jboolean
_waitCond(iLock** lkp, jlong timeout)
{
	iLock* lk;
	void* holder;
	Hjava_lang_Thread* tid;
	Hjava_lang_Thread** ptr;
	jboolean r;
//	nativeThread* cur = GET_CURRENT_THREAD( &cur); 
	lk = getHeavyLock(lkp);
	holder = lk->holder;

	/* I must be holding the thing */
	if (!onCurrentThreadStack(holder)) {
		putHeavyLock(lkp, holder);
		postAndThrowException("java.lang.IllegalMonitorStateException", "");
	}

	tid = Kaffe_TI_currentJava();

	DBG( vm_lock, ("Wait @ %x by %x\n", lkp, unhand(tid)->PrivateInfo));

	unhand(tid)->nextlk = lk->cv;
        GC_WRITE( tid, lk->cv ); /* unhand(tid)->nextlk */
	lk->cv = tid;
	putHeavyLock(lkp, lk);
	slowUnlockMutex(lkp, holder);

//	printf("JUST BEFORE SEMGET with timeout = %d, thread is %p\n", (int) timeout,cur );
	r = Kaffe_LI_semget(unhand(tid)->sem, timeout);
//	printf("JUST AFTER SEMGET with timeout = %d, thread is %p\n",(int) timeout, cur );

	/* Timeout */
	if (r == false) {
		lk = getHeavyLock(lkp);
		/* Remove myself from CV or MUX queue - if I'm * not on either
		 * then I should wait on myself to remove any pending signal.
		 */
		for (ptr = &lk->cv; *ptr != 0; ptr = &unhand(*ptr)->nextlk) {
			if ((*ptr) == tid) {
				*ptr = unhand(tid)->nextlk;
				goto found;
			}
		}
		for (ptr = &lk->mux; *ptr != 0; ptr = &unhand(*ptr)->nextlk) {
			if ((*ptr) == tid) {
				*ptr = unhand(tid)->nextlk;
				goto found;
			}
		}
		/* Not on list - so must have been signalled after all -
		 * decrease the semaphore to avoid problems.
		 */
		 	
//		printf("JUST BEFORE SEMGET without timeout, thread is %p\n",cur );
                Kaffe_LI_semget(unhand(tid)->sem, NOTIMEOUT);
//	        printf("JUST AFTER SEMGET without timeout, thread is %p\n",cur );


		found:;
		putHeavyLock(lkp, lk);
	}

	slowLockMutex(lkp, holder);

	return (r);
}

void
_signalCond(iLock** lkp)
{
	iLock* lk;
	Hjava_lang_Thread* tid;

	lk = getHeavyLock(lkp);

	if (!onCurrentThreadStack(lk->holder)) {
		putHeavyLock(lkp, lk);
		postAndThrowException("java.lang.IllegalMonitorStateException", "");
	}

	/* Move one CV's onto the MUX */
	tid = lk->cv;
	DBG( vm_lock, ("Signal @ %x to %x\n", lkp, unhand(tid)->PrivateInfo));

	if (tid != 0) {
		lk->cv = unhand(tid)->nextlk;
		unhand(tid)->nextlk = lk->mux;
                GC_WRITE( tid, lk->mux ); /* unhand(tid)->nextlk */
		lk->mux = tid;
	}

	putHeavyLock(lkp, lk);
}

void
_broadcastCond(iLock** lkp)
{
	iLock* lk;
	Hjava_lang_Thread* tid;

	lk = getHeavyLock(lkp);

	if (!onCurrentThreadStack(lk->holder)) {
		putHeavyLock(lkp, lk);
		postAndThrowException("java.lang.IllegalMonitorStateException", "");
	}

	DBG( vm_lock, ("Broadcast @ %x to: \n", lkp));

	/* Move all the CV's onto the MUX */
	while (lk->cv != 0) {
		tid = lk->cv;
		DBG( vm_lock, ("  %x \n", unhand(tid)->PrivateInfo));
		lk->cv = unhand(tid)->nextlk;
		unhand(tid)->nextlk = lk->mux;
                GC_WRITE( tid, lk->mux ); /* unhand(tid)->nextlk */
		lk->mux = tid;
	}

	putHeavyLock(lkp, lk);
}


/*
 * Lock a mutex - try to do this quickly but if we failed because
 * we can't determine whether this is a multiple entry (recursive) lock or we've got
 * contention (mutex is already held) then fall back on a slow lock.
 */
 
/* 08/03/2000 Philippe changed criteria for recursive lock to STACKREDZONE / 2
   instead of hardcoded 1024. Seems reasonable + hopefully improve performance by reducing 
   number of slowLockMutex calls
   
   Also hardcoded 1024 would have likely failed on SMX which uses 1024 for its STACKREDZONE
*/


/* Explanation of what is happening:

As you see - lkp is not protected at all.  If val != 0, this doesn't mean that in the next 
statement it's != 0, i.e. in this time another thread may execute some code to change lkp. 

If *lkp becomes null, that means some other thread 
unlocked the mutex since - we can still safely call slowLockMutex. 
(Worst thing that happens will be that we'll put a heavy lock in 
place where we don't need one - a small optimization in slowLockMutex 
avoids that ...) 

The only situation in which we rely on the value read from *lkp is when 
it's != 0 && *lkp - where <= 1024. However, in this situation 
we're holding the lock already so no other thread can write *lkp. 
This is a recursive enter. 

There are two undocumented assumptions here: namely that 
LOCK_INPROGRESS - where > 1024, because *lkp can be LOCK_INPROGRESS, 
and that the stack grows down. This is all perfectly 64bit safe and 
architecture-independent. 

*/

void
_lockMutex(iLock** lkp, void* where)
{
	uintp val;

	val = (uintp)*lkp;
	
	 

	if (val == 0) { /* no none has it at this point */
		if (!COMPARE_AND_EXCHANGE(lkp, 0, (iLock*)where)) {
			slowLockMutex(lkp, where);  /* got contention */
		}
		else {
			DBG( vm_lock, ("Fast Lock @ %x\n", lkp));
		}
	}
	else if ((val - (uintp)where) > (STACKREDZONE / 2) ) { 
		/* can't determine if it is recursive and got contention */
		slowLockMutex(lkp, where);
	}
	
	
	/* else we had it already (recursive) */
	
}

/*
 * Unlock a mutex - try to do this quickly but if we failed then
 * we've got contention so fall back on a slow unlock.
 */
void
_unlockMutex(iLock** lkp, void* where)
{
	uintp val;

	val = (uintp)*lkp;

	if ((val & 1) != 0) { /* was not a stack lock */
		slowUnlockMutex(lkp, where);
	}
	/* notice critical use of lazy evaluation */
	else if (val == (uintp)where && !COMPARE_AND_EXCHANGE(lkp, (iLock*)where, LOCKFREE)) { 
	        slowUnlockMutex(lkp, where);
	}
	
}

void
lockObject(Hjava_lang_Object* obj)
{
	_lockMutex(&obj->lock, &obj);
}

void
unlockObject(Hjava_lang_Object* obj)
{
	_unlockMutex(&obj->lock, &obj);
}

void
slowLockObject(Hjava_lang_Object* obj, void* where)
{
	slowLockMutex(&obj->lock, where);
}

void
slowUnlockObject(Hjava_lang_Object* obj, void* where)
{
	slowUnlockMutex(&obj->lock, where);
}

/******************************************************************************
 *
 *  The following routines are used to convert POSIX style locks into the
 *  semaphores we actually need here.
 *
 ******************************************************************************/

/* Defining and using our own sems because Posix sems don't have timeouts */

/* The following are binary sems */


static
jboolean
_SemGet(void* sem, jlong timeout)
{
	jboolean r;
	sem2posixLock* lk;
//	nativeThread* cur = GET_CURRENT_THREAD( &cur);

	r = true;
	lk = (sem2posixLock*)sem;
       // printf("In semget before lock, thread = %p\n", cur);
	Kaffe_LI_lock(lk);
//	printf("In semget after lock, thread = %p\n", cur);
#if 0
	/* This was the old way to do this - we get true or false back
	 * from the wait - true means we got a signal, false means we
	 * timed out.  Trouble is that this doesn't seem to work, possibly
	 * due to bugs in the pthread implementation.
	 */
	if (lk->count == 0) {
		r = Kaffe_LI_wait(lk, timeout);
	}
	if (r == true) {
		assert(lk->count == 1);
		lk->count--;
	}
	else {
		assert(lk->count == 0);
	}
#endif
	/* This is the new scheme.  Since count is in fact a binary flag
	 * (it should never be any value other than 0 or 1) we can use this
	 * to determine if we got a signal or timed out.  Hence we can
	 * return true or false based on that.
	 */
//	printf("In semget before wait, thread = %p, lk->cv = %p, thread_priority = %d, lk->count = %d\n", cur, lk->cv,
//	                                             unhand(cur->thread)->priority            ,lk->count);
	if (lk->count == 0) {
		Kaffe_LI_wait(lk, timeout);//if we don't timeout, lk->count was incremented in _SemPut 
	}
//	printf("In semget after wait, thread = %p, lk->cv = %p\n", cur, lk->cv);
	if (lk->count == 1) {
		lk->count = 0;
		r = true;
	}
	else {
		assert(lk->count == 0);  
		r = false;
	}
	
	Kaffe_LI_unlock(lk);
//	printf("In semget after unlock, thread = %p\n", cur);

	
	return (r);
}

static
void
_SemPut(void* sem)
{
	sem2posixLock* lk;
//	nativeThread* cur = GET_CURRENT_THREAD( &cur); 


	lk = (sem2posixLock*)sem;

      //  printf("lk->count = %d\n", lk->count);
        //while(lk->count != 0)
	/*	if(sched_yield() < 0)
			perror("sched_yield");*/
			//sleep(1);
       
      //  printf("In semput before lock, thread = %p\n", cur);
	Kaffe_LI_lock(lk);
//	printf("In semput after lock, thread = %p, lk->count =%d, lk->cv = %p, thread_priority = %d\n", cur, lk->count, lk->cv, unhand(cur->thread)->priority );

	assert(lk->count == 0);
	lk->count++;
	Kaffe_LI_signal(lk);
	Kaffe_LI_unlock(lk);
//	printf("In semput after unlocked, thread = %p, lk->cv = %p\n", cur, lk->cv);

}
