/*
 * lookup.c
 * Various lookup calls for resolving objects, methods, exceptions, etc.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "jtypes.h"
#include "classMethod.h"
#include "slots.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "lookup.h"
#include "exception.h"
#include "threads.h"
#include "baseClasses.h"
#include "machine.h"
#include "locks.h"
#include "soft.h"
#include "strings.h"
#include "jdebug.h"

/*
 * Lookup a method reference and get the various interesting bits.
 */
int
getMethodSignatureClass(constIndex idx, Hjava_lang_Class* this, int isSpecial, callInfo* call)
{
	constants* pool;
	constIndex ci;
	constIndex ni;
	Hjava_lang_Class* class;
	char* name;
	char* signature;
	Method* mptr;
	int i;
	int ret;

	pool = CLASS_CONSTANTS(this);
	if (pool->tags[idx] != CONSTANT_Methodref &&
	    pool->tags[idx] != CONSTANT_InterfaceMethodref) {
		DBG( vm_lookup,("No Methodref found\n"));
		postException("java.lang.NoSuchMethodError", "");
		return (0);
	}

	ni = METHODREF_NAMEANDTYPE(idx, pool);
	name = WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]);
	signature = WORD2UTF(pool->data[NAMEANDTYPE_SIGNATURE(ni, pool)]);

	call->name = name;
	call->signature = signature;
	call->class = 0;
	call->method = 0;
	call->cname = 0;

	ci = METHODREF_CLASS(idx, pool);
	class = getClass(ci, this, &call->cname);
	if (class != 0) {
		call->cname = CLASS_CNAME(class);
		ret = processClass(class, CSTATE_LINKED);
		if (ret != 0) {
			if (isSpecial == 1) {
				if (strcmp(name, constructor_name) != 0 && class != this && instanceof(class, this)) {
					class = this->superclass;
				}
			}

			DBG( vm_lookup,("getMethodSignatureClass(%s,%s,%s)\n",
							CLASS_CNAME(class), name, signature));

			call->class = class;
			mptr = findMethod(class, name, signature);
			if (mptr != 0) {
				call->method = mptr;
			}
			/* If we've not found anything and we're searching
			 * interfaces, search them too.
			 */
			if (mptr == 0 && isSpecial == 2) {
				class = call->class;
				for (i = class->total_interface_len - 1; i >= 0; i--) {
					mptr = findMethodLocal(class->interfaces[i], name, signature);
					if (mptr != 0) {
						call->method = mptr;
						break;
					}
				}
			}

			/* If we get to the end of the classes then we've not found
			 * the method.
			 */
			if (class == 0) {
				call->method = 0;
			}
		}
	}

	/* Calculate in's and out's */
	countInsAndOuts(signature, &call->in, &call->out, &call->rettype);

	if (call->class == 0) {
		postException("java.lang.NoClassDefFoundError", (call->cname == 0 ? "" : call->cname));
		return (0);
	}
	else if (call->method == 0) {
		postException("java.lang.NoSuchMethodError", name);
		return (0);
	}
	else {
		return (1);
	}
}

Hjava_lang_Class*
getClass(constIndex idx, Hjava_lang_Class* this, char** cname)
{
	constants* pool;
	char* name;
	char* tmp;
	Hjava_lang_Class* class;
	int tag;
	int iLockRoot;

	name = 0;
	pool = CLASS_CONSTANTS(this);

	tag = pool->tags[idx];

	switch (tag) {
	case CONSTANT_ResolvedClass:
		class = CLASS_CLASS(idx, pool);
		if (cname != 0) {
			*cname = CLASS_CNAME(class);
		}
		return (class);

	case CONSTANT_Class:
		/* The class may be resolved by another thread so we better
		 * lock and get the tag & name again just in case.  If we
		 * have been resolved then we just return the answer.
		 */
		lockMutex(&this->head);
		tag = pool->tags[idx];
		name = WORD2UTF(pool->data[idx]);
		unlockMutex(&this->head);

		if (tag == CONSTANT_ResolvedClass) {
			class = CLASS_CLASS(idx, pool);
			if (cname != 0) {
				*cname = CLASS_CNAME(class);
			}
			return (class);
		}
		break;

	default:
		postException("java.lang.ClassFormatError", CLASS_CNAME(this));
		return (0);
	}

	UTF_REPLACE(tmp, name, '/', '.');
	class = loadClass(tmp, this->loader, CSTATE_UNLOADED);

	/* Lock the class while we update the constant pool.  Someone
	 * may have done this already but we don't care.
	 */
	if (class != 0) {
		lockMutex(&this->head);
		pool->data[idx] = (ConstSlot)class;
		pool->tags[idx] = CONSTANT_ResolvedClass;
		GC_WRITE(this, (void*)pool->data[idx]); /* class->constants.data[idx] */
		unlockMutex(&this->head);
		if (cname != 0) {
			*cname = CLASS_CNAME(class);
		}
		return (class);
	}

	if (cname != 0) {
		*cname = addUtfCopy(tmp, -1);
	}

	return (class);
}

int
getField(constIndex idx, Hjava_lang_Class* this, fieldInfo* ret)
{
	constants* pool;
	constIndex ci;
	constIndex ni;
	Field* fld;

	ret->cname = 0;
	ret->name = 0;
	ret->signature = 0;
	ret->field = 0;
	ret->class = 0;

	pool = CLASS_CONSTANTS(this);
	if (pool->tags[idx] == CONSTANT_ResolvedFieldref) {
		fld = (Field*)pool->data[idx];
		ret->name = fld->name;
		ret->signature = fld->ftype;
		ret->field = fld;
		ret->class = this;
		return (1);
	}

	if (pool->tags[idx] != CONSTANT_Fieldref) {
		DBG( vm_lookup, ("No Fieldref found\n"));
		postException("java.lang.NoSuchFieldError", "");
		return (0);
	}

	ci = FIELDREF_CLASS(idx, pool);
	ni = FIELDREF_NAMEANDTYPE(idx, pool);
	ret->name = WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]);
	ret->signature = WORD2UTF(pool->data[NAMEANDTYPE_SIGNATURE(ni, pool)]);
	ret->class = getClass(ci, this, 0);
	if (ret->class == 0 || processClass(ret->class, CSTATE_LINKED) == 0) {
		ret->cname = WORD2UTF(pool->data[ci]);
		return (0);
	}

DBG(vm_lookup, ("getField(%s,%s,%s)\n", ret->class->name, ret->name, ret->signature));

	fld = lookupClassField(ret->class, ret->name);

DBG_ACTION(vm_lookup, { if (fld == 0) DBG( vm_lookup,("field not found\n")) });

	if (fld == 0) {
		postException("java.lang.NoSuchFieldError", ret->name);
		return (0);
	}
	else {
		if (ret->class == this) {
			pool->data[idx] = (ConstSlot)fld;
			pool->tags[idx] = CONSTANT_ResolvedFieldref;
		}
		ret->field = fld;
		return (1);
	}
}

/*
 * Lookup a method (and translate) in the specified class.
 */
Method*
findMethodLocal(Hjava_lang_Class* class, char* name, char* sig)
{
	Method* mptr;
	int n;

	/*
	 * Lookup method - this could be alot more efficient but never mind.
	 * Also there is no attempt to honour PUBLIC, PRIVATE, etc.
	 */
	n = CLASS_NMETHODS(class);
	for (mptr = CLASS_METHODS(class); --n >= 0; ++mptr) {
		if (strcmp(name, mptr->name) == 0 && strcmp(sig, mptr->signature) == 0) {
			return (mptr);
		}
	}
	return (NULL);
}

/*
 * Lookup a method (and translate) in the specified class.
 */
Method*
findMethod(Hjava_lang_Class* class, char* name, char* sig)
{
	/*
	 * Lookup method - this could be alot more efficient but never mind.
	 * Also there is no attempt to honour PUBLIC, PRIVATE, etc.
	 */
	for (; class != 0; class = class->superclass) {
		Method* mptr = findMethodLocal(class, name, sig);
		if (mptr != NULL) {
			return mptr;
		}
	}
	return (0);
}

/*
 * Find the matching exception handler in a method (if there is one)
 */
uintp
findExceptionHandlerInMethod(Hjava_lang_Class* class, Method* meth, uintp pc, Hjava_lang_Class* eclass)
{
	char* catch;
	Hjava_lang_Class* cptr;
	jexceptionEntry* eptr;

	/* No exceptions - can be me */
	if (meth->exception_table == 0) {
		return (0);
	}

	if (!CLASS_IS_PRECOMPILED(class)) {
		for (eptr = &meth->exception_table[0]; eptr->start_pc != (uintp)-1; eptr++) {
			uintp start_pc = eptr->start_pc;
			uintp end_pc = eptr->end_pc;
			uintp handler_pc = eptr->handler_pc;

			DBG( vm_lookup,("Exceptions %x (%x-%x)\n", pc, start_pc, end_pc));
			if (pc < start_pc || pc > end_pc) {
				continue;
			}
			DBG( vm_lookup, ("Found exception 0x%x\n", handler_pc));

			/* Found exception - is it right type */
			if (eptr->catch == 0) {
				return (handler_pc);
			}
			catch = eptr->catch;
			for (cptr = eclass; cptr != 0; cptr = cptr->superclass) {
				if (strcmp(catch, cptr->name) == 0) {
					return (handler_pc);
				}
			}
		}
		return (0);
	}
	else {
		int dist;
		jexceptionEntry* selected;
		uintp* aptr;

		dist = 0x7FFFFFFF;
		selected = 0;

		/* Look though the exception table and find the exception where
		 * the distance between the start_pc and the current pc
		 * is the smallest and the catch matches the exception
		 * object.
		 */
		for (eptr = &meth->exception_table[0]; eptr->start_pc != (uintp)-1; eptr++) {
			int diff = (int)(pc - eptr->start_pc);
			if (diff >= 0 && diff < dist) {
				catch = eptr->catch;
				for (cptr = eclass; cptr != 0; cptr = cptr->superclass) {
					if (catch == 0 || strcmp(catch, cptr->name) == 0) {
						dist = diff;
						selected = eptr;
						break;
					}
				}
			}
		}
		/* If we didn't find anything then no exceptions here */
		if (selected == 0) {
			return (0);
		}
		/* Now we look for a simple basic-block match which is better
		 * than the exception match.  If we find one then we're
		 * really in a basic-block which is not in an appropriate
		 * try section.  Note that if we find the same block we
		 * don't consider it a match because the distance won't
		 * be better.
		 */
		for (aptr = (uintp*)(eptr+1); *aptr != (uintp)-1; aptr++) {
			int diff = (int)(pc - *aptr);
			if (diff >= 0 && diff < dist) {
				return (0);
			}
		}
		return (selected->handler_pc);
	}
}
