/* gc-incremental.c
 * The incremental garbage collector.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "gc.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "locks.h"
#include "threads.h"
#include "support.h"
#include "exception.h"
#include "external.h"
#include "lookup.h"
#include "md.h"
#include "access.h"
#include "java_lang_ref_Reference.h"
#include "gc-mem.h"
#include "strings.h"

#include "jni.h"
#include "jdebug.h"
#include "jmeasure.h"

/* JNIEXPORT void Kaffe_ExceptionClear(JNIEnv*); */
extern void Kaffe_CallVoidMethodA(JNIEnv*, jobject, jmethodID, jvalue*);

static gcList gclists[6];
const static int finalise_class = 5;
const static int mustfree = 4;
const static int white = 3;
const static int grey = 2;
const static int black = 1;
const static int finalise = 0;
static Hjava_lang_ref_Reference* refQueue;
static Hjava_lang_ref_Reference* refDispatchQueue;


static bool gcRunning = false;
static bool finalRunning = false;
static int urgentState = 0;

#if defined(GC_STATS)

/* Class and object space statistics */
static struct {
	int	classNr;
	int	objectNr;
	int	arrayNr;
	int	methodNr;
	int	fieldNr;
	int	staticNr;
	int	dispatchNr;
	int	bytecodeNr;
	int	exceptionNr;
	int	constantNr;
	int	utf8constNr;
	int	interfaceNr;
	int	jitNr;
	int	lockNr;
	int	refNr;
	int	threadCtxNr;
	int	threadNr;
	int	classloaderNr;
	int	jreferenceNr;
	int	analyseNr;
	int	inflateNr;
	int	stringsNr;
	int	engineNr;
	int	estackNr;
	int	jarNr;
	int	normalNr;
	int	stringNr;
	int	awtNr;
	
	int	fixedNr;
	int	otherNr;
	int	totalFixedNr;
	
	int	classMem;
	int	objectMem;
	int	arrayMem;
	int	methodMem;
	int	fieldMem;
	int	staticMem;
	int	dispatchMem;
	int	bytecodeMem;
	int	exceptionMem;
	int	constantMem;
	int	utf8constMem;
	int	interfaceMem;
	int	jitMem;
	int	lockMem;
	int	refMem;
	int	threadCtxMem;
	int	threadMem;
	int	classloaderMem;
	int	jreferenceMem;
	int	analyseMem;
	int	inflateMem;
	int	stringsMem;
	int	engineMem;
	int	estackMem;
	int	jarMem;
	int	normalMem;
	int	stringMem;
	int	awtMem;
	
	int	fixedMem;
	int	otherMem;
	int	totalFixedMem;
} objectStats;

static struct {
	int size;
	int count;
} objectSizes[] = {

#define	OBJECTSIZE(SZ)	{ SZ, 0 },

	OBJECTSIZE(32)
	OBJECTSIZE(40)
	OBJECTSIZE(48)
	OBJECTSIZE(56)
	OBJECTSIZE(64)
	OBJECTSIZE(80)
	OBJECTSIZE(96)
	OBJECTSIZE(112)
	OBJECTSIZE(128)
	OBJECTSIZE(160)
	OBJECTSIZE(192)
	OBJECTSIZE(224)
	OBJECTSIZE(240)
	OBJECTSIZE(496)
	OBJECTSIZE(1000)
	OBJECTSIZE(2016)
	OBJECTSIZE(4040)
	OBJECTSIZE(8192)
	OBJECTSIZE(12288)
	OBJECTSIZE(16483)
	OBJECTSIZE(32768)
	OBJECTSIZE(65536)
	OBJECTSIZE(131072)
	OBJECTSIZE(262144)
	OBJECTSIZE(524288)
	OBJECTSIZE(1048576)
	OBJECTSIZE(0x7FFFFFFF)

#undef	OBJECTSIZE

	{ -1, -1 }

};

static void objectSizesPrint(void);
static void objectSizesAdd(size_t);

static void objectStatsChange(gc_unit*, int);

#define	OBJECTSTATSADD(M)	objectStatsChange(M, 1)
#define	OBJECTSTATSREMOVE(M)	objectStatsChange(M, -1)
#define	OBJECTSIZESADD(S)	objectSizesAdd(S)

#else

#define	OBJECTSTATSADD(M)
#define	OBJECTSTATSREMOVE(M)
#define	OBJECTSIZESADD(S)

#endif

/* We run the GC after allocating 1Mbyte of memory.  If we are
* doing this incrementally, then we will have GC our entire heap
* by the time we've allocated this much new space.
*/
#define	ALLOCCOUNTGC	(1024*1024)

static GcAdvice* advice;

extern iLock* gcman1;
extern iLock* gcman2;
extern iLock* finman1;
extern iLock* finman2;
extern iLock* gc_lock;

int gc_mode = GC_DISABLED;	/* GC will be enabled after the first
				 * thread is setup.
				 */

extern Hjava_lang_Class* ArrayInterfaces[];

static void gcFree(void*);

static void finalizeObject(void*);
static void finalizeClass(void*);

void walkConservative(void*, uint32);
void walkMemory(void*);

static void walkNull(void*, uint32);
static void walkClass(void*, uint32);
static void walkObject(void*, uint32);
static void walkRefArray(void*, uint32);
static void walkJReference(void*, uint32);
static void walkThread(void*, uint32);

/* Standard GC function sets */
static gcFuncs gcFunctions[] = {
	{ walkConservative, GC_OBJECT_NORMAL },	/* GC_ALLOC_NORMAL */
	{ walkNull,	    GC_OBJECT_NORMAL },	/* GC_ALLOC_ESTACK */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_FIXED */
	{ walkObject,	    finalizeObject   },	/* GC_ALLOC_NORMALOBJECT */
	{ walkNull,	    GC_OBJECT_NORMAL },	/* GC_ALLOC_PRIMARRAY */
	{ walkRefArray,     GC_OBJECT_NORMAL },	/* GC_ALLOC_REFARRAY */
	{ walkClass,	    finalizeClass    },	/* GC_ALLOC_CLASSOBJECT */
	{ walkThread,	    finalizeObject   },	/* GC_ALLOC_THREAD */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_METHOD */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_FIELD */
	{ walkConservative, GC_OBJECT_NORMAL },	/* GC_ALLOC_STATICDATA */
	{ walkNull,         GC_OBJECT_FIXED  },	/* GC_ALLOC_DISPATCHTABLE */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_BYTECODE */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_EXCEPTIONTABLE */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_CONSTANT */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_UTF8CONST */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_INTERFACE */
	{ walkNull,         GC_OBJECT_FIXED  },	/* GC_ALLOC_JITCODE */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_LOCK */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_THREADCTX */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_REF */
	{ walkConservative, GC_OBJECT_NORMAL }, /* GC_ALLOC_CLASSPOOL */
	{ walkJReference,   finalizeObject   },	/* GC_ALLOC_JREFERENCE */
	{ walkConservative, GC_OBJECT_NORMAL }, /* GC_ALLOC_CLASSLOADER */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_ANALYSE */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_INFLATE */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_STRINGS */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_ENGINE */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_JAR */
	{ walkNull,	    GC_OBJECT_FIXED  },	/* GC_ALLOC_AWT */
	{ walkObject,	    GC_OBJECT_NORMAL },	/* GC_ALLOC_STRING */
	{ walkObject,	    GC_OBJECT_NORMAL },	/* GC_ALLOC_NORMALOBJECT_NOFINALIZE */
};

#define	REFOBJALLOCSZ	 8
#define	REFOBJHASHSZ	 8
typedef struct _refObject {
	void*			start;
	void*			end;
	uintp			ref;
	struct _refObject*	next;
} refObject;
typedef struct _refTable {
	refObject*		hash[REFOBJHASHSZ];
} refTable;
static refTable			refObjects;
#define	REFOBJHASH(V)		((((uintp)(V))/(2*sizeof(uintp)))%REFOBJHASHSZ)

struct _gcStats gcStats;

extern gc_block* gc_objecthash[];
extern size_t gc_heap_total;
extern size_t gc_heap_limit;

static void startGC(void);
static void finishGC(void);
static void mustfreeGC(void);
static void startFinalizer(void);
static void calculateAdviceNow(void);

inline
static
void
calculateAdvice(void)
{
	if (advice != 0) {
		calculateAdviceNow();
	}
}

/*
* Initalise the Garbage Collection system.
*/
static
void
initGc(void)
{
	URESETLIST(gclists[white]);
	URESETLIST(gclists[grey]);
	URESETLIST(gclists[black]);
	URESETLIST(gclists[finalise]);
	URESETLIST(gclists[mustfree]);
	URESETLIST(gclists[finalise_class]);
}

/* inline */
int
isObject(gc_unit* unit, gc_block* info)
{
	uintp hidx;
	gc_block* hptr;
	uint32 idx;

	/* Get hash index for this block */
	hidx = GC_OBJECT_HASHIDX(info);
	for (hptr = gc_objecthash[hidx]; hptr != 0; hptr = hptr->next) {
		if (hptr == info) {
		/* Make sure 'unit' refers to the beginning of an
		* object.  We do this by making sure it is correctly
		* aligned within the block.
			*/
			idx = GCMEM2IDX(info, unit);
			if (idx < info->nr && GCBLOCK2MEM(info, idx) == unit && (GC_GET_COLOUR(info, idx) & GC_COLOUR_INUSE) == GC_COLOUR_INUSE) {
				/* It is a real object ... */
				return (1);
			}
			break;
		}
	}
	/* Not found */
	return (0);
}

/*
* Mark the memory as reached if it really is an object.
*/
void
markObject(void* mem)
{
	gc_block* info;
	uint32 idx;
	gc_unit* unit;
	
	if (mem == 0) {
		return;
	}
	
	/*
	* First we check to see if the memory 'mem' is in fact the
	* beginning of an object.  If not we just return.
	*/
	
	/* Get block info for this memory - if it exists */
	unit = UTOUNIT(mem);
	info = GCMEM2BLOCK(unit);

	/* If it's not a real object, then return */
	if (isObject(unit, info) == 0) {
		return;
	}
	
	/* It is a real object ... */
	idx = GCMEM2IDX(info, unit);

	/* This object better not be free alread */
	assert(GC_GET_COLOUR(info, idx) != GC_COLOUR_FREE);

	/* If object's been traced before, don't do it again */
	if (GC_GET_COLOUR(info, idx) != GC_COLOUR_WHITE) {
		return;
	}

	/* If we found a new white object, mark it as grey and
	 * move it into the grey list.
	 */
	GC_SET_COLOUR(info, idx, GC_COLOUR_GREY);
	UREMOVELIST(unit);
	UAPPENDLIST(gclists[grey], unit);
}

/*
* Mark the memory as reached and walked (black) if it really is an object.
*/
void
markAsBlackObject(void* mem)
{
	gc_block* info;
	uint32 idx;
	gc_unit* unit;
	
	if (mem == 0) {
		return;
	}
	
	/*
	* First we check to see if the memory 'mem' is in fact the
	* beginning of an object.  If not we just return.
	*/
	
	/* Get block info for this memory - if it exists */
	unit = UTOUNIT(mem);
	info = GCMEM2BLOCK(unit);

	/* If it's not a real object, then return */
	if (isObject(unit, info) == 0) {
		return;
	}
	
	/* It is a real object ... */
	idx = GCMEM2IDX(info, unit);

	/* This object better not be free alread */
	assert(GC_GET_COLOUR(info, idx) != GC_COLOUR_FREE);

	/* If object's been traced before, don't do it again */
	if (GC_GET_COLOUR(info, idx) == GC_COLOUR_BLACK) {
		return;
	}

	/* If we found a new white object, mark it as grey and
	 * move it into the grey list.
	 */
	GC_SET_COLOUR(info, idx, GC_COLOUR_BLACK);
	UREMOVELIST(unit);
	UAPPENDLIST(gclists[black], unit);
}

/*
* Walk a bit of memory.
*/
void
walkMemory(void* mem)
{
	gc_block* info;
	int idx;
	gc_unit* unit;
	
	unit = UTOUNIT(mem);
	info = GCMEM2BLOCK(unit);
	idx = GCMEM2IDX(info, unit);
	
	UREMOVELIST(unit);
	UAPPENDLIST(gclists[black], unit);
	GC_SET_COLOUR(info, idx, GC_COLOUR_BLACK);
	
	(*gcFunctions[GC_GET_FUNCS(info, idx)].walk)(mem, GCBLOCKSIZE(info));
}

void
walkConservative(void* base, uint32 size)
{
	int8* mem;
	
	DBG( vm_mem, ("walkConservative: %x-%x\n", base, ((uintp)base)+size));
	
	gcStats.markedobj += 1;
	
	if (size > 0) {
		gcStats.markedmem += size;

		for (mem = ((int8*)base) + (size & -ALIGNMENTOF_VOIDP) - sizeof(void*); (void*)mem >= base; mem -= ALIGNMENTOF_VOIDP) {
			if (*(void**)mem != 0) {
				markObject(*(void**)mem);
			}
		}
	}
}

/*
* Walk an object.
*/
static
void
walkObject(void* base, uint32 size)
{
#if 0
	int n;
	unsigned int* refs;
	Hjava_lang_Class* cls;
	Hjava_lang_Object* obj;
	void* mem;

	obj = (Hjava_lang_Object*)base;

	/* If this object has no dtable, then its just being created and
	 * isn't yet in use - don't need to walk it.
	 */
	if (obj->dtable == 0) {
		return;
	}
	markObject(obj->dtable);

	cls = OBJECT_CLASS(obj);
	n = cls->nrefs;
	if (n > 0) {
		assert(size > 0);
		refs = cls->refs;
		while (--n >= 0) {
			mem = *(void**)((uintp)base+refs[n]);
			if (mem != 0) {
				markObject(mem);
			}
		}
	}
#else
	walkConservative(base, size);
#endif
}

/*
* Walk a thread object.
*/
static
void
walkThread(void* base, uint32 size)
{
	walkConservative(base, size);
	Kaffe_TI_GcWalkThread((Hjava_lang_Thread*)base);
}

/*
* Walk an object which cannot reference anything.
*/
static
void
walkNull(void* base, uint32 size)
{
	gcStats.markedobj += 1;
	gcStats.markedmem += size;
	
	markObject(((Hjava_lang_Object*)base)->dtable);
}

/* 
 * Define GC_SAFE to walk some extra nodes of Class, if
 * you're having problems.
 */
#define GC_SAFE 1

 /*
 * Walk a class object.
 * The class object is currently a hierarchy of fixed objects,
 * so we must be careful and mark "exit" points of the
 * class meta-data.
 */
static
void
walkClass(void* base, uint32 size)
{
	Hjava_lang_Class* class;
	struct _constants* pool;
	int ix;
	Method* meth;
	Field* fld;
#if defined(GC_SAFE)
	jexceptionEntry* eentry;
#endif
	
	class = (Hjava_lang_Class*)base;

	/* If this class has no dtable, its in the proces of being created
	 * and doesn't yet contains anything interesting.
	 */
	if (class->head.dtable == 0) {
		return;
	}

	markObject(CLASS_CNAME(class));
	/* Mark strings referenced in the constant pool */
	pool = CLASS_CONSTANTS (class);
        if (pool->tags) {
                for (ix = pool->size; --ix >= 0 ;) {
#if !defined(GC_SAFE)
			char type = pool->tags[ix];
			if (type == CONSTANT_Utf8 ||
				type == CONSTANT_String ||
				type == CONSTANT_ResolvedString ||
				type == CONSTANT_Class ||
				type == CONSTANT_ResolvedClass)
#endif
                        markObject((void*)pool->data[ix]);
                }
	}

#if defined(GC_SAFE)
        /* We might want to mark the superclass... */
	if (class->sdata && class->superclass) {
		markObject(class->superclass);
	}
#endif
		
        /* Handle methods */
	meth = CLASS_METHODS(class);
	for (ix = CLASS_NMETHODS(class); --ix >= 0; meth++) {
		markObject(meth->name);
		markObject(meth->signature);
#if defined(GC_SAFE)
                /* We might want to mark exception handler classes... */
                if (meth->exception_table) {
			for (eentry = meth->exception_table; eentry->start_pc != (uintp)-1; eentry++) {
                                markObject((void*)eentry->catch);
			}
		}
#endif
	}

        /* Handle fields */
	fld = CLASS_FIELDS(class);
	for (ix = CLASS_NFIELDS(class); --ix >= 0; fld++) {
		markObject(fld->name);
			
#if defined(GC_SAFE)
		/* We might want to mark field type name ... */
		markObject(fld->ftype);
#endif
	}
		
#if defined(GC_SAFE)
        /* We might want to mark the class loader... */
        markObject(class->loader);
		
        /* We might want to mark the class interfaces... */
	for (ix = class->total_interface_len; --ix >= 0;) {
		markObject(class->interfaces[ix]);
	}
#endif
	if (!CLASS_IS_PRIMITIVE(class)) {
		markObject(class->dtable);
	}
		
	/* Mark the static data */
	markObject(class->sdata);
}

/*
* Walk an array object objects.
*/
static
void
walkRefArray(void* base, uint32 size)
{
	Hjava_lang_Object* arr;
	int i;
	Hjava_lang_Object** ptr;
	
	gcStats.markedobj += 1;
	gcStats.markedmem += size;
	
	arr = (Hjava_lang_Object*)base;
	
	markObject(arr->dtable);
	
	ptr = ARRAY_DATA(HArrayOfObject, arr);
	for (i = ARRAY_SIZE(arr); --i>= 0; ) {
		Hjava_lang_Object* el = *ptr++;
		if (el != NULL) {
			markObject(el);
		}
	}
}

static
void
walkJReference(void* base, uint32 size)
{
	Hjava_lang_ref_Reference* obj;
	
	obj = (Hjava_lang_ref_Reference*)base;
	
	markObject(obj->base.dtable);
	
	/* Queue reference for later examination by GC */
	if (unhand(obj)->obj != 0) {
		unhand(obj)->next = refQueue;
		refQueue = obj;
	}
	
	/* Only mark the queue, not the object referenced */
	markObject(unhand(obj)->que);
	
	/* Walk the rest of the object */
	walkConservative(obj+1, size-sizeof(Hjava_lang_ref_Reference));
}

/*
* Now handle any java.lang.ref.Reference objects
*/
static
void
markReferences(void)
{
	Hjava_lang_ref_Reference* ref;
	Hjava_lang_ref_Reference* nref;
	gc_block* info;
	gc_unit* unit;
	int idx;
	
	for (ref = refQueue; ref != 0; ref = nref) {
		nref = unhand(ref)->next;
		
		/* No object?  No operation */
		if (unhand(ref)->obj == 0) {
			continue;
		}
		
		unit = UTOUNIT(unhand(ref)->obj);
		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);
		
		/* If object isn't white then it's not interesting */
		if (GC_GET_COLOUR(info, idx) != GC_COLOUR_WHITE) {
			continue;
		}
		
		/* Soft references are dispatched before finalization but only
		* when memory is low.
		*/
		if (unhand(ref)->type == java_lang_ref_Reference_SOFT) {
			if (urgentState != 0) {
				unhand(ref)->obj = 0;
			}
			else {
				markObject(unhand(ref)->obj);
				continue;
			}
		}
		/* Weak references are dispatched before finalization */
		else if (unhand(ref)->type == java_lang_ref_Reference_WEAK) {
			unhand(ref)->obj = 0;
		}
		/* Recycle references are dispatch and make the object
		 * reachable again.
		 */
		else if (unhand(ref)->type == java_lang_ref_Reference_RECYCLE) {
			markObject(unhand(ref)->obj);
		}
		/* Phantom references are dispatched after finalization */
		else if (unhand(ref)->type == java_lang_ref_Reference_PHANTOM) {
			if (GC_GET_STATE(info, idx) == GC_STATE_FINALIZED ||
				GC_GET_STATE(info, idx) == GC_STATE_NORMAL) {
				unhand(ref)->obj = 0;
			}
			else {
				continue;
			}
		}
		/* Unknown reference type */
		else {
			continue;
		}
		
		/* Queue reference for dispatch */
		unhand(ref)->next = refDispatchQueue;
		refDispatchQueue = ref;
	}
}

/*
* The Garbage Collector sits in a loop starting a collection, waiting
* until it's finished incrementally, then tidying up before starting
* another one.
*/
void
gcMan(void* arg)
{
	gc_unit* unit;
	gc_unit* nunit;
	gc_block* info;
	int idx;
	int iLockRoot;
	
	lockStaticMutex(&gcman1);
	
	for(;;) {
                MSR_START(gc_sweep_time);

		lockStaticMutex(&gc_lock);

		Kaffe_TI_enterCritSect();
		startGC();
		
		for (unit = gclists[grey].cnext; unit != &gclists[grey]; unit = gclists[grey].cnext) {
			walkMemory(UTOMEM(unit));
		}
		
		/* Handle any java.lang.ref.References */
		markReferences();
		
		/* Now walk any white objects which will be finalied.  They
		* may get reattached, so anything they reference must also
		* be live just in case.
		*/
		for (unit = gclists[white].cnext; unit != &gclists[white]; unit = nunit) {
			nunit = unit->cnext;
			info = GCMEM2BLOCK(unit);
			idx = GCMEM2IDX(info, unit);
			
			if (GC_GET_STATE(info, idx) == GC_STATE_NEEDFINALIZE) {
				GC_SET_STATE(info, idx, GC_STATE_INFINALIZE);
				markObject(UTOMEM(unit));
			}
		}
		
		/* We may now have more grey objects, so walk them */
		for (unit = gclists[grey].cnext; unit != &gclists[grey]; unit = gclists[grey].cnext) {
			walkMemory(UTOMEM(unit));
		}
		
		finishGC();
		
		/* Calculate memory advice */
		calculateAdvice();
		
		Kaffe_TI_exitCritSect();

		unlockStaticMutex(&gc_lock);

		/* Free any memory we can - this can (and must) be done
		 * outside the critical section.
		 */
		mustfreeGC();
		
		/* Kick of the finalizer if necessary */
		startFinalizer();
		
#if defined(GC_STATS)
		OBJECTSTATSPRINT(Kaffe_JavaVMArgs[0].enableVerboseGC);
#endif
		
		gcStats.allocobj = 0;
		gcStats.allocmem = 0;
		
		/* Wake up anyone waiting for the GC to finish */
		lockStaticMutex(&gcman2);
		broadcastStaticCond(&gcman2);
		unlockStaticMutex(&gcman2);
		MSR_STOP(gc_sweep_time);
		
		urgentState = 0;
		gcRunning = false;
		while (gcRunning == false) {
			waitStaticCond(&gcman1, NOTIMEOUT);
		}
		assert(gcRunning == true);
	}
}

/*
* Start the GC process by scanning the root and thread stack objects.
*/
static
void
startGC(void)
{
	int i;
	refObject* robj;
	gc_unit* unit;
	gc_unit* nunit;
	
	gcStats.freedmem = 0;
	gcStats.freedobj = 0;
	gcStats.markedobj = 0;
	gcStats.markedmem = 0;
	gcStats.finalmem = 0;
	gcStats.finalobj = 0;
	
	refQueue = 0;
	refDispatchQueue = 0;
	
	/* Walk referenced memory */
	for (i = 0; i < REFOBJHASHSZ; i++) {
		for (robj = refObjects.hash[i]; robj != 0; robj = robj->next) {
			walkConservative(robj->start, robj->end - robj->start);
		}
	}
	
	/* Walk all objects on the finalizer list */
	for (unit = gclists[finalise].cnext; unit != &gclists[finalise]; unit = nunit) {
		nunit = unit->cnext;
		markObject(UTOMEM(unit));
	}
	
	/* Walk the thread objects */
	Kaffe_TI_GcWalkThreads();
}

/*
* Finish off the GC process.  Any unreached (white) objects are moved
* for finalising and the finaliser woken.
* The reached (black) objects are moved onto the now empty white list
* and recoloured white.
*/
static
void
finishGC(void)
{
	gc_unit* unit;
	gc_block* info;
	int idx;
	
	/* There shouldn't be any grey objects at this point */
	assert(gclists[grey].cnext == &gclists[grey]);
	
	/*
	* Any white objects should now be freed, but we cannot call
	* gc_heap_free here because we might block in gc_heap_free,
	* which would leave the white list unprotected.
	* So we move them to a 'mustfree' list from where we'll pull them
	* off later.
	*/
	while (gclists[white].cnext != &gclists[white]) {
		unit = gclists[white].cnext;
		UREMOVELIST(unit);
		
		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);
		
		assert(GC_GET_COLOUR(info, idx) == GC_COLOUR_WHITE);
		assert(GC_GET_STATE(info, idx) == GC_STATE_NORMAL);
		UAPPENDLIST(gclists[mustfree], unit);
	}
	
	/* Now move the black objects back to the white queue for next time.
	*/
	while (gclists[black].cnext != &gclists[black]) {
		unit = gclists[black].cnext;
		UREMOVELIST(unit);
		
		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);
		
		assert(GC_GET_COLOUR(info, idx) == GC_COLOUR_BLACK);
		
		if (GC_GET_STATE(info, idx) == GC_STATE_INFINALIZE) {
			UAPPENDLIST(gclists[finalise], unit);
			gcStats.finalmem += GCBLOCKSIZE(GCMEM2BLOCK(unit));
			gcStats.finalobj += 1;
		}
		else {
			UAPPENDLIST(gclists[white], unit);
		}
		GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
	}
}

static
void
mustfreeGC(void)
{
	gc_unit* unit;
	gc_block* info;
	int idx;
	int iLockRoot;
	
	/*
	* Now free the objects.  We can block here since we're the only
	* thread manipulating the "mustfree" list.
	*/
	while (gclists[mustfree].cnext != &gclists[mustfree]) {
		unit = gclists[mustfree].cnext;
		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);
		UREMOVELIST(unit);
		gcStats.freedmem += GCBLOCKSIZE(info);
		gcStats.freedobj += 1;
		OBJECTSTATSREMOVE(unit);

		/* Handle strings */
		if (GC_GET_FUNCS(info, idx) == GC_ALLOC_STRING) {
			uninternJavaString((Hjava_lang_String*)UTOMEM(unit));
		}

		lockStaticMutex(&gc_lock);
		/* Keep the stats correct */
		gcStats.totalmem -= GCBLOCKSIZE(info);
		gcStats.totalobj -= 1;
		gc_heap_free(unit);
		unlockStaticMutex(&gc_lock);
	}
	DBG( vm_mem, ("Freed %d objects of %dK\n", gcStats.freedobj,
		gcStats.freedmem/1024));
}

static
void
startFinalizer(void)
{
	int iLockRoot;
	int start;
	
	/* If there's stuff to be finalised or dispatched then we'd
	 * better do it.  Note that we mustn't hold the gc_lock and finman1
	 * lock at the same time!!.
	 */
	lockStaticMutex(&gc_lock);
	start = 0;
	if (gclists[finalise].cnext != &gclists[finalise] || refDispatchQueue != 0) {
		start = 1;
	}
	unlockStaticMutex(&gc_lock);

	if (start != 0) {
		lockStaticMutex(&finman1);
		if (finalRunning == false) {
			finalRunning = true;
			signalStaticCond(&finman1);
		}
		unlockStaticMutex(&finman1);
	}
}

/*
* The finaliser sits in a loop waiting to finalise objects.  When a
* new finalised list is available, it is woken by the GC and finalises
* the objects in turn.  An object is only finalised once after which
* it is deleted.
*/
void
finaliserMan(void* arg)
{
	gc_block* info;
	gc_unit* unit;
	int idx;
	Hjava_lang_ref_Reference* ref;
	int iLockRoot;
	jmethodID dispatch;
	
	lockStaticMutex(&finman1);
	
	dispatch = 0;
	for (;;) {
		
		finalRunning = false;
		while (!finalRunning)
			waitStaticCond(&finman1, NOTIMEOUT);
/*		assert(finalRunning == true); */
		
		/* Dispatch any java.lang.ref.References */
		for (ref = refDispatchQueue; ref != 0; ref = unhand(ref)->next) {
			/* Use JNI to make dispatch call to protect finalizer
			 * against exceptions.
			 */
			if (dispatch == 0) {
				dispatch = findMethod(ReferenceClass, dispatch_name, void_sig);
				assert(dispatch != 0);
			}
			Kaffe_CallVoidMethodA(0, (jobject)ref, dispatch, 0);
			Kaffe_ExceptionClear(0);
		}
		
		/* Finalize any pending objects, defer ClassObjects for a bit */
		while (gclists[finalise].cnext != &gclists[finalise]) {
			lockStaticMutex(&gc_lock);
			unit = gclists[finalise].cnext;
			info = GCMEM2BLOCK(unit);
			idx = GCMEM2IDX(info, unit);
                        if (GC_GET_FUNCS(info,idx) == GC_ALLOC_CLASSOBJECT) {
                                UREMOVELIST(unit);
                                UAPPENDLIST(gclists[finalise_class], unit);
                                unlockStaticMutex(&gc_lock);

                        } else {
                                UREMOVELIST(unit);
                                GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
                                UAPPENDLIST(gclists[white], unit);
                                assert(GC_GET_STATE(info,idx) == GC_STATE_INFINALIZE);
                                /* Objects are only finalised once */
                                GC_SET_STATE(info, idx, GC_STATE_FINALIZED);
                                unlockStaticMutex(&gc_lock);
                                /* Call finaliser */
                                unlockStaticMutex(&finman1);
                                (*gcFunctions[GC_GET_FUNCS(info,idx)].final)(UTOMEM(unit));
                                lockStaticMutex(&finman1);
                        }
		}
		
		/* Finalize ClassObjects now */
		while (gclists[finalise_class].cnext != &gclists[finalise_class]) {
			lockStaticMutex(&gc_lock);
			unit = gclists[finalise_class].cnext;
			info = GCMEM2BLOCK(unit);
			idx = GCMEM2IDX(info, unit);

			UREMOVELIST(unit);
			GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
			UAPPENDLIST(gclists[white], unit);
			assert(GC_GET_STATE(info,idx) == GC_STATE_INFINALIZE);
			GC_SET_STATE(info, idx, GC_STATE_FINALIZED);
			unlockStaticMutex(&gc_lock);
			unlockStaticMutex(&finman1);
			(*gcFunctions[GC_GET_FUNCS(info,idx)].final)(UTOMEM(unit));
			lockStaticMutex(&finman1);
		}
		
		/* Wake up anyone waiting for the finalizer to finish */
		lockStaticMutex(&finman2);
		broadcastStaticCond(&finman2);
		unlockStaticMutex(&finman2);
	}
}

/*
* Explicity invoke the garbage collecto and wait for it to complete.
*/
static
void
gcInvokeGC(int urgent)
{
#if defined(GC_ENABLE)
	int iLockRoot;

	lockStaticMutex(&gcman1);
	lockStaticMutex(&gcman2);
	
	if (gcRunning == false) {
		gcRunning = true;
		urgentState = urgent;
		signalStaticCond(&gcman1);
	}
	unlockStaticMutex(&gcman1);
	
	waitStaticCond(&gcman2, NOTIMEOUT);
	unlockStaticMutex(&gcman2);
#endif
}

/*
* Explicity invoke the finalizer.
*/
static
void
gcInvokeFinalizer(void)
{
#if defined(GC_ENABLE)
	int iLockRoot;
	
	/* First invoke the GC */
	invokeGC(0);
	
	/* Run the finalizer (if might already be running as a result of
	 * the GC)
	 */
	lockStaticMutex(&finman1);
	lockStaticMutex(&finman2);

	if (finalRunning == false) {
		finalRunning = true;
		signalStaticCond(&finman1); 
	}
	unlockStaticMutex(&finman1);

	waitStaticCond(&finman2, NOTIMEOUT);
	unlockStaticMutex(&finman2);
#endif
}       

/*
* Allocate a new object.  The object is attached to the white queue.
* After allocation, if incremental collection is active we peform
* a little garbage collection.  If we finish it, we wakeup the garbage
* collector.
*/
static
void*
gcMalloc(size_t size, int fidx)
{
	static int gc_init = 0;
	gc_block* info;
	gc_unit* unit;
	int i;
	int iLockRoot;
	void* mem;

	MSR_INC(gc_alloc);
	MSR_START(gc_alloc_time);
	
	/* Initialise GC */
	if (gc_init == 0) {
		gc_init = 1;
		initGc();
	}

	lockStaticMutex(&gc_lock);

	unit = gc_heap_malloc(size + sizeof(gc_unit));
	
	if (unit == 0) {
		postOutOfMemory();
		return (0);
	}
	DBG( vm_mem, ("gcMalloc: 0x%x (%d)\n", unit, size));

	info = GCMEM2BLOCK(unit);
	i = GCMEM2IDX(info, unit);

#define DBGTYPE(xx) \
	   case GC_ALLOC_ ## xx: \
DBG( vm_trace_alloc, ("0x%x (%d -> %d) %d " #xx "\n", UTOMEM(unit), size, memsize, fidx)); \
	   break; 

	DBG_ACTION( vm_trace_alloc, {
	  int memsize = GCBLOCKSIZE(info);

	  switch (fidx) {
		DBGTYPE(NORMAL)
		DBGTYPE(ESTACK)
		DBGTYPE(FIXED)
		DBGTYPE(NORMALOBJECT)
		DBGTYPE(PRIMARRAY)
		DBGTYPE(REFARRAY)
		DBGTYPE(CLASSOBJECT)
		DBGTYPE(THREAD)
		DBGTYPE(METHOD)
		DBGTYPE(FIELD)
		DBGTYPE(STATICDATA)
		DBGTYPE(DISPATCHTABLE)
		DBGTYPE(BYTECODE)
		DBGTYPE(EXCEPTIONTABLE)
		DBGTYPE(CONSTANT)
		DBGTYPE(UTF8CONST)
		DBGTYPE(INTERFACE)
		DBGTYPE(JITCODE)
		DBGTYPE(LOCK)
		DBGTYPE(THREADCTX)
		DBGTYPE(REF)
		DBGTYPE(CLASSPOOL)
		DBGTYPE(JREFERENCE)
		DBGTYPE(CLASSLOADER)
		DBGTYPE(ANALYSE)
		DBGTYPE(INFLATE)
		DBGTYPE(STRINGS)
		DBGTYPE(ENGINE)
		DBGTYPE(JAR)
		DBGTYPE(AWT)
		DBGTYPE(STRING)
		DBGTYPE(NORMALOBJECT_NOFINALIZE)
	  default:
		DBG( vm_trace_alloc, ("0x%x (%d -> %d) %d ??\n", UTOMEM(unit), size, memsize, fidx));
	  }

	});
	
	
	MSR_ADD(gc_alloc_mem, GCBLOCKSIZE(info));
	
	gcStats.totalmem += GCBLOCKSIZE(info);
	gcStats.totalobj += 1;
	gcStats.allocmem += GCBLOCKSIZE(info);
	gcStats.allocobj += 1;
	
	GC_SET_FUNCS(info, i, fidx);
	
	OBJECTSTATSADD(unit);
	OBJECTSIZESADD(size);
	
	/* Determine whether we need to finalise or not */
	if ( gcFunctions[fidx].final == GC_OBJECT_NORMAL || gcFunctions[fidx].final == GC_OBJECT_FIXED) {
		GC_SET_STATE(info, i, GC_STATE_NORMAL);
	}
	else {
		GC_SET_STATE(info, i, GC_STATE_NEEDFINALIZE);
	}
	
	/* If object is fixed, we give it the fixed colour and do not
	* attach it to any lists.  This object is not part of the GC
	* regieme and must be freed explicitly.
	*/
	if (gcFunctions[fidx].final == GC_OBJECT_FIXED) {
		GC_SET_COLOUR(info, i, GC_COLOUR_FIXED);
	}
	else {
		GC_SET_COLOUR(info, i, GC_COLOUR_WHITE);
		UAPPENDLIST(gclists[white], unit);
	}

	/* Make sure we have a reference to the base of the object
	 * before we release the gc_lock, otherwise it might be
	 * garbage collected.
	 */
	mem = UTOMEM(unit);
	unlockStaticMutex(&gc_lock);
	
	/* Calculate memory advice */
	calculateAdvice();
	
	MSR_STOP(gc_alloc_time);
	return (mem);
}

/*
* Reallocate an object.
*/
static
void*
gcRealloc(void* mem, size_t size, int fidx)
{
	gc_block* info;
	int idx;
	void* newmem;
	gc_unit* unit;
	int iLockRoot;
	
	assert(gcFunctions[fidx].final == GC_OBJECT_FIXED);
	
	/* If nothing to realloc from, just allocate */
	if (mem == NULL) {
		return (gcMalloc(size, fidx));
	}
	
	lockStaticMutex(&gc_lock);
	unit = UTOUNIT(mem);
	info = GCMEM2BLOCK(unit);
	idx = GCMEM2IDX(info, unit);
	
	/* Can only handled fixed objects at the moment */
	assert(GC_GET_COLOUR(info, idx) == GC_COLOUR_FIXED);
	
	unlockStaticMutex(&gc_lock);

	/* If we'll fit into the current space, just send it back */
	if (info->size >= size + sizeof(gc_unit)) {
		return (mem);
	}
	
	/* Allocate new memory, copy data, and free the old */
	newmem = gcMalloc(size, fidx);
	if (newmem != 0) {
		memcpy(newmem, mem, info->size);
	}
	gcFree(mem);
	
	return (newmem);
}

/*
* Explicitly free an object.
*/
static
void
gcFree(void* mem)
{
	gc_block* info;
	int idx;
	gc_unit* unit;
	int iLockRoot;
	
	if (mem != 0) {
		lockStaticMutex(&gc_lock);
		unit = UTOUNIT(mem);
		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);
		
		if (GC_GET_COLOUR(info, idx) == GC_COLOUR_FIXED) {
			
			OBJECTSTATSREMOVE(unit);
			
			/* Keep the stats correct */
			gcStats.totalmem -= GCBLOCKSIZE(info);
			gcStats.totalobj -= 1;
			gcStats.allocmem -= GCBLOCKSIZE(info);
			gcStats.allocobj -= 1;
			
			gc_heap_free(unit);
			
			/* Calculate memory advice */
			calculateAdvice();
		}
		else {
			/* We just ignore this - it'll get GCed */
		}
		unlockStaticMutex(&gc_lock);
	}
}

/*
* Add a presistent reference to an object.
*/
static
void*
gcAddRef(void* start, void* end)
{
	uint32 idx;
	refObject* obj;
	
	idx = REFOBJHASH(start);
	for (obj = refObjects.hash[idx]; obj != 0; obj = obj->next) {
		/* Found it - just increase reference */
		if (obj->start == start) {
			obj->ref++;
			if (end > obj->end) {
				obj->end = end;
			}
			return (obj);
		}
	}
	
	/* Not found - create a new one */
	obj = (refObject*)gcMalloc(sizeof(refObject), GC_ALLOC_REF);
	if (obj != 0) {
		obj->start = start;
		obj->end = end;
		obj->ref = 1;
		obj->next = refObjects.hash[idx];
		refObjects.hash[idx] = obj;
		return (obj);
	}
	else {
		return (0);
	}
}

/*
* Remove a persistent reference to an object.  If the count becomes
* zero then the reference is removed.
*/
static
void
gcRmRef(void* start)
{
	uint32 idx;
	refObject** objp;
	refObject* obj;
	
	idx = REFOBJHASH(start);
	for (objp = &refObjects.hash[idx]; *objp != 0; objp = &obj->next) {
		obj = *objp;
		/* Found it - just increase reference */
		if (obj->start == start) {
			obj->ref--;
			if (obj->ref == 0) {
				*objp = obj->next;
				gcFree(obj);
			}
			break;
		}
	}
}

static
void
finalizeObject(void* ob)
{
	Hjava_lang_Object* obj;
	Method* final;

	obj = (Hjava_lang_Object*)ob;
	/* If this object has no dtable then it's been completely created yet.
	 * which is worrying since we're finalizing it!!!
	 */
	assert(obj->dtable != 0);
	final = OBJECT_CLASS(obj)->finalizer;
	if (final != 0) {
		/* Use JNI to make finalizer call to protect GC against exceptions */
		Kaffe_CallVoidMethodA(0, (jobject)ob, (jmethodID)final, 0);
		Kaffe_ExceptionClear(0);
	}
}

static
void
finalizeClass(void* obj)
{
	Hjava_lang_Class* class;

	class = (Hjava_lang_Class*)obj;

	if (Kaffe_JavaVMArgs[0].enableVerboseGC > 0) {
		kprintf("<GC: Unloading %s>\n", CLASS_CNAME(class));
	}
	
	/* Unload class */
	unloadClass(class);
}

/*
* Register a GC advice object - we only support one at the moment.
*/
static
void
gcRegister(GcAdvice* adv)
{
	advice = adv;
}

/*
* Calculate memory advice and signal anyone interested.
*/
static
void
calculateAdviceNow(void)
{
	static int lastused = 0;
	int used;
	int col;
	
	col = advice->color;
	used = (gcStats.totalmem * 100) / gc_heap_limit;
	
	if (used > lastused) {
		if (col < ADVICE_RED && used > advice->hi_red) {
			col = ADVICE_RED;
		}
		else if (col < ADVICE_ORANGE && used > advice->hi_orange) {
			col = ADVICE_ORANGE;
		}
		else if (col < ADVICE_YELLOW && used > advice->hi_yellow) {
			col = ADVICE_YELLOW;
		}
	} else if (used < lastused) {
		if (col > ADVICE_GREEN && used < advice->lo_yellow) {
			col = ADVICE_GREEN;
		}
		else if (col > ADVICE_YELLOW && used < advice->lo_orange) {
			col = ADVICE_YELLOW;
		}
		else if (col > ADVICE_ORANGE && used < advice->lo_red) {
			col = ADVICE_ORANGE;
		}
	}
	lastused = used;
	
	/* If we get a color change, wake anyone interested in it */
	if (col != advice->color) {
		lockObject(&advice->head);
		advice->color = col;
		broadcastCond(&advice->head);
		unlockObject(&advice->head);
	}
}

#if defined(GC_STATS)

/* --------------------------------------------------------------------- */
/* The following functions are strictly for statistics gathering	 */

static
void
objectStatsChange(gc_unit* unit, int diff)
{
	gc_block* info;
	int idx;
	int objdiff;
	int memdiff;
	int func;
	
	info = GCMEM2BLOCK(unit);
	idx = GCMEM2IDX(info, unit);
	objdiff = diff * 1;
	memdiff = diff * GCBLOCKSIZE(info);
	func = GC_GET_FUNCS(info, idx);

      	DBG_ACTION( vm_trace_gcfree_object, {
       	  Hjava_lang_Object* obj;

       	  obj = (Hjava_lang_Object*) UTOMEM(unit);
	  if ((func == GC_ALLOC_NORMALOBJECT_NOFINALIZE) ||
	    (func == GC_ALLOC_NORMALOBJECT)) {
		if (diff < 0) {
			Hjava_lang_Class* class = obj->dtable->class;

			DBG( vm_trace_gcfree_object, ("0x%x: %s %d\n", obj, (class ? class->name: "<none>"), memdiff));
		}
          }

	});

	
#define	CHECK_GC_TYPE(GCFUNCS,STAT) \
	if (func == (GCFUNCS)) { \
		objectStats.STAT##Nr += objdiff; \
		objectStats.STAT##Mem += memdiff; \
	} else
	
	CHECK_GC_TYPE(GC_ALLOC_NORMALOBJECT, object)
	CHECK_GC_TYPE(GC_ALLOC_NORMALOBJECT_NOFINALIZE, object)
	CHECK_GC_TYPE(GC_ALLOC_REFARRAY, array)
	CHECK_GC_TYPE(GC_ALLOC_PRIMARRAY, array)
	CHECK_GC_TYPE(GC_ALLOC_CLASSOBJECT, class)
	CHECK_GC_TYPE(GC_ALLOC_METHOD, method)
	CHECK_GC_TYPE(GC_ALLOC_FIELD, field)
	CHECK_GC_TYPE(GC_ALLOC_STATICDATA, static)
	CHECK_GC_TYPE(GC_ALLOC_DISPATCHTABLE, dispatch)
	CHECK_GC_TYPE(GC_ALLOC_BYTECODE, bytecode)
	CHECK_GC_TYPE(GC_ALLOC_EXCEPTIONTABLE, exception)
	CHECK_GC_TYPE(GC_ALLOC_CONSTANT, constant)
	CHECK_GC_TYPE(GC_ALLOC_UTF8CONST, utf8const)
	CHECK_GC_TYPE(GC_ALLOC_INTERFACE, interface)
	CHECK_GC_TYPE(GC_ALLOC_JITCODE, jit)
	CHECK_GC_TYPE(GC_ALLOC_LOCK, lock)
	CHECK_GC_TYPE(GC_ALLOC_FIXED, fixed)
	CHECK_GC_TYPE(GC_ALLOC_REF, ref)
	CHECK_GC_TYPE(GC_ALLOC_THREADCTX, threadCtx)
	CHECK_GC_TYPE(GC_ALLOC_THREAD, thread)
	CHECK_GC_TYPE(GC_ALLOC_CLASSLOADER, classloader)
	CHECK_GC_TYPE(GC_ALLOC_JREFERENCE, jreference)
	CHECK_GC_TYPE(GC_ALLOC_ANALYSE, analyse)
	CHECK_GC_TYPE(GC_ALLOC_INFLATE, inflate)
	CHECK_GC_TYPE(GC_ALLOC_STRINGS, strings)
	CHECK_GC_TYPE(GC_ALLOC_ENGINE, engine)
	CHECK_GC_TYPE(GC_ALLOC_ESTACK, estack)
	CHECK_GC_TYPE(GC_ALLOC_JAR, jar)
	CHECK_GC_TYPE(GC_ALLOC_NORMAL, normal)
	CHECK_GC_TYPE(GC_ALLOC_STRING, string)
	CHECK_GC_TYPE(GC_ALLOC_AWT, awt)
	{
		/* Else .... */
		objectStats.otherNr += objdiff;
		objectStats.otherMem += memdiff;
	}
	
#undef	CHECK_GC_TYPE

	/* We keep a seperate count of all fixed memory */
	if (gcFunctions[func].final == GC_OBJECT_FIXED) {
		objectStats.totalFixedNr += objdiff;
		objectStats.totalFixedMem += memdiff;
	}
	
}

void
objectStatsPrint(int lvl)
{
	int cnt = 0;
	uint32 markedmem;

	if (lvl == 0) {
		return;
	}

	markedmem = gcStats.markedmem + objectStats.totalFixedMem;
	kprintf("<GC: heaptotal %dK, heaplimit %dK, inuse %dK, alloc %dK, marked %dK, freed %dK final %dK>\n", gc_heap_total/1024, gc_heap_limit/1024, gcStats.totalmem/1024, gcStats.allocmem/1024, markedmem/1024, gcStats.freedmem/1024, gcStats.finalmem/1024);

	if (lvl == 1) {
		return;
	}
	
#define	PRINT_GC_STAT(STAT)			\
	kprintf("%14.14s: Nr %6d  Mem %6dK",	\
	#STAT, objectStats.STAT##Nr, objectStats.STAT##Mem / 1024); \
	cnt++;					\
	if (cnt % 2 != 0) {			\
		kprintf("   ");			\
	}					\
	else {					\
		kprintf("\n");			\
	}
	
	kprintf("Memory statistics:\n");
	kprintf("------------------\n");
	
	PRINT_GC_STAT(object);
	PRINT_GC_STAT(array);
	PRINT_GC_STAT(class);
	PRINT_GC_STAT(method);
	PRINT_GC_STAT(field);
	PRINT_GC_STAT(static);
	PRINT_GC_STAT(dispatch);
	PRINT_GC_STAT(bytecode);
	PRINT_GC_STAT(exception);
	PRINT_GC_STAT(constant);
	PRINT_GC_STAT(utf8const);
	PRINT_GC_STAT(interface);
	PRINT_GC_STAT(jit);
	PRINT_GC_STAT(fixed);
	PRINT_GC_STAT(lock);
	PRINT_GC_STAT(ref);
	PRINT_GC_STAT(thread);
	PRINT_GC_STAT(threadCtx);
	PRINT_GC_STAT(classloader);
	PRINT_GC_STAT(jreference);
	PRINT_GC_STAT(analyse);
	PRINT_GC_STAT(inflate);
	PRINT_GC_STAT(strings);
	PRINT_GC_STAT(engine);
	PRINT_GC_STAT(jar);
	PRINT_GC_STAT(estack);
	PRINT_GC_STAT(awt);
	PRINT_GC_STAT(normal);
	PRINT_GC_STAT(string);
	PRINT_GC_STAT(other);
	
	if (cnt % 2 != 0) {
		kprintf("\n");
	}

#undef	PRINT_GC_STAT

	objectSizesPrint();
}

static
void
objectSizesAdd(size_t sz)
{
	int i;
	for (i = 0; sz > objectSizes[i].size; i++)
		;
	objectSizes[i].count++;
}

static
void
objectSizesPrint(void)
{
	int i;
	uint64 total;
	int count;
	int nr;
	int j;

	total = 0;
	count = 0;
	for (i = 0; objectSizes[i].size != -1; i++) {
		total += (uint64)objectSizes[i].size * (uint64)objectSizes[i].count;
		count += objectSizes[i].count;
	}
	nr = i;

	if (total == 0) {
		return;
	}

	kprintf("Percentage size allocations:\n");
	kprintf("----------------------------\n");


	for (i = 0; i < nr; ) {
		for (j = 0; j < 3 && i < nr; j++, i++) {
			kprintf("%10d :%5.1f /%5.1f  ", objectSizes[i].size, (float)(objectSizes[i].count * 100 / (float)count), (float)((uint64)objectSizes[i].size * (uint64)objectSizes[i].count * 100 / (float)total));
		}
		kprintf("\n");
	}
}

#endif

GarbageCollectorInterface Kaffe_GarbageCollectorInterface = {
	gcMalloc,
	gcRealloc,
	gcFree,
		
	gcInvokeGC,
	gcInvokeFinalizer,
	
	gcAddRef,
	gcRmRef,
	
	gcRegister,		
};
