/* 
 * collect.c
 * The incremental garbage collector.
 *
 * Copyright (c) 1999 - 2001
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "locks.h"
#include "gc.h"
#include "gc-env.h"
#include "threads.h"
#include "lookup.h"
#include "java_lang_ref_Reference.h"
#include "support.h"
#include "jmeasure.h"

#define DBG(_t, _m)

extern void Kaffe_CallVoidMethodA(JNIEnv*, jobject, jmethodID, jvalue*);
extern void Kaffe_ExceptionClear(JNIEnv*);

iLock* gcman1;
iLock* gcman2;
iLock* finman1;
iLock* finman2;
extern iLock* gc_lock;

int gc_mode = GC_DISABLED;	/* GC will be enabled after the first
				 * thread is setup.
				 */

static bool gcRunning = false;
static bool finalRunning = false;
int urgentState = 0;

#define	REFOBJALLOCSZ	8
#define	REFOBJHASHSZ	8
typedef struct _refObject {
	void*			start;
	void*			end;
	uintp			ref;
	struct _refObject*	next;
} refObject;
typedef struct _refTable {
	refObject*		hash[REFOBJHASHSZ];
} refTable;
static refTable			refObjects;
#define	REFOBJHASH(V)		((((uintp)(V))/(2*sizeof(uintp)))%REFOBJHASHSZ)

#if defined(GC_INCREMENTAL)
static int gc_firsttime = 1;
#endif
#if defined(GC_VERIFY)
int in_verify = 0;
#endif

int gc_count = 0;
#if defined(GC_EXP)
uint black_val;
uint white_val;
heap_obj black_list;
heap_obj white_list;
heap_obj free_list;
heap_obj finalize_list;
heap_obj finalize_class_list;
#else
heap_obj* white_list;
heap_obj* finalize_list;
heap_obj* finalize_class_list;
#endif
heap_obj* grey_list;

extern Hjava_lang_ref_Reference* refQueue;
extern Hjava_lang_ref_Reference* refDispatchQueue;

int gc_grey_cnt;

#if defined(GC_INCREMENTAL)
static int gc_inc_limit = 100;
static int gc_inc_calls;
static int gc_inc_objs;

inline void
gcWriteRefFast(void* from, void* to)
{
        heap_obj* obj = MEM2HEAP_OBJ(to);
        uint32 n = 0;
        int iLockRoot;

        lockStaticMutex(&gc_lock);
                        
        DBG(gc_incremental,("%p -> %p\n",from,to));

        REMOVE_OBJ(obj);
        INSERT_GREYLIST(obj);

        while (grey_list && n < gc_inc_limit) {
                obj = grey_list;
                grey_list = GET_NEXTGREY(obj);
                _walkMemory( obj );
#if defined(GC_EXP)
                INSERT_OBJ(black_list, obj);
#endif
                n++;
        }
        gc_inc_calls++; 
        gc_inc_objs += n;

        unlockStaticMutex(&gc_lock);
}

void
gcWriteRef(void* from, void* to)
{
        heap_obj* from_obj = MEM2HEAP_OBJ(from);
        heap_obj* to_obj;
        MSR_TIME(gc_writeref);

        MSR_START(gc_writeref);
        if (!IS_COLOR(from_obj, COLOR_WHITE) && to) {
                page_blk* header = MEM2PAGE_HDR(to);

                if (to < (void*)(void*)gc_heap_min || to >= (void*)gc_heap_max ||
                    ((uint32)to - (uint32)header) < sizeof(page_blk) ||
                        ((uint)to & 0x3)) return;
                if (IS_FIXED_PAGE(header)) return;
                to_obj = MEM2HEAP_OBJ(to);
                if (IS_COLOR(to_obj, COLOR_WHITE)) {
                        if (IS_BLOB_PAGE(header)) {
                                if (MEM2HEAP_BLOB(to)->magic != GC_MAGIC) return;
                        } else {
                                if (to_obj->gctable == 0) return;
                                if ( (uint32)((void*)to_obj - ((void*)header + sizeof(page_blk))) % 
                                     (PAGE_OBJSIZE(header) + sizeof(heap_obj)) )
                                        return;
                        }
                        if (header->magic != GC_MAGIC) return;

                        gcWriteRefFast(from, to);
                }
        }
        MSR_STOP(gc_writeref);
}
#else
void
gcWriteRef(void* from, void* to)
{
}
#endif

inline static
void
finalizerStart(void)
{
        int iLockRoot;
        int start;

	/* If there's stuff to be finalised or dispatched then we'd
	 * better do it.  Note that we mustn't hold the gc_lock and finman1
	 * lock at the same time!!.
	 */
        lockStaticMutex(&gc_lock);
        start = 0;
#if defined(GC_EXP)
	if (finalize_list.next != 0 || refDispatchQueue != 0) {
#else
	if (finalize_list != 0 || refDispatchQueue != 0) {
#endif
                start = 1;
        }
        unlockStaticMutex(&gc_lock);

        if (start != 0) {
		lockStaticMutex(&finman1);
                if (finalRunning == false) {
                        finalRunning = true;
                        signalStaticCond(&finman1);
                }
		unlockStaticMutex(&finman1);
	}

}

inline static 
void
gcStart(void)
{
	int i;
	refObject* robj;
        heap_obj * obj;
#ifdef GC_EXP
        heap_obj * next;
#endif

        /* Walk the referenced objects */
	for (i = 0; i < REFOBJHASHSZ; i++) {
		for (robj = refObjects.hash[i]; robj != 0; robj = robj->next) {
                        markObject(robj->start);
			walkConservative(robj->start, robj->end - robj->start);
		}
	}

        /* Walk all objects on the finalizer list */
#if defined(GC_EXP)
        obj = finalize_list.next;
        while (obj) {
                next = GET_NEXT(obj);
                _walkMemory( obj );
                obj = next;
        }
#else
        obj = finalize_list;
        while (obj) {
                _walkMemory( obj );
                obj = GET_NEXT(obj);
        }
#endif

#if defined(GC_EXP)
        obj = finalize_class_list.next;
        while (obj) {
                next = GET_NEXT(obj);
                _walkMemory( obj );
                obj = next;
        }
#else
        obj = finalize_class_list;
        while (obj) {
                _walkMemory( obj );
                obj = GET_NEXT(obj);
        }
#endif

        /* Walk the thread objects */
	Kaffe_TI_GcWalkThreads();

}

inline static 
void
gcFinish(void)
{
	/* 
	 * Move white objects to finalize list
	 */
#if defined(GC_EXP)
        heap_obj *obj = black_list.next;
        int iLockRoot;
        MSR_TIME(gc_free_time);

        white_list.next = black_list.next;
        white_list.next->prev = &white_list;
        black_list.next = 0;

        if (white_val == 0) {
                white_val = 0x2;
                black_val = 0x0;
        } else {
                white_val = 0x0;
                black_val = 0x2;
        }

        lockStaticMutex(&gc_lock);
        MSR_START(gc_free_time);
        while (free_list.next) {
                obj = free_list.next;
                REMOVE_OBJ(obj);
                heap_free(HEAP_OBJ2MEM(obj));
        }
        MSR_STOP(gc_free_time);
        unlockStaticMutex(&gc_lock);
#else
        heap_obj *next;
        heap_obj *obj, **list = &white_list;
        int iLockRoot;
        MSR_TIME(gc_free_time);

        lockStaticMutex(&gc_lock);
        MSR_START(gc_free_time);
        while( (obj = *list) != NULL ){
                if( IS_FINAL( obj ) && IS_COLOR( obj, COLOR_WHITE ) ){
                        next = GET_NEXT( obj );
                        REMOVE_OBJ( obj );
                        INSERT_OBJ( finalize_list, obj );
                        *list = next;
                } else if( IS_COLOR( obj, COLOR_WHITE ) ){
                        next = GET_NEXT( obj );
                        REMOVE_OBJ( obj );
                        heap_free( HEAP_OBJ2MEM( obj ) );
                        *list = next;
                } else {
                        SET_COLOR( obj, COLOR_WHITE );
                        list = &GET_NEXT( obj );
                }
        }
        MSR_STOP(gc_free_time);
        unlockStaticMutex(&gc_lock);
#endif
}

inline static
void
gcMustFree(void)
{
}

/*
 * The Garbage Collector sits in a loop starting a collection, waiting
 * until it's finished incrementally, then tidying up before starting
 * another one.
 */
void
gcMan(void* arg)
{
        heap_obj *obj, *next;
        int gc_alloc;
        int iLockRoot;
        int n;
        MSR_TIME(gc_sweep_time)
        MSR_TIME(gc_roots_time)

	lockStaticMutex(&gcman1);

	for(;;) {
                lockStaticMutex(&gc_lock);

		(*Kaffe_ThreadInterface.enterCritSect)();

                gc_alloc = gc_current_alloc;

#if defined(GC_INCREMENTAL)
DBG(gc,         printf("GC INC start: %db\n",gc_current_alloc);)
                if (!gc_firsttime) {
                        MSR_START(gc_roots_time);
                        gcStart();
                        MSR_STOP(gc_roots_time);

                        MSR_START(gc_sweep_time);
                        /* Cleanup remaining objects on greylist */
                        while (grey_list) {
                                obj = grey_list;
                                grey_list = 0;
                                n = 0;
                                while (obj) {
                                        next = GET_NEXTGREY(obj);
                                        _walkMemory( obj );
#if defined(GC_EXP)
                                        INSERT_OBJ(black_list, obj);
#endif
                                        obj = next;
                                        n++;
                                }
                                gc_inc_objs += n;
                        }
                        gc_inc_limit = gc_inc_objs / gc_inc_calls;
                        
                        markReferences();
                        
                        /* 
                         * Now walk any white objects which will be finalied.  They
                         * may get reattached, so anything they reference must also
                         * be live just in case.
                         */
#if defined(GC_EXP)
                        while (white_list.next) {
                                void* final;
                                obj = white_list.next;
                                REMOVE_OBJ(obj);
                                final = obj->gctable->final;
                                if( final != 0 ){
                                        if (final == finalizeObject) {
                                                assert( OBJECT_CLASS((Hjava_lang_Object*)HEAP_OBJ2MEM(obj))->finalizer != NULL );
                                        }
                                        MARK_FINAL(obj);
                                        _walkMemory( obj );
                                        INSERT_OBJ(finalize_list, obj);
                                }
                                else {
                                        INSERT_OBJ(free_list, obj);
                                }
                        }
#else
                        obj = white_list;
                        while (obj) {
                                if( IS_COLOR(obj, COLOR_WHITE) ) {
                                        void* final = obj->gctable->final;
                                        if (final) {
                                                if (final == finalizeObject) {
                                                        assert( OBJECT_CLASS((Hjava_lang_Object*)HEAP_OBJ2MEM(obj))->finalizer != NULL );
                                                }
                                                MARK_FINAL(obj);
                                                _walkMemory( obj );
                                        }
                                }
                                obj = GET_NEXT(obj);
                        }
#endif
                        
                        /* We may now have more grey objects, so walk them */
                        while (grey_list) {
                                obj = grey_list;
                                grey_list = 0;
                                while (obj) {
                                        next = GET_NEXTGREY(obj);
                                        _walkMemory( obj );
#if defined(GC_EXP)
                                        INSERT_OBJ(black_list, obj);
#endif
                                        obj = next;
                                }
                        }

                        MSR_STOP(gc_sweep_time);
                        gcFinish();
                }
                gc_firsttime = 0;

                gc_inc_calls++;
		if (Kaffe_JavaVMArgs[0].enableVerboseGC > 0) {
                        kprintf("<GC: incremental obj %d, calls %d, o/c %d>\n", gc_inc_objs, gc_inc_calls, gc_inc_objs/gc_inc_calls);
                }

                refQueue = 0;
                refDispatchQueue = 0;
#if !defined(GC_EXP)
                grey_list = 0;
#endif
                gc_inc_calls = 1;
                gc_inc_objs = 0;
                gc_grey_cnt = 0;

                /* Mark all root objects */
                MSR_START(gc_roots_time);
                gcStart();
                MSR_STOP(gc_roots_time);
#else
DBG(gc,         printf("GC MS start: %db\n",gc_current_alloc);)
                refQueue = 0;
                refDispatchQueue = 0;
                grey_list = 0;

                MSR_START(gc_roots_time);
                gcStart();
                MSR_STOP(gc_roots_time);

                MSR_START(gc_sweep_time);
                while (grey_list) {
                        obj = grey_list;
                        grey_list = 0;
                        while (obj) {
                                next = GET_NEXTGREY(obj);
                                _walkMemory( obj );
                                obj = next;
                        }
                }

		markReferences();

                /* 
                 * Now walk any white objects which will be finalied.  They
                 * may get reattached, so anything they reference must also
                 * be live just in case.
                 */
                obj = white_list;
                while (obj) {
                        if( IS_COLOR(obj, COLOR_WHITE) ){
                                void* final = obj->gctable->final;
                                if (final) {
                                        if (final == finalizeObject) {
                                                assert( OBJECT_CLASS((Hjava_lang_Object*)HEAP_OBJ2MEM(obj))->finalizer != NULL );
                                        }
                                        MARK_FINAL(obj);
                                        _walkMemory( obj );
                                }
                        }
                        obj = GET_NEXT(obj);
                }

		/* We may now have more grey objects, so walk them */
                while (grey_list) {
                        obj = grey_list;
                        grey_list = 0;
                        while (obj) {
                                next = GET_NEXTGREY(obj);
                                _walkMemory( obj );
                                obj = next;
                        }
                }

                MSR_STOP(gc_sweep_time);
                gcFinish();
#endif
                calculateAdvice();

		(*Kaffe_ThreadInterface.exitCritSect)();

                unlockStaticMutex(&gc_lock);

                gcMustFree();

                finalizerStart();

DBG(gc,         printf("GC end: %db\n",gc_current_alloc);)
                if (((float)gc_current_alloc/(float)gc_alloc) > 0.67F && 
                    (gc_panic_level == 0)) {
                        gc_heap_allocation_size += Kaffe_JavaVMArgs[0].allocHeapSize;
                         if( gc_heap_allocation_size > Kaffe_JavaVMArgs[0].maxHeapSize ){
                                 gc_heap_allocation_size = Kaffe_JavaVMArgs[0].maxHeapSize;
                         }
                }

		if (Kaffe_JavaVMArgs[0].enableVerboseGC > 0) {
			kprintf("<GC: before %dK, after %dK, limit %dK>\n", gc_alloc/1024, gc_current_alloc/1024, gc_heap_allocation_size/1024);
		}
#if defined(GC_STATS)
		if (Kaffe_JavaVMArgs[0].enableVerboseGC > 1) {
                        gcPrintStats();
		}
#endif
		
		/* Wake up anyone waiting for the GC to finish */
		lockStaticMutex(&gcman2);
		broadcastStaticCond(&gcman2);
		unlockStaticMutex(&gcman2);

                urgentState = 0;
		gcRunning = false;
		waitStaticCond(&gcman1, NOTIMEOUT);
		assert(gcRunning == true);
	}
}

/*
 * The finaliser sits in a loop waiting to finalise objects.  When a
 * new finalised list is available, it is woken by the GC and finalises
 * the objects in turn.  An object is only finalised once after which
 * it is deleted.
 */
void
finaliserMan(void* arg)
{
        heap_obj *obj;
	Hjava_lang_ref_Reference* ref;
        int iLockRoot;
        jmethodID dispatch;
        int n;
        gcTable * gctable;
        void * mem;
        finalFunc final;
        MSR_TIME(gc_finalize_time);

	lockStaticMutex(&finman1);

        dispatch = 0;
	for (;;) {
		finalRunning = false;
                while( !finalRunning ){
                        waitStaticCond(&finman1, NOTIMEOUT);
                }
		assert(finalRunning == true);

                n = 0;
                MSR_START(gc_finalize_time);

		/* Dispatch any java.lang.ref.References */
		for (ref = refDispatchQueue; ref != 0; ref = unhand(ref)->next) {
                        unhand(ref)->obj = 0;
                        if( dispatch == 0 ){
                                dispatch = findMethod( ReferenceClass, dispatch_name, void_sig );
                                assert(dispatch != 0);
                        }
                        Kaffe_CallVoidMethodA(0, (jobject)ref, dispatch, 0);
                        Kaffe_ExceptionClear(0);
		}

#if defined(GC_EXP)
                while (finalize_list.next) {
                        lockStaticMutex(&gc_lock);
                        obj = finalize_list.next;
                        REMOVE_OBJ(obj);
#else
                while (finalize_list) {
                        lockStaticMutex(&gc_lock);
                        obj = finalize_list;
                        finalize_list = GET_NEXT(obj);
                        REMOVE_OBJ(obj);
#endif
                        n++;
                        gctable = obj->gctable;
                        if( gctable == GC_ALLOC_CLASSOBJECT ){
                                INSERT_OBJ(finalize_class_list, obj);
                                unlockStaticMutex(&gc_lock);

                        } else {
                                unlockStaticMutex(&gc_lock);
                                assert( gctable > gc_table && gctable < &gc_table[MAX_GC_FUNCS] );
                                final = gctable->final;
                                mem = HEAP_OBJ2MEM(obj);
                                if( final !=  NULL ){
                                        unlockStaticMutex(&finman1);
                                        (*final)( mem );
                                        lockStaticMutex(&finman1);
                                }
                                lockStaticMutex(&gc_lock);
                                heap_free( mem );
                                unlockStaticMutex(&gc_lock);
                        }
                }
#if defined(GC_EXP)
                while (finalize_list.next) {
                        lockStaticMutex(&gc_lock);
                        obj = finalize_list.next;
                        REMOVE_OBJ(obj);
#else
                while (finalize_class_list) {
                        lockStaticMutex(&gc_lock);
                        obj = finalize_class_list;
                        finalize_class_list = GET_NEXT(obj);
                        REMOVE_OBJ(obj);
#endif
                        unlockStaticMutex(&gc_lock);
                        final = obj->gctable->final;
                        mem = HEAP_OBJ2MEM(obj);
                        if( final != NULL ){
                                unlockStaticMutex(&finman1);
                                (*final)( mem );
                                lockStaticMutex(&finman1);
                        }
                        lockStaticMutex(&gc_lock);
                        heap_free( mem );
                        unlockStaticMutex(&gc_lock);
                }
                MSR_STOP(gc_finalize_time);
		if (Kaffe_JavaVMArgs[0].enableVerboseGC > 0) {
			kprintf("<GC: finalized %d obj>\n",n);
		}

		/* Wake up anyone waiting for the finalizer to finish */
		lockStaticMutex(&finman2);
		broadcastStaticCond(&finman2);
		unlockStaticMutex(&finman2);
	}
}

/*
 * Explicity invoke the garbage collector and wait for it to complete.
 */
void
invokeGC(int urgent)
{
        int iLockRoot;
	lockStaticMutex(&gcman1);
	lockStaticMutex(&gcman2);

	if (gcRunning == false) {
		gcRunning = true;
                urgentState = urgent;
		signalStaticCond(&gcman1);
	}
	unlockStaticMutex(&gcman1);
	waitStaticCond(&gcman2, NOTIMEOUT);
	unlockStaticMutex(&gcman2);
}

/*
 * Explicity invoke the finalizer.
 */
void
invokeFinalizer(void)
{
        int iLockRoot;

	/* First invoke the GC */
	invokeGC(0);

	/* Run the finalizer (if might already be running as a result of
	 * the GC)
	 */
	lockStaticMutex(&finman1);
	lockStaticMutex(&finman2);

	if (finalRunning == false) {
		finalRunning = true;
		signalStaticCond(&finman1); 
	}
	unlockStaticMutex(&finman1);

	waitStaticCond(&finman2, NOTIMEOUT);
	unlockStaticMutex(&finman2);
}       


/*
* Add a presistent reference to an object.
*/
void*
gcAddRootRef(void* start, void* end)
{
	uint32 idx;
	refObject* obj;
	
	idx = REFOBJHASH(start);
	for (obj = refObjects.hash[idx]; obj != 0; obj = obj->next) {
		/* Found it - just increase reference */
		if (obj->start == start) {
			obj->ref++;
			if (end > obj->end) {
				obj->end = end;
			}
			return (obj);
		}
	}
	
	/* Not found - create a new one */
	obj = (refObject*)gc_malloc_fixed_typed(sizeof(refObject), GC_ALLOC_REF);
	if (obj != 0) {
		obj->start = start;
		obj->end = end;
		obj->ref = 1;
		obj->next = refObjects.hash[idx];
		refObjects.hash[idx] = obj;
		return (obj);
	}
	else {
		return (0);
	}
}

/*
* Remove a persistent reference to an object.  If the count becomes
* zero then the reference is removed.
*/
void
gcDelRootRef(void* start)
{
	uint32 idx;
	refObject** objp;
	refObject* obj;
	
	idx = REFOBJHASH(start);
	for (objp = &refObjects.hash[idx]; *objp != 0; objp = &obj->next) {
		obj = *objp;
		/* Found it - just increase reference */
		if (obj->start == start) {
			obj->ref--;
			if (obj->ref == 0) {
				*objp = obj->next;
				fixed_free(obj);
			}
			break;
		}
	}
}

GcAdvice* advice;

/*
 * Register a GC advice object - we only support one at the moment.
 */
void
gcRegisterAdvice(GcAdvice* adv)
{
	advice = adv;
}

/*
 * Calculate memory advice and signal anyone interested.
 */
void
calculateAdvice(void)
{
	static int lastused = 0;
	int used = 0;
	int col;
	if (!advice) {
		return;
	}

	col = advice->color;
	// used = (gcStats.totalmem * 100) / gc_heap_limit;

	if (used > lastused) {
		if (col < ADVICE_RED && used > advice->hi_red) {
			col = ADVICE_RED;
		}
		else if (col < ADVICE_ORANGE && used > advice->hi_orange) {
			col = ADVICE_ORANGE;
		}
		else if (col < ADVICE_YELLOW && used > advice->hi_yellow) {
			col = ADVICE_YELLOW;
		}
	}
	else {
		if (col > ADVICE_GREEN && used < advice->lo_yellow) {
			col = ADVICE_GREEN;
		}
		else if (col > ADVICE_YELLOW && used < advice->lo_orange) {
			col = ADVICE_YELLOW;
		}
		else if (col > ADVICE_ORANGE && used < advice->lo_red) {
			col = ADVICE_ORANGE;
		}
	}
	lastused = used;

	/* If we get a color change, wake anyone interested in it */
	if (col != advice->color) {
		lockObject(&advice->head);
		advice->color = col;
		broadcastCond(&advice->head);
		unlockObject(&advice->head);
	}
}
