/*
 * gc.h
 * The garbage collector.
 *
 * Copyright (c) 1999 - 2001
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __gc_h
#define __gc_h

/* Enable the garbage collector - recommended */
#define	GC_ENABLE 1

/* #define GC_EXP 1 */
#define GC_INCREMENTAL 1
#define GC_SAFE 1
#define GC_64B 1

#include "java_lang_Object.h"
#include "../gtypes.h"

/* This should match the kaffe.lang.MemoryAdvice class */

#define	ADVICE_GREEN		0
#define	ADVICE_YELLOW		1
#define	ADVICE_ORANGE		2
#define	ADVICE_RED		3

typedef struct {
	Hjava_lang_Object	head;
	jint			color;
	jint			hi_yellow;
	jint			lo_yellow;
	jint			hi_orange;
	jint			lo_orange;
	jint			hi_red;
	jint			lo_red;
} GcAdvice;

#define	gc_malloc(A,B)		heap_malloc(A,B)
#define	gc_malloc_fixed_typed(A,B)	fixed_malloc(A)
#define	gc_malloc_opt(A,B,C)	heap_malloc_obj(B,C)
#define	gc_calloc(A,B,C)	heap_malloc((A)*(B),C)
#define	gc_calloc_fixed_typed(A,B,C)	fixed_malloc((A)*(B))
#define	gc_calloc_fixed(A,B)	fixed_malloc((A)*(B))
#define	gc_malloc_fixed(A)	fixed_malloc(A)
#define	gc_realloc_fixed(A,B)	fixed_realloc(A,B)
#define	gc_free(A)		heap_free(A)
#define	gc_free_fixed(A)	fixed_free(A)

#define	gc_add_ref(A,B)		gcAddRootRef(A,B)
#define	gc_rm_ref(A)		gcDelRootRef(A)

#define	gc_register(A)		gcRegisterAdvice(A)

#define object_gc_color		((-2)*sizeof(void*))

#define heapFree()		(gc_heap_limit - gc_heap_total)
#define heapTotal()		(gc_heap_limit)

#if defined(GC_INCREMENTAL)
#define GC_WRITE(_f, _t)	gcWriteRef(_f, _t)
#define GC_WRITE_SAFE(_f, _t)	gcWriteRef(_f, _t)
#else
#define GC_WRITE(_o, _p)
#define GC_WRITE_SAFE(_f, _t)
#endif

struct _gcTable;
struct _page_blk;

/* Exported from page.h */

typedef struct _page_blk {
	uint16 			info;
	uint16 			n_pages;
	uint32			magic;
	uint32			used;
	struct _page_blk* 	prev;
	struct _page_blk* 	next;
	void* 			freelist;
} page_blk;

#define MIN_HEAPSIZE		0
#define MAX_HEAPSIZE		(64*1024*1024)
#define ALLOC_HEAPSIZE		(1024*1024)
#define MIN_BLOB_SIZE		200
#define ALIGN_OBJ		3
#define SZ2FREELIST(_sz)	(((_sz) + 7) >> ALIGN_OBJ)

JNIEXPORT void* JNICALL fixed_malloc(int);
JNIEXPORT void* JNICALL heap_malloc(int, struct _gcTable*);
JNIEXPORT void* JNICALL heap_malloc_obj(struct _page_blk**, struct _gcTable*);
JNIEXPORT void* JNICALL fixed_realloc(void*, int);
JNIEXPORT void  JNICALL fixed_free(void*);
JNIEXPORT void  JNICALL heap_free(void*);
JNIEXPORT void  JNICALL mm_init();

JNIEXPORT unsigned int gc_pgsize;
JNIEXPORT struct _page_blk * free_fixed_obj[];
JNIEXPORT struct _page_blk * free_heap_obj[];
JNIEXPORT uint32 gc_heap_total;
JNIEXPORT uint32 gc_heap_limit;

/* Exported from gc-funcs.h */

typedef void (*walkFunc)( void *, unsigned int );
typedef void (*finalFunc)( void * );

typedef struct _gcTable {
	walkFunc		walk;
	finalFunc		final;
} gcTable;

extern gcTable gc_table[];

#define	GC_ALLOC_NORMAL		(&gc_table[0])
#define	GC_ALLOC_ESTACK		(&gc_table[1])
#define	GC_ALLOC_FIXED		(&gc_table[2])
#define	GC_ALLOC_NORMALOBJECT	(&gc_table[3])
#define	GC_ALLOC_PRIMARRAY	(&gc_table[4])
#define	GC_ALLOC_REFARRAY	(&gc_table[5])
#define	GC_ALLOC_CLASSOBJECT	(&gc_table[6])
#define	GC_ALLOC_THREAD		(&gc_table[7])
#define	GC_ALLOC_STATICDATA	(&gc_table[8])
#define	GC_ALLOC_BAG		(&gc_table[9])
#define	GC_ALLOC_JREFERENCE	(&gc_table[10])
#define	GC_ALLOC_CLASSLOADER	(&gc_table[11])
#define	GC_ALLOC_NORMALOBJECT_NOFINALIZE	(&gc_table[12])
#define	GC_ALLOC_STRING		(&gc_table[13])
#define	GC_ALLOC_STRINGS	0
#define	GC_ALLOC_CLASSPOOL	GC_ALLOC_BAG
#define MAX_GC_FUNCS		16

extern void walkConservative(void*, unsigned int);

/* Exported from collect.h */

extern int gc_mode;

#define	GC_DISABLED		0
#define	GC_ENABLED		1

extern void* gcAddRootRef(void*, void*);
extern void gcDelRootRef(void*);
extern void gcWriteRef(void*,void*);
extern void calculateAdvice();
extern void invokeGC(int);
extern void invokeFinalizer();
extern void gcRegisterAdvice(GcAdvice*);

#endif
