/* 
 * page.c
 * The heap manager.
 *
 * Copyright (c) 1999 - 2001
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "classMethod.h"
#include "gtypes.h"
#include "locks.h"
#include "gc.h"
#include "gc-env.h"
#include "jsyscall.h"
#include "soft.h"
#include "exception.h"
#include "support.h"
#include "page.h"

//#define DBG(_t, _m) if (dbg) {printf _m ; fflush(stdout);}
//static int dbg = 1;
#define DBG(_t, _m) 

page_blk* free_fixed_obj[SZ2FREELIST(MIN_BLOB_SIZE)+1];
page_blk* free_fixed_blob;
page_blk* free_heap_obj[SZ2FREELIST(MIN_BLOB_SIZE)+1];
page_blk* free_heap_blob;

uint32 page_size;
uint32 collected = 0;
page_blk free_pages;
static page_blk* last_page;
uintp gc_heap_min;
uintp gc_heap_max;
iLock* gc_lock;

/*
 * Compatibility
 */
size_t gc_pgsize;
size_t gc_heap_limit;
size_t gc_heap_total;
size_t gc_heap_allocation_size;
int32 gc_current_alloc;
extern struct Hjava_lang_Thread* garbageman;
int gc_panic_level = 0;
int n_req_pages;

int fake_memory;
#define MALLOC_ZERO_BYTES 	(void*)(&fake_memory)

extern void free_freelists();

static void 
outOfMemory() 
{
        switch (gc_panic_level) {
        case 1:
                invokeGC(0);
                break;
        case 2:
                free_freelists();
                break;
        case 3:
                invokeGC(0);
                break;
        case 4:
                free_freelists();
                break;
        case 5:
                kprintf( "PANIC: OutOfMemory !!!\n" );
                kprintf( "page_alloc: panic %d: limit %dK: current %dK\n", gc_panic_level, gc_heap_limit/1000, gc_heap_total/1000 );
                postException("java/lang/OutOfMemoryError",0);
                break;
        }
DBG(mm, ("page_alloc: panic %d: limit %dK: current: %dK\n", gc_panic_level, gc_heap_limit/1000, gc_heap_total/1000));
}

/*
 * The foundation of the GC is a paged allocation system.
 * Each page is assigned a specific function when it is
 * allocated to the memory system.
 */
page_blk *
page_alloc(int n_pages) 
{
        page_blk * rem, * new, * tmp;
        int n_rem_pages;
        page_blk * cur = &free_pages;

        n_rem_pages = 0;
        if( gc_heap_total > gc_heap_limit ){
                gc_panic_level++;
                n_req_pages = n_pages;
                return NULL;
        } 
        else{
                gc_panic_level = 0;
        }
        while( cur->next != NULL ){
                if( (n_rem_pages = cur->next->n_pages - n_pages) >= 0 ){
                        break;
                }
                cur = cur->next;
        }
        gc_heap_total += n_pages * page_size;
        if( cur->next != NULL ){
                /* Found block in freelist */
DBG(mm, 	("page_alloc: %d freelist pages @ %x from %d pages\n",n_pages,cur->next,cur->next->n_pages));
                new = cur->next;
                if( n_rem_pages > 0 ){
                        rem = PAGE_BLK_OFFSET( new, n_pages );
                        rem->next = new->next;
			rem->prev = new;
                        rem->n_pages = n_rem_pages;
                        cur->next = rem;
                        tmp = PAGE_BLK_OFFSET( rem, n_rem_pages );
                        if( (void*)tmp < (void*)gc_heap_max ){
                                tmp->prev = rem;
                        }
                } 
                else {
                        cur->next = cur->next->next;
                }
        } 
        else {
                /* Else, create page */
                new = (page_blk*)pagealloc( n_pages * page_size );
                new->prev = last_page;
                new->next = 0;
                last_page = new;
                gc_heap_max = (uintp)( (void*)new + n_pages * page_size );
DBG(mm, 	("page_alloc: %d new pages @ %x\n",n_pages,new));
        }
        new->used = 0;
        new->freelist = NULL;
        new->n_pages = n_pages;
        new->info = 0;
        SET_ALLOCATED_PAGE( new );
        SET_PAGE_MAGIC( new );
        return new;
}

/*
 * This routine returns free pages back to the page allocator.
 * Pointers assigned during allocation try to make this efficient.
 */
void
page_free( page_blk * free )
{
        int merged = 0;
        page_blk * next, * prev, * tmp;

        /* Merge free page to adjacent blk, if possible */
        free->info = 0;
        free->magic = 0;
        gc_heap_total -= free->n_pages * page_size;
        next = PAGE_BLK_OFFSET( free, free->n_pages );
        prev = free->prev;

        if( next <= last_page ){
                if( !IS_ALLOCATED_PAGE( next ) ){
DBG(mm, 		("page_free: merge %d pages @ %x to next %x\n",free->n_pages,free,next));
                        free->n_pages += next->n_pages;
                        free->next = next->next;
                        tmp = PAGE_BLK_OFFSET( free, free->n_pages );
                        if( (void*)tmp < (void*)gc_heap_max ){
                                tmp->prev = free;
                        }
                        tmp = free;
                        while( (tmp = tmp->prev) != NULL ) {
                                if( !IS_ALLOCATED_PAGE( tmp ) ){
                                        tmp->next = free;
                                        break;
                                }
                        }
                        merged = 1;
                }
        }
        if( !IS_ALLOCATED_PAGE( prev ) && prev != &free_pages ){
DBG(mm, 	("page_free: merge %d pages @ %x to prev %x\n",free->n_pages,free,prev));
                prev->n_pages += free->n_pages;
                tmp = PAGE_BLK_OFFSET( prev, prev->n_pages );
                if( (void*)tmp < (void*)gc_heap_max ){
                        tmp->prev = prev;
                }
                if( merged ){
                        prev->next = free->next;
                }
                return;
        }

        /* Else, put it back on the freelist, in order */
        if( !merged ){
DBG(mm, 	("page_free: %d pages @ %x into freelist\n",free->n_pages,free));
		prev = free;
                while( (prev = prev->prev) != NULL ){ 
			if( !IS_ALLOCATED_PAGE( prev ) ) {
DBG(mm, 			("page_free: insert between %x & %x\n",prev,prev->next));
				free->next = prev->next;
				prev->next = free;
				return;
                        }
                }
        }
}

/*
 * Allocate a new page that will be used to hold small objects
 * of the same size.  This allows us to minimize overheads.
 */
static int
new_obj_blk( page_blk ** list, int hdr_sz, int is_fixed ) 
{
        int n;
	page_blk ** freelist = is_fixed ? free_fixed_obj : free_heap_obj;
        int obj_sz = ( ( (void*)list - (void*)freelist ) / sizeof(page_blk*) ) << ALIGN_OBJ;
#if 0
        int align_sz = (obj_sz + hdr_sz + 7) & ~0x7;
#else
        int align_sz = obj_sz + hdr_sz;
#endif
        int n_obj = ( page_size - sizeof(page_blk) )/ align_sz;
        page_blk * blk = (page_blk*)page_alloc( 1 );
        base_obj * obj = (base_obj*)( (void*)blk + sizeof(page_blk) );
DBG(mm, ("malloc: build page: %d objs of %db\n",n_obj,obj_sz));

        if( blk == NULL ){
                return 0;
        }
        blk->next = (*list);
        (*list) = blk;
        blk->freelist = obj;
        SET_PAGE_OBJSIZE( blk,obj_sz );
        if( is_fixed ){
                SET_FIXED_PAGE( blk );
        }
        for( n = n_obj; --n > 0; ){
                obj->next = (base_obj*)( (void*)obj + align_sz );
                if( !is_fixed ){
                        ((heap_obj*)obj)->gctable = NULL;
                }
                obj = obj->next;
        }
	obj->next = NULL;
        return 1;
}

/*
 * Allocate a block of page(s) that will be used to hold 
 * large, variable sized objects (blob).
 */
static int
new_blob_blk( int sz, int hdr_sz, int is_fixed )
{
        int obj_sz = sz + hdr_sz + sizeof(page_blk);
        int n_pages = obj_sz % page_size ? obj_sz / page_size + 1 : obj_sz / page_size;
        page_blk* blk = (page_blk*)page_alloc( n_pages );
        base_blob* obj = (base_blob*)( (void*)blk + sizeof(page_blk) );
	page_blk** freelist = is_fixed ? &free_fixed_blob : &free_heap_blob;

DBG(mm, ("malloc: build page: 1 objs of %db\n",n_pages*page_size));
        if( blk == NULL ){
                return 0;
        }
        blk->next = (*freelist);
        (*freelist) = blk;
        blk->freelist = obj;
        SET_BLOB_PAGE(blk);
        obj->size = n_pages*page_size - hdr_sz - sizeof(page_blk);
        obj->next = 0;
        if( !is_fixed ){
                ((heap_blob*)obj)->gctable = 0;
        }
        return 1;
}

/*
 * Malloc a blob, doing different things if the blob is
 * to be fixed (not part of the GC heap).
 */
void*
malloc_blob( int sz, int is_fixed, gcTable * gctable )
{
        page_blk ** list, * page;
        base_blob * new, * rem, * obj = NULL;
        base_blob ** cur = 0;
        void* mem;
        int found = 0;
        int hdr_sz = is_fixed ? sizeof(base_blob) : sizeof(heap_blob);
#if defined(GC_64B)
        int size = (sz & -8) + 2*sizeof(void*);
#else
        int size = (sz & -ALIGNMENTOF_VOIDP) + sizeof(void*);
#endif
	int blob_sz = hdr_sz + size;
        int iLockRoot;
        list = is_fixed ? &free_fixed_blob : &free_heap_blob;

	lockStaticMutex( &gc_lock );
        page = NULL;
        while( !found && (page = *list) != NULL ){
                if( page->n_pages * ( page_size - sizeof(page_blk) ) - page->used > blob_sz ){
                        cur = (base_blob**)&page->freelist;
                        while( (obj = *cur) != NULL ){
                                if( obj->size >= blob_sz
                                   && (void*)obj + hdr_sz < (void*)page + page_size ){
                                        found = 1;
                                        break;
                                }
                                else{
                                        cur = &obj->next;
                                }
                        }
                } 
                if( !found ){
                        list = &page->next;
                }
        }
        if( !found ){
                NOTIFY_COLLECTOR();
                while( new_blob_blk( size, hdr_sz, is_fixed ) == 0 ){
                        unlockStaticMutex( &gc_lock );
                        outOfMemory();
                        lockStaticMutex( &gc_lock );
                }
                list = is_fixed ? &free_fixed_blob : &free_heap_blob;
                page = *list;
                if( is_fixed ){
                        SET_FIXED_PAGE( page );
                }
                cur = (base_blob**)&page->freelist;
                obj = *cur;
        }
        page->used += blob_sz;
	gc_current_alloc += size;
DBG(mm, ("malloc: %x %db @ %x %s\n",page,size,obj, IS_FIXED_PAGE(page) ? "fixed" : ""));

        new = obj;
        if( new->size > blob_sz ){
                rem = (base_blob*)( (void*)new + blob_sz );
                rem->next = new->next;
                rem->size = new->size - blob_sz;
                new->size = size;
DBG(mm, 	("malloc: split into %db @ %x & %db @ %x\n",new->size,new,rem->size,rem));
                (*cur) = rem;
        } 
        else {
                (*cur) = new->next;
        }
        
        new->magic = is_fixed ? 0 : GC_MAGIC;
        
        if( is_fixed ){
                unlockStaticMutex( &gc_lock );
                ALTER_STATS( GC_ALLOC_FIXED, 1, size );
                mem = BLOB2MEM( new );
                memset( mem, 0, size );
                CHECK_ALIGNMENT( mem );
                return mem;
        } 
        else {
                heap_obj * obj = (heap_obj*)( (void*)new + 2 * sizeof(uint32) );
                obj->gctable = gctable;
                SET_COLOR( obj, COLOR_WHITE );
                INSERT_OBJ( white_list, obj );
                unlockStaticMutex( &gc_lock );
                ALTER_STATS( gctable, 1, size );
                mem = HEAP_OBJ2MEM( obj );
                memset( mem, 0, size );
                CHECK_ALIGNMENT( mem );
                return mem;
        }
}

/*
 * Malloc a fixed object, using a pointer to a freelist
 */
void*
fixed_malloc_obj( page_blk ** list)
{
        page_blk * page;
        base_obj * new;
        void * mem;
        int sz;
        int iLockRoot;

	lockStaticMutex( &gc_lock );
        if( (page = *list) == NULL ){
                NOTIFY_COLLECTOR();
                while( new_obj_blk( list, sizeof(base_obj), 1 ) == 0 ){
                        unlockStaticMutex( &gc_lock );
                        outOfMemory();
                        lockStaticMutex( &gc_lock );
                }
                page = *list;
        }

        /* Get from freelist */
        new = (base_obj*)page->freelist;
        page->freelist = (void*)new->next;
        page->used++;
        sz = PAGE_OBJSIZE( page );
	gc_current_alloc += sz;
DBG(mm, ("malloc: %db @ %x fixed\n",sz,new));
        if( page->freelist == NULL ){
                *list = page->next;
        }
	unlockStaticMutex( &gc_lock );
        ALTER_STATS( GC_ALLOC_FIXED, 1, sz );
        mem = OBJ2MEM( new );
        memset( mem, 0, sz );
        CHECK_ALIGNMENT( mem );
        return mem;
}

/*
 * Malloc a heap object, using a pointer to a freelist
 */
void*
heap_malloc_obj( page_blk ** list, gcTable * gctable )
{
        page_blk * page;
        heap_obj * new;
        void * mem;
        int sz;
        int iLockRoot;

        assert( gctable != NULL );
	lockStaticMutex( &gc_lock );
        if( (page = *list) == NULL ){
                NOTIFY_COLLECTOR();
                while( new_obj_blk( list, sizeof(heap_obj), 0 ) == 0 ){
                        unlockStaticMutex( &gc_lock );
                        outOfMemory();
                        lockStaticMutex( &gc_lock );
                }
                page = *list;
        }

        /* Get from freelist */
        new = (heap_obj*)page->freelist;
        page->freelist = (void*)new->next;
        page->used++;
        sz = PAGE_OBJSIZE( page );
	gc_current_alloc += sz;
DBG(mm, ("malloc: %db @ %x\n",sz,new));
        if( page->freelist == NULL ){
                *list = page->next;
        }

        new->gctable = gctable;
        SET_COLOR( new, COLOR_WHITE );
        INSERT_OBJ( white_list, new );
	unlockStaticMutex( &gc_lock );

        ALTER_STATS( gctable, 1, sz );
        mem = HEAP_OBJ2MEM( new );
        memset( mem, 0, sz );
        CHECK_ALIGNMENT( mem );
        return mem;
}

/*
 * Malloc a fixed object of given size
 */
void*
fixed_malloc( int sz )
{
        void * mem;

        if( sz == 0 ){
                return MALLOC_ZERO_BYTES;
        }
        mem = sz < MIN_BLOB_SIZE ? 
                fixed_malloc_obj( &free_fixed_obj[SZ2FREELIST( sz )] ) : 
                malloc_blob( sz, 1, 0 );
        return mem;
}

void*
fixed_realloc( void * mem, int sz )
{
        int old_sz;
        page_blk * header = MEM2PAGE_HDR( mem );
        void * new = sz < MIN_BLOB_SIZE ? 
                fixed_malloc_obj( &free_fixed_obj[SZ2FREELIST( sz )] ) : 
                malloc_blob( sz, 1, 0 );
        if( mem != NULL ){
                old_sz = IS_BLOB_PAGE( header ) ? 
                        MEM2BLOB(mem)->size : PAGE_OBJSIZE( header );
                memcpy( new, mem, old_sz );
                fixed_free( mem );
        }
	return new;
}

/*
 * Malloc a heap object of given size and type
 */
void*
heap_malloc( int sz, gcTable * gctable )
{
        void * mem;

        assert( gctable );
        if( sz == 0 ){
                return MALLOC_ZERO_BYTES;
        }
        mem = (sz < MIN_BLOB_SIZE) ? 
                heap_malloc_obj( &free_heap_obj[SZ2FREELIST(sz)], gctable ) : 
                malloc_blob( sz, 0, gctable );
        return mem;
}

/*
 * Free a heap object, returning it to the page's freelist
 */
static void
free_obj( page_blk * header, base_obj * obj )
{
        uint32 has_free_obj = header->freelist ? 1 : 0;
	uint32 sz = PAGE_OBJSIZE( header );

        /* Return object to head of object freelist */
DBG(mm, ("free: %db @ %x\n",PAGE_OBJSIZE(header),obj));
        header->used--;
	gc_current_alloc -= sz;
#if defined(DEBUG)
        memset( OBJ2MEM(obj), 0xf4, sz );
#endif
        obj->next = has_free_obj ? (base_obj*)header->freelist : 0;
        header->freelist = (void*)obj;
        if( !has_free_obj ){
                /* Relink page to page freelist */
                page_blk ** freelist = header->info & PAGE_FIXED_FLAG ? 
			&free_fixed_obj[SZ2FREELIST( sz )] : 
                        &free_heap_obj[SZ2FREELIST( sz )];
DBG(mm, 	("free: relink to freelist\n"));
                header->next = (*freelist);
                (*freelist) = header;
        }
}

/*
 * Free a heap blob, returning it to the page's freelist
 */
static void
free_blob( page_blk * header, base_blob * blob, uint32 hdr_sz )
{
	base_blob * cur, ** list;
	uint32 merged = 0;

DBG(mm, ("free: %db @ %x\n",blob->size,blob));
	header->used -= blob->size + hdr_sz;
	gc_current_alloc -= blob->size;
#if defined(DEBUG)
        memset( BLOB2MEM(blob), 0xf5, blob->size );
#endif
	list = (base_blob**)&header->freelist;

        /* Figure out where this blob belongs in the page block */
        if( (cur = *list) == NULL ){
                *list = (void*)blob;
                blob->next = 0;
		return;
        }
	if( (void*)list < (void*)blob && (void*)cur > (void*)blob ){
		if( (void*)blob + blob->size + hdr_sz == cur ){
DBG(mm, 		("free: join %d @ %x\n",cur->size,cur));
			blob->size += cur->size + hdr_sz;
			blob->next = cur->next;
		} 
                else {
DBG(mm, 		("free: link btwn head & %x\n",cur));
			blob->next = cur;
		}
	        *list = blob;
		return;
	}
        cur = (base_blob*)header->freelist;
	while( cur != NULL ){
		if( cur->next == NULL || ( cur < blob && cur->next > blob ) ){
			break;
                }
		cur = cur->next;
	}

        /* Try to merge the blob to adjacent blobs that are free */
	if( (void*)blob + blob->size + hdr_sz == cur->next ){
DBG(mm, 	("free: join %d @ %x\n",cur->next->size,cur->next));
		blob->size += cur->next->size + hdr_sz;
		blob->next = cur->next->next;
		cur->next = blob;
		merged = 1;
	}
	if( (void*)cur + cur->size + hdr_sz == blob ){
DBG(mm, 	("free: join %d @ %x\n",cur->size,cur));
		cur->size += blob->size + hdr_sz;
		if( merged ){
			cur->next = blob->next;
                }
		merged = 1;
	}

        /* Otherwise, just link it in order */
	if( !merged ){
DBG(mm, 	("free: link btwn %x & %x\n",cur,cur->next));
		blob->next = cur->next;
		cur->next = blob;
	}
}

/*
 * Free a fixed object
 */
void
fixed_free( void * mem )
{
        page_blk * header = MEM2PAGE_HDR( mem );
        int iLockRoot;

        if( mem == NULL || mem == MALLOC_ZERO_BYTES ){
                return;
        }
	lockStaticMutex( &gc_lock );
        assert( IS_FIXED_PAGE( header ) );
        if( header->info & PAGE_BLOB_FLAG ){
                base_blob * blob = MEM2BLOB(mem);
                ALTER_STATS( GC_ALLOC_FIXED, -1, -blob->size );
                free_blob( header, blob, sizeof(base_blob) );
        } 
        else {
                ALTER_STATS( GC_ALLOC_FIXED, -1, -PAGE_OBJSIZE(header) );
                free_obj( header, MEM2OBJ(mem) );
        }        
	unlockStaticMutex( &gc_lock );
}

/*
 * Free a heap object
 */
void
heap_free( void * mem )
{
        page_blk * header = MEM2PAGE_HDR( mem );

        if( mem == NULL || mem == MALLOC_ZERO_BYTES ){
                return;
        }
        assert( !IS_FIXED_PAGE( header ) );
        if( header->info & PAGE_BLOB_FLAG ){
                heap_blob * blob = MEM2HEAP_BLOB(mem);
                ALTER_STATS( blob->gctable, -1, -blob->size );
                blob->magic = 0;
                blob->gctable = 0;
                free_blob( header, (base_blob*)blob, sizeof(heap_blob) );
        } 
        else{
                heap_obj* obj = MEM2HEAP_OBJ( mem );
                ALTER_STATS( obj->gctable, -1, -PAGE_OBJSIZE(header) );
                obj->gctable = 0;
                free_obj( header, (base_obj*)obj );
        }
}

/*
 * Walk freelists, look for empty pages and return to page allocator.
 */
static void
free_freelist( page_blk ** list )
{
        page_blk * page;
        while( (page = *list) != NULL && n_req_pages > 0 ){
                if( page->used == 0 ){
                        page_blk * next = page->next;
                        n_req_pages -= page->n_pages;
                        page_free( page );
                        (*list) = next;
                } 
                else{
                        list = &page->next;
                }
        }
}

void
free_freelists() 
{
        int n;
        int iLockRoot;

        lockStaticMutex( &gc_lock );
        free_freelist( &free_fixed_blob );
        free_freelist( &free_heap_blob );
        for( n = SZ2FREELIST( MIN_BLOB_SIZE ) + 1; --n >= 0; ){
                free_freelist( &free_fixed_obj[n] );
                free_freelist( &free_heap_obj[n] );
        }
        unlockStaticMutex( &gc_lock );
}

void
mm_init() 
{
        gc_pgsize = page_size = getpagesize();
        gc_heap_limit =  1.5*Kaffe_JavaVMArgs[0].maxHeapSize;
        gc_heap_allocation_size = Kaffe_JavaVMArgs[0].maxHeapSize;
        gc_heap_total = 0;
	gc_current_alloc = 0;

DBG(mm, ("page size = %x (%db)\n",page_size,page_size));
        last_page = &free_pages;

        gc_heap_min = (uintp)pagealloc( 0 );

        memset( free_fixed_obj, 0, (SZ2FREELIST(MIN_BLOB_SIZE)+1) * sizeof(page_blk*) );
        memset( free_heap_obj, 0, (SZ2FREELIST(MIN_BLOB_SIZE)+1) * sizeof(page_blk*) );
        memset( &free_pages, 0, sizeof(page_blk));

#if defined(GC_EXP)
        white_val = 0x2;
        black_val = 0x0;
        white_list.next = 0;
        white_list.prev = 0;
        free_list.next = 0;
        free_list.prev = 0;
        finalize_list.next = 0;
        finalize_list.prev = 0;
        finalize_class_list.next = 0;
        finalize_class_list.prev = 0;
        grey_list = 0;
        black_list.next = 0;
        black_list.prev = 0;
#else
        white_list = 0;
        finalize_list = 0;
        finalize_class_list = 0;
#endif
}

#if defined(DEBUG)
void
print_freelist(page_blk* page) {
        base_obj* obj = (base_obj*)page->freelist;
        printf ("%x ",(uint)page->freelist);
        while( obj != NULL ){
                obj = obj->next;
                printf ("%x ",(uint)obj);
                if( (void*)obj >= (void*)page + page_size ){
                        break;
                }
        }
        printf ("\n");
}

void
verify_freelist(void) {
        page_blk ** list = free_heap_obj;
        page_blk * page;
        base_obj * obj;
        page_blk ** limit = &free_heap_obj[SZ2FREELIST( MIN_BLOB_SIZE )];
        for( ; list < limit; list++ ){
                if( *list != NULL ){
                        page = (*list);
                        obj = (base_obj*)page->freelist;
                        while( obj != NULL ){
                                obj = obj->next;
                                if( (void*)obj >= ((void*)page + page_size) ){
                                        printf ("PROBLEM!\n");
                                        break;
                                }
                        }
                }
        }
}
#endif
