/*
 * gc2.c
 * This interfaces the VM to the Hans-J. Boehm Incremental Garbage
 * Collector (version 6.0 alpha 7).
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "gc2.h"
#include "access.h"
#include "classMethod.h"
#include "locks.h"

extern void Kaffe_CallVoidMethodA(JNIEnv*, jobject, jmethodID, jvalue*);

typedef void* (*gcfunc)(size_t);

iLock* finman1;
static boolean finalRunning;

static
void*
GC_malloc_atomic_and_clear(size_t sz)
{
	void* mem;
	mem = GC_malloc_atomic(sz);
	if (mem != 0) {
		memset(mem, 0, sz);
	}
	return (mem);
}

static
void*
GC_malloc_atomic_uncollectable_and_clear(size_t sz)
{
	void* mem;
	mem = GC_malloc_atomic_uncollectable(sz);
	if (mem != 0) {
		memset(mem, 0, sz);
	}
	return (mem);
}

static gcfunc calls[] = {
	/* GC_ALLOC_NORMAL         */ GC_malloc,
	/* GC_ALLOC_ESTACK         */ GC_malloc_atomic_and_clear,
	/* GC_ALLOC_FIXED          */ GC_malloc_atomic_uncollectable_and_clear, 
	/* GC_ALLOC_NORMALOBJECT   */ GC_malloc,
	/* GC_ALLOC_PRIMARRAY      */ GC_malloc_atomic_and_clear,
	/* GC_ALLOC_REFARRAY       */ GC_malloc,
	/* GC_ALLOC_CLASSOBJECT    */ GC_malloc,
	/* GC_ALLOC_THREAD         */ GC_malloc,
	/* GC_ALLOC_METHOD         */ GC_malloc,
	/* GC_ALLOC_FIELD          */ GC_malloc,
	/* GC_ALLOC_STATICDATA     */ GC_malloc,
	/* GC_ALLOC_DISPATCHTABLE  */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_BYTECODE       */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_EXCEPTIONTABLE */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_CONSTANT       */ GC_malloc,
	/* GC_ALLOC_UTF8CONST      */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_INTERFACE      */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_JITCODE        */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_LOCK           */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_THREADCTX      */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_REF            */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_CLASSPOOL      */ GC_malloc,
	/* GC_ALLOC_JREFERENCE     */ GC_malloc,
	/* GC_ALLOC_CLASSLOADER    */ GC_malloc,
	/* GC_ALLOC_ANALYSE        */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_INFLATE        */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_STRINGS        */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_ENGINE         */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_JAR            */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_AWT            */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_SIGNATURE      */ GC_malloc_atomic_uncollectable_and_clear,
	/* GC_ALLOC_NORMALOBJECT_NOFINALIZE */ GC_malloc,
};

static
void
finalizeObject(void* ob, void* ignore)
{
	Hjava_lang_Object* obj;
	Method* final;

	obj = (Hjava_lang_Object*)ob;
	/* If this object has no dtable then it's been completely created yet.
	 * which is worrying since we're finalizing it!!!
	 */
	assert(obj->dtable != 0);
	final = OBJECT_CLASS(obj)->finalizer;
	if (final != 0) {
		/* Use JNI to make finalizer call to protect GC against exceptions */
		Kaffe_CallVoidMethodA(0, (jobject)ob, (jmethodID)final, 0);
		Kaffe_ExceptionClear(0);
	}
}

static
void
finalizeClass(void* obj, void* ignore)
{
	Hjava_lang_Class* class;

	class = (Hjava_lang_Class*)obj;

	if (Kaffe_JavaVMArgs[0].enableVerboseGC > 0) {
		kprintf("<GC: Unloading %s>\n", CLASS_CNAME(class));
	}
	
	/* Unload class */
	unloadClass(class);
}

void
finaliserMan(void* dummy)
{
	for (;;) {
		for (finalRunning = false; !finalRunning;) {
			waitStaticCond(&finman1, NOTIMEOUT);
		}
		GC_invoke_finalizers();
	}
}

static
void
signalFinalizer(void)
{
	int iLockRoot;

	lockStaticMutex(&finman1);
	if (finalRunning == false) {
		finalRunning = true;
		signalStaticCond(&finman1);
	}
	unlockStaticMutex(&finman1);
}

void*
__kaffe_malloc(int sz)
{
	return GC_malloc_atomic_uncollectable(sz);
}

void*
__kaffe_realloc(void* mem, int sz)
{
	return GC_REALLOC(mem, sz);
}

void
__kaffe_free(void* mem)
{
	GC_FREE(mem);
}

void*
__kaffe_awt_malloc(size_t sz)
{
	return GC_malloc_atomic_uncollectable(sz);
}

void
__kaffe_awt_free(void* mem)
{
	GC_FREE(mem);
}

void*
gc_malloc(size_t sz, int type)
{
	static int first = 1;
	void* mem;

	if (first) {
		first = 0;
		GC_finalizer_notifier = signalFinalizer;
		GC_java_finalization = 1;
		GC_finalize_on_demand = 1;
	}

	// Allocate memory
	mem = (*calls[type])(sz);

	// Attach finalizer
	if (mem != 0) {
		switch (type) {
		case GC_ALLOC_NORMALOBJECT:
		case GC_ALLOC_THREAD:
		case GC_ALLOC_JREFERENCE:
			GC_REGISTER_FINALIZER_NO_ORDER(mem, finalizeObject, 0, 0, 0);
			break;
		case GC_ALLOC_CLASSOBJECT:
			GC_REGISTER_FINALIZER_NO_ORDER(mem, finalizeClass, 0, 0, 0);
			break;
		}
	}
	return (mem);
}

void*
gc_malloc_fixed(size_t sz)
{
	return (*calls[GC_ALLOC_FIXED])(sz);
}

void*
gc_realloc_fixed(void* mem, size_t sz)
{
	if (mem == 0) {
		return GC_malloc_atomic_uncollectable(sz);
	}
	else {
		return GC_REALLOC(mem, sz);
	}
}

void*
gc_add_ref(void* mem)
{
	void** ref;

	ref = (void**)GC_malloc_uncollectable(sizeof(void*));
	if (ref != 0) {
		*ref = mem;
	}
	return (ref);
}

void
gc_rm_ref(void* ref)
{
	GC_FREE(ref);
}
