/*
 * Handle create and subsequent garbage collection of objects.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: objects.c,v 1.18 2001/11/10 04:12:16 samc Exp $
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "lookup.h"
#include "itypes.h"
#include "baseClasses.h"
#include "exception.h"
#include "itypes.h"
#include "md.h"
#include "external.h"
#include "gc.h"
#include "locks.h"
#include "jdebug.h"
#include "precompile.h"
#include "engine.h"

static iLock* traceLock;

#ifdef DEBUG
static int dontTrace = 0;

void
dumpObjectStackTrace()
{
#if defined(FP_OFFSET)
	int iLockRoot;

	Hjava_lang_Throwable* eobj;

       	lockStaticMutex(&traceLock);

	dontTrace = 1;

	eobj = (Hjava_lang_Throwable*)Kaffe_AllocObject(0, Kaffe_FindClass(0, "java.lang.Throwable"));
	unhand(eobj)->backtrace = (Hjava_lang_Object*)buildStackTraceGC();
        GC_WRITE( eobj, unhand(eobj)->backtrace );
	printStackTrace(eobj, 0, 1);

	dontTrace = 0;

       	unlockStaticMutex(&traceLock);


#endif
}
#endif

/*
 * Create a new Java object.
 * If the object is a Class then we make it a root (for the moment).
 * Otherwise, if the object has a finalize method we look it up and store
 * it with the object for it's later use.  Otherwise the object is normal.
 */
Hjava_lang_Object*
newObject(Hjava_lang_Class* class)
{
	Hjava_lang_Object* obj;
#if defined(MEM2)
	gcTable* type = class->gc_type;
        page_blk** freelist = class->freelist;
	if (freelist) {
                obj = gc_malloc_opt(0, freelist, type);
        }
	else {
                obj = gc_malloc(CLASS_FSIZE(class), type);
        }
#else
        obj = gc_malloc(CLASS_FSIZE(class), class->flags & FLAG_TYPE);
#endif

	DBG_ACTION( vm_trace_new_object, {
//        if (!dontTrace)
	    DBG( vm_trace_new_object, ("0x%x: %s\n", obj, (class ? class->name: "<none>")));
	});

        /* Fill in object information */
	if (obj == 0) {
		postOutOfMemory();
	}
	else {
		obj->dtable = class->dtable;
		GC_WRITE(obj, obj->dtable); /* object->dtable */
	}
 
	DBG_ACTION( vm_trace_new_object_stack, {
	  if (!dontTrace) 
		  dumpObjectStackTrace();
	});
       return (obj);
}

/*
 * Allocate a new class object.
 */
Hjava_lang_Class*
newClass(void)
{
	Hjava_lang_Class* cls;

	cls = gc_malloc(sizeof(Hjava_lang_Class), GC_ALLOC_CLASSOBJECT);

	if (cls == 0) {
		postOutOfMemory();
	}
	else {
		/* Fill in object information. */
		cls->head.dtable = (dispatchTable*)&KaffeDispatchTable_java_lang_Class;
		GC_WRITE(cls, cls->head.dtable); /* object->dtable */

// CDBG(		kprintf("newClass %x\n", cls);				)
	}

        return (cls);
}

/*
 * Allocate a new array, of whatever types.
 */
Hjava_lang_Object*
newArray(Hjava_lang_Class* elclass, int count)
{
	Hjava_lang_Object* obj;
	Hjava_lang_Class* class;

	class = lookupArrayProcess(elclass, CSTATE_COMPLETE);
	if (class == 0) {
		return (0);
	}
	if (elclass == &_Jv_intClass || elclass == &_Jv_floatClass) {
		obj = gc_malloc((TYPE_PRIM_SIZE(elclass) * count) + ARRAY_DATA_OFFSET(HArrayOfInt), GC_ALLOC_PRIMARRAY);
	}
	else if (elclass == &_Jv_longClass || elclass == &_Jv_doubleClass) {
		obj = gc_malloc((TYPE_PRIM_SIZE(elclass) * count) + ARRAY_DATA_OFFSET(HArrayOfLong), GC_ALLOC_PRIMARRAY);
	}
	else if (elclass == &_Jv_booleanClass || elclass == &_Jv_byteClass) {
		obj = gc_malloc((TYPE_PRIM_SIZE(elclass) * count) + ARRAY_DATA_OFFSET(HArrayOfByte), GC_ALLOC_PRIMARRAY);
	}
	else if (elclass == &_Jv_charClass || elclass == &_Jv_shortClass) {
		obj = gc_malloc((TYPE_PRIM_SIZE(elclass) * count) + ARRAY_DATA_OFFSET(HArrayOfShort), GC_ALLOC_PRIMARRAY);
	}
	else {
		obj = gc_malloc((PTR_TYPE_SIZE * count) + ARRAY_DATA_OFFSET(HArrayOfObject), GC_ALLOC_REFARRAY);
	}
// ADBG(	kprintf("0x%x: newArray %s\n", obj, elclass->name->data);	)
	if (obj == 0) {
		postOutOfMemory();
	}
	else {
		obj->dtable = class->dtable;
		GC_WRITE(obj, obj->dtable); /* object->dtable */
		ARRAY_SIZE(obj) = count;
	}
	return (obj);
}

/*
 * Allocate a new multi-dimensional array.
 */
Hjava_lang_Object*
newMultiArray(Hjava_lang_Class* clazz, int* dims)
{
	Hjava_lang_Object* obj;
	Hjava_lang_Object** array;
	int i;
	
	obj = newArray(CLASS_ELEMENT_TYPE(clazz), dims[0]);
	if (obj == 0) {
		return (0);
	}
	if (dims[1] >= 0) {
		array = ARRAY_DATA(HArrayOfObject, obj);
		for (i = 0; i < dims[0]; i++) {
			array[i] = newMultiArray(CLASS_ELEMENT_TYPE(clazz), &dims[1]);
                        GC_WRITE(obj, array[i]); /* object->data[i] */
			if (array[i] == 0) {
				return (0);
			}
		}
	}
	return (obj);
}
