/*
 * soft.c
 * Soft instruction support.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	MDBG(s)
#define	ADBG(s)
#define	CDBG(s)
#define	IDBG(s)
#define	TDBG(s)
#define	SOFTDBG(s)


#if MDBG(1) - 1 == 0
#undef CDBG
#define	CDBG(s) s
#endif

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include <stdarg.h>
#include "gtypes.h"
#include "bytecode.h"
#include "slots.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "support.h"
#include "baseClasses.h"
#include "lookup.h"
#include "exception.h"
#include "locks.h"
#include "soft.h"
#include "external.h"
#include "threads.h"
#include "baseClasses.h"
#include "patch.h"
#include "itypes.h"
#include "jni.h"
#include "md.h"
#include "fp.h"

#if defined(SMX)
#define inline
#endif

#undef	SOFTDBG
#define	SOFTDBG(s)

/*
 * Define various float/double constants.
 */

static jvalue finfp = { 0x7f800000 };
static jvalue finfm = { 0xff800000 };
static jvalue fnan  = { 0x7fc00000 };
#if defined(DOUBLE_WORDS_BIG_ENDIAN) && !defined(WORDS_BIGENDIAN) && 0
static jvalue dinfp = { 0x000000007ff00000 };
static jvalue dinfm = { 0x00000000fff00000 };
static jvalue dnan  = { 0x000000007ff80000 };
#else
static jvalue dinfp = { (uint64)0x7ff00000 << 32 };
static jvalue dinfm = { (uint64)0xfff00000 << 32 };
static jvalue dnan  = { (uint64)0x7ff80000 << 32 };
#endif

/*
 * Pointer to stack overflow exception is used by precompiled
 */
void* soft_stackoverflow_ptr = soft_stackoverflow;

/*
 * _Jv_AllocObject
 */
void*
_Jv_AllocObject(Hjava_lang_Class* c)
{
	void* ret;

SOFTDBG(kprintf("New object of type %s (%d)\n", c->name->data, CLASS_FSIZE(c));)

	if (processClass(c, CSTATE_COMPLETE) == 0) {
		throwPostedException();
		ret = 0;
	}
	else {
		ret = newObject(c);
		if (ret == 0) {
			throwPostedException();
		}
	}
	return (ret);
}

/*
 * _Jv_NewArray
 */
void*
_Jv_NewArray(jint type, jint size)
{
	void* ret;

SOFTDBG(kprintf("New object of %d type (%d)\n", type, size); )

	if (size < 0) {
		postAndThrowException("java.lang.NegativeArraySizeException", "");
	}
	ret = newArray(TYPE_CLASS(type), size);
	if (ret == 0) {
		throwPostedException();
	}
	return (ret);
}

/*
 * _Jv_NewObjectArray
 */
void*
_Jv_NewObjectArray(jint size, Hjava_lang_Class* elclass)
{
	void* ret;

SOFTDBG( kprintf("soft_anewarray %x %s %d\n", elclass, elclass->name->data, size); )

	if (size < 0) {
		postAndThrowException("java.lang.NegativeArraySizeException", "");
	}
	ret = newArray(elclass, size);
	if (ret == 0) {
		throwPostedException();
	}
	return (ret);
}

/*
 * soft_multianewarray.
 */
void*
soft_multianewarray(Hjava_lang_Class* class, jint dims, slots* args)
{
	Hjava_lang_Object* obj;
	jint arg;
	int i;
	int* arraydims;

SOFTDBG( kprintf("soft_multianewarray\n");	)

	arraydims = alloca((dims+1) * sizeof(int));

	/* Extract the dimensions into an array */
	for (i = 0; i < dims; i++) {
		arg = args[-i].v.tint;
		if (arg < 0) {
			postException("java.lang.NegativeArraySizeException", "");
			obj = 0;
			goto error;
		}
		arraydims[dims-i-1] = arg;
	}
	arraydims[i] = -1;

	/* Mmm, okay now build the array using the wonders of recursion */
        obj = newMultiArray(class, arraydims);

	error:;

	if (obj == 0) {
		throwPostedException();
	}
	/* Return the base object */
	return (obj);
}

/*
 * soft_lookupmethod.
 */
void*
soft_lookupmethod(Hjava_lang_Object* obj, Hjava_lang_Class* iface, int idx)
{
	Hjava_lang_Class* cls;
	Hjava_lang_Class* icls;
	int i;
        dispatchTable * dtable;

	if (obj == 0) {
		soft_nullpointer();
	}
	cls = OBJECT_CLASS(obj);
        dtable = obj->dtable;
	idx += dtable->vsize;
	for (i = 0; cls->total_interface_len; i++) {
		icls = cls->interfaces[i];
		if (icls == iface) {
#ifdef DIRECT_CALL
			return dtable->method[idx];
#else
			return (*dtable->method[idx]);
#endif
		}
		idx += CLASS_NMETHODS(icls);
	}

	postAndThrowException("java.lang.ClassFormatError", "interface not found");
	return (0);
}

void*
soft_lazylookupmethod(Hjava_lang_Object* obj, char* methodname, char* methodsig)
{
	Method* meth;

	if (obj == 0) {
		soft_nullpointer();
	}

	meth = findMethod(OBJECT_CLASS(obj), methodname, methodsig);

	if (meth == 0) {
		soft_nosuchmethod(CLASS_CNAME(OBJECT_CLASS(obj)), methodname, methodsig);
	}

	return (meth->ncode);
}

inline
jint
instanceof_class(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	for (; oc != 0; oc = oc->superclass) {
		if (c == oc) {
			return (1);
		}
	}
	return (0);
}

inline
jint
instanceof_interface(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	int i;

	/* Check 'total' interface list */
	for (i = oc->total_interface_len - 1; i >= 0; i--) {
		if (c == oc->interfaces[i]) {
			return (1);
		}
	}
	return (0);
}

inline
jint
instanceof_array(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	/* Skip as many arrays of arrays as we can.  We stop when we find
	 * a base class in either. */
	while (CLASS_IS_ARRAY(c) && CLASS_IS_ARRAY(oc)) {
		c = CLASS_ELEMENT_TYPE(c);
		oc = CLASS_ELEMENT_TYPE(oc);
	}

	/* If we are still casting to an array then we have failed already */
	if (CLASS_IS_ARRAY(c))
		return (0);

	/* If a base type, they must match exact. */
	if (CLASS_IS_PRIMITIVE(c)) {
		return (c == oc);
	}

	/* Casting to an object of some description. */
	if (CLASS_IS_ARRAY(oc)) {
		/* The only thing we can cast an array to is java.lang.Object.
		 * Checking this here willl save time.
		 */
		return (c == ObjectClass);
	}

	/* Cannot cast to a primitive class. */
	if (CLASS_IS_PRIMITIVE(oc)) {
		return (0);
	}

	/* Casting one object to another.  */
	return (instanceof(c, oc));
}

jint
instanceof(Hjava_lang_Class* c, Hjava_lang_Class* oc)
{
	/* Handle the simplest case first - they are the same */
	if (c == oc) {
		return (1);
	}

	if (processClass(c, CSTATE_LINKED) == 0 || processClass(oc, CSTATE_LINKED) == 0) {
		throwPostedException();
	}

	/* Else if an array check that */
	if (CLASS_IS_ARRAY(c)) {
		return (instanceof_array(c, oc));
	}
	/* Else if an interface check that */
	else if (CLASS_IS_INTERFACE(c)) {
		return (instanceof_interface(c, oc));
	}
	/* Else is must to a class */
	else {
		return (instanceof_class(c, oc));
	}
}

/*
 * _Jv_IsInstanceOf.
 */
jint
_Jv_IsInstanceOf(Hjava_lang_Object* o, Hjava_lang_Class* c)
{
SOFTDBG( kprintf("_Jv_IsInstanceOf %s %s\n", CLASS_CNAME(c), CLASS_CNAME(OBJECT_CLASS(o))); )
	/* Null object are never instances of anything */
	if (o == 0) {
		return (0);
	}

	return (instanceof(c, OBJECT_CLASS(o)));
}

/*
 * _Jv_CheckCast.
 */
void*
_Jv_CheckCast(Hjava_lang_Class* c, Hjava_lang_Object* o)
{
SOFTDBG( kprintf("_Jv_CheckCast\n"); )

	if (o != 0 && !instanceof(c, OBJECT_CLASS(o))) {
		postAndThrowException("java.lang.ClassCastException", CLASS_CNAME(c));
	}
	return (o);
}

jint
soft_lazyinstanceof(Hjava_lang_Object* o, Hjava_lang_ClassLoader* loader, char* c)
{
	Hjava_lang_Class* cls;

	cls = loadClass(c, loader, CSTATE_UNLOADED);
	if (cls == 0) {
		soft_nosuchclass(c);
	}

	return (_Jv_IsInstanceOf(o, cls));
}

void*
soft_lazycheckcast(char* c, Hjava_lang_ClassLoader* loader, Hjava_lang_Object* o)
{
	Hjava_lang_Class* cls;

	cls = loadClass(c, loader, CSTATE_UNLOADED);
	if (cls == 0) {
		soft_nosuchclass(c);
	}

	return (_Jv_CheckCast(cls, o));
}

/*
 * soft_athrow.
 */
void
soft_athrow(Hjava_lang_Throwable* o)
{
SOFTDBG( kprintf("soft_athrow\n"); )

	if (o == 0) {
		soft_nullpointer();
	}
	else {
		postAndThrowExternalException(o);
	}
}

/*
 * soft_badarrayindex.
 */
void
soft_badarrayindex(void)
{
SOFTDBG( kprintf("soft_badarrayindex\n"); )
	postAndThrowException("java.lang.ArrayIndexOutOfBoundsException", "");
}

/*
 * soft_stackoverflow.
 */
void
soft_stackoverflow(void)
{
	postAndThrowException("java.lang.StackOverflowError", "");
}

/*
 * soft_nullpointer.
 */
void
soft_nullpointer(void)
{
SOFTDBG( kprintf("soft_nullpointer\n"); )
	postAndThrowException("java.lang.NullPointerException", "");
}

/*
 * soft_divzero.
 */
void
soft_divzero(void)
{
SOFTDBG( kprintf("soft_divzero\n"); )
	postAndThrowException("java.lang.ArithmeticException", "");
}

/*
 * soft_nosuchclass.
 */
void            
soft_nosuchclass(char* name)
{
SOFTDBG( kprintf("soft_nosuchclass\n"); )
	if (name == 0) {
		name = "";
	}
	postAndThrowException("java.lang.NoClassDefFoundError", name);
}

/*
 * soft_nosuchmethod.
 */
void            
soft_nosuchmethod(char* c, char* n, char* s)
{
	char buf[100];

SOFTDBG( kprintf("soft_nosuchmethod\n"); )

	strncpy(buf, c, sizeof(buf));
	strncat(buf, ".", sizeof(buf));
	strncat(buf, n, sizeof(buf));
	strncat(buf, s, sizeof(buf));

	postAndThrowException("java.lang.NoSuchMethodError", buf);
}

/*
 * soft_nosuchfield.
 */
void            
soft_nosuchfield(char* c, char* n)
{
	char buf[100];

SOFTDBG( kprintf("soft_nosuchfield"); )

	if (c != 0) {
		strncpy(buf, c, sizeof(buf));
		strncat(buf, ".", sizeof(buf));
	}
	else {
		strncpy(buf, "<unknown class>", sizeof(buf));
	}
	strncat(buf, n, sizeof(buf));

	postAndThrowException("java.lang.NoSuchFieldError", buf);
}

/*
 * _Jv_InitClass.
 */
void
_Jv_InitClass(Hjava_lang_Class* c)
{
SOFTDBG( kprintf("soft_init_class %x %s\n", c, c->name->data); )

	if (processClass(c, CSTATE_COMPLETE) == 0) {
		throwPostedException();
	}
#ifdef DIRECT_CALL
        finishClassPatches( c );
#endif
}

/*
 * Check we can store 'obj' into the 'array'.
 */
void
soft_checkarraystore(Hjava_lang_Object* array, Hjava_lang_Object* obj)
{
SOFTDBG( kprintf("soft_checkarraystore\n"); )
	if (obj != 0 && _Jv_IsInstanceOf(obj, CLASS_ELEMENT_TYPE(OBJECT_CLASS(array))) == 0) {
		postAndThrowException("java.lang.ArrayStoreException", "");
	}
}

/*
 * Return the exception object.
 */
void*
soft_get_thrown_exception(void)
{
	return (getPostedException());
}

/*
 * soft_dcmpg
 */
jint
soft_dcmpg(jdouble v1, jdouble v2)
{
	jint ret;
	if ((!isinf(v1) && DISNAN(doubleToLong(v1))) || (!isinf(v2) && DISNAN(doubleToLong(v2)))) {
		ret = 1;
	}
	else if (v1 > v2) {
		ret = 1;
	}
	else if (v1 == v2) {
		ret = 0;
	}
	else {
		ret = -1;
	}

	return (ret);
}

/*
 * soft_dcmpl
 */
jint
soft_dcmpl(jdouble v1, jdouble v2)
{
        jint ret;
	if ((!isinf(v1) && DISNAN(doubleToLong(v1))) || (!isinf(v2) && DISNAN(doubleToLong(v2)))) {
		ret = -1;
	}
        else if (v1 > v2) {
                ret = 1;
        }
        else if (v1 == v2) {
                ret = 0;
        }
        else {
                ret = -1;
        }
	return (ret);
}

/*
 * soft_fcmpg
 */
jint
soft_fcmpg(jfloat v1, jfloat v2)
{
        jint ret;
	if ((!isinf(v1) && FISNAN(floatToInt(v1))) || (!isinf(v2) && FISNAN(floatToInt(v2)))) {
		ret = 1;
	}
        else if (v1 > v2) {
                ret = 1;
        }
        else if (v1 == v2) {
                ret = 0;
        }
        else {
                ret = -1;
        }
	return (ret);
}

/*
 * soft_fcmpg
 */
jint
soft_fcmpl(jfloat v1, jfloat v2)
{
        jint ret;
	if ((!isinf(v1) && FISNAN(floatToInt(v1))) || (!isinf(v2) && FISNAN(floatToInt(v2)))) {
		ret = -1;
	}
        else if (v1 > v2) {
                ret = 1;
        }
        else if (v1 == v2) {
                ret = 0;
        }
        else {
                ret = -1;
        }
	return (ret);
}

jlong
soft_lmul(jlong v1, jlong v2)
{
	return (v1 * v2);
}

jlong
soft_ldiv(jlong v1, jlong v2)
{
	return (v1 / v2);
}

jlong
soft_lrem(jlong v1, jlong v2)
{
	return (v1 % v2);
}

jfloat
soft_fadd(jfloat v1, jfloat v2)
{
	return (v1 + v2);
}

jfloat
soft_fsub(jfloat v1, jfloat v2)
{
	return (v1 - v2);
}

jfloat
soft_fmul(jfloat v1, jfloat v2)
{
	return (v1 * v2);
}

jfloat
soft_fdiv(jfloat v1, jfloat v2)
{
	if (v2 == 0.0) {
		if (FISNAN(floatToInt(v1)) || v1 == 0.0) {
			return (fnan.f);
		}
		else if ((v1 > 0.0) ^ FISPOS(floatToInt(v2))) {
			return (finfm.f);
		}
		else {
			return (finfp.f);
		}
	}
	return (v1 / v2);
}

jfloat
soft_frem(jfloat v1, jfloat v2)
{
	return (javaRemainderf(v1, v2));
}

jdouble
soft_faddl(jdouble v1, jdouble v2)
{
	return (v1 + v2);
}

jdouble
soft_fsubl(jdouble v1, jdouble v2)
{
	return (v1 - v2);
}

jdouble
soft_fmull(jdouble v1, jdouble v2)
{
	return (v1 * v2);
}

jdouble
soft_fdivl(jdouble v1, jdouble v2)
{
	if (v2 == 0.0) {
		if (DISNAN(doubleToLong(v1)) || v1 == 0.0) {
			return (dnan.d);
		}
		else if ((v1 > 0.0) ^ DISPOS(doubleToLong(v2))) {
			return (dinfm.d);
		}
		else {
			return (dinfp.d);
		}
	}
	return (v1 / v2);
}

jdouble
soft_freml(jdouble v1, jdouble v2)
{
	return (javaRemainder(v1, v2));
}

jlong
soft_lshll(jlong v1, jint v2)
{
	return (v1 << (v2 & 63));
}

jlong
soft_ashrl(jlong v1, jint v2)
{
	return (v1 >> (v2 & 63));
}

jlong
soft_lshrl(jlong v1, jint v2)
{
	return (((uint64)v1) >> (v2 & 63));
}

jint
soft_lcmp(jlong v1, jlong v2)
{
	if (v2 < v1) {
		return (-1);
	}
	else if (v2 > v1) {
		return (1);
	}
	else {
		return (0);
	}
}

jint
soft_mul(jint v1, jint v2)
{
	return (v1*v2);
}

jint
soft_div(jint v1, jint v2)
{
	return (v1/v2);
}

jint
soft_rem(jint v1, jint v2)
{
	return (v1%v2);
}

jfloat
soft_cvtlf(jlong v)
{
	return ((jfloat)v);
}

jfloat
soft_cvtif(jint v)
{
	return ((jfloat)v);
}

jdouble
soft_cvtid(jint v)
{
	return ((jdouble)v);
}

jdouble
soft_cvtld(jlong v)
{
	return ((jdouble)v);
}

jdouble
soft_cvtfd(jfloat v)
{
	return ((jdouble)v);
}

jfloat
soft_cvtdf(jdouble v)
{
	return ((jfloat)v);
}

/*
 * The following functions round the float/double to an int/long.
 * They round the value toward zero.
 */

jlong
soft_cvtfl(jfloat v)
{
	if (v < 0.0) {
		return ((jlong)ceil(v));
	}
	else {
		return ((jlong)floor(v));
	}
}

jlong
soft_cvtdl(jdouble v)
{
	if (v < 0.0) {
		return ((jlong)ceil(v));
	}
	else {
		return ((jlong)floor(v));
	}
}

jint
soft_cvtfi(jfloat v)
{
	if (v < 0.0) {
		return ((jint)ceil(v));
	}
	else {
		return ((jint)floor(v));
	}
}

jint
soft_cvtdi(jdouble v)
{
	if (v < 0.0) {
		return ((jint)ceil(v));
	}
	else {
		return ((jint)floor(v));
	}
}

jlong
doubleToLong(jdouble val)
{
	jvalue v;
	v.d = val;
#if defined(DOUBLE_WORDS_BIG_ENDIAN) && !defined(WORDS_BIGENDIAN)
	{
		jsize t;
		t = v.w.f;
		v.w.f = v.w.s;
		v.w.s = t;
	}
#endif
	return (v.j);
}

jdouble
longToDouble(jlong val)
{
	jvalue v;
	if (DISNAN(val)) {
		return (dnan.d);
	}
	v.j = val;
#if defined(DOUBLE_WORDS_BIG_ENDIAN) && !defined(WORDS_BIGENDIAN)
	{
		jsize t;
		t = v.w.f;
		v.w.f = v.w.s;
		v.w.s = t;
	}
#endif
	return (v.d);
}

jint
floatToInt(jfloat val)
{
	jvalue v;
	v.f = val;
	return (v.i);
}

jfloat
intToFloat(jint val)
{
	jvalue v;
	if (FISNAN(val)) {
		return (fnan.f);
	}
	else {
		v.i = val;
		return (v.f);
	}
}

/*
 * soft_init_precompile
 */
jint
soft_init_precompiled(Hjava_lang_Class* c)
{
	if (processClass(c, CSTATE_COMPLETE) == 0) {
		throwPostedException();
		return (0);
	}
	else {
		return (1);
	}
}

void
soft_debug1(void* a0, void* a1, void* a2)
{
}

void
soft_debug2(void* a0, void* a1, void* a2)
{
}

void
soft_trace(Method* meth, Hjava_lang_Class* clazz, void* args)
{
        kprintf("C %s.%s%s @ %p\n", CLASS_CNAME(clazz), meth->name, meth->signature, METHOD_NATIVECODE(meth));
#if defined(TRACE_METHOD_ARGS)
	if (args != 0) {
		TRACE_METHOD_ARGS();
	}
#endif
}

void
soft_end( uint8 type, void * rv )
{
#if defined(TRACE_METHOD_END)
        TRACE_METHOD_END();
#endif
}
