/*
 * string.c
 * Manage Java and char* (utf) strings
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "classMethod.h"
#include "jtypes.h"
#include "constants.h"
#include "objects.h"
#include "itypes.h"
#include "locks.h"
#include "baseClasses.h"
#include "strings.h"
#include "support.h"
#include "bag.h"

static unsigned int utfHashcode(void* key);
static int utfEquals(void* hashdata, void* key);

static bag utfbag;

static unsigned int stringHashcode(void* key);
static int stringEquals(void* hashdata, void* key);

#if !defined(MEM2)
static bag _stringbag;
#endif
static bag * stringbag;

static iLock* utflock;
static iLock* stringlock;

void
initStrings(void)
{
	bagStart(&utfbag, 128, 75, utfHashcode, utfEquals, GC_ALLOC_STRINGS);
#if defined(MEM2)
        stringbag = gc_malloc(sizeof(bag), GC_ALLOC_BAG);
        gc_add_ref( stringbag, stringbag+1 );
#else
        stringbag = &_stringbag;
#endif
	bagStart(stringbag, 128, 75, stringHashcode, stringEquals, GC_ALLOC_STRINGS);
}

static
unsigned int
stringHashcode(void* key)
{
	Hjava_lang_String* str;

	str = (Hjava_lang_String*)key;
	if (STRING_HASH(str) == 0) {
		int code = 0;
		int i;
		for (i = 0; i < STRING_SIZE(str); i++) {
			code = code * 31 + STRING_DATA(str)[i];
		}
		STRING_HASH(str) = code;
	}
	return (STRING_HASH(str));
}

static
int
stringEquals(void* hashdata, void* key)
{
	Hjava_lang_String* stra;
	Hjava_lang_String* strb;

	stra = (Hjava_lang_String*)hashdata;
	strb = (Hjava_lang_String*)key;

	if (STRING_SIZE(stra) != STRING_SIZE(strb) || STRING_HASH(stra) != STRING_HASH(strb)) {
		return (1);
	}
	else {
		return (memcmp(STRING_DATA(stra), STRING_DATA(strb), STRING_SIZE(stra)));
	}
}

Hjava_lang_String*
internJavaString(Hjava_lang_String* str)
{
	Hjava_lang_String* nstr;
	int iLockRoot;

	lockStaticMutex(&stringlock);

	if (unhand(str)->interned == 0) {
		nstr = (Hjava_lang_String*)bagGet(stringbag, str);
		while (nstr == 0) {
			nstr = (Hjava_lang_String*)bagPut(stringbag, str, str);
			if (nstr != 0) {
				GC_WRITE(stringbag, str);
				unhand(nstr)->interned = 1;
			}
			else {
				unlockStaticMutex(&stringlock);
				bagRehash(stringbag);
				lockStaticMutex(&stringlock);
			}
		}
	}
	else {
		nstr = str;
	}

	unlockStaticMutex(&stringlock);

	return (nstr);
}

void
uninternJavaString(Hjava_lang_String* str)
{
	int iLockRoot;

	lockStaticMutex(&stringlock);

	if (unhand(str)->interned == 1) {
		bagRemove(stringbag, str);
	}

	unlockStaticMutex(&stringlock);
}

static
unsigned int
utfHashcode(void* key)
{
	unsigned char* ptr;
	unsigned int code;
	
	code = 0;
	for (ptr = (unsigned char*)key; ptr[0] != 0; ptr++) {
		code = code * 31 + *ptr;
	}
	return (code);
}

static
int
utfEquals(void* hashdata, void* key)
{
	return (strcmp((char*)hashdata, (char*)key));
}

char*
addUtf(char* str)
{
	int iLockRoot;

	lockStaticMutex(&utflock);

	for (;;) {
		str = bagPut(&utfbag, str, str);
		if (str != 0) {
			break;
		}
		unlockStaticMutex(&utflock);
		bagRehash(&utfbag);
		lockStaticMutex(&utflock);
	}

	unlockStaticMutex(&utflock);

	return (str);
}

char*
addUtfCopy(char* str, int len)
{
	char* nstr;
	char* nnstr;
	int iLockRoot;

	lockStaticMutex(&utflock);

	/* Quick check for the string */
	if (len == -1) {
		nstr = (char*)bagGet(&utfbag, str);
		if (nstr != 0) {
			nstr = bagPut(&utfbag, nstr, nstr);
			unlockStaticMutex(&utflock);
			return (nstr);
		}
		len = strlen(str);
	}

	unlockStaticMutex(&utflock);

	nstr = gc_malloc_fixed_typed(len + 1, GC_ALLOC_UTF8CONST);
	strncpy(nstr, str, len);
	nstr[len] = 0;

	for (;;) {
		lockStaticMutex(&utflock);
		nnstr = (char*)bagPut(&utfbag, nstr, nstr);
		unlockStaticMutex(&utflock);
		if (nnstr != 0) {
			break;
		}
		bagRehash(&utfbag);
	}

	if (nnstr != nstr) {
		gc_free_fixed(nstr);
	}

	return (nnstr);
}

void
rmUtf(char* str)
{
	int iLockRoot;

	lockStaticMutex(&utflock);
	bagRemove(&utfbag, str);
	unlockStaticMutex(&utflock);
}

Hjava_lang_String*
makeReplaceJavaStringFromUtf(char* str, int len, int search, int replace)
{
	Hjava_lang_String* obj;
	unsigned char* ptr;
	unsigned char* limit;
	jchar* chrs;
	int count;

	if (len == -1) {
		len = strlen(str);
	}
	limit = str + len;

	obj = (Hjava_lang_String*)newObject(StringClass);

	count = 0;
	chrs = STRING_DATA(obj);
	for (ptr = (unsigned char*)str; ptr < limit;) {
		int ch = UTF8_GET(ptr, limit);
		if (ch < 0) {
			break;
		}
		count++;
	}
	unhand(obj)->count = count;

	unhand(obj)->value = (HArrayOfChar*)newArray(&_Jv_charClass, len);
        GC_WRITE(obj, unhand(obj)->value); /* string->value */

	unhand(obj)->offset = 0;
	unhand(obj)->interned = 0;

	chrs = STRING_DATA(obj);
	for (ptr = (unsigned char*)str; ptr < limit;) {
		jchar ch = UTF8_GET(ptr, limit);
		if (ch == search) {
			ch = replace;
		}
		*chrs++ = ch;
	}

	return (obj);
}
