/*
 * lock-impl.c - pthread-based LockInterface implementation (Posix style)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
#define PTD 0

#include <pthread.h>

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "threads.h"
#include "locks.h"


/*
Lynx: The _cond_ lock functions wake up when we're being killed,
	but pthread_testcancel() doesn't seem to care.  So, we'll kill
	ourselves.
*/
extern int nonDaemons;

/***********************************************************************
 * Interface functions
 *
 * All of them should eventually turned into compile time interfaces
 * (exported inlined funcs)
 */

/*
 * Lock the mutex associated with the sem2posixLock
 */
void  
Llock ( sem2posixLock* lk )
{
  lLock( lk);
}

/*
 * Unlock the mutex associated with the sem2posixLock
 */
void    
Lunlock ( sem2posixLock* lk )
{
  lUnlock( lk);
}




/*
 * Wait for the condvar associated with the sem2posixLock, with
 * a given relative timeout in ms (which we have to convert into
 * a absolute timespec now)
 */
jboolean
Lwait ( sem2posixLock* lk, jlong timeout )
{
  nativeThread    *cur = GET_CURRENT_THREAD(&cur);
  int             stat;
  struct timespec abst;
  struct timeval  now;
  jlong ms ;

  
  cur->stackCur  = (void*)&cur;

  if ( timeout == 0 ) {
	/* we handle this as "wait forever"	*/
	cur->blockState |= BS_CV;
	

	stat = pthread_cond_wait( (pthread_cond_t*)lk->cv, (pthread_mutex_t*)lk->mux);
	
	
	
	cur->blockState &= ~BS_CV;

  }
  else {
#ifndef __Lynx__

       // printf("__Lynx__\n");
	/* timeout is in millisecs, timeval in microsecs, and timespec in nanosecs */
	if(gettimeofday( &now, 0) < 0)
		perror("gettimeofday"), ABORT();
/*
	abst.tv_sec = now.tv_sec + (timeout / 1000);
	abst.tv_nsec = (now.tv_usec * 1000) + (timeout % 1000) * 1000000;
*/
/* pjo, 10 apr 2000
- Don't use that computation--if nsec rolls over to the next second then its
	operation is unpredictable.
*/
	ms = (now.tv_usec + 999) / 1000 ;
	ms += timeout ;
	abst.tv_sec = now.tv_sec + (ms / 1000) ;
	ms %= 1000 ;
	ms *= 1000 ;	/* now it's us. */
	abst.tv_nsec = ms * 1000 ; /* Now it's ns. */
#else
/* pjo, 11 apr 2000
- Lynx posix compliance means that the time is *relative*, not absolute.
*/
     	abst.tv_sec = timeout / 1000 ;
	timeout %= 1000 ;
	abst.tv_nsec = timeout * 1000 ;		/* ms -> us */
	abst.tv_nsec *= 1000 ;				/* us -> ns */
#endif
 
	cur->blockState |= BS_CV_TO;


#if defined(SA_RESTART)

	 stat = pthread_cond_timedwait( (pthread_cond_t*)lk->cv, (pthread_mutex_t*)lk->mux, &abst);

#else

	do{
	   stat = pthread_cond_timedwait( (pthread_cond_t*)lk->cv, (pthread_mutex_t*)lk->mux, &abst);
        }
	while(stat == EINTR); 
	
#endif



	cur->blockState &= ~BS_CV_TO;
  }

/* Did we get cancelled?  Usually we'd expect to never find out.  */
  pthread_testcancel() ;
  if (nonDaemons == 0)
  	pthread_exit(0) ;

  return (stat == 0);
}

/*
 * Signal the condvar associated with the sem2posixLock, to wake one of
 * the threads waiting on it
 */
void    
Lsignal ( sem2posixLock* lk )
{
  lSignal( lk);
}       



/***********************************************************************
 * create and initialize the LockInterface
 */

LockInterface Kaffe_LockInterface = {
  Llock,
  Lunlock,
  Lwait,
  Lsignal,

  0,
  0,
};
