/*
 * internal.c
 *
 * Copyright (c) 1996, 1997
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

/*
 * This file links the ThreadInterface as defined by Tim to the revised
 * version of the internal thread interface ("jthreads").
 *
 * This file implements the functions in ThreadInterface by invoking
 * functions declared in jthread.h
 * It is responsible for unhanding, as well as invoking functions in other
 * parts of the VM.
 *
 * Note that jthreads - while heavily based on Tim's code - can be used 
 * independently of the Kaffe JVM.  That is, they don't have any compile or
 * link time dependencies on it.  This file knows about all the JVM related 
 * types, whereas jthread does not.
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-signal.h"
#include "config-net.h"
#include "jtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "lookup.h"
#include "threads.h"
#include "locks.h"
#include "exception.h"
#include "support.h"
#include "external.h"
#include "errors.h"
#include "gc.h"
#include "md.h"
#include "lerrno.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"
#include "jthread.h"

static void 
broadcastDeath(void *jlThread)
{
        Hjava_lang_Thread *tid = jlThread;
	iLock* iLockRoot;

        /* Notify on the object just in case anyone is waiting */
        lockMutex(&tid->base);
        broadcastCond(&tid->base);
        unlockMutex(&tid->base);
}

static void *
thread_malloc(size_t s)
{
	return gc_malloc_fixed_typed(s, GC_ALLOC_THREADCTX);
}

static void
thread_free(void *p)
{
	gc_free_fixed(p);
}

int flag_preemption = true;
int runFinalizerOnExit;

static
void
runfinalizer()
{
	if (runFinalizerOnExit) {
		invokeFinalizer();
	}
}

static
void
TcreateFirst(Hjava_lang_Thread* tid)
{
	sem2posixLock* lk;

	jthread_init((jthread_t*)&(unhand(tid)->PrivateInfo),
		flag_preemption,
		java_lang_Thread_MAX_PRIORITY+1,
		java_lang_Thread_MIN_PRIORITY,
		java_lang_Thread_NORM_PRIORITY,
		tid,
		threadStackSize,
		thread_malloc,
		thread_free,
		0 /* broadcastDeath */ ); /* PM: death notification done in Thread.finish() */
	jthread_atexit(runfinalizer);
	GC_WRITE(tid, unhand(tid)->PrivateInfo);

	lk = (sem2posixLock*)unhand(tid)->sem;
	lk->mux = thread_malloc(sizeof(jmutex));
	lk->cv = thread_malloc(sizeof(jcondvar));
	jmutex_initialise(lk->mux);
	jcondvar_initialise(lk->cv);
}

static
void
Tcreate(Hjava_lang_Thread* tid, void* func)
{
	sem2posixLock* lk;

	lk = (sem2posixLock*)unhand(tid)->sem;
	lk->mux = thread_malloc(sizeof(jmutex));
	lk->cv = thread_malloc(sizeof(jcondvar));
	jmutex_initialise(lk->mux);
	jcondvar_initialise(lk->cv);

	jthread_create((jthread_t*)&(unhand(tid)->PrivateInfo),
		unhand(tid)->priority,
		func,
		unhand(tid)->daemon,
		tid,
		threadStackSize);
	GC_WRITE(tid, unhand(tid)->PrivateInfo);
}

static  
void    
Tprio(Hjava_lang_Thread* tid, jint prio)
{       
	/* no native thread yet */
	if (unhand(tid)->PrivateInfo == 0)
		return;

	jthread_setpriority((jthread_t)unhand(tid)->PrivateInfo, prio);
}               

static
jint    
Tframes(Hjava_lang_Thread* tid)
{
	return jthread_frames((jthread_t)unhand(tid)->PrivateInfo);
}       

static
void
Texit(Hjava_lang_Thread* tid)
{
	jthread_exit((jthread_t)unhand(tid)->PrivateInfo);
}
 
static
void    
Tfinalize(Hjava_lang_Thread* tid)
{
	if (unhand(tid)->PrivateInfo != 0) {
		thread_free(((sem2posixLock*)unhand(tid)->sem)->mux);
		thread_free(((sem2posixLock*)unhand(tid)->sem)->cv);
		jthread_destroy((jthread_t)unhand(tid)->PrivateInfo);
		unhand(tid)->PrivateInfo = 0;
	}
}       

static          
Hjava_lang_Thread*      
TcurrentJava(void)      
{               
	return GET_COOKIE();
}

static
void
TenterCritSect (void)
{
}

static
void
TexitCritSect (void)
{
}

static          
void
TwalkThreads(void)
{               
	/* this is from gc-incremental.c */
	extern void walkMemory(void*);
	jthread_walkLiveThreads(walkMemory);
}

/*      
 * Walk the thread's internal context.
 */
static
void
TwalkThread(Hjava_lang_Thread* tid)
{
	void *from;
	unsigned len;
	jthread_t jtid = (jthread_t)unhand(tid)->PrivateInfo;

        if (jtid == 0)
                return;
 
        markObject(unhand(tid)->exceptObj);

	/* ask threading system what the interesting stack range is */
	jthread_extract_stack(jtid, &from, &len);

	/* and walk it */
	walkConservative(from, len);
}

/* called when unrolling exceptions */
static  
void
TnextFrame(stackFrame* fm)
{
        machineStackFrame* nfm;

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

        if (jthread_on_current_stack(STACK_FRAME(nfm))) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = STACK_FRAME(nfm);
        }
        else {
		fm->return_pc = 0;
		fm->return_frame = 0;
        }
}

static  
void    
Llock(sem2posixLock* lk)
{
	jmutex_lock(lk->mux);
}

static  
void    
Lunlock(sem2posixLock* lk)
{       
	jmutex_unlock(lk->mux);
}

static
jboolean
Lwait(sem2posixLock* lk, jlong timeout)
{
	return (jcondvar_wait(lk->cv, lk->mux, timeout));
}
 
static  
void    
Lsignal(sem2posixLock* lk)
{
	jcondvar_signal(lk->cv, lk->mux);
}       

/*
 * Wait for a child process.
 */
static
int
threadedWaitpid(int wpid, int* status, int options)
{
	int pid = jthreadedWaitpid(wpid, status, options);
	if (pid == -1)
		SignalError("java.lang.InternalError", 
				"no waitpid() available");

	return pid;
}

static
int
jthreadedMkdir(const char* path)
{
	return mkdir(path, 0777);
}

static
int
jthreadedFileInfo(const char* path, int type, void* extra)
{
	struct stat buf;
	int r;

	switch (type) {
	case JINFO_EXISTS:
		r = stat(path, &buf);
		if (r == 0) {
			return (1);
		}
		break;

	case JINFO_ISFILE:
		r = stat(path, &buf);
		if (r == 0 && S_ISREG(buf.st_mode)) {
			return (1);
		}
		break;

	case JINFO_ISDIR:
		r = stat(path, &buf);
		if (r == 0 && S_ISDIR(buf.st_mode)) {
			return (1);
		}
		break;

	case JINFO_CANREAD:
		r = access(path, R_OK);
		if (r >= 0) {
			return (1);
		}
		break;

	case JINFO_CANWRITE:
		r = access(path, W_OK);
		if (r >= 0) {
			return (1);
		}
		break;

	case JINFO_LENGTH:
		r = stat(path, &buf);
		if (r == 0) {
			*((jlong*)extra) = buf.st_size;
			return (1);
		}
		break;

	case JINFO_LASTMODIFIED:
		r = stat(path, &buf);
		if (r == 0) {
			*((jlong*)extra) = (jlong)buf.st_mtime * (jlong)1000;
			return (1);
		}
		break;
	}

	return (0);
}

static
int
jthreadedFileInfoFD(int fd, int type, void* extra)
{
	int r;
	int nr;
	struct stat buf;
#if !(defined(HAVE_IOCTL) && defined(FIONREAD)) && defined(HAVE_SELECT)
	static struct timeval tm = { 0, 0 };
	fd_set rd;
#endif

	switch (type) {
	case JINFO_LENGTH:
		r = fstat(fd, &buf);
		if (r == 0) {
			*((jlong*)extra) = buf.st_size;
			return (1);
		}
		break;

	case JINFO_AVAILABLE:
#if defined(HAVE_IOCTL) && defined(FIONREAD)
		r = ioctl(fd, FIONREAD, &nr);
		if (r >= 0) {
			return (nr);
		}
		break;
#elif defined(HAVE_SELECT)
		/* This uses select() to work out if we can read - but what
		 * happens at the end of file?
		 */
		FD_ZERO(&rd);
		FD_SET(fd, &rd);
		r = select(fd+1, &rd, NULL, NULL, &tm);
		if (r == 1) {
			return (1);
		}
		break;
#endif
	}

	return (0);
}

static
int
jthreadedReaddir(const char* path, jdirEntry** list)
{
        DIR* dir;
        struct dirent* entry;
        jdirEntry* dirlist;
        jdirEntry* mentry;
	int count;

        dir = opendir(path);
        if (dir == 0) {
                return (-1);
        }

        dirlist = 0;
        count = 0;
        while ((entry = readdir(dir)) != 0) {
                /* We skip '.' and '..' */
                if (strcmp(".", entry->d_name) == 0 ||
                    strcmp("..", entry->d_name) == 0) {
                        continue;
                }
                mentry = malloc(sizeof(jdirEntry) + NAMLEN(entry));
                assert(mentry != 0);
                strcpy(mentry->name, entry->d_name);
                mentry->next = dirlist;
                dirlist = mentry;
                count++;
        }
        closedir(dir);

	*list = dirlist;
	return (count);
}

static
void
notSupported(void)
{
	SignalError("java.lang.Error", "not supported");
}

static
void
Tinit(void)
{
}

static
void
Finit(void)
{
}

static
void 
Tstop(Hjava_lang_Thread* tid, HObject* exc)
{
}

static
void
TstackInfo(threadStackInfo* info)
{
	jthread* jtid = jthread_current();
	info->low = jtid->stackBase;
	info->high = jtid->stackEnd;
	info->curr = info->low;
}

static
void*
Mpagealloc(int size)
{
	void* ptr;

#define	CHECK_OUT_OF_MEMORY(P)	if ((P) == 0) throwOutOfMemory();

#if defined(HAVE_SBRK)

	/* Our primary choice for basic memory allocation is sbrk() which
	 * should avoid any unsee space overheads.
	 */
	for (;;) {
		int missed;
		ptr = sbrk(size);
		if (ptr == (void*)-1) {
			ptr = 0;
			break;
		}
		if ((uintp)ptr % gc_pgsize == 0) {
			break;
		}
		missed = gc_pgsize - ((uintp)ptr % gc_pgsize);
		sbrk(-size + missed);
	}
	CHECK_OUT_OF_MEMORY(ptr);

#elif defined(HAVE_MEMALIGN)

        ptr = memalign(gc_pgsize, size);
	CHECK_OUT_OF_MEMORY(ptr);

#elif defined(HAVE_VALLOC)

        ptr = valloc(size);
	CHECK_OUT_OF_MEMORY(ptr);

#else

	/* Fallback ...
	 * Allocate memory using malloc and align by hand.
	 */
	size += gc_pgsize;
#ifdef malloc
#undef malloc
#endif
        ptr = malloc(size);
	CHECK_OUT_OF_MEMORY(ptr);
	ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -gc_pgsize);

#endif

	return (ptr);
}

static
int
Kinput(void)
{
	char ch;
	if ((*Kaffe_FileSystemCallInterface._read)(0, &ch ,1) == 1) {
		return (ch);
	}
	else {
		return (-1);
	}
}

static
int
Kinputavailable(void)
{
	return (0);
}

static
void
Koutput(int chr)
{
	char ch = (char)chr;
	(*Kaffe_FileSystemCallInterface._write)(1, &ch ,1);
}

static
void
Kerror(int chr)
{
	char ch = (char)chr;
	(*Kaffe_FileSystemCallInterface._write)(2, &ch ,1);
}

static
jlong
Ktime(void)
{
	jlong tme;
#if defined(HAVE_GETTIMEOFDAY)
	struct timeval tm;
	gettimeofday(&tm, 0);
	tme = (((jlong)tm.tv_sec * (jlong)1000) + ((jlong)tm.tv_usec / (jlong)1000));
#elif defined(HAVE_FTIME)
	struct timeb tm;
	ftime(&tm);
	tme = (((jlong)tm.time * (jlong)1000) + (jlong)tm.millitm);
#elif defined(HAVE__FTIME)
	struct _timeb tm;
	_ftime(&tm);
	tme = (((jlong)tm.time * (jlong)1000) + (jlong)tm.millitm);
#elif defined(HAVE_TIME)
	tme = (jlong)1000 * (jlong)time(0);
#else
	tme = 1;
#endif
	return (tme);
}

/*
 * Define the thread interface.
 */
ThreadInterface Kaffe_ThreadInterface = {

	Tinit,

	TcreateFirst,
	Tcreate,
	jthread_yield,
	Tprio, 
	Texit,
	Tstop,
	Tfinalize,

	TcurrentJava,
	jthread_current,
	
	TenterCritSect,
	TexitCritSect,

	TwalkThreads,
	TwalkThread,

	TnextFrame,

	TstackInfo
};

/*
 * Define the lock interface.
 */
LockInterface Kaffe_LockInterface = {

  Llock,
  Lunlock,
  Lwait,
  Lsignal,

  0,
  0,

};

KernelCallInterface Kaffe_KernelCallInterface = {

	jthreadedFileDescriptor,
        select,	
        jthreadedWaitpid,

	Mpagealloc,
	0, /* pagefree */
	Kinput,
	Koutput,
	Kerror,
	Ktime,
	Kinputavailable,

};


FileSystemCallInterface Kaffe_FileSystemCallInterface = {

	Finit,

        jthreadedOpen,
        jthreadedRead,	
        jthreadedWrite, 
        jthreadedLseek,
        close,
        jthreadedFileInfo,
        jthreadedFileInfoFD,

        jthreadedReaddir,

        jthreadedMkdir,
        rmdir,
        rename,
        remove,

};

#if defined(HAVE_LIBRARY_NET) && defined(HAVE_SYS_SOCKET_H)

static int Init;

static
void
Ninit (void)
{
	if (!Init)
	{
		// Need to move this elsewhere
		if (SocketInit () != 0)
		{
			printf ("Failed to init network\n") ;
			return;
		}
		Init = 1 ;
	}
}

static
int
Nsocket(int stream)
{
  if ( !Init )
		return -1 ;

	if (stream == 0) {
		return (socket(AF_INET, SOCK_DGRAM, 0));
	}
	else {
		return (socket(AF_INET, SOCK_STREAM, 0));
	}
}

static
int
Nbind(int fd, struct sockaddr* addr, int addrlen)
{
	return (bind(fd, addr, addrlen));
}

static
int
Nlisten(int fd, int qlength)
{
	return (listen(fd, qlength));
}

static
int
Nconnect(int fd, struct sockaddr* addr, size_t addrlen)
{
	return (connect(fd, addr, addrlen));
}

static
int
Naccept(int fd, struct sockaddr* addr, size_t* addrlen)
{
	return (accept(fd, addr, (int*)addrlen));
}

static
ssize_t
Nsockread(int fd, void* buf, size_t len)
{
	return (recv(fd, buf, len, 0));
}

static
ssize_t
Nrecvfrom(int fd, void* buf, size_t len, int flags, struct sockaddr* from, int* fromlen)
{
	return (recvfrom(fd, buf, len, flags, from, fromlen));
}

static
ssize_t
Nsockwrite(int fd, const void* buf, size_t len)
{
	return (send(fd, (char *)buf, len, 0));
}

static
ssize_t
Nsendto(int fd, const void* buf, size_t len, int flags, const struct sockaddr* to, int tolen)
{
	return (sendto(fd, (char *)buf, len, flags, (struct sockaddr*)to, tolen));
}

static
int
Nclose(int fd)
{
	return (soclose(fd));
}

static
int
Nsetsockopt(int fd, int level, int optname, const void* optval, int optlen)
{
	notSupported () ;
	//return (setsockopt(fd, level, optname, (char*)optval, optlen));
}

static
int
Ngetsockopt(int fd, int level, int optname, void* optval, int* optlen)
{
	notSupported () ;
	//return (getsockopt(fd, level, optname, (char*)optval, optlen));
}

static
int
Ngetsockname(int fd, struct sockaddr* addr, int* addrlen)
{
	return (getsockname(fd, addr, addrlen));
}

static
int
Ngetpeername(int fd, struct sockaddr* addr, int* addrlen)
{
	return (getpeername(fd, addr, addrlen));
}

static
char*
Ngethostname(void)
{
	return ("localhost");
}

static
int*
Ngethostbyname(char* name)
{
	struct hostent* ent;
	int* addrs;

	if (!NetInit ())
		return -1 ;

	ent = gethostbyname(name);
	if (ent == 0) {
		return (0);
	}
	addrs = (int*)gc_malloc_fixed(sizeof(int) * 2);
	addrs[0] = *(int*)ent->h_addr;
	addrs[1] = 0;
	return (addrs);
}

static
char*
Ngethostbyaddr(int addr)
{
	struct hostent* ent;

	if (!NetInit ())
		return NULL ;

	ent = gethostbyaddr((char*)&addr, 4, AF_INET);
	if (ent == 0) {
		return (0);
	}
	return (ent->h_name);
}

static
int
Nioctlsocket(int fd, int opt, int* val)
{
	return (soioctl(fd, opt, (char *)val));
}

#else

static
void
Ninit (void)
{
}

static
int
Nsocket(int stream)
{
	return -1 ;
}

static
int
Nbind(int fd, struct sockaddr* addr, int addrlen)
{
	return -1 ;
}

static
int
Nlisten(int fd, int qlength)
{
	return -1 ;
}

static
int
Nconnect(int fd, struct sockaddr* addr, size_t addrlen)
{
	return -1 ;
}

static
int
Naccept(int fd, struct sockaddr* addr, size_t* addrlen)
{
	return -1 ;
}

static
ssize_t
Nsockread(int fd, void* buf, size_t len)
{
	return -1 ;
}

static
ssize_t
Nrecvfrom(int fd, void* buf, size_t len, int flags, struct sockaddr* from, int* fromlen)
{
	return -1 ;
}

static
ssize_t
Nsockwrite(int fd, const void* buf, size_t len)
{
	return -1 ;
}

static
ssize_t
Nsendto(int fd, const void* buf, size_t len, int flags, const struct sockaddr* to, int tolen)
{
	return -1 ;
}

static
int
Nclose(int fd)
{
	return -1 ;
}

static
int
Nsetsockopt(int fd, int level, int optname, const void* optval, int optlen)
{
	return -1 ;
}

static
int
Ngetsockopt(int fd, int level, int optname, void* optval, int* optlen)
{
	return -1 ;
}

static
int
Ngetsockname(int fd, struct sockaddr* addr, int* addrlen)
{
	return -1 ;
}

static
int
Ngetpeername(int fd, struct sockaddr* addr, int* addrlen)
{
	return -1 ;
}

static
char*
Ngethostname(void)
{
	return 0;
}

static
int*
Ngethostbyname(char* name)
{
	return 0 ;
}

static
char*
Ngethostbyaddr(int addr)
{
	return 0 ;
}

static
int
Nioctlsocket(int fd, int opt, int* val)
{
	return -1 ;
}

#endif

NetworkCallInterface Kaffe_NetworkCallInterface = {
	Ninit,

	Nsocket,
	Nbind,
	Nlisten,
	Nconnect,
	Naccept,
	Nsockread,
	Nrecvfrom,
	Nsockwrite,
	Nsendto,
	Nclose,
	Nsetsockopt,
	Ngetsockopt,
	Ngetsockname,
	Ngetpeername,
	Ngethostname,
	Ngethostbyname,
	Ngethostbyaddr,
	Nioctlsocket,

	0, /* setreadtimeout */
	0, /* getreadtimeout */

};
