/*
 * kpsos.c
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "thread.h"
#include "locks.h"
#include "jsyscall.h"
#include "gc.h"
#include <psos.h>
#include "kpsos.h"

/*
 * Do we run finalizer on exit?  Well this VM probably never exits anyhow.
 */
bool runFinalizerOnExit;

static Hjava_lang_Thread* activeThreads;
static unsigned long spinlock = 0xDEADDEAD;
static unsigned long sleeplock = 0xDEADDEAD;

static
void
TcreateFirst(Hjava_lang_Thread* tid)
{
	nativeThread* ptr;

	ptr = gc_malloc_fixed(sizeof(nativeThread));
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	t_ident((char*)0, 0, &ptr->tid);

	/* Add thread to list of known java threads */
	ptr->next = activeThreads;
	activeThreads = tid;

	ptr->stackBase = ...
	ptr->stackEnd = ...
}

static
void
Tcreate(Hjava_lang_Thread* tid, void* func)
{
	nativeThread* ptr;
	static unsigned long args[4];

	ptr = gc_malloc_fixed(sizeof(nativeThread));
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	t_create(
		"JAVA",
		JAVAPRIO2PSOSPRIO(unhand(tid)->priority),
		threadStackSize,
		0,
		T_LOCAL|T_FPU,
		&ptr->tid);

	/* Add thread to list of known java threads */
	ptr->next = activeThreads;
	activeThreads = ptr;

	ptr->stackBase = ...
	ptr->stackEnd = ...

	t_start(ptr->tid, T_PREEMPT|T_TSLICE, func, args);
}

static  
void    
Tprio(Hjava_lang_Thread* tid, jint prio)
{
	unsigned long old;

	if (unhand(tid)->PrivateInfo != 0) {
		t_setpri(NATIVE_THREAD(tid)->tid, JAVAPRIO2PSOSPRIO(prio), &old);
	}
}               

/*
 * Sleep current thread for given period.
 */
static
void
Tsleep(jlong period)
{
	/* We do this by acquiring a semaphore which we can never get with
	 * a timeout.
	 */
	if (sleeplock == 0xDEADDEAD) {
		sm_create("SLEP", 0, SM_LOCAL|SM_PRIOR, &sleeplock);
	}
	sm_p(sleeplock, T_WAIT, JAVATIMEOUT2PSOSTIMEOUT(period));
}

/*
 * Terminate current thread.
 */
static
void
Texit(void)
{
	unsigned long tid;
	Hjava_lang_Thread** tptr;
	nativeThread* ntid;

	t_ident((char*)0, 0, &tid);

	t_delete(tid);

	/* Remove thread from active list */
	for (tptr = &activeThreads; *tptr != 0; tptr = &ntid->next) {
		ntid = NATIVE_THREAD(*tptr);
		if (ntid->threadx == tid) {
			*tptr = ntid->next;
			break;
		}
	}
}

/*
 * Yield processor
 */
static
void
Tyield(void)
{
	/* Don't know how to yield on pSOS */
}

static
jint    
Tframes(Hjava_lang_Thread* tid)
{
	return (0);
}
 
/*
 * Thread is being finalized - free any held resource.
 */
static
void    
Tfinalize(Hjava_lang_Thread* tid)
{
	gc_free_fixed(NATIVE_THREAD(tid));
}       

static          
Hjava_lang_Thread*      
TcurrentJava(void)      
{
	unsigned long tid;
	Hjava_lang_Thread* tptr;
	nativeThread* nptr;

	t_ident((char*)0, 0, &tid);

	/* Find Java thread */
	for (tptr = activeThreads; tptr != 0; tptr = nptr->next) {
		nptr = NATIVE_THREAD(tptr);
		if (nptr->tid == tid) {
			return (tptr);
		}
	}
	return (0);	/* Shouldn't happen */
}

static          
void*
TcurrentNative(void)      
{
	unsigned long tid;

	t_ident((char*)0, 0, &tid);

	return (tid);
}

static          
void
TwalkThreads(void)
{               
	Hjava_lang_Thread* tid;

	for (tid = activeThreads; tid != NULL; tid = NATIVE_THREAD(tid)->next) {
		walkMemory(tid);
	}
}

/*      
 * Walk the thread's internal context.
 */
static
void
TwalkThread(Hjava_lang_Thread* tid)
{
	nativeThread* native;
	void* stackStart;
	void* stackEnd;

	if (unhand(tid)->PrivateInfo == 0) {
		return;
	}

	native = NATIVE_THREAD(tid);
	markObject(unhand(tid)->jnireferences);
	markObject(unhand(tid)->exceptObj);
	stackStart = ...
	stackEnd = ...
	walkConservative(native->stackBase, native->stackEnd);
}

static
void
TnextFrame(stackFrame* fm)
{
	machineStackFrame* nfm;
	void* fp;
	nativeThread* native;

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

	fp = STACK_FRAME(nfm);
	native = NATIVE_THREAD(TcurrentJava());

	if (fp > native->stackBase && fp < native->stackEnd) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = fp;
	}
	else {
		fm->return_pc = 0;
		fm->return_frame = 0;
	}
}

static
void  
Linit(iLock* lk)
{
	unsigned long mux;
	unsigned long cv;

	sm_create("MUX ", 1, SM_LOCAL|SM_PRIOR, &mux);
	sm_create("CV  ", 0, SM_LOCAL|SM_PRIOR, &cv);

	lk->mux = (void*)mux;
	lk->cv = (void*)cv;
}               

/*
 * Get a semaphore (with timeout).
 */
static
void
Lsemget(void* sem, jlong timeout)
{
	int tout;

	tout = JAVATIMEOUT2PSOSTIMEOUT(timeout);

	sm_p((unsigned long)sem, SM_WAIT, tout);
}

/*
 * Put a semaphore.
 */
static
void
Lsemput(void* sem)
{
	sm_v((unsigned long)sem);
}

void    
Tspinon(void* arg)
{
	if (spinlock == 0xDEADDEAD) {
		sm_create("SPIN", 1, SM_LOCAL|SM_PRIOR, &spinlock);
	}
	sm_p(spinlock, SM_WAIT, 0);
}       
        
void            
Tspinoff(void* arg)
{
	sm_v(spinlock);
}               

void
Tspinoffall(void* arg)
{
}

static
void
Tinit(void)
{
}

static
void
notSupported()
{
	printf("Operation not supported\n");
}

/*
 * Define the thread interface.
 */
ThreadInterface Kaffe_ThreadInterface = {

	Tinit,
        TcreateFirst,
        Tcreate,
        Tsleep,
        Tyield,
        Tprio, 
        Texit,
        Tframes,
        Tfinalize,
        TcurrentJava,
	TcurrentNative,
        TwalkThreads,
        TwalkThread,
        TnextFrame,

};

/*
 * Define the lock interface.
 */
LockInterface Kaffe_LockInterface = {

        Linit,
        Linit,

        0,
        0,
        0,
        0,
        0,

	Lsemget,
	Lsemput,

        Tspinon,
        Tspinoff,

};

/*
 * Define the system call interface.
 */
SystemCallInterface Kaffe_SystemCallInterface = {

	notSupported,
        notSupported,
        notSupported,	
        notSupported, 
        notSupported,
        notSupported,
        notSupported,
        notSupported,

        notSupported,

        notSupported,
        notSupported,
        notSupported,
        notSupported,

        notSupported,
        notSupported,
        notSupported, 
        notSupported,	
        notSupported,
        notSupported, 
        notSupported,	
        notSupported,	
        notSupported,
        notSupported, 
        notSupported,

        notSupported,	

        notSupported
};
