/*
 * smx-fs.c
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#define DBG(s)

#define EXTERN_C extern "C"

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "threads.h"
#include "locks.h"
#include "gc.h"
#include "smx-smx.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"
#include "support.h"

extern BOOLEAN floppy_open(word);

/*
 * Init file system.
 */
void
Sinit(void)
{
    /* no need to change the default drive if it is set in properly /smx/xfile/pcconf.h */
}

/*
 * Open a file.
 */
static
int
Sopen(const char* path, int mode)
{
	PCFD fd;

DBG( kprintf("Sopen: %s (%d)\n", path, mode); )

	switch (mode) {
	case JOPEN_READ:
		fd = po_open((char*)path, PO_BINARY|PO_RDONLY, PS_IREAD|PS_IWRITE);
		break;
	case JOPEN_WRITE:
		fd = po_open((char*)path, PO_BINARY|PO_WRONLY|PO_CREAT|PO_TRUNC, PS_IREAD|PS_IWRITE);
		break;
	case JOPEN_APPEND:
		fd = po_open((char*)path, PO_BINARY|PO_RDWR|PO_CREAT, PS_IREAD|PS_IWRITE);
		/* Simulate append by moving to end of file */
		po_lseek(fd, 0L, PSEEK_END);
		break;
	case JOPEN_READWRITE:
		fd = po_open((char*)path, PO_BINARY|PO_RDWR|PO_CREAT, PS_IREAD|PS_IWRITE);
		break;
	default:
		fd = -1;
		break;
	}

	return (fd);
}

/*
 * Get information about a file handle.
 */
static
int
SfileinfoFD(int fd, int type, void* extra)
{
	STAT statbuf;
	long here;

	switch (type) {
	case JINFO_LENGTH:
		if (pc_fstat(fd, &statbuf) == -1) {
			return (0);
		}
		*((jlong*)extra) = statbuf.st_size;
		return (1);

	case JINFO_AVAILABLE:
		if (pc_fstat(fd, &statbuf) == -1) {
			return (0);
		}
		here = po_lseek(fd, 0L, PSEEK_CUR);
		if (here == -1) {
			return (0);
		}
		return ((int)(statbuf.st_size - here));

	default:
		return (0);
	}
}

static
int
Sfileinfo(const char* name, int type, void* extra)
{
	STAT statbuf;

DBG( kprintf("Sfileinfo: %s\n", name); )

	switch (type) {
	case JINFO_EXISTS:
		if (pc_stat((char*)name, &statbuf) == -1) {
			return (0);
		}
		return (1);

	case JINFO_ISFILE:
		if (pc_stat((char*)name, &statbuf) == -1) {
			return (0);
		}
		if ((statbuf.st_mode & S_IFREG) == 0) {
			return (0);
		}
		return (1);

	case JINFO_ISDIR:
		if (pc_stat((char*)name, &statbuf) == -1) {
			return (0);
		}
		if ((statbuf.st_mode & S_IFDIR) == 0) {
			return (0);
		}
		return (1);

	case JINFO_CANREAD:
		/* We can always read on a DOS compatible file system */
		return (1);

	case JINFO_CANWRITE:
		if (pc_stat((char*)name, &statbuf) == -1) {
			return (0);
		}
		if ((statbuf.st_mode & S_IWRITE) == 0) {
			return (0);
		}
		return (1);

	case JINFO_LENGTH:
		if (pc_stat((char*)name, &statbuf) == -1) {
			return (0);
		}
		*((jlong*)extra) = statbuf.st_size;
		return (1);

	case JINFO_LASTMODIFIED:
		if (pc_stat((char*)name, &statbuf) == -1) {
			return (0);
		}
		/*
		*((jlong*)extra) = (jlong)(statbuf.st_mtime)* (jlong)1000;
		*/
		*((jlong*)extra) = 0;
		return (1);

	default:
		return (0);
	}
}

static
ssize_t
Sread(int fd, void* buf, size_t len)
{
	return (po_read(fd, (byte*)buf, (ushort)len));
}

static
ssize_t
Swrite(int fd, const void* buf, size_t len)
{
	/* Truncate the length if it's too long - it's okay to write less
	 * than we're asked to.
	 */
	if (len > 0xFFFF) {
		len = 0xFFFF;
	}
	return (po_write(fd, (byte*)buf, (ushort)len));
}

static
int
Slseek(int fd, off_t offset, int whence)
{
	short origin;

	switch (whence) {
	case JSEEK_END:
		origin = PSEEK_END;
		break;
	case JSEEK_CUR:
		origin = PSEEK_CUR;
		break;
	case JSEEK_SET:
	default:
		origin = PSEEK_SET;
		break;
	}

	/* po_lseek returns the new position, or -1 on error.
	 */
	return (po_lseek(fd, (long)offset, origin));
}

static
int
Sclose(int fd)
{
	return (po_close(fd));
}

static
int
Sreaddir(const char* name, struct jdirEntry** data)
{
	DSTAT statbuf;
    jdirEntry* dirlist;
    jdirEntry* mentry;
    int count;
	char* dirname;

DBG( kprintf("Sreaddir: %s\n", name); )

	dirname = (char*)gc_malloc_fixed(strlen(name) + 5);
	strcpy(dirname, name);
	strcat(dirname, "\\*.*");

	if (pc_gfirst(&statbuf, dirname) == FALSE) {
		gc_free_fixed(dirname);
		return (-1);
	}

        dirlist = 0;
        count = 0;
	do {
                /* We skip '.' and '..' */
                if (strcmp(".", statbuf.filename) == 0 ||
                    strcmp("..", statbuf.filename) == 0) {
                        continue;
                }
                mentry = (jdirEntry*)gc_malloc_fixed(sizeof(jdirEntry) + strlen(statbuf.filename));
                assert(mentry != 0);
                strcpy(mentry->name, statbuf.filename);
                mentry->next = dirlist;
                dirlist = mentry;
                count++;

        } while (pc_gnext(&statbuf) == TRUE);

        pc_gdone(&statbuf);
		gc_free_fixed(dirname);

        *data = dirlist;
        return (count);
}

static
int
Smkdir(const char* name)
{
	if (pc_mkdir((char*)name) == TRUE) {
		return (0);
	}
	else {
		return (-1);
	}
}

static
int
Srmdir(const char* name)
{
	if (pc_rmdir((char*)name) == TRUE) {
		return (0);
	}
	else {
		return (-1);
	}
}

static
int
Srename(const char* from, const char* to)
{
	if (pc_mv((char*)from, (char*)to) == TRUE) {
		return (0);
	}
	else {
		return (-1);
	}
}

static
int
Sremove(const char* name)
{
	if (pc_unlink((char*)name) == TRUE) {
		return (0);
	}
	else {
		return (-1);
	}
}

extern "C" {
FileSystemCallInterface Kaffe_FileSystemCallInterface = {

	Sinit,	

	Sopen,
	Sread,
	Swrite,
	Slseek,
	Sclose,
	Sfileinfo,
	SfileinfoFD,

	Sreaddir,

	Smkdir,
	Srmdir,
	Srename,
	Sremove,

};
}
