/*
 * smx-os.c
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

extern "C" {

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "threads.h"   /* will get us thread-impl.h */
#include "locks.h"     /* and this brings in lock-impl.h */
#include "support.h"

#include "gc.h"

#include "smx-smx.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"

int                runFinalizerOnExit;
int                critSection;

/* we need a linked list of all active threads for GC purposes */
nativeThread       *activeThreads;


#define LOCKSLOT   int iLockRoot
#define LOCK()     lockStaticMutex( &tLock)
#define UNLOCK()   unlockStaticMutex( &tLock)
static iLock	   *tLock;


#if !defined(STACKREDZONE)
#define	STACKREDZONE	1024
#endif



void
Tinit(void)
{
	nativeThread* ptr;

	ptr = (nativeThread*)gc_malloc_fixed(sizeof(nativeThread));
	ptr->task = ct; /* ct - current task global variable */
	TASK_SPECIFIC_DATA(ct) = ptr;
}

void
TcreateFirst(Hjava_lang_Thread* thread)
{
	nativeThread* ptr;

	ptr = (nativeThread*)TASK_SPECIFIC_DATA(ct);
	unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	ptr->thread = thread;

	unhand(thread)->sem = (Hkaffe_util_Ptr*) create_sema(1, 0);
	assert(unhand(thread)->sem != 0);

	ptr->next = activeThreads;
	activeThreads = ptr;

	bump_task(ptr->task, JAVAPRIO2SMXPRIO(unhand(thread)->priority));
}

void
Tcreate(Hjava_lang_Thread* thread, void* func)
{
	LOCKSLOT;
	nativeThread* ptr;

	ptr = (nativeThread*)gc_malloc_fixed(sizeof(nativeThread));
	unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	/*
	 * Note that tasks have to be created in a suspend state - otherwise
	 * we would have to revert to a "tRun" wrapper for the task user func
	 * (and some syncing) in order to ensure activeThreads consistency
	 */
	ptr->task = create_taskn((void (*)(void))func,
		                     JAVAPRIO2SMXPRIO(unhand(thread)->priority), threadStackSize);
	assert(ptr->task != 0);
	TASK_SPECIFIC_DATA(ptr->task) = ptr;
	ptr->thread = thread;

	unhand(thread)->sem = (Hkaffe_util_Ptr*) create_sema(1, 0);
	assert(unhand(thread)->sem != 0);

	LOCK();    /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
	ptr->next = activeThreads;
	activeThreads = ptr;
	UNLOCK();  /* --------------------------------------------------------------------- */

	resume(ptr->task);
}

static  
void    
Tprio (Hjava_lang_Thread* thread, jint prio)
{
	tPrio( thread, prio);
}               


/*
 * Terminate current thread.
 */
void
Texit (Hjava_lang_Thread* thread)
{
	LOCKSLOT;
	nativeThread    *t, *nt = 0;

	LOCK();   /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

	if ( activeThreads->thread == thread ) {
		nt = activeThreads;
		activeThreads = nt->next;
	}
	else {
		for (t = activeThreads; t->next; t = t->next) {
			if ( t->next->thread == thread ){
				nt = t->next;
				t->next = nt->next;
				break;
			}
		}
	}

	UNLOCK(); /* ----------------------------------------------------------------------- */

	assert( nt != 0 );

	unhand(thread)->PrivateInfo = 0;
	nt->thread = 0;
	TASK_SPECIFIC_DATA(nt->task) = 0;

	delete_sem((SCB_PTR far&)unhand(thread)->sem);
	delete_task((TCB_PTR far&)nt->task);

	gc_free_fixed( nt);
}

/*
 * Yield processor
 */
void
Tyield(void)
{
	tYield();
}

        
/*
 * Thread is being finalized - free any held resource.
 */
void    
Tfinalize(Hjava_lang_Thread* thread)
{
	/* Does nothing */
}       


Hjava_lang_Thread*      
TcurrentJava(void)      
{
	return (tCurrentJava());
}
          
void
TenterCritSect(void)
{
	/* nothing to do, SMX has strict fifo scheduling, and the "sp"
	 * stackpointer fields are set each time the scheduler switches away from
	 * a task */

	critSection++;
}


void
TexitCritSect(void)
{
	/* nothing to do, strict fifo scheduling (see TenterCritSect) */

	critSection--;
}

void
TwalkThreads (void)
{               
	nativeThread *nt;

	assert( critSection > 0 );
	/* no need to lock, we are in a critical section */

	for ( nt = activeThreads; nt != 0; nt = nt->next ){
		walkMemory( nt->thread);
	}
}

/*      
 * Walk the thread's internal context.
 */
void
TwalkThread (Hjava_lang_Thread* thread)
{
	nativeThread   *nt = (nativeThread*)unhand(thread)->PrivateInfo;
	unsigned long  smin, smax;
	long           ssize;

	/* never walk a thread outside a crit section */
	assert( critSection > 0 );

	if ( nt == 0 ) {
		/* Hmm, walking a already exited thread? */
		return;
	}

	markObject(unhand(nt->thread)->jnireferences);
	markObject(unhand(nt->thread)->exceptObj);

	/* let's just deal with downward growing stacks */
	smax = (unsigned long) nt->task->sh->sx;
	smin = (nt->task == ct) ? (unsigned long) &smax : (unsigned long) nt->task->sp;
	//smin = (unsigned long) nt->task->sh->si;

	ssize = smax - smin + 4;
	//assert( (ssize < 0) || (ssize > threadStackSize) );

	walkConservative( (void*)smin, ssize);
}


void
TnextFrame(stackFrame* fm)
{
	machineStackFrame* nfm;
	void* fp;

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

	fp = STACK_FRAME(nfm);
	if (fp > ct->sh->si && fp < ct->sh->sx) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = fp;
	}
	else {
		fm->return_pc = 0;
		fm->return_frame = 0;
	}
}

void
TstackInfo (threadStackInfo* info)
{
	nativeThread* nt = (nativeThread*) TASK_SPECIFIC_DATA(ct);

	info->low = (void*)((uintp)nt->task->sh->si + STACKREDZONE);
	info->high = (void*)nt->task->sh->sx;
	info->curr = info->low;
}


/*********************************************************************************
 * lock interface implementation
 */

/*
 * Get a semaphore (with timeout).
 */
jboolean
Lsemget(void* sem, jlong timeout)
{
	return (lSemget( sem, timeout));
}

/*
 * Put a semaphore.
 */
void
Lsemput(void* sem)
{
	lSemput( sem);
}


/*
 * Define the thread interface.
 */

ThreadInterface Kaffe_ThreadInterface = {

	Tinit,
	TcreateFirst,
	Tcreate,
	Tyield,
	Tprio, 
	Texit,
	0,
	Tfinalize,
	TcurrentJava,
	0,
	TenterCritSect,
	TexitCritSect,
	TwalkThreads,
	TwalkThread,
	TnextFrame,
	TstackInfo

};

void
notSupported()
{
        kprintf("Operation not supported\n");
}

void*
Mpagealloc(int size)
{
	void* ptr;

	size += gc_pgsize;

#undef malloc

	ptr = malloc(size);
	if (ptr == 0) {
		return (0);
	}
	ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -(int)gc_pgsize);

	return (ptr);
}

void
Mpagefree(void* mem)
{
	/* Freeing memory back to system not currently supported */
}

static
int
Kinput(void)
{
	/* Input not currently supported */
	return (-1);
}

static
int
Kinputavailable(void)
{
	return (0);
}

static char  outBuf[256];
static char  *outLine = outBuf;
static int   nOut;

static
void
Koutput(int chr)
{
	if ( (nOut >= sizeof( outBuf)) || (chr == '\n') ){
		nOut = 0;
	}
	else {
		outBuf[nOut++] = chr;
		outBuf[nOut] = 0;
	}

	/* Print character to console */
	kprintf("%c", chr);
}

static
jlong
Ktime(void)
{
        dword tm;
	dword st;
	dword et;
        jlong rtm;

	/* Get the two time - and keep getting them if the etime changes */
	do {
		et = get_etime();
		st = get_stime();
	} while (et != get_etime());

        /* This give us he time in cf.sec units */
	tm = (st * cf.sec) + (et % cf.sec);

        /* We now convert this into milliseconds */
        rtm = (((jlong)tm) * (jlong)1000) / (jlong)cf.sec;

        return (rtm);
}



/*
 * Define the lock interface.
 */

LockInterface Kaffe_LockInterface = {

        0,
        0,
        0,
        0,

	Lsemget,
	Lsemput,

};

/*
 * Define the system call interface.
 */

KernelCallInterface Kaffe_KernelCallInterface = {

	0,
	0,
	0,
	Mpagealloc,
	Mpagefree,
	Kinput,
	Koutput,
	Koutput,
	Ktime,
	Kinputavailable,

};

}
