/*
 * internal.c
 *
 * Copyright (c) 1996, 1997
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

/*
 * This file links the ThreadInterface as defined by Tim to the revised
 * version of the internal thread interface ("jthreads").
 *
 * This file implements the functions in ThreadInterface by invoking
 * functions declared in jthread.h
 * It is responsible for unhanding, as well as invoking functions in other
 * parts of the VM.
 *
 * Note that jthreads - while heavily based on Tim's code - can be used 
 * independently of the Kaffe JVM.  That is, they don't have any compile or
 * link time dependencies on it.  This file knows about all the JVM related 
 * types, whereas jthread does not.
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-net.h"
#include "config-signal.h"
#include "jtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "lookup.h"
#include "threads.h"
#include "locks.h"
#include "exception.h"
#include "support.h"
#include "external.h"
#include "gc.h"
#include "md.h"
#include "lerrno.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"
#include "jthread.h"
#if defined(HAVE_SYS_TIMEB_H)
#include <sys/timeb.h>
#endif
#if defined(HAVE_REALTIME_FIFO) || defined(HAVE_REALTIME_RR)
#include <sched.h>
#endif

static void *
thread_malloc(size_t s)
{
	return gc_malloc_fixed_typed(s, GC_ALLOC_THREADCTX);
}

static void
thread_free(void *p)
{
	gc_free_fixed(p);
}

int flag_preemption = true;
int runFinalizerOnExit;

static
void
runfinalizer()
{
	if (runFinalizerOnExit) {
		invokeFinalizer();
	}
}

static
void
Tattach(Hjava_lang_Thread* tid)
{
	sem2posixLock* lk;

        jthread_attach((jthread_t*)&(unhand(tid)->PrivateInfo),
		tid, java_lang_Thread_NORM_PRIORITY, threadStackSize);
	GC_WRITE(tid, unhand(tid)->PrivateInfo);

	lk = (sem2posixLock*)unhand(tid)->sem;
	lk->mux = thread_malloc(sizeof(jmutex));
	lk->cv = thread_malloc(sizeof(jcondvar));
        jmutex_initialise(lk->mux);
        jcondvar_initialise(lk->cv);
}

static
void
Tcreate(Hjava_lang_Thread* tid, void* func)
{
	sem2posixLock* lk;

	lk = (sem2posixLock*)unhand(tid)->sem;
	lk->mux = thread_malloc(sizeof(jmutex));
	lk->cv = thread_malloc(sizeof(jcondvar));
        jmutex_initialise(lk->mux);
        jcondvar_initialise(lk->cv);

        jthread_create((jthread_t*)&(unhand(tid)->PrivateInfo),
		unhand(tid)->priority,
		func,
		unhand(tid)->daemon,
		tid,
		threadStackSize);
	GC_WRITE(tid, unhand(tid)->PrivateInfo);
}

static  
void    
Tprio(Hjava_lang_Thread* tid, jint prio)
{       
	if (unhand(tid)->PrivateInfo != 0) {
		jthread_setpriority((jthread_t)unhand(tid)->PrivateInfo, prio);
	}
}               

static
void
Texit(Hjava_lang_Thread* tid)
{
	jthread_t t;

	t = (jthread_t)unhand(tid)->PrivateInfo;
	unhand(tid)->PrivateInfo = 0;
	thread_free(((sem2posixLock*)unhand(tid)->sem)->mux);
	thread_free(((sem2posixLock*)unhand(tid)->sem)->cv);
	jthread_exit(t);
}

static
void    
Tfinalize(Hjava_lang_Thread* tid)
{
	/* Doesn't do anything - all taken care of in exit */
}       

static          
Hjava_lang_Thread*      
TcurrentJava(void)      
{               
	return GET_COOKIE();
}

static
void
TenterCritSect(void)
{
	intsDisable();
}

static
void
TexitCritSect(void)
{
	intsRestore();
}

#if defined(MEM1) || defined(MEM2)

static          
void
TwalkThreads(void)
{               
	/* this is from gc-incremental.c */
	extern void walkMemory(void*);
	jthread_walkLiveThreads(walkMemory);
}

/*      
 * Walk the thread's internal context.
 */
static
void
TwalkThread(Hjava_lang_Thread* tid)
{
	jthread_t jtid;

	jtid = (jthread_t)unhand(tid)->PrivateInfo;
        if (jtid == 0) {
                return;
	}

	/* Walk the register in the jmpbuf since these values may not
	 * be saved on the stack.
	 */
	walkConservative(&jtid->env, sizeof(jmp_buf));

	/* Walk the stack */
#if defined(STACK_GROWS_UP)
	walkConservative(jtid->stackBase, jtid->stackCurr - jtid->stackBase);
#else
	walkConservative(jtid->stackCurr, jtid->stackEnd - jtid->stackCurr);
#endif
}

#endif

/* called when unrolling exceptions */
static  
void
TnextFrame(stackFrame* fm)
{
#if defined(STACK_NEXT_FRAME)
	STACK_NEXT_FRAME(fm);
        if (!jthread_on_current_stack(fm->return_frame)) {
		fm->return_pc = 0;
		fm->return_frame = 0;
        }
#else
        machineStackFrame* nfm;

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

        if (jthread_on_current_stack(STACK_FRAME(nfm))) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = STACK_FRAME(nfm);
        }
        else {
		fm->return_pc = 0;
		fm->return_frame = 0;
        }
#endif
}

static  
void    
Llock(sem2posixLock* lk)
{
	jmutex_lock(lk->mux);
}

static  
void    
Lunlock(sem2posixLock* lk)
{       
	jmutex_unlock(lk->mux);
}

static
boolean
Lwait(sem2posixLock* lk, jlong timeout)
{
	return (jcondvar_wait(lk->cv, lk->mux, timeout));
}
 
static  
void    
Lsignal(sem2posixLock* lk)
{
	jcondvar_signal(lk->cv, lk->mux);
}       
        
static
int
jthreadedMkdir(const char* path)
{
	return mkdir(path, 0777);
}

static
int
jthreadedFileInfo(const char* path, int type, void* extra)
{
	struct stat buf;
	int r;

	switch (type) {
	case JINFO_EXISTS:
		r = stat(path, &buf);
		if (r == 0) {
			return (1);
		}
		break;

	case JINFO_ISFILE:
		r = stat(path, &buf);
		if (r == 0 && S_ISREG(buf.st_mode)) {
			return (1);
		}
		break;

	case JINFO_ISDIR:
		r = stat(path, &buf);
		if (r == 0 && S_ISDIR(buf.st_mode)) {
			return (1);
		}
		break;

	case JINFO_CANREAD:
		r = access(path, R_OK);
		if (r >= 0) {
			return (1);
		}
		break;

	case JINFO_CANWRITE:
		r = access(path, W_OK);
		if (r >= 0) {
			return (1);
		}
		break;

	case JINFO_LENGTH:
		r = stat(path, &buf);
		if (r == 0) {
			*((jlong*)extra) = buf.st_size;
			return (1);
		}
		break;

	case JINFO_LASTMODIFIED:
		r = stat(path, &buf);
		if (r == 0) {
			*((jlong*)extra) = (jlong)buf.st_mtime * (jlong)1000;
			return (1);
		}
		break;
	}

	return (0);
}

static
int
jthreadedFileInfoFD(int fd, int type, void* extra)
{
	int r;
	int nr;
	struct stat buf;
	off_t cur;
#if !(defined(HAVE_IOCTL) && defined(FIONREAD)) && defined(HAVE_DECLARED_SELECT)
	static struct timeval tm = { 0, 0 };
	fd_set rd;
#endif

	switch (type) {
	case JINFO_LENGTH:
		r = fstat(fd, &buf);
		if (r == 0) {
			*((jlong*)extra) = buf.st_size;
			return (1);
		}
		break;

	case JINFO_AVAILABLE:
		cur = lseek(fd, (off_t)0, SEEK_CUR);
		if (cur != (off_t)-1) {
			if (fstat(fd, &buf) != -1 && S_ISREG(buf.st_mode)) {
				return (buf.st_size - cur);
			}
		}
#if defined(HAVE_IOCTL) && defined(FIONREAD)
		r = ioctl(fd, FIONREAD, &nr);
		if (r >= 0) {
			return (nr);
		}
		break;
#elif defined(HAVE_DECLARED_SELECT)
		/* This uses select() to work out if we can read - but what
		 * happens at the end of file?
		 */
		FD_ZERO(&rd);
		FD_SET(fd, &rd);
		r = select(fd+1, &rd, NULL, NULL, &tm);
		if (r == 1) {
			return (1);
		}
		break;
#endif
	}

	return (0);
}

static
int
jthreadedReaddir(const char* path, jdirEntry** list)
{
        DIR* dir;
        struct dirent* entry;
        jdirEntry* dirlist;
        jdirEntry* mentry;
	int count;

        dir = opendir(path);
        if (dir == 0) {
		*list = 0;
                return (-1);
        }

        dirlist = 0;
        count = 0;
        while ((entry = readdir(dir)) != 0) {
                /* We skip '.' and '..' */
                if (strcmp(".", entry->d_name) == 0 ||
                    strcmp("..", entry->d_name) == 0) {
                        continue;
                }
                mentry = gc_malloc_fixed(sizeof(jdirEntry) + NAMLEN(entry));
                assert(mentry != 0);
                strcpy(mentry->name, entry->d_name);
                mentry->next = dirlist;
                dirlist = mentry;
                count++;
        }
        closedir(dir);

	*list = dirlist;
	return (count);
}

static
char*
Ngethostname(void)
{
	char* buf;

	buf = gc_malloc_fixed(64);
	if (gethostname(buf, 64) == 0) {
		return (buf);
	}
	else {
		return ("localhost");
	}
}

static
int*
Ngethostbyname(char* name)
{
	struct hostent* ent;
	int* addrs;

	ent = gethostbyname(name);
	if (ent == 0) {
		return (0);
	}
	addrs = (int*)gc_malloc_fixed(sizeof(int) * 2);
	addrs[0] = *(int*)ent->h_addr;
	addrs[1] = 0;
	return (addrs);
}

static
char*
Ngethostbyaddr(int addr)
{
	struct hostent* ent;

	ent = gethostbyaddr((char*)&addr, 4, AF_INET);
	if (ent == 0) {
		return (0);
	}
	return (ent->h_name);
}

static
void
Tinit(void)
{
#if defined(HAVE_REALTIME_FIFO) || defined(HAVE_REALTIME_RR)
	struct sched_param sp;
        sp.sched_priority = 10; /* arbitrary */
#if defined(HAVE_REALTIME_FIFO)
	if(sched_setscheduler(getpid(), SCHED_FIFO, &sp) < 0) {
              PerrorAbort("sched_setscheduler");
	}
#else
        if(sched_setscheduler(getpid(), SCHED_RR, &sp) < 0) {
              PerrorAbort("sched_setscheduler");
	}
#endif
#endif	
        jthread_init(flag_preemption,
		java_lang_Thread_MAX_PRIORITY+1,
		java_lang_Thread_MIN_PRIORITY,
		thread_malloc,
		thread_free,
		0);
	jthread_atexit(runfinalizer);
}

static
void
TstackInfo(threadStackInfo* info)
{
	jthread* jtid = jthread_current();
	info->low = jtid->stackBase;
	info->high = jtid->stackEnd;
	info->curr = jtid->stackCurr;
}

static
void
Finit(void)
{
}

static
void
Ninit(void)
{
}

static
void*
Mpagealloc(int size)
{
#if defined(MEM3)
	return (0);
#else
	void* ptr;

#if defined(HAVE_SBRK)

	/* Our primary choice for basic memory allocation is sbrk() which
	 * should avoid any unsee space overheads.
	 */
	for (;;) {
		int missed;
		ptr = sbrk(size);
		if (ptr == (void*)-1) {
			ptr = 0;
			break;
		}
		if ((uintp)ptr % gc_pgsize == 0) {
			break;
		}
		missed = gc_pgsize - ((uintp)ptr % gc_pgsize);
		sbrk(-size + missed);
	}
	if (ptr == 0) {
		return (0);
	}

#elif defined(HAVE_MEMALIGN)

        ptr = memalign(gc_pgsize, size);
	CHECK_OUT_OF_MEMORY(ptr);

#elif defined(HAVE_VALLOC)

        ptr = valloc(size);
	if (ptr == 0) {
		return (0);
	}

#else

	/* Fallback ...
	 * Allocate memory using malloc and align by hand.
	 */
	size += gc_pgsize;
#ifdef malloc
#undef malloc
#endif
        ptr = malloc(size);
	if (ptr == 0) {
		return (0);
	}
	ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -gc_pgsize);

#endif

	return (ptr);
#endif
}

static
jlong
Ktime(void)
{
	jlong tme;
#if defined(HAVE_GETTIMEOFDAY)
	struct timeval tm;
	gettimeofday(&tm, 0);
	tme = (((jlong)tm.tv_sec * (jlong)1000) + ((jlong)tm.tv_usec / (jlong)1000));
#elif defined(HAVE_FTIME)
	struct timeb tm;
	ftime(&tm);
	tme = (((jlong)tm.time * (jlong)1000) + (jlong)tm.millitm);
#elif defined(HAVE__FTIME)
	struct _timeb tm;
	_ftime(&tm);
	tme = (((jlong)tm.time * (jlong)1000) + (jlong)tm.millitm);
#elif defined(HAVE_TIME)
	tme = (jlong)1000 * (jlong)time(0);
#else
	tme = 1;
#endif
	return (tme);
}

/*
 * Define the thread interface.
 */
ThreadInterface Kaffe_ThreadInterface = {

  Tinit,

  Tattach,
  Tcreate,
  jthread_yield,
  Tprio, 
  Texit,
  0,
  Tfinalize,

  TcurrentJava,
  0,

  TenterCritSect,
  TexitCritSect,

#if defined(MEM1) || defined(MEM2)
  TwalkThreads,
  TwalkThread,
#else
  0,
  0,
#endif

  TnextFrame,
  
  TstackInfo,
};

/*
 * Define the lock interface.
 */
LockInterface Kaffe_LockInterface = {

  Llock,
  Lunlock,
  Lwait,
  Lsignal,

  0,
  0,
};

KernelCallInterface Kaffe_KernelCallInterface = {

  jthreadedFileDescriptor,
  jthreadedSelect,	
  jthreadedWaitpid,

  Mpagealloc,
  0, /* pagefree */
  0,
  0,
  0,
  Ktime,
  0,

};

FileSystemCallInterface Kaffe_FileSystemCallInterface = {

  Finit,

  jthreadedOpen,
  jthreadedRead,	
  jthreadedWrite, 
  jthreadedLseek,
  close,
  jthreadedFileInfo,
  jthreadedFileInfoFD,

  jthreadedReaddir,
  
  jthreadedMkdir,
  rmdir,
  rename,
  remove,

};

NetworkCallInterface Kaffe_NetworkCallInterface = {
  Ninit,
  jthreadedSocket,
  bind,
  listen,
  jthreadedConnect,
  jthreadedAccept, 
  jthreadedRead,	
  jthreadedRecvfrom,
  jthreadedWrite, 
  sendto,	
  close,
  setsockopt,	
  getsockopt,
  getsockname, 
  getpeername,
  
  Ngethostname,
  Ngethostbyname,
  Ngethostbyaddr,
  ioctl,

  setreadtimeout,
  getreadtimeout,

};
