/*
 * jthread.c
 * Java thread package - derived from thread-internal.c
 *
 * Internal threading system support
 *
 * Copyright (c) 1996, 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Written by Godmar Back <gback@cs.utah.edu> and 
 *            Tim Wilkinson <tim@transvirtual.com>
 */

#define DBG(s)
#define SDBG(s)
#define SELDBG(s)
#define EDBG(s)
#define SOCKDBG(s)
#define FILEDBG(s)
#define SWITCHDBG(s)
#define LOCKDBG(s)

#include "jthread.h"
#include "support.h"
#include "jnidefs.h"
#include "locks.h"

#if defined(NO_ASYNC_EVENTS)
#undef SIGALRM
#undef SIGVTALRM
#endif
#if defined(NO_ASYNC_IO)
#undef SIGIO
#endif
#if defined(NO_IO_SIGNAL)
#undef SIGIO
#endif

#if defined(ASYNC_DNS)
adns_state adns;
adns_callback adns_vec;
#endif

/* Flags used for threading I/O calls */
#define TH_READ                         0
#define TH_WRITE                        1
#define TH_ACCEPT                       TH_READ
#define TH_CONNECT                      TH_WRITE

/* thread status */
#define THREAD_SUSPENDED                0
#define THREAD_RUNNING                  1
#define THREAD_DEAD                     2

/* thread flags */
#define THREAD_FLAGS_GENERAL            0x00000000
#define THREAD_FLAGS_NOSTACKALLOC       0x00000001
#define THREAD_FLAGS_ALARMTRIGGERED	0x00000002
#define THREAD_FLAGS_ALARM              0x00000004
#define THREAD_FLAGS_USERSUSPEND        0x00000008
#define THREAD_FLAGS_STOPPED        	0x00000010
#define THREAD_FLAGS_DYING        	0x00000020
#define THREAD_FLAGS_BLOCKEDEXTERNAL	0x00000040
#define THREAD_FLAGS_MUX		0x00000080
#define THREAD_FLAGS_WAIT		0x00000100
#define THREAD_FLAGS_READ		0x00000200
#define THREAD_FLAGS_WRITE		0x00000400
#define THREAD_FLAGS_SELECT		0x00000800
#define THREAD_FLAGS_WAITPID		0x00001000
#define THREAD_FLAGS_INTERRUPTABLE	0x00002000
#define THREAD_FLAGS_INTERRUPTED	0x00004000

#define THREAD_FLAGS_DEBUG0		0x10000000
#define THREAD_FLAGS_DEBUG1		0x20000000
#define THREAD_FLAGS_DEBUG2		0x40000000
#define THREAD_FLAGS_DEBUG3		0x80000000

/* Clear these flags when we resume a thread */
#define	THREAD_FLAGS_CLEAR \
	(THREAD_FLAGS_BLOCKEDEXTERNAL | \
	 THREAD_FLAGS_MUX | \
	 THREAD_FLAGS_WAIT | \
	 THREAD_FLAGS_READ | \
	 THREAD_FLAGS_WRITE | \
	 THREAD_FLAGS_SELECT | \
	 THREAD_FLAGS_WAITPID | \
	 THREAD_FLAGS_INTERRUPTABLE)

/*
 * Variables.
 * These should be kept static to ensure encapsulation.
 */
static int preemptive = true;	/* enable preemptive scheduling */
static int talive; 		/* number of threads alive */
static int tdaemon;		/* number of daemons alive */
static void (*runOnExit)();	/* function to run when all non-daemon die */

static jthread**threadQhead;	/* double-linked run queue */ 
static jthread**threadQtail;
static jthread* liveThreads;	/* list of all live threads */
static jthread* alarmList;	/* list of all threads on alarm queue */
static jthread* waitForList;	/* list of all threads waiting for a child */
static jthread* recycledList;

static int maxFd = 0;		/* highest known fd + 1 */
static fd_set readsPending;	/* fds we want to read from */
static fd_set writesPending;	/* fds we want to write to */
static jthread* readQ[FD_SETSIZE];	/* threads blocked on read */
static jthread* writeQ[FD_SETSIZE];	/* threads blocked on write */
static jlong readQtimeout[FD_SETSIZE];	/* threads timeout on read */
static jmutex threadLock;	/* static lock to protect liveThreads etc. */
static int activeThreads;	/* Number of active (non-blocked) threads */

/* from old version, it is not clear that this is still important 
 * see comments near use
 */
static int alarmBlocked;

static int sigPending;		/* flags that says whether a intr is pending */
static int pendingSig[NSIG];	/* array that says which intrs are pending */

static int blockInts;		/* counter that says whether irqs are blocked */
static int needReschedule;	/* is a change in the current thread required */

/*
 * the following variables are set by jthread_init, and show how the
 * threading system is parametrized.
 */
static void *(*allocator)(size_t); 	/* malloc */
static void (*deallocator)(void*);	/* free */
static void (*destructor1)(void*);	/* call when a thread exits */
static int  max_priority;		/* maximum supported priority */
static int  min_priority;		/* minimum supported priority */

/*
 * Info on threaded selects.
 */
static fd_set* select_rd;
static fd_set* select_wr;
static jthread* selectQ;

/* 
 * Note: this comment applied to 0.9.2:
 *
 * The following is so that currentJava() returns NULL before the threading
 * system is initialized - some locks, as in processClass, are acquired before
 * the threading system is initialized (for instance:
 * initialiseKaffe->initBaseClasses->loadStaticClass->processClass
 */
static jthread dummy_initializer = {
	0, /* status */
	0, /* priority */
	0, /* stackCurr */
	0, /* stackBase */
	(void*)0xFFFFFFFF, /* stackEnd */

	/* ... the rest can be zero ... */
};
jthread* currentJThread = &dummy_initializer;	/* the current thread */

#define	INIT_THREAD_ALIGN	(0x2000)

/*
 * Function declarations.
 * Again, keep these static to ensure encapsulation.
 */
static void handleInterrupt(int sig);
static void interrupt(EXCEPTIONPROTO);
static void childDeath();
static void handleIO(int);
static bool blockOnFile(int fd, int op);
static void killThread(jthread *jtid);
static void resumeThread(jthread* jtid);
static void reschedule(void);
static void restore_fds();
static void restore_fds_and_exit();
static void processSignals();
static boolean suspendOnQThread(jthread*, jthread**, jlong);
static void nullException();
static void arithmeticException();
static void catchSignal(int, void*);
void deadlock(void);

/* Some systems need special setups for the exception handling */
#if !defined(EXCEPTIONSTART)
#define EXCEPTIONSTART()
#endif
#if !defined(EXCEPTIONEND)
#define EXCEPTIONEND()
#endif

extern void dumpObjectLocks(void);

/*
 * macros to set and extract stack pointer from jmp_buf
 * make sure SP_OFFSET has the correct value for your architecture!
 */
#define GET_SP(E)       (((void**)(E))[SP_OFFSET])
#define SET_SP(E, V)    ((void**)(E))[SP_OFFSET] = (V)
#define GET_FP(E)       (((void**)(E))[FP_OFFSET])
#define SET_FP(E, V)    ((void**)(E))[FP_OFFSET] = (V)

/* amount of stack space to be duplicated at stack creation time */
#if !defined(STACK_COPY)
#define STACK_COPY      (64*4)
#endif

#if defined(HAVE_SYS_WAIT_H)
#include <sys/wait.h>
#endif

/* Select an alarm system */
#if !defined(NO_ASYNC_EVENTS)
#if defined(HAVE_SETITIMER) && defined(ITIMER_REAL)
#define	MALARM(_mt)							\
	{								\
		struct itimerval tm;					\
		tm.it_interval.tv_sec = 0;				\
		tm.it_interval.tv_usec = 0;				\
		tm.it_value.tv_sec = (_mt) / 1000;			\
		tm.it_value.tv_usec = ((_mt) % 1000) * 1000;		\
		setitimer(ITIMER_REAL, &tm, 0);				\
	}
#elif defined(HAVE_ALARM)
#define	MALARM(_mt)	alarm((int)(((_mt) + 999) / 1000))
#endif
#endif

/*============================================================================
 * interrupt dis/enabling - these functions might get called quite frequently,
 * so we inline them at least here
 */

/*
 * check whether interrupts are disabled
 */
inline
int
intsDisabled()
{       
	return (blockInts > 0);
}       

/*
 * disable interrupts
 *
 * Instead of blocking signals, we increment a counter.
 * If a signal comes in while the counter is non-zero, we set a pending flag
 * and mark the signal as pending.
 *
 * intsDisable may be invoked recursively. (is that really a good idea? - gb)
 */
inline
void 
intsDisable()
{
	blockInts++;
}

/*
 * restore interrupts
 *
 * If interrupts are about to be reenabled, execute the handlers for all
 * signals that are pending.
 */
inline
void
intsRestore()
{ 
	/* DEBUG */
	assert(blockInts >= 1);

	if (blockInts == 1) {   
		if (sigPending) {
			processSignals();
		}
 
		/* reschedule if necessary */
		if (needReschedule == true) {
			reschedule(); 
		}
	}
	blockInts--;
}

/*
 * reenable interrupts, non-recursive version.
 */
inline
void
intsRestoreAll()
{
	blockInts = 1;
	intsRestore();
}


/*============================================================================
 *
 * Functions related to run and alarm queue manipulation
 *
 * note that the functions with 'i' in their name assume that the caller
 * has interrupts disabled.
 */
/*
 * yield to another thread
 */
static
inline
void
internalYield()
{
        int priority = currentJThread->priority; 
   
        if (threadQhead[priority] != 0 && threadQhead[priority] != threadQtail[priority]) {
                /* Get the next thread and move me to the end */
		jthread* firstThread = threadQhead[priority];
                threadQhead[priority] = firstThread->nextQ;  
                threadQtail[priority]->nextQ = firstThread;
                threadQtail[priority] = firstThread;
                firstThread->nextQ = 0;
                needReschedule = true;
        }
}

static
void
addToAlarmQ(jthread* jtid, jlong timeout)
{
	jthread** tidp;
	static int inalarmq;
	jlong ctime;

	assert(intsDisabled());
	assert(inalarmq == 0);
	inalarmq = 1;

	jtid->flags |= THREAD_FLAGS_ALARM;

	/* Get absolute time */
	ctime = currtime();
	jtid->time = timeout + ctime;
	/* If the time overflows then just set it to the max */
	if (jtid->time < 0) {
		jtid->time = 0x7FFFFFFFFFFFFFFFLL;
	}

	/* Find place in alarm list and insert it */
	for (tidp = &alarmList; (*tidp) != 0; tidp = &(*tidp)->nextalarm) {
		if ((*tidp)->time > jtid->time) {
			break;
		}
	}
	jtid->nextalarm = *tidp;
	*tidp = jtid;

	inalarmq = 0;

#if !defined(NO_ASYNC_EVENTS)
	/* If I'm head of alarm list, restart alarm */
	if (tidp == &alarmList) {
		MALARM(jtid->time - ctime);
	}
#endif
}

static
void
removeFromAlarmQ(jthread* jtid)
{
	jthread** tidp;

	assert(intsDisabled());

	jtid->flags &= ~THREAD_FLAGS_ALARM;

	/* Find thread in alarm list and remove it */
	for (tidp = &alarmList; (*tidp) != 0; tidp = &(*tidp)->nextalarm) {
		if ((*tidp) == jtid) {
			(*tidp) = jtid->nextalarm;
			jtid->nextalarm = 0;
			jtid->time = 0;
			break;
		}
	}
}


/*
 * Resume a thread running.
 * This routine has to be called only from locations which ensure
 * run / block queue consistency. There is no check for illegal resume
 * conditions (like explicitly resuming an IO blocked thread). There also
 * is no update of any blocking queue. Both has to be done by the caller
 */
static
void
resumeThread(jthread* jtid)
{
	jthread** ntid;

DBG(	kprintf("resumeThread %x %s\n", jtid, makeJavaArray2CString(unhand((Hjava_lang_Thread*)jtid->jlThread)->name)); )
	intsDisable();

	if (jtid->status != THREAD_RUNNING) {

		if (jtid->flags & THREAD_FLAGS_BLOCKEDEXTERNAL) {
			activeThreads--;
		}
		jtid->flags &= ~THREAD_FLAGS_CLEAR;

		/* Remove from alarmQ if necessary */
		if ((jtid->flags & THREAD_FLAGS_ALARM) != 0) {
			removeFromAlarmQ(jtid);
		}
		/* Remove from lockQ if necessary */
		if (jtid->blockqueue != 0) {
			for (ntid = jtid->blockqueue; *ntid != 0; ntid = &(*ntid)->nextQ) {
				if (*ntid == jtid) {
					*ntid = jtid->nextQ;
					break;
				}
			}
			jtid->blockqueue = 0;
		}

		jtid->nextQ = 0;
		jtid->status = THREAD_RUNNING;

		/* Place thread on the end of its queue */
		if (threadQhead[jtid->priority] == 0) {
			threadQhead[jtid->priority] = jtid;
			threadQtail[jtid->priority] = jtid;
			if (jtid->priority > currentJThread->priority) {
				needReschedule = true;
			}
		}
		else {
			threadQtail[jtid->priority]->nextQ = jtid;
			threadQtail[jtid->priority] = jtid;
		}
	}
SDBG(	else {
		kprintf("Re-resuming 0x%x\n", jtid);
	}							)
	intsRestore();
}

/*
 * Suspend a thread on a queue.
 */
static
boolean
suspendOnQThread(jthread* jtid, jthread** queue, jlong timeout)
{
	jthread** ntid;
	jthread* last;
	boolean ret;

SDBG(	kprintf("suspendOnQThread %p %p (%qd) bI %d\n", 
	jtid, queue, timeout, blockInts); )

	assert(intsDisabled());

	ret = true;
	if (jtid->status != THREAD_SUSPENDED) {
		jtid->status = THREAD_SUSPENDED;

		last = 0;
		for (ntid = &threadQhead[jtid->priority]; *ntid != 0; ntid = &(*ntid)->nextQ) {
			if (*ntid == jtid) {
				/* Remove thread from runq */
				*ntid = jtid->nextQ;
				if (*ntid == 0) {
					threadQtail[jtid->priority] = last;
				}
				jtid->nextQ = 0;

				/* Insert onto tail of lock wait Q */
				if (queue != 0) {
					jtid->blockqueue = queue;
					while (*queue != 0) {
						queue = &(*queue)->nextQ;
					}
					*queue = jtid;
				}

				/* If I have a timeout, insert into alarmq */
				if (timeout > NOTIMEOUT) {
					addToAlarmQ(jtid, timeout);
				}

				/* If I was running, reschedule */
				if (jtid == currentJThread) {
					reschedule();
					/* If I returned with an alarm
					 * triggered or was interrupted,
					 * we'll return false.
					 */
					if ((jtid->flags & (THREAD_FLAGS_ALARMTRIGGERED|THREAD_FLAGS_INTERRUPTED)) != 0) {
						jtid->flags &= ~(THREAD_FLAGS_ALARMTRIGGERED|THREAD_FLAGS_INTERRUPTED);
						ret = false;
					}
					
				}
				break;
			}
			last = *ntid;
		}
	}
SDBG(	else {
		kprintf("Re-suspending 0x%x on %x\n", jtid, *queue);
	}							)

	return (ret);
}

/*
 * Kill thread.
 */
static void 
killThread(jthread *tid)
{
	jthread** ntid;
	jthread* last;

	intsDisable();

	/* allow thread to perform any action before being killed -
	 * such as notifying on the object 
	 */
	if (destructor1) {
		(*destructor1)(tid->jlThread);
	}

DBG(	kprintf("killThread %x kills %x\n", currentJThread, tid); )

	if (tid->status != THREAD_DEAD) {

		/* 
		 * If we're asked to terminate another thread, make it
		 * runnable first before killing it.  Otherwise, it may be
		 * resumed later if it's waiting for an alarm of I/O.
		 *
		 * If it's waiting for a lock, we're still hosed, however,
		 * as the lock is left in an inconsistent state.
		 */
		if (tid->status != THREAD_RUNNING) {
			resumeThread(tid);
		}

		/* Get thread off runq (if it needs it) */
		if (tid->status == THREAD_RUNNING) {

			int prio = tid->priority;
			last = 0;
			for (ntid = &threadQhead[prio]; *ntid != 0; ntid = &(*ntid)->nextQ) {
				if (*ntid == tid) {
					*ntid = tid->nextQ;
					if (*ntid == 0) {
						threadQtail[prio] = last;
					}
					tid->nextQ = 0;
					break;
				}
				last = *ntid;
			}
		}

		/* Run something else */
		if (currentJThread == tid) {
			needReschedule = true;
			blockInts = 1;
		}

		/* Put on recycle list (but not main thread) */
		if ((tid->flags & THREAD_FLAGS_NOSTACKALLOC) == 0) {
			jmutex_lock(&threadLock);
			tid->nextlive = recycledList;
			recycledList = tid;
			jmutex_unlock(&threadLock);
		}

		/* Dead Jim - let the GC pick up the remains */
		tid->status = THREAD_DEAD;
	}

	intsRestore();
}


/*============================================================================
 *
 * Functions related to interrupt handling
 *
 */

static void
processSignals()
{
  int i;
  for (i = 1; i < NSIG; i++) {
	if (pendingSig[i]) {
	  pendingSig[i] = 0;
	  handleInterrupt(i);
	}
  }
  sigPending = 0;
}

/*
 * Handle a signal/interrupt.
 *
 * This is the handler given to catchSignal.
 */
static void
interrupt(EXCEPTIONPROTO)
{
        static int withchild = false;

	EXCEPTIONSTART();

#if defined(SIGCHLD)
        /* This bizzare bit of code handles the SYSV machines which re-throw 
	 * SIGCHLD when you reset the handler.  It also works for those that 
	 * don't.
         * Perhaps there's a way to detect this in the configuration process?
         */
        if (sig == SIGCHLD) {
		if (withchild == true) {
			return;
		}
		withchild = true;
	}
#endif
	/* Re-enable signal - necessary for SysV */
        catchSignal(sig, interrupt);
	withchild = false;

	/*
	 * If ints are blocked, this might indicate an inconsistent state of
	 * one of the thread queues (either alarmList or threadQhead/tail).
	 * We better don't touch one of them in this case and come back later.
	 */
	if (blockInts > 0) {
		pendingSig[sig] = 1;
		sigPending = 1;
		return;
	}
	intsDisable();

	pendingSig[sig] = 0;
	handleInterrupt(sig);

	/*
	 * The next bit is rather tricky.  If we don't reschedule then things
	 * are fine, we exit this handler and everything continues correctly.
	 * On the otherhand, if we do reschedule, we will schedule the new
	 * thread with alarms blocked which is wrong.  However, we cannot
	 * unblock them here incase we have just set an alarm which goes
	 * off before the reschedule takes place (and we enter this routine
	 * recusively which isn't good).  So, we set a flag indicating alarms
	 * are blocked, and allow the rescheduler to unblock the alarm signal
	 * after the context switch has been made.  At this point it's safe.
	 */
	alarmBlocked = true;
	intsRestore();
	alarmBlocked = false;

	EXCEPTIONEND();
}

/*
 * handle a SIGVTALRM alarm.
 *
 * If preemption is disabled, we have the current thread so that it is
 * scheduled in a round-robin fashion with its peers who have the same
 * priority.
 */
static void 
handleVtAlarm()
{
	static int c;

	if (preemptive) {
		intsDisable();
		internalYield();
		intsRestore();
	}

	/*
	 * This is kind of ugly: some fds won't send us SIGIO.
	 * Example: the pseudo-tty driver in FreeBSD won't send a signal
	 * if we blocked on a write because the output buffer was full, and
	 * the output buffer became empty again.
	 *
	 * So we check periodically, every 0.2 seconds virtual time.
	 */
	if (++c % 20 == 0) {
		handleIO(false);
	}
}

/*
 * handle a SIGALRM alarm.
 */
static void 
alarmException()
{
	jthread* jtid;
	jlong time;

	/* Wake all the threads which need waking */
	time = currtime();
	while (alarmList != 0 && alarmList->time <= time) {
		/* Restart thread - this will tidy up the alarm and blocked
		 * queues.
		 */
		jtid = alarmList;
		alarmList = alarmList->nextalarm;
		jtid->nextalarm = 0;
		jtid->time = 0;
		jtid->flags |= THREAD_FLAGS_ALARMTRIGGERED;
		resumeThread(jtid);
	}

#if !defined(NO_ASYNC_EVENTS)
	/* Restart alarm */
	if (alarmList != 0) {
		time = alarmList->time - time;
		MALARM(time);
	}
#endif
}

/*
 * handle an interrupt.
 * 
 * this function is either invoked from within a signal handler, or as the
 * result of intsRestore.
 */
static void 
handleInterrupt(int sig)
{
	switch(sig) {
#if !defined(NO_ASYNC_EVENTS)
	case SIGALRM:
		alarmException();
		break;

	case SIGVTALRM:
		handleVtAlarm();
		break;
#endif
#if !defined(NO_ASYNC_IO)
#if defined(SIGIO)
	case SIGIO:
		handleIO(false);
		break;
#endif
#endif
#if defined(SIGCHLD)
	case SIGCHLD:
		childDeath();
		break;
#endif
#if defined(SIGQUIT)
	case SIGQUIT:
		deadlock();
		break;
#endif
	default:
		kprintf("unknown signal %d\n", sig);
		exit(-1);
	}
}

/*============================================================================
 *
 * Functions dealing with thread contexts and the garbage collection interface
 *
 */

/*
 * Allocate a new thread context and stack.
 */
static jthread*
newThreadCtx(int stackSize)
{
	jthread* ct;

	if (recycledList != 0) {
		ct = recycledList;
		recycledList = recycledList->nextlive;
	}
	else {
		ct = allocator(sizeof(jthread) + stackSize);
	}
	ct->stackBase = (ct + 1);
	ct->stackEnd = ct->stackBase + stackSize;
#if defined(STACK_GROWS_UP)
#if defined(FRAMEALIGN)
        ct->stackBase = (void*)STACKALIGN((int)ct->stackBase);
#endif
	ct->stackCurr = ct->stackBase;
#else
#if defined(FRAMEALIGN)
        ct->stackEnd = (void*)(-FRAMEALIGN & (int)ct->stackEnd);
#endif
	ct->stackCurr = ct->stackEnd;
#endif
	ct->status = THREAD_SUSPENDED;

DBG( 	kprintf("allocating new thread, stack base %p-%p\n", 
	    ct->stackBase, ct->stackEnd); )

	return (ct);
}

/*
 * free a thread context and its stack
 */
void    
jthread_destroy(jthread *jtid)
{
	/* Does nothing - thread recycled by jthread_exit */
}

/*
 * iterate over all live threads
 */
void
jthread_walkLiveThreads(void (*func)(void *jlThread))
{
        jthread* tid;

        for (tid = liveThreads; tid != NULL; tid = tid->nextlive) {
                func(tid->jlThread);
        }
}

/*
 * determine whether an address lies on your current stack frame
 */
int
jthread_on_current_stack(void *bp)
{
        return bp >= currentJThread->stackBase && bp < currentJThread->stackEnd;
}       

/* 
 * XXX this is supposed to count the number of stack frames 
 */
int
jthread_frames(jthread* thrd)
{
        return 0;
}

/*============================================================================
 *
 * Functions for initialization and thread creation
 *
 */

/*
 * Initialize the threading system. 
 */
void
jthread_init(int pre, int maxpr, int minpr,
	void *(*_allocator)(size_t), 
	void (*_deallocator)(void*),
	void (*_destructor1)(void*))
{
	struct itimerval tm;

	/* set stdin, stdout, and stderr in async mode */
	assert(0 == jthreadedFileDescriptor(0));
	assert(1 == jthreadedFileDescriptor(1));
#if (DBG(1) + 1) > 0
	/* for debugging, you leave stderr in blocking mode */
	assert(2 == jthreadedFileDescriptor(2));
#endif
	atexit(restore_fds);
	/*
	 * On some systems, it is essential that we put the fds back
	 * in their non-blocking state
	 */
	catchSignal(SIGINT, restore_fds_and_exit);
	catchSignal(SIGHUP, restore_fds_and_exit);
	catchSignal(SIGTERM, restore_fds_and_exit);

	preemptive = pre;
	max_priority = maxpr;
	min_priority = minpr;
	allocator = _allocator;
	deallocator = _deallocator;
	destructor1 = _destructor1;
	threadQhead = allocator((maxpr + 1) * sizeof (jthread *));
	threadQtail = allocator((maxpr + 1) * sizeof (jthread *));

#if defined(SIGVTALRM)
	if (preemptive) {
		catchSignal(SIGVTALRM, interrupt);
		tm.it_interval.tv_sec = tm.it_value.tv_sec = 0;
		tm.it_interval.tv_usec = tm.it_value.tv_usec = 10000;/* 10 ms */
		setitimer(ITIMER_VIRTUAL, &tm, 0);
	}
#endif
#if !defined(NO_ASYNC_EVENTS)
	catchSignal(SIGALRM, interrupt);
#endif
#if !defined(NO_ASYNC_IO)
#if defined(SIGIO)
	catchSignal(SIGIO, interrupt);
#endif
#endif
#if defined(SIGCHLD)
	catchSignal(SIGCHLD, interrupt);
#endif  
#if defined(SIGSEGV)
	catchSignal(SIGSEGV, nullException);
#endif
#if defined(SIGBUS)
	catchSignal(SIGBUS, nullException);
#endif
#if defined(SIGFPE)
	catchSignal(SIGFPE, arithmeticException);
#endif
#if defined(SIGPIPE)
	catchSignal(SIGPIPE, SIG_IGN);
#endif
#if defined(SIGQUIT)
	catchSignal(SIGQUIT, interrupt);
#endif

#if defined(ASYNC_DNS)
	adns_init(&adns, adns_if_noautosys|adns_if_nosigpipe|adns_if_noerrprint|adns_if_noserverwarn, 0);
#endif
}

/*
 * Initialize the threading system. 
 */
void
jthread_attach(jthread **jtid_slot,
        void *jlmainThread,
	int prio,
	size_t threadStackSize)
{
        jthread *jtid; 
	uintp stkp;

	jtid = newThreadCtx(0);
	if (!jtid) {
                *jtid_slot = jtid;
		return;
        }

        jtid->jlThread = jlmainThread;
        jtid->priority = prio;
        jtid->status = THREAD_SUSPENDED;
        jtid->flags = THREAD_FLAGS_NOSTACKALLOC;
        jtid->func = (void (*)(void*))jthread_init;
        jtid->nextlive = liveThreads;
        jtid->time = 0;
        liveThreads = jtid;

#if defined(STACK_GROWS_UP)
	stkp = (uintp)&jtid & -INIT_THREAD_ALIGN;
	jtid->stackBase = (void*)stkp;
	jtid->stackEnd = jtid->stackBase + mainThreadStackSize;
#if defined(FRAMEALIGN)
        jtid->stackBase = (void*)STACKALIGN((int)jtid->stackBase);
#endif
        jtid->stackCurr = jtid->stackBase;
#else
	stkp = (((uintp)&jtid) + INIT_THREAD_ALIGN - 1) & -INIT_THREAD_ALIGN;
        jtid->stackEnd = (void*)stkp;
        jtid->stackBase = jtid->stackEnd - threadStackSize;
#if defined (FRAMEALIGN)
        jtid->stackEnd = (void*)(-FRAMEALIGN & (int)jtid->stackEnd);
#endif
        jtid->stackCurr = jtid->stackEnd;
#endif
        talive++;
        currentJThread = jtid;
        *jtid_slot = jtid;
        resumeThread(jtid);
}

/*
 * set a function to be run when all non-daemon threads have exited
 */
void 
jthread_atexit(void (*f)())
{
	runOnExit = f;
}

static void
start_this_sucker_on_a_new_frame()
{
	/* all threads start with interrupts turned off */
	intsRestore();
	currentJThread->func(currentJThread->jlThread);
	jthread_exit(currentJThread); 
}


/*
 * create a new jthread
 */
void
jthread_create(jthread **jtid_slot, unsigned char pri, 
               void (*func)(void *), int daemon,
               void *jlThread, size_t threadStackSize)
{
	jthread *jtid; 
	void	*oldstack, *newstack;

	activeThreads++;

	jmutex_lock(&threadLock);
        jtid = newThreadCtx(threadStackSize);
        assert(jtid != 0);      /* XXX */
        jtid->priority = pri;
        jtid->jlThread = jlThread;
        jtid->status = THREAD_SUSPENDED;
        jtid->flags = THREAD_FLAGS_GENERAL|THREAD_FLAGS_BLOCKEDEXTERNAL;

        jtid->nextlive = liveThreads;
        liveThreads = jtid;

        talive++;       
        if ((jtid->daemon = daemon) != 0) {
                tdaemon++;
        }
DBG(	kprintf("creating thread %p, daemon=%d\n", jtid, daemon); )
	jmutex_unlock(&threadLock);

        assert(func != 0); 
        jtid->func = func;

	/* 
	 * set the first jmp point 
	 *
	 * Note that when we return from setjmp in the context of
	 * a new thread, we must no longer access any local variables
	 * in this function.  The reason is that we didn't munge
	 * the base pointer that is used to access these variables.
	 *
	 * To be safe, we immediately call a new function.
	 */
        if (setjmp(jtid->env)) {
		/* new thread */
		start_this_sucker_on_a_new_frame();
		assert(!"Never!");
		/* NOT REACHED */
	} 

	/* set up context for new thread */
	oldstack = GET_SP(jtid->env);

#if defined(STACK_GROWS_UP)
	newstack = jtid->stackBase+STACK_COPY;
	memcpy(newstack-STACK_COPY, oldstack-STACK_COPY, STACK_COPY);
#else /* !STACK_GROWS_UP */
	newstack = jtid->stackEnd-STACK_COPY;
	memcpy(newstack, oldstack, STACK_COPY);
#endif /* !STACK_GROWS_UP */

	SET_SP(jtid->env, newstack);

#if defined(FP_OFFSET)
	/* needed for: IRIX */
	SET_FP(jtid->env, newstack + ((void *)GET_FP(jtid->env) - oldstack));
#endif

        *jtid_slot = jtid;
        resumeThread(jtid);
}

/*============================================================================
 *
 * Functions that are part of the user interface
 *
 */

/*
 * yield to a thread of equal priority
 */
void
jthread_yield()
{
        intsDisable();
        internalYield();
        intsRestore();
}

/* 
 * Check whether a thread is alive.
 *
 * Note that threads executing their onstop function are not alive.
 */
int
jthread_alive(jthread *jtid)
{
        int status = true;
        intsDisable();
        if (jtid == 0 || jtid->status == THREAD_DEAD) {
                status = false;
	}
        intsRestore();
        return status;
}

/*
 * Change thread priority.
 */
void
jthread_setpriority(jthread* jtid, int prio)
{
	jthread** ntid;
	jthread* last;

	if (jtid->status == THREAD_SUSPENDED) {
		jtid->priority = (unsigned char)prio;
		return;
	}

	intsDisable();

	/* Remove from current thread list */
	last = 0;
	for (ntid = &threadQhead[jtid->priority]; *ntid != 0; ntid = &(*ntid)->nextQ) {
		if (*ntid == jtid) {
			*ntid = jtid->nextQ;
			if (*ntid == 0) {
				threadQtail[jtid->priority] = last;
			}
			break;
		}
		last = *ntid;
	}

	/* Insert onto a new one */
	jtid->priority = (unsigned char)prio;
	if (threadQhead[prio] == 0) {
		threadQhead[prio] = jtid;
		threadQtail[prio] = jtid;
	}
	else {
		threadQtail[prio]->nextQ = jtid;
		threadQtail[prio] = jtid;
	}
	jtid->nextQ = 0;

	/* If I was rescheduled, or something of greater priority was,
	 * insist on a reschedule.
	 */
	if (jtid == currentJThread || prio > currentJThread->priority) {
		needReschedule = true;
	}

	intsRestore();
}

/*
 * Have a thread exit.
 */
void
jthread_exit(jthread* ttid)
{
	jthread** ntid;
	jthread* tid;

DBG(	kprintf("jthread_exit %x %s\n", ttid, makeJavaArray2CString(unhand((Hjava_lang_Thread*)ttid->jlThread)->name)); )

	jmutex_lock(&threadLock);

	talive--;
	if (ttid->daemon) {
		tdaemon--;
	}

	/* Remove thread from live list so it can be garbaged */
	for (ntid = &liveThreads; *ntid != 0; ntid = &(*ntid)->nextlive) 
	{
		if (ttid == (*ntid)) {
			(*ntid) = ttid->nextlive;
			break;
		}
	}

	jmutex_unlock(&threadLock);

	/* If we only have daemons left, then we should exit. */
	if (talive == tdaemon) {
DBG( 	kprintf("all done, closing shop\n");	)
		if (runOnExit != 0) {
		    runOnExit();
		}

		for (tid = liveThreads; tid != 0; tid = tid->nextlive) {
		    killThread(tid);
		}
		/* Am I suppose to close things down nicely ?? */
		exit(0);
	}
	else if (ttid != currentJThread) {
		killThread(ttid);
	}
	else {
		for (;;) {
			killThread(ttid);
			intsDisable();
			suspendOnQThread(currentJThread, 0, 1000);
			intsRestore();
		}
	}
}

/*
 * have main thread wait for all threads to finish
 */
void jthread_exit_when_done()
{
        while (talive > 1) {
		jthread_yield();
	}
	jthread_exit(currentJThread);
}

/*
 * Reschedule the thread.
 * Called whenever a change in the running thread is required.
 */
void
reschedule(void)
{
	int i;
	jthread* lastThread;
	int b;
	sigset_t nsig;

	/* A reschedule in a non-blocked context is half way to hell */
	assert(intsDisabled());

	b = blockInts;

	for (;;) {
		for (i = max_priority; i >= min_priority; i--) {
			if (threadQhead[i] != 0) {
				if (threadQhead[i] != currentJThread) {
					lastThread = currentJThread;
					currentJThread = threadQhead[i];

SWITCHDBG( 				kprintf("switch from %p to %p\n", lastThread, currentJThread); )

					if (setjmp(lastThread->env) == 0) {
						lastThread->stackCurr = GET_SP(lastThread->env);
						longjmp(currentJThread->env, 1);
					}

					/* Alarm signal may be blocked - if so
					 * unblock it. - XXX
					 */
					if (alarmBlocked == true) {
						alarmBlocked = false;
						sigemptyset(&nsig);
#if defined(SIGALRM)
						sigaddset(&nsig, SIGALRM);
#endif
#if defined(SIGVTALRM)
						sigaddset(&nsig, SIGVTALRM);
#endif
#if defined(SIGIO)
						sigaddset(&nsig, SIGIO);
#endif
						if(sigprocmask(SIG_UNBLOCK, &nsig, 0) < 0)
							PerrorAbort("sigprocmask");
					}

					/* Restore ints */
					blockInts = b;

					assert(currentJThread == lastThread);

					if ((currentJThread->flags & THREAD_FLAGS_STOPPED) != 0) {
						currentJThread->flags &= ~THREAD_FLAGS_STOPPED;
						postAndThrowException("java.lang.ThreadDeath", 0);
					}
				}
				/* Now kill the schedule */
				needReschedule = false;
				return;
			}
		}

		if (activeThreads == 0) {
			deadlock();
			kprintf("DEADLOCK: all threads are blocked.\n");
			ABORT();
		}

		handleIO(true);
	}
}

/*============================================================================
 * 
 * I/O interrupt related functions
 *
 */


/*
 * Process incoming SIGIO
 * return 1 if select was interrupted
 */
static
void
handleIO(int sleep)
{
	int r;
	fd_set rd;
	fd_set wr;
	fd_set ex;
	jthread* tid;
	jthread* ntid;
	struct timeval tm;
	struct timeval* ptm;
	int i;
	jlong timeout;
	int mfd;

	assert(intsDisabled());

SELDBG(	kprintf("handleIO(%s)\n", sleep ? "sleep" : "poll");		)

	FD_COPY(&readsPending, &rd);
	FD_COPY(&writesPending, &wr);
	FD_ZERO(&ex);
	mfd = maxFd;
	ptm = &tm;

	tm.tv_usec = 0;
	tm.tv_sec = 0;
	if (sleep) {
		/* If we're got not pending alarms then sleep for 10 seconds. */
		if (alarmList == 0) {
			tm.tv_usec = 0;
			tm.tv_sec = 10;
		}
		else {
			timeout = alarmList->time - currtime();
			if (timeout > 0) {
				tm.tv_usec = (timeout % 1000) * 1000;
				tm.tv_sec = timeout / 1000;
			}
			else {
				/* If we've missed an alarm signal then
				 * we could get here.  We'll just pretend
				 * an alarm just happened.
				 */
				alarmException();
			}
		}
	}

	/* If someone is doing a select, add in those file handles */
	if (selectQ != 0) {
SELDBG(		kprintf("Adding threaded select fds: ");	)
		for (i = 0; i < maxFd; i++) {
			if (FD_ISSET(i, select_rd)) {
SELDBG(				kprintf("r%d ", i);		)
				FD_SET(i, &rd);
			}
			if (FD_ISSET(i, select_wr)) {
SELDBG(				kprintf("w%d ", i);		)
				FD_SET(i, &wr);
			}
		}
SELDBG(		kprintf("\n");					)
	}

	/*
	 * figure out which fds are ready
	 */
#if defined(ASYNC_DNS)
	adns_beforeselect(adns, &mfd, &rd, &wr, &ex, &ptm, 0, 0);
	r = select(mfd, &rd, &wr, &ex, ptm);
	adns_afterselect(adns, mfd, &rd, &wr, &ex, 0);
	if (adns_vec != 0) {
		for (;;) {
			adns_answer* answer;
			void* context;
			int ar;
			int ip;
			char* name;
			adns_query query;

			/* Look for any completed query */
			query = 0;
			answer = 0;
			context = 0;
			ar = adns_check(adns, &query, &answer, &context);
			if (ar != 0) {
				break;
			}
			name = 0;
			ip = 0;
			if (answer->status == adns_s_ok) {
				if (answer->type == adns_r_ptr) {
					name = *answer->rrs.str;
				}
				else if (answer->type == adns_r_a) {
					ip = GET_ADDRESS(*answer->rrs.inaddr);
				}
			}
			(*adns_vec)(context, name, ip);
		}
	}
	else
#else
	r = select(mfd, &rd, &wr, &ex, ptm);
#endif

	if (sigPending) {
		processSignals();
	}
	if (r <= 0) {
		return;
	}

SELDBG(	kprintf("Select returns %d\n", r);			)

	for (i = 0; i < maxFd; i++) {
		if (readQ[i] != 0 && FD_ISSET(i, &rd)) {
			needReschedule = true;
			for (tid = readQ[i]; tid != 0; tid = ntid) {
				ntid = tid->nextQ;
				tid->blockqueue = 0;
				resumeThread(tid);
			}
			readQ[i] = 0;
		}
		if (writeQ[i] != 0 && FD_ISSET(i, &wr)) {
			needReschedule = true;
			for (tid = writeQ[i]; tid != 0; tid = ntid) {
				ntid = tid->nextQ;
				tid->blockqueue = 0;
				resumeThread(tid);
			}
			writeQ[i] = 0;
		}
	}


	/* If someone is waiting on a threaded select, fixup the
	 * select sets and wake it.
	 */
	if (selectQ != 0) {
SELDBG(		kprintf("Checking threaded select fds: ");	)
		for (i = 0; i < maxFd; i++) {
			if ((FD_ISSET(i, &rd) && FD_ISSET(i, select_rd)) ||
			    (FD_ISSET(i, &wr) && FD_ISSET(i, select_wr))) {
SELDBG(				kprintf("found %d", i);		)
				for (i = 0; i < maxFd; i++) {
					if (!FD_ISSET(i, &rd)) {
						FD_CLR(i, select_rd);
					}
					if (!FD_ISSET(i, &wr)) {
						FD_CLR(i, select_wr);
					}
				}

				/* Wake the selecting thread */
				needReschedule = true;
				selectQ->blockqueue = 0;
				resumeThread(selectQ);
				selectQ = 0;
				break;
			}
		}
SELDBG(		kprintf("\n");					)
	}
}

#if !defined(NO_ASYNC_IO)
/*
 * A file I/O operation could not be completed. Sleep until we are woken up
 * by the SIGIO handler.
 *
 * Interrupts are disabled on entry and exit.
 * fd is assumed to be valid.
 */
static
bool
blockOnFile(int fd, int op)
{
	bool ret;

FILEDBG(	kprintf("blockOnFile(%d,%s)\n", fd, op == TH_READ ? "r":"w"); )

	assert(intsDisabled());

	activeThreads++;

	if (fd >= maxFd) {
		maxFd = fd + 1;
	}
	if (op == TH_READ) {
		FD_SET(fd, &readsPending);
		currentJThread->flags |= THREAD_FLAGS_BLOCKEDEXTERNAL|THREAD_FLAGS_READ|THREAD_FLAGS_INTERRUPTABLE;
		ret = suspendOnQThread(currentJThread, &readQ[fd], readQtimeout[fd]);
		FD_CLR(fd, &readsPending);
	}
	else {
		FD_SET(fd, &writesPending);
		currentJThread->flags |= THREAD_FLAGS_BLOCKEDEXTERNAL|THREAD_FLAGS_WRITE|THREAD_FLAGS_INTERRUPTABLE;
		ret = suspendOnQThread(currentJThread, &writeQ[fd], NOTIMEOUT);
		FD_CLR(fd, &writesPending);
	}

	return (ret);
}
#endif

/*============================================================================
 * 
 * locking subsystem
 *
 */

void 
jmutex_initialise(jmutex *lock)
{
	lock->holder = NULL;
	lock->waiting = NULL;
}

void
jmutex_lock(jmutex *lock)
{
	intsDisable();
	while (lock->holder != NULL) {
		currentJThread->flags |= THREAD_FLAGS_MUX;
		suspendOnQThread(currentJThread, &lock->waiting, NOTIMEOUT);
	}
	lock->holder = currentJThread;
	intsRestore();
}

void
jmutex_unlock(jmutex *lock)
{
	intsDisable();
	lock->holder = NULL;
	if (lock->waiting != 0) {
		jthread* tid;
		tid = lock->waiting;
		lock->waiting = tid->nextQ;
		assert(tid->status != THREAD_RUNNING);
		tid->blockqueue = 0;
		resumeThread(tid);
	}
	intsRestore();
}

void
jcondvar_initialise(jcondvar *cv)
{
	cv = NULL;
}

boolean
jcondvar_wait(jcondvar *cv, jmutex *lock, jlong timeout)
{
	boolean ret;

	intsDisable();

	/* give up mutex */
	lock->holder = NULL;			
	if (lock->waiting != NULL) {
		jthread* tid;
		tid = lock->waiting;
		lock->waiting = tid->nextQ;
		assert(tid->status != THREAD_RUNNING);
		tid->blockqueue = 0;
		resumeThread(tid);
	}
	if (timeout != 0) {
		activeThreads++;
		currentJThread->flags |= THREAD_FLAGS_BLOCKEDEXTERNAL;
	}
	/* wait to be signaled */
	currentJThread->flags |= THREAD_FLAGS_WAIT;
	ret = suspendOnQThread(currentJThread, cv, timeout);
	/* reacquire mutex */
	while (lock->holder != NULL) {
		currentJThread->flags |= THREAD_FLAGS_MUX;
		suspendOnQThread(currentJThread, &lock->waiting, NOTIMEOUT);
	}
	lock->holder = currentJThread;
	intsRestore();

	return (ret);
}

void
jcondvar_signal(jcondvar *cv, jmutex *lock)
{
	intsDisable();
	if (*cv != NULL) {
		jthread* tid;
		/* take off condvar queue */
		tid = *cv;
		*cv = tid->nextQ;
		/* put on lock queue */
		tid->nextQ = lock->waiting;
		lock->waiting = tid;
	}
	intsRestore();
}

void
jcondvar_broadcast(jcondvar *cv, jmutex *lock)
{
	intsDisable();
	if (*cv != NULL) {
		/* splice the lists `*cv' and `lock->waiting' */

		jthread** tidp;
		/* advance to last element in cv list */
		for (tidp = cv; *tidp != 0; tidp = &(*tidp)->nextQ)
			;
		(*tidp) = lock->waiting;
		lock->waiting = *cv;
		*cv = NULL;
	}
	intsRestore();
}

/*============================================================================
 * 
 * I/O routines that are exported to the user
 *
 */

/*
 * Create a threaded file descriptor.
 *
 * We try various fcntl and ioctl to put the file descriptor in non-blocking
 * mode and to enable asynchronous notifications.  If it fails, well too bad.
 */
int
jthreadedFileDescriptor(int fd)
{
#if !defined(NO_ASYNC_IO)
	int on = 1;
	int r;

	/* Make non-blocking */
	r = fcntl(fd, F_GETFL, 0);
	fcntl(fd, F_SETFL, r | O_NONBLOCK 
#if defined(O_ASYNC)
		| O_ASYNC
#elif defined(FASYNC)
		| FASYNC
#endif
	);
#if defined(FIOSSAIOSTAT)
	/* on hpux */
	ioctl(fd, FIOSSAIOSTAT, &on);
#elif defined(FIOASYNC)
	/* 
	 * On some systems, this will not work if a terminal fd is redirected.
	 * (Solaris sets errno to ENXIO in this case.
	 */
	ioctl(fd, FIOASYNC, &on);
#endif

#if !(defined(O_ASYNC) || defined(FIOASYNC) ||  \
      defined(FASYNC) || defined(FIOSSAIOSTAT))
#error	Could not put socket in async mode
#endif

#if defined(F_SETOWN)
	/* Allow socket to signal this process when new data is available */
	/* On some systems, this will flag an error if fd is not a socket */
	fcntl(fd, F_SETOWN, getpid());
#endif
#endif
	readQtimeout[fd] = NOTIMEOUT;
	return (fd);
}

/*
 * In SVR4 systems (notably AIX and HPUX 9.x), putting a file descriptor 
 * in non-blocking mode affects the actual terminal file.  
 * Thus, the shell we see the fd in
 * non-blocking mode when we exit and log the user off.
 *
 * Under Solaris, this happens if you use FIONBIO to get into non-blocking 
 * mode.  (as opposed to O_NONBLOCK)
 */
static void
restore_fds()
{
	int i;
	/* clear non-blocking flag for file descriptor stdin, stdout, stderr */
	for (i = 0; i < 3; i++)
	    if(fcntl(i, F_SETFL, fcntl(i, F_GETFL, 0) & ~O_NONBLOCK) < 0)
	    	PerrorAbort("fcntl");
}

static void
restore_fds_and_exit()
{
       
	restore_fds();
	/* technically, we should restore the original handler and rethrow
	 * the signal.
	 */
	exit(-1);		/* XXX */
}

/*
 * Threaded socket create.
 */
int
jthreadedSocket(int stream)
{
	int fd;
	if (stream == 0) {
		fd = socket(AF_INET, SOCK_DGRAM, 0);
	}
	else {
		fd = socket(AF_INET, SOCK_STREAM, 0);
	}
	return (jthreadedFileDescriptor(fd));
}

/*
 * Threaded file open.
 */
int
jthreadedOpen(const char* path, int mode)
{
	int fd;

	switch (mode) {
	case JOPEN_READ:
		fd = open(path, O_RDONLY|O_BINARY);
		break;
	case JOPEN_WRITE:
		fd = open(path, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0666);
		break;
	case JOPEN_APPEND:
		fd = open(path, O_RDWR|O_CREAT|O_APPEND|O_BINARY, 0666);
		break;
	case JOPEN_READWRITE:
		fd = open(path, O_RDWR|O_CREAT|O_BINARY, 0666);
		break;
	default:
		fd = -1;
		break;
	}

	if (fd < 0) {
		return (fd);
	}
	else {
		return (jthreadedFileDescriptor(fd));
	}
}

int
jthreadedLseek(int fd, off_t pos, int whence)
{
	return (lseek(fd, pos, whence));
}

/*
 * Threaded socket connect.
 */
int
jthreadedConnect(int fd, struct sockaddr* addr, size_t len)
{
#if !defined(NO_ASYNC_IO)
	int r;

	intsDisable();	
	for (;;) {
SOCKDBG(printf("Connecting fd=%d\n", fd);			)
		r = connect(fd, addr, len);
		if (r == 0) {
			break;
		}
		if (r < 0 && !(errno == EINPROGRESS || errno == EINTR)) {
		      break;
		}
		if (blockOnFile(fd, TH_CONNECT) == false) {
			intsRestore();
			unhand(Kaffe_TI_currentJava())->interrupting = false;
			postException("java.io.InterruptedIOException", "Connect aborted");
			return (0);
		}
	}
	intsRestore();

SOCKDBG(printf("Connected fd=%d\n", fd);			)

	return (r);
#else
	return (connect(fd, addr, len));
#endif
}

/*
 * Threaded socket accept.
 */
int
jthreadedAccept(int fd, struct sockaddr* addr, size_t* len)
{
#if !defined(NO_ASYNC_IO)
	int r;

	intsDisable();
	for (;;) {
		r = accept(fd, addr, len);
		if (r >= 0 || !(errno == EWOULDBLOCK || errno == EINTR))
			break;
		if (blockOnFile(fd, TH_ACCEPT) == false) {
			intsRestore();
			unhand(Kaffe_TI_currentJava())->interrupting = false;
			postException("java.io.InterruptedIOException", "Accept timeout/aborted");
			return (0);
		}
	}
	intsRestore();
SOCKDBG(printf("Accepted on fd=%d, r=%d\n", fd, r);			)
	return (r >= 0 ? jthreadedFileDescriptor(r) : r);
#else
	return (accept(fd, addr, len));
#endif
}

/*
 * Threaded block until available
 */
void
jthreadedPollFd(int fd)
{
#if !defined(NO_ASYNC_IO)
	int r;

	intsDisable();
	/* jthreadedFileinfoFD is static, jsyscalls.h prototypes undefed.. */
	for ( ;; ) {
		r = Kaffe_FileSystemCallInterface._infoFD(fd, JINFO_AVAILABLE, 0);
		if ( r > 0 ) {
			break;
		}
		blockOnFile( fd, TH_READ);
	}
	intsRestore();
#else
	Kaffe_FileSystemCallInterface._infoFD(fd, JINFO_AVAILABLE, 0);
#endif
}

/*
 * Threaded read
 */
ssize_t
jthreadedRead(int fd, void* buf, size_t len)
{
#if !defined(NO_ASYNC_IO)
	ssize_t r;

	intsDisable();

	/* It would appear that we should block on a read of length
	 * zero if there's still data to be read but none happens
	 * to be available (to JDK 1.1.6 spec)
	 */
	if (len == 0) {
		blockOnFile(fd, TH_READ);
		return (0);
	}

	for (;;) {
SOCKDBG(printf("Reading %d on fd=%d\n", len, fd);			)
		r = read(fd, buf, len);
		if (r >= 0 || !(errno == EWOULDBLOCK || errno == EINTR)) {
			break;
		}
		if (blockOnFile(fd, TH_READ) == false) {
			intsRestore();
			unhand(Kaffe_TI_currentJava())->interrupting = false;
			postException("java.io.InterruptedIOException", "Read timeout/aborted");
			return (0);
		}
	}
	intsRestore();
SOCKDBG(printf("Read %d on fd=%d\n", r, fd);				)
	return r;
#else
	return (read(fd, buf, len));
#endif
}

/*
 * Threaded write
 */
ssize_t
jthreadedWrite(int fd, const void* buf, size_t len)
{
#if !defined(NO_ASYNC_IO)
	ssize_t r = 1;
	const void* ptr;

	ptr = buf;

	intsDisable();
	while (len > 0 && r > 0) {
SOCKDBG(printf("Writing %d on fd=%d\n", len, fd);			)
		r = (ssize_t)write(fd, ptr, len);
		if (r >= 0) {
			ptr += r;
			len -= r;
			r = ptr - buf;
			continue;
		}
		if (!(errno == EWOULDBLOCK || errno == EINTR)) {
			break;
		}
		if (blockOnFile(fd, TH_WRITE) == false) {
			intsRestore();
			unhand(Kaffe_TI_currentJava())->interrupting = false;
			postException("java.io.InterruptedIOException", "Write aborted");
			return (ptr - buf);
		}
		r = 1;
	}
	intsRestore();
SOCKDBG(printf("Wrote %d on fd=%d\n", r, fd);				)
	return r; 
#else
	return ((ssize_t)write(fd, buf, len));
#endif
}

/*
 * Threaded recvfrom 
 */
ssize_t 
jthreadedRecvfrom(int fd, void* buf, size_t len, int flags, 
	struct sockaddr* from, int* fromlen)
{
#if !defined(NO_ASYNC_IO)
	ssize_t r;
 
	intsDisable();
	for (;;) {
SOCKDBG(printf("Receiving %d on fd=%d\n", len, fd);			)
		r = recvfrom(fd, buf, len, flags, from, fromlen);
		if (r >= 0 || !(errno == EWOULDBLOCK || errno == EINTR))
			break;
		if (blockOnFile(fd, TH_READ) == false) {
			intsRestore();
			unhand(Kaffe_TI_currentJava())->interrupting = false;
			postException("java.io.InterruptedIOException", "Receive abort/timeout");
			return (0);
		}
	}
	intsRestore();
SOCKDBG(printf("Received %d on fd=%d\n", r, fd);			)
	return r;
#else
	return (recvfrom(fd, buf, len, flags, from, fromlen));
#endif
}

/*
 * Threaded select.
 */
int
jthreadedSelect(int nfds, fd_set* readfds, fd_set* writefds, fd_set* exceptfds, struct timeval* timeout)
{
	static fd_set nullfds[1];
	int i;
	int count;

	intsDisable();

	/* Only one at once */
	if (selectQ != 0) {
		intsRestore();
		return (-1);
	}

	/* If we're haven't got an fd_set use a blank one */
	if (readfds == 0) {
		readfds = nullfds;
	}
	if (writefds == 0) {
		writefds = nullfds;
	}

	select_rd = readfds;
	select_wr = writefds;
	if (nfds > maxFd) {
		maxFd = nfds;
	}

	activeThreads++;
	currentJThread->flags |= THREAD_FLAGS_BLOCKEDEXTERNAL|THREAD_FLAGS_SELECT;

	suspendOnQThread(currentJThread, &selectQ, NOTIMEOUT);

	/* Count the file descriptors which are now active */
	count = 0;
	for (i = 0; i < nfds; i++) {
		if (FD_ISSET(i, readfds)) {
			count++;
		}
		else if (FD_ISSET(i, writefds)) {
			count++;
		}
	}

	intsRestore();

	return (count);
}

void
jthreadedInterrupt(struct Hjava_lang_Thread* tid)
{
	jthread_t jtid;

	jtid = (jthread_t)unhand(tid)->PrivateInfo;

	if (jtid != 0 && (jtid->flags & THREAD_FLAGS_INTERRUPTABLE) != 0) {
		jtid->flags |= THREAD_FLAGS_INTERRUPTED;
		resumeThread(jtid);
	}
}

/*
 * Set the socket timeout.
 */
int
setreadtimeout(int fd, int time)
{
	if (time == 0) {
		readQtimeout[fd] = NOTIMEOUT;
	}
	else {
		readQtimeout[fd] = (jlong)time;
	}
	return (0);
}

/*
 * Get the socket timeout.
 */
int
getreadtimeout(int fd)
{
	jlong time;

	time = readQtimeout[fd];
	if (time == NOTIMEOUT) {
		return (0);
	}
	else {
		return ((int)time);
	}
}

/*============================================================================
 * 
 * Routines dealing with Process::waitFor. 
 *
 */

/*
 * Child process has died.
 */
static
void
childDeath()
{
	if (waitForList) {
		resumeThread(waitForList);
	}
}

/* 
 * Wait for a child process.
 */
int
jthreadedWaitpid(int wpid, int* status, int options)
{
#if defined(HAVE_WAITPID)
	int npid;

DBG(	kprintf("waitpid %d current=%p\n", wpid, currentJThread); )

	intsDisable();
	for (;;) {
		npid = waitpid(wpid, status, options|WNOHANG);
		if (npid > 0) {
			break;
		}
		activeThreads++;
		currentJThread->flags |= THREAD_FLAGS_BLOCKEDEXTERNAL|THREAD_FLAGS_WAITPID;

		suspendOnQThread(currentJThread, &waitForList, NOTIMEOUT);
	}
	intsRestore();
	return (npid);
#else
	return -1;
#endif
}

#if defined(SIGSEGV) || defined(SIGBUS)
/*
 * Null exception - catches bad memory accesses.
 */
static
void
nullException(EXCEPTIONPROTO)
{
	EXCEPTIONSTART();

#if !defined(DEBUG)
	catchSignal(sig, nullException);
#endif
	EXCEPTIONFIXRETURN();

EDBG(	kprintf("NullPointerException %p\n", EXCEPTIONPC());	)

	postAndThrowException("java.lang.NullPointerException", "");

	EXCEPTIONEND();
}
#endif

#if defined(SIGFPE)
/*
 * Division by zero.
 */
static
void
arithmeticException(EXCEPTIONPROTO)
{
	EXCEPTIONSTART();

#if !defined(DEBUG)
	catchSignal(sig, arithmeticException);
#endif

	EXCEPTIONFIXRETURN();

EDBG(	kprintf("ArithmeticException\n");	)

	postAndThrowException("java.lang.ArithmeticException", "");

	EXCEPTIONEND();
}
#endif

#if defined(SIGSEGV) || defined(SIGBUS) || defined(SIGFPE) || defined(SIGPIPE)
/*
 * Setup a signal handler.
 */
static
void
catchSignal(int sig, void* handler)
{
	sigset_t nsig;

#if defined(HAVE_SIGACTION)

	struct sigaction newact;

	newact.sa_handler = (SIG_T)handler;
	sigemptyset(&newact.sa_mask);
	newact.sa_flags = 0;
#if defined(SA_SIGINFO)
	newact.sa_flags |= SA_SIGINFO;
#endif
#if defined(SA_RESTART)
	newact.sa_flags |= SA_RESTART;
#endif
	if(sigaction(sig, &newact, NULL) < 0)
		PerrorAbort("sigaction");

#elif defined(HAVE_SIGNAL)

	signal(sig, (SIG_T)handler);

#else
        ABORT();
#endif

	/* Unblock this signal */
	sigemptyset(&nsig);
	sigaddset(&nsig, sig);
	if(sigprocmask(SIG_UNBLOCK, &nsig, 0) < 0)
		PerrorAbort("sigprocmask");
}
#endif

static
char*
nameThread(Hjava_lang_Thread* tid)
{
	static char buf[80];
	int i;
	HArrayOfChar* name;

	name = unhand(tid)->name;
	for (i = 0; i < sizeof buf - 1 && i < ARRAY_SIZE(name); i++) {
		buf[i] = ((short*)ARRAY_DATA(HArrayOfShort, name))[i];
	}
	buf[i] = 0;

	return (buf);
}

static
char*
printflags(int f)
{
	int i;
	static char b[256];
	static struct {
		int flagvalue;
		char *flagname;
	} flags[] = {
		{ THREAD_FLAGS_GENERAL, "GENERAL" },
		{ THREAD_FLAGS_NOSTACKALLOC, "NOSTACKALLOC" },
		{ THREAD_FLAGS_ALARM, "ALARM" },
		{ THREAD_FLAGS_ALARMTRIGGERED, "ALARMTRIGGERED" },
		{ THREAD_FLAGS_USERSUSPEND, "USERSUSPENDED" },
		{ THREAD_FLAGS_STOPPED, "STOPPED" },
		{ THREAD_FLAGS_DYING, "DYING" },
		{ THREAD_FLAGS_BLOCKEDEXTERNAL, "BLOCKEDEXTERNAL" },
		{ THREAD_FLAGS_MUX, "MUTEX" },
		{ THREAD_FLAGS_WAIT, "WAIT" },
		{ THREAD_FLAGS_READ, "READ" },
		{ THREAD_FLAGS_WRITE, "WRITE" },
		{ THREAD_FLAGS_SELECT, "SELECT" },
		{ THREAD_FLAGS_WAITPID, "WAITPID" },
		{ THREAD_FLAGS_INTERRUPTED, "INTERRUPTED" },
		{ THREAD_FLAGS_INTERRUPTABLE, "INTERRUPTABLE" },
		{ THREAD_FLAGS_DEBUG0, "DEBUG0" },
		{ THREAD_FLAGS_DEBUG1, "DEBUG1" },
		{ THREAD_FLAGS_DEBUG2, "DEBUG2" },
		{ THREAD_FLAGS_DEBUG3, "DEBUG3" },
		{ 0, NULL }
	};

        b[0] = 0;
	for (i = 0; flags[i].flagname != 0; i++) {
                if (f & flags[i].flagvalue) {
                        strcat(b, flags[i].flagname);
                        strcat(b, " ");
                }
        }
	if (b[0] == 0) {
		return ("<NONE>");
	}
	else {
		return (b);
	}
}

static
void
dumpStackTrace(jthread* tid)
{
#if defined(FP_OFFSET)
	jthread* old;
	Hjava_lang_Throwable* eobj;

	eobj = (Hjava_lang_Throwable*)Kaffe_AllocObject(0, Kaffe_FindClass(0, "java.lang.Throwable"));
	old = currentJThread;
	currentJThread = tid;
	unhand(eobj)->backtrace = (Hjava_lang_Object*)buildStackTraceFrom((void*)GET_FP(tid->env), 1);
	currentJThread = old;
	kprintf("   stacktrace:\n");
	printStackTrace(eobj, 0, 1);
#endif
}

void
deadlock(void)
{
	jthread* tid;
	jthread* t;

#if defined(DEBUG)
	dumpObjectLocks();
#endif

	kprintf("Dumping threads:\n");

        for (tid = liveThreads; tid != NULL; tid = tid->nextlive) {
		kprintf("'%s' ", nameThread((Hjava_lang_Thread*)(tid->jlThread)));
		kprintf("tid %p status %s flags (%x) %s\n", tid,
			(tid->status == THREAD_SUSPENDED ? "SUSPENDED" :
			 tid->status == THREAD_RUNNING ? "RUNNING" :
			 tid->status == THREAD_DEAD ? "DEAD" : "UNKNOWN"),
			tid->flags, printflags(tid->flags));
		if (tid->blockqueue != 0) {
			kprintf("   blocked @ %p (", tid->blockqueue);
			t = *tid->blockqueue;
			for (; t != 0; t = t->nextQ) {
				kprintf("%p->", t);
			}
			kprintf("END)\n");
		}
#if defined(DEBUG)
		kprintf("   debug val: %x\n", tid->dval);
#endif
		dumpStackTrace(tid);
        }
}
