/*
 * jthread.h
 *
 * Thread support using internal system.
 *
 * Copyright (c) 1996, 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Written by Godmar Back <gback@cs.utah.edu> and
 *            Tim Wilkinson <tim@transvirtual.com>
 */

#ifndef __jthread_h
#define __jthread_h

#include "config.h"
#include "config-setjmp.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-net.h"
#include "config-signal.h"
#include "jtypes.h"
#include "objects.h"
#include "gtypes.h"
#if defined(DEBUG)
#include "classMethod.h"
#endif
#include "threads.h"
#include "support.h"
#include "md.h"
#include "lerrno.h"
#include "java_lang_Throwable.h"
#include "engine.h"
#include "exception.h"

/* 
 * make sure jsyscall doesn't redefine open, read, etc. 
 * XXX - this is not very clean
 */
#define NOUNIXPROTOTYPES
#include "jsyscall.h"

/*
 * a mutex is represented by a two-word structure
 */
typedef struct _jmutex
{
        struct _jthread *holder;
        struct _jthread *waiting;
} jmutex;

/*
 * a condition variable is simply expressed as a list of threads
 * waiting to be notified
 */
typedef struct _jthread *jcondvar;

/*
 * This is our internal structure representing the "native" threads.
 * This used to be called "ctx".
 */
typedef struct _jthread {
	unsigned char			status;
	unsigned char			priority;
	void*				stackCurr;
	void*				stackBase;
	void*				stackEnd;
	jlong				time;
	struct _jthread*		nextQ;
	struct _jthread*		nextlive;
	struct _jthread*		nextalarm;
	struct _jthread**		blockqueue;
	unsigned int			flags;
	void				(*func)(void *);
	int				daemon;
#if defined(DEBUG)
	void*				waitobj;
	int				dval;
#endif

	/* this one is simply thread specific data, or a cookie -
	 * used to hold the current Java thread
	 */
	void*				jlThread;
	jmp_buf				env;
	/* 
	 * note that this causes gdb under Solaris to crash when trying to
	 * print a struct jthread
	 */
	/* for alignment (Gcc extension) */
	double				align[0];
} jthread;
typedef jthread* jthread_t;

#define GET_COOKIE()	(jthread_current()->jlThread)

extern jthread_t currentJThread;

/****************************************************************************
 *
 * The following functions form the interface provided by the 
 * revised internal threading system.
 */

/* 
 * initialize the threading system
 */
void
jthread_init(
	int preemptive,			/* preemptive scheduling */
	int maxpr, 			/* maximum priority */
	int minpr, 			/* minimum priority */
        void *(*_allocator)(size_t),	/* memory allocator */
	void (*_deallocator)(void*),	/* memory deallocator */
	void (*_destructor1)(void*));	/* called when a thread exits */ 

/*
 * Attach the current thread.
 */
void
jthread_attach(
	jthread**,			/* where to store jthread pointer */
	void*,				/* cookie */
	int,				/* prio */
	size_t);			/* stack size */

/*
 * create a thread with a given priority
 */
void
jthread_create(jthread**,
               unsigned char pri,       /* initial priority */
               void (*func)(void *), 		/* start function */
               int daemon, 			/* is this thread a daemon? */
               void *jlThread, 		/* cookie for this thread */
               size_t threadStackSize);	/* stack size to be allocated */

/*
 * set a function to be run when last non-daemon dies 
 * this is used to run the finalizer on exit.
 */
void 	jthread_atexit(void (*f)());

/*
 * walk all live threads, and invoke `func', passing in their cookie
 * this is used with `func' set to walkMemory
 */
void 	jthread_walkLiveThreads(void (*func)(void *jlThread));

/* 
 * destroy this jthread structure 
 */
void	jthread_destroy(jthread_t jtid);

/*
 * set the priority of a thread 
 */
void	jthread_setpriority(jthread_t jtid, int prio);

/*
 * yield to another thread
 */
void 	jthread_yield();

/*
 * sleep for time milliseconds
 */
void 	jthread_sleep(jlong time);

/* 
 * return the current thread 
 */
static inline jthread_t
jthread_current() 
{ 
	return currentJThread; 
}

/* 
 * count the number of stack frames - unimplemented 
 */
int 	jthread_frames(jthread_t thrd);

/* 
 * return whether this thread is alive or not
 */
int 	jthread_alive(jthread_t jtid);

/*
 * stop this thread
 */
void 	jthread_stop(jthread_t jtid);

/*
 * interrupt this thread
 */
void 	jthread_interrupt(jthread_t jtid);

/*
 * have the current thread exit
 */
void 	jthread_exit();

/*
 * determine whether a location is on the stack of the current thread
 * (was FRAMEOKAY)
 */
int 	jthread_on_current_stack(void *bp);

/*
 * determine the "interesting" stack range a conservative gc must walk
 */
void jthread_extract_stack(jthread_t jtid, void **from, unsigned *len);

/*
 * Disallow cancellation for current thread
 */
void jthread_disable_stop();

/*
 * Reallow cancellation for current thread
 * If a cancellation is pending, the stop method will be called
 */
void jthread_enable_stop();

/*
 * functions to disable and restore interrupts
 */
void 	intsDisable();
void 	intsRestore();
void	intsRestoreAll();
int 	intsDisabled();

/*
 * API related to I/O
 */
int jthreadedFileDescriptor(int fd);
int jthreadedOpen(const char* path, int mode);
int jthreadedLseek(int, off_t, int);
int jthreadedSocket(int);
int jthreadedConnect(int fd, struct sockaddr* addr, size_t len);
int jthreadedAccept(int fd, struct sockaddr* addr, size_t* len);
ssize_t jthreadedRead(int fd, void* buf, size_t len);
ssize_t jthreadedWrite(int fd, const void* buf, size_t len);
ssize_t jthreadedRecvfrom(int fd, void* buf, size_t len, int flags,
        struct sockaddr* from, int* fromlen);
int jthreadedWaitpid(int wpid, int* status, int options);
int jthreadedSelect(int, fd_set*, fd_set*, fd_set*, struct timeval*);
void jthreadedPollFd(int fd);
int setreadtimeout(int, int);
int getreadtimeout(int);

#if defined(DEBUG)
#define	jthread_debug_set(D,V)	currentJThread->flags |= 0x10000000 << (D);\
				currentJThread->dval = (int)(V)
#define	jthread_debug_clr(D)	currentJThread->flags &= ~(0x10000000 << (D))
#endif

/* 
 * Locking API
 */
void jmutex_initialise(jmutex *lock);
void jmutex_lock(jmutex *lock);
void jmutex_unlock(jmutex *lock);

void jcondvar_initialise(jcondvar *cv);
boolean jcondvar_wait(jcondvar *cv, jmutex *lock, jlong timeout);
void jcondvar_signal(jcondvar *cv, jmutex *lock);
void jcondvar_broadcast(jcondvar *cv, jmutex *lock);

#if defined(__WIN32__)
#define SIG_T   void(*)()
#else
#define SIG_T   void*
#endif

#endif
