/*
 * fs-impl.c - pthread-based FileSystemCallInterface implementation
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "gtypes.h"
#include "gc.h"

/* we need this for the various Jxx defines */
#include "jfilesystem.h"
/* but make sure we don't use it's macros */
#undef open
#undef mkdir

/* not sure if this is a standard one */
#if !defined (O_BINARY)
#define O_BINARY 0
#endif

/***********************************************************************
 * Interface functions
 */

static
void
Finit(void)
{
}

/*
 * File open. Note that our modes (from jfilesystem.h) do not directly
 * map to the fcntl.h ones, hence we need the wrapper
 */
int
FOpen (const char* path, int mode)
{
  int fd;

  switch (mode) {
  case JOPEN_READ:
	fd = open(path, O_RDONLY|O_BINARY);
	break;
  case JOPEN_WRITE:
	fd = open(path, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0666);
	break;
  case JOPEN_APPEND:
	fd = open(path, O_RDWR|O_CREAT|O_APPEND|O_BINARY, 0666);
	break;
  case JOPEN_READWRITE:
	fd = open(path, O_RDWR|O_CREAT|O_BINARY, 0666);
	break;
  default:
	fd = -1;
	break;
  }

  return fd;
}



static
int
FFileInfo (const char* path, int type, void* extra)
{
  struct stat buf;
  int r;

  switch (type) {
  case JINFO_EXISTS:
	r = stat(path, &buf);
	if (r == 0) {
	  return (1);
	}
	break;

  case JINFO_ISFILE:
	r = stat(path, &buf);
	if (r == 0 && S_ISREG(buf.st_mode)) {
	  return (1);
	}
	break;

  case JINFO_ISDIR:
	r = stat(path, &buf);
	if (r == 0 && S_ISDIR(buf.st_mode)) {
	  return (1);
	}
	break;
	
  case JINFO_CANREAD:
	r = access(path, R_OK);
	if (r >= 0) {
	  return (1);
	}
	break;

  case JINFO_CANWRITE:
	r = access(path, W_OK);
	if (r >= 0) {
	  return (1);
	}
	break;

  case JINFO_LENGTH:
	r = stat(path, &buf);
	if (r == 0) {
	  *((jlong*)extra) = buf.st_size;
	  return (1);
	}
	break;

  case JINFO_LASTMODIFIED:
	r = stat(path, &buf);
	if (r == 0) {
	  *((jlong*)extra) = (jlong)buf.st_mtime * (jlong)1000;
	  return (1);
	}
	break;
  }

  return (0);
}

static
int
FFileInfoFD (int fd, int type, void* extra)
{
  int nr, r;
  struct stat buf;
  off_t cur;
#if !(defined(HAVE_IOCTL) && defined(FIONREAD)) && defined(HAVE_SELECT)
  static struct timeval tm = { 0, 0 };
  fd_set rd;
#endif

  switch (type) {
  case JINFO_LENGTH:
	r = fstat(fd, &buf);
	if (r == 0) {
	  *((jlong*)extra) = buf.st_size;
	  return (1);
	}
	break;

  case JINFO_AVAILABLE:
	cur = lseek(fd, (off_t)0, SEEK_CUR);
	if (cur != (off_t)-1) {
	  if (fstat(fd, &buf) != -1) {
		return (buf.st_size - cur);
	  }
	}
#if defined(HAVE_IOCTL) && defined(FIONREAD)
	r = ioctl(fd, FIONREAD, &nr);
	if (r >= 0) {
	  return (nr);
	}
	break;
#elif defined(HAVE_SELECT)
	/* This uses select() to work out if we can read - but what
	 * happens at the end of file?
	 */
	FD_ZERO(&rd);
	FD_SET(fd, &rd);
	r = select(fd+1, &rd, NULL, NULL, &tm);
	if (r == 1) {
	  return (1);
	}
	break;
#endif
  }

  return (0);
}


static
int
FReaddir (const char* path, jdirEntry** list)
{
  DIR* dir;
  struct dirent entry;
  struct dirent *result;
  jdirEntry* dirlist;
  jdirEntry* mentry;
  int count;

  dir = opendir( path);
  if ( dir == 0 ) {
	*list = 0;
	return (-1);
  }

  dirlist = 0;
  count = 0;

  /* we use readdir_r to avoid a global entry buffer, which would impose
   * a potential deadlock resource
   */
  while ( (readdir_r( dir, &entry, &result) == 0) && (result != 0) ) {
	/* We skip '.' and '..' */
	if ( result->d_name[0] == '.' ) {
	  if ( result->d_name[1] == '.' ) {
		if ( result->d_name[2] == 0 ) {
		  continue;
		}
	  }
	  else if ( result->d_name[1] == 0 ) {
		continue;
	  }
	}

	mentry = gc_malloc_fixed( sizeof(jdirEntry) + NAMLEN(result));
	assert( mentry != 0 );
	strcpy( mentry->name, result->d_name);
	mentry->next = dirlist;
	dirlist = mentry;
	count++;
  }
  closedir( dir);

  *list = dirlist;
  return count;
}

static
int
FMkdir ( const char* path )
{
  return mkdir(path, 0777);
}


/***********************************************************************
 * create and initialize the FileSystemCallInterface
 */


FileSystemCallInterface Kaffe_FileSystemCallInterface = {
  Finit,

  FOpen,
  read,              /* pread would be nicer, but needs explicit offset */  
  write,             /* same goes for pwrite */
  (int(*)(int,off_t,int)) lseek,
  close,
  FFileInfo,
  FFileInfoFD,

  FReaddir,

  FMkdir,
  rmdir,
  rename,
  remove,
};
