/*
 * kernel-impl.c - pthread-based KernelCallInterface implementation
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include <sys/wait.h>
#if defined(HAVE_FTIME)
#include <sys/timeb.h>
#endif
#if defined(HAVE_TIME)
#include <time.h>
#endif
#include "gtypes.h"
#include "jsyscall.h"
#include "gc.h"


/*
 * This is here because it depends on the platform whether process
 * resources have to be cleaned up during exit
 */
int runFinalizerOnExit = 0;


/***********************************************************************
 * Interface functions
 */

/*
 * Turn the provided file descriptor in something that conforms
 * to our threading system (eg. set non-blocking modes for user
 * thread systems). Since we talk kernel-threads, we don't care
 * for blocking
 */
jint
KFileDescriptor ( int fd )
{
  return fd;
}


#if defined(MEM1) || defined(MEM2)
/*
 * This is the low(est) level allocator our system provides. It should
 * map to the most raw (low level overhead) address space expansion
 * that is available. It is not used frequently, on small objects, i.e.
 * malloc is just a last retreat, not the preferred choice
 */
static
void*
Kpagealloc(int size)
{
  void* ptr;

#if defined(HAVE_SBRK)

  /* Our primary choice for basic memory allocation is sbrk() which
   * should avoid any unseen space overheads.
   */
  for (;;) {
	int missed;
	ptr = sbrk(size);
	if (ptr == (void*)-1) {
	  ptr = 0;
	  break;
	}
	if ((uintp)ptr % gc_pgsize == 0) {
	  break;
	}
	missed = gc_pgsize - ((uintp)ptr % gc_pgsize);
	sbrk(-size + missed);
  }
  if (ptr == 0) {
	return (0);
  }

#elif defined(HAVE_MEMALIGN)

  ptr = memalign(gc_pgsize, size);
  CHECK_OUT_OF_MEMORY(ptr);

#elif defined(HAVE_VALLOC)

  ptr = valloc(size);
  if (ptr == 0) {
	return (0);
  }

#else
  
  /* Fallback ...
   * Allocate memory using malloc and align by hand.
   */
  size += gc_pgsize;
#ifdef malloc
#undef malloc
#endif
  ptr = malloc(size);
  if (ptr == 0) {
	return (0);
  }
  ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -gc_pgsize);
	
#endif

  return (ptr);
}
#endif

/*
 * Get a byte from whatever is supposed to be our 'stdin' stream.
 */
static
int
Kinput(void)
{
  char c;
  if (read( 0, &c, 1) != 1) {
	return (-1);
  }
  else {
	return (c);
  }
}

static
int
Kinputavailable(void)
{
	return (fileinfoFD(0, JINFO_AVAILABLE, 0));
}

/*
 * put a byte on whatever is supposed to be our 'stdout' stream
 */
static
void
Koutput(int chr)
{
  char c = (char) chr;
  write( 1, &c, 1);
}

/*
 * put a byte on whatever is supposed to be our 'stderr' stream
 */
static
void
Kerror(int chr)
{
  char c = (char) chr;
  write( 2, &c, 1);
}

/*
 * Return the current time in milliseconds. Prefer the time which has
 * the highest resolution
 */
static
jlong
Ktime (void)
{
  jlong tme;
#if defined(HAVE_GETTIMEOFDAY)
  struct timeval tm;
  gettimeofday(&tm, 0);
  tme = (((jlong)tm.tv_sec * (jlong)1000) + ((jlong)tm.tv_usec / (jlong)1000));
#elif defined(HAVE_FTIME)
  struct timeb tm;
  ftime(&tm);
  tme = (((jlong)tm.time * (jlong)1000) + (jlong)tm.millitm);
#elif defined(HAVE__FTIME)
  struct _timeb tm;
  _ftime(&tm);
  tme = (((jlong)tm.time * (jlong)1000) + (jlong)tm.millitm);
#elif defined(HAVE_TIME)
  tme = (jlong)1000 * (jlong)time(0);
#else
  tme = 1;
#endif

  return (tme);
}


/***********************************************************************
 * create and initialize the KernelCallInterface
 */

KernelCallInterface Kaffe_KernelCallInterface = {
  KFileDescriptor,
  select, 
  waitpid,

#if defined(MEM1) || defined(MEM2)
  Kpagealloc,
#else
  0,
#endif
  0, /* pagefree */
  Kinput,
  Koutput,
  Kerror,
  Ktime,
  Kinputavailable,
};
