/*
 * thread-impl.c - pthread based ThreadInterface implementation
 *  
 * Kernel Threads are assumed in some places
 *  
 * Copyright (c) 1999,2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-signal.h"
#include "gc.h"

#include <pthread.h>
#include <semaphore.h>
#include <sched.h>

#include "lerrno.h"

#include "java_lang_Throwable.h"
#include "gtypes.h"
#include "threads.h"
#include "gc.h"
#include "jni.h"
#include "locks.h"

#include "jdebug.h"

/* these are required for handling exceptions */
#include "engine.h"
#include "exception.h"

#include "thread-impl.h"
#include "thread-internal.h"

/* Some systems need special setups for the exception handling */
#if !defined(EXCEPTIONSTART)
#define EXCEPTIONSTART()
#endif
#if !defined(EXCEPTIONEND)
#define EXCEPTIONEND()
#endif

/* for debugging, needed to compile even if DEBUG is not defined */
char stat_act[]   = { ' ', 'a' };
char stat_susp[]  = { ' ', 's', ' ', 'r', ' ', ' ', ' ', ' ', ' ', ' ', ' ' };
char stat_block[] = { ' ', 'T', 'm', ' ', 'c', ' ', ' ', ' ', 't', ' ', ' ' };



/***********************************************************************
 * typedefs & defines
 */

#include <bits/local_lim.h>

/* our upper create limit, to ensure we don't blow the system */
#define MAX_SYS_THREADS     _POSIX_THREAD_THREADS_MAX - 1

/* our maximum number of (temporarily) cached native threads */
#define MAX_CACHED 8 

/* number of seconds we keep native threads cached for recycling */ 
#define CACHE_SEC 10

/* Stack alignment */
#define	STACK_ALIGN	0x1000

/*
 * Now it starts to get hackish - we have to pick some signals
 * for suspend/resume (enter/exitCritSect) which don't interfere
 * with pthread implementations. Note that we can't rely on when
 * a suspend signal is delivered, and it's therefore not safe
 * to mulitplex a single signal for both suspend & resume purposes
 */
 

 
#if !defined(__SIGRTMIN) || (__SIGRTMAX - __SIGRTMIN < 3)
#define          SIG_SUSPEND   SIGURG
#define          SIG_RESUME    SIGTSTP
#define          SIG_DUMP      SIGXCPU

/*
 * Sneak these signal in from the thread library.
 */
#define		 PSIG_RESTART	SIGUSR1
#define		 PSIG_CANCEL	SIGUSR2

#else
/* included part on linux i386 */

#define          SIG_SUSPEND   SIGUSR1
#define          SIG_RESUME    SIGUSR2
#define          SIG_DUMP      SIGXCPU


/*
 * Sneak these signal in from the thread library.
 */
#define		 PSIG_RESTART	(__SIGRTMIN)
#define		 PSIG_CANCEL	(__SIGRTMIN+1)

#endif


/***********************************************************************
 * global data
 */

/* We keep a list of all active threads, so that we can enumerate them */
nativeThread     *activeThreads;

/* We don't throw away threads when their user func terminates, but suspend
 * and cache them for later re-use */
nativeThread     *cache;

/* When we dispose of a thread we add it to the disposeQ because we cannot
 * remove it out from under us.
 */
nativeThread	*disposeQ;

/* The notorious first thread, which has to be handled differently because
 * it isn't created explicitly */
nativeThread     *firstThread;

/* Number of active non-daemon threads (the last terminating nonDaemon
 * causes the process (kaffe) to shut down */
int              nonDaemons;

/* Number of system threads (either running (activeThreads) or
 * blocked (cache)). We need this to implement our own barrier, since
 * many kernel thread systems don't behave gracefully on exceeding their limit */ 
int              nSysThreads;

/* number of currently cached threads */
int              nCached;

/* map the Java priority levels to whatever the pthreads impl gives us */
/* Java priorities are 1-10 + 0 + 11 (for gc thread) Philippe 7-27-2000 */
/* + 1 for highest-> critical sections, real-time */ 
int              priorities[java_lang_Thread_MAX_PRIORITY+1+1+1]; 

/* thread-specific-data key to retrieve 'nativeData' */
pthread_key_t    ntKey;

/* our lock to protect thread list manipulation/iteration */
iLock     *tLock;

/* a hint to avoid unnecessary pthread_creates (with pending exits) */
volatile int     pendingExits;

/* level of critical sections (0 = none) */
int              critSection;

/* helper semaphore to signal completion of critical section enter/exit */
sem_t            critSem;

sigset_t         suspendSet;

/* an optional deadlock watchdog thread (not in the activeThread list),
 * activated by DEBUG topic vm_thread or vm_thread_deadlock_detect */
pthread_t        deadlockWatchdog;


/* last queried current thread (to optionally speed up currentXX queries) */
nativeThread     *lastCur;


/* tDump support: was a signal received to dump the threading system's state
   since we cannot call it directlty from a sig handler cause it calls mutexlock and others  
   so watchdog checks for this setting in a loop*/

short pleaseDump = 0;  


#define DEADLOCK_OR_DUMPREQUEST_DETECT_INTERVAL  2    /* how long do we pause before checking for another 
                                                         deadlock or SIG_DUMP signal */
	
#include "objects.h"
static void tInitLock ( nativeThread*  );
static void tSetLock ( nativeThread*  );
						 																														 									 					 
																																												 
/*Support for real time */
			
/*Determine scheduling policy */
    				

#if defined(HAVE_REALTIME_FIFO) || defined(HAVE_REALTIME_RR)
        
	
#ifdef 	HAVE_REALTIME_FIFO
	
  unsigned long SchedulePolicy = SCHED_FIFO;
#else
        
  unsigned long	SchedulePolicy = SCHED_RR;
#endif


		 
#else
       
   unsigned long SchedulePolicy = SCHED_OTHER;
#endif       
																																														 
																																																																																					 																																									 
void suspend_signal_handler ( int sig );
void resume_signal_handler ( int sig );
static void tDispose ( nativeThread* nt );
static void tDisposeAll ( void );


#define LOCKSLOT  int iLockRoot

#define LOCK(_nt)                \
  _nt->blockState |= BS_THREAD;  \
  lockStaticMutex( &tLock)
  
#define UNLOCK(_nt)              \
  unlockStaticMutex( &tLock);    \
  _nt->blockState &= ~BS_THREAD

#define LOCKFORDUMP   lockStaticMutex( &tLock)
#define UNLOCKFORDUMP  unlockStaticMutex( &tLock)
  


/***********************************************************************
 * internal functions
 */

/*
 * get and store the invariant part of a stack address
 */
void
tPrepareStackLookup ( nativeThread* nt )
{
	int i;
	unsigned long a = (unsigned long) nt->stackMin +1;
	unsigned long b = (unsigned long) nt->stackMax -1;
	unsigned long a0 = a;

	for ( i=0; a != b; a>>=1, b>>=1, i++ )
		;

	nt->stackMask = (0xffffffff >> i) << i;
	nt->stackBase = a0 & nt->stackMask;
}


/*
 * On demand debug signal to dump the state of the threading system (requested
 * by a external "kill -s <SIG_DUMP> <proc-id>", where proc-id is ANY thread of the VM (so this
 * excludes pthreads' manager thread)
 */

void
dump_signal_handler ( int sig )  
{
 /* must set a flag to be checked by watchdog . 
    Must also resuspend if was called while suspended,
    since sigsuspend returns when any signal is received */
 /* Can't have pthread calls in signal handlers */ 
 
   pleaseDump = 1;
}



/*
 * SIGSEGV handler (mapping system exceptions to Java exceptions)
 */
void
segv_signal_handler ( EXCEPTIONPROTO ) 
{
	EXCEPTIONSTART();

	EXCEPTIONFIXRETURN();

	postAndThrowException("java.lang.NullPointerException", 0);
	
	EXCEPTIONEND();
}



/*
 * dump a thread list
 */
void
tDumpList (nativeThread* list)
{
	int             i;
	char            a;
	nativeThread    *t;

	for ( t=list, i=0; t; t=t->next, i++ ){
		/* the first one? */
		a = (t == firstThread) ? '1' : ' ';

		kprintf ("%4d: %c %c%c%c   %p [tid: %4d, java: %p]  "
			 "lock: [mux: %p cv: %p] stack: [%p..%p..%p]\n",
			 i, a,
			 stat_act[t->active], stat_susp[t->suspendState], 
			 stat_block[t->blockState],
			 t, t->tid, t->thread,
			 &t->mux, &t->cv,
			 t->stackMin, t->stackCur, t->stackMax);
	}
	
}



/*
 * dump the state of the threading system
 */
void
tDump (void)
{
	DBG_ACTION_OR( vm_thread,
	               vm_thread_deadlock_detect, {
		LOCKSLOT;
		void             *lock;
	        void             *holder = 0; /* take care of warnings */
		void             *mux = 0;
		void             *muxNat = 0;
		void             *cv = 0;
		void             *cvNat = 0;
  
  
  	        lock   = tLock;
		
		if(lock != LOCKFREE){
	        	holder = tLock->holder;
			mux    = tLock->mux;
			muxNat = tLock->mux ? unhand(tLock->mux)->PrivateInfo : 0;
			cv     = tLock->cv;
			cvNat  = tLock->cv ? unhand(tLock->cv)->PrivateInfo : 0;
                }

  
                /* want a snapshot so lock all thread lists */  
       
                LOCKFORDUMP;/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */


		       
		kprintf("======================== thread dump =========================\n");
		
		kprintf("state:  nonDaemons: %d, critSection: %d\n",
						 nonDaemons, critSection);
	        if(lock != LOCKFREE){
	                kprintf ("tLock:       %p [holder: %p, mux: %p (native: %p), cv: %p (native: %p)]\n",
	                                       lock, holder, mux, muxNat, cv, cvNat);
                }
		else{
			kprintf("tLock: currently free\n");
		}
		
		kprintf("active threads:\n");
		
		tDumpList(activeThreads);

		
		kprintf("\ncached threads:\n");
		
		tDumpList(cache);

		
		kprintf ("====================== end thread dump =======================\n");
		
	
		UNLOCKFORDUMP;/* ------------------------------------------------------ tLock */
	
	})
}


static
void*
tWatchdogRun (void* p)
{
      	nativeThread *t;
	int life;
	int            oldCancelType, oldCancelState;
        nativeThread* nt = (nativeThread*)p;
	size_t           ss;

	
        /* need the following stuff to be able to get locks later */
	
	if((errno = pthread_setspecific( ntKey, nt)) != 0)
		PerrorAbort("pthread_setspecific");


        // set cancellability
	
	if((errno = pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, &oldCancelState)) != 0)
                PerrorAbort("pthread_setcancelstate");

	
	if((errno = pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, &oldCancelType)) != 0)
		PerrorAbort("pthread_setcanceltype");


        
	/* get the stack boundaries (no way of knowing if setstacksize succeeded to set
	   exactly what we asked) */
#if defined(_POSIX_THREAD_ATTR_STACKSIZE)
	   
	if((errno = pthread_attr_getstacksize( &nt->attr, &ss)) != 0)
		PerrorAbort("pthread_attr_getstacksize");
#endif


#if defined(STACK_GROWS_UP)
	nt->stackMin  = (void*) ((uintp)&nt & -STACK_ALIGN);
	nt->stackMax  = (void*) ((uintp) nt->stackMin + ss);
#else
	nt->stackMax  = (void*) (((uintp)&nt + STACK_ALIGN - 1) & -STACK_ALIGN);
	nt->stackMin  = (void*) ((uintp) nt->stackMax - ss);
#endif		
	
	
	
	tPrepareStackLookup( nt);
	
        			
	/* main loop */						
							
	while ( nonDaemons ) {
	

                sleep(DEADLOCK_OR_DUMPREQUEST_DETECT_INTERVAL) ;
		
		
			
DBG_ACTION( vm_thread_deadlock_detect, {

		life = 0;
		
	
		for ( t=activeThreads; t != NULL; t = t->next ){
			/*
			 * if we have a single thread that is not blocked at all, is in a
			 * timeout wait, and is not suspended, we are still safe (even though
			 * the timeout value might effectively be a deadlock). Also, holding
			   a mutex is fine. 
			 */
						
			if ( (!t->blockState || (t->blockState == BS_CV_TO) 
			     || (t->blockState == BS_THREAD)) && !t->suspendState ){
				life = 1;
				break;
			}
		}
		

		if ( !life ) {
			DBG( vm_thread_deadlock_detect, ("DEADLOCK\n"));
			tDump();
			ABORT();
		}
	        
 })		
 
 
				
DBG_ACTION( vm_thread,	{

                /* atomic test and set so miss less signals 
		   This  stuff means:
		   if(pleaseDump) pleaseDump = 0;

		*/
		
	        if(COMPARE_AND_EXCHANGE(&pleaseDump, 1, 0)){
			tDump();
                }
		
})	
			  	
		
		 
	} /* while */
	

	return 0;
}






void
tStartDeadlockWatchdog ()
{
	struct sched_param sp;
	
	nativeThread* nt;
        
	extern Hjava_lang_Class* ThreadClass;

		
	
	/* must "fake" that this is like another thread so that can get locks later */
	
	
	nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);
	assert(nt != 0);

	nt->tid = deadlockWatchdog;
       
      
      
        nt->thread = (Hjava_lang_Thread*)newObject(ThreadClass);
	assert(nt->thread != 0);

        unhand(nt->thread)->sem = gc_malloc_fixed_typed(sizeof(sem2posixLock), GC_ALLOC_THREADCTX);

        nt->suspendState = 0;

	 
	
	/* init our cv and mux fields for locking */
	tInitLock( nt);
	tSetLock( nt);



        /* end of "fake" stuff */
	

	sp.sched_priority = priorities[0];  /* looow */

	if((errno = pthread_attr_init( &nt->attr)) != 0)
		PerrorAbort("pthread_attr_init");

       	
	if((errno = pthread_attr_setschedpolicy( &nt->attr, SchedulePolicy)) !=0)
		PerrorAbort("pthread_attr_setschedpolicy");


        if((errno = pthread_attr_setschedparam( &nt->attr, &sp)) != 0)
		PerrorAbort("pthread_attr_setschedparam");
	
							
#if defined(_POSIX_THREAD_ATTR_STACKSIZE)

        if((errno = pthread_attr_setstacksize( &nt->attr, PTHREAD_STACK_MIN)) != 0)
                PerrorAbort("pthread_attr_setstacksize");

#endif

        if((errno = pthread_attr_setdetachstate( &nt->attr, PTHREAD_CREATE_DETACHED)) != 0)
		PerrorAbort("pthread_attr_setdetachstate");


	if((errno = pthread_create( &deadlockWatchdog,  &nt->attr, tWatchdogRun, nt)) != 0)
		PerrorAbort("pthread_create");
		
	
	
		
}



/***********************************************************************
 * thread system initialization
 */

/*
 * static init of signal handlers
 */
void
tInitSignalHandlers ()
{
	struct sigaction sigSuspend, sigResume, sigSegv;
	unsigned int flags = 0;

#if defined(SA_RESTART)
	flags |= SA_RESTART;
#endif

#if !defined(HAVE_REALTIME_FIFO) && !defined(HAVE_REALTIME_RR)

	sigSuspend.sa_flags = flags;
	sigSuspend.sa_handler = suspend_signal_handler;
	sigemptyset( &sigSuspend.sa_mask);
	sigaddset( &sigSuspend.sa_mask, SIG_SUSPEND);
	sigaddset( &sigSuspend.sa_mask, SIG_RESUME);
	sigaddset( &sigSuspend.sa_mask, PSIG_RESTART);
	sigaddset( &sigSuspend.sa_mask, PSIG_CANCEL);

	sigaddset( &sigSuspend.sa_mask, SIGSTOP);
	sigaddset( &sigSuspend.sa_mask, SIGCONT);
	sigaddset( &sigSuspend.sa_mask, SIGWINCH);

	if(sigaction( SIG_SUSPEND, &sigSuspend, NULL) < 0)
		PerrorAbort("sigaction");


	sigResume.sa_flags = flags;
	sigResume.sa_handler = resume_signal_handler;
	sigResume.sa_mask = sigSuspend.sa_mask;
	if(sigaction( SIG_RESUME, &sigResume, NULL) < 0)
		PerrorAbort("sigaction");

#endif


DBG_ACTION( vm_thread, {
        
	                struct sigaction sigDump;
			
		        sigDump.sa_flags = flags;
	                sigDump.sa_handler = dump_signal_handler;
	                sigemptyset( &sigDump.sa_mask);
	                if(sigaction( SIG_DUMP, &sigDump, NULL) < 0)
		        	PerrorAbort("sigaction");

	  });



	sigSegv.sa_flags = flags;
#if defined(SA_SIGINFO)
	sigSegv.sa_flags |= SA_SIGINFO;
#endif
#if defined(SA_NOMASK)
	sigSegv.sa_flags |= SA_NOMASK;
#endif
	sigSegv.sa_handler = (void(*)(int)) segv_signal_handler;
	sigemptyset( &sigSegv.sa_mask);
	if(sigaction( SIGSEGV, &sigSegv, NULL) < 0)
		PerrorAbort("sigaction");
}

/*
 * static init set up of Java-to-pthread priority mapping (pthread prioritiy levels
 * are implementation dependent)
 */
 
 
static
void
tMapPriorities ()
{
	int     d, min, max, n, i;
	float   r;

	if((min = sched_get_priority_min(SchedulePolicy)) < 0)
		PerrorAbort("sched_get_priority_min");
	if((max = sched_get_priority_max(SchedulePolicy)) < 0)
               	PerrorAbort("sched_get_priority_max");

	if((d = max - min) == 0)
	    return; /* priorities array already filled with 0 since it is a global var*/
	n = sizeof( priorities) / sizeof(priorities[0]);
	r = (float)d / (float)n;
	

	for ( i=0; i<n; i++ ) {
		priorities[i] = (int)(i*r + 0.5) + min;
	}
	
}


/*
 * per-native thread init of condvars, mutexes and semaphores
 */
static
void
tInitLock ( nativeThread* nt )
{
        pthread_condattr_t    cvAttr;
	pthread_mutexattr_t   muxAttr;

	/* init a process private condvar */
	pthread_condattr_init( &cvAttr);
#if defined(_POSIX_THREAD_PROCESS_SHARED)
        if((errno = pthread_condattr_setpshared( &cvAttr, PTHREAD_PROCESS_PRIVATE)) != 0)
		PerrorAbort("pthread_condattr_setpshared");
#endif
    	pthread_cond_init( &nt->cv, &cvAttr);

	/* init a process private mutex. We deal with priority inversion
	 * by means of inheriting the highest priority of any thread which
	 * requests the mutex
	 */
	pthread_mutexattr_init( &muxAttr);
#if defined(_POSIX_THREAD_PROCESS_SHARED)
        if((errno = pthread_mutexattr_setpshared( &muxAttr, PTHREAD_PROCESS_PRIVATE)) != 0)
		PerrorAbort("pthread_mutexattr_setpshared");
#endif
#if defined(_POSIX_THREAD_PRIO_INHERIT)
	if((errno = pthread_mutexattr_setprotocol( &muxAttr, PTHREAD_PRIO_INHERIT)) != 0)
		PerrorAbort("pthread_mutexattr_setprotocol");
#endif
	pthread_mutex_init( &nt->mux, &muxAttr);
	
	
	/* init a non-shared (process-exclusive) semaphore with value '0' */
	if(sem_init( &nt->sem, 0, 0) < 0)
		PerrorAbort("sem_init");

}

/*
 * init the Java thread locking facilities (sem) from the native thread
 * condvar and mutex
 */
static
void
tSetLock ( nativeThread* nt )
{
	sem2posixLock *lk;

	lk = (sem2posixLock*)unhand(nt->thread)->sem;
	lk->cv  = &nt->cv;
	lk->mux = &nt->mux;
	lk->thd = nt;
}

/*
 * We must have a certain amount of credible thread information setup
 * as soon as possible.
 */
static
void
tSetupFirstNative(void)
{
	nativeThread* nt;

	/*
	 * We need to have a native thread context available as soon as possible.
	 */
	nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);
	nt->tid = pthread_self(); 
        if((errno = pthread_setspecific( ntKey, nt)) != 0)
		PerrorAbort("pthread_setspecific");
	nt->stackMin  = (void*)0;
	nt->stackMax  = (void*)-1;
	nt->stackBase = 0;
	nt->stackMask = 0;
	nt->thread = (Hjava_lang_Thread*)1;	/* Non-zero dummy */
	lastCur = nt;
}

/*
 * The global, one-time initialization goes here. This is a
 * alternative to scattered pthread_once() calls
 */
void
Tinit(void)
{
       	/* The following hack is necessary because the pthread standard does not require
	   errno to be set in the lib call, only to return the error number 
	   errno is thread specific when _REENTRANT is defined*/
	   
	if((errno = pthread_key_create( &ntKey, NULL)) != 0)
		PerrorAbort("pthread_create");
	    
	if(sem_init( &critSem, 0, 0) < 0)
		PerrorAbort("sem_init");
		
	
	tMapPriorities();
	
	tInitSignalHandlers();
	

#if !defined(HAVE_REALTIME_FIFO) && !defined(HAVE_REALTIME_RR)

        sigfillset( &suspendSet);
		  
	sigdelset( &suspendSet, SIG_RESUME);
	
	sigdelset( &suspendSet, SIGINT); /* so that can end the program from terminal in 
	                                    case of internal deadlock  */
        sigdelset( &suspendSet, SIGQUIT);
	 
	
DBG_ACTION( vm_thread, 
	                
        sigdelset( &suspendSet, SIG_DUMP);
	  
)


#endif

	tSetupFirstNative();
	
	
	DBG( vm_thread, ("initialized\n"));


       
}


/***********************************************************************
 * thread creation
 */

void
Tattach ( Hjava_lang_Thread* thread )
{
	nativeThread   *nt;
	int            oldCancelType, oldCancelState;
	LOCKSLOT;
        struct sched_param   sp;

	nt = GET_CURRENT_THREAD( &nt );
	
	if (firstThread != 0) {
		/* If this thread is already attached then ignore */
		if (nt != 0) {
			return;
		}
		nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);
	}
	assert(nt != 0);

        nt->tid = pthread_self();


	/* we can't use nt->attr, because it wasn't used to create this thread */
	if((errno = pthread_attr_init( &nt->attr)) != 0)
		PerrorAbort("pthread_attr_init");


	nt->thread = thread;
	nt->suspendState = 0;
	nt->func   = NULL;
	nt->next   = NULL;

	/* Get stack boundaries. Note that this is just an approximation
	 * which should cover all gc-relevant stack locations
	 */
#if defined(STACK_GROWS_UP)
	nt->stackMin  = (void*) ((uintp)&nt & -STACK_ALIGN);
	nt->stackMax  = (void*) ((uintp) nt->stackMin + threadStackSize);
#else
	nt->stackMax  = (void*) (((uintp)&nt + STACK_ALIGN - 1) & -STACK_ALIGN);
	nt->stackMin  = (void*) ((uintp) nt->stackMax - threadStackSize);
#endif

	tPrepareStackLookup( nt);
	lastCur = nt;

	DBG( vm_thread, TMSG_SHORT( "attach ", nt));

	/* init our cv and mux fields for locking */
	tInitLock( nt);
	tSetLock( nt);

	/* Link native and Java thread objects
	 * We already are executing in the right thread, so we can set the specific
	 * data right away
	 */
	if((errno = pthread_setspecific( ntKey, nt)) != 0)
		PerrorAbort("pthread_setspecific");
		
		
 
	unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)nt;
	
        if((errno = pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, &oldCancelState)) != 0)
                PerrorAbort("pthread_setcancelstate");

	
	if((errno = pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, &oldCancelType)) != 0)
		PerrorAbort("pthread_setcanceltype");
		
		
	if((errno = pthread_detach( nt->tid)) != 0)
		PerrorAbort("pthread_detach");
	

  
	/* Link the new one into the activeThreads list. We lock until
	 * the newly created thread is set up correctly (i.e. is walkable)
	 */
	LOCK( nt); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

	nt->active = 1;
	nt->next = activeThreads;
	activeThreads = nt;
	
	nonDaemons++;

	if (firstThread == 0) {
		firstThread = nt;
				
               	DBG_ACTION_OR( vm_thread,
		               vm_thread_deadlock_detect, { tStartDeadlockWatchdog(); });
		
	}
		
    
	nSysThreads++;
	
	sp.sched_priority = priorities[unhand(nt->thread)->priority];
        if((errno = pthread_setschedparam( nt->tid, SchedulePolicy, &sp)) != 0)
                PerrorAbort("pthread_setschedparam");


	
	UNLOCK( nt); /* ---------------------------------------------------- tLock */
}


/*
 * This is our thread function wrapper, which we need because of two
 * reasons. First, there is no way to set thread specific data from
 * outside the current thread (and it should be properly set before we
 * enter the *real* thread func). Second, this is a better place to handle
 * the recycle looping than Texit (would be strange to loop in there)
 */
static
void*
tRun ( void* p )
{
	LOCKSLOT;
	nativeThread     *cur = (nativeThread*)p;
	nativeThread     *t;
	nativeThread     **tp;
	size_t           ss;
	int              oldCancelType, oldCancelState;

	int              dispose, cached;
	struct timespec  abst;
	struct timeval   now;


	/* get the stack boundaries (no way of knowing if setstacksize succeeded to set
	   exactly what we asked) */
#if defined(_POSIX_THREAD_ATTR_STACKSIZE)

	if((errno = pthread_attr_getstacksize( &cur->attr, &ss)) != 0)
		PerrorAbort("pthread_attr_getstacksize");
		
#endif

				
  
#if defined(STACK_GROWS_UP)
	// cur->stackMin = &cur;
	cur->stackMin  = (void*) ((uintp)&cur & -STACK_ALIGN);
	cur->stackMax = (void*) ((unsigned long)cur->stackMin + ss);
#else
	cur->stackMax  = (void*)(((uintp)&cur + STACK_ALIGN - 1) & -STACK_ALIGN);
	// cur->stackMax = &cur;
	cur->stackMin = (void*) ((unsigned long)cur->stackMax - ss);
#endif

	tPrepareStackLookup( cur);
	
	if((errno = pthread_setspecific( ntKey, cur)) != 0)
		PerrorAbort("pthread_setspecific");
		
	if((errno = pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, &oldCancelState)) != 0)
                PerrorAbort("pthread_setcancelstate");
	

	if((errno = pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, &oldCancelType)) != 0)
		PerrorAbort("pthread_setcanceltype");


	/* We set this here explicity since the creator doesn't always get to set this in time.
	 */
	cur->tid = pthread_self();/*pthread_self never fails */


	/* We are reasonably operational now, flag our creator that it's safe to give
	 * up the thread list lock. Make sure we don't try to acquire the LOCK before
	 * that, since Tcreate still holds it (-> deadlock )
	 */
	if(sem_post( &cur->sem) < 0)
		PerrorAbort("sem_post");

	for(;;) {
		DBG( vm_thread, TMSG_LONG( "calling user func of: ", cur));

		/* Now call our thread function, which happens to be firstStartThread(),
		 * which will call TExit before it returns */
		cur->func();
	
		DBG( vm_thread, TMSG_LONG( "exiting user func of: ", cur));

		LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

		  
		/* remove from active list */
		if ( cur == activeThreads ){
			activeThreads = cur->next;
		}
		else {
			for ( t=activeThreads; t->next && (t->next != cur); t=t->next );
			assert( t->next != 0 );
			t->next = cur->next;
		}

		/* unlink Java thread */
		unhand(cur->thread)->PrivateInfo = 0;
		cur->suspendState = 0;

		if ( nCached < MAX_CACHED ){
			/* add this thread to the cached list */
			nCached++;
			cur->next = cache;
			cache = cur;
			DBG( vm_thread, TMSG_SHORT( "cached thread ", cur));	
			cached = 1;
		}
		else {
			cached = 0;
		}
		
		pendingExits--;

		if ( cached == 0 ) {
			/* if it hasn't been cached, it can't be resurrected, bail out */
			break;
		}

		UNLOCK( cur); /* ---------------------------------------------------- tLock */
		
		/* Keep this thread for CACHE_SEC seconds before throwing it away.
		 * If it is resurrected during that time (by Tcreate), we
		 * recycle it, otherwise it's going to be disposed.
		 * No need to update the blockState, since we aren't active anymore
		 */
		if(gettimeofday( &now, 0) < 0)
		   PerrorAbort("gettimeofday");
		abst.tv_sec  = now.tv_sec + CACHE_SEC;
		abst.tv_nsec = 0;
 
		if((errno = pthread_mutex_lock( &cur->mux)) != 0)
			PerrorAbort("pthread_mutex_lock");

		if (cur->active == 0) {
		
		        int stat;
			
			#if defined(SA_RESTART)

	                	pthread_cond_timedwait( &cur->cv, &cur->mux, &abst);

			#else
        		        // problem: race condition ->if signaled in between two wait calls
				do{
	   				stat = pthread_cond_timedwait( &cur->cv, &cur->mux, &abst);        		
				}
				while(stat == EINTR); 
	
			#endif
		}
		if((errno = pthread_mutex_unlock( &cur->mux)) != 0)
			PerrorAbort("pthread_mutex_unlock");


		LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */
		
		dispose = 0;	
		
		if ( cur->active == 0) { /* has not been activated within CACHE_SEC */
		
			/* remove it from the cache list (can be at any position in there) */
			for ( tp = &cache; *tp != 0; tp = &t->next ) {
				t = *tp;
				if ( t == cur ){
					*tp = cur->next;
					nCached--;
					dispose = 1;
					break;
				}
			}
			assert(dispose == 1);
			break;
		}

		UNLOCK( cur); /* -------------------------------------------------- tLock */
			

		/*
		 * we have already been moved back to the activeThreads list by
		 * Tcreate (saves us a lock)
		 */
		DBG( vm_thread, TMSG_SHORT( "reused thread ", cur));
	}

			DBG( vm_thread, TMSG_SHORT( "dispose thread ", cur));
	/* Add to the disposeQ */
	cur->next = disposeQ;
	disposeQ = cur;

        
	nSysThreads--;

	UNLOCK( cur); /* -------------------------------------------------- tLock */
	

	return 0;
}


/*
 * Create a new native thread for a given Java Thread object. Note
 * that we are called from Thread.start(), i.e. we are already allowed to
 * call func() (which happens to be threads.c:firstStartThread)
 *
 * Our call graph looks like this:
 *
 *     Thread.start
 *       startThread
 *         Kaffe_ThreadInterface.create (firstStartThread) == Tcreate
 *           pthread_create(..tRun..)
 *                                                 creator-thread
 *   --------------------------------------------------------------
 *                                                 created-thread
 *     tRun
 *       firstStartThread
 *         Thread.run
 *         exitThread
 *           Thread.finish
 *           Kaffe_ThreadInterface.exit == Texit
 */

void
Tcreate ( Hjava_lang_Thread* thread, void* func )
{
	LOCKSLOT;
	nativeThread         *cur = GET_CURRENT_THREAD( &cur); 
	nativeThread         *nt;
	struct sched_param   sp;

	/* if we are the first one, it's seriously broken */
	assert( activeThreads != 0 );

	/*
	 * This is a safeguard to avoid creating too many new threads
	 * because of a high Tcreate call frequency from a high priority
	 * thread (which doesn't give exiters a chance to acquire the lock
	 * to update the cache list).
	 */
	while ( (cache == 0) && pendingExits ){
		if(sched_yield() < 0)
			perror("sched_yield");
	}
	
	sp.sched_priority = priorities[unhand(thread)->priority];
        
	
	LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

              
	if ( !unhand(thread)->daemon ) {
		nonDaemons++;
	}

	if ( cache ) {
		/* move thread from the cache to the active list */
		nt = cache;
		cache = cache->next;
		nCached--;

		nt->next = activeThreads;
		activeThreads = nt;

		nt->thread = thread;
		nt->func = func;
		nt->stackCur = 0;
		unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)nt;
		tSetLock( nt);

		if((errno = pthread_setschedparam( nt->tid, SchedulePolicy, &sp)) != 0)
			PerrorAbort("pthread_setschedparam");

		
		DBG( vm_thread, TMSG_SHORT( "create recycled ", nt));

		/* resurrect it */
		if((errno = pthread_mutex_lock( &nt->mux)) != 0)
			PerrorAbort("pthread_mutex_lock");

		nt->active = 1;
		pthread_cond_signal( &nt->cv);
		if((errno = pthread_mutex_unlock( &nt->mux)) != 0)
			PerrorAbort("pthread_mutex_unlock");

		
		UNLOCK( cur); /* ---------------------------------------------------- tLock */
	}
	else {
		UNLOCK( cur); /* ---------------------------------------------------- tLock */

		if ( nSysThreads++ > MAX_SYS_THREADS ){		
			/* fail, we exceeded our physical thread limit*/
			DBG( vm_thread, ( "too many threads (%d)\n", nSysThreads));
			postAndThrowException("java.lang.InternalError", "Attempt to create more threads than the system can handle");
			return;
		}

		nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);


	        if((errno = pthread_attr_init( &nt->attr)) != 0)
			PerrorAbort("pthread_attr_init");

                if((errno = pthread_attr_setschedpolicy( &nt->attr, SchedulePolicy)) != 0)
			PerrorAbort("pthread_attr_setschedpolicy");
			
		if((errno = pthread_attr_setschedparam( &nt->attr, &sp)) != 0)
			PerrorAbort("pthread_attr_setschedparam");
 
#if defined(_POSIX_THREAD_ATTR_STACKSIZE)
 
		if((errno = pthread_attr_setstacksize( &nt->attr, threadStackSize)) != 0)
			PerrorAbort("pthread_attr_setstacksize");
#endif

                if((errno = pthread_attr_setdetachstate( &nt->attr, PTHREAD_CREATE_DETACHED)) != 0)
			PerrorAbort("pthread_attr_setdetachstate");
		
		nt->thread       = thread;
		nt->func         = func;
		nt->suspendState = 0;
		nt->stackMin     = 0;
		nt->stackMax     = 0;
		nt->stackCur     = 0;

		unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)nt;

		DBG( vm_thread, TMSG_SHORT( "create new ", nt));

		/* init our cv and mux fields for locking */
		tInitLock( nt);
		tSetLock( nt);
		
		/* Link the new one into the activeThreads list. We lock until
		 * the newly created thread is set up correctly (i.e. is walkable)
		 */
		LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

		nt->active = 1;
		nt->next = activeThreads;
		activeThreads = nt;
		
		/* Note that we don't directly start 'func' because we (a) still need to
		 * set the thread specifics, and (b) we need a looper for our thread
		 * recycling (pooling). We create the new thread while still holding the lock, because
		 * we otherwise might have an invalid tid in the activeList. The new thread
		 * in turn doesn't need the lock until it exits
		 */
		if((errno = pthread_create( &nt->tid, &nt->attr, tRun, nt)) != 0)
			PerrorAbort("pthread_create");


		/* wait until the thread specific data has been set, and the new thread
		 * is in a suspendable state (posted from tRun)*/
		if(sem_wait( &nt->sem) < 0)
			PerrorAbort("sem_wait");

		
		UNLOCK( cur); /* ---------------------------------------------------- tLock */
	}
}


/***********************************************************************
 * thread exit & cleanup
 */

/*
 * Native thread cleanup. This is just called in case a native thread
 * is not cached
 */
static
void
tDispose ( nativeThread* nt )
{
      
	assert(nt->active == 0);

        //all threads created in detached state  

        if((errno = pthread_mutex_destroy( &nt->mux)) != 0)
		PerrorAbort("pthread_mutex_destroy");
        if((errno = pthread_cond_destroy( &nt->cv)) != 0)
		PerrorAbort("pthread_cond_destroy");
	if(sem_destroy( &nt->sem) < 0)
		perror("sem_destroy");

	gc_free_fixed( nt);
}

/*
 * We dispose of all the native threads that have died recently.
 */
static
void
tDisposeAll( void )
{
	nativeThread* nt;
	nativeThread* nnt;

	for (nt = disposeQ; nt != 0; nt = nnt) {
		nnt = nt->next;
		tDispose (nt);
	}

	disposeQ = 0;

	// Set lastCur to something safe 
	lastCur = GET_CURRENT_THREAD( &nt);
}

/*
 * Function to be called (by threads.c firstStartThread) when the thread leaves
 * the user thread function
 */
void
Texit ( Hjava_lang_Thread* thread )
{
	LOCKSLOT;
	nativeThread         *cur = NATIVE_THREAD( thread);
	nativeThread         *t;

	/*
	 * We are leaving the thread user func, which means we are not
	 * subject to GC, anymore (has to be marked here because the first thread
	 * never goes through tRun)
	 */

	cur->active = 0;

	DBG( vm_thread, TMSG_SHORT( "exit ", cur));  

	
	if ( !unhand(thread)->daemon ){
		/* the last non daemon should shut down the process */
		if ( --nonDaemons == 0 ) {
		      
			
		        LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */
			
			DBG( vm_thread, ("exit on last nonDaemon\n"));

			/*
			 * be a nice citizen, try to cancel all other threads before we
			 * bail out, to give them a chance to run their cleanup handlers
			 */
			for ( t=cache; t != NULL; t = t->next ){
				if((errno = pthread_cancel( t->tid)) != 0)
					PerrorAbort("pthread_cancel");
			}

			for ( t=activeThreads; t != NULL; t = t->next ){
				if ( t != cur ) {
					if((errno = pthread_cancel( t->tid)) != 0)
						PerrorAbort("pthread_cancel");

				}
			}
			
			if ( deadlockWatchdog ){
				if((errno = pthread_cancel( deadlockWatchdog)) != 0)
					PerrorAbort("pthread_cancel");
			}

			if ( (cur != firstThread) && (firstThread->active == 0) ) {
				/* if the firstThread has already been frozen, it's not in the cache list */
				if((errno = pthread_cancel( firstThread->tid)) != 0)
					PerrorAbort("pthread_cancel");
			}

			pthread_exit( 0);

			/* pretty useless, but clean */
			UNLOCK( cur); /* -------------------------------------------------- tLock */

			/* we shouldn't get here, this is a last safeguard */
			EXIT(0);
		}
	}

	if ( cur == firstThread ) {
	
			/*
		 * We don't cache this one, but we have to remove it from the active list. Note
		 * that the firstThread always has to be the last entry in the activeThreads list
		 * (we just add new entries at the head)
		 */
		LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

		/* if we would be the head, we would have been the last, too (and already exited) */
		assert( cur != activeThreads);

		for ( t=activeThreads; (t != NULL) && (t->next != cur); t=t->next );
		assert( t != NULL);
		t->next = cur->next;
		
		UNLOCK( cur); /* ---------------------------------------------------- tLock */  

		/*
		 * Put us into a permanent freeze to avoid shut down of the whole process, since the main routine was
		   the first thread  
		 
		 */
		if(sem_wait( &cur->sem) < 0)
			PerrorAbort("sem_wait");

	}
	else {
	        /* flag that we soon will get a new cache entry (would be annoying to
		 * create a new thread in the meantime) */
		pendingExits++;
	}
}


/*
 * Thread is being finalized - free any held resource.
 */
void    
Tfinalize (Hjava_lang_Thread* thread)
{
	DBG_ACTION( vm_thread, {
		nativeThread         *cur = NATIVE_THREAD( thread);
		DBG( vm_thread, TMSG_SHORT( "finalize ", cur));
	});
}       


/***********************************************************************
 * scheduling and query
 */

/*
 * Change the scheduler priority of a running thread. Since we aren't
 * called directly, we can assume that 'prio' is within [MIN_PRIORITY..MAX_PRIORITY]
 */
void    
Tprio (Hjava_lang_Thread* thread, jint prio)
{
	nativeThread         *cur = NATIVE_THREAD( thread);
	struct sched_param   sp;


        // assertion cur != 0 failed, check this stuff out
	if(cur){
	
		sp.sched_priority = priorities[prio];
		
		DBG( vm_thread, ("set priority: %p [tid: %d, java: %x) to %d (%d)\n",
					cur, cur->tid, cur->thread, prio, priorities[prio]));
		if((errno = pthread_setschedparam( cur->tid, SchedulePolicy, &sp)) != 0)
			PerrorAbort("pthread_setschedparam");
        } 
	
}            


/*
 * Yield - reschedule current native thread at end of all running
 * threads with same or higher priority
 */
void
Tyield (void)
{
	tYield();
}

/*
 * Get me the Java Thread object of the current native thread
 * This should be exported as a compile time interface (potential
 * high frequency call).
 */
static          
Hjava_lang_Thread*      
TcurrentJava (void)      
{
	return tCurrentJava();
}


/*
 * This is used by the execution engine to build stack traces (we don't
 * have to care for speed, here)
 */
void
TnextFrame (stackFrame* fm)
{
	nativeThread* nt = pthread_getspecific( ntKey);
	machineStackFrame* nfm;
	uintp fp;

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

	fp = (uintp)STACK_FRAME(nfm);
	if ( (fp > (uintp) nt->stackMin) && (fp < (uintp) nt->stackMax) ){
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = (void*)fp;
	}
	else {
		fm->return_pc = 0;
		fm->return_frame = 0;
	}
}

/*******************************************************************************
 * the suspend/resume mechanism
 */

/* MAYBE an issue for real-time mode here, read the comment */

/*
 * The suspend signal handler, which we need to implement critical sections.
 * It is used for two purposes: (a) to block all active threads which might
 * get rescheduled during a critical section for SCHED_OTHER), and (b) 
   to get a GC-limit of the current
 * thread stack so that we don't have to scan the whole thing. Since this
 * assumes that we execute on the threads normal stack, make sure we have no
 * SA_ONSTACK / signalstack() in effect
 */
 
void
suspend_signal_handler ( int sig )
{
	nativeThread  *cur = GET_CURRENT_THREAD(&cur);

	DBG( vm_thread_sig, ("suspend signal handler: %p\n", cur));

	/* signals are global things and might come from anywhere (outside our purpose), anytime */
	if ( !cur || !cur->active ){
		return;
	}
	
	if ( cur->suspendState == SS_PENDING_SUSPEND ){

		/* assuming we are executing on the thread stack, we record our current pos */
		cur->stackCur     = (void*)&cur; /* just the adress of some local */
		cur->suspendState = SS_SUSPENDED;

		/* notify the critSect owner that we are now suspending in the handler */
		if(sem_post( &critSem) < 0)
			PerrorAbort("sem_post");
			
DBG_ACTION_AND_ELSE( vm_thread, {

               	/* freeze until we get a subsequent SIG_RESUME (and not a SIG_DUMP) */
		while(!cur->resumed)
	          sigsuspend( &suspendSet);

},
             	/* freeze until we get a subsequent SIG_RESUME */
		sigsuspend( &suspendSet);


)

               	
		DBG( vm_thread_sig, ("sigsuspend returned: %p\n", cur));

		cur->stackCur     = 0;
		cur->suspendState = 0;

		/* notify the critSect owner we are leaving the handler */
	        if(sem_post( &critSem) < 0)
			PerrorAbort("sem_post");
	}
}


/*
 * The resume signal handler, which we mainly need to get the implicit sigreturn
 * call (i.e. to unblock a preceeding sigsuspend).
 */
void
resume_signal_handler ( int sig )
{

DBG_ACTION( vm_thread, {
	/* set flag so know sigsuspend returns from receiving SIG_RESUME and not SIG_DUMP */
                
        nativeThread  *cur = GET_CURRENT_THREAD(&cur);
	cur->resumed = 1;
	  
})

	
}



/*
   If running in real-time, set all other threads's state to suspended, and increase
   our priority to higher than any other thread's. 
   If not in real-time, temporarily suspend all threads but the current one. This is
   an expensive operation. 
   Make sure no suspended thread blocks on a resource (mutexed non-reentrant
   lib func) we use within the critical section, or we have a classical deadlock.
   Critical sections can be nested, but that's usually not a good idea (increases
   chances for deadlocks or trojan threads)
*/
 
void
TenterCritSect ()
{
	LOCKSLOT;
	nativeThread  *cur = GET_CURRENT_THREAD( &cur);
	nativeThread  *t;
        
	
#if defined(HAVE_REALTIME_FIFO) || defined(HAVE_REALTIME_RR)

        struct sched_param sp; 
	
	LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */
        
	if ( ++critSection == 1 ){
        	
		DBG( vm_thread, ("enter crit section[%d] from: %p [tid:%d, java:%p)\n",
					 critSection, cur, cur->tid, cur->thread));
	        
		for ( t=activeThreads; t; t = t->next )
			if ( (t != cur) && (t->suspendState == 0) && (t->active) ) 
				t->suspendState = SS_SUSPENDED;
				
				
		sp.sched_priority =priorities[THREAD_MAXPRIO + 1];
	
		if((errno = pthread_setschedparam(cur->tid, SchedulePolicy, &sp)) !=0)
			PerrorAbort("pthread_setschedparam");
	}
	
	UNLOCK( cur); /* ------------------------------------------------------ tLock */

#else		

	

	/* don't allow any new thread to be created or recycled until this is done */
	LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */
	
	DBG( vm_thread, ("enter crit section[%d] from: %p [tid:%d, java:%p)\n",
					 critSection, cur, cur->tid, cur->thread));

	if ( ++critSection == 1 ){
	
		for ( t=activeThreads; t; t = t->next ){
			/*
			 * make sure we don't suspend ourselves, and we don't expect suspend
			 * signals to be handled by threads which are blocked on someting else
			 * than the thread lock (which we soon release)
			 */
			if ( (t != cur) && (t->suspendState == 0) && (t->active) ) {
				DBG( vm_thread_sig, ("signal suspend: %p (susp: %d blk: %d)\n",
									 t, t->suspendState, t->blockState));
				
				t->suspendState = SS_PENDING_SUSPEND;
				
				DBG_ACTION( vm_thread, 

                                	t->resumed = 0; 
                                
				)                              
				
				
				if ((errno = pthread_kill( t->tid, SIG_SUSPEND)) != 0 ){
					DBG( vm_thread, ("error sending SUSPEND signal to %p\n", t));
					PerrorAbort("pthread_kill");
				}
				else {
			
					/* BAD: Empirical workaround for lost signals (with accumulative syncing)
					 * It shouldn't be necessary (posix sems are accumulative), and it
					 * is bad, performance wise (at least n*2 context switches per suspend)
					 * but it works more reliably on linux 2.2.x */
					if(sem_wait( &critSem) < 0)
						PerrorAbort("sem_wait");

				}
			}
		}

	}

	UNLOCK( cur); /* ------------------------------------------------------ tLock */
	
#endif /* else */

	DBG( vm_thread, ( "critical section (%d) established\n", critSection))
}



/*
 * Resume all temporarily suspended threads. Just take action if this
 * is the outmost exit
 */
void
TexitCritSect ()
{
	nativeThread *t;
	
	
	/* Now would be a good time to dispose of any dead native threads */
        tDisposeAll();
	
	
	if ( !critSection )
		return;

	
	
#if defined(HAVE_REALTIME_FIFO) || defined(HAVE_REALTIME_RR)

        
        if ( --critSection == 0 ){
	
	        struct sched_param sp;
		nativeThread  *cur = GET_CURRENT_THREAD( &cur);

                /* No need to sync, there's nobody else running */
 	  	for ( t=activeThreads; t; t = t->next )
			if ( t->suspendState & (SS_PENDING_SUSPEND | SS_SUSPENDED) )
				t->suspendState = 0;
				
		sp.sched_priority = priorities[unhand(cur->thread)->priority];
		if((errno = pthread_setschedparam(cur->tid, SchedulePolicy, &sp)) !=0)
			PerrorAbort("pthread_setschedparam");
        }

	
		
#else

	if ( --critSection == 0 ){
	
		/* No need to sync, there's nobody else running */
		
		for ( t=activeThreads; t; t = t->next ){
			if ( t->suspendState & (SS_PENDING_SUSPEND | SS_SUSPENDED) ){
				
				
				DBG( vm_thread, ("signal resume: %p (sus: %d blk: %d)\n",
								 t, t->suspendState, t->blockState));

				t->suspendState = SS_PENDING_RESUME;
				
				if ((errno = pthread_kill( t->tid, SIG_RESUME)) != 0 ){
					DBG( vm_thread, ("error sending RESUME signal to %p\n", t));
					PerrorAbort("pthread_kill");			
				}

				/* ack wait workaround, see TentercritSect remarks */
				if(sem_wait( &critSem) < 0)
					PerrorAbort("sem_wait");

			}
		}
	}
	
#endif /*#else */	

	DBG( vm_thread, ("exit crit section (%d)\n", critSection));
}

#if defined(MEM1) || defined(MEM2)

/*******************************************************************************
 * GC related stuff
 */

/*
 * Walk stacks of all threads, except of the current one (doesn't
 * make much sense since that's the GC itself, and it's task is to
 * get rid of garbage, not to pin it down - besides the fact that
 * its stack is a moving target).
 *
 * The call chain looks like this
 *
 *  startGC
 *    Kaffe_ThreadInterface.GcWalkThreads == TwalkThreads
 *      walkMemory
 *        gcFunctions[..].walk == walkThread
 *        
 */
void
TwalkThreads (void)
{
	nativeThread *t;

	DBG( vm_thread, ("start walking threads\n"));
	
	for ( t = activeThreads; t != NULL; t = t->next) {
		walkMemory((void*)t->thread);
	}
	
	DBG( vm_thread, ("end walking threads\n"));

}

/*      
 * Walk the thread's internal context. The stack is actually "walked down"
 * [base+size..base].
 * ? how useful is this for the GC thread itself ?
 */
void
TwalkThread ( Hjava_lang_Thread* thread )
{
	nativeThread  *t;
	nativeThread  *nt = NATIVE_THREAD( thread);
	void          *base;
	long          size;

	if ( !nt || !nt->active || !nt->stackMin ){
		return;
	}

	DBG( vm_thread, TMSG_LONG( "walking ", nt));

	if ( !nt->suspendState && !nt->blockState ){
		t = GET_CURRENT_THREAD(&t);
		if ( t == nt ){
			/* Ok, we do walk the gcMan for now, but just up to this point */
			nt->stackCur = &t;
		}
		else {
			/* everything else should be blocked or suspended by now */
			DBG( vm_thread, ("walking a running thread %p\n", nt));
			//tDump();
			//ABORT();
			return;
		}
	}
	else if ( nt->suspendState == SS_PENDING_SUSPEND ){
		DBG( vm_thread, ("pending suspend, walk whole stack\n"));
		/*
		 * Assuming the very next thing after a context switch to this thread
		 * would be calling the signal handler, we accept that case. Unfortunately,
		 * we don't have a stackCur, yet
		 */
#if defined(STACK_GROWS_UP)
		nt->stackCur = nt->stackMax;
#else
		nt->stackCur = nt->stackMin;
#endif	
	}

	if ( ((uintp) nt->stackCur < (uintp) nt->stackMin) ||
		(((uintp) nt->stackCur > (uintp) nt->stackMax)) ) {
		DBG( vm_thread, ("inconsistent stack\n"));
		tDump();
		ABORT();
	}

	markObject( unhand(thread)->jnireferences);
	markObject( unhand(thread)->exceptObj);

#if defined(STACK_GROWS_UP)
	base = nt->stackMin;
	size = (unsigned long)nt->stackCur - (unsigned long) base;
#else
	base = nt->stackCur;
	size = (uintp)nt->stackMax - (uintp) base;
#endif

	walkConservative( base, size);
}

#endif

/*
   Return the current stack boundaries. Note that 'low' and 'high'
   are arithmetical, i.e. not with respect to stack usage (not the bounds that 
   we ever attained in this run). Since it just works on the current thread, 
   this function cannot be used to GC-walk stacks, hence 'curr' is rather useless 
   in this context
 */
void
TstackInfo ( threadStackInfo* info )
{
	nativeThread* nt = GET_CURRENT_THREAD( &nt);
	
	info->low  = nt->stackMin;
	info->high = nt->stackMax;
	info->curr = &nt;
}


/***********************************************************************
 * create and initialize the ThreadInterface
 */

ThreadInterface Kaffe_ThreadInterface = {
	Tinit,

	Tattach,
	Tcreate,
	Tyield,
	Tprio, 
	Texit,
	0,
	Tfinalize,

	TcurrentJava,
	0,

	TenterCritSect,
	TexitCritSect,

#if defined(MEM1) || defined(MEM2)
	TwalkThreads,
	TwalkThread,
#else
	0,
	0,
#endif

	TnextFrame,

	TstackInfo
};

