/*
 * thread-impl.h - this is the part of the ThreadInterface implementation
 * we want to expose (by means of injection into the abstraction threads.h)
 *
 * Use it only for two purposes:
 *  (1) redefine generic interface function macros of threads.h in order
 *      to turn the runtime interface (completely or partly) into a
 *      compile-time interface
 *  (2) VERY RESTRICTIVE - define macros which can be used in clients to
 *      check for a certain subsystem (but JUST if there are strong reasons
 *      against using a clean interface function instead)
 *
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __thread_impl_h
#define __thread_impl_h

/* always make sure we see the generic macros first */
#include "threads.h"

/* Get us the required internal types we need for our inline funcs. This seems
 * to be the wrong order (impl including/exposing internal), but there's no way
 * around that if we want any benefit from inlines/macros
 */
#include "thread-internal.h"

#include "jdebug.h"

/***********************************************************************************
 * define the interface funtions we would like to provide as inlines
 */

static inline
void
tYield (void)
{
  DBG_ACTION( vm_thread, {
	nativeThread *cur = GET_CURRENT_THREAD( &cur);
	DBG( vm_thread, TMSG_SHORT( "yield ", cur));
  });

  if(sched_yield() < 0)
    perror("sched_yield"); /* convince me this is fatal, for now, not! */
}


/*
 * Note that any use of cached query results has to avoid locking,
 * or we are going to make things worse. Fortunately, we don't need
 * locking if we store the invariant stack address part (sBase) in
 * the nativeThread data itself (it's also constant). By using a local
 * var 'cur', we don't even assume that lastCur doesn't change while
 * we are executing in tCurrentXX. Our only requirements are 
 * atomic global pointer updates, and a fast, reliable way to identify
 * stacks (the weak part)
 */

static inline          
Hjava_lang_Thread*      
tCurrentJava (void)      
{
  nativeThread *cur = lastCur;
  char         stackAdr[6];  /* maybe gcc keeps this in a reg, screwing up & */

  /* check if lastcur is cur so may avoid calling GET_CURRENT_THREAD */
  if ( !ON_THREAD_STACK( cur, &(stackAdr[0])) ){ 
	  lastCur = cur = GET_CURRENT_THREAD( &cur);
	  assert( cur != NULL );
	  assert( cur->thread != NULL );
  }

  return cur->thread;
}


/* these are the function prototypes we export in our interface macros */
extern void Tinit (void);
extern void Tattach (Hjava_lang_Thread* thread);
extern void Tcreate ( Hjava_lang_Thread* thread, void* func );
extern void Tprio (Hjava_lang_Thread* thread, jint prio);
extern void Texit ( Hjava_lang_Thread* thread );
extern void Tfinalize (Hjava_lang_Thread* thread);
extern void TnextFrame (stackFrame* fm);
extern void TenterCritSect (void);
extern void TexitCritSect (void);
extern void TwalkThreads (void);
extern void TwalkThread ( Hjava_lang_Thread* thread );
extern void TstackInfo ( threadStackInfo* info );



/***********************************************************************************
 * turn the runtime interface into a compiletime interface (isn't really
 * required for all funcs, but will eventually replace the old interface)
 */

#undef  Kaffe_TI_init
#define Kaffe_TI_init()                 Tinit()

#undef  Kaffe_TI_attach
#define Kaffe_TI_attach(_thread)   	Tattach(_thread)

#undef  Kaffe_TI_create
#undef  Kaffe_TI_create
#define Kaffe_TI_create(_thread,_func)  Tcreate(_thread,_func)

#undef  Kaffe_TI_yield
#define Kaffe_TI_yield()                tYield()

#undef  Kaffe_TI_setPriority
#define Kaffe_TI_setPriority(_thread,_prio) Tprio(_thread,_prio)

#undef  Kaffe_TI_exit
#define Kaffe_TI_exit(_thread)          Texit(_thread)

#undef  Kaffe_TI_finalize
#define Kaffe_TI_finalize(_thread)      Tfinalize(_thread)

#undef  Kaffe_TI_currentJava
#define Kaffe_TI_currentJava()          tCurrentJava()

#undef  Kaffe_TI_enterCritSect
#define Kaffe_TI_enterCritSect()        TenterCritSect()

#undef  Kaffe_TI_exitCritSect
#define Kaffe_TI_exitCritSect()         TexitCritSect()

#undef  Kaffe_TI_GcWalkThreads
#define Kaffe_TI_GcWalkThreads()        TwalkThreads()

#undef  Kaffe_TI_GcWalkThread
#define Kaffe_TI_GcWalkThread(_thread)  TwalkThread(_thread)

#undef  Kaffe_TI_nextFrame
#define Kaffe_TI_nextFrame(_frame)      TnextFrame(_frame)

#undef  Kaffe_TI_stackInfo
#define Kaffe_TI_stackInfo(_stckInfo)   TstackInfo(_stckInfo)

#undef  Kaffe_TI_interrupt
#define Kaffe_TI_interrupt(_thread)     /* not supported yet */

/***********************************************************************************
 * defines for identifying the subsystem. NOTE - using this violates the interface
 * encapsulation, and is only justified if we otherwise would end up with new,
 * non-orthogonal interface functions
 */

#define UNIX_PTHREADS

#endif /* __thread_impl_h */
