/*
 * threads.c
 * Thread support.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-signal.h"
#include "jtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "lookup.h"
#include "threads.h"
#include "locks.h"
#include "exception.h"
#include "support.h"
#include "external.h"
#include "gc.h"
#include "md.h"
#include "jni.h"
#include "jnidefs.h"

#include "jdebug.h"

Hjava_lang_Thread* garbageman;
Hjava_lang_Thread* finalman;
Hjava_lang_ThreadGroup* standardGroup;

uint32 threadStackSize;

extern jboolean usingPosixLocks;

static void firstStartThread(void*);
static void createSystemDaemons(void);

Hjava_lang_Thread* createDaemon(void*, char*, int);


/*
 * Initialise the native thread system.
 */
void
initNativeThreads(void)
{
	Kaffe_TI_init();
}

/*
 * Initialise threads.
 */
void
initThreads(void)
{
	threadStackSize = Kaffe_JavaVMArgs[0].nativeStackSize;

#if defined(MEM2)
	ThreadClass->gc_type = GC_ALLOC_THREAD;
#else
	ThreadClass->flags |= GC_ALLOC_THREAD;
#endif

	/* Create base group */
	standardGroup = (Hjava_lang_ThreadGroup*)newObject(ThreadGroupClass);

	assert(standardGroup != 0);
	unhand(standardGroup)->parent = 0;
	unhand(standardGroup)->name = Kaffe_NewStringUTF(0, "main");
        GC_WRITE(standardGroup, unhand(standardGroup)->name); /* standardGroup->detailMessage */
	unhand(standardGroup)->maxPriority = java_lang_Thread_MAX_PRIORITY;
	unhand(standardGroup)->destroyed = 0;
	unhand(standardGroup)->daemon = 0;
	unhand(standardGroup)->vmAllowSuspension = 0;
	unhand(standardGroup)->nthreads = 0;
	unhand(standardGroup)->threads = 0;
	unhand(standardGroup)->ngroups = 0;
	unhand(standardGroup)->groups = (HArrayOfObject*)newArray(ThreadGroupClass, 0);
        GC_WRITE(standardGroup, unhand(standardGroup)->groups); /* standardGroup->groups */

	/* Allocate a thread to be the main thread */
	attachCurrentThread("main");

	/* Now create the system daemons */
	createSystemDaemons();
}

/*
 * Start a new thread running.
 */
void
startThread(Hjava_lang_Thread* tid)
{
	if (usingPosixLocks == true) {
		unhand(tid)->sem = gc_malloc_fixed_typed(sizeof(sem2posixLock), GC_ALLOC_THREADCTX);
		if (unhand(tid)->sem == 0) {
			throwPostedException();
		}
	}
	unhand(tid)->stackOverflowError = (Hjava_lang_Throwable*)newObject(SOClass);
        GC_WRITE( tid, unhand(tid)->stackOverflowError ); /* tid->stackOverflowError */
	unhand(tid)->outOfMemoryError = (Hjava_lang_Throwable*)newObject(OOMClass);
        GC_WRITE( tid, unhand(tid)->outOfMemoryError ); /* tid->outOfMemoryError */
	unhand(tid)->started = true;
	Kaffe_TI_create(tid, &firstStartThread);
}

/*
 * Attach a native thread to a java thread.
 */
void
attachCurrentThread(char* nm)
{
	Hjava_lang_Thread* tid;

	tid = (Hjava_lang_Thread*)newObject(ThreadClass);
	assert(tid != 0);

	unhand(tid)->name = Kaffe_NewStringUTF(0, nm);
        GC_WRITE( tid, unhand(tid)->name ); /* tid->name */
	unhand(tid)->priority = java_lang_Thread_NORM_PRIORITY;
	unhand(tid)->threadQ = 0;
	unhand(tid)->daemon = 0;
	unhand(tid)->interrupting = 0;
	unhand(tid)->target = 0;
	unhand(tid)->group = standardGroup;
        GC_WRITE( tid, unhand(tid)->group ); /* tid->group */
	unhand(tid)->started = true;
	if (usingPosixLocks == true) {
		unhand(tid)->sem = gc_malloc_fixed_typed(sizeof(sem2posixLock), GC_ALLOC_THREADCTX);
	}

	Kaffe_TI_attach(tid);

	unhand(tid)->stackOverflowError = (Hjava_lang_Throwable*)newObject(SOClass);
        GC_WRITE( tid, unhand(tid)->stackOverflowError ); /* tid->stackOverflowError */
	unhand(tid)->outOfMemoryError = (Hjava_lang_Throwable*)newObject(OOMClass);
        GC_WRITE( tid, unhand(tid)->outOfMemoryError ); /* tid->outOfMemoryError */

	/* Attach thread to threadGroup */
	Kaffe_CallVoidMethod(0, unhand(tid)->group, addThreadMethod, tid);
}

static
void
createSystemDaemons(void)
{
#if defined(GC_ENABLE) && (defined(MEM1) || defined(MEM2))
	/* Breaks encapsulation */
	{
		extern void gcMan(void*);
		extern void finaliserMan(void*);

		/* Start the GC daemons we need */
		finalman = createDaemon(&finaliserMan, "finaliser", THREAD_MAXPRIO-1);
		garbageman = createDaemon(&gcMan, "gc", THREAD_MAXPRIO);
		gc_mode = GC_ENABLED;
	}
#if defined(GC_ENABLE) && defined(MEM3)
	/* Breaks encapsulation */
	{
		extern void finaliserMan(void*);

		/* Start the GC daemons we need */
		finalman = createDaemon(&finaliserMan, "finaliser", THREAD_MAXPRIO-1);
		gc_mode = GC_ENABLED;
	}
#endif
#endif
#if defined(TIMELIMIT)
	{
		static void timelimitMain(void*);
		createDaemon(&timelimitMain, "timelimit", THREAD_MAXPRIO);
		fprintf(stderr, "THIS IS A TIME LIMITED COPY OF KAFFE OPENVM - IT WILL RUN FOR %d MINUTES\n", TIMELIMIT);

	}
#endif
}

/*
 * Start a daemon thread.
 */
Hjava_lang_Thread*
createDaemon(void* func, char* nm, int prio)
{
	Hjava_lang_Thread* tid;

	DBG( vm_thread, ("createDaemon %s\n",nm));

	/* Keep daemon threads as root objects */
	tid = (Hjava_lang_Thread*)newObject(ThreadClass);
	assert(tid != 0);

	unhand(tid)->name = Kaffe_NewStringUTF(0, nm);
        GC_WRITE( tid, unhand(tid)->name ); /* tid->name */
	unhand(tid)->priority = prio;
	unhand(tid)->threadQ = 0;
	unhand(tid)->daemon = 1;
	unhand(tid)->interrupting = 0;
	unhand(tid)->target = 0;
	unhand(tid)->group = 0;
	unhand(tid)->stackOverflowError = (Hjava_lang_Throwable*)newObject(SOClass);
        GC_WRITE( tid, unhand(tid)->stackOverflowError ); /* tid->stackOverflowError */
	unhand(tid)->outOfMemoryError = (Hjava_lang_Throwable*)newObject(OOMClass);
        GC_WRITE( tid, unhand(tid)->outOfMemoryError ); /* tid->outOfMemoryError */
	if (usingPosixLocks == true) {
		unhand(tid)->sem = gc_malloc_fixed_typed(sizeof(sem2posixLock), GC_ALLOC_THREADCTX);
	}

	Kaffe_TI_create( tid, func);

	return (tid);
}

/*
 * All threads start here.
 */
static
void
firstStartThread(void* arg)
{
	JNIEnv *env = (JNIEnv*)&Kaffe_JNIEnv;
	Hjava_lang_Thread* tid;
	jmethodID meth;
	jclass clazz;
	jobject eobj;
	jobject group;

	tid  = Kaffe_TI_currentJava();

	DBG( vm_thread, ("firstStartThread %x %s\n", tid, Kaffe_GetStringUTFChars(0, unhand(tid)->name)));

	clazz = (*env)->GetObjectClass(env, tid);
	meth = (*env)->GetMethodID(env, clazz, "run", "()V");
	if (meth != 0) {
		(*env)->CallVoidMethod(env, tid, meth);
	}
	eobj = (*env)->ExceptionOccurred(env);

	(*env)->ExceptionClear(env);
	if (eobj != 0 && unhand(tid)->dying == false) {
                group = unhand(tid)->group;
                clazz = (*env)->GetObjectClass(env, group);
                meth = (*env)->GetMethodID(env, clazz, "uncaughtException", "(Ljava/lang/Thread;Ljava/lang/Throwable;)V");

                unhand(tid)->dying = true;
                (*env)->CallVoidMethod(env, group, meth, tid, eobj);
                (*env)->ExceptionClear(env);
	}

	exitThread();
}


/*
 * Change thread priority.
 */
void
setPriorityThread(Hjava_lang_Thread* tid, int prio)
{
  /* since we set the java Thread field, we have to make sure it
   * has a legal value
   */
  if ( prio < java_lang_Thread_MIN_PRIORITY ){
	prio = java_lang_Thread_MIN_PRIORITY;
  }
  else if ( prio > java_lang_Thread_MAX_PRIORITY ){
	prio = java_lang_Thread_MAX_PRIORITY;
  }

  unhand(tid)->priority = prio;
  Kaffe_TI_setPriority(tid, prio);
}

/*
 * Terminate a thread.
 */
void
exitThread(void)
{
	JNIEnv *env = (JNIEnv*)&Kaffe_JNIEnv;
	Hjava_lang_Thread* tid;
	jclass clazz;
	jmethodID meth;

	tid  = Kaffe_TI_currentJava();
	clazz = (*env)->GetObjectClass(env, tid);
	meth = (*env)->GetMethodID(env, clazz, "finish", "()V");
	(*env)->CallVoidMethod(env, tid, meth);
	(*env)->ExceptionClear(env);

	Kaffe_TI_exit(tid);
}

/*
 * Terminate the given thread with violence.
 */
void
destroyThread(Hjava_lang_Thread* tid)
{
	Kaffe_TI_exit(tid);
}


/*
 * Get current thread's stack limit.
 */
void*
getCurrentStackLimit(void)
{
	threadStackInfo info;

	Kaffe_TI_stackInfo(&info);
#if defined(STACK_GROWS_UP)
	return ((void*)((uintp)info.high - STACKREDZONE));
#else
	return ((void*)((uintp)info.low + STACKREDZONE));
#endif
}

/*
 * Finalize a thread.
 *  This is to free the native thread context.
 */
void
finalizeThread(Hjava_lang_Thread* tid)
{
	Kaffe_TI_finalize(tid);
	
	if (usingPosixLocks == true) {
		gc_free_fixed(unhand(tid)->sem);
	}
}

/*
 * Interrupt a thread if we can.
 */
void
interruptThread(Hjava_lang_Thread* tid)
{
	Kaffe_TI_interrupt(tid);
}

/*
 * Is the given address on the current thread's stack?
 */
jboolean
onCurrentThreadStack(void* ptr)
{
	threadStackInfo info;

	Kaffe_TI_stackInfo(&info);
	if (ptr > info.low && ptr < info.high) {
		return (true);
	}
	else {
		return (false);
	}
}

#if defined(TIMELIMIT)

static iLock* timeout;

static
void
timelimitMain(void* arg)
{
	int iLockRoot;
	lockStaticMutex(&timeout);
	waitStaticCond(&timeout, (jlong)(TIMELIMIT * 60 * 1000));
	unlockStaticMutex(&timeout);
	fprintf(stderr, "TIME LIMIT EXPIRED - TERMINATING\n");
	exit(0);
}
#endif
