/**
 * clr.c - color management
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"
#include <limits.h>


/********************************************************************************
 * 256 PseudoColor visual
 * 
 * this is something that can't be done right - we don't know what
 * colors are installed, and which ones are essential for other apps. Without
 * disturbing other X citizens, we can't do much to strive for optimal RGB
 * distribution. What we ca do is
 * 
 *  - map web colors correctly (if they are in the colormap), this is
 *    why we use a 12x12x12 lookup cube resolution (6x6x6 would not enable
 *    us to reach colors > 216)
 *  - serve explicitly requested RGBs
 *  - do a best match search for populating our lookup color cube
 * 
 * If you need more colors, define IMPROVE_CMAP, which will grab as much
 * cells as possible (RW for now) to serve the worst mismatches
 */

#define MAX_VAL 255
#define MAX_REQUESTS 16
#define CLR_ACCEPT 64
#define CLR_RECONCILE 2000
#define MAX_GRAB 64

#undef IMPROVE_CMAP

#ifdef IMPROVE_CMAP
static void
updateBadList ( int* len, unsigned int* bad, unsigned int* dist, int d, int rgb )
{
	int i;

	for ( i=0; i<*len; i++ ){
		if ( dist[i] < d ){
			memmove( dist+i+1, dist+i, (*len - i) * sizeof(int));
			memmove( bad+i+1, bad+i, (*len - i) * sizeof(int));
			break;
		}
	}
	bad[i] = rgb;
	dist[i] = d;

	if ( *len < MAX_GRAB ) {
		(*len)++;
	}	
}
#endif

static void
setPaletteEntry ( Rgb2Pseudo* map, int idx,
				  unsigned char* lo, unsigned char* lb,
				  int r, int g, int b )
{
	int i;
	
	map->r[idx] = (unsigned char) r;
	map->g[idx] = (unsigned char) g;
	map->b[idx] = (unsigned char) b;

	/* update ordered lookup table and backlink table */
	for ( i=0; i<map->len; i++ ){
		if ( lo[i] > r ){
			memmove( lo+i+1, lo+i, map->len - i);
			memmove( lb+i+1, lb+i, map->len - i);
			break;
		}
	}
	lo[i] = r;
	lb[i] = idx;

	if ( map->len < 255 ) {
		map->len++;
	}
}

static inline
int binSearch ( unsigned char* list, int n, unsigned char v )
{
	int b = 0, t = n, c;
	
	while ( t > (b+1) ){
		c = (b + t) / 2;
		if ( v == list[c] ){
			return c;
		}
		else if ( v < list[c] ){
			t = c;
		}
		else {
			b = c;
		}
	}
	
	return b;
}


static inline
int exhaustiveSearch ( Rgb2Pseudo* map, unsigned char* list, int n,
					   unsigned char r, unsigned char g, unsigned char b )
{
	unsigned int d, dmin = 0xffffffff;
	int cr, cg, cb;
	int i, l;
	int idx = 0;
	
	for ( i=0; i<n; i++ ){
		l = list[i];
		cr = map->r[l];
		cg = map->g[l];
		cb = map->b[l];

		cr -= r;
		cg -= g;
		cb -= b;
		
		if ( (cr | cg | cb) == 0 ){
			return l;
		}
		
		cr *= cr;
		cg *= cg;
		cb *= cb;
		
		d = cr + cg + cb;
		
		if ( d < dmin ){
			dmin = d;
			idx = l;
		}
	}

	return idx;
}


static int
nearestColor ( Rgb2Pseudo* map, unsigned char* lo, unsigned char* lb,
			   int r, int g, int b, int dist )
{
	unsigned char   c[128];
	int  m=0;
	int  i, j, k, l;
	int  r0, r1, g0, g1, b0, b1;
	int  dr, dg, db;
	
	dr = dg = db = dist;

	r0 = r - dr;     if ( r0 < 0 ) r0 = 0;
	r1 = r + dr;     if ( r1 > MAX_VAL ) r1 = MAX_VAL;

	g0 = g - dg;     if ( g0 < 0 ) g0 = 0;
	g1 = g + dg;     if ( g1 > MAX_VAL ) g1 = MAX_VAL;

	b0 = b - db;     if ( b0 < 0 ) b0 = 0;
	b1 = b + db;     if ( b1 > MAX_VAL ) b1 = MAX_VAL;

	i = binSearch( lo, map->len, r);
	j = i;
	k = i-1;

	while ( 1 )  {
		for ( ; (j < map->len) && (lo[j] <= r1); j++ ){
			l = lb[j];
			if ( (map->g[l] >= g0) && (map->g[l] <= g1) ){
				if ( (map->b[l] >= b0) && (map->b[l] <= b1) ){
					c[m++] = l;
				}
			}
		}
		for ( ; (k >= 0) && (lo[k] >= r0); k-- ){
			l = lb[k];
			if ( (map->g[l] >= g0) && (map->g[l] <= g1) ){
				if ( (map->b[l] >= b0) && (map->b[l] <= b1) ){
					c[m++] = l;
				}
			}
		}
		
		if ( m > 0 ) {
			break;
		}
		
		/* hmm, nothing found, increase range  */
		r0 -= 32;         if ( r0 < 0 ) r0 = 0;
		r1 += 32;         if ( r1 > MAX_VAL ) r1 = MAX_VAL;
	}
	
	/* now go through our candidate list to find the nearest color */	
	return exhaustiveSearch( map, c, m, r, g, b);
} 

static inline unsigned int
colorDist ( int r0, int g0, int b0, int r1, int g1, int b1 )
{
	int dr = r0 - r1;
	int dg = g0 - g1;
	int db = b0 - b1;
	
	return (dr*dr + dg*dg + db*db);
}



/* our lookup points are d=255/11 based, since we have 12 elements available */
static unsigned char a12[12] = { 0, 23, 46, 70, 93, 116, 139, 162, 185, 209, 232, 255 };

static void
initColormap(JNIEnv * env, Toolkit * X, Colormap cm, Rgb2Pseudo * map)
{
	XColor  xclr;
	Visual  *v = DefaultVisualOfScreen(DefaultScreenOfDisplay(X->dsp));
	int     i, j, k, pix;
	unsigned int  blackSeen = 0;
	
	unsigned char lo[256];
	unsigned char lb[256];

	jclass clazz;
	jfieldID fid;
	jarray rgbRequests = 0;
	jboolean isCopy;
	jint *jrgbs = 0;
	int navail, nReq = 0, m;
	int r, g, b;
	unsigned long pixels[MAX_GRAB];
	jint req[MAX_REQUESTS];
	unsigned long planeMasks[1];
	
	int alloc;
	
 #ifdef IMPROVE_CMAP
	int d, firstPass = 1;
	unsigned int dist[MAX_GRAB];
	unsigned int bad[MAX_GRAB];
	int nbad = 0;
#endif

	/* get the kaffe.awt.DefaultsRGB.RgbRequests field */
	if ((clazz = (*env)->FindClass(env, "kaffe.awt.Defaults"))) {
		if ((fid = (*env)->GetStaticFieldID(env, clazz, "RgbRequests", "[I"))) {
			if ((rgbRequests = (*env)->GetStaticObjectField(env, clazz, fid))) {
				jrgbs = (*env)->GetIntArrayElements(env, rgbRequests, &isCopy);
				nReq = (*env)->GetArrayLength(env, rgbRequests);

				if (nReq > MAX_REQUESTS)
				  nReq = MAX_REQUESTS;

				memcpy(req, jrgbs, nReq * sizeof(jint));
				(*env)->ReleaseIntArrayElements(env, rgbRequests, jrgbs, JNI_ABORT);
			}
		}
	}
	
	/* get our X color map and store it away */
	for (i = 0; i < v->map_entries; i++) {
		xclr.pixel = i;
		XQueryColor( X->dsp, cm, &xclr);
		
		if ( (xclr.red | xclr.green | xclr.blue) == 0 ){
			if ( blackSeen ) {
				break;
			}
			blackSeen = 1;
		}
		
		setPaletteEntry( map, i, lo, lb, (xclr.red>>8), (xclr.green>>8), (xclr.blue>>8));

		/* check if this resolves a request */
		for (m = 0; m < nReq; m++) {
			if (req[m] &&
				colorDist( JRED(req[m]), JGREEN(req[m]), JBLUE(req[m]),
						   xclr.red>>8, xclr.green>>8, xclr.blue>>8) < CLR_ACCEPT) {
				req[m] = 0;                  /* mark color request as satisfied */
				break;
			}
		}		
	}
	
	/*
	 * Determine how many RW cells there are available
	 * (maybe we should switch over to read-only color cells, to
	 * make results more reproducable)
	 */
	for (navail = MAX_GRAB; navail; navail--) {
		if (XAllocColorCells(X->dsp, cm, True, planeMasks, 0, pixels, navail))
		  break;
	}
	/* don't touch already set cells */
	for ( j=0; j<navail && (pixels[j] < v->map_entries); j++ );
	navail -= j;
	alloc = j;
	
	
	/* set cells of not-yet satisfied rgb requests */
	for (j = 0; navail && (j < nReq); j++) {
		if (req[j]) {
			r = JRED(req[j]);
			g = JGREEN(req[j]);
			b = JBLUE(req[j]);

			xclr.pixel = pixels[alloc++];
			xclr.red = r << 8;
			xclr.green = g << 8;
			xclr.blue = b << 8;
			if ( !XStoreColor(X->dsp, cm, &xclr) ) {
				navail = 0;
				break;
			}

			setPaletteEntry( map, xclr.pixel, lo, lb, r, g, b);
			navail--;
		}
	}

#ifdef IMPROVE_CMAP
	again:
#endif

	/* now initialize our lookup cube from it */
	for ( i=0; i<N_CELLS; i++ ){
		for ( j=0; j<N_CELLS; j++ ){
			for ( k=0; k<N_CELLS; k++ ){
				pix = nearestColor( map, lo, lb, a12[i], a12[j], a12[k], 51);
#ifdef IMPROVE_CMAP
				d = colorDist( a12[i], a12[j], a12[k], map->r[pix], map->g[pix], map->b[pix]);
			    if ( firstPass && navail && (d > CLR_RECONCILE) ) { 
					updateBadList( &nbad, bad, dist, d,
								  JRGB(a12[i],a12[j],a12[k]));
				}
#endif	
				map->pix[i][j][k] = pix;
			}
		}
	}
	
#ifdef IMPROVE_CMAP
	if ( firstPass ) {
		firstPass = 0;
		for ( i=0; navail && (i < nbad); i++ ){
			r = JRED( bad[i]);
			g = JGREEN( bad[i]);
			b = JBLUE( bad[i]);

			xclr.pixel = pixels[alloc++];
			xclr.red = r << 8;
			xclr.green = g << 8;
			xclr.blue = b << 8;
			xclr.flags = DoRed | DoGreen | DoBlue;
			if ( !XStoreColor(X->dsp, cm, &xclr) ){
				break;
			}

			setPaletteEntry( map, xclr.pixel, lo, lb, r, g, b);
			navail--;
		}
	
		goto again;
	}
#endif
}


static Rgb2Pseudo *
initRgb2Pseudo(JNIEnv * env, jclass clazz, Toolkit * X)
{
	Colormap dcm;
	Rgb2Pseudo *map;

	dcm = DefaultColormapOfScreen(DefaultScreenOfDisplay(X->dsp));
	map = (Rgb2Pseudo *) AWT_MALLOC(sizeof(Rgb2Pseudo));

	initColormap(env, X, dcm, map);
	X->colorMode = CM_PSEUDO;

	return map;
}


/********************************************************************************
 * TrueColor (RGB) visual
 */

static Rgb2True *
initRgb2True(JNIEnv * env, jclass clazz, Toolkit * X)
{
	Visual *v = DefaultVisualOfScreen(DefaultScreenOfDisplay(X->dsp));
	unsigned int m;
	int nRed, nGreen, nBlue;
	int iRed, iGreen, iBlue;
	int n;
	Rgb2True *map = 0;

	if ((v->blue_mask == 0xff) && (v->green_mask == 0xff00) && (v->red_mask == 0xff0000)) {
		/*
		 * This is our favourite case - a direct 8-8-8 native rgb. It could be handled as
		 * a 0,0 TrueColor conversion, but (esp. for image manipulations) we can save a lot
		 * of computation by a special TrueColor mode
		 */
		DBG(awt_clr, ("AWT color mode: CM_TRUE_888\n"));
		X->colorMode = CM_TRUE_888;
	}
	else {
		/*
		 * There is either a rearrangement or a non-8 bit color component involved,
		 * get start index and length of each color component. Note that the Rgb2True
		 * struct is used to compute pixelvalues from Java rgbs, i.e. the mask and shift
		 * values are relative to the Java 8-8-8 RGB.
		 */
		map = (Rgb2True *) AWT_MALLOC(sizeof(Rgb2True));

		for (iBlue = 0, m = v->blue_mask; (m & 1) == 0; iBlue++, m >>= 1);
		for (nBlue = 0; m; nBlue++, m >>= 1);

		for (iGreen = 0, m = v->green_mask; (m & 1) == 0; iGreen++, m >>= 1);
		for (nGreen = 0; m; nGreen++, m >>= 1);

		for (iRed = 0, m = v->red_mask; (m & 1) == 0; iRed++, m >>= 1);
		for (nRed = 0; m; nRed++, m >>= 1);

		map->blueShift = 8 - (iBlue + nBlue);
		if (nBlue < 8) {                   /* color reduction */
			n = 8 - nBlue;
			map->blueMask = (0xff >> n) << n;
		}
		else {                             /* color expansion */
			map->blueMask = 0xff;
		}

		map->greenShift = 16 - (iGreen + nGreen);
		if (nGreen < 8) {                  /* color reduction */
			n = 8 + (8 - nGreen);
			map->greenMask = (0xff00 >> n) << n;
		}
		else {                             /* color expansion */
			map->greenMask = 0xff00;
		}

		map->redShift = 24 - (iRed + nRed);
		if (nRed < 8) {                    /* color reduction */
			n = 16 + (8 - nRed);
			map->redMask = (0xff0000 >> n) << n;
		}
		else {                             /* color expansion */
			map->redMask = 0xff0000;
		}

		X->colorMode = CM_TRUE;

		DBG(awt_clr, ("AWT color mode: CM_TRUE\n"));
		DBG(awt_clr, ("    red:   %8x, %d\n", map->redMask, map->redShift));
		DBG(awt_clr, ("    green: %8x, %d\n", map->greenMask, map->greenShift));
		DBG(awt_clr, ("    blue:  %8x, %d\n", map->blueMask, map->blueShift));
	}

	return map;
}


/********************************************************************************
 * DirectColor (RGB) visual: each pixel value is composed of three, non-overlapping
 * colormap indices.
 * Again, our policy is to go with the default colormap, and not to swap it (so that
 * we don't disturb other desktop citizens).
 * This rather simple implementation does not allocate cells for mismatches yet
 *
 * We could save some space in case we have a DirectColor visual with a depth < 24,
 * but that would complicate pixelValue() and probably is rather unusual, anyway
 * (since DirectVisual is a expensive beast to be found on high-end HW)
 *
 * However, we do make some assumptions here which might be a bit too optimistic
 * (black being at index 0, free cells being concentracted at the end of the cmap,
 * with a 0-value)
 */

static void 
fillUpPartMap(unsigned char *pix, unsigned char *val)
{
	int i, j, k, i2;

	for (i = 1, k = 0; i < N_DIRECT; i++) {
		if (pix[i] == 0) {
			for (j = i + 1; (j < N_DIRECT) && (pix[j] == 0); j++);
			if (j == N_DIRECT) {             /* last one, fill up rest */
				for (; i < j; i++) {
					pix[i] = pix[k];
					val[pix[i]] = k;
				}
			}
			else {
				i2 = (i + j) / 2;
				for (; i < i2; i++) {
					pix[i] = pix[k];
					val[pix[i]] = k;
				}
				for (; i < j; i++) {
					pix[i] = pix[j];
					val[pix[i]] = j;
					k = j;
				}
			}
		}
		else {
			k = i;
		}
	}
}

static void
setPartMapFromDMap(Toolkit * X, Colormap dcm,
                   int component, int nIdx, int idxShift,
                   unsigned char *pix, unsigned char *val)
{
	XColor xclr;
	int i, idx;
	unsigned short *v;

	if (component == 0)                  /* red */
	  v = &xclr.red;
	else if (component == 1)             /* green */
	  v = &xclr.green;
	else                                 /* blue */
	  v = &xclr.blue;

	for (i = 0; i < nIdx; i++) {
		xclr.pixel = i << idxShift;
		XQueryColor(X->dsp, dcm, &xclr);

		if (i && !*v)
		  break;                           /* skip free cells (assuming 0: black) */

		idx = ROUND_SHORT2CHAR(*v);
		pix[idx] = i;
		val[i] = idx;
	}
}

static Rgb2Direct *
initRgb2Direct(JNIEnv * env, jclass clazz, Toolkit * X)
{
	Visual *v = DefaultVisualOfScreen(DefaultScreenOfDisplay(X->dsp));
	Rgb2Direct *map = (Rgb2Direct *) AWT_MALLOC(sizeof(Rgb2Direct));
	Colormap dcm = DefaultColormapOfScreen(DefaultScreenOfDisplay(X->dsp));
	int iBlue, nBlue, iGreen, nGreen, iRed, nRed;
	int i, m;

	for (i = 0; i < N_DIRECT; i++) {
		map->bluePix[i] = map->greenPix[i] = map->redPix[i] = 0;
	}

	for (iBlue = 0, m = v->blue_mask; (m & 1) == 0; iBlue++, m >>= 1);
	for (nBlue = 0; m; nBlue++, m >>= 1);
	
	for (iGreen = 0, m = v->green_mask; (m & 1) == 0; iGreen++, m >>= 1);
	for (nGreen = 0; m; nGreen++, m >>= 1);

	for (iRed = 0, m = v->red_mask; (m & 1) == 0; iRed++, m >>= 1);
	for (nRed = 0; m; nRed++, m >>= 1);

	/* we simply get the sizes of component colormaps from the visual masks */
	map->nBlue = v->blue_mask >> iBlue;
	map->nGreen = v->green_mask >> iGreen;
	map->nRed = v->red_mask >> iRed;

	/* the shifts are to convert component indices into a compound pixelvalue */
	map->blueShift = iBlue;
	map->greenShift = iGreen;
	map->redShift = iRed;

	/*
	 * phase 1: start to populate our component maps (R/G/B -> index) from dcm
	 */
	setPartMapFromDMap(X, dcm, 0, map->nRed, map->redShift, map->redPix, map->red);
	setPartMapFromDMap(X, dcm, 1, map->nGreen, map->greenShift, map->greenPix, map->green);
	setPartMapFromDMap(X, dcm, 2, map->nBlue, map->blueShift, map->bluePix, map->blue);

	/*
	 * phase 2: fill up missing entries
	 * (based on the somewhat optimistic assumption that free cells are at the upper end)
	 */
	fillUpPartMap(map->bluePix, map->blue);
	fillUpPartMap(map->greenPix, map->green);
	fillUpPartMap(map->redPix, map->red);

#ifdef NEVER
	for (i = 0; i < N_DIRECT; i++) {
		printf("%2x :   %3d,%2x  %3d,%2x  %3d,%2x\n", i,
			   map->redPix[i], map->red[map->redPix[i]],
			   map->greenPix[i], map->green[map->greenPix[i]],
			   map->bluePix[i], map->blue[map->bluePix[i]]);
	}
#endif

	X->colorMode = CM_DIRECT;

	return map;
}


/********************************************************************************
 * common funcs
 */

int
modPixelValue ( Toolkit* X, int pix, int mod )
{
	jint modPix=0;
	int r, g, b;

	switch ( mod ){
	case BRIGHTER: mod = 4; break;
	case DARKER:   mod = 2; break;
	default:       return pix;
	}

	rgbValues( X, pix, &r, &g, &b);

	r = r * mod / 3;
	g = g * mod / 3;
	b = b * mod / 3;
	if ( r > 0xff ) r = 0xff;
	if ( g > 0xff ) g = 0xff;
	if ( b > 0xff ) b = 0xff;

	modPix = pixelValue( X, JRGB( r, g, b));

	return modPix;
}


void
initColorMapping(JNIEnv * env, jclass clazz, Toolkit * X)
{
	Visual *v = DefaultVisualOfScreen(DefaultScreenOfDisplay(X->dsp));

	DBG(awt_clr, ("X visual:\n"));
	DBG(awt_clr, ("  id:          %d\n", v->visualid));
	DBG(awt_clr, ("  class:       %d\n", v->class));
	DBG(awt_clr, ("  red_mask     %x\n", v->red_mask));
	DBG(awt_clr, ("  green_mask   %x\n", v->green_mask));
	DBG(awt_clr, ("  blue_mask    %x\n", v->blue_mask));
	DBG(awt_clr, ("  bits_per_rgb %x\n", v->bits_per_rgb));
	DBG(awt_clr, ("  map_entries  %d\n", v->map_entries));

	/* check for directly supported color modes / visuals */
	switch (v->class) {
	 case DirectColor:
		X->dclr = initRgb2Direct(env, clazz, X);
		break;
	 case TrueColor:
		X->tclr = initRgb2True(env, clazz, X);
		break;
	 case PseudoColor:
		X->pclr = initRgb2Pseudo(env, clazz, X);
		break;
	 default:
		X->colorMode = CM_GENERIC;
	}

	X->black = BlackPixel( X->dsp, DefaultScreen( X->dsp));
	X->white = WhitePixel( X->dsp, DefaultScreen( X->dsp));

	DBG(awt_clr, ("colorMode: %d\n", X->colorMode));
}

jobject
getSystemColorModel ( JNIEnv * env, jclass clazz, Toolkit* X, int trans )
{
	static jarray rgbArray;
	
	jobject cm = 0;
	jclass cmClazz;
	jmethodID cmCtorId;
	Visual *v = DefaultVisualOfScreen(DefaultScreenOfDisplay(X->dsp));
	jint   *rgbs;
	jboolean isCopy;
	int    i;
	

	if (!X->colorMode){
		initColorMapping(env, clazz, X);
	}
	
	switch (v->class) {
	 case DirectColor:
		break;

	 case TrueColor:
		cmClazz = (*env)->FindClass(env, "java.awt.image.DirectColorModel");
		cmCtorId = (*env)->GetMethodID(env, cmClazz, "<init>", "(IIIII)V");
		cm = (*env)->NewObject(env, cmClazz, cmCtorId,
							   v->bits_per_rgb, v->red_mask, v->green_mask, v->blue_mask, 0);
		break;

	 case PseudoColor:
		cmClazz = (*env)->FindClass(env, "java.awt.image.IndexColorModel");
		cmCtorId = (*env)->GetMethodID(env, cmClazz, "<init>", "(I[II)V");
		
		if ( !rgbArray ) {
			/* we need to do this just once, and subsequently share
			 * the array between all SystemColorModel based instances
			 */
			rgbArray = (*env)->NewIntArray( env, 256);
			rgbs = (*env)->GetIntArrayElements( env, rgbArray, &isCopy);

			for ( i=0; i<256; i++ ){
				rgbs[i] = JRGB( X->pclr->r[i], X->pclr->g[i], X->pclr->b[i]) | 0xff000000;
			}

			(*env)->ReleaseIntArrayElements( env, rgbArray, rgbs, 0);
			(*env)->NewGlobalRef( env, rgbArray);
		}

		cm = (*env)->NewObject( env, cmClazz, cmCtorId, 8, rgbArray, trans);
		break;

	 default:
	}

	
	return cm;
}

/********************************************************************************
 * exported functions
 */

jint
Java_java_awt_Toolkit_clrGetPixelValue(JNIEnv * env, jclass clazz, jint rgb)
{
	jint pix;

	/*
	 * We do this deferred to avoid class init problems with Defaults<->Color
	 * (the notorious DefaultsRGB workaround)
	 */
	if (!X->colorMode) {
		initColorMapping(env, clazz, X);
	}

	pix = pixelValue(X, rgb);
	DBG(awt_clr, ("clrGetPixelValue: %8x -> %x (%d)\n", rgb, pix, pix));

	return pix;
}

void
Java_java_awt_Toolkit_clrSetSystemColors(JNIEnv * env, jclass clazz, jarray sysClrs)
{
#ifdef NEVER                           /* maybe this could be initialized via X resources */
	jboolean isCopy;
	jint *rgbs = (*env)->GetIntArrayElements(env, sysClrs, &isCopy);
	int len = (*env)->GetArrayLength(env, sysClrs);

	rgbs[0] = 0;                         /* desktop */
	rgbs[1] = 0;                         /* active_caption */
	rgbs[2] = 0;                         /* active_caption_text */
	rgbs[3] = 0;                         /* active_caption_border */
	rgbs[4] = 0;                         /* inactive_caption */
	rgbs[5] = 0;                         /* inactive_caption_text */
	rgbs[6] = 0;                         /* inactive_caption_border */
	rgbs[7] = 0;                         /* window */
	rgbs[8] = 0;                         /* window_border */
	rgbs[9] = 0;                         /* window_text */
	rgbs[10] = 0;                        /* menu */
	rgbs[11] = 0;                        /* menu_text */
	rgbs[12] = 0;                        /* text */
	rgbs[13] = 0;                        /* text_text */
	rgbs[14] = 0;                        /* text_highlight */
	rgbs[15] = 0;                        /* text_highlight_text */
	rgbs[16] = 0;                        /* text_inactive_text */
	rgbs[17] = 0;                        /* control */
	rgbs[18] = 0;                        /* control_text */
	rgbs[19] = 0;                        /* control_highlight */
	rgbs[20] = 0;                        /* control_lt_highlight */
	rgbs[21] = 0;                        /* control_shadow */
	rgbs[22] = 0;                        /* control_dk_shadow */
	rgbs[23] = 0;                        /* scrollbar */
	rgbs[24] = 0;                        /* info */
	rgbs[25] = 0;                        /* info_text */
	
	(*env)->ReleaseIntArrayElements(env, sysClrs, rgbs, 0);
#endif
}

/*
 * we need to implement brighter / darker in the native layer because the usual
 * arithmetic Rgb brightening/darkening isn't really useful if it comes to
 * limited PseudoColor visuals (e.g. a primitive 16 color VGA wouldn't produce any
 * usable results). Even 256 colormaps suffer from that
 */
jlong
Java_java_awt_Toolkit_clrModBrightness(JNIEnv * env, jclass clazz, jint pix, int mod)
{
	unsigned int r, g, b;
	jint modRgb, modPix;

	modPix = modPixelValue( X, pix, mod);

	rgbValues( X, pix, &r, &g, &b);
	modRgb = JRGB(r, g, b);

	return (((jlong) modPix << 32) | modRgb);
}

							 
jobject
Java_java_awt_Toolkit_clrGetColorModel ( JNIEnv * env, jclass clazz )
{
	return getSystemColorModel( env, clazz, X, -1);
}
