/**
 * fnt.c - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

/*******************************************************************************
 * Font support
 */


/*
 * these are (hopefully) the most usual XFLD font weight names, ordered
 * in our preferred lookup sequence (from PLAIN to BOLD)
 */
char *weight[] =
{"medium", "normal", "regular", "thin", "light",
 "black", "demibold", "heavy", "extrabold", "extrablack", "bold"};
#define NWEIGHT 11

/*
 * these are (more confident) the XFLD slant names, ordered in our
 * preferred lookup sequence (from PLAIN to ITALIC)
 */
char *slant[] =
{"r", "i", "o"};
#define NSLANT 3

/*
 * these are finally the size offsets we want to try
 */
int dsize[] = { 0, 10, -10, 20, -20, 30, -30, 40, -40 };
#define NDSIZE 9


/* our ultima ratio if we really can't find a suitable font */
XFontStruct *defaultFont;

/*
 * since we use XFontStructs for both Font and FontMetrics objects, we
 * need to determine when it is safe to free the XFontStruct. The alternative
 * would be to use a WeakReferences in Font/FontMetrics, but our classes have
 * to deal with environments where there are no WeakReferences.
 * The alternative would be to wrap font objects, but who says that Xlib
 * doesn't do some internal caching and returns the same XFontStruct twice
 * (unlikely, but possible, hence Murphy).
 * Since it it's unlikely that we have programs using hundreths of resident
 * font objects, we are probably better of if we use a simple array instead
 * of a hashtable
 */

#define MAX_FONTS 40
XFontStruct*          inUseFs[MAX_FONTS];
unsigned short        inUseCount[MAX_FONTS];

static inline
int registerFs ( XFontStruct* fs )
{
	register int i;

	/* check if there already is an entry */
	for ( i=0; i<MAX_FONTS; i++ ){
		if ( inUseFs[i] == fs ){
			inUseCount[i]++;
			return 1;
		}
	}

	/* no, there isn't - add a new entry */
	for ( i=0; i<MAX_FONTS; i++ ){
		if ( !inUseFs[i] ){
			inUseFs[i] = fs;
			inUseCount[i] = 1;
			return 1;
		}
	}

	/* outch, out of font registry space */
	return 0;
}

static inline
void unregisterFs ( XFontStruct* fs )
{
	register int i;

	for ( i=0; i<MAX_FONTS; i++ ){
		if ( inUseFs[i] == fs ){
			if ( (--inUseCount[i]) == 0 ){
				DBG( awt_fnt, ("free native font: %p\n", fs));
				inUseFs[i] = 0;
				XFreeFont(X->dsp, fs);
			}
			return;
		}
	}	
}


/*******************************************************************************
 * internal funcs
 */

/*
 * set up everything we need to subsequentlt map the Java-side font specs (XFLD
 * template, attributes, and point sizes) into X fonts. Ultimately, we should
 * use XDefaults to override the specs passes in from Defaults.java, but let's
 * go for a long way to avoid any user config.
 */
void
initFonts ( Toolkit* X )
{
	char buf[128];
	char spec[] = "-adobe-courier-medium-r-*-*-*-%d-%d-%d-*-*-*-*";
	XFontStruct *fs = 0;
	unsigned long res;

	/* Check which dpi font resolution we should look for. We need this,
	 * because it's a bad idea to leave it up to the X font server, which
	 * font class to use (usually ends up in font discontinuities, i.e. a
	 * enlarged point-size suddenly producing a smaller font).
	 * Unfortunately, we have to use some heuristics, here, since the 100/75 dpi
	 * distinction doesn't really match existing monitors, anymore (most of them
	 * are much closer to 100dpi, with X set up to use 75dpi fonts).
	 * Even more unfortunately, "resx" and "resy" seem to be used rather as a hint
	 * by existing font servers, i.e. they don't enforce them. We better check
	 * if our preferred resolution is available, and if we have to adjust "ptSz"
	 * requests.
	 */

	X->dpi = (DisplayWidth(X->dsp, DefaultScreen(X->dsp)) >= 1024) ? 100 : 75;
	X->ptSzDelta = 0;

	sprintf( buf, spec, 120, X->dpi, X->dpi);
	fs = XLoadQueryFont(X->dsp, buf);

	if ( fs ){
		XGetFontProperty( fs, XA_RESOLUTION, &res);
		/* XA_RESOLUTION is the nominal pixel-per-point res (in hundredths),
		 * a inch has 72.27 points, (-> 100 dpi: 138, 75dpi: 103)
		 */
		res = (res > 120) ? 100 : 75;
		XFreeFont( X->dsp, fs);
	}
	else {
		/* hmm, if we don't have such a basic beast, it's probably because this
		 * font server *is* picky about resolutions - swap them
		 */
		res = (X->dpi == 100) ? 75 : 100;
	}

	/* the "ptSz" adjustment is another heuristic measure, based on the XFree86
	 * fonts
	 */
	if ( res < X->dpi ){
		/* we obviously don't have the 100 dpi fonts, switch to
		 * 75 dpi and decrease size requests (this is the common case,
		 * the server is still on the default 75 dpi setting, 100 dpi fonts
		 * have not been installed (do it!)
		 */
		X->dpi = 75;
		X->ptSzDelta = -40;
	}
	else if ( res > X->dpi ){
		/* rare, but let's check if somebody went solely with 100 dpi on
		 * a supposed to be 75 dpi environment
		 */
		X->dpi = 100;
		X->ptSzDelta += 40;
	}

	/* now get us a default font (which has to be guaranteed as a fallback) */
	defaultFont = XLoadQueryFont( X->dsp, "fixed");
	if ( !defaultFont ){
		DBG( awt_fnt, ("font panic - no defaultFont 'fixed' found\n"));
	}
	else {
		registerFs( defaultFont);
	}
}


/*
 * Our font lookup strategy is like this: every mapping that can be known
 * a priori goes into kaffe.awt.Defaults, everything that might be specified
 * at runtime (sizes, slant, weight) has to be handled here. In detail:
 * 
 * (1) foundry, family, encoding and everything EXCEPT of weight,slant,size
 *     are specified in kaffe.awt.Defaults FontSpec strings, which are used
 *     as simple XLFD C format specifications
 *     (e.g."-adobe-helvetica-%s-%s-*-*-*-%d-%d-%d-*-*-*-*")
 * (2) since both BOLD and ITALIC map to more than one XFLD attribute value
 *     (e.g. ITALIC-> "o", "i"), and many fonts are not available in all
 *     sizes, we have to make up a sequence of alternate specs to search
 * (3) the search first tries to vary sizes, then weights, then slants
 * (4) in case all of this fails, it directly tries the font spec
 * (5) if this fails, too, it backs up to what is supposed to be a safe X
 *     standard font (fix size)
 *
 * Once again - we don't try to deduce XFLD family names from the Java names,
 * that's in kaffe.awt.Defaults, and Defaults is meant to be the thing which
 * adapts the AWT to your X installation/preferences. Don't lament, modify it!
 */

void*
Java_java_awt_Toolkit_fntInitFont ( JNIEnv* env, jclass clazz, jstring jSpec,
jint style, jint size )
{
	int i, j, k, i0, i1, j0, j1, di, dj;
	char buf[160];
	char *spec = java2CString(env, X, jSpec);
	XFontStruct *fs = 0;

	DBG( awt_fnt, ("initFont: spec:%s style:%x, size:%d\n", spec, style, size));

	size *= 10;                          /* convert into X pointsizes */
	size += X->ptSzDelta;

	if (style & 0x1) {                   /* we have a Font.BOLD request */
		i0 = NWEIGHT - 1;
		i1 = -1;
		di = -1;
	}
	else {
		i0 = 0;
		i1 = NWEIGHT;
		di = 1;
	}

	if (style & 0x2) {                   /* we have a Font.ITALIC request */
		j0 = NSLANT - 1;
		j1 = -1;
		dj = -1;
	}
	else {
		j0 = 0;
		j1 = NSLANT;
		dj = 1;
	}

	for (j = j0; !fs && (j != j1); j += dj) {
		for (i = i0; !fs && (i != i1); i += di) {
			for (k = 0; !fs && (k < NDSIZE); k++) {
				sprintf(buf, spec, weight[i], slant[j], size + dsize[k], X->dpi, X->dpi);
				DBG(awt_fnt, ("look up font: %s\n", buf));
				fs = XLoadQueryFont(X->dsp, buf);
			}
		}
	}

	if (!fs) {
		/* now we are getting desperate, try the spec directly (without vars) */
		fs = XLoadQueryFont(X->dsp, spec);
	}

	if (!fs) {
		/* if we still have no font, our last line of defense is the default one */
		DBG(awt_fnt, ("cannot load font: %s (backup to default)\n", buf));
		fs = defaultFont;
	}

	registerFs( fs);

	return (void *) fs;
}

void
Java_java_awt_Toolkit_fntFreeFont(JNIEnv * env, jclass clazz, XFontStruct * fs)
{
	unregisterFs( fs);
}


/*******************************************************************************
 * FontMetrics support
 */

void *
Java_java_awt_Toolkit_fntInitFontMetrics( JNIEnv * env, jclass clazz,
										  XFontStruct* fs, jobject metrics )
{
	jclass fmClazz = (*env)->GetObjectClass( env, metrics);
	jfieldID id;

	DBG( awt_fnt, ("initFontMetrics: %x\n", fs));

	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAdvance", "I")) ){
		(*env)->SetIntField( env, metrics, id, fs->max_bounds.width);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "leading", "I")) ){
		(*env)->SetIntField( env, metrics, id, 0);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "ascent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fs->ascent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAscent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fs->max_bounds.ascent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "descent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fs->descent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxDescent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fs->max_bounds.descent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "height", "I")) ){
		(*env)->SetIntField( env, metrics, id,
							 fs->ascent + fs->descent + 1);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "fixedWidth", "I")) ){
		(*env)->SetIntField( env, metrics, id,
							 (fs->min_bounds.width == fs->max_bounds.width) ?
							 fs->max_bounds.width : 0);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "isWideFont", "Z")) ){
		(*env)->SetBooleanField( env, metrics, id,
								 (fs->min_byte1 | fs->max_byte1));
	}

	DBG( awt_fnt, (" -> metrics: %x\n", fs));

	registerFs( fs);

	return fs;
}

void
Java_java_awt_Toolkit_fntFreeFontMetrics(JNIEnv * env, jclass clazz, XFontStruct * fs)
{
	unregisterFs( fs);
}


jobject
Java_java_awt_Toolkit_fntGetWidths(JNIEnv * env, jclass clazz, XFontStruct * fs)
{
	int n = 256;
	jarray widths;
	jint *jw;
	jboolean isCopy;
	register int i, j;

	widths = (*env)->NewIntArray(env, 256);
	jw = (*env)->GetIntArrayElements(env, widths, &isCopy);

	if (fs->max_char_or_byte2 < n)
		n = fs->max_char_or_byte2;

	if (fs->min_bounds.width == fs->max_bounds.width) {
		for (i = fs->min_char_or_byte2, j = 0; i < n; i++, j++)
			jw[i] = fs->max_bounds.width;
	}
	else {
		for (i = fs->min_char_or_byte2, j = 0; i < n; i++, j++)
			jw[i] = fs->per_char[j].width;
	}

	(*env)->ReleaseIntArrayElements(env, widths, jw, 0);

	return widths;
}


jint
Java_java_awt_Toolkit_fntBytesWidth(JNIEnv * env, jclass clazz, XFontStruct * fs,
                                    jarray jBytes, jint off, jint len)
{
	jboolean isCopy;
	jbyte *jb = (*env)->GetByteArrayElements(env, jBytes, &isCopy);
	int n = (*env)->GetArrayLength(env, jBytes);
	int w;

	if (off + len > n)
		len = n - off;

	w = XTextWidth(fs, jb + off, len);

	(*env)->ReleaseByteArrayElements(env, jBytes, jb, JNI_ABORT);
	return w;
}

jint
Java_java_awt_Toolkit_fntCharWidth(JNIEnv * env, jclass clazz, XFontStruct * fs, jchar jChar)
{
#ifndef WORDS_BIGENDIAN
	jChar = (jChar << 8) | (jChar >> 8);
#endif

	return XTextWidth16(fs, (XChar2b *) & jChar, 1);
}

jint
Java_java_awt_Toolkit_fntCharsWidth(JNIEnv * env, jclass clazz, XFontStruct * fs,
                                    jarray jChars, jint off, jint len)
{
	jboolean isCopy;
	jchar *jc = (*env)->GetCharArrayElements(env, jChars, &isCopy);
	int n = (*env)->GetArrayLength(env, jChars);
	XChar2b *b;
	int w;

	if (off + len > n)
		len = n - off;

#ifndef WORDS_BIGENDIAN
	n = sizeof(XChar2b) * len;
	b = (XChar2b *) getBuffer(X, n);
	swab((jc + off), b, n);
#else
	b = (XChar2b *) (jc + off);
#endif

	w = XTextWidth16(fs, b, len);

	(*env)->ReleaseCharArrayElements(env, jChars, jc, JNI_ABORT);
	return w;
}

jint
Java_java_awt_Toolkit_fntStringWidth(JNIEnv * env, jclass clazz, XFontStruct * fs, jstring jStr)
{
	jboolean isCopy;
	const jchar *jc = (*env)->GetStringChars(env, jStr, &isCopy);
	int len = (*env)->GetStringLength(env, jStr);
	int w, n;
	XChar2b *b;

#ifndef WORDS_BIGENDIAN
	n = sizeof(XChar2b) * len;
	b = (XChar2b *) getBuffer(X, n);
	swab(jc, b, n);
#else
	b = (XChar2b *) jc;
#endif

	w = XTextWidth16(fs, b, len);

	(*env)->ReleaseStringChars(env, jStr, jc);
	return w;
}
