/**
 * img.c - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

#define SINGLEFRAMEDONE 2
#define STATICIMAGEDONE 3
#define TOPDOWNLEFTRIGHT 2
#define COMPLETESCANLINES 4

/************************************************************************************
 * own auxiliary funcs
 */

static inline void
freeXImage(Image*img)
{
	if ( (img->flags & IMG_EXTDATA) == 0 ){
		AWT_FREE(img->xImg->data);
		img->xImg->data = 0;
	}
	XDestroyImage( img->xImg);
	img->xImg = 0;
}

static inline void
freeXMask(Image*img)
{
	if ( (img->flags & IMG_EXTDATA) == 0 ){
		AWT_FREE(img->xMask->data);
		img->xMask->data = 0;
	}
	XDestroyImage( img->xMask);
	img->xMask = 0;
}



/*
 * The basic Image (*not* XImage) alloc, which leaves the X representation
 * still open (it's just the envelope). We don't use 'isDrawable' here, but want to
 * keep it compatible with no-native-wm
 */
Image *
createImage(int width, int height, int isDrawable )
{
  Image *img = AWT_CALLOC(1, sizeof(Image));

  if ( img ) {
    img->trans = -1;             /* default to no alpha */
    img->width = width;          /* we need to (temp) store them for subsequent X image creation */
    img->height = height;
    
    if ( isDrawable ) {
      int depth = DefaultDepth(X->dsp, DefaultScreen(X->dsp));
      img->pix = XCreatePixmap(X->dsp, X->root, width, height, depth);
      img->flags |= (IMG_PIXMAP|IMG_OFFSCREEN);
      
      if ( !img->pix ){
	AWT_FREE( img);
	return 0;
      }
    }
    else {
      createXImage( X, img);
      if ( !img->xImg ){
	AWT_FREE( img);
	return 0;
      }
    }
  }
	
  return img;
}


#if defined(HAVE_LIBXEXT)

static int
createShmXImage(Toolkit * X, Image * img, int depth, int isMask)
{
	Visual *vis = DefaultVisual(X->dsp, DefaultScreen(X->dsp));
	XShmSegmentInfo *shmi = (XShmSegmentInfo *) AWT_MALLOC(sizeof(XShmSegmentInfo));
	XImage *xim;
	unsigned int nBytes;

	if ( !shmi ){
	    return 0;  /* out of memory */
	}
	
	if (isMask) {
		xim = XShmCreateImage(X->dsp, vis, depth, XYBitmap, 0, shmi, img->width, img->height);
	}
	else {
		xim = XShmCreateImage(X->dsp, vis, depth, ZPixmap, 0, shmi, img->width, img->height);
	}

	nBytes = xim->bytes_per_line * img->height;
	shmi->shmid = shmget(IPC_PRIVATE, nBytes, (IPC_CREAT | 0777));

	/*
	 * It is essential to check if shmget failed, because shared memory is usually
	 * a scarce resource
	 */
	if (shmi->shmid == -1) {
		XShmDetach(X->dsp, shmi);
		xim->data = 0;
		XDestroyImage(xim);
		AWT_FREE(shmi);

		X->shm = SUSPEND_SHM;
		return 0;
	}

	xim->data = shmi->shmaddr = shmat(shmi->shmid, 0, 0);
	shmi->readOnly = False;
	XShmAttach(X->dsp, shmi);
	/*
	 * make sure it will be freed automatically once the attachment count comes
	 * down to 0 (either by explicit imgFreeImage or by process termination)
	 */
	shmctl(shmi->shmid, IPC_RMID, 0);

	if (isMask) {
		memset(xim->data, 0xff, nBytes);
		img->shmiMask = shmi;
		img->xMask = xim;
	}
	else {
		img->shmiImg = shmi;
		img->xImg = xim;
	}

	img->flags |= IMG_SHARED;

	return 1;
}


static void
destroyShmXImage(Toolkit * X, Image * img, int isMask)
{
	XShmSegmentInfo *shmi;
	XImage *xim;

	if (isMask) {
		shmi = img->shmiMask;
		xim = img->xMask;
		img->shmiMask = 0;
	}
	else {
		shmi = img->shmiImg;
		xim = img->xImg;
		img->shmiImg = 0;
	}

	XShmDetach(X->dsp, shmi);
	xim->data = 0;
	XDestroyImage(xim);
	/* we created it as 'deleted', so we just have to detach here */
	shmdt(shmi->shmaddr);
	AWT_FREE(shmi);

	/* if we have suspended shm, give it a try again */
	if (X->shm == SUSPEND_SHM)
		X->shm = USE_SHM;

	img->flags &= ~IMG_SHARED;
}

#endif  /* HAVE_LIBXEXT */


/*
 * Create a (possibly shared) blank XImage object, which involves a properly
 * initialized XImage structure *and* a explicitly allocated pixel buffer
 */
void
createXImage(Toolkit * X, Image * img)
{
	int bitmap_pad;
	int bytes_per_line;
	int bytes_per_pix;
	unsigned int nPix;
	char *data;
	Visual *vis = DefaultVisual(X->dsp, DefaultScreen(X->dsp));
	int depth = DefaultDepth(X->dsp, DefaultScreen(X->dsp));

	if (depth <= 8){
		bytes_per_pix = 1;
	}
	else if (depth <= 16){
		bytes_per_pix = 2;
	}
	else {
		bytes_per_pix = 4;
	}

	bytes_per_line = bytes_per_pix * img->width;
	bitmap_pad = bytes_per_pix * 8;
	nPix = img->width * img->height;

	img->flags |= IMG_XIMAGE;

#if defined(HAVE_LIBXEXT)
	if ((X->shm == USE_SHM) && (nPix > X->shmThreshold) ) {
		if (createShmXImage(X, img, depth, False)) {
			DBG(awt_img, ("alloc Shm: %x %x %x (%dx%d) \n", img, img->xImg, img->shmiImg,
						  img->width, img->height));
			return;
		}
	}
#endif /* HAVE_LIBEXT */

	data = AWT_CALLOC(nPix, bytes_per_pix);
	img->xImg = XCreateImage(X->dsp, vis, depth, ZPixmap, 0,
							 data, img->width, img->height, bitmap_pad, bytes_per_line);
	DBG(awt_img, ("alloc: %x %x (%dx%d)\n", img, img->xImg, img->width, img->height));
}

/*
 * create a (possibly shared) XImage to hold the mask info for a reduced alpha
 * image (with transparent pixels)
 */
void
createXMaskImage(Toolkit * X, Image * img)
{
	int bytes_per_line;
	unsigned int nBytes, nPix;
	char *data;
	Visual *vis = DefaultVisual(X->dsp, DefaultScreen(X->dsp));

	bytes_per_line = (img->width + 7) / 8;
	nPix = img->width * img->height;
	nBytes = bytes_per_line * img->height;

#if defined(HAVE_LIBXEXT)
	if ((X->shm == USE_SHM) && (nPix > X->shmThreshold)) {
		if (createShmXImage(X, img, 1, True)) {
			DBG(awt_img, ("alloc Shm mask: %x %x %x (%dx%d) \n", img, img->xMask, img->shmiMask,
						  img->width, img->height));
			return;
		}
	}
#endif /* HAVE_LIBEXT */

	data = AWT_MALLOC(nBytes);
	memset(data, 0xff, nBytes);

	img->xMask = XCreateImage(X->dsp, vis, 1, XYBitmap, 0,
							  data, img->width, img->height, 8, bytes_per_line);
	DBG(awt_img, ("alloc mask: %x %x (%dx%d)\n", img, img->xMask, img->width, img->height));
}


/*
 * Convert a Image which has a (client-side) XImage representation into
 * a (server-side) pixmap representation. Note that we don't do that for
 * shared images, since we use them just for big things, which shouldn't
 * gobble up precious server offscreen space
 */
static int
convertToPixmap(Toolkit * X, Image * img, int keepImg)
{
	int depth = DefaultDepth(X->dsp, DefaultScreen(X->dsp));

	DBG( awt_img, ("conertToPixmap: %p [%p] ", img, img->xImg));

#if defined(HAVE_LIBXEXT)
	if ( img->shmiImg || img->shmiMask || img->alpha ){
		return 0; /* nope, this is a biggy (keep it out of the server) */
	}
#endif /* HAVE_LIBEXT */

	if (img->xImg) {
		if (img->pix) {
			XFreePixmap(X->dsp, img->pix);
		}

		img->pix = XCreatePixmap(X->dsp, X->root, img->width, img->height, depth);

		if (img->pix) {
			// reset (maybe we had a drawMaskedPixmap, before */
			XSetClipOrigin(X->dsp, X->gcScreen, 0, 0);
			XSetClipMask( X->dsp, X->gcScreen, 0);

			XPutImage(X->dsp, img->pix, X->gcScreen, img->xImg,
					  0, 0, 0, 0, img->width, img->height);

			img->flags |= IMG_PIXMAP;

			if ( !keepImg ){
				freeXImage( img);
				img->flags &= ~IMG_XIMAGE;
			}
		}
	}

	if (img->xMask) {
		if (img->pixMask) {
			XFreePixmap(X->dsp, img->pixMask);
		}

		img->pixMask = XCreatePixmap(X->dsp, X->root, img->width, img->height, 1);

		if (img->pixMask) {
			/* note that gcBitmap has flipped foreground/background colors because
			 * we actually use pixMask as a (temporary) clipMask
			 */
			XPutImage(X->dsp, img->pixMask, X->gcBitmap, img->xMask,
					  0, 0, 0, 0, img->width, img->height);

			if ( !keepImg ){
				freeXMask( img);
			}
		}
	}

	DBG( awt_img, (" -> %p, %p\n", img->pix, img->pixMask));

	return (img->pix != 0);
}


/*
 * Convert a pixmap (server-side) image representation into a XImage (client-side)
 * one. This is required if we need access to the image pels (e.g. for production
 * or scaling). Note that we don't convert to shared images
 */
int
convertToXImage(Toolkit* X, Image* img, int keepPix)
{
	if ( img->pix ){
		if ( img->xImg ){
			freeXImage( img);
		}

		/* watch out, the image data is going to be allocated by Xlib */
		img->xImg = XGetImage(X->dsp, img->pix, 0, 0, img->width, img->height,
							  0xffffffff, ZPixmap);

		img->flags |= (IMG_EXTDATA | IMG_XIMAGE);

		if ( !keepPix ){
			XFreePixmap( X->dsp, img->pix);
			img->pix = 0;
			img->flags &= ~IMG_PIXMAP;
		}
	}

	if ( img->pixMask ) {
		if ( img->xMask ){
			freeXMask( img);
		}

		img->xMask = XGetImage(X->dsp, img->pixMask, 0, 0, img->width, img->height,
							   1, XYPixmap);
		if ( !keepPix ){
			XFreePixmap( X->dsp, img->pix);
			img->pix = 0;
		}
	}

	return (img->xImg != 0);
}

/*
 * This is our policy routine - non-offscreen images start out as XImages (possibly
 * shared if they exceed a certain size). When they got drawn the first time, without
 * being scaled or produced before, we convert all non-shareds into pixmaps (to speed
 * up drawing). In order to save space, we free the XImage structures in that case.
 * If a already pixmap converted image gets scaled, or produced, afterwards, we restore
 * the XImage (and keep it, since it's likely that it will be drawn, again)
 */
void
setImageState(Toolkit *X, Image* img, int state)
{
	switch ( state ){
	case IMG_DRAWN:
		/* in case this is the first draw request, the image is not a pixmap, is not
		 * shared, and it hasn't been produced or scaled yet, convert it into a pixmap
		 */
		if ( (img->flags & (IMG_DRAWN | IMG_PIXMAP)) == 0 ){
			if ( ((img->flags & (IMG_SCALED | IMG_PRODUCED | IMG_SHARED | IMG_ANIMATION))  == 0) ){
				convertToPixmap( X, img, False);  /* free XImage(s) to save resources */
			}
			img->flags |= IMG_DRAWN;
		}
		break;

	case IMG_SCALED:
	case IMG_PRODUCED:
		/* we need to have pixel data - if this is not a XImage, convert it, but
		 * keep the pixmap (it either has been drawn, already, or was a screen
		 * image, anyway)
		 */
		if ( img->flags & IMG_PIXMAP ){
			convertToXImage( X, img, True);
		}
		img->flags |= state;
		break;
	}
}


/******************************************************************************
 * full alpha channel support
 */

/*
 * Create a simple alpha channel buffer (one byte per pixel)
 */
void
createAlphaImage(Toolkit * X, Image * img)
{
	int nBytes = img->width * img->height;

	img->alpha = AWT_MALLOC(sizeof(AlphaImage));

	DBG( awt_img, ("createAlphaImage %p [%d,%d] -> %p\n",
				   img, img->width, img->height, img->alpha));
	
	img->alpha->width = img->width;
	img->alpha->height = img->height;
	img->alpha->buf = AWT_MALLOC(nBytes);
	memset(img->alpha->buf, 0x0, nBytes);
}

#if defined (DRAWABLE_ALPHA)
void
fillAlphaImage ( AlphaImage* aimg, int x0, int y0, int x1, int y1, unsigned char a )
{
	int x, y;
	unsigned char* p = aimg->buf + (y0 * aimg->width) + x0;
	int d = aimg->width - x1 + x0-1;

	for ( y=y0; y <= y1; y++ ){
		for ( x=x0; x <= x1; x++ ){
			*p++ = a;
		}
		p += d;
	}
}
#endif

/*
 * For images with a full alpha channel, check if we really need an alpha byte for
 * each pel, or if a mask bitmap (alpha 0x00 / 0xff) will be sufficient
 */
int
needsFullAlpha(Toolkit * X, Image * img, double threshold)
{
	int i, j, a;
	int n = 0, max;

	if (!img->alpha)
		return 0;

	max = (img->width * img->height) * threshold;

	for (i = 0; i < img->height; i++) {
		for (j = 0; j < img->width; j++) {
			a = GetAlpha(img->alpha, j, i);
			if ((a != 0) && (a != 0xff)) {
				if (++n > max){
					DBG( awt_img, ("needsFullAlpha %p : true\n", img));
					return 1;
				}
			}
		}
	}

	DBG( awt_img, ("needsFullAlpha %p : false\n", img));

	return 0;
}

/*
 * get statistics about how many non-transparent, full transparent, and
 * partly transparent pixels we have in an image
 */
static void
countAlphas(Image * img, int *noAlpha, int *partAlpha, int *fullAlpha)
{
	int i, j, a;

	if (!img->alpha)
		return;

	for (i = 0; i < img->height; i++) {
		for (j = 0; j < img->width; j++) {
			a = GetAlpha(img->alpha, j, i);
			if (a == 0) {
				(*noAlpha)++;
			}
			else if (a == 0xff) {
				(*fullAlpha)++;
			}
			else {
				(*partAlpha)++;
			}
		}
	}
}

/*
 * A full alpha image channel is way slower than using a mask bitmap (= 0 / 0xff alpha).
 * This function provides a simple alpha-to-mask translation
 */
void
reduceAlpha(Toolkit * X, Image * img, int threshold)
{
	int i, j, a;

	if (!img->alpha)
		return;

	DBG( awt_img, ("reduceAlpha %p\n", img));
	
	createXMaskImage(X, img);

	for (i = 0; i < img->height; i++) {
		for (j = 0; j < img->width; j++) {
			a = GetAlpha(img->alpha, j, i);
			if (a < threshold) {
				XPutPixel(img->xImg, j, i, 0);
				XPutPixel(img->xMask, j, i, 0);
			}
		}
	}

	AWT_FREE(img->alpha->buf);
	AWT_FREE(img->alpha);
	img->alpha = 0;
}


/******************************************************************************
 * image scaling support
 */

static inline int
interpolate(int ul, int ur, int ll, int lr, double dx, double dy)
{
	double u = ul + (double) (ur - ul) * dx;
	double l = ll + (double) (lr - ll) * dx;

	return (int) (u + (l - u) * dy + 0.5);
}

static unsigned int
getScaledAlpha(Toolkit * X, Image * img, int x, int y, double dx, double dy)
{
	int ul, ur, ll, lr, a;
	int xi = (dx) ? x + 1 : x;
	int yi = (dy) ? y + 1 : y;

	if (img->alpha) {
		ul = GetAlpha(img->alpha, x, y);
		ur = GetAlpha(img->alpha, xi, y);
		ll = GetAlpha(img->alpha, x, yi);
		lr = GetAlpha(img->alpha, xi, yi);
		a = (unsigned int) interpolate(ul, ur, ll, lr, dx, dy);
		return a;
	}

	return 0xff;
}

static long
getScaledPixel(Toolkit * X, Image * img, int x, int y, double dx, double dy)
{
	unsigned long ul, ur, ll, lr;
	int ulR, urR, llR, lrR, ulG, urG, llG, lrG, ulB, urB, llB, lrB, r, g, b;
	int xi = (dx) ? x + 1 : x;
	int yi = (dy) ? y + 1 : y;

	if (img->xMask) {
		ul = XGetPixel(img->xMask, x, y);
		ur = XGetPixel(img->xMask, xi, y);
		ll = XGetPixel(img->xMask, x, yi);
		lr = XGetPixel(img->xMask, xi, yi);

		if (!interpolate(ul, ur, ll, lr, dx, dy))
			return -1;
	}

	ul = XGetPixel(img->xImg, x, y);
	ur = XGetPixel(img->xImg, xi, y);
	ll = XGetPixel(img->xImg, x, yi);
	lr = XGetPixel(img->xImg, xi, yi);


	if ((ul == ur) && (ll == ul) && (lr == ll)) {
		rgbValues(X, ul, &r, &g, &b);
	}
	else {
		rgbValues(X, ul, &ulR, &ulG, &ulB);
		rgbValues(X, ur, &urR, &urG, &urB);
		rgbValues(X, ll, &llR, &llG, &llB);
		rgbValues(X, lr, &lrR, &lrG, &lrB);

		r = interpolate(ulR, urR, llR, lrR, dx, dy);
		g = interpolate(ulG, urG, llG, lrG, dx, dy);
		b = interpolate(ulB, urB, llB, lrB, dx, dy);
	}

	return pixelValue(X, (r << 16) | (g << 8) | b);
}



void
initScaledImage(Toolkit * X, Image * tgt, Image * src,
                int dx0, int dy0, int dx1, int dy1,
                int sx0, int sy0, int sx1, int sy1)
{
	double xScale, yScale, sX, sY, sxDelta, syDelta;
	int dx, dy, dxInc, dyInc, sx, sy;
	long c;

	dxInc = (dx1 > dx0) ? 1 : -1;
	dyInc = (dy1 > dy0) ? 1 : -1;

	dx1 += dxInc;
	dy1 += dyInc;

	xScale = (double) (dx1 - dx0) / (double) (sx1 - sx0 + 1);
	yScale = (double) (dy1 - dy0) / (double) (sy1 - sy0 + 1);

	for (dy = dy0; dy != dy1; dy += dyInc) {
		sY = sy0 + (dy - dy0) / yScale;
		sy = (int) sY;
		syDelta = (sy < sy1) ? sY - sy : 0;

		for (dx = dx0; dx != dx1; dx += dxInc) {
			sX = sx0 + (dx - dx0) / xScale;
			sx = (int) sX;
			sxDelta = (sx < sx1) ? sX - sx : 0;

			if ((c = getScaledPixel(X, src, sx, sy, sxDelta, syDelta)) != -1) {
				XPutPixel(tgt->xImg, dx, dy, c);
				if (src->alpha)
					PutAlpha(tgt->alpha, dx, dy, getScaledAlpha(X, src, sx, sy, sxDelta, syDelta));
			}
			else {
				XPutPixel(tgt->xMask, dx, dy, 0);
				XPutPixel(tgt->xImg, dx, dy, 0);
			}
		}
	}
}


/************************************************************************************
 * exported native methods
 */

/*
 * this is called for non-NativeImageProducers, to initialize the native part
 * so that we can subsequently do setPixel() calls
 */
Image*
Java_java_awt_Toolkit_imgCreateImage(JNIEnv * env, jclass clazz, jint width, jint height)
{
	Image *img = createImage(width, height, 0);
	
	DBG( awt_img, ("createImage: %d,%d -> %p [%p,%p]\n", width, height, img, img->pix, img->xImg));
	return img;
}


Image*
Java_java_awt_Toolkit_imgCreateScreenImage(JNIEnv * env, jclass clazz, jint width, jint height)
{
	Image *img = createImage(width, height, 1);
	int depth = DefaultDepth(X->dsp, DefaultScreen(X->dsp));

	DBG( awt_img, ("createScreenImage: %d,%d -> %p\n", width, height, img));
	img->pix = XCreatePixmap(X->dsp, X->root, width, height, depth);
	img->flags |= (IMG_PIXMAP|IMG_OFFSCREEN);
	return img;
}


/* generic (ImageProducer) based image construction */

void
Java_java_awt_Toolkit_imgSetIdxPels(JNIEnv * env, jclass clazz, Image * img,
                                    jint x, jint y, jint w, jint h,
									jarray clrMap, jarray idxPels, jint trans,
                                    jint off, jint scan)
{
	register int row, col;
	unsigned long pix;
	jint rgb;
	jboolean isCopy;
	jint *clr = (*env)->GetIntArrayElements(env, clrMap, &isCopy);
	jbyte *pel = (*env)->GetByteArrayElements(env, idxPels, &isCopy);
	int maxCol = x + w;
	int maxRow = y + h;
	unsigned char val;
 	int i = off;
	int di = scan - w;

	DBG( awt_img, ("setIdxPels: %p, %d,%d - %d,%d,  %p,%p,  %d, %d,%d\n",
				   img, (int)x,(int)y,(int)w,(int)h, clrMap,idxPels, (int)trans,
				   (int)off,(int)scan));
	
	if ((trans >= 0) && !img->xMask){
		createXMaskImage(X, img);
	}

	for ( row = y; row < maxRow; row++, i+=di ) {
		for (col = x; col < maxCol; col++, i++ ) {
			val = pel[i];
			rgb = clr[val];
			pix = pixelValue(X, rgb);
			if ( (trans != -1) && (val == trans) ) {
				pix = 0;
				XPutPixel(img->xMask, col, row, 0);
			}
			XPutPixel(img->xImg, col, row, pix);
		}
	}

	(*env)->ReleaseIntArrayElements(env, clrMap, clr, JNI_ABORT);
	(*env)->ReleaseByteArrayElements(env, idxPels, pel, JNI_ABORT);
}


void
Java_java_awt_Toolkit_imgSetRGBPels(JNIEnv * env, jclass clazz, Image * img,
                                    jint x, jint y, jint w, jint h,
                                    jarray rgbPels, jboolean hasAlpha, jint off, jint scan)
{
	register int row, col;
	unsigned long pix = 0;
	jboolean isCopy;
	jint *rgbs = (*env)->GetIntArrayElements(env, rgbPels, &isCopy);
	int maxCol = x + w;
	int maxRow = y + h;
	jint val;
	int i = off;
	int di = scan - w;


	DBG( awt_img, ("setRGBPels: %p [%p,%p], %d,%d - %d,%d,  %p, %d,%d\n",
		       img, img->pix, img->xImg, (int)x,(int)y,(int)w,(int)h, rgbPels, (int)off,(int)scan));

	for (row = y; row < maxRow; row++, i+=di) {
		for (col = x; col < maxCol; col++, i++) {
		        val = rgbs[i];
			
			/* 	
			 * look pix got never initialized since if alpha bits are not ff
			 * pix was not assigned a value. Therefore we have get value for
			 * it using pixelValue. Now we can use XPutPixel after setting the
			 * value for pix
			 */

			pix = pixelValue(X, val);

			if ((val & 0xff000000) == 0xff000000) XPutPixel(img->xImg, col, row, pix);
			else {
#if defined(USE_FULL_ALPHA)	
				/*	
				 * No way to tell for now if this will be a reduced (on/off) or a
				 * full alpha channel. We have to be prepared for the "worst", and reduce
				 * later	
				 */

				if ( hasAlpha && !img->alpha){
					createAlphaImage(X, img);
				}

				PutAlpha(img->alpha, col, row, (unsigned char)(val >> 24));
				XPutPixel(img->xImg, col, row, pix);
#else	
				if ( !img->xMask ){
					createXMaskImage( X, img);
				}
				XPutPixel( img->xMask, col, row, 0);
				XPutPixel( img->xImg, col, row, 0);
#endif
			}
		}
	}

	(*env)->ReleaseIntArrayElements(env, rgbPels, rgbs, JNI_ABORT);
}

void
Java_java_awt_Toolkit_imgComplete(JNIEnv * env, jclass clazz, Image * img, jint status)
{
	/*
	 * Check for alpha channel reduction. Note that full alpha images aren't created
	 * with Shm (by policy), so you might loose the Shm speed factor. This method is just
	 * called for external (generic) production, since our own prod facilities usually
	 * know better if and how to do alpha support
	 */
	DBG( awt_img, ("complete: %p %x\n", img, (unsigned)status));

	if ( status == SINGLEFRAMEDONE ){
		/* treat this a animation, regardless of img->next (it
		 * might produce several times into the same surface)
		 */
		img->flags |= IMG_ANIMATION;
	}

	if (img->alpha && !needsFullAlpha(X, img, 0.0)) {
		reduceAlpha(X, img, 128);
	}
}

void
Java_java_awt_Toolkit_imgFreeImage(JNIEnv * env, jclass clazz, Image * img)
{
	Image *next, *first = img;

	/* we do this just once, since an image ring shares the same buffer */
	if ( img->buffer ){
		Java_java_awt_Toolkit_imgFreeImage( env, clazz, img->buffer);
	}
	
	/* we have to be aware of image rings (GIF movies), iterate */
	do {
		if (img->pix) {
			XFreePixmap(X->dsp, img->pix);
			img->pix = 0;
		}

		if (img->pixMask) {
			XFreePixmap(X->dsp, img->pixMask);
			img->pixMask = 0;
		}

		/*
		 * note that XDestroyImage automatically frees any non-NULL data pointer
		 * (since we explicitly allocated the data, we better free it explicitly, too)
		 * malloc, free might be resolved
		 */

		if (img->xImg) {
#if defined(HAVE_LIBXEXT)
			if (img->shmiImg) {
				DBG(awt_img, ("free Shm: %x %x %x (%dx%d)\n", img, img->xImg, img->shmiImg,
							  img->width, img->height));
				destroyShmXImage(X, img, False);
				img->xImg = 0;
			}
			else
#endif /* HAVE_LIBXEXT */
				{
					DBG(awt_img, ("free: %x %x (%dx%d)\n", img, img->xImg, img->width, img->height));
					freeXImage( img);
				}
		}

		if (img->xMask) {
#if defined(HAVE_LIBXEXT)
			if (img->shmiMask) {
				DBG(awt_img, ("free Shm mask: %x %x %x (%dx%d)\n", img, img->xMask, img->shmiMask,
							  img->width, img->height));
				destroyShmXImage(X, img, True);
				img->xMask = 0;
			}
			else
#endif /* HAVE_LIBXEXT */
				{
					DBG(awt_img, ("free mask: %x %x (%dx%d)\n", img, img->xMask,
								  img->width, img->height));
					freeXMask( img);
				}
		}

		if (img->alpha) {
			AWT_FREE(img->alpha->buf);
			AWT_FREE(img->alpha);
			img->alpha = 0;
		}

		if ( img->gr ){
			Java_java_awt_Toolkit_graFreeGraphics( env, clazz, img->gr);
			img->gr = 0;
		}
		
		next = img->next;
		AWT_FREE(img);
		img = next;
	} while ((img != 0) && (img != first));
}


void*
Java_java_awt_Toolkit_imgCreateScaledImage(JNIEnv * env, jclass clazz,
										   Image * img, int width, int height)
{
	int depth;
	Image *scaledImg = createImage(width, height, JNI_FALSE);
	
	DBG( awt_img, ("createScaledImage: %p  %d,%d -> %d,%d\n",
				   img, img->width, img->height, width, height));
	
	if ( !scaledImg ) {
	    return 0;
	}
				
	if ( (img->flags & IMG_SCALED) == 0 ){
		setImageState( X, img, IMG_SCALED);
	}

	if (img->xImg) {
		if (img->xMask)
			createXMaskImage(X, scaledImg);

		initScaledImage(X, scaledImg, img,
						0, 0, width - 1, height - 1,
						0, 0, img->width - 1, img->height - 1);
	}
	else if (img->pix) {
		depth = DefaultDepth(X->dsp, DefaultScreen(X->dsp));
		scaledImg->pix = XCreatePixmap(X->dsp, X->root, width, height, depth);
	}

	return scaledImg;
}


void
Java_java_awt_Toolkit_imgProduceImage(JNIEnv * env, jclass clazz,
									  jobject producer, Image * img,
									  jboolean firstFrame)
{
	int i, j;
	int r, g, b;
	unsigned long pix;
	jclass prodClazz, modelClazz;
	jmethodID modelCtor, setPix, imgCompl;
	jmethodID setDim=0, setCM=0, setHints=0;
	jobject model;
	jarray pixArray;
	jint *ipix = 0;
	jbyte *bpix = 0;
	jboolean isCopy;

	DBG( awt_img, ("produceImage: %p  prod: %p first: %d\n", img, producer, firstFrame));
	
	if ( (img->flags & IMG_PRODUCED) == 0 ){
		setImageState( X, img, IMG_PRODUCED);
	}

	prodClazz = (*env)->GetObjectClass(env, producer);
	
	if ( X->colorMode == CM_PSEUDO ) {
		/* produce with a IndexColorModel, to mimic JDK behavior (there are
		 * apps out there using RgbFilters relying on getting IndexColorModel
		 * based setPixel() calls)
		 */
		model = getSystemColorModel( env, clazz, X, img->trans);
		setPix = (*env)->GetMethodID(env, prodClazz, "setPixels",
									 "(IIIILjava/awt/image/ColorModel;[BII)V");

 		/* allocate enough buffer space for a single scanline */
		pixArray = (*env)->NewByteArray(env, img->width);
		bpix = (*env)->GetByteArrayElements(env, pixArray, &isCopy);
	}
	else {
		modelClazz = (*env)->FindClass(env, "kaffe.awt.JavaColorModel");
		modelCtor = (*env)->GetStaticMethodID(env, modelClazz, "getSingleton", 
											  "()Lkaffe/awt/JavaColorModel;");
		model = (*env)->CallStaticObjectMethod(env, modelClazz, modelCtor);

		setPix = (*env)->GetMethodID(env, prodClazz, "setPixels", 
									 "(IIIILjava/awt/image/ColorModel;[III)V");

		pixArray = (*env)->NewIntArray(env, img->width);
		ipix = (*env)->GetIntArrayElements(env, pixArray, &isCopy);
	}

	imgCompl = (*env)->GetMethodID(env, prodClazz, "imageComplete", "(I)V");

	if ( firstFrame ) {
		setDim = (*env)->GetMethodID(env, prodClazz, "setDimensions", "(II)V");
		setCM = (*env)->GetMethodID(env, prodClazz, "setColorModel",
									"(Ljava/awt/image/ColorModel;)V");
		setHints = (*env)->GetMethodID(env, prodClazz, "setHints", "(I)V");
	}
	
	if (img->pix && !img->xImg) {
		img->xImg = XGetImage(X->dsp, img->pix, 0, 0, 
							  img->width, img->height, 0xffffffff, ZPixmap);
	}

	UNBLOCK_EXECUTE( env, clazz, {
		if ( firstFrame ) {
			(*env)->CallVoidMethod(env, producer, setDim, img->width, img->height);
			(*env)->CallVoidMethod(env, producer, setCM, model);
			(*env)->CallVoidMethod(env, producer, setHints, TOPDOWNLEFTRIGHT | COMPLETESCANLINES);
		}

		if (img->xImg) {
			for (j = 0; j < img->height; j++) {
				if ( X->colorMode == CM_PSEUDO ) {
					/* we are using a SystemColorModel derived one, so there is no
					 * need to translate pixels other than the transparent one
					 */
					for (i = 0; i < img->width; i++) {
						if ((img->xMask == 0) || XGetPixel(img->xMask, i, j)) {
							bpix[i] = XGetPixel(img->xImg, i, j);
						}
						else {
							bpix[i] = img->trans;
						}
					}
					(*env)->ReleaseByteArrayElements(env, pixArray, bpix, JNI_COMMIT);
				}
				else {
					for (i = 0; i < img->width; i++) {
						if ((img->xMask == 0) || XGetPixel(img->xMask, i, j)) {
							pix = XGetPixel(img->xImg, i, j);
							rgbValues(X, pix, &r, &g, &b);
							ipix[i] = (0xff000000 | (r << 16) | (g << 8) | b);
						}
						else {
							ipix[i] = 0;
						}
					}
					(*env)->ReleaseIntArrayElements(env, pixArray, ipix, JNI_COMMIT);
				}

				(*env)->CallVoidMethod( env, producer, setPix, 0, j, img->width,
									    1, model, pixArray, 0, img->width);
			}
		}

		(*env)->CallVoidMethod(env, producer, imgCompl,
							   (img->next) ? SINGLEFRAMEDONE : STATICIMAGEDONE);
	});

	if (img->pix && img->xImg) {
		XDestroyImage(img->xImg);
		img->xImg = 0;
	}

	if ( X->colorMode == CM_PSEUDO ){
		(*env)->ReleaseByteArrayElements(env, pixArray, bpix, JNI_ABORT);
	}
	else {
		(*env)->ReleaseIntArrayElements(env, pixArray, ipix, JNI_ABORT);
	}
}



/************************************************************************************
 * native format production dispatcher routines (GIF, jpeg, png, ..)
 */
Image* readGifFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readGifData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readJpegFile ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readJpegData ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readPngFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readPngData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readTiffFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readTiffData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readXbmFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readXbmData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );

#define SIG_LENGTH 4
#define SIG_GIF    1
#define SIG_JPEG   2
#define SIG_PNG    3
#define SIG_TIFF   4
#define SIG_XBM    5

static Image *unknownImage = 0;               /* fill in some default image here */

static int
imageFormat(unsigned char *sig)
{
	if ((sig[0] == 'G') && (sig[1] == 'I') && (sig[2] == 'F'))
		return SIG_GIF;

	if ((sig[0] == 0xff) && (sig[1] == 0xd8) && (sig[2] == 0xff) &&
		((sig[3] == 0xe0) || (sig[3] == 0xc4)) )
		return SIG_JPEG;

	if ((sig[0] == 0x89) && (sig[1] == 'P') && (sig[2] == 'N') && (sig[3] == 'G'))
		return SIG_PNG;

	if ( (((sig[0] == 'I') && (sig[1] == 'I' )) ||
		  ((sig[0] == 'M') && (sig[1] == 'M'))) && *(short*)(sig+2) == 42 )
		return SIG_TIFF;

	if ( (sig[0] == '#') && (sig[1] == 'd') &&
		 (sig[2] == 'e') && (sig[3] == 'f') )
		return SIG_XBM;

	return 0;
}

void*
Java_java_awt_Toolkit_imgCreateFromFile(JNIEnv * env, jclass clazz,
										jobject producer, jstring fileName)
{
    jclass prodCls;
	jmethodID setDim;

	Image *img = 0;
	int infile;
	char *fn = java2CString(env, X, fileName);
	unsigned char sig[SIG_LENGTH];

	DBG( awt_img, ("createFromFile: %s\n", fn));

	if ( !(prodCls = (*env)->GetObjectClass( env, producer)) ){
	    return 0;
	}
	if ( !(setDim = (*env)->GetMethodID( env, prodCls, "setDimensions", "(II)V")) ){
	    return 0;
	}	

	if ((infile = AWT_OPEN(fn)) >= 0) {
		if (AWT_READ(infile, sig, sizeof(sig)) == sizeof(sig)) {
			AWT_REWIND(infile);              /* some native converters can't skip the signature read */

			switch (imageFormat(sig)) {
			case SIG_GIF:
				img = readGifFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_JPEG:
				img = readJpegFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_PNG:
				img = readPngFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_TIFF:
				img = readTiffFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_XBM:
				img = readXbmFile( env, clazz, producer, setDim, infile, fn);
				break;
			default:
				img = unknownImage;
			}
		}
		AWT_CLOSE(infile);
	}

	DBG( awt_img, (" -> img: %p %d,%d\n", img,
				   (img) ? img->width : 0, (img) ? img->height : 0));

	return img;
}

void*
Java_java_awt_Toolkit_imgCreateFromData ( JNIEnv * env, jclass clazz,
										  jobject producer, 
										  jarray jbuffer, jint off, jint len)
{
    jclass prodCls;
    jmethodID fillBuf;
	jmethodID setDim;

	Image *img = 0;
	jboolean isCopy;
	jint length = (*env)->GetArrayLength(env, jbuffer);
	jbyte *jb;

	DBG( awt_img, ("createFromData: %p, %d,%d\n", jbuffer, (int)len));

	if ( !(prodCls = (*env)->GetObjectClass( env, producer)) ){
	    return 0;
	}
	if ( !(fillBuf = (*env)->GetMethodID( env, prodCls, "fillBuffer", "()I")) ){
	    return 0;
	}
	if ( !(setDim = (*env)->GetMethodID( env, prodCls, "setDimensions", "(II)V")) ){
	    return 0;
	}

	jb = (*env)->GetByteArrayElements(env, jbuffer, &isCopy);

	/* in case of a buffer overrun, we probably have a JPEG read error, anyway */
	if ( (off+len) <= length) {
		switch (imageFormat(jb+off)) {
		case SIG_GIF:
			img = readGifData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_JPEG:
			img = readJpegData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_PNG:
			img = readPngData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_TIFF:
			img = readTiffData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_XBM:
			img = readXbmData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		default:
			img = unknownImage;
		}
	}

	(*env)->ReleaseByteArrayElements(env, jbuffer, jb, JNI_ABORT);

	DBG( awt_img, (" -> img: %p %d,%d\n", img,
				   (img) ? img->width : 0, (img) ? img->height : 0));

	return img;
}

void*
Java_java_awt_Toolkit_imgSetFrame(JNIEnv * env, jclass clazz, Image * img, int frameNo)
{
	Image *imgCur = img;

	if (!img->next)
		return img;

	while (img->frame != frameNo) {
		img = img->next;
		if (img == imgCur) {
			break;
		}
	}

	return img;
}


/************************************************************************************
 * field access
 */

jint
Java_java_awt_Toolkit_imgGetWidth(JNIEnv * env, jclass clazz, Image * img)
{
	return img->width;
}

jint
Java_java_awt_Toolkit_imgGetHeight(JNIEnv * env, jclass clazz, Image * img)
{
	return img->height;
}

jint
Java_java_awt_Toolkit_imgGetFrameCount(JNIEnv * env, jclass clazz, Image * img)
{
	return img->nFrames;
}

jint
Java_java_awt_Toolkit_imgGetLoopCount(JNIEnv * env, jclass clazz, Image * img)
{
	return img->nLoops;
}

jint
Java_java_awt_Toolkit_imgGetLatency(JNIEnv * env, jclass clazz, Image * img)
{
	return img->latency;
}

void*
Java_java_awt_Toolkit_imgGetNextFrame(JNIEnv * env, jclass clazz, Image * img)
{
	return img->next;                    /* next in the ring */
}
