/**
 * toolkit.h - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __toolkit_h
#define __toolkit_h

#include "config.h"
#include "jtypes.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-setjmp.h"
#define	__gc_incremental_h
#include "../../../kaffe/kaffevm/gc.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <sys/ipc.h>

#if defined(HAVE_LIBXEXT)
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif

#include <jni.h>
#include "jmalloc.h"
#include "../../../kaffe/kaffevm/jdebug.h"

/* contains the UNBLOCK_EXECUTE macro */
#include "jlocks.h"


/* no idea why gcc complains about this not being found no matter if
 * unistd.h or string.h are included
 */
extern void swab(const void *from, void *to, size_t n);


/*
 * configure support for offscreen image alpha channels
 */
#define DRAWABLE_ALPHA 1

/*
 * do we want to use full alphas in non-offscreen images?
 */
#define USE_FULL_ALPHA

/*
 * percentile of non-fully transparent/opaque pixels (alpha: ]0..255[ )
 * which will force alpha channel reduction (to a simple transparency mask)
 */
#define ALPHA_MASK_THRESHOLD 0.08

/*
 * the alpha value which will determine of a pixel alpha will be reduced
 * to either transparent or opaque (alpha>=threshold)
 */
#define ALPHA_VALUE_THRESHOLD 200



/*******************************************************************************
 * color conversion structures
 */

#define N_DIRECT 256

typedef struct _Rgb2Direct {
	unsigned char red[N_DIRECT];
	unsigned char redPix[N_DIRECT];
	int redShift;
	int nRed;
	unsigned char green[N_DIRECT];
	unsigned char greenPix[N_DIRECT];
	int greenShift;
	int nGreen;
	unsigned char blue[N_DIRECT];
	unsigned char bluePix[N_DIRECT];
	int blueShift;
	int nBlue;
} Rgb2Direct;

typedef struct _Rgb2True {
	unsigned int redMask;
	unsigned int greenMask;
	unsigned int blueMask;
	int blueShift;
	int redShift;
	int greenShift;
} Rgb2True;


#define N_CELLS      12
#define MAX_CELL_IDX 11

typedef struct _Rgb2Pseudo {
	unsigned char r[256];
	unsigned char g[256];
	unsigned char b[256];
	unsigned char pix[N_CELLS][N_CELLS][N_CELLS];
	unsigned char len;
} Rgb2Pseudo;



/*******************************************************************************
 * image handling structures
 */

typedef struct _AlphaImage {           /* storage for full alpha channel images */
	unsigned char *buf;
	int width, height;
} AlphaImage;

#define NO_SHM       0                 /* we don't have MIT-Shm support in the X server */
#define USE_SHM      1                 /* we have support, use it */
#define SUSPEND_SHM  2                 /* we have support, but it ran out of space */

#define IMG_XIMAGE     0x0001          /* X image rep is XImage */
#define IMG_PIXMAP     0x0002          /* X image rep is Pixmap */
#define IMG_EXTDATA    0x0100          /* xImg->data was allocated by Xlib */
#define IMG_SHARED     0x0200          /* XImage is shared (Shm) */
#define IMG_OFFSCREEN  0x0400
#define IMG_DRAWN      0x1000          /* image has been drawn */
#define IMG_SCALED     0x2000          /* image has been scaled */
#define IMG_PRODUCED   0x4000          /* image has been produced */
#define IMG_IS_DELTA   0x10000         /* draw this frame superposed */
#define IMG_BUFFER     0x20000         /* image needs rendering buffer */
#define IMG_ANIMATION  0x40000         /* image used as in-situ animation (no ring) */

#define NO_ALPHA   0                   /* image has no alpha channel at all */
#define MASK_ALPHA 1                   /* just opaque/transparent */
#define FULL_ALPHA 2                   /* full 0..255 alpha channel */

#define NO_TRANS_INDEX_AVAILABLE -2    /* we can't use a transparency index */


struct _Graphics;
struct _Image;


typedef struct _Image {
	unsigned int flags;

	Pixmap pix;                          /* pixmap for screen images */
	Pixmap pixMask;                      /* mask pixmap */

	XImage *xImg;                        /* "real" image */
	XImage *xMask;                       /* mask image for reduced alpha (on/off) images */

#if defined(HAVE_LIBXEXT)
	XShmSegmentInfo *shmiImg;            /* Shm info for shared mem real image */
	XShmSegmentInfo *shmiMask;           /* Shm info for shared mem mask image */
#endif

	AlphaImage *alpha;                   /* full alpha channel (for alpha != 0x00 or 0xff) */

	int trans;                           /* transparent index */
	int left, top;                       /* some GIF movies require drawing offsets */
	int width, height;                   /* we need this in case we are a pixmap */

	/* invariant animation data */
	struct _Image    *buffer;            /* doublebuffer for delta-frame animations */
	struct _Graphics *gr;                /* graphics to update the buffer */
	short nLoops;                        /* number of animation loops */
	short nFrames;                       /* number of frames for animation */

	/* frame specific animation data */
	short latency;                       /* between image flips, for "gif-movies" */
	short frame;                         /* frame number for animations */

	struct _Image *next;                 /* for animations, imag queues, caches.. */
} Image;


/*******************************************************************************
 * Graphics state cache
 */

#define WINDOW   0
#define IMAGE    1
#define GRAPHICS 2

typedef struct _Graphics {
	GC gc;
	Drawable drw;

	int fg;
	int bg;
	char xor;
	int xclr;
	int x0;
	int y0;
	XRectangle clip;

	// the added support for adding clipping rectangles under
	// X architecture. we need to store clip mask ourselves
	// since Xlib does not gives us any chance to access its
	// clipmask. It is explicitely stated in the documentation
	// that clip mask cannot be obtained for a given GC.
	Pixmap clip_mask;
  

#if defined (DRAWABLE_ALPHA)
	Image* img;
	int alpha;
#endif
} Graphics;



/*******************************************************************************
 * structure to store guessed and computed Frame/Dialog insets (titlebar, borders)
 */

typedef struct _DecoInset {
	int left;
	int top;
	int right;
	int bottom;
	char guess;
} DecoInset;


/*******************************************************************************
 * We use  our own structure instead of directly storing X window handles, to
 * enable us to attach and query attributes (owners, states, flags..) to particular
 * X window instances. We could do this by means of X properties, but this might
 * end up in costly round-trips and even more memory (than we trade for speed here).
 * It's a must have for our popup key/focus event forwarding (see wnd.c)
 */

typedef struct _WindowRec {
	Window w;
	unsigned flags;
	Window owner;
} WindowRec;

/*******************************************************************************
 * this is the master AWT structure (singleton object), glueing it al together
 */

typedef struct _Toolkit {
	Display *dsp;
	Window root;

	char *buf;
	unsigned int nBuf;

	int colorMode;                       /* refers to CM_xx constants, not X visuals */
	Rgb2True *tclr;
	Rgb2Pseudo *pclr;
	Rgb2Direct *dclr;

	unsigned long black, white;
	
	int shm;
	int shmThreshold;

	GC gcScreen;                         /* default GC for screen ops */
	GC gcBitmap;                         /* default GC for bitmap ops */

	int dpi;                             /* font resolution class to use (75/100 dpi) */
	int ptSzDelta;                       /* point size delta (in case we have a dpi mismatch)) */

	Cursor cursors[14];

	DecoInset frameInsets;
	DecoInset dialogInsets;

	XEvent event;
	char preFetched;
	char blocking;
	int pending;
	int evtId;
	fd_set rfds;                         /* used to select-check availability of X input (pthreads) */

	Window lastWindow;
	int srcIdx;
	WindowRec *windows;
	int nWindows;

	Window cbdOwner;
	Window wakeUp;

	Window focus;                        /* this is the real focus, if it is in our process */
	Window focusFwd;                     /* this might be a (owned) window we forward the focus to */
	int fwdIdx;                          /* cached index of the focus forward window */
} Toolkit;


/*******************************************************************************
 * global data def/decl
 */
#ifdef MAIN

Toolkit XTk;
Toolkit *X = &XTk;

Atom WM_PROTOCOLS;
Atom WM_DELETE_WINDOW;
Atom WM_TAKE_FOCUS;
Atom WAKEUP;
Atom RETRY_FOCUS;
Atom FORWARD_FOCUS;
Atom SELECTION_DATA;
Atom JAVA_OBJECT;

jclass AWTError;
JNIEnv *JniEnv;

#else

extern Toolkit *X;

extern Atom WM_PROTOCOLS;
extern Atom WM_DELETE_WINDOW;
extern Atom WM_TAKE_FOCUS;
extern Atom WAKEUP;
extern Atom RETRY_FOCUS;
extern Atom FORWARD_FOCUS;
extern Atom SELECTION_DATA;
extern Atom JAVA_OBJECT;

extern jclass AWTError;
extern JNIEnv *JniEnv;

#endif                          /* MAIN */

extern long StdEvents, PopupEvents;


/*****************************************************************************************
 * heap wrapper macros
 */

/**********
static inline void* _awt_malloc_wrapper ( size_t size )
{
	void *adr = malloc( size);
	DBG( awt_mem, ("malloc: %d  -> %p\n", size, adr));
	return adr;
}

static inline void* _awt_calloc_wrapper ( int n, size_t size )
{
	void *adr = calloc( n, size);
	DBG( awt_mem, ("calloc: %d,%d  -> %p\n", n, size, adr));
	return adr;
}

static inline void _awt_free_wrapper ( void* adr )
{
	DBG( awt_mem, ("free: %p\n", adr));
	free( adr);
}
*************/

static inline void *
_awt_malloc_wrapper(size_t size)
{
	void *adr = gc_malloc_fixed_typed(size, GC_ALLOC_AWT);
	DBG(awt_mem, ("malloc: %d  -> %x\n", size, adr));
	return adr;
}

static inline void *
_awt_calloc_wrapper(int n, size_t size)
{
	void *adr = gc_calloc_fixed_typed(n, size, GC_ALLOC_AWT);
	DBG(awt_mem, ("calloc: %d,%d  -> %x\n", n, size, adr));
	return adr;
}

static inline void 
_awt_free_wrapper(void *adr)
{
	DBG(awt_mem, ("free: %x\n", adr));
	gc_free_fixed(adr);
}

#define AWT_MALLOC(_n) \
  _awt_malloc_wrapper( _n)

#define AWT_CALLOC(_n,_sz) \
  _awt_calloc_wrapper( _n, _sz)

#define AWT_FREE(_adr) \
  _awt_free_wrapper( _adr);


#ifndef AWT_ALLOCA
#  if defined(HAVE_ALLOCA)
#    define AWT_ALLOCA(_n)           alloca(_n)
#    define AWT_ALLOCA_FREE(_adr)
#  else
#    define AWT_ALLOCA(_n)           AWT_MALLOC(_n)
#    define AWT_ALLOCA_FREE(_adr)    AWT_FREE(_adr)
#  endif
#endif


/*******************************************************************************
 *
 */


static inline char *
java2CString(JNIEnv * env, Toolkit * X, jstring jstr)
{
	jboolean isCopy;
	register int i;
	int n = (*env)->GetStringLength(env, jstr);
	const jchar *jc = (*env)->GetStringChars(env, jstr, &isCopy);

	if (n >= (int)X->nBuf) {
		if (X->buf)
			AWT_FREE(X->buf);
		X->buf = AWT_MALLOC(n + 1);
		X->nBuf = n + 1;
	}

	for (i = 0; i < n; i++)
		X->buf[i] = (char) jc[i];
	X->buf[i] = 0;
	(*env)->ReleaseStringChars(env, jstr, jc);

	return X->buf;
}

static inline char *
jchar2CString(Toolkit * X, jchar * jc, int len)
{
	register int i;
	int n = len + 1;

	if (n > (int)X->nBuf) {
		if (X->buf)
			AWT_FREE(X->buf);
		X->buf = AWT_MALLOC(n);
		X->nBuf = n;
	}

	for (i = 0; i < len; i++)
		X->buf[i] = (char) jc[i];
	X->buf[i] = 0;

	return X->buf;
}

static inline void *
getBuffer(Toolkit * X, unsigned int nBytes)
{
	if (nBytes > X->nBuf) {
		if (X->buf)
			AWT_FREE(X->buf);
		X->buf = AWT_MALLOC(nBytes);
		X->nBuf = nBytes;
	}
	return X->buf;
}


/*****************************************************************************************
 * color functions & defines
 */

/*
 * These are our directly supported visuals / color modes. Note that there is
 * no more 1-1 correspondence to X visuals, since we do a categorisation with
 * respect to our internal RGB <-> pixel conversion. All visuals not listed
 * explicitly are handled via the generic XAllocColor/XQueryColor (which might
 * slow down images considerably)
 *
 * NOTE: these values have to be != 0, since '0' is used to trigger color init
 */
#define CM_PSEUDO       1              /* PseudoColor visual */
#define CM_TRUE         2              /* general TrueColor visual */
#define CM_TRUE_888     3              /* special 8-8-8 bit TrueColor visual */
#define CM_DIRECT       4
#define CM_GENERIC      5              /* grays, DirectColor (packed) etc. */


void initColorMapping(JNIEnv * env, jclass clazz, Toolkit * X);
jobject getSystemColorModel(JNIEnv * env, jclass clazz, Toolkit * X, int trans);

#define DARKER   1
#define BRIGHTER 2
int modPixelValue ( Toolkit* X, int pix, int mod );

#define JRGB(_r,_g,_b)  (_r<<16 | _g<<8 | _b)
#define JRED(_rgb)      ((_rgb & 0xff0000) >> 16)
#define JGREEN(_rgb)    ((_rgb & 0x00ff00) >> 8)
#define JBLUE(_rgb)     (_rgb & 0x0000ff)

/* map [0..255] channel value to [0..11] cell index */
#define CELL_INDEX(_n)   ((((int)(_n)) * 4 - 1) / 86)

#define ROUND_SHORT2CHAR(_n) \
  ((unsigned short)_n >= 0xff70 ) ? 0xff : (unsigned char)(((unsigned short)_n + 0x80) >> 8)

/* shift ops with negative values have undefined results */
#define SIGNED_RSHIFT(_var,_val,_shift)   \
   _var = _val;                           \
   if ( _shift > 0 )	 _var >>= _shift; \
   else                  _var <<= -_shift;

#define SIGNED_LSHIFT(_var,_val,_shift)   \
   _var = _val;                           \
   if ( _shift > 0 )	 _var <<= _shift; \
   else                  _var >>= -_shift;


/*
 * add alpha to native pixel value
 */
static inline jint addAlpha ( jint rgb, unsigned int pix )
{
#if defined (DRAWABLE_ALPHA)
	return ((rgb & 0xff000000) | pix);
#else
	return pix;
#endif
}

/*
 * get alpha channel [0..255] from jint pixel value
 */
static inline unsigned char getAlpha ( unsigned int pix )
{
#if defined (DRAWABLE_ALPHA)
	return (pix & 0xff000000) >> 24;
#else
	return 0xff;
#endif
}

/*
 * remove alpha channel from jint pixel value
 */
static inline unsigned int stripAlpha ( unsigned int pix )
{
#if defined (DRAWABLE_ALPHA)
	return (pix & 0x00ffffff);
#else
	return pix;
#endif
}


static inline jint
pixelValue(Toolkit * X, jint rgb)
{
	int r, g, b;
	XColor xclr;
	jint pix;

	switch (X->colorMode) {
	case CM_PSEUDO:
		pix = X->pclr->pix[ CELL_INDEX( JRED(rgb)) ]
			              [ CELL_INDEX( JGREEN(rgb)) ]
			              [ CELL_INDEX( JBLUE(rgb)) ];
		break;

	case CM_TRUE:
		SIGNED_RSHIFT(b, (rgb & X->tclr->blueMask), X->tclr->blueShift);
		SIGNED_RSHIFT(g, (rgb & X->tclr->greenMask), X->tclr->greenShift);
		SIGNED_RSHIFT(r, (rgb & X->tclr->redMask), X->tclr->redShift);
		pix = (b | g | r);
		break;

	case CM_TRUE_888:
		pix = (rgb & 0xffffff);
		break;

	case CM_DIRECT:
		pix = (((jint) X->dclr->redPix[JRED(rgb)] << X->dclr->redShift) |
			   ((jint) X->dclr->greenPix[JGREEN(rgb)] << X->dclr->greenShift) |
			   ((jint) X->dclr->bluePix[JBLUE(rgb)] << X->dclr->blueShift));
		break;

	default:
		/*
		 * this is a generic fallback for "exotic" visuals and might be *awefully*
		 * slow (esp. for images) because XAllocColor is a roundtrip
		 */
		xclr.red = (rgb & 0xff0000) >> 8;
		xclr.green = (rgb & 0xff00);
		xclr.blue = (rgb & 0xff) << 8;
		xclr.flags = DoRed | DoGreen | DoBlue;
		XAllocColor(X->dsp, DefaultColormapOfScreen(DefaultScreenOfDisplay(X->dsp)), &xclr);
		pix = xclr.pixel;
	}

#if defined(DRAWABLE_ALPHA)
	return addAlpha( rgb, pix);
#else
	return pix;
#endif
}

static inline void
rgbValues(Toolkit * X, unsigned long pixel, int *r, int *g, int *b)
{
	Visual *v;
	XColor xclr;

	switch (X->colorMode) {
	case CM_PSEUDO:
		*r = X->pclr->r[(unsigned char) pixel];
		*g = X->pclr->g[(unsigned char) pixel];
		*b = X->pclr->b[(unsigned char) pixel];
		break;

	case CM_TRUE:
		v = DefaultVisual(X->dsp, DefaultScreen(X->dsp));
		SIGNED_LSHIFT(*r, (pixel & v->red_mask), X->tclr->redShift);
		SIGNED_LSHIFT(*g, (pixel & v->green_mask), X->tclr->greenShift);
		SIGNED_LSHIFT(*b, (pixel & v->blue_mask), X->tclr->blueShift);
		*r >>= 16;
		*g >>= 8;
		break;
	case CM_TRUE_888:
		*r = JRED(pixel);
		*g = JGREEN(pixel);
		*b = JBLUE(pixel);
		break;

	case CM_DIRECT:
		v = DefaultVisual(X->dsp, DefaultScreen(X->dsp));
		*r = X->dclr->red[((pixel & v->red_mask) >> X->dclr->redShift)];
		*g = X->dclr->green[((pixel & v->green_mask) >> X->dclr->greenShift)];
		*b = X->dclr->blue[((pixel & v->blue_mask) >> X->dclr->blueShift)];
		break;

	default:
		/*
		 * this is a generic fallback for "exotic" visuals and might be *awefully*
		 * slow (esp. for images) because XAllocColor is a roundtrip
		 */
		xclr.pixel = pixel;
		XQueryColor(X->dsp, DefaultColormapOfScreen(DefaultScreenOfDisplay(X->dsp)), &xclr);
		*r = xclr.red >> 8;
		*g = xclr.green >> 8;
		*b = xclr.blue >> 8;

		break;
	}
}


/*****************************************************************************************
 * image functions
 */

/* all from img.c */
int needsFullAlpha(Toolkit *, Image *, double);
Image *createImage(int width, int height, int isDrawable);
void createXMaskImage(Toolkit * X, Image * img);
void createXImage(Toolkit * X, Image * img);
void createAlphaImage(Toolkit * X, Image * img);
void fillAlphaImage( AlphaImage* aimg, int x0, int y0, int x1, int y1, unsigned char a);
void initScaledImage(Toolkit * X, Image * tgt, Image * src,
                     int dx0, int dy0, int dx1, int dy1,
                     int sx0, int sy0, int sx1, int sy1);
void reduceAlpha ( Toolkit* X, Image* img, int threshold );
Image* Java_java_awt_Toolkit_imgCreateScreenImage(JNIEnv * env, jclass clazz, jint width, jint height);
void setImageState(Toolkit* X, Image * img, int state);
int  convertToXImage( Toolkit* X, Image* img, int);
void Java_java_awt_Toolkit_imgFreeImage(JNIEnv * env, jclass clazz, Image* img);
void Java_java_awt_Toolkit_graFreeGraphics(JNIEnv * env, jclass clazz, Graphics* gr);
void Java_java_awt_Toolkit_graDrawImage( JNIEnv * env, jclass clazz, Graphics* gr, Image * img,
                                         jint srcX, jint srcY,
                                         jint dstX, jint dstY,
                                         jint width, jint height, jint bgval);

static inline void
PutAlpha(AlphaImage * img, int col, int row, unsigned char alpha)
{
	img->buf[row * img->width + col] = alpha;
}

static inline int
GetAlpha(AlphaImage * img, int col, int row)
{
	return img->buf[row * img->width + col];
}

/*****************************************************************************************
 * clipboard functions
 */

jobject selectionClear(JNIEnv * env, Toolkit * X);     /* from cbd.c */
jobject selectionRequest(JNIEnv * env, Toolkit * X);   /* from cbd.c */


/*****************************************************************************************
 * async (multithreaded) macros
 * this can be used to solve the problem of deferred drawing requests, not being
 * flushed because of infrequent (non-polled) XNextEvent calls.
 * (for now, we go with a backgound flush thread)
 */

#define XFLUSH(_X,_force)



/*****************************************************************************************
 * file io wrapper macros (for image production)
 */

#include "jfilesystem.h"

#define AWT_OPEN(_file)               open(_file, JOPEN_READ)
#define AWT_REWIND(_fd)               lseek(_fd, 0, JSEEK_SET)
#define AWT_SETPOS(_fd,_off)          lseek(_fd, _off, JSEEK_CUR)
#define AWT_READ(_fd,_buf,_count)     read(_fd,_buf,_count)
#define AWT_CLOSE(_fd)                close(_fd)



/*****************************************************************************************
 * macros to manage the source table (conversion of X windows to/from indices, which
 * are consistent with the AWTEvent.sources array)
 */

#define WND_FRAME      0x01
#define WND_WINDOW     0x02
#define WND_DIALOG     0x04
#define WND_EMBEDDED   0x08

#define WND_MAPPED     0x10
#define WND_DESTROYED  0x20

static inline int 
getFreeSourceIdx(Toolkit * X, Window wnd)
{
	register int i, n;

	/*
	 * we don't use a double hashing here because collisions are very unlikely
	 * (window IDs usually already are hashed, so it does not make sense to
	 * hash them again - we just could make it worse
	 */
	for (i = (unsigned long) wnd, n = 0; n < X->nWindows; i++, n++) {
		i %= X->nWindows;
		if ((int) (X->windows[i].w) <= 0) {
			X->srcIdx = i;
			X->lastWindow = wnd;

			return i;
		}
	}

	return -1;
}

static inline int 
getSourceIdx(Toolkit * X, Window w)
{
	int n;
	register int i;

	if (w == X->lastWindow) {
		return X->srcIdx;
	}
	else {
		for (i = (unsigned long) w, n = 0; n < X->nWindows; i++, n++) {
			i %= X->nWindows;
			if (X->windows[i].w == w) {
				X->srcIdx = i;
				X->lastWindow = w;
				return X->srcIdx;
			}
			else if (X->windows[i].w == 0) {
				return -1;
			}
		}
		return -1;
	}
}

static inline int 
checkSource(Toolkit * X, int idx)
{
	return ((idx >= 0) && (idx < X->nWindows) && (X->windows[idx].w));
}


/*****************************************************************************************
 * focus forwarding
 */

#define FWD_SET    0                   /* set focus forwarding */
#define FWD_CLEAR  1                   /* reset focus forwarding */
#define FWD_REVERT 2                   /* reset focus on owner */

/*
 * Macros to implement keyboard handling for owned windows (which don't ever get
 * the X focus). This is done by means of forwarding and generation of "artificial"
 * focus events (generated from clientMessages.
 */

static inline void 
resetFocusForwarding(Toolkit * X)
{
	X->fwdIdx = -1;
	X->focusFwd = 0;
}

void forwardFocus(int cmd, Window wnd);  /* from wnd.c */

/*****************************************************************************************
 * font prototypes
 */

void initFonts (Toolkit*X);             /* from fnt.c */


/*****************************************************************************************
 * graphics prototypes
 */
extern void
Java_java_awt_Toolkit_graFillRect(JNIEnv * env, jclass clazz, Graphics * gr,
                                  jint x, jint y, jint width, jint height);


/*****************************************************************************************
 * include macros for no-native-wm comptibility
 */

#include "grlib.h"

#endif
