/**
 * dither.h - utility structs and macros for 2-dim error-diffusion dithering
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


typedef struct {
	signed char r;
	signed char g;
	signed char b;
} pix_err;


static inline void setErr ( pix_err* e, int i, int er, int eg, int eb )
{
	e[i].r += er;  e[i].g = eg;  e[i].b = eb;
}

static inline int normVal ( int v, int* err )
{
	if ( v > 0xff ) {
		*err = v - 0xff;   v = 0xff; 
	}
	else if ( v < 0 ) {
		*err = v;   v = 0;
	}

	return v;
}

static inline g_pixel_t setDiffusedPixel ( Image* img, int x, int y,
										   int r, int g, int b,
										   int* er, int* eg, int* eb  )
{
	g_pixel_t  pix;
	int        rr, gg, bb;

	*er = *eg = *eb = 0;
	r = normVal( r, er);
	g = normVal( g, eg);
	b = normVal( b, eb);

	/* this is the error-corrected rgb value we want */
	pix = g_pixelValue( X, ((r << 16) | (g << 8) | b));
	g_putImagePixel( img, x, y, pix, 0xff);
		
	/* query the channel values we actually got, and compute errors */
	g_rgbValues( X, pix, &rr, &gg, &bb);
	*er += r - rr;
	*eg += g - gg;
	*eb += b - bb;

	return pix;
}

static inline void propagatePixelError ( int col, int row, int width,
										 int er, int eg, int eb,
										 pix_err* e0, pix_err* e1 )
{
	int i;

	/* distribute the channel errors according to the scheme:
	 * (with alternating propagation direction)
	 *
	 *     -->   P    7/16  
	 *    3/16  5/16  1/16
	 */

	if ( row & 1 ){         /* backward propagation */
		if ( col > 0 ){
			i = col - 1;

			/* left neighbor same scanline */
			setErr( e0, i, (er * 7/16), (eg * 7/16), (eb * 7/16));

			/* left neighbor, next scanline (reset first) */
			setErr( e1, i, (er /16), (eg /16), (eb /16));
		}

		if ( col < width-1 ) {
			/* right neightbor, next scanline */
			setErr( e1, col+1, (er * 3/16), (eg * 3/16), (eb * 3/16));
		}
	}
	else {                  /* forward propagation */
		if ( col < width-1 ){
			i = col + 1;

			/* right neighbor same scanline */
			setErr( e0, i, (er * 7/16), (eg * 7/16), (eb * 7/16));

			/* right neighbor, next scanline (reset first) */
			setErr( e1, i, (er /16), (eg /16), (eb /16));
		}

		if ( col > 0 ) {
			/* left neightbor, next scanline */
			setErr( e1, col-1, (er * 3/16), (eg * 3/16), (eb * 3/16));
		}
	}

	/* bottom neightbor */
	setErr( e1, col, (er * 5/16), (eg * 5/16), (eb * 5/16));
}
