/**
 * imggif.c - interface for libgif input routines
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <stdio.h>
#include "config.h"

#if defined(HAVE_GIF_LIB_H) && (defined(HAVE_LIBGIF) || defined(HAVE_LIBUNGIF))
#define HAVE_GIF 1
#endif

#if defined(HAVE_GIF)
#include "gif_lib.h"
#endif

#include "toolkit.h"

#include "dither.h"


/**************************************************************************************
 * internal functions
 */

#if defined(HAVE_GIF)

static int iOffset[] = { 0, 4, 2, 1 };
static int iJumps[] = { 8, 8, 4, 2 };

static void
writeRow ( Image* img, GifPixelType* rowBuf, GifColorType* cm,
		   int row, unsigned char* usedClrs )
{
	int             idx, col;
	jint            rgb;
	g_pixel_t       pix;

	for ( col=0; col < _(img)->width; col++ ) {
		idx = (unsigned int) rowBuf[col];

		rgb = (cm[idx].Red << 16) | (cm[idx].Green << 8) | (cm[idx].Blue);
		pix = g_pixelValue( X, rgb);
		
		if ( usedClrs ){
			usedClrs[pix]=1;
		}
		
		g_putImagePixel( img, col, row, pix, (idx == _(img)->trans) ? 0 : 0xff);
	}
}

static void
writeDitheredRow ( Image* img, GifPixelType* rowBuf, GifColorType* cm,
				   int row, unsigned char* usedClrs,
				   pix_err* e0, pix_err* e1 )
{
	int             idx, col;
	int             r, g, b, er, eg, eb;
	jint            rgb;
	g_pixel_t       pix;

	if ( row & 1 ) {                     /* odd row: process backwards */
		for ( col= _(img)->width-1; col >= 0; col-- ) {
			idx = rowBuf[col];

			if ( idx != _(img)->trans ) { 
				r = cm[idx].Red    + e0[col].r;
				g = cm[idx].Green  + e0[col].g;
				b = cm[idx].Blue   + e0[col].b;

				pix = setDiffusedPixel( img, col, row, r, g, b, &er, &eg, &eb);
				if ( usedClrs ){
					usedClrs[pix]=1;
				}

				propagatePixelError( col, row, _(img)->width,
									 er, eg, eb, e0, e1);
			}
			else {
				rgb = (cm[idx].Red << 16) | (cm[idx].Green << 8) | (cm[idx].Blue);
				pix = g_pixelValue( X, rgb);
				g_putImagePixel( img, col, row, pix, 0);
			}
		}
	}
	else {                               /* even row: process forwards */
		for ( col=0; col < _(img)->width; col++ ) {
			idx = rowBuf[col];

			if ( idx != _(img)->trans ) { 			
				r = cm[idx].Red    + e0[col].r;
				g = cm[idx].Green  + e0[col].g;
				b = cm[idx].Blue   + e0[col].b;

				pix = setDiffusedPixel( img, col, row, r, g, b, &er, &eg, &eb);
				if ( usedClrs ){
					usedClrs[pix]=1;
				}

				propagatePixelError( col, row, _(img)->width,
									 er, eg, eb, e0, e1);
			}
			else {
				rgb = (cm[idx].Red << 16) | (cm[idx].Green << 8) | (cm[idx].Blue);
				pix = g_pixelValue( X, rgb);
				g_putImagePixel( img, col, row, pix, 0);
			}
		}
	}
}



#define CHECK(gifOp)  \
   if ( gifOp == GIF_ERROR ) { error = 1; goto exit; }

#define MAX_ICON_SIZE   64
#define MAX_ICON_COLORS 32


static Image*
readGif ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, GifFileType *gf )
{
	int             error = 0;
	Image           *firstImg = 0, *img;
	int             i, extCode, width, height, row, cmapSize;
	int             trans = -1, nFrames=0, delay=0, disp = 0, nLoops=-1;
	GifRecordType   rec;
	GifByteType     *ext;
	ColorMapObject  *cmap = 0;
	GifColorType    *clrs = 0;
	GifPixelType    *rowBuf = 0;
	int             rowBufSize = 0;
	unsigned char   *usedClrs = 0;
	pix_err         *e0 = 0, *e1 = 0, *p;
	int             dither = 0; 
	g_contextSav_t  sav;
	/* variables added for setting transparent color on the native side */
	int             rgb;
	g_pixel_t       pix;
	
	g_saveContext( &sav);

	DBG( awt_img, ("readGif start\n"));

	if ( _(X)->colorMode == CM_PSEUDO ) {
		/* we use this as a temporary storage to remap transparency indices */
		usedClrs = AWT_ALLOCA( 256);
	}
	
	img = 0;
	do {
		CHECK( DGifGetRecordType( gf, &rec));

		switch ( rec ) {
		case IMAGE_DESC_RECORD_TYPE:
			CHECK( DGifGetImageDesc( gf));

			width    = gf->Image.Width;
			height   = gf->Image.Height;

			cmap     = (gf->Image.ColorMap) ? gf->Image.ColorMap : gf->SColorMap;
			clrs     = cmap->Colors;
			cmapSize = cmap->ColorCount;

			/* this is bloody heuristics - if we exceed a certain size or 
			 * color number, we assume this is a picture that should be dithered
			 */
			dither = (_(X)->colorMode == CM_PSEUDO) && 
				(width > MAX_ICON_SIZE) && (height > MAX_ICON_SIZE) && 
				(cmapSize > MAX_ICON_COLORS);

			/* do a little Gif consistency check (SWidth should be hull, but who knows) */
			if ( !rowBuf || (gf->Image.Width > rowBufSize) ){
				if ( !rowBuf ) {
					rowBufSize = (gf->SWidth > gf->Image.Width)? gf->SWidth : gf->Image.Width;
					rowBuf = (GifPixelType*) AWT_MALLOC( rowBufSize);
				}
				else {
					AWT_FREE( rowBuf);
					rowBufSize = gf->Image.Width;
					rowBuf = (GifPixelType*) AWT_MALLOC( rowBufSize);

					if ( e0 ) AWT_FREE( e0);
					if ( e1 ) AWT_FREE( e1);					
				}
				if ( rowBuf == 0 ){
					/* if we fail to allocate (maybe insane gif width) exit cleanly */
					CHECK( GIF_ERROR);
				}

				/* create the error diffusion cache in case we want to dither */
				if ( dither ){
					if ( !(e0 = AWT_CALLOC( 1, rowBufSize * sizeof(pix_err))) ||
						 !(e1 = AWT_CALLOC( 1, rowBufSize * sizeof(pix_err))) ){
						CHECK( GIF_ERROR);
					}
				}
			}
		
			/*
			 * create our image objects and keep track of frames 
			 */
			if ( !firstImg ) {     /* this is the first (maybe only) frame */
				firstImg = img = createImage( width, height, FALSE);
				if ( !img ){
				    CHECK( GIF_ERROR);
				}

				/* We call back to set Dimensions as soon as possible, to avoid
				 * timing dependent anomalies of various clients like browsers
				 * (make sure we unblock the Toolkit because we might end up
				 * in application specific synchronization)
				 */
				UNBLOCK_EXECUTE( env, sync, {
					(*env)->CallVoidMethod( env, prod, setDim, (jint)width, (jint)height);
				});
							   
			}
			else {                 /* this is a subsequent gif-movie frame, link it in */
				DBG( awt_img, ("readGif - add frame\n"));
				g_completeImage( img);
				
				_(img)->next = createImage( width, height, FALSE);
				if ( !img ) {
				    CHECK( GIF_ERROR);
				}
				img = _(img)->next;
			}

 			g_setAlphaType( img, (trans != -1) ? MASK_ALPHA : NO_ALPHA);
			_(img)->trans = trans;

			if ( disp != 0 ){
				switch ( disp ){
				case 1:
					/* we need to mark the first image, so that the buffer gets
					 * created in time
					 */
					_(firstImg)->flags |= IMG_BUFFER;
					_(img)->flags |= IMG_IS_DELTA;
					break;
				}
				disp = 0;
			}

			_(img)->latency = delay;
			delay = 0;

			_(img)->nLoops = nLoops;
			_(img)->left = gf->Image.Left;
			_(img)->top = gf->Image.Top;
			_(img)->frame = nFrames;

			g_setCurrentImage( X, img);
			nFrames++;

			/*
			 * start reading in the image data
			 */
			if ( gf->Image.Interlace ) {
				/* Need to perform 4 passes on the images: */
				for ( i = 0; i < 4; i++ ) {
					for (row = iOffset[i]; row < height; row += iJumps[i]) {
						memset( rowBuf, gf->SBackGroundColor, width);
						CHECK( DGifGetLine( gf, rowBuf, width));

						writeRow( img, rowBuf, clrs, row, usedClrs);
						/* sorry, no way to dither this unless we want to buffer the
						 * whole thing
						 */
					}
				}
			}
			else {
				for ( row = 0; row < height; row++) {
					memset( rowBuf, gf->SBackGroundColor, width);
					CHECK( DGifGetLine(gf, rowBuf, width));

					if ( !dither ) {
						writeRow( img, rowBuf, clrs, row, usedClrs);
					}
					else {
						writeDitheredRow( img, rowBuf, clrs, row, usedClrs, e0, e1);

						/* rotate error diffusion buffers, zero next scanline buffer */
						p = e0;  e0 = e1;  e1 = p;
						memset( e1, 0, width * sizeof(pix_err));
					}
				}
			}
			break;

		case EXTENSION_RECORD_TYPE:
			CHECK( DGifGetExtension( gf, &extCode, &ext));

			if ( extCode == 0xf9 ) {   /* graphics extension */
				/*
				 * extension record with transparency spec are preceeding description records
				 * (which create new Images), so just cache the tranp index, here
				 */
				if ( ext[1] & 1 ) {      /* transparent index following */
					trans = ext[4];
					DBG( awt_img, ("readGif transparent: %d\n", trans));
				}

				delay = ((ext[3] << 8) | ext[2]) * 10; /* delay in 1/100 secs */

				/* dispose spec */
				if ( ext[1] & 0x1c ) {
					disp = (ext[1] >> 2) & 3;
					DBG( awt_img, ("readGif dispose: %d\n", disp));
				}
			}
			else if ( extCode == 0xff ) {  /* application extension block */
				/* Check for a netscape loop extension. It's not yet clear
				 * if this is the only way Netscape extracts the loop spec,
				 * but it's the only documented info
				 */
				if ( strncmp( &ext[1], "NETSCAPE2.0", 11) == 0 ){
					/* the info is actually in the next extension */
					CHECK( DGifGetExtensionNext( gf, &ext));
					nLoops = ((ext[3] << 8) | ext[2]);
					if ( !nLoops ){
						nLoops = -1; /* we interpret this as 'forever' */
					}
					DBG( awt_img, ("readGif loop extension %d\n", nLoops));
				}
			}

			while ( ext != NULL ) {
				CHECK( DGifGetExtensionNext( gf, &ext));
			}
			break;

		case TERMINATE_RECORD_TYPE:
			break;

		default:                /* Should be traps by DGifGetRecordType. */
			break;
		}
	} while ( rec != TERMINATE_RECORD_TYPE );

	g_completeImage( img);
	g_restoreContext( &sav);

	if ( firstImg ){
		_(firstImg)->nFrames = nFrames;		

		if ( img != firstImg ){
			_(img)->next = firstImg;   /* link it together (as a ring) */
		}
		else {
			/* yes, there are gifs with a single frame and combine spec */
			_(firstImg)->flags &= ~IMG_BUFFER;
		}
	}

	rgb = (clrs[trans].Red << 16) | (clrs[trans].Green << 8) | (clrs[trans].Blue);
	pix = g_pixelValue( X, rgb);
	g_setTranspValue( img, pix );

	if ( usedClrs ) {
		/* if it ever comes to a production again, we have to provide
		 * a transparency index into the SystemColorMap that isn't used.
		 */
		_(img)->trans = NO_TRANS_INDEX_AVAILABLE;
		for ( i=0; i<256; i++ ){
			if ( usedClrs[i] == 0 ) {
				_(img)->trans = i;
				break;
			}
		}
	}
	
	DBG( awt_img, ("readGif end: %p\n", firstImg));

 exit:
    if ( error ) {
		Java_java_awt_Toolkit_imgFreeImage( 0, 0, firstImg);
		firstImg = 0;
	}
    if ( rowBuf )     AWT_FREE( rowBuf);
	if ( usedClrs )   AWT_ALLOCA_FREE( usedClrs);
	if ( e0 )         AWT_FREE( e0);
	if ( e1 )         AWT_FREE( e1);

    g_restoreContext( &sav);

	return firstImg;
}

/**************************************************************************************
 * memory buffer IO
 */

typedef struct {
	unsigned char *buf;
	unsigned char *p;
	int           remain;

	JNIEnv        *env;
	jobject       sync;

	jobject       prod;
    jmethodID     fillBuf;
} BufferSource;

static int
readGifBuffer ( GifFileType *gf, GifByteType* buf, int length )
{
	BufferSource *src;
	GifByteType  *p = buf;
	int n, m = length;

	if ( !(src = (BufferSource*)gf->UserData) ){
		return 0;
	}

	if ( m > src->remain ){
		if ( src->remain ){
			memcpy( p, src->p, src->remain);
			m -= src->remain;
			p += src->remain;
		}

		while ( m ) {
			UNBLOCK_EXECUTE( src->env, src->sync, {
				n = (*(src->env))->CallIntMethod( src->env, src->prod, src->fillBuf);
			});
			if ( n <= 0 ){
				return 0;
			}
			src->p = src->buf;
			src->remain = n;
			
			if ( n > m ) {
				break;
			}
			else {
				memcpy( p, src->p, n);
				m -= n;
				p += n;
			}
		}
	}


	memcpy( p, src->p, m);
	src->p += m;
	src->remain -= m;

	return length;
}

/**************************************************************************************
 * file descriptor IO
 * (intercept all file io and map it to AWT_xx file io maros, in order to
 * obtain thread safety)
 */
typedef struct {
	int fd;
} FileSource;

static int
readGifFileSource( GifFileType *gf, GifByteType* buf, int length )
{
	FileSource *psource = (FileSource*)gf->UserData;
	int n =  AWT_READ( psource->fd, buf, length);

	return (n < 0) ? 0 : n;
}

#endif /* HAVE_GIF */

/**************************************************************************************
 * these are the "exported" production interface functions
 */

Image*
readGifFile ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim,
			  int infile, char* fname )
{
	Image          *img = 0;

#if defined(HAVE_GIF)
	/*
	 * we don't use DGifOpenFile because file io might be intercepted
	 * (because of threading)
	 */
	FileSource     fileSrc;
	GifFileType    *gf;

	fileSrc.fd = infile;

	if ( !(gf = DGifOpen( &fileSrc, readGifFileSource)) )
		return 0;

	img = readGif( env, sync, prod, setDim, gf);

	DGifCloseFile( gf);
#else
	DBG( awt_img, ("readGifFile not supported\n"));
#endif

	return img;
}

Image*
readGifData ( JNIEnv* env, jobject sync,
			  jobject prod, jmethodID setDim, jmethodID fillBuf,
			  unsigned char* buf, int off, int len )
{
	Image          *img = 0;

#if defined(HAVE_GIF)
	BufferSource   bufSrc;
	GifFileType    *gf;

	DBG( awt_img, ("readGifData: %p %d\n", buf, (int)len));

	/* we depend on libungif > 3.01 here (with our alternate input extensions) */
	bufSrc.buf     = buf;
	bufSrc.p       = buf + off;
	bufSrc.remain  = len;

	bufSrc.env     = env;
	bufSrc.sync    = sync;
	bufSrc.prod    = prod;
	bufSrc.fillBuf = fillBuf;

	if ( !(gf = DGifOpen( &bufSrc, readGifBuffer)) )
		return 0;

	img = readGif( env, sync, prod, setDim, gf);

	DGifCloseFile( gf);

#else
	DBG( awt_img, ("readGifData not supported\n"));  
#endif /* HAVE_GIF */


	return img;
}
