/**
 * imgjpeg.c - interface for libjpeg GIF input routines
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "toolkit.h"

#if defined(HAVE_JPEGLIB_H) && defined(HAVE_LIBJPEG)
#define HAVE_JPEG 1
#endif

#if defined(HAVE_JPEG)
#include <setjmp.h>
#undef HAVE_STDDEF_H   /* To avoid conflict in JPEG */
#include <jpeglib.h>
#include <jerror.h>
#endif

#include "dither.h"

#if defined(HAVE_JPEG)

/*******************************************************************************************
 * required typedefs and structs
 */

struct error_mgr {
	struct jpeg_error_mgr pub;
	jmp_buf setjmp_buffer;
};

typedef struct error_mgr * error_ptr;

METHODDEF(void)
error_exit ( j_common_ptr cinfo)
{
	error_ptr myerr = (error_ptr) cinfo->err;
	(*cinfo->err->output_message)(cinfo);
	longjmp(myerr->setjmp_buffer, 1);
}

typedef struct {
	struct jpeg_source_mgr pub;	/* public fields */

	JNIEnv    *env;
	jobject   sync;
	jobject   prod;
	jmethodID fill;

	int       len;          /* length of external buffer */
	JOCTET    *buffer;		/* start of buffer */
} bufsrc_mgr;

typedef bufsrc_mgr * bufsrc_ptr;


#define BUF_SIZE 2048

typedef struct {
	struct jpeg_source_mgr pub;	/* public fields */

	int fd;
	JOCTET *buffer;
} filesrc_mgr;

typedef filesrc_mgr * filesrc_ptr;


/*******************************************************************************************
 * memory buffer extension for the standard jpeg lib input management (jdatasrc.c)
 */

METHODDEF(void)
init_buf_source (j_decompress_ptr cinfo)
{
    /* nothing to do */
}


static JOCTET end_marker[2] = { (JOCTET) 0xFF, (JOCTET) JPEG_EOI };

METHODDEF(boolean)
get_buf_source (j_decompress_ptr cinfo)
{
	bufsrc_ptr src = (bufsrc_ptr) cinfo->src;
	
	UNBLOCK_EXECUTE( src->env, src->sync, {
		src->len = (*(src->env))->CallIntMethod( src->env, src->prod, src->fill);
	});
	if ( src->len <= 0 ){
		src->pub.next_input_byte = end_marker;
		src->pub.bytes_in_buffer = 0; /* sizeof(end_marker); */
		return TRUE;
	}

	/* we read in one sweep (no need to use two buffers) */
	src->pub.next_input_byte = src->buffer;
	src->pub.bytes_in_buffer = src->len;

	return TRUE;
}


METHODDEF(void)
skip_buf_source ( j_decompress_ptr cinfo, long num_bytes )
{
	bufsrc_ptr src = (bufsrc_ptr) cinfo->src;

	if (num_bytes > 0) {
		while (num_bytes > (long) src->pub.bytes_in_buffer) {
			num_bytes -= (long) src->pub.bytes_in_buffer;
			get_buf_source (cinfo);
		}
		src->pub.next_input_byte += (size_t) num_bytes;
		src->pub.bytes_in_buffer -= (size_t) num_bytes;
	}
	else {
		/* we can't do back jumps, anymore */
		ERREXITS( cinfo, JERR_NOTIMPL, "back-skip not supported");
	}
}


METHODDEF(void)
term_buf_source (j_decompress_ptr cinfo)
{
	bufsrc_ptr src = (bufsrc_ptr) cinfo->src;

	/*
	 * Take care the (external) buffer isn't deleted by jpeg_destroy_decompress.
	 * This shouldn't be necessary since jpeg_destroy cleans up by means of the 
	 * jpeg memory manager, but we better play it safe
	 */
	src->buffer = 0;
}


static GLOBAL(void)
jpeg_buffer_src ( j_decompress_ptr cinfo,
				  JNIEnv* env, jobject sync, jobject prod, jmethodID fill,
				  unsigned char* buf, int off, int len )
{
	bufsrc_ptr src;

	if (cinfo->src == NULL) {	/* first time for this JPEG object? */
		cinfo->src = (struct jpeg_source_mgr *)
			(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
										sizeof(bufsrc_mgr));
	}

	src = (bufsrc_ptr) cinfo->src;
	src->pub.init_source = init_buf_source;
	src->pub.fill_input_buffer = get_buf_source;
	src->pub.skip_input_data = skip_buf_source;
	src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
	src->pub.term_source = term_buf_source;

	src->env  = env;
	src->sync = sync;

	src->prod = prod;
	src->fill = fill;
	
	src->buffer = buf;
	src->len    = len;

	src->pub.bytes_in_buffer = src->len;
	src->pub.next_input_byte = buf + off;
}


/*******************************************************************************************
 * file desriptor read extension for the standard jpeg lib input management (jdatasrc.c)
 */

METHODDEF(void)
init_file_source (j_decompress_ptr cinfo)
{
	/* nothing to do */
}

METHODDEF(boolean)
get_file_source (j_decompress_ptr cinfo)
{
	filesrc_ptr src = (filesrc_ptr) cinfo->src;
	int         n = AWT_READ( src->fd, src->buffer, BUF_SIZE);

	if ( n < 0 ) {
		WARNMS(cinfo, JWRN_JPEG_EOF);
		src->pub.next_input_byte = end_marker;
		src->pub.bytes_in_buffer = sizeof(end_marker);
	}
	else {
		src->pub.next_input_byte = src->buffer;
		src->pub.bytes_in_buffer = n;
	}

	return TRUE;
}


METHODDEF(void)
skip_file_source ( j_decompress_ptr cinfo, long num_bytes )
{
	filesrc_ptr src = (filesrc_ptr) cinfo->src;

	if ( num_bytes < src->pub.bytes_in_buffer ){
		src->pub.next_input_byte += num_bytes;
		src->pub.bytes_in_buffer -= num_bytes;
	}
	else {
		num_bytes -= src->pub.bytes_in_buffer;

		AWT_SETPOS( src->fd, num_bytes);
		get_file_source( cinfo);
	}
}


METHODDEF(void)
term_file_source (j_decompress_ptr cinfo)
{
	/* nothing to do, here
	 * the file is closed by our caller, the buffer is allocated on the stack
	 */
}


static GLOBAL(void)
jpeg_file_src (j_decompress_ptr cinfo, int fd )
{
	filesrc_ptr src;

	if (cinfo->src == NULL) {	/* first time for this JPEG object? */
		src = (filesrc_ptr)
			(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
										sizeof(filesrc_mgr));
		src->buffer = (JOCTET *)
			(*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
										BUF_SIZE * sizeof(JOCTET));
		cinfo->src = (struct jpeg_source_mgr *) src;
	}
	else {
		src = (filesrc_ptr) cinfo->src;
	}

	src->pub.init_source = init_file_source;
	src->pub.fill_input_buffer = get_file_source;
	src->pub.skip_input_data = skip_file_source;
	src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
	src->pub.term_source = term_file_source;

	src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
	src->pub.next_input_byte = NULL; /* until buffer loaded */

	src->fd = fd;
}


/************************************************************************************
 * JPEG auxiliary funcs (production callbacks)
 */

static void
jscan_to_img( Image * img, JSAMPROW buf, struct jpeg_decompress_struct * cinfo)
{
	g_pixel_t  pix;
	register int  col, rgb;

	for ( col = 0; col < cinfo->output_width; col++) {

		/* we have requested a JCS_RGB transformation, i.e. we get three
		 * R/G/B output components (each one byte) per pixel
		 */
		rgb  = ((*buf++)<<16);
		rgb |= ((*buf++)<<8);
		rgb |= *buf++;

		pix = g_pixelValue( X, rgb);
		g_putImagePixel( img, col, cinfo->output_scanline-1, pix, 0xff);
	}
}



static void
jscan_to_dithered_img( Image* img, JSAMPROW buf,
					   struct jpeg_decompress_struct* cinfo,
					   pix_err* e0, pix_err* e1 )
{
	int  col, row;
	int  r, g, b, er, eg, eb;
	
	row = cinfo->output_scanline -1;     /* jpeg has already advanced it */

	if ( row & 1 ) {                     /* odd row: process backwards */
		buf += 3 * cinfo->output_width -1;	   
		for ( col = cinfo->output_width-1; col>=0; col-- ){
			b = (*buf--) + e0[col].b;
			g = (*buf--) + e0[col].g;
			r = (*buf--) + e0[col].r;
			
			setDiffusedPixel( img, col, row, r, g, b, &er, &eg, &eb);
			propagatePixelError( col, row, cinfo->output_width,
								 er, eg, eb, e0, e1);
		}
	}
	else {                               /* even row: process forwards */
		for ( col=0; col < cinfo->output_width; col++ ) {
			r = (*buf++) + e0[col].r;
			g = (*buf++) + e0[col].g;
			b = (*buf++) + e0[col].b;
			
			setDiffusedPixel( img, col, row, r, g, b, &er, &eg, &eb);
			propagatePixelError( col, row, cinfo->output_width,
								 er, eg, eb, e0, e1);
		}
	}
}

/************************************************************************************
 * Jpeg production
 */

static Image*
readJpeg ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim,
		   struct jpeg_decompress_struct* cinfo )
{
	Image  *img = 0;
	struct error_mgr jerr;
	JSAMPARRAY buffer = 0;
	int        row_stride;
	pix_err    *e0 = 0, *e1 = 0, *p;
	g_contextSav_t sav;

	g_saveContext( &sav);

	/* set up our error handler/exit */
	cinfo->err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = error_exit;

	if ( setjmp(jerr.setjmp_buffer)) {
		if ( img )    Java_java_awt_Toolkit_imgFreeImage( 0, 0, img);
		if ( e0 )     AWT_FREE( e0);
		if ( e1 )     AWT_FREE( e1);

		g_restoreContext( &sav);
		return 0;
	}

	/* now start to process the jpeg source */
	jpeg_read_header( cinfo, TRUE);

	/* We call back to set Dimensions as soon as possible, to avoid
	 * timing dependent anomalies of various clients like browsers
	 * (make sure we unblock the Toolkit because we might end up
	 * in application specific synchronization)
	 */
	UNBLOCK_EXECUTE( env, sync, {
		(*env)->CallVoidMethod( env, prod, setDim,
							   (jint)cinfo->image_width, (jint)cinfo->image_height);
	});

	/* we don't want to quantize because of two reasons:
	 * (a) it internally requires a full image buffer (w*h*3)
	 * (b) we want to dither the results for non-TrueColor visuals, anyway
	 */
	cinfo->quantize_colors = FALSE;
	cinfo->out_color_space = JCS_RGB;

	jpeg_start_decompress( cinfo);
	row_stride = cinfo->output_width * cinfo->output_components;

	/* create the error diffusion cache in case we want to dither */
	if ( _(X)->colorMode != CM_TRUE ){
		if ( !(e0 = AWT_CALLOC( 1, cinfo->output_width * sizeof(pix_err))) ||
			 !(e1 = AWT_CALLOC( 1, cinfo->output_width * sizeof(pix_err))) ){
			ERREXIT1( cinfo, JERR_OUT_OF_MEMORY, 1);
		}
	}

	/* it's time to create the target image */
	if ( !(img = createImage( cinfo->output_width, cinfo->output_height, FALSE)) ){
		ERREXIT1( cinfo, JERR_OUT_OF_MEMORY, 2);
	}
	g_setAlphaType( img, NO_ALPHA);
	g_setCurrentImage( X, img);

	buffer = (*cinfo->mem->alloc_sarray)((j_common_ptr) cinfo, JPOOL_IMAGE, row_stride, 1);
		
	while (cinfo->output_scanline < cinfo->output_height) {
		jpeg_read_scanlines( cinfo, buffer, 1);

		if ( _(X)->colorMode == CM_TRUE ){
			jscan_to_img( (Image*)img, buffer[0], cinfo);
		}
		else {
			jscan_to_dithered_img( (Image*)img, buffer[0], cinfo, e0, e1);
			
			/* rotate error diffusion buffers, zero next scanline buffer */
			p = e0;  e0 = e1;  e1 = p;
			memset( e1, 0, cinfo->output_width * sizeof(pix_err));
		}
	}

	jpeg_finish_decompress( cinfo);

	if ( _(X)->colorMode != CM_TRUE ){
		AWT_FREE(e0);
		AWT_FREE(e1);
	}

	g_completeImage( img);
	g_restoreContext( &sav);

	return img;
}

#endif /* HAVE_JPEGLIB_H */


/**************************************************************************************
 * these are the "exported" production interface functions
 */

Image*
readJpegFile ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim,
			   int infile, char* fname )
{
	Image *img = 0;

#ifdef HAVE_JPEG
	struct jpeg_decompress_struct cinfo;

	jpeg_create_decompress( &cinfo);
	jpeg_file_src( &cinfo, infile);

	img = readJpeg( env, sync, prod, setDim, &cinfo );

	jpeg_destroy_decompress( &cinfo);
#else
	DBG( awt_img, ("readJpegFile not supported\n"));
#endif /* HAVE_JPEG */

	return img;
}


Image*
readJpegData ( JNIEnv* env, jobject sync,
			   jobject prod, jmethodID setDim, jmethodID fillBuf,
               unsigned char* buf, int off, int len )
{
	Image *img = 0;

#ifdef HAVE_JPEG
	struct jpeg_decompress_struct cinfo;

	jpeg_create_decompress( &cinfo);
	jpeg_buffer_src( &cinfo, env, sync, prod, fillBuf, buf, off, len);

	img = readJpeg( env, sync, prod, setDim, &cinfo );

	jpeg_destroy_decompress( &cinfo);
#else
  DBG( awt_img, ("readJpegData not supported\n"));
#endif /* HAVE_JPEG */

	return img;
}
