/**
 * imgpng.c - interface for libpng input routines
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"

#if defined(HAVE_PNG_H) && defined(HAVE_LIBPNG) && defined(HAVE_LIBZ)
#define INCLUDE_PNG 1
#endif

#if defined(INCLUDE_PNG)
#include "png.h"
#endif

#include "toolkit.h"


#if defined (INCLUDE_PNG)

/**************************************************************************************
 * auxiliary funtions
 */

static inline int
hasAlpha ( png_structp png_ptr )
{
	return  (png_ptr->color_type == PNG_COLOR_TYPE_GRAY_ALPHA ||
			 png_ptr->color_type == PNG_COLOR_TYPE_RGB_ALPHA);

}

static inline jint
readARGB ( unsigned char** p, int hasAlpha )
{
	jint argb;

	argb = *(*p)++;
	argb <<= 8;
	argb += *(*p)++;
	argb <<= 8;
	argb += *(*p)++;

	if ( hasAlpha ) {
		argb <<= 8; 
		argb += *(*p)++;
	}

	return argb;
}


static void
readRowData ( png_structp png_ptr, png_infop info_ptr, png_bytep row,
			  Image *img, unsigned char* usedClrs )
{
	int            i, j;
	jint           argb;
	int            trans = g_getAlphaType( img);
	g_pixel_t      pix;
	unsigned char  *p;

	for ( i = 0; i < info_ptr->height; i++ ) {
		png_read_row( png_ptr, row, 0);

		for ( j=0, p=(unsigned char*)row; j<info_ptr->width; j++ ) {
			argb = readARGB( &p, 1 /*trans*/);
			pix = g_pixelValue(X, argb);
			g_putImagePixel( img, j, i, pix, (unsigned char)(argb>>24));
			if ( usedClrs ){
				usedClrs[pix]=1;
			}
		}
	}
}

static void
readImageData ( png_structp png_ptr, png_infop info_ptr, png_bytep rows[],
				Image *img, unsigned char* usedClrs )
{
	int            i, j;
	jint           argb;
	unsigned char  *p;
	g_pixel_t      pix;
	int            trans = g_getAlphaType( img);

	png_read_image( png_ptr, rows);

	for ( i=0; i<info_ptr->height; i++ ) {
		p = (unsigned char*) rows[i];

		for ( j=0; j<info_ptr->width; j++ ) {
			argb = readARGB( &p, trans);
			pix = g_pixelValue(X, argb);
			g_putImagePixel( img, j, i, pix, (unsigned char)(argb>>24));

			if ( usedClrs ){
				usedClrs[pix]=1;
			}
		}
	}
}


static void
readInterlacedData ( png_structp png_ptr, png_infop info_ptr, png_bytep rows[],
                     Image *img, unsigned char* usedClrs, int passes )
{
	// not implemented
}


static Image*
readPng ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim,
		  png_structp png_ptr, png_infop info_ptr )
{
	Image          *img = 0;
	png_bytepp     rows = 0;
	png_bytep      data = 0;
	double         screen_gamma = 1.2, file_gamma;
	int            i, number_passes;
	int            row_bytes;
	unsigned char  *usedClrs = 0;
	g_contextSav_t sav;

	g_saveContext( &sav);

	if ( _(X)->colorMode == CM_PSEUDO ) {
		/* we use this as a temporary storage to remap transparency indices */
		usedClrs = AWT_ALLOCA( 256);
	}

	/* set up the error exit */
	if ( setjmp(png_ptr->jmpbuf) ) {
		if ( img )
			Java_java_awt_Toolkit_imgFreeImage( 0, 0, img);
		if ( rows )
			AWT_FREE( rows);
		if ( data )
			AWT_FREE( data);
		if ( usedClrs )
			AWT_ALLOCA_FREE( usedClrs);
		g_restoreContext( &sav);
		return 0;
	}

	png_read_info( png_ptr, info_ptr);

	/* handle gamma correction  */   
	{
#if defined(PNG_READ_sRGB_SUPPORTED)
		int intent;
		if ( png_get_sRGB( png_ptr, info_ptr, &intent) ) {
			png_set_sRGB( png_ptr, info_ptr, intent);
		}
		else
#endif
			{
				if ( png_get_gAMA( png_ptr, info_ptr, &file_gamma) )
					png_set_gamma( png_ptr, screen_gamma, file_gamma);
				else
					png_set_gamma( png_ptr, screen_gamma, 0.50);
			}
	}

	png_set_strip_16( png_ptr);
	png_set_expand( png_ptr);
	png_set_gray_to_rgb( png_ptr);
	png_set_swap_alpha( png_ptr); /* we need ARGB instead of the standard RGBA */

	/* Optional call to gamma correct and add the background to the palette */
	png_read_update_info( png_ptr, info_ptr);

	row_bytes  = png_get_rowbytes( png_ptr, info_ptr);
	number_passes = png_set_interlace_handling( png_ptr);

	/* We call back to set Dimensions as soon as possible, to avoid
	 * timing dependent anomalies of various clients like browsers
	 */
	UNBLOCK_EXECUTE( env, sync, {
		(*env)->CallVoidMethod( env, prod, setDim, 
								(jint)info_ptr->width, (jint)info_ptr->height);
	});

	/* time to create the image */
	if ( !(img = createImage( info_ptr->width, info_ptr->height, 0)) ){
	    return 0;
	}
	
	/* at this point, there's no telling how much alpha this image will need */
	g_setAlphaType( img, hasAlpha( png_ptr) ? FULL_ALPHA : NO_ALPHA);
	g_setCurrentImage( X, img);


	if ( info_ptr->interlace_type != 0 ) {
		rows = AWT_MALLOC( sizeof(png_bytep) * info_ptr->height);		
		data = AWT_MALLOC( row_bytes * info_ptr->height);
		
		if ( !rows || !data ){
		    longjmp( png_ptr->jmpbuf, 1);
		}
		
		for ( i=0; i<info_ptr->height; i++ ){
			rows[i] = (data + i*row_bytes);
		}

#ifndef OPTIMIZE_SPACE
		/*
		 * This is bad: to read an interlaced image, we need enough space to (temporarily)
		 * store the whole transformed data (passes need prev. results). Unfortunately,
		 * interlacing is used for large images, and this might require a LOT of memory.
		 */
		readImageData( png_ptr, info_ptr, rows, img, usedClrs);
#else
		readInterlacedData( png_ptr, info_ptr, rows, img, usedClrs, number_passes);
#endif
		AWT_FREE( rows);	
		AWT_FREE( data);
	}
	else {
		data = AWT_MALLOC( row_bytes);
		
		if ( !data ) {
		    longjmp( png_ptr->jmpbuf, 1);
		}
		
		readRowData( png_ptr, info_ptr, data, img, usedClrs);
		AWT_FREE( data);
	}

	/* read rest of file, and get additional chunks in info_ptr */
	png_read_end( png_ptr, info_ptr);
	
	/* try to reduce alphas in the name of efficiency */
	g_completeImage( img);
	g_restoreContext( &sav);

	if ( usedClrs ) {
		/* if it ever comes to a production again, we have to provide
		 * a transparency index into the SystemColorMap that isn't used.
		 */
		_(img)->trans = NO_TRANS_INDEX_AVAILABLE;
		for ( i=0; i<256; i++ ){
			if ( usedClrs[i] == 0 ) {
				_(img)->trans = i;
				break;
			}
		}
		AWT_ALLOCA_FREE( usedClrs);
	}

	return img;
}


/**************************************************************************************
 * memory buffer IO
 */

typedef struct {
	unsigned char *buf;
	unsigned char *p;
	int           remain;

	JNIEnv        *env;
	jobject       sync;

	jobject       prod;
    jmethodID     fillBuf;
} BufferSource;

static void 
bufferRead ( png_structp png_ptr, png_bytep data, png_size_t length )
{
	BufferSource *src;
	unsigned char *p = data;
	int n, m = length;

	if ( !(src = png_get_io_ptr( png_ptr)) ){
		png_error( png_ptr, "illegal data read");
		return;
	}

	if ( m > src->remain ){
		if ( src->remain ){
			memcpy( p, src->p, src->remain);
			m -= src->remain;
			p += src->remain;
		}

		while ( m ) {
			UNBLOCK_EXECUTE( src->env, src->sync, {
				n = (*(src->env))->CallIntMethod( src->env, src->prod, src->fillBuf);
			});
			if ( n <= 0 ){
				png_error( png_ptr, "illegal data read");
				return;
			}
			src->p = src->buf;
			src->remain = n;
			
			if ( n > m ) {
				break;
			}
			else {
				memcpy( p, src->p, n);
				m -= n;
				p += n;
			}
		}
	}


	memcpy( p, src->p, m);
	src->p += m;
	src->remain -= m;
}

/**************************************************************************************
 * file descriptor IO
 * (intercept all file io and map it to AWT_xx file io maros, in order
 * to obtain thread safety)
 */
typedef struct {
	int fd;
} FileSource;

static void
fileRead ( png_structp png_ptr, png_bytep data, png_size_t length )
{
	FileSource *psource = png_get_io_ptr( png_ptr);
	int        n = AWT_READ( psource->fd, data, length);
  
	if ( n != length )
		png_error( png_ptr, "illegal data read");
}

#endif /* HAVE_PNG_H */

/**************************************************************************************
 * these are the "exported" production interface functions
 */

Image*
readPngFile ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim,
			  int infile, char* fname )
{
	Image          *img = 0;
#ifdef INCLUDE_PNG
	png_structp    png_ptr;
	png_infop      info_ptr;
	FileSource     source;

	source.fd = infile;

	png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING,
									  (png_voidp)0, (png_error_ptr)0, (png_error_ptr)0);
	if ( !png_ptr )
		return 0;

	info_ptr = png_create_info_struct( png_ptr);
	png_set_read_fn( png_ptr, &source, fileRead);

	img = readPng( env, sync, prod, setDim, png_ptr, info_ptr );

	png_destroy_read_struct( &png_ptr, &info_ptr, (png_infopp)NULL);
#else
	DBG( awt_img, ("readPngFile not supported\n"));
#endif /* INCLUDE_PNG */

	return img;
}


Image*
readPngData ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
			  unsigned char* buf, int off, int len )
{
	Image          *img = 0;

#ifdef INCLUDE_PNG
	png_structp    png_ptr;
	png_infop      info_ptr;
	BufferSource   bufSrc;

	DBG( awt_img, ("readPngData: %p %d %d\n", buf, off, len));

	bufSrc.buf     = buf;
	bufSrc.p       = buf + off;
	bufSrc.remain  = len;

	bufSrc.env     = env;
	bufSrc.sync    = sync;

	bufSrc.prod    = prod;
	bufSrc.fillBuf = fillBuf;

	png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING,
									  (png_voidp)0, (png_error_ptr)0, (png_error_ptr)0);
	if ( !png_ptr )
		return 0;

	info_ptr = png_create_info_struct( png_ptr);
	png_set_read_fn( png_ptr, &bufSrc, bufferRead);

	img = readPng( env, sync, prod, setDim, png_ptr, info_ptr);

	png_destroy_read_struct( &png_ptr, &info_ptr, (png_infopp)NULL);
#else
	DBG( awt_img, ("readPngData not supported\n"));
#endif /* INCLUDE_PNG */

	return img;
}
