/**
 * imggif.c - interface for libgif input routines
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <stdio.h>
#include "config.h"

#if defined(HAVE_TIFFIO_H) && defined(HAVE_LIBTIFF)
#define	HAVE_TIFF 1
#endif

#if defined(HAVE_TIFF)
#include "tiffio.h"
#endif

#include "toolkit.h"


/**************************************************************************************
 * internal functions
 */

#if defined(HAVE_TIFF)

static Image*
readTiff ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, TIFF* tif )
{
	Image         *img = 0;
	int           width, height, x, y;
	short         extraSamples;
	uint32        *raster, *p;
	int           trans = NO_ALPHA;
	int           r,g,b,a, i;
	g_pixel_t     pix;
	unsigned char *usedClrs = 0;
	g_contextSav_t  sav;

	TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, &width);
	TIFFGetField( tif, TIFFTAG_IMAGELENGTH, &height);

	if ( width <= 0 || height <= 0 ) {
		return 0;
	}

	if ( _(X)->colorMode == CM_PSEUDO ) {
		/* we use this as a temporary storage to remap transparency indices */
		usedClrs = AWT_ALLOCA( 256);
	}

	/* That doesn't help us much, here. Because TIFF needs a lot of data upfront,
	 * we don't come back with the dimensions much faster than what it takes to
	 * produce the whole thing. For the sake of consistency (with the other
	 * decoders) we do it, anyway
	 */
	UNBLOCK_EXECUTE( env, sync, {
		(*env)->CallVoidMethod( env, prod, setDim, (jint)width, (jint)height);
	});

	if ( !(raster = (uint32*) AWT_MALLOC( width*height * sizeof (uint32))) ){
		goto exit;
	}

	if ( TIFFReadRGBAImage( tif, width, height, raster, 0) ) {
		g_saveContext( &sav);
		
		if ( !(img = createImage( width, height, 0)) ){
			goto exit;
		}

		g_setCurrentImage( X, img);

		TIFFGetField( tif, TIFFTAG_EXTRASAMPLES, &extraSamples);
		if ( extraSamples == EXTRASAMPLE_ASSOCALPHA ){
			trans = MASK_ALPHA;
		}
		else if ( extraSamples == EXTRASAMPLE_UNASSALPHA ){
			trans = FULL_ALPHA;
		}
		g_setAlphaType( img, trans);

		for ( y = 0, p=raster; y < height; y++ ){
			for ( x = 0; x < width; x++ ){
				r = TIFFGetR( *p);
				g = TIFFGetG( *p);
				b = TIFFGetB( *p);
				a = TIFFGetA( *p);

				pix = g_pixelValue( X, (r << 16) | (g << 8) | b);
				g_putImagePixel( img, x, y, pix, a);

				if ( usedClrs ) {
					usedClrs[pix] = 1;
				}

				p++;
			}
		}

		g_completeImage( img);

		if ( usedClrs ) {
			/* if it ever comes to a production again, we have to provide
			 * a transparency index into the SystemColorMap that isn't used.
			 */
			_(img)->trans = NO_TRANS_INDEX_AVAILABLE;
			for ( i=0; i<256; i++ ){
				if ( usedClrs[i] == 0 ) {
					printf( "alpha idx: %d\n", i);
					_(img)->trans = i;
					break;
				}
			}
		}
	}

 exit:
	if ( raster )
		AWT_FREE( raster);
	if ( usedClrs )
		AWT_ALLOCA_FREE( usedClrs);
	g_restoreContext( &sav);

	return img;
}

#endif /* HAVE_TIFF */

/**************************************************************************************
 * these are the "exported" production interface functions
 */
Image*
readTiffFile ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim,
			   int infile, char* filename )
{
	Image          *img = 0;
#if defined(HAVE_TIFF)
	TIFF* tif = TIFFFdOpen( infile, filename, "r");

	if ( tif ){
		img = readTiff( env, sync, prod, setDim, tif);
		TIFFClose(tif);
	}
#else
	DBG( awt_img, ("readTiffFile not supported\n"));
#endif

	return img;
}


Image*
readTiffData ( JNIEnv* env, jobject sync,
			   jobject prod, jmethodID setDim, jmethodID fillBuf,
			   unsigned char* buf, int off, int len )
{
	Image          *img = 0;
#if defined(HAVE_TIFF)

	/* Hmm, trouble. Tiff is pretty file oriented, i.e. it depends on
	 * mmap or seek functions. Since we no longer read in the whole data,
	 * and our buffer is probably too small to do this, we can't go with the
	 * usual BufferSource approach. To mimic seek/mmap for URL connections would be
	 * terribly inefficient. We backup to local, temporary files, here
	 */

	TIFF  *tif;
	int   infile;
	int    n = len;
	unsigned char *p = buf + off;
	char  fname[32];

	strcpy( fname, "tfXXXXXX");
	if ( (infile = mkstemp( fname)) < 0 ){
		return 0;
	}

	UNBLOCK_EXECUTE( env, sync, {
		while ( n > 0 ){
			write( infile, p, n);
			n = (*env)->CallIntMethod( env, prod, fillBuf);
			p = buf;
		}
	});

	/* libtiff is picky about that - if the file isn't positioned at the
	 * beginning, it complains about not being able to read the TIFF header
	 */
	lseek( infile, 0, SEEK_SET);
	tif = TIFFFdOpen( infile, fname, "r");

	if ( tif ){
		img = readTiff( env, sync, prod, setDim, tif);
		TIFFClose(tif);
	}

	close( infile);
	remove( fname);
#else
	DBG( awt_img, ("readTiffData not supported\n"));
#endif /* HAVE_TIFF */

	return img;
}
