/**
 * imgxbm.c - xbm image support
 *            This is a pretty inefficient format. Since Java doesn't
 *            support monochrome (color swappable) images, this is
 *            mapped to a full Image with a mask bitmap (the unset
 *            xbm pixels are transparent, the set ones are black), which
 *            is a terrible waste of space. It wouldn't be here if not for
 *            the PersonalJava spec
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <stdio.h>
#include "config.h"

#include "toolkit.h"

typedef struct {
	JNIEnv        *env;
	jobject       sync;
	jobject       prod;
    jmethodID     fillBuf;
} javaCallback;

typedef struct _readBuf {
	char  *buf;
	int   bsize;
	int   len;
	int   i;

	int   (*fill)( struct _readBuf* );
	void* dUsr;
} readBuf;

static inline char
nextByte ( readBuf* b )
{
	if ( b->i < b->len ){
		return b->buf[b->i++];
	}
	else {
		if ( b->fill && ((b->len = b->fill( b)) > 0) ){
			b->i = 1;
			return b->buf[0];
		}
		return 0;
	}
}

static int
nextNumber ( readBuf* b )
{
	char c, cLast = 0;
	int n;

	while ( (c=nextByte(b)) ) {
		/* skip C comments */
		if ( (c == '/') && (nextByte(b) == '*') ){
			c = nextByte(b);
			do {
				while ( c && (c != '*') ){
					c = nextByte(b);
				}
				if ( !c )
					return 0;
			} while ( (c = nextByte(b)) != '/' );
		}


		if ( (c >= '0') && (c <= '9') ){

			/* if the last char wasn't alpha, this is a number */
			if ( cLast <= 'A' ){
				cLast = c;
				c = nextByte(b);

				if ( (cLast == '0') && (c == 'x') ){
					/* hex byte */
					c = nextByte(b);
					n = ( c > '9' ) ? (10 + (c|0x20) - 'a') : c - '0';
					c = nextByte(b);
					n *= 16;
					n += ( c > '9' ) ? (10 + (c|0x20) - 'a') : c - '0';
				}
				else {
					/* decimal number */
					n = (cLast-'0') * 10 + (c - '0');
					for ( c=nextByte(b); (c >= '0') && (c <= '9'); c=nextByte(b)){
						n *= 10;
						n += (c - '0');
					}
				}

				return n;
			}
		}

		cLast = c;
	}

	return -1;
}

static int
fillBufFromFile ( readBuf* b )
{
	int n;

	n = AWT_READ( (int)b->dUsr, b->buf, b->bsize);
	if ( n > 0 ){
		b->i = 0;
		b->len = b->bsize = n;

		return n;
	}

	return 0;
}

static int
fillBufFromJava ( readBuf* b )
{
	int n;
	javaCallback *cb = (javaCallback*) b->dUsr;

	UNBLOCK_EXECUTE( cb->env, cb->sync, {
		n = (*(cb->env))->CallIntMethod( cb->env, cb->prod, cb->fillBuf);
	});

	if ( n > 0 ){
		b->i = 0;
		b->len = b->bsize = n;

		return n;
	}

	return 0;
}

static Image*
readXbm ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, readBuf* b )
{
	int n, width, height, i, j, l;
	Image *img;

	width = nextNumber( b);
	height = nextNumber( b);

	if ( (width == 0) || (height == 0) ){
		return 0;
	}

	/* We call back to set Dimensions as soon as possible, to avoid
	 * timing dependent anomalies of various clients like browsers
	 */
	UNBLOCK_EXECUTE( env, sync, {
		(*env)->CallVoidMethod( env, prod, setDim, width, height);
	});

	if ( !(img = createImage( width, height, 0)) ){
	    return 0;
	}
	DBG( awt_img, ("readXbm: %p  %d,%d\n", img, width, height));

	g_setAlphaType( img, MASK_ALPHA);
	g_setCurrentImage( X, img);
			
	for ( j=0; j<height; j++){
		n = nextNumber( b);
		for ( i=0, l=0; i<width; ){
			if ( n & 0x01 ) {
				g_putImagePixel( img, i, j, _(X)->black, 0xff);
			}
			else {
				g_putImagePixel( img, i, j, _(X)->white, 0);
			}
			i++; l++;

			if ( (l == 8) && (i < width) ){
				n = nextNumber( b);
				l = 0;
			}
			else {
				n >>= 1;
			}
		}
	}

	/* this is easy - since we just use black as a non-transparent color, we can
	 * simply choose white as the transparent one (for subsequent productions)
	 */
	_(img)->trans = _(X)->white;

	g_completeImage( img);
	return img;
}

/**************************************************************************************
 * these are the "exported" production interface functions
 */

Image*
readXbmFile ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim,
			  int infile, char* fname )
{
	Image          *img = 0;
	readBuf        b;

	b.len = b.i = 0;
	b.bsize = 512; 
	b.dUsr = (void*) infile;
	b.fill = &fillBufFromFile;
	if ( !(b.buf = AWT_ALLOCA( b.bsize)) ){
	    return 0;
	}

	DBG( awt_img, ("readXbmFile: %s %d\n", fname, infile));

	img = readXbm( env, sync, prod, setDim, &b);

	AWT_ALLOCA_FREE( b.buf);

	return img;
}


Image*
readXbmData ( JNIEnv* env, jobject sync, 
			  jobject prod, jmethodID setDim, jmethodID fillBuf,
			  unsigned char* buf, int off, int len )
{
	Image          *img = 0;
	readBuf        b;
	javaCallback   cb;

	cb.env     = env;
	cb.sync    = sync;

	cb.prod    = prod;
	cb.fillBuf = fillBuf;	

	b.buf  = buf;
	b.i    = off;
	b.len  = b.bsize = off+len;
	b.dUsr = &cb;
	b.fill = &fillBufFromJava;

	DBG( awt_img, ("readXbmData: %p %d %d\n", buf, off, len));

	img = readXbm( env, sync, prod, setDim, &b);

	return img;
}
