/**
 * fnt.c - font support
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <sys/stat.h>
#include <unistd.h>
#include <grx20.h>
#include <string.h>

#include "toolkit.h"


/*******************************************************************************
 * internal functions
 */

void
initFonts ( Toolkit* X )
{
  char *path;
  struct stat fs;

  path = get_config_string( "fonts", "path", "/allegro/fonts");
  if ( (stat( path, &fs) < 0) || !(fs.st_mode & S_IFDIR) ){
	DBG( awt_fnt, ("Allegro fontpath not found: %s\n", path));
  }
  else {
	X->fsPath   = path;
	X->fsOffset  = get_config_int( "fonts", "offset", 3);
	X->fsBigger  = get_config_int( "fonts", "bigger", 0);
	X->fsSmaller = get_config_int( "fonts", "smaller", 4);
  }

  X->fsDefault = malloc( sizeof(MFONT));
  X->fsDefault->font   = font;
  X->fsDefault->height = font->height;
  X->fsDefault->ascent = 5;               /* who knows */
  X->fsDefault->maxAdvance = 8;
  X->fsDefault->leading = 1;
}


MFONT*
loadMFONT ( char* fname )
{
  MFONT              *mfont = 0;
  DATAFILE           *df;
  char               *prop;
  int                height, baseline, ascent, descent, maxAdvance, leading, proportional;

  if ( (df = load_datafile( fname)) ){
	if ( !(prop = get_datafile_property( df, DAT_ID( 'H','G','H','T'))) ) return 0;
	height = atoi( prop);

	if ( !(prop = get_datafile_property( df, DAT_ID( 'B','A','S','E'))) ) return 0;
	baseline = atoi( prop);

	if ( !(prop = get_datafile_property( df, DAT_ID( 'A','S','C','T'))) ) return 0;
	ascent = atoi( prop);

	if ( !(prop = get_datafile_property( df, DAT_ID( 'D','S','C','T'))) ) return 0;
	descent = atoi( prop);

	if ( !(prop = get_datafile_property( df, DAT_ID( 'M','A','D','V'))) ) return 0;
	maxAdvance = atoi( prop);

	if ( !(prop = get_datafile_property( df, DAT_ID( 'P','R','O','P'))) ) return 0;
	proportional = atoi( prop);

	if ( !(prop = get_datafile_property( df, DAT_ID( 'L','E','A','D'))) ) return 0;
	leading = atoi( prop);

	mfont = malloc( sizeof(MFONT));
	mfont->font     = (FONT*) df[0].dat;
	mfont->height   = height;
	mfont->baseline = baseline;
	mfont->ascent   = ascent;
	mfont->descent  = descent;
	mfont->maxAdvance = maxAdvance;
	mfont->leading = leading;
	mfont->proportional = proportional;
  }

  return mfont;
}

int
text_nlength ( MFONT* mfont, void* data, int len, int isWideChar )
{
  FONT_PROP *fp;
  char      *c;
  short     *wc;
  register  i, w;

  if ( mfont->proportional ){
	fp = mfont->font->dat.dat_prop;

	if ( isWideChar ) {
	  for ( i=0, w=0, c=data; i<len; i++ ){
		w += fp->dat[ c[i] ]->w;
	  }
	}
	else {
	  for ( i=0, w=0, wc=data; i<len; i++ ){
		w += fp->dat[ wc[i] ]->w;
	  }
	}
  }
  else {
	w = len * mfont->maxAdvance;
  }
  return w;
}


/*******************************************************************************
 * exported functions
 */

void*
Java_java_awt_Toolkit_fntInitFont ( JNIEnv* env, jclass clazz,
									jstring jSpec, jint style, jint size )
{
  MFONT       *mfont;
  char        buf[256], specBuf[128];
  char        *spec = java2CString( env, X, jSpec);
  char        *weight = "";
  char        *slant  = "";
  int         i, sz;

  if ( style & 0x1 ) weight = "b";
  if ( style & 0x2 ) slant  = "i";

  strcpy( specBuf, "%s/");
  strcat( specBuf, spec);

  size += X->fsOffset;

  sprintf( buf, specBuf, X->fsPath, size, weight, slant);
  mfont = loadMFONT( buf);

  if ( !mfont ){
	for ( i=-X->fsBigger; i<X->fsSmaller && !font; i++ ) {
	  if ( i != 0 ) {
		DBG( awt_fnt, ("font not found: <%s>, looking for size: %d\n", buf, size-i));
		sprintf( buf, specBuf, X->fsPath, size-i, weight, slant);
		mfont = loadMFONT( buf);
	  }
	}

	if ( !mfont ) {
	  DBG( awt_fnt, ("font not found: <%s>, backup to BIOS font\n", buf));
	  mfont = X->fsDefault;
	}
  }
  
  return (void*) mfont;
}

void
Java_java_awt_Toolkit_fntFreeFont ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  if ( fnt != X->fsDefault ) {
	destroy_font( fnt->font);
	free( fnt);
  }
}


/*******************************************************************************
 * FontMetrics support
 */

void*
Java_java_awt_Toolkit_fntInitFontMetrics ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt;
}

void
Java_java_awt_Toolkit_fntFreeFontMetrics ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
}


jint
Java_java_awt_Toolkit_fntGetAscent ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt->ascent;
}

jint
Java_java_awt_Toolkit_fntGetDescent ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt->descent;
}

jint
Java_java_awt_Toolkit_fntGetFixedWidth ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return (fnt->proportional) ? 0 : fnt->maxAdvance;
}

jint
Java_java_awt_Toolkit_fntGetHeight ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt->height;
}

jint
Java_java_awt_Toolkit_fntGetLeading ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt->leading;
}

jint
Java_java_awt_Toolkit_fntGetMaxAdvance ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt->maxAdvance;
}

jint
Java_java_awt_Toolkit_fntGetMaxAscent ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt->ascent;
}

jint
Java_java_awt_Toolkit_fntGetMaxDescent ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return fnt->descent;
}

jboolean
Java_java_awt_Toolkit_fntIsWideFont ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  return 0;  /* not supported by Allegro */
}

jobject
Java_java_awt_Toolkit_fntGetWidths ( JNIEnv* env, jclass clazz, MFONT* fnt )
{
  FONT_PROP *fp = fnt->font->dat.dat_prop;
  int       n = FONT_SIZE, i0=32;  /* fixed values in Allegro */
  jarray    widths = (*env)->NewIntArray( env, 256);
  jboolean  isCopy;
  jint      *jw = (*env)->GetIntArrayElements( env, widths, &isCopy);
  register  i, j;

  if ( fnt->proportional ) {
	for ( i=i0, j=0; j < n; i++, j++ )
	  jw[i] = fp->dat[j]->w;
  }
  else {
	for ( i=0; i < 256; i++ )
	  jw[i] = fnt->maxAdvance;
  }

  (*env)->ReleaseIntArrayElements( env, widths, jw, 0);

  return widths;
}


jint
Java_java_awt_Toolkit_fntBytesWidth ( JNIEnv* env, jclass clazz, MFONT* fnt,
									  jarray jBytes, jint off, jint len )
{
  jboolean  isCopy;
  jbyte    *jb = (*env)->GetByteArrayElements( env, jBytes, &isCopy);
  int       n = (*env)->GetArrayLength( env, jBytes);
  int       w;

  if ( off+len > n ) len = n - off;

  w = text_nlength( fnt, jb+off, len, 0);

  (*env)->ReleaseByteArrayElements( env, jBytes, jb, JNI_ABORT);
  return w;
}

jint
Java_java_awt_Toolkit_fntCharWidth ( JNIEnv* env, jclass clazz, MFONT* fnt, jchar jc )
{
  FONT_PROP *fp = fnt->font->dat.dat_prop;

  if ( fnt->proportional ) {
	return fp->dat[(unsigned char)jc - 32]->w;
  }
  else
	return fnt->maxAdvance;
}

jint
Java_java_awt_Toolkit_fntCharsWidth ( JNIEnv* env, jclass clazz, MFONT* fnt,
									  jarray jChars, jint off, jint len )
{
  jboolean  isCopy;
  jchar    *jc = (*env)->GetCharArrayElements( env, jChars, &isCopy);
  int      n = (*env)->GetArrayLength( env, jChars);
  int      w;

  if ( off+len > n ) len = n - off;

  w = text_nlength( fnt, (jchar*)(jc+off), len, 1);

  (*env)->ReleaseCharArrayElements( env, jChars, jc, JNI_ABORT);
  return w;
}

jint
Java_java_awt_Toolkit_fntStringWidth ( JNIEnv* env, jclass clazz, MFONT* fnt, jstring jStr )
{
  jboolean isCopy;
  const jchar    *jc = (*env)->GetStringChars( env, jStr, &isCopy);
  int      len = (*env)->GetStringLength( env, jStr);
  int      w, n;

  w = text_nlength( fnt, (void*)jc, len, 1);

  (*env)->ReleaseStringChars( env, jStr, jc);
  return w;
}
