/**
 * flash.c - MacroMedia flash interface
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

#include "flash.h"

/* borrowed from kaffevm/locks.h */
extern jboolean _waitCond ( void* obj, jlong timeout );
extern void     _signalCond ( void* obj );
extern void     _broadcastCond ( void* obj ); 
#define	waitCond(O,T)			_waitCond(&(O)->lock, (T))
#define	signalCond(O)			_signalCond(&(O)->lock)
#define broadcastCond(O)        _broadcastCond(&(O)->lock)


#define MAX_WAIT   5000
#define QUEUE_SIZE 16

typedef struct _Flash {
	jobject               tlkLock;
	jobject               player;
	jmethodID             callback;
	JNIEnv                *env;

	Image                 *background;

	struct {
		unsigned transparent  :1;
		unsigned sound        :1;
	} attrs;

	int                   x, y, width, height;

	char                  *fileSource;
	char                  *sourceBase;

	Image                 *canvas;
	Graphics              *gr;

	struct {
		unsigned loaded       :1;
		unsigned windowed     :1;
		unsigned running      :1;
		unsigned stop         :1;
		unsigned drawAll      :1;
		unsigned suspended    :1;    
		unsigned rewind       :1;
		unsigned waitForInput :1;
	} state;

	FlashDisplay          fd;
	struct FlashInfo      fi;
	FlashHandle           fh;

	FlashEvent            *queue;
	int                   head, next;
} Flash;



/***********************************************************************************
 * internal helpers
 */

static inline int
lastIndexOf ( char* s, char c ) 
{
	char *p;
	int i, n = -1;

	for ( i=0, p=s; *p; i++, p++ ) {
		if ( *p == c ){
			n = i;
		}
	}

	return n;
}

static char*
getSourceBase ( char* filename )
{
	char *sb = 0;
	int  n = lastIndexOf( filename, '/');

	if ( n >= 0 ) {
		sb = AWT_MALLOC( n+1);
		strncpy( sb, filename, n+1);
	}

	return sb;
}

static int
readFile ( char *filename, char **buffer, long *size )
{
	FILE *in;
	char *buf;
	long length;
 
	in = fopen( filename,"r");
	if ( in == 0 ) {
		return -1;
	}

	fseek( in,0,SEEK_END);
	length = ftell( in);
	rewind( in);
	buf = AWT_MALLOC( length);
	fread( buf,length,1,in);
	fclose( in);
	
	*size = length;
	*buffer = buf;
 
	return length;
} 

static void
showUrl ( char *url, char *target, void *clientData )
{
	Flash *f = (Flash*) clientData;
	JNIEnv *env = f->env;
	jstring sCmd, sArg;

	DBG( awt_fla, ("showUrl [Flash: %p, url:%s, target:%s]\n", f, url, target));

	if ( env && f->callback ) {
		sCmd = (*env)->NewStringUTF( env, url);
		sArg = (*env)->NewStringUTF( env, target);
		(*env)->CallObjectMethod( env, f->player, f->callback, sCmd, sArg);
	}
}

static void
getSwf ( char *url, int level, void *clientData )
{
	Flash *f = (Flash*) clientData;
	char *buffer = 0;
	long size;
	char *src = url;

	DBG( awt_fla, ("getSwf [Flash: %p, url:%s, level:%d base: %s]\n",
				   f, url, level, f->sourceBase));

	if ( (lastIndexOf( url, '/') < 0) && f->sourceBase){
		src = AWT_ALLOCA( 128);
		strcpy( src, f->sourceBase);
		strcat( src, url);
	}

	if ( readFile( src, &buffer, &size) > 0 ) {
		FlashParse( f->fh, level, buffer, size);
		AWT_FREE( buffer);
	}

	if ( src != url ){
		AWT_ALLOCA_FREE( src);
	}
}


static void
initImage ( Flash* f, Image* img )
{
	Graphics      *gr = f->gr;
	fgl_context_t gc;
	
	fgl_setBmp( &gc, img->bmp);
	fgl_setAlpha( &gc, 0xff);

	fgl_hideCursor();
	fgl_bitBlt( &gc,
				0, 0,
				gr->gc.bmp,
			    f->x + _(gr)->x0, f->y + _(gr)->y0,
			    f->x + _(gr)->x0 + f->width -1,
			    f->y + _(gr)->y0 + f->height -1);
	fgl_showCursor();
}


static void
saveBackground ( Flash* f )
{
	Image         *oldBmp = 0;

	if ( f->background ) {
		oldBmp = f->background;
	}

	DBG( awt_fla,("saveBackground [Flash: %p, old:%p]\n", f, oldBmp));

	f->background = createImage( f->width, f->height, JNI_TRUE);
	initImage( f, f->background);

	FlashSetBackground( f->fh, f->background->bmp->surface);

	if ( oldBmp ) {
		/* don't do that before the new one was set */
		Java_java_awt_Toolkit_imgFreeImage( f->env, 0, oldBmp);
	}

	DBG( awt_fla,("  -> saveBackground: %p\n", f->background));
}

static void
restoreBackground ( Flash* f, int dispose )
{
	if ( f->background ){

		DBG( awt_fla,("restoreBackground [Flash: %p, img:%p, disp:%d]\n", 
					  f, f->background, dispose));

		Java_java_awt_Toolkit_graDrawImage( f->env, 0, f->gr, f->background,
											0, 0,
											f->x, f->y,
											f->width, f->height, -1);
		
		if ( dispose ) {
			Java_java_awt_Toolkit_imgFreeImage( f->env, 0, f->background);
			f->background = 0;
			FlashSetBackground( f->fh, 0);
		}
	}
}

static void
drawImage ( JNIEnv* env, Flash* f )
{
	int x, y, w, h;

	if ( f->state.drawAll ) {
		Java_java_awt_Toolkit_graDrawImage( env, 0, f->gr, f->canvas,
										   0, 0,
										   f->x, f->y,
										   f->width, f->height, -1);
		f->state.drawAll = 0;
	}
	else {
 #ifdef ROT_90
		// flashlib doesn't know anything about rotation
		x = f->fd.clip_y;
		y = f->height - f->fd.clip_x - f->fd.clip_width;
		w = f->fd.clip_height;
		h = f->fd.clip_width;					
#else
		x = f->fd.clip_x;
		y = f->fd.clip_y;
		w = f->fd.clip_width;
		h = f->fd.clip_height;
#endif
		Java_java_awt_Toolkit_graDrawImage( env, 0, f->gr, f->canvas, x,y, x,y,w,h, -1);
	}
}


/* these are the handled keys */
#define VK_ENTER		 '\n'
#define VK_TAB           '\t'
#define VK_LEFT		     0x25
#define VK_UP		     0x26
#define VK_RIGHT		 0x27
#define VK_DOWN		     0x28

/* these are the relevant Java event types */
#define    KEY_PRESSED          401
#define    KEY_RELEASED         402

#define    MOUSE_PRESSED        501
#define    MOUSE_RELEASED       502
#define    MOUSE_MOVED          503
#define    MOUSE_ENTERED        504
#define    MOUSE_EXITED         505
#define    MOUSE_DRAGGED        506

static inline int
translateEvent ( Flash* f, int* type, int* x, int* y, int* key )
{
	if ( *type > 500 ){
		*x -= f->x;
		*y -= f->y;
	}
	else {
		switch ( *key ) {
		case VK_ENTER:   *key = FeKeyEnter;        break;
		case VK_TAB:     *key = FeKeyNext;         break;
		case VK_LEFT:    *key = FeKeyLeft;         break;
		case VK_UP:      *key = FeKeyUp;           break;
		case VK_RIGHT:   *key = FeKeyRight;        break;
		case VK_DOWN:    *key = FeKeyDown;         break;
		}
	}

	switch ( *type ) {
	case KEY_PRESSED:     *type = FeKeyPress;      break;
	case KEY_RELEASED:    *type = FeKeyRelease;    break;
	case MOUSE_MOVED:
	case MOUSE_DRAGGED:   /* Hmm, not supported by flashlib, yet */
		                  *type = FeMouseMove;     break;
	case MOUSE_PRESSED:   *type = FeButtonPress;   break;
	case MOUSE_RELEASED:  *type = FeButtonRelease; break;
	default:              return 0;
	}

	return 1;
}


/***********************************************************************************
 * exported APIs
 */

void*
Java_kaffe_awt_player_FlashPlayer_flaInit ( JNIEnv* env, jobject player, jobject tlkLock )
{
	Flash *f = AWT_CALLOC( 1, sizeof(Flash));
	jclass clazz;
	
	DBG( awt_fla,("flaInit [player: %p] -> Flash: %p\n", player, f));

	f->player  = (*env)->NewGlobalRef( env, player);
	f->tlkLock = (*env)->NewGlobalRef( env, tlkLock);

	clazz    = (*env)->GetObjectClass( env, player);
	f->callback = (*env)->GetMethodID( env, clazz,
									   "flashCallback",
									   "(Ljava/lang/String;Ljava/lang/String;)V");

	f->fh = FlashNew();

	f->queue = AWT_MALLOC( sizeof(FlashEvent) * QUEUE_SIZE);
	f->head  = -1;
	f->next  = -1;

	FlashSetGetUrlMethod( f->fh, showUrl, f);
	FlashSetGetSwfMethod( f->fh, getSwf, f);

	return f;
}

void
Java_kaffe_awt_player_FlashPlayer_flaInitSound ( JNIEnv* env, jobject player, Flash* f )
{
	DBG( awt_fla,("flaInitSound [player: %p flash: %p]\n", player, f));
	FlashSoundInit( f->fh, "/dev/dsp");
}

void
Java_kaffe_awt_player_FlashPlayer_flaSetGraphics ( JNIEnv* env, jobject player, Flash* f,
											Graphics* gr )
{
	DBG( awt_fla,("flaSetGraphics [player: %p flash: %p gr: %p]\n", player, f, gr));
	f->gr = gr;
}

jint
Java_kaffe_awt_player_FlashPlayer_flaSetWindow ( JNIEnv* env, jobject player,
										  Flash* f, int x, int y, int width, int height,
										  jboolean isTransparent )
{
	/* we just create the buffer image and set the 'fd' device params, all
	 * other action has to wait until we have a parsed source (flashlib doesn't
	 * like to call FlashGraphicInit before FlashParse is done)
	 */

	DBG( awt_fla,("flaSetWindow [Flash: %p  x: %d, y: %d, w: %d, h: %d]\n",
				  f, x, y, width, height));

	if ( f->canvas ) {
		Java_java_awt_Toolkit_imgFreeImage( env, 0, f->canvas);
	}

	f->x = x;
	f->y = y;
	f->width  = width;
	f->height = height;

	f->canvas = createImage( width, height, JNI_TRUE);

	f->fd.pixels   = fgl_getSlc( f->canvas->bmp, 0, 0);
	f->fd.width    = width;
	f->fd.height   = height;
#ifdef ROT_90
	f->fd.bpl      = fgl_getScanlineLength( f->canvas->bmp->height) * fgl_getSlcSize();
#else
	f->fd.bpl      = fgl_getScanlineLength( f->canvas->bmp->width) * fgl_getSlcSize();
#endif
	f->fd.depth    = fgl_getBitsPerPixel() / 8;
	f->fd.bpp      = f->fd.depth;

	f->attrs.transparent = isTransparent;

	f->state.windowed = 1;

	if ( f->state.loaded ) {
		/* re-initialize the flash graphics */
		FlashGraphicInit( f->fh, &f->fd);
	}

	return 1;
}

jint
Java_kaffe_awt_player_FlashPlayer_flaSetFileSource ( JNIEnv* env, jobject player,
											  Flash* f, jstring fileName )
{
	jbyte *fname;
	
	fname = (jbyte*)(*env)->GetStringUTFChars(env, fileName, 0);

	DBG( awt_fla,("flaSetFileSource [Flash: %p, file: %s]\n", f, fname));

	/* we don't do much here, the real action has to be postponed until
	 * we start to play the source (in its own thread)
	 */
	f->fileSource = fname;

	/* keep that around for second level swfs */
	f->sourceBase = getSourceBase( f->fileSource);

	return 1;
}

void
Java_kaffe_awt_player_FlashPlayer_flaSetDataSource ( JNIEnv* env, jobject player, Flash* f,
											  jarray jbuffer, jint off, jint len )
{

}

jint
Java_kaffe_awt_player_FlashPlayer_flaLoadSource ( JNIEnv* env, jobject player, Flash* f )
{
	int   ret = 0;
	char  *buffer;
	long  size;
	int   status;

	/* This is a flashlib pitfall (we eventually might fix) - FlashGraphicInit
	 * can't be called before FlashParse is done, i.e. we have to hard synchronize
	 * them. But defering the FlashParse up to this point is apropriate anyway, 
	 * because by now we should be running in our own thread (load might be lengthy)
	 */

	DBG( awt_fla,("flaLoadSource [Flash: %p]\n", f));

	if ( f->fileSource ) {
		if ( readFile( f->fileSource, &buffer, &size) > 0 ) {
			do {
				status = FlashParse( f->fh, 0, buffer, size);
			} while ( status & FLASH_PARSE_NEED_DATA );
		
			AWT_FREE( buffer);
		
			if ( status != FLASH_PARSE_ERROR ){
				FlashGetInfo( f->fh, &f->fi);
				
				DBG( awt_fla,("    rate: %d, count: %d, width: %d, height: %d\n",
							  f->fi.frameRate, f->fi.frameCount, f->fi.frameWidth, f->fi.frameHeight));
				ret = 1;
			}

			f->state.loaded = 1;

			if ( f->state.windowed ) {
				ret = FlashGraphicInit( f->fh, &f->fd);
			}
		}
	}

	DBG( awt_fla,("   -> flaLoadSource ret: %d\n", ret));

	return ret;
}

jint
Java_kaffe_awt_player_FlashPlayer_flaGetWidth ( JNIEnv* env, jobject player, Flash* f )
{
	int w = (f && f->state.loaded) ? (f->fi.frameWidth / 20) : 0; 

	DBG( awt_fla,("flaGetWidth [player: %p flash: %p] -> %d\n", player, f, w));
	return w;
}

jint
Java_kaffe_awt_player_FlashPlayer_flaGetHeight ( JNIEnv* env, jobject player, Flash* f )
{
	int h = (f && f->state.loaded) ? (f->fi.frameHeight / 20) : 0;

	DBG( awt_fla,("flaGetWidth [player: %p flash: %p] -> %d\n", player, f, h));
	return h;
}

void
Java_kaffe_awt_player_FlashPlayer_flaInputEvent ( JNIEnv* env, Hjava_lang_Object* player, Flash* f,
										   jint type, jint x, jint y, jint key )
{
	FlashEvent *fe;

	if ( !translateEvent( f, &type, &x, &y, &key) ){
		return;
	}

	/* solicitate equal event types */
	if ( type != (f->queue[f->head]).type ){
		f->head++;
		f->head %= QUEUE_SIZE;

		if ( f->head == f->next ){
			/* queue overrun, we loose one */
			f->next++;
			f->next %= QUEUE_SIZE;
		}
	}

	if ( f->next < 0 ){
		f->next = f->head;
	}

	fe = &(f->queue[f->head]);
	fe->type = type;
#ifdef ROT_90
	fe->x    = f->height - y;
	fe->y    = x;
#else
	fe->x    = x;
	fe->y    = y;
#endif
	fe->key  = key;

	/* wake up in case flaPlay waits for input */
	if ( f->state.waitForInput ) {
		f->state.waitForInput = 0;
		broadcastCond( player);
	}

	DBG( awt_fla,("flaInputEvent [Flash: %p, type: %d, x: %d, y: %d, key: %d,  queue: %d,%d]\n",
				  f, type, x, y, key, f->head, f->next));
}


void
Java_kaffe_awt_player_FlashPlayer_flaDrawAll ( JNIEnv* env, Hjava_lang_Object* player, Flash* f )
{
	DBG( awt_fla,("flaDrawAll [Flash: %p]\n", f));

	f->state.drawAll = 1;
	
	/* wake it up in case the runloop is either suspended or sleeping */
	broadcastCond( player);
}

jint
Java_kaffe_awt_player_FlashPlayer_flaPlay ( JNIEnv* env, Hjava_lang_Object* player, Flash* f )
{
	FlashEvent *fe;
	struct timeval wd, now, ed;
	int    wakeUp = 1;
	int    delay = 0;

	f->state.stop = 0;
	f->state.running = 1;
	f->env     = env;  /* watch out - has to be the thread from which we do callbacks */

	DBG( awt_fla,("flaPlay started [Flash: %p]\n", f));

	/* by now, we should be on the screen - do we have to save our background? */
	if ( f->attrs.transparent ){
		saveBackground( f);
	}
	initImage( f, f->canvas);

	/* get back to square 1, we might replay the same script */
	FlashExec( f->fh, FLASH_REWIND, 0, &wd);
	FlashExec( f->fh, FLASH_CONT, 0, &wd);

	while ( !f->state.stop ) {
		while ( f->state.suspended ) {
			DBG( awt_fla, ("..suspended %p\n", f));
			broadcastCond( player);

			waitCond( player, 0);
			wakeUp = 1;
			DBG( awt_fla, ("..resumed %p\n", f));
		}
		if ( f->state.stop ) {
			DBG( awt_fla, ("..stopping %p\n", f));
			break;
		}
		if ( f->state.rewind ) {
			DBG( awt_fla,("..rewind\n"));
			FlashExec( f->fh, FLASH_REWIND, 0, &wd);
			wakeUp = FlashExec( f->fh, FLASH_CONT, 0, &wd);
			f->state.rewind = 0;
		}

		if ( wakeUp ) {
			/* prepare the next frame */
			wakeUp = FlashExec( f->fh, FLASH_WAKEUP, 0, &wd);

			if ( f->fd.flash_refresh ) {
				//(*env)->MonitorEnter( env, f->tlkLock);
				drawImage( env, f);
				f->fd.flash_refresh = 0;
				//(*env)->MonitorExit( env, f->tlkLock);
			}
		}

		if ( f->next >= 0 ) {
			/* we have input events, pass them down */
			fe = &(f->queue[f->next]);
		
			if ( f->next == f->head ){
				f->next = -1;
			}
			else {
				f->next++;
				f->next %= QUEUE_SIZE;
			}

			/* send the input event down and then unconditionally wake, since we
			 * might have a rendering response to it (which is NOT returned
			 * by a FLASH_EVENT exec
			 */
			FlashExec( f->fh, FLASH_EVENT, fe, &ed);

 			if ( f->fd.flash_refresh ) {
				drawImage( env, f);
			}
				
			wakeUp = 1;
		}

		if ( wakeUp ) {
			gettimeofday( &now,0);
			delay = (wd.tv_sec-now.tv_sec)*1000 + (wd.tv_usec-now.tv_usec)/1000;

			if ( delay < 20 ) {
				delay = 20;
			}

			/* Hmmm, this is a workaround for the anomaly that jthread wakep latency
			 * seems to be much higher than the specified sleep time
			 */
			if ( delay < 90 ) {
				waitCond( player, (jlong)delay/3);
				//usleep( delay);
			}
			else {
				waitCond( player, (jlong)delay/2);
			}
		}
		else {
			if ( f->state.drawAll ) {
				Java_java_awt_Toolkit_graDrawImage( env, 0, f->gr, f->canvas,
												   0, 0,
												   f->x, f->y,
												   f->width, f->height, -1);
				f->state.drawAll = 0;
			}
			
			/* Hmm, no event, no wakeUp, we should suspend ourself for a while to
			 * avoid spinning at full speed
			 */
			f->state.waitForInput = 1;
			waitCond( player, MAX_WAIT);
		}
	}

	if ( f->attrs.transparent ){
		restoreBackground( f, JNI_TRUE);
	}

	if ( f->state.stop ) {
		DBG( awt_fla,("..stopped\n"));
		FlashExec( f->fh, FLASH_STOP, 0, &wd);
		f->state.stop = 0;
	}

	f->env = 0;
	f->state.running = 0;
	broadcastCond( player);

	DBG( awt_fla,("flaPlay finished [Flash: %p] stop: %d wakeUp: %d\n",
				  f, f->state.stop, wakeUp));

	return 1;
}


jboolean
Java_kaffe_awt_player_FlashPlayer_flaStop ( JNIEnv* env, Hjava_lang_Object* player, Flash* f )
{
	jboolean ret = JNI_TRUE;

	DBG( awt_fla,("flaStop [Flash: %p] request\n", f));

	f->state.stop = 1;

	if ( f->state.suspended ) {
		f->state.suspended = 0;
	}

	/* wake it up in case the runloop is either suspended or sleeping */
	broadcastCond( player);
	
	if ( f->state.running ) {
		ret = waitCond( player, MAX_WAIT);
	}

	DBG( awt_fla,("  -> stopped [Flash: %p] %d\n", f, ret));
	return ret;
}



void
Java_kaffe_awt_player_FlashPlayer_flaRewind ( JNIEnv* env, jobject player, Flash* f )
{
	struct timeval wd;

	DBG( awt_fla,("flaRewind [Flash: %p]\n", f));

	if ( f->state.running ) {
		f->state.rewind = 1;
	}
	else {
		FlashExec( f->fh, FLASH_REWIND, 0, &wd);
	}
}

jboolean
Java_kaffe_awt_player_FlashPlayer_flaSuspend ( JNIEnv* env, Hjava_lang_Object* player, Flash* f )
{
	jboolean ret = JNI_TRUE;

	DBG( awt_fla,("flaSuspend [Flash: %p]\n", f));

	if ( f->state.running && !f->state.suspended ){
		f->state.suspended = 1;

		/* wake it up in case the runloop is either suspended or sleeping */
		broadcastCond( player);
		
		ret = waitCond( player, MAX_WAIT);
	}

	DBG( awt_fla,("  -> suspended [Flash: %p] %d\n", f, ret));
	return ret;
}

void
Java_kaffe_awt_player_FlashPlayer_flaResume ( JNIEnv* env, Hjava_lang_Object* player, Flash* f )
{
	DBG( awt_fla,("flaResume [Flash: %p]\n", f));

	f->state.suspended = 0;
	broadcastCond( player);
}

void
Java_kaffe_awt_player_FlashPlayer_flaBlank ( JNIEnv* env, jobject player, Flash* f )
{
	DBG( awt_fla,("flaBlank [Flash: %p]\n", f));

	restoreBackground( f, JNI_FALSE);
}

void
Java_kaffe_awt_player_FlashPlayer_flaDispose ( JNIEnv* env, jobject player, Flash* f )
{
	DBG( awt_fla,("flaDispose [Flash: %p]\n", f));

	if ( !f ) {
		return;
	}

	if ( f->canvas ){
		Java_java_awt_Toolkit_imgFreeImage( env, 0, f->canvas);
		f->canvas = 0;
	}

	if ( f->background ) {
		Java_java_awt_Toolkit_imgFreeImage( env, 0, f->background);
		f->background = 0;
	}

	if ( f->tlkLock ){
		(*env)->DeleteGlobalRef( env, f->tlkLock);
		f->tlkLock = 0;
	}

	if ( f->player ){
		(*env)->DeleteGlobalRef( env, player);
		f->player = 0;
	}

	if ( f->queue ){
		AWT_FREE( f->queue);
		f->queue = 0;
	}

	if ( f->fileSource ){
		AWT_FREE( f->fileSource);
		f->fileSource = 0;
	}

	if ( f->sourceBase ){
		AWT_FREE( f->sourceBase);
		f->sourceBase = 0;
	}

	FlashClose( f->fh);

	AWT_FREE( f);
}
