/**
 * fnt.c - font support
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#ifndef DEBUG
#define FGL_INLINE_FONT
#endif
#include "fgl.h"

#include "toolkit.h"


/*******************************************************************************
 * internal functions
 */

void
initFonts ( Toolkit* X )
{
	/* nothing to do */
}



/*******************************************************************************
 * exported functions
 */

void*
EXPORT(fntInitFont) ( JNIEnv* env, jclass clazz,
					  jstring jSpec, jint style, jint size )
{
	fgl_font_t *fnt;
	char *spec = java2CString( env, _(X), jSpec);
	char *weight = "";
	char *slant  = "";
	char buf[64];

	/* this is pretty stupid, we first mangle attributes in, just to have
	 * them subsequently extracted in fgl, again
	 */
	if ( style & 0x2 ) {
		if ( style & 0x01 )
			weight = "b";
		slant  = "i";
	}
	else if ( style & 0x01 ){
		weight = "bd";
	}
	sprintf( buf, spec, weight, slant);

	DBG( awt_fnt, ("initFont: %s,%d,%d -> %s\n", spec, style, size, buf));

	fnt = fgl_getFont( buf, size);

	DBG( awt_fnt, (" -> font: %x\n", fnt));
  
	return (void*) fnt;
}

void
EXPORT(fntFreeFont) ( JNIEnv* env, jclass clazz, g_font_t* fnt )
{
	DBG( awt_fnt, ("freeFont: %x\n", fnt));

	/* this just decrements the nRefs counter, and only frees the font if there
	 * is no other corresponding Font/FontMetrics object around
	 */
	fgl_disposeFont( fnt);
}


/*******************************************************************************
 * FontMetrics support
 */
void*
EXPORT(fntInitFontMetrics) ( JNIEnv* env, jclass clazz,
							 fgl_font_t *fnt, jobject metrics )
{
	jclass fmClazz = (*env)->GetObjectClass( env, metrics);
	jfieldID id;


	DBG( awt_fnt, ("initFontMetrics: %x\n", fnt));

	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAdvance", "I")) ){
		(*env)->SetIntField( env, metrics, id, fgl_getMaxAdvance( fnt));
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "leading", "I")) ){
		(*env)->SetIntField( env, metrics, id, 0);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "ascent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->ascend);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAscent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->ascend);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "descent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->descend);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxDescent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->descend);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "height", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->fontHeight);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "fixedWidth", "I")) ){
		(*env)->SetIntField( env, metrics, id, fgl_getFixedWidth( fnt));
	}
#if defined(LOCALE_INT)
	if ( (id = (*env)->GetFieldID( env, fmClazz, "isWideFont", "Z")) ){
		(*env)->SetBooleanField( env, metrics, id, JNI_TRUE);
	}
#endif /* LOCALE_INT */
	DBG( awt_fnt, (" -> metrics: %x\n", fnt));

	/* make sure a subsequent fntFreeFont doesn't kill our font reference */
	fgl_addFontReference( fnt);
	
	return fnt;
}

void
EXPORT(fntFreeFontMetrics) ( JNIEnv* env, jclass clazz, g_font_t* fnt )
{
	DBG( awt_fnt, ("initFontMetrics: %x\n", fnt));
	
	/* this just decrements the nRefs counter, and only frees the font if there
	 * is no other corresponding Font/FontMetrics object around
	 */
	fgl_disposeFont( fnt);
}


jobject
EXPORT(fntGetWidths) ( JNIEnv* env, jclass clazz, g_font_t* fnt )
{
	int          n = 256;
	jarray       widths = (*env)->NewIntArray( env, 256);
	jboolean     isCopy;
	jint         *jw = (*env)->GetIntArrayElements( env, widths, &isCopy);
	register int i;

	DBG( awt_fnt, ("getWidths: %x\n", fnt));

	for ( i=0; i < fnt->nGlyphs ; i++ ){
		jw[i] = fnt->advance[i];
	}
	for ( ; i < n ; i++ ){
		jw[i] = fnt->advance[0];
	}


	(*env)->ReleaseIntArrayElements( env, widths, jw, 0);

	return widths;
}


jint
EXPORT(fntBytesWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt,
						jarray jBytes, jint off, jint len )
{
	jboolean  isCopy;
	jbyte     *jb = (*env)->GetByteArrayElements( env, jBytes, &isCopy);
	int       n = (*env)->GetArrayLength( env, jBytes);
	int       w;

	if ( off+len > n )
		len = n - off;

	w = fgl_getTextLength( fnt, (jb+off), len);

	(*env)->ReleaseByteArrayElements( env, jBytes, jb, JNI_ABORT);

	DBG( awt_fnt, ("bytesWidth: %x \"%3s..\" -> %d\n", fnt, jb+off, w));

	return w;
}

jint
EXPORT(fntCharWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt, jchar jChar )
{
	int i = jChar - fnt->uniOffset;
	jint w = 0;

#if defined(LOCALE_INT)

	int j;
	int match = 0;

	if ( i < fnt->nGlyphs ) {
		w = fnt->advance[i];
		match = 1;
	}
	else { 
		for ( j = 0 ; j < fnt->ext_number; j ++) {
			if ( i > fnt->ext[j]->uniOffset && i <= fnt->ext[j]->uniLast) {
				w = fnt->ext[j]->advance[ i- fnt->ext[j]->uniOffset ];
				match = 1;
			}
		}
	}
	if (!match) w = fnt->advance[0];
#else
	if ( i >= fnt->nGlyphs ) i = 0;
	w = fnt->advance[i];
#endif
	DBG( awt_fnt, ("charWidth: %x \'%c\' -> %d\n", fnt, jChar, w));

	return w;
}

jint
EXPORT(fntCharsWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt,
						jarray jChars, jint off, jint len )
{
	jboolean  isCopy;
	jchar    *jc = (*env)->GetCharArrayElements( env, jChars, &isCopy);
	int      n = (*env)->GetArrayLength( env, jChars);
	int      w;

	if ( off+len > n )
		len = n - off;

    w = fgl_getWTextLength( fnt, jc+off, len);

	(*env)->ReleaseCharArrayElements( env, jChars, jc, JNI_ABORT);

	DBG( awt_fnt, ("charWidths: %x \"%s\" -> %d\n", fnt,
				   jchar2CString( _(X),(jc+off),len), w));

	return w;
}

jint
EXPORT(fntStringWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt, jstring jStr )
{
	jboolean isCopy;
	jchar    *jc = (jchar*) (*env)->GetStringChars( env, jStr, &isCopy);
	int      len = (*env)->GetStringLength( env, jStr);
	int      w;

    w = fgl_getWTextLength( fnt, jc, len);

	(*env)->ReleaseStringChars( env, jStr, jc);

	DBG( awt_fnt, ("stringWidth: %x \"%s\" -> %d\n", fnt,
				   java2CString( env,_(X),jStr), w));

	return w;
}
