/**
 * grlib.h - target driver library interface layer for no-native-wm AWTs
 *           (fgl/ipl version)
 *
 * grlib.h contains most of AWT's "bottom interface" (aka DRIVER LIB
 * INTERFACE), which encapsulates the platform specific driver functions.
 *
 *  +----------------------------------------------------------------------+
 *  |Java class libraries                                                  |
 *  |                                                                      |
 *  | java.awt.Graphics (, Image, Window, Color, Font, ..)                 |
 *  |   drawLine() { Toolkit.graDrawLine() }                               |
 *  |                     |                                                |
 *  |  +----------------- | ---------------------------------------------+ |
 *  |  |java.awt.Toolkit  V                    [NATIVE AWT LIB INTERFACE]| |
 *  |  |  native graDrawLine(), wndCreateFrame(), clrGetPixelValue() ... | |
 *  |  +----------------- | -------------- | ---------- | ---------- | --+ |
 *  |                     |                |            |            |     |
 *  +-------------------- | -------------- | ---------- | ---------- | ----+
 *                        |                |            |            |
 *  +-------------------- | -------------- | ---------- | ---------- | ----+
 *  | native AWT library  |                V            V            V     |
 *  |              gra.c  V               wnd.c        clr.c        ...    |
 *  |                                                                      |
 *  |   Java_java_awt_Toolkit_graDrawLine(){ g_drawLine.. }                |
 *  |                                  +--------- | -----------------------+
 *  |                                  | grlib.h  V  [DRIVER LIB INTERFACE]|
 *  +----------------------------------+---- | ----------------------------+
 *                                           |
 *  +--------------------------------------- | ----------------------------+
 *  | driver library (GRX, MGL, UGL, ..)     V                             |
 *  |  (provides rasterizer)                                               |
 *  +----------------------------------------------------------------------+
 *
 * Unlike the "top interface" (NATIVE AWT LIB INTERFACE containing
 * native methods used from the unified AWT class library), it can be
 * implemented as a subset, i.e.there is no need to implement all it's
 * functions and types, if you don't use the relevant common module(s)
 * (like gra-common.c). All native methods are exported (and can be
 * overriden) from the platform specific ones (like gra.c).
 * 
 * This header is included from toolkit.h, which is in turn the glue header
 * for all native AWT modules.
 *
 * +-----------------------------------------------------------------------+
 * | gra.c (, wnd.c, img.c clr.c fnt.c..) (*)                              |
 * |  +------------------------------------------------------------------+ |
 * |  | toolkit.h (*)                                                    | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |  | toolkit-common.h                                            | | |
 * |  |  |   typedef struct CommonImage, CommonGraphics, CommonToolkit | | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |                                                                  | |
 * |  |  typedef struct Image, Graphics, Toolkit {CommonToolkit ct..}    | |
 * |  |                                                                  | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |  | grlib.h (*)                                                 | | |
 * |  |  |   g_setColor(){.}, g_drawLine(){ GrDrawLine() }, .       |  | | |
 * |  |  +--------------------------------------------------------- | -+ | |
 * |  +------------------------------------------------------------ | ---+ |
 * |                                                                |      |
 * |  // exported functions with generic implementation             |      |
 * |  #define _graDrawLine Java_java_awt_Toolkit_graDrawLine..      |      |
 * |                                                                |      |
 * |  +------------------------------------------------------------ | ---+ |
 * |  | gra-common.c (, wnd-common.c, img-common.c, clr-common.c..) |    | |
 * |  |  _graDrawLine() { .. g_drawLine().. }                       V    | |
 * |  +------------------------------------------------------------------+ |
 * |                                                                       |
 * |  // platform specific exported functions                              |
 * |  Java_java_awt_Toolkit_graDrawImageScaled() {..} ..                   |
 * |                                                                       |
 * +-----------------------------------------------------------------------+
 *
 * The (*) marked files are platform dependent
 *
 * Note that this header is included in two parts - the top half defining
 * just generic types (which can be used in the CommonXX structures, the
 * bottom half defining the driver interface functions (which are usually
 * inlined).
 *
 * Functions and types starting with a "g_" prefix (e.g. g_drawLine) are
 * used in the standard common modules (have a fixed meaning), all other
 * definitions are just internal helpers, and can only be used from the
 * platform specific modules (e.g. gra.c), or from within grlib.h itself.
 *
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#if !defined (GRLIB_FUNCS)
/* this section is inlcuded before toolkit-common.h, to define
 * generic types which can be used in the common structure definitions
 */

/***************************************************************************
 * scalar & reference types 
 */

/* the type of native pixel values */
typedef  fgl_color_t       g_pixel_t;

/* the type to set color palettes */
typedef  fgl_color_t       g_paletteEntry_t;

/* (optional) transparent graphics context used for context save/restore
 * operations (if the driver lib provides/requires them)
 */
typedef  void*             g_contextSav_t;

/* (transparent) font type */
typedef  fgl_font_t        g_font_t;

/* the 'point' type of polygon related operations */
typedef  fgl_point_t       g_point_t;

/* function to (type transparent) set point coordinates */
static inline void g_setPoint ( g_point_t* p, int x, int y )
{ 
	p->x = x; p->y = y;
}

static inline void g_setScreenRotation ( char rot )
{
#ifdef LIB_FGL_0_0
        fgl_setScreenRotation ( rot );
#endif
}
#else /* defined GRLIB_FUNCS */
/* this is included before toolkit-common-post.h, i.e. we can
 * use these functions/macros there
 */

/***************************************************************************
 * internal helper (can be called from platform specific modules, but
 * NOT from common/generic code)
 */


/******************************************************************************
 * visual detection
 */

/*
 * get the number of bits per display pixel (= visual)
 */
static inline unsigned int g_getBitsPerPixel ( Toolkit* X )
{
	return fgl_getBitsPerPixel();
}

/*
 * (optional) function to set system into RGB mode (some drivers
 * have to be switched explicitly)
 */
static inline void g_initRgbMode ( Toolkit* X )
{
	/* not required, all pre-determined */
}


/******************************************************************************
 * pseudocolor palette (colormap) access & manipulation
 *
 * These macros are used (from clr.c) to set the system palette (we
 * need to evenly distribute RGBs, with additional capabilities to
 * allocate a few exact color requests).
 *
 * Note that some systems require to 'allocate' system color cells, i.e.
 * don't give us direct r/w access to a certain system palette entry/index.
 */

/*
 * Initialize system palette modification (reset, allocation etc.)
 */
static inline void g_initPalette ( CommonIdxMap* map )
{
	/* not required, all done automatically at fgl_init() time */
}

/*
 * Set a system palette cell to the provided RGB components [0..255],
 * answer palette index where cell is stored (if the driver lib can't
 * change palette cells in-situ, and has to 'allocate' them).
 * 'idx' is the requested cell index (for in-situ changeable palettes)
 */
static inline int g_setPaletteEntry ( CommonIdxMap* map, int idx,
									  int r, int g, int b )
{
	/* not required, all done automatically at fgl_init() time */
	return 0;
}

/*
 * Query RGB components of palette cell at 'idx'
 */
static inline void g_queryPaletteEntry ( CommonIdxMap* map, int idx,
										 int* pr, int* pg, int* pb )
{
	/* watch out, we don't have a generic CommonIdxMap _(X)->map.. */
	fgl_cmap_t* cmap = fgl_getCmap();

	*pr = cmap->r[idx];
	*pg = cmap->g[idx];
	*pb = cmap->b[idx];
}

/*
 * Make (modified) palette active (some drivers refer to this as
 * 'realizing' the palette)
 */
static inline void g_setPalette ( CommonIdxMap* map )
{
	/* not required, all done automatically at fgl_init() time */
}


/*
 * add alpha channel [0..255] to jint pixel value
 */
static inline jint g_addAlpha ( jint rgb, jint pix )
{
	return ((rgb & 0xff000000) | pix);
}

/*
 * get alpha channel [0..255] from jint pixel value
 */
static inline unsigned char g_getAlpha ( jint pix )
{
	return (pix >> 24);
}

/*
 * remove alpha channel from jint pixel value
 */
static inline g_pixel_t g_stripAlpha ( jint pix )
{
	return (pix & 0x00ffffff);
}

/*
 * Answer pixel value for given 8-8-8-8 ARGB
 * (remove switch for speed optimization)
 */
static inline jint g_pixelValue ( Toolkit* X, jint rgb )
{
	jint pix = fgl_getPixelValue( rgb);
	
	return g_addAlpha( rgb, pix);
}


/*
 * Answer RGB components for given pixel value
 */
static inline void g_rgbValues ( Toolkit* X, unsigned long pixel, int* r, int* g, int* b )
{
	fgl_getRgbValues( pixel, r, g, b);
}


static inline jint g_modPixelValue ( Toolkit* X, int pix, int mod )
{
	switch ( mod ){
	case BRIGHTER: mod = 1; break;
	case DARKER: mod = -1; break;
	default: return pix;
	}

	return g_addAlpha( pix, fgl_modPixelValue( pix, mod));
}


/******************************************************************************
 * Graphics state - these functions are used to set the state of our
 * native cache 'Graphics' objects
 */

/*
 * Set global draw context according to the provided Image
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentImage ( Toolkit* X, Image* img )
{
	/* not required, gc always passed into primitives */
}

/*
 * Set global draw context according to the provided Graphics
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentGraphics ( Toolkit* X, Graphics* gr )
{
	/* not required, gc always passed into primitives */
}

/*
 * Temporarily store the current context (if the graphics driver
 * can't draw into aribtrary contexts)
 */
static inline void g_saveContext ( g_contextSav_t* cs )
{
	/* not required, gc always passed into primitives */
}

/*
 * Restore the current context from a temporarily saved one
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_restoreContext ( g_contextSav_t* cs )
{
	/* not required, gc always passed into primitives */
}

/*
 * Set the active clip rectangle according to the provided Graphics object
 */
static inline void g_setClip ( Graphics* gr )
{
	fgl_setClip( &(gr->gc),
				 _(gr)->curClip->x0, _(gr)->curClip->y0,
				 _(gr)->curClip->x1, _(gr)->curClip->y1);
}

/*
 * one time init for new Graphics objects
 */
static inline void g_initGraphics ( Graphics* gr )
{
	/* nothing to do */
}


/*
 * free resources for Graphics object
 */
static inline void g_freeGraphics ( Graphics* g )
{
	/* nothing to do */
}


/*
 * Set the draw context of a Graphics object 'gr' from a Window object
 */
static inline void g_setWindowGraphics ( Graphics* gr, AwtWnd* wnd )
{
	fgl_setBmp( &(gr->gc), fgl_screenBmp);
}

/*
 * Set the draw context of a Graphics object 'gr' from a Image object
 */
static inline void g_setImageGraphics ( Graphics* gr, Image* img )
{
	fgl_setBmp( &(gr->gc), img->bmp);
}

/*
 * Set the draw context of a Graphics object 'gr' from a Graphics object
 */
static inline void g_setGraphicsGraphics ( Graphics* gr, Graphics *ogr )
{
	fgl_setBmp( &(gr->gc), ogr->gc.bmp);
}

/*
 * Set the foreground color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setColor ( Graphics* gr, u32 clr )
{
#if defined (USE_FULL_ALPHA)
	fgl_setAlpha( &(gr->gc), g_getAlpha( clr));
#else
	fgl_setAlpha( &(gr->gc), 0xff);
#endif
	
	fgl_setColor( &(gr->gc), g_stripAlpha( clr));
}

/*
 * Answer foreground color of Graphics object
 */
static inline u32 g_getColor ( Graphics* gr )
{
#if defined (USE_FULL_ALPHA)
#if defined (LIB_FGL_0_0)
	return (gr->gc.alphar << 24) | gr->gc.clr;
#else
	return (gr->gc.alpha << 24) | gr->gc.clr;
#endif
#else
	return gr->gc.clr;
#endif
}

/*
 * Set the background color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 * This is a optional function for drivers which can't draw
 * text with transparent background
 */
static inline void g_setBackColor ( Graphics* gr, g_pixel_t clr )
{
	gr->bg = clr;
}

/*
 * Set the font of a Graphics object
 * Note that we also might have to set the active font in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setFont ( Graphics* gr, g_font_t* fnt )
{
	fgl_setFont( &(gr->gc), fnt);
}

/*
 * Set the draw mode of a Graphics object to 'overpaint'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setPaintMode ( Graphics* gr )
{
	fgl_setPaintMode( &(gr->gc));
}

/*
 * Set the draw mode of a Graphics object to 'xor'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setXORMode ( Graphics* gr, u32 xclr )
{
	fgl_setXorMode( &(gr->gc), xclr);
}


/******************************************************************************
 * drawing operations
 */

/*
 * Set color of pixel
 * (Make sure it is clipped according to the Graphics, since many
 * graphics drivers don't provide clipping for this operation)
 */
static inline void g_putPixel ( Graphics* gr, int x, int y)
{
	fgl_setPixel( &(gr->gc), x, y, gr->gc.clr);
}

/*
 * Fill rectangle with background color of Graphics object
 */
static inline void g_clearRect ( Graphics* gr, int x0, int y0, int x1, int y1)
{
	fgl_fillRectClr( &(gr->gc), x0, y0, x1, y1, gr->bg);
}

/*
 * Draw a line
 */
static inline void g_drawLine ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	fgl_drawLine( &(gr->gc), x0, y0, x1, y1);
}

static inline void g_drawHLine ( Graphics* gr, int x0, int y0, int x1 )
{
	fgl_drawHLine( &(gr->gc), x0, y0, x1);
}

static inline void g_drawVLine ( Graphics* gr, int x0, int y0, int y1 )
{
	fgl_drawVLine( &(gr->gc), x0, y0, y1);
}

/*
 * Draw a rectangle (corner exclusive)
 */
static inline void g_drawRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	fgl_drawRect( &(gr->gc), x0, y0, x1, y1);
}

/*
 * Fill a rectangle (corner inclusive)
 */
static inline void g_fillRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	fgl_fillRect( &(gr->gc), x0, y0, x1, y1);
}

/*
 * Draw open arc (i.e. no lines from arc endpoints to center of oval).
 * Note that this actually is a oval, not just a circle
 */
static inline void g_drawOpenArc ( Graphics* gr, int x, int y,
								   int ra, int rb, int a1, int a2 )
{
	/* not yet */
}


/*
 * Fill closed arc (pie chart)
 */
static inline void g_fillArc ( Graphics* gr, int x, int y,
							   int ra, int rb, int a1, int a2 )
{
	/* not yet */
}

/*
 * Draw oval (ellipse)
 */
static inline void g_drawOval ( Graphics* gr, int cx, int cy, int ra, int rb )
{
	fgl_drawEllipse( &(gr->gc), cx, cy, ra, rb);
}

/*
 * Fill oval (ellispe)
 */
static inline void g_fillOval ( Graphics* gr, int cx, int cy, int ra, int rb )
{
	fgl_fillEllipse( &(gr->gc), cx, cy, ra, rb);
}

/*
 * Draw single-byte text
 */
static inline void g_drawByteText ( Graphics* gr, char* str,
									int len, int x, int y )
{
	fgl_drawText( &(gr->gc), x, y, str, len);
}

/*
 * Draw unicode text
 */
static inline void g_drawWideText ( Graphics* gr, jchar* str,
									int len, int x, int y )
{
	fgl_drawWText( &(gr->gc), x, y, str, len);
}

/*
 * Draw unicode vertical text
 */
static inline void g_drawWideVertText ( Graphics* gr, jchar* str,
									int len, int x, int y )
{
	fgl_drawWVertText( &(gr->gc), x, y, str, len);
}

/*
 * Draw polygon (lines connecting vertices in 'points')
 */
static inline void g_drawPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	fgl_drawPolygon( &(gr->gc), points, nPoints);
}

/*
 * Fill polygon
 */
static inline void g_fillPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	fgl_fillPolygon( &(gr->gc), points, nPoints);
}


/******************************************************************************
 * image drawing (bitblt related functions)
 */

/*
 * Bit block transfer in same context (Image or display)
 */
static inline void g_copyArea ( Graphics* gr, int x, int y,
								int xw, int yh, int xdelta, int ydelta)
{
	fgl_moveArea( &(gr->gc), x, y, xw, yh, xdelta, ydelta);
}

/*
 * Bit block transfer from Image context source (to Image or display)
 * This operation has to handle alpha channels. In case of a non-negative
 * 'bg' color, the transparent image pixels have to be set to this value
 */
static inline void g_drawImage ( Graphics* gr, Image* img, int dx, int dy,
								 int sx, int sy, int sxw, int syh, g_pixel_t bg )
{
	if ( (bg != (g_pixel_t)-1) && (img->mask) ){
		fgl_fillRect( &(gr->gc), dx, dy, dx+(sxw-sx), dy+(syh-sy));
	}

	if ( img->alpha ){
		fgl_alphaMaskBlt( &(gr->gc), dx, dy, img->bmp, sx, sy, sxw, syh, img->alpha);
	}
	else if ( img->mask ) {
		fgl_maskBlt( &(gr->gc), dx, dy, img->bmp, sx, sy, sxw, syh, img->mask);
	}
	else {
		fgl_bitBlt( &(gr->gc), dx, dy, img->bmp, sx, sy, sxw, syh);
	}
}

/*
 * grab pixels from a source Graphics, and punch them into a destination
 * Graphics (this is a generalized blit with a source that is not a
 * Java image)
 */
static inline void g_grabPixels ( Graphics* dgr, Graphics* sgr,
								  int sx, int sy, int w, int h )
{
	fgl_bitBlt( &(dgr->gc),
			    _(dgr)->x0, _(dgr)->y0,
			    sgr->gc.bmp,
			    sx + _(sgr)->x0, sy + _(sgr)->y0,
			    sx + _(sgr)->x0 + w -1,
			    sy + _(sgr)->y0 + h -1);
}


/******************************************************************************
 * image processing
 *
 * The protocol for creating, initializing and rendering native images
 * has been adapted to support subsequent alpha channel processing (which
 * can be extended to offscreen images). The involved steps are:
 *
 *  g_initImage      Prepares Image object so that native pixel values
 *                   can be (permanently or temporarily) stored
 *
 *  g_setAlphaType+  Has to be called as soon as information about
 *                   alpha channels (none, masked, full) is available. 
 *                   After the call, (temporary or permanent) alpha
 *                   values can be stored for each pixel. Note that
 *                   it is possible to subsequently "increase" the
 *                   alpha type (none->masked->full, i.e. the function
 *                   can be called several times)
 *
 * g_putImagePixel+  set color, alpha information for a pixel
 *
 * g_completeImage   check if alpha info can be reduced (full->masked->none)
 *                   and transform any temporary (generic) information
 *                   into a platform specific format which is suitable
 *                   for rendering
 */

/*
 * Initialize a Image object so that we can set it's pixel values.
 * Note that we don't have alpha information, yet (i.e. the image
 * subsequently might be set to masked or full alpha)
 */
static inline int g_initImage ( Image* img )
{
	img->bmp = fgl_createBitmap( _(img)->width, _(img)->height, -1);
	return (img->bmp != 0);
}


/*
 * Set alpha channel type of Image object. The supported 'transType'
 * values are NO_ALPHA, MASKED_ALPHA (transparent pixels), FULL_ALPHA.
 * After this operation, we have to be able to store pixel alpha
 * information (at least temporarily, to be later-on processed by
 * completeImage)
 */
static inline void g_setAlphaType ( Image* img, int alphaType )
{
	if ( alphaType && ! (img->mask || img->alpha) ){
		if ( alphaType == MASK_ALPHA ){
			img->mask = fgl_createMonoBitmap( _(img)->width, _(img)->height, 0xff);
		}
		else if ( alphaType == FULL_ALPHA ){
#if defined(USE_FULL_ALPHA)
			img->alpha = fgl_createA8Bitmap( _(img)->width, _(img)->height, 0xff);
#else
			img->mask = fgl_createMonoBitmap( _(img)->width, _(img)->height, 0xff);
#endif
		}
	}
}

/*
 * set transparency color of the image in the native layer
 */
static inline void g_setTranspValue ( Image* img, g_pixel_t clr )
{
  img->bmp->transp = clr;
}

/*
 * Transform temporarily stored image information (both pixel color and
 * alpha values) into a representation which is suitable for drawing
 * the image. Free any temporary information.
 */
static inline void g_completeImage ( Image* img )
{
	/* nope, we don't reduce (if we have full alpha) */
}

/*
 * Clean up platform specific Image components
 */
static inline void g_freeImage ( Image* img )
{
	if ( img->bmp ){
		fgl_disposeBitmap( img->bmp);
		img->bmp = 0;
	}

	if ( img->mask ) {
		fgl_disposeMonoBitmap( img->mask);
		img->mask = 0;
	}

	if ( img->alpha ) {
		fgl_disposeA8Bitmap( img->alpha);
		img->alpha = 0;
	}
}


/*
 * Get alpha value of image pixel [0..255]
 */
static inline int g_getPixelAlpha ( Image* img, int x, int y )
{
	if ( img->mask ){
		return fgl_isMonoBitSet( img->mask, x, y) ? 0xff : 0;
	}
	else if ( img->alpha ){
		return fgl_getPixelA8( img->alpha, x, y);
	}
	else {
		return 0xff;
	}
}

/*
 * Answer alpha type of image. There are three supported
 * alpha type values
 *  - NO_ALPHA     : no alpha channel at all
 *  - MASKED_ALPHA : alpha = [0,255] (transparent pixels)
 *  - FULL_ALPHA   : alpha = [0..255]
 */
static inline int g_getAlphaType ( Image* img )
{
	if ( img->mask )
		return MASK_ALPHA;
	else if ( img->alpha )
		return FULL_ALPHA;
	else
		return NO_ALPHA;
}


/*
 * Set pixel value in Image (with optional alpha value [0..255]).
 * This is called before g_completeImage(), i.e. it can use
 * temporary pixel/alpha storage
 */
static inline void g_putImagePixel ( Image* img , int x, int y,
									 g_pixel_t pix, unsigned char alpha )
{
	if ( alpha != 0xff ){
		if ( img->alpha ){
			fgl_setPixelA8( img->alpha, x, y, alpha);
		}
		else if ( img->mask ){
			fgl_unsetMonoBit( img->mask, x, y);
		}
	}

	fgl_setBmpPixel( img->bmp, x, y, pix);
}

/*
 * Answer pixel value of already completed Image
 */
static inline g_pixel_t  g_getImagePixel ( Image* img, int x, int y )
{
	return fgl_getBmpPixel( img->bmp, x, y);
}


/******************************************************************************
 * event processing
 */

/*
 * Initialize keyboard input
 */
static inline void g_initKeyboard ( Toolkit* X )
{
	/* nothing to do, it's in ipl_init() */
}

/*
 * Initialize moude input
 */
static inline void g_initMouse ( Toolkit* X )
{
	/* mouse already initialized in ipl_init() */

	/* cursor is already initialized by tlkInit */
	if ( !fgl_cursor ){
		DBG( awt_evt, ("no cursor"));
		return;
	}

    fgl_moveCursor( ipl_xPointer, ipl_yPointer);
    fgl_showCursor();
}

/*
 * Get current horizontal pointer location
 */
static inline int g_getMouseX ( Toolkit* X )
{
	return X->event.x;
}

/*
 * Get current vertical pointer location
 */
static inline int g_getMouseY ( Toolkit* X )
{
	return X->event.y;
}

/*
 * Get nmber of pressed mouse button
 */
static inline int g_getMouseBtn ( Toolkit* X )
{
	return X->event.buttons;
}

/*
 * Get Java keycode of current input event
 */
static inline int g_getKeyCode ( Toolkit* X )
{
	return X->event.keyCode;
}

/*
 * Get Java keychar of current input event
 */
static inline int g_getKeyChar ( Toolkit* X )
{
	return X->event.keyChar;
}

/*
 * Make sure the next event will be a (artificial) mouse motion event
 * (with the last known pointer position). Used to retarget mouse for popups.
 * Note that at this point we might not have had a valid _(X)->event, so
 * we must acquire the coords elsewhere
 */
static inline void g_queueMouseMotion ( JNIEnv* env, Toolkit* X )
{
	X->event.type = EVT_MOUSE_MOVE;
	X->event.x = ipl_xPointer;
	X->event.y = ipl_yPointer;
	X->event.next = 0;

	ipl_queueEvent( &(X->event));
}

static inline void g_queuePaint ( JNIEnv* env, Toolkit* X)
{
	ipl_wakeUp();
}

#endif /* GRLIB_FUNCS */
