/*
 * geometry.c - geometric routines based on LinkedRects (inlined methods are in
 *              toolkit-common-pre.h)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <stdlib.h>

#include "toolkit.h"

#define LEFT    1
#define TOP     2
#define RIGHT   4
#define BOTTOM  8


LinkedRect *rectCache;


/*****************************************************************************************
 * geometric functions
 */

LinkedRect*
getDiff ( LinkedRect* below, LinkedRect* above )
{
  LinkedRect     *r, *list = 0;
  unsigned char  edge = 0;

  /* not obscured at all */
  if ( (above->x1 <= below->x0) || (above->x0 >= below->x1) ||
	   (above->y1 <= below->y0) || (above->y0 >= below->y1) )
	return newRect( below->x0, below->y0, below->x1, below->y1, 0, 0);

  if ( (above->x0 > below->x0) && (above->x0 < below->x1) ) edge |= LEFT;
  if ( (above->y0 > below->y0) && (above->y0 < below->y1) ) edge |= TOP;
  if ( (above->x1 > below->x0) && (above->x1 < below->x1) ) edge |= RIGHT;
  if ( (above->y1 > below->y0) && (above->y1 < below->y1) ) edge |= BOTTOM;

  switch ( edge ) {
  case 0:
	return 0; /* completely obscured */
  case LEFT:
	list = newRect( below->x0, below->y0, above->x0-1, below->y1, 0, 0);
	break;
  case TOP:
	list = newRect( below->x0, below->y0, below->x1, above->y0-1, 0, 0);
	break;
  case LEFT | TOP:
	list = newRect( below->x0, below->y0, above->x0-1, below->y1, 0, 0);
	newRect( above->x0, below->y0, below->x1, above->y0-1, list, 0);
	break;
  case RIGHT:
	list = newRect( above->x1+1, below->y0, below->x1, below->y1, 0, 0);
	break;
  case RIGHT | LEFT:
	list = newRect( below->x0, below->y0, above->x0-1, below->y1, 0, 0);
	newRect( above->x1+1, below->y0, below->x1, below->y1, list, 0);
	break;
  case RIGHT | TOP:
	list = newRect( below->x0, below->y0, above->x1, above->y0-1, 0, 0);
	newRect( above->x1+1, below->y0, below->x1, below->y1, list, 0);
	break;
  case LEFT | TOP | RIGHT:
	list = newRect( below->x0, above->y0, above->x0-1, below->y1, 0, 0);
	r = newRect( below->x0, below->y0, below->x1, above->y0-1, list, 0);
	newRect( above->x1+1, above->y0, below->x1, below->y1, r, 0);
	break;
  case BOTTOM:
	list = newRect( below->x0, above->y1+1, below->x1, below->y1, 0, 0);
	break;
  case LEFT | BOTTOM:
	list = newRect( below->x0, below->y0, above->x0-1, below->y1, 0, 0);
	newRect( above->x0, above->y1+1, below->x1, below->y1, list, 0);
	break;
  case TOP | BOTTOM:
	list = newRect( below->x0, below->y0, below->x1, above->y0-1, 0, 0);
	newRect( below->x0, above->y1+1, below->x1, below->y1, list, 0);
	break;
  case LEFT | TOP | BOTTOM:
	list = newRect( below->x0, below->y0, above->x0-1, below->y1, 0, 0);
	r = newRect( above->x0, below->y0, below->x1, above->y0-1, list, 0);
	newRect( above->x0, above->y1+1, below->x1, below->y1, r, 0);
	break;
  case RIGHT | BOTTOM:
	list = newRect( below->x0, above->y1+1, above->x1, below->y1, 0, 0);
	newRect( above->x1+1, below->y0, below->x1, below->y1, list, 0);
	break;
  case LEFT | BOTTOM | RIGHT:
	list = newRect( below->x0, below->y0, above->x0-1, above->y1, 0, 0);
	r = newRect( below->x0, above->y1+1, below->x1, below->y1, list, 0);
	newRect( above->x1+1, below->y0, below->x1, above->y1, r, 0);
	break;
  case TOP | RIGHT | BOTTOM:
	list = newRect( below->x0, below->y0, above->x1, above->y0-1, 0, 0);
	r = newRect( above->x1+1, below->y0, below->x1, below->y1, list, 0);
	newRect( below->x0, above->y1+1, above->x1, below->y1, r, 0);
	break;
  case LEFT | TOP | RIGHT | BOTTOM:
	list = newRect( below->x0, below->y0, above->x0-1, below->y1, 0, 0);
	r = newRect( above->x0, below->y0, above->x1, above->y0-1, list, 0);
	r = newRect( above->x0, above->y1+1, above->x1, below->y1, r, 0);
	newRect( above->x1+1, below->y0, below->x1, below->y1, r, 0);
	break;
  }
	
  return list;
}

LinkedRect*
getIntersection ( LinkedRect *b, LinkedRect *a )
{
  int x0, y0, x1, y1;

  if ( (a->x1 <= b->x0) || (a->x0 >= b->x1) || (a->y1 <= b->y0) || (a->y0 >= b->y1) )
	return 0;  /* no intersection at all */

  x0 = (a->x0 > b->x0) ? a->x0 : b->x0;
  y0 = (a->y0 > b->y0) ? a->y0 : b->y0;
  x1 = (a->x1 < b->x1) ? a->x1 : b->x1;
  y1 = (a->y1 < b->y1) ? a->y1 : b->y1;

  return newRect( x0, y0, x1, y1, 0, 0);
}

LinkedRect*
getUnion ( LinkedRect *a, LinkedRect *b )
{
  int x0, y0, x1, y1;

  x0 = (a->x0 < b->x0) ? a->x0 : b->x0;
  y0 = (a->y0 < b->y0) ? a->y0 : b->y0;
  x1 = (a->x1 > b->x1) ? a->x1 : b->x1;
  y1 = (a->y1 > b->y1) ? a->y1 : b->y1;

  return newRect( x0, y0, x1, y1, 0, 0);
}

LinkedRect*
getListUnion ( LinkedRect *list )
{
  LinkedRect *r = newRect( list->x0, list->y0, list->x1, list->y1, 0, 0);

  for ( list=list->next; list; list=list->next ){
    r->x0 = (r->x0 < list->x0) ? r->x0 : list->x0;
    r->y0 = (r->y0 < list->y0) ? r->y0 : list->y0;
    r->x1 = (r->x1 > list->x1) ? r->x1 : list->x1;
    r->y1 = (r->y1 > list->y1) ? r->y1 : list->y1;
  }

  return r;
}

LinkedRect*
getDiffPart ( LinkedRect* list, LinkedRect* w )
{
  LinkedRect  *r, *v, *p;

  for ( v=list, list=p=0; v; r=v, v=v->next, freeRect( r) ) {
	if ( (r = getDiff( v, w)) ) {      /* we have a difference */
	  if ( !p ){
		list = p = r;
	  }
	  else {
		p->next = r;
	  }
	  for ( p=r; p->next; p=p->next );
	}
  }

  return list;
}

LinkedRect*
getDiffs ( LinkedRect* r0, LinkedRect* r1 )
{
  LinkedRect  *w;
  LinkedRect  *list;

  list = newRect( r1->x0, r1->y0, r1->x1, r1->y1, 0, 0);

  for ( w=r0; w; w=w->next ) {
	list = getDiffPart( list, w);
  }
  return list;
}

#ifdef DEBUG
void dumpLinkedRect ( LinkedRect* r )
{
  int i;

  for ( i=0; r; r=r->next, i++ )
	printf( "[%d] : %d,%d  %d,%d\n", i, r->x0, r->y0, r->x1, r->y1);
}
#endif /* DEBUG */


#ifdef TEST
void
main ( int argc, char* argv[] )
{
  /**
	 LinkedRect *below, *above, *seg;

	 below = newRect( atoi( argv[1]), atoi( argv[2]), atoi( argv[3]), atoi( argv[4]), 0, 0);
	 above = newRect( atoi( argv[5]), atoi( argv[6]), atoi( argv[7]), atoi( argv[8]), 0, 0);

	 seg = getDiff( below, above);
	 dumpLinkedRect( seg);
  **/

  LinkedRect *r1 = newRect( -2, 3,  1, 6, 0, 0);
  LinkedRect *r2 = newRect( 1, -1, 4, 2, r1, 0);
  LinkedRect *r3 = newRect( 3, 1, 7, 5, r2, 0);

  LinkedRect *r  = newRect( 0, 0, 5, 4, 0, 0);
  LinkedRect *s = getDiffs( r1, r);

  //LinkedRect *s = getIntersection( r3, r);

  dumpLinkedRect( s);
}
#endif

