/**
 * gra-common.c - common graphics support (heavily depends on grlib.def macros)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


/**************************************************************************
 * auxiliary functions
 */

static inline int isGrVisible ( Graphics* gr )
{
	return (_(gr)->x0 >= -10000);
}

static inline void clipToScreen ( Graphics* gr )
{
	/* a little bit of sanity check, clip to screen */
	if ( _(gr)->clip.x0 < 0 ) _(gr)->clip.x0 = 0;
	if ( _(gr)->clip.y0 < 0 ) _(gr)->clip.y0 = 0;
	if ( _(gr)->clip.x1 >= _(X)->width )  _(gr)->clip.x1 = _(X)->width-1;
	if ( _(gr)->clip.y1 >= _(X)->height ) _(gr)->clip.y1 = _(X)->height-1;
}

static inline void setGrVisible ( Graphics *gr, int isVisible )
{
	/*
	 * This is rather a hack to "defuse" a Graphics object, but we don't want to
	 * add checks on every draw op to test if the target is invisible. Make sure
	 * this is clipped correctly by the grlib (and doesn't produce segfaults
	 */
	if ( !isVisible ) {
		if ( _(gr)->x0 >= -10000 ) 
			_(gr)->x0 -= 100000;
	}
	else {
		if ( _(gr)->x0 <= -10000 )
			_(gr)->x0 += 100000;
	}
}


/*
 * clip segments are calculated from the clip rect of the Graphics object, and
 * the visSegments (rectangle list) of the target Window, i.e. have to be recalculated
 * each time either one of them changes (which is detected via corresponding
 * visTag values)
 */
void
recalcClipSegments ( Graphics* gr )
{
	AwtWnd     *wnd = (AwtWnd*) _(gr)->tgt;
	LinkedRect *v, *r, *p;
	LinkedRect *visSegs = wnd->visSegments;

	if ( _(gr)->clipSegments ) {
		freeRectList( _(gr)->clipSegments);
	}

	_(gr)->clipSegments = 0;
	_(gr)->curClip = 0;
	_(gr)->visTag = wnd->visTag;

	if ( wnd->mapped && wnd->visibility ) {
		p = 0;
		for ( v=visSegs; v; v=v->next ){
			if ( (r = getIntersection( v, &_(gr)->clip) ) ){
				if ( !_(gr)->clipSegments ){
					_(gr)->clipSegments = p = r;
				}
				else {
					p->next = r;
					p = r;
				}
			}
		}
	}
}


void
updateWindowClipping ( Graphics* gr )
{
	int isVisible = isGrVisible( gr);
	int oldX = _(gr)->x0;
	int oldY = _(gr)->y0;

	_(gr)->x0 = ((AwtWnd*)_(gr)->tgt)->bounds.x0 + _(gr)->xOff;
	_(gr)->y0 = ((AwtWnd*)_(gr)->tgt)->bounds.y0 + _(gr)->yOff;

	translateRect( &_(gr)->clip, (_(gr)->x0 - oldX), (_(gr)->y0 -oldY));
		
	recalcClipSegments( gr);
	
	if ( !isVisible ){
		setGrVisible( gr, 0);
	}
}

/*
 * Make sure the graphics object has the right state (color, font, clip segments)
 * before it is used. This is part of the prolog of each draw function
 */
inline int
setContext ( Toolkit* X, Graphics* gr )
{
	if ( _(X)->lastGr != gr ) {
		_(X)->lastGr = gr;
		_(gr)->curClip = 0;

		/* this will set the propper context, colors, font and mode*/
		g_setCurrentGraphics( X, gr);
	}

	if ( _(gr)->tgtType == WINDOW ) {
		if ( ((AwtWnd*)_(gr)->tgt)->visTag != _(gr)->visTag ){
			updateWindowClipping( gr);
			_(gr)->curClip = 0;
		}
	}

	if ( _(gr)->clipSegments ) {
		if ( _(gr)->curClip != _(gr)->clipSegments ) {
			/* "activate" Graphics clipping */
			_(gr)->curClip = _(gr)->clipSegments;
			g_setClip( gr);
		}
		return 1;
	}
	return 0;
}

inline int
nextClipSegment ( Graphics* gr )
{
	if ( _(gr)->curClip->next ){
		_(gr)->curClip = _(gr)->curClip->next;
		g_setClip( gr);
		return 1;
	}
	else {
		return 0;
	}
}


int
jarray2Points ( JNIEnv* env, Toolkit* X, g_point_t** pp, int x0, int y0,
				jarray xPoints, jarray yPoints, int nPoints, int closeIt )
{
	register int  i;
	int           n;
	jboolean      isCopy;
	jint          *jx, *jy;

	if ( !xPoints || !yPoints ) return (0);

	jx = (*env)->GetIntArrayElements( env, xPoints, &isCopy);
	jy = (*env)->GetIntArrayElements( env, yPoints, &isCopy);

	if ( (n = (*env)->GetArrayLength( env, xPoints)) < nPoints ) nPoints = n;
	if ( (n = (*env)->GetArrayLength( env, yPoints)) < nPoints ) nPoints = n;

	*pp = getBuffer( _(X), (nPoints+1)*sizeof(g_point_t));

	for ( i=0; i<nPoints; i++){
		g_setPoint( &(*pp)[i], jx[i] + x0, jy[i] + y0);
	}

	if ( closeIt && ((jx[i] != jx[0]) || (jy[i] != jy[0])) ){
		g_setPoint( &(*pp)[nPoints], jx[0]+x0, jy[0]+y0);
		nPoints++;
	}

	(*env)->ReleaseIntArrayElements( env, xPoints, jx, JNI_ABORT);
	(*env)->ReleaseIntArrayElements( env, yPoints, jy, JNI_ABORT);

	return nPoints;
}



/**************************************************************************
 * initialization and finalization
 */

void*
_graInitGraphics ( JNIEnv* env, jclass clazz,
				   Graphics* gr, jobject tgt, jint tgtType, jboolean isVisible,
				   jint xOff, jint yOff,
				   jint xClip, jint yClip, jint wClip, jint hClip,
				   jobject fnt, jint fg, jint bg )
{
	DBG( awt_gra, ("initGraphics: %p (%p, %d) %d,%d  %d,%d - %d,%d  %p %x %x\n",
				   gr, tgt,(int)tgtType, (int)xOff, (int)yOff,
				   (int)xClip, (int)yClip, (int)wClip, (int)hClip,
				   fnt, (unsigned)fg, (unsigned)bg));

	if ( !gr ) {
		gr = (Graphics*) AWT_CALLOC( 1, sizeof(Graphics));
		g_initGraphics( gr);
		DBG( awt_gra, (" -> gr: %p\n", gr));
	}
	else {
		if ( _(X)->lastGr == gr )
			_(X)->lastGr = 0;
	}

	if ( tgt ) {
		switch ( tgtType ) {
		case WINDOW:
			g_setWindowGraphics( gr, (AwtWnd*)tgt);
			_(gr)->tgtType = tgtType;
			_(gr)->tgt = tgt;
			_(gr)->xOff = xOff;
			_(gr)->yOff = yOff;
			xOff += ((AwtWnd*)tgt)->bounds.x0;
			yOff += ((AwtWnd*)tgt)->bounds.y0;
			break;
		case IMAGE:
			g_setImageGraphics( gr, (Image*)tgt);
			_(gr)->tgtType = tgtType;
			_(gr)->tgt = tgt;
			_(gr)->xOff = 0;
			_(gr)->yOff = 0;
			break;
		case GRAPHICS:
			g_setGraphicsGraphics( gr, (Graphics*)tgt);
			_(gr)->tgtType = ((Graphics*)tgt)->common.tgtType;
			_(gr)->tgt = ((Graphics*)tgt)->common.tgt;

			if ( _(gr)->tgtType == WINDOW ) {
				_(gr)->xOff = xOff;
				_(gr)->yOff = yOff;
				xOff += ((AwtWnd*)(_(gr)->tgt))->bounds.x0;
				yOff += ((AwtWnd*)(_(gr)->tgt))->bounds.y0;
			}
			else if ( _(gr)->tgtType == IMAGE ) {
				_(gr)->xOff = 0;
				_(gr)->yOff = 0;
			}

			/* the visibility of our ref graphics is unknown to Java - we have to deduce it */
			isVisible = isGrVisible( (Graphics *)tgt);

			break;
		default: 
			DBG( awt_gra,("illegal Graphics target: %d\n", (int)tgtType));
		}
	}
	else {
		if ( _(gr)->tgtType == WINDOW ) {
			if ( _(gr)->tgt ) {
				_(gr)->xOff = xOff;
				_(gr)->yOff = yOff;
				xOff += ((AwtWnd*)(_(gr)->tgt))->bounds.x0;
				yOff += ((AwtWnd*)(_(gr)->tgt))->bounds.y0;
			}
			else {
				DBG( awt_gra, ("Graphics target not initialized yet\n"));
			}
		}
	}

	_(gr)->x0   = xOff;
	_(gr)->y0   = yOff;

	g_setColor( gr, fg);
	g_setBackColor( gr, bg);
	g_setPaintMode( gr);
	if ( fnt ){
		/* we might not have one, make sure the driver doesn't go bonkers on '0' */
		g_setFont( gr, (g_font_t*)fnt);
	}

	_(gr)->clip.x0 = xOff + xClip;
	_(gr)->clip.y0 = yOff + yClip;
	_(gr)->clip.x1 = _(gr)->clip.x0 + wClip - 1;
	_(gr)->clip.y1 = _(gr)->clip.y0 + hClip - 1;

	if ( _(gr)->clip.x1 < _(gr)->clip.x0 )  
	    _(gr)->clip.x1 = _(gr)->clip.x0; 
	if ( _(gr)->clip.y1 < _(gr)->clip.y0 )  
	    _(gr)->clip.y1 = _(gr)->clip.y0; 
	
	clipToScreen( gr);

	_(gr)->curClip = 0;

	if ( _(gr)->tgtType == WINDOW ) {
		recalcClipSegments( gr);                      /* this frees the current clipSegs */
		_(gr)->visTag = ((AwtWnd*)(_(gr)->tgt))->visTag;
	}
	else if ( _(gr)->tgtType == IMAGE ){
		if ( _(gr)->clipSegments ){                       /* recycle clipSegs */
			setRect( _(gr)->clipSegments,
					 _(gr)->clip.x0, _(gr)->clip.y0,
					 _(gr)->clip.x1, _(gr)->clip.y1, 0);
		}
		else {
			_(gr)->clipSegments = newRect( _(gr)->clip.x0, _(gr)->clip.y0,
										   _(gr)->clip.x1, _(gr)->clip.y1, 0, 0);
		}
	}

	if ( isVisible == JNI_FALSE ){
		setGrVisible( gr, 0);
	}

	return gr;
}

void
_graSetVisible ( JNIEnv* env, jclass clazz, Graphics* gr, jint isVisible )
{
	DBG( awt_gra, ("setVisble: %p  %d\n", gr, (int)isVisible));

	setGrVisible( gr, isVisible);
}


void
_graFreeGraphics ( JNIEnv* env, jclass clazz, Graphics* gr )
{
	DBG( awt_gra, ("freeGraphics: %p\n", gr));

	/*
	 * Keep your fingers off the gr->gc, it's either the screen gc or an
	 * image. The first one is there to stay, and the last one has to be freed
	 * in the target (Image) context.
	 */

	if ( _(gr)->clipSegments )
		freeRectList( _(gr)->clipSegments);

	g_freeGraphics( gr);

	AWT_FREE( gr);
}


/**************************************************************************
 * simple drawing operations
 */

void
_graCopyArea ( JNIEnv* env, jclass clazz, Graphics* gr, 
			   jint x, jint y, jint width, jint height,
			   jint xDelta, jint yDelta )
{
	int wScr, hScr, xw, yh, clipped = 0;
	LinkedRect r, s, *u;

	DBG( awt_gra, ("copyArea: %p, %d,%d-%d,%d  %d,%d\n",
				   gr, (int)x, (int)y, (int)width, (int)height,
				   (int)xDelta, (int)yDelta));
	
	if ( (width <= 0) || (height <= 0) ){
	    return;
	}

	x += _(gr)->x0; y += _(gr)->y0;
	
	if ( _(gr)->tgtType == WINDOW ){
		/*
		 * Since we cannot make assumptions about offscreen drawing space, we have to take
		 * care of blits which exceed our screen, and to generate update events for
		 * the rects which would have to be blitted in from offscreen.
		 * Moreover, an amazing number of graphics drivers don't handle screen clipping
		 * correctly for blits, and because we need to check screen dims, anyway, we
		 * take care of this, too (even though it really should go into the driver)
		 */
		wScr = _(X)->width;
		hScr = _(X)->height;
		xw = x + width-1;
		yh = y + height-1;

		/* check if the source is outside the screen */
		setRect( &s, x, y, xw, yh, 0);
		if ( (x < 0) || (y < 0) || (xw > wScr) || (yh > hScr) ) {
			clipRect( &s, 0, 0, wScr, hScr);
			clipped = 1;
		}
		
		/*
		 * This is questionable - what do we do if there is a copyArea request
		 * which is partly obscured by other windows? Lightweight-wise, we
		 * may simply ignore it, assuming that apps don't rely on native
		 * doublebuffering for that purpose. But then there are all these
		 * dreaded Panels and Canvases, where people carelessly got used to
		 * exactly this hidden, native feature. Same consequence for us - we
		 * have to redraw the parts for which we have no valid source data,
		 * but this costs us in terms of speed
		 */
#if defined (BACKGROUND_SCROLL)
		if ( (u = getDiffs( _(gr)->clipSegments, &s)) ){
			AwtWnd *w = (AwtWnd*)_(gr)->tgt;
			translateRect( u, xDelta, yDelta);
			updateWindow( env, X, w, u);
		}
#endif
		/* check if the destination is outside the screen */
		translateRect( &s, xDelta, yDelta);
		if ( (s.x0 < 0) || (s.y0 < 0) || (s.x1 > wScr) || (s.y1 > hScr) ){
			clipRect( &s, 0, 0, wScr, hScr);
			clipped = 1;
		}

		if ( clipped ) {
			setRect( &r, x, y, xw, yh, 0);
			translateRect( &r, xDelta, yDelta);
			clipRect( &r, 0, 0, wScr, hScr);

			/* redraw parts which are blitted in */
			if ( (u = getDiffs( &s, &r)) ) {
				AwtWnd *w = (AwtWnd*)_(gr)->tgt;
				translateRect( u, -w->bounds.x0, -w->bounds.y0);
				updateWindow( env, X, w, u);
			}

			/* crop parts which would be blitted out */
			translateRect( &s, -xDelta, -yDelta);
			x = s.x0;
			y = s.y0;
			width  = s.x1 - s.x0;
			height = s.y1 - s.y0;
		}
	}
	
	if ( setContext( X, gr) ) {
		do {
			g_copyArea( gr, x, y, x+width-1, y+height-1, xDelta, yDelta);
		} while ( nextClipSegment( gr) );
	}
}


void
_graClearRect ( JNIEnv* env, jclass clazz, Graphics* gr,
				jint x, jint y, jint width, jint height, jint bg )
{
	DBG( awt_gra, ("clearRect: %p, %d,%d-%d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	g_setBackColor( gr, bg);
	
	if ( setContext( X, gr) ) {
		x += _(gr)->x0; y += _(gr)->y0;

		do {
			g_clearRect( gr, x, y, x+width-1, y+height-1);
		} while ( nextClipSegment( gr) );
	}
}

void
_graDrawLine ( JNIEnv* env, jclass clazz, Graphics* gr,
			   jint x1, jint y1, jint x2, jint y2 )
{
	DBG( awt_gra, ("drawLine: %p, %d,%d - %d,%d\n", gr,
				   (int)x1, (int)y1, (int)x2, (int)y2));

	if ( setContext( X, gr) ) {

		x1 += _(gr)->x0; y1 += _(gr)->y0;
		x2 += _(gr)->x0; y2 += _(gr)->y0;

		do {
			if ( (x1 == x2) && (y1 == y2) ){ /* another swing gimmick */
				g_putPixel( gr, x1, y1);
			}
			else {
				g_drawLine( gr, x1, y1, x2, y2);
			}
		} while ( nextClipSegment( gr) );
	}
}


void
_graDrawArc ( JNIEnv* env, jclass clazz, Graphics* gr,
			  jint x, jint y, jint width, jint height,
			  jint startAngle, jint arcAngle )
{
	int xa = width/2;
	int ya = height/2;

	DBG( awt_gra, ("drawArc: %p, %d,%d-%d,%d  %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height,
				   (int)startAngle, (int)arcAngle));

	
	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	x += _(gr)->x0;
	y += _(gr)->y0;

	if ( setContext( X, gr) ) {
		do {
			g_drawOpenArc( gr, x+xa, y+ya, xa, ya, startAngle, arcAngle);
		} while ( nextClipSegment( gr) );
	}
}


void
_graFillArc ( JNIEnv* env, jclass clazz, Graphics* gr,
			  jint x, jint y, jint width, jint height,
			  jint startAngle, jint arcAngle )
{
	int xa = width/2;
	int ya = height/2;

	DBG( awt_gra, ("fillArc: %p, %d,%d-%d,%d  %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height,
				   (int)startAngle, (int)arcAngle));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	x += _(gr)->x0;
	y += _(gr)->y0;

	if ( setContext( X, gr) ) {
		do {
			g_fillArc( gr, x+xa, y+ya, xa, ya, startAngle, arcAngle);
		} while ( nextClipSegment( gr) );
	}
}

void
_graDrawOval ( JNIEnv* env, jclass clazz, Graphics* gr,
			   jint x, jint y, jint width, jint height )
{
	int xa = width/2;
	int ya = height/2;

	DBG( awt_gra, ("drawOval: %p, %d,%d - %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	x += _(gr)->x0;
	y += _(gr)->y0;

	if ( setContext( X, gr) ) {
		do {
			g_drawOval( gr, x+xa, y+ya, xa, ya);
		} while ( nextClipSegment( gr) );
	}
}

void
_graFillOval ( JNIEnv* env, jclass clazz, Graphics* gr,
			   jint x, jint y, jint width, jint height )
{
	int xa = width/2;
	int ya = height/2;

	DBG( awt_gra, ("fillOval: %p, %d,%d - %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	x += _(gr)->x0;
	y += _(gr)->y0;

	if ( setContext( X, gr) ) {
		do {
			g_fillOval( gr, x+xa, y+ya, xa, ya);
		} while ( nextClipSegment( gr) );
	}
}

void
_graDrawRect ( JNIEnv* env, jclass clazz, Graphics* gr,
			   jint x, jint y, jint width, jint height )
{
	DBG( awt_gra, ("drawRect: %p, %d,%d - %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height));

	if ( (width <= 0) || (height <= 0) )
		return;

	if ( setContext( X, gr) ) {
		x += _(gr)->x0; y += _(gr)->y0;

		do {
			g_drawRect( gr, x, y, x+width, y+height);
		} while ( nextClipSegment( gr) );
	}
}

void
_graFillRect ( JNIEnv* env, jclass clazz, Graphics* gr,
			   jint x, jint y, jint width, jint height )
{
	DBG( awt_gra, ("fillRect: %p, %d,%d - %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height));

	if ( (width <= 0) || (height <= 0) )
		return;

	if ( setContext( X, gr) ) {
		x += _(gr)->x0; y += _(gr)->y0;

		do {
			if ( width == 1 ) {                  /* some swing gimmicks */
				g_drawLine( gr, x, y, x, y+height-1);
			}
			else if ( height == 1 ) {
				g_drawLine( gr, x, y, x+width-1, y);
			}
			else {
				g_fillRect( gr, x, y, x+width-1, y+height-1);
			}
		} while ( nextClipSegment( gr) );
	}
}

void
_graDrawPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
				  jarray xPoints, jarray yPoints, jint nPoints )
{
	g_point_t *p;

	DBG( awt_gra, ("drawPolygon: %p, %p,%p  %d\n", gr, xPoints, yPoints, (int)nPoints));

	if ( !xPoints || !yPoints ) return;

	if ( setContext( X, gr) ) {
		nPoints = jarray2Points( env, X, &p, _(gr)->x0, _(gr)->y0,
								 xPoints, yPoints, nPoints, TRUE);
		do {
			g_drawPoly( gr, p, nPoints);
		} while ( nextClipSegment( gr) );
	}
}

void
_graDrawPolyline ( JNIEnv* env, jclass clazz, Graphics* gr,
				   jarray xPoints, jarray yPoints, jint nPoints )
{
	g_point_t *p;

	DBG( awt_gra, ("drawPolyline: %p, %p,%p  %d\n", gr, xPoints, yPoints, (int)nPoints));

	if ( !xPoints || !yPoints ) return;

	if ( setContext( X, gr) ) {
		nPoints = jarray2Points( env, X, &p, _(gr)->x0, _(gr)->y0,
								 xPoints, yPoints, nPoints, FALSE);
		do {
			g_drawPoly( gr, p, nPoints);
		} while ( nextClipSegment( gr) );
	}
}

void
_graFillPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
				  jarray xPoints, jarray yPoints, jint nPoints )
{
	g_point_t *p;

	DBG( awt_gra, ("fillPolygon: %p, %p,%p  %d\n", gr, xPoints, yPoints, (int)nPoints));

	if ( !xPoints || !yPoints ) return;

	if ( setContext( X, gr) ) {
		nPoints = jarray2Points( env, X, &p, _(gr)->x0, _(gr)->y0,
								 xPoints, yPoints, nPoints, FALSE);
		do {
			g_fillPoly( gr, p, nPoints);
		} while ( nextClipSegment( gr) );
	}
}


/**************************************************************************
 * compound drawing operations
 */

void
_graDrawRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
					jint x, jint y, jint width, jint height, 
					jint wArc, jint hArc )
{
	int x1, x2, y1, y2, a, b;

	DBG( awt_gra, ("drawRoundRect: %p, %d,%d - %d,%d  %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height, (int)wArc, (int)hArc));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	if ( setContext( X, gr) ) {

		x += _(gr)->x0;  y += _(gr)->y0;
		a = wArc / 2; b = hArc / 2;

		do {
			x1 = x + a;   x2 = x + width - a;  y2 = y + height;
			g_drawLine( gr, x1, y, x2, y);
			g_drawLine( gr, x1, y2, x2, y2);

			y1 = y + b;   y2 = y + height - b; x2 = x + width;
			g_drawLine( gr, x, y1, x, y2);
			g_drawLine( gr, x2, y1, x2, y2);

			g_drawOpenArc( gr, x+a, y+b, a, b, 90, 90); /* ul */

			x2 = x + width - a;
			g_drawOpenArc( gr, x2, y+b, a, b, 0, 90);   /* ur */

			y2 = y + height - b;
			g_drawOpenArc( gr, x2, y2, a, b, 270, 90);  /* lr */

			g_drawOpenArc( gr, x+a, y2, a, b, 180, 90); /* ll */
		} while ( nextClipSegment( gr) );
	}
}

void
_graFillRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
					jint x, jint y, jint width, jint height, 
					jint wArc, jint hArc )
{
	int x1, x2, y1, y2, a, b;

	DBG( awt_gra, ("fillRoundRect: %p, %d,%d - %d,%d  %d,%d\n", gr,
				   (int)x, (int)y, (int)width, (int)height, (int)wArc, (int)hArc));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	if ( setContext( X, gr) ) {
		x += _(gr)->x0;  y += _(gr)->y0;
		a = wArc / 2; b = hArc / 2;

		do {
			y1 = y + b;  y2 = y + height - b;
			if ( y2 > y1 )
				g_fillRect( gr, x, y1, x+width-1, y2);
  
			x1 = x + a;  x2 = x + width - a;
			if ( x2 > x1 ) {
				g_fillRect( gr, x1, y, x2, y+b);
				g_fillRect( gr, x1, y2, x2, y2+b);
			}

			/* check out if arc fill isn't faster (once we have it) */
			g_fillOval( gr, x+a, y+b, a, b);

			x2 = x + width - a;
			g_fillOval( gr, x2, y+b, a, b);

			y2 = y + height - b;
			g_fillOval( gr, x2, y2, a, b);

			g_fillOval( gr, x+a, y2, a, b);
		} while ( nextClipSegment( gr) );
	}
}


void
_graDraw3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
				 jint x, jint y, jint width, jint height,
				 jboolean raised, jint pixelValue )
{
	int        bright, dark, clr;
	int        xw, yh;

	DBG( awt_gra, ("draw3DRect: %p, %d,%d - %d,%d  %d %x\n", gr,
				   (int)x, (int)y, (int)width, (int)height, raised,
				   (unsigned)pixelValue));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	if ( setContext( X, gr) ) {
		dark   = (int) g_modPixelValue( X, pixelValue, DARKER);
		bright = (int) g_modPixelValue( X, pixelValue, BRIGHTER);
		clr    = g_getColor( gr);
	
		x += _(gr)->x0;     y += _(gr)->y0;
		xw = x + width;  yh = y + height;

		do {
			g_setColor( gr, raised ? bright : dark);
			g_drawHLine( gr, x, y, xw-1);
			g_drawVLine( gr, x, y, yh);

			g_setColor( gr, raised ? dark : bright);
			g_drawHLine( gr, x+1, yh, xw);
			g_drawVLine( gr, xw, y, yh);
		} while ( nextClipSegment( gr) );
	
		g_setColor( gr, clr);
	}
}

void
_graFill3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
				 jint x, jint y, jint width, jint height,
				 jboolean raised, jint pixelValue )
{
	int        bright, dark, clr;
	int        xw, yh;

	DBG( awt_gra, ("fill3DRect: %p, %d,%d - %d,%d  %d %x\n", gr,
				   (int)x, (int)y, (int)width, (int)height, raised,
				   (unsigned)pixelValue));

	if ( (width <= 0) || (height <= 0) ) {
	    return;
	}

	if ( setContext( X, gr) ) {
		dark   = (int) g_modPixelValue( X, pixelValue, DARKER);
		bright = (int) g_modPixelValue( X, pixelValue, BRIGHTER);
		clr    = g_getColor( gr);

		x += _(gr)->x0;     y += _(gr)->y0;
		xw = x + width-1;  yh = y + height-1;

		do {
			g_setColor( gr, raised ? bright : dark);
			g_drawHLine( gr, x, y, xw-1);
			g_drawVLine( gr, x, y, yh);

			g_setColor( gr, raised ? dark : bright);
			g_drawHLine( gr, x+1, yh, xw);
			g_drawVLine( gr, xw, y, yh);

			g_setColor( gr, clr);
			g_fillRect( gr, x+1, y+1, xw-1, yh-1);
		} while ( nextClipSegment( gr) );
	}
}


/**************************************************************************
 * text rendering
 */

void
_graDrawBytes ( JNIEnv* env, jclass clazz, Graphics* gr,
				jarray jBytes, jint offset, jint len,
				jint x, jint y )
{
	jboolean  isCopy;
	int       n;
	jbyte     *jb, *jbo;

	if ( !jBytes ) return;

	x += _(gr)->x0;
	y += _(gr)->y0;

	n   = (*env)->GetArrayLength( env, jBytes);
	jb = (*env)->GetByteArrayElements( env, jBytes, &isCopy);
	jbo = jb + offset;

	DBG( awt_gra, ("drawBytes: %p, %p,%d,%d  \"%s\"  %d,%d\n",
				   gr, jBytes, (int)offset, (int)len, jb, (int)x, (int)y));

	if ( offset+len > n )
		len = n - offset;

	if ( setContext( X, gr) ) {
		do {
			g_drawByteText( gr, jbo, len, x, y);
		} while ( nextClipSegment( gr) );
	}

	(*env)->ReleaseByteArrayElements( env, jBytes, jb, JNI_ABORT);
}

void
_graDrawChars ( JNIEnv* env, jclass clazz, Graphics* gr,
				jarray jChars, jint offset, jint len,
				jint x, jint y )
{
	jboolean    isCopy;
	int         n;
	jchar       *jc, *jco;

	if ( !jChars ) return;

	x += _(gr)->x0;
	y += _(gr)->y0;

	n = (*env)->GetArrayLength( env, jChars);
	jc = (*env)->GetCharArrayElements( env, jChars, &isCopy);
	jco = jc + offset;

	DBG( awt_gra, ("drawChars: %p, %p,%d,%d  \"%s\"  %d,%d\n",
				   gr, jChars, (int)offset, (int)len,
				   jchar2CString( _(X), jco, len), (int)x, (int)y));

	if ( offset+len > n )
		len = n - offset;

	if ( setContext( X, gr) ) {
		do {
			g_drawWideText( gr, jco, len, x, y);
		} while ( nextClipSegment( gr) );
	}

	(*env)->ReleaseCharArrayElements( env, jChars, jc, JNI_ABORT);
}

void
_graDrawString ( JNIEnv* env, jclass clazz, Graphics* gr,
				 jstring str, jint x, jint y )
{
	jboolean    isCopy;
	int         len;
	const jchar *jc;

	DBG( awt_gra, ("drawString: %p  \"%s\"  %d,%d\n",
				   gr, java2CString( env, _(X), str), (int)x, (int)y));

	if ( !str ) return;

	x += _(gr)->x0;
	y += _(gr)->y0;

	len   = (*env)->GetStringLength( env, str);
	jc = (*env)->GetStringChars( env, str, &isCopy);

	if ( setContext( X, gr) ) {
		do {
			g_drawWideText( gr, (jchar*)jc, len, x, y);
		} while ( nextClipSegment( gr) );
	}

	(*env)->ReleaseStringChars( env, str, jc);
}

void
_graDrawVertString ( JNIEnv* env, jclass clazz, Graphics* gr,
				 jstring str, jint x, jint y )
{
	jboolean    isCopy;
	int         len;
	const jchar *jc;

	DBG( awt_gra, ("drawVertString: %p  \"%s\"  %d,%d\n",
				   gr, java2CString( env, _(X), str), (int)x, (int)y));

	if ( !str ) return;

	x += _(gr)->x0;
	y += _(gr)->y0;

	len   = (*env)->GetStringLength( env, str);
	jc = (*env)->GetStringChars( env, str, &isCopy);

	if ( setContext( X, gr) ) {
		do {
			g_drawWideVertText( gr, (jchar*)jc, len, x, y);
		} while ( nextClipSegment( gr) );
	}

	(*env)->ReleaseStringChars( env, str, jc);
}



/*******************************************************************************
 * graphics state modification
 */

void
_graAddClip ( JNIEnv* env, jclass clazz, Graphics* gr,
			  jint xClip, jint yClip, jint wClip, jint hClip )
{
	LinkedRect r;
	int x = xClip + _(gr)->x0;
	int y = yClip + _(gr)->y0;
	int xw = x + wClip -1;
	int yh = y + hClip -1;

	DBG( awt_gra, ("addClip: %p, %d,%d - %d,%d\n", gr,
				   (int)xClip, (int)yClip, (int)wClip, (int)hClip));

	/* clip the clip rect to our gr clip bounds, since we are going
	 * to compute a diff (which shouldn't exceed our bounds
	 */
	if ( x <= _(gr)->clip.x0 )  x  = _(gr)->clip.x0;
	if ( y <= _(gr)->clip.y0 )  y  = _(gr)->clip.y0;
	if ( xw >= _(gr)->clip.x1 ) xw = _(gr)->clip.x1;
	if ( yh >= _(gr)->clip.y1 ) yh = _(gr)->clip.y1;

	setRect( &r, x, y, xw, yh, 0);
	_(gr)->clipSegments = getDiffPart( _(gr)->clipSegments, &r);
}

void
_graSetClip ( JNIEnv* env, jclass clazz, Graphics* gr,
			  jint xClip, jint yClip, jint wClip, jint hClip )
{
	DBG( awt_gra, ("setClip: %p, %d,%d - %d,%d\n", gr,
				   (int)xClip, (int)yClip, (int)wClip, (int)hClip));

	
	_(gr)->clip.x0 = _(gr)->x0 + xClip;
	_(gr)->clip.y0 = _(gr)->y0 + yClip;
	_(gr)->clip.x1 = _(gr)->clip.x0 + wClip - 1;
	_(gr)->clip.y1 = _(gr)->clip.y0 + hClip - 1;

	if ( _(gr)->clip.x1 < _(gr)->clip.x0 )  
	    _(gr)->clip.x1 = _(gr)->clip.x0; 
	if ( _(gr)->clip.y1 < _(gr)->clip.y0 )  
	    _(gr)->clip.y1 = _(gr)->clip.y0; 

	clipToScreen( gr);

	if ( _(gr)->tgtType == WINDOW ){
		recalcClipSegments( gr);
	}
	else if ( _(gr)->tgtType == IMAGE ) {
		setRect( _(gr)->clipSegments,
				 _(gr)->clip.x0, _(gr)->clip.y0,
				 _(gr)->clip.x1, _(gr)->clip.y1, 0);

		_(gr)->curClip = 0; /* otherwise we fail the test in setContext */
	}

	/* no need to call g_setClip because a subsequent setContext()
	 * will do that for us
	 */
}

void
_graSetOffset ( JNIEnv* env, jclass clazz, Graphics* gr,
				jint xOff, jint yOff )
{
	DBG( awt_gra, ("setOffset: %p, %d,%d\n", gr, (int)xOff, (int)yOff));

	_(gr)->x0 = _(gr)->xOff = xOff;
	_(gr)->y0 = _(gr)->yOff = yOff;

	if ( _(gr)->tgtType == WINDOW ){
		_(gr)->x0 += ((AwtWnd*)(_(gr)->tgt))->bounds.x0;
		_(gr)->y0 += ((AwtWnd*)(_(gr)->tgt))->bounds.y0;
	}

	_(gr)->curClip = 0; /* force new clip in next setContext() */
}


void
_graSetColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
{
	DBG( awt_gra, ("setColor: %p, %x\n", gr, (unsigned)clr));

	g_setColor( gr, clr);
}

void
_graSetBackColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
{
	DBG( awt_gra, ("setBackColor: %p, %x\n", gr, (unsigned)clr));

	g_setBackColor( gr, clr);
}

void
_graSetFont ( JNIEnv* env, jclass clazz, Graphics* gr, g_font_t* fnt )
{
	DBG( awt_gra, ("setFont: %p, %p\n", gr, fnt));

	g_setFont( gr, fnt);
}

void
_graSetPaintMode ( JNIEnv* env, jclass clazz, Graphics* gr )
{
	DBG( awt_gra, ("setPaintMode: %p\n", gr));

	g_setPaintMode( gr);
}

void
_graSetXORMode ( JNIEnv* env, jclass clazz, Graphics* gr, jint xorClr )
{
	DBG( awt_gra, ("setXORMode: %p, %x\n", gr, (unsigned)xorClr));

	g_setXORMode( gr, xorClr);
}


/*******************************************************************************
 * image drawing
 */

/*
 * create a offscreen buffer image suitable for rendering delta frames
 */
static inline Image*
getBufferImage ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
				 int x, int y )
{
	Image    *buffer, *im;

	/* do we already have a buffer offscreen image? If not, create it and
	 * set it for all frames
	 */
	if ( !(buffer = _(img)->buffer) ){
		_(img)->buffer = buffer = createImage( _(img)->width, _(img)->height, 1);
		g_setAlphaType( buffer, NO_ALPHA);
		for ( im = _(img)->next; im != img; im=_(im)->next){
			_(im)->buffer = _(img)->buffer;
		}
		
		/* create a graphics for this buffer, so that we don't have to re-init
		 * each and every time (might be costly, and it's all invariant)
		 */
		_(buffer)->gr = _graInitGraphics( env, clazz, 0, (jobject)buffer, IMAGE, 1, 0, 0,
		                                  0, 0, _(img)->width, _(img)->height,
		                                  0, 0xff000000, 0xffffffff);

		/* check if the first frame is already transparent, and fill with
		 * the target background if it is
		 */
		if ( _(img)->trans != -1 ) {
			g_grabPixels( _(buffer)->gr, gr,
						  x, y, _(img)->width, _(img)->height);
		}
	}

	/* draw the current frame into the buffer */
	g_drawImage( _(buffer)->gr, img, _(img)->left, _(img)->top,
	             0, 0, _(img)->width-1, _(img)->height-1, -1);

	return buffer;
}


void
_graDrawImage ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
				jint srcX, jint srcY, jint dstX, jint dstY,
				jint width, jint height, jint bgval )
{
	int     x0, y0, x1, y1;
	int     dX, dY, sX, sY, sXw, sYh;

	if ( (!img) || (_(img)->width == -1) || !isGrVisible(gr) ){ /* last defense sanity check */
		return;
	}

	DBG( awt_gra, ("drawImage: %p %p (%p,%p %d,%d) %d,%d, %d,%d, %d,%d, %x\n",
				   gr, img,  _(img)->mask,_(img)->alpha, 
				   _(img)->width,_(img)->height,
				   (int)srcX, (int)srcY, (int)dstX, (int)dstY, 
				   (int)width, (int)height, (unsigned)bgval));

	if ( (width <= 0) || (height <= 0) ){
	    return;
	}

	/* If this is a delta frame animation, we need to draw the new frame into
	 * a buffer, and then blit this buffer onto the screen. This results in
	 * two blits per frame, but the delta is probably small enough to be fast,
	 * and the overall costs significantly less than drawing from frame0.
	 * Note that this scheme is also useful for video frames, but doesn't work
	 * if the background is changed between frames
	 */
	if ( (_(img)->flags & IMG_BUFFER) || (_(img)->buffer) ) {
		img = getBufferImage( env, clazz, gr, img, dstX, dstY);
	}
	
	if ( srcX + width  > _(img)->width  ){
		width  = _(img)->width  - srcX;
	}
	if ( srcY + height > _(img)->height ){
		height = _(img)->height - srcY;
	}

	if ( setContext( X, gr) ) {
		x0  = _(gr)->x0 + dstX + _(img)->left;
		y0  = _(gr)->y0 + dstY + _(img)->top;
		x1  = x0 + width;
		y1  = y0 + height;

		do {
			if ( x0 < _(gr)->curClip->x0 ) {
				dX = _(gr)->curClip->x0;
				sX = srcX + (dX - x0);
			}
			else {
				dX = x0;
				sX = srcX;
			}
			if ( y0 < _(gr)->curClip->y0 ) {
				dY = _(gr)->curClip->y0;
				sY = srcY + (dY - y0);

			}
			else {
				dY = y0;
				sY = srcY;
			}

			if ( x1 > _(gr)->curClip->x1 ){
				sXw = sX + (_(gr)->curClip->x1 - dX) + 1;
			}
			else {
				sXw = sX + (x1 - dX);
			}
			if ( y1 > _(gr)->curClip->y1 ){
				sYh = sY + (_(gr)->curClip->y1 - dY) + 1;

				//DBG( awt_gra, ("clipping y1: %d,%d, %d,%d, %d\n",
				//     y1, _(gr)->curClip->y1, sY, dY, sYh ));

			}
			else {
				sYh = sY + (y1 - dY);
			}

			if ( (sXw > 0) && (sYh > 0) && (sX >=0) && (sY >=0) ) {
				g_drawImage( gr, img, dX, dY, sX, sY, sXw, sYh, bgval);

			}

		} while ( nextClipSegment( gr) );
	}
}






