/**
 * evt.c - platform specific event handling (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

/******************************************************************************
 * generic exported methods
 */

#define _evtInit              EXPORT(evtInit)
#define _evtRegisterSource    EXPORT(evtRegisterSource)
#define _evtUnregisterSource  EXPORT(evtUnregisterSource)

/******************************************************************************
 * import generic functionality
 */

#include "../evt-common.c"


#include "keytab.h"


/******************************************************************************
 * platform specific event functions
 */

/*
 * determine keyboard state changes (to generate the required modifier keys)
 */
inline void
getStateChanges ( Toolkit* X )
{
	if ( X->event.kbstat & GR_KB_SHIFT ){
		if ( !(X->lastKbStat & GR_KB_SHIFT) )  X->addedKbStat |= GR_KB_SHIFT;
	}
	else {
		if ( X->lastKbStat & GR_KB_SHIFT )   X->removedKbStat |= GR_KB_SHIFT;
	}

	if ( X->event.kbstat & GR_KB_CTRL ){
		if ( !(X->lastKbStat & GR_KB_CTRL) )  X->addedKbStat |= GR_KB_CTRL;
	}
	else {
		if ( X->lastKbStat & GR_KB_CTRL )   X->removedKbStat |= GR_KB_CTRL;
	}
 
	if ( X->event.kbstat & GR_KB_ALT ){
		if ( !(X->lastKbStat & GR_KB_ALT) )  X->addedKbStat |= GR_KB_ALT;
	}
	else {
		if ( X->lastKbStat & GR_KB_ALT )   X->removedKbStat |= GR_KB_ALT;
	}
}


/*
 * emit (artificial) shift/control/alt key events
 */
jobject
modKeyNotify ( JNIEnv* env, Toolkit* X, int id )
{
	int keyCode;
	int *stat = (id == KEY_PRESSED) ? &X->addedKbStat : &X->removedKbStat;

	if ( *stat & GR_KB_SHIFT ){
		keyCode = VK_SHIFT;
		*stat &= ~GR_KB_SHIFT;
	}
	else if ( *stat & GR_KB_CTRL ){
		keyCode = VK_CTRL;
		*stat &= ~GR_KB_CTRL;
	}
	else if ( *stat & GR_KB_ALT ){
		keyCode = VK_ALT;
		*stat &= ~GR_KB_ALT;
	}
	else {
		return 0;  /* what? why have we been called (shouldn't happen) */
	}

	DBG( awt_evt, ("modKeyNotify: %d %d\n", id, keyCode));

	return (*env)->CallStaticObjectMethod( env, KeyEvent, getKeyEvent,
										   0, id, keyCode, 0, -1);  
}


/*******************************************************************************
 * exported event functions
 */

/*
 * GRX has a somewhat restricted input model, where key press/release
 * events are reported as single KEY_PRESS events, and modifier keys
 * are not reported at all. We have to generate the KEY_PRESSED and
 * KEY_RELEASED artificially, with the modifier keys extracted from
 * the modifier field of 'ordinary' key events
 */

jobject
EXPORT(evtGetNextEvent) ( JNIEnv* env, jclass clazz )
{
	jobject jEvt = NULL;

	DBG( awt_evt, ("getNextEvent\n"));

	while ( 1 ) {
		/* check for already queued events, first */
		if ( X->removedKbStat ) {            /* synth. modifier KeyRelease */
			return modKeyNotify( env, X, KEY_RELEASED);
		}
		else if ( X->addedKbStat ) {         /* synth. modifier KeyPress */
			return modKeyNotify( env, X, KEY_PRESSED);
		}
		else if ( X->mouseMotion ) {
			X->mouseMotion = 0;
			return motionNotify( env, X);
		}
		else if ( X->keyDown ) {             /* char/func key KeyPress */
			X->keyDown = 0;
			return keyNotify( env, X, KEY_PRESSED);
		}
		else if ( X->keyUp ) {               /* synth. char/func key KeyRelease */
			X->keyUp = 0;
			return keyNotify( env, X, KEY_RELEASED);
		}
		else if ( X->buttonDown ) {
			X->buttonDown = 0;
			return buttonNotify( env, X, MOUSE_PRESSED);
		}
		else if ( X->buttonUp ) {
			X->buttonUp = 0;
			return buttonNotify( env, X, MOUSE_RELEASED);
		}
		else {
			/* we have eventually reached the real input query */
			GrMouseGetEventT( GR_M_EVENT | GR_M_POLL, &X->event, 0);

			if ( X->event.flags & GR_M_EVENT ) {
				/* check if we have a new/removed modifier we should turn into
				 * a event of its own
				 */
				if ( X->event.kbstat != X->lastKbStat ){
					getStateChanges( X);
					X->lastKbStat = X->event.kbstat;
				}
			
				/* We might get several Java events from a single native one,
				 * don't be exclusive
				 */
				if ( X->event.flags & GR_M_MOTION ) {
					X->mouseMotion = 1;
				}
				if ( X->event.flags & GR_M_BUTTON_DOWN ) {
					X->buttonDown = 1;
				}
				if ( X->event.flags & GR_M_BUTTON_UP ) {
					X->buttonUp = 1;
				}
				if ( X->event.flags & GR_M_KEYPRESS ) {
					X->keyDown = X->keyUp = 1;
				}
				
				continue;
			}
			else {
				/* process paint events after all input events */
				if ( X->common.pendingUpdates ) {
					return paintNotify( env, X);
				}
				else {
					return 0;
				}
			}
		}
	}

	return jEvt;
}


jobject
EXPORT(evtPeekEvent) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_evt, ("peekEvent\n"));

	return NULL;
}


/*
 * We can't do a XPeekIfEvent here because it would block until such an event
 * comes in
 */
jobject
EXPORT(evtPeekEventId) ( JNIEnv* env, jclass clazz, jint id )
{
	DBG( awt_evt, ("peekEventId: %d\n", id));

	return NULL;
}


/*
 * This is just used to wakeup the getNextEvent call if we are multithreaded,
 * and we post an event from outside the dispatcher thread (which might be blocked
 * in a native getNextEvent). This is a tribute to the fact that we actually have
 * two different queues - the native one, and the queue for self-posted events
 */
void
EXPORT(evtWakeup) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_evt, ("evtWakeup\n"));
}

/*
 * This is a helper for (IS_DISPATCH_EXCLUSIVE | NATIVE_DISPATCHER_LOOP) windowing
 * systems, where we need to execute certain requests (e.g. window creation) to
 * the native event thread. Not required here
 */
void
EXPORT(evtSendWMEvent) ( JNIEnv* env, jclass clazz, jobject wmEvt )
{
}
