/**
 * fnt.c - font support
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include "toolkit.h"

#include "../fnt-common.c"

/*******************************************************************************
 * internal functions
 */

void
initFonts ( Toolkit* X )
{
	char *fspec;
	char path[128];
	int  off = 3, bigger = 0, smaller = 4;
	struct stat fs;

	if ( (fspec = getenv( "GRXFONT")) ) {
		DBG( awt_fnt, ("initFonts: GRXFONT: \"%s\"\n", fspec));

		if ( sscanf( fspec, "%s %d %d %d", path, &off, &smaller, &bigger) ) {
	  
			if ( (stat( path, &fs) < 0) || !(fs.st_mode & S_IFDIR) ){
				fprintf( stderr, "GRXFONT fontpath not found: %s\n", path);
			}
			else {
				GrSetFontPath( path);
			}
		}
	}
	else {
		DBG( awt_fnt, ("initFonts: no GRXFONT env specifed\n"));
	}

	X->fsOffset = off;
	X->fsBigger = bigger;
	X->fsSmaller = smaller;
}



/*******************************************************************************
 * exported functions
 */

static char* defFont = "pc8x14.fnt";

void*
EXPORT(fntInitFont) ( JNIEnv* env, jclass clazz,
					  jstring jSpec, jint style, jint size )
{
	GrFont *fnt;
	char buf[128];
	char *spec = java2CString( env, _(X), jSpec);
	char *weight = "";
	char *slant  = "";
	int  i;

	DBG( awt_fnt, ("initFont: %x (%s),  %d,%d\n", jSpec, spec, style, size));

	if ( style & 0x1 ) weight = "b";
	if ( style & 0x2 ) slant  = "i";

	/*
	 * GRX fonts are simple bitmap fonts, which are looked up by filename - there
	 * is no font server that knows about point sizes or compensates different screen
	 * resolutions. We therefore need a mechanism to deal with unresolved
	 * requests. This is done by a environment variable
	 *  "GRXFONTS = <font-path> [<size-offset> [<max-decrease> [<max-increase>]]]"
	 * We first apply <size-offset> and see if we find a matching font file. If not,
	 * we iterate downwards from <size> + <size-offset> + <max-increase>..<max-decrease>
	 * in order to find a matching file
	 */

	size += X->fsOffset;
	sprintf( buf, spec, size, weight, slant);

	fnt = GrLoadFont( buf);
	DBG( awt_fnt, (" lookup: %s -> %x\n", buf, fnt));

	if ( !fnt ){
		for ( i=-X->fsBigger; i<X->fsSmaller && !fnt; i++ ) {
			if ( i != 0 ) {
				/* fprintf( stderr, "font not found: <%s>, looking for size: %d\n", buf, size-i); */
				sprintf( buf, spec, size-i, weight, slant);
				fnt = GrLoadFont( buf);
			}
		}

		if ( !fnt ) {
			/* fprintf( stderr, "font not found: <%s>, backup to: <%s>\n", buf, defFont); */
			fnt = GrLoadFont( defFont);
			if ( !fnt ) {  /* our last resort (in the hope that the GRX driver has a default) */
				fnt = &GrDefaultFont;
			}
		}
	}

	DBG( awt_fnt, (" -> font: %x\n", fnt));
  
	if ( !registerFs( fnt) ){
		return 0;  /* too many fonts in use */
	}

	return (void*) fnt;
}

void
EXPORT(fntFreeFont) ( JNIEnv* env, jclass clazz, GrFont* fnt )
{
	DBG( awt_fnt, ("freeFont: %x\n", fnt));

	if ( unregisterFs( fnt) ){
		GrUnloadFont( fnt);
	}
}


/*******************************************************************************
 * FontMetrics support
 */

void*
EXPORT(fntInitFontMetrics) ( JNIEnv* env, jclass clazz,
							 GrFont* fnt, jobject metrics )
{
	jclass fmClazz = (*env)->GetObjectClass( env, metrics);
	jfieldID id;

	DBG( awt_fnt, ("initFontMetrics: %x\n", fnt));

	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAdvance", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->maxwidth);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "leading", "I")) ){
		(*env)->SetIntField( env, metrics, id, 0);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "ascent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->h.baseline);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAscent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->h.baseline);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "descent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->h.height - fnt->h.baseline);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxDescent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->h.height - fnt->h.baseline);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "height", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->h.height);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "fixedWidth", "I")) ){
		(*env)->SetIntField( env, metrics, id,
							 (fnt->h.proportional) ? 0 : fnt->h.width);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "isWideFont", "Z")) ){
		(*env)->SetBooleanField( env, metrics, id,
								 (fnt->h.numchars > 256) ? JNI_TRUE:JNI_FALSE);
	}

	DBG( awt_fnt, (" -> metrics: %x\n", fnt));

	if ( !registerFs( fnt) ){
		return 0;  /* can't really happen, there must be a font for it */
	}

	return fnt;
}

void
EXPORT(fntFreeFontMetrics) ( JNIEnv* env, jclass clazz, GrFont* fnt )
{
	DBG( awt_fnt, ("initFontMetrics: %x\n", fnt));

	if ( unregisterFs( fnt) ){
		GrUnloadFont( fnt);
	}
}

jobject
EXPORT(fntGetWidths) ( JNIEnv* env, jclass clazz, GrFont* fnt )
{
	int          n = (fnt->h.numchars > 256) ? 256 : fnt->h.numchars;
	jarray       widths = (*env)->NewIntArray( env, 256);
	jboolean     isCopy;
	jint         *jw = (*env)->GetIntArrayElements( env, widths, &isCopy);
	register int i, j;

	DBG( awt_fnt, ("getWidths: %x\n", fnt));

	for ( i=fnt->h.minchar, j=0; j < n; i++, j++ )
		jw[i] = fnt->chrinfo[j].width;

	(*env)->ReleaseIntArrayElements( env, widths, jw, 0);

	return widths;
}

jint
EXPORT(fntBytesWidth) ( JNIEnv* env, jclass clazz, GrFont* fnt,
						jarray jBytes, jint off, jint len )
{
	jboolean  isCopy;
	jbyte    *jb = (*env)->GetByteArrayElements( env, jBytes, &isCopy);
	int       n = (*env)->GetArrayLength( env, jBytes);
	int       w;

	if ( off+len > n ) len = n - off;

	w = GrFontStringWidth( fnt, jb+off, len, GR_BYTE_TEXT);

	(*env)->ReleaseByteArrayElements( env, jBytes, jb, JNI_ABORT);

	DBG( awt_fnt, ("bytesWidth: %x \"%3s..\" -> %d\n", fnt, jb+off, w));

	return w;
}

jint
EXPORT(fntCharWidth) ( JNIEnv* env, jclass clazz, GrFont* fnt, jchar jChar )
{
	jint w = GrFontCharWidth( fnt, jChar);

	DBG( awt_fnt, ("charWidth: %x \'%c\' -> %d\n", fnt, jChar, w));

	return w;
}

jint
EXPORT(fntCharsWidth) ( JNIEnv* env, jclass clazz, GrFont* fnt,
						jarray jChars, jint off, jint len )
{
	jboolean  isCopy;
	jchar    *jc = (*env)->GetCharArrayElements( env, jChars, &isCopy);
	int      n = (*env)->GetArrayLength( env, jChars);
	int      w;

	if ( off+len > n ) len = n - off;

	w = GrFontStringWidth( fnt, (jchar*)(jc+off), len, GR_WORD_TEXT);

	(*env)->ReleaseCharArrayElements( env, jChars, jc, JNI_ABORT);

	DBG( awt_fnt, ("charWidths: %x \"%s\" -> %d\n", fnt,
				   jchar2CString( _(X),(jc+off),len), w));

	return w;
}

jint
EXPORT(fntStringWidth) ( JNIEnv* env, jclass clazz, GrFont* fnt, jstring jStr )
{
	jboolean isCopy;
	const jchar    *jc = (*env)->GetStringChars( env, jStr, &isCopy);
	int      len = (*env)->GetStringLength( env, jStr);
	int      w;

	w = GrFontStringWidth( fnt, (void*)jc, len, GR_WORD_TEXT);

	(*env)->ReleaseStringChars( env, jStr, jc);

	DBG( awt_fnt, ("stringWidth: %x \"%s\" -> %d\n", fnt,
				   java2CString( env,_(X),jStr), w));

	return w;
}
