/**
 * gra.c - platform specific drawing support (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <string.h>

#include "toolkit.h"

/*****************************************************************************
 * generic exported methods
 */

#define _graInitGraphics   EXPORT(graInitGraphics)
#define _graFreeGraphics   EXPORT(graFreeGraphics)

#define _graCopyArea       EXPORT(graCopyArea)
#define _graClearRect      EXPORT(graClearRect)
#define _graDrawLine       EXPORT(graDrawLine)
#define _graDrawArc        EXPORT(graDrawArc)
#define _graFillArc        EXPORT(graFillArc)
#define _graDrawOval       EXPORT(graDrawOval)
#define _graFillOval       EXPORT(graFillOval)
#define _graDrawRect       EXPORT(graDrawRect)
#define _graFillRect       EXPORT(graFillRect)
#define _graDrawPolygon    EXPORT(graDrawPolygon)
#define _graDrawPolyline   EXPORT(graDrawPolyline)
#define _graFillPolygon    EXPORT(graFillPolygon)

#define _graDrawRoundRect  EXPORT(graDrawRoundRect)
#define _graFillRoundRect  EXPORT(graFillRoundRect)
#define _graDraw3DRect     EXPORT(graDraw3DRect)
#define _graFill3DRect     EXPORT(graFill3DRect)

#define _graDrawBytes      EXPORT(graDrawBytes)
#define _graDrawChars      EXPORT(graDrawChars)
#define _graDrawString     EXPORT(graDrawString)

#define _graAddClip        EXPORT(graAddClip)
#define _graSetClip        EXPORT(graSetClip)
#define _graSetOffset      EXPORT(graSetOffset)
#define _graSetVisible     EXPORT(graSetVisible)
#define _graSetColor       EXPORT(graSetColor)
#define _graSetBackColor   EXPORT(graSetBackColor)
#define _graSetFont        EXPORT(graSetFont)
#define _graSetPaintMode   EXPORT(graSetPaintMode)
#define _graSetXORMode     EXPORT(graSetXORMode)

#define _graDrawImage      EXPORT(graDrawImage)

/*****************************************************************************
 * import generic Graphics functionality
 */

#include "../gra-common.c"



/******************************************************************************
 * platform specific image rendering
 */

#if defined(HAVE_X) /* workaround for grxX GetPixel bug */
void
drawAlphaImage ( Graphics* gr, Image* img,
				 jint srcX, jint srcY, jint dstX, jint dstY,
				 jint width, jint height, jint bgval )
{
  int       i, j, si, sj, alpha;
  unsigned long dpix, spix;
  int       sr, sg, sb, dr, dg, db;
  GrContext *dstImg;
  GrContext save;

  DBG( awt_gra, ("drawAlphaImage: %x %x (%x, %x),  %d,%d  %d,%d  %d,%d  %x\n",
				 gr, img, img->gc, img->common.alpha, srcX, srcY, dstX, dstY, width, height, bgval));

  GrSaveContext( &save);

  if ( (dstImg = GrCreateContext( width, height, 0, 0)) ){
	/*
	 * the GrBitBlt highly depends on the current GrContext, i.e. the
	 * GrBitBlt, GrSetContext shouldn't be changed (GRX bug??)
	 */
	GrBitBlt( dstImg, 0, 0, gr->gc, dstX, dstY, dstX+width, dstY+height, GrWRITE);
	GrSetContext( dstImg);

	for ( j=0, sj=srcY; j<height; j++, sj++ ) {
	  for ( i=0, si=srcX; i<width; i++, si++ ) {
		dpix = GrPixelC( dstImg, i, j);
		spix = GrPixelC( img->gc, si, sj);
		alpha = getAlpha( _(img)->alpha, si, sj);

		g_rgbValues( X, dpix, &dr, &dg, &db);
		g_rgbValues( X, spix, &sr, &sg, &sb);

		if ( alpha ) {
		  /*
		   * ints still considered to be substantially faster than floats on most
		   * architectures (avoid div by 255), map all 0..255 -> 1..256
		   */
		  alpha++;
		  dr++; dg++; db++;
		  sr++; sg++; sb++;

		  dr = ((alpha * sr + (256 - alpha) * dr) + 128) >> 8;
		  dg = ((alpha * sg + (256 - alpha) * dg) + 128) >> 8;
		  db = ((alpha * sb + (256 - alpha) * db) + 128) >> 8;

		  dr--; dg--; db--;

		  dpix = g_pixelValue( X, (dr << 16)|(dg << 8)|(db));
		  GrPlotNC( i, j, g_stripAlpha( dpix));
		}
	  }
	}

	GrSetContext( &save); /* order matters, see above */
	GrBitBlt( gr->gc, dstX, dstY, dstImg, 0, 0, width, height, GrWRITE);
	GrDestroyContext( dstImg);
  }
}

#else  /* !HAVE_X - this should be much faster with direct framebuffer access */

void
drawAlphaImage ( Graphics* gr, Image* img,
				 jint srcX, jint srcY, jint dstX, jint dstY,
				 jint width, jint height, jint bgval )
{
	int       di, dj, si, sj, alpha;
	unsigned long dpix, spix;
	int       sr, sg, sb, dr, dg, db;

	int       dimax = dstX + width;
	int       djmax = dstY + height;
	unsigned char ai;

	DBG( awt_gra, ("drawAlphaImage: %x %x (%x, %x),  %d,%d  %d,%d  %d,%d  %x\n",
				   gr, img, img->gc, img->common.alpha, srcX, srcY, dstX, dstY, width, height, bgval));

	for ( dj=dstY, sj=srcY; dj<djmax; dj++, sj++ ) {
		for ( di=dstX, si=srcX; di<dimax; di++, si++ ) {
			dpix = GrPixelC( gr->gc, di, dj);
			spix = GrPixelC( img->gc, si, sj);
			alpha = getAlpha( _(img)->alpha, si, sj);

			g_rgbValues( X, dpix, &dr, &dg, &db);
			g_rgbValues( X, spix, &sr, &sg, &sb);

			if ( alpha ) {
				/*
				 * ints still considered to be substantially faster than floats on most
				 * architectures (avoid div by 255), map all 0..255 -> 1..256
				 */
				alpha++;
				dr++; dg++; db++;
				sr++; sg++; sb++;

				ai = 256 - alpha;

				dr = ((alpha * sr  +  ai * dr) + 128) >> 8;
				dg = ((alpha * sg  +  ai * dg) + 128) >> 8;
				db = ((alpha * sb  +  ai * db) + 128) >> 8;

				dr--; dg--; db--;

				dpix = g_pixelValue( X, (dr << 16)|(dg << 8)|(db));
				GrPlotNC( di, dj, dpix);
			}
		}
	}
}

#endif /* HAVE_X */


void
EXPORT(graDrawImageScaled) ( JNIEnv* env, jclass clazz, 
							 Graphics* gr, Image* img,
							 jint dx0, jint dy0, jint dx1, jint dy1,
							 jint sx0, jint sy0, jint sx1, jint sy1,
							 jint bgval )
{
  int        x0, y0, x1, y1;
  int        iw = img->common.width;
  Image      tgt;

  if ( (!img) || (_(img)->width == -1) ){ /* last defense sanity check */
	return;
  }

  DBG( awt_gra, ("drawImageScaled: %x %x (%x), %d,%d,%d,%d, %d,%d,%d,%d, %x\n",
				 gr, img, img->gc, dx0, dy0, dx1, dy1, sx0, sy0, sx1, sy1, bgval));

  if ( dx1 > dx0 ) {
	x0 = dx0; x1 = dx1;
  }
  else {
	x0 = dx1; x1 = dx0;
  }
  if ( dy1 > dy0 ) {
	y0 = dy0; y1 = dy1;
  }
  else {
	y0 = dy1; y1 = dy0;
  }

  if ( sx0 < 0 ) sx0 = 0;
  if ( sx1 < 0 ) sx1 = 0;
  if ( sx0 >= iw ) sx0 = iw-1;
  if ( sx1 >= iw ) sx1 = iw-1;

  memset( (void*)&tgt, 0, sizeof(Image));
  tgt.common.width = (x1 - x0 +1);
  tgt.common.height = (y1 - y0 +1);

  tgt.gc = GrCreateContext( tgt.common.width, tgt.common.height, 0, 0);

  g_setAlphaType( &tgt, g_getAlphaType( img));
  initScaledImage( X, &tgt, img, dx0-x0, dy0-y0, dx1-x0, dy1-y0, sx0, sy0, sx1, sy1);

  Java_java_awt_Toolkit_graDrawImage ( env, clazz, gr, &tgt, 0, 0,
									   x0, y0, tgt.common.width, tgt.common.height,
									   bgval);

  GrDestroyContext( tgt.gc);

  if ( tgt.mask ){
	 GrDestroyContext( tgt.mask);
  }
  if ( tgt.common.mask ){
    freeImageMask( tgt.common.mask);
  }
  if ( tgt.common.alpha ) {
    freeImageAlpha( tgt.common.alpha);
  }
}
