/**
 * grlib.h - target driver library interface layer for no-native-wm AWTs
 *           (GRX 2.3 version)
 *
 * grlib.h contains most of AWT's "bottom interface" (aka DRIVER LIB
 * INTERFACE), which encapsulates the platform specific driver functions.
 *
 *  +----------------------------------------------------------------------+
 *  |Java class libraries                                                  |
 *  |                                                                      |
 *  | java.awt.Graphics (, Image, Window, Color, Font, ..)                 |
 *  |   drawLine() { Toolkit.graDrawLine() }                               |
 *  |                     |                                                |
 *  |  +----------------- | ---------------------------------------------+ |
 *  |  |java.awt.Toolkit  V                    [NATIVE AWT LIB INTERFACE]| |
 *  |  |  native graDrawLine(), wndCreateFrame(), clrGetPixelValue() ... | |
 *  |  +----------------- | -------------- | ---------- | ---------- | --+ |
 *  |                     |                |            |            |     |
 *  +-------------------- | -------------- | ---------- | ---------- | ----+
 *                        |                |            |            |
 *  +-------------------- | -------------- | ---------- | ---------- | ----+
 *  | native AWT library  |                V            V            V     |
 *  |              gra.c  V               wnd.c        clr.c        ...    |
 *  |                                                                      |
 *  |   Java_java_awt_Toolkit_graDrawLine(){ g_drawLine.. }                |
 *  |                                  +--------- | -----------------------+
 *  |                                  | grlib.h  V  [DRIVER LIB INTERFACE]|
 *  +----------------------------------+---- | ----------------------------+
 *                                           |
 *  +--------------------------------------- | ----------------------------+
 *  | driver library (GRX, MGL, UGL, ..)     V                             |
 *  |  (provides rasterizer)                                               |
 *  +----------------------------------------------------------------------+
 *
 * Unlike the "top interface" (NATIVE AWT LIB INTERFACE containing
 * native methods used from the unified AWT class library), it can be
 * implemented as a subset, i.e.there is no need to implement all it's
 * functions and types, if you don't use the relevant common module(s)
 * (like gra-common.c). All native methods are exported (and can be
 * overriden) from the platform specific ones (like gra.c).
 * 
 * This header is included from toolkit.h, which is in turn the glue header
 * for all native AWT modules.
 *
 * +-----------------------------------------------------------------------+
 * | gra.c (, wnd.c, img.c clr.c fnt.c..) (*)                              |
 * |  +------------------------------------------------------------------+ |
 * |  | toolkit.h (*)                                                    | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |  | toolkit-common.h                                            | | |
 * |  |  |   typedef struct CommonImage, CommonGraphics, CommonToolkit | | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |                                                                  | |
 * |  |  typedef struct Image, Graphics, Toolkit {CommonToolkit ct..}    | |
 * |  |                                                                  | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |  | grlib.h (*)                                                 | | |
 * |  |  |   g_setColor(){.}, g_drawLine(){ GrDrawLine() }, .       |  | | |
 * |  |  +--------------------------------------------------------- | -+ | |
 * |  +------------------------------------------------------------ | ---+ |
 * |                                                                |      |
 * |  // exported functions with generic implementation             |      |
 * |  #define _graDrawLine Java_java_awt_Toolkit_graDrawLine..      |      |
 * |                                                                |      |
 * |  +------------------------------------------------------------ | ---+ |
 * |  | gra-common.c (, wnd-common.c, img-common.c, clr-common.c..) |    | |
 * |  |  _graDrawLine() { .. g_drawLine().. }                       V    | |
 * |  +------------------------------------------------------------------+ |
 * |                                                                       |
 * |  // platform specific exported functions                              |
 * |  Java_java_awt_Toolkit_graDrawImageScaled() {..} ..                   |
 * |                                                                       |
 * +-----------------------------------------------------------------------+
 *
 * The (*) marked files are platform dependent
 *
 * Note that this header is included in two parts - the top half defining
 * just generic types (which can be used in the CommonXX structures, the
 * bottom half defining the driver interface functions (which are usually
 * inlined).
 *
 * Functions and types starting with a "g_" prefix (e.g. g_drawLine) are
 * used in the standard common modules (have a fixed meaning), all other
 * definitions are just internal helpers, and can only be used from the
 * platform specific modules (e.g. gra.c), or from within grlib.h itself.
 *
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#if !defined (GRLIB_FUNCS)
/* this section is inlcuded before toolkit-common.h, to define
 * generic types which can be used in the common structure definitions
 */

/***************************************************************************
 * scalar & reference types 
 */

/* the type of native pixel values */
typedef  GrColor    g_pixel_t;

/* the type to set color palettes */
typedef  GrColor    g_paletteEntry_t;

/* (optional) transparent graphics context used for context save/restore
 * operations (if the driver lib provides/requires them)
 */
typedef  GrContext  g_contextSav_t;

/* (transparent) font type */
typedef  GrFont     g_font_t;

/* the 'point' type of polygon related operations */
typedef struct _g_point_t {
	int x, y;
} g_point_t;

/* function to (type transparent) set point coordinates */
static inline void g_setPoint ( g_point_t* p, int x, int y )
{ 
	p->x = x; p->y = y;
}


#else /* defined GRLIB_FUNCS */
/* this is included before toolkit-common-post.h, i.e. we can
 * use these functions/macros there
 */

/***************************************************************************
 * internal helper (can be called from platform specific modules, but
 * NOT from common/generic code)
 */

/*
 * Draw a pixel in the current context. Note that - even if we explicitly
 * provide the gc - this HAS TO BE the current context (GRX pitfall)
 */
static inline void  _put_pixel_ ( GrContext* gc, int x, int y, g_pixel_t pixel )
{
	gc->gc_frame.gf_driver->drawpixel( x + gc->gc_xoffset,
									   y + gc->gc_yoffset, pixel);
}


/******************************************************************************
 * visual detection
 */

/*
 * get the number of bits per display pixel (= visual)
 */
static inline unsigned int g_getBitsPerPixel ( Toolkit* X )
{
	int bpp;
	unsigned int n = GrNumColors();

	for ( bpp=1; (1 << bpp) < n; bpp++ );
	return bpp;
}

/*
 * (optional) function to set system into RGB mode (some drivers
 * have to be switched explicitly)
 */
static inline void g_initRgbMode ( Toolkit* X )
{
	GrSetRGBcolorMode();
}


/******************************************************************************
 * pseudocolor palette (colormap) access & manipulation
 *
 * These macros are used (from clr.c) to set the system palette (we
 * need to evenly distribute RGBs, with additional capabilities to
 * allocate a few exact color requests).
 *
 * Note that some systems require to 'allocate' system color cells, i.e.
 * don't give us direct r/w access to a certain system palette entry/index.
 */

/*
 * Initialize system palette modification (reset, allocation etc.)
 */
static inline void g_initPalette ( CommonIdxMap* map )
{
	GrResetColors ();
}

/*
 * Set a system palette cell to the provided RGB components [0..255],
 * answer palette index where cell is stored (if the driver lib can't
 * change palette cells in-situ, and has to 'allocate' them).
 * 'idx' is the requested cell index (for in-situ changeable palettes)
 */
static inline int g_setPaletteEntry ( CommonIdxMap* map, int idx,
									  int r, int g, int b )
{
	return GrAllocColor( r, g, b);
}

/*
 * Query RGB components of palette cell at 'idx'
 */
static inline void g_queryPaletteEntry ( CommonIdxMap* map, int idx,
										 int* pr, int* pg, int* pb )
{
	GrQueryColor( idx, pr, pg, pb);
}

/*
 * Make (modified) palette active (some drivers refer to this as
 * 'realizing' the palette)
 */
static inline void g_setPalette ( CommonIdxMap* map )
{
	GrRefreshColors();
}

/*
 * add alpha channel [0..255] to jint pixel value
 */
static inline jint g_addAlpha ( jint rgb, g_pixel_t pix )
{
	return ((rgb & 0xff000000) | pix);
}

/*
 * get alpha channel [0..255] from jint pixel value
 */
static inline unsigned char g_getAlpha ( g_pixel_t pix )
{
	return (pix & 0xff000000) >> 24;
}

/*
 * remove alpha channel from jint pixel value
 */
static inline g_pixel_t g_stripAlpha ( g_pixel_t pix )
{
	return (pix & 0x00ffffff);
}


/*
 * Answer pixel value for given 8-8-8-8 ARGB
 * (remove switch for speed optimization)
 */
static inline jint g_pixelValue ( Toolkit* X, jint rgb )
{
	jint pix;

	switch ( _(X)->colorMode ) {
	case CM_TRUE:
		pix = (((rgb & _(X)->map.tclr->blueMask)  >> _(X)->map.tclr->blueShift) +
			   ((rgb & _(X)->map.tclr->greenMask) >> _(X)->map.tclr->greenShift) +
			   ((rgb & _(X)->map.tclr->redMask)   >> _(X)->map.tclr->redShift));
		break;
	case CM_PSEUDO:
		pix = _(X)->map.pclr->pix[ CELL_INDEX( JRED(rgb))  ]
			                     [ CELL_INDEX( JGREEN(rgb))]
			                     [ CELL_INDEX( JBLUE(rgb)) ];
		break;
	case CM_VGA:
		pix = getVGAPixelValue( X, rgb, 0);
		break;
	default:
		pix = 0;
	}

	return g_addAlpha( rgb, pix);
}


/*
 * Answer RGB components for given pixel value
 */
static inline void g_rgbValues ( Toolkit* X, unsigned long pixel, int* r, int* g, int* b )
{
	switch ( _(X)->colorMode ) {
	case CM_TRUE:
		*r = ((pixel & _(X)->map.tclr->pixRedMask)   << _(X)->map.tclr->redShift)   >> 16;
		*g = ((pixel & _(X)->map.tclr->pixGreenMask) << _(X)->map.tclr->greenShift) >> 8;
		*b = ((pixel & _(X)->map.tclr->pixBlueMask)  << _(X)->map.tclr->blueShift);
		break;
	default:
		GrQueryColor( pixel, r, g, b);
	}
}

/*
 * Answer pixel value with modified brightness
 */
static inline jint g_modPixelValue ( Toolkit* X, int pix, int mod )
{
	return modPixelValue( X, pix, mod);
}


/******************************************************************************
 * Graphics state - these functions are used to set the state of our
 * native cache 'Graphics' objects
 */

/*
 * Set global draw context according to the provided Image
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentImage ( Toolkit* X, Image* img )
{
	GrSetContext( img->gc);
}

/*
 * Set global draw context according to the provided Graphics
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentGraphics ( Toolkit* X, Graphics* gr )
{
	if ( X->lastGc != gr->gc ){
		GrSetContext( gr->gc);
		X->lastGc = gr->gc;
	}
}

/*
 * Temporarily store the current context (if the graphics driver
 * can't draw into aribtrary contexts)
 */
static inline void g_saveContext ( g_contextSav_t* cs )
{
	GrSaveContext( cs);
}

/*
 * Restore the current context from a temporarily saved one
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_restoreContext ( g_contextSav_t* cs )
{
	GrSetContext( cs);
}

#ifdef SPEED_OPTIMIZED
static void g_setClip ( Graphics* gr )
{
	if ( (_(gr)->curClip->x0 <= gr->gc->gc_xmax) && (_(gr)->curClip->x1 >= 0) &&
		 (_(gr)->curClip->y0 <= gr->gc->gc_ymax) && (_(gr)->curClip->y1 >= 0) ) {

		if ( _(gr)->curClip->x0 < 0 )
			_(gr)->curClip->x0 = 0;
		if ( _(gr)->curClip->y0 < 0 )
			_(gr)->curClip->y0 = 0;
		if ( _(gr)->curClip->x1 > gr->gc->gc_xmax )
			_(gr)->curClip->x1 = gr->gc->gc_xmax;
		if ( _(gr)->curClip->y1 > gr->gc->gc_ymax )
			_(gr)->curClip->y1 = gr->gc->gc_ymax;

		gr->gc->gc_xcliplo = _(gr)->curClip->x0;
		gr->gc->gc_ycliplo = _(gr)->curClip->y0;
		gr->gc->gc_xcliphi = _(gr)->curClip->x1;
		gr->gc->gc_ycliphi = _(gr)->curClip->y1;

		/* if this is the current GC, set it in effect now */
		if ( X->lastGc == gr->gc ){
			GrContext* gc = (GrContext*) GrCurrentContext();
			gc->gc_xcliplo = _(gr)->curClip->x0;
			gc->gc_ycliplo = _(gr)->curClip->y0;
			gc->gc_xcliphi = _(gr)->curClip->x1;
			gc->gc_ycliphi = _(gr)->curClip->y1;
		}
	}
}
#endif

/*
 * Set the active clip rectangle according to the provided Graphics object
 */
static inline void g_setClip ( Graphics* gr )
{
	if ( X->lastGc == gr->gc ){
		GrSetClipBox( _(gr)->curClip->x0, _(gr)->curClip->y0,
					  _(gr)->curClip->x1, _(gr)->curClip->y1);
	}
}

/*
 * one time init for new Graphics objects
 */
static inline void g_initGraphics ( Graphics* gr )
{
	/* nothing to do */
}

/*
 * free resources for Graphics object
 */
static inline void g_freeGraphics ( Graphics* g )
{
	/* nothing to do */
}


/*
 * Set the draw context of a Graphics object 'gr' from a Window object
 */
static inline void g_setWindowGraphics ( Graphics* gr, AwtWnd* wnd )
{
	gr->gc = (GrContext*) GrScreenContext();
}

/*
 * Set the draw context of a Graphics object 'gr' from a Image object
 */
static inline void g_setImageGraphics ( Graphics* gr, Image* img )
{
	gr->gc = img->gc;
}

/*
 * Set the draw context of a Graphics object 'gr' from a Graphics object
 */
static inline void g_setGraphicsGraphics ( Graphics* gr, Graphics *ogr )
{
	gr->gc = ogr->gc;
}

/*
 * Set the foreground color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setColor ( Graphics* gr, g_pixel_t clr )
{
	gr->alpha = g_getAlpha( clr);

	gr->fg = g_stripAlpha( clr);
	gr->clr = (gr->xclr != -1) ? ((gr->fg ^ gr->xclr) | GrXOR) : gr->fg;
	gr->txo.txo_fgcolor.v = gr->fg;

	/* we support some alpha operations for drawable Images */
	if ((_(gr)->tgtType ==IMAGE) && !_(IMAGE_TGT(gr))->alpha && (gr->alpha<0xff)){
		Image* img = IMAGE_TGT(gr);
		_(img)->alpha = createImageAlpha( X, _(img)->width, _(img)->height);
	}

	/* no need to set current color, it's a draw op parameter */
}

/*
 * Answer foreground color of Graphics object
 */
static inline g_pixel_t g_getColor ( Graphics* gr )
{
	return ((gr->alpha << 24) | gr->fg);
}

/*
 * Set the background color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 * This is a optional function for drivers which can't draw
 * text with transparent background
 */
static inline void g_setBackColor ( Graphics* gr, g_pixel_t clr )
{
	gr->bg = g_stripAlpha( clr);
}

/*
 * Set the font of a Graphics object
 * Note that we also might have to set the active font in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setFont ( Graphics* gr, g_font_t* fnt )
{
	gr->txo.txo_font      = fnt;
	gr->txo.txo_fgcolor.v = gr->fg;
	gr->txo.txo_bgcolor.v = GrNOCOLOR;
	gr->txo.txo_chrtype   = GR_WORD_TEXT;
	gr->txo.txo_direct    = GR_TEXT_DEFAULT;
	gr->txo.txo_xalign    = GR_ALIGN_LEFT;
	gr->txo.txo_yalign    = GR_ALIGN_BASELINE;

	/* no need to set current color, it's a draw op parameter */
}

/*
 * Set the draw mode of a Graphics object to 'overpaint'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setPaintMode ( Graphics* gr )
{
	gr->xclr = -1;
	gr->clr  = gr->fg;
}

/*
 * Set the draw mode of a Graphics object to 'xor'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setXORMode ( Graphics* gr, g_pixel_t xclr )
{
	gr->xclr = g_stripAlpha( xclr);
	gr->clr  = (gr->fg ^ gr->xclr) | GrXOR;
}


/******************************************************************************
 * drawing operations
 */

/*
 * Set color of pixel
 * (Make sure it is clipped according to the Graphics, since many
 * graphics drivers don't provide clipping for this operation)
 */
static inline void g_putPixel ( Graphics* gr, int x, int y)
{
	if ( (x >= gr->gc->gc_xcliplo) && (x <= gr->gc->gc_xcliphi) &&
		 (y >= gr->gc->gc_ycliplo) && (y <= gr->gc->gc_ycliphi) ){
		_put_pixel_( gr->gc, x, y, gr->clr);
	}
}

/*
 * Fill rectangle with background color of Graphics object
 */
static inline void g_clearRect ( Graphics* gr, int x0, int y0, int x1, int y1)
{
	GrFilledBox( x0, y0, x1, y1, gr->bg);
}

/*
 * Draw a line
 */
static inline void g_drawLine ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	GrLine ( x0, y0, x1, y1, gr->clr);
}

static inline void g_drawHLine ( Graphics* gr, int x0, int y0, int x1 )
{
	GrHLine ( x0, x1, y0, gr->clr);
}

static inline void g_drawVLine ( Graphics* gr, int x0, int y0, int y1 )
{
	GrVLine ( x0, y0, y1, gr->clr);
}


/*
 * Draw a rectangle (corner exclusive)
 */
static inline void g_drawRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	GrBox ( x0, y0, x1, y1, gr->clr);
}

/*
 * Fill a rectangle (corner inclusive)
 */
static inline void g_fillRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	GrFilledBox ( x0, y0, x1, y1, gr->clr);

	if ( (_(gr)->tgtType == IMAGE) && _(IMAGE_TGT(gr))->alpha ){
		fillImageAlpha( _(IMAGE_TGT(gr))->alpha, x0, y0, x1, y1, gr->alpha);
	}
}

/*
 * Draw open arc (i.e. no lines from arc endpoints to center of oval).
 * Note that this actually is a oval, not just a circle
 */
static inline void g_drawOpenArc ( Graphics* gr, int x, int y,
								   int ra, int rb, int a1, int a2 )
{
	GrEllipseArc( x, y, ra, rb, a1*10, (a1+a2)*10, GR_ARC_STYLE_OPEN, gr->clr);
}


/*
 * Fill closed arc (pie chart)
 */
static inline void g_fillArc ( Graphics* gr, int x, int y,
							   int ra, int rb, int a1, int a2 )
{
	GrFilledEllipseArc( x, y, ra, rb, a1*10, (a1+a2)*10,
						GR_ARC_STYLE_CLOSE2, gr->clr);
}

/*
 * Draw oval (ellipse)
 */
static inline void g_drawOval ( Graphics* gr, int x, int y, int ra, int rb )
{
	GrEllipse( x, y, ra, rb, gr->clr);
}

/*
 * Fill oval (ellispe)
 */
static inline void g_fillOval ( Graphics* gr, int x, int y, int ra, int rb )
{
	GrFilledEllipse( x, y, ra, rb, gr->clr);
}

/*
 * Draw single-byte text
 */
static inline void g_drawByteText ( Graphics* gr, char* str,
									int len, int x, int y )
{
	gr->txo.txo_chrtype = GR_BYTE_TEXT;
	GrDrawString( str, len, x, y, &gr->txo);
}

/*
 * Draw unicode text
 */
static inline void g_drawWideText ( Graphics* gr, jchar* str,
									int len, int x, int y )
{
	gr->txo.txo_chrtype = GR_WORD_TEXT;
	GrDrawString( str, len, x, y, &gr->txo);
}

/*
 * Draw polygon (lines connecting vertices in 'points')
 */
static inline void g_drawPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	GrPolyLine( nPoints, (void*)points, gr->clr);
}

/*
 * Fill polygon
 */
static inline void g_fillPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	GrFilledPolygon( nPoints, (void*)points, gr->clr);
}


/******************************************************************************
 * image drawing (bitblt related functions)
 */

/*
 * Bit block transfer in same context (Image or display)
 */
static inline void g_copyArea ( Graphics* gr, int x, int y,
								int xw, int yh, int xdelta, int ydelta)
{
	GrBitBlt( gr->gc, x+xdelta, y+ydelta, gr->gc, x, y, xw, yh, GrWRITE);
}

/*
 * Bit block transfer from Image context source (to Image or display)
 * This operation has to handle alpha channels. In case of a non-negative
 * 'bg' color, the transparent image pixels have to be set to this value
 */
static inline void g_drawImage ( Graphics* gr, Image* img, int dx, int dy,
								 int sx, int sy, int sxw, int syh, g_pixel_t bg )
{
	int dxw = dx + sxw - sx;
	int dyh = dy + syh - sy;

	int bc = GrMouseBlock( 0, dx, dy, dxw, dyh);

	if ( img->mask ){
		if ( bg != -1 )
			GrFilledBox( dx, dy, dxw, dyh, bg);

		GrBitBlt( gr->gc, dx, dy, img->mask, sx, sy, sxw, syh, GrAND);
		GrBitBlt( gr->gc, dx, dy, img->gc, sx, sy, sxw, syh, GrOR);
	}
	else if ( _(img)->alpha ){
		drawAlphaImage( gr, img, sx, sy, dx, dy, sxw-sx, syh-sy, bg);
	}
	else {
		GrBitBlt( gr->gc, dx, dy, img->gc, sx, sy, sxw, syh, GrWRITE);
	}

	GrMouseUnBlock( bc);
}

/*
 * grab pixels from a source Graphics, and punch them into a destination
 * Graphics (this is a generalized blit with a source that is not a
 * Java image)
 */
static inline void g_grabPixels ( Graphics* dgr, Graphics* sgr,
								  int sx, int sy, int w, int h )
{
	/* not yet */
}


/******************************************************************************
 * image processing
 *
 * The protocol for creating, initializing and rendering native images
 * has been adapted to support subsequent alpha channel processing (which
 * can be extended to offscreen images). The involved steps are:
 *
 *  g_initImage      Prepares Image object so that native pixel values
 *                   can be (permanently or temporarily) stored
 *
 *  g_setAlphaType+  Has to be called as soon as information about
 *                   alpha channels (none, masked, full) is available. 
 *                   After the call, (temporary or permanent) alpha
 *                   values can be stored for each pixel. Note that
 *                   it is possible to subsequently "increase" the
 *                   alpha type (none->masked->full, i.e. the function
 *                   can be called several times)
 *
 * g_putImagePixel+  set color, alpha information for a pixel
 *
 * g_completeImage   check if alpha info can be reduced (full->masked->none)
 *                   and transform any temporary (generic) information
 *                   into a platform specific format which is suitable
 *                   for rendering
 */

/*
 * Initialize a Image object so that we can set it's pixel values.
 * Note that we don't have alpha information, yet (i.e. the image
 * subsequently might be set to masked or full alpha)
 */
static inline int g_initImage ( Image* img )
{
	img->gc = GrCreateContext( _(img)->width, _(img)->height, 0, 0);
	return (img->gc != 0);
}


/*
 * Set alpha channel type of Image object. The supported 'transType'
 * values are NO_ALPHA, MASKED_ALPHA (transparent pixels), FULL_ALPHA.
 * After this operation, we have to be able to store pixel alpha
 * information (at least temporarily, to be later-on processed by
 * completeImage)
 */
static inline void g_setAlphaType ( Image* img, int alphaType )
{
	if ( alphaType && ! (_(img)->mask || _(img)->alpha) ){
		if ( alphaType == MASK_ALPHA ){
			_(img)->mask = createImageMask(  X, _(img)->width, _(img)->height);
		}
		else if ( alphaType == FULL_ALPHA ){
#if defined(USE_FULL_ALPHA)
			_(img)->alpha = createImageAlpha( X, _(img)->width, _(img)->height);
#else
			_(img)->mask = createImageMask(  X, _(img)->width, _(img)->height);
#endif
		}
	}
}

/*
 * set transparency color of the image in the native layer
 */
static inline void g_setTranspValue ( Image* img, g_pixel_t clr )
{
 
}

/*
 * Transform temporarily stored image information (both pixel color and
 * alpha values) into a representation which is suitable for drawing
 * the image. Free any temporary information.
 */
static inline void g_completeImage ( Image* img )
{
	extern void createMaskContext( Image* img );

	if ( _(img)->mask || (_(img)->alpha && !(needsFullAlpha(_(img)->alpha, 0.0))) ){
		createMaskContext( img);
	}
}


/*
 * Clean up the platform specific Image components
 */
static inline void g_freeImage ( Image* img )
{
	if ( img->gc ){
		GrDestroyContext( img->gc);
		img->gc = 0;
	}

	if ( img->mask ) {
		GrDestroyContext( img->mask);
		img->mask = 0;
	}	
}


/*
 * Get alpha value of image pixel [0..255]
 */
static inline int g_getPixelAlpha ( Image* img, int x, int y )
{
	if ( img->mask ){
		/* looks counter-intuitive, but we GrAND the mask source to
		 * blacken the visible parts
		 */
		return GrPixelC( img->mask, x, y) ? 0 : 0xff;
	}
	else if ( _(img)->alpha ){
		return getAlpha( _(img)->alpha, x, y);
	}
	else {
		return 0xff;
	}
}

/*
 * Answer alpha type of image. There are three supported
 * alpha type values
 *  - NO_ALPHA     : no alpha channel at all
 *  - MASKED_ALPHA : alpha = [0,255] (transparent pixels)
 *  - FULL_ALPHA   : alpha = [0..255]
 */
static inline int g_getAlphaType ( Image* img )
{
	if ( _(img)->mask || img->mask )
		return MASK_ALPHA;
	else if ( _(img)->alpha )
		return FULL_ALPHA;
	else
		return NO_ALPHA;
}


/*
 * Set pixel value in Image (with optional alpha value [0..255]).
 * This is called before g_completeImage(), i.e. it can use
 * temporary pixel/alpha storage
 */
static inline void g_putImagePixel ( Image* img , int x, int y,
									 g_pixel_t pix, unsigned char alpha )
{
	if ( alpha != 0xff ){
		if ( _(img)->alpha ){
			putAlpha( _(img)->alpha, x, y, alpha);
		}
		else if ( _(img)->mask ){
			unsetMask( _(img)->mask, x, y);
			pix = 0;
		}
	}

	_put_pixel_( img->gc, x, y, pix);
}

/*
 * Answer pixel value of already completed Image
 */
static inline g_pixel_t  g_getImagePixel ( Image* img, int x, int y )
{
	return GrPixelC( img->gc, x, y);
}


/******************************************************************************
 * event processing
 */

extern jchar keyChars[];
extern int   nKeyChars;
extern jbyte keyCodes[];
extern int   nKeyCodes;

/*
 * Initialize keyboard input
 */
static inline void g_initKeyboard ( Toolkit* X )
{
	X->keyChars  = keyChars;
	X->nKeyChars = nKeyChars;

	X->keyCodes  = keyCodes;
	X->nKeyCodes = nKeyCodes;
}

/*
 * Initialize moude input
 */
static inline void g_initMouse ( Toolkit* X )
{
	int spmult=1, spdiv=1, thresh=0, accel=0;
	char *mode;

	GrMouseEventMode( 1);
	GrMouseInit();
	GrMouseSetColors( _(X)->black, _(X)->white);
	GrMouseSetCursor( X->cursors[0]);
	GrMouseDisplayCursor();
	/*
	 * Since GRX does a lot of mouse mickey translation, we provide an
	 * optional GRXMOUSE=<mult> <div> <threshold> <accel> setting to utilize this
	 */
	if ( (mode = getenv( "GRXMOUSE")) ) {
		if ( sscanf( mode, "%d %d %d %d\n", &spmult, &spdiv, &thresh, &accel) != 4 ){
			fprintf( stderr, "illegal mouse mode setting: GRXMOUSE= %s\n", mode);	  
		}
		if ( spmult && spdiv ) {
			GrMouseSetSpeed( spmult, spdiv);
		}
		if ( thresh && accel ){
			GrMouseSetAccel( thresh, accel);
		}
	}
}

/*
 * Get current horizontal pointer location
 */
static inline int g_getMouseX ( Toolkit* X )
{
	return X->event.x;
}

/*
 * Get current vertical pointer location
 */
static inline int g_getMouseY ( Toolkit* X )
{
	return X->event.y;
}

/*
 * Get nmber of pressed mouse button
 */
static inline int g_getMouseBtn ( Toolkit* X )
{
	if ( X->event.buttons & GR_M_LEFT )        return 1;
	else if ( X->event.buttons & GR_M_MIDDLE ) return 2;
	else if ( X->event.buttons & GR_M_RIGHT )  return 3;
	else                                       return 0;
}

/*
 * Get Java keycode of current input event
 */
static inline int g_getKeyCode ( Toolkit* X )
{
	if ( X->event.key < X->nKeyCodes ){
		return X->keyCodes[X->event.key];
	}
	else {
		return ((X->event.key >= 'a') && (X->event.key <= 'z')) ?
			X->event.key ^ 0x20 : X->event.key;
	} 
}

/*
 * Get Java keychar of current input event
 */
static inline int g_getKeyChar ( Toolkit* X )
{
	if ( X->event.key < X->nKeyChars ){
		return X->keyChars[X->event.key];
	}
	else {
		return ( X->event.key < ' ' || X->event.key > 'z' ) ? 0 : X->event.key;
	}
}

/*
 * Make sure the next event will be a (artificial) mouse motion event
 * (with the last known pointer position). Used to retarget mouse for popups
 */
static inline void g_queueMouseMotion ( JNIEnv* env, Toolkit* X )
{
	X->event.x = _(X)->xMouse;
	X->event.y = _(X)->yMouse;

	X->mouseMotion = 1;
}

/*
 * notify event system that we have a pending paint request
 */
static inline void g_queuePaint ( JNIEnv* env, Toolkit* X)
{
	/* nothing to do */
}

#endif /* GRLIB_FUNCS */
