/**
 * img-common.c - common image manipulation functions
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define SINGLEFRAMEDONE 2
#define STATICIMAGEDONE 3
#define TOPDOWNLEFTRIGHT 2
#define COMPLETESCANLINES 4


Image*
createImage ( int width, int height, int isDrawable )
{
	Image *img = 0;
	
	if ( !(img = AWT_CALLOC( 1, sizeof( Image))) ){
	    return 0;
	}

	_(img)->trans      = -1;
	_(img)->width      = width;
	_(img)->height     = height;

    if ( isDrawable ){
		_(img)->flags |= IMG_IS_DRAWABLE;
	}

	DBG( awt_img, ("create: %d,%d\n", width, height));

	if ( (width > 0) && (height > 0) ){
		if ( ! g_initImage( img) ){
			AWT_FREE( img);
			img = 0;
		}
	}

	DBG( awt_img, (" -> img: %p\n", img));

	return img;
}


#if !defined(EXTERNAL_MASK)

ImageMask*
createImageMask ( Toolkit* X, int width, int height )
{
	int nBytes;
	int scanline = (width + MASKWORDSIZE-1) / MASKWORDSIZE;
	int nBytes = (MASKWORDSIZE/8) * scanline * height;  
	ImageMask *mimg;

	DBG( awt_img, ("createImageMask: %d,%d\n", width, height));

	if ( !(mimg = AWT_MALLOC( sizeof( ImageMask) + nBytes)) ){
	    return 0;
	}
	
	mimg->scanline = scanline;
	mimg->width    = width;
	mimg->height   = height;
	mimg->buf      = (char*)mimg + sizeof(ImageMask);
	
	memset( mimg->buf, 0xff, nBytes);

	DBG( awt_img, (" -> maskImage: %p\n", mimg));

	return mimg;
}

void
freeImageMask ( ImageMask* img )
{
	AWT_FREE( img);
}

#endif /* EXTERNAL_MASK */



#if !defined(EXTERNAL_ALPHA)

ImageAlpha*
createImageAlpha ( Toolkit* X, int width, int height )
{
	int nBytes = width * height;
	ImageAlpha    *aimg;

	DBG( awt_img, ("createImageAlpha: %d,%d\n", width, height));

	if ( !(aimg = AWT_MALLOC( sizeof( ImageAlpha) + nBytes)) ){
	    return 0;
	}

	aimg->width  = width;
	aimg->height = height;
	aimg->buf    = (char*)aimg + sizeof(ImageAlpha);

	memset( aimg->buf, 0xff, nBytes);

	DBG( awt_img, (" -> alphaImage: %p\n", aimg));

	return aimg;
}

void
freeImageAlpha ( ImageAlpha* img )
{
	AWT_FREE( img);
}

#endif /* EXTERNAL ALPHA */

void
fillImageAlpha ( ImageAlpha* aimg, int x0, int y0, int x1, int y1,
                 unsigned char a )
{
	int x, y;
	unsigned char* p = aimg->buf + (y0 * aimg->width) + x0;
	int d = aimg->width - x1 + x0-1;

	for ( y=y0; y <= y1; y++ ){
		for ( x=x0; x <= x1; x++ ){
			*p++ = a;
		}
		p += d;
	}
}

/*
 * For images with a full alpha channel, check if we really need an alpha byte for
 * each pel, or if a mask bitmap (alpha 0x00 / 0xff) will be sufficient
 */
int
needsFullAlpha ( ImageAlpha *aimg, double threshold )
{
	int i, j, a;
	int n = 0, max;

	max = (aimg->width * aimg->height) * threshold;

	for ( i=0; i < aimg->height; i++ ) {
		for ( j=0; j < aimg->width; j++ ) {
			a = getAlpha( aimg, j, i);
			if ( (a != 0) && (a != 0xff) ) {
				if ( ++n > max )
					return 1;
			}
		}
	}

	return 0;
}

void
countAlphas ( ImageAlpha *aimg, int* noAlpha, int* partAlpha, int* fullAlpha )
{
	int i, j, a;

	for ( i=0; i < aimg->height; i++ ) {
		for ( j=0; j < aimg->width; j++ ) {
			a = getAlpha( aimg, j, i);
			if ( a == 0 ) {
				(*noAlpha)++;
			}
			else if ( a == 0xff ) {
				(*fullAlpha)++;
			}
			else {
				(*partAlpha)++;
			}
		}
	}
}

/*
 * A full alpha image channel is way slower than using a mask bitmap (= 0 / 0xff alpha).
 * This function provides a simple alpha-to-mask translation. We pass the ImageAlpha
 * explicitly so that we don't have to drag it up to common (make sure it's
 * consistent)
 */
void
reduceAlpha ( Image* img, ImageAlpha* aimg, int threshold )
{
	int    i, j, a;
	g_pixel_t pix;

	DBG( awt_img, ("reduce image alpha: %p, [%p, %p] %p, %d\n",
				   img, _(img)->mask, _(img)->alpha,  aimg, threshold));

	if ( !aimg )
		return;

	for ( i=0; i < aimg->height; i++ ) {
		for ( j=0; j < aimg->width; j++ ) {
			a = (getAlpha( aimg, j, i) < threshold) ? 0 : 0xff;

			pix = g_getImagePixel( img, j, i);
			g_putImagePixel( img, j, i, pix, a);
		}
	}

	/* deallocation of the ImageAlpha has to be done by the client */
}


inline int
interpolate ( int ul, int ur, int ll, int lr, double dx, double dy )
{
	double u = ul + (double)(ur - ul) * dx;
	double l = ll + (double)(lr - ll) * dx;

	return (int) (u + (l - u) * dy  + 0.5);
}

void
getScaledPixel ( Toolkit* X, Image* img, int x, int y, double dx, double dy,
                 int trans, g_pixel_t* pix, unsigned char* alpha )
{
	g_pixel_t ul, ur, ll, lr;
	unsigned char   a0, a1, a2, a3;
	int    ulR, urR, llR, lrR, ulG, urG, llG, lrG, ulB, urB, llB, lrB, r, g, b;
	int    xi = (dx) ? x+1 : x;
	int    yi = (dy) ? y+1 : y;

	if ( xi >= _(img)->width )  xi = x;
	if ( yi >= _(img)->height ) yi = y;

	if ( trans ) {
		a0 = g_getPixelAlpha( img, x,   y);
		a1 = g_getPixelAlpha( img, xi,  y);
		a2 = g_getPixelAlpha( img, x,  yi);
		a3 = g_getPixelAlpha( img, xi, yi);

		*alpha = (unsigned char) interpolate( a0, a1, a2, a3, dx, dy);
	
		if ( (trans == MASK_ALPHA) && (*alpha < 128) ){
			*alpha = 0;
		}
		if ( ! *alpha ){  /* shortcut, but we loose color info */
			*pix = 0;
			return;
		}
	}
	else {
		a0 = a1 = a2 = a3 = 0;  /* make silly compiler happy */
		*alpha = 0xff;
	}

	ul = g_getImagePixel( img, x,  y);
	ur = g_getImagePixel( img, xi, y);
	ll = g_getImagePixel( img, x,  yi);
	lr = g_getImagePixel( img, xi, yi);

	if ( trans ) {
		if ( !a0 )	   ul = a1 ? ur : (a2 ? ll : lr);
		if ( !a1 )	   ur = a0 ? ul : (a3 ? lr : ll);
		if ( !a2 )    ll = a3 ? lr : (a0 ? ul : ur);
		if ( !a3 )    lr = a2 ? ll : (a1 ? ur : ul);
	}

	if ( (ul == ur) && (ll == ul) && (lr == ll) ) {
		g_rgbValues( X, ul, &r, &g, &b);
	}
	else {
		g_rgbValues( X, ul, &ulR, &ulG, &ulB);
		g_rgbValues( X, ur, &urR, &urG, &urB);
		g_rgbValues( X, ll, &llR, &llG, &llB);
		g_rgbValues( X, lr, &lrR, &lrG, &lrB);

		r = interpolate( ulR, urR, llR, lrR, dx, dy);
		g = interpolate( ulG, urG, llG, lrG, dx, dy);
		b = interpolate( ulB, urB, llB, lrB, dx, dy);
	}

	*pix = g_pixelValue( X, (r << 16) | (g << 8) | b);
}


void
initScaledImage ( Toolkit* X, Image *tgt, Image *src,
				  int dx0, int dy0, int dx1, int dy1,
				  int sx0, int sy0, int sx1, int sy1 )
{
	double         xScale, yScale, sX, sY, sxDelta, syDelta;
	int            dx, dy, dxInc, dyInc, sx, sy;
	unsigned char  alpha;
	int            trans = g_getAlphaType( src);
	g_pixel_t      pix;
	g_contextSav_t sav;

	g_setAlphaType( tgt, trans);

	/*
	 * this is a tribute to libs not providing a general PUT_PIXEL(context)
	 * function (like GRX)
	 */
	g_saveContext( &sav);
	g_setCurrentImage( X, tgt);

	DBG( awt_img, ("initScaledImage: tgt: %p, src: %p,  %d,%d,%d,%d %d,%d,%d,%d\n",
				   tgt, src, dx0,dy0,dx1,dy1, sx0,sy0,sx1,sy1));

	dxInc = (dx1 > dx0) ? 1 : -1;
	dyInc = (dy1 > dy0) ? 1 : -1;

	dx1 += dxInc;
	dy1 += dyInc;

	xScale = (double) (dx1 - dx0) / (double) (sx1 - sx0 +1);
	yScale = (double) (dy1 - dy0) / (double) (sy1 - sy0 +1);

	for ( dy=dy0; dy != dy1; dy += dyInc ) {
		sY = sy0 + (dy - dy0) / yScale;
		sy = (int) sY;
		syDelta = (sy < sy1) ? sY - sy : 0;

		for ( dx=dx0; dx != dx1; dx += dxInc ) {
			sX = sx0 + (dx - dx0) / xScale;
			sx = (int) sX;
			sxDelta = (sx < sx1) ? sX - sx : 0;
	
			getScaledPixel( X, src, sx, sy, sxDelta, syDelta, trans, &pix, &alpha);
			g_putImagePixel( tgt, dx, dy, pix, alpha);
		}
	}

	g_completeImage( tgt);
	g_restoreContext( &sav);
}


/************************************************************************************
 * exported image creation & destruction
 */

/*
 * this is called for non-NativeImageProducers, to initialize the native part
 * so that we can subsequently do setPixel() calls
 */
void*
_imgCreateImage ( JNIEnv* env, jclass clazz, jint width, jint height )
{
	Image *img = createImage(width, height, 0);
	
	DBG( awt_img, ("createImage: %d,%d -> %p\n", width, height, img));
	return img;
}


void
_imgFreeImage ( JNIEnv * env, jclass clazz, Image * img)
{
	Image *next, *first = img;

	/* we do this just once, since an image ring shares the same buffer */
	if ( _(img)->buffer ){
		Java_java_awt_Toolkit_imgFreeImage( env, clazz, _(img)->buffer);
	}
	
	/* we have to be aware of image rings (GIF movies), iterate */
	do {
		DBG( awt_img, ("freeImage: %p (%d,%d  %p%p,  %p,%p, %p\n",
					   img, _(img)->width, _(img)->height,
					   _(img)->mask, _(img)->alpha,
					   _(img)->buffer, _(img)->gr, _(img)->next));

#if !defined(EXTERNAL_MASK)					
		if ( _(img)->mask ){
			freeImageMask( _(img)->mask);
			_(img)->mask = 0;
		}
#endif

#if !defined(EXTERNAL_ALPHA)	
		if ( _(img)->alpha ) {
			freeImageAlpha( _(img)->alpha);
			_(img)->alpha = 0;
		}
#endif
		
		if ( _(img)->gr ){
			Java_java_awt_Toolkit_graFreeGraphics( env, clazz, _(img)->gr);
			_(img)->gr = 0;
		}

		g_freeImage( img); /* platform specific fields are cleaned up here */
				
		next = _(img)->next;
		AWT_FREE( img);
		img = next;
	} while ( (img != 0) && (img != first) );
}


void*
_imgCreateScreenImage ( JNIEnv* env, jclass clazz, jint width, jint height )
{
	Image *img;

	img = createImage( width, height, 1);
	g_setAlphaType( img, NO_ALPHA);

	DBG( awt_img, ("createScreenImage: %d,%d -> %p\n", (int)width, (int)height));

	return img;
}

void*
_imgCreateScaledImage ( JNIEnv* env, jclass clazz,
						Image* img, int width, int height )
{
	Image *scaledImg = createImage( width, height, FALSE);

	DBG( awt_img, ("createScaledImage: %p  %d,%d\n", img, width, height));

	initScaledImage ( X, scaledImg, img,
					  0, 0, width-1, height-1,
					  0, 0, _(img)->width-1, _(img)->height-1);

	return scaledImg;
}

/************************************************************************************
 * generic (ImageProducer) based image construction
 */

void
_imgSetIdxPels ( JNIEnv * env, jclass clazz, Image * img,
				 jint x, jint y, jint w, jint h,
				 jarray clrMap, jarray idxPels, jint trans,
				 jint off, jint scan)
{
	register int    row, col;
	jint            rgb;
	jboolean        isCopy;
	jint            *clr = (*env)->GetIntArrayElements( env, clrMap, &isCopy);
	jbyte           *pel = (*env)->GetByteArrayElements( env, idxPels, &isCopy);
	int             maxCol = x + w;
	int             maxRow = y + h;
	unsigned char   val;
	g_pixel_t       pix;
 	int             i = off;
	int             di = scan - w;
	g_contextSav_t  sav;

	g_setAlphaType( img, (trans != -1) ? MASK_ALPHA : NO_ALPHA);

	/*
	 * this is a tribute to libs not providing a general PUT_PIXEL(context)
	 * function (like GRX)
	 */
	g_saveContext( &sav);
	g_setCurrentImage( X, img);

	DBG( awt_img, ("setIdxPels: %p, %d,%d - %d,%d,  %p,%p,  %d, %d,%d\n",
				   img, (int)x,(int)y,(int)w,(int)h, clrMap,idxPels, (int)trans,
				   (int)off,(int)scan));

	for ( row = y; row < maxRow; row++, i+=di ) {
		for ( col = x; col < maxCol; col++, i++ ) {
			val = pel[i];
			rgb = clr[val];
			pix = g_pixelValue( X, rgb);

			g_putImagePixel( img, col, row, pix, 
							((trans != -1) && (val == trans)) ? 0 : 0xff);
		}
	}

	g_restoreContext( &sav);

	(*env)->ReleaseIntArrayElements( env, clrMap, clr, JNI_ABORT);
	(*env)->ReleaseByteArrayElements( env, idxPels, pel, JNI_ABORT);
}



void
_imgSetRGBPels ( JNIEnv * env, jclass clazz, Image * img,
				 jint x, jint y, jint w, jint h,
				 jarray rgbPels, jboolean hasAlpha, jint off, jint scan)
{
	int             row, col;
	jboolean        isCopy;
	jint            *rgbs = (*env)->GetIntArrayElements( env, rgbPels, &isCopy);
	int             maxCol = x + w;
	int             maxRow = y + h;
	jint		    val;
	unsigned char   alpha;
	g_pixel_t       pix;
 	int             i = off;
	int             di = scan - w;
	g_contextSav_t  sav;

#if defined (USE_FULL_ALPHA)
	g_setAlphaType( img, hasAlpha ? FULL_ALPHA : NO_ALPHA);
#else
	g_setAlphaType( img, hasAlpha ? MASK_ALPHA : NO_ALPHA);
#endif

	/*
	 * this is a tribute to libs not providing a general PUT_PIXEL(context)
	 * function (like GRX)
	 */
	g_saveContext( &sav);
	g_setCurrentImage( X, img);

	DBG( awt_img, ("setRGBPels: %p, %d,%d - %d,%d,  %p, %d,%d\n",
				   img, (int)x,(int)y,(int)w,(int)h, rgbPels, (int)off,(int)scan));

	for ( row = y; row < maxRow; row++, i+=di ) {
		for ( col = x; col < maxCol; col++, i++ ) {
			val = rgbs[i];
			pix = g_pixelValue( X, val);
			alpha = (unsigned char) (val >> 24);

			g_putImagePixel( img, col, row, pix, alpha);
		}
	}

	g_restoreContext( &sav);

	(*env)->ReleaseIntArrayElements( env, rgbPels, rgbs, JNI_ABORT);
}

void
_imgComplete( JNIEnv* env, jclass clazz, Image * img, jint status )
{
	DBG( awt_img, ("complete: %p %x\n", img, (unsigned)status));

	if ( status == SINGLEFRAMEDONE ){
		/* treat this a animation, regardless of img->next (it
		 * might produce several times into the same surface)
		 */
		_(img)->flags |= IMG_ANIMATION;
	}

	/*
	 * Check for alpha channel reduction, and/or transform generic
	 * (temporary) mask/alpha information into device dependent formats
	 */
	g_completeImage( img);	
}


/************************************************************************************
 * internal format production
 */

void
_imgProduceImage ( JNIEnv* env, jclass clazz,
				   jobject producer, Image* img, jboolean firstFrame )
{
	int       i, j;
	int	      r, g, b, alpha;
	g_pixel_t pix;
	jclass    prodClazz, modelClazz;
	jmethodID modelCtor, setPix, imgCompl;
	jmethodID setDim=0, setCM=0, setHints=0;
	jobject   model;
	jarray    pixArray;
	jint      *ipix = 0;
	jbyte     *bpix = 0;
	jboolean  isCopy;
	int       useIndexColorModel = 0;

	DBG( awt_img, ("produceImage: %p  prod: %p first: %d\n", img, producer, firstFrame));

	if ( (_(X)->colorMode == CM_PSEUDO) && (_(img)->trans != NO_TRANS_INDEX_AVAILABLE) ) {
		/* produce with a IndexColorModel, to mimic JDK behavior (there are
		 * apps out there using RgbFilters relying on getting IndexColorModel
		 * based setPixel() calls). Don't do that if the external image is not
		 * colortable based and uses all available system colors (trans=-2)
		 */
		useIndexColorModel = 1;
	}

	prodClazz = (*env)->GetObjectClass(env, producer);

	if ( useIndexColorModel ) {
		model = getSystemColorModel( env, clazz, X, _(img)->trans);
		setPix = (*env)->GetMethodID(env, prodClazz, "setPixels",
									 "(IIIILjava/awt/image/ColorModel;[BII)V");
 		/* allocate enough buffer space for a single scanline */
		pixArray = (*env)->NewByteArray(env, _(img)->width);
		bpix = (*env)->GetByteArrayElements(env, pixArray, &isCopy);
	}
	else {
		modelClazz = (*env)->FindClass(env, "kaffe.awt.JavaColorModel");
		modelCtor = (*env)->GetStaticMethodID(env, modelClazz, "getSingleton", 
											  "()Lkaffe/awt/JavaColorModel;");
		model = (*env)->CallStaticObjectMethod(env, modelClazz, modelCtor);

		setPix = (*env)->GetMethodID(env, prodClazz, "setPixels", 
									 "(IIIILjava/awt/image/ColorModel;[III)V");

		pixArray = (*env)->NewIntArray(env, _(img)->width);
		ipix = (*env)->GetIntArrayElements(env, pixArray, &isCopy);
	}

	imgCompl = (*env)->GetMethodID(env, prodClazz, "imageComplete", "(I)V");

	if ( firstFrame ) {
		setDim = (*env)->GetMethodID(env, prodClazz, "setDimensions", "(II)V");
		setCM = (*env)->GetMethodID(env, prodClazz, "setColorModel",
									"(Ljava/awt/image/ColorModel;)V");
		setHints = (*env)->GetMethodID(env, prodClazz, "setHints", "(I)V");
	}


	UNBLOCK_EXECUTE( env, clazz, {
		if ( firstFrame ) {
			(*env)->CallVoidMethod( env, producer, setDim, _(img)->width, _(img)->height);
			(*env)->CallVoidMethod( env, producer, setCM, model);
			(*env)->CallVoidMethod( env, producer, setHints, TOPDOWNLEFTRIGHT | COMPLETESCANLINES);
		}	

		for ( j=0; j < _(img)->height; j++ ){
			if ( useIndexColorModel ) {
				/* we are using a SystemColorModel derived one, so there is no
				 * need to translate pixels other than the transparent one
				 */
				for (i = 0; i < _(img)->width; i++) {
					if ( g_getPixelAlpha( img, i, j) != 0 ) {
						bpix[i] = g_getImagePixel( img, i, j);
					}
					else {
						bpix[i] = _(img)->trans;
					}
				}
				(*env)->ReleaseByteArrayElements(env, pixArray, bpix, JNI_COMMIT);
			}
			else {
				for ( i=0; i<_(img)->width; i++ ) {
					alpha = g_getPixelAlpha( img, i, j);
					pix   = g_getImagePixel( img, i, j);
					
					g_rgbValues( X, pix, &r, &g, &b);
					ipix[i] = ((alpha <<24) | (r << 16) | (g << 8) | b);
				}
				(*env)->ReleaseIntArrayElements(env, pixArray, ipix, JNI_COMMIT);
			}

			(*env)->CallVoidMethod( env, producer, setPix, 0, j,
									_(img)->width, 1, model,
									pixArray, 0, _(img)->width);
		}

		(*env)->CallVoidMethod(env, producer, imgCompl,
							   (_(img)->next) ? SINGLEFRAMEDONE : STATICIMAGEDONE);
	});

	if ( useIndexColorModel ){
		(*env)->ReleaseByteArrayElements(env, pixArray, bpix, JNI_ABORT);
	}
	else {
		(*env)->ReleaseIntArrayElements(env, pixArray, ipix, JNI_ABORT);
	}
}


/************************************************************************************
 * native format production dispatcher routines (GIF, jpeg, png, ..)
 */

/* interfaces of image conversion functions */
Image* readGifFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readGifData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readJpegFile ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readJpegData ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readPngFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readPngData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readTiffFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readTiffData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );
Image* readXbmFile  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, int fd, char* fname );
Image* readXbmData  ( JNIEnv* env, jobject sync, jobject prod, jmethodID setDim, jmethodID fillBuf,
					  unsigned char*, int off, int len );

#define SIG_LENGTH 4
#define SIG_GIF    1
#define SIG_JPEG   2
#define SIG_PNG    3
#define SIG_TIFF   4
#define SIG_XBM    5
#define SIG_SWF    6

Image *unknownImage = 0;  /* fill in some default image here */

int imageFormat ( unsigned char* sig ) {
	if ( (sig[0] == 'G') && (sig[1] == 'I') && (sig[2] == 'F') )
		return SIG_GIF;

	if ( (sig[0] == 0xff) && (sig[1] == 0xd8) && (sig[2] == 0xff) &&
		 ((sig[3] == 0xe0) || (sig[3] == 0xc4)) )
		return SIG_JPEG;

	if ( (sig[0] == 0x89) && (sig[1] == 'P') && (sig[2] == 'N') && (sig[3] == 'G') )
		return SIG_PNG;

	if ( (((sig[0] == 'I') && (sig[1] == 'I' )) ||
		  ((sig[0] == 'M') && (sig[1] == 'M'))) && *(short*)(sig+2) == 42 )
		return SIG_TIFF;
	
	if ( (sig[0] == '#') && (sig[1] == 'd') &&
		 (sig[2] == 'e') && (sig[3] == 'f') )
		return SIG_XBM;

	if ( (sig[0] == 'F') && (sig[1] == 'W') && (sig[2] == 'S') )
		return SIG_SWF;

	return 0;
}

void*
_imgCreateFromFile ( JNIEnv* env, jclass clazz, 
					 jobject producer, jstring fileName )
{
    jclass prodCls;
	jmethodID setDim;

	Image *img = 0;
	int   infile;
	char  *fn = java2CString( env, _(X), fileName);
	unsigned char  sig[SIG_LENGTH];

	DBG( awt_img, ("createFromFile: %s\n", fn));

	if ( !(prodCls = (*env)->GetObjectClass( env, producer)) ){
	    return 0;
	}
	if ( !(setDim = (*env)->GetMethodID( env, prodCls, "setDimensions", "(II)V")) ){
	    return 0;
	}	

	if ((infile = AWT_OPEN(fn)) >= 0) {
		if (AWT_READ(infile, sig, sizeof(sig)) == sizeof(sig)) {
			AWT_REWIND(infile);              /* some native converters can't skip the signature read */

			switch (imageFormat(sig)) {
			case SIG_GIF:
				img = readGifFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_JPEG:
				img = readJpegFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_PNG:
				img = readPngFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_TIFF:
				img = readTiffFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_XBM:
				img = readXbmFile( env, clazz, producer, setDim, infile, fn);
				break;
			case SIG_SWF:
				/* not yet */
			default:
				img = unknownImage;
			}
		}
		AWT_CLOSE(infile);
	}

	DBG( awt_img, (" -> img: %p %d,%d\n", img,
				   (img) ? _(img)->width : 0, (img) ? _(img)->height : 0));

	return img;
}


void *
_imgCreateFromData( JNIEnv * env, jclass clazz,
					jobject producer, jarray jbuffer, jint off, jint len)
{
    jclass prodCls;
    jmethodID fillBuf;
	jmethodID setDim;

	Image *img = 0;
	jboolean isCopy;
	jint   length = (*env)->GetArrayLength( env, jbuffer);
	jbyte *jb;

	DBG( awt_img, ("createFromData: %p, %d\n", jbuffer, (int)len));

	if ( !(prodCls = (*env)->GetObjectClass( env, producer)) ){
	    return 0;
	}
	if ( !(fillBuf = (*env)->GetMethodID( env, prodCls, "fillBuffer", "()I")) ){
	    return 0;
	}
	if ( !(setDim = (*env)->GetMethodID( env, prodCls, "setDimensions", "(II)V")) ){
	    return 0;
	}

	jb = (*env)->GetByteArrayElements(env, jbuffer, &isCopy);

	/* in case of a buffer overrun, we probably have a JPEG read error, anyway */
	if ( (off+len) <= length) {
		switch (imageFormat(jb+off)) {
		case SIG_GIF:
			img = readGifData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_JPEG:
			img = readJpegData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_PNG:
			img = readPngData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_TIFF:
			img = readTiffData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_XBM:
			img = readXbmData( env, clazz, producer, setDim, fillBuf, jb, off, len);
			break;
		case SIG_SWF:
			/* not yet */
		default:
			img = unknownImage;
		}
	}

	(*env)->ReleaseByteArrayElements( env, jbuffer, jb, JNI_ABORT);

	DBG( awt_img, (" -> img: %p %d,%d\n", img,
				   (img) ? _(img)->width : 0, (img) ? _(img)->height : 0));

	return img;  
}


/************************************************************************************
 * get / set image attributes
 */

void*
_imgSetFrame ( JNIEnv* env, jclass clazz, Image* img, int frameNo )
{
	Image *imgCur = img;

	if ( !_(img)->next )
		return img;

	while ( _(img)->frame != frameNo ) {
		img = _(img)->next;

		if ( img == imgCur )
			break;
	}

	return img;
}

jint
_imgGetWidth ( JNIEnv * env, jclass clazz, Image * img)
{
	return _(img)->width;
}

jint
_imgGetHeight ( JNIEnv * env, jclass clazz, Image * img)
{
	return _(img)->height;
}

jint
_imgGetFrameCount (JNIEnv * env, jclass clazz, Image * img)
{
	return _(img)->nFrames;
}

jint
_imgGetLoopCount (JNIEnv * env, jclass clazz, Image * img)
{
	return _(img)->nLoops;
}

jint
_imgGetLatency ( JNIEnv* env, jclass clazz, Image* img)
{
	return _(img)->latency;
}

void*
_imgGetNextFrame ( JNIEnv* env, jclass clazz, Image* img )
{
	return _(img)->next;   /* next in the ring */
}
