/**
 * fnt.c - font support
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include "toolkit.h"


/*******************************************************************************
 * internal functions
 */

void
initFonts ( Toolkit* X )
{
	/* nothing to do */
}

/*******************************************************************************
 * exported functions
 */

static char* defFont = "pc8x14.fnt";
static int szDeltas[] = { 0, -1, +1, -2, +2 };
static int nDeltas = (sizeof szDeltas / sizeof( szDeltas[0]));

void*
Java_java_awt_Toolkit_fntInitFont ( JNIEnv* env, jclass clazz,
									jstring jSpec, jint style, jint size )
{
	font_lib_t *fontLib;
	font_t     *font = 0;
	char buf[128];
	char *spec = java2CString( env, _(X), jSpec);
	char *mod = 0;
	float ptSize;
	float slant = 0.0;
	float angle = 0.0;
	ibool antiAliased = (_(X)->colorMode == CM_TRUE);
	int  i;

	i = sprintf( buf, spec, mod);

	DBG( awt_fnt, ("initFont: %x (%s),  %d,%d\n", jSpec, spec, style, size));

	if ( style == 3 )    mod = "bi";
	else if (style == 2) mod = "i";
	else if (style == 1) mod = "b";
	else                 mod = "";

	/* If we want pointSize offsets (because of different screen resolutions),
	 * we have to append a "+n" or "-n" to the font.properties entry. We
	 * assume that we have just a one digit offset
	 */
	if ( buf[i-2] == '-' ){
		size -= (buf[i-1]) - '0';
		buf[i-2] = 0;
	}
	else if ( buf[i-2] == '+' ){
		size += (buf[i-1]) - '0';
		buf[i-2] = 0;
	}

	/* ?? maybe we should cache fonts so that we don't have to
	 * reopen fontLibs all the time
	 */
	if ( (fontLib = MGL_openFontLib( buf)) ){
		/* in case we don't have a direct match, try to vary the size */
		for ( i=0; i<nDeltas && !font; i++ ){
			ptSize = size + szDeltas[i];
			font = MGL_loadFontInstance( fontLib, ptSize, slant, angle, antiAliased);
		}

		MGL_closeFontLib( fontLib);
	}

	if ( !font ){
		/* try the defFont as a last resort */
		/* fprintf( stderr, "font not found: <%s>, backup to: <%s>\n", buf, defFont); */
		font = MGL_loadFont( defFont);
	}

	DBG( awt_fnt, (" -> font: %x\n", font));
  
	return (void*) font;
}

void
Java_java_awt_Toolkit_fntFreeFont ( JNIEnv* env, jclass clazz, font_t* fnt )
{
	DBG( awt_fnt, ("freeFont: %x\n", fnt));

	MGL_unloadFontInstance( fnt);
}


/*******************************************************************************
 * FontMetrics support
 */

void*
Java_java_awt_Toolkit_fntInitFontMetrics ( JNIEnv* env, jclass clazz, font_t* fnt )
{
	jclass fmClazz = (*env)->GetObjectClass( env, metrics);
	jfieldID id;

	DBG( awt_fnt, ("initFontMetrics: %x\n", fnt));

	if ( fnt->fontType == MGL_VECTORFONT ){
		return 0;  /* not supported */
	}

	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAdvance", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->maxWidth);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "leading", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->leading);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "ascent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->ascent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAscent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->ascent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "descent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->descent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxDescent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->descent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "height", "I")) ){
		(*env)->SetIntField( env, metrics, id, fnt->fontHeight);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "fixedWidth", "I")) ){
		(*env)->SetIntField( env, metrics, id,
							 (fnt->fontType == MGL_FIXEDFONT) ? 0:fnt->maxWidth);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "isWideFont", "Z")) ){
		(*env)->SetBooleanField( env, metrics, id, JNI_FALSE); /* not yet */
	}

	DBG( awt_fnt, (" -> metrics: %x\n", fnt));

	return fnt;
}

void
Java_java_awt_Toolkit_fntFreeFontMetrics ( JNIEnv* env, jclass clazz, font_t* fnt )
{
	DBG( awt_fnt, ("freeFontMetrics: %x\n", fnt));
	/* nothing to do, it's a font_t */
}

/*
 * with respect to width calculation, we suffer a bit from two MGL drawbacks:
 * (1) - we can just query for the current font
 * (2) - all string-based queries reflect on '0' terminated strings
 */

jobject
Java_java_awt_Toolkit_fntGetWidths ( JNIEnv* env, jclass clazz, font_t* fnt )
{
	int          n = 256;
	font_t       *curFont;
	jarray       widths = (*env)->NewIntArray( env, 256);
	jboolean     isCopy;
	jint         *jw = (*env)->GetIntArrayElements( env, widths, &isCopy);
	register int i;

	DBG( awt_fnt, ("getWidths: %x\n", fnt));

	if ( (curFont = MGL_getFont()) != fnt ){
		MGL_useFont( fnt);
	}

	for ( i=0; i < n; i++ ){
		jw[i] = MGL_charWidth( i);
	}

	if ( curFont != fnt ){
		MGL_useFont( curFont);
	}

	(*env)->ReleaseIntArrayElements( env, widths, jw, 0);

	return widths;
}


jint
Java_java_awt_Toolkit_fntBytesWidth ( JNIEnv* env, jclass clazz, font_t* fnt,
									  jarray jBytes, jint off, jint len )
{
	font_t    *curFont;
	jboolean  isCopy;
	char      *buf;
	jbyte     *jb = (*env)->GetByteArrayElements( env, jBytes, &isCopy);
	int       n = (*env)->GetArrayLength( env, jBytes);
	int       w;

	if ( off+len > n )
		len = n - off;

	buf = jbyte2CString( _(X), jb+off, len);

	if ( (curFont = MGL_getFont()) != fnt ){
		MGL_useFont( fnt);
		w = MGL_textWidth( buf);
		MGL_useFont( curFont);
	}
	else {
		w = MGL_textWidth( buf);
	}

	(*env)->ReleaseByteArrayElements( env, jBytes, jb, JNI_ABORT);

	DBG( awt_fnt, ("bytesWidth: %x \"%3s..\" -> %d\n", fnt, jb+off, w));

	return w;
}

jint
Java_java_awt_Toolkit_fntCharWidth ( JNIEnv* env, jclass clazz, font_t* fnt, jchar jChar )
{
	int       w;
	font_t    *curFont;

	if ( (curFont = MGL_getFont()) != fnt ){
		MGL_useFont( fnt);
		w = MGL_charWidth_W( jChar);
		MGL_useFont( curFont);
	}
	else {
		w = MGL_charWidth_W( jChar);
	}

	DBG( awt_fnt, ("charWidth: %x \'%c\' -> %d\n", fnt, jChar, w));

	return w;
}

jint
Java_java_awt_Toolkit_fntCharsWidth ( JNIEnv* env, jclass clazz, font_t* fnt,
									  jarray jChars, jint off, jint len )
{
	jboolean  isCopy;
	jchar    *jc = (*env)->GetCharArrayElements( env, jChars, &isCopy);
	int      n = (*env)->GetArrayLength( env, jChars);
	int      w;
	wchar_t  *buf;
	font_t   *curFont;

	if ( off+len > n )
		len = n - off;

	buf = (wchar_t*) jchar2WString( _(X), jc+off, len);

	if ( (curFont = MGL_getFont()) != fnt ){
		MGL_useFont( fnt);
		w = MGL_textWidth_W( buf);
		MGL_useFont( curFont);
	}
	else {
		w = MGL_textWidth_W( buf);
	}

	(*env)->ReleaseCharArrayElements( env, jChars, jc, JNI_ABORT);

	DBG( awt_fnt, ("charWidths: %x \"%s\" -> %d\n", fnt,
				   jchar2CString( _(X),(jc+off),len), w));

	return w;
}

jint
Java_java_awt_Toolkit_fntStringWidth ( JNIEnv* env, jclass clazz, font_t* fnt, jstring jStr )
{
	jboolean isCopy;
	jchar    *jc = (jchar*) (*env)->GetStringChars( env, jStr, &isCopy);
	int      len = (*env)->GetStringLength( env, jStr);
	int      w;
	wchar_t  *buf;
	font_t   *curFont;

	/* oh yes, we have to - it might be a substring */
	buf = (wchar_t*) jchar2WString( _(X), jc, len);

	if ( (curFont = MGL_getFont()) != fnt ){
		MGL_useFont( fnt);
		w = MGL_textWidth_W( buf);
		MGL_useFont( curFont);
	}
	else {
		w = MGL_textWidth_W( buf);
	}

	(*env)->ReleaseStringChars( env, jStr, jc);

	DBG( awt_fnt, ("stringWidth: %x \"%s\" -> %d\n", fnt,
				   java2CString( env, _(X), jStr), w));

	return w;
}
