/**
 * grlib.h - target graphics library interface layer for no-native-wm AWTs
 *
 * Copyright (c) 1998
 *		Transvirtual Technologies, Inc.	 All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#if !defined (GRLIB_FUNCS)
/* this section is inlcuded before toolkit-common-pre.h, to define
 * generic types which can be used in the common structure definitions
 */

/***************************************************************************
 * scalar & reference types 
 */

/* the type of native pixel values */
typedef	 color_t	g_pixel_t;

/* the type to set color palettes */
typedef	 palette_t	g_paletteEntry_t;

/* (optional) transparent graphics context used for context save/restore
 * operations (if the driver lib provides/requires them)
 */
typedef	 void*	g_contextSav_t;

/* (transparent) font type */
typedef	 font_t		g_font_t;

/* the 'point' type of polygon related operations */
typedef point_t	 g_point_t;

/* function to (type transparent) set point coordinates */
static inline void g_setPoint ( g_point_t* p, int x, int y )
{ 
	p->x = x; p->y = y;
}


#else /* defined GRLIB_FUNCS */
/* this is included before toolkit-common-post.h, i.e. we can
 * use these functions/macros there
 */

typedef struct {
    unsigned high : 4;
    unsigned low  : 4;
} halfbytes;

/*
 * internal macro to set (display depth) bitmap pixels
 * (eliminate switch for maximum performance)
 */
static inline void _put_bitmap_pixel_ ( bitmap_t* bmp, int x, int y, g_pixel_t c )
{
	switch ( _(X)->nColorPlanes ){
	case 4: {
		halfbytes *p = (halfbytes*) ((u8*)(bmp->surface) +
									 (y * bmp->bytesPerLine) + x/2);
		if ( x & 0x1 ) p->low  = (c & 0xf);
		else           p->high = (c & 0xf);
		break;
	}
	case 8: {
		u8 *p = (u8*)(bmp->surface) + (y * bmp->bytesPerLine) + x;
		*p = (u8) c;
		break;
	}
	case 16: {
		u16 *p = (u16*)(bmp->surface) + (y * bmp->bytesPerLine/2) + x;
		*p = (u16) c;
		break;
	}
	case 24: {
		u8  *pb;
		u16 *p = (u16*)((u8*)(bmp->surface) + (y * bmp->bytesPerLine) + (x *3));
		*p = (u16) (c & 0x00ffff00) >> 8;
		
		pb = (u8*)p + 2;
		*pb = (u8) (c & 0xff);
		break;
	}
	case 32: {
		u32 *p = (u32*)(bmp->surface) + (y * bmp->bytesPerLine/4) + x;
		*p = (u32)c;
		break;
	}
	default:
	}
}

/*
 * internal macro to query (display depth) bitmap pixels
 */
static inline g_pixel_t _get_bitmap_pixel_ ( bitmap_t* bmp, int x, int y )
{
	switch ( _(X)->nColorPlanes ){
	case 4: {
		halfbytes *p = (halfbytes*) ((u8*)(bmp->surface)
									 + (y * bmp->bytesPerLine) + x/2);
		if ( x & 0x1 )  return p->low;
		else            return p->high;
	}
	case 8: {
		u8 *p = (u8*)(bmp->surface) + (y * bmp->bytesPerLine) + x;
		return *p;
	}
	case 16: {
		u16 *p = (u16*)(bmp->surface) + (y * bmp->bytesPerLine/2) + x;
		return *p;
	}
	case 24: {
		u16 *p = (u16*)((u8*)(bmp->surface) + (y * bmp->bytesPerLine) + (x *3));
		u8  *pb = (u8*)p + 2;
		return ((*p << 8) | *pb);
	}
	case 32: {
		u32 *p = (u32*)(bmp->surface) + (y * bmp->bytesPerLine/4) + x;
		return *p;
	}
	default:
		return 0;
	}
}

static inline bitmap_t* _create_bitmap_ ( int width, int height )
{
	bitmap_t *bmp;
	unsigned bytesPerLine = ( width * _(X)->nColorPlanes) + 31 / 8;
	unsigned nBytes = bytesPerLine * height;

	bmp = AWT_MALLOC( sizeof(bitmap_t) + nBytes);
	bmp->surface = bmp + sizeof(bitmap_t);
	bmp->width   = width;
	bmp->height  = height;
	bmp->bytesPerLine = bytesPerLine;
	bmp->bitsPerPixel = _(X)->nColorPlanes;
	bmp->pf = X->pf;
	bmp->pal = X->pal;

	return bmp;
}

/******************************************************************************
 * visual detection
 */

/*
 * get the number of bits per display pixel (= visual)
 */
static inline unsigned int g_getBitsPerPixel ( Toolkit* X )
{
	return MGL_getBitsPerPixel( X->gc);
}

/*
 * (optional) function to set system into RGB mode (some drivers
 * have to be switched explicitly)
 */
static inline void g_initRgbMode ( Toolkit* X )
{
}


/******************************************************************************
 * pseudocolor palette (colormap) access & manipulation
 *
 * These macros are used (from clr.c) to set the system palette (we
 * need to evenly distribute RGBs, with additional capabilities to
 * allocate a few exact color requests).
 *
 * Note that some systems require to 'allocate' system color cells, i.e.
 * don't give us direct r/w access to a certain system palette entry/index.
 */

/*
 * Initialize system palette modification (reset, allocation etc.)
 */
static inline void g_initPalette ( CommonIdxMap* map )
{
	map->clrs = AWT_ALLOCA( map->nClrs * sizeof(g_paletteEntry_t));
}

/*
 * Set a system palette cell to the provided RGB components [0..255],
 * answer palette index where cell is stored (if the driver lib can't
 * change palette cells in-situ, and has to 'allocate' them).
 * 'idx' is the requested cell index (for in-situ changeable palettes)
 */
static inline int g_setPaletteEntry ( CommonIdxMap* map, int idx,
									  int r, int g, int b )
{
	map->clrs[idx].blue	 = b;
	map->clrs[idx].green = g;
	map->clrs[idx].red	 = r;

	return idx;
}

/*
 * Query RGB components of palette cell at 'idx'
 */
static inline void g_queryPaletteEntry ( CommonIdxMap* map, int idx,
										 int* pr, int* pg, int* pb )
{
	*pr = map->clrs[idx].red;
	*pg = map->clrs[idx].green;
	*pb = map->clrs[idx].blue;
}

/*
 * Make (modified) palette active (some drivers refer to this as
 * 'realizing' the palette)
 */
static inline void g_setPalette ( CommonIdxMap* map )
{
	MGL_setPalette( X->gc, map->clrs, map->nClrs, 0);
	/* we keep the palette around for bitmap creation */
}

/*
 * add alpha channel [0..255] to jint pixel value
 */
static inline jint g_addAlpha ( jint rgb, g_pixel_t pix )
{
  return ((rgb & 0xff000000) | pix);
}

/*
 * get alpha channel [0..255] from jint pixel value
 */
static inline unsigned char g_getAlpha ( g_pixel_t pix )
{
  return (pix & 0xff000000) >> 24;
}

/*
 * remove alpha channel from jint pixel value
 */
static inline g_pixel_t g_stripAlpha ( g_pixel_t pix )
{
  return (pix & 0x00ffffff);
}

/*
 * Answer pixel value for given 8-8-8-8 ARGB
 * (remove switch for speed optimization)
 */
static inline jint g_pixelValue ( Toolkit* X, jint rgb )
{
	jint pix;

	switch ( _(X)->colorMode ) {
	case CM_TRUE:
		pix = (((rgb & _(X)->map.tclr->blueMask)  >> _(X)->map.tclr->blueShift) +
			   ((rgb & _(X)->map.tclr->greenMask) >> _(X)->map.tclr->greenShift) +
			   ((rgb & _(X)->map.tclr->redMask)   >> _(X)->map.tclr->redShift));
		break;
	case CM_PSEUDO:
		pix = _(X)->map.pclr->pix[(int) ((JRED(rgb)   + D16) / D8)]
			                    [(int) ((JGREEN(rgb) + D16) / D8)]
			                    [(int) ((JBLUE(rgb)  + D16) / D8)];
		break;
	case CM_VGA:
		pix = getVGAPixelValue( X, rgb, 0);
		break;
	default:
		pix = 0;
	}

	return g_addAlpha( rgb, pix);
}


/*
 * Answer RGB components for given pixel value
 */
static inline void g_rgbValues ( Toolkit* X, unsigned long pixel, int* r, int* g, int* b )
{
	u8 rr, gg, bb;
	
	switch ( _(X)->colorMode ) {
	case CM_TRUE:
		*r = ((pixel & _(X)->map.tclr->pixRedMask)   << _(X)->map.tclr->redShift)   >> 16;
		*g = ((pixel & _(X)->map.tclr->pixGreenMask) << _(X)->map.tclr->greenShift) >> 8;
		*b = ((pixel & _(X)->map.tclr->pixBlueMask)  << _(X)->map.tclr->blueShift);
		break;
	default:
	        MGL_getPaletteEntry( X->gc, pixel, &rr, &gg, &bb);
	        *r = rr; *g = gg; *b = bb;
	}
}

/*
 * Answer pixel value with modified brightness
 */
static inline jint g_modPixelValue ( Toolkit* X, int pix, int mod )
{
	return modPixelValue( X, pix, mod);
}

/******************************************************************************
 * Graphics state - these functions are used to set the state of our
 * native cache 'Graphics' objects
 */

/*
 * Set global draw context according to the provided Image
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentImage ( Toolkit* X, Image* img )
{
	MGL_makeCurrentDC( img->gc);
}

/*
 * Set global draw context according to the provided Graphics
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentGraphics ( Toolkit* X, Graphics* gr )
{
	if ( X->lastGc != gr->gc ){
		MGL_makeCurrentDC( gr->gc);
		X->lastGc = gr->gc;
	}

	MGL_useFont( gr->fnt);
	MGL_setColor( gr->clr);
	MGL_setWriteMode( (gr->xclr == -1) ? MGL_REPLACE_MODE : MGL_XOR_MODE );
}

/*
 * Temporarily store the current context (if the graphics driver
 * can't draw into aribtrary contexts)
 */
static inline void g_saveContext ( g_contextSav_t* cs )
{
}

/*
 * Restore the current context from a temporarily saved one
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_restoreContext ( g_contextSav_t* cs )
{
}

/*
 * Set the active clip rectangle according to the provided Graphics object
 */
static inline void g_setClip ( Graphics* gr )
{
	rect_t r;

	if ( X->lastGc == gr->gc ){
		r.left	 = _(gr)->curClip->x0;
		r.top	 = _(gr)->curClip->y0;
		r.right	 = _(gr)->curClip->x1;
		r.bottom = _(gr)->curClip->y1;

		MGL_setClipRectDC( gr->gc, r);
	}
}

/*
 * one time init for new Graphics objects
 */
static inline void g_initGraphics ( Graphics* gr )
{
	/* nothing to do */
}

/*
 * Set the draw context of a Graphics object 'gr' from a Window object
 */
static inline void g_setWindowGraphics ( Graphics* gr, AwtWnd* wnd )
{
	gr->gc = X->gc;
}

/*
 * Set the draw context of a Graphics object 'gr' from a Image object
 */
static inline void g_setImageGraphics ( Graphics* gr, Image* img )
{
	gr->gc = img->gc;
}

/*
 * Set the draw context of a Graphics object 'gr' from a Graphics object
 */
static inline void g_setGraphicsGraphics ( Graphics* gr, Graphics *ogr )
{
	gr->gc = ogr->gc;
}

/*
 * Set the foreground color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setColor ( Graphics* gr, g_pixel_t clr )
{
	gr->fg = clr;
	gr->clr = (gr->xclr != -1) ? (clr ^ gr->xclr) : clr;

	if ( gr == X->lastGr ){
		MGL_setColor( gr->clr);
	}
}

/*
 * Answer foreground color of Graphics object
 */
static inline g_pixel_t g_getColor ( Graphics* gr )
{
	return gr->fg;
}

/*
 * Set the background color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 * This is a optional function for drivers which can't draw
 * text with transparent background
 */
static inline void g_setBackColor ( Graphics* gr, g_pixel_t clr )
{
	gr->bg = clr;
}

/*
 * Set the font of a Graphics object
 * Note that we also might have to set the active font in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setFont ( Graphics* gr, g_font_t* fnt )
{
	gr->fnt = fnt;

	if ( gr == X->lastGr ) {
		MGL_useFont( fnt);
	}
}

/*
 * Set the draw mode of a Graphics object to 'overpaint'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setPaintMode ( Graphics* gr )
{
	gr->xclr = -1;
	gr->clr	 = gr->fg;

	if ( gr == X->lastGr ) {
		MGL_setWriteMode( MGL_REPLACE_MODE);
	}
}

/*
 * Set the draw mode of a Graphics object to 'xor'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setXORMode ( Graphics* gr, g_pixel_t xclr )
{
	gr->xclr = xclr;
	gr->clr	 = (gr->fg ^ xclr);

	if ( gr == X->lastGr ) {
		MGL_setWriteMode( MGL_XOR_MODE);
	}
}


/******************************************************************************
 * drawing operations
 */

/*
 * Set color of pixel
 * (Make sure it is clipped according to the Graphics, since many
 * graphics drivers don't provide clipping for this operation)
 */
static inline void g_putPixel ( Graphics* gr, int x, int y)
{
	MGL_pixelCoord( x, y);
}

/*
 * Fill rectangle with background color of Graphics object
 */
static inline void g_clearRect ( Graphics* gr, int x0, int y0, int x1, int y1)
{
	g_pixel_t clr = gr->clr;

	MGL_setColor( gr->bg);
	MGL_fillRectCoord( x0, y0, x1, y1);
	MGL_setColor( clr);
}

/*
 * Draw a line
 */
static inline void g_drawLine ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	MGL_lineCoord( x0, y0, x1, y1);
}

static inline void g_drawHLine ( Graphics* gr, int x0, int y0, int x1 )
{
	MGL_lineCoord( x0, y0, x1, y0);
}

static inline void g_drawVLine ( Graphics* gr, int x0, int y0, int y1 )
{
	MGL_lineCoord( x0, y0, x0, y1);
}


/*
 * Draw a rectangle (corner exclusive)
 */
static inline void g_drawRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	MGL_rectCoord( x0, y0, x1, y1);
}

/*
 * Fill a rectangle (corner inclusive)
 */
static inline void g_fillRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	MGL_fillRectCoord( x0, y0, x1, y1);
}

/*
 * Draw open arc (i.e. no lines from arc endpoints to center of oval).
 * Note that this actually is a oval, not just a circle
 */
static inline void g_drawOpenArc ( Graphics* gr, int x, int y,
								   int ra, int rb, int a1, int a2 )
{
	MGL_ellipseArcCoord( x, y, ra, rb, a1, ( a1+a2));
}


/*
 * Fill closed arc (pie chart)
 */
static inline void g_fillArc ( Graphics* gr, int x, int y,
							   int ra, int rb, int a1, int a2 )
{
	MGL_fillEllipseArcCoord( x, y, ra, rb, a1,(a1+a2));
}

/*
 * Draw oval (ellipse)
 */
static inline void g_drawOval ( Graphics* gr, int x, int y, int ra, int rb )
{
	MGL_ellipseCoord( x, y, ra, rb);
}

/*
 * Fill oval (ellispe)
 */
static inline void g_fillOval ( Graphics* gr, int x, int y, int ra, int rb )
{
	MGL_fillEllipseCoord( x, y, ra, rb);
}

/*
 * Draw single-byte text
 */
static inline void g_drawByteText ( Graphics* gr, char* str,
									int len, int x, int y )
{
	char *c = jbyte2CString(  _(X),  str, len);
	MGL_drawStrXY( x, y, c);
}

/*
 * Draw unicode text
 */
static inline void g_drawWideText ( Graphics* gr, jchar* str,
									int len, int x, int y )
{
	wchar_t *wc = (wchar_t*) jchar2WString( _(X), str, len);
	MGL_drawStrXY_W( x, y, wc);
}

/*
 * Draw polygon (lines connecting vertices in 'points')
 */
static inline void g_drawPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	MGL_polyLine( nPoints, points);
}

/*
 * Fill polygon
 */
static inline void g_fillPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	MGL_fillPolygon( nPoints, points, 0, 0);
}


/******************************************************************************
 * image drawing (bitblt related functions)
 */

/*
 * Bit block transfer in same context (Image or display)
 */
static inline void g_copyArea ( Graphics* gr, int x, int y,
								int xw, int yh, int xdelta, int ydelta)
{
	MGL_bitBltCoord( gr->gc, gr->gc, x, y, xw, yh, 
					 x+xdelta, y+ydelta, MGL_REPLACE_MODE);
}

/*
 * Bit block transfer from Image context source (to Image or display)
 * This operation has to handle alpha channels. In case of a non-negative
 * 'bg' color, the transparent image pixels have to be set to this value
 */
static inline void g_drawImage ( Graphics* gr, Image* img, int dx, int dy,
								 int sx, int sy, int sxw, int syh, g_pixel_t bg )
{
	if ( img->mask ){
		if ( bg != -1 ){
			int dxw = dx + sxw - sx;
			int dyh = dy + syh - sy;
			g_pixel_t clr = gr->clr;

			MGL_setColor( gr->bg);
			MGL_fillRectCoord( dx, dy, dxw, dyh);
			MGL_setColor( clr);
		}

		MGL_putBitmapSection( X->gc,  sx, sy, sxw, syh,
							  dx, dy, img->mask, MGL_XOR_MODE);
		MGL_putBitmapSection( X->gc,  sx, sy, sxw, syh,
							  dx, dy, img->bmp, MGL_AND_MODE);
	}
	else if ( _(img)->alpha ){
		drawAlphaImage( gr, img, sx, sy, dx, dy, sxw-sx, syh-sy, bg);
	}
	else {
		MGL_putBitmapSection( X->gc,  sx, sy, sxw, syh,
							  dx, dy, img->bmp, MGL_REPLACE_MODE);
	}
}

/*
 * grab pixels from a source Graphics, and punch them into a destination
 * Graphics (this is a generalized blit with a source that is not a
 * Java image)
 */
static inline void g_grabPixels ( Graphics* dgr, Graphics* sgr,
								  int sx, int sy, int w, int h )
{
	/* not yet */
}


/******************************************************************************
 * image processing
 */

/*
 * Initialize a Image object so that we can set it's pixel values.
 * Note that we don't have alpha information, yet (i.e. the image
 * subsequently might be set to masked or full alpha)
 */
static inline int g_initImage ( Image* img )
{
	if ( (_(img)->flags & IMG_IS_DRAWABLE) ){
		img->gc = MGL_createMemoryDC( _(img)->width, _(img)->height,
									  _(X)->nColorPlanes, X->pf);
		return (img->gc != 0);
	}
	else {
		img->bmp = _create_bitmap_( _(img)->width, _(img)->height);
		return (img->bmp != 0);
	}
}


/*
 * Set alpha channel type of Image object. The supported 'transType'
 * values are NO_ALPHA, MASKED_ALPHA (transparent pixels), FULL_ALPHA.
 * After this operation, we have to be able to store pixel alpha
 * information (at least temporarily, to be later-on processed by
 * completeImage)
 */
static inline void g_setAlphaType ( Image* img, int alphaType )
{
	if ( alphaType && ! (_(img)->mask || _(img)->alpha) ){
		if ( alphaType == MASK_ALPHA ){
			_(img)->mask = createImageMask(	X, _(img)->width, _(img)->height);
		}
		else if ( alphaType == FULL_ALPHA ){
#if defined(USE_FULL_ALPHA)
			_(img)->alpha = createImageAlpha( X, _(img)->width, _(img)->height);
#else
			_(img)->mask = createImageMask(	X, _(img)->width, _(img)->height);
#endif
		}
	}
}

/*
 * set transparency color of the image in the native layer
 */
static inline void g_setTranspValue ( Image* img, g_pixel_t clr )
{
 
}

/*
 * Transform temporarily stored image information (both pixel color and
 * alpha values) into a representation which is suitable for drawing
 * the image. Free any temporary information.
 */
static inline void g_completeImage ( Image* img )
{
	if ( _(img)->mask ){
		/* the mask data (surface) is already in the right format,
		 * we just have to wrap it up in a bitmap_t
		 */
		img->mask = AWT_CALLOC( 1, sizeof(bitmap_t));
		img->mask->width  = _(img)->width;
		img->mask->height = _(img)->height;
		img->mask->bitsPerPixel = 1;
		img->mask->bytesPerLine = _(img)->mask->scanline;
		img->mask->pf = NULL;
		img->mask->pal = NULL;
		img->mask->surface = _(img)->mask;
	}
}

/*
 * Clean up platform specific Image components
 */
static inline void g_freeImage ( Image* img )
{
	if ( img->gc ){
		MGL_destroyDC( img->gc);
		img->gc = 0;
	}

	if ( img->bmp ) {
		/* we allocated bitmap and surface in one chunk */
		AWT_FREE( img->bmp);
		img->bmp = 0;
	}

	if ( img->mask ) {
		AWT_FREE( img->mask);
		img->mask = 0;
	}
}


/*
 * Get alpha value of image pixel [0..255]
 */
static inline int g_getPixelAlpha ( Image* img, int x, int y )
{
	if ( img->mask ){
		/* looks counter-intuitive, but we AND the mask source to
		 * blacken the visible parts
		 */
		return getMask( _(img)->mask, x, y) ? 0 : 0xff;
	}
	else if ( _(img)->alpha ){
		return getAlpha( _(img)->alpha, x, y);
	}
	else {
		return 0xff;
	}
}

/*
 * Answer alpha type of image. There are three supported
 * alpha type values
 *	- NO_ALPHA	   : no alpha channel at all
 *	- MASKED_ALPHA : alpha = [0,255] (transparent pixels)
 *	- FULL_ALPHA   : alpha = [0..255]
 */
static inline int g_getAlphaType ( Image* img )
{
	if ( _(img)->mask )
		return MASK_ALPHA;
	else if ( _(img)->alpha )
		return FULL_ALPHA;
	else
		return NO_ALPHA;
}


/*
 * Set pixel value in Image (with optional alpha value [0..255]).
 * This is called before g_completeImage(), i.e. it can use
 * temporary pixel/alpha storage.
 * Note that we don't call this for drawable images (which are modified
 * via a Graphics)
 */
static inline void g_putImagePixel ( Image* img , int x, int y,
									 g_pixel_t pix, unsigned char alpha )
{
	if ( alpha != 0xff ){
		if ( _(img)->alpha ){
			putAlpha( _(img)->alpha, x, y, alpha);
		}
		else if ( _(img)->mask ){
			unsetMask( _(img)->mask, x, y);
			pix = 0;
		}
	}

	_put_bitmap_pixel_( img->bmp, x, y, pix);
}

/*
 * Answer pixel value of already completed Image
 */
static inline g_pixel_t	 g_getImagePixel ( Image* img, int x, int y )
{
	return _get_bitmap_pixel_( img->bmp, x, y);
}


/******************************************************************************
 * event processing
 */

/*
 * Initialize keyboard input
 */
static inline void g_initKeyboard ( Toolkit* X )
{
}

/*
 * Initialize moude input
 */
static inline void g_initMouse ( Toolkit* X )
{
}

/*
 * Get current horizontal pointer location
 */
static inline int g_getMouseX ( Toolkit* X )
{
	return X->event.where_x;
}

/*
 * Get current vertical pointer location
 */
static inline int g_getMouseY ( Toolkit* X )
{
	return X->event.where_y;
}

/*
 * Get nmber of pressed mouse button
 */
static inline int g_getMouseBtn ( Toolkit* X )
{
	if ( X->event.modifiers & EVT_LEFTBUT )		   return 1;
	else if ( X->event.modifiers & EVT_MIDDLEBUT ) return 2;
	else if ( X->event.modifiers & EVT_RIGHTBUT )  return 3;
	else									   return 0;
}

/*
 * Get Java keycode of current input event
 */
static inline int g_getKeyCode ( Toolkit* X )
{
	unsigned int sc = EVT_scanCode( X->event.message);
	return (sc < nKeyCodes) ? keyCodes[sc] : 0; 
}

/*
 * Get Java keychar of current input event
 */
static inline int g_getKeyChar ( Toolkit* X )
{
	return EVT_asciiCode( X->event.message);
}

/*
 * Make sure the next event will be a (artificial) mouse motion (with the
 * last known pointer position). Used to retarget mouse for popups
 */
static inline void g_queueMouseMotion ( JNIEnv* env, Toolkit* X )
{
	X->event.where_x = _(X)->xMouse;
	X->event.where_y = _(X)->yMouse;

	X->event.what = EVT_MOUSEMOVE;

	X->preFetched = 1;
}

/*
 * notify event system that we have a pending paint request
 */
static inline void g_queuePaint ( JNIEnv* env, Toolkit* X)
{
	/* nothing to do */
}

#endif /* GRLIB_FUNCS */
