/*
 * toolkit-common-pre.h - common toolkit header
 *
 * All things which go in here should NOT use any struct intrinsics from toolkit.h
 * (esp. not struct _Toolkit itself), since it is included IN FRONT of these
 * struct defs. However, target specific macros from grlib.def (e.g. for PIXEL type)
 * can be used
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


/***********************************************************************************
 * most basic things
 */

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif


#define EXPORT(_m) Java_java_awt_Toolkit_##_m

/*
 * Macro to save typing for common field access. We want to have the real
 * types (not the common subsets) available (even in the xxx-common.c modules)
 * so that we can pass them into grlib macros (which should be able to get all
 * fields)
 */
#define _(_o)   (&((_o)->common))


/*
 * default DEBUG macro, in case we don't have the JDEBUG system
 */
#if !defined(DBG)
#  if defined(DEBUG)
#    define DBG(_key,_msg) if (1) {printf( "%s : ", #_key ); printf _msg;}
#  else
#    define DBG(_key,_msg)
#  endif
#endif 


/***********************************************************************************
 * heap macros
 */

/*
 * a failed allocation attempt should throw a OutOfMemoryException in the allocator,
 * we just report it here
 */

extern void* __kaffe_awt_malloc(size_t);
extern void __kaffe_awt_free(void*);

static inline void* _awt_malloc_wrapper ( size_t size )
{
	void *adr = __kaffe_awt_malloc( size);
	DBG( awt_mem, ("malloc: %d  -> %p\n", size, adr));
	DBG_ACTION( awt_mem, if ( !adr ) {
	    DBG( awt_mem, ("ALLOCATION FAILED"));
	});
	return adr;
}

static inline void* _awt_calloc_wrapper ( int n, size_t size )
{
	void *adr = __kaffe_awt_malloc( n * size);
	DBG( awt_mem, ("calloc: %d,%d  -> %p\n", n, size, adr));
	DBG_ACTION( awt_mem, if ( !adr ) {
	    DBG( awt_mem, ("ALLOCATION FAILED"));
	});
	return adr;
}

static inline void _awt_free_wrapper ( void* adr )
{
	DBG( awt_mem, ("free: %p\n", adr));
	__kaffe_awt_free( adr);
}


#ifndef AWT_MALLOC
#  define AWT_MALLOC(_n) \
     _awt_malloc_wrapper( _n)
#endif

#ifndef AWT_CALLOC
#  define AWT_CALLOC(_n,_sz) \
     _awt_calloc_wrapper( _n, _sz)
#endif

#ifndef AWT_FREE
#  define AWT_FREE(_adr) \
     _awt_free_wrapper( _adr);
#endif

#ifndef AWT_ALLOCA
#  if defined(HAVE_ALLOCA)
#    define AWT_ALLOCA(_n)           alloca(_n)
#    define AWT_ALLOCA_FREE(_adr)
#  else
#    define AWT_ALLOCA(_n)           AWT_MALLOC(_n)
#    define AWT_ALLOCA_FREE(_adr)    AWT_FREE(_adr)
#  endif
#endif




/***********************************************************************************
 * file io wrapper macros (used for ext. format image production)
 */

#ifndef AWT_OPEN
#  define AWT_OPEN(_file)               open(_file, JOPEN_READ)
#endif

#ifndef AWT_REWIND
#  define AWT_REWIND(_fd)               lseek(_fd, 0, JSEEK_SET)
#endif

#ifndef AWT_SETPOS
#  define AWT_SETPOS(_fd,_off)          lseek(_fd, _off, JSEEK_CUR)
#endif

#ifndef AWT_READ
#  define AWT_READ(_fd,_buf,_count)     read(_fd,_buf,_count)
#endif

#ifndef AWT_CLOSE
#  define AWT_CLOSE(_fd)                close(_fd)
#endif


/***********************************************************************************
 * LinkedRect support
 */

typedef struct _LinkedRect {
	short               x0, y0, x1, y1;
	struct _LinkedRect* next;
} LinkedRect;


extern LinkedRect* rectCache;
extern LinkedRect* getIntersection ( LinkedRect *b, LinkedRect *a );
extern LinkedRect* getUnion ( LinkedRect *b, LinkedRect *a );
extern LinkedRect* getDiff ( LinkedRect* below, LinkedRect* above );
extern LinkedRect* getDiffPart ( LinkedRect* list, LinkedRect* w );
extern LinkedRect* getDiffs ( LinkedRect* r0, LinkedRect* r1 );

extern void dumpLinkedRect ( LinkedRect* r );

static inline int
clipInterval ( int x, int x0, int x1 )
{
	return ((x < x0) ? x0 : ((x > x1) ? x1 : x));
}

static inline char
equalsRect ( LinkedRect*r, int x0, int y0, int x1, int y1 )
{
	return ((r->x0 == x0) && (r->y0 == y0) && (r->x1 == x1) && (r->y1 == y1));
}

static inline void
setRect ( LinkedRect*r, int x0, int y0, int x1, int y1, LinkedRect* next )
{
	r->x0 = x0; r->y0 = y0; r->x1 = x1; r->y1 = y1; r->next = next;
}

static inline void
translateRect ( LinkedRect*r, int dx, int dy )
{
	r->x0 += dx; r->y0 += dy; r->x1 += dx; r->y1 += dy;
}

static inline void
clipRect ( LinkedRect*r, int x0, int y0, int x1, int y1 )
{
	r->x0 = clipInterval( r->x0, x0, x1);
	r->x1 = clipInterval( r->x1, x0, x1);
	r->y0 = clipInterval( r->y0, y0, y1);
	r->y1 = clipInterval( r->y1, y0, y1);
}

static inline void
appendRect ( LinkedRect* list, LinkedRect* r )
{
	for ( ; list->next; list=list->next );
	list->next = r;
}

static inline LinkedRect*
newRect ( int x0, int y0, int x1, int y1, LinkedRect* prev, LinkedRect* next )
{
	LinkedRect *r;

	if ( rectCache ) {
		r = rectCache;
		rectCache = r->next;
	}
	else {
		r = (LinkedRect*) AWT_MALLOC( sizeof( LinkedRect));
	}

	r->x0 = x0;  r->y0 = y0;  r->x1 = x1;  r->y1 = y1;

	if ( prev )
		prev->next = r;
	r->next = next;

	return r;
}

static inline LinkedRect*
newDim ( LinkedRect* r )
{
	return newRect( 0, 0, r->x1 - r->x0, r->y1 - r->y0, 0, 0);
}

static inline void
freeRect ( LinkedRect* r )
{
	r->next = rectCache;
	rectCache = r;
}

static inline void
freeRectList ( LinkedRect* r )
{
	LinkedRect *v;

	for ( v=r; v->next; v=v->next );
	v->next = rectCache;

	rectCache = r;
}


/***********************************************************************************
 * common image types + funcs
 */

typedef struct _ImageAlpha {
	unsigned char *buf;
	int           width, height;
} ImageAlpha;


static inline void
putAlpha ( ImageAlpha* img, int col, int row, unsigned char alpha )
{
	img->buf[ row*img->width + col ] = alpha;
}


static inline int
getAlpha ( ImageAlpha* img, int col, int row )
{
	return img->buf[ row*img->width + col];
}



#define MASKWORDSIZE 8
#define MASKWORD     unsigned char

#if defined (MAIN)
unsigned int imageMaskBits[] = {
#if (MASKWORDSIZE == 32)
	0x80000000, 0x40000000, 0x20000000, 0x10000000,
	0x8000000, 0x4000000, 0x2000000, 0x1000000,
	0x800000, 0x400000, 0x200000, 0x100000,
	0x80000, 0x40000, 0x20000, 0x10000,
#endif
#if (MASKWORDSIZE >= 16)
	0x8000, 0x4000, 0x2000, 0x1000,
	0x800, 0x400, 0x200, 0x100,
#endif
	0x80, 0x40, 0x20, 0x10,
	0x8, 0x4, 0x2, 0x1
};
#else
extern unsigned int imageMaskBits[];
#endif

typedef struct _ImageMask {
	MASKWORD      *buf;
	int           width, height;
	int           scanline;  /* number of "MASKWORDS" in scanline */
} ImageMask;

static inline void
setMask ( ImageMask* img, int col, int row )
{
	MASKWORD *p = ((MASKWORD*)img->buf) + (row * img->scanline) + (col / MASKWORDSIZE);
	*p |= imageMaskBits[col % MASKWORDSIZE];
}

static inline void
unsetMask ( ImageMask* img, int col, int row )
{
	MASKWORD *p = ((MASKWORD*)img->buf) + (row * img->scanline) + (col / MASKWORDSIZE);
	*p &= ~imageMaskBits[col % MASKWORDSIZE];
}

static inline int
getMask( ImageMask* img, int col, int row )
{
	MASKWORD *p = img->buf + (row * img->scanline) + (col / MASKWORDSIZE);
	return (*p & imageMaskBits[col % MASKWORDSIZE]);
}

/***********************************************************************************
 * common Window types + defines
 */

#define INVISIBLE       0
#define PARTLY_VISIBLE  1
#define FULLY_VISIBLE   2

typedef struct _AwtWnd {
	LinkedRect      bounds;

	char            mapped;
	char            visibility;

	unsigned        visTag;
	LinkedRect*     visSegments;
	
	char            update;
	LinkedRect*     updateRect;

	struct _AwtWnd* next;
	struct _AwtWnd* prev;
} AwtWnd;


/***********************************************************************************
 * common color types + defines
 */

#define MAX_RGB_REQUESTS 40  /* 256 - 216 web colors */

/* has to be != 0, since '0' is used to trigger the color int */
#define CM_VGA    1
#define CM_PSEUDO 2
#define CM_TRUE   3
#define CM_MONO   4
#define CM_GRAY   5

typedef struct _RgbTrueMap {
	unsigned int    redMask;
	unsigned int    greenMask;
	unsigned int    blueMask;
	char            blueShift;
	char            redShift;
	char            greenShift;

	unsigned int    pixRedMask;
	unsigned int    pixGreenMask;
	unsigned int    pixBlueMask;
} RgbTrueMap;

typedef struct _Common_IdxMap {
	g_paletteEntry_t *clrs;
	int              nClrs;
} CommonIdxMap;

typedef struct _RgbPseudoMap {
	CommonIdxMap     common;
	unsigned char    pix[12][12][12]; /* pix[r][g][b] -> system palette index */
} RgbPseudoMap;

typedef struct RgbVgaMap {
	CommonIdxMap     common;
} RgbVgaMap;

typedef union _RgbMap {
	RgbTrueMap     *tclr;
	RgbPseudoMap   *pclr;
	RgbVgaMap      *vclr;
} RgbMap;


/* map [0..255] channel value to [0..11] cell index */
#define CELL_INDEX(_n)   ((((int)(_n)) * 4 - 1) / 86)


#define JRGB(_r,_g,_b)  (_r<<16 | _g<<8 | _b)
#define JALPHA(_rgb)    (((_rgb) & 0xff000000) >> 24)
#define JRED(_rgb)      (((_rgb) & 0xff0000) >> 16)
#define JGREEN(_rgb)    (((_rgb) & 0x00ff00) >> 8)
#define JBLUE(_rgb)     ((_rgb) & 0x0000ff)


/* VGA indices */
#define  CLR_BLACK     0
#define  CLR_WHITE     1
#define  CLR_RED       2
#define  CLR_GREEN     3
#define  CLR_BLUE      4
#define  CLR_LIGHTGRAY 5
#define  CLR_GRAY      6
#define  CLR_DARKGRAY  7
#define  CLR_YELLOW    8
#define  CLR_MAGENTA   9
#define  CLR_CYAN      10
#define  CLR_PINK      11
#define  CLR_ORANGE    12
#define  CLR_GRAY1     13
#define  CLR_GRAY2     14
#define  CLR_RED1      15

#define DARKER    1
#define BRIGHTER  2


/***********************************************************************************
 * These will be our main data structures (all including a common part
 * defined subsequently). Forward declare them so that we can link objects and
 * prototype functions here
 */
struct _Image;
struct _Graphics;
struct _Toolkit;



/***********************************************************************************
 * common Image structure
 */

#define NO_ALPHA   0
#define MASK_ALPHA 1
#define FULL_ALPHA 2

#define IMG_IS_DRAWABLE  0x01         /* offscreen image we can draw into */
#define IMG_IS_DELTA     0x02         /* has to be drawn superposed */
#define IMG_BUFFER       0x04         /* image needs a rendering buffer */
#define IMG_ANIMATION    0x10         /* image used as in-situ animation (no ring) */

#define NO_TRANS_INDEX_AVAILABLE -2   /* we can't use a transparency index */


typedef struct _Common_	 {
	ImageAlpha       *alpha;         /* full alpha channel (for alpha != 0x00 or 0xff) */
	ImageMask        *mask;          /* on/off alpha channel (GIFS) */
	int              trans;          /* intermediate transparency index */

	int              flags;          /* drawable, delta etc. */

	int              left, top;      /* some GIF movies require drawing offsets */
	int              width, height;

	/* per frame animation data */
	short            latency;        /* between image flips, for "gif-movies" */
	short            frame;          /* frame number for animations */
	
	/* invariant animation data */
	struct _Image    *buffer;        /* doublebuffer for delta-frame animations */
	struct _Graphics *gr;            /* graphics to update the buffer */
	short            nLoops;         /* number of loops to go */
	short            nFrames;        /* number of frames in animation */

	struct _Image    *next;          /* link for animations, queues etc. */
} CommonImage;


/***********************************************************************************
 * common Graphics structure
 */
#define  WINDOW   0
#define  IMAGE    1
#define  GRAPHICS 2

typedef struct _Common_Graphics {
	void            *tgt;
	char            tgtType;
	
	int             x0, y0;
	short           xOff, yOff;

	LinkedRect      clip;

	unsigned        visTag;
	LinkedRect      *clipSegments;
	LinkedRect      *curClip;
 } CommonGraphics;


/***********************************************************************************
 * common Toolkit structure
 */

typedef struct _Common_Toolkit {
	int            width;
	int            height;

	int            xoff;
	int            yoff;

	char           *buf;
	unsigned int   nBuf;

	int            colorMode;
	int            nColorPlanes;  
	RgbMap         map;
	int            black, white;

	int            nWindows;
	AwtWnd         **windows;

	AwtWnd         *top;
	AwtWnd         *bottom;
	AwtWnd         *root;

	AwtWnd         *lastWindow;
	int            srcIdx;
	int            rootIdx;

	AwtWnd         *mouseWindow;
	int            xMouse, yMouse;
	int            button;
	int            dragging;

	int            pendingUpdates;
	LinkedRect     *rootUpdates;

	void*          lastGr;          /* last Graphics used in a drawing op */
} CommonToolkit;



/***********************************************************************************
 * buffered conversion functions
 */

static inline void*
getBuffer ( CommonToolkit* X, unsigned int nBytes )
{
	if ( nBytes > X->nBuf ) {
		if ( X->buf )
			AWT_FREE( X->buf);
		X->buf  = AWT_MALLOC( nBytes);
		X->nBuf = nBytes;
	}
	return X->buf;
}

static inline char*
java2CString ( JNIEnv *env, CommonToolkit* X, jstring jstr )
{
	jboolean      isCopy;
	register int  i;
	int           n = (*env)->GetStringLength( env, jstr);
	const jchar   *jc = (*env)->GetStringChars( env, jstr, &isCopy);
	register char *c;
	char          *c0 = (char*) getBuffer( X, n+1);

	for ( i=0, c=c0; i<n; i++ ){
		*c++ = (char) jc[i];
	}
	*c = 0;

	(*env)->ReleaseStringChars( env, jstr, jc);

	return c0;
}

static inline char*
jchar2CString ( CommonToolkit* X, jchar* jc, int len )
{
	register int   i;
	register char  *c;
	char           *c0 = (char*) getBuffer( X, len+1);
  
	for ( i=0, c=c0; i<len; i++ ){
		*c++ = (char) jc[i];
	}
	*c = 0;

	return c0;
}

/*
 * this is just here to avoid local buffers, it doesn't have any other effect
 * than just to append a trailing \0
 */
static inline char*
jbyte2CString ( CommonToolkit* X, jbyte* jb, int len )
{
	register int   i;
	register char  *c;
	char           *c0 = (char*) getBuffer( X, len+1);
 
	for ( i=0, c=c0; i<len; i++ ){
		*c++ = (char) jb[i];
	}
	*c = 0;

	return c0;
}

/*
 * this is just here to avoid local buffers, it doesn't have any other effect
 * than just to append a trailing \0
 */
static inline jchar*
jchar2WString ( CommonToolkit* X, jchar* jc, int len )
{
	register int   i;
	register jchar *c;
	jchar          *c0 = (jchar*) getBuffer( X, (len+1)*sizeof(jchar));
 
	for ( i=0, c=c0; i<len; i++ ){
		*c++ = jc[i];
	}
	*c = 0;

	return c0;
}


/*************************************************************************************
 * commonly used function prototypes
 */

void           drawAlphaImage      ( struct _Graphics* gr, struct _Image* img,
									 jint srcX, jint srcY, jint dstX, jint dstY,
									 jint width, jint height, jint bgval );
void           retargetMouse       ( JNIEnv* env, struct _Toolkit* X );
void           initCursors         ( struct _Toolkit* X);
void           initScaledImage     ( struct _Toolkit* X, struct _Image *tgt, struct _Image *src,
									 int dx0, int dy0, int dx1, int dy1,
									 int sx0, int sy0, int sx1, int sy1 );
struct _Image* createImage         ( int width, int height, int isDrawable );
ImageAlpha*    createImageAlpha    ( struct _Toolkit* X, int width, int height );
void           freeImageAlpha      ( ImageAlpha* img );
ImageMask*     createImageMask     ( struct _Toolkit* X, int width, int height );
void           freeImageMask       ( ImageMask* img );
int            needsFullAlpha      ( ImageAlpha *aimg, double threshold );
void           fillImageAlpha      ( ImageAlpha *aimg, int x0, int y0,
									 int x1, int y1, unsigned char a);
void           reduceAlpha         ( struct _Image* img, ImageAlpha* aimg, int threshold );
void           Java_java_awt_Toolkit_imgFreeImage( JNIEnv* env, jclass clazz,
												   struct _Image* img);

void           Java_java_awt_Toolkit_graDrawImage( JNIEnv * env, jclass clazz,
												   struct _Graphics* gr,
												   struct _Image * img,
												   jint srcX, jint srcY,
												   jint dstX, jint dstY,
												   jint width, jint height, jint bgval);

void           Java_java_awt_Toolkit_graFreeGraphics( JNIEnv* env, jclass clazz, 
													  struct _Graphics* gr);

void           Java_java_awt_Toolkit_graFillRect( JNIEnv * env, jclass clazz,
												  struct _Graphics* gr,
                                                  jint x, jint y,
												  jint width, jint height);

void           updateWindow        ( JNIEnv* env, struct _Toolkit* X, AwtWnd*, LinkedRect* );

void           initColorMapping    ( JNIEnv* env, struct _Toolkit* X);
jobject        getSystemColorModel ( JNIEnv * env, jclass clazz, struct _Toolkit* X, int trans );
jint           getVGAPixelValue    ( struct _Toolkit* X, jint rgb, int mod );
int            modPixelValue       ( struct _Toolkit* X, jint rgb, int mod );
