/**
 * evt.c - platform specific event handling (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define NO_DEBUG

#include "toolkit.h"

#include "keytab_uni.h"

/******************************************************************************
 * generic exported methods
 */

#define _evtInit              EXPORT(evtInit)
#define _evtRegisterSource    EXPORT(evtRegisterSource)
#define _evtUnregisterSource  EXPORT(evtUnregisterSource)

/******************************************************************************
 * import generic functionality
 */

#include "../evt-common.c"


/******************************************************************************
 * platform specific event functions
 */

int stopEventLoop;

inline int
getButtonEventId ( Toolkit *X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);

	if ( evt->type.pointer.buttonChange & UGL_POINTER_BUTTON1 ){
		return (evt->type.pointer.buttonState & UGL_POINTER_BUTTON1) ?
			MOUSE_PRESSED : MOUSE_RELEASED; 
	}

	if ( evt->type.pointer.buttonChange & UGL_POINTER_BUTTON2 ){
		return (evt->type.pointer.buttonState & UGL_POINTER_BUTTON2) ?
			MOUSE_PRESSED : MOUSE_RELEASED; 
	}

	if ( evt->type.pointer.buttonChange & UGL_POINTER_BUTTON3 ){
		return (evt->type.pointer.buttonState & UGL_POINTER_BUTTON3) ?
			MOUSE_PRESSED : MOUSE_RELEASED; 
	}

	return 0;
}

inline int
getKeyEventId ( Toolkit *X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);

	return (evt->modifiers & UGL_KEYBOARD_KEYDOWN) ? KEY_PRESSED : KEY_RELEASED;
}


/*******************************************************************************
 * exported event functions
 */

void
Java_java_awt_NativeCollector_ncLoop ( JNIEnv * env, jclass clazz )
{

#if defined (USE_NATIVE_DISPATCHER_LOOP)
	UGL_STATUS stat;
	UGL_INPUT_EVENT *evt;

	/*taskPrioritySet( taskIdSelf(), 250);*/

	while ( !stopEventLoop ){
		stat = uglEventGet( X->qid, &(X->event), sizeof(UGL_EVENT),
							UGL_WAIT_FOREVER );

		if ( (stat == UGL_STATUS_OK) &&
			 (X->event.header.category == UGL_EVENT_CATEGORY_INPUT) ){
			evt = (UGL_INPUT_EVENT*) &(X->event);
			
			switch ( evt->header.type ){
			case UGL_EVENT_TYPE_KEYBOARD:
				DBG( awt_evt, ("received key event: %d\n", evt->header.type));
				
				keyNotify( env, X, getKeyEventId( X));
				break;

			case UGL_EVENT_TYPE_POINTER:
				DBG( awt_evt, ("received abs pointer event: %d\n", evt->header.type));

				if ( evt->type.pointer.buttonChange ){
					buttonNotify( env, X, getButtonEventId( X));
				}
				else {
					if ( (evt->x != _(X)->xMouse) || (evt->y != _(X)->yMouse) ){
						motionNotify( env, X);
					}
				}

				break;

			case TERMINATE_EVENT:
				uglEventQDestroy( X->eid, X->qid);
				uglDeinitialize();
				return;

			default:
				DBG( awt_evt, ("unknown event: %d\n", evt->header.type));
			}
		}
		else if ( stat == UGL_STATUS_ERROR ){
			DBG( awt_evt, ("error reading event\n"));
		}
		else if ( stat == UGL_STATUS_Q_EMPTY ) {
			DBG( awt_evt, ("queue empty\n")); /* ?? we WAIT_FOREVER ?? */
		}
	}
#endif /* USE_NATIVE_DISPATCHER_LOOP */
}


jobject
EXPORT(evtGetNextEvent) ( JNIEnv* env, jclass clazz )
{
#if !defined (USE_NATIVE_DISPATCHER_LOOP)
	UGL_STATUS stat;
	UGL_INPUT_EVENT *evt;

	/* not used, we have a native-dispatch AWT */
	DBG( awt_evt, ("getEvent (1x) \n"));

	if ( _(X)->pendingUpdates ) {
		DBG( awt_evt, ("   paintNotify\n"));
		return paintNotify( env, X);
	}

	UNBLOCK_EXECUTE( env, clazz, {
		stat = uglEventGet( X->qid, &(X->event),
							sizeof(UGL_EVENT),
							UGL_WAIT_FOREVER );
	});

	if ( (stat == UGL_STATUS_OK) &&
		 (X->event.header.category == UGL_EVENT_CATEGORY_INPUT) ){
		evt = (UGL_INPUT_EVENT*) &(X->event);
			
		switch ( evt->header.type ){
		case UGL_EVENT_TYPE_KEYBOARD:
			DBG( awt_evt, ("received key event: %d\n", evt->header.type));
			return keyNotify( env, X, getKeyEventId( X));

		case UGL_EVENT_TYPE_POINTER:
			DBG( awt_evt, ("received abs pointer event: %d\n", evt->header.type));

			if ( evt->type.pointer.buttonChange ){
				return buttonNotify( env, X, getButtonEventId( X));
			}
			else {
				if ( (evt->x != _(X)->xMouse) || (evt->y != _(X)->yMouse) ){
					return motionNotify( env, X);
				}
			}

			break;

		case TERMINATE_EVENT:
			DBG( awt_evt, ("received terminate event\n"));
			uglEventQDestroy( X->eid, X->qid);
			uglDeinitialize();
			return 0;

		case WAKEUP_EVENT:
			DBG( awt_evt, ("received wakeup event\n"));
			return 0;

		default:
			DBG( awt_evt, ("unknown event: %d\n", evt->header.type));
		}
	}
	else if ( stat == UGL_STATUS_ERROR ){
		DBG( awt_evt, ("error reading event\n"));
	}
	else if ( stat == UGL_STATUS_Q_EMPTY ) {
		DBG( awt_evt, ("queue empty\n")); /* ?? we WAIT_FOREVER ?? */
	}
#endif

	return NULL;
}


jobject
EXPORT(evtPeekEvent) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_evt, ("peekEvent\n"));
	return NULL;
}


/*
 * We can't do a XPeekIfEvent here because it would block until such an event
 * comes in
 */
jobject
EXPORT(evtPeekEventId) ( JNIEnv* env, jclass clazz, jint id )
{
	DBG( awt_evt, ("peekEventId: %d\n", (int)id));
	return NULL;
}


/*
 * This is just used to wakeup the getNextEvent call if we have a TLK_IS_BLOCKING,
 * non- NATIVE_DISPATCHER_LOOP style AWT, and we post an event from outside the
 * dispatcher thread (which might be blocked in a native getNextEvent).
 * This is a tribute to the fact that we actually have two different queues - the
 * native one, and the queue for self-posted events
 */
void
EXPORT(evtWakeup) ( JNIEnv* env, jclass clazz )
{
#if !defined (USE_NATIVE_DISPATCHER_LOOP)
	UGL_EVENT event;

	DBG( awt_evt, ("evtWakeup\n"));

	/* we use an artificial terminate event to make sure the
	 * AWT-event and driver tasks really terminate. The real
	 * termination is done from within ncLoop in respond to this event
	 */
	event.header.type = WAKEUP_EVENT;
	event.header.category = UGL_EVENT_CATEGORY_INPUT;
	uglEventPost( X->eid, X->qid, &event, sizeof( event));
#endif /* !USE_NATIVE_DISPATCHER_LOOP */
}

/*
 * This is a helper for (IS_DISPATCH_EXCLUSIVE | NATIVE_DISPATCHER_LOOP) windowing
 * systems, where we need to execute certain requests (e.g. window creation) in
 * the native event thread. Not required, here
 */
void
EXPORT(evtSendWMEvent) ( JNIEnv* env, jclass clazz, jobject wmEvt )
{
	DBG( awt_evt, ("evtSendWMEvent\n"));
	/* not required, we have a NATIVE_DISPATCHER_LOOP */
}
