/**
 * fnt.c - font support
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

#include "../fnt-common.c"

/*******************************************************************************
 * internal functions
 */

/* multiplication to transform point sizes into pixel sizes (value can be fraction) */
#define POINTS2PIXEL 1


void
initFonts ( Toolkit* X )
{
	/* nothing to do, we rely on uglFindFont */
}


/*
 * We should ultimately use this, but it currently hangs in uglFontFind, and
 * we have to guarantee that a valid font def is returned, no matter how bad
 * the match is
 */
void findFont ( char* faceName, int pixelSize, int bold, int italic, UGL_FONT_DEF* fd )
{
	UGL_FONT_DESC fDesc;
	UGL_FONT_DESC_PRIORITY fDescPrio;

    fDescPrio.pixelSize = 7;
    fDescPrio.weight    = 9;
    fDescPrio.serif     = UGL_FONT_DONT_CARE;
    fDescPrio.italic    = 8;
    fDescPrio.spacing   = UGL_FONT_DONT_CARE;
    fDescPrio.charSet   = UGL_FONT_DONT_CARE;
    fDescPrio.faceName  = 10;

    fDesc.faceName[0] = 0;
    strncat( fDesc.faceName, faceName, sizeof(fDesc.faceName));
	fDesc.pixelSize.min = 0;
	fDesc.pixelSize.max = pixelSize*2;

	if ( bold ){
		fDesc.weight.min = UGL_FONT_BOLD_MIN;
		fDesc.weight.max = UGL_FONT_BOLD_MAX;
	}
	else {
		fDesc.weight.min = UGL_FONT_BOLD_OFF;
		fDesc.weight.max = UGL_FONT_BOLD_MIN;
	}

	if ( italic ) {
		fDesc.italic = UGL_FONT_ITALIC;
	}
	else {
		fDesc.italic = UGL_FONT_UPRIGHT;
	}

    fDesc.charSet = UGL_FONT_ISO_8859_1;
    fDesc.spacing = UGL_FONT_DONT_CARE;

    uglFontFind( X->fid, &fDesc, &fDescPrio, fd);
}


/*******************************************************************************
 * exported functions
 */

void*
EXPORT(fntInitFont) ( JNIEnv* env, jclass clazz,
					  jstring jSpec, jint style, jint size )
{
	char *spec = java2CString( env, _(X), jSpec);
	UGL_FONT_ID fnt;
	UGL_FONT_DEF fd;

	DBG( awt_fnt, ("initFont: %s,  %d,%d\n", spec, (int)style, (int)size));

	/* map pointsize into pixelsize */

	size = (int) (size * POINTS2PIXEL);

	findFont( spec, size, (style & 0x01), (style & 0x02), &fd);
	fnt = uglFontCreate( X->fid, &fd);

	DBG( awt_fnt, (" -> font: %p\n", fnt));
  
	if ( !registerFs( fnt) ){
		return 0;  /* outch, we have too many fonts in use */
	}

	return (void*) fnt;
}

void
EXPORT(fntFreeFont) ( JNIEnv* env, jclass clazz, g_font_t* fnt )
{
	DBG( awt_fnt, ("freeFont: %p\n", fnt));

	if ( unregisterFs( fnt) ){
		uglFontDestroy( fnt);
	}
}


/*******************************************************************************
 * FontMetrics support
 */

void*
EXPORT(fntInitFontMetrics) ( JNIEnv* env, jclass clazz,
							 g_font_t* fnt, jobject metrics )
{
	UGL_FONT_METRICS fm;
	jclass fmClazz = (*env)->GetObjectClass( env, metrics);
	jfieldID id;
	int w, h;

	uglFontMetricsGet( fnt, &fm);

	DBG( awt_fnt, ("initFontMetrics: %p\n", fnt));

	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAdvance", "I")) ){
		(*env)->SetIntField( env, metrics, id, fm.maxAdvance);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "leading", "I")) ){
		(*env)->SetIntField( env, metrics, id, fm.leading);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "ascent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fm.maxAscent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxAscent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fm.maxAscent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "descent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fm.maxDescent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "maxDescent", "I")) ){
		(*env)->SetIntField( env, metrics, id, fm.maxDescent);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "height", "I")) ){
		(*env)->SetIntField( env, metrics, id, fm.height);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "fixedWidth", "I")) ){
		if ( fm.spacing == UGL_FONT_MONO_SPACED ){
			uglTextSizeGet( fnt, &w, &h, 1, " ");
		}
		else {
			w = 0;
		}
		(*env)->SetIntField( env, metrics, id, w);
	}
	if ( (id = (*env)->GetFieldID( env, fmClazz, "isWideFont", "Z")) ){
		(*env)->SetBooleanField( env, metrics, id, JNI_TRUE);
	}

	DBG( awt_fnt, (" -> metrics: %p\n", fnt));

	if ( !registerFs( fnt) ){
		return 0;
	}

	return fnt;
}

void
EXPORT(fntFreeFontMetrics) ( JNIEnv* env, jclass clazz, g_font_t* fnt )
{
	DBG( awt_fnt, ("initFontMetrics: %p\n", fnt));

	if ( unregisterFs( fnt) ){
		uglFontDestroy( fnt);
	}
}


jobject
EXPORT(fntGetWidths) ( JNIEnv* env, jclass clazz, g_font_t* fnt )
{
	int          n = 256, h;
	jarray       widths = (*env)->NewIntArray( env, 256);
	jboolean     isCopy;
	jint         *jw = (*env)->GetIntArrayElements( env, widths, &isCopy);
	register int i, j;
	char         c;

	DBG( awt_fnt, ("getWidths: %p\n", fnt));

	for ( i=0, j=0; j < n; i++, j++ ){
		c= i;
		uglTextSizeGet( fnt, (int*) &(jw[i]), &h, 1, &c);
	}

	(*env)->ReleaseIntArrayElements( env, widths, jw, 0);

	return widths;
}


jint
EXPORT(fntBytesWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt,
						jarray jBytes, jint off, jint len )
{
	jboolean  isCopy;
	jbyte     *jb = (*env)->GetByteArrayElements( env, jBytes, &isCopy);
	int       n = (*env)->GetArrayLength( env, jBytes);
	int       w, h;

	if ( off+len > n )
		len = n - off;

	if ( uglTextSizeGet( fnt, &w, &h, len, (jb+off)) == UGL_STATUS_ERROR ){
		w = 0;
	}

	(*env)->ReleaseByteArrayElements( env, jBytes, jb, JNI_ABORT);

	DBG( awt_fnt, ("bytesWidth: %p \"%3s..\" -> %d\n", fnt, jb+off, w));

	return w;
}

jint
EXPORT(fntCharWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt, jchar jc )
{
	int w, h;

	if ( uglTextSizeGetW( fnt, &w, &h, 1, &jc) == UGL_STATUS_ERROR ){
		w = 0;
	}

	DBG( awt_fnt, ("charWidth: %p \'%c\' -> %d\n", fnt, jc, w));

	return w;
}

jint
EXPORT(fntCharsWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt,
						jarray jChars, jint off, jint len )
{
	jboolean  isCopy;
	jchar    *jc = (*env)->GetCharArrayElements( env, jChars, &isCopy);
	int      n = (*env)->GetArrayLength( env, jChars);
	int      w, h;

	if ( off+len > n )
		len = n - off;

	if ( uglTextSizeGetW( fnt, &w, &h, len, (jc+off)) == UGL_STATUS_ERROR ){
		w = 0;
	}

	(*env)->ReleaseCharArrayElements( env, jChars, jc, JNI_ABORT);

	DBG( awt_fnt, ("charWidths: %p \"%s\" -> %d\n", fnt,
				   jchar2CString( _(X),(jc+off),len), w));

	return w;
}

jint
EXPORT(fntStringWidth) ( JNIEnv* env, jclass clazz, g_font_t* fnt, jstring jStr )
{
	jboolean isCopy;
	const jchar    *jc = (*env)->GetStringChars( env, jStr, &isCopy);
	int      len = (*env)->GetStringLength( env, jStr);
	int      w, h;

	if ( uglTextSizeGetW( fnt, &w, &h, len, jc) == UGL_STATUS_ERROR ){
		w = 0;
	}

	(*env)->ReleaseStringChars( env, jStr, jc);

	DBG( awt_fnt, ("stringWidth: %p \"%s\" -> %d\n", fnt,
				   java2CString( env,_(X),jStr), w));

	return w;
}
