/**
 * grlib.h - target driver library interface layer for no-native-wm AWTs
 *           (UGL 2 version)
 *
 * grlib.h contains most of AWT's "bottom interface" (aka DRIVER LIB
 * INTERFACE), which encapsulates the platform specific driver functions.
 *
 *  +----------------------------------------------------------------------+
 *  |Java class libraries                                                  |
 *  |                                                                      |
 *  | java.awt.Graphics (, Image, Window, Color, Font, ..)                 |
 *  |   drawLine() { Toolkit.graDrawLine() }                               |
 *  |                     |                                                |
 *  |  +----------------- | ---------------------------------------------+ |
 *  |  |java.awt.Toolkit  V                    [NATIVE AWT LIB INTERFACE]| |
 *  |  |  native graDrawLine(), wndCreateFrame(), clrGetPixelValue() ... | |
 *  |  +----------------- | -------------- | ---------- | ---------- | --+ |
 *  |                     |                |            |            |     |
 *  +-------------------- | -------------- | ---------- | ---------- | ----+
 *                        |                |            |            |
 *  +-------------------- | -------------- | ---------- | ---------- | ----+
 *  | native AWT library  |                V            V            V     |
 *  |              gra.c  V               wnd.c        clr.c        ...    |
 *  |                                                                      |
 *  |   Java_java_awt_Toolkit_graDrawLine(){ g_drawLine.. }                |
 *  |                                  +--------- | -----------------------+
 *  |                                  | grlib.h  V  [DRIVER LIB INTERFACE]|
 *  +----------------------------------+---- | ----------------------------+
 *                                           |
 *  +--------------------------------------- | ----------------------------+
 *  | driver library (GRX, MGL, UGL, ..)     V                             |
 *  |  (provides rasterizer)                                               |
 *  +----------------------------------------------------------------------+
 *
 * Unlike the "top interface" (NATIVE AWT LIB INTERFACE containing
 * native methods used from the unified AWT class library), it can be
 * implemented as a subset, i.e.there is no need to implement all it's
 * functions and types, if you don't use the relevant common module(s)
 * (like gra-common.c). All native methods are exported (and can be
 * overriden) from the platform specific ones (like gra.c).
 * 
 * This header is included from toolkit.h, which is in turn the glue header
 * for all native AWT modules.
 *
 * +-----------------------------------------------------------------------+
 * | gra.c (, wnd.c, img.c clr.c fnt.c..) (*)                              |
 * |  +------------------------------------------------------------------+ |
 * |  | toolkit.h (*)                                                    | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |  | toolkit-common.h                                            | | |
 * |  |  |   typedef struct CommonImage, CommonGraphics, CommonToolkit | | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |                                                                  | |
 * |  |  typedef struct Image, Graphics, Toolkit {CommonToolkit ct..}    | |
 * |  |                                                                  | |
 * |  |  +-------------------------------------------------------------+ | |
 * |  |  | grlib.h                                                     | | |
 * |  |  |   g_setColor(){.}, g_drawLine(){ GrDrawLine() }, .       |  | | |
 * |  |  +--------------------------------------------------------- | -+ | |
 * |  +------------------------------------------------------------ | ---+ |
 * |                                                                |      |
 * |  // exported functions with generic implementation             |      |
 * |  #define _graDrawLine Java_java_awt_Toolkit_graDrawLine..      |      |
 * |                                                                |      |
 * |  +------------------------------------------------------------ | ---+ |
 * |  | gra-common.c (, wnd-common.c, img-common.c, clr-common.c..) |    | |
 * |  |  _graDrawLine() { .. g_drawLine().. }                       V    | |
 * |  +------------------------------------------------------------------+ |
 * |                                                                       |
 * |  // platform specific exported functions                              |
 * |  Java_java_awt_Toolkit_graDrawImageScaled() {..} ..                   |
 * |                                                                       |
 * +-----------------------------------------------------------------------+
 *
 * The (*) marked files are platform dependent
 *
 * Note that this header is included in two parts - the top half defining
 * just generic types (which can be used in the CommonXX structures, the
 * bottom half defining the driver interface functions (which are usually
 * inlined).
 *
 * Functions and types starting with a "g_" prefix (e.g. g_drawLine) are
 * used in the standard common modules (have a fixed meaning), all other
 * definitions are just internal helpers, and can only be used from the
 * platform specific modules (e.g. gra.c), or from within grlib.h itself.
 *
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#if !defined (GRLIB_FUNCS)
/* this section is inlcuded before toolkit-common.h, to define
 * generic types which can be used in the common structure definitions
 */

/***************************************************************************
 * scalar & reference types 
 */

/* the type of native pixel values */
typedef  UGL_COLOR       g_pixel_t;

/* the type to set color palettes */
typedef  UGL_ARGB         g_paletteEntry_t;

/* (optional) transparent graphics context used for context save/restore
 * operations (if the driver lib provides/requires them)
 */
typedef  void*           g_contextSav_t;

/* (transparent) font type */
typedef  UGL_FONT        g_font_t;

/* the 'point' type of polygon related operations */
typedef UGL_POINT        g_point_t;

/* function to (type transparent) set point coordinates */
static inline void g_setPoint ( g_point_t* p, int x, int y )
{ 
	p->x = x; p->y = y;
}


#else /* defined GRLIB_FUNCS */
/* this is included before toolkit-common-post.h, i.e. we can
 * use these functions/macros there
 */

/***************************************************************************
 * internal helper (can be called from platform specific modules, but
 * NOT from common/generic code)
 */

static inline void _setBmpGc ( Image* img )
{
	if ( X->bmp != img->bmp ){
		uglDefaultBitmapSet( X->bmpGc, img->bmp);
		X->bmp = img->bmp;
	}
}

/******************************************************************************
 * visual detection
 */

/*
 * get the number of bits per display pixel (= visual)
 */
static inline unsigned int g_getBitsPerPixel ( Toolkit* X )
{
	UGL_COLOR_INFO ci;

	uglInfo( X->did, UGL_COLOR_INFO_REQ, &ci);

	return ci.depth;
}

/*
 * (optional) function to set system into RGB mode (some drivers
 * have to be switched explicitly)
 */
static inline void g_initRgbMode ( Toolkit* X )
{
	/* not required */
}


/******************************************************************************
 * pseudocolor palette (colormap) access & manipulation
 *
 * These macros are used (from clr.c) to set the system palette (we
 * need to evenly distribute RGBs, with additional capabilities to
 * allocate a few exact color requests).
 *
 * Note that some systems require to 'allocate' system color cells, i.e.
 * don't give us direct r/w access to a certain system palette entry/index.
 */

/*
 * Initialize system palette modification (reset, allocation etc.)
 */
static inline void g_initPalette ( CommonIdxMap* map )
{
	map->clrs = AWT_MALLOC( sizeof(UGL_ARGB)*map->nClrs);
}

/*
 * Set a system palette cell to the provided RGB components [0..255],
 * answer palette index where cell is stored (if the driver lib can't
 * change palette cells in-situ, and has to 'allocate' them).
 * 'idx' is the requested cell index (for in-situ changeable palettes)
 */
static inline int g_setPaletteEntry ( CommonIdxMap* map, int idx,
									  int r, int g, int b )
{
	map->clrs[idx] = UGL_MAKE_ARGB( 0xff, r, g, b);
	return idx;
}

/*
 * Query RGB components of palette cell at 'idx'
 */
static inline void g_queryPaletteEntry ( CommonIdxMap* map, int idx,
										 int* pr, int* pg, int* pb )
{
	*pr = UGL_RGB_RED( map->clrs[idx]);
	*pg = UGL_RGB_GREEN( map->clrs[idx]);
	*pb = UGL_RGB_BLUE( map->clrs[idx]);
}

/*
 * Make (modified) palette active (some drivers refer to this as
 * 'realizing' the palette)
 */
static inline void g_setPalette ( CommonIdxMap* map )
{
	int *idx = AWT_ALLOCA( sizeof(int) * map->nClrs);
	int i;

	UGL_COLOR *c = AWT_MALLOC( map->nClrs * sizeof(UGL_COLOR));
	
	for ( i=0; i<map->nClrs; i++ ){
		idx[i] = i;
	}

	if ( uglColorAlloc( X->did, map->clrs, idx, c, map->nClrs) == UGL_STATUS_ERROR ){
		DBG( awt_clr, ("failed to set colormap\n"));
	}

	AWT_FREE(c);
	AWT_ALLOCA_FREE( idx);
}

/*
 * add alpha channel [0..255] to jint pixel value
 */
static inline jint g_addAlpha ( jint rgb, g_pixel_t pix )
{
	return ((rgb & 0xff000000) | pix);
}

/*
 * get alpha channel [0..255] from jint pixel value
 */
static inline unsigned char g_getAlpha ( g_pixel_t pix )
{
	return (pix & 0xff000000) >> 24;
}

/*
 * remove alpha channel from jint pixel value
 */
static inline g_pixel_t g_stripAlpha ( g_pixel_t pix )
{
	return (pix & 0x00ffffff);
}


/*
 * Answer pixel value for given 8-8-8-8 ARGB
 * (remove switch for speed optimization)
 */
static inline jint g_pixelValue ( Toolkit* X, jint rgb )
{
	jint i, pix;

	switch ( _(X)->colorMode ) {
	case CM_TRUE:
		pix = (((rgb & _(X)->map.tclr->blueMask)  >> _(X)->map.tclr->blueShift) +
			   ((rgb & _(X)->map.tclr->greenMask) >> _(X)->map.tclr->greenShift) +
			   ((rgb & _(X)->map.tclr->redMask)   >> _(X)->map.tclr->redShift));
		break;

	case CM_PSEUDO:
		rgb &= 0xffffff;
		for ( i=0; i<_(X)->map.pclr->nExtra; i++ ){
			if ( _(X)->map.pclr->extra[i] == rgb ){
				return 216 + i;
			}
		}
		pix = _(X)->map.pclr->pix[ CELL_INDEX_8( JRED(rgb))  ]
			                     [ CELL_INDEX_8( JGREEN(rgb)) ]
			                     [ CELL_INDEX_8( JBLUE(rgb)) ];
		break;

	case CM_VGA:
		pix = getVGAPixelValue( X, rgb, 0);
		break;

	case CM_MONO:
		pix = (JRED(rgb)> 128) || (JGREEN(rgb)>128) || (JBLUE(rgb)>128) ? 1 : 0;
		break;

	default:
		pix = 0;
	}

	return g_addAlpha( rgb, pix);
}


/*
 * Answer RGB components for given pixel value
 */
static inline void g_rgbValues ( Toolkit* X, unsigned long pix, int* r, int* g, int* b )
{
	UGL_ARGB rgb;

	switch ( _(X)->colorMode ){
	case CM_TRUE:
		*r = ((pix & _(X)->map.tclr->pixRedMask)   << _(X)->map.tclr->redShift)   >> 16;
		*g = ((pix & _(X)->map.tclr->pixGreenMask) << _(X)->map.tclr->greenShift) >> 8;
		*b = ((pix & _(X)->map.tclr->pixBlueMask)  << _(X)->map.tclr->blueShift);
		break;

	case CM_PSEUDO:
		rgb = _(X)->map.pclr->common.clrs[pix & 0xff];
		*r = UGL_RGB_RED( rgb);
		*g = UGL_RGB_GREEN( rgb);
		*b = UGL_RGB_BLUE( rgb);	
		break;

	case CM_VGA:
		rgb = _(X)->map.vclr->common.clrs[pix & 0xf];
		*r = UGL_RGB_RED( rgb);
		*g = UGL_RGB_GREEN( rgb);
		*b = UGL_RGB_BLUE( rgb);	
		break;

	case CM_MONO:
		/* black = full intesity, white = no intesity */
		if ( pix ){
			*r = *g = *b = 0;
		}
		else {
			*r = *g = *b = 255;
		}
		break;
	}
}

/*
 * Answer pixel value with modified brightness
 */
static inline jint g_modPixelValue ( Toolkit* X, int pix, int mod )
{
	return modPixelValue( X, pix, mod);
}


/******************************************************************************
 * Graphics state - these functions are used to set the state of our
 * native cache 'Graphics' objects
 */

/*
 * Set global draw context according to the provided Image
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentImage ( Toolkit* X, Image* img )
{
	/* not required, gc always passed into primitives */
}

/*
 * Set global draw context according to the provided Graphics
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentGraphics ( Toolkit* X, Graphics* gr )
{
	/* not required, gc always passed into primitives */
}

/*
 * Temporarily store the current context (if the graphics driver
 * can't draw into aribtrary contexts)
 */
static inline void g_saveContext ( g_contextSav_t* cs )
{
	/* not required, gc always passed into primitives */
}

/*
 * Restore the current context from a temporarily saved one
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_restoreContext ( g_contextSav_t* cs )
{
	/* not required, gc always passed into primitives */
}

/*
 * Set the active clip rectangle according to the provided Graphics object
 */
static inline void g_setClip ( Graphics* gr )
{
	uglClipRectSet( gr->gc,
					_(gr)->curClip->x0, _(gr)->curClip->y0,
					_(gr)->curClip->x1, _(gr)->curClip->y1);
}

/*
 * one time init for new Graphics objects
 */
static inline void g_initGraphics ( Graphics* gr )
{
	gr->gc = uglGcCreate( X->did);
	uglBackgroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglLineStyleSet( gr->gc, UGL_LINE_STYLE_SOLID);
	uglLineWidthSet( gr->gc, 1);
}

/*
 * free resources for Graphics object
 */
static inline void g_freeGraphics ( Graphics* g )
{
	if ( gr->gc ){
		uglGcDestroy( gr->gc);
		gr->gc = 0;
	}
}


/*
 * Set the draw context of a Graphics object 'gr' from a Window object
 */
static inline void g_setWindowGraphics ( Graphics* gr, AwtWnd* wnd )
{
	uglDefaultBitmapSet( gr->gc, X->screenBmp);
}

/*
 * Set the draw context of a Graphics object 'gr' from a Image object
 */
static inline void g_setImageGraphics ( Graphics* gr, Image* img )
{
	uglDefaultBitmapSet( gr->gc, img->bmp);
}

/*
 * Set the draw context of a Graphics object 'gr' from a Graphics object
 */
static inline void g_setGraphicsGraphics ( Graphics* gr, Graphics *ogr )
{
	UGL_DDB_ID bmp;

	uglDefaultBitmapGet( ogr->gc, &bmp);
	uglDefaultBitmapSet( gr->gc, bmp);
}

/*
 * Set the foreground color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setColor ( Graphics* gr, g_pixel_t clr )
{
	gr->alpha = g_getAlpha( clr);

	gr->fg = g_stripAlpha( clr);
	gr->clr = (gr->xclr != -1) ? (gr->fg ^ gr->xclr) : gr->fg;

#if defined(USE_FULL_ALPHA)
	/* we support some alpha operations for drawable Images */
	if ((_(gr)->tgtType ==IMAGE) && !_(IMAGE_TGT(gr))->alpha && (gr->alpha<0xff)){
		Image* img = IMAGE_TGT(gr);
		_(img)->alpha = createImageAlpha( X, _(img)->width, _(img)->height);
	}
#endif

	uglForegroundColorSet( gr->gc, gr->clr);
}

/*
 * Answer foreground color of Graphics object
 */
static inline g_pixel_t g_getColor ( Graphics* gr )
{
	return ((gr->alpha << 24) | gr->fg);
}

/*
 * Set the background color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 * This is a optional function for drivers which can't draw
 * text with transparent background
 */
static inline void g_setBackColor ( Graphics* gr, g_pixel_t clr )
{
	gr->bg = g_stripAlpha( clr);

	/* watch out, we shouldn't set uglBackgroundColor here, because it
	 * switches all primitives to fill mode
	 */
}

/*
 * Set the font of a Graphics object
 * Note that we also might have to set the active font in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setFont ( Graphics* gr, g_font_t* fnt )
{
	gr->fnt = fnt;
	uglFontSet( gr->gc, fnt);
}

/*
 * Set the draw mode of a Graphics object to 'overpaint'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setPaintMode ( Graphics* gr )
{
	gr->xclr = -1;
	gr->clr  = gr->fg;

	uglRasterModeSet( gr->gc, UGL_RASTER_OP_COPY);
}

/*
 * Set the draw mode of a Graphics object to 'xor'
 * Note that we also might have to set the active draw mode in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setXORMode ( Graphics* gr, g_pixel_t xclr )
{
	gr->xclr = g_stripAlpha( xclr);
	gr->clr  = (gr->fg ^ gr->xclr);

	uglRasterModeSet( gr->gc, UGL_RASTER_OP_XOR);
}


/******************************************************************************
 * drawing operations
 */

/*
 * Set color of pixel
 * (Make sure it is clipped according to the Graphics, since many
 * graphics drivers don't provide clipping for this operation)
 */
static inline void g_putPixel ( Graphics* gr, int x, int y)
{
	uglPixelSet( gr->gc, x, y, gr->clr);
}

/*
 * Fill rectangle with background color of Graphics object
 */
static inline void g_clearRect ( Graphics* gr, int x0, int y0, int x1, int y1)
{
	uglForegroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglBackgroundColorSet( gr->gc, gr->bg);	
	uglRectangle( gr->gc, x0, y0, x1, y1);
	uglBackgroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglForegroundColorSet( gr->gc, gr->clr);
}

/*
 * Draw a line
 */
static inline void g_drawLine ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	uglLine ( gr->gc, x0, y0, x1, y1);
}

static inline void g_drawHLine ( Graphics* gr, int x0, int y0, int x1 )
{
	uglLine ( gr->gc, x0, y0, x1, y0);
}

static inline void g_drawVLine ( Graphics* gr, int x0, int y0, int y1 )
{
	uglLine ( gr->gc, x0, y0, x0, y1);
}


/*
 * Draw a rectangle (corner exclusive)
 */
static inline void g_drawRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	uglRectangle ( gr->gc, x0, y0, x1, y1);
}

/*
 * Fill a rectangle (corner inclusive)
 */
static inline void g_fillRect ( Graphics* gr, int x0, int y0, int x1, int y1 )
{
	uglForegroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglBackgroundColorSet( gr->gc, gr->clr);
	uglRectangle ( gr->gc, x0, y0, x1, y1);
	uglBackgroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglForegroundColorSet( gr->gc, gr->clr);

	if ( (_(gr)->tgtType == IMAGE) && _(IMAGE_TGT(gr))->alpha ){
		fillImageAlpha( _(IMAGE_TGT(gr))->alpha, x0, y0, x1, y1, gr->alpha);
	}
}

/*
 * Draw open arc (i.e. no lines from arc endpoints to center of oval).
 * Note that this actually is a oval, not just a circle
 */
static inline void g_drawOpenArc ( Graphics* gr, int x, int y,
								   int ra, int rb, int a1, int a2 )
{
	double r1, r2;
	int x1, y1, x2, y2;

	a1 = (a1 + 90) % 360;
	a2 = (a1 + a2) % 360;

	r1 = (double)a1 * M_PI/180.0;
	r2 = (double)a2 * M_PI/180.0;
	x1 = x + (ra * sin( r1));
	y1 = y + (rb * cos( r1));
	x2 = x + (ra * sin( r2));
	y2 = y + (rb * cos( r2));

	uglEllipse( gr->gc, x-ra, y-rb, x+ra, y+rb,
				x1, y1, x2, y2);
}


/*
 * Fill closed arc (pie chart)
 */
static inline void g_fillArc ( Graphics* gr, int x, int y,
							   int ra, int rb, int a1, int a2 )
{
	uglForegroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglBackgroundColorSet( gr->gc, gr->clr);
	g_drawOpenArc( gr, x, y, ra, rb, a1, a2);
	uglBackgroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglForegroundColorSet( gr->gc, gr->clr);
}

/*
 * Draw oval (ellipse)
 */
static inline void g_drawOval ( Graphics* gr, int x, int y, int ra, int rb )
{
	int xra = x+ra;
	uglEllipse( gr->gc, x-ra, y-rb, xra, y+rb, xra, y, xra, y);
}

/*
 * Fill oval (ellispe)
 */
static inline void g_fillOval ( Graphics* gr, int x, int y, int ra, int rb )
{
	uglForegroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglBackgroundColorSet( gr->gc, gr->clr);
	g_drawOval( gr, x, y, ra, rb);
	uglBackgroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglForegroundColorSet( gr->gc, gr->clr);
}

/*
 * Draw single-byte text
 */
static inline void g_drawByteText ( Graphics* gr, char* str,
									int len, int x, int y )
{
	uglTextDraw( gr->gc, x, y, len, str);
}

/*
 * Draw unicode text
 */
static inline void g_drawWideText ( Graphics* gr, jchar* str,
									int len, int x, int y )
{
	uglTextDrawW( gr->gc, x, y, len, str);
}

/*
 * Draw polygon (lines connecting vertices in 'points')
 */
static inline void g_drawPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	uglPolygon( gr->gc, nPoints, (UGL_POS*)points);
}

/*
 * Fill polygon
 */
static inline void g_fillPoly ( Graphics* gr, g_point_t* points, int nPoints)
{
	uglForegroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglBackgroundColorSet( gr->gc, gr->clr);
	g_drawPoly( gr, points, nPoints);
	uglBackgroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	uglForegroundColorSet( gr->gc, gr->clr);
}


/******************************************************************************
 * image drawing (bitblt related functions)
 */

/*
 * Bit block transfer in same context (Image or display)
 */
static inline void g_copyArea ( Graphics* gr, int x, int y,
								int xw, int yh, int xdelta, int ydelta)
{
	uglBitmapBlt( gr->gc, UGL_DEFAULT_ID, x, y, xw, yh,
				   UGL_DEFAULT_ID, x+xdelta, y+ydelta);
}

/*
 * Bit block transfer from Image context source (to Image or display)
 * This operation has to handle alpha channels. In case of a non-negative
 * 'bg' color, the transparent image pixels have to be set to this value
 */
static inline void g_drawImage ( Graphics* gr, Image* img, int dx, int dy,
								 int sx, int sy, int sxw, int syh, g_pixel_t bg )
{
	if ( (bg != -1) && (img->bmp->type == UGL_TDDB_TYPE) ){
		uglBackgroundColorSet( gr->gc, gr->clr);
		uglRectangle ( gr->gc, dx, dy, dx + (sxw-sx), dy + (syh-sy));
		uglBackgroundColorSet( gr->gc, UGL_COLOR_TRANSPARENT);
	}

	if ( _(img)->alpha ){
		drawAlphaImage( gr, img, sx, sy, dx, dy, sxw-sx, syh-sy, bg);
	}
	else {
		uglBitmapBlt( gr->gc, img->bmp, sx, sy, sxw, syh,
					   UGL_DEFAULT_ID, dx, dy);
	}
}

/*
 * grab pixels from a source Graphics, and punch them into a destination
 * Graphics (this is a generalized blit with a source that is not a
 * Java image)
 */
static inline void g_grabPixels ( Graphics* dgr, Graphics* sgr,
								  int sx, int sy, int w, int h )
{
	/* not yet */
}

/******************************************************************************
 * image processing
 *
 * The protocol for creating, initializing and rendering native images
 * has been adapted to support subsequent alpha channel processing (which
 * can be extended to offscreen images). The involved steps are:
 *
 *  g_initImage      Prepares Image object so that native pixel values
 *                   can be (permanently or temporarily) stored
 *
 *  g_setAlphaType+  Has to be called as soon as information about
 *                   alpha channels (none, masked, full) is available. 
 *                   After the call, (temporary or permanent) alpha
 *                   values can be stored for each pixel. Note that
 *                   it is possible to subsequently "increase" the
 *                   alpha type (none->masked->full, i.e. the function
 *                   can be called several times)
 *
 * g_putImagePixel+  set color, alpha information for a pixel
 *
 * g_completeImage   check if alpha info can be reduced (full->masked->none)
 *                   and transform any temporary (generic) information
 *                   into a platform specific format which is suitable
 *                   for rendering
 */

/*
 * Initialize a Image object so that we can set it's pixel values.
 * Note that we don't have alpha information, yet (i.e. the image
 * subsequently might be set to masked or full alpha)
 */
static inline int g_initImage ( Image* img )
{
	/* got too large to be inlined, it's in img.c */
	extern void initImage( Image* img );
	return initImage( img);
}


/*
 * Set alpha channel type of Image object. The supported 'transType'
 * values are NO_ALPHA, MASKED_ALPHA (transparent pixels), FULL_ALPHA.
 * After this operation, we have to be able to store pixel alpha
 * information (at least temporarily, to be later-on processed by
 * completeImage)
 */
static inline void g_setAlphaType ( Image* img, int alphaType )
{
	if ( alphaType && ! (_(img)->mask || _(img)->alpha) ){
		if ( alphaType == MASK_ALPHA ){
			_(img)->mask = createImageMask(  X, _(img)->width, _(img)->height);
		}
		else if ( alphaType == FULL_ALPHA ){
#if defined(USE_FULL_ALPHA)
			_(img)->alpha = createImageAlpha( X, _(img)->width, _(img)->height);
#else
			_(img)->mask = createImageMask(  X, _(img)->width, _(img)->height);
#endif
		}
	}
}

/*
 * set transparency color of the image in the native layer
 */
static inline void g_setTranspValue ( Image* img, g_pixel_t clr )
{
 
}

/*
 * Transform temporarily stored image information (both pixel color and
 * alpha values) into a representation which is suitable for drawing
 * the image. Free any temporary information.
 */
static inline void g_completeImage ( Image* img )
{
	/* got too large to be inlined, it's in img.c */
	extern void completeImage( Image* img);
	completeImage( img);
}

/*
 * Clean up platform specific Image components
 */
static inline void g_freeImage ( Image* img )
{
	if ( img->bmp ){
		if ( img->bmp->type == UGL_TDDB_TYPE ){
			uglTransBitmapDestroy( X->did, (UGL_TDDB_ID)img->bmp);
		}
		else {
			uglBitmapDestroy( X->did, img->bmp);
		}
		img->bmp = 0;
	}
}


/*
 * Get alpha value of image pixel [0..255]
 */
static inline int g_getPixelAlpha ( Image* img, int x, int y )
{
	/* we actually need just the full alpha image value (it's used in
	 * the generic scaling)
	 */
	UGL_COLOR clr;

    if ( img->bmp->type == UGL_TDDB_TYPE ) {
		_setBmpGc( img);
		uglPixelGet( X->bmpGc, x, y, &clr);
		return (clr == UGL_COLOR_TRANSPARENT) ? 0 : 0xff;
	}
	else if ( _(img)->alpha ){
		return getAlpha( _(img)->alpha, x, y);
	}
	else {
		return 0xff;
	}
}

/*
 * Answer alpha type of image. There are three supported
 * alpha type values
 *  - NO_ALPHA     : no alpha channel at all
 *  - MASKED_ALPHA : alpha = [0,255] (transparent pixels)
 *  - FULL_ALPHA   : alpha = [0..255]
 */
static inline int g_getAlphaType ( Image* img )
{
	if ( img->bmp->type == UGL_TDDB_TYPE )
		return MASK_ALPHA;
	else if ( _(img)->alpha )
		return FULL_ALPHA;
	else
		return NO_ALPHA;
}


/*
 * Set pixel value in Image (with optional alpha value [0..255]).
 * This is called before g_completeImage(), i.e. it can use
 * temporary pixel/alpha storage
 */
static inline void g_putImagePixel ( Image* img , int x, int y,
									 g_pixel_t pix, unsigned char alpha )
{
	if ( alpha != 0xff ){
		if ( _(img)->alpha ){
			putAlpha( _(img)->alpha, x, y, alpha);
		}
		else if ( _(img)->mask ){
			unsetMask( _(img)->mask, x, y);
			pix = 0;
		}
	}

	if ( _(X)->nColorPlanes <= 8 ){
		((UINT8*)img->pels)[ (_(img)->width * y) + x ] = (UINT8)pix;
	}
	else if ( _(X)->nColorPlanes <= 16 ){
		((UINT16*)img->pels)[ (_(img)->width * y) + x ] = (UINT16)pix;
	}
	else {
		((UGL_COLOR*)img->pels)[ (_(img)->width * y) + x ] = (UGL_COLOR)pix;
	}
}

/*
 * Answer pixel value of already completed Image
 */
static inline g_pixel_t  g_getImagePixel ( Image* img, int x, int y )
{
	UGL_COLOR pix;

	_setBmpGc( img);
	uglPixelGet( X->bmpGc, x, y, &pix);

	return pix;
}


/******************************************************************************
 * event processing
 */

/*
 * Initialize keyboard input
 */
static inline void g_initKeyboard ( Toolkit* X )
{
	/* nothing to do */
}

/*
 * Initialize moude input
 */
static inline void g_initMouse ( Toolkit* X )
{
	uglCursorInit( X->did, _(X)->width, _(X)->height,
				   _(X)->width/2, _(X)->height/2); 
	uglCursorImageSet( X->did, X->cursors[0]);
	uglCursorOn( X->did);
	uglCursorMove( X->did, _(X)->width/2, _(X)->height/2);
}

/*
 * Get current horizontal pointer location
 */
static inline int g_getMouseX ( Toolkit* X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);
	return evt->x;
}

/*
 * Get current vertical pointer location
 */
static inline int g_getMouseY ( Toolkit* X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);
	return evt->y;
}

/*
 * Get nmber of pressed mouse button
 */
static inline int g_getMouseBtn ( Toolkit* X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);

	if ( evt->type.pointer.buttonState & UGL_POINTER_BUTTON1 )
		return 1;
#ifdef THREE_BUTTON_MOUSE
	if ( evt->type.pointer.buttonState & UGL_POINTER_BUTTON2 )
		return 2;
	if ( evt->type.pointer.buttonState & UGL_POINTER_BUTTON3 )
		return 3;
#else
	if ( evt->type.pointer.buttonState & (UGL_POINTER_BUTTON2|UGL_POINTER_BUTTON3) )
		return 3;
#endif

	return 1;  /* default is to act like left button */
}

/*
 * Get Java keycode of current input event
 */

/* we can finally assume this is constant (hence not linked into X) */
extern unsigned char keyCodes[], ckeyCodes[], fkeyCodes[];

static inline int g_getKeyCode ( Toolkit* X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);
	int k = evt->type.keyboard.key;

	if ( k < 0xff ){
		return keyCodes[k];
	}
	else if ( k >= UGL_UNI_PRIVATE ){
		if ( k <= UGL_UNI_PROGRAM ){
			return ckeyCodes[ k - UGL_UNI_PRIVATE ];
		}
		else if ( (k > UGL_UNI_FUNC_BASE) && (k <= UGL_UNI_F12) ){
			return fkeyCodes[ k - UGL_UNI_FUNC_BASE-1 ];
		}
	}

	return 0;
}

/*
 * Get Java keychar of current input event
 */
static inline int g_getKeyChar ( Toolkit* X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);

	if ( (unsigned)evt->type.keyboard.key < (unsigned)UGL_UNI_PRIVATE ){
		return evt->type.keyboard.key;
	}

	return 0;
}

/*
 * Make sure the next event will be a (artificial) mouse motion event
 * (with the last known pointer position). Used to retarget mouse for popups
 */
static inline void g_queueMouseMotion ( JNIEnv* env, Toolkit* X )
{
	UGL_INPUT_EVENT *evt = (UGL_INPUT_EVENT*) &(X->event);

	evt->x = _(X)->xMouse;
	evt->y = _(X)->yMouse;

	motionNotify( env, X);
}

/*
 * notify event system that we have a pending paint request
 */
static inline void g_queuePaint ( JNIEnv* env, Toolkit* X)
{
	UGL_EVENT event;

	DBG( awt_evt, ("queuePaint\n"));

#if defined (USE_NATIVE_DISPATCHER_LOOP)
	while ( paintNotify( env, X) );
#else  /* !USE_NATIVE_DISPATCHER_LOOP */
	event.header.type = WAKEUP_EVENT;
	event.header.category = UGL_EVENT_CATEGORY_INPUT;
	uglEventPost( X->eid, X->qid, &event, sizeof( event));
#endif /* USE_NATIVE_DISPATCHER_LOOP */
}

#endif /* GRLIB_FUNCS */
