/**
 * img.c - platform specific image support (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define NO_DEBUG

#include "toolkit.h"

/******************************************************************************
 * generic (common) exported methods
 */
#define _imgCreateImage         EXPORT(imgCreateImage)
#define _imgFreeImage           EXPORT(imgFreeImage)
#define _imgCreateScreenImage   EXPORT(imgCreateScreenImage)

#define _imgSetIdxPels          EXPORT(imgSetIdxPels)
#define _imgSetRGBPels          EXPORT(imgSetRGBPels)
#define _imgComplete            EXPORT(imgComplete)

#define _imgCreateFromFile      EXPORT(imgCreateFromFile)
#define _imgCreateFromData      EXPORT(imgCreateFromData)
#define _imgProduceImage        EXPORT(imgProduceImage)
#define _imgSetFrame            EXPORT(imgSetFrame)
#define _imgGetWidth            EXPORT(imgGetWidth)
#define _imgGetHeight           EXPORT(imgGetHeight)
#define _imgGetFrameCount       EXPORT(imgGetFrameCount)
#define _imgGetLoopCount        EXPORT(imgGetLoopCount)
#define _imgGetLatency          EXPORT(imgGetLatency)
#define _imgGetNextFrame        EXPORT(imgGetNextFrame)


/******************************************************************************
 * import generic functionality
 */
#include "../img-common.c"


/******************************************************************************
 * platform specific things
 */

int
initImage ( Image* img )
{
	UGL_DIB dib;

	if ( !(_(img)->flags & IMG_IS_DRAWABLE) ) {
		dib.width  = _(img)->width;
		dib.height = _(img)->height;
		dib.stride = _(img)->width;

		if ( _(X)->nColorPlanes == 4 ) {  /* VGA */
			dib.imageFormat = UGL_INDEXED_4;
			dib.colorFormat = UGL_DEVICE_COLOR;
			dib.clutSize = 0;
			dib.pClut  = 0;		

		else if ( _(X)->nColorPlanes == 8 ) {
			dib.imageFormat = UGL_INDEXED_8;
			dib.colorFormat = UGL_DEVICE_COLOR;
			dib.clutSize = 0;
			dib.pClut  = 0;
		}
		else if ( _(X)->nColorPlanes == 16 ) {
			dib.imageFormat = UGL_DIRECT;
			dib.colorFormat = UGL_RGB565;
			dib.clutSize = 0;
			dib.pClut  = 0;

		}
		else {
			dib.imageFormat = UGL_DIRECT;
			dib.colorFormat = UGL_DEVICE_COLOR;
			dib.clutSize = 0;
			dib.pClut  = 0;
		}
	
		if ((img->bmp = uglBitmapCreate( X->did, &dib, UGL_DIB_INIT_NONE, 0, 0)) == NULL) {
			DBG (awt_mem, ("memory allocation failure in initImage\n"));
			return 0;
		}

		/* here we get pretty close to the inner workings of UGL, but
		 * that's justified by the fct that we save temporary pel space
		 */
		img->pels = ((UGL_GEN_DDB *) img->bmp)->image;
		DBG( awt_img, ("initImage: img->pels == %d\n", img->pels));

		return 1;
	}
	else {
		dib.width  = _(img)->width;
		dib.height = _(img)->height;
		dib.stride = _(img)->width;
		dib.imageFormat = UGL_DIB_DIRECT;
		dib.colorFormat = UGL_DEVICE_COLOR;
		dib.clutSize = 0;
		dib.pClut  = 0;
		dib.pImage = 0;

		img->bmp = uglBitmapCreate( X->did, &dib, UGL_DIB_INIT_NONE, 0, 0);

		if (img->bmp == NULL) {
		  return 0;
		}
		else {
		  return 1;
		}
	}
}


void
completeImage ( Image* img )
{
	/* if we have a full alpha mask, try to reduce it to a simple mono mask
	 * (it's more efficient to blit and saves memory)
	 */
	if ( _(img)->alpha && !needsFullAlpha(_(img)->alpha, 0.05) ){
		if ( (_(img)->mask = createImageMask(  X, _(img)->width, _(img)->height)) ){
    		reduceAlpha( img, _(img)->alpha, 128);
    		freeImageAlpha( _(img)->alpha);
    		_(img)->alpha = 0;
        }
	}

	if ( _(img)->mask ) {
		UGL_MDIB			mdib;
	    UGL_MDDB_ID			mddb;
		UGL_TDDB_ID			tddb;
		
		mdib.width  = _(img)->mask->width;
		mdib.height = _(img)->mask->height;
		mdib.stride = _(img)->mask->scanline * MASKWORDSIZE;
		mdib.pImage = _(img)->mask->buf;           /* bit order ? */

		if ((mddb = uglMonoBitmapCreate (X->did, &mdib, UGL_DIB_INIT_DATA, 0, 0)) == UGL_NULL) {
			return;
		}

		if ((tddb = uglTransBitmapCreateFromDDB (X->did, img->bmp, mddb, 0)) == UGL_NULL) {
			uglBitmapDestroy (X->did, mddb);
			return;
		}

		/* Switch the img->bmp, which used to point to the DDB for the main
		 * image, to point to the new masked image (the original image pointer
		 * is now tucked inside.
		 */
		img->bmp = tddb;

		freeImageMask( _(img)->mask); /* it's in the TDDB now */
		_(img)->mask = 0;
	}
}


/************************************************************************************
 * exported native methods
 */

void*
EXPORT(imgCreateScaledImage) ( JNIEnv* env, jclass clazz,
							   Image* img, int width, int height )
{
	UGL_DIB dib;
	UGL_MDIB mdib;
	Image * scaledImg;

	DBG( awt_img, ("createScaledImage: %p  %d,%d\n", img, width, height));

	/* unfortunately, we can't use createImage(w,h,TRUE), because it
	 * isn't aware of TDDBs
	 */

	if ( ! scaledImg = AWT_CALLOC( 1, sizeof( Image)) ){
		return 0;
	}

	_(img)->trans      = -1;
	_(img)->width      = width;
	_(img)->height     = height;

	dib.width  = width;
	dib.height = height;
	dib.stride = width;
	dib.imageFormat = UGL_DEVICE_COLOR;
	dib.colorFormat = UGL_DEVICE_COLOR;
	dib.clutSize = 0;
	dib.pClut  = 0;
	dib.pImage = 0;

	if ( img->bmp->type == UGL_TDDB_TYPE ){
		mdib.width  = dib.width;
		mdib.height = dib.height;
		mdib.stride = dib.width;
		mdib.pImage = 0;

		scaledImg->bmp = uglTransBitmapCreate( X->did, &dib, &mdib,
										 UGL_DIB_INIT_NONE, 0, 0);
	}
	else {
		scaledImg->bmp = uglBitmapCreate( X->did, &dib, UGL_DIB_INIT_NONE, 0, 0);
	}

	if ( !_(img)->alpha ) {
		_setBmpGc( scaledImg);
		uglBitmapStretchBlt ( X->bmpGc, img->bmp,
							  0, 0, _(img)->width, _(img)->height, 
							  scaledImg->bmp,
							  0, 0, width, height);
	}
	else {
		/* scaled alpha images not yet supported */
	}

	return scaledImg;
}
