/**
 * wnd.c - common native toplevel window related functions
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


/**********************************************************************************
 * AwtWnd ctor/dtor and list functions
 */

static AwtWnd *windowCache;

void
initWindows ( Toolkit* X )
{
	LinkedRect *r;
	int        n = 32;

	/* LinkedRects are high frequency resources, better get some of them beforehand */
	r = (LinkedRect*) AWT_CALLOC( n, sizeof(LinkedRect));
	while ( n-- )
		freeRect( r++);
}


void*
newWindow ( int x, int y, int width, int height )
{
	AwtWnd *w;

	if ( windowCache ) {
		w = windowCache;
		windowCache = w->next;
	}
	else {
		w = (AwtWnd*) AWT_MALLOC( sizeof( AwtWnd));
	}

	setRect( &w->bounds, x, y, x+width-1, y+height-1, 0);

	w->mapped = 0;
	w->visibility = INVISIBLE;
	w->visSegments = 0;
	w->visTag = 0;

	w->update = 0;
	w->updateRect = 0;

	w->next = 0;
	w->prev = 0;

	return w;
}

void
freeWindow ( AwtWnd* wnd )
{
	if ( wnd->visSegments )
		freeRectList( wnd->visSegments);

	wnd->next = windowCache;
	windowCache = wnd;
}


void
addWindowToList ( Toolkit* X, AwtWnd* wnd )
{
	wnd->next = _(X)->top;

	if ( _(X)->top )
		_(X)->top->prev = wnd;
	else
		_(X)->bottom = wnd;

	_(X)->top = wnd;
}

#ifdef DEBUG
void
dumpWindowList ( Toolkit* X )
{
	AwtWnd* w;
	int     i;

	for ( i=0, w=_(X)->top; w; w=w->next, i++ ) {
		printf( "# %d -------- %p (%d,%d %d,%d) map: %d, tag: %d, vis: %d, upd: %d  n: %p, p: %p\n", i, w,
				w->bounds.x0, w->bounds.y0, w->bounds.x1, w->bounds.y1,
				w->mapped, w->visTag, w->visibility, w->update, w->next, w->prev);
		dumpLinkedRect( w->visSegments);
	}
}
#endif

void
setTop ( Toolkit* X, AwtWnd* wnd )
{
	if ( wnd != _(X)->top ) {
		wnd->prev->next = wnd->next;

		if ( wnd == _(X)->bottom )
			_(X)->bottom = wnd->prev;
		else
			wnd->next->prev = wnd->prev;

		wnd->prev = 0;
		wnd->next = _(X)->top;
		_(X)->top->prev = wnd;
		_(X)->top = wnd;
	}
}

void
setBottom ( Toolkit* X, AwtWnd* wnd )
{
	if ( wnd != _(X)->bottom ) {
		wnd->next->prev = wnd->prev;

		if ( wnd == _(X)->top )
			_(X)->top = wnd->next;
		else
			wnd->prev->next = wnd->next;

		wnd->next = 0;
		wnd->prev = _(X)->bottom;
		_(X)->bottom->next = wnd;
		_(X)->bottom = wnd;
	}
}

void
removeWindowFromList ( Toolkit* X, AwtWnd* wnd )
{
	if ( wnd->prev ){
		wnd->prev->next = wnd->next;
	}
	else {
		_(X)->top = wnd->next;
		if ( _(X)->top )
			_(X)->top->prev = 0;
	}

	if ( wnd->next ){
		wnd->next->prev = wnd->prev;
	}
	else {
		_(X)->bottom = wnd->prev;
		if ( _(X)->bottom )
			_(X)->bottom->next = 0;
	}
}

/*
 * this is used to update the visSegments list of a Window, which is solely based
 * on its stacking order and dimensions, i.e. gets changed infrequently, just by
 * toFront(), toBack(), setBounds(), setVisible(). The visSegments are subsequently
 * used to efficiently calculate the clipSegments for Graphics objects (which can
 * be a performance critical operation).
 * Note that the RootWindow has just one visSegment, and that's the entire window,
 * i.e. doesn't need to get updated. This is because the RootWindow is always has the
 * lowest stacking order, but it's Graphics objects can be used to draw on the whole
 * screen (like it would be on top).
 */
int
recalcVisibility ( AwtWnd *wTop, AwtWnd* wnd )
{
	AwtWnd      *w;
	LinkedRect  *list;

	if ( wnd->visSegments ) {
		freeRectList( wnd->visSegments);
	}

	list = newRect( wnd->bounds.x0, wnd->bounds.y0, wnd->bounds.x1, wnd->bounds.y1, 0, 0);

	for ( w=wTop; w && (w != wnd); w=w->next ) {
		if ( w->mapped ) {
			list = getDiffPart( list, &w->bounds);
		}
	}

	wnd->visTag++;
	wnd->visSegments = list;

	if ( !wnd->visSegments ){
		wnd->visibility = INVISIBLE;
	}
	else if ( (wnd->visSegments->x0 != wnd->bounds.x0) ||
			  (wnd->visSegments->y0 != wnd->bounds.y0) ||
			  (wnd->visSegments->x1 != wnd->bounds.x1) ||
			  (wnd->visSegments->y1 != wnd->bounds.y1) ) {
		wnd->visibility = PARTLY_VISIBLE;
	}
	else {
		wnd->visibility = FULLY_VISIBLE;
	}

	return wnd->visibility;
}

/*
 * This is used to explicitly determine which parts of the RootWindow have to be
 * redrawn after a toplevel got resized or mapped/unmapped.
 * Note that this is required because RootWindow graphics objects don't get clipped
 * (can draw on top of everything), i.e. we can't simply fill the whole screen to
 * redraw the background.
 */
void
checkRootUpdates ( JNIEnv* env, Toolkit* X, int x0, int y0, int x1, int y1 )
{
	AwtWnd      *w;
	LinkedRect  *list, *r;

	list = newRect( x0, y0, x1, y1, 0, 0);

	for ( w=_(X)->top; w; w=w->next ) {
		if ( w->mapped ) {
			list = getDiffPart( list, &w->bounds);
		}
	}

	if ( list ){
 		if ( _(X)->rootUpdates ) {
			for ( r=_(X)->rootUpdates; r->next; r=r->next );
			r->next = list;
		}
		else {
			_(X)->rootUpdates = list;
		}
		_(X)->pendingUpdates++;
		g_queuePaint( env, X);
	}
}


/*
 * This is a accumulative version, which is more suitable in case it
 * is better to do one big redraw instead of multiple (overlapping) ones
 */
void
updateWindow ( JNIEnv* env, Toolkit* X, AwtWnd* w, LinkedRect* r )
{
	LinkedRect *ur;

	if ( w->updateRect ) {
		ur = w->updateRect;
		w->updateRect = getUnion( ur, r);
		freeRect( ur);
		freeRect( r);
	}
	else {
		w->updateRect = r;
	}

	_(X)->pendingUpdates++;
	g_queuePaint( env, X);
}

/*
 * This is for cases where the update rects are supposed to be
 * small and disjunct (e.g. for scrolling in a offscreen blit in
 */
void
addWindowUpdate ( JNIEnv* env, Toolkit* X, AwtWnd* w, LinkedRect* r )
{
	if ( w->updateRect ) {
		appendRect( w->updateRect, r);
	}
	else {
		w->updateRect = r;
	}

	_(X)->pendingUpdates++;
	g_queuePaint( env, X);
}

/**********************************************************************************
 * exported functions
 */


void
_wndSetTitle ( JNIEnv* env, jclass clazz, void* wnd, jstring jTitle )
{
	DBG( awt_wnd, ("setTitle: %p %s\n", wnd, java2CString( env, _(X), jTitle)));
}


void
_wndSetResizable ( JNIEnv* env, jclass clazz,
				   void* wnd, jboolean isResizable,
				   int x, int y, int width, int height )
{
	DBG( awt_wnd, ("setResizable: %p  %d  %d,%d,%d,%d\n", wnd,isResizable, x,y,width,height));
}

void*
_wndCreateFrame ( JNIEnv* env, jclass clazz, jstring jTitle,
				  jint x, jint y, jint width, jint height,
				  jint jCursor, jint clrBack, jboolean isResizable )
{
	AwtWnd *w;

	DBG( awt_wnd, ("createFrame: \"%s\"  %d,%d,%d,%d\n",
				   java2CString( env, _(X), jTitle),
				   (int)x, (int)y, (int)width, (int)height));
  
	if ( width < 0 ) width = 0;
	if ( height < 0 ) height = 0;

	w =  newWindow( x, y, width, height);
	addWindowToList( X, w);
	
	return w;
}


void*
_wndCreateWindow ( JNIEnv* env, jclass clazz, void* owner,
				   jint x, jint y, jint width, jint height,
				   jint jCursor, jint clrBack )
{
	AwtWnd *w;

	DBG( awt_wnd, ("createWindow: %p, %d,%d,%d,%d\n", 
				   owner, (int)x, (int)y, (int)width, (int)height));

	if ( width < 0 ) width = 0;
	if ( height < 0 ) height = 0;

	w = newWindow( x, y, width, height);

	if ( !_(X)->root ) {
		/*
		 * this is a deviation to ordinary Windows: even if all windows are in front
		 * of it, we want to be able to (over)paint everything with a root Graphics
		 * (e.g. for rubber bands, icons etc.). We therefore pretend a (not chained)
		 * top window instead.
		 */
		_(X)->root = w;
		w->mapped = 1;
		w->visibility = FULLY_VISIBLE;

		setRect( &w->bounds, x, y, x+width-1, y+height-1, 0);
		w->visSegments = newRect( w->bounds.x0, w->bounds.y0, w->bounds.x1, w->bounds.y1, 0, 0);
		w->visTag = 1;
	}
	else {
		addWindowToList( X, w);
	}

	return w;
}


void*
_wndCreateDialog ( JNIEnv* env, jclass clazz,
				   void* owner, jstring jTitle,
				   jint x, jint y, jint width, jint height,
				   jint jCursor, jint clrBack, jboolean isResizable )
{
	AwtWnd *w;

	DBG( awt_wnd, ("createDialog: \"%s\"  %p  %d,%d,%d,%d\n",
				   java2CString( env,_(X), jTitle), owner,
				   (int)x, (int)y, (int)width, (int)height));

	if ( width < 0 ) width = 0;
	if ( height < 0 ) height = 0;

	w =  newWindow( x, y, width, height);

	addWindowToList( X, w);
	return w;
}


void
_wndRequestFocus ( JNIEnv* env, jclass clazz, void* wnd )
{
	DBG( awt_wnd, ("requestFocus: %p\n", wnd));  
}



void
_wndSetBounds ( JNIEnv* env, jclass clazz, AwtWnd* wnd,
				jint x, jint y, jint width, jint height,
				jboolean isResizable )
{
	AwtWnd *w;
	LinkedRect b;
	LinkedRect *r;

	DBG( awt_wnd, ("setBounds:  %p  %d,%d %d,%d  %d\n",
				   wnd, (int)x, (int)y, (int)width, (int)height, isResizable));

	/* remember old coords, we might have to repaint parts of the RootWindow */
	if ( wnd->mapped ){
		setRect( &b, wnd->bounds.x0, wnd->bounds.y0, wnd->bounds.x1, wnd->bounds.y1, 0);
	}

	if ( width < 0 ) width = 0;
	if ( height < 0 ) height = 0;

	/* set new coords first, it might affect lower sibling visibility */
	setRect( &wnd->bounds, x, y, x+width-1, y+height-1, 0);

	for ( w=wnd; w; w=w->next ){
		if ( w->mapped && recalcVisibility( _(X)->top, w) ) {
			if ( w == wnd ) {                                    /* redraw completely */
				updateWindow( env, X, w, newRect( 0, 0, width, height, 0,0));
			}
			else if ( (r = getIntersection( &b, &w->bounds)) ){  /* redraw unobscured parts */
				translateRect( r, -w->bounds.x0, -w->bounds.y0);
				updateWindow( env, X, w, r);
			}
		}
	}

    /* we don't have to recalc the RootWindow visSegments (see above - its Graphics
	 * objects never get clipped), but have to make sure we issue explicit paint events
	 * for all parts which become un-obscured because of setBounds
	 */
	if ( wnd->mapped ) {
		checkRootUpdates( env, X, b.x0, b.y0, b.x1, b.y1);
		retargetMouse( env, X);
	}
}


void
_wndRepaint ( JNIEnv* env, jclass clazz, AwtWnd* wnd,
			  jint x, jint y, jint width, jint height )
{
	DBG( awt_wnd, ("repaint: %p, %d,%d,%d,%d\n",
				   wnd, (int)x, (int)y, (int)width, (int)height));

	updateWindow( env, X, wnd, newRect( x, y, x+width, y+height, 0, 0));
}


void
_wndSetIcon ( JNIEnv* env, jclass clazz, AwtWnd* wnd, void* img )
{
	DBG( awt_wnd, ("setIcon: %p  %p\n", wnd, img));
}


void
_wndSetVisible ( JNIEnv* env, jclass clazz, AwtWnd* wnd, jboolean showIt )
{
	AwtWnd      *w;
	LinkedRect  *r;

	DBG( awt_wnd, ("setVisible: %p  %d\n", wnd, showIt));
	wnd->mapped = showIt;

	if ( showIt ) {
		for ( w=wnd; w; w=w->next ){
			if ( w->mapped )
				recalcVisibility( _(X)->top, w);
		}
		if ( wnd->visibility )
			updateWindow( env, X, wnd, newDim( &wnd->bounds));
	}
	else {
		for ( w=wnd->next; w; w=w->next ){
			if ( w->mapped && recalcVisibility( _(X)->top, w) ) {
				if ( (r = getIntersection( &wnd->bounds, &w->bounds)) ){
					translateRect( r, -w->bounds.x0, -w->bounds.y0);
					updateWindow( env, X, w, r);
				}
			}
		}

	    /* see wndSetBounds reg. why we explcitly have to redraw parts of the RootWindow */
		checkRootUpdates( env, X, wnd->bounds.x0, wnd->bounds.y0, wnd->bounds.x1, wnd->bounds.y1);
	}

	retargetMouse( env, X);
}


void
_wndToBack ( JNIEnv* env, jclass clazz, AwtWnd* wnd )
{
	AwtWnd *w;
	AwtWnd *w0= wnd->next;
	LinkedRect *r;

	DBG( awt_wnd, ("toBack: %p\n", wnd));

	setBottom( X, wnd);

	for ( w=w0; w; w=w->next ) {
		if ( w->mapped && recalcVisibility( _(X)->top, w) ) {
			if ( (w != wnd) && (r = getIntersection( &wnd->bounds, &w->bounds)) ){
				translateRect( r, -w->bounds.x0, -w->bounds.y0);
				updateWindow( env, X, w, r);
			}
		}
	}

	retargetMouse( env, X);
}


void
_wndToFront ( JNIEnv* env, jclass clazz, AwtWnd* wnd )
{
	AwtWnd *w;
	AwtWnd *w1 = wnd->next;

	DBG( awt_wnd, ("toFront: %p\n", wnd));

	setTop( X, wnd);

	for ( w=_(X)->top; w != w1; w=w->next ) {
		if ( w->mapped )
			recalcVisibility( _(X)->top, w);
	}

	if ( wnd->mapped )
		updateWindow( env, X, wnd, newDim( &wnd->bounds));

	retargetMouse( env, X);
}


jint
_wndDestroyWindow ( JNIEnv* env, jclass clazz, AwtWnd* wnd )
{
	if ( wnd ) {
		DBG( awt_wnd, ("destroy: %p\n", wnd));

		_wndSetVisible( env, clazz, wnd, 0);

		if ( wnd == _(X)->mouseWindow )
			_(X)->mouseWindow = 0;
		if ( wnd == _(X)->lastWindow )
			_(X)->lastWindow = 0;

		removeWindowFromList( X, wnd);
		freeWindow( wnd);
	}
	return (1);
}


void
_wndSetFrameInsets ( JNIEnv* env, jclass clazz,
					 jint top, jint left, jint bottom, jint right )
{
	/* we happily can ignore this, handling all decos in Java itself */
	DBG( awt_wnd, ("setFrameInsets:  %d,%d,%d,%d\n",
				   (int)top, (int)left, (int)bottom, (int)right));
}


void
_wndSetDialogInsets ( JNIEnv* env, jclass clazz,
					  jint top, jint left, jint bottom, jint right )
{
	/* we happily can ignore this, handling all decos in Java itself */
	DBG( awt_wnd, ("setDialogInsets:  %d,%d,%d,%d\n",
				   (int)top, (int)left, (int)bottom, (int)right));
}
