/**
 * gra.cpp - drawing support for PEG AWT 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"

struct Graphics;
Graphics *lastGr;

#define  WINDOW   0
#define  IMAGE    1
#define  GRAPHICS 2

// the PegThing* seems to be just a dummy in case we draw in a bitmap..
#define  PT  (wnd ? wnd->pt : X->presman)

struct Graphics {
	MappedWindow *wnd;
	PegBitmap    *bmp;

	int          fg;
	int          bg;
	int          flags;

	PegFont      *fnt;
	PegRect      clip;         // our own clipping
	int          x0;
	int          y0;

	// These functions have to wrap each primitive draw operation. This is to
	// ensure the propper PEG context (invalid area, clipping,..)

	int startDraw()
	{
		if ( wnd ) {
			if ( !wnd->isVisible )
				return 0;

			// we still need this if we don't want to handle mouse pointer redrawing explicitly
			X->screen->Invalidate( wnd->pt->mClip);

			wnd->pt->mClip = clip;
			wnd->pt->BeginDraw();
		}
		else if ( bmp ) {
			X->screen->BeginDraw( X->presman, bmp);  // no invalidation required for bitmaps
		}

		return 1;
	}

	void endDraw()
	{
		if ( wnd /* && !wnd->isDrawing */ ) {
			wnd->pt->EndDraw();
			wnd->pt->mClip = wnd->tgtClip;
		}
		else if ( bmp ) {
			X->screen->EndDraw( bmp);
		}
	}


	// coordinate translation
	int transX ()
	{
		return wnd ? (x0 + wnd->pt->mReal.wLeft) : x0;
	}

	int transY ()
	{
		return wnd ? (y0 + wnd->pt->mReal.wTop) : y0;
	}

	void translate ( int& x, int&y )
	{
		 x += transX();
		 y += transY(); 
	}


	// These are our drawing primitives. Here, we adapt to different target types
	// (PegThing, PegBitmap), and care for temporary flags. We don't
	// modify any coordinates

    void moveRect( PegRect& get, PegPoint& put);



	void drawText ( PegPoint& p, char* s, int len )
	{
		PegColor pc( fg);

		// we have to apply the baseline correction by ourselves
		p.y -= fnt->uAscent;

		X->screen->DrawText( PT, p, s, pc, fnt, len);
	}

	void drawLine ( int x1, int y1, int x2, int y2 )
	{
		PegColor pc( fg, fg, flags);
		X->screen->Line( PT, x1, y1, x2, y2, pc);
	}

	void drawPolygon ( PegPoint* poly, int nPoints )
	{
		PegColor pc( fg, fg, flags);
		X->screen->Polygon( PT, poly, nPoints, pc);
	}

	void fillPolygon ( PegPoint* poly, int nPoints )
	{
		PegColor pc( fg, fg, flags | CF_FILL);
		X->screen->Polygon( PT, poly, nPoints, pc);
	}

	void drawRect ( PegRect& r )
	{
		PegColor pc( fg, fg, flags);
		X->screen->Rectangle( PT, r, pc);
	}

	void fillRect ( PegRect& r )
	{
		PegColor pc( fg, fg, flags | CF_FILL);
		X->screen->Rectangle( PT, r, pc);
	}

	void clearRect ( PegRect& r )
	{
		PegColor pc( bg, bg, CF_FILL);
		X->screen->Rectangle( PT, r, pc);
	}

	void drawOval ( PegRect& r )
	{
		PegColor pc( fg, fg, flags);
		X->screen->Ellipse( PT, r, pc, 1);
	}

	void fillOval ( PegRect& r )
	{
		PegColor pc( fg, fg, flags | CF_FILL);
		X->screen->Ellipse( PT, r, pc, 1);
	}

	void drawImage ( PegBitmap* bmp, int x, int y )
	{
		PegPoint p;
		p.x = x;
		p.y = y;

		X->screen->Bitmap( PT, p, bmp);
	}
};


void Graphics::moveRect ( PegRect& get, PegPoint& put )
	{
		if ( wnd ) {
			PegRect r, s;

			s.Set( 0, 0, X->screen->GetXRes(), X->screen->GetYRes());
			r.Set( get.wLeft, get.wTop, get.wRight, get.wBottom);
			r.MoveTo( put.x, put.y);

			if ( (get.wLeft < 0) || (get.wRight > s.wRight) || (get.wTop < 0) || (get.wBottom > s.wBottom)
				   || (r.wLeft < 0) || (r.wRight > s.wRight) || (r.wTop < 0) || (r.wBottom > s.wBottom) ){
				r &= s;

				X->presman->Invalidate( r);
				wnd->pt->Draw();
			}
			else {
				X->screen->ViewportMove( wnd->pt, get, put);
			}
		}
		else {
			X->screen->RectMove( PT, get, put);
		}
	}


/*******************************************************************************
 *
 */

extern "C" {

void* _cdecl
Java_java_awt_Toolkit_graInitGraphics ( JNIEnv* env, jclass clazz,
										Graphics* gr, void* tgt, jint tgtType,
										jint xOff, jint yOff,
										jint xClip, jint yClip, jint wClip, jint hClip,
										PegFont* fnt, jint fg, jint bg, jboolean blank )
{
	Graphics      *go;
	MappedWindow  *wnd = 0;
	PegBitmap     *bmp = 0;
	
	DBG( awt_gra, ("initGraphics: %x (%x, %d) %d,%d  %d,%d - %d,%d  %x %x %x\n",
		gr, tgt,tgtType, xOff,yOff, xClip,yClip,wClip,hClip, fnt,fg,bg));
	
	if ( tgt ) {
		switch ( tgtType ) {
		case 0:                                            /* Window */
			wnd = (MappedWindow*)tgt;
			break;
			
		case 1:                                            /* Image */
			bmp = (PegBitmap*)((Image*)tgt)->bmp;
			break;
			
		case 2:                                            /* Graphics */
			go = (Graphics*)tgt;
			if ( go->wnd )
				wnd = go->wnd;
			else if ( go->bmp )
				bmp = go->bmp;
			break;
			
		default:
			DBG( awt_gra, ("illegal Graphics target: %d\n", tgtType));
		}
	}
	else {
		if ( gr ){
			wnd = gr->wnd;
			bmp = gr->bmp;
		}
		else {
			DBG( awt_gra, ("attempt to set an uninitialized graphics"));
		}
	}
	
	if ( !gr ) {
		gr = new Graphics();		
		DBG( awt_gra, (" ->gr: %x (%x)\n", gr, gr->gc));
	}
	
	gr->wnd = wnd;
	gr->bmp = bmp;
	
	gr->bg = bg;
	gr->fg = fg;
	gr->fnt = fnt;
	gr->x0   = xOff;
	gr->y0   = yOff;
	
	gr->translate( xClip, yClip);
	if ( wClip < 0 ) wClip = 0;
	if ( hClip < 0 ) hClip = 0;
	gr->clip.Set( xClip, yClip, xClip + wClip -1, yClip + hClip -1);
	if ( gr->wnd )
		gr->clip &= gr->wnd->pt->mClient;
	
	if ( blank ) {
		gr->startDraw();
		gr->clearRect( gr->clip);
		gr->endDraw();
	}
	
	gr->flags = CF_NONE;
	
	return gr;
}


void _cdecl
Java_java_awt_Toolkit_graFreeGraphics ( JNIEnv* env, jclass clazz, Graphics* gr )
{
	DBG( awt_gra, ("freeGraphics: %x\n", gr));
	
	delete gr;
}


void _cdecl
Java_java_awt_Toolkit_graCopyArea ( JNIEnv* env, jclass clazz, Graphics* gr, 
								   jint x, jint y, jint width, jint height,
								   jint xDelta, jint yDelta )
{
	PegRect  get;
	PegPoint put;
	
	DBG( awt_gra, ("copyArea: %x, %d,%d-%d,%d  %d,%d\n", gr, x,y,width,height, xDelta,yDelta));
	
	gr->translate( x, y);
	get.Set( x, y, x+width, y+height);
	
	x += xDelta;
	y += yDelta;
	put.x = x;
	put.y = y;
	
	if ( gr->startDraw() ){
		gr->moveRect( get, put);
		gr->endDraw();
	}
}


void _cdecl
Java_java_awt_Toolkit_graClearRect ( JNIEnv* env, jclass clazz, Graphics* gr,
									jint x, jint y, jint width, jint height )
{
	PegRect r;
	
	DBG( awt_gra, ("clearRect: %x, %d,%d-%d,%d\n", gr, x,y,width,height));
	
	gr->translate( x, y);
	r.Set( x, y, x+width-1, y+height-1);
	
	if ( gr->startDraw() ){
		gr->clearRect( r);
		gr->endDraw();
	}
}


void _cdecl
Java_java_awt_Toolkit_graDrawBytes ( JNIEnv* env, jclass clazz, Graphics* gr,
									jarray jBytes, jint offset, jint len, jint x, jint y )
{
	jboolean isCopy;
	int      n;
	jbyte    *jb, *jbo;
	PegPoint p;
	
	if ( !jBytes ) return;
	
	n = env->GetArrayLength( jBytes);
	jb = env->GetByteArrayElements( jBytes, &isCopy);
	jbo = jb + offset;
	
	DBG( awt_gra, ("drawBytes: %x, %x,%d,%d  \"%s\"  %d,%d\n", gr,
		jBytes,offset,len, jb, x,y));
	
	if ( offset+len > n )
		len = n - offset;
	
	gr->translate( x, y);
	p.x = x;
	p.y = y;
	
	if ( gr->startDraw() ){
		gr->drawText( p, jbo, len);
		gr->endDraw();
	}
	
	env->ReleaseByteArrayElements( jBytes, jb, JNI_ABORT);
}


void _cdecl
Java_java_awt_Toolkit_graDrawChars ( JNIEnv* env, jclass clazz, Graphics* gr,
									jarray jChars, jint offset, jint len, jint x, jint y )
{
	PegPoint p;
	jboolean isCopy;
	int      n;
	jchar    *jc, *jco;
	char     *buf;
	
	if ( !jChars ) return;
	
	n   = env->GetArrayLength( jChars);
	jc = env->GetCharArrayElements( jChars, &isCopy);
	jco = jc + offset;
	
	if ( offset+len > n )
		len = n - offset;
	
	buf = jchar2CString( X, jco, len);
	
	DBG( awt_gra, ("drawChars: %x, %x,%d,%d  \"%s\"  %d,%d\n", gr, jChars,offset,len,
		buf, x,y));
	
	gr->translate( x, y);
	p.x = x;
	p.y = y;
	
	if ( gr->startDraw() ){
		gr->drawText( p, buf, -1);
		gr->endDraw();
	}
	
	env->ReleaseCharArrayElements( jChars, jc, JNI_ABORT);
}


void _cdecl
Java_java_awt_Toolkit_graDrawString ( JNIEnv* env, jclass clazz,
									 Graphics* gr, jstring str, jint x, jint y )
{
	PegPoint     p;
	jboolean     isCopy;
	int          len;
	const jchar  *jc;
	char         *buf;
	
	if ( !str ) return;
	
	len = env->GetStringLength( str);
	jc = env->GetStringChars( str, &isCopy);
	
	buf = jchar2CString( X, (jchar*)jc, len);
	
	DBG( awt_gra, ("drawString: %x  \"%s\"  %d,%d\n", gr, buf, x,y));
	
	gr->translate( x, y);
	p.x = x;
	p.y = y;
	
	if ( gr->startDraw() ){
		gr->drawText( p, buf, -1);
		gr->endDraw();
	}
	
	env->ReleaseStringChars( str, jc);
}


void _cdecl
Java_java_awt_Toolkit_graDrawLine ( JNIEnv* env, jclass clazz, Graphics* gr,
								   jint x1, jint y1, jint x2, jint y2 )
{
	DBG( awt_gra, ("drawLine: %x, %d,%d - %d,%d\n", gr, x1,y1, x2,y2));
	
	gr->translate( x1, y1);
	gr->translate( x2, y2);
	
	if ( gr->startDraw() ){
		gr->drawLine( x1, y1, x2, y2);
		gr->endDraw();
	}
}


void _cdecl
Java_java_awt_Toolkit_graDrawArc ( JNIEnv* env, jclass clazz, Graphics* gr,
								   jint x, jint y, jint width, jint height,
								   jint startAngle, jint arcAngle )
{
	DBG( awt_gra, ("drawArc: %x, %d,%d-%d,%d  %d,%d\n", gr,
				 x,y,width,height, startAngle,arcAngle));

	// we are still missing the EllispeArc in Peg (which just has a circular arc)
}


void _cdecl
Java_java_awt_Toolkit_graFillArc ( JNIEnv* env, jclass clazz, Graphics* gr,
								   jint x, jint y, jint width, jint height,
								   jint startAngle, jint arcAngle )
{
	DBG( awt_gra, ("fillArc: %x, %d,%d-%d,%d  %d,%d\n", gr,
				 x,y,width,height, startAngle, arcAngle));

	// we are still missing the EllispeArc in Peg (which just has a circular arc)
}


void _cdecl
Java_java_awt_Toolkit_graDrawOval ( JNIEnv* env, jclass clazz, Graphics* gr,
									jint x, jint y, jint width, jint height )
{
	PegRect r;
	DBG( awt_gra, ("drawOval: %x, %d,%d - %d,%d\n", gr, x,y,width,height));

	if ( (width >= 0) && (height >= 0) ){
		gr->translate( x, y);
		r.Set( x, y, x+width, y+height);

		if ( gr->startDraw() ){
			gr->drawOval( r);
			gr->endDraw();
		}
	}
}


void _cdecl
Java_java_awt_Toolkit_graFillOval ( JNIEnv* env, jclass clazz, Graphics* gr,
									jint x, jint y, jint width, jint height )
{
	PegRect r;
	DBG( awt_gra, ("fillOval: %x, %d,%d - %d,%d\n", gr, x,y,width,height));

	if ( (width >= 0) && (height >= 0) ){
		gr->translate( x, y);
		r.Set( x, y, x+width, y+height);

		if ( gr->startDraw() ){
			gr->fillOval( r);
			gr->endDraw();
		}
	}

}


int
jarray2Points ( JNIEnv* env, Toolkit* X, PegPoint** pp, int x0, int y0,
			   jarray xPoints, jarray yPoints, int nPoints )
{
	register int i;
	int      n;
	jboolean isCopy;
	jint     *jx = env->GetIntArrayElements( xPoints, &isCopy);
	jint     *jy = env->GetIntArrayElements( yPoints, &isCopy);
	
	if ( (n = env->GetArrayLength( xPoints)) < nPoints ) nPoints = n;
	if ( (n = env->GetArrayLength( yPoints)) < nPoints ) nPoints = n;
	
	*pp = (PegPoint*) getBuffer( X, (nPoints+1)*sizeof(PegPoint));
	
	for ( i=0; i<nPoints; i++){
		(*pp)[i].x = jx[i] + x0;
		(*pp)[i].y = jy[i] + y0;
	}
	
	env->ReleaseIntArrayElements( xPoints, jx, JNI_ABORT);
	env->ReleaseIntArrayElements( yPoints, jy, JNI_ABORT);
	
	return nPoints;
}


void _cdecl
Java_java_awt_Toolkit_graDrawPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
									  jarray xPoints, jarray yPoints, jint nPoints )
{
	int n;
	PegPoint   *p;
	
	DBG( awt_gra, ("drawPolygon: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));
	
	if ( !xPoints || !yPoints ) return;
	
	nPoints = jarray2Points( env, X, &p,
		gr->transX(), gr->transY(),
		xPoints, yPoints, nPoints);
	
	n = nPoints-1;
	if ( (p[0].x != p[n].x) || (p[0].y != p[n].y) ) {
		p[nPoints].x = p[0].x;
		p[nPoints].y = p[0].y;
		nPoints++;
	}
	
	if ( gr->startDraw() ){
		gr->drawPolygon( p, nPoints);
		gr->endDraw();
	}
}


void _cdecl
Java_java_awt_Toolkit_graDrawPolyline ( JNIEnv* env, jclass clazz, Graphics* gr,
									   jarray xPoints, jarray yPoints, jint nPoints )
{
	PegPoint   *p;
	
	DBG( awt_gra, ("drawPolyline: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));
	
	if ( !xPoints || !yPoints ) return;
	
	nPoints = jarray2Points( env, X, &p,
		gr->transX(), gr->transY(),
		xPoints, yPoints, nPoints);
	
	if ( gr->startDraw() ){
		gr->drawPolygon( p, nPoints);
		gr->endDraw();
	}
}


void _cdecl
Java_java_awt_Toolkit_graFillPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
									  jarray xPoints, jarray yPoints, jint nPoints )
{
	PegPoint   *p;
	
	DBG( awt_gra, ("fillPolygon: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));
	
	if ( !xPoints || !yPoints ) return;
	
	nPoints = jarray2Points( env, X, &p,
		gr->transX(), gr->transY(),
		xPoints, yPoints, nPoints);
	
	if ( gr->startDraw() ){
		gr->fillPolygon( p, nPoints);
		gr->endDraw();
	}
}


void _cdecl
Java_java_awt_Toolkit_graDrawRect ( JNIEnv* env, jclass clazz, Graphics* gr,
								   jint x, jint y, jint width, jint height )
{
	PegRect r;
	
	DBG( awt_gra, ("drawRect: %x, %d,%d - %d,%d\n", gr, x,y,width,height));
	
	if ( (width >= 0) && (height >= 0) ){
		gr->translate( x, y);
		r.Set( x, y, x+width, y+height);
		
		if ( gr->startDraw() ){
			gr->drawRect( r);
			gr->endDraw();
		}
	}
}


void _cdecl
Java_java_awt_Toolkit_graFillRect ( JNIEnv* env, jclass clazz, Graphics* gr,
								   jint x, jint y, jint width, jint height )
{
	PegRect r;
	
	DBG( awt_gra, ("fillRect: %x, %d,%d - %d,%d\n", gr, x,y,width,height));
	
	if ( (width >= 0) && (height >= 0) ) {
		gr->translate( x, y);
		r.Set( x, y, x+width-1, y+height-1);
		
		if ( gr->startDraw() ){
			gr->fillRect( r);
			gr->endDraw();
		}
	}
}


void _cdecl
Java_java_awt_Toolkit_graDrawRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
										jint x, jint y, jint width, jint height, 
										jint wArc, jint hArc )
{
	int x1, x2, y1, y2, a, b;
	
	DBG( awt_gra, ("drawRoundRect: %x, %d,%d - %d,%d  %d,%d\n", gr,
		x,y,width,height, wArc, hArc));
	
	if ( gr->startDraw() ){
		
		gr->translate( x, y);
		a = wArc / 2; b = hArc / 2;
		
		x1 = x + a;   x2 = x + width - a;  y2 = y + height;
		gr->drawLine( x1, y, x2, y);
		gr->drawLine( x1, y2, x2, y2);
		
		y1 = y + b;   y2 = y + height - b; x2 = x + width;
		gr->drawLine( x, y1, x, y2);
		gr->drawLine( x2, y1, x2, y2);
		
		//XDrawArc( X->dsp, gr->drw, gr->gc, x, y, wArc, hArc, 90*64, 90*64);
		
		x2 = x + width - wArc;
		//XDrawArc( X->dsp, gr->drw, gr->gc, x2, y, wArc, hArc, 0, 90*64);
		
		y2 = y + height - hArc;
		//XDrawArc( X->dsp, gr->drw, gr->gc, x2, y2, wArc, hArc, 0, -90*64);
		
		//XDrawArc( X->dsp, gr->drw, gr->gc, x, y2, wArc, hArc, 180*64, 90*64);
		
		gr->endDraw();
	}
}

void _cdecl
Java_java_awt_Toolkit_graFillRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
										jint x, jint y, jint width, jint height, 
										jint wArc, jint hArc )
{
	PegRect  r;
	int      x1, x2, y1, y2, a, b;
	
	DBG( awt_gra, ("fillRoundRect: %x, %d,%d - %d,%d  %d,%d\n", gr,
		x,y,width,height, wArc, hArc));
	
	if ( gr->startDraw() ){
		
		gr->translate( x, y);
		a = wArc / 2; b = hArc / 2;
		
		y1 = y + b;  y2 = y + height - b;
		if ( y2 > y1 ) {
			r.Set( x, y1, x+width, y2);
			gr->fillRect( r);
		}
		
		x1 = x + a;  x2 = x + width - a;
		if ( x2 > x1 ) {
			r.Set( x1, y, x2, y+b);
			gr->fillRect( r);
			
			r.Set( x1, y2, x2, y2+b);
			gr->fillRect( r);
		}
		
		//XFillArc( X->dsp, gr->drw, gr->gc, x, y, wArc, hArc, 90*64, 90*64);
		
		x2 = x + width - wArc;
		//XFillArc( X->dsp, gr->drw, gr->gc, x2, y, wArc, hArc, 0, 90*64);
		
		y2 = y + height - hArc;
		//XFillArc( X->dsp, gr->drw, gr->gc, x2, y2, wArc, hArc, 0, -90*64);
		
		//XFillArc( X->dsp, gr->drw, gr->gc, x, y2, wArc, hArc, 180*64, 90*64);
		
		gr->endDraw();
	}
}


void _cdecl
Java_java_awt_Toolkit_graDraw3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
									 jint x, jint y, jint width, jint height,
									 jboolean raised, jint rgb )
{
	int      bright, dark;
	int      xw, yh;
	int      fg = gr->fg;
	
	DBG( awt_gra, ("draw3DRect: %x, %d,%d - %d,%d  %d %x\n", gr,
		x,y,width,height, raised, rgb));
	
	if ( gr->startDraw() ){
		
		/* we should use the same mechanism like ordinary Graphics brighter()/darker() here */
		dark   = (int) (Java_java_awt_Toolkit_clrDark( env, clazz, rgb) >> 32);
		bright = (int) (Java_java_awt_Toolkit_clrBright( env, clazz, rgb) >> 32);
		
		gr->translate( x, y);
		xw = x + width;
		yh = y + height;
		
		gr->fg = (raised ? bright : dark);
		gr->drawLine( x, y, xw-1, y);
		gr->drawLine( x, y, x, yh);
		
		gr->fg = (raised ? dark : bright);
		gr->drawLine( x+1, yh, xw, yh);
		gr->drawLine( xw, y, xw, yh);
		
		gr->endDraw();
	}
	
	gr->fg = fg;
}


void _cdecl
Java_java_awt_Toolkit_graFill3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
									 jint x, jint y, jint width, jint height,
									 jboolean raised, jint rgb )
{
	PegRect r;
	int     u = x, v = y;
	
	DBG( awt_gra, ("fill3DRect: %x, %d,%d - %d,%d  %d %x\n", gr,
		x,y,width,height, raised, rgb));
	
	gr->translate( u, v);
	r.Set( u+1, v+1, u+width-2, v+height-2); 
	
	if ( gr->startDraw() ){
		gr->fillRect( r);
		gr->endDraw();
	}
	
	Java_java_awt_Toolkit_graDraw3DRect( env, clazz, gr, x, y, width-1, height-1, raised, rgb);
}


void _cdecl
Java_java_awt_Toolkit_graAddClip ( JNIEnv* env, jclass clazz, Graphics* gr,
																	jint xClip, jint yClip, jint wClip, jint hClip )
{
	/* not implemented yet */
}


void _cdecl
Java_java_awt_Toolkit_graSetClip ( JNIEnv* env, jclass clazz, Graphics* gr,
								  jint xClip, jint yClip, jint wClip, jint hClip )
{
	DBG( awt_gra, ("setClip: %x, %d,%d - %d,%d\n", gr, xClip, yClip, wClip, hClip));
	
	if ( wClip < 0 ) wClip = 0;
	if ( hClip < 0 ) hClip = 0;
	
	gr->translate( xClip, yClip);
	gr->clip.Set( xClip, yClip, xClip + wClip -1, yClip + hClip -1);
	if ( gr->wnd ){
		gr->clip &= gr->wnd->pt->mClient;
	}
}

void _cdecl
Java_java_awt_Toolkit_graSetColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
{
	DBG( awt_gra, ("setColor: %x, %x\n", gr, clr));
	gr->fg = clr;
}

void _cdecl
Java_java_awt_Toolkit_graSetBackColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
{
	DBG( awt_gra, ("setBackColor: %x, %x\n", gr, clr));
	gr->bg = clr;
}

void _cdecl
Java_java_awt_Toolkit_graSetFont ( JNIEnv* env, jclass clazz, Graphics* gr, PegFont* fnt )
{
	DBG( awt_gra, ("setFont: %x, %x\n", gr, fnt));
	
	gr->fnt = fnt;
}

void _cdecl
Java_java_awt_Toolkit_graSetOffset ( JNIEnv* env, jclass clazz, Graphics* gr, jint xOff, jint yOff )
{
	DBG( awt_gra, ("setOffset: %x, %d,%d\n", gr, xOff, yOff));
	
	gr->x0 = xOff;
	gr->y0 = yOff;
}

void _cdecl
Java_java_awt_Toolkit_graSetPaintMode ( JNIEnv* env, jclass clazz, Graphics* gr )
{
	DBG( awt_gra, ("setPaintMode: %x\n", gr));
	
	gr->flags = CF_NONE;
}

void _cdecl
Java_java_awt_Toolkit_graSetXORMode ( JNIEnv* env, jclass clazz, Graphics* gr, jint xorClr )
{
	DBG( awt_gra, ("setXORMode: %x, %x\n", gr, xorClr));
	
	gr->flags = CF_XOR;
}

void _cdecl
Java_java_awt_Toolkit_graSetVisible ( JNIEnv* env, jclass clazz, Graphics* gr, jint isVisible )
{
	DBG( awt_gra, ("setVisble: %x  %d\n", gr, isVisible));
	/* we don't need that here, since we check visibility in startDraw */
}


/************************************************************************************
 * image rendering
 */


void _cdecl
Java_java_awt_Toolkit_graDrawImage ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
									jint srcX, jint srcY,
									jint dstX, jint dstY,
									jint width, jint height, jint bgval )
{
	int     isPart = 0;
	PegRect clip;
	
	DBG( awt_gra, ("drawImage: %x %x (%x,%x,%x %d,%d) %d,%d, %d,%d, %d,%d, %x\n",
		gr, img,  img->xImg,img->xMask,img->alpha,  img->width,img->height,
		srcX,srcY,  dstX,dstY,	 width,height, bgval));
	
	gr->translate( dstX, dstY);
	
	dstX += img->left;
	dstY += img->top;
	
	if ( srcX || srcY || width < img->width || height < img->height ) {
		clip = gr->clip;
		gr->clip.Set( dstX, dstY, width, height);
		dstX -= srcX;
		dstY -= srcY;
		isPart = 1;
	}
	
	if ( gr->startDraw() ){
		if ( bgval != -1 ) {
			int     fg = gr->fg;
			PegRect r;
			
			r.Set( dstX, dstY, dstX + width, dstY + height);
			
			gr->fg = bgval;
			gr->fillRect( r);
			gr->fg = fg;
		}
		
		gr->drawImage( img->bmp, dstX, dstY);
		gr->endDraw();
	}
	
	if ( isPart ) {
		gr->clip = clip;
	}
}


void _cdecl
Java_java_awt_Toolkit_graDrawImageScaled ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
										   jint dx0, jint dy0, jint dx1, jint dy1,
										   jint sx0, jint sy0, jint sx1, jint sy1, jint bgval )
{
	Image sImg;
	int   dw, dh;

	DBG( awt_gra, ("drawImageScaled: %x %x (%x), %d,%d,%d,%d, %d,%d,%d,%d, %x\n",
				 gr, img, img->xImg, dx0, dy0, dx1, dy1, sx0, sy0, sx1, sy1, bgval));

	dw = dx1 - dx0;
	dh = dy1 - dy0;

	sImg.top    = 0;
	sImg.left   = 0;
	sImg.width  = dw;
	sImg.height = dh;
	sImg.next   = 0;
	sImg.alpha  = 0;
	sImg.bmp    = X->screen->ResizeImage( img->bmp, dw, dh);

	if ( sImg.bmp && sImg.bmp->pStart) {
		Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, &sImg,
		                                  sx0, sy0, dx0, dy0, dw, dh, bgval);
		delete[] sImg.bmp->pStart;
		delete sImg.bmp;
		sImg.bmp = 0;
	}
}

}