/**
 * img.cpp - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.hpp"
#include "jsyscall.h"


Image*
createImage ( int width, int height )
{
	Image     *img = (Image*)AWT_CALLOC( 1, sizeof( Image));
	img->width = width;
	img->height = height;
	
	// everything else is zeroed, for now
	return img;
}

Image*
createOffScreenImage ( jint width, jint height )
{
	Image  *img = createImage( width, height);
	img->bmp = X->screen->CreateBitmap( width, height);
	
	if ( !img->bmp || !img->bmp->pStart ){
		if ( img->bmp )
			AWT_FREE( img->bmp);
		AWT_FREE( img);
		
		return 0;
	}
	
	return img;
}


void
createAlphaImage ( Toolkit* X, Image *img )
{
	int nBytes = img->width * img->height;
	
	img->alpha = (AlphaImage*)AWT_MALLOC( sizeof( AlphaImage));
	
	img->alpha->width  = img->width;
	img->alpha->height = img->height;
	img->alpha->buf = (unsigned char*)AWT_MALLOC( nBytes);
	memset( img->alpha->buf, 0xff, nBytes);
}



/************************************************************************************
 * exported native methods
 */

extern "C" {

void* __cdecl
Java_java_awt_Toolkit_imgCreateImage ( JNIEnv* env, jclass clazz, jint width, jint height )
{
	return createOffScreenImage( width, height);
}
	
	
void* _cdecl
Java_java_awt_Toolkit_imgCreateScreenImage ( JNIEnv* env, jclass clazz, jint width, jint height )
{
	return createOffScreenImage( width, height);
}


/* generic (ImageProducer) based image construction */

#define TRANSP_PIXEL  255  // PEG way to say 'transparent'

void _cdecl
Java_java_awt_Toolkit_imgSetIdxPels ( JNIEnv* env, jclass clazz, Image * img,
									 jint x, jint y, jint w, jint h,
									 jarray clrMap, jarray idxPels, jint trans,
									 jint off, jint scan)
{
	register int    row, col;
	COLORVAL        pix;
	jint            rgb;
	jboolean        isCopy;
	jint            *clr = env->GetIntArrayElements( clrMap, &isCopy);
	jbyte           *pel = env->GetByteArrayElements( idxPels, &isCopy);
	unsigned char   *idx = (unsigned char*)(pel + off);
	int             maxCol = x + w;
	int             maxRow = y + h;
	unsigned char   curPel;
	
	X->screen->BeginDraw( X->presman, img->bmp);
	
	for ( row = y; row < maxRow; row++) {
		for ( col = x; col < maxCol; col++) {
			curPel = idx[col + row * scan];
			
			if ( (trans >= 0) && (curPel == trans) ){
				pix = TRANSP_PIXEL;
			}
			else {
				rgb = clr[curPel];
				pix = pixelValue( X, rgb);
			}
			X->screen->PutPixel( X->presman, col, row, pix);
		}
		x = 0;
	}
	
	X->screen->EndDraw( img->bmp);
	
	env->ReleaseIntArrayElements( clrMap, clr, JNI_ABORT);
	env->ReleaseByteArrayElements( idxPels, pel, JNI_ABORT);
}


void _cdecl
Java_java_awt_Toolkit_imgSetRGBPels ( JNIEnv* env, jclass clazz, Image * img,
									 jint x, jint y, jint w, jint h,
									 jarray rgbPels, jboolean hasAlpha, jint off, jint scan)
{
	register int    row, col;
	COLORVAL        pix;
	jboolean        isCopy;
	jint            *rgbs = env->GetIntArrayElements( rgbPels, &isCopy);
	jint            *rgb = rgbs + off;
	int             maxCol = x + w;
	int             maxRow = y + h;
	jint		        val;
	
	X->screen->BeginDraw( X->presman, img->bmp);
	
	for ( row = y; row < maxRow; row++) {
		for ( col = x; col < maxCol; col++) {
			val = rgb[col + row * scan];
			if ( (val & 0xff000000) == 0xff000000 ) {
				pix = pixelValue( X, val);
				X->screen->PutPixel( X->presman, col, row, pix);
			}
			else {
				/*
				 * No way to tell for now if this will be a reduced (on/off) or a full
				 * alpha channel. We have to be prepared for the "worst", and reduce later
				 */
				if ( hasAlpha && !img->alpha ){
					createAlphaImage( X, img);
				}
				PutAlpha( img->alpha, col, row, (val >> 24));
				
				X->screen->PutPixel( X->presman, col, row, pix);
			}
		}
		x = 0;
	}
	
	X->screen->EndDraw( img->bmp);
	
	env->ReleaseIntArrayElements( rgbPels, rgbs, JNI_ABORT);
}

void _cdecl
Java_java_awt_Toolkit_imgComplete( JNIEnv* env, jclass clazz, Image * img, jint status )
{
	/* to stay consistent with the image production (conversion) mechanism,
	 * alpha channels need to be supported by PEG (driver specific) */
}

void _cdecl
Java_java_awt_Toolkit_imgFreeImage( JNIEnv* env, jclass clazz, Image * img)
{
	Image *next, *first = img;
	
	/* we have to be aware of image rings (GIF movies), iterate */
	do {
		next = img->next;
		
		if ( img->bmp ){
			if ( img->bmp->pStart )
				delete[] img->bmp->pStart;
			delete img->bmp;
			img->bmp = 0;
		}
		
		if ( img->alpha ){
			if ( img->alpha->buf )
				AWT_FREE( img->alpha->buf);
			AWT_FREE( img->alpha);
			img->alpha = 0;
		}
		
		AWT_FREE( img);
		img = next;
	} while ( (img != 0) && (img != first) );
}


void* _cdecl
Java_java_awt_Toolkit_imgCreateScaledImage ( JNIEnv* env, jclass clazz,
											Image* img, int width, int height )
{
	Image *scaledImg = createImage( width, height);
	
	if ( img->bmp ) {
		scaledImg->bmp = X->screen->ResizeImage( img->bmp, width, height);
	}
	
	return scaledImg;
}


void _cdecl
Java_java_awt_Toolkit_imgProduceImage ( JNIEnv* env, jclass clazz, jobject producer, Image* img )
{
	int       i, j, r, g, b;
	jint      pix;
	jboolean  isCopy;
	jclass    prodClazz  = env->GetObjectClass( producer);
	jclass    modelClazz = env->FindClass( "java.awt.image.ColorModel");
	
	jmethodID modelCtor = env->GetStaticMethodID( modelClazz, "getRGBdefault", "()Ljava/awt/image/ColorModel;");
	jmethodID setDim    = env->GetMethodID( prodClazz, "setDimensions", "(II)V");
	jmethodID setCM     = env->GetMethodID( prodClazz, "setColorModel", "(Ljava/awt/image/ColorModel;)V");
	jmethodID setHints  = env->GetMethodID( prodClazz, "setHints", "(I)V");
	jmethodID setPix    = env->GetMethodID( prodClazz, "setPixels", "(IIIILjava/awt/image/ColorModel;[III)V");
	jmethodID imgCompl  = env->GetMethodID( prodClazz, "imageComplete", "(I)V");
	
	jobject   model     = env->CallStaticObjectMethod( modelClazz, modelCtor);
	
	/* for JDK compat, the pixel buffer has to be large enough to hold the *complete* image */
	jarray    pelArray  = env->NewIntArray( img->width * img->height);
	jint*     pels = env->GetIntArrayElements( pelArray, &isCopy);
	
	env->CallVoidMethod( producer, setDim, img->width, img->height);
	env->CallVoidMethod( producer, setCM, model);
	env->CallVoidMethod( producer, setHints, 6); /* TOPDOWNLEFTRIGHT | COMPLETESCANLINES */
	
	// set the GetPixel target
	X->screen->BeginDraw( X->presman, img->bmp);

	for ( j=0; j<img->height; j++ ) {
		for ( i=0; i<img->width; i++ ) {
			pix = X->screen->GetPixel( X->presman, i, j);
			if ( pix == TRANSP_PIXEL ){
				rgbValues( X, pix, &r, &g, &b);
				pels[j*img->width+i] = (0xff000000 | (r << 16) | (g << 8) | b);
			}
			else {
				pels[j*img->width+i] = 0;
			}
		}
	}

	X->screen->EndDraw();

	if ( isCopy ) {
		env->ReleaseIntArrayElements( pelArray, pels, JNI_COMMIT);
	}
	
	env->CallVoidMethod( producer, setPix,
		0, 0, img->width, img->height, model, pelArray, 0, img->width);
	env->CallVoidMethod( producer, imgCompl, 3); /* 3 = STATICIMAGEDONE */
}



/************************************************************************************
 * native format production dispatcher routines (GIF, jpeg, png, ..)
 */

#define SIG_LENGTH 4

#define SIG_GIF    1
#define SIG_JPEG   2
#define SIG_PNG    3

Image *unknownImage = 0;  /* fill in some default image here */

int imageFormat ( unsigned char* sig ) {
  if ( (sig[0] == 'G') && (sig[1] == 'I') && (sig[2] == 'F') )
		return SIG_GIF;
	
  if ( (sig[0] == 0xff) && (sig[1] == 0xd8) && (sig[2] == 0xff) && (sig[3] == 0xe0) )
		return SIG_JPEG;
	
  if ( (sig[0] == 0x89) && (sig[1] == 'P') && (sig[2] == 'N') && (sig[3] == 'G') )
		return SIG_PNG;
	
  return 0;
}


Image*
getImages ( PegImageConvert* conv )
{
	int         i, n;
	Image       *first = (Image*)0, *img = (Image*)0;
	PegBitmap   **bmps;
	WORD        state = conv->GetState();
	int         type;
	GIF_IMAGE_INFO  *gifInfo = (GIF_IMAGE_INFO*) 0;

	if ( (state & (PIC_COMPLETE | PIC_ERROR)) == PIC_COMPLETE ) {
		// Ok, we got it - turn it into an Image
		bmps = conv->GetBitmapList();
		n = conv->GetBitmapCount();
		type = conv->ImageType();

		if ( type == PIC_TYPE_GIF ){
			// Hrmm, a downcast without rtti
			gifInfo = ((PegGifConvert*)conv)->GetGifInfo();
		}

		for ( i=0; i<n; i++ ) {
			if ( bmps[i] && bmps[i]->pStart ) {
				if ( img ) {
					img->next = createImage( bmps[i]->wWidth, bmps[i]->wHeight);
					img = img->next;
				}
				else {
					first = img = createImage(  bmps[i]->wWidth, bmps[i]->wHeight);
				}
				img->bmp = bmps[i];
				img->frame = i;

				if ( gifInfo ){
					img->left = gifInfo[i].xOffset;
					img->top = gifInfo[i].yOffset;
					img->latency = gifInfo[i].tDelay * 10; // delay is in sec/100
				}
			}
		}

		if ( img != first ){
			img->next = first;
		}
	}

	return first;
}

void
runConv ( PegImageConvert* conv, BOOL (*inputFunction) (WORD,WORD,PegImageConvert*))
{
	DWORD         npal;
	UCHAR         *pal;

	pal = X->screen->GetPalette( &npal);
	conv->SetSystemPalette( pal, (WORD)npal);

	conv->SetMode( PIC_INLINE_REMAP);  // because of icons
	conv->SetIdleCallback( inputFunction);
	conv->ReadImage();                 // kick it off
}


BOOL
readFileImageData ( WORD id, WORD convState, PegImageConvert* conv )
{
	UCHAR buf[1024];
	int   n, nMax;

	if ( convState & PIC_IDLE ) {             // more data, quick..
		nMax = conv->InputFreeSpace();
		if ( nMax > sizeof(buf) )
			nMax = sizeof(buf);
		n = read( id, buf, nMax);               // don't read more than we can handle at a time
		if ( n ) {
			conv->SendData( buf, n);
			return TRUE;
		}
		else {
			return FALSE;
		}
	}
	else if ( convState & PIC_COMPLETE ) {    // we are done with it
		return FALSE;
	}
	else {                                    // rest is supposed to be an error
		return FALSE;
	}
}

void* _cdecl
Java_java_awt_Toolkit_imgCreateFromFile ( JNIEnv* env, jclass clazz, jstring fileName )
{
	int              infile;
	Image            *img = (Image*)0;
	char             *fn = java2CString( env, X, fileName);
	unsigned char    sig[SIG_LENGTH];
	PegImageConvert  *conv = NULL;

	if ( (infile = AWT_OPEN( fn)) >= 0 ) {
		if ( AWT_READ( infile, sig, sizeof(sig)) == sizeof(sig) ) {
			AWT_REWIND( infile);  /* some native converters can't skip the signature read */
			
			switch ( imageFormat( sig) ) {
			case SIG_GIF:
				conv = new PegGifConvert( infile);
				break;
			case SIG_JPEG:
				conv = new PegJpgConvert( infile);
				break;
			default:
				conv = new PegBmpConvert( infile);
				break;
			}

			if ( conv != NULL ) {
				runConv( conv, readFileImageData);
				img = getImages( conv);

				delete conv;
			}
		}
		AWT_CLOSE( infile);
	}
	
	return img;
}


/* can be global, since we use it from a synced Toolkit method, anyway */
UCHAR *picDataBuf;
int   picDataLen;

BOOL
readBufferImageData ( WORD bmpId, WORD convState, PegImageConvert* conv )
{
  int n;

	if ( convState &  PIC_IDLE ) {
		if ( picDataLen ) {
			n = conv->SendData( picDataBuf, picDataLen);  /* returns PEGs buffer size */
			if ( n < picDataLen ) {
				picDataLen -= n;
				picDataBuf += n;
			}
			else {
				picDataBuf = 0;
				picDataLen = 0;
			}
			return TRUE;
		}
	}
	else if ( convState & (PIC_ERROR | PIC_COMPLETE) ) {
		return FALSE;
	}
	else if ( convState & (PIC_HEADER_KNOWN | PIC_PALETTE_KNOWN) ) {
		return TRUE;
	}

  return FALSE;
}

void* _cdecl
Java_java_awt_Toolkit_imgCreateFromData ( JNIEnv* env, jclass clazz,
																				 jarray jbuffer, jint off, jint len )
{
	Image            *img = (Image*)0;
	PegImageConvert  *conv;
	jboolean         isCopy;
	jint             length = env->GetArrayLength( jbuffer);
	jbyte             *jb = env->GetByteArrayElements( jbuffer, &isCopy);
	
	picDataBuf = (unsigned char*) jb + off;
	picDataLen = ((off + len) > length) ? (length - off) : len;
	
	switch ( imageFormat( picDataBuf) ){
	case SIG_GIF:
		conv = new PegGifConvert( 0);
		break;
		
	case SIG_JPEG:
		conv = new PegJpgConvert( 0);
		break;
		
	default:
		conv = new PegBmpConvert( 0);
		break;
	}
	
	runConv( conv, readBufferImageData); 
	img = getImages( conv);
	delete conv;
	
	env->ReleaseByteArrayElements( jbuffer, jb, JNI_ABORT);
	return img;
}

void*
Java_java_awt_Toolkit_imgSetFrame ( JNIEnv* env, jclass clazz, Image* img, int frameNo )
{
  Image *imgCur = img;

  if ( !img->next )
		return img;

  while ( img->frame != frameNo ) {
		img = img->next;

		if ( img == imgCur )
			break;
  }

  return img;
}

/************************************************************************************
 * field access
 */

jint _cdecl
Java_java_awt_Toolkit_imgGetWidth ( JNIEnv* env, jclass clazz, Image* img)
{
  return img->width;
}

jint _cdecl
Java_java_awt_Toolkit_imgGetHeight ( JNIEnv* env, jclass clazz, Image* img)
{
  return img->height;
}

jint _cdecl
Java_java_awt_Toolkit_imgIsMultiFrame ( JNIEnv* env, jclass clazz, Image* img)
{
  int res = ((void*)(img->next) != (void*)0);
	return res;
}

jint _cdecl
Java_java_awt_Toolkit_imgGetLatency ( JNIEnv* env, jclass clazz, Image* img)
{
  return img->latency;
}

void* _cdecl
Java_java_awt_Toolkit_imgGetNextFrame ( JNIEnv* env, jclass clazz, Image* img )
{
  return img->next;   /* next in the ring */
}

}
