/**
 * toolkit.h - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: toolkit.hpp,v 1.14 2001/07/23 23:03:12 samc Exp $
 */

#ifndef __toolkit_h
#define __toolkit_h

#include <jni.h>

#if defined(HAVE_STRING_H)
#include <string.h>
#endif

#pragma pack(push)
#define PEG_FULL_GRAPHICS
#pragma pack(1)
#include <peg.hpp>
#pragma pack(pop)

#include "config.h"
#include "config-io.h"

extern "C" void kprintf( const char*, ...); 
#define DBG(_topic,_msgSpec)
#define DBG_ACTION(_topic,__action)

/*******************************************************************************
 * color conversion structures
 */

typedef struct _Rgb2True {
  unsigned int   redMask;
  unsigned int   greenMask;
  unsigned int   blueMask;
  int            blueShift;
  int            redShift;
  int            greenShift;

	unsigned int   pixRedMask;
	unsigned int   pixGreenMask;
	unsigned int   pixBlueMask;
} Rgb2True;


typedef struct _RgbColor {
  unsigned char r;
  unsigned char g;
  unsigned char b;
} RgbColor;


typedef struct _Rgb2Pseudo {
  RgbColor       rgb[256];
  unsigned char  pix[8][8][8];
} Rgb2Pseudo;

typedef struct _Rgb2Vga {
	jint clrs[16];
} Rgb2Vga;

/*******************************************************************************
 * image handling structures
 */

typedef struct _AlphaImage {       /* storage for full alpha channel images */
  unsigned char *buf;
  int           width, height;
} AlphaImage;


typedef struct _Image {
	PegBitmap        *bmp;
	AlphaImage       *alpha;         /* full alpha channel (for alpha != 0x00 or 0xff) */
	
	int              left, top;      /* for gif animations */
	int              width, height;  /* a bit redundant, but there might be non-bmp images */
	short            latency;        /* between image flips, for "gif-movies" */
	short            frame;          /* frame number for gif animations */
	struct _Image    *next;          /* next movie-frame */
} Image;


/*******************************************************************************
 * structure to store guessed and computed Frame/Dialog insets (titlebar, borders)
 */

typedef struct _DecoInset {
  int            left;
  int            top;
  int            right;
  int            bottom;
  char           guess;
} DecoInset;           

/*****************************************************************************************
 * window wrappers
 */

// This following is just an odd workaround for missing rtti, which isn't available
// for the MS C++ compiler under SMX

typedef union _prealthing {
	PegWindow           *window;
	PegDecoratedWindow  *deco;
	PegDialog           *dialog;
} prealthing;

#define PEG_TYPE_WINDOW 1
#define PEG_TYPE_DECO   2
#define PEG_TYPE_DIALOG 3

class MappedWindow {
public:
	int          srcIdx;
	char         isDrawing;
	char         isVisible;

	prealthing   prt;
	int          pegType;
	PegThing     *pt;

	// we need to save the original PegThing mClip in order to restore it after
	// drawing operations
	PegRect      tgtClip;

	void emitJavaEvents ( const PegMessage& msg );
	void emitKeyEvent ( const PegMessage& msg, int id );

	MappedWindow () {
		srcIdx = -1;
		isDrawing = FALSE;
		isVisible = FALSE;
	}
};

class JPegDialog : public PegDialog, public MappedWindow {
public:
	JPegDialog ( PegRect& bounds )
		: PegDialog( bounds) {
		pt = (PegThing*) this;
		prt.dialog  = (PegDialog*)this;
		pegType = PEG_TYPE_DIALOG;
	}

	virtual SIGNED Message ( const PegMessage &msg );
	virtual void Resize ( PegRect newSize);
	virtual void Draw ();
};

class JPegDecoWindow : public PegDecoratedWindow, public MappedWindow {
public:
	JPegDecoWindow ( PegRect& bounds )
		: PegDecoratedWindow( bounds, FF_THICK) {
		pt = (PegThing*) this;
		prt.deco = (PegDecoratedWindow*) this;
		pegType = PEG_TYPE_DECO;
	}

	virtual SIGNED Message ( const PegMessage &msg );
	virtual void Resize ( PegRect newSize);
	virtual void Draw ();
};

class JPegWindow : public PegWindow, public MappedWindow {
public:
	JPegWindow ( PegRect& bounds )
		: PegWindow( bounds, FF_NONE) {
		pt = (PegThing*) this;
		prt.window = (PegWindow*) this;
		pegType = PEG_TYPE_WINDOW;
	}

	virtual SIGNED Message ( const PegMessage &msg );
	virtual void Resize ( PegRect newSize);
	virtual void Draw ();
};

/*******************************************************************************
 * this is the master AWT structure (singleton object), glueing it all together
 */

typedef struct _Toolkit {
	char           *buf;
	unsigned int   nBuf;
	
	PegScreen      *screen;
	PegPresentationManager *presman;
	
	int            colorMode;      /* refers to CM_xx constants, not X visuals */
	Rgb2True       *tclr;
	Rgb2Pseudo     *pclr;
	Rgb2Vga        *vclr;
	
	UCHAR          pCursor;        /* PEG cursor in effect */
	MappedWindow   *current;
	
	DecoInset      frameInsets;
	DecoInset      dialogInsets;
	
	int            evtId;
	
	MappedWindow   **windows;
	int            nWindows;
} Toolkit;


/*******************************************************************************
 * global data def/decl
 */
#ifdef MAIN

Toolkit   XTk;
Toolkit   *X = &XTk;

jclass  InitClass;
jclass  AWTError;
JNIEnv  *JniEnv;            // unsafe!!

#else

extern Toolkit* X;

extern jclass Tlk;
extern jclass AWTError;
extern JNIEnv* JniEnv;

#endif /* MAIN */

/*******************************************************************************
 *
 */


static inline char* java2CString ( JNIEnv *env, Toolkit* X, jstring jstr ) {
  jboolean       isCopy;
  unsigned int   i, n = env->GetStringLength( jstr);
  const jchar    *jc = env->GetStringChars( jstr, &isCopy);
	
  if ( n >= X->nBuf ) {
		if ( X->buf )
			free( X->buf);
		X->buf = (char*) malloc( n+1);
		X->nBuf = n+1;
  }
	
  for ( i=0; i<n; i++ ) X->buf[i] = (char) jc[i];
  X->buf[i] = 0;
  env->ReleaseStringChars( jstr, jc);
	
  return X->buf;
}

static inline char* jchar2CString ( Toolkit* X, jchar* jc, int len ) {
  register int  i;
  unsigned int  n = len+1;
  
  if ( n > X->nBuf ) {
		if ( X->buf )
			free( X->buf);
		X->buf  = (char*)malloc( n);
		X->nBuf = n;
  }
	
  for ( i=0; i<len; i++ ) X->buf[i] = (char) jc[i];
  X->buf[i] = 0;
	
  return X->buf;
}

static inline char* jbyte2CString ( Toolkit* X, jbyte* jb, int len ) {
  register int  i;
  unsigned int  n = len+1;
  
  if ( n > X->nBuf ) {
		if ( X->buf )
			free( X->buf);
		X->buf  = (char*)malloc( n);
		X->nBuf = n;
  }
	
  for ( i=0; i<len; i++ ) X->buf[i] = jb[i];
  X->buf[i] = 0;
	
  return X->buf;
}

static inline void* getBuffer ( Toolkit* X, unsigned int nBytes ) {
  if ( nBytes > X->nBuf ) {
		if ( X->buf )
			free( X->buf);
		X->buf  = (char*) malloc( nBytes);
		X->nBuf = nBytes;
  }
  return X->buf;
}


/*****************************************************************************************
 * color functions & defines
 */

/* VGA indices */
#define  CLR_BLACK     0
#define  CLR_RED1      1
#define  CLR_GRAY1     2
#define  CLR_ORANGE    3
#define  CLR_DARKGRAY  4
#define  CLR_PINK      5
#define  CLR_GRAY2     6
#define  CLR_LIGHTGRAY 7
#define  CLR_GRAY      8
#define  CLR_RED       9
#define  CLR_GREEN     10
#define  CLR_YELLOW    11
#define  CLR_BLUE      12
#define  CLR_MAGENTA   13
#define  CLR_CYAN      14
#define  CLR_WHITE     15


/*
 * These are our directly supported visuals / color modes. Note that there is
 * no more 1-1 correspondence to X visuals, since we do a categorisation with
 * respect to our internal RGB <-> pixel conversion. All visuals not listed
 * explicitly are handled via the generic XAllocColor/XQueryColor (which might
 * slow down images considerably)
 *
 * NOTE: these values have to be != 0, since '0' is used to trigger color init
 */
#define CM_PSEUDO_256   1  /* PseudoColor visual */
#define CM_TRUE         2  /* general TrueColor visual */
#define CM_TRUE_888     3  /* special 8-8-8 bit TrueColor visual */
#define CM_VGA          5  /* wonderful 16 fixed colors */
#define CM_GENERIC      6  /* grays, DirectColor (packed) etc. */


void initColorMapping ( JNIEnv* env, Toolkit* X);
jint getVGAPixelValue ( Toolkit* X, jint rgb, int mod );

extern "C" {
jlong Java_java_awt_Toolkit_clrBright ( JNIEnv* env, jclass clazz, jint rgb );
jlong Java_java_awt_Toolkit_clrDark ( JNIEnv* env, jclass clazz, jint rgb );
}

#define JRGB(_r,_g,_b)  (_r<<16 | _g<<8 | _b)
#define JRED(_rgb)      ((_rgb & 0xff0000) >> 16)
#define JGREEN(_rgb)    ((_rgb & 0x00ff00) >> 8)
#define JBLUE(_rgb)     (_rgb & 0x0000ff)

#define D8   36.43  /* 255 / 7 */
#define D16  18.21

#define JI8(_v) (int)((_v + D16) / D8)
#define XI8(_v) (int)(((_v>>8) + D16) / D8)

#define ROUND_SHORT2CHAR(_n) \
  ((unsigned short)_n >= 0xff70 ) ? 0xff : (unsigned char)(((unsigned short)_n + 0x80) >> 8)

static inline jint
pixelValue ( Toolkit* X, jint rgb )
{	
  switch ( X->colorMode ) {
	case CM_VGA:
		return getVGAPixelValue( X, rgb, 0);

  case CM_PSEUDO_256:
		return X->pclr->pix [JI8(JRED(rgb))] [JI8(JGREEN(rgb))] [JI8(JBLUE(rgb))];
		
  case CM_TRUE:
		return (((rgb & X->tclr->blueMask)  >> X->tclr->blueShift) |
			((rgb & X->tclr->greenMask) >> X->tclr->greenShift) |
			((rgb & X->tclr->redMask)   >> X->tclr->redShift));
		
  case CM_TRUE_888:
		return (rgb & 0xffffff);
		
  default:
		return 0;
  }
}

static inline void
rgbValues ( Toolkit* X, unsigned long pixel, int* r, int* g, int* b )
{	
  switch ( X->colorMode ) {
	case CM_VGA:
		*r = (X->vclr->clrs[pixel & 0xf]) >> 16;
		*g = (X->vclr->clrs[pixel & 0xf] & 0xff00) >> 8;
		*b = X->vclr->clrs[pixel] & 0xff;
		break;

  case CM_PSEUDO_256:
		*r = X->pclr->rgb[(unsigned char)pixel].r;
		*g = X->pclr->rgb[(unsigned char)pixel].g;
		*b = X->pclr->rgb[(unsigned char)pixel].b;
		break;
		
  case CM_TRUE:
		*r = ((pixel & X->tclr->pixRedMask)   << X->tclr->redShift) >> 16;
		*g = ((pixel & X->tclr->pixGreenMask) << X->tclr->greenShift) >> 8;
		*b = ((pixel & X->tclr->pixBlueMask)  << X->tclr->blueShift);
		break;
		
  case CM_TRUE_888:
		*r = JRED( pixel);
		*g = JGREEN( pixel);
		*b = JBLUE( pixel);
		break;
		
  default:
		*r = *g = *b = 0;
  }
}


/*****************************************************************************************
 * image functions
 */

Image* createImage ( int width, int height);
void createAlphaImage ( Toolkit* X, Image* img );
void initScaledImage ( Toolkit* X, Image *tgt, Image *src,
											int dx0, int dy0, int dx1, int dy1,
											int sx0, int sy0, int sx1, int sy1 );


static inline void
PutAlpha ( AlphaImage* img, int col, int row, unsigned char alpha )
{
  img->buf[ row*img->width + col ] = alpha;
}

static inline int
GetAlpha ( AlphaImage* img, int col, int row )
{
  return img->buf[ row*img->width + col];
}



/*****************************************************************************************
 * heap wrapper macros
 */


static inline void* _awt_malloc_wrapper ( size_t size )
{
  void *adr = malloc( size);
  DBG( awt_mem, ("malloc: %d  -> %x\n", size, adr));
  return adr;
}

static inline void* _awt_calloc_wrapper ( int n, size_t size )
{
  void *adr = calloc( n, size);
  DBG( awt_mem, ("calloc: %d,%d  -> %x\n", n, size, adr));
  return adr;
}

static inline void _awt_free_wrapper ( void* adr )
{
  DBG( awt_mem, ("free: %x\n", adr));
  free( adr);
}


#define AWT_MALLOC(_n) \
  _awt_malloc_wrapper( _n)

#define AWT_CALLOC(_n,_sz) \
  _awt_calloc_wrapper( _n, _sz)

#define AWT_FREE(_adr) \
  _awt_free_wrapper( _adr);


/*****************************************************************************************
 * file io wrapper macros (for image production)
 */

extern "C" {
#include "jfilesystem.h"
}
#define AWT_OPEN(_file)               open(_file, JOPEN_READ)
#define AWT_REWIND(_fd)               lseek(_fd, 0, JSEEK_SET)
#define AWT_SETPOS(_fd,_off)          lseek(_fd, _off, JSEEK_CUR)
#define AWT_READ(_fd,_buf,_count)     read(_fd,_buf,_count)
#define AWT_CLOSE(_fd)                close(_fd)


#endif
