/**
 * theme.c - generic theme drawing support
 * 
 *           This is the swiss army knife for image based theme implementation.
 *           We implement these drawing policy functions here, because they
 *           would require a LOT more native calls if done from Java
 *           (drawImage iterations), and they are generic enough to be used
 *           in a huge number of combinations. Moreover, these functions
 *           are implemented so that they don't require clipping, and don't
 *           rely on the (target specific) Graphics state, which means we can
 *           use a parent Graphics to draw child borders/backgrounds
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


/*
 * public flag values, to be used from Java
 */

/* these flags describe how border images should be aligned (not acumulative) */
#define ALIGN_INNER  0 /* align border images along interior edges */
#define ALIGN_CENTER 1 /* align border images centered */
#define ALIGN_OUTER  2 /* align border images along exterior edges */

/* these flags control how the background is painted (color is accumulative) */
#define FILL_NONE    0 /* do not fill at all */
#define FILL_COLOR   1 /* fill with solid color (accumulative) */
#define FILL_TILE    2 /* tile fill image */
#define FILL_CENTER  4 /* center fill image */


#include "toolkit.h"

#define TL 0
#define BR 1


static inline int max3 ( int a, int b, int c )
{
	if ( a >= b ) {
		return (c > a) ? c : a;
	}
	else {
		return (c > b) ? c : b;
	}
}

static inline int align ( int v, int d, int dmax, int align, int edge )
{	
	if ( d == dmax ) {
		if (edge != TL) v -= dmax;
	}
	else {
		if ( align == ALIGN_OUTER ) {
			if (edge != TL) v -= d;
		}
		else if ( align == ALIGN_CENTER ) {
			v += (dmax - d) / 2;
			if ( edge != TL ) v -= dmax;
		}
		else {
			if ( edge == TL ) v += dmax - d;
			else              v -= dmax;
		}
	}
	
	return v;
}

static inline void getImageSize ( Image* im, int* w, int* h )
{
	if ( im ) {
		*w = _(im)->width;
		*h = _(im)->height;
	}
	else {
		*w = 0;
		*h = 0;
	}
}

void
Java_java_awt_Toolkit_thmDrawBorder ( JNIEnv* env, jclass clazz,
									  Graphics* gr, jint x, jint y, jint w, jint h,
									  jint alignment,
									  Image* itl, Image* itm, Image* itr,
									  Image* iml, Image* imr,
									  Image* ibl, Image* ibm, Image* ibr )
{
	int x0, x1, y0, y1;
	int wtl, wtm, wtr, htl, htm, htr, wml, hml, wmr, hmr, wbl, wbm, wbr, hbl, hbm, hbr;
	int xtl, ytl, xtr, ytr, xbl, ybl, xbr, ybr, xr, yb;
	int ht, hb, wl, wr, d;
	int xaml=0, xamr=0;
	
	/* this is the bounding box of the decorated region */
	x0 = x;
	x1 = x + w;
	y0 = y;
	y1 = y + h;
	
	getImageSize( itl, &wtl, &htl);
	getImageSize( itm, &wtm, &htm);
	getImageSize( itr, &wtr, &htr);

	getImageSize( iml, &wml, &hml);
	getImageSize( imr, &wmr, &hmr);

	getImageSize( ibl, &wbl, &hbl);
	getImageSize( ibm, &wbm, &hbm);
	getImageSize( ibr, &wbr, &hbr);
	
	/* max top, bottom, left, right border extends */
	ht = max3( htl, htm, htr);
	hb = max3( hbl, hbm, hbr);
	wl = max3( wtl, wml, wbl);
	wr = max3( wtr, wmr, wbr);	
	
	/* default inner coords of corners */
	xtl = x0;
	ytl = y0;
	xtr = x1;
	ytr = y0;
	xbl = x0;
	ybl = y1;
	xbr = x1;
	ybr = y1; 
	
	/* upper left corner */
	if ( itl ) {
		x = align( x0, wtl, wl, alignment, TL);
		y = align( y0, htl, ht, alignment, TL);
		Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, itl,
										   0, 0, x, y, wtl, htl, -1);
		xtl = x + wtl;
		ytl = y + htl;
	}

	/* upper right corner */
	if ( itr ) {
		x = align( x1, wtr, wr, alignment, BR);
		y = align( y0, htr, ht, alignment, TL);
		Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, itr,
										   0, 0, x, y, wtr, htr, -1);
		xtr = x;
		ytr = y + htr;
	}

	/* lower left corner */
	if ( ibl ) {
		x = align( x0, wbl, wl, alignment, TL);
		y = align( y1, hbl, hb, alignment, BR);
		Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, ibl,
										   0, 0, x, y, wbl, hbl, -1);
		xbl = x + wbl;
		ybl = y;
	}

	/* lower right corner */
	if ( ibr ) {
		x = align( x1, wbr, wr, alignment, BR);
		y = align( y1, hbr, hb, alignment, BR);
		Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, ibr,
										   0, 0, x, y, wbr, hbr, -1);
		xbr = x;
		ybr = y;
	}

	/* compute x alignment of middle borders in case we don't have corners
	 * (at least two borders have to be aligned to each other)
	 */
	if ( iml )   xaml = align( x0, wml, wl, alignment, TL);
	if ( imr )   xamr = align( x1, wmr, wr, alignment, BR);
		
 	/* upper border, left to right */
	if ( itm ) {
		if ( !itl && iml) xtl = xaml + wml;
		if ( !itr && imr) xtr = xamr;
		
		y = align( y0, htm, ht, alignment, TL);
		d = wtm;
		for ( x = xtl; x < xtr; x = xr ) {
			xr = x + wtm;
			if ( xr > xtr ) d = xtr - x;
		
			Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, itm,
											   0, 0, x, y, d,  htm, -1);
		}
		
		y += htm;
		if ( ytl < y)  ytl = y;
		if ( ytr < y ) ytr = y;
	}

	/* lower border, left to right */
	if ( ibm ) {
		if ( !ibl && iml) xbl = xaml + wml;
		if ( !ibr && imr) xbr = xamr;
		
		y = align( y1, hbm, hb, alignment, BR);
		d = wbm;
		for ( x = xbl; x < xbr; x = xr ) {
			xr = x + wbm;
			if ( xr > xbr ) d = xbr - x;
		
			Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, ibm,
											   0, 0, x, y, d, hbm, -1);
		}
		
		if ( ybl > y ) ybl = y;
		if ( ybr > y ) ybr = y;
	}

	
	/* left border top to bottom */
	if ( iml ) {
		x = xaml;
		d = hml;
		for ( y = ytl; y < ybl; y = yb ) {
			yb = y + hml;
			if ( yb > ybl ) d = ybl - y;
			Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, iml,
											   0, 0, x, y, wml, d, -1);
		}
	}

	/* right border top to bottom */
	if ( imr ) {		
		x = xamr;
		d = hmr;
		for ( y = ytr; y < ybr; y = yb ) {
			yb = y + hmr;
			if ( yb > ybr ) d = ybr - y;
		
			Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, imr,
											    0, 0, x, y, wmr, d, -1);
		}
	}			
}

void
Java_java_awt_Toolkit_thmDrawBackground ( JNIEnv* env, jclass clazz, Graphics* gr,
										  jint cx, jint cy, jint cw, jint ch,
										  jint x, jint y, jint w, jint h,
										  jint xOff, jint yOff,
										  Image* img, jint clrFill, jint fillMode )
{
	int u, v, uw, vh, us, vs, du, dv, u0, v0, u1, v1;
	int cxw  = cx + cw;
	int cyh  = cy + ch;
	int xw   = x + w;
	int yh   = y + h;
	unsigned int clr;
	int iw, ih;
	
	/* be paranoid, clip to bounding rectangle */
	if ( x < cx ) {
		w -= (cx - x);
		x = cx;
	}
	if ( y < cy ) {
		h -= (cy - y);
		y = cy;
	}
	if ( xw > cxw ) {
		w -= (xw - cxw);
		xw = cxw;
	}
	if ( yh > cyh ) {
		h -= (yh - cyh);
		yh = cyh;
	}

	if ( fillMode & FILL_COLOR ) {
		clr = g_getColor( gr);
		g_setColor( gr, clrFill);
		Java_java_awt_Toolkit_graFillRect( env, clazz, gr, x, y, w, h);
		g_setColor( gr, clr);
	}
		
	if ( img ) {
		iw = _(img)->width;
		ih = _(img)->height;
		
		if ( (fillMode & FILL_TILE) ) {

			/* upper left tile origin */
			//u0 = (x - (cx - xOff)) / iw * iw - xOff + cx;
			//v0 = (y - (cy - yOff)) / ih * ih - yOff + cy;

			/* this computes the first tile origin under the
			 * assumption that background tiling extends into infinity
			 * (is not constrained by the content extend
			 * */
			if ( xOff < 0 ){
				u0 = x - (iw + (xOff + (x + cx)) % iw);
			}
			else {
				u0 = x - ((x - cx) + xOff) % iw;
			}			
			
			if ( yOff < 0 ){
				//v0 = y - (ih - (-yOff - (y - cy)) % ih);
				v0 = y - (ih + (yOff + y - cy) % ih);
			}
			else {
				v0 = y - ((y - cy) + yOff) % ih;
			}
			
			for ( v=v0; v < yh; v = vh ){
				vh = v + ih;
				
				/* clip top */
				if ( v < y ) {
					vs = y - v;
					dv = ih - vs;
					v = y;
				}
				else {
					vs = 0;
					dv = ih;
				}
				
				/* clip bottom */
				if ( vh > yh ) {
					dv -= (vh - yh);
				}
				
				for ( u=u0; u < xw; u = uw ) {
					uw = u + iw;
					
					/* clip left */
					if ( u < x ) {
						us = x - u;
						du = iw - us;
						u = x;
					}
					else {
						us = 0;
						du = iw;
					}
					
					/* clip right */
					if ( uw > xw ) {
						du -= (uw - xw);
					}

					Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, img,
													    us, vs, u, v, du, dv, -1);
				}
			}			
		}
		else if ( fillMode & FILL_CENTER ) {

			u0 = cx + (cw - iw) / 2 - xOff;
			v0 = cy + (ch - ih) / 2 - yOff;
			u1 = u0 + iw;
			v1 = v0 + ih;
			
			if ( (u1 < x) || (v1 < y) || (u0 > xw) || (v0 > yh) ) return;
			
			if ( u0 < x ) {
				us = x - u0;
			}
			else {
				us = 0;
				x  = u0;
			}
			if ( (us + w) > iw ) {
				w = iw - us;
			}
			
			if ( v0 < y ) {
				vs = y - v0;
			}
			else {
				vs = 0;
				y = v0;
			}
			if ( (vs + h) > ih ) {
				h = ih - vs;
			}

			Java_java_awt_Toolkit_graDrawImage( env, clazz, gr, img,
											   us, vs, x, y, w, h, -1);
		}
	}
}
