/*
 * kaffe.media.VideoCtrl.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "files.h"
#include "jsyscall.h"
#include "defs.h"
#include "jni.h"
#include "../../../kaffe/kaffevm/support.h"

#include <sys/ipc.h>
#include <sys/shm.h>

#include "VideoCtrl.h"


#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#if defined(TARGET_DEVICE_IPAQ90)
#define	ROT_90 1
#endif

#define err_ctrl(format, arg...)  kprintf( format, ## arg)

typedef struct _vctlr {
  int filedes[2];
  video_smem *vsync;
  char pipe_buf[256];
} VideoCtrl;

/*****************************
 * helper functions
 *****************************/

void write_pipe( VideoCtrl *ctrl) {
  write( ctrl->filedes[1], ctrl->pipe_buf, strlen( ctrl->pipe_buf) + 1);
}

int run_decoder( VideoCtrl *ctrl, const char *decoder) {
  pid_t pid;
  int sid;
  void *addr;
  char buf0[20];
  char buf1[20];

  if ( pipe( ctrl->filedes) == -1 ) {
    err_ctrl( "failed to open pipes\n");
    return 1;
  }
  
  sid = shmget( IPC_PRIVATE, sizeof( video_smem),IPC_CREAT | 0777);

  if ( sid == -1 ){
    err_ctrl( "failed to create shared memory\n");
    return 2;
  }

  addr = shmat( sid, 0, 0);
  if ( addr == (void*)-1 ) {
    err_ctrl( "failed to access shared memory: %d\n", sid); 
    return 3;
  }

  ctrl->vsync = (video_smem*)addr;

  if ( (pid = fork() ) == 0 ) {
    sprintf( buf0, "%s=%d", "--pipe", ctrl->filedes[0]);
    sprintf( buf1, "%s=%d", "--shmem", sid);
 
    if ( -1 == execl( decoder, decoder, buf0, buf1, NULL) ) {
      shmdt( addr);
      err_ctrl( "failed to start decoder: %s (reason: %d)\n", decoder, errno);
      return 4;
    }
  }

  return 0;
}

/*********************************
 * exported functions
 *********************************/
 
JNIEXPORT void* JNICALL
Java_kaffe_media_VideoCtrl_open0( JNIEnv *env, jclass clazz, jstring decoder)
{
  VideoCtrl *ctrl;
  const char* str;

  if ( decoder ) {
    ctrl = calloc( 1, sizeof( VideoCtrl));
    str = (*env)->GetStringUTFChars(env, decoder, 0);

    if ( ! run_decoder( ctrl, str)) {
      (*env)->ReleaseStringUTFChars(env, decoder, str);
      return ctrl;
    }

    (*env)->ReleaseStringUTFChars(env, decoder, str);
    free( ctrl);
  }

  return NULL;
}

JNIEXPORT jint JNICALL
Java_kaffe_media_VideoCtrl_getPictureWidth( JNIEnv* env, jclass clazz, VideoCtrl *ctrl )
{
    int w;
    LOCK_SMEM( ctrl->vsync);
#if defined(ROT_90)
    w = ctrl->vsync->encoded_height;
#else
    w = ctrl->vsync->encoded_width;
#endif
    UNLOCK_SMEM( ctrl->vsync);

    return w;
}

JNIEXPORT jint JNICALL
Java_kaffe_media_VideoCtrl_getPictureHeight( JNIEnv* env, jclass clazz, VideoCtrl *ctrl )
{
    int h;
    LOCK_SMEM( ctrl->vsync);
#if defined(ROT_90)
    h = ctrl->vsync->encoded_width;
#else
    h = ctrl->vsync->encoded_height;
#endif
    UNLOCK_SMEM( ctrl->vsync);

    return h;
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_stop ( JNIEnv* env, jclass clazz, VideoCtrl *ctrl) 
{
  LOCK_SMEM( ctrl->vsync);
  ctrl->vsync->stop_production = 1;
  UNLOCK_SMEM( ctrl->vsync);
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_start ( JNIEnv* env, jclass clazz, VideoCtrl *ctrl, jstring name )
{
    const char* str;

    if ( name ) {
	str = (*env)->GetStringUTFChars(env, name, 0);
	strcpy( ctrl->pipe_buf, str);
	write_pipe( ctrl);
	(*env)->ReleaseStringUTFChars(env, name, str);
    }

 }

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_suspend( JNIEnv* env, jclass clazz, VideoCtrl *ctrl)
{
    LOCK_SMEM( ctrl->vsync);
    ctrl->vsync->modifiers |= VSM_SUSPEND_MODE;
    ctrl->vsync->suspended = 1;
    UNLOCK_SMEM( ctrl->vsync);
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_resume( JNIEnv* env, jclass clazz, VideoCtrl *ctrl)
{
    LOCK_SMEM( ctrl->vsync);
    ctrl->vsync->modifiers |= VSM_SUSPEND_MODE;
    ctrl->vsync->suspended = 0;
    UNLOCK_SMEM( ctrl->vsync);
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_setChromaKey( JNIEnv* env, jclass clazz, VideoCtrl *ctrl, jint color) 
{
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_setBounds( JNIEnv* env, jclass clazz, VideoCtrl *ctrl,jint x, jint y, jint width, jint height)
{
    video_smem *s = ctrl->vsync;

    LOCK_SMEM( s);
    ctrl->vsync->modifiers |= VSM_BOUNDS;

#if defined(ROT_90)
    s->x = y;
    s->y = x;
    s->width = height;
    s->height = width;
#else
    s->x = x;
    s->y = y;
    s->width = width;
    s->height = height;
#endif

    UNLOCK_SMEM( s);
}

JNIEXPORT jint JNICALL
Java_kaffe_media_VideoCtrl_getBrightness( JNIEnv* env, jclass clazz, VideoCtrl *ctrl ) 
{
    int val;
    LOCK_SMEM( ctrl->vsync);
    val = ctrl->vsync->brightness;
    UNLOCK_SMEM( ctrl->vsync);
    return val;
}

JNIEXPORT jint JNICALL
Java_kaffe_media_VideoCtrl_getVolume( JNIEnv* env, jclass clazz, VideoCtrl *ctrl) 
{
    int val;
    LOCK_SMEM( ctrl->vsync);
    val = ctrl->vsync->volume;
    UNLOCK_SMEM( ctrl->vsync);
    return val;
}

JNIEXPORT jint JNICALL
Java_kaffe_media_VideoCtrl_getContrast( JNIEnv* env, jclass clazz, VideoCtrl *ctrl ) 
{
    int val;
    LOCK_SMEM( ctrl->vsync);
    val = ctrl->vsync->contrast;
    UNLOCK_SMEM( ctrl->vsync);
    return val;
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_setBrightness( JNIEnv* env, jclass clazz, VideoCtrl *ctrl, jint brightness ) 
{
    LOCK_SMEM( ctrl->vsync);
    ctrl->vsync->modifiers |= VSM_BRIGHTNESS;
    ctrl->vsync->brightness = brightness;
    UNLOCK_SMEM( ctrl->vsync);
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_setVolume( JNIEnv* env, jclass clazz, VideoCtrl *ctrl, jint volume ) 
{
    LOCK_SMEM( ctrl->vsync);
    ctrl->vsync->modifiers |= VSM_VOLUME;
    ctrl->vsync->volume = volume;
    UNLOCK_SMEM( ctrl->vsync);
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_setContrast( JNIEnv* env, jclass clazz, VideoCtrl *ctrl, jint contrast )
{
    LOCK_SMEM( ctrl->vsync);
    ctrl->vsync->modifiers |= VSM_CONTRAST;
    ctrl->vsync->contrast = contrast;
    UNLOCK_SMEM( ctrl->vsync);
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_close0( JNIEnv* env, jclass clazz, VideoCtrl *ctrl ) {
    shmdt( ctrl->vsync);
    free( ctrl);
}

JNIEXPORT void JNICALL
Java_kaffe_media_VideoCtrl_nextFrame( JNIEnv* env, jclass clazz, VideoCtrl *ctrl, jint timeout) 
{
  unsigned long ut = 1000L * timeout;

  do {
    if ( ! ctrl->vsync->rendering ) {
      break;
    }
    usleep( 10000);
    ut -= 10000;
  }while ( ut > 10000 );

}


