/*
 * java.util.Arrays.c
 *
 * Copyright (c) 1999
 *	Archie L. Cobbs. All rights reserved.
 * Copyright (c) 1999 - 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * Author: Archie L. Cobbs <archie@whistle.com>
 */

#include "config.h"
#include "config-std.h"
#include <jni.h>
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/fp.h"
#include "../../../kaffe/kaffevm/gc.h"
#include "../../../kaffe/kaffevm/exception.h"
#include "md.h"

#if defined(HAVE_MODULE_COLLECTIONS)

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static JNIEnv* genv;
static jobject gcomp;
static jmethodID gmeth;

/*
 * Sorting functions
 *
 * We rely on libc routines to do the sorting because they are already
 * highly optimized and debugged.
 *
 * Note that these routines assume the bounds checking has already been
 * done in Java.
 */

static int
cmpByte(const void *p1, const void *p2)
{
	return (int)*((const jbyte *) p1) - (int)*((const jbyte *) p2);
}

static int
cmpChar(const void *p1, const void *p2)
{
	return (int)*((const jchar *) p1) - (int)*((const jchar *) p2);
}

static int
cmpDouble(const void *p1, const void *p2)
{
	const jlong bits1 = doubleToLong(*((const jdouble *) p1));
	const jlong bits2 = doubleToLong(*((const jdouble *) p2));

	return (bits1 == bits2) ? 0 : (bits1 < bits2) ? -1 : 1;
}

static int
cmpFloat(const void *p1, const void *p2)
{
	const jint bits1 = floatToInt(*((const jfloat *) p1));
	const jint bits2 = floatToInt(*((const jfloat *) p2));

	return (bits1 == bits2) ? 0 : (bits1 < bits2) ? -1 : 1;
}

static int
cmpInt(const void *p1, const void *p2)
{
	const jint int1 = *((const jint *) p1);
	const jint int2 = *((const jint *) p2);

	return (int1 == int2) ? 0 : (int1 < int2) ? -1 : 1;
}

static int
cmpShort(const void *p1, const void *p2)
{
	return (int)*((const jshort *) p1) - (int)*((const jshort *) p2);
}

static int
cmpLong(const void *p1, const void *p2)
{
	const jlong long1 = *((const jlong *) p1);
	const jlong long2 = *((const jlong *) p2);

	return (long1 == long2) ? 0 : (long1 < long2) ? -1 : 1;
}

static
int
cmpObject(const void *p1, const void *p2)
{
	return ((*genv)->CallIntMethod(genv, gcomp, gmeth, *(jobject*)p1, *(jobject*)p2));
}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortByte(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex)
{
	jbyte* ja = (*env)->GetByteArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jbyte), cmpByte);
	(*env)->ReleaseByteArrayElements(env, a, ja, JNI_COMMIT);

}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortChar(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex)
{
	jchar* ja = (*env)->GetCharArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jchar), cmpChar);
	(*env)->ReleaseCharArrayElements(env, a, ja, JNI_COMMIT);
}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortDouble(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex)
{
	jdouble* ja = (*env)->GetDoubleArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jdouble), cmpDouble);
	(*env)->ReleaseDoubleArrayElements(env, a, ja, JNI_COMMIT);
}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortFloat(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex)
{
	jfloat* ja = (*env)->GetFloatArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jfloat), cmpFloat);
	(*env)->ReleaseFloatArrayElements(env, a, ja, JNI_COMMIT);
}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortInt(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex)
{
	jint* ja = (*env)->GetIntArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jint), cmpInt);
	(*env)->ReleaseIntArrayElements(env, a, ja, JNI_COMMIT);
}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortShort(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex)
{
	jshort* ja = (*env)->GetShortArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jshort), cmpShort);
	(*env)->ReleaseShortArrayElements(env, a, ja, JNI_COMMIT);
}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortLong(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex)
{
	jlong* ja = (*env)->GetLongArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jlong), cmpLong);
	(*env)->ReleaseLongArrayElements(env, a, ja, JNI_COMMIT);
}

JNIEXPORT void JNICALL
Java_java_util_Arrays_sortObject(JNIEnv* env, jclass array, jarray a, jint fromIndex, jint toIndex, jobject c)
{
	/* This method is sync - so this stacking will work okay */
	jobject* ja;
	jobject oldcomp;
	jmethodID oldmeth;

	genv = env;
	oldcomp = gcomp;
	gcomp = c;
	oldmeth = gmeth;
	gmeth = (*env)->GetMethodID(env, (*env)->GetObjectClass(env, c), "compare", "(Ljava/lang/Object;Ljava/lang/Object;)I");

	/* This should be GetObjectArrayElements but there isn't one in JNI
	 * for some daft reason.  So we get them as int's instead and hope
	 * for the best.
	 */
	ja = (jobject*)(*env)->GetIntArrayElements(env, a, 0);
	qsort(&ja[fromIndex], toIndex - fromIndex, sizeof(jobject), cmpObject);
	(*env)->ReleaseIntArrayElements(env, a, (jint*)ja, JNI_COMMIT);

	gcomp = oldcomp;
	gmeth = oldmeth;
}

#endif
