/*
 * java.lang.ClassReflection.c
 *
 * Copyright (c) 1996 - 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/soft.h"
#include "../../../kaffe/kaffevm/engine.h"
#include "../../../kaffe/kaffevm/exception.h"
#include <jni.h>
#include "defs.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static jclass classc;

static jclass constructorc;
static jclass methodc;
static jclass fieldc;

static jmethodID mconstructor;
static jmethodID mmethod;
static jmethodID mfield;

static jfieldID cclazzf;
static jfieldID cslotf;
static jfieldID cparamf;
static jfieldID cexceptf;

static jfieldID mclazzf;
static jfieldID mslotf;
static jfieldID mnamef;
static jfieldID mparamf;
static jfieldID mexceptf;
static jfieldID mreturnf;

static jfieldID fclazzf;
static jfieldID fslotf;
static jfieldID fnamef;
static jfieldID ftypef;

static
void
init(JNIEnv* env)
{
	if (classc == 0) {
		classc = (*env)->FindClass(env, "java.lang.Class");

		constructorc = (*env)->FindClass(env, "java.lang.reflect.Constructor");
		methodc = (*env)->FindClass(env, "java.lang.reflect.Method");
		fieldc = (*env)->FindClass(env, "java.lang.reflect.Field");

		mconstructor = (*env)->GetMethodID(env, constructorc, "<init>", "()V");
		mmethod = (*env)->GetMethodID(env, methodc, "<init>", "()V");
		mfield = (*env)->GetMethodID(env, fieldc, "<init>", "()V");

		cclazzf = (*env)->GetFieldID(env, constructorc, "clazz", "java.lang.Class");
		cslotf = (*env)->GetFieldID(env, constructorc, "slot", "I");
		cparamf = (*env)->GetFieldID(env, constructorc, "parameterTypes", "[java.lang.Class");
		cexceptf = (*env)->GetFieldID(env, constructorc, "exceptionTypes", "[java.lang.Class");

		mclazzf = (*env)->GetFieldID(env, methodc, "clazz", "java.lang.Class");
		mslotf = (*env)->GetFieldID(env, methodc, "slot", "I");
		mnamef = (*env)->GetFieldID(env, methodc, "name", "java.lang.String");
		mparamf = (*env)->GetFieldID(env, methodc, "parameterTypes", "[java.lang.Class");
		mexceptf = (*env)->GetFieldID(env, methodc, "exceptionTypes", "[java.lang.Class");
		mreturnf = (*env)->GetFieldID(env, methodc, "returnType", "java.lang.Class");

		fclazzf = (*env)->GetFieldID(env, fieldc, "clazz", "java.lang.Class");
		fslotf = (*env)->GetFieldID(env, fieldc, "slot", "I");
		fnamef = (*env)->GetFieldID(env, fieldc, "name", "java.lang.String");
		ftypef = (*env)->GetFieldID(env, fieldc, "type", "java.lang.Class");
	}
}

typedef struct _methodSig {
	int	len;
	char*	ret;
	char*	elem[1];
} methodSig;

/*
 * Build a signature (as an array of classes) based on the
 * textual signature
 */
static
methodSig*
makeSignature(char* sig)
{
	char* data;
	methodSig* msig;
	int len;
	int i;

	/* Work out how many elements in signature */
	data = sig + 1;
	for (len = 0; *data != ')'; len++) {
		if (*data == '[') {
			while (*data == '[') {
				data++;
			}
			if (*data == 'L') {
				while (*data != ';') {
					data++;
				}
			}
		}
		else if (*data == 'L') {
			while (*data != ';') {
				data++;
			}
		}
		data++;
	}

	msig = gc_malloc_fixed(sizeof(methodSig) + (len-1)*sizeof(Utf8Const*));
	msig->len = len;
	data = sig + 1;
	for (i = 0; i < len; i++) {
		msig->elem[i] = utfFromSig(&data);
                /* GC_WRITE(class, msig->elem[i]); */
                /* FIXME: GC_WRITE needs class pointer here */
	}
	data++;	/* Skip ')' */
	msig->ret = utfFromSig(&data);
        /* GC_WRITE(class, msig->ret); */
        /* FIXME: GC_WRITE needs class pointer here */

	return (msig);
}

static
jboolean
checkParameters(JNIEnv* env, jarray arr, char* signature)
{
	int i;
	char* name;
	Hjava_lang_Class* one;
	Hjava_lang_Class* two;
	methodSig* sig;

	sig = makeSignature(signature);

	if (arr == 0 && sig->len == 0) {
		gc_free_fixed(sig);
		return (JNI_TRUE);
	}
	if (sig->len != (*env)->GetArrayLength(env, arr)) {
		gc_free_fixed(sig);
		return (JNI_FALSE);
	}
	for (i = 0; i < sig->len; i++) {
		one = (Hjava_lang_Class*)(*env)->GetObjectArrayElement(env, arr, i);
		name = sig->elem[i];
		two = classFromSigProcess(&name, one->loader, CSTATE_LINKED);
		if (one != two) {
			gc_free_fixed(sig);
			return (JNI_FALSE);
		}
	}
	gc_free_fixed(sig);
	return (JNI_TRUE);
}

static
jarray
makeParameters(JNIEnv* env, Method* meth, methodSig* sig, Hjava_lang_ClassLoader* loader)
{
	int i;
	jarray array;
	char* name;

	array = (*env)->NewObjectArray(env, sig->len, classc, 0);
	for (i = 0; i < sig->len; i++) {
		name = sig->elem[i];
		(*env)->SetObjectArrayElement(env, array, i, (jobject)classFromSigProcess(&name, loader, CSTATE_LINKED));
	}

        return (array);
}

static
jobject
makeReturn(Method* meth, methodSig* sig, Hjava_lang_ClassLoader* loader)
{
	char* name;

	name = sig->ret;
	return ((jobject)classFromSigProcess(&name, loader, CSTATE_LINKED));
}

static
jarray
makeExceptions(JNIEnv* env, Method* meth, Hjava_lang_Class* clazz)
{
        int nr;
        int i;
        jarray array;

	if (meth->exception_declared == 0) {
		nr = 0;
	}
	else {
		nr = meth->exception_declared->length;
	}

	array = (*env)->NewObjectArray(env, nr, classc, 0);
        for (i = 0; i < nr; i++) {
		(*env)->SetObjectArrayElement(env, array, i, (jobject)getClass(meth->exception_declared->entry[i], clazz, 0));
        }

        return (array);
}

static
jobject
makeConstructor(JNIEnv* env, Hjava_lang_Class* clazz, int slot)
{
	jobject con;
	Method* mth;
	methodSig* sig;

	mth = &CLASS_METHODS(clazz)[slot];
	sig = makeSignature(mth->signature);
	con = (*env)->NewObject(env, constructorc, mconstructor);

	(*env)->SetObjectField(env, con, cclazzf, (jobject)clazz);
	(*env)->SetIntField(env, con, cslotf, slot);
	(*env)->SetObjectField(env, con, cparamf, makeParameters(env, mth, sig, clazz->loader));
	(*env)->SetObjectField(env, con, cexceptf, makeExceptions(env, mth, clazz));

	gc_free_fixed(sig);

	return (con);
}

static
jobject
makeMethod(JNIEnv* env, Hjava_lang_Class* clazz, int slot)
{
	jobject meth;
	Method* mth;
	methodSig* sig;

	mth = &CLASS_METHODS(clazz)[slot];
	sig = makeSignature(mth->signature);
	meth = (*env)->NewObject(env, methodc, mmethod);

	(*env)->SetObjectField(env, meth, mclazzf, (jobject)clazz);
	(*env)->SetIntField(env, meth, mslotf, slot);
	(*env)->SetObjectField(env, meth, mnamef, (*env)->NewStringUTF(env, mth->name));
	(*env)->SetObjectField(env, meth, mparamf, makeParameters(env, mth, sig, clazz->loader));
	(*env)->SetObjectField(env, meth, mexceptf, makeExceptions(env, mth, clazz));
	(*env)->SetObjectField(env, meth, mreturnf, makeReturn(mth, sig, clazz->loader));

	gc_free_fixed(sig);

	return (meth);
}

static
jobject
makeField(JNIEnv* env, Hjava_lang_Class* clazz, int slot)
{
	jobject field;
	Field* fld;

	fld = &CLASS_FIELDS(clazz)[slot];
	field = (*env)->NewObject(env, fieldc, mfield);

	(*env)->SetObjectField(env, field, fclazzf, (jobject)clazz);
	(*env)->SetIntField(env, field, fslotf, slot);
	(*env)->SetObjectField(env, field, fnamef, (*env)->NewStringUTF(env, fld->name));
	(*env)->SetObjectField(env, field, ftypef, (jobject)resolveFieldType(fld, clazz));

	return (field);
}
 
static
jarray
getClassMethods(JNIEnv* env, jclass this, jint mask, jint flags)
{
	int count;
	Hjava_lang_Class* clas;
	int i;
	Method* mth;
	jarray array;
	int constructor;

	count = 0;
	for (clas = (Hjava_lang_Class*)this; clas != NULL; clas = clas->superclass) {
		mth = CLASS_METHODS(clas);
		i = CLASS_NMETHODS(clas);
		for (i = CLASS_NMETHODS(clas)-1 ; i >= 0; i--) {
			if ((mth[i].accflags & mask) == flags) {
				count++;
			}
		}
		if ((flags & ACC_PUBLIC) == 0 || (flags & ACC_CONSTRUCTOR) != 0) {
			break;
		}
	}

	constructor = (flags & ACC_CONSTRUCTOR);
	array = (*env)->NewObjectArray(env, count, (constructor != 0 ? constructorc : methodc), 0);

	count = 0;
	for (clas = this; clas != NULL; clas = clas->superclass) {
		mth = CLASS_METHODS(clas);
		for (i = CLASS_NMETHODS(clas)-1; i >= 0; i--) {
			if ((mth[i].accflags & mask) == flags) {
				if (constructor != 0) {
					(*env)->SetObjectArrayElement(env, array, count, makeConstructor(env, clas, i));
				}
				else {
					(*env)->SetObjectArrayElement(env, array, count, makeMethod(env, clas, i));
				}
				count++;
			}
		}
		if ((flags & ACC_PUBLIC) == 0 || (flags & ACC_CONSTRUCTOR) != 0) {
			break;
		}
	}

	return (array);
}

static
jarray
getInterfaceMethods(JNIEnv* env, jclass this, jint mask, jint flags)
{
	int count;
	Hjava_lang_Class* clas;
	jarray array;
	int i;
	Method* mth;
	int c;

	count = 0;
	for (c = -1; c < ((Hjava_lang_Class*)this)->total_interface_len; c++) {
		if (c == -1) {
			clas = ((Hjava_lang_Class*)this);
		}
		else {
			clas = ((Hjava_lang_Class*)this)->interfaces[c];
		}
		mth = CLASS_METHODS(clas);
		i = CLASS_NMETHODS(clas);
		for (i = CLASS_NMETHODS(clas)-1 ; i >= 0; i--) {
			if ((mth[i].accflags & mask) == flags) {
				count++;
			}
		}
		if ((flags & ACC_PUBLIC) == 0) {
			break;
		}
	}
	array = (*env)->NewObjectArray(env, count, methodc, 0);

	count = 0;
	for (c = -1; c < ((Hjava_lang_Class*)this)->total_interface_len; c++) {
		if (c == -1) {
			clas = ((Hjava_lang_Class*)this);
		}
		else {
			clas = ((Hjava_lang_Class*)this)->interfaces[c];
		}
		mth = CLASS_METHODS(clas);
		for (i = CLASS_NMETHODS(clas)-1; i >= 0; i--) {
			if ((mth[i].accflags & mask) == flags) {
				(*env)->SetObjectArrayElement(env, array, count, makeMethod(env, clas, i));
				count++;
			}
		}
		if ((flags & ACC_PUBLIC) == 0) {
			break;
		}
	}

	return (array);
}

JNIEXPORT jarray JNICALL
Java_java_lang_Class_getMethods0(JNIEnv* env, jclass this, jint mask, jint flags)
{
	init(env);

	if (CLASS_IS_INTERFACE((Hjava_lang_Class*)this)) {
		return (getInterfaceMethods(env, this, mask, flags));
	}
	else {
		return (getClassMethods(env, this, mask, flags));
	}
}

JNIEXPORT jarray JNICALL
Java_java_lang_Class_getFields0(JNIEnv* env, jclass clazz, jint mask, jint flags)
{
	int count;
	Hjava_lang_Class* clas;
	Field* fld;
	jarray array;
	int i;

	init(env);

	count = 0;
	for (clas = clazz; clas != NULL; clas = clas->superclass) {
		fld = CLASS_FIELDS(clas);
		for (i = CLASS_NFIELDS(clas)-1; i >= 0; i--) {
			if ((fld[i].accflags & mask) == flags) {
				count++;
			}
		}
		if ((flags & ACC_PUBLIC) == 0) {
			break;
		}
	}

	array = (*env)->NewObjectArray(env, count, fieldc, 0);

	count = 0;
	for (clas = clazz; clas != NULL; clas = clas->superclass) {
		fld = CLASS_FIELDS(clas);
		for (i = CLASS_NFIELDS(clas)-1; i >= 0;  i--) {
			if ((fld[i].accflags & mask) == flags) {
				(*env)->SetObjectArrayElement(env, array, count, makeField(env, clas, i));
				count++;
			}
		}
		if ((flags & ACC_PUBLIC) == 0) {
			break;
		}
	}

	return (array);
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getMethod0(JNIEnv* env, jclass this, jstring name, jarray arr, jint mask, jint flags)
{
	Hjava_lang_Class* clas;
	char* cname;

	init(env);

	cname = (char*)(*env)->GetStringUTFChars(env, name, 0);

	clas = this;
	do {
		Method* mth = CLASS_METHODS(clas);
		int n = CLASS_NMETHODS(clas);
		int i;
		for (i = 0;  i < n;  ++mth, ++i) {
			if ((mth->accflags & mask) == flags && strcmp(mth->name, cname) == 0 && checkParameters(env, arr, mth->signature)) {
				(*env)->ReleaseStringUTFChars(env, name, cname);
				if ((flags & ACC_CONSTRUCTOR) == 0) {
					return ((jobject)makeMethod(env, clas, i));
				}
				else {
					return ((jobject)makeConstructor(env, clas, i));
				}
			}
		}
		clas = clas->superclass;
	} while ((flags & ACC_PUBLIC) != 0 && clas != 0);

	(*env)->ReleaseStringUTFChars(env, name, cname);
	(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.NoSuchMethodException"), "");
	return (0);
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getField0(JNIEnv* env, jclass clazz, jstring name, int mask, int flags)
{
	Hjava_lang_Class* clas;
	char* cname;

	init(env);

	cname = (char*)(*env)->GetStringUTFChars(env, name, 0);

	clas = (Hjava_lang_Class*) clazz;
	do {
		Field* fld = CLASS_FIELDS(clas);
		int n = CLASS_NFIELDS(clas);
		int i;
		for (i = 0;  i < n;  ++fld, ++i) {
			if ((fld->accflags & mask) == flags && strcmp(fld->name, cname) == 0) {
				(*env)->ReleaseStringUTFChars(env, name, cname);
				return ((jobject)makeField(env, clas, i));
			}
		}
		clas = clas->superclass;
	} while ((flags & ACC_PUBLIC) != 0 && clas != 0);

	(*env)->ReleaseStringUTFChars(env, name, cname);
	(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.NoSuchFieldException"), "");
	return 0;
}
