package java.awt;

import java.awt.event.WindowEvent;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;
import kaffe.util.Condition;
import kaffe.util.Ptr;

/**
 * Dialog - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author P. Mehlitz
 */
public class Dialog
  extends Frame
{
	final private static long serialVersionUID = 5920926903803293709L;
	static Insets dialogInsets;
	static int named;
	static boolean guessedInsets = true;

static {
	// these are just the guesses (to be overwritten by a subsequent
	// setDialogInsets() from the native layer)
	Insets guess = Defaults.DialogInsets;
	
	dialogInsets = new Insets( guess.top, guess.left, guess.bottom, guess.right);
	
	Toolkit.wndSetDialogInsets( guess.top, guess.left, guess.bottom, guess.right);
}

public Dialog ( Frame owner ) {
	this( owner, null, false);
}

public Dialog ( Frame owner, String title ) {
	this( owner, title, false);
}

public Dialog ( Frame owner, String title, boolean isModal) {
	super( owner, title);

	insets = dialogInsets;

	if ( isModal ){
		flags |= IS_MODAL;
	}
}

public Dialog ( Frame owner, boolean isModal ) {
	this( owner, null, isModal);
}

void createNative () {
	// This is the terminal class addNotify() part. DANGER: ptr isn't a real object
	int u = x;
	int v = y;
	int w = width;
	int h = height;

	// insets seem to be set by the JDK during addNotify
	// (no need to create fresh objects since they are insets()-copied anyway <sigh>)
	insets = dialogInsets;

	if ( (Toolkit.flags & Toolkit.EXTERNAL_DECO) != 0 ) {
		// we just pretend to own the deco space, subtract it before going native
		u += insets.left;
		v += insets.top;
		w -= (insets.left + insets.right);
		h -= (insets.top + insets.bottom);
	}

	nativeData = Toolkit.wndCreateDialog( (owner != null) ? owner.nativeData : null,
	                                      title, u, v, w, h,
	                                      cursor.type, bgClr.nativeValue,
	                                      ((flags & IS_RESIZABLE) != 0));
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "dialog" + (named++);
}

public void hide () {
	if ( (flags & IS_MODAL) != 0 )
		dispose();
	else
		super.hide();
}

public boolean isModal() {
	return ((flags & IS_MODAL) != 0);
}

boolean isSystemModal (){
	return ((flags & IS_MODAL) != 0) && (owner == null);
}

public Dimension preferredSize () {
	Dimension d = super.preferredSize();
	
	d.width  += (Defaults.DialogInsets.left + Defaults.DialogInsets.right);
	d.height += (Defaults.DialogInsets.top + Defaults.DialogInsets.bottom);

	return (d);
}

void process ( WindowEvent e ) {
	super.process( e);
	
	// some windowing systems don't support global dialogs - we have to mimic
	// this in terms of never let this one get obscured (otherwise we have a perfect
	// UI lockout, Java wise)
	if ( e.getID() == WindowEvent.WINDOW_DEACTIVATED ){
		if ( isSystemModal() ){
			toFront();
			//requestFocus();
			Toolkit.eventQueue.postEvent( WMEvent.getEvent( this, WMEvent.WM_SET_FOCUS));
		}
	}
}

static void setDecoInsets ( int top, int left, int bottom, int right, int srcIdx  ){
	// this is the native callBack to set exact (calculated) dialog deco extends

	guessedInsets = false;

	dialogInsets.top    = top;
	dialogInsets.left   = left;
	dialogInsets.bottom = bottom;
	dialogInsets.right  = right;

	if ( Frame.guessedInsets ) {
		// let's use that as a better guess for Frame insets (to avoid
		// subsequent resizes)
		Frame.setDecoInsets( top, left, bottom, right, -1);
	}

	// if we got the correction in the context of a initial Window positioning
	// we have to make sure a subsequent ComponentEvt.getEvent() invalidates
	// this instance (which wouldn't be the case if we let its (faked) dimension alone)
	if ( srcIdx != -1 ) {
		Component src = AWTEvent.sources[srcIdx];
		src.width = src.height = 0;
	}
}

public void setModal ( boolean isModal ) {
	if ( isModal ) 
		flags |= IS_MODAL;
	else
		flags &= ~IS_MODAL;
}

public void show () {
	// DEP - should be in setVisible()
	super.show ();
	
	if ( (flags & IS_MODAL) != 0 ) {
		if ( owner != null ) {
			// We don't use native modal dialogs, so let's at least temporarily
			// disable the owner
			owner.propagateTempEnabled( false);
		}
		else {
			// modal and no owner is interpreted as system modal
			AWTEvent.tempEnableAll( false, this);
		}
		
		// Hrmm, can't implement this directly because we can't add public
		// methods to Dialog
		Toolkit.eventThread.runUntil( new Condition() {
			public boolean evaluate (){
				return ((Dialog.this.flags & Component.IS_ADD_NOTIFIED) == 0);
			}
		});
		
		if ( owner != null ){
			owner.propagateTempEnabled( true);
		}
		else {
			// was system modal, get all back (except of the ones which were tempDisabled)
			AWTEvent.tempEnableAll( true, this);
		}
	}
}
}
