package java.awt;

import java.awt.event.FocusEvent;
import java.awt.event.PaintEvent;
import kaffe.awt.Defaults;

/**
 *
 * Copyright (c) 1998
 *   Transvirtual Technologies Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 * @author P.C.Mehlitz
 */
public class EventQueue
{
	AWTEvent localQueue;
	AWTEvent localEnd;
/**
 * The deferredLocalQueue is actually a third priority class of events, to
 * be processed after the localQueue, and after any native-generated events.
 * It is mainly for stuff which can/has to be accumulated (e.g. scroll messages).
 * we don't need a queue end-ref here, since the deferredLocalQueue is for
 * events we want to accumulate (i.e. we don't append too frequently, and
 * probably traverse the whole thing anyway (to solicit events).
 */
	AWTEvent deferredLocalQueue;

static {
	// force static init of AWTEvent (need that for native event system
	// initialization)
	Class c = AWTEvent.class;
}

public EventQueue () {
}

synchronized void dropAll ( Object source ) {
	AWTEvent e = localQueue;
	AWTEvent del, last = null;

	while ( e != null ) {
		if ( AWTEvent.getSource( e) == source ) {
			if ( localEnd == e )
				localEnd = last;

			if ( last == null )
				localQueue = e.next;
			else
				last.next = e.next;

			del = e;
			e = e.next;
			del.recycle();  // watch out - recycle mutates the 'next' field
		}
		else {
			last = e;
			e = e.next;
		}
	}
}

synchronized void dropLiveEvents ( Object source ) {
	AWTEvent e = localQueue;
	AWTEvent del, last = null;

	while ( e != null ) {
		if ( e.isLiveEventFor( source) ) {
			if ( localEnd == e )
				localEnd = last;

			if ( last == null )
				localQueue = e.next;
			else
				last.next = e.next;

			del = e;
			e = e.next;
			del.recycle();  // watch out - recycle mutates the 'next' field
		}
		else {
			last = e;
			e = e.next;
		}
	}
}

synchronized void dropPaintEvents ( Object source, int x, int y, int w, int h ) {
	AWTEvent e = localQueue;
	AWTEvent del, last = null;

	while ( e != null ) {
		if ( (e.id == PaintEvent.UPDATE) && e.isObsoletePaint( source, x, y, w, h) ) {
			if ( localEnd == e )
				localEnd = last;

	 		if ( last == null )
				localQueue = e.next;
			else
				last.next = e.next;

			del = e;
			e = e.next;
			del.recycle();  // watch out - recycle mutates the 'next' field
		}
		else {
			last = e;
			e = e.next;
		}
	}
}

synchronized void dropPendingEvents ( Component src, int id ) {
	AWTEvent e = localQueue;
	AWTEvent del, last = null;

	while ( e != null ) {
		if ( ((id != 0) && (e.id != id)) ||
		     ((src != null) && (e.getSource() != src)) ) {
			last = e;
			e = e.next;
			continue;
		}
	
		if ( localEnd == e )
			localEnd = last;

 		if ( last == null )
			localQueue = e.next;
		else
			last.next = e.next;

		del = e;
		e = e.next;
		del.recycle();  // watch out - recycle mutates the 'next' field
	}
}

public AWTEvent getNextEvent () {
	AWTEvent e;

	if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) != 0 ) {
		// We only have one source of Java events - our localQueue (every
		// native event is posted to this queue by the native layer). If
		// it is empty, we can't do anything else than go sleeping
		synchronized ( this) {
			// block until we get something in
			while ( localQueue == null ) {
				try {
					wait();
				} catch ( InterruptedException xx ) {}
			}

			e = localQueue;
			localQueue = e.next;
			e.next = null;
			if ( e == localEnd )
				localEnd = null; // just to avoid a temp mem leak

			return e;
		}
	}
	else {
		// Slightly more complex - we have two sources of events: our local queue
		// and some native mechanism accessed via evtGetNextEvent(). We start with
		// our local queue, but don't block in case it is empty. Blocking (if any) is done
		// in evtGetNextEvent(), which means we have to create some native event traffic
		// (getting back as a null event) in case we subsequently post a event to
		// the localQueue
		while ( true ) {
			synchronized ( this ) {
				if ( localQueue != null ) {
					e = localQueue;
					localQueue = e.next;
					e.next = null;
					if ( e == localEnd )
						localEnd = null; // just to avoid a temp mem leak

					return e;
				}
			}

			// this is the sync point in case we have a blocking AWT (suspending
			// the dispatcher thread until the next event becomes available)
			if ( (e = Toolkit.evtGetNextEvent()) != null ) {
				e.next = null;
				
				if ( deferredLocalQueue != null ){
					Toolkit.evtWakeup(); // make sure the next evtGetNextEvent doesn't block
				}
				return e;
			}
			else {
				// this didn't block because there was a pending wakeUp() - check if
				// we have something in the deferredLocalQueue
				synchronized ( this ){
					if ( deferredLocalQueue != null ){
						e = deferredLocalQueue;
						deferredLocalQueue = e.next;
						e.next = null;
						return e;
					}
				}				
			}

			// we don't have to check Toolkit.IS_BLOCKING here, since we reach
			// this point only in case it is not blocked, or evtGetNextEvent()
			// returned 'null'
			
			// if we got here, we either have a non-blocking AWT, and there are no
			// more native events (which means we should sleep to give other threads a chance),
			// or we have a pending wakeUp/local event for a blocking AWT (which means we
			// shouldn't sleep)
			if ( (localQueue == null) && !AWTEvent.accelHint ) {
				try {
					Thread.sleep( Defaults.EventPollingRate);
				} catch ( InterruptedException xx ) {}
			}
		}
	}
}

synchronized boolean hasPendingEvents ( Component c, int id ) {
	AWTEvent e;

	for ( e=localQueue; e != null; e = e.next ) {
		if ( (c != null) && (c != e.getSource()) )
			continue;
	
		if ( (id != 0) && (id != e.id) )
			continue;
	
		return true;
	}

	return false;
}

public synchronized AWTEvent peekEvent () {
	if ( localQueue != null ){
		return localQueue;
	}
	else if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) == 0 ) {
		return Toolkit.evtPeekEvent();
	}

	return null;
}

public synchronized AWTEvent peekEvent ( int id ) {
	for ( AWTEvent e=localQueue; e != null; e = e.next ) {
		if ( e.id == id )
			return e;
	}
	
	if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) == 0 )
		return Toolkit.evtPeekEventId( id);
		
	return null;
}

synchronized boolean postAdjustmentEvent ( int id, Adjustable c, int op, int val ) {
	AWTEvent e = deferredLocalQueue;

	// again, it's redundant, but there is no real generic way to solicit events
	// w/o turning the posted one into a new event, upfront (to avoid the specific
	// parameters). Maybe we will do that if more come along.
	if ( e != null ) {
		do {
			if ( (e instanceof AdjustmentEvt) && 
			     ((AdjustmentEvt)e).solicitAdjustment( id, c, op, val) ){
				return false;
			}
			if ( e.next == null ) {
				break;
			}
			else {
				e = e.next;
			}
		} while ( true );

		e.next = AdjustmentEvt.getEvent( c, id, op, val);
	}
	else {
		deferredLocalQueue = AdjustmentEvt.getEvent( c, id, op, val);

		if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) != 0 ) {
			notify();  // wake up any waiter
		}
		else if ( (Toolkit.flags & Toolkit.IS_BLOCKING) != 0 ){
			Toolkit.evtWakeup();
		}
	}

	return true;
}

public synchronized void postEvent ( AWTEvent e ) {
	if ( localQueue == null ) {
		localQueue = localEnd = e;

		if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) != 0 ) {
			notify();  // wake up any waiter
		}
		else {
			// If we use blocked IO, and this is not the eventThread, wake it up by creating
			// some IO traffic. No need to do that if we have a native dispatcher loop
			if ( ((Toolkit.flags & Toolkit.IS_BLOCKING) != 0)
		        && (Thread.currentThread() != Toolkit.eventThread) ){
				Toolkit.evtWakeup();
			}
		}
	}
	else {
		localEnd.next = e;
		localEnd = e;
				
		// there is no need to wakeup the eventThread, since local events are
		// always processed *before* blocking on a native event inquiry (and
		// the localQueue isn't empty)
	}
}

void postFocusEvent ( FocusEvent evt ) {
	if ( evt.id == FocusEvent.FOCUS_GAINED ) {
		dropPendingEvents( null, FocusEvent.FOCUS_GAINED);
	}
	
	postEvent( evt);
}

synchronized boolean postPaintEvent ( int id, Component c, int x, int y, int width, int height ) {
	AWTEvent e = localQueue;

	// PaintEvents are mapped to the next native-like component by means of
	// recursive repaint() calls, so we don't have to do that here
	
	// OK, this is pretty redundant to postEvent, but we don't want to
	// scan the localQueue twice (it might get large)
	if ( e != null ) {
		do {
			if ( (e instanceof PaintEvt) && 
			     ((PaintEvt)e).solicitRepaint( id, c, x, y, width, height) ){
				return false;
			}
			if ( e.next == null ) {
				break;
			}
			else {
				e = e.next;
			}
		} while ( true );

		e.next = localEnd = PaintEvt.getEvent( c, id, 0, x, y, width, height);
	}
	else {
		localQueue = localEnd = PaintEvt.getEvent( c, id, 0, x, y, width, height);

		if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) != 0 ) {
			notify();  // wake up any waiter
		}
		else if ( ((Toolkit.flags & Toolkit.IS_BLOCKING) != 0) &&
		     (Thread.currentThread() != Toolkit.eventThread) ) {
			Toolkit.evtWakeup();
		}
	}

        return true;
}

synchronized boolean processEvents ( long timeout ) {
	AWTEvent next;

	// Loop until we have no more events in the queue. Make sure we don't loop forever
	// if we've got some event processing that casues new events

	long tend = System.currentTimeMillis() + timeout;

	do {
		while ( localQueue != null ) {
			// dispatch might recycle the thing, so remember the next
			next = localQueue.next;
			localQueue.dispatch();
			localQueue = next;
			
			if ( System.currentTimeMillis() > tend ){
				return false;
			}
		}

		while ( Toolkit.evtPeekEvent() != null ) {
			Toolkit.evtGetNextEvent().dispatch();

			if ( System.currentTimeMillis() > tend ){
				return false;
			}		
		}
	} while ( (localQueue != null) || (Toolkit.evtPeekEvent() != null) );

	return true;
}

synchronized void processPaintEvents ( ) {
	AWTEvent e = localQueue;
	AWTEvent del, last = null;

	while ( e != null ) {
		if ( (e.id == PaintEvent.UPDATE) || (e.id == PaintEvent.PAINT) ) {
			if ( localEnd == e )
				localEnd = last;

	 		if ( last == null )
				localQueue = e.next;
			else
				last.next = e.next;

			del = e;
			e = e.next;
			del.dispatch();
		}
		else {
			last = e;
			e = e.next;
		}
	}
}
}
