package java.awt;

import java.util.Hashtable;
import kaffe.awt.Defaults;
import kaffe.util.Ptr;

/**
 * Component - abstract root of all widgets
 *
 * Copyright (c) 1998
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */
public class FontMetrics
  implements java.io.Serializable
{
	transient Ptr nativeData;
	protected String fontSpec;
	transient int height;
	transient int fixedWidth;
	transient int ascent;
	transient int maxAscent;
	transient int descent;
	transient int maxDescent;
	transient int maxAdvance;
	transient int leading;
	transient boolean isWideFont;
	transient int[] widths;
	static transient Hashtable cache;
	final private static long serialVersionUID = 1681126225205050147L;

FontMetrics ( Font font ) {
	// FontMetrics objects can be quite expensive to initialize, which is why
	// we might want to cache them. But if we do that, and keep a reference to the
	// generating font, we might waste a lot of memory if the number of actually
	// used fonts is small. It's rather unusual to query fonts via fontmetrics
	// (which we slow down here)
	fontSpec = font.encode();

	// this initializes all scalar int/boolean fields, but we defer
	// widths[] initialization until it's explicitly requested
	nativeData = Toolkit.fntInitFontMetrics( font.nativeData, this);
}

public int bytesWidth ( byte data[], int off, int len ) {
	if ( fixedWidth != 0 ) {
		return len*fixedWidth;
	}
	else if ( !isWideFont ) {
		int i, w, n=off+len;
		if ( widths == null ) {
			widths = Toolkit.fntGetWidths( nativeData);
		}
		try {
			for ( i=off, w=0; i<n; i++ )
				w += widths[data[i]];
		}
		catch ( ArrayIndexOutOfBoundsException x ) { return 0; }
		return w;
	}
	else {
		return Toolkit.fntBytesWidth( nativeData, data, off, len);
	}
}

public int charWidth ( char c ) {
	if ( fixedWidth != 0 )
		return fixedWidth;
	else if ( c < 256 ){
		if ( widths == null ) {
			widths = Toolkit.fntGetWidths( nativeData);
		}
		return widths[c];
	}
	else
		return Toolkit.fntCharWidth( nativeData, c);
}

public int charWidth ( int c ) {
	return charWidth( (char) c);
}

public int charsWidth ( char data[], int off, int len ) {
	if ( fixedWidth != 0 ) {
		return len*fixedWidth;
	}
	else if ( !isWideFont ) {
		int i, w, n=off+len;
		if ( widths == null ) {
			widths = Toolkit.fntGetWidths( nativeData);
		}
		try {
			for ( i=off, w=0; i<n; i++ )
				w += widths[data[i]];
		}
		catch ( ArrayIndexOutOfBoundsException x ) { return 0; }
		return w;
	}
	else {
		return Toolkit.fntCharsWidth( nativeData, data, off, len);
	}
}

protected void finalize () throws Throwable {
	if ( nativeData != null ) {
		Toolkit.fntFreeFontMetrics( nativeData);
		nativeData = null;
	}
	super.finalize();
}

public int getAscent() {
	return ascent;
}

public int getDescent() {
	return descent;
}

public Font getFont() {
	return Font.decode( fontSpec);
}

static FontMetrics getFontMetrics ( Font font ) {
	FontMetrics metrics = null;
	if (Defaults.CacheFontMetrics) { 
		String key = font.encode ();
		if ( cache == null ) {
			cache = new Hashtable(20);
		}
		if ((metrics = (FontMetrics) cache.get (key)) == null) {
			// cache missed: create new metrics & install in cache
			cache.put (key, (metrics = new FontMetrics (font)));
		}
	} 
	else {
		// not caching at all
		metrics = new FontMetrics (font);
	}

	return metrics;
}

public int getHeight() {
	return height;
}

public int getLeading() {
	return leading;
}

public int getMaxAdvance() {
	return maxAdvance;
}

public int getMaxAscent() {
	return maxAscent;
}

/**
 * @deprecated, use getMaxDescent()
 */
public int getMaxDecent() {
	return (getMaxDescent());
}

public int getMaxDescent() {
	return maxDescent;
}

public int[] getWidths () {
	if ( widths == null ) {
		widths = Toolkit.fntGetWidths( nativeData);
	}

	return widths;
}

public int stringWidth ( String s ) {
	return Toolkit.fntStringWidth( nativeData, s);
}

public String toString () {
	return getClass().getName() +
	" [" + fontSpec + ']';
}
}
