package java.awt;

import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.IndexColorModel;
import java.awt.image.MemoryImageSource;
import java.io.File;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import kaffe.awt.Defaults;
import kaffe.awt.JavaColorModel;
import kaffe.net.URLConnectionWrapper;
import kaffe.util.Ptr;
import kaffe.util.VectorSnapshot;

/**
 * Copyright (c) 1998
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */
public class Image
{
	Ptr nativeData;
	int width = -1;
	int height = -1;
	ImageProducer producer;
	Object observers;
	int flags;
	Hashtable props;
	final public static int SCALE_DEFAULT = 1;
	final public static int SCALE_FAST = 2;
	final public static int SCALE_SMOOTH = 4;
	final public static int SCALE_REPLICATE = 8;
	final public static int SCALE_AREA_AVERAGING = 16;
	final static int MASK = ImageObserver.WIDTH|ImageObserver.HEIGHT|
			ImageObserver.PROPERTIES|ImageObserver.SOMEBITS|
			ImageObserver.FRAMEBITS|ImageObserver.ALLBITS|
			ImageObserver.ERROR|ImageObserver.ABORT;
	final static int PRODUCING = 0x200;
	final static int READY = ImageObserver.WIDTH | ImageObserver.HEIGHT | ImageObserver.ALLBITS | PRODUCING;
	final static int SCREEN = 0x400;
	final static int BLOCK_FRAMELOADER = 0x800;
	final static int IS_ANIMATION = 0x1000;
	final static int IS_SYNC_PRODUCED = 0x2000;
	final static int IS_DEFERRED_PRODUCED = 0x4000;
	final static int SUSPENDED = 0x8000;
	final public static Object UndefinedProperty = ImageLoader.class;
	static Image unknownImage;

Image ( File file ) {
	producer = new NativeFileProducer( file, this);

	if ( file.exists() && (file.length() < Defaults.FileImageThreshold) ){
		produceSync();
	}
}

Image ( Image img, int w, int h ) {
	nativeData = Toolkit.imgCreateScaledImage(img.nativeData, w, h);
	width = w;
	height = h;
	flags = READY;
}

Image ( ImageProducer ip ) {
	producer = ip;

	if ( ip instanceof MemoryImageSource ) {
		// JDK doesn't produce MemoryImageSources sync, but it does fancy things
		// with the production thread priority (starting "fast" to get dimensions back
		// as soon as possible, then bumping it down below NORM, to put the expensive
		// data processing into the background). Since thread priorities are just a
		// hint, this is leading us nowhere if the thread implementation is different to
		// the JDK (green?) threads. If the app implicitly relies on the production
		// timing, we consider this to be wrong (there are explicit mechanisms like
		// MediaTracker and checkImage). javanoid is a bad app in that sense.
		flags |= IS_DEFERRED_PRODUCED;
	}
	else if ( ip instanceof URLConnectionWrapper ){
		// this is a workaround for the java.net design problem of
		// returning ImageProducers from url.getContent(). We don't
		// want to couple java.net to all our internal AWT knowledge,
		// but we also can't open a URL connection twice for every
		// image to produce via URL getContent (too slow)
		producer = new NativeURLProducer( ((URLConnectionWrapper)ip).getConnection(), this);
	}
	else {
		// it's probably a PiggybackProducer, but we don't care - it's async produced
		// via ImageLoader and setPixels()
	}
}

Image ( URL url ) {
	String protocol = url.getProtocol();

	if ( protocol.equals( "file") ){
		// If we know upfront that it's a file, produce it from the file
		// (URLs are async and more expensive). This might not just be a "file:/.."
		// URL, but also one of these "jar:file:/.." monsters
		File file = new File( url.getFile());
		producer = new NativeFileProducer( file, this);
		
		if ( file.exists() && (file.length() < Defaults.FileImageThreshold) ){
			produceSync();
		}
	}
	else if ( protocol.equals( "jar") && url.getFile().startsWith( "file:") ){
		// this is a file which is in a jar, we can produce that sync if it's not
		// too big (the "subprotocol" 'file:' means it is in the filesystem) 
		try {
			URLConnection conn = url.openConnection();
			int sz = conn.getContentLength();
			producer = new NativeURLProducer( conn, this);
			
			if ( sz < Defaults.FileImageThreshold ){
				produceSync();
			}
		}
		catch ( Exception x ) {
			producer = new NativeURLProducer( url, this);
		}
	}
	else if ( protocol.equals( "tv")) {		
		producer = new TVDummyProducer( url, this);
		produceSync();
	}
	else {
		producer = new NativeURLProducer( url, this);
	}
}

Image ( byte[] data, int offs, int len) {
	producer = new NativeDataProducer( data, offs, len, this);

	// this should be pretty safe, data is avail, and it can't be animation	
	if  ( len < Defaults.DataImageThreshold ){
		produceSync();
	}
}

Image ( int w, int h ) {
	nativeData = Toolkit.imgCreateScreenImage( w, h );
	
	width = w;
	height = h;
	flags = READY | SCREEN;
}

// -------------------------------------------------------------------------
synchronized void addObserver ( ImageObserver observer ) {
	if ( observer == null ) {
		return;
	}

	if ( observers == null ) {
		observers = observer;
	}	
	else if ( observers instanceof Vector ) {
		Vector v = (Vector) observers;
		if ( !v.contains(observer) )
			v.addElement(observer);
	}
	else {
		if ( observer != observers ) {
			Vector v = new Vector( 3);
			v.addElement( observers);
			v.addElement( observer);
			observers = v;
		}
	}
}

synchronized int checkImage ( int w, int h, ImageObserver obs, boolean load ){
	if ( (flags & (ImageObserver.ALLBITS | ImageObserver.FRAMEBITS)) != 0 ) {
		if ( ((w > 0) || (h > 0)) && (w != width) && (h != height) ){
			scale( w, h);
		}
		
		if ( (obs != null) && ((flags & ImageObserver.FRAMEBITS) != 0) ){
			addObserver( obs);
		}
	}
	else if ( load && ((flags & PRODUCING) == 0) ) {
		loadImage( w, h, obs);
	}
	else if ( obs != null ) {
		addObserver( obs);
	}

	return (flags & MASK);
}

protected void finalize () throws Throwable {
	flush();
	super.finalize();
}

public synchronized void flush () {
	// This is the bad guy - the spec demands that it reverts us into a "not-yet-loaded"
	// state, "to be recreated or fetched again from its source" when a subsequent
	// rendering takes place. Since "its source" doesn't mean the *original* source
	// (see getSource() - "an object that *produces* the image"), we could stay with the
	// native representation (probably the most efficient storage and prod environment),
	// but apps explicitly using flush most probably use it because the original
	// image source has changed, i.e. they use flush() as a way to in-situ modify
	// otherwise constant images (by means of changed sources). Only heaven knows why
	// they don't replace the image (this is like a Graphics that can be brought back
	// to life after a dispose()). That means we have to keep a ref of byte[] data, and
	// we can't produce on-screen images

	if ( nativeData != null){
		Toolkit.imgFreeImage( nativeData);
		nativeData = null;
	}
	
	flags = 0;
	width = -1;
	height = -1;
}

public Graphics getGraphics () {
	if ((flags & SCREEN) == 0 || nativeData == null) {
		return null;
	}
	else {
		return NativeGraphics.getGraphics( this,
		                   nativeData, NativeGraphics.TGT_TYPE_IMAGE, true,
		                   0, 0, 0, 0, width, height,
		                   Color.black, Color.white, Defaults.WndFont);
	}
}

public synchronized int getHeight ( ImageObserver observer ) {
	if ( (flags & ImageObserver.HEIGHT) != 0 ) {
		return height;
	}
	else {
		loadImage( -1, -1, observer);
		return (-1);
	}
}

public Object getProperty ( String name, ImageObserver observer ) {
	if ( props == null ) {
		if ( (flags & (ImageObserver.FRAMEBITS | ImageObserver.ALLBITS)) != 0 )
			return UndefinedProperty;
		else if ( (flags & PRODUCING) == 0 )
			loadImage( -1, -1, observer);

		return null;
	}
	else {
		Object val = props.get( name);
		if ( val == null )
			return UndefinedProperty;
			
		return val;
	}
}

public Image getScaledInstance (int width, int height, int hints) {

	if ( (flags & (ImageObserver.ERROR|ImageObserver.FRAMEBITS)) == 0 ){
		MediaTracker mt = new MediaTracker( null);
		mt.addImage( this, 0);
		try {
			mt.waitForID( 0);
		}
		catch ( InterruptedException x ){
			return null;
		}
	}

	if ( (flags & (ImageObserver.ERROR|ImageObserver.ABORT)) != 0 ){
		return null;
	}

	if ( (width < 0) && (height < 0) ){
		throw new AWTError( "illegal image scaling request: , w: "
		                        + width + ", h: " + height);
	}

	// a negative value is computed from the other dimension and
	// a constant aspect ratio
	if ( width < 0 ){
		width = (int) (((float)height / this.height) * this.width);
	}
	else if ( height < 0 ){
		height = (int) (((float)width / this.width) * this.height);
	}

	return (new Image(this, width, height));
}

public ImageProducer getSource () {
/**
	if ( nativeData != null ){
		// that's way faster than URL or File
		return new NativeSelfProducer( this);
	}
**/
	// we can't give out our real NativeImageProducer, since this is a very
	// specialized thing which only deals with ImageLoader consumers (and one
	// fixed target Image)	
	if ( producer instanceof NativeImageProducer ){
		return new PiggybackProducer( (NativeImageProducer) producer);
	}
	else if ( producer instanceof PiggybackProducer ){
		return new PiggybackProducer( ((PiggybackProducer)producer).refProducer);
	}
	else {
		return producer;
	}
}

static Image getUnknownImage (){
	if ( unknownImage == null ){
		int d = Toolkit.screenSize.height / 64;

		unknownImage = new Image( d, d);

		Graphics g = unknownImage.getGraphics();
		if ( g != null ){
			g.setColor( Color.white);
			g.fillRect( 0, 0, d, d);
			g.setColor( Color.red);
			g.draw3DRect( 0, 0, d, d, false);
			
			g.dispose();
		}
		
		// maybe we have a tracker waiting on this - make sure it flags the app
		// that it didn't find the Image
		unknownImage.flags |= (ImageObserver.ERROR|ImageObserver.ABORT);
	}
	
	return unknownImage;
}

public synchronized int getWidth ( ImageObserver observer ) {
	if ( (flags & ImageObserver.WIDTH) != 0 ) {
		return (width);
	}
	else {
		loadImage( -1, -1, observer);
		return (-1);
	}
}

boolean loadImage ( int w, int h, ImageObserver obs ) {
	if ( (flags & (ImageObserver.FRAMEBITS|ImageObserver.ALLBITS)) != 0 ){
		// it's (partly?) loaded already, but do we have to scale it?
		if ( ((w > 0) || (h > 0)) && (w != width) && (h != height) ) {
			scale( w, h);
		}

		if ( (flags & PRODUCING) != 0 ){		
			if ( (flags & ImageObserver.FRAMEBITS) != 0 ){
				// if it's a already producing animation, register the new observer
				if ( obs != null ){
					addObserver( obs);
				}
			}
			return true;
		}
	}
	else if ( (flags & ImageObserver.ABORT) != 0 ) {
	  // we ultimately failed
		return false;
	}
	
	// it's either not in production yet, or not yet finished, so
	// we have to add another observer
	if ( obs != null ) {
		addObserver( obs);
	}
	
	if ( (flags & PRODUCING) == 0 ) {
		// has to be produced or animated, kick it off
		flags |= PRODUCING;
		if ((flags & IS_DEFERRED_PRODUCED) != 0) {
			produceSync();			
		} else {
			ImageLoader.load( this);
		}
	}
	
	return ((flags & (ImageObserver.FRAMEBITS|ImageObserver.ALLBITS)) != 0);
}

void produceSync () {
	flags |= (Image.PRODUCING|IS_SYNC_PRODUCED);
	ImageLoader ld = new ImageLoader();
	ld.loadSync( this);
}

synchronized void removeAllObservers () {
	observers = null;
}

synchronized void removeObserver ( ImageObserver observer ) {
	if ( (observers == null) || (observer == null) )
		return;

	if ( observers == observer ){
		observers = null;
	}
	else if ( observers instanceof Vector ) {
		Vector v = (Vector)observers;
		v.removeElement( observer);
		
		if ( v.size() == 1 )
			observers = v.firstElement();
	}
}

private boolean scale (int w, int h) {
	if ((flags & SCREEN) != 0 || (flags & ImageObserver.ALLBITS) == 0 || nativeData == null) {
		return (false);
	}
	if (w != width || h != height) {
		Ptr oldNativeData = nativeData;
		nativeData = Toolkit.imgCreateScaledImage( nativeData, w, h );
		Toolkit.imgFreeImage( oldNativeData);
		width = w;
		height = h;
	}
	return (true);
}

void stateChange(int flags, int x, int y, int w, int h) {
	ImageObserver obs = null;
	Enumeration e = null;

	synchronized ( this ) {
		this.flags |= flags;
		
		if (observers == null) {
			return;
		}

		if ( observers instanceof Vector ){
			// we can't use a (index-based) standard Vector enumerator because we have
			// to notify *all* elements regardless of any removals
			e = VectorSnapshot.getCached( (Vector)observers);
		}
		else if ( observers instanceof ImageObserver ) {
			obs = (ImageObserver) observers;
		}
	}

	// Since we notify sync from the production thread, make sure that imageUpdate
	// is called outside the synchronized block (observers live in user land, if
	// they block - for whatever reason - we shouldn't leave the image locked)
	
	// If we get false returned by an imageUpdate, this indicates the observer is no longer
	// interested in updates. This is *NOT* what is said in the Addison-Wesley
	// documentation, but is what is says in the JDK javadoc documentation.	
	if ( obs != null ){
		if ( obs.imageUpdate( this, flags, x, y, w, h) == false ) {
			removeObserver( obs);
		}
	}
	else if ( e != null ) {
		for ( ; e.hasMoreElements(); ) {
			obs = (ImageObserver) e.nextElement();
			if ( obs.imageUpdate( this, flags, x, y, w, h) == false ){
				removeObserver( obs);
			}
		}
	}
}

synchronized void stopAnimation () {
	if ( (flags & (IS_ANIMATION | PRODUCING)) == (IS_ANIMATION | PRODUCING) ){
		removeAllObservers();
		flags &= ~PRODUCING;
		if ( (flags & SUSPENDED) != 0 ){
			flags &= ~SUSPENDED;
			synchronized ( ImageLoader.animLock ){
				ImageLoader.animLock.notifyAll();
			}
		}
	}
}

synchronized void stopProduction () {
	if ( (flags & PRODUCING) != 0 ){
		if ( (flags & IS_ANIMATION) != 0 ){
			stopAnimation();
		}
		else {
			removeAllObservers();
			
			if ( producer instanceof NativeImageProducer ){
				// don't depend on instant reaction - if the production is inside
				// a Socket ctor, this might take a loong time to run into the
				// system specific timeout
				((NativeImageProducer)producer).stopProduction();
			}
		}
		
		flush();
		
		// not clear if we should keep ABORT - it could block susequent load requests
		flags = ImageObserver.ABORT;
	}
}

public String toString() {
	String s = getClass().getName() + " [" + width + ',' + height + 
	              ((nativeData != null) ? ", native" : ", non-native") +
                ", flags: 0x"  + Integer.toHexString( flags);

	if ( (flags & PRODUCING) != 0 )
		s += " producing";
	if ( (flags & ImageObserver.ALLBITS) != 0 )
		s += " allbits";
	else if ( (flags & ImageObserver.FRAMEBITS) != 0 )
		s += " framebits";
	else if ( (flags & ImageObserver.ERROR) != 0 )
		s += " error";

	s += "]";

	return s;		
}
}
