/**
 * Copyright (c) 1999
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */

package java.awt;

import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.IndexColorModel;
import java.awt.image.MemoryImageSource;
import java.util.Hashtable;
import kaffe.awt.Defaults;
import kaffe.awt.JavaColorModel;
import kaffe.util.Ptr;
import kaffe.util.Queue;
import kaffe.util.Timer;
import kaffe.util.TimerClient;

/**
 * Copyright (c) 1999
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */
class ImageLoader
  implements Runnable, ImageConsumer, TimerClient
{
	Image img;
	int[] data = null;
	boolean isActive = false;
	boolean isAsync = false;
	static Queue queue = new Queue();
	static int idleTime = 3000;
	static int active;
	static int notifying;
	static int deferred;
	static ImageLoader checker;
	static int ilcount = 1;
	final static int PRIO_START = Thread.NORM_PRIORITY + 3;
	final static int PRIO_DATA = Thread.NORM_PRIORITY - 2;
	static Object animLock = new Object();

static void checkQueue ( boolean defer ) {

	if ( queue.isEmpty() ){
		// nothing to do at all
		return;
	}

	if ( active > 0 ){
		// if we still have at least one active loader, this one will check the queue before
		// it (maybe) blocks or animates
		return;
	}

	if ( defer ){
		if ( notifying > 0 ){
			// There is a chance that a notifying loader soon will come back, and then we shouldn't
			// spawn a new one. On the other hand, we have to take provisions in case a
			// ImageObserver blocks, which shouldn't affect other image productions

			// we could use the eventdispatcher to timeout wait, but that would block event processing
			// (might be even good for deferred repaints, but that depends on the app)
			if ( checker != null ){
				deferred = 1;
				Timer.getDefaultTimer().addClient( checker, 100, 0);
				return;
			}
		}
	}

	// Ok, nobody is available, we have to spawn a new loader. Note that we increment
	// the active count here, so that subsequent checks don't have to wait until the
	// beast actually runs
	active++;
	ImageLoader ld = new ImageLoader();
	if ( checker == null ){
		checker = ld;
	}
	ld.isActive = true;
	ld.isAsync = true;

	Thread t = new Thread( ld, "AWT-ImageLoader-" + ilcount++);
	t.setPriority( PRIO_START);
	t.start();
}

public void imageComplete ( int status ) {
	// producer callback - this is the one we get from every producer

	int s = 0;
	int latency;
	data = null;

	// note that the ImageConsumer status values are not bit flags!

	if ( status == STATICIMAGEDONE ){
		s = ImageObserver.ALLBITS;
	}
	else if ( status == SINGLEFRAMEDONE ) {
		img.flags |= Image.IS_ANIMATION;
		s = ImageObserver.FRAMEBITS;
	}
	else {
		if ( (status & IMAGEERROR) != 0 )       s |= ImageObserver.ERROR;
		if ( (status & IMAGEABORTED) != 0 )     s |= ImageObserver.ABORT;
	}

	// give native layer a chance to do some solicitation (alpha channel reduction etc.)
	if ( !(img.producer instanceof NativeImageProducer) && img.nativeData != null) {
		Toolkit.imgComplete( img.nativeData, status);
	}

	imageStateChange( img, s, 0, 0, img.width, img.height);

	// if it's an animation, this isn't the last imageComplete
	if ( status == SINGLEFRAMEDONE ){
		if ( img.producer instanceof NativeImageProducer ){
			// if it's a NativeImageProducer, all frames are already processed, and we
			// simply loop through them until the last observer is gone
			// make sure we don't loop if this is sync (we are still in the img ctor)
			if ( (img.flags & Image.IS_SYNC_PRODUCED) == 0 ){
				runAnimation();
			}
			else {
				// make sure this thing subsequently gets scheduled again
				img.flags &= ~Image.PRODUCING;
			}
		}
		else if ( img.producer instanceof MemoryImageSource ) {
			// We are not running in our own thread, MemoryImageSources
			// are sync produced. Hence there is no observer yet if this is
			// the first frame. Make sure we don't pull ourselves as a consumer (this is
			// a producer private loader), because that would stop MemoryImageSource
			// notification
		}
		else if ( img.observers == null ){
			// not sure about this one - arbitrary animation producers tend to continue
			// sending us imageCompletes for each frame automatically (i.e. they do the
			// looping by themselves). But how do we tell them to stop w/o the
			// sledgehammer approach of throwing an exception. The best guess seems to
			// be pulling the consumer
			
			img.producer.removeConsumer( this);
			
			if ( !isActive ){
				resumeActive();
			}
		}
		else {
			suspendActive();
		}
	}
}

void imageStateChange ( Image img, int flags, int x, int y, int w, int h ){
	active--;
	isActive = false;
	notifying++;
	
	checkQueue( true);
	img.stateChange( flags, x, y, w, h);

	notifying--;
	isActive = true;
	active++;
}

static void load ( Image img ) {
	queue.put( img);	
	checkQueue( true);
}

void loadSync ( Image img ) {
	this.img = img;
	img.producer.startProduction( this);
	this.img = null;
}

void resumeActive () {
	active++;
	isActive = true;
}

public void run () {
	while ( (img = (Image)queue.get( idleTime)) != null ){
		if ( (img.flags & Image.PRODUCING) != 0 ){
			try {
				// This will give us at least one imageComplete() callback.
				// Note that we don't get out here for animations (until the animation
				// is stopped, which seems to be done when the last observer goes away)	

				if ( (img.flags & Image.IS_ANIMATION) != 0 ){
					// this was already sync produced, just loop around
					runAnimation();
				}
				else {
					if ( img.nativeData == null ){
						// note that this might as well end up in runAnimation (called
						// from imageComplete), since it runs through the whole
						// Consumer callback chain
						img.producer.startProduction( this);
					}
				}
			}
			catch ( Exception x ) {
				// we don't want to fall out of the production thread
				x.printStackTrace();
			}
		}
		
		img = null;   // avoid a temp memory leak
		
		// bump it back to start priority
		Thread.currentThread().setPriority( PRIO_START);
	}

	if ( this == checker ){
		checker = null;
	}

	active--;
	isActive = false;
}

void runAnimation () {
	int latency;
	int nLoops;
	Ptr firstFrame = img.nativeData;
	Ptr nextFrame;

	// while we are running in an animation, there's no use waiting
	// for this thread to become available for new productions
	suspendActive();

	nLoops = Toolkit.imgGetLoopCount( img.nativeData);

outer:
	while ( img.observers != null ){
	
		// We might have to temporarily suspend the production in case the animated
		// image was scrolled out of the visible area.
		// Note that we shouldn't sync on any object which is visible in user land, or
		// we risk deadlocks (in fact, there is a chance that apps sync on imageUpdate/paint
		// which is why we have to Defaults parameterize it)
		while ( Defaults.SuspendAnimations && ((img.flags & Image.SUSPENDED) != 0) ){
			try {
				synchronized ( animLock ){
					animLock.wait();
					if ( img.observers == null ){ // somebody stopped us
						continue outer;
					}
				}
			}
			catch ( InterruptedException xx ) {}
		}
	
		latency = Toolkit.imgGetLatency( img.nativeData);
		if ( latency == 0 ) latency = Defaults.FrameLatency;
		try {
			Thread.sleep( latency);
		} catch ( InterruptedException x ) {}
		
		if ( img.producer instanceof NativeImageProducer ) {
			// we already have all frames in the native image, switch to the next one
			
			nextFrame = Toolkit.imgGetNextFrame( img.nativeData);
			
			if ( nextFrame == firstFrame ){
				if ( nLoops-- == 0 ){
					break;
				}
			}
			
			img.nativeData = nextFrame;
			img.stateChange( ImageObserver.FRAMEBITS, 0, 0, img.width, img.height);
		}
		else {
			// Outch, we shouldn't get here for app animation producers
			throw new AWTError( "unsupported animation producer");
		}
	}

	// JDK resets the flags to 'non-produced' when a animation is stopped, but it doesn't
	// propagate this change (makes sense, we stop if there is no observer, anymore)
	img.flags = (ImageObserver.WIDTH|ImageObserver.HEIGHT|ImageObserver.PROPERTIES|Image.IS_ANIMATION);	
	resumeActive();
}

public void setColorModel ( ColorModel model ) {
	// producer callback (in case it's not a NativeProducer)

	// Note that according to the specs, this is just the color model
	// of the "majority" of pixels, i.e. it might change in subsequent
	// setPixel calls.
	// It's therefore not very useful for us
}

public void setDimensions ( int width, int height ) {
	// producer callback (this might be called from NativeProducers
	// and app specific producers)

	img.width = width;
	img.height = height;

	// If we have a NativeImageProducer, the nativeData field will automatically be
	// set when the production is done. For an arbitrary producer, we have to init
	// the native part explicitly, before we get any setPixels()
	if ( (img.nativeData == null) && !(img.producer instanceof NativeImageProducer) ){
		img.nativeData = Toolkit.imgCreateImage( width, height);
	}

	imageStateChange( img, ImageObserver.WIDTH|ImageObserver.HEIGHT, 0, 0, width, height);
	
	if ( isAsync ) {
		Thread.currentThread().setPriority( PRIO_DATA);
	}
}

public void setHints ( int hints ) {
	// producer callback (in case it's not a NativeProducer)
	// we don't honor hints
}

public void setPixels ( int x, int y, int w, int h,
			ColorModel model, byte pixels[], int offset, int scansize ) {
  // producer callback (in case it's not a NativeProducer)
  
  if ( img.nativeData == null ) {
    // producer trouble, we haven't got dimensions yet
    return;
  }
  
  if ( model instanceof IndexColorModel ) {
    IndexColorModel icm = (IndexColorModel) model;
    Toolkit.imgSetIdxPels( img.nativeData, x, y, w, h, icm.rgbs,
			   pixels, icm.trans, offset, scansize);
    imageStateChange( img, ImageObserver.SOMEBITS, x, y, w, h);
  }
  else {
    System.err.println("Unhandled colorModel: " + model.getClass());
  }
}

public void setPixels ( int x, int y, int w, int h,
			ColorModel model, int pixels[], int offset, int scansize ) {
  // producer callback (in case it's not a NativeProducer)
  
  boolean alpha = false;

  if ( img.nativeData == null ) {
    // producer trouble, we haven't got dimensions yet
    return;
  }
  
  if ( model instanceof JavaColorModel ) {
    alpha = true;
    // OK, nothing to convert here
    Toolkit.imgSetRGBPels( img.nativeData, x, y, w, h, pixels, alpha, offset, scansize);
  }
  else {
    // any color model should also be converted.
    // We unfortunately cannot overide the original source. We need to create
    // the new source for the memory.

    if (data == null) data = new int[w*h];
    int xw = x + w;
    int yh = y + h;
    int i, j, idx;
    int i0 = y * scansize + offset;
    for ( j = y; j < yh; j++, i0 += scansize ) {
      for ( i = x, idx = i+i0; i < xw; i++, idx++) {
				// are we allowed to change the array in-situ?
				// Most likely not. there seems to be a problem.
				data[idx] = model.getRGB( pixels[idx]);
      }
    }

    // check for the alpha. If direct we know what it is, otherwise set to
    // true and worry later on.
    if ( model instanceof DirectColorModel ){
			alpha = ((DirectColorModel)model).getAlphaMask() != 0;
		}
    else {
			alpha = true;
		}

    // set the pixels
    Toolkit.imgSetRGBPels( img.nativeData, x, y, w, h, data, alpha, offset, scansize);
  }
  imageStateChange( img, ImageObserver.SOMEBITS, x, y, w, h);
}

public void setProperties ( Hashtable properties ) {
	// producer callback (in case it's not a NativeProducer)

	img.props = properties;
	imageStateChange( img, img.flags | ImageObserver.PROPERTIES, 0, 0, img.width, img.height);
}

void suspendActive () {
	isActive = false;
	active--;
	checkQueue( true);
}

public void timerExpired ( Timer timer ){
	deferred = 0;
//System.out.println( "expired");
	checkQueue( true);
}

public String toString () {
	String s = "ImageLoader " + Integer.toHexString(hashCode());
	
	if ( img != null ){
		s += " [" + Integer.toHexString(img.hashCode()) + ']';
	}
	else {
		s += " [null]";
	}
	
	return s;		
}
}
