/**
 * Copyright (c) 2000
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */

package java.awt;

import java.awt.image.ColorModel;
import java.awt.image.ImageConsumer;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.Vector;
import kaffe.io.AccessibleBIStream;
import kaffe.util.Ptr;
import kaffe.util.VectorSnapshot;

/**
 * This shouldn't be a inner class since you can easily grab the source of an
 * image and use it outside of this image (e.g. to create other images - whatever
 * it might be good for)
 */
abstract class NativeImageProducer
  extends GenericProducer
{
	Image img;
	byte[] buf;
	InputStream in;
	Vector dependents;
	static int BUFSIZE = 4096;

NativeImageProducer () {
}

NativeImageProducer ( Image target ) {
	img = target;
}

boolean addDependent ( PiggybackProducer piggy ) {
	if ( img.nativeData != null ){
		piggy.imageAvailable();
		return false;
	}
	else {
		if ( dependents == null ) {
			dependents = new Vector( 3);
		}
	
		dependents.addElement( piggy);
		return true;
	}
}

int fillBuffer (){
	int n;

	if ( (in == null) || ((img.flags & Image.PRODUCING) == 0) ){
		return 0;
	}

	try {
		if ( in instanceof AccessibleBIStream ){
			n = ((AccessibleBIStream)in).refillBuffer();
		}
		else {
			n = in.read( buf, 0, buf.length);
		}
	}
	catch ( IOException x ){
		return 0;
	}

	return n;
}

void flagProductionFailure () {
	if ( (img.flags & img.PRODUCING) != 0 ){
		imageComplete( ImageConsumer.IMAGEERROR|ImageConsumer.IMAGEABORTED);
	}
	else {
		imageComplete( ImageConsumer.IMAGEABORTED); // deliberately
	}
}

void notifyDependents () {
	int n = dependents.size();
	int i;
	
	for ( i=0; i<n; i++ ){
		((PiggybackProducer)dependents.elementAt( i)).imageAvailable();
	}
	
	dependents = null;
}

abstract void produce ();

void produce ( Ptr ptr ) {
	if ( consumer instanceof ImageLoader ) {
		ImageLoader loader = (ImageLoader)consumer;
		Image img = loader.img;
	
		img.nativeData = ptr;

		if ( dependents != null ){
			notifyDependents();
		}

		loader.imageComplete( (Toolkit.imgGetFrameCount( ptr) > 1) ?
		                           ImageConsumer.SINGLEFRAMEDONE : ImageConsumer.STATICIMAGEDONE);		
	}
	else {
		Toolkit.imgProduceImage( this, ptr, true);
		Toolkit.imgFreeImage( ptr);
	}
}

void produce ( URLConnection conn ) {
	int len, sz, pos, count;

	try {
		if (conn != null) {
			in = conn.getInputStream();
			if ( in != null ) {
				if ( in instanceof AccessibleBIStream ) {
					AccessibleBIStream ais = (AccessibleBIStream) in;
					buf = ais.getBuffer();
					pos = ais.getPos();
					count = ais.getCount();
					if ( count > pos ){
						len = count - pos;
					}
					else {
						len = ais.refillBuffer();
						pos = 0;
					}
				}
				else {
					sz = conn.getContentLength();
					if ( (sz <= 0) || (sz > BUFSIZE) ) {  // it's unknown, let's assume some size
						sz = BUFSIZE;
					}
					buf = new byte[sz];
					pos = 0;
					len = in.read( buf, 0, sz);
				}

				if ( len > 0 ){
					// watch out, this might call back fillBuffer() several times
					Ptr ptr = Toolkit.imgCreateFromData( this, buf, pos, len);

					if ( ptr != null ){
						produce( ptr);
					}
					else {
						flagProductionFailure();
					}
				}
			}
		}
	}
	catch ( Exception x ) {
		flagProductionFailure();
	}
	finally {
		if ( in != null ){
			try {
				in.close();
			}
			catch ( IOException xx ) {
				// nothing we can do about it, just proceed
			}		
			in = null;
		}
	}
}

void removeDependent ( PiggybackProducer dep ){
	if ( dependents != null ){
		dependents.removeElement( dep);
	}
}

public void startProduction ( ImageConsumer ic ) {
	addConsumer( ic);
	produce();
	removeConsumer( ic);
}

void stopProduction (){
	// implemented by subclass if stoppable
}
}
