package java.io;

import java.io.Serializable;
import java.lang.Comparable;
import java.lang.String;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;
import kaffe.lang.Application;
import kaffe.lang.ApplicationResource;
import kaffe.io.FileName;

/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class File
	implements Serializable, Comparable {

final private static long serialVersionUID = 301077366599181567L;
final public static String separator = System.getProperty("file.separator");
final public static char separatorChar = separator.charAt(0);
final public static String pathSeparator = System.getProperty("path.separator");
final public static char pathSeparatorChar = pathSeparator.charAt(0);

final private static Random random = new Random();

private FileName fd;
private ApplicationResource res;

public File(File dir, String name) {
        this((dir == null) ? null : dir.getPath(), name);
}

public File(String path) {
	if (path == null) {
		throw new NullPointerException();
	}
	fd = new FileName(path);
}

public File(String path, String name) {
	if (name == null) {
		throw new NullPointerException();
	}
	if (path == null) {
		path = name;
	}
	else if (path.length() == 0) {
		path = separatorChar + name;
	}
	else {
		char last = path.charAt(path.length() - 1);
		if (last != separatorChar) {
			path += separatorChar + name;
		}
		else {
			path += name;
		}
	}
	fd = new FileName(path);
}

public boolean canRead() {
	checkReadAccess();
	return (fd.checkMode(FileName.CANREAD));
}

public boolean canWrite() {
	checkWriteAccess();
	return (fd.checkMode(FileName.CANWRITE));
}

private void checkReadAccess() {
        SecurityManager sm = System.getSecurityManager();
        if (sm != null)
                sm.checkRead(getPath());
}

private void checkWriteAccess() {
        SecurityManager sm = System.getSecurityManager();
        if (sm != null)
                sm.checkWrite(getPath());
}

// XXX some filesystems are case-insensitive
public int compareTo(File that) {
	return getPath().compareTo(that.getPath());
}

public int compareTo(Object that) {
	return compareTo((File)that);
}

public static File createTempFile(String prefix, String suffix) throws IOException {
	return createTempFile(prefix, suffix, null);
}

public static File createTempFile(String prefix, String suffix, File dir) throws IOException {
	if (prefix.length() < 3) {
		throw new IllegalArgumentException(prefix);
	}
	if (suffix == null) {
		suffix = ".tmp";
	}
	if (dir == null) {
		dir = new File(System.getProperties().getProperty("java.io.tmpdir", "."));
	}
	while (true) {
		File f = new File(dir, prefix + Integer.toHexString(random.nextInt(0x100000)).toUpperCase() + suffix);
		if (f.createNewFile())
			return f;
	}
}

public boolean delete() {
	System.getSecurityManager().checkDelete(getPath());
	if (res != null) {
		Application.removeResource(res);
	}
	if (fd.checkMode(FileName.DIRECTORY)) {
		return (fd.rmdir());
	}
	else {
		return (fd.delete());
	}
}

public void deleteOnExit() {
	res = new ApplicationResource() {
		public void freeResource() {
			File.this.delete();
		}
	};
	Application.addResource(res);
}

// XXX some filesystems are case-insensitive
public boolean equals(Object that) {
	return (that instanceof File) ? getPath().equals(((File)that).getPath()) : false;
}

public boolean exists() {
	checkReadAccess();
	if (fd.checkMode(FileName.FILE) || fd.checkMode(FileName.DIRECTORY)) {
		return (true);
	}
	else {
		return (false);
	}
}

public String getAbsolutePath() {
	if (isAbsolute()) {
		return getPath();
	}
	else {
		return System.getProperty("user.dir") + separatorChar + getPath();
	}
}

public String getCanonicalPath() throws IOException {
        
        StringTokenizer tok = new StringTokenizer(getAbsolutePath(), separator);
        int len = tok.countTokens();
        String[] array = new String[len+2];
        int j = 0;
        for (int i = 0; i < len; i++) {
                String str = tok.nextToken();
                if (str.equals("..")) {
			if (j > 0)
				j--;
                } else if (!str.equals(".")) {
			array[j] = str;
			j++;
                }
        }

	// If the length is 0 then just return the base.
	if (j == 0) {
		return (separator);
	}

        // Build a string of the remaining elements.
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < j; i++) {
		buf.append(separatorChar);
                buf.append(array[i]);
        }
        return (buf.toString());

}

public String getName() {
	String path = getPath();
	return path.substring(path.lastIndexOf(separatorChar)+1);
}

public String getParent() {
	String p = null;
	String path = getPath();
	int slashIndex = path.lastIndexOf(separatorChar);
	if (slashIndex > 0) {
		p = path.substring(0, slashIndex);
	}
	if (slashIndex == 0 && path.length() > 1) {
		p = separator;
	}
	return (p);
}

public String getPath() {
	return fd.getPath();
}

public int hashCode() {
	return getPath().hashCode();
}

public boolean isAbsolute() {
	return (fd.isAbsolute());
}

public boolean isDirectory() {
	checkReadAccess();
	return (fd.checkMode(FileName.DIRECTORY));
}

public boolean isFile() {
	checkReadAccess();
	return (fd.checkMode(FileName.FILE));
}

public long lastModified() {
	checkReadAccess();
	return (fd.lastModified());
}

public long length() {
	checkReadAccess();
	return (fd.length());
}

public String[] list() {
	checkReadAccess();
	if (fd.checkMode(FileName.DIRECTORY)) {
		return (fd.readdir());
	}
	else {
		return (null);
	}
}

public String[] list(FilenameFilter filter) {
	String[] all = list();
	if (all == null || filter == null) {
		return (all);
	}

	Vector filtered = new Vector();
	
	for (int idx = 0; idx < all.length; idx++) {
		String fn = all[idx];
		if (filter.accept(this, fn)) {
			filtered.addElement(fn);
		}
	}

	String[] result = new String[filtered.size()];
	filtered.copyInto((Object [])result);

	return (result);
}

public File[] listFiles() {
	return listFiles((FilenameFilter)null);
}

public File[] listFiles(final FileFilter filter) {
	return listFiles(new FilenameFilter() {
		public boolean accept(File dir, String name) {
			return filter.accept(new File(dir, name));
		}
	});
}

public File[] listFiles(FilenameFilter filter) {
	String[] names = list(filter);
	if (names == null) {
		return null;
	}
	File[] files = new File[names.length];
	for (int idx = 0; idx < names.length; idx++) {
		files[idx] = new File(this, names[idx]);
	}
	return files;
}

public boolean mkdir() {
	checkWriteAccess();
	if (isDirectory()) {
		return (true);
	}
	else {
		return (fd.mkdir());
	}
}

public boolean mkdirs() {
	if (!toString().equals(separator)) {
		String parentString = getParent();
		if (parentString != null) {
			File parent = new File(parentString);
			if (parent.mkdirs() == false) {
				return (false);
			}
		}
	}
	return (mkdir());
}

public boolean renameTo(File dest) {
	System.getSecurityManager().checkWrite(getPath());
	System.getSecurityManager().checkWrite(dest.getPath());
	return (fd.rename(dest.fd));
}

public String toString() {
	return getPath();
}
public boolean createNewFile() {
        checkWriteAccess();
        if (exists()) {
                return (true);
        }
        else {
                return (fd.createNewFile());
        }        
}

}
