/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.lang;

import java.io.InputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import kaffe.lang.SystemClassLoader;

final public class Class
	implements Serializable {

final private static long serialVersionUID = 3206093459760846163L;
final private static int CONSTRUCTOR = 0x0800;
final private static Class[] nullParams = new Class[0];

private Class() { /* this class is not instantiable by the general public */ }

public static Class forName(String className) throws ClassNotFoundException {
	return (forName(className, true, SystemClassLoader.getCallerClassLoader(0)));
}

public static Class forName(String className, boolean initialize, ClassLoader loader) throws ClassNotFoundException {
	if (loader == null) {
		loader = SystemClassLoader.getClassLoader();
	}
	/* We must check to see if this class is loaded before we try to load it again */
	Class cls = loader.getLoadedClass(className, initialize);
	if (cls == null) {
		cls = loader.loadClass(className, initialize);
	}
	return (cls);
}

private String fullResourceName(String name) {
	if (name.charAt(0) == '/') {
		return (name.substring(1));
	}
	String cname = getName();
	StringBuffer buf = new StringBuffer();
	int tail = cname.lastIndexOf('.');
	if (tail != -1) {
		buf.append(cname.substring(0, tail+1).replace('.', '/'));
	}
	buf.append(name);
	return (buf.toString());
}

/**
 * Determines the class loader for the class.
 *
 * @returns   the class loader that created the class or interface
 *            represented by this object, or null if the class was not
 *            created by a class loader.
 * @see       java.lang.ClassLoader
 */
public ClassLoader getClassLoader() {
	ClassLoader loader = getClassLoader0();
	if (loader == null) {
		loader = SystemClassLoader.getClassLoader();
	}
	return (loader);
}

native private ClassLoader getClassLoader0();

public Class[] getClasses() {
	return (getClasses0(false));
}

native private Class[] getClasses0(boolean declared);

native public Class getComponentType();

public Constructor getConstructor(Class parameterTypes[]) throws NoSuchMethodException, SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.PUBLIC);
	if (parameterTypes == null) {
		parameterTypes = nullParams;
	}
	return ((Constructor)getMethod0("<init>", parameterTypes, CONSTRUCTOR|Modifier.PUBLIC, CONSTRUCTOR|Modifier.PUBLIC));
}

public Constructor[] getConstructors() throws SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.PUBLIC);
	return ((Constructor[])getMethods0(CONSTRUCTOR|Modifier.PUBLIC, CONSTRUCTOR|Modifier.PUBLIC));
}

public Class[] getDeclaredClasses() throws SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.DECLARED);
	return (getClasses0(true));
}

public Constructor getDeclaredConstructor(Class parameterTypes[]) throws NoSuchMethodException, SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.DECLARED);
	if (parameterTypes == null) {
		parameterTypes = nullParams;
	}
	return ((Constructor)getMethod0("<init>", parameterTypes, CONSTRUCTOR, CONSTRUCTOR));
}

public Constructor[] getDeclaredConstructors() throws SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.DECLARED);
	return ((Constructor[])getMethods0(CONSTRUCTOR, CONSTRUCTOR));
}

public Field getDeclaredField(String name) throws NoSuchFieldException, SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.DECLARED);
	return (getField0(name, 0, 0));
}

public Field[] getDeclaredFields() throws SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.DECLARED);
	return (getFields0(0, 0));
}

public Method getDeclaredMethod(String name, Class parameterTypes[]) throws NoSuchMethodException, SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.DECLARED);
	if (parameterTypes == null) {
		parameterTypes = nullParams;
	}
	return ((Method)getMethod0(name, parameterTypes, CONSTRUCTOR, 0));
}

public Method[] getDeclaredMethods() throws SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.DECLARED);
	return ((Method[])getMethods0(CONSTRUCTOR, 0));
}

public Class getDeclaringClass() {
	// Not sure what this is about.
	return (null);
}

public Field getField(String name) throws NoSuchFieldException, SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.PUBLIC);
	return (getField0(name, Modifier.PUBLIC, Modifier.PUBLIC));
}

native private Field getField0(String name, int mask, int flags);

public Field[] getFields() throws SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.PUBLIC);
	return (getFields0(Modifier.PUBLIC, Modifier.PUBLIC));
}

native private Field[] getFields0(int mask, int flags);

native public Class[] getInterfaces();

public Method getMethod(String name, Class parameterTypes[]) throws NoSuchMethodException, SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.PUBLIC );
	if (parameterTypes == null) {
		parameterTypes = nullParams;
	}
	return ((Method)getMethod0(name, parameterTypes, CONSTRUCTOR|Modifier.PUBLIC, Modifier.PUBLIC));
}

native private Member getMethod0(String name, Class[] args, int mask, int flags);

public Method[] getMethods() throws SecurityException {
	System.getSecurityManager().checkMemberAccess( this, Member.PUBLIC);
	return ((Method[])getMethods0(Modifier.PUBLIC|CONSTRUCTOR, Modifier.PUBLIC));
}

native private Member[] getMethods0(int mask, int flags);

native public int getModifiers();

native public String getName();

native static Class getPrimitiveClass(String name);

/**
 * Finds a resource with the specified name.  
 * The rules for searching for resources associated with a given class 
 * are implemented by the class loader of the class. 
 * <p>
 * The Class methods delegate to ClassLoader methods, after applying a 
 * naming convention: if the resource name starts with "/", it is used as 
 * is. Otherwise, the name of the package is prepended, after converting 
 * "." to "/". 
 *
 * @param 	name the string representing the resource to be found. 
 * @return 	the URL object having the specified name, or null if no 
 *		resource with the specified name is found. 
 */
public URL getResource(String name) {
	return (getClassLoader().getResource(fullResourceName(name)));
}

/**
 * Finds a resource with a given name.  
 * Will return null if no resource with this name is found. The rules
 * for searching a resources associated with a given class are implemented 
 * by the ClassLoader of the class.
 *
 * The Class methods delegate to ClassLoader methods, after applying a 
 * naming convention: if the resource name starts with "/", it is used 
 * as is. Otherwise, the name of the package is prepended, after 
 * converting "." to "/". 
 *
 * @param 	name the string representing the resource to be found 
 * @return 	the InputStream object having the specified name, or null 
 *		if no resource with the specified name is found. 
 */
public InputStream getResourceAsStream(String name) {
	return (getClassLoader().getResourceAsStream(fullResourceName(name)));
}

native public Object[] getSigners();

native public Class getSuperclass();

native public boolean isArray();

native public boolean isAssignableFrom(Class cls);

native public boolean isInstance(Object obj);

native public boolean isInterface();

native public boolean isPrimitive();

public Object newInstance() throws InstantiationException, IllegalAccessException {
	try {
		if (isInterface() /* || isAbstract() */) {
			throw new InstantiationException(getName());
		}
		return (newInstance0());
	}
	catch (NoSuchMethodError _) {
		throw new IllegalAccessException();
	}
}

native public Object newInstance0() throws NoSuchMethodError;

/*
 * toString() 
 */
public String toString() {
	return (isInterface() ? "interface " : "class ") + getName();
}
}
