/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.lang;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.CodeSource;
import java.security.ProtectionDomain;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;
import kaffe.lang.SystemClassLoader;
import kaffe.util.Ptr;

abstract public class ClassLoader
{
	/**
	 * To prevent any classes from being gc'd before we are gc'd, we keep
	 * them in this native hashtable.  This table only contains classes
	 * for which we are the defining loader (ie, not classes for which we
	 * are merely the initiating loader).
	 */
	private Ptr classPool;

	/**
	 * We keep a list of native libraries we own.
	 */
	private Ptr nativeLibs;

	private ProtectionDomain defaultProtectionDomain;
	final private ClassLoader parent;

protected ClassLoader() {
	System.getSecurityManager().checkCreateClassLoader();
	this.parent = getSystemClassLoader();
	SystemClassLoader.alert(this);
}

protected ClassLoader(ClassLoader parent) {
	System.getSecurityManager().checkCreateClassLoader();
	this.parent = parent;
	SystemClassLoader.alert(this);
}

final protected Class defineClass(String name, byte data[], int off, int len) throws ClassFormatError {
	if (defaultProtectionDomain == null) {
		// XXX FIXME..
		defaultProtectionDomain = new ProtectionDomain(null, null);
	}
	return defineClass(name, data, off, len, defaultProtectionDomain);
}

final protected Class defineClass(String name, byte data[], int off, int len, ProtectionDomain pd) throws ClassFormatError {
	if (off < 0 || len < 0 || off + len > data.length) {
		throw new IndexOutOfBoundsException();
	}
	// We load in this loader, but if this loader is the System we
	// change to 'null'.
	ClassLoader ldr = this;
	if (ldr == getSystemClassLoader()) {
		ldr = null;
	}
	return (defineClass0(ldr, name, data, off, len));
}

/**
 * @deprecated
 */
final protected Class defineClass(byte data[], int off, int len) throws ClassFormatError {
	return (defineClass(null, data, off, len));
}

native static private Class defineClass0(ClassLoader loader, String name, byte data[], int off, int len);

protected Package definePackage(String name, String specTitle, String specVersion, String specVendor, String implTitle, String implVersion, String implVendor, URL sealBase) throws IllegalArgumentException {
	throw new kaffe.util.NotImplemented(getClass().getName() + ".definePackage()");
}

protected Class findClass(String name) throws ClassNotFoundException {
	throw new ClassNotFoundException(name);
}

protected String findLibrary(String libname) {
	return null;
}

final protected Class findLoadedClass(String name) {
	return findLoadedClass0(name);
}

native private Class findLoadedClass0(String name);

protected URL findResource(String name) {
	try {
		Enumeration e = findResources(name);
		if ( e.hasMoreElements() ) {
			return (URL)e.nextElement();
		}
	}
	catch (IOException e) {
	}

	return null;
}

protected Enumeration findResources(String name) throws IOException {
	return new Vector().elements();		// ie, an empty Enumeration
}

final protected Class findSystemClass(String name) throws ClassNotFoundException {
	return getSystemClassLoader().findClass(name);
}

/**
 * Return the class if it's loaded already.
 */
Class getLoadedClass(String name, boolean init) {
	Class cls = findLoadedClass(name);
	if (cls != null && init) {
		resolveClass(cls);
	}
	return (cls);
}

protected Package getPackage(String name) {
	throw new kaffe.util.NotImplemented(getClass().getName() + ".getPackage()");
}

protected Package[] getPackages() {
	throw new kaffe.util.NotImplemented(getClass().getName() + ".getPackages()");
}

final public ClassLoader getParent() {
	return parent;
}

public URL getResource(String name) {
	try {
		Enumeration e = getResources(name);
		if ( e.hasMoreElements() ) {
			return (URL)e.nextElement();
		}
	}
	catch (IOException e) {
	}

	return null;
}

public InputStream getResourceAsStream(String name) {
	URL url = getResource(name);

	if (url != null) {
		try {
			return url.openStream();
		}
		catch (IOException e) {
		}
	}
	return null;
}

final public Enumeration getResources(String name) throws IOException {
	Vector v = new Vector();

	if (this != getSystemClassLoader()) {
		Enumeration e;
		if (parent != null) {
			e = parent.getResources(name);
		}
		else {
			e = getSystemClassLoader().getResources(name);
		}
		while (e.hasMoreElements()) {
			v.addElement(e.nextElement());
		}
	}
	for (Enumeration e = findResources(name); e.hasMoreElements(); ) {
		v.addElement(e.nextElement());
	}

	return v.elements();
}

public static ClassLoader getSystemClassLoader() {
	return SystemClassLoader.getClassLoader();
}

public static URL getSystemResource(String name) {
	return getSystemClassLoader().getResource(name);
}

public static InputStream getSystemResourceAsStream(String name) {
	return getSystemClassLoader().getResourceAsStream(name);
}

public static Enumeration getSystemResources(String name) throws IOException {
	return getSystemClassLoader().getResources(name);
}

public Class loadClass(String name) throws ClassNotFoundException {
	return (loadClass(name, false));
}

protected Class loadClass(String name, boolean resolve) throws ClassNotFoundException {
	Class c;

	// First, see if already loaded by this class
	c = findLoadedClass(name);
	if (c == null) {
		// Second, try the parent class loader
		try {
			if (parent == null) {
				c = getSystemClassLoader().loadClass(name, resolve);
			}
			else {
				c = parent.loadClass(name, resolve);
			}
		}
		catch (ClassNotFoundException e) {
			// Third, try findClass()
			c = findClass(name);
		}
	}

	// Now, optionally resolve the class
	if (resolve) {
		resolveClass(c);
	}
	return (c);
}

final protected void resolveClass(Class c) {
	resolveClass0(c);
}

native private void resolveClass0(Class cls);

final protected void setSigners(Class cl, Object signers[]) {
	throw new kaffe.util.NotImplemented(getClass().getName() + ".setSigners()");
}

protected void finalize() throws Throwable {
	finalize0();
	super.finalize();
}

native private void finalize0();

}
