/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.lang;

import kaffe.lang.Softcall;

public final class StringBuffer implements java.io.Serializable {

private static final int MINLEN = 16;
static final char[] nullBuf = new char[0];

static final int NORMAL = 0;
static final int TOSTRING = 1;
static final int TOSTRINGBUFFER = 2;

static final int POOLSIZE = 4096;
static char[] pool = nullBuf;
static int poolpos;

char[] buffer;				// character buffer
int bstart;				// start offset in buffer
int blength;				// buffer length
int used;				// # chars used in buffer
String myString;			// My string if I've been
					// made into one
int state;

// This is what Sun's JDK1.1 "serialver java.lang.StringBuffer" says
private static final long serialVersionUID = 3388685877147921107L; 

public StringBuffer() {
	buffer = nullBuf;
}

public StringBuffer(String str) {
	this();
	append(str);
}

public StringBuffer(int length) {
	if (length < MINLEN) {
		length = MINLEN;
	}
	alloc(length);
}

public StringBuffer append(Object obj) {
	return append(String.valueOf(obj));
}

public synchronized StringBuffer append (String str) {
	if (str == null) {
		str = "null";
	}

	int len = str.count;
	int offset = str.offset;
	char[] val = str.value;

	if (used == 0 && str.myStringBuffer != null) {
		used = len;
		buffer = val;
		bstart = offset;
		blength = str.myStringBuffer.blength;
		state = TOSTRING;

		str.myStringBuffer.state = TOSTRINGBUFFER;
		str.myStringBuffer = null;
		myString = null;

		return (this);
	}

	if (state == TOSTRINGBUFFER || used + len > blength) {
		normalize(used + len, state == TOSTRINGBUFFER);
	}
	Softcall.memcpy(buffer, bstart+used, val, offset, len, 2);
	used += len;
	return this;
}

public StringBuffer append(boolean b) {
	if (b) {
		return (append("true"));
	}
	else {
		return (append("false"));
	}
}

public synchronized StringBuffer append(char c) {
	if (state == TOSTRINGBUFFER || used + 1 > blength) {
		normalize(used + 1, state == TOSTRINGBUFFER);
	}
	buffer[bstart + used] = c;
	used++;
	return this;
}

public StringBuffer append(char str[]) {
	return (append(str, 0, str.length));
}

public synchronized StringBuffer append(char[] str, int offset, int len) {
	if (offset < 0 || len < 0 || offset + len > str.length) {
		throw new StringIndexOutOfBoundsException();
	}
	if (state == TOSTRINGBUFFER || used + len > blength) {
		normalize(used + len, state == TOSTRINGBUFFER);
	}
	Softcall.memcpy(buffer, bstart+used, str, offset, len, 2);
	used += len;
	return this;
}

public StringBuffer append(double d) {
	return append(String.valueOf(d));
}

public StringBuffer append(float f) {
	return append(String.valueOf(f));
}

public StringBuffer append(int i) {
	return Integer.convertIntoStringBuffer(this, i, 10);
}

public StringBuffer append(long l) {
	return append(String.valueOf(l));
}

public int capacity() {
	return blength;
}

public synchronized char charAt(int index) {
	if (index >= used) {
		throw new StringIndexOutOfBoundsException();
	}
	return buffer[bstart+index];
}

public synchronized StringBuffer deleteCharAt(int index) {
	if (index < 0 || index >= used) {
		throw new StringIndexOutOfBoundsException();
	}
	if (state != NORMAL) {
		normalize(used, true);
	}
	System.arraycopy(buffer, index + 1, buffer, index, --used - index);
	return this;
}

public synchronized StringBuffer delete(int start, int end) {
	if (state != NORMAL) {
		normalize(used, true);
	}
	if (end > used) {
		end = used;
	}
	if (start < 0 || start > end) {
		throw new StringIndexOutOfBoundsException();
	}
	System.arraycopy(buffer, end, buffer, start, used - end);
	used -= end - start;
	return this;
}

public synchronized StringBuffer replace(int start, int end, String str) {
	return delete(start, end).insert(start, str);
}

public synchronized String substring(int start) {
	return substring(start, used);
}

public synchronized String substring(int start, int end) {
	int length = end - start;

	if (start < 0 || end > used || length < 0) {
		throw new StringIndexOutOfBoundsException();
	}
	else {
		return (toString().substring(start, end));
	}
}

public synchronized void ensureCapacity(int minCapacity) {
	if (minCapacity > blength) {
		normalize(minCapacity, true);
	}
}

public synchronized void getChars(int srcBegin, int srcEnd, char dst[], int dstBegin) {
	if (srcEnd > used) {
		throw new StringIndexOutOfBoundsException();
	}
	Softcall.memcpy(dst, dstBegin, buffer, srcBegin+bstart, (srcEnd-srcBegin), 2);
}

public synchronized StringBuffer insert(int offset, Object obj) {
	return insert(offset, String.valueOf(obj));
}

public synchronized StringBuffer insert(int offset, String str) {
	if (str == null) {
		str = "null";
	}
	return insert(offset, str.value, str.offset, str.count);
}

public StringBuffer insert(int offset, boolean b) {
	return insert(offset, String.valueOf(b));
}

public synchronized StringBuffer insert(int offset, char c) {
	return insert(offset, new char[] { c }, 0, 1);
}

public StringBuffer insert(int offset, char[] str) {
	return insert(offset, str, 0, str.length);
}

public synchronized StringBuffer insert(int index, char[] str,
		int offset, int len) {
	if (index < 0 || index > used) {
		throw new StringIndexOutOfBoundsException();
	}
	if (offset < 0 || len < 0 || offset + len > str.length) {
		throw new StringIndexOutOfBoundsException();
	}
	if (state != NORMAL || used + len > blength) {
		normalize(used + len, true);
	}

	// Shift buffer rightward to make space
	System.arraycopy(buffer, index, buffer, index + len, used - index);

	// Copy inserted chars into place
	System.arraycopy(str, offset, buffer, index, len);

	// Update used count
	used += len;
	return this;
}

public StringBuffer insert(int offset, double d) {
	return insert(offset, String.valueOf(d));
}

public StringBuffer insert(int offset, float f) {
	return insert(offset, String.valueOf(f));
}

public StringBuffer insert(int offset, int i) {
	return insert(offset, String.valueOf(i));
}

public StringBuffer insert(int offset, long l) {
	return insert(offset, String.valueOf(l));
}

public int length () {
	return used;
}

public synchronized StringBuffer reverse() {
	if (state != NORMAL) {
		normalize(used, true);
	}
	for (int pos = used/2 - 1; pos >= 0; pos--) {
		char ch = buffer[bstart+pos];
		buffer[bstart+pos] = buffer[bstart+used - pos - 1];
		buffer[bstart+used - pos - 1] = ch;
	}
	return this;
}

public synchronized void setCharAt(int index, char ch) {
	if (index < 0 || index >= used) {
		throw new StringIndexOutOfBoundsException();
	}
	if (state != NORMAL) {
		normalize(used, true);
	}
	buffer[bstart+index] = ch;
}

public synchronized void setLength(int newLength) {
	if (newLength < 0) {
		throw new StringIndexOutOfBoundsException();
	}
	if (state != NORMAL) {
		normalize(newLength, true);
	}
	if (newLength > used) {
		if (newLength > blength) {
			ensureCapacity(newLength);
		}
		for (int index = used; index < newLength; index++) {
			buffer[bstart+index] = '\0';
		}
	}
	used = newLength;
}

public String toString() {
	synchronized (this) {
		if (myString == null) {
			new String(this); // myString set in constructor
		}
		else if (myString.count != used) {
			myString.myStringBuffer = null;
			new String(this); // myString set in constructor
		}
		return (myString);
	}
}

private void normalize(int minCapacity, boolean noextend) {
	// Increase buffer size in powers of two to
	// avoid O(n^2) behavior

	int len = blength;
	if (len < MINLEN) {
		len = MINLEN;
	}
	for (; len < minCapacity; len <<= 1)
		;

	if (!noextend && bstart + blength == poolpos && bstart + len < pool.length) {
		poolpos += len - blength;
		blength = len;
		return;
	}

	// Allocate a new buffer and copy data over
	char[] oldBuffer = buffer;
	int oldbstart = bstart;
	alloc(len);
	Softcall.memcpy(buffer, bstart, oldBuffer, oldbstart, used, 2);

	if (myString != null) {
		myString.myStringBuffer = null;
		myString = null;
	}
	state = NORMAL;
}

private void alloc(int length) {
	if (length > pool.length - poolpos) {
		if (length > POOLSIZE) {
			buffer = new char[length];
			bstart = 0;
			blength = length;
			return;
		}
		pool = new char[POOLSIZE];
		poolpos = 0;
	}
	buffer = pool;
	bstart = poolpos;
	blength = length;
	poolpos += length;
}

}
