/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.lang.reflect;

import java.lang.Class;
import java.lang.String;

public final class Method
	implements Member {

static private final Object[] nullArgs = new Object[0];
Class clazz;
int slot;
String name;
Class returnType;
Class[] parameterTypes;
Class[] exceptionTypes;

static {
	init0();
}

native private static void init0();

Method() {
}
    
public boolean equals(Object obj) {
	// Catch the simple case where they're really the same
	if ((Object)this == obj) {
		return (true);
	}
	// if obj is null or not Method, then they are not the same
	if (!(obj instanceof Method)) {
		return (false);
	}

	Method mobj = (Method)obj;

	if (clazz != mobj.clazz ||
	    returnType != mobj.returnType ||
	    parameterTypes.length != mobj.parameterTypes.length ||
	    ! name.equals(mobj.name)) {
		return (false);
	}

	for (int i = parameterTypes.length; i-- > 0; ) {
		if (parameterTypes[i] != mobj.parameterTypes[i]) {
			return (false);
		}
	}

	return (true);
}

public Class getDeclaringClass() {
	return (clazz);
}

public Class[] getExceptionTypes() {
	return (exceptionTypes);
}

public int getModifiers() {
	return (getModifiers0(clazz, slot));
}

native static int getModifiers0(Class clazz, int slot);

public String getName() {
	return (name);
}

public Class[] getParameterTypes() {
	return (parameterTypes);
}

public Class getReturnType() {
	return (returnType);
}

public int hashCode() {
	return (clazz.getName().hashCode() ^ name.hashCode());
}

public Object invoke(Object obj, Object args[]) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
	if (args == null) {
		args = nullArgs;
	}
	if (parameterTypes.length != args.length) {
		throw new IllegalAccessException("incorrect number of parameters");
	}
	else if (obj == null && !"<init>".equals(name) && (getModifiers() & Modifier.STATIC) == 0) {
		throw new NullPointerException();
	}

	/* Process arguments to get them 'correct' */
	for (int i = 0; i < args.length; i++) {
		Class pt = parameterTypes[i];
		Object arg = args[i];
		if (arg == null || pt.isAssignableFrom(arg.getClass())) {
			/* Arg okay */
		}
		else if (pt.isPrimitive()) {
			/* Might need fixup */
			if (pt == Boolean.TYPE) {
				if (arg instanceof Boolean) {
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else if (pt == Byte.TYPE) {
				if (arg instanceof Byte) {
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else if (pt == Short.TYPE) {
				if (arg instanceof Short) {
				}
				else if (arg instanceof Byte) {
					args[i] = new Short((short)((Byte)arg).byteValue());
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else if (pt == Character.TYPE) {
				if (arg instanceof Character) {
				}
				else if (arg instanceof Byte) {
					args[i] = new Character((char)((Byte)arg).byteValue());
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else if (pt == Integer.TYPE) {
				if (arg instanceof Integer) {
				}
				else if (arg instanceof Short) {
					args[i] = new Integer((int)((Short)arg).shortValue());
				}
				else if (arg instanceof Character) {
					args[i] = new Integer((int)((Character)arg).charValue());
				}
				else if (arg instanceof Byte) {
					args[i] = new Integer((int)((Byte)arg).byteValue());
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else if (pt == Long.TYPE) {
				if (arg instanceof Long) {
				}
				if (arg instanceof Integer) {
					args[i] = new Long((long)((Integer)arg).intValue());
				}
				else if (arg instanceof Short) {
					args[i] = new Long((long)((Short)arg).shortValue());
				}
				else if (arg instanceof Character) {
					args[i] = new Long((long)((Character)arg).charValue());
				}
				else if (arg instanceof Byte) {
					args[i] = new Long((long)((Byte)arg).byteValue());
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else if (pt == Float.TYPE) {
				if (arg instanceof Float) {
				}
				else if (arg instanceof Long) {
					args[i] = new Float((float)((Long)arg).longValue());
				}
				if (arg instanceof Integer) {
					args[i] = new Float((float)((Integer)arg).intValue());
				}
				else if (arg instanceof Short) {
					args[i] = new Float((float)((Short)arg).shortValue());
				}
				else if (arg instanceof Character) {
					args[i] = new Float((float)((Character)arg).charValue());
				}
				else if (arg instanceof Byte) {
					args[i] = new Float((float)((Byte)arg).byteValue());
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else if (pt == Double.TYPE) {
				if (arg instanceof Float) {
				}
				else if (arg instanceof Float) {
					args[i] = new Double((double)((Float)arg).floatValue());
				}
				else if (arg instanceof Long) {
					args[i] = new Double((double)((Long)arg).longValue());
				}
				if (arg instanceof Integer) {
					args[i] = new Double((double)((Integer)arg).intValue());
				}
				else if (arg instanceof Short) {
					args[i] = new Double((double)((Short)arg).shortValue());
				}
				else if (arg instanceof Character) {
					args[i] = new Double((double)((Character)arg).charValue());
				}
				else if (arg instanceof Byte) {
					args[i] = new Double((double)((Byte)arg).byteValue());
				}
				else {
					throw new IllegalAccessException();
				}
			}
			else {
				throw new Error("cannot happen");
			}
		}
		else {
			throw new IllegalAccessException("incompatible argument");
		}
	}

	return (invoke0(obj, args, clazz, slot, parameterTypes, returnType));
}

native private Object invoke0(Object obj, Object args[], Class clazz, int slot, Class[] paramTypes, Class returnType) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException;

public String toString() {
	StringBuffer str = new StringBuffer();
	int mod = getModifiers();

	if (Modifier.isPublic(mod)) {
		str.append("public ");
	}
	else if (Modifier.isPrivate(mod)) {
		str.append("private ");
	}
	else if (Modifier.isProtected(mod)) {
		str.append("protected ");
	}

	if (Modifier.isAbstract(mod)) {
		str.append("abstract ");
	}
	if (Modifier.isStatic(mod)) {
		str.append("static ");
	}
	if (Modifier.isFinal(mod)) {
		str.append("final ");
	}
	if (Modifier.isSynchronized(mod)) {
		str.append("synchronized ");
	}
	if (Modifier.isNative(mod)) {
		str.append("native ");
	}

	// Return type
	str.append(getPrettyName(returnType));
	str.append(" ");

	// Class name
	str.append(clazz.getName());
	str.append(".");

	// Method name
	str.append(name);
	str.append("(");

	// Signature
	for (int i = 0; i < parameterTypes.length; i++) {
		str.append(getPrettyName(parameterTypes[i]));
		if (i+1 < parameterTypes.length) {
			str.append(",");
		}
	}
	str.append(")");

        if (exceptionTypes.length > 0) {
                str.append(" throws ");
                for (int i = 0; i < exceptionTypes.length; i++) {
                        str.append(exceptionTypes[i].getName());
                        if (i+1 < exceptionTypes.length) {
                                str.append(",");
                        }
                }
        }

	return (new String(str));
}

static String getPrettyName(Class cls) {
	StringBuffer str = new StringBuffer();
	for (int count = 0;; count++) {
		if (!cls.isArray()) {
			str.append(cls.getName());
			for (; count > 0; count--) {
				str.append("[]");
			}
			return (str.toString());
		}
		cls = cls.getComponentType();
	}
}

}
