/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.net;

import java.io.Serializable;
import java.lang.String;
import java.util.StringTokenizer;
import kaffe.net.DNS;

final public class InetAddress
	implements Serializable {

final private static long serialVersionUID = 3286316764910316507L;
private static InetAddress loopback = new InetAddress("loopback", 0x7F000001);

private static InetAddress last;
private static String lasthost;
private static int lastaddr;

DNS dns;

private InetAddress(String name, int addr) {
	dns = new DNS(name, addr);
}

private InetAddress(String name) {
	dns = new DNS(name);
}

private InetAddress(int addr) {
	dns = new DNS(addr);
}

static synchronized InetAddress newInstance(int addr) {
	if (addr != lastaddr) {
		lasthost = null;
		lastaddr = addr;
		last = new InetAddress(addr);
	}
	return (last);
}

static synchronized InetAddress newInstance(String host) {
	if (!host.equals(lasthost)) {
		lasthost = host;
		lastaddr = 0;
		last = new InetAddress(host);
	}
	return (last);
}

void flushFromCache() {
	synchronized (InetAddress.class) {
		if (last == this) {
			lasthost = null;
			lastaddr = 0;
			last = null;
		}
	}
}

public boolean equals(Object obj) {
	if (this == obj) {
		return (true);
	}
	else if (obj instanceof InetAddress) {
		if (dns.equals(((InetAddress)obj).dns)) {
			return (true);
		}
	}
	return (false);
}

int getAddressAsInt() {
	try {
		return (dns.getAddress());
	}
	catch (UnknownHostException _) {
		return (0);
	}
}

public byte[] getAddress() {
	int address = getAddressAsInt();
	byte result[] = new byte[4];
	result[0] = (byte)((address >> 24) & 0xFF);
	result[1] = (byte)((address >> 16) & 0xFF);
	result[2] = (byte)((address >>  8) & 0xFF);
	result[3] = (byte) (address        & 0xFF);
	return (result);
}

public static InetAddress[] getAllByName(String host) throws UnknownHostException {
	InetAddress[] addrs = new InetAddress[1];
	addrs[0] = getByName(host);
	return (addrs);
}

public static synchronized InetAddress getByName(String host) throws UnknownHostException {
	if (host == null || host.equals("")) {
		return (getLoopback());
	}

	InetAddress addr;
	if (Character.isDigit(host.charAt(0))) {
		int ip = 0;
		StringTokenizer tok = new StringTokenizer(host, ".");
		while (tok.hasMoreElements()) {
			String s = tok.nextToken();
			ip <<= 8;
			try {
				ip |= Integer.parseInt(s);
			}
			catch (NumberFormatException _) {
			}
		}
		addr = newInstance(ip);
		addr.dns.getHostname();
	}
	else {
		addr = newInstance(host);
		addr.dns.getAddress();
	}

	return (addr);
}

static InetAddress getAnyAddress() {
	return (new InetAddress(null, 0));
}

public String getHostAddress() {
	int addr = 0;
	try {
		addr = dns.getAddress();
	}
	catch (UnknownHostException _) {
	}
	return (DNS.addressToString(addr));
}

public String getHostName() {
	try {
		return (dns.getHostname());
	}
	catch (UnknownHostException _) {
		return (getHostAddress());
	}
}

public static InetAddress getLocalHost() throws UnknownHostException {
	try {
		String name = DNS.getLocalHostname();
		if (name != null) {
			System.getSecurityManager().checkConnect(name, 0);
			return (getByName(name));
		}
	}
	catch (SecurityException se) {
	}
	return (getLoopback());
}

static InetAddress getLoopback() {
	return (loopback);
}

public int hashCode() {
	try {
		return (dns.getAddress());
	}
	catch (UnknownHostException _) {
		return (0);
	}
}

public boolean isMulticastAddress() {
	try {
		if ((dns.getAddress() & 0xE0000000) == 0xE0000000) {
			return (true);
		}
	}
	catch (UnknownHostException _) {
	}
	return (false);
}

public String toString() {
	StringBuffer result = new StringBuffer();

	result.append(getHostName());
	result.append("/");
	result.append(getHostAddress());

	return (result.toString());
}
}
