package java.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.lang.String;
import java.util.Hashtable;
import java.util.StringTokenizer;
import kaffe.net.DefaultURLStreamHandlerFactory;

/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
final public class URL
  implements Serializable
{
	final private static long serialVersionUID = -7627629688361524110L;
	private static URLStreamHandlerFactory userFactory = null;
	private static URLStreamHandlerFactory sysFactory = new DefaultURLStreamHandlerFactory();
	private URLStreamHandler handler;
	private String protocol;
	private String host = "";
	private int port = -1;
	private String file = "";
	private String ref;
	private URLConnection conn;

public URL(String spec) throws MalformedURLException {
	this(null, spec);
}

public URL(String protocol, String host, String file)
		throws MalformedURLException {
	this(protocol, host, -1, file, null);
}

public URL(String protocol, String host, int port, String file)
		throws MalformedURLException {
	this(protocol, host, port, file, null);
}

public URL(String protocol, String host, int port, String file,
		URLStreamHandler handler) throws MalformedURLException {
	setHandler(protocol, handler);
	
	if (file != null) {
		//normalize to unix like separators
		file = file.replace( '\\', '/');
	}
	this.handler.setURL(this, protocol, host, port, file, "");
}

public URL(URL context, String spec) throws MalformedURLException {
	this(context, spec, null);
}

//
// Create a new URL from an (optional) existing context and a String.
//
// Certain parts of the way an URL is parsed are universal for all protocols,
// and others are protocol specific. This is the way we parse an URL:
//
//	URL   ->   [<protocol>:]<protocol-specific-stuff>[#<reference>]
//
// Note that handler.parseURL() is responsible for merging the URL together
// with any inherited context. This is because different protocols have
// different rules for how a relative URL is interpreted with context.
//
public URL(URL context, String spec, URLStreamHandler handler)
		throws MalformedURLException {

	//normalize to unix like separators
	spec = spec.replace( '\\', '/');
	
	int firstColon = spec.indexOf(':');
	int firstSlash = spec.indexOf('/');

	// See if "spec" specifies a protocol
	if (firstColon != -1 && (firstSlash == -1 || firstColon < firstSlash)) {
		// Save protocol
		protocol = spec.substring(0, firstColon);

		// If protocols are the same, inherit from context
		if (context != null && context.protocol.equals(protocol)) {
			host = context.host;
			port = context.port;
			file = context.file;
		}
	}
	else {
		// Need to inherit the protocol from the context
		if (context == null) {
			throw new MalformedURLException("no protocol: " + spec);
		}
		protocol = context.protocol;
		host = context.host;
		port = context.port;
		file = context.file;
		handler = context.handler;
	}

	// Use handler to parse the rest, within
	// the context defined for this URL so far
	setHandler(protocol, handler);

	try {
		// In theory we don't have to substring here, we can just
		// pass the start & len on - but some handlers don't
		// do this right (lazy) so we do this to help them out.
		String nspec = spec.substring(firstColon + 1);
		this.handler.parseURL(this, nspec, 0, nspec.length());
	}
	catch (Error e) {
		throw new MalformedURLException(e.getMessage());
	}
}

public boolean equals(Object obj) {
	if (!(obj instanceof URL)) {
		return (false);
	}
	URL that = (URL)obj;

	if (this.protocol.equals(that.protocol) &&
	    this.host.equals(that.host) &&
	    this.port == that.port &&
	    this.file.equals(that.file) &&
	    (this.ref == null ?
	      (that.ref == null) : this.ref.equals(that.ref)) ) {
		return (true);
	}
	return (false);
}

final public Object getContent() throws IOException {
	return openConnection().getContent();
}

public String getFile() {
	return (file);
}

public String getHost() {
	return (host);
}

public int getPort() {
	return (port);
}

public String getProtocol() {
	return (protocol);
}

public String getRef() {
	return (ref);
}

public int hashCode() {
	return (protocol.hashCode() ^ host.hashCode() ^ file.hashCode());
}

public URLConnection openConnection() throws IOException {
	// We *ALWAYS* open a new connection even if we already have
	// one open.
	conn = handler.openConnection(this);
	// conn.connect(); - not anymore!!
	return (conn);
}

final public InputStream openStream() throws IOException {
	// We *ALWAYS* open a new connection even if we already have
	// one open.
	return openConnection().getInputStream();
}

public boolean sameFile(URL that) {
	if (this.protocol.equals(that.protocol) &&
	    this.host.equals(that.host) &&
	    this.port == that.port &&
	    this.file.equals(that.file) ) {
		return (true);
	}
	return (false);
}

protected void set(String protocol, String host, int port, String file, String ref) {
	this.protocol = protocol;
	this.host = host;
	this.port = port;
	this.file = file;
	this.ref = ref;
}

// Check the supplied handler if any, otherwise compute default handler
private void setHandler(String protocol, URLStreamHandler handler) throws MalformedURLException {
	// User-supplied handler?
	if (handler != null) {
		System.getSecurityManager().checkPermission(new NetPermission("specifyStreamHandler"));
		this.handler = handler;
		return;
	}

	if (userFactory != null) {
		handler = userFactory.createURLStreamHandler(protocol);
	}
	if (handler == null) {
		handler = sysFactory.createURLStreamHandler(protocol);
	}
	if (handler != null) {
		this.handler = handler;
		return;
	}

	// None found
	throw new MalformedURLException("unknown protocol: " + protocol);
}

public static synchronized void setURLStreamHandlerFactory(URLStreamHandlerFactory fac) {
	if (userFactory == null) {
		System.getSecurityManager().checkSetFactory();
		userFactory = fac;
	}
	else {
		throw new Error("factory already set");
	}
}

public String toExternalForm() {
	return (handler.toExternalForm(this));
}

public String toString() {
	return (toExternalForm());
}
}
