/*
 * Java core library component.
 *
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.security;

import java.util.Random;

public class SecureRandom
  extends Random
{
	final static String ENGINE_CLASS = "SecureRandom";
	final private Provider provider;
	final private SecureRandomSpi engine;

public SecureRandom() { 
	try {
		Security.Engine e = Security.getCryptInstance(ENGINE_CLASS);
		provider = e.provider;
		engine = (SecureRandomSpi)e.engine;
	}
	catch (NoSuchAlgorithmException e) {
		throw new Error("no " + ENGINE_CLASS + " found");
	}
}

protected SecureRandom(SecureRandomSpi engine, Provider provider) {
	this.engine = engine;
	this.provider = provider;
}

public SecureRandom(byte[] seed){
	this();
	setSeed(seed);
}

public byte[] generateSeed(int numBytes) {
	return engine.engineGenerateSeed(numBytes);
}

public static SecureRandom getInstance(String alg)
		throws NoSuchAlgorithmException {
	Security.Engine e = Security.getCryptInstance(ENGINE_CLASS, alg);
	return new SecureRandom((SecureRandomSpi)e.engine, e.provider);
}

public static SecureRandom getInstance(String alg, String prov) 
		throws NoSuchAlgorithmException, NoSuchProviderException {
	Security.Engine e = Security.getCryptInstance( ENGINE_CLASS, alg, prov);
	return new SecureRandom((SecureRandomSpi)e.engine, e.provider);
}

final public Provider getProvider() {
	return provider;
}

public static byte[] getSeed(int numBytes) {
	return new SecureRandom().getSeed(numBytes);
}

final protected int next(int numBits) {
	byte[] res = new byte[(numBits + 7) / 8];
	nextBytes(res);
	return ( (res[0] & 0xFF) |
		((res[1] <<  8) & 0xFF00) |
		((res[2] << 16) & 0xFF0000) |
		((res[3] << 24) & 0xFF000000));
}

public void nextBytes(byte[] bytes) {
	engine.engineNextBytes(bytes);
}

public void setSeed(byte[] seed) {
	if (engine != null) {
		engine.engineSetSeed(seed);
	}
}

public void setSeed(long seed) {
	byte[] nseed = new byte[8];
	for (int i = 0; i < 8; i++) {
		nseed[i] = (byte)(seed >> (i * 8));
	}
	setSeed(nseed);
}
}
