
/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.util.zip;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.Integer;
import kaffe.util.UTF8;

public class ZipInputStream extends InflaterInputStream
	implements ZipConstants {

  private byte sigbuf[] = new byte[4];
  private byte zheader[] = new byte[LOC_RECSZ];
  private byte dheader[] = new byte[DATA_RECSZ];
  private boolean gotSig;
  private ZipEntry entry;
  private InputStream din;

  public ZipInputStream(InputStream inp) {
    super(inp);
  }

  public ZipEntry getNextEntry() throws IOException {

    closeEntry();

    synchronized (this) {

      // Read next signature
      int sig = readSig();
      switch (sig) {
      case (int)CEN_HEADSIG:		// central dir: no more entries
	return null;
      case (int)LOC_HEADSIG:		// another entry
	break;
      default:				// unexpected
	throw new IOException("Bogus signature: 0x" + Integer.toHexString(sig));
      }

      // Read remainder of local header
      readFully(zheader, sigbuf.length, zheader.length - sigbuf.length);

      // Read filename; assume UTF-8 encoding
      byte[] nameBuf = new byte[get16(zheader, LOC_FILENAMELEN)];
      readFully(nameBuf, 0, nameBuf.length);

      // Read extra field
      byte[] extra = new byte[get16(zheader, LOC_EXTRAFIELDLEN)];
      readFully(extra, 0, extra.length);

      // Setup new entry
      entry = new ZipEntry(UTF8.decode(nameBuf));
      entry.version = get16(zheader, LOC_VERSIONEXTRACT);
      entry.flag    = get16(zheader, LOC_FLAGS);

      entry.setMethod(get16(zheader, LOC_METHOD));
      entry.setDosTime(get32(zheader, LOC_TIME));

      int datacrc = get32(zheader, LOC_CRC);
      int datacsize = get32(zheader, LOC_COMPRESSEDSIZE);
      int datasize = get32(zheader, LOC_UNCOMPRESSEDSIZE);

      if (datacrc == 0 && datasize != 0) {
	throw new IOException( "CRC error: crc=0, csize=" + datacsize + ", size=" + datasize);
      }

      entry.setCrc(datacrc & 0xffffffffL);
      entry.setCompressedSize(datacsize & 0xffffffffL);
      entry.setSize(datasize & 0xffffffffL);

      entry.setExtra( extra );
      entry.setComment("");
      entry.offset = 0;

      // We read the data into a buffer for later handling.
      int l = (int)entry.csize;
      if (l > buf.length) {
        buf = new byte[l];
      }
      din = null;
      readFully(buf, 0, l);

      // Wrap it up as a stream (inflating if required)
      din = new ByteArrayInputStream(buf, 0, l);
      if (entry.method != STORED) {
	din = new InflaterInputStream(din);
      }

      // System.out.println("ZipEntry (Post LOC): " + entry + ", meth="
      //  + entry.method + ", size=" + entry.size + ", csize=" + entry.csize
      //  + ", crc=" + entry.crc);

      return entry;
    }
  }

  public void closeEntry() throws IOException {
    synchronized (this) {
      // Don't do this twice
      if (entry != null) {
        entry = null;
        din = null;
      }
    }
  }

  public int read(byte[] b, int off, int len) throws IOException {
    if (din != null) {
      return din.read(b, off, len);
    }
    else {
      // Bypass the superclass - we don't want to use the Inflater
      return in.read(b, off, len);
    }
  }

  public long skip(long n) throws IOException {
    if (din != null) {
      return din.skip(n);
    }
    else {
      // Bypass the superclass - we don't want to use the Inflater
      return in.skip(n);
    }
  }

  public void close() throws IOException {
    closeEntry();
    super.close();
  }

  private int get16(byte[] buf, int base) {
    int val = (int)buf[base] & 0xFF;
    val |= ((int)buf[base+1] & 0xFF) << 8;
    return (val);
  }

  private int get32(byte[] buf, int base) {
    return  buf[base] & 0xFF
	| ((buf[base+1] & 0xFF) << 8)
	| ((buf[base+2] & 0xFF) << 16)
	| ((buf[base+3] & 0xFF) << 24);
  }

  private void readFully(byte[] buf, int off, int len) throws IOException {
    int nread;

    for (int remain = len; remain > 0; remain -= nread) {
      nread = read(buf, off, remain);
      if (nread == -1) {
	if (remain != 0) {
	  throw new IOException("truncated file (only read "
	    + (len - remain) + "/" + len + ")");
	}
	return;
      }
      off += nread;
    }
  }

  private int readSig() throws IOException {
    readFully(sigbuf, 0, sigbuf.length);
    return (get32(sigbuf, 0));
  }

  protected ZipEntry createZipEntry(String name) {
	// XXX FIXME what is this supposed to do?
	return null;
  }
}

