/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 *
 */
package javax.sound.sampled;


/**
  * A FloatControl object provides control over a range of floating-point values. 
  * Float controls are often represented in graphical user interfaces by continuously 
  * adjustable objects such as sliders or rotary knobs. Concrete subclasses of FloatControl 
  * implement controls, such as gain and pan, that affect a line's audio signal in some way 
  * that an application can manipulate. The FloatControl.Type inner class provides static 
  * instances of types that are used to identify some common kinds of float control. 
  *
  * The FloatControl abstract class provides methods to set and get the control's current 
  * floating-point value. Other methods obtain the possible range of values and the control's 
  * resolution (the smallest increment between returned values). Some float controls allow 
  * ramping to a new value over a specified period of time. FloatControl also includes methods 
  * that return string labels for the minimum, maximum, and midpoint positions of the control. 
  */
public abstract class FloatControl
  extends Control
{

  private Type type;
  private float maximum;
  private float minimum;
  private String maxLabel;
  private String midLabel;
  private String minLabel;
  private float precision;
  private String units;
  private int updatePeriod;
  private float value;


  protected FloatControl (Type type, float minimum,
			  float maximum, float precision, int updatePeriod,
			  float initialValue, String units)
  {
    super (type);
    this.type = type;
    this.maximum = maximum;
    this.minimum = minimum;
    this.precision = precision;
    this.units = units;
    this.updatePeriod = updatePeriod;
    this.value = initialValue;
  }

  protected FloatControl (Type type, float minimum,
			  float maximum, float precision, int updatePeriod,
			  float initialValue, String units, String minLabel,
			  String midLabel, String maxLabel)
  {
    super (type);
    this.type = type;
    this.maximum = maximum;
    this.minimum = minimum;
    this.maxLabel = maxLabel;
    this.midLabel = midLabel;
    this.minLabel = minLabel;
    this.precision = precision;
    this.units = units;
    this.updatePeriod = updatePeriod;
    this.value = initialValue;
  }

  
  public float getMaximum ()
  {
    return this.maximum;
  }

  public String getMaxLabel ()
  {
    return this.maxLabel;
  }

  public String getMidLabel ()
  {
    return this.midLabel;
  }

  public float getMinimum ()
  {
    return this.minimum;
  }

  public String getMinLabel ()
  {
    return this.minLabel;
  }

  public float getPrecision ()
  {
    return this.precision;
  }

  public String getUnits ()
  {
    return this.units;
  }

  public int getUpdatePeriod ()
  {
    return this.updatePeriod;
  }

  public float getValue ()
  {
    return this.value;
  }

  public void setValue (float newValue)
  {
    this.value = newValue;
  }

  /**
   * Changes the control value from the initial value to the final value linearly over the 
   * specified time period, specified in microseconds.   
   */
  public void shift ( float from, float to, int microseconds)
  {

    float difference = to - from;
    float current = 0;

    float increments = difference / getPrecision ();

    setValue (from);
    
    int counter = 0;
    while (counter < microseconds)
    {
      current = getValue () + getPrecision ();
      setValue (current);
      //wait (1); //wait one microsecond
      counter++;
    }


  }

  public String toString ()
  {
    return "Float Control of type [" + this.type.floatType + "]";
  }

  /**
   * An instance of the FloatControl.Type inner class identifies one kind of float control.
   *
   */
  public static class Type
    extends Control.Type
  {

    public static final Type AUX_RETURN = new Type ("AUX_RETURN");
    public static final Type AUX_SEND = new Type ("AUX_SEND");
    public static final Type BALANCE = new Type ("BALANCE");
    public static final Type MASTER_GAIN = new Type ("MASTER_GAIN");
    public static final Type PAN = new Type ("PAN");
    public static final Type REVERB_RETURN = new Type ("REVERB_RETURN");
    public static final Type REVERB_SEND = new Type ("REVERB_SEND");
    public static final Type SAMPLE_RATE = new Type ("SAMPLE_RATE");
    public static final Type VOLUME = new Type ("VOLUME");

    String floatType;


    protected Type (String type)
    {
      super (type);
      this.floatType = type;
    }


  }


}
