/**
 * Copyright (c) 2001
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */

package kaffe.awt;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.image.ImageObserver;

public class ImageDrawer
  implements ImageObserver, ComponentListener
{
	Component target;
	Graphics gr;
	Image img;
	int x;
	int y;
	Color bg;
	int flags;
	ImageDrawer next;
	static ImageDrawer cache;
	static ImageDrawer active;
	final static int RECYCLE = 1;
	final static int BOUNCE = 2;
	final static int BACKGROUND = 4;

public ImageDrawer ( Component target, Graphics g, Image img, int x, int y, Color bg ) {
	this.target = target;
	this.img = img;
	this.x = x;
	this.y = y;
	this.bg = bg;
	
	target.addComponentListener( this);
}

public void componentHidden ( ComponentEvent evt ){
	flags |= BOUNCE;
}

public void componentMoved ( ComponentEvent evt ){
	flags |= BOUNCE;
}

public void componentResized ( ComponentEvent evt ){
	flags |= BOUNCE;
}

public void componentShown ( ComponentEvent evt ){
	flags |= BOUNCE;
}

public void dispose () {
	ImageDrawer dr, d;

	if ( active != null ){
		// unlink from active list
		if ( this == active ){
			active = next;
		}
		else {
			for ( dr=active; (d = dr.next) != null; dr = d ){
				if ( d == this ){
					dr.next = next;
					break;
				}
			}
		}
	}

	target.removeComponentListener( this);

	target = null;
	gr = null;
	img = null;
	bg = null;
	flags = 0;

	synchronized ( ImageDrawer.class ){
		next = cache;
		cache = this;
	}
}

void draw ( Image img ){
	Graphics g;

	if ( gr != null ){
		if ( (flags & BACKGROUND) != 0 ){
			if ( (g = gr.create()) != null ){
				g.clipRect( x, y, img.getWidth(null), img.getHeight(null));
				target.paint(g);
				g.dispose();
			}
		}
		else {
			gr.drawImage( img, x, y, bg, this);
		}
	}
	else {
		if ( (g = target.getGraphics()) != null ){
			if ( (flags & BACKGROUND) != 0 ){
				g.clipRect( x, y, img.getWidth(null), img.getHeight(null));
				target.paint( g);
			}
			else {
				g.drawImage( img, x, y, bg, this);
			}
			g.dispose();
		}
	}
}

public static void drawImage ( Component target, Graphics g, boolean isResident,
                 Image img, int x, int y, Color bgColor, boolean isBackground ){
	int flags = target.checkImage( img, -1, -1, null);

	if ( (flags & ImageObserver.ALLBITS) != 0 ){
		// no need to redirect, draw it right away
		g.drawImage( img, x, y, bgColor, null);
	}
	else {
		ImageDrawer dr;
		
		// check if we already have a drawer registered for this request. If we
		// have, skip this one (avoid double redraws)
		for ( dr=active; dr != null; dr = dr.next ){
			if ( (dr.img == img) && (dr.target == target) && (dr.x == x) && (dr.y == y) ){
				if ( isBackground && ((flags & ImageObserver.FRAMEBITS) != 0) ){
					// we got here recursively from paint()
					// just draw the image, but keep the observer
					g.drawImage( img, x, y, bgColor, dr);
				}

				return;
			}
		}
		
		dr = getDrawer( target, (isResident ? g : null), img, x, y, bgColor);
		dr.flags |= RECYCLE;
		
		if ( isBackground ) {
			dr.flags |= BACKGROUND;
		}

		dr.next = active;
		active=dr;

		g.drawImage( img, x, y, bgColor, dr);
	}
}

protected void finalize () {
	dispose();
}

public static synchronized ImageDrawer getDrawer ( Component target, Graphics g, Image img, int x, int y, Color bgColor ){
	ImageDrawer dr = null;
	
	if ( cache == null ){
		return new ImageDrawer( target, g, img, x, y, bgColor);
	}
	else {
		dr = cache;
		cache = cache.next;
		
		dr.target = target;
		dr.gr = g;
		dr.img = img;
		dr.x = x;
		dr.y = y;
		dr.bg = bgColor;
		dr.next = null;
		
		target.addComponentListener( dr);
	}
	
	return dr;
}

public boolean imageUpdate (Image img, int infoflags, int x, int y, int width, int height ) {
	Component tgt;

	if ( (target == null) || (img != this.img) ){
		return false;
	}

	if ( (flags & BOUNCE) != 0 ){
		// some change in the target - hand it back to the target for further processing
		flags &= ~BOUNCE;
		tgt = target;
		
		if ( (flags & RECYCLE) != 0 ){
			dispose();
		}
		
		return tgt.imageUpdate( img, infoflags, x, y, width, height);
	}

	if ( (infoflags & (ALLBITS | FRAMEBITS)) != 0 ) {
		if ( target.isShowing() ){
			draw( img);
			
			if ( (infoflags & ALLBITS) != 0 ){
				if ( (flags & RECYCLE) != 0 ){
					dispose();
				}
				return false;
			}
		}
	}

	return ( (infoflags & (ALLBITS | ABORT | ERROR)) == 0 );
}
}
