package kaffe.awt.flash;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Panel;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import kaffe.io.AccessibleBAOStream;
import kaffe.util.Ptr;

public class FlashImage
	extends Panel {

private static final int STOPPED	     = 0;
private static final int PARSING	     = 1;
private static final int PENDING	     = 2;
private static final int PLAYING	     = 3;

private static final int FLASH_WAKEUP        = 1;
private static final int FLASH_EVENT         = 2;

private static final int FLASH_KEY_UP        = 1;
private static final int FLASH_KEY_DOWN      = 2;
private static final int FLASH_KEY_LEFT      = 3;
private static final int FLASH_KEY_RIGHT     = 4;
private static final int FLASH_KEY_ENTER     = 5;
private static final int FLASH_KEY_NEXT      = 6;

private static final int FLASH_MOUSE_MOVE    = 1;
private static final int FLASH_MOUSE_PRESS   = 2;
private static final int FLASH_MOUSE_RELEASE = 3;

private static final String SOUNDDEVICE = "/dev/dsp";

private static FlashImagePlayer player = FlashImagePlayer.getInstance();

private Ptr handle;
private int state;
private Image canvas;

private byte[] fdata;
private int foffset;
private int flength;

private int eventKey;
private int eventMouse;
private int eventX;
private int eventY;
private long eventTime;

public FlashImage() {
}

public void setSrc(URL url) throws IOException {
	AccessibleBAOStream data = new AccessibleBAOStream();
	InputStream in = url.openStream();
	if (in == null) {
		throw new IOException("failed to open " + url);
	}
	if (!data.readFrom(in)) {
		throw new IOException("failed to read " + url);
	}
	setSrc(data.getBuffer(), 0, data.size());
}

public void setSrc(byte[] data, int offset, int length) throws IOException {
	if (state != STOPPED) {
		changeState(STOPPED);
	}
	fdata = data;
	foffset = offset;
	flength = length;
	changeState(PARSING);
}

public void play() {
	changeState(PENDING);
}

public void stop() {
	changeState(STOPPED);
}

public void restart() {
}

private void changeState(int newstate) {
	while (state != newstate) {
//System.out.println(state + " -> " + newstate);
		synchronized (player) {
			switch (state) {
			case STOPPED:
				switch (newstate) {
				case PARSING:
				case PENDING:
					parseFlash();
					state = PARSING;
					break;

				default:
					throw new Error("bad state transition");
				}
				break;

			case PARSING:
				switch (newstate) {
				case PENDING:
					state = PENDING;
					if (getPeer() != null) {
						newstate = PLAYING;
					}
					break;

				case STOPPED:
					state = STOPPED;
					break;

				default:
					throw new Error("bad state transition");
				}
				break;

			case PENDING:
				switch (newstate) {
				case PLAYING:
					if (getPeer() != null) {
						state = PLAYING;
						enableEvents(AWTEvent.MOUSE_EVENT_MASK|AWTEvent.MOUSE_MOTION_EVENT_MASK|AWTEvent.KEY_EVENT_MASK);
						openFlashDisplay();
						player.add(this);
						eventTime = -1;
						player.notify();
					}
					break;

				case STOPPED:
					state = STOPPED;
					break;

				default:
					throw new Error("bad state transition");
				}
				break;

			case PLAYING:
				switch (newstate) {
				case STOPPED:
				case PENDING:
					state = STOPPED;
					player.remove(this);
					disableEvents(AWTEvent.MOUSE_EVENT_MASK|AWTEvent.MOUSE_MOTION_EVENT_MASK|AWTEvent.KEY_EVENT_MASK);
					closeFlashDisplay();
					break;

				default:
					throw new Error("bad state transition");
				}
				break;

			default:
				throw new Error("bad state");
			}
		}
	}
}

public void addNotify() {
	switch (state) {
	case STOPPED:
	case PARSING:
		super.addNotify();
		break;

	case PENDING:
		super.addNotify();
		changeState(PLAYING);
		break;

	case PLAYING:
		changeState(STOPPED);
		super.addNotify();
		changeState(PENDING);
		break;
	}
}

public void removeNotify() {
	switch (state) {
	case STOPPED:
	case PARSING:
	case PENDING:
		super.removeNotify();
		break;

	case PLAYING:
		changeState(STOPPED);
		super.removeNotify();
		changeState(PENDING);
		break;
	}
}

public boolean isFocusTraversable() {
	return true;
}
	  
public void reshape(int x, int y, int w, int h) {
	Dimension d = getSize();
	if (w == d.width && h == d.height) {
		super.reshape(x, y, w, h);
	}
	else {
		switch (state) {
		case STOPPED:
		case PARSING:
			super.reshape(x, y, w, h);
			break;

		case PLAYING:
		case PENDING:
			stop();
			super.reshape(x, y, w, h);
			play();
			break;
		
		default:
			break;
		}
	}
}

long handleEvents() {
	long timeout = -1;

	if (eventMouse != 0 || eventKey != 0) {
		timeout = flashExec0(handle, FLASH_EVENT, eventKey, eventMouse, eventX, eventY);
	}
	eventMouse = 0;
	eventKey = 0;

	/* Timeout last event */
	if (System.currentTimeMillis() > eventTime) {
		eventTime = -1;
	}

	/* Calculate time to next event */
	if (timeout == -1) {
	}
	else if (eventTime == -1) {
		eventTime = timeout;
	}
	else if (timeout < eventTime) {
		eventTime = timeout;
	}

	/* No event time known - get a new time */
	if (eventTime == -1) {
		eventTime = flashExec0(handle, FLASH_WAKEUP, 0, 0, 0, 0);
	}

	return (eventTime);
}

void redraw() {
	switch (state) {
	case PLAYING:
		if (flashNeedRedraw0(handle)) {
			repaint();
		}
		break;

	default:
		break;
	}
}

protected void processKeyEvent(KeyEvent evt) {
	keyEvent(evt);
}

protected void processMouseEvent(MouseEvent evt) {
	mouseEvent(evt);
}

protected void processMouseMotionEvent(MouseEvent evt) {
	mouseEvent(evt);
}

public void update(Graphics g) {
	paint(g);
}

public void paint(Graphics g) {
	if (canvas != null) {
		g.drawImage(canvas, 0, 0, null);
	}
}

public Dimension getPreferredSize() {
	switch (state) {
	case PENDING:
	case PLAYING:
	case PARSING:
		return (new Dimension(flashWidth0(handle), flashHeight0(handle)));

	default:
	case STOPPED:
		return (new Dimension(32, 32));
	}
}

private void parseFlash() {
	handle = flashNew0();
	flashParse0(handle, fdata, foffset, flength);
}

private void openFlashDisplay() {
	if (getPeer() == null) {
		throw new Error("not showing");
	}
	Dimension d = getSize();
	canvas = createImage(d.width, d.height);
	flashOpenDisplay0(handle, canvas, d.width, d.height);
	flashOpenSound0(handle, SOUNDDEVICE);
}

private void closeFlashDisplay() {
	if (canvas != null) {
		canvas.flush();
		canvas = null;
	}
	if (handle != null) {
		flashCloseDisplay0(handle);
		handle = null;
	}
}

private void keyEvent(KeyEvent evt) {
	synchronized (player) {
		if (evt.getID() != KeyEvent.KEY_PRESSED) {
			return;
		}
		switch (evt.getKeyCode()) {
		case KeyEvent.VK_UP:
			eventKey = FLASH_KEY_UP;
			break;
		case KeyEvent.VK_DOWN:
			eventKey = FLASH_KEY_DOWN;
			break;
		case KeyEvent.VK_LEFT:
			eventKey = FLASH_KEY_LEFT;
			break;
		case KeyEvent.VK_RIGHT:
			eventKey = FLASH_KEY_RIGHT;
			break;
		case KeyEvent.VK_ENTER:
			eventKey = FLASH_KEY_ENTER;
			break;
		case KeyEvent.VK_TAB:
			eventKey = FLASH_KEY_NEXT;
			break;
		default:
			eventKey = 0;
			break;
		}
		if (eventKey != 0) {
			player.notify();
		}
	}
}

private void mouseEvent(MouseEvent evt) {
	synchronized (player) {
		switch (evt.getID()) {
		case MouseEvent.MOUSE_PRESSED:
			eventMouse = FLASH_MOUSE_PRESS;
			break;
		case MouseEvent.MOUSE_RELEASED:
			eventMouse = FLASH_MOUSE_RELEASE;
			break;
		case MouseEvent.MOUSE_MOVED:
			eventMouse = FLASH_MOUSE_MOVE;
			eventX = evt.getX();
			eventY = evt.getY();
			break;
		case MouseEvent.MOUSE_ENTERED:
			requestFocus();
			break;
		default:
			eventMouse = 0;
			break;
		}
		if (eventMouse != 0) {
			player.notify();
		}
	}
}

protected void getURLCallback(String url, String target) {
}

protected void getSwfCallback(String url, int level) {
}

private native long flashExec0(Ptr handle, int flags, int key, int mouse, int x, int y);
private native Ptr flashNew0();
private native boolean flashParse0(Ptr handle, byte[] data, int offset, int length);
private native boolean flashNeedRedraw0(Ptr handle);
private native int flashWidth0(Ptr handle);
private native int flashHeight0(Ptr handle);
private native void flashOpenSound0(Ptr handle, String device);
private native void flashOpenDisplay0(Ptr handle, Image canvas, int width, int height);
private native void flashCloseDisplay0(Ptr handle);

// ----------------------------------------------------------------------

public static void main(String[] a) throws Throwable {
	FlashImage flash = new FlashImage();
	java.awt.Frame frame = new java.awt.Frame();
	frame.add(flash);
	flash.setSrc(new URL(a[0]));
	frame.pack();
	frame.show();
	flash.play();
}

}
