/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.lang;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.ClassLoader;
import java.lang.ClassNotFoundException;
import java.lang.String;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;
import kaffe.io.AccessibleBAIStream;

public class SystemClassLoader
  extends Application {

final private static SystemClassLoader singleton;
static boolean otherLoaders;
private File last;

static {
	singleton = new SystemClassLoader();
	otherLoaders = false;
}

private SystemClassLoader() {
	super(null);		// this line is not really necessary!
}

protected Class findClass(String name) throws ClassNotFoundException {
	return (Class.forName(name, true, null));
}

protected URL findResource(String name) {
	return ((URL)findMyResources(name, true));
}

/**
 * This allows us to add a class to the system class loader from outside.
 */
public Class addClass(byte data[], int start, int len) {
	return (defineClass(null, data, start, len));
}

/*
 * Search through the CLASSPATH directories and ZIP files to find
 * the named resource (which may appear more than once). Make sure
 * it really exists in each place before adding it.
 */
protected Enumeration findResources(String name) {
	return ((Enumeration)findMyResources(name, false));
}

/*
 * Search for resources - if 'one' is true then we return the first
 * match we find, otherwise we return an enumeration of matches.
 */
private Object findMyResources(String name, boolean one) {
	Vector v;

	/* If we're looking for a single resource and we have a 'last' place
	 * where we found the last resource, chances are we'll find this
	 * resource there too.  So look there first and if we find it
	 * return it immediately.
	 */
	if (one) {
		if (last != null) {
			URL u = findResourceURL(last, name);
			if (u != null) {
				return (u);
			}
		}
		v = null;
	}
	else {
		v = new Vector();
	}

	String classpath = System.getProperties().getProperty("java.class.path");
	StringTokenizer t = new StringTokenizer(classpath, File.pathSeparator);
	while (t.hasMoreTokens()) {
		File file = new File(t.nextToken());
		URL u = findResourceURL(file, name);
		if (u != null) {
			if (one) {
				last = file;
				return (u);
			}
			v.addElement(u);
		}
	}
	if (one) {
		return (null);
	}
	return (v.elements());
}

private URL findResourceURL(File file, String name) {
	if ( file.exists() ) {
		if (file.isDirectory()) {
			file = new File(file, name);				
			if (file.isFile()) {
				try {
					String pn = file.getCanonicalPath().replace(File.separatorChar, '/');
					return (new URL("file", "", pn));
				} catch (IOException e) {
				}
			}
		}
		else if (file.isFile()) {
			ZipFile zip = null;
			try {
				zip = new ZipFile(file);
				ZipEntry entry = zip.getEntry( name);
				if (entry != null && !entry.isDirectory()) {
					String pn = file.getCanonicalPath().replace(File.separatorChar, '/')
						+ "!/" + entry.getName();
					return (new URL("zip:file:" + pn));
				}
			}
			catch (IOException e) {
			}
			finally {
				if (zip != null) {
					try {
						zip.close();
					}
					catch (IOException e) {
					}
				}
			}
		}
	}
	return (null);
}

public static SystemClassLoader getClassLoader() {
	return (singleton);
}

public static Class[] getClassStack() {
	return (getClassStack0());
}

native private static Class[] getClassStack0();

native private static Class getCurrentClass0();

public static ClassLoader getCurrentClassLoader() {
	if (otherLoaders) {
		Class cls = getCurrentClass0();
		if (cls != null) {
			return (cls.getClassLoader());
		}
	}
	return (singleton);
}

public static ClassLoader getCallerClassLoader(int depth) {
	if (otherLoaders) {
		ClassLoader loader = getCallerClassLoader0(depth + 3);
		if (loader != null) {
			return (loader);
		}
	}
	return (singleton);
}

native private static ClassLoader getCallerClassLoader0(int depth);

/*
 * Bypass all the class loaders and URL stuff and get the file
 * directly from the system.
 */
public InputStream getResourceAsStream(String name) {
	byte[] data = getSystemResourceAsBytes0(name);
	if (data != null) {
		return (new AccessibleBAIStream(data));
	}
	else {
		return (null);
	}
}

native private static byte[] getSystemResourceAsBytes0(String filename);

/**
 * @deprecated, Deprecated in 1.2
 */
public static boolean inClassLoader() {
	if (otherLoaders && getCurrentClassLoader() != singleton) {
		return (true);
	}
	else {
		return (false);
	}
}

native private static int getCurrentClassLoaderDepth0();

/**
 * @deprecated, Deprecated in 1.2
 */
public static int classLoaderDepth() {
	if (otherLoaders) {
		return (getCurrentClassLoaderDepth0());
	}
	return (-1);
}

public static Class currentLoadedClass() {
	if (otherLoaders) {
		return (getCurrentClass0());
	}
        return (null);
}

/**
 * Alert the system that we're creating a new class loader so
 * we can turn on some of the more expensive checking stuff.
 */
public static void alert(ClassLoader loader) {
	otherLoaders = true;
}

public native Class loadClass(String name, boolean resolve) throws ClassNotFoundException;

public native String findLibrary(String name);

}
