/*
 * cmap_p8_256.c - fgl module for pixelvalue/RGB mapping
 *                 (8bpp PseudoColors with equal distributed 
 *                  6x6x6 "web safe" color cube)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 09/01/00  Peter Mehlitz
 *
 * $tvt: cmap_p8.h,v 1.1 2001/11/20 00:21:44 damian Exp $
 */

#ifndef FGL_CMAP_H_INCLUDED
#define FGL_CMAP_H_INCLUDED

#include "libfgl.h"
#include <fgl.h>

extern u8 rgbCube[12][12][12];
extern fgl_cmap_t* fglCmap;
extern fgl_cmap_t* oldCmap;
extern u8 a6[6];

/*
 * these are our public APIs for RGB<->pixVal conversion
 */


static inline
fgl_color_t _fgl_getPixelValue ( u32 argb ) 
{
	int i, j, k;
	
	i = (((int)R_8(argb)) * 4 - 1) / 86;
	j = (((int)G_8(argb)) * 4 - 1) / 86;
	k = (((int)B_8(argb)) * 4 - 1) / 86;
	
	return rgbCube[i][j][k];
}

static inline
void _fgl_getRgbValues ( fgl_color_t pix, int* r, int* g, int* b )
{
	*r = fglCmap->r[pix];
	*g = fglCmap->g[pix];
	*b = fglCmap->b[pix];
}

static inline
fgl_color_t _fgl_modPixelValue( fgl_color_t pix, int mod )
{
	u32 r, g, b;
	fgl_getRgbValues( pix, &r, &g, &b);

	mod = 3 + mod;
	if ( mod < 0 ) mod = 0;

	r = (r * mod) / 3;
	g = (g * mod) / 3;
	b = (b * mod) / 3;

	if ( r > 0xff ) r = 0xff;
	if ( g > 0xff ) g = 0xff;
	if ( b > 0xff ) b = 0xff;

	return _fgl_getPixelValue( RGB_888( r, g, b));
}

static inline
fgl_color_t _fgl_blendPixelValues ( u32 alpha, fgl_color_t d, fgl_color_t s )
{
	u32 ialpha = 255 - alpha;
	u32  dr, dg, db, sr, sg, sb;
	
	fgl_getRgbValues( d, &dr, &dg, &db);
	fgl_getRgbValues( s, &sr, &sg, &sb);
	
	alpha++;
	
	dr = ((alpha * sr + ialpha * dr) + 255) >> 8;
	dg = ((alpha * sg + ialpha * dg) + 255) >> 8;
	db = ((alpha * sb + ialpha * db) + 255) >> 8;
	
	return _fgl_getPixelValue( RGB_888(dr, dg, db));
}

static inline
int _addPaletteEntry ( fgl_cmap_t* map, u8 r, u8 g, u8 b )
{
  int n = map->len;

  map->r[n] = r;
  map->g[n] = g;
  map->b[n] = b;

  map->len++;

  return n;
}

static inline
void _restoreOldCmap ()
{
  if ( oldCmap ){
      fgl_setHWCmap( oldCmap);
  }
}

int _fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld )
{
  int  i, j, k, i2, j2, k2;
  int  r, g, b; 
  u8   pix;

  fglCmap = fgl_allocCmap();

  /* allocate the 216 web colors, and init the lookup cube cells from them */
  for ( i=0, i2=0; i<6; i++, i2=i*2 ){
      for ( j=0, j2=0; j<6; j++, j2=j*2 ){
          for ( k=0, k2=0; k<6; k++, k2=k*2 ){
              pix = _addPaletteEntry( fglCmap, a6[i], a6[j], a6[k]);

              rgbCube [i2  ][j2  ][k2] = pix;
              rgbCube [i2+1][j2  ][k2] = pix;
              rgbCube [i2+1][j2+1][k2] = pix;
              rgbCube [i2  ][j2+1][k2] = pix;

              k2++;
              rgbCube [i2  ][j2  ][k2] = pix;
              rgbCube [i2+1][j2  ][k2] = pix;
              rgbCube [i2+1][j2+1][k2] = pix;
              rgbCube [i2  ][j2+1][k2] = pix;
          }
      }
  }


  /* there might be up to 40 colors left, try to distribute them most evenly among
  * the edges and diagonals of the RGB cube
  */
  if ( fglCmap->len + nReq < 251 ) {
      /* grays */
      rgbCube[10][10][10] = _addPaletteEntry( fglCmap, 238, 238, 238);
      rgbCube[ 8][ 8][ 8] = _addPaletteEntry( fglCmap, 178, 178, 178);
      rgbCube[ 6][ 6][ 6] = _addPaletteEntry( fglCmap, 128, 128, 128);
      rgbCube[ 5][ 5][ 5] = _addPaletteEntry( fglCmap, 116, 116, 116);
      rgbCube[ 3][ 3][ 3] = _addPaletteEntry( fglCmap,  76,  76,  76);
      rgbCube[ 1][ 1][ 1] = _addPaletteEntry( fglCmap,  26,  26,  26);
  }

  if ( fglCmap->len + nReq < 251 ) {
  /* some more reds */
  rgbCube[10][0][0] = _addPaletteEntry( fglCmap, 230, 0, 0);
  rgbCube[ 8][0][0] = _addPaletteEntry( fglCmap, 179, 0, 0);
  rgbCube[ 6][0][0] = _addPaletteEntry( fglCmap, 128, 0, 0);
  rgbCube[ 5][0][0] = _addPaletteEntry( fglCmap, 116, 0, 0);
  rgbCube[ 3][0][0] = _addPaletteEntry( fglCmap,  76, 0, 0);
  rgbCube[ 1][0][0] = _addPaletteEntry( fglCmap,  26, 0, 0);
  }
  if ( fglCmap->len + nReq < 251 ) {
      /* and greens */
      rgbCube[0][10][0] = _addPaletteEntry( fglCmap, 0, 230, 0);
      rgbCube[0][ 8][0] = _addPaletteEntry( fglCmap, 0, 179, 0);
      rgbCube[0][ 6][0] = _addPaletteEntry( fglCmap, 0, 128, 0);
      rgbCube[0][ 5][0] = _addPaletteEntry( fglCmap, 0, 116, 0);
      rgbCube[0][ 3][0] = _addPaletteEntry( fglCmap, 0,  76, 0);
      rgbCube[0][ 1][0] = _addPaletteEntry( fglCmap, 0,  26, 0);
  }


  if ( fglCmap->len + nReq < 251 ) {
      /* and blue */
      rgbCube[0][0][10] = _addPaletteEntry( fglCmap, 0, 0, 230);
      rgbCube[0][0][ 8] = _addPaletteEntry( fglCmap, 0, 0, 179);
      rgbCube[0][0][ 6] = _addPaletteEntry( fglCmap, 0, 0, 128);
      rgbCube[0][0][ 5] = _addPaletteEntry( fglCmap, 0, 0, 116);
      rgbCube[0][0][ 3] = _addPaletteEntry( fglCmap, 0, 0,  76);
      rgbCube[0][0][ 1] = _addPaletteEntry( fglCmap, 0, 0,  26);
  }


  if ( fglCmap->len + nReq < 254 ) {
      /* and yellow */
      rgbCube[10][10][0] = _addPaletteEntry( fglCmap, 230, 230, 0);
      rgbCube[ 8][ 8][0] = _addPaletteEntry( fglCmap, 179, 179, 0);
      rgbCube[ 6][ 6][0] = _addPaletteEntry( fglCmap, 128, 128, 0);
      //rgbCube[0][ 5][0] = _addPaletteEntry( fglCmap, 116, 116, 0);
      //rgbCube[ 3][ 3][0] = _addPaletteEntry( fglCmap, 76,  76, 0);
      //rgbCube[ 1][ 1][0] = _addPaletteEntry( fglCmap, 26,  26, 0);
  }

  if ( fglCmap->len + nReq < 256 ) {
      rgbCube[10][ 9][0] = _addPaletteEntry( fglCmap, 238, 204, 0);
  }
  
  if ( fglCmap->len + nReq < 253 ) {
      /* and cyan */
      rgbCube[0][10][10] = _addPaletteEntry( fglCmap, 0, 230, 230);
      rgbCube[0][ 8][ 8] = _addPaletteEntry( fglCmap, 0, 179, 179);
      rgbCube[0][ 6][ 6] = _addPaletteEntry( fglCmap, 0, 128, 128);
      rgbCube[0][0][ 5]  = _addPaletteEntry( fglCmap, 0, 116, 116);
      //rgbCube[0][ 3][ 3] = _addPaletteEntry( fglCmap, 0, 76,  76);
      //rgbCube[0][ 1][ 1] = _addPaletteEntry( fglCmap, 0, 26,  26);
  }
  
  if ( fglCmap->len + nReq < 255 ) {
      /* sky */
      rgbCube[9][10][10] = _addPaletteEntry( fglCmap, 204, 220, 238);
      rgbCube[9][ 9][10] = _addPaletteEntry( fglCmap, 204, 204, 238);
  }
  
  if ( fglCmap->len + nReq < 255 ) {
      /* flesh */
      rgbCube[10][9][10] = _addPaletteEntry( fglCmap, 238, 204, 220);
      rgbCube[10][9][ 9] = _addPaletteEntry( fglCmap, 234, 204, 204);
  }

  if ( fglCmap->len + nReq < 252 ) {
      /* magenta */
      rgbCube[10][0][10] = _addPaletteEntry( fglCmap, 230, 0, 230);
      rgbCube[ 8][0][ 8] = _addPaletteEntry( fglCmap, 178, 0, 178);
      rgbCube[ 6][0][ 6] = _addPaletteEntry( fglCmap, 128, 0, 128);
      //rgbCube[ 5][0][ 5] = _addPaletteEntry( fglCmap, 116, 0, 116);
      rgbCube[ 3][0][ 3] = _addPaletteEntry( fglCmap,  76, 0,  76);
      rgbCube[ 1][0][ 1] = _addPaletteEntry( fglCmap,  26, 0,  26);
  }


  /* set cells of explicit rgb requests (watch out, this might overwrite all previous,
  * cells, so choose them carefully
   */
  if ( nReq > 40 ) nReq = 40;   /* 256 - 216 web colors */
  for (j = 0; (j < nReq); j++) {
      r = R_8(rgbReq[j]);
      g = G_8(rgbReq[j]);
      b = B_8(rgbReq[j]);

      i = (r * 4 - 1) / 86;
      j = (g * 4 - 1) / 86;
      k = (b * 4 - 1) / 86;

      rgbCube[i][j][k] = _addPaletteEntry( fglCmap, r, g, b);
  }

  /* Ok, we have set up fglCmap and rgbCube, now install it on the device */
  if ( saveOld ){
      /* save current cmap so that we can restore it on termination */
      oldCmap = fgl_allocCmap();
      if ( fgl_getHWCmap( oldCmap) ){
          atexit( _restoreOldCmap);
      }
  }

  if ( fgl_setHWCmap( fglCmap) ){
      fgl_clrBlack     = fgl_getPixelValue( 0);
      fgl_clrWhite     = fgl_getPixelValue( 0xffffff);
      fgl_clrRed       = fgl_getPixelValue( 0xff0000);
      fgl_clrGreen     = fgl_getPixelValue( 0x00ff00);
      fgl_clrBlue      = fgl_getPixelValue( 0x0000ff);
      fgl_clrYellow    = fgl_getPixelValue( 0xffff00);
      fgl_clrCyan      = fgl_getPixelValue( 0x00ffff);
      fgl_clrGray      = fgl_getPixelValue( 0x808080);
      fgl_clrLightGray = fgl_getPixelValue( 0xc0c0c0);
      fgl_clrOrange    = fgl_getPixelValue( 0xffc800);

      return 1;
  }
  return 0;
}


#endif /* FGL_CMAP_H_INCLUDED */
