/*
 * feng_ttf.c - fgl module to implement a truetype font file (*.ttf)
 *              font engine (using the freetype library)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: feng_ttf.c,v 1.6 2001/11/20 18:45:46 damian Exp $
 */

#include "feng.h"

#include <ft2build.h>
#include FT_FREETYPE_H

char *fntDir;

static int delta[] = { 0, -1, 1, -2, 2 };
fgl_fontEntry_t            *_fgl_fontCache;
int                        _fgl_fontInitialized = 0;
fgl_fontEntry_t            *_fgl_setCache;
int                        _fgl_setInitialized = 0;
fgl_font_t                 *fgl_defaultFont = 0;

int                        _fgl_antialias = 1;
int                        _fgl_setCount;
char                       **_fgl_setName;

FT_Library library;

int xres = 96;
int yres = 96;

#define MAX_NGLYPHS 256


int fgl_initFontEngine () 
{
        int error;
  
	if ( (error = FT_Init_FreeType( &library)) ){
		warnx("FT_Init_FreeType failed: %d\n", error);
		return 0;
	}

	if ( !(fntDir = getenv( "FGL_FONTS")) ){
#if defined(DEF_FONTPATH)
		fntDir = DEF_FONTPATH;
#else
		fntDir = "/usr/local/lib/fgl/fonts";
#endif
	}
        if ( getenv( "FGL_FONTS_MONO" )) {
            _fgl_antialias = 0;
        }
	return 1;
}

fgl_font_t* fgl_loadFont ( char* fname, u8 ptSize ) /*{{{*/
{
  FT_Error             error;
  FT_Face              face;

  char                 buf[128];
  int                  i, j, k, w, h, m;
  u16                  platform, encoding, nGlyphs;
  u8                   *pSurf;
  fgl_font_t           *fnt = 0;
  u16                  *map;


  sprintf( buf, "%s/%s.ttf", fntDir, fname);

  if ( (error = FT_New_Face( library, buf, 0, &face)) ){
      warnx("FT_New_Face \"%s\" failed: %d\n", buf, error);
      return 0;
  }

  if ( (error = FT_Set_Char_Size( face, 0, ptSize * 64, xres, yres)) ){
      warnx("FT_Set_Char_Size failed: %d\n", error);
      return 0;
  }

  /* there can be quite a number of glyphs in a ttf file, check if we have to cap */
  nGlyphs = ( face->num_glyphs > MAX_NGLYPHS) ? MAX_NGLYPHS : face->num_glyphs;

  /* BAD! reverse mapping eats up memory, and it probably is used
  * just for gaps (all referring to the same void glyph). But how
  * can we detect if we already processed a certain glyph index,
  * without walking the whole unicode array?
   */
  map = FGL_MALLOC( face->num_glyphs * sizeof(u16));
  memset( map, 0, face->num_glyphs * sizeof(u16));	

  fnt = allocFont( nGlyphs);
  m = 0;

  /* first pass, fill in font data */
  for ( i = 0; i < nGlyphs; i++ ){
      k = FT_Get_Char_Index( face, i );

      if ( map[k] && (map[k] < nGlyphs) ){
          /* if we already have a map entry for k, simply copy the data (if it
          * doesn't refer to a out-of-range glyph
           */
          j = map[k];

          fnt->width[i]    = fnt->width[j];
          fnt->height[i]   = fnt->height[j];			
          fnt->xBearing[i] = fnt->xBearing[j];
          fnt->yBearing[i] = fnt->yBearing[j];
          fnt->advance[i]  = fnt->advance[j];			
      }
      else {
          /* Ok, we don't have this one in our glyph space yet, query it */

          if ( !(error = FT_Load_Glyph( face, k, FT_LOAD_RENDER)) ){

              w = face->glyph->bitmap.width;
              h = face->glyph->bitmap.rows;
              
              if ( _fgl_antialias ) {
                  m += w * h;
              }
              else {
                  m += h * MONO_SLENGTH( w );
              }
              fnt->width[i]    = w;
              fnt->height[i]   = h;
              fnt->xBearing[i] = face->glyph->bitmap_left;
              fnt->yBearing[i] = face->glyph->bitmap_top;
              fnt->advance[i]  = face->glyph->advance.x >> 6;

              map[k] = i;
          }
      }
  }

  //fnt->ascend  = face->size->metrics.ascender / 64;
  //fnt->descend = face->size->metrics.descender/ 64;

  fnt->ascend  = FT_MulFix ( face->ascender , face->size->metrics.y_scale) / 64;
  fnt->descend = - FT_MulFix ( face->descender , face->size->metrics.y_scale) / 64; 
  fnt->fontHeight = FT_MulFix ( face->height , face->size->metrics.y_scale) / 64;
  fnt->descend += fnt->fontHeight - fnt->ascend - fnt->descend;
  pSurf = fnt->surface = FGL_CALLOC(1, m);

  /* second pass, render glyphs */
  for ( i = 0; i < nGlyphs; i++ ){
      k = FT_Get_Char_Index( face, i );

      if ( map[k] == i ) {
          /* this is a new one, render the glyph */

            if ( fnt->width[i] ) {
              face->glyph->bitmap.width = fnt->width[i];
              face->glyph->bitmap.rows = fnt->height[i];
              if ( _fgl_antialias ) {
                  face->glyph->bitmap.pitch = fnt->width[i];
              }
              else {
                  face->glyph->bitmap.pitch = MONO_SLENGTH( fnt->width[i] );
              }
              fnt->surfOff[i]  = (unsigned long)pSurf - (unsigned long)fnt->surface;

              if ( !(error = FT_Load_Glyph( face, k, FT_LOAD_NO_BITMAP )) ) {
                  if ( _fgl_antialias ) {
                      FT_Render_Glyph ( face->glyph, ft_render_mode_normal );
                      memcpy ( pSurf, face->glyph->bitmap.buffer, fnt->width[i] * fnt->height[i]);
                      pSurf += (fnt->width[i] * fnt->height[i]);
                  }
                  else {
                      FT_Render_Glyph ( face->glyph, ft_render_mode_mono );
                      memcpy ( pSurf, face->glyph->bitmap.buffer, MONO_SLENGTH(fnt->width[i]) * fnt->height[i]);
                      pSurf += (MONO_SLENGTH(fnt->width[i]) * fnt->height[i]);
                  }
              }
            }
      }
      else {
          /* We already had this one, copy the offset. Note that map[k]
          * can't be outside our glyph space because of the way we set it in pass 1
           */
          fnt->surfOff[i] = fnt->surfOff[ map[k] ];
      }
  }

  fnt->sLength = fnt->width;
  fnt->uniOffset =0;
  fnt->state = 0;
  if ( _fgl_antialias ) fnt->state |= AAF_8;
  FT_Done_Face( face);

  FGL_FREE( map);

  fnt->fname  = strdup( fname);
  fnt->ptSize = ptSize;

  return fnt;
}

void fgl_unloadFont ( fgl_font_t* fnt ) /*{{{*/
{
  FGL_FREE( fnt->surface);
  FGL_FREE( fnt);
}
