/* -*- mode: C; mode: fold -*- */
/*
 * fgld_16  - fgl header to handle framebuffer access and scanline cells
 *           for 16-bpp visuals (read/write value, compute/alter scanline 
 *           cell addresses), generic part (might be partly overriden by
 *           platform specific optimizations)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FG_FGLD_H_INCLUDED
#define FG_FGLD_H_INCLUDED

/*****************************************************************************
 * linear framebuffer access for 5-6-5 truecolor visual
 */
#if defined (TARGET_CPU_I386)
#include "fgld_16_i386.h"
#endif /* defined TARGET_CPU_I386 */

#if defined (TARGET_CPU_ARM)
#include "fgld_16_arm.h"
#endif /* defined TARGET_CPU_ARM */

#if defined (TARGET_CPU_MIPS)
#include "fgld_16_mipsel.h"
#endif /* defined TARGET_CPU_MIPS */

/***************************************************************************
 * public functions
 */


static inline
u8 _fgl_getSlcSize () /*{{{*/
{
	return 2;
}
/*}}}*/

// return length of scanline in cell units
static inline
int _fgl_getScanlineLength ( int length ) /*{{{*/
{
	return length;
}
/*}}}*/

static inline
fgl_slc_t* _fgl_getSlc ( fgl_bitmap_t* bmp, u32 x, u32 y ) /*{{{*/
{
  u8* surface;
  if ( bmp->state & IS_SHAREDMEM ) {
       surface = (u8 *)bmp + sizeof ( fgl_bitmap_t );
       return ((fgl_slc_t *)surface + y * bmp->sLength + x );
  }
  else {
       return (bmp->surface + (y * bmp->sLength) + x);
  }
}
/*}}}*/

static inline
void _fgl_incSlc ( fgl_slc_t **p, u32 x ) /*{{{*/
{
	(*p)++;
}
/*}}}*/

static inline
void _fgl_decSlc ( fgl_slc_t **p, u32 x ) /*{{{*/
{
	(*p)--;
}
/*}}}*/

static inline
void _fgl_addSlc ( fgl_slc_t **p, u32 x, i32 pixels ) /*{{{*/
{
	(*p) += pixels;
}
/*}}}*/

static inline
void _fgl_subSlc ( fgl_slc_t **p, u32 x, i32 pixels ) /*{{{*/
{
	(*p) -= pixels;
}
/*}}}*/

static inline
fgl_color_t _fgl_getSlcPixel ( fgl_slc_t *p, u32 x )  /*{{{*/
{
	return (fgl_color_t) *p;
}
/*}}}*/

static inline
void _fgl_setSlcPixel ( fgl_slc_t *p, u32 x, fgl_color_t clr )  /*{{{*/
{
	*p = clr;
}
/*}}}*/

static inline
void _fgl_setSlcPixelXor ( fgl_slc_t *p, u32 x, fgl_color_t clr ) /*{{{*/
{
	*p ^= clr;
}
/*}}}*/

#ifndef HAVE_SETSLCCOLS
static inline
void _fgl_setSlcCols ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	if ( len < 32 ){
		fgl_slc_t *pmax = p + len;

		while ( p < pmax  ){
			*p++ = clr;
		}
	}
	else {
		/* do a little loop unrolling */
		u32 *pp;
		u32 cc = ((u32)clr << 16) | clr;

		if ( x & 1 ){ 			/* align */
			*p++ = clr;
			len--;
		}

		pp = (u32*) p;
		while (len > 15){
 			*pp++ = cc;
			*pp++ = cc;
 			*pp++ = cc;
			*pp++ = cc;
 			*pp++ = cc;
			*pp++ = cc;
 			*pp++ = cc;
			*pp++ = cc;

			len -= 16;
		}

		if ( len ){
			p = ((fgl_slc_t*)pp);
			while ( len-- ){
				*p++ = clr;
			}
		}
	}
}
/*}}}*/
#endif /* HAVE_SETSLCCOLS */

static inline
void _fgl_setSlcColsXor ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	fgl_slc_t *pmax = p + len;

	while ( p < pmax  ){
		*p++ ^= clr;
	}
}
/*}}}*/

static inline
void _fgl_setSlcRows ( fgl_slc_t *p, i32 inc, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	do {
		*p = clr;
		p += inc;
	} while ( --len );
}
/*}}}*/

static inline
void _fgl_setSlcRowsXor ( fgl_slc_t *p, i32 inc, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	do {
		*p ^= clr;
		p += inc;
	} while ( --len );
}
/*}}}*/

#ifndef HAVE_INITBMP
static inline
void _fgl_initBmp ( fgl_bitmap_t* bmp, i32 clr ) /*{{{*/
{
	fgl_slc_t* p = bmp->surface;
	u32 len = bmp->width * bmp->height;
	u32 *pp;
	u32 cc = ((u32)clr << 16) | (clr & 0xffff);

	pp = (u32*) p;
	while (len > 15){
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		
		len -= 16;
	}

	if ( len ){
		p = ((fgl_slc_t*)pp);
		while ( len-- ){
			*p++ = (fgl_slc_t)clr;
		}
	}
}
/*}}}*/
#endif /* HAVE_INITBMP */


#ifndef HAVE_CPYSLC
static inline
void _fgl_cpySlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len ) /*{{{*/
{
	memcpy( dest, src, len*2); 
}
/*}}}*/
#endif /* HAVE_CPYSLC */

#ifndef HAVE_MOVESLC
static inline
void _fgl_moveSlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len ) /*{{{*/
{
	memmove( dest, src, len*2);
}
/*}}}*/
#endif /* HAVE_MOVESLC */
#endif /* FGL_FGLD_H_INCLUDED */
