/*
 * font.c - fgl module for generic font handling (load/unload
 *          display, metrics)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "font.h"

static int delta[] = { 0, -1, 1, -2, 2 };

fgl_font_t* fgl_getFontSet ( char* fname, u8 ptSize ) 
{
  fgl_fontEntry_t *fe;
  fgl_font_t *fnt = NULL;
  int i;

  if ( !_fgl_fontInitialized && !(_fgl_fontInitialized = fgl_initFontEngine()) )
    {
      warnx("font initialization failed\n");
      return 0;
    }

  if ( fname )
    {
      /* find a match in the table */
      for ( i=0; i < sizeof(delta)/sizeof(delta[0]); i++ )
        {
          for ( fe = _fgl_setCache; fe; fe = fe->next )
            {
              if ( (*fname == *fe->fname) &&
                  !strcmp( fname, fe->fname) && (ptSize + delta[i] == fe->ptSize) )
                {
                  fe->fnt->nRefs++;
                  return fe->fnt;
                }
            }
          if ( (fnt = fgl_loadFont( fname, ptSize + delta[i])) )
            {
              fe = (fgl_fontEntry_t*) FGL_MALLOC( sizeof(fgl_fontEntry_t));

              /* update font cache */
              fe->fnt    = fnt;
              fe->fname  = font_strdup( fname);
              fe->ptSize = ptSize;
              fe->next   = _fgl_fontCache;

              _fgl_setCache = fe;

              return fnt;
            }
        }
    }

  return fnt;

}

void fgl_addFontReference ( fgl_font_t* fnt ) /*{{{*/
{
  if ( fnt ) {
      fnt->nRefs++;
  }
}

fgl_font_t* fgl_getFont ( char* fname, u8 ptSize ) 
{
  fgl_fontEntry_t *fe;
  fgl_font_t *fnt = NULL;
  int i,j;

  if ( !_fgl_fontInitialized && !(_fgl_fontInitialized = fgl_initFontEngine()) )
    {
      warnx("font initialization failed\n");
      return 0;
    }

  if ( fname )
    {
      /* find a match in the table */
      for ( i=0; i < sizeof(delta)/sizeof(delta[0]); i++ )
        {
          for ( fe = _fgl_fontCache; fe; fe = fe->next )
            {
              if ( (*fname == *fe->fname) &&
                  !strcmp( fname, fe->fname) && (ptSize + delta[i] == fe->ptSize) )
                {
                  fe->fnt->nRefs++;
                  return fe->fnt;
                }
            }
          if ( (fnt = fgl_loadFont( fname, ptSize + delta[i])) )
            {
              fe = (fgl_fontEntry_t*) FGL_MALLOC( sizeof(fgl_fontEntry_t));

              /* update font cache */
              fe->fnt    = fnt;
              fe->fname  = font_strdup( fname);
              fe->ptSize = ptSize;
              fe->next   = _fgl_fontCache;

              _fgl_fontCache = fe;

              if (_fgl_setInitialized )
                {
                  fnt->set_number = _fgl_setCount;
                  fnt->set = (struct _fgl_font_t **) FGL_MALLOC( _fgl_setCount * sizeof(struct _fgl_font_set *));
                  for ( j = 0 ; j < _fgl_setCount; j++ )
                    {
                      fnt->set[j] = (struct _fgl_font_t *) fgl_getFont( _fgl_setName[j], ptSize );
                    }
                }
              
              return fnt;
            }
        }
      if ( fgl_defaultFont )
        {
          fnt = fgl_defaultFont;
          fnt->nRefs++;
        } 
      else 
        {
          fnt = fgl_defaultFont = fgl_loadFont( "default", ptSize);
        }
      if ( fnt )
        {
          fe = (fgl_fontEntry_t*) FGL_MALLOC( sizeof(fgl_fontEntry_t));

          /* update font cache */
          fe->fnt    = fnt;
          fe->fname  = font_strdup( fname);
          fe->ptSize = ptSize;
          fe->next   = _fgl_fontCache;

          _fgl_fontCache = fe;
        }
      return fnt;
    }
    
  if ( fgl_defaultFont )
    {
      fnt = fgl_defaultFont;
      fnt->nRefs++;
    }
  else
    {
      fnt = fgl_defaultFont = fgl_loadFont( "default", ptSize);
    }

  return fnt;

}

int fgl_getMaxAdvance ( fgl_font_t* fnt ) 
{
	int i, a = 0;
	/*int j;*/
	if (fnt->maxAdvance != -1) return fnt->maxAdvance;
	for ( i=0; i<fnt->nGlyphs; i++ ){
		if ( fnt->advance[i] > a ){
			a = fnt->advance[i];
		}
	}
	fnt->maxAdvance = a;
	return a;
}

void fgl_disposeFont ( fgl_font_t* fnt ) /*{{{*/
{
	fgl_fontEntry_t *fe, *fe1=0, *fe2;
	
	if ( fnt && fnt->nRefs ){
		fnt->nRefs--;
		
		if ( fnt->nRefs == 0 ){
			/* free the font and all it's entries */
			for ( fe = _fgl_fontCache; fe; ){
				if ( fe->fnt == fnt ){
					if ( fe == _fgl_fontCache ){
						_fgl_fontCache = fe->next;
					}
					else {
						fe1->next = fe->next;
					}

					fe2 = fe;
					fe = fe->next;

					FGL_FREE( fe2->fname);
					FGL_FREE( fe2);
				}
				else {
					fe1 = fe;
					fe = fe->next;
				}
			}

			fgl_unloadFont( fnt);
		}
	}

}

void fgl_drawText ( fgl_context_t* ctx, i32 x, i32 y, char* text, i32 len ) 
{
  	char* s;
        u8 *slength;
	int  i;
	fgl_font_t *fnt = ctx->fnt;
	fgl_a8Bitmap_t bmp1;
        fgl_monoBitmap_t bmp2;

	if ( !fnt || !text )
		return;

        if ( fnt->state & AAF_8 ) 
          {
            fgl_rotateA8Font( fnt );
            bmp1.state = fnt->state & 0x3;
          }
        else 
          {
            fgl_rotateMFont( fnt );
            bmp2.state = fnt->state & 0x3;
          }
          
          {
            SET_ROTATION( ctx->bmp );
            if (rot & 1)
              {
                slength = fnt->height;
              }
            else
              {
                slength = fnt->width;
              }
          }

        for ( s=text; (u32)len && *s; s++, len-- ){
            i = (*s - fnt->uniOffset);
            if ( fnt->state & AAF_8 ) {
                if ( (bmp1.width = fnt->width[i]) ){
                    bmp1.surface = fnt->surface + fnt->surfOff[i];
                    bmp1.height  = fnt->height[i];
                    bmp1.sLength = slength[i];
                    fgl_a8BltBmp( ctx,
                        x + fnt->xBearing[i],
                        y - fnt->yBearing[i],
                        &bmp1, ctx->clr);
                }
            }
            else {
                if ( (bmp2.width = fnt->width[i]) ){
                    bmp2.surface = fnt->surface + fnt->surfOff[i];
                    bmp2.height  = fnt->height[i];
                    bmp2.sLength = MONO_SLENGTH( slength[i] );
                    fgl_monoBltBmp( ctx,
                        x + fnt->xBearing[i],
                        y - fnt->yBearing[i],
                        &bmp2, ctx->clr);
                }
            }
            x += fnt->advance[i];
        }
}

void fgl_drawWText ( fgl_context_t* ctx, i32 x, i32 y, u16* text, u32 len ) 
{
	u16  *s, i;
	fgl_font_t *fnt = ctx->fnt;
	fgl_a8Bitmap_t bmp1;
        fgl_monoBitmap_t bmp2;
	u16  j;

	if ( !fnt || !text )
		return;

        if ( fnt->state & AAF_8 ) 
          {
            fgl_rotateA8Font( fnt );
            bmp1.state = fnt->state & 0x3;
            if ( _fgl_setInitialized )
              {
                for ( i =0 ; i < _fgl_setCount; i++ )
                  {
                    fgl_rotateA8Font( (fgl_font_t *)(fnt->set[i]) );
                  }
              }
          }
        else 
          {
            fgl_rotateMFont( fnt );
            bmp2.state = fnt->state & 0x3;
            if ( _fgl_setInitialized )
              {
                for ( i =0 ; i < _fgl_setCount; i++ )
                  {
                    fgl_rotateMFont( (fgl_font_t *)(fnt->set[i]) );
                  }
              }
          }

	for ( s=(u16*)text; len && *s; s++, len-- ){
          fnt = ctx->fnt;
	  i = *s - fnt->uniOffset;
	  if ( i >= fnt->nGlyphs ) 
            {
              i = 0;
              for ( j = 0 ; j < fnt->set_number ; j++ )
                {
                  if ( *s >= fnt->set[j]->uniOffset && *s - fnt->set[j]->uniOffset < fnt->set[j]->nGlyphs )
                    {
                      fnt = (fgl_font_t *)fnt->set[j];
                      i = *s - fnt->uniOffset;
                    }
                }
            }
          if ( fnt->state & AAF_8 ) {
              if ( (bmp1.width = fnt->width[i]) ){
                  bmp1.surface = fnt->surface + fnt->surfOff[i];
                  bmp1.height  = fnt->height[i];
                  bmp1.sLength = fnt->sLength[i];
                  fgl_a8BltBmp( ctx,
                      x + fnt->xBearing[i],
                      y - fnt->yBearing[i],
                      &bmp1, ctx->clr);
              }
          }
          else {
              if ( (bmp2.width = fnt->width[i]) ){
                  bmp2.surface = fnt->surface + fnt->surfOff[i];
                  bmp2.height  = fnt->height[i];
                  bmp2.sLength = MONO_SLENGTH( fnt->sLength[i] );
                  fgl_monoBltBmp( ctx,
                      x + fnt->xBearing[i],
                      y - fnt->yBearing[i],
                      &bmp2, ctx->clr);
              }
          }
          x += fnt->advance[i];
        }
}

void fgl_drawWVertText ( fgl_context_t* ctx, i32 x, i32 y, u16* text, u32 len ) 
{
	u16  *s, i,j;
	fgl_font_t *fnt = ctx->fnt;
	fgl_a8Bitmap_t bmp1;
	fgl_monoBitmap_t bmp2;
	int xoff;
	int xadvance;
	y += fnt->fontHeight;
	xadvance = fgl_getMaxAdvance(fnt);

	if ( !fnt || !text )
		return;

        if ( fnt->state & AAF_8 ) 
          {
            fgl_rotateA8Font( fnt );
            bmp1.state = fnt->state & 0x3;
            if ( _fgl_setInitialized )
              {
                for ( i =0 ; i < _fgl_setCount; i++ )
                  {
                    fgl_rotateA8Font( (fgl_font_t *)(fnt->set[i]) );
                  }
              }
          }
        else 
          {
            fgl_rotateMFont( fnt );
            bmp2.state = fnt->state & 0x3;
            if ( _fgl_setInitialized )
              {
                for ( i =0 ; i < _fgl_setCount; i++ )
                  {
                    fgl_rotateMFont( (fgl_font_t *)(fnt->set[i]) );
                  }
              }
          }
        
	for ( s=(u16*)text; len && *s; s++, len-- ){
	  i = *s - fnt->uniOffset;
	  if ( i >= fnt->nGlyphs ) 
            {
              i = 0;
              for ( j = 0 ; j < fnt->set_number ; j++ )
                {
                  if ( *s >= fnt->set[j]->uniOffset && *s - fnt->set[j]->uniOffset < fnt->set[j]->nGlyphs )
                    {
                      fnt = (fgl_font_t *)fnt->set[j];
                      i = *s - fnt->uniOffset;
                    }
                }
            }
          if ( fnt->state & AAF_8 ) {
              if ( (bmp1.width = fnt->width[i]) ){
                  xoff = ( xadvance - fnt->width[i] - fnt->xBearing[i])/2;
                  bmp1.surface = fnt->surface + fnt->surfOff[i];
                  bmp1.height  = fnt->height[i];
                  bmp1.sLength = fnt->sLength[i];
                  fgl_a8BltBmp( ctx,
                      x + fnt->xBearing[i]+xoff,
                      y - fnt->yBearing[i],
                      &bmp1, ctx->clr);

              }
          }
          else {
              if ( (bmp2.width = fnt->width[i]) ){
                  xoff = ( xadvance - fnt->width[i] - fnt->xBearing[i])/2;
                  bmp2.surface = fnt->surface + fnt->surfOff[i];
                  bmp2.height  = fnt->height[i];
                  bmp2.sLength = MONO_SLENGTH( fnt->sLength[i]);
                  fgl_monoBltBmp( ctx,
                      x + fnt->xBearing[i]+xoff,
                      y - fnt->yBearing[i],
                      &bmp2, ctx->clr);

              }
          }
          y += fnt->fontHeight;
        }
}

int fgl_getTextLength ( fgl_font_t* fnt, char* text, u32 len ) 
{
	int w = 0;
	u16 i;
	char *s;
	
	for ( s=text; len && *s; s++, len-- ) {
		i = *s - fnt->uniOffset;
		if ( i < fnt->nGlyphs ){
			w += fnt->advance[i];
		}
	}

	return w;
}

int fgl_getWTextLength ( fgl_font_t* fnt, u16* text, u32 len ) 
{
	int w = 0;
	u16 *s, i;
	for ( s=(u16*)text; len && *s; s++, len-- ) {
		i = *s - fnt->uniOffset;
		if ( i < fnt->nGlyphs ){
			w += fnt->advance[i];
		}
	}
	return w;
}

int fgl_getFixedWidth ( fgl_font_t* fnt ) 
{
	int i;
        int a = fnt->advance[1];

	for ( i=1; i<fnt->nGlyphs; i++ ){
		if ( fnt->advance[i] != a ){
			return 0;
		}
	}
	return a;
}

