/*
 * init_fb.c - FGL_INIT_MODULE to initialize the framebuffer device and
 *             global fgl variables (Linux version)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */
#include "libfgl.h"
#include "../include/fgl.h"
#include "accel.h"
#include "fgld_16.h"
#include "cmap_t565.h"

#include <asm/page.h>
#include <linux/fb.h>
#include <sys/user.h>

#if defined(USE_VT_CTRL)
#include <linux/kd.h>
#endif

#define DEFAULT_FB "/dev/fb0"
#define DEFAULT_VT "/dev/stdin"
static int fdvt = -1;

fgl_bitmap_t               _fgl_screenBmpObject;
fgl_bitmap_t               *fgl_screenBmp = &_fgl_screenBmpObject;

fgl_context_t              _fgl_screenCtxObject;
fgl_context_t              *fgl_screenCtx = &_fgl_screenCtxObject;

int                        _fgl_fh;
struct fb_var_screeninfo   _fgl_varInfo;
struct fb_fix_screeninfo   _fgl_fixInfo;
void* scr;

int fgl_getScreenWidth () 
{
  return fgl_screenCtx->bmp->width;
}

int fgl_getScreenHeight () 
{
  return fgl_screenCtx->bmp->height;
}

int fgl_getScreenXoff () 
{
  return fgl_screenCtx->xoff;
}

int fgl_getScreenYoff ()
{
  return fgl_screenCtx->yoff;
}

/*
 * fgl_terminate - clean up system resources
 * 
 * This is automatically called during process exit, so don't inline.
 * It also has to be able to detect multiple calls
 */
void fgl_terminate () 
{
  char str[20];
  
  if ( scr ){
      munmap( scr, _fgl_fixInfo.smem_len);
      scr = 0;
  }

  if ( _fgl_fh >= 0 ) {
      close( _fgl_fh);
      _fgl_fh = -1;
  }

  if (fdvt >= 0) {
#if defined(USE_VT_CTRL)
      if (ioctl(fdvt, KDSETMODE, KD_TEXT) < 0) {
          warnx( "failed to set KD_TEXT mode\n");
      }
#endif
      sprintf(str, "\033[?25h"); // Turn on cursor
      write(fdvt, str, 6);
      sprintf(str, "\0338"); // Restore saved state
      write(fdvt, str, 2);
      sprintf(str, "\033[9;15]"); // Set screen blank timeout
      // to 15 minutes
      write(fdvt, str, 7);
      close(fdvt);
      fdvt = -1;
  }
}

/*
 * This is our color mapping set up defined in the cmap_XX module. We
 * can't inline because cmap uses our colormap functions (which are
 * device dependent)
 */
extern int  fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld );

/*
 * fgl_init - initialize system resources, screenBmp and color map
 * 
 *   nReq    : number of requested RGB cells (pseudocolor only)
 *   rgbReqs : 888 RGB values of requested colors (pseudocolor only)
 *   restore : 1: restore colormap when process exits (pseudocolor only)
 */
static inline
int _fgl_init ( int nReq, u32* rgbReqs, int restore ) 
{
  char  *fbdev;
  fgl_config_t cfg;

  char  *vtdev;
  char  str[20];

  vtdev = getenv( "VTCTRL");
  if ( !vtdev ){
      vtdev = DEFAULT_VT;
  }
  fdvt = open(vtdev, O_RDWR|O_NDELAY, 0);
  if (fdvt < 0) {
      warnx( "failed to open device: '%s'\n", vtdev);
      return 0;
  }
  sprintf(str, "\0337"); // Save state
  write(fdvt, str, 2);
  sprintf(str, "\033[9;0]"); // Set screen blank timeout
  // to 0 minutes (disable)
  write(fdvt, str, 6);
  sprintf(str, "\033[?25l"); // Turn off cursor
  write(fdvt, str, 6);
                           
#if defined(USE_VT_CTRL)
  if (ioctl(fdvt, KDSETMODE, KD_GRAPHICS) < 0) {
      warnx( "failed to set KD_GRAPHICS mode\n");
      return 0;
  }
#endif

  fbdev = getenv( "FRAMEBUFFER");
  if ( !fbdev ){
      fbdev = DEFAULT_FB;
  }

  if ( (_fgl_fh = open( fbdev, O_RDWR)) < 0 ){
      warnx( "failed to open device: '%s'\n", fbdev);
      return 0;
  }

  /* get framebuffer info */
  if ( ioctl( _fgl_fh, FBIOGET_VSCREENINFO, &_fgl_varInfo) ){
      warnx( "FBIOGET_VSCREENINFO failed: %s\n", strerror(errno));
      return 0;
  }

  if ( ioctl( _fgl_fh, FBIOGET_FSCREENINFO, &_fgl_fixInfo) ){
      warnx( "FBIOGET_FSCREENINFO failed: %s\n", strerror(errno));
      return 0;
  }

  /* check if it's compatible with our build-settings */
  if ( _fgl_varInfo.bits_per_pixel != fgl_getBitsPerPixel() ){
      warnx( "wrong framebuffer driver %d (display is %d) ",
          fgl_getBitsPerPixel(), _fgl_varInfo.bits_per_pixel);
      return 0;
  }

  /* map framebuffer device into our memory */
  scr = mmap( NULL, _fgl_fixInfo.smem_len,
      PROT_READ | PROT_WRITE, MAP_SHARED,
      _fgl_fh, 0);
  if ( scr == MAP_FAILED ){
      puts( "framebuffer mmap failed");
      return 0;
  }

  if ( atexit( fgl_terminate) == -1 ){
      puts( "failed to install exit handler");
  }

  /*
  * in case of MobilePro the actual memory address of the fb 
  * is not aligned on the page boundary. Calling mmap therefore
  * returns memory aligned address to which we have to use
  * offset based on the actual memory location. The only case
  * so far is MobilePro 780 but we will do it for the benefit
  * of other new platforms that may exhibit same problem
   */
  fgl_screenBmp->surface = (fgl_slc_t*) ((char *)scr + ((long)_fgl_fixInfo.smem_start % (int)PAGE_SIZE));
  fgl_screenBmp->width   = _fgl_varInfo.xres_virtual; /* might be bigger than xres */
  fgl_screenBmp->height  = _fgl_varInfo.yres_virtual;
  fgl_screenBmp->state  |= IMAGE_ON_SCREEN;
  /*
  * The actual length of the fb line and screen with may differ. The line
  * length calcualtion should be based therefore on the fb linegth returned
  * in fixScreenInfo structure. for multiple byte pixels sLnegth holds number
  * of pixels, otherwise it holds numnber of bytes.
   */

  fgl_screenBmp->sLength = _fgl_fixInfo.line_length * 8 / _fgl_varInfo.bits_per_pixel;

  fgl_screenCtx->bmp     = fgl_screenBmp;
  fgl_screenCtx->alphar   = 0xff;
  fgl_screenCtx->alphag   = 0xff;
  fgl_screenCtx->alphab   = 0xff;
  fgl_screenCtx->x1Clip  = fgl_screenBmp->width-1;
  fgl_screenCtx->y1Clip  = fgl_screenBmp->height-1;
  fgl_screenCtx->xoff    = 0;
  fgl_screenCtx->yoff    = 0;
  
/* set up any color mapping we might need */
  fgl_initRgbCmap ( nReq, rgbReqs, restore );

  /* reset the screen dimensions if config file exists */
  if (fgl_readConfig( &cfg)) {
      if ( cfg.xoff != -1 && cfg.yoff != -1 &&
          cfg.width != -1 && cfg.height != -1 ) {

          fgl_screenBmp->surface += cfg.xoff + 
          cfg.yoff * fgl_screenBmp->sLength;
          fgl_screenBmp->width   = cfg.width;
          fgl_screenBmp->height  = cfg.height;


          fgl_screenCtx->xoff    = cfg.xoff;
          fgl_screenCtx->yoff    = cfg.yoff;

          fgl_screenCtx->x1Clip  = fgl_screenBmp->width-1;
          fgl_screenCtx->y1Clip  = fgl_screenBmp->height-1;
      }
  }
  /* run accel init if any */
#ifdef USE_ACCEL
  if ( fgl_accel_init )
    {
      fgl_accel_init( );
    }
#endif /* USE_ACCEL */
  return 1;
}


int fgl_init ( int nReq, u32* rgbReqs, int restore )
{
  return _fgl_init ( nReq, rgbReqs, restore );
}

int fgl_initialize () 
{	
	return _fgl_init( 0, 0, 1);
}

/*
 * fgl_initialize - convenience wrapper for fgl_init w/o RGB requests
 * specifying size
 */
int fgl_open (u32 x, u32 y, u32 w, u32 h) 
{	
  int r;

  r = fgl_init( 0, 0, 1);

  fgl_screenBmp->width   = w; 
  fgl_screenBmp->height  = h;

  fgl_screenCtx->x1Clip  = w-1;
  fgl_screenCtx->y1Clip  = h-1;
  
  fgl_screenBmp->surface = _fgl_getSlc( fgl_screenBmp, x, y);
  return r;
}
/*}}}*/

/****************************************************************************
 * These are the low level functions to manage device specific colormaps.
 * Note that we can't inline them, because they are used from the cmap_XX
 * module, which is included
 */

static inline
int fgl_hasCmap () /*{{{*/
{
#if defined (IS_CMAP_MODE)
  return 1;
#else
  return 0;
#endif
}/*}}}*/

fgl_cmap_t* fgl_allocCmap () /*{{{*/
{
  fgl_cmap_t* cm = 0;

#if defined (IS_CMAP_MODE)
  int i, n;

  if ( fgl_hasCmap() ){
      for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
      cm = FGL_CALLOC( 1, sizeof(fgl_cmap_t) + 3 * n);

      cm->r = (u8*)cm + sizeof(fgl_cmap_t);
      cm->g = cm->r + n;
      cm->b = cm->g + n;
  }
#endif /* IS_CMAP_MODE */

  return cm;
}

void fgl_freeCmap ( fgl_cmap_t* cm ) /*{{{*/
{
	FGL_FREE( cm);
}

int fgl_getHWCmap ( fgl_cmap_t* cm ) /*{{{*/
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i, n;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;

	for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
	cmap.len = n;

	buf = (u16*) FGL_MALLOC( n * 3 * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + n;
	cmap.blue  = cmap.green + n;
	cmap.transp = 0;

	if ( ioctl( _fgl_fh, FBIOGETCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}

	cm->len = cmap.len;
	for ( i=0; i<cmap.len; i++ ){
		cm->r[i] = cmap.red[i] >> 8;
		cm->g[i] = cmap.green[i] >> 8;
		cm->b[i] = cmap.blue[i] >> 8;
	}

	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}

int fgl_setHWCmap ( fgl_cmap_t* cm ) /*{{{*/
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;
	cmap.len = cm->len;

	buf = (u16*) FGL_MALLOC( 3 * cm->len * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + cmap.len;
	cmap.blue  = cmap.green + cmap.len;
	cmap.transp = 0;

	for ( i=0; i<cm->len; i++ ){
		cmap.red[i]   = (u16)(cm->r[i]) << 8;
		cmap.green[i] = (u16)(cm->g[i]) << 8;
		cmap.blue[i]  = (u16)(cm->b[i]) << 8;
	}

	if ( ioctl( _fgl_fh, FBIOPUTCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}
	
	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}
