/*
 * mbmp.c - fgl module to handle monochrome bitmaps (depth 1)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "mbmp.h"

u8 _monoBits[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

void fgl_rotateMBitmap( fgl_monoBitmap_t* bmp )
{
#if defined(DYNAMIC)

  u8 brot,srot;
  u8 rotation;
  u8 *old_surface, *new_surface;
  u8 pix;
  u32 slength2;
  u32 w,h,u,v,x,y;

  u32 size = ( bmp->width + 7 ) * ( bmp->height + 7 ) / 8;

  /* we can always assume that the image is rotated from its current 
  *    * to the screen orientation
  *       */
  srot = fgl_screenBmp->state & 0x3;
  brot = bmp->state & 0x3;

  if ( srot == brot ) return; /* if nothing ton do do not waste time */

  rotation = ( 4 + srot - brot ) % 4; /* calculate the counter-clockwise */

  new_surface = (u8 *) FGL_MALLOC ( size );
  old_surface = _fgl_getMonoSurface( bmp );

  /* setup the slengths for the rotation 
  *    * slength1 is the slength of the original image
  *       * and slength2 is for the rotated one.
  *          */

  if ( brot & 1) {
      w = bmp->height;
      h = bmp->width;
  }
  else {
      h = bmp->height;
      w = bmp->width;
  }

  switch( rotation ) {
    case 1:
      slength2 = ( h + 7 ) / 8;
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              pix = old_surface[ y * bmp->sLength + x / 8 ] & _monoBits[ x % 8 ] ;
              u = h - y - 1;
              v = x;
              if ( pix )
                {
                  new_surface[ v * slength2 + u / 8 ] |= _monoBits[ u % 8];
                }
              else
                {
                  new_surface[ v * slength2 + u / 8 ] &= ~_monoBits[ u % 8 ];
                }
          }
      }
      bmp->sLength = slength2;
      break;
    case 2:
      slength2 = ( w +7 ) / 8;
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              pix = old_surface[ y * bmp->sLength + x / 8 ] & _monoBits[ x % 8 ] ;
              u = w - x - 1;
              v = h - y - 1;
              if ( pix )
                {
                  new_surface[ v * slength2 + u / 8 ] |= _monoBits[ u % 8];
                }
              else
                {
                  new_surface[ v * slength2 + u / 8 ] &= ~_monoBits[ u % 8 ];
                }
          }
      }
      break;
    case 3:
      slength2 = ( h + 7 ) / 8;
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              pix = old_surface[ y * bmp->sLength + x / 8 ] & _monoBits[ x % 8 ] ;
              u = y;
              v = w - x - 1;
              if ( pix )
                {
                  new_surface[ v * slength2 + u / 8 ] |= _monoBits[ u % 8];
                }
              else
                {
                  new_surface[ v * slength2 + u / 8 ] &= ~_monoBits[ u % 8 ];
                }
          }
      }
      bmp->sLength = slength2;
      break;
    default:
      return;
  }
  if ( bmp->state & IS_SHAREDMEM )
    {
      memcpy( old_surface, new_surface, size );
      FGL_FREE( new_surface );
    }
  else
    {
      FGL_FREE( bmp->surface );
      bmp->surface = new_surface;
    }
  BMP_SET_ROTATION( bmp, srot );

#endif /* DYNAMIC */
}

void fgl_setMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) 
{
    {
      SET_ROTATION( ((fgl_bitmap_t *)bmp) );
      POINT_ROTATE( rot, (fgl_bitmap_t *)bmp, col, row );
    }

  _fgl_setMonoBit( bmp, col, row );
}

void fgl_unsetMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) 
{
    {
      SET_ROTATION( ((fgl_bitmap_t *)bmp) );
      POINT_ROTATE( rot, (fgl_bitmap_t *)bmp, col, row );
    }

  _fgl_unsetMonoBit ( bmp, col, row ); 
}

int fgl_isMonoBitSet ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) 
{
    {
      SET_ROTATION( ((fgl_bitmap_t *)bmp) );
      POINT_ROTATE( rot, (fgl_bitmap_t *)bmp, col, row );
    }
  return _fgl_isMonoBitSet ( bmp, col, row );
}

fgl_monoBitmap_t* fgl_createMonoBitmap ( u32 width, u32 height, int init ) 
{
  u32 size =  ( width + 7 ) * ( height + 7 ) / 8;
  fgl_monoBitmap_t *bmp = (fgl_monoBitmap_t *) FGL_MALLOC ( sizeof ( fgl_monoBitmap_t ));
  if ( !bmp )
    {
      return 0;
    }

  bmp->surface = (u8 *) FGL_MALLOC ( sizeof ( u8 ) * size );
  bmp->width  = width;
  bmp->height = height;
  bmp->state = 0;
  if ( init == -1 )
    {
      SET_ROTATION( fgl_screenBmp );
      BMP_SET_ROTATION( ((fgl_bitmap_t *)bmp), rot );
      
      if ( rot & 1 )
        {
          bmp->sLength = (height + 7) / 8;
        }
      else
        {
          bmp->sLength = (width + 7) / 8;
        }
    }
  else
    {
      BMP_SET_ROTATION( ((fgl_bitmap_t *)bmp), 0 );
      bmp->sLength = (width + 7) / 8;
      memset ( bmp->surface, init, size );
      fgl_rotateMBitmap( bmp ); 
    }

  return bmp;
}

void fgl_disposeMonoBitmap ( fgl_monoBitmap_t* bmp ) 
{
  FGL_FREE ( bmp->surface );
  FGL_FREE ( bmp );
}

void fgl_monoBlt ( fgl_context_t* dst, i32 u0, i32 v0,
    fgl_monoBitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, 
    fgl_color_t clr ) 
{
  u8 *ps;
  fgl_slc_t *pd;
  i32 u, u1, v1, sinc, dinc;
  i32 w, h;
  register u8 a, i;
  u8 i0, off;

  if ( x0 > x1 )  swap32( &x0, &x1);
  if ( y0 > y1 )  swap32( &y0, &y1);

  if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
        src->width, src->height, &x0, &y0, &x1, &y1) ){
      return;
  }

  fgl_rotateMBitmap( src );
  fgl_rotateCBitmap( dst->bmp );

    {
      SET_ROTATION( ((fgl_bitmap_t *)src) );
      RECT_ROTATE( rot, ((fgl_bitmap_t *)src), x0, y0, x1, y1 );
    }

    {
      SET_ROTATION( dst->bmp);
      RECT_ROTATE( rot, dst->bmp, u0, v0, u1, v1 );
    }

#ifdef USE_ACCEL
  if ( fgl_accel_monoBlt && ( dst->bmp->state & IMAGE_ON_SCREEN )) 
    {
      fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_ACCEL_MONOBLT_OP );
      fgl_accel_monoBlt( fgl_screenCtx, u0, v0, src, x0, y0, x1, y1, clr);
      fgl_endDraw( dst->bmp, u0, v0, u1, v1);
      return ;
    }
#endif /* USE_ACCEL */

  w = x1 - x0;
  h = y1 - y0;
  off = x0 % 8;

  ps = src->surface + (y0 * src->sLength) + (x0 / 8);
  pd = _fgl_getSlc( dst->bmp, u0, v0);

  a = *ps;
  u = u0;
  u1 = u0 + w;
  i = i0 = 8-off;

  dinc = (dst->bmp->sLength - u1) + u0;                 /* in pixels */
  sinc = src->sLength + (x0>>3) - (x1>>3);  /* in bytes */

  fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

  while ( h-- >= 0 ){
      a <<= off;

      while ( 1 ){
          if ( a & 0x80 ){
              _fgl_setSlcPixel( pd, u, clr);
          }

#ifdef DEBUG_MBMP
          putchar( (a & 0x80) ? 'O' : '.');
          if ( u == u1 ) putchar( '\n');
#endif

          if ( u == u1 ){
              break;
          }

          if ( --i == 0 ){
              /* next source byte */
              i = 8;
              a = *(++ps);
          }
          else {
              a <<= 1;
          }

          _fgl_incSlc( &pd, u);
          u++;
      }

      /* next source and destination scanlines */
      _fgl_addSlc( &pd, u, dinc);
      ps += sinc;
      a = *ps;
      u = u0;
      i = i0;
  }

  fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}


