/*
 * poly.h - fgl module to draw and fill polygons
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_POLY_H_INCLUDED
#define FGL_POLY_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>
#include <limits.h>

/* module specific includes */
#include "draw.h"
#include "rot.h"

typedef struct _edge_t { 
	int ymin;
	int ymax;
	int xmin;
	int x;
	int dx2, dy;
	int k;
	struct _edge_t *nextYmin;
	struct _edge_t *nextScan;
	struct _edge_t *nextActive;
} edge_t;

static inline
void boundingBoxPoly ( fgl_point_t* points, u32 nPoints,
					   int* xmin, int* ymin, int* xmax, int* ymax )
{
	u32 i;

	*xmin = *xmax = points[0].x;
	*ymin = *ymax = points[0].y;

	for ( i=1; i<nPoints; i++ ){
		if ( points[i].x < *xmin )
			*xmin = points[i].x;
		else if ( points[i].x > *xmax )
			*xmax = points[i].x;

		if ( points[i].y < *ymin )
			*ymin = points[i].y;
		else if ( points[i].y > *ymax )
			*ymax = points[i].y;		
	}
}

static inline
int nextIdx ( int i, int n ) 
{
	if ( ++i == n ) return 0;
	else            return i;
}

static inline
int prevIdx ( int i, int n ) 
{
	if ( --i < 0 ) return n-1;
	else           return i;
}


static inline
int isMonotoneInc ( int i, fgl_point_t* pt, u32 nPt ) 
{
	int j;
	for ( j=nextIdx(i, nPt); pt[i].y == pt[j].y; j=nextIdx(j, nPt));
	return ( pt[j].y > pt[i].y );
}

static inline
int isMonotoneDec ( int i, fgl_point_t* pt, u32 nPt ) 
{
	int j;
	for ( j=prevIdx(i, nPt); pt[i].y == pt[j].y; j=prevIdx(j, nPt));
	return ( pt[j].y > pt[i].y );
}

/*
 * Build edge list. Each index holds all edges starting at the corresponding
 * scanline, sorted in increasing x order
 */
static inline
void initEdgeList( fgl_point_t* pt, u32 nPt,
				   edge_t** list, edge_t* edge, int ymin ) 
{
	u32 i, j;
	edge_t *e, *e1, *s, *s1;

	for ( i=0; i<nPt; i++ ) {     /* edge pt[i] -> pt[j] */
		j = nextIdx( i, nPt);
		if ( pt[i].y != pt[j].y ) {     /* skip horizontal lines */
			if ( pt[i].y > pt[j].y ) {  /* up */
				edge->ymin = pt[j].y;
				edge->ymax = pt[i].y;
				edge->x    = pt[j].x;
				edge->dy   = pt[i].y - pt[j].y;
				edge->dx2  = (pt[i].x - pt[j].x) * 2;
				if ( isMonotoneDec( i, pt, nPt) ){
					edge->ymax--;
				}
			}
			else {                              /* down */
				edge->ymin = pt[i].y;
				edge->ymax = pt[j].y;
				edge->x    = pt[i].x;
				edge->dy   = pt[j].y - pt[i].y;
				edge->dx2  = (pt[j].x - pt[i].x) * 2;
				if ( isMonotoneInc( j, pt, nPt) ){
					edge->ymax--;
				}
			}

			edge->xmin = (pt[i].x > pt[j].x) ? pt[j].x : pt[i].x;
			edge->k    = 0;

			/* build edge list in increasing ymin order */
			for ( e1=0, e=*list; e && (e->ymin < edge->ymin); e1=e, e=e->nextYmin );
			if ( !e || (e->ymin > edge->ymin) ) {  /* insert/append new */
				if ( !e1 ){
					edge->nextYmin = *list;
					*list = edge;
				}
				else {
					e1->nextYmin = edge;
					edge->nextYmin = e;
				}
			}
			else if ( e->ymin == edge->ymin ) { /* sort in at same y */
				/* we could do better for 'peaks' (i.e. egde->x == e->x, to avoid subsequent
				 * reordering in updateActiveList), but would need a (scaled) slope for that.
				 * It's not clear if this buys us something, performance-wise (and we have
				 * to order-check in updateActiveList anyway because of crossing lines)
				 */
				
				if ( edge->x < e->x ) {
					/* sort in at beginning */
					if ( e == *list ){
						*list = edge;
					}
					else {
						e1->nextYmin = edge;
					}
					edge->nextYmin = e->nextYmin;
					e->nextYmin = 0;
					edge->nextScan = e;
				}
				else {
					edge->nextYmin = 0;

					for ( s1=e, s=e->nextScan; s && (s->x < edge->x); s1=s, s=s->nextScan );

					if ( !s ) {
						s1->nextScan = edge;
					}
					else {
						s1->nextScan = edge;
						edge->nextScan = s;
					}
				}
			}

			edge++;
		}
	}
}

/*
 * sort node into active list in increasing x order
 */
static inline
void sortInActive ( edge_t** active, edge_t* p ) 
{
	edge_t  *e;
	
	if ( p->x < (*active)->x ){
		p->nextActive = *active;
		*active = p;
	}
	else {
		for ( e=*active; e->nextActive; e = e->nextActive ){
			if ( p->x < e->nextActive->x ){
				p->nextActive = e->nextActive;
				e->nextActive = p;
				break;
			}
		}
		if ( !e->nextActive ){
			e->nextActive = p;
			p->nextActive = 0;
		}
	}
}

/*
 * update active list for the next scanline (y increased by 1):
 *   (1) remove all completed lines (e->ymax < y)
 *   (2) update e->x values for all actives, and re-order active list (if required)
 *   (3) sort in all new edges (y == e->ymin)
 */
static inline
int updateActiveList( edge_t** active, edge_t** nextYmin, int y ) 
{
	edge_t *e, *p;

	/* remove all edges with ymax < scanline i from head of activeList */
	while ( (*active) && ((*active)->ymax < y) ){
		*active = (*active)->nextActive;
	}

	/* do we still have active edges? */
	if ( *active ) {
		/* yep, remove all ymax < scanline edges from tail of activeList */
		for ( e = *active; e->nextActive; ){
			if ( e->nextActive->ymax < y ){
				e->nextActive = e->nextActive->nextActive;
			}
			else {
				e = e->nextActive;
			}
		}

		/* update x for all remaining active edges */
		for ( e = *active, p=0; e; p=e, e = e->nextActive ){
			e->k += e->dx2;
			
			if ( e->dx2 > 0 ){           /* x is increasing */
				while ( e->k > e->dy ){
					e->x++;
					e->k -= (e->dy <<1);
				}
			}
			else {                       /* x is decreasing */ 
				while ( e->k < -e->dy ){
					e->x--;
					e->k += (e->dy <<1);
				}
			}

			/* do we have to re-order (lines might cross) ? */
			if ( p && (e->x < p->x) ) {
				p->nextActive = e->nextActive;
				sortInActive( active, e);
				e = p;
			}
		}
		
		if ( *nextYmin && (y == (*nextYmin)->ymin) ){
			/* sort in new edges from the edgeList */
			for ( p = *nextYmin; p; p = p->nextScan ){
				sortInActive( active, p);
			}
			*nextYmin = (*nextYmin)->nextYmin;
		}
	}
	else {
		if ( *nextYmin ) {
			/* no actives left, start with a new activeList from the edge list */
			/* (but just if this isn't the last line, i.e. (y == ymin)  */
			*active = *nextYmin;
			for ( e = *active; e; e = e->nextScan ){
				e->nextActive = e->nextScan;
			}
			*nextYmin = (*nextYmin)->nextYmin;
		}
	}
	return (*active != 0);
}

#define MAX_EDGES 10

#endif /* FGL_POLY_H_INCLUDED */

