/*
 * sprite.c - fgl module to handle sprites (transparent bitmaps with
 *            screen save/restore) and cursor (global sprite)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "sprite.h"

fgl_sprite_t               *fgl_cursor;
fgl_sprite_t               *fgl_client_cursor;
u8                         _fgl_redrawCursor;
u8                         _fgl_inCursorOp;


void fgl_hideSprite ( fgl_context_t* ctx, fgl_sprite_t* sp ) 
{
  _fgl_hideSprite ( ctx, sp );
}

void fgl_showSprite ( fgl_context_t* ctx, fgl_sprite_t* sp ) 
{	
  _fgl_showSprite ( ctx, sp );
}

fgl_sprite_t* fgl_createSprite ( fgl_spriteData_t *spd ) 
{
	u32         i, x, y;
	fgl_slc_t   *p;
	u8          *q;
	fgl_color_t *cmap;
	fgl_color_t clr;
	fgl_monoBitmap_t *mask = fgl_createMonoBitmap( spd->width, spd->height, 0xff);
	fgl_bitmap_t *bmp = fgl_createBitmap( spd->width, spd->height, -1);
	fgl_sprite_t *sp  = (fgl_sprite_t*) FGL_MALLOC( sizeof(fgl_sprite_t));
#ifdef USE_ACCEL
	fgl_hardware_sprite_t *hsp;
#endif /* USE_ACCEL */
	sp->bmp  = bmp;
	sp->mask = mask;
	sp->sav  = fgl_createBitmap( spd->width, spd->height, -1);
        sp->state = 0;

	cmap = (fgl_color_t*) alloca( sizeof(fgl_color_t) * spd->nColors);
	for ( i=0; i<spd->nColors; i++ ){
		cmap[i] = fgl_getPixelValue( spd->cmap[i]);
	}

	bmp->transp = cmap[spd->transp];

	p = bmp->surface;
	q = spd->data;

	for ( y=0; y<spd->height; y++ ){
		for ( x=0; x<spd->width; x++ ){
			clr = (*q < spd->nColors) ? cmap[*q] : fgl_clrBlack;
			fgl_setBmpPixel( bmp, x, y, clr);
			// fgl_incSlc( &p, x);

			if ( *q == spd->transp ){
				fgl_unsetMonoBit( mask, x, y);
			}
			q++;
		} 
	}

	sp->x0 = sp->y0 = 0;
	sp->x1 = bmp->width-1;
	sp->y1 = bmp->height-1;

	sp->xhot = spd->xhot;
	sp->yhot = spd->yhot;

	sp->isVisible = 0;
#ifdef USE_ACCEL 
	/* 
	 * hardware cursor create function is defined to
	 * to early to be aware of the FGL_MALLOC macro
	 * so we have to allocate the memory now and pass
	 * it to the function.
	 */
        if ( fgl_hardware_createSprite )
          {
            hsp = FGL_MALLOC( sizeof(fgl_hardware_sprite_t));
            fgl_hardware_createSprite( spd, sp, hsp );
          }
#endif /* USE_ACCEL */
        return sp;
}

void fgl_moveSprite ( fgl_context_t* ctx, fgl_sprite_t* sp, i32 x, i32 y ) 
{
	i32 u, v, xw, yh; 
        fgl_bitmap_t *bmp = 0;
        fgl_bitmap_t *sav = 0;

        if ( sp->state & IS_SHAREDMEM ) {
            bmp      = (fgl_bitmap_t *) ((u8 *)sp  + sizeof ( fgl_sprite_t ));
            sav      = (fgl_bitmap_t *) ((u8 *)bmp + sizeof ( fgl_bitmap_t ) +
                bmp->width * bmp->height * sizeof ( fgl_slc_t ));
            xw = sav->width-1;
            yh = sav->height-1;
        } else {
            xw = sp->sav->width-1;
            yh = sp->sav->height-1;
        }

	u = x - sp->xhot;
	v = y - sp->yhot;
	sp->x1 = u+xw;
	sp->y1 = v+yh;

 	/* restore old contents (if it's not the first time) */
	if ( sp->isVisible )
          {
            if ( sp->state & IS_SHAREDMEM )
              {
                fgl_monoBitmap_t *mask = (fgl_monoBitmap_t *) ((u8 *)sav + sizeof ( fgl_bitmap_t ) +
                    sav->width * sav->height * sizeof ( fgl_slc_t ));

                /* restore old area */
                fgl_bitBltBmp( ctx, sp->x0, sp->y0, sav);
                /* save new area */
                fgl_saveArea( sav, 0, 0, ctx->bmp, u, v, sp->x1, sp->y1);
                /* draw sprite */
                fgl_maskBltBmp( ctx, u, v, bmp, mask);
              }
            else
              {
                /* restore old area */
                fgl_bitBltBmp( ctx, sp->x0, sp->y0, sp->sav);
                /* save new area */
                fgl_saveArea( sp->sav, 0, 0, ctx->bmp, u, v, sp->x1, sp->y1);
                /* draw sprite */
                fgl_maskBltBmp( ctx, u, v, sp->bmp, sp->mask);
              }
          }
        sp->x0 = u;
        sp->y0 = v;
}

/*
 * Here comes the cursor support, which is just a global sprite
 */

void fgl_moveCursor ( i32 x, i32 y ) 
{
  _fgl_moveCursor( x, y );
}

void fgl_showCursor ()
{
  _fgl_showCursor();
}

void fgl_hideCursor () 
{
  _fgl_hideCursor();
}

void fgl_setCursor ( fgl_sprite_t* sp ) 
{
  _fgl_setCursor( sp );
}

