/*
 * sprite.h - fgl module to handle sprites (transparent bitmaps with
 *            screen save/restore) and cursor (global sprite)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_SPRITE_H_INCLUDED
#define FGL_SPRITE_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "draw.h"

static inline
void _fgl_hideSprite ( fgl_context_t* ctx, fgl_sprite_t* sp ) 
{
   if ( sp->isVisible )
    {
      if ( sp->state & IS_SHAREDMEM ) 
        {
          fgl_bitmap_t *bmp      = (fgl_bitmap_t *) ((u8 *)sp  + sizeof ( fgl_sprite_t ));
          fgl_bitmap_t *sav      = (fgl_bitmap_t *) ((u8 *)bmp + sizeof ( fgl_bitmap_t ) +
              bmp->width * bmp->height * sizeof ( fgl_slc_t ));
          /*
          fgl_monoBitmap_t *mask = (fgl_monoBitmap_t *) ((u8 *)sav + sizeof ( fgl_bitmap_t ) +
              sav->width * sav->height * sizeof ( fgl_slc_t ));
              */

          sp->isVisible = 0;
          fgl_waitReady( ctx->bmp );
          fgl_bitBltBmp( ctx, sp->x0, sp->y0, sav);
        }
      else 
        {
          sp->isVisible = 0;
          fgl_waitReady( ctx->bmp );
          fgl_bitBltBmp( ctx, sp->x0, sp->y0, sp->sav);
        }                
    }
}

static inline
void _fgl_showSprite ( fgl_context_t* ctx, fgl_sprite_t* sp ) 
{
  if ( !sp->isVisible )
    {
      if ( sp->state & IS_SHAREDMEM )
        {
          fgl_bitmap_t *bmp      = (fgl_bitmap_t *) ((u8 *)sp  + sizeof ( fgl_sprite_t ));
          fgl_bitmap_t *sav      = (fgl_bitmap_t *) ((u8 *)bmp + sizeof ( fgl_bitmap_t ) +
              bmp->width * bmp->height * sizeof ( fgl_slc_t ));
          fgl_monoBitmap_t *mask = (fgl_monoBitmap_t *) ((u8 *)sav + sizeof ( fgl_bitmap_t ) +
              sav->width * sav->height * sizeof ( fgl_slc_t ));

          sp->isVisible = 1;

          fgl_waitReady( ctx->bmp );
          fgl_saveArea( sav, 0, 0,
              ctx->bmp, sp->x0, sp->y0, sp->x1, sp->y1);
          fgl_waitReady( ctx->bmp );
          fgl_maskBlt( ctx, sp->x0, sp->y0, bmp, 0, 0,
              sp->x1 - sp->x0, sp->y1 - sp->y0, mask);
        }
      else 
        {
          sp->isVisible = 1;

          fgl_waitReady( ctx->bmp );
          fgl_saveArea( sp->sav, 0, 0,
              ctx->bmp, sp->x0, sp->y0, sp->x1, sp->y1);
          fgl_waitReady( ctx->bmp );
          fgl_maskBlt( ctx, sp->x0, sp->y0, sp->bmp, 0, 0,
              sp->x1 - sp->x0, sp->y1 - sp->y0, sp->mask);
        }
    }
}

static inline
void _fgl_moveCursor ( i32 x, i32 y ) 
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
  if ( x <= 0 ) x = 0;
  if ( y <= 0 ) y = 0;

  _fgl_inCursorOp = 1;
  fgl_moveSprite( fgl_screenCtx, fgl_cursor, x, y);
  _fgl_inCursorOp = 0;
#endif
}

static inline
void _fgl_showCursor ()
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
  _fgl_inCursorOp = 1;
  _fgl_showSprite( fgl_screenCtx, fgl_cursor);
  _fgl_inCursorOp = 0;
#endif
}

static inline
void _fgl_hideCursor ()
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
  _fgl_inCursorOp = 1;
  _fgl_hideSprite( fgl_screenCtx, fgl_cursor);
  _fgl_inCursorOp = 0;
#endif
}

static inline
void _fgl_setCursor ( fgl_sprite_t* sp )
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
  int x, y;
  if ( fgl_cursor && fgl_cursor->isVisible ){
      
      x = fgl_cursor->x0 + fgl_cursor->xhot;
      y = fgl_cursor->y0 + fgl_cursor->yhot;

      sp->x0 = x - sp->xhot;
      sp->y0 = y - sp->yhot;
      sp->x1 = sp->x0 + sp->bmp->width;
      sp->y1 = sp->y0 + sp->bmp->height;

      fgl_hideCursor();
      if ( PServer ) 
        {
         fgl_PServerUpdateCursor( sp ); 
        }
      else
        {
          fgl_cursor = sp;
        }
      fgl_showCursor();
  }
  else {
      if ( PServer ) 
        {
          fgl_PServerUpdateCursor( sp ); 
        }
      else
        {
          fgl_cursor = sp;
        }
  }
#endif /* USE_xx_CURSOR */
}

static inline
int _fgl_getSpriteSize( fgl_sprite_t *sp )
{
  int s = 0;
  s += sizeof( *sp );
  s += sizeof( *(sp->bmp) );
  s += sizeof( fgl_slc_t) * sp->bmp->width * sp->bmp->height ;
  s += sizeof( *(sp->sav) );
  s += sizeof( fgl_slc_t) * sp->sav->width * sp->sav->height ;
  s += sizeof( *(sp->mask) );
  s += (sp->bmp->width + 7) * ( sp->bmp->height + 7 ) /8;

  return s;
}

static inline
void _fgl_copySprite( fgl_sprite_t *sp, void *dest)
{
  char *c = (char *)dest;
  fgl_sprite_t *nsp;
  fgl_bitmap_t *nbmp;
  fgl_bitmap_t *nsav;
  fgl_monoBitmap_t *nmask;
  
  nsp = ( fgl_sprite_t *) c;
  memcpy(c, sp, sizeof( fgl_sprite_t ));
  c += sizeof( fgl_sprite_t );

  nbmp = ( fgl_bitmap_t *) c;
  memcpy(c, sp->bmp, sizeof( fgl_bitmap_t));
  c += sizeof( fgl_bitmap_t );

  nbmp->surface = ( fgl_slc_t *)c;
  memcpy(c, sp->bmp->surface, sizeof( fgl_slc_t ) * sp->bmp->width * sp->bmp->height );
  c += sizeof( fgl_slc_t ) * sp->bmp->width * sp->bmp->height;
  
  nsav = ( fgl_bitmap_t *) c;
  memcpy(c, sp->sav, sizeof( fgl_bitmap_t));
  c += sizeof( fgl_bitmap_t);

  nsav->surface = ( fgl_slc_t *)c;
  memcpy(c, sp->sav->surface, sizeof( fgl_slc_t ) * sp->sav->width * sp->sav->height );
  c += sizeof( fgl_slc_t ) * sp->sav->width * sp->sav->height;
  
  nmask = (fgl_monoBitmap_t *) c;
  memcpy(c, sp->mask, sizeof( fgl_monoBitmap_t));
  c += sizeof( fgl_monoBitmap_t);

  nmask->surface = ( u8 *)c;
  memcpy(c, sp->mask->surface, (sp->mask->width + 7) * (sp->sav->height + 7) / 8 );
  c += (sp->mask->width + 7) * (sp->sav->height + 7) / 8;

  nsp->bmp  = nbmp;
  nsp->sav  = nsav;
  nsp->mask = nmask;

  nsp->state       |= IS_SHAREDMEM;
  nsp->bmp->state  |= IS_SHAREDMEM;
  nsp->sav->state  |= IS_SHAREDMEM;
  nsp->mask->state |= IS_SHAREDMEM;

}


#endif /* FGL_SPRITE_H_INCLUDED */
