/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Engine.java,v 1.23 2001/11/27 19:27:06 tim Exp $
 */
package org.xoe.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Vector;
import java.util.ResourceBundle;
import java.util.MissingResourceException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.Constants;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.packages.Core;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.packages.InstallOption;
import org.xoe.core.packages.PackageException;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.packages.PackageQuery;
import org.xoe.core.packages.SearchResults;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


/**
 * Primary controling class for the XOE framework.  This class
 * control the bootstraping of services and packages and starts
 * the initial launcher.
 */
public class Engine
  implements ISubsystem, Constants
{
  private static Document m_bootstrapDoc;
  private static File m_rootDir = null;

  private static Document m_config;

  private static String NAME;
  private static String VERSION;
  private static String BUILD_DATE;
  private static String BUILD_USER;
  private static String BUILD_HOST;
  private static String BUILD_ID;
  private static String SYSTEM_ARCH;

  // make sure all buildInfo Strings are initializsed before
  // the Engine can be used.
  static
  {
    SYSTEM_ARCH = System.getProperty ("xoe.arch");
    if (SYSTEM_ARCH == null)
      {
        // attempt to guess the system archtecture.
        String arch = System.getProperties ().getProperty ("os.arch");
        String os   = System.getProperties ().getProperty ("os.name");
        // [FIXME] hack.
        if (arch.endsWith ("86"))
          {
            arch = "i386";
          }
        SYSTEM_ARCH = arch + "-" + os.toLowerCase ();
      }
    try
      {
        ResourceBundle buildInfo = ResourceBundle.getBundle ("org.xoe.core.buildinfo");
        NAME         = buildInfo.getString ("name");
        VERSION      = buildInfo.getString ("version");
        BUILD_DATE   = buildInfo.getString ("date");
        BUILD_ID     = buildInfo.getString ("id");
        BUILD_HOST   = buildInfo.getString ("host");
        BUILD_USER   = buildInfo.getString ("user");
      } 
    catch (MissingResourceException e)
      {
        Logger.log (Engine.class, "missing or incomplete resource file: org.xoe.core.buildinfo - using dummy values");
        NAME         = "Unknown";
        VERSION      = "Unknown";
        BUILD_DATE   = "Unknown";
        BUILD_ID     = "Unknown";
        BUILD_HOST   = "Unknown";
        BUILD_USER   = "Unknown";
      }
  }

  // accessor functions for Engine constants.
  public static String getVersion    () { return VERSION;     }
  public static String getName       () { return NAME;        }
  public static String getBuildId    () { return BUILD_ID;    }
  public static String getBuildDate  () { return BUILD_DATE;  }
  public static String getBuildUser  () { return BUILD_USER;  }
  public static String getBuildHost  () { return BUILD_HOST;  }
  public static String getSystemArch () { return SYSTEM_ARCH; }

  /**
   * main () function that starts the XOE framework.
   */
  public static void main (String[] args)
    throws IOException
  {
    if (args.length != 0)
      {
        Logger.log (Engine.class, "duh, main don't take no arrrrgs");
        Logger.log (Engine.class, "set the xoe.bootstrap property if you want to pass settings to XOE");
        System.exit (1);
      }
    bootstrap ();
    start ();
  }

  public static void shutdown (int returnCode)
  {
    Logger.log (Engine.class, "shutting down XOE session");
    System.exit (returnCode);
  }

  /**
   * Will return true if the bootstrap had no failures, false otherwise
   * (Normally we will continue to boot despite errors)
   */
  public static boolean bootstrap ()
    throws IOException
  {
    Logger.log (Engine.class, "loading " + NAME + " v" + VERSION + "\nbuilt: " + BUILD_DATE);
    Logger.log (Engine.class, "system architecture is: " + SYSTEM_ARCH);

    String bootstrapFile = System.getProperty ("xoe.bootstrap");
    if (bootstrapFile == null)
      {
        bootstrapFile = (System.getProperty (HOMEDIR_PROPERTY)
		+ File.separator
		+ DEFAULT_BOOTSTRAP);
      }

    long startTime = System.currentTimeMillis ();

    boolean success = true;

    File bsFile = new File (bootstrapFile);
    m_rootDir = new File (bsFile.getParent ());

    m_bootstrapDoc = DOMlet.create (new FileInputStream (bsFile), bsFile.toString ());
    NodeList nl = XPath.findNodes (m_bootstrapDoc, "bootstrap/load");

    // Optimization: accumulate all packages that are specified in order and
    // try to process them as a batch. If that fails, go back and do them all
    // individually.
    Vector packages = new Vector ();
    for (int i = 0; i < nl.getLength (); i++)
      {
	Element el = (Element) nl.item (i);

	String subsystem = el.getAttribute ("subsystem");
	String service   = el.getAttribute ("service");
	String confPath  = el.getAttribute ("config");

	String pname = el.getAttribute ("package");

	if (subsystem != null)
	  {
	    success = installSubsystem (subsystem, confPath) && success;
	  }
	else if (service != null)
	  {
	    success = installService (service, confPath) && success;
	  }
	else if (pname != null)
	  {
	    //success = installPackage (pname) && success;
	    packages.addElement (pname);
	  }
	else
	  {
	    Logger.log (Engine.class, "WARNING: Unrecognized load command in " + bootstrapFile + ": " + XMLWriter.getXML(el));
	    success = false;
	  }

	if ((i + 1 == nl.getLength () || ((Element)nl.item (i + 1)).getAttribute ("package") == null) && packages.size () > 0)
	  {
	    String[] pnames = (String[]) packages.toArray (new String[packages.size ()]);
	    try
	      {
		PackageManager.batchInstall (pnames);
	      }
	    catch (PackageException ex)
	      {
		Logger.log (Engine.class, ex.toString ());
		Logger.log (Engine.class, "tried to install several packages as a batch but it failed. Will try individually.");
		for (int j = 0; j < pnames.length; j++)
		  {
		    success = installPackage (pnames[j]) && success;
		  }
	      }

	    packages = new Vector ();
	  }
      }

    Logger.log (Engine.class, "bootstrap time: " + (System.currentTimeMillis () - startTime) + " ms");
    Logger.log (Engine.class, "\tverify:\t\t" + Core.verifyTimer + " ms");
    Logger.log (Engine.class, "\torder handlers:\t" + Core.orderedHandlersTimer + " ms");
    Logger.log (Engine.class, "\tinstall:\t" + Core.installTimer + " ms");
    Logger.log (Engine.class, "\tbatch calcs:\t" + PackageManager.batchInstallTimer + " ms");
    Logger.log (Engine.class, "\tlog time:\t" + Logger.logTime + " ms");

    return success;
  }

  private static boolean installService (String cname, String conf)
  {
    IService provider;
    try
      { // so many things can fail on this line
	provider = (IService)Class.forName (cname).newInstance ();
      }
    catch (Exception ex)
      {
	Logger.log (Engine.class, "WARNING: Failed to instantiate service");
	Logger.log (Engine.class, "         Class: " + cname);
	Logger.log (Engine.class, ex);
	
	return false;
      }

    try
      {
	Document confdoc = null;
	if (conf != null)
	  {
	    File f = new File (m_rootDir, conf);
	    confdoc = DOMlet.create (new FileInputStream (f), f.toString ());
	  }
	provider.init (confdoc);
	ServiceLocator.registerService (provider);
      }
    catch (Exception ex)
      {
	Logger.log (Engine.class, "WARNING: Failed to initialize service");
	Logger.log (Engine.class, "         Class: " + cname);
	Logger.log (Engine.class, "         Config: " + (conf == null ? "none" : conf));
	Logger.log (Engine.class, ex);
	return false;
      }

    return true;
  }


  private static boolean installSubsystem (String cname, String conf)
  {
    ISubsystem subsystem;
    try
      { // so many things can fail on this line
	subsystem = (ISubsystem) Class.forName (cname).newInstance ();
      }
    catch (Exception ex)
      {
	Logger.log (Engine.class, "WARNING: Failed to instantiate subsystem");
	Logger.log (Engine.class, "         Class: " + cname);
	Logger.log (Engine.class, ex);
	
	return false;
      }

    try
      {
	Document confdoc = null;
	if (conf != null)
	  {
	    File f = new File (m_rootDir, conf);
	    confdoc = DOMlet.create (new FileInputStream (f), f.toString ());
	  }
	subsystem.initSubsystem (confdoc);
      }
    catch (Exception ex)
      {
	Logger.log (Engine.class, "WARNING: Failed to initialize subsystem");
	Logger.log (Engine.class, "         Class: " + cname);
	Logger.log (Engine.class, "         Config: " + (conf == null ? "none" : conf));
	Logger.log (Engine.class, ex);
	return false;
      }

    return true;
  }

  // will return true if all packages were installed
  private static boolean installPackage (String pname)
  {
    PackageQuery query = new PackageQuery ();
    query.setName (pname);
    try
      {
        SearchResults res = PackageManager.search (query);
        if (res.getInstalledPackages ().length != 0)
          {
            return true;
          }
        InstallOption[] options = res.getInstallablePackages ();

        if (options.length > 0)
          {
            Logger.status ("installing " + pname);
            PackageManager.commit (options[0]);
          }
        else
          {
            Logger.log (Engine.class, "no installable packages for: " + pname);
            return false;
          }
      }
    catch (Throwable ex)
      {
        Logger.log (Engine.class, "failed to install package: " + pname);
        Logger.log (Engine.class, ex);
        return false;
      }

    return true;
  }

  /**
   * Start the XOE framework by looking for an installed IShell service
   * and start()ing it.
   */
  public static void start ()
  {
    IShell shell = (IShell) ServiceLocator.findImplementor (IShell.INTERFACE);
    if (shell == null)
      {
        Logger.log (Engine.class, "no IShell found in system");
        Logger.log (Engine.class, "failed ot find IShell service (XOE requires at least one IShell service)");
        shutdown (1);
      }
    shell.start ();
  }

  public static File getBootstrapDirectory ()
  {
    return m_rootDir;
  }

  public static String getConfig (String configName)
  {
    return getConfig (configName, null);
  }

  public static String getConfig (String configName, String defaultValue)
  {
    String configValue = System.getProperty (PROPERTY_PREFIX + configName);
    if (configValue == null)
      {
	if (m_config != null)
	  {
	    configValue = XPath.findString (m_config.getDocumentElement (),
					   configName);

	    if (configValue.length () <= 0)
	      {
		return defaultValue;
	      }
	  }
      }
    return configValue;
  }

  public static int getConfig (String configName, int defaultValue)
  {
    return Integer.parseInt (getConfig (configName, Integer.toString (defaultValue)));
  }

  public void initSubsystem (Document configDoc)
    throws Exception
  {
    m_config = configDoc;
  }
}
