/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: ContentCache.java,v 1.5 2001/12/18 06:06:15 colin Exp $
 */
package org.xoe.core.content; 

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.Constants;
import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


/**
 * The <code>ContentCache</code> class provides a local cache for
 * all content, whether it be from the local virtual filesystem 
 * (the Stash) or the web.  Accesses to the Stash, using with
 * the <code>stash:/</code> protocol, are passed on to the Stash
 * for handling.
 *
 */
public class ContentCache 
  implements Constants
{
  private static final Document cacheDocument;
  private static final Element cacheElement;
  private static final ContentElementFactory factory;

  static
  {
    cacheDocument = new XDocument ();
    cacheElement = cacheDocument.createElementNS (null, "contentCache");
    cacheDocument.appendChild (cacheElement);

    factory = new ContentElementFactory (cacheDocument);
  }

  /**
   *
   */
  public static ContentElementFactory getFactory ()
  {
    return factory;
  }

  public static URL buildURL (String src)
    throws MalformedURLException
  {
    return buildURL (src, null);
  }

  public static URL buildURL (String src, XDocument doc)
    throws MalformedURLException
  {
    if (doc != null && src.indexOf (':') == -1) 
      {
        return (new URL (doc.getURL (), src));
      }
    else
      {
        return (new URL (src));
      }
  }

  public static boolean contains (URL url)
    throws IOException
  {
    if (url == null)
      {
        throw new IOException ("null url in contains");
      }

    // lookup the content element in the stash if the protocol is stash:/
    if (url.getProtocol ().equals (STASH_PROTOCOL))
      {
        return (Stash.contains (url));
      }
    return (justFetch (url.toString()) != null);
  }

  public static boolean contains (String src)
    throws IOException
  {
    if (src == null)
      {
        throw new IOException ("null src in contains");
      }

    // lookup the content element in the stash if the protocol is stash:/
    if (src.startsWith (STASH_PROTOCOL))
      {
        return (Stash.contains (src));
      }

    return (justFetch (src) != null);
  }

  public static ContentElement fetch (URL u)
    throws IOException
  {
    if (u == null)
      {
        throw new IOException ("null url in fetch");
      }

    // lookup the content element in the stash if the protocol is stash:/
    if (u.getProtocol ().equals (STASH_PROTOCOL))
      {
        return (Stash.fetch (u.toString ()));
      }

    ContentElement content = justFetch (u.toString ());
    if (content != null)
      {
        return (content);
      }

    return justPut (u);
  }

  public static ContentElement fetch (String src)
    throws IOException
  {
    if (src == null)
      {
        throw new IOException ("null src in fetch");
      }

    // lookup the content element in the stash if the protocol is stash:/
    if (src.startsWith (STASH_PROTOCOL))
      {
        return (Stash.fetch (src));
      }

    ContentElement content = justFetch (src);
    if (content != null)
      {
        return (content);
      }

    return justPut (buildURL (src));
  } 


  public static void clear (String src)
  {
    if (src != null)
      {
	ContentElement content = justFetch (src);
	Logger.log (ContentCache.class, "src = " + src);
	if (content != null)
	  {
	    Logger.log (ContentCache.class, "content = " + content);
	    content.clearCachedData ();
	    return;
	  }
      }
  }

  /**
   * justFetch looks up the ContentElement by src string.  if it's not 
   * found it simply returns null and let's the calling method handle
   * adding new content to the cache.
   */
  private static ContentElement justFetch (String src)
  {
    return (ContentElement) XPath.findFirstNode (cacheElement, ContentElement.TAGNAME 
                                                       + "[@" + ContentElement.URL_ATTR 
                                                       + "='" + src + "']");
  }

  private static ContentElement justPut (URL u)
    throws IOException
  {
    return justPut (factory.createContentElement (u));
  }

  private static ContentElement justPut (ContentElement content)
  {
    if (content != null)
      {
        // add the new element
        cacheElement.appendChild (content);
      }
    return (content);
  }

  public static ContentElement remove (URL u)
    throws IOException
  {
    if (u == null)
      {
        throw new IOException ("null url in remove");
      }      
    return (remove (u.toString ()));
  }

  public static ContentElement remove (String src)
    throws IOException
  {
    if (src == null)
      {
        throw new IOException ("null src in remove");
      }

    // remove the content element from the stash if the protocol is stash:/
    if (src.startsWith (STASH_PROTOCOL))
      {
        return (Stash.remove (src));
      }

    ContentElement content = justFetch (src);
    if (content != null)
      {
        return ((ContentElement)cacheElement.removeChild (content));
      }
    return (content);
  }

  public static ContentElement findFirstNode (String query)
  {
    return (ContentElement) XPath.findFirstNode (cacheElement, query);
  }

  /**
   * Returns the first node satisfying the xpath query
   * <code>query</code> using this node as the context.
   */
  public static ContentElement findFirstNode (XPath query)
  {
    return (ContentElement) XPath.findFirstNode (cacheElement, query);
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using this node as the context.
   */
  public static NodeList findNodes (String query)
  {
    return XPath.findNodes (cacheElement, query);
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using this node as the context.
   */
  public static NodeList findNodes (XPath query) 
  {
    return XPath.findNodes (cacheElement, query);
  }

  public static String getXML ()
  {
    return XMLWriter.getXML (cacheElement);
  }

} // ContentCache
