/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: ContentElement.java,v 1.5 2001/12/18 06:06:15 colin Exp $
 */
package org.xoe.core.content;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.util.Date;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.dom.XElement;
import org.xoe.core.util.Logger;


/**
 * The <code>ContentElement</code> class represents the atomic unit
 * of content handling in the XOE system. <br>  All content used by
 * the system, whether local or on the Web, is at some point in it's
 * lifecycle managed as an instance of this class.
 *
 */
public class ContentElement 
  extends XElement 
{
  static final String TAGNAME = "content";
  static final String URL_ATTR = "url";
  static final String LENGTH_ATTR = "length";

  // data buffer of the content represented by this element 
  private byte[] data;

  // location of the content 
  private URL url;

  // the content type 
  private ContentType type;

  /**
   * the date this object was last modified 
   */
  protected Date lastModified;

 
  // -------------------------------------------------------------------------
  // constructors

  ContentElement (URL u, Document owner, Date date)
  {
    super (owner, TAGNAME,  null);
    setURL (u);
    if (date == null)	
      {
	lastModified = new Date ();
      }
    else 
      {
	lastModified = date;
      }
  }

 
  // -------------------------------------------------------------------------
  // property access methods

  private void setURL (URL u)
  {
    // add the url attribute
    if (u != null)
      {
	url = u;
	setAttributeNS (null, URL_ATTR, url.toString ());
      }
  }

  /**
   * Returns the URL associated with this content.
   *
   * @return the URL associated with this content
   */
  public URL getURL ()
  {
    return (url);
  }

  void setData (byte[] d)
  {
    if (d != null)
      {
        data = d;
	setLength (data.length);
      }
  }

  public void clearCachedData ()
  {
    data = null;
    setLength (0);
  }

  /**
   * Returns the InputStream associated with this content.
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the InputStream associated with this content
   */
  public InputStream getInputStream ()
    throws IOException
  {
    if (data == null)
      {
        setData (getBytes (url));
      }

    return (new ByteArrayInputStream (data));
  }

  /**
   * Sets the size of the data associated with this content.
   *
   * @param len the size, in bytes, of the content's data
   */
  protected void setLength (long len)
  {
    // add the content length attribute
    setAttributeNS (null, LENGTH_ATTR, Long.toString (len));
  }

  /**
   * Returns the size of the data associated with this content. 
   *
   * @return the size, in bytes, of the associated content's data; 
   *         0 if there is an error
   */
  public int getContentLength ()
  {
    String len = getAttributeNS (null, LENGTH_ATTR);
    if (len != null)
      {
        try
          {
	    return (Long.valueOf (len).intValue ());
          }
        catch (NumberFormatException nfe)
          {
            Logger.log (this, nfe);
          }
      }
    else
      {
        Logger.log (this, "getContentLength: length attr not found");
        Logger.log (this, "getContentLength: url = " + url);
      }

    return (0);
  }

  void setType (ContentType t)
  {
    type = t;    
  }

  /**
   * Returns the type of the data associated with this content. 
   *
   * @return the type of the data associated with this content
   */
  public ContentType getType ()
    throws IOException
  {
    // if we don't have a type yet, or we haven't yet resolved the suffix
    // to a mimetype, try and resolve the mimetype again.
    if (type == null || type.getMimetype () == null)
      {
	type = DefaultContentTypeResolver.getContentType (this);
      }
    return (type);
  }

  /**
   * Returns the date this object was last changed
   *
   * @return the date this object was last changed
   */
  public Date getLastModified ()
  {
    return lastModified;
  }    


  // -------------------------------------------------------------------------
  // member methods

  /*
   *  Returns an alias content element - essentially a clone of this
   */
  ContentElement makeAlias (URL u)
  {
    ContentElement alias = (ContentElement) super.cloneNode (true);
    alias.setURL (u);
    alias.setData ((byte[]) data.clone ());
    alias.setType ((ContentType) type.clone ());
    return alias;
  }

  /*
   *  Returns a byte array built from the data associated with
   *  the url passed in 
   */
  byte[] getBytes (URL url)
    throws IOException
  {
    // get the connection for the url passed in
    URLConnection urlc = url.openConnection ();

    if (urlc.getInputStream () == null)
      {
        return (null);
      }

    // get the content length
    int len = urlc.getContentLength ();
    byte[] data = null;
    try
      {
	// create a new content element
	data =  getBytes (urlc.getInputStream (), len);
      }
    finally
      {
	// close the url connections input stream    
	urlc.getInputStream ().close ();
      }

    this.url = urlc.getURL ();
    return (data);    
  }

  /*
   *  Returns a byte array built from the data associated with
   *  the input stream passed in
   */
  static byte[] getBytes (InputStream dataStream, int len)
    throws IOException
  {
    if (dataStream == null)
      {
        return null;
      }

    byte[] streamCopy;

    // is length valid?
    if (len > 0)
      {
        // valid length so use it to build the array and read all at once
	streamCopy = new byte[len];
        int pos = 0;
	while (len > 0)
	  {
	    int actual = dataStream.read (streamCopy, pos, len);
	    if (actual == -1) break;
	    pos += actual;
	    len -= actual;
	  }
      }
    else
      {
        // invalid length, so read from the stream until we get a -1 length back
	ByteArrayOutputStream baos = new ByteArrayOutputStream ();
	byte buf[] = new byte[1024];

	while (true)
	  {
	    len = dataStream.read (buf);
	    if (len == -1) break;
	    baos.write (buf, 0, len);
	  }

	baos.flush ();
	streamCopy = baos.toByteArray ();
      }

    return (streamCopy);
  }


  // -------------------------------------------------------------------------
  // static methods

  /**
   * Returns a string which is the encoded representation of the
   * string passed in.  Used to encode filenames that may not be
   * legal element names according to the DOM spec.
   *
   * @param s a string representing the name to be encoded
   *
   * @return the encoded string
   */
  public static String encodeName (String s)
  {
    StringBuffer result = new StringBuffer ();
    try
      {
        byte b[] = s.getBytes ("UTF8");
	for (int pos = 0; pos <b.length; pos++) 
	  {
	    char chr = (char)(b[pos] & 0x00ff);
	    if ((chr >= 'A' && chr <= 'Z') 
                || (chr >= 'a' && chr <= 'z') 
                || (chr >= '0' && chr <= '9') 
                || chr == '.' 
                || chr == '-' 
                || chr == ':' 
                || chr == '_') 
              {
		result.append (chr);
	      }
	    else if (chr == ' ') 
              {
	        result.append ('+');
	      }
	    else 
              {
	        result.append ('_' + ((int)chr < 16 ? "0" : "") + Integer.toString ((int)chr, 16));
	      }
	  }
      }
    catch (java.io.UnsupportedEncodingException uee)
      {
      }
    return (result.toString ());
  }

 
  // -------------------------------------------------------------------------
  // org.w3c.dom.Node

  public Node cloneNode (boolean deep)
  {
    ContentElement clonedContent = (ContentElement)super.cloneNode (deep);
    try
      {
        clonedContent.setURL (new URL (url.toString ()));
      }
    catch (MalformedURLException e) 
      {
        Logger.log (this, "cloneNode: " + e);
      }
    clonedContent.setData ((byte[])data.clone ());
    clonedContent.setType ((ContentType)type.clone ());
    return clonedContent;
  }

} // ContentElement
