/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: ContentElementFactory.java,v 1.7 2001/12/18 06:06:15 colin Exp $
 */
package org.xoe.core.content;

import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Date;

import org.xoe.core.util.Logger;

import org.w3c.dom.Document;


/**
 * The <code>ContentElementFactory</code> class is used to construct
 * content elements for the Content Cache and the Stash.
 */
public class ContentElementFactory
{
  /*
   *  the Document to use as the owner document for newly created 
   *  content elements.
   */
  private Document cacheDocument;

 
  // -------------------------------------------------------------------------
  // constructors

  ContentElementFactory (Document doc) 
  {
    cacheDocument = doc;
  }
 
 
  // -------------------------------------------------------------------------
  // methods for creating content elements

  /**
   * Returns the content element associated with the content at <code>url</code>. 
   * 
   * @param url the URL associated with the desired content
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the content element associated with <code>url</code>
   */
  public ContentElement createContentElement (URL url)
    throws IOException
  {
    return  new ContentElement (url, cacheDocument, null);
  }

  /**
   * Returns the content element associated with the content at <code>url</code>. 
   * Uses <code>dataStream</code> to build the content element's internal byte 
   * array of data. <br /><i>Note:</i> closing <code>dataStream</code> 
   * must be handled by the caller.
   *    
   * @param url the URL associated with the desired content
   * @param dataStream the InputStream representing the content's binary data
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the content element associated with <code>url</code>
   */
  public ContentElement createContentElement (URL url, InputStream dataStream)
    throws IOException
  {
    return (createContentElement (url, dataStream, -1));
  }

  /**
   * Returns the content element associated with the content at <code>url</code>. 
   * Uses <code>dataStream</code> to build the content element's internal byte 
   * array of data. <br /><i>Note:</i> closing <code>dataStream</code> 
   * must be handled by the caller.
   *    
   * @param url the URL associated with the desired content
   * @param dataStream the InputStream representing the content's data
   * @param len the length, in bytes, of the data
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the content element associated with <code>url</code>
   */
  public ContentElement createContentElement (URL url, InputStream dataStream, int len)
    throws IOException
  {
    return (createContentElement (url, ContentElement.getBytes (dataStream, len), null));
  }

  /**
   * Returns the content element associated with the content at <code>url</code>. 
   * Uses <code>dataStream</code> to build the content element's internal byte 
   * array of data. <br /><i>Note:</i> closing <code>dataStream</code> 
   * must be handled by the caller.
   *    
   * @param url the URL associated with the desired content
   * @param dataStream the InputStream representing the content's binary data
   * @param d the Date to use as the last modified for this content
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the content element associated with <code>url</code>
   */
  public ContentElement createContentElement (URL url, InputStream dataStream, Date d)
    throws IOException
  {
    return (createContentElement (url, dataStream, -1, d));
  }

  /**
   * Returns the content element associated with the content at <code>url</code>. 
   * Uses <code>dataStream</code> to build the content element's internal byte 
   * array of data. <br /><i>Note:</i> closing <code>dataStream</code> 
   * must be handled by the caller.
   *    
   * @param url the URL associated with the desired content
   * @param dataStream the InputStream representing the content's binary data
   * @param len the length, in bytes, of the data
   * @param d the Date to use as the last modified for this content
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the content element associated with <code>url</code>
   */
  public ContentElement createContentElement (URL url, InputStream dataStream, int len, Date d)
    throws IOException
  {
    return (createContentElement (url, ContentElement.getBytes (dataStream, len), d));
  }

  /**
   * Returns the content element associated with the content at <code>url</code>. 
   * Uses <code>dataArray</code> for the content's internal byte array of data. 
   *    
   * @param url the URL associated with the desired content
   * @param dataArray the content's binary data
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the content element associated with <code>url</code>
   */
  public ContentElement createContentElement (URL url, byte[] dataArray)
    throws IOException
  {
    return createContentElement (url, dataArray, null);
  }
  
  /**
   * Returns the content element associated with the content at <code>url</code>. 
   * Uses <code>dataArray</code> for the content's internal byte array of data. 
   *    
   * @param url the URL associated with the desired content
   * @param dataArray the content's binary data
   * @param d the Date to use as the last modified for this content
   *    
   * @exception IOException if an I/O error occurs
   *
   * @return the content element associated with <code>url</code>
   */
  public ContentElement createContentElement (URL url, byte[] dataArray, Date d)
    throws IOException
  {
    if (url == null)
      {
        return (null);
      }

    // get a new ContentElement from the cacheDocument
    ContentElement content = new ContentElement (url, cacheDocument, d);

    // set the content's data
    content.setData (dataArray);

    return (content);
  }  

} // ContentElement
