/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: ContentType.java,v 1.5 2001/12/04 22:15:47 colin Exp $
 */
package org.xoe.core.content;

import org.xoe.core.Constants;
import org.xoe.core.packages.IPackageQuery;
import org.xoe.core.services.IService;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;


/**
 * This class is designed to support finding basic information about the data
 * type of a document or resource, using IContentTypeResolver service
 * providers.
 * 
 * The default service provider will attempt to find the mimetype based on the
 * suffix of the file.
 * 
 * If the file does not have a suffix, it will attempt to get the information
 * from the URL by calling <code>URL.openConnection ().getContentType ()</code>.
 * In this case, calling <code>getSuffix ()</code> will return null.
 * 
 * (If the resource is described as "text/plain", this will result in retrieval
 * of the first several bytes of the file for comparison by a
 * <code>kaffe.net.StreamMap</code> implementation. The default implementation,
 * <code>kaffe.net.DefaultStreamMap</code> can identify PNG, GIF, JPEG, and
 * class files based on the first four bytes of the stream.)
 * 
 * If the mimetype is "text/xml", it will attempt to retrieve the Document Type
 * name from the Document Type Declaration. Otherwise, calling
 * <code>getXMLDoctypeNameNS ()</code> will return null.
 * 
 * Service providers may subclass ContentType to provide additional information.
 */
public class ContentType
  implements Constants
{
  private static final IService[] nullIService = new IService[0];

  String m_suffix = null;
  String m_mimetype = null;
  String m_doctypeNameNS = null;

  // Package query for finding a handler for the 
  // ContentType.  
  private class ContentServicePackageQuery 
    implements IPackageQuery
  {
    // the java interface that we are looking for.
    private String m_interface = null;

    // the internal XPath.  We can cache this because contenttypes
    // are not changeable.
    private String m_xpath = null;
    
    public ContentServicePackageQuery (String iface)
    {
      m_interface = iface;
    }
    
    public String toXPath ()
    {
      if (m_xpath != null)
        {
          return m_xpath;
        }

      if (m_interface == null)
	{ // being explicit here
	  m_xpath = null;
	}
      else if (m_doctypeNameNS != null)
	{ // suffix and mimetype are not important
	  m_xpath = ( "package" + 
		  ( "[provides/proviso" +
		    ( "[implements='" + m_interface + "']") +
		    ( "[feature" +
	              ( "[@name='" + FEATURE_DOCTYPE_NAME_NS + "']") +
                      ( "[@value='" + m_doctypeNameNS + "']") +
		      "]") +
	            "]") 
                );
	}
      else if (m_mimetype != null)
	{ // suffix is not useful in this case
	  m_xpath = ( "package" + 
		  ( "[provides/proviso" +
		      ( "[implements='" + m_interface + "']") +
		      ( "[feature[@name='mimetype']" +
			( "[@value='" + m_mimetype + "']") +
			"]") +
		    "]")
	        );
	}
      else if (m_suffix != null)
	{ // this is still a mimetype query, but now the mimetype comes from inside the
	  // context package itself. Isn't XPath amazing?
	  m_xpath = ( "package" + 
		    ( "[provides/proviso" +
		        ( "[implements='" + m_interface + "']") +
		        ( "[feature[@name='extension']" +
		          ( "[@value='" + m_suffix + "']") +
			"]") +
		      "]")
                  );
	}
      else
	{ // that explicit stuff again ;)
	  m_xpath = null;
	}

      Logger.log (ContentServicePackageQuery.class, m_xpath);
      return m_xpath;
    }
  }

  public ContentType ()
  {
    this (null, null, null, null);
  }

  public ContentType (String mimetype)
  {
    this (null, mimetype, null, null);
  }

  public ContentType (String mimetype, String docTypeName)
  {
    this (null, mimetype, docTypeName, null);
  }

  public ContentType (String mimetype, String docTypeName, String suffix)
  {
    this (null, mimetype, docTypeName, suffix);
  }

  public ContentType (ContentType base, String mimetype)
  {
    this (base, mimetype, null, null);
  }

  public ContentType (ContentType base, String mimetype, String docTypeName)
  {
    this (base, mimetype, docTypeName, null);
  }

  public ContentType (ContentType base, String mimetype, String docTypeName, String suffix)
  {
    this (base);

    if (suffix != null)
      {
	m_suffix = suffix;
      }
    if (docTypeName != null)
      {
	m_doctypeNameNS = docTypeName;
      }
    if (mimetype != null)
      {
	m_mimetype = mimetype;
      }
  }

  public ContentType (ContentType base)
  {
    if (base == null)
      {
	return;
      }
    m_suffix = base.getSuffix ();
    m_mimetype = base.getMimetype ();
    m_doctypeNameNS = base.getXMLDoctypeNameNS ();
  }

  public Object clone ()
  {
    return new ContentType (this);
  }

  public String getSuffix ()
  {
    return m_suffix;
  }

  public String getMimetype ()
  {
    return m_mimetype;
  }

  public String getXMLDoctypeNameNS ()
  {
    return m_doctypeNameNS;
  }
  
  public IPackageQuery getPackageQuery (String iface)
  {
    return new ContentServicePackageQuery (iface);
  }

  public IService getService (String iface)
  {
    IService[] services = getServices (iface);
    if (services.length > 0)
      {
	return services[0];
      }
    return null;
  }

  public IService[] getServices (String iface)
  {
    SimpleServiceQuery q = SimpleServiceQuery.alloc ();

    if (m_doctypeNameNS != null)
      {
	q.setInterface (iface);
	q.addFeature (FEATURE_DOCTYPE_NAME_NS, m_doctypeNameNS);
	IService[] services = ServiceLocator.findServices (q);
	if (services.length != 0)
	  {
	    SimpleServiceQuery.free (q);
	    return (services);
	  }
      }
    
    if (m_mimetype != null)
      {
	q.reset ();
	q.setInterface (iface);
	q.addFeature (FEATURE_MIMETYPE, m_mimetype);
	IService[] services = ServiceLocator.findServices (q);
	if (services.length != 0)
	  {
	    SimpleServiceQuery.free (q);
	    return (services);
	  }
      }
      
    if (m_suffix != null)
      {
	q.reset ();
	q.setInterface (iface);
	q.addFeature (FEATURE_EXTENSION, m_suffix);
	IService[] services = ServiceLocator.findServices (q);
	if (services.length != 0)
	  {
	    SimpleServiceQuery.free (q);
	    return (services);
	  }
      }

    SimpleServiceQuery.free (q);
    return nullIService;
  }

  public String toExternalForm ()
  {
    if (m_doctypeNameNS != null)
      {
        return "XML doctype is " + m_doctypeNameNS;
      }
    if (m_mimetype != null)
      {
        return "mimetype is " + m_mimetype;
      }
    if (m_suffix != null)
      {
        return "extension is " + m_suffix;
      }
    return "unknown type";
  }

  public String toString ()
  {
    return "{'" + m_suffix + "','" + m_mimetype + "','" + m_doctypeNameNS + "'}";
  }

  public int hashCode ()
  {
    if (m_mimetype != null)
      {
	return m_mimetype.hashCode ();
      }
    if (m_suffix != null)
      {
	return m_suffix.hashCode ();
      }
    return super.hashCode ();
  }

  public boolean equals (Object o)
  {
    if (!(o instanceof ContentType)) 
      {
        return false;
      }
    
    ContentType other = (ContentType) o;
    return (this.toString ().equals (other.toString ()));
  }
}
