/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: DefaultContentTypeResolver.java,v 1.8 2001/12/04 22:18:34 colin Exp $
 */
package org.xoe.core.content;

import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.DOMlet;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceInstanceElement;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.IServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.core.util.ProviderHashtable;
import org.xoe.core.xpath.XPath;


/** 
 * Keeps track of basic suffix->mimetype mappings.
 * 
 * These can be installed as part of a package along with a "provider key" that
 * can be used to remove all the mappings associated with a package.
 *
 * The provider key should typically be a reference to the package, or a unique
 * String identifier.
 */
public class DefaultContentTypeResolver 
  implements IContentTypeResolver
{
  private static DefaultContentTypeResolver m_instance = null;

  private ProviderHashtable m_hash = new ProviderHashtable ();

  public static DefaultContentTypeResolver getInstance ()
  {
    return m_instance;
  }

  public static class CTRMimetypeQuery 
    implements IServiceQuery
  {
    private String m_mimetype = null;

    public CTRMimetypeQuery ()
    {
    }

    public CTRMimetypeQuery (String mimetype)
    {
      m_mimetype = mimetype;
    }

    public String toXPath ()
    {
      String query =
	("xoeservices:service[xoeservices:functionality" 
         + "[@type='" + IContentTypeResolver.FUNCTIONALITY + "']" 
         + "[xoeservices:implements='" + IContentTypeResolver.INTERFACE + "']" 
         + "[xoeservices:feature[@name='" + FEATURE_MIMETYPE + "']" 
         + (m_mimetype != null 
            ? "[(@value='" + m_mimetype + "') or (@value='" + MIMETYPE_ALL + "')]]]" 
            : "[@value='" + MIMETYPE_UNKNOWN + "']]]"));

      return query;
    }
  }


  public static class CTRSuffixQuery 
    implements IServiceQuery
  {
    private String m_suffix = null;

    public CTRSuffixQuery ()
    {
      m_suffix = "*";
    }

    public CTRSuffixQuery (String filename)
    {
      m_suffix = DefaultContentTypeResolver.getSuffix (filename);
    }

    public String toXPath ()
    {
      String query =
	("xoeservices:service[xoeservices:functionality" 
         + "[@type='" + IContentTypeResolver.FUNCTIONALITY + "']" 
         + "[xoeservices:implements='" + IContentTypeResolver.INTERFACE + "']" 
         + "[xoeservices:feature[@name='" + FEATURE_EXTENSION + "']"
         + "[(@value='" + m_suffix + "') or (@value='" + EXTENSION_ALL + "')]]]");

      return query;
    }
  }

  public void processMimetypeFile (Document doc, Object provider)
  {
    NodeList list = XPath.findNodes (doc, "mimetype-mappings/mapping");
    int len = list.getLength ();
    
    len = list.getLength ();
    for (int i = 0; i < len; i++) 
      {
	Element elem = (Element) list.item (i);
	DefaultContentTypeResolver.getInstance ().putMapping (elem.getAttribute ("suffix"),
							      elem.getAttribute ("mimetype"),
							      provider);
      }

  }
  public static IService [] ctrs = new IService [0];

  public void init (Document configDoc)
  {
    m_instance = this;
    ctrs = new IService [] { this };
    if (configDoc != null)
      {
	processMimetypeFile (configDoc, null);
      }
  }

  public String[] getFunctionalities ()
  {
    String[] ret = {IContentTypeResolver.FUNCTIONALITY};
    return ret;
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IContentTypeResolver.FUNCTIONALITY))
      {
	String[] ret = { IContentTypeResolver.INTERFACE };
	return ret;
      }
    return null;
  }

  public String[] getFeatures (String func)
  {
    if (func.equals (IContentTypeResolver.FUNCTIONALITY))
      {
	String[] ret = { FEATURE_EXTENSION, EXTENSION_ALL,
			 FEATURE_MIMETYPE, MIMETYPE_TEXT_PLAIN,
			 FEATURE_MIMETYPE, MIMETYPE_TEXT_XML,
			 FEATURE_MIMETYPE, MIMETYPE_UNKNOWN };

	return ret;
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public void putMapping (String suffix, String mimetype)
  {
    putMapping (suffix, mimetype, null);
  }

  public void putMapping (String suffix, String mimetype, Object provider)
  {
    // removes the '.' if there is one
    suffix = getSuffix (suffix);

    m_hash.put (suffix, mimetype, provider);
  }

  public void removeProvider (Object provider)
  {
    m_hash.removeProvider (provider);
  }

  public void removeMapping (String suffix)
  {
    removeMapping (suffix, null);
  }
  
  public void removeMapping (String suffix, Object provider)
  {
    // removes the '.' if there is one
    suffix = getSuffix (suffix);
    
    m_hash.remove (suffix, provider);
  }

  public ContentType resolveStream (InputStream in, String mimetype)
    throws IOException
  {
    return (resolveStream (in, new ContentType (mimetype)));
  }

  public ContentType resolveStream (InputStream in, ContentType type)
    throws IOException
  {
    String mimetype = null;
    if (type != null)
      {
        mimetype = type.getMimetype ();
      }

    if (mimetype == null || !mimetype.equals (MIMETYPE_TEXT_XML))
      {
	mimetype = URLConnection.guessContentTypeFromStream (in);
      }

    if (mimetype == null || mimetype.equals (MIMETYPE_TEXT_PLAIN) || mimetype.equals (MIMETYPE_TEXT_XML) )
      {
	DOMlet doc = DOMlet.create (in, "tmpFile");
	if (doc != null)
	  {
	    type = new ContentType (type, MIMETYPE_TEXT_XML, doc.getDoctypeNameNS ());
	  }
      }

    return (type);
  }

  // get a ContentType for a suffix
  public ContentType resolveSuffix (String suffix)
  {
    if (suffix == null)
      {
	return null;
      }

    String mimetype = (String)m_hash.get (suffix);

    if (mimetype == null)
      {
	return null;
      }

    return new ContentType (mimetype, null, suffix);
  }

  public static String getSuffix (String filename)
  {
    if (filename == null)
      {
	return null;
      }

    int lastSlash = filename.lastIndexOf ('/');
    int lastDot = filename.lastIndexOf ('.');

    String suffix = null;

    if (lastDot != -1 && lastDot > lastSlash && lastDot < filename.length () - 1)
      {
	suffix = filename.substring (lastDot + 1);
      }
    // no slash, they might have just passed in the suffix
    else if (lastSlash == -1)
      {
	suffix = filename;
      }

    return suffix;
  }

  public static ContentType getContentType (String suffix)
  {

    ContentType ct = null;
//     IServiceQuery q = new CTRSuffixQuery (suffix);

    IService[] ctr = ctrs ; // ServiceLocator.findServices (q);


    for (int i = 0; i < ctr.length; i++)
      {
	ContentType temp = ((IContentTypeResolver)ctr[i]).resolveSuffix (suffix);
	if (temp != null && temp.getMimetype () != null)
	  {
	    ct = temp;

	    if (temp.getMimetype ().equals (MIMETYPE_TEXT_PLAIN) 
                || temp.getMimetype ().equals (MIMETYPE_TEXT_XML))
	      { 
                // this will do, but we'll keep looking to see if we
		// find something more specific
	      }
	    else
	      { 
                // as good as we can hope for
		break;
	      }
	  }
      }

    if (ct == null)
      {
	ct = new ContentType ((String)null, (String)null, suffix);
      }
    return ct;
  }

  public static ContentType getContentType (URL url, String filename)
  {
    boolean isText = false;
    String mimetype = null;
    if (url != null && url.getProtocol ().equals ("http"))
      {
        try 
          {
	    mimetype = url.openConnection ().getContentType ();
          }
        catch (IOException ioe)
          {
            Logger.log (DefaultContentTypeResolver.class, ioe);
          }

        if (mimetype != null && !(isText = mimetype.equals ("text/plain")))
          {
            return (new ContentType (mimetype));
          } 
      }

    if (filename == null)
      {
        if (url != null)
          {
	    filename = url.getFile ();
          }
      }
    ContentType ct = getContentType (getSuffix (filename));

    mimetype = ct.getMimetype ();
    try
      {
	if (mimetype == null)
	  {
            // if the http headers said this was plain text, then use that for now
            if (isText)
              {
                return (new ContentType ("text/plain"));
              }

            if (url != null)
              {
                mimetype = url.openConnection ().getContentType ();
		ct = new ContentType (ct, mimetype);
              }
	  }
      }
    catch (IOException ioe)
      {
	Logger.log (DefaultContentTypeResolver.class, ioe);
      }
    return (ct);
  }


  public static ContentType getContentType (ContentType type, InputStream in)
    throws IOException
  {
    if (in != null)
      {
	IService[] ctr = ServiceLocator.findServices (new CTRMimetypeQuery (type.getMimetype ()));
	for (int i = 0; i < ctr.length; i++)
	  {
	    ContentType temp = ((IContentTypeResolver)ctr[i]).resolveStream (in, type);
	    if (temp != null)
	      {
		return temp;
	      }
	  }
      }
    return (type);
  }

  public static ContentType getContentType (URL url)
    throws IOException
  {
    if (url == null)
      {
	return (new ContentType ((String)null));
      }
    return (getContentType (url, url.getFile (), null));
  }

  public static ContentType getContentType (String filename, InputStream in)
    throws IOException
  {
    return getContentType (null, filename, in);
  } 

  public static ContentType getContentType (URL url, String filename, InputStream in)
    throws IOException
  {
    ContentType type = getContentType (url, filename);

    // if no input stream was passed in, get it from the url if we can
    if (in == null)
      {
	if (url != null)
	  {
	    in = url.openStream ();
	  }
      }

    // at this point, we need to determine if we have enough information
    // or if we need to analyze the stream to really figure out this type.
    if (checkInputStream (type.getMimetype ()))
      {
	type =  getContentType (type, in);
      }

    return (type);
  }

  public static ContentType getContentType (ContentElement content)
    throws IOException
  {
    ContentType type = getContentType (content.getURL (), null);

    // at this point, we need to determine if we have enough information
    // or if we need to analyze the stream to really figure out this type.
    if (checkInputStream (type.getMimetype ()))
      {
	type = getContentType (type, content.getInputStream ());
      }

    return (type);
  }

  public static boolean checkInputStream (String mimetype)
  {
    // when we are able to accurately guess text/plain types we'll do that here
    // but for now, only continue if it's text/xml 
    return (MIMETYPE_TEXT_XML.equals (mimetype));
  }

//  public static InputStream getInputStream (URLConnection urlc)
//    throws IOException
//  {
//    URL url = urlc.getURL ();
//    InputStream in = null;
//
//    // if this resource is in the cache, use it's input stream
//    if (ContentCache.contains (url))
//      { 
//	in = ContentCache.fetch (url).getInputStream ();
//      }
//    // otherwise, use the url connection's input stream
//    else
//      {
//	in = urlc.getInputStream ();
//      }
//
//    return in;
//  }

}
