/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: DocumentCache.java,v 1.7 2001/12/18 06:06:15 colin Exp $
 */
package org.xoe.core.content; 

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.Constants;
import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;


/**
 * The <code>DocumentCache</code> class provides a convenient front-end to the
 * Content Cache for handling XML documents.  The <i>fetch</i> and <i>remove</i>
 * operations apply to the Content Cache while the <i>put</i> operation only 
 * applies to the Stash.
 */
public class DocumentCache 
  implements Constants
{
  private static Hashtable documents = new Hashtable();


  /**
   * Returns the XML document specified by <code>uri</code>.
   *
   * @param uri the String specifying the location of XML document to return.
   *
   * @return the XML document specified by <code>uri</code>.
   */
  public static XDocument fetchDocument (String uri)
  {
    try
      {
	return (fetchDocument (ContentCache.buildURL (uri)));
      }
    catch (MalformedURLException mue) 
      {
        Logger.log (DocumentCache.class, "Bad URL: " + uri);
        Logger.log (DocumentCache.class, mue);
      }
    return (null);
  }

  /**
   * Returns the XML document specified by <code>uri</code> resolved relative
   * to <code>doc</code>.
   *
   * @param uri the String specifying the location, relative to <code>doc
   *              </code> of the XML document to return.
   * @param doc the XML document to use as the base for resolving <code>
   *              uri</code>.
   *
   * @return the XML document specified by <code>uri</code>.
   */
  public static XDocument fetchDocument (String uri, XDocument doc)
  {
    try
      {
	return fetchDocument (ContentCache.buildURL (uri, doc));
      }
    catch (MalformedURLException mue) 
      {
        Logger.log (DocumentCache.class, "Bad URL: " + uri);
        Logger.log (DocumentCache.class, mue);
      }
    return (null);
  }

  /**
   * Returns the XML document specified by <code>url</code>.
   *
   * @param url the URL specifying the location of XML document to return.
   *
   * @return the XML document specified by the URL passed in.
   */
  public static XDocument fetchDocument (URL url) 
  {
    if (url == null)
      {
        return (null);
      }

    XDocument doc = null;
    try
      {
        ContentElement content = ContentCache.fetch (url);
        if (content == null)
          {
            return null;
          }

        IDocumentBuilder db = (IDocumentBuilder)content.getType ().getService (IDocumentBuilder.INTERFACE);
        if (db != null) 
          {
            doc = db.getDocument (content);
          }
        else 
          {
            Logger.log (DocumentCache.class, "no IDocumentBuilder found for " + content.getURL ());
          }
      }
    catch (IOException ioe)
      {
        Logger.log (DocumentCache.class, "fetchDocument: while fetching " + url);
        Logger.log (DocumentCache.class, ioe);
      }

    return (doc);
  }

  /**
   * Adds the XML document <code>doc</code> to the Stash using <code>url</code>
   * as the index.  
   * 
   * @param url the URL to use as the index into the Stash for this document.
   * @param doc the XML document to put in the Stash
   *    
   * @exception IOException if <code>doc</code> is null.
   */
  public static void putDocument (URL url, Document doc) 
    throws IOException
  {
    if (doc == null) 
      {
        throw new IOException ("null document in putDocument");
      }

    // call getXML with create namespaces set to true to create
    // a byte array representing the document passed in.
    byte[] data = XMLWriter.getXML (doc, true, true).getBytes ("UTF8");

    // create a content element with the url passed in and the byte array 
    // reperesentation of the document passed in
    ContentElement content = Stash.getFactory ().createContentElement (url, data);

    // add the new content element to the stash
    Stash.put (content);
  }

  /**
   * Adds the XML document <code>doc</code> to the Stash using the document's
   * URL as the index.  
   * 
   * @param doc the XML document to put in the Stash
   *    
   * @exception IOException if <code>doc</code> is null.
   */
  public static void putDocument (XDocument doc) 
    throws IOException
  {
    if (doc == null) 
      {
        throw new IOException ("null document in putDocument");
      }

    putDocument(doc.getURL(), doc);
  }
  
  /**
   * Removes the XML document <code>doc</code> from the Content Cache.
   * 
   * @param doc the XML document to remove from the Content Cache.
   *    
   * @exception IOException if <code>doc</code> is null.
   */
  public static void removeDocument (XDocument doc) 
    throws IOException
  {
    if (doc == null) 
      {
        throw new IOException ("null document in removeDocument");
      }

    ContentCache.remove (doc.getURL ());
  }

  /**
   * Returns the XML document refered to by <code>filename</code> from the
   * package workspace.  If the file isn't found, the read-only instance of 
   * it is copied from the package archive, placed in the package workspace, 
   * and returned.
   *
   * @param filename the name of the XML document to return
   * @param pkgWorkspace the workspace location of the package
   * @param pkgArchive the archive location of the package
   *    
   * @exception IOException if an I/O error occurs.
   */
  public static XDocument fetchFromPackageWorkspace (String filename, String pkgWorkspace, String pkgArchive)
    throws IOException
  {
    if (!pkgWorkspace.endsWith ("/"))
      {
        pkgWorkspace += "/";
      } 
    String pkgPath = pkgWorkspace + filename;

    XDocument doc = fetchDocument (pkgPath);
    if (doc == null)
      {
	if (!pkgArchive.endsWith ("/"))
	  {
	    pkgArchive += "/";
	  } 

        // copy the document from the install location to pkg space
        XDocument readOnlyDoc = fetchDocument (pkgArchive + filename);
        if (readOnlyDoc != null)
          {
	    doc = new XDocument ();
	    Node n = doc.importNode (readOnlyDoc.getDocumentElement (), true);
	    doc.appendChild (n);
	    doc.setURL (new URL (pkgPath));
	    DocumentCache.putDocument (doc);
          }
      }

    return (doc);
  }

} // DocumentCache
