/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Stash.java,v 1.14 2001/12/18 06:05:50 colin Exp $
 */
package org.xoe.core.content; 

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.Constants;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.ResourceInstaller;
import org.xoe.core.packages.ResourceInstallerData;
import org.xoe.core.packages.XoeClassLoader;
import org.xoe.core.packages.XoePackage;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


/** 
 * The class <code>Stash</code> represents a virtual filesystem implemented 
 * as a DOM tree of content elements.  The filesystem may be read in from or 
 * flushed out to disk.  Reading from disk is done in a lazy fashion as the 
 * DOM tree is expanded.
 */
public class Stash 
  implements Constants
{
  // the document representing the VFS 
  private static final Document stashDocument;
  // the document element 
  private static final StashDirectory stashElement;
  // the factory that makes stash elements
  private static final ContentElementFactory factory;
  // the special place in the stash where packages are installed
  private static URL pkgArchiveRootURL;

  // build the basic DOM skeleton for the stash
  static
  {
    stashDocument = new XDocument ();
    stashElement = createStashRoot ();
    stashDocument.appendChild (stashElement);
    factory = new ContentElementFactory (stashDocument);    
  }

  /**
   * Returns the factory to use for creating a new Stash content element.
   * 
   * @return the factory to use for creating a new Stash content element.
   */
  public static ContentElementFactory getFactory ()
  {
    return factory;
  }

  /**
   * Returns 
   */
  public static URL getPkgArchiveRoot ()
  {
    if (pkgArchiveRootURL == null)
      {
	try
	  {
	    pkgArchiveRootURL = new URL (PACKAGE_ARCHIVE_ROOT);
	  }
	catch (MalformedURLException mue)
	  {
            // if we fail here, no point in going on...
	    Logger.log (Stash.class, "encountered a fatal error while establishing the package root:");
	    Logger.log (Stash.class, mue);
            System.exit (1);           
	  }
      }
    return (pkgArchiveRootURL);
  }

  public static URL getWorkingDirectory (XoePackage pkg)
  {
    ResourceInstallerData data = (ResourceInstallerData)pkg.getExData (ResourceInstaller.NAMESPACE);
    if (data == null)
      {
	return null;
      }

    return (data.getWorkingDirectory ());
  }

  public static URL getWorkingDirectory (Object o)
  {
    return (getWorkingDirectory (o.getClass ()));
  }

  public static URL getWorkingDirectory (Class c)
  {
    return (getWorkingDirectory (c.getClassLoader ()));
  }

  public static URL getWorkingDirectory (ClassLoader cl)
  {
    if (cl instanceof XoeClassLoader)
      {
	return (getWorkingDirectory (((XoeClassLoader)cl).getPackage ()));
      }

    return (null);
  }

  public static String getUserWorkspaceURI (String username)
  {
    return (STASH_PROTOCOL + ":/home/" + username + "/data/");
  }

  public static String getPackageArchiveURI (String pkgURI)
  {
    return (PACKAGE_ARCHIVE_ROOT + pkgURI + "/");
  }

  public static String getPackageWorkspaceURI (String username, String pkgName)
  {
    return (getUserWorkspaceURI (username) + pkgName + "/");
  }

  public static void setPath (String path)
    throws StashException
  {
    if (stashElement.expanded == true)
      {
        throw new StashException ("cannot set stash path after read from disc");
      }

    if (path == null)
      {
        throw new StashException ("invalid stash location: " + path);
      }

    File dir = new File (path);
    stashElement.setFile (dir);

    if (!dir.exists ())
      {
        throw new StashException ("invalid stash location: " + path);
      }

    if (!dir.isDirectory ())
      {
        throw new StashException ("stash location is not a directory: " + path);
      }

    Logger.log (Stash.class, "stash directory is: " + dir.getAbsolutePath ());
  }

  public static String getPath ()
  {
    return stashElement.getFile ().getPath ();
  }

  public static boolean contains (URL url)
    throws IOException
  {
    return (fetch (url) != null);
  }

  public static boolean contains (String src)
    throws IOException
  {
    return contains (new URL (src));
  }

  /**
   * fetch looks up the ContentElement by URL.
   */
  public static ContentElement fetch (URL url)
    throws StashException
  {
    if (url == null)
      {
        throw new StashException ("null url in fetch request");
      }
    return (fetch (url.toString ()));
  }

  /**
   * fetch looks up the ContentElement by src string.  
   * If it's not found it simply returns null.
   */
  public static ContentElement fetch (String src)
    throws StashException
  {
    if (src == null)
      {
        throw new StashException ("null src in fetch request");
      }

    if (!src.startsWith (STASH_PROTOCOL))
      {
        throw new StashException ("Bad protocol in fetch request: " + src);
      }

    String path = src.substring (src.indexOf (":/") + 1);

    // special case for the root directory.
    if (path.equals ("/"))
      {
        return (stashElement);      
      }

    // chop off initial separator
    if (path.startsWith (File.separator))
      {
        path = path.substring (File.separator.length ());
      }

    // chop off trailing separator
    if (path.endsWith (File.separator))
      {
        path = path.substring (0, path.length () - File.separator.length ());
      }

    return (findFirstNode (path));
  }

  /**
   * Add a ContentElement to the stash.  Only elements whose protocol
   * is stash: can be added.  Any existing element at the same URL is
   * removed and returned.  All subdirectories in the path are created
   * if they do not exist already.
   */
  public static ContentElement put (ContentElement content)
    throws IOException
  {
    String path = content.getURL ().getFile ();
    if (path.indexOf (" ") > -1)//file paths should not contain spaces
	{
	  throw new IOException ("No space allowed in ContentElement file path: [" + path + "]");
	}

    // get & remove the existing element
    ContentElement oldElement = remove (content.getURL ());

    StashDirectory parent = stashElement;

    if (path.startsWith (File.separator))
      {
        path = path.substring (File.separator.length ());
      }

    int pos = path.lastIndexOf (File.separator);

    String nodeName = path.substring (pos+1); 

    if (pos != -1)
      {
        path = path.substring (0, pos);
        if ((parent = (StashDirectory) findFirstNode (path)) == null)
          {
            parent = putDirectory (path);
          }
      }

    content.setNodeName (nodeName); 
    
    // append the content element as a child of the final directory.
    parent.appendChild (content);

    // ensure that the type has been resolve for this content element
    content.getType ();

    return (oldElement);
  }

  /**
   * Add a directory hierarchy to the stash
   */
  public static StashDirectory putDirectory (String dirName)
    throws IOException
  {
    StashDirectory parent = stashElement;
    StashDirectory newDir = null;

    // construct a directory node 
    int j = 0; 
    int i = 0;
    int len = dirName.length ();
    while (i < len)
      {
        // find the next directory name in path
        String singleDir;
        j = dirName.indexOf (File.separator, i);
        if (j == -1) 
          { 
            j = len; 
          }

        singleDir = dirName.substring (i, j); 
        i = j + File.separator.length ();

        newDir = (StashDirectory) XPath.findFirstNode (parent, singleDir);
        if (newDir == null)
          {
            newDir = createDirectory (parent, singleDir);
          }

        parent = newDir;
      }

    return (newDir);
  }

  /**
   * Create a the stash root StashDirectory (ContentElement)
   */
  static StashDirectory createStashRoot ()
  {
    try
      {
        return (createDirectory (null, "stash"));
      }
    catch (IOException ioe)
      {
        Logger.log (Stash.class, ioe);
        System.exit (1);
      } 
    return (null);
  }

  /**
   * Create a single StashDirectory (ContentElement) and add it to the given parent.
   * Passing in 'null' for the parent will create the root stash directory.
   */
  static StashDirectory createDirectory (StashDirectory parent, String dirName)
    throws IOException
  {
    File file = null;
    URL url = null;
    StashDirectory newDir = null;

    if (parent == null)
      {
	// create the stash root directory
	file = new File ("/");
	url = new URL (STASH_PROTOCOL + ":/");
      }
    else
      {
	// insert the new directory into the tree
	//Logger.log (Stash.class, "adding dir:" + dirName);
	file = new File (parent.getFile ().getPath () + File.separator + dirName);
	String parentURL = parent.getURL ().toString ();
	// [FIXME] we have to so this because the file protocol fails to construct
	// URLs will two initial slashed.  (and the stash protocol handler inherits this
	// bug.
	if (parentURL.endsWith (File.separator))
	  {
	    url = new URL (parentURL + dirName);
	  }
	else
	  {
	    url = new URL (parentURL + File.separator + dirName);
	  }
      }

    newDir = new StashDirectory (file, url, stashDocument);

    if (parent != null)
      {
	parent.appendChild (newDir);
      }

    return (newDir);
  }

  public static void putAlias (String name, String src)
    throws IOException
  {
    // get the content element pointed to by the src arg
    ContentElement content = fetch (src);
    if (content == null)
      {
        throw new StashException("content '" + src + "' not found");
      }
    addAlias (name, content);    
  }

  /**
   * aliases are copies for now.
   */
  public static ContentElement addAlias (String name, ContentElement content)
    throws IOException
  {
    ContentElement alias = content.makeAlias (new URL (name));
    return (put (alias));
  }

  /**
   * not implemented.  aliases are copies for now.
   */
  public static String getAlias (String name)
  {
    return (null);
  }

  /** 
   * Remove the content element at the given URL.
   * [FIXME] should we remove parents who are now empty?
   */
  public static ContentElement remove (URL url)
    throws StashException
  {
    if (url == null)
      {
        throw new StashException ("null url in remove request");
      }
    return (remove (url.toString ()));
  }

  public static ContentElement remove (String src)
    throws StashException
  {
    if (src == null)
      {
        throw new StashException ("null src string in remove request");
      }

    if (!src.startsWith (STASH_PROTOCOL))
      {
        throw new StashException ("Bad protocol in remove request: " + src);
      }

    ContentElement content = fetch (src);
    if (content != null)
      {
        Node parent = content.getParentNode ();
        if (parent != null)
          {
            return ((ContentElement)parent.removeChild (content));
          }
      }
    return (null);
  }

  public static void removeAlias (String name)
    throws StashException
  {
    remove (name);
  }

  public static ContentElement findFirstNode (String query)
  {
    return (ContentElement) XPath.findFirstNode (stashElement, query);
  }

  /**
   * Returns the first node satisfying the xpath query
   * <code>query</code> using the stashElement as the context.
   */
  public static ContentElement findFirstNode (XPath query)
  {
    return (ContentElement) XPath.findFirstNode (stashElement, query);
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using this node as the context.
   */
  public static NodeList findNodes (String query)
  {
    return XPath.findNodes (stashElement, query);
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using this node as the context.
   */
  public static NodeList findNodes (XPath query) 
  {
    return XPath.findNodes (stashElement, query);
  }

  public static URL createUniqueFilename (URL baseURL, String filename, String suffix)
    throws IOException
  {
    URL u = new URL (baseURL, filename + "." + suffix);
    ContentElement c = fetch (u);
    while (c != null)
      {
	u = new URL (baseURL, filename + Math.random () + "." + suffix);
	c = ContentCache.fetch (u);
      }

    return (u);
  }

  /** 
   * Returns the XML representation of the Stash document.
   *
   * @return the XML representation of the Stash document.
   */ 
  public static String getXML ()
  {
    return XMLWriter.getXML (stashElement);
  }

  public static FileContentElement createFileContentElement (URL url, File file)
    throws FileNotFoundException, IOException
  {
    if (url == null)
      {
        return (null);
      }

    // get a new ContentElement from the cacheDocument
    return (new FileContentElement (file, url, stashDocument));
  }
} 
