/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: DOMlet.java,v 1.2 2001/10/21 04:13:28 colin Exp $
 */
package org.xoe.core.dom;

import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import org.xoe.core.content.Stash;
import org.xoe.core.util.Logger;


/**
 * DOM Document representing an XML file.
 */
public class DOMlet 
  extends XDocument
{    
  /**
   * parser object for this document
   */ 
  private Parser parser;


  // -------------------------------------------------------------------------

  private void expand () 
  {
    if (parser != null)
      {
	try 
          {
	    Parser p = parser;
            parser = null;
	    p.expandNode (this);
          }
	catch (IOException e) 
	  {
            throw new DOMException (e);
	  }
      }
  }


  //----------------------------------------------------------------------
  // All the following functions simply add a call the expand () 

  //----------------------------------------------------------------------
  // org.w3c.dom.Node

  public String getValue (String name)
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.getValue (name));
  }

  public Node cloneNode (boolean deep)
  {
    // ensure that the document is fully parsed
    if (parser != null)
      {
	expand ();
      }
    return (super.cloneNode (deep));
  }
  

  //----------------------------------------------------------------------
  // org.xoe.core.dom.ParentNode

  public Node getFirstChild ()
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.getFirstChild ());
  }

  public Node getLastChild ()
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.getLastChild ());
  }

  public Node insertBefore (Node newChild, Node refChild) 
    throws DOMException 
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.insertBefore (newChild, refChild));
  }

  public Node replaceChild (Node newChild, Node oldChild) 
    throws DOMException
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.replaceChild (newChild, oldChild));
  }

  public Node removeChild (Node oldChild) 
    throws DOMException
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.removeChild (oldChild));
  }

  public Node appendChild (Node newChild) 
    throws DOMException 
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.appendChild (newChild));
  }

  public boolean hasChildNodes ()  
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.hasChildNodes ());
  }

  public Node item (int index)   
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.item (index));
  }

  public int getLength () 
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.getLength ());
  }


  // -------------------------------------------------------------------------
  // org.w3c.dom.Document

  public DocumentType getDoctype ()
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.getDoctype ());
  }

  public Element getDocumentElement ()
  {
    if (parser != null)
      {
	expand ();
      }
    return (super.getDocumentElement ());
  }

  public String getDoctypeNameNS ()
  {
    //if the node is expanded, call the super because we already know the namespace.
    //if it's not expanded, call parser.getDocType, which wil find the namepsace efficiently.

    if (parser == null)
      {
	return super.getDoctypeNameNS ();
      }
    else
      {
	try 
	  {
	    return parser.getDoctypeNameNS (this);
	  }
	catch (IOException ioe)
	  {
	    Logger.log (this, ioe);
	  }
      }
    //this should never happen. maybe we should return something else
    return (null);
  }


  // -------------------------------------------------------------------------
  // static create methods

  public static DOMlet create (String url)
  {
    try 
      {
	if (url.indexOf (':') == -1)
	  {
            // no protocol was passed in so build it relative to the stash path
 
            // is the first character a file separator?
            if (!url.startsWith (File.separator))
              {
                // no leading slash, so add one
                url =  Stash.getPath () + File.separator + url;
              }
            else
              {
                // already have the leading slash, so just append
                url =  Stash.getPath () + url;
              }
	  }

	return create (new URL (url));
      }
    catch (MalformedURLException mue) 
      {
	Logger.log (DOMlet.class, "Bad URL: " + url);
	Logger.log (DOMlet.class, mue);
      }
    return null;
  }

  public static DOMlet create (URL url)
  {
    DOMlet newDOMlet = new DOMlet ();
    try 
      {
	newDOMlet.parser = new Parser (url, newDOMlet);
	newDOMlet.setURL (url);
	return newDOMlet;
      }
    catch (IOException ioe) 
      {
	Logger.log (newDOMlet, ioe);
	return (null);
      }
   
  }

  public static DOMlet create (char[] data, URL url)
  {
    DOMlet newDOMlet = create (data, url.toString ());
    newDOMlet.setURL (url);
    return newDOMlet;
  }

  /**
   * Create a DOM tree from the char array data. The docName is used
   * primarily for printing error messages in LazyDOM.
   */
  public static DOMlet create (char[] data, String docName)
  {
    DOMlet newDOMlet = new DOMlet ();
    try 
      {
	newDOMlet.parser = new Parser (data, newDOMlet, docName);
	return newDOMlet;
      }
    catch (IOException ioe) 
      {
	return (null);
      }
  }

  public static DOMlet create (InputStream in, URL url)
  {
    DOMlet newDOMlet = create(in, url.toString ());
    newDOMlet.setURL (url);
    return newDOMlet;
  }
  
  public static DOMlet create (InputStream in, URL url, int estSize)
  {
    DOMlet newDOMlet = create(in, url.toString (), estSize);
    newDOMlet.setURL (url);
    return newDOMlet;
  }

  /**
   * Create a DOM tree from an InputStream. The docName is used
   * primarily for printing error messages in LazyDOM.
   */

  public static DOMlet create (InputStream in, String docName)
  {
    return create (in, docName, -1);
  }

  public static DOMlet create (InputStream in, String docName, int estSize)
  {
    DOMlet newDOMlet = new DOMlet ();
    try 
      {
	newDOMlet.parser = new Parser (in, newDOMlet, docName, estSize);
	return newDOMlet;
      }
    catch (IOException ioe) 
      {
      Logger.log (DOMlet.class, ioe);
	return (null);
      }
  }


  // -------------------------------------------------------------------------
  // org.xoe.dom.core.Cacheable

  public void reset ()
  {
    super.reset ();
    parser = null;
  }
}
