/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: ElementNamedNodeMapAdaptor.java,v 1.4 2001/12/13 21:53:57 colin Exp $
 */
package org.xoe.core.dom;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import org.xoe.core.Constants;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;


public class ElementNamedNodeMapAdaptor
  implements NamedNodeMap
{
  XElement parent;
  XAttr[] attrs;
  int nr;

  ElementNamedNodeMapAdaptor ()
  {
  }

  ElementNamedNodeMapAdaptor (XElement newParent)
  {
    parent = newParent;
  }
  
  public Node getNamedItem (String itemName)
  {
    return getNamedItemNS (null, itemName);
  }

  public Node setNamedItem (Node newItem)
  {
    return setNamedItemNS (newItem);
  }

  public Node removeNamedItem (String name) 
    throws DOMException
  {
    return removeNamedItemNS (null, name);
  }

  public Node item (int position)
  {
    if (position < 0 || position >= nr)
      {
	return (null);
      }
    else
      {
	return (attrs[position]);
      }
  }

  public int getLength ()
  {
    return (nr);
  }

  public Node getNamedItemNS (String nameSpaceURI, String itemName)
  {
    for (int i = 0; i < nr; i++)
      {
	XAttr attr = attrs[i];
	if (attr.localname.equals (itemName))
	  {
	    if (attr.namespace == nameSpaceURI 
                || (attr.namespace != null 
                    && attr.namespace.equals (nameSpaceURI)))
	      {
		return (attr);
	      }
	  }
      }
    return (null);
  }

  /** 
   * this method now throws exceptions if new Item is a part of a different 
   * document than that which created this, or if newItem is an attribute 
   * that already belongs to an element.
   */
  public Node setNamedItemNS (Node newItem)
  {
    XAttr item = (XAttr) newItem;

    if (parent != null 
        && item.element != null 
        && parent.getOwnerDocument () != item.element.getOwnerDocument ())
      {
	throw new DOMException (DOMException.WRONG_DOCUMENT_ERR, 
                                (parent == null ? 
                                   "null" : parent.getOwnerDocument ().getNodeName ()));
      }

    if (item.getNodeType() == Node.ATTRIBUTE_NODE 
        && item.element != null && item.element != parent)
      {
	throw new DOMException (DOMException.INUSE_ATTRIBUTE_ERR, 
                                "Attr name = " 
                                 + item.getNodeName ());
      }

    // Look for 'xmlns' attribute which is treated specially
    if (Constants.XMLNS.equals (item.getPrefix ()) 
        || (item.getPrefix () == null 
            && Constants.XMLNS.equals (item.localname)))
      {
	item.namespace = Constants.XMLNS_NAMESPACE;
      }

    item.element = parent;

    for (int i = 0; i < nr; i++)
      {
	XAttr attr = attrs[i];
	if (attr.localname.equals (item.localname))
	  {
	    if (attr.namespace == item.namespace 
                || (attr.namespace != null 
                    && attr.namespace.equals (item.namespace)))
	      {
		attrs[i] = item;
		return (attr);
	      }
	  }
      }

    if (nr == 0)
      {
	attrs = new XAttr[8];
      }
    else if (nr == attrs.length)
      {
	XAttr[] old = attrs;
	attrs = new XAttr[nr * 2];
	System.arraycopy(old, 0, attrs, 0, nr);
      }

    attrs[nr] = item;
    nr++;
    return (null);
  }

  public Node removeNamedItemNS (String nameSpaceURI, String name) 
    throws DOMException
  {
    for (int i = 0; i < nr; i++)
      {
	XAttr attr = attrs[i];
	if (attr.localname.equals (name))
	  {
	    if (attr.namespace == nameSpaceURI 
                || (attr.namespace != null 
                    && attr.namespace.equals (nameSpaceURI)))
	      {
		System.arraycopy(attrs, i+1, attrs, i, nr-i-1);
		nr--;
		return (attr);
	      }
	  }
      }

    throw new DOMException (DOMException.NOT_FOUND_ERR, 
                            "No item found for name: " 
                             + name);
  }

  // ----------------------------------------------------------------------

  ElementNamedNodeMapAdaptor cloneMap (XElement newOwner)
  {
    ElementNamedNodeMapAdaptor n = new ElementNamedNodeMapAdaptor (newOwner);
    n.nr = nr;
    n.attrs = new XAttr[n.nr];
    for (int i = 0; i < nr; i++)
      {
	n.attrs[i] = (XAttr) attrs[i].cloneNode (true);
	n.attrs[i].element = newOwner;
      }

    return (n);
  }

  public String getPrefix (String prefix)
  {
    return getPrefix (this, prefix);
  }

  public static String getPrefix (NamedNodeMap map, String prefix)
  {
    String ns = Constants.XMLNS;
    if (prefix == null)
      {
	prefix = ns;
	ns = null;
      }

    Attr attr = (Attr) map.getNamedItemNS (Constants.XMLNS_NAMESPACE, prefix);
    if (attr != null 
        && (ns == null 
            || ns.equals (attr.getPrefix ())))
      {
	return (attr.getValue ());
      }

    return (null);
  }

  void fixup (XElement elem, Node parent)
  {
    elem.attributes = this;
    this.parent = elem;
    for (int i = 0; i < nr; i++)
      {
	XAttr attr = attrs[i];
	attr.element = elem;
	String nsPrefix = attr.getPrefix ();
	if (nsPrefix != null)
	  {
	    String nskeyAttr = Constants.XMLNS + ':' + nsPrefix;
	    XAttr nsattr = (XAttr)getNamedItemNS (null, nskeyAttr);
	    String nsAttr;
	    if (nsattr != null)
	      {
		nsAttr = nsattr.getValue ();
	      }
	    else
	      {
		nsAttr = DOMUtils.lookupNSForPrefix (parent, nsPrefix, null);
	      }
	    attr.namespace = nsAttr;
	  }
      }
  }

  public XAttr getItemByValue (String value)
  {
    for (int i = 0; i < nr; i++)
      {
	XAttr attr = attrs[i];
	if (attr.getValue().equals (value))
	  {
	    return (attr);
	  }
      }
    return (null);
  }

}
