/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: ParentNode.java,v 1.1 2001/10/21 03:56:01 colin Exp $
 */
package org.xoe.core.dom;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.util.Logger;


/**
 *  ParentNode is a DOM node which supports children.
 */
abstract public class ParentNode
  extends XNode
  implements NodeList
{
  // Used to manage DOM hierarchy.
  XNode tail;

  // -------------------------------------------------------------------------
  // org.w3c.dom.Node

  public NodeList getChildNodes () 
  {
    return this;
  }

  public Node getFirstChild () 
  {
    if (tail == null) 
      {
	return (null);
      }
    return (tail.nextchild);
  }

  public Node getLastChild () 
  {
    if (tail == null) 
      {
	return (null);
      }
    return (tail);
  }

  public NamedNodeMap getAttributes () 
  {
    return null;
  }

  public Node insertBefore (Node newChild, Node refChild) 
    throws DOMException 
  {
    if (refChild == null)
      {
        return (appendChild (newChild));
      }

    XNode newNode = (XNode)newChild;

    //CODE FOR THROWING EXCEPTIONS
    if ((newNode.getOwnerDocument () != this.getOwnerDocument ())
        && newNode.getOwnerDocument () != null
        && newNode.getOwnerDocument () != this)
      {
        Logger.log (this, "The newChild is " 
                           + newNode.toString () 
                           + ". It's owner is " 
		           + newNode.getOwnerDocument ());

        Logger.log (this, "This node is " 
                           + this.toString () 
                           + ". It's owner is " 
                           + this.getOwnerDocument ());

        throw new DOMException (DOMException.WRONG_DOCUMENT_ERR, 
                                "Node " 
                                + newNode 
                                + " with owner "
                                + newNode.getOwnerDocument () 
                                + " cannot be inserted into "
                                + this.getOwnerDocument ());
      }

    if (((XNode)refChild).parent != this)
      {
	Logger.log (this, "refChild has parent: " 
                           + ((XNode)refChild).parent 
		           + ".  Parent should be this node: " 
                           + this);

	throw new DOMException (DOMException.NOT_FOUND_ERR, 
                                "Node " 
				 + refChild 
                                 + "is not a child of " 
                                 +  this);
      }
    
    for (XNode n = this; n != null; n = n.parent)
      {
	if (n == newNode)
	  {
	    throw new DOMException (DOMException.HIERARCHY_REQUEST_ERR, 
                                    "insertBefore is trying to insert an ancestor of this node");
	  }
      }
    
     if (newNode.getNodeType () == Node.DOCUMENT_FRAGMENT_NODE)
      {
	if (newNode.hasChildNodes ())
	  {
	    Node child = newNode.getFirstChild ();
	    Node temp;
	    while (child != null)
	      {
		temp = child.getNextSibling ();
		newNode.removeChild (child);
		insertBefore (child, refChild);
		child = temp;
	      }
	  }

	return (newNode);
      }
     else 
       {
	 if (tail != null) 
	   {
	     newNode.parent = this;
	     XNode currentNode = tail;
	     do 
	       {
		 if (currentNode.nextchild == refChild) 
		   {
		     newNode.nextchild = (XNode)refChild;
		     currentNode.nextchild = newNode;
		     return (newNode);
		   }

		 currentNode = currentNode.nextchild;
	       } 
	     while (currentNode != tail);
	   }

	 return (null);
       }
  }

  public Node replaceChild (Node newChild, Node oldChild) 
    throws DOMException 
  {
  
    XNode nc = (XNode)newChild;
    XNode oc = (XNode) oldChild;

    if ((nc.getOwnerDocument () != this.getOwnerDocument ())
        && nc.getOwnerDocument () != null
        && nc.getOwnerDocument () != this)
      {
        Logger.log (this, "newChild is "  
                           + nc.toString () 
                           + ". It's owner is " 
                           + nc.getOwnerDocument ().toString ());

        Logger.log (this, "this is " 
                           + this.toString () 
                           + ". It's owner is " 
                           + this.getOwnerDocument ().toString ());

        throw new DOMException (DOMException.WRONG_DOCUMENT_ERR, 
                                "Node " 
                                 + nc 
                                 + "with owner"
                                 + nc.getOwnerDocument () 
                                 + " cannot replace children of "
                                 + this.getOwnerDocument ());
      }

    if (oc.parent != this)
      {
	Logger.log (this, "oldChild has parent: " 
                           + oc.parent 
		           + ". Parent should be this node: " 
                           + this);

	throw new DOMException (DOMException.NOT_FOUND_ERR, 
                                "Node " 
				 + oc 
                                 + "is not a child of " 
                                 +  this);
      }

    if (nc.getNodeType () == Node.DOCUMENT_FRAGMENT_NODE)
      {
	if (nc.hasChildNodes ())
          {
	    Node child = nc.getLastChild ();
	    nc.removeChild (child);
	    replaceChild (child, oc);
	    insertBefore (nc, child);
	  }

	return (nc);
      }
    else
      {
	if (tail != null) 
	  {
	    nc.parent = this;
	    oc.parent = null;
	    XNode n = tail;
	    if (n.nextchild == n 
                && n == oc) 
	      {
		nc.nextchild = nc;
		tail = nc;
		return (nc);
	      }
	    else 
              {
		do 
                  {
		    if (n.nextchild == oc) 
		      {
			nc.nextchild = n.nextchild.nextchild;
			n.nextchild = nc;
			if (tail == oc) 
                          {
			    tail = nc;
			  }
			return (nc);
		      }
		    n = n.nextchild;
		  } while (n != tail);
	      }
	  }
	return (null);
      }
  }

  public Node removeChild (Node oldChild) 
    throws DOMException 
  {
    if (tail != null) 
      {
	XNode n = tail;
	if (n.nextchild == n && n == oldChild) 
	  {
	    tail = null;
	    ((XNode) oldChild).parent = null;	     
	    return (oldChild);
	  } 
	else 
	  { 
	    do 
              {
		if (n.nextchild == oldChild) 
		  {
		    n.nextchild = n.nextchild.nextchild;
		    if (tail == oldChild) 
		      {
			tail = n;
		      }
		    ((XNode) oldChild).parent = null;
		    return (oldChild);
		  }
		n = n.nextchild;
	      } 
	    while (n != tail);
	  }
      }
    return (null);
  }

  public Node appendChild (Node newChild) 
    throws DOMException 
  {
    XNode nc = (XNode)newChild;
    if (nc.parent != null)
      {
	throw new DOMException (DOMException.HIERARCHY_REQUEST_ERR, 
			       "Node " 
                                + nc 
                                + " with parent " 
                                + nc.parent 
                                + " cannot be appended to " 
                                + this); 
      }

    if ((nc.getOwnerDocument () != this.getOwnerDocument ())
	&& nc.getOwnerDocument () != null
	&& nc.getOwnerDocument () != this)
      {
	Logger.log (this, "newChild is " 
			   + nc.toString () 
			   + ". It's owner is " 
			   + nc.getOwnerDocument ());

	Logger.log (this, "This node is " 
			   + this.toString () 
			   + ". It's owner is " 
			   + this.getOwnerDocument ());

	throw new DOMException (DOMException.WRONG_DOCUMENT_ERR, 
                                "Node " 
                                + nc 
                                + "with owner"
				+ nc.getOwnerDocument () 
                                + " cannot be appended to children of "
				+ this.getOwnerDocument ());
      }

    if (nc.getNodeType () == Node.DOCUMENT_FRAGMENT_NODE)
      {
	if (nc.hasChildNodes ())
	  {
	    Node child = nc.getFirstChild ();
	    Node temp;
	    while (child != null)
	      {
		temp = child.getNextSibling();
		nc.removeChild (child);
		appendChild (child);
		child = temp;
	      }
	  }
	return (nc);
	
      }
    else
      {
	nc.parent = this;
	if (tail == null) 
	  {
	    nc.nextchild = nc;
	  }
	else 
	  {
	    nc.nextchild = tail.nextchild;
	    tail.nextchild = nc;
	  }
	tail = nc;
	return (nc);
      }
  }

  public boolean hasChildNodes() 
  {
    return (tail != null);
  }

  public Node cloneNode (boolean deep) 
  {
    ParentNode nnode = (ParentNode)super.cloneNode (deep);
    nnode.tail = null;
    if (deep == true && tail != null) 
      {
	for (Node child = getFirstChild (); 
             child != null; 
             child = child.getNextSibling ()) 
	  {
	    Node nchild = child.cloneNode (true);
	    if (nchild != null) 
	      {
		nnode.appendChild (nchild);
	      }
	  }
      }
    return (nnode);
  }


  // -------------------------------------------------------------------------
  // org.w3c.dom.NodeList

  public Node item (int index)
  {
    if (tail == null) 
      {
	return (null);
      }

    XNode n = tail.nextchild;
    for (int i = 0; i < index; i++) 
      {
	if (n == tail) 
	  {
	    return (null);
	  }
	n = n.nextchild;
      }

    return (n);
  }

  public int getLength ()
  {
    if (tail == null) 
      {
	return (0);
      }

    int i = 1;
    for (XNode n = tail.nextchild; n != tail; n = n.nextchild) 
      {
	i++;
      }

    return (i);
  }

  // -------------------------------------------------------------------------
  // org.xoe.core.dom.Cacheable

  //  reset is called by classes its subclasses and is used by the 
  //  element cache to tell nodes to clear all state information
  public void reset ()
  {
    tail = null;
    super.reset ();
  }

}
