/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: Parser.java,v 1.6 2001/11/10 02:44:03 samc Exp $
 */
package org.xoe.core.dom;

import java.io.CharArrayReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.Reader;
import java.net.URL;
import java.net.URLConnection;
import java.util.Vector;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Document;

import org.xoe.core.Constants;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;


public class Parser 
  implements Constants 
{
  private final static String XML_STYLESHEET_PI = "xml-stylesheet";

  private char[] buffer;

  // after a call to read the char that was read is at buffer[blen]
  // the real end of the buffer - blen may not be the end because of unreads
  // current line number for read (), doesn't decrease on unread ()s
  private int blen;        
  private int blenMax;            
  private int lineNum=1;          

  // help report xml errors in files
  private String fileName;               

  private EntityResolver defaultResolver = DefaultEntityResolver.getInstance ();

  /** 
   * the document being built
   */
  private XDocument doc;

  // Reused. Often. In recursive code!
  // Use with caution.
  private TokenInfo token = new TokenInfo ();

  /**
   * Creates a new <code>Parser</code> instance, based on the file at
   * that URL.
   *
   * @param file an <code>URL</code> value which is the adress of the
   * document to be parsed
   * @exception IOException if an error occurs reading the URL
   */
  public Parser (URL file, XDocument d) 
    throws IOException 
  {
    this (file, "UTF8", d);
  }

  public Parser (URL file, String enc, XDocument d) 
    throws IOException 
  {
    this.doc = d;
    this.fileName = file.toString ();

    URLConnection conn = file.openConnection ();
    conn.connect ();
    int contentLength;
    try 
      { 
	contentLength = Integer.parseInt (conn.getHeaderField ("content-length"));
      }
    catch (NumberFormatException e)
      {
	Logger.log(this, "content-length header for " + file + " is not a Number: " + conn.getHeaderField ("content-length"));
	contentLength = 256;
      }

    fillBuffer(conn.getInputStream (), enc, contentLength);
  }

  public Parser (char[] data, XDocument d, String fileName) 
    throws IOException 
  {
    //     Logger.log (this, "constructing Parser object for " + fileName);
    this.doc = d;
    this.fileName = fileName;
    buffer = data;
    blenMax = data.length;
  }
    
  public Parser (InputStream file, XDocument d, String fileName, int estSize) 
    throws IOException 
  {
    this (file, "UTF8", d, fileName, estSize);
  }

  public Parser (InputStream file, String enc, XDocument d, String fileName, int estSize) 
    throws IOException 
  {
    this.doc = d;
    this.fileName = fileName;
    fillBuffer(file, enc, estSize);
  }

  private void fillBuffer (InputStream is, String enc, int estSize) 
    throws IOException
  {
    Reader reader = new InputStreamReader (is, enc);

    if ( reader == null) 
      {
 	throw new IOException ( "file doesn't exist");
      }
    
    if (estSize < 1) 
      {
	estSize = 256;
      }
    buffer = new char[estSize];
    blenMax=0;
    int ch;
    if (reader != null) 
      {
	while ((ch =reader.read ()) != -1)
	  {
	    if (blenMax == buffer.length) 
	      {
//  		Logger.log (this, " file " + fileName + " increasing buffer size. Old length = " + buffer.length );
		char[] o = buffer;
		buffer = new char[o.length * 2];
		System.arraycopy (o, 0, buffer, 0, o.length);
	      }
	    buffer[blenMax++]= (char)ch;
	  }
      }
    reader.close ();
  }

  private void setDocument (XDocument doc) 
  {
    this.doc = doc;
  }

  private synchronized String getString (int start, int end) 
  {
    if (end > start) 
      {
	return new String (buffer, start, end-start);
      }
    else 
      {
	return "";
      }
  }

  public void expandNode (Node parent) 
    throws IOException
  {
    long start = System.currentTimeMillis ();
    expandNode(parent, -1);
    long end = System.currentTimeMillis ();
    //Logger.log (this, fileName + " expandNodeTime = " + (end-start));
  }

  private void expandNode (Node parent, int depth) 
    throws IOException
  {    
    // sanity check
    if (buffer == null) 
      {
	Logger.log (this, " Parser. exandNode() called with null buffer");
	return;
      }

    boolean angle = false;
    boolean endtag = false;
    boolean intag = false;
    boolean intext = false;
    int ch = 0;
    int lch = 0;
    int last_open_tag_loc=0;

    int startText=-1; 
    int startElement = -1; 
    
    for (;;) 
      {
	lch = ch+1-1;
	ch = read ();
	if (ch == -1) 
	  {
	    buffer = null;
	    break;
	  }
	out : switch (ch) 
	  {
	  case ' ' :
	  case '\t': 
	  case '\n':
	    break out;

	  case '<':
	    if (intag) 
	      {  // < inside tag, warn
		Logger.log (this, "XML parse error in " + fileName + ": line " + lineNum);
		Logger.log (this, "unexpected '<' inside start or end tag at offset " + blen); 
		break out;
	      }
	    else if (intext)  // text node just ended
	      {
		// create text node
		parent.appendChild (doc.createTextNode (getString (startText, blen-1)));
	      }
	    startElement = blen;
	    intag = true;
	    intext = false;
	    break out;

	  case '>':
	    if (intag) 
	      {
		if (!endtag)  // end of start or <xxx/> tag
		  {
		    Element e = createElementNode(startElement, blen-2, parent, doc);
		    parent.appendChild (e);
		    if (lch != '/')
		      {
			if (depth -1 == 0)
			  return;
			
			if (e.getNodeName () == "script")
			  {
			    startText = blen;
			    readUntil ("</script>", "<script> tag");
			    e. appendChild(doc.createComment(getString (startText, blen- 10)));
			  }
			else
			  {
			    expandNode (e, depth == 0 ? 0 : depth -1);
			  }
		      }
		  } 
		else 
		  { 
		    // check node name against end tag 
		    readToken(startElement+1, token);
		    if (!token.equals(parent.getNodeName()))
		      { // end/start tag mismatch
			Logger.log (this, fileName + ": line: " + lineNum + " got </" + token + ">, expected </" + parent.getNodeName() + ">");
			// check if one of the ancestors node name matches this tag
			String parentNodeName = parent.getNodeName();
			parent = parent.getParentNode();
			while (parent != null)
			  {
		            if (token.equals(parent.getNodeName()))
			      {
				Logger.log (this, " Inserting </" + parentNodeName + "> and continuing");
				rewind(startElement - 2);
				return;
			      }
			    parent = parent.getParentNode();			    
			  }
			Logger.log (this, " Assuming typo and continuing");
		      }
		    return ;
		  }
		intag = false;
	      }
	    break;
	    
	  default : 
	    if (!intag && !intext)
	      {
		intext = true;
		startText = blen-1;
		if (Character.isWhitespace((char) lch))
		  startText --;
	      }
	    switch (ch)
	      {
	      case '\'':
	      case '"':
		int quote = ch;
		if (!intag) 
		  {
		    break out;
		  }
		for (;;) // Skip to the next quote
		  {
		    int startLine = lineNum;
		    ch = read ();
		    switch (ch)
		      {
		      case '&':
			handleEntities ( /* do not use default resolver */ false); // char entities are expanded in getAttributes,
			break;
		      case '\\':
			read ();
			break;
		      case -1:
			Logger.log (this, "Unexpected eof : unterminated quoted string started at line " + startLine);
			quote = -1;
		      }
		    if (ch == quote) 
		      {
			break;
		      }
		  }
		break out;

	      case '!':	// comment ENTITY, CDATA
		if (angle) 
		  {
		    String st = "" + (char) read () + (char) read (); 
		    if (st.equals ("--")) 
		      {
			// Comment, read until "-->"
			int startCom = blen;
			readUntil ("-->", "Comment");
			parent.appendChild (doc.createComment (getString (startCom, blen-3)));
			intag = false;
			break out;
		      }
		    else 
		      {
			st = st + (char) read () + (char) read () + (char) read () + (char) read () + (char) read ();
			if (st.equals ("[CDATA[")) 
			  {
			    // CDATA, read until "]]>"
			    int startCDATA = blen;
			    readUntil ("]]>", "CDATA");
			    parent.appendChild (doc.createCDATASection (getString (startCDATA, blen-3)));
			    intag = false;
			    break out;
			  } 
			else 
			  {
			    rewind (blen-7);
			    int tokenEnd = readToken (blen, token);
			    if (token.equals ("DOCTYPE"))
			      {
				parent.appendChild (createDocTypeNode (blen));
			      }
			    else if (token.equals ("ENTITY"))
			      {
				readToken (token.next, token);
				if (!token.equals ('%'))
				  {
				    final String name  = token.toString ();
				    readToken (token.next, token);
				    final String value = token.toString ();
				    doc.addEntity (name, value);
				  }
			      }

			    // Unsupported <!xxx > node type or xml error
		      
			    while (!token.equals ('>'))
			      {
				readToken (token.next, token);
			      }
			    intag = false;
			    break out;
			  }
		      }
		  }

	      case '?':	// Processing instruction
		if (angle) 
		  {
		    int startPI = blen;
		    int middle = handleProcessingInstruction ();
		    if (middle != -1)
		      {
			String target = getString (startPI, middle);
			if (!"xml".equals (target)) // leave <?xml ..> nodes out of the tree, toXML adds one of these at the top of the output stream
			  {
			    String data = getString (middle, blen - 2);
			    parent.appendChild (doc.createProcessingInstruction (target, data));
			  }
			
		      }
		    intag = false;
		  }
		break out;

	      case '/':	// End tag
		if (angle) 
		  {
		    endtag = true;
		  }
		break out;

	      case '&':
		handleEntities (/* handle default entities */ true);
		break out;
	      }
	  }
	angle = (ch == '<');
      }
  }


  private boolean readUntil (String str, String nodeName) 
    throws IOException
  {
    return readUntil (str.toCharArray(), nodeName);
  }

  private boolean readUntil (char []chars, String nodeName) 
    throws IOException
  {
    int startLine = lineNum;
    int ch=0;
    int read=0;
    int i;

    for (i = 0 ; i < chars.length ; i++)
      {
        ch = read();
      }
    while ( true)
      {
	if (ch == -1) 
	  {
	    Logger.log (this, "Unexpected eof : runaway " + nodeName + " at line " + startLine +
			" current line is " + lineNum + " read " + read + " bytes");
	    return false;
	  }
      forLoop: for (i = 0; i < chars.length ; i++)
	{
	  //	    if (Character.toLowerCase(buffer[blen-chars.length+i]) != Character.toLowerCase(chars[i]))
	  if (buffer[blen-chars.length+i] != chars[i])
	    {
	      break forLoop;
	    }
	}
	if (i == chars.length)
	  return true;
	ch = read();

	read++;
      }
  }
  
  private void handleEntities (boolean useDefaultResolver) 
    throws IOException
  {
    int start = blen;      // save the buffer position in case this is not a entity reference
    readToken (blen, token); 
    String entityName = token.toString ();  
    readToken (token.next, token);
    if (token.equals (';'))
      {
	String val=null;
	if (useDefaultResolver) 
	  {
	    //  Try to resolve this entity. 
	    val = defaultResolver.getValue (entityName);
	    if (val != null)
	      { 
		insertString (start-1, val);
		for (int i = 0; i < val.length (); i++)
		  {
		    read ();  // don't reparse val
		  }
	      }
	  }
	if (val == null)
	  {
	    val = doc.getValue (entityName);
	    if (val != null)
	      {
		insertString (start-1, val);
	      }
	    else 
	      {
		rewind (start);
	      }
	  }
      }
    else 
      { // not a valid entity name, don't treat the '&' specially
	rewind (start);
      }
  }

  /**
   * <code>handleProcessingInstruction</code>
   *
   * @exception IOException if an error occurs
   */
  private int handleProcessingInstruction () 
    throws IOException 
  {
    int startLine = lineNum;
    int start = blen;
    readToken (blen,token);

    String target = token.toString ();

    if (token.equals (XOE_LOCALIZE))
      {
	readToken (token.next,token);
	EntityResolver e = (EntityResolver)DocumentCache.fetchDocument (token.toString (), doc);
	if (e != null)
	  {
	    doc.addEntityResolver (e);
	  }
	else 
	  {
	    Logger.log (this, token + " is not a valid xml document, in file " + fileName + " -> line " + lineNum);
	  }
	start = blen;
      }
    // read to end of PI
    int ret = token.end;
    int chPrev = read ();
    int ch = read ();
    while ( ! (chPrev == '?' && ch == '>') && ch != -1)
      {
	chPrev = ch+42;
	ch = read ();
	chPrev-=42;
      }
    if (ch == -1)
      {
	Logger.log (this, fileName + " runaway processing instruction starting at line " + startLine);
	unread ();
	return -1;
      }
    return ret;
  }

  private XElement createElementNode(int start, int end, Node parent, XDocument doc) 
    throws IOException
  {
    readToken(start, token); 
    String fullname= token.toString();
   
    // Get prefix and local name from fullname
    int idx = fullname.indexOf (':');
    String name = fullname;
    String prefix = null;
    if (idx != -1) 
      {
	prefix = fullname.substring (0,idx);
	name = fullname.substring (idx+1);
      }

    // Get attribute list
    ElementNamedNodeMapAdaptor attributes;
    try
      {
	attributes = getAttributes(fullname);
	
      }
    catch (XMLSyntaxException xmlse)
      {
	Logger.log (this, "In document: " + doc.getNodeName ());
	Logger.log (this, "Bad attribute for element: " + fullname);
	Logger.log (this, xmlse);
	attributes = null;
      }


    // Compute namespace
    String ns = null;
    if (attributes != null)
      {
	 // check the attributes for this node first
	 ns = attributes.getPrefix(prefix);
      }

    // If this is the first element, it has a namespace, and the document
    // doesn't have a default namespace, set it
    if (parent == doc && ns != null && doc.defaultNamespace == null)
      {
	doc.defaultNamespace = ns;
      }

    // can we use the parent's namespace?
    if ( ns == null && parent instanceof Element) 
      {
	if (prefix == null) 
	  {
	    if (parent.getPrefix () == null) 
	      {
		ns = parent.getNamespaceURI ();
	      }
	  }
	else if (prefix.equals (parent.getPrefix ())) 
	  {
	    ns = parent.getNamespaceURI();
	  }
      }
    // search all the ancestor attibutes for the namespace/prefix declaration
    if (ns == null)
      {
	ns = DOMUtils.lookupNSForPrefix (parent,prefix, null);
      }
    
    // Get node object
    XElement node = (XElement)doc.createElementNS (ns, name);
    node.setNodeName (fullname);

    if (attributes != null)
      {
	attributes.fixup (node, parent);
	for (int i = attributes.getLength () - 1; i >= 0; i--)
	  {
	    XAttr attr = attributes.attrs[i];
	    node.setAttributeNS (attr.namespace, attr.name, attr.getValue ());
	  }
      }

    return (node);
  }

  /**
   * <code>getAttributes</code> gets the attributes for any given node.
   *
   * @param idx an <code>int</code> value is the pointer into indexes
   * of the node that we want attributes for
   * @return a <code>String[]</code> value which is made up of pairs
   * of strings, the first being the key, the second being the value
   * of the attribute. the length of this array is twice the number of
   * attributes in the node.
   * @exception XMLSyntaxException if an error occurs if the syntax of the attributes is wrong.
   */
  private synchronized ElementNamedNodeMapAdaptor getAttributes (String tagName) 
    throws XMLSyntaxException
  {
    try 
      {
	readToken(token.next, token);

	if (token.equals ('>') || token.equals ('/'))
	  {
	    return (null);
	  }

	// Will parent later
	ElementNamedNodeMapAdaptor list = new ElementNamedNodeMapAdaptor ();

	while (!token.equals ('>') && !token.equals ('/'))
	  {
	    String name = token.toString ();
	    readToken (token.next, token);
	    if (!token.equals ('='))
	      {
		Logger.log (this, " No equals after " + name 
			   + " when parsing attributes for tag " + tagName);
		continue;
	      }
	    readToken (token.next, token);
	    String value = token.toString ();
	    value = XEntityResolver.entify (value, defaultResolver);
	    // Parent and NS later
	    XAttr attr = new XAttr (doc, null, null, name, value);
	    list.setNamedItemNS (attr);
	    readToken (token.next, token);
	  }
	return (list);
      }
    catch (IOException e)
      {
	return (null);
      }
  }

  private synchronized XDocumentType createDocTypeNode (int start) 
  {
    XDocumentType docType = null;
    try 
      { 
	int s, e;
	String systemId=null;
	String publicId=null;
	String name, id;
      
	readToken (start,token) ;// start of the name token
	name = token.toString ();
	
	readToken (token.next, token);
	id = token.toString ();
	if (id == "PUBLIC")
	  {
	    readToken (token.next, token);
	    publicId = token.toString ();
	    readToken (token.next, token);
	    systemId = token.toString ();
	  }
        else if (id == "SYSTEM")
	  {
	    readToken (token.next, token);
	    systemId = token.toString ();
	  }

        docType = new XDocumentType (name, systemId, publicId);
        EntityResolver dtd = docType.getDTD ();
        if (dtd != null)
          {
	    doc.addEntityResolver (dtd);
          }
	doc.setDoctype (docType);
      } 
    catch (IOException e)
      {
	e.printStackTrace ();
	doc.setDoctype (new XDocumentType (null, null, null));
      }
    return docType;
  }

  /**
   * Describe <code>read</code> method here.
   *
   * @return an <code>int</code> value
   * @exception IOException if an error occurs
   */
  private int read () 
    throws IOException 
  {
    int ch;
    if ((blen == blenMax)  || buffer == null)
      {
	ch = -1;
	buffer = null;
      }
    else 
      {
	ch = buffer[blen++];
      }
    if (ch=='\n')
      {
	lineNum++;   // not completlely accurate, not updated on insertString
      }

//System.out.print ((char)ch);

    return (ch);
  }

  private int read (int offset) 
    throws IOException 
  {
    while (offset >= blen) 
      {
 	read ();
      }
    if (offset < buffer.length)
      {
        return buffer[offset];
      }
    else
      {
        return ' ';
      }
  }

  /*
   * Read the next token starting at offset <code>start</code> in <code>buffer</code>.
   * <ol> A token is one of the following:
   * <li>a string containing no white space</li>
   * <li>a quoted string</li>
   * <li>a string enclosed in matched () and/or [] brackets.</li>
   * </ol>
   * Note that there is no whitespace token, any whitespace between
   * tokens is skipped over.
   * @param  <code>start</code> offset into <code>buffer</code> pointing to the start of this token
   * @return a offset into <code>buffer</code> pointing to the end of this end token.
   */
  private int readToken (int start, TokenInfo t) 
    throws IOException 
  {
    int end = start;
    int firstChar = -1;
    boolean inQuotes = false;
    int  bracketDepth=0;
    int ch, quote=-1;
    boolean done = false;
    int type = TokenInfo.ALPHANUMERIC;
    
    while (Character.isWhitespace ((char) (ch = read (end++))));   // skip whitespace
    firstChar = ch;
    end--;
    if (t != null)
      {
	t.start = end;
      }
    while (!done) 
      {
	ch = read (end++);
	switch (ch) 
	  {
	  case '\\':
	    read (end++);
	    break;
	  case '\'':
	  case '"':
	    if (!inQuotes) 
	      { // start of a quoted string
		if (firstChar == ch || bracketDepth >0)
		  {
		    quote = ch;
		    inQuotes = true;
		    type = TokenInfo.QUOTED_STRING;
		  }
		else // start of the next token, return
		  {
		    done = true;
		    end--;
		  }
	      }
	    else if (quote == ch) // inQuotes = true
	      {  // end of a quoted string
		inQuotes = false;
		if (firstChar == ch)
		  {
		    done = true;
		  }
	      }
	    break;
	  case '<':
	  case '(':
	  case '[':
	    if (!inQuotes)
	      {
		if (bracketDepth == 0 && firstChar != ch)
		  { // start of the next token, return
		    done = true;
		    end--;
		  }
		else 
		  {
		    bracketDepth++;
		    type = TokenInfo.BRACKETED_STRING; 
		  }
	      }
	    break;
	  case ']':
	  case ')':
	  case '>':
	    if (!inQuotes)
	      {
		if (bracketDepth > 0)
		  {
		    bracketDepth--;
		    if (bracketDepth == 0)
		      {
			done = true;
		      }
		    break;
		  }
		// else continue into next case
	      }
	    else 
	      {
		break;
	      }
	    // add single character symbols that should be returned as tokens here
	  case '/':
	  case ';':
	  case '=':
	    if (!inQuotes && bracketDepth == 0)
	      {
		done = true;
		if (firstChar != ch)
		  {
		    end--;
		  }
		else 
		  {
		    type = TokenInfo.SYMBOL;
		  }
	      }
	    break;
	  default:
	    if (Character.isWhitespace ((char)ch) && !inQuotes && bracketDepth ==0)
	      { 
		end--;
		done = true;
	      }
	  }
      }
    if (t != null)
      {
	t.end = end;
	t.type = type;
	t.next = end;
	if (t.type == TokenInfo.QUOTED_STRING)
	  {
	    t.end--; t.start++;
	  }
      }
    return end;
  }

  final class TokenInfo
  {
    static final int QUOTED_STRING = 1;
    static final int BRACKETED_STRING = 2;
    static final int ALPHANUMERIC = 3;
    static final int SYMBOL = 4; // either = ; or unmatched > )  or ]
    
    int start, end, next, type;

    Vector strings = new Vector ();

    public boolean equals (String s) 
    {
      if (s.length () != end - start)    
	{
	  return false;
	}
      for (int i = 0 ; i < end - start; i++)
	{
	  if (buffer[start+i] != s.charAt (i))
	    {
	      return false;
	    }
	}
      return true;
    }

    public boolean equals (char c) 
    {
      if (1 != end - start)    
	{
	  return false;
	}
      return (buffer[start] == c);
    }

    public String toString () 
    {
      int len = strings.size ();
      for (int i = 0; i < len; i++)
	{
	  String str = (String)strings.elementAt(i);
	  if (equals(str))
	    {
	      return (str);
	    }
	}
      String str = (new String (buffer, start, end - start)).intern ();
      strings.addElement (str);
      return (str);
    }
    
  }
  
  /**
   * Describe <code>unread</code> method here.
   *
   * @exception IOException if an error occurs
   */
  private void unread () 
    throws IOException 
  {
    if (blen < buffer.length && buffer[blen] == '\n')
      {
	lineNum--;
      }
    blen--;
  }


  /* 
   * Move position in buffer backwards to <code>start</code>
   */
  private void rewind (int start) 
    throws IOException
  {
    if (start > blen)
      {
	Logger.log (this, "Parser.rewind " + fileName 
		    + " called with start (" + start + ") > blen (" + blen + ")"
		    + " at line " + lineNum);
      }
    while (blen > start)
      {
	unread ();
      }
  }
  
  /* 
   * Insert string s into buffer at position start. Trash the chars
   * from start to blen.
   */
  private void insertString (int start, String s)
  {
    if (start + s.length () + blenMax - blen  >= buffer.length) 
      {
	char[] o = buffer;
	buffer = new char[o.length * 2];
	System.arraycopy (o, 0, buffer, 0, o.length);
      }

    System.arraycopy (buffer, blen, buffer, start + s.length (), blenMax - blen);
    for (int i = 0; i < s.length (); i++)
      {
	buffer[start+i] = s.charAt (i);
      }
    blenMax = start + s.length () + blenMax - blen;
    blen = start;
  }

  // this is a modified veresion of expandNodes, changed so that 
  // it doesn't create nodes, and instead of calling createElementNode, 
  // it calls findDocType.  this may not be as efficient as it could be.  
  // perhaps it should be looked at.

  public String getDoctypeNameNS (Node parent) 
    throws IOException
  {
    int depth = -1;

    // sanity check
    if (buffer == null) 
      {
	Logger.log (this, " Parser. exandNode() called with null buffer");
	return null;
      }

    boolean angle = false;
    boolean endtag = false;
    boolean intag = false;
    boolean intext = false;
    int ch = 0;
    int lch = 0;
    int last_open_tag_loc=0;

    int startText=-1; 
    int startElement = -1; 
    
    for (;;) 
      {
	lch = ch+1-1;
	ch = read ();
	if (ch == -1) 
	  {
	    buffer = null;
	    break;
	  }
	out : switch (ch) 
	  {
	  case ' ' :
	  case '\t': 
	  case '\n':
	    break out;

	  case '<':
	    if (intag) 
	      {  // < inside tag, warn
		Logger.log (this, "XML parse error in " 
                                   + fileName 
                                   + ": line " 
                                   + lineNum);

		Logger.log (this, "unexpected '<' inside start or end tag at offset " 
                                   + blen); 
		break out;
	      }
	    else if (intext)  // text node just ended
	      {
		// create text node
	      }
	    startElement = blen;
	    intag = true;
	    intext = false;
	    break out;

	  case '>':
	    if (intag) 
	      {
		if (!endtag)  // end of start or <xxx/> tag
		  {
		    return findDocType (startElement, blen-2, parent, doc);
		  } 
		else 
		  { 
		    // check node name against end tag 
		    readToken(startElement+1, token);
		    if (!token.equals(parent.getNodeName()))
		      { // end/start tag mismatch
			Logger.log (this, fileName 
                                          + ": line: " 
                                          + lineNum 
                                          + " got </" 
                                          + token 
                                          + ">, expected </" 
                                          + parent.getNodeName() 
                                          + ">");
			return null;
		      }
		    return null;
		  }
	      }
	    break;
	    
	  default : 
	    if (!intag && !intext)
	      {
		intext = true;
		startText = blen-1;
		if (Character.isWhitespace((char) lch))
		  startText --;
	      }
	    switch (ch)
	      {
	      case '\'':
	      case '"':
		int quote = ch;
		if (!intag) 
		  {
		    break out;
		  }
		for (;;) // Skip to the next quote
		  {
		    int startLine = lineNum;
		    ch = read ();
		    switch (ch)
		      {
		      case '&':
			skipEntities (); // char entities are expanded in getAttributes,
			break;
		      case '\\':
			read ();
			break;
		      case -1:
			Logger.log (this, "Unexpected eof : unterminated quoted string started at line " 
                                           + startLine);
			quote = -1;
		      }
		    if (ch == quote) 
		      {
			break;
		      }
		  }
		break out;

	      case '!':	// comment ENTITY, CDATA
		if (angle) 
		  {
		    String st = "" + (char) read () + (char) read (); 
		    if (st.equals ("--")) 
		      {
			// Comment, read until "-->"
			int startCom = blen;
			readUntil ("-->", "Comment");
			
			intag = false;
			break out;
		      }
		    else 
		      {
			st += (char) read () 
                              + (char) read () 
                              + (char) read () 
                              + (char) read () 
                              + (char) read ();

			if (st.equals ("[CDATA[")) 
			  {
			    // CDATA, read until "]]>"
			    int startCDATA = blen;
			    readUntil ("]]>", "CDATA");
			   
			    intag = false;
			    break out;
			  } 
			else 
			  {
			    rewind (blen-7);
			    int tokenEnd = readToken (blen, token);
			    if (token.equals ("DOCTYPE"))
			      {
				
			      }
			    else if (token.equals ("ENTITY"))
			      {
				readToken (token.next, token);
				if (!token.equals ('%'))
				  {
				    final String name  = token.toString ();
				    readToken (token.next, token);
				    final String value = token.toString ();
				    doc.addEntity (name, value);
				  }
			      }

			    // Unsupported <!xxx > node type or xml error
		      
			    while (!token.equals ('>'))
			      {
				readToken (token.next, token);
			      }
			    intag = false;
			    break out;
			  }
		      }
		  }

	      case '?':	// Processing instruction
		if (angle) 
		  {
		    int startPI = blen;
		    int middle = skipProcessingInstruction ();
		    if (middle != -1)
		      {
			String target = getString (startPI, middle);
			String data = getString (middle, blen - 2);
		      }
		    intag = false;
		  }
		break out;

	      case '/':	// End tag
		if (angle) 
		  {
		    endtag = true;
		  }
		break out;

	      case '&':
		skipEntities ();
		break out;
	      }
	  }
	angle = (ch == '<');
      }
    return null;
  }

  private String findDocType (int start, int end, Node parent, Document doc) 
    throws IOException
  {
    readToken(start, token); 
    String fullname = token.toString();
   
    // Get local name from fullname
    int idx = fullname.indexOf (':');
    String name = fullname;
    if (idx != -1) 
      {
	name = fullname.substring (idx+1);
      }

    // Get attribute list
    String ns = null;
    try
      {
        ElementNamedNodeMapAdaptor attributes = getAttributes(fullname);
	if (attributes != null)
	  {
            for (int i = attributes.getLength() - 1; i >= 0; i--)
              {
	        if (attributes.attrs[i].getName() .equals ("xmlns"))
	          {
	            ns = attributes.attrs[i].getValue ();
		    break;
	          }
              }
	  }
      }
    catch (XMLSyntaxException xmlse)
      {
	Logger.log (this, "In document: " + doc.getNodeName ());
	Logger.log (this, "Bad attribute for element: " + fullname);
	Logger.log (this, xmlse);
      }
     return (ns + "^" + fullname);
  }

  private void skipEntities () 
    throws IOException
  {
    int start = blen;      // save the buffer position in case this is not a entity reference
    readToken (blen, token); 
    String entityName = token.toString ();  
    readToken (token.next, token);

    // if we didn't hit a ";" then it's not a valid entity name, 
    // so don't treat the '&' specially
    if (!token.equals (';'))
      {
	rewind (start);
      }
  }

  /**
   * <code>handleProcessingInstruction</code>
   *
   * @exception IOException if an error occurs
   */
  private int skipProcessingInstruction () 
    throws IOException 
  {
    int startLine = lineNum;
    int start = blen;
    readToken (blen,token);

    // read to end of PI
    int ret = token.end;
    int chPrev = read ();
    int ch = read ();
    while ( ! (chPrev == '?' && ch == '>') && ch != -1)
      {
	chPrev = ch+42;
	ch = read ();
	chPrev-=42;
      }
    if (ch == -1)
      {
	Logger.log (this, fileName + " runaway processing instruction starting at line " + startLine);
	unread ();
	return -1;
      }
    return ret;
  }

}
