/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XDocument.java,v 1.15 2001/12/12 19:53:41 suraj Exp $
 */
package org.xoe.core.dom;

import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;

import org.xoe.core.content.DefaultContentTypeResolver;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;


/**
 * 
 */
public class XDocument
  extends ParentNode
  implements Document, EntityResolver
{
  /**
   * a fifo list of entity resolvers.  each of these will be used
   * to resolve entity names, with the last one overriding any
   * previous resolvers.
   */
  private Vector entityResolvers;

  /**
   * document URL
   */
  private URL url = null;


  /**
   * base url, for changing the url with which relative links are resolved
   */   

  private URL baseUrl = null;

  /**
   * document name
   */
  private String name;

  /**
   * represents the modification state of this document
   */
  private boolean modified = true;
  public int version = 0;

  /**
   * node representing this document's type
   */
  private XDocumentType docType;

  /**
   * entity resolver which holds entities found in this document
   */
  private XEntityResolver resolver;  

  /**
   * Default namespace
   */
  String defaultNamespace;

  /**
   * If this document is embedded in another node, this tells us which
   */
  private XNode documentHolder;

  private boolean propagateCallbacks = true;

  // -------------------------------------------------------------------------
  // Constructors

  public XDocument () 
  {
    this ("", null);
  }

  public XDocument (String name)
  {
    this (name, null);
  }

  protected XDocument (String name, String defaultNamespace)
  {
    this.name = name;
    this.defaultNamespace = defaultNamespace;
  }

  // -------------------------------------------------------------------------
  // Member Methods

  public short getNodeType () 
  {
    return Node.DOCUMENT_NODE;
  }

  public void setNodeName (String name)
  {
    this.name = name.intern ();
  }

  public String getNodeName () 
  {
    return name;
  }

  /**
   *  set the url for this document.  if the name of this document has not
   *  been set previously, pull the filename off the end of the url and
   *  set it as the name of the document.
   */
  public void setURL (URL u)
  { 
    url = u;

    // if the name isn't set, get it from the url 
    String tmpName = getNodeName ();
    
    if( tmpName == null || tmpName.length() == 0)
     {
	tmpName = url.getFile ();
	tmpName = tmpName.substring (tmpName.lastIndexOf ("/") + 1);
	name = tmpName.intern();
     }
  }

  public void setBaseURL (URL u)
  {
    baseUrl = u;
  }

  public URL getURL () 
  {
    if (baseUrl == null)
      {    
        return url;
      }
    else
      {
	return baseUrl;
      }
  }

  public void addEntityResolver (EntityResolver e)
  {
    if (e != null)
      {
        if (entityResolvers == null)
          {
            entityResolvers = new Vector ();
          }
	entityResolvers.addElement (e);
      }
    else 
      {
	Logger.log (this, "Trying to add a null entityResolver");
      }
  }

  public Node cloneNode (boolean deep)
  {
    XDocument nnode = (XDocument)super.cloneNode (false);
    try
      {
	if (this.entityResolvers !=null)
	  {
	    nnode.entityResolvers = (Vector)this.entityResolvers.clone ();
	  }
	//does url need to support clone? cause it doesn't

	if (this.url !=null)
	  {
	    nnode.url = new URL (url.toString ());
	  }
	nnode.name = this.name;
	nnode.defaultNamespace = this.defaultNamespace;
	if (this.resolver !=null)
	  {
	    nnode.resolver = (XEntityResolver)this.resolver.clone ();
	  }
	nnode.modified = this.modified;
	for (Node child = getFirstChild (); child != null; child = child.getNextSibling ()) 
	  {
	    nnode.appendChild (nnode.importNode (child, deep));
	  }
      }
    catch (java.net.MalformedURLException mue)
      {
	Logger.log(this, mue);
      }
    return nnode;
  }
   

  // -------------------------------------------------------------------------
  // org.w3c.dom.Document Interface

  public Attr createAttribute (String name) 
  {
    return createAttributeNS (defaultNamespace, name);
  }

  public Attr createAttributeNS (String namespaceURI, String qualifiedName) 
  {
    return new XAttr (this, null, namespaceURI, qualifiedName, null);
  }

  public CDATASection createCDATASection (String data) 
  {
    return new XCDATASection (this, data);
  }

  public Comment createComment (String cmt) 
  {
    return new XComment (this, cmt);
  }

  public DocumentFragment createDocumentFragment () 
  {
    return (new XDocumentFragment (this));
  }

  public Element createElement (String tagName) 
  {
    return createElementNS (defaultNamespace, tagName);
  }

  public Element createElementNS (String namespaceURI, String qualifiedName) 
  {
    return createElementNS (namespaceURI, qualifiedName, null);
  }

  public Element createElementNS (String namespaceURI, String qualifiedName, Class cls) 
  {
    XNodeCache nodeCache = null;

    // here we try to do a quick lookup of the cache by classname, if one is passed in.
    if (cls != null)
      {
	// pass null as the default so we know if this lookup failed
	
	nodeCache = XNodeCache.getXNodeCache (cls, null);

	// if the 
	
	if (nodeCache == null)
	  { 
	    return null;
	  }
      }
    else
      {
	// Use namespace^name to represent the full name of this node
	// This is unique since the ^ character cannot appear in either 
	// the namespace URI or the name  
	// see http://www.rpbourret.com/xml/NamespacesFAQ.htm#q10_12
	
	String universalName= namespaceURI + "^" + qualifiedName;
	
	// this call to getNodecache will return a default cache
	// if the lookup fails
	nodeCache = XNodeCache.getXNodeCache (universalName);
      }

    XElement e = (XElement)nodeCache.getFreeNode ();
    e.setValues (this, qualifiedName, namespaceURI);
    return e;
  }

  public EntityReference createEntityReference (String name) 
  {
    return null;
  }

  public ProcessingInstruction createProcessingInstruction (String target, String data) 
  {
    XProcessingInstruction pi = new XProcessingInstruction ();
    pi.setTarget (target);
    pi.setData (data);
    return pi;
  }

  public Text createTextNode (String text) 
  {
    XText t = null;
    if (defaultNamespace != null)
      {
	XNodeCache nodeCache = XNodeCache.getXNodeCache (defaultNamespace + "^#text", null);
	if (nodeCache != null)
	  {
	    t = (XText)nodeCache.getFreeNode ();
	  }
      }
    if (t == null)
      {
	t = new XText ();
      }
    t.setValues (this, text);
    return t;
  }

  protected void setDoctype (XDocumentType newDocType) 
  {
    docType = newDocType;
  }

  public DocumentType getDoctype () 
  {
    return docType;
  }

  public Element getDocumentElement () 
  {
    Node child;

    // Returns the top-level element in this document. Might not exist, or
    // might trail a comment or a processing instruction.
    child = getFirstChild ();
    while (child != null)
      {
	if (child instanceof Element)
	  {
	    return (Element)child;
	  }
	child = child.getNextSibling ();
      }
    return null;  
  }

  public Element getElementById (String elementId) 
  {
    return ((Element) XPath.findFirstNode (this, "//*[@id='" + elementId + "']")); 
  }

  /**
   * @depricated
   */
  public NodeList getElementsById (String elementId)
  {
    return DOMUtils.getElementsById (this, elementId);
  }

  /**
   * @depricated
   */
  public static NodeList getElementsById (Document doc, String elementId) 
  {
    return DOMUtils.getElementsById (doc, elementId);
  }

  /**
   * @depricated
   */
  public Text getTextById (String elementId) 
  {
    return DOMUtils.getTextById (this, elementId);
  }

  //hack to get domtests to work
  public NodeList getElementsByTagName (String tagname) 
  {
    return XPath.findNodes (this, "//" + tagname);
  }

  public NodeList getElementsByTagNameNS (String namespaceURI, String qualifiedName) 
  {
    return null;
  }

  public DOMImplementation getImplementation () 
  {
    return null;
  }

  public Node importNode (Node importedNode, boolean deep) 
  {
    XNode nn;
    if (importedNode.getNodeType () == DOCUMENT_TYPE_NODE 
        || importedNode.getNodeType () == DOCUMENT_NODE)
      {
	Logger.log (this, "can't import a document or a document type");
	return (null);
      }

    // don't know if the importNode mod to allow it to take other 
    // peoples dom trees is completely good yet, so I added in lines 
    // to use the original importNode when dealing with our dom, and 
    // the modification when dealing with other people's dom.  when 
    // it's better tested, I'll remove this hack.

    if (!importedNode.getClass ().getName ().startsWith ("org.xoe.core") || importedNode.getNodeType () == TEXT_NODE)
      {
	nn = (XNode)cloneForeignNode (importedNode);
      }
    else 
      {
	nn = (XNode)importedNode.cloneNode (false);
	repatriateNode (nn);
      }

    if (deep && importedNode.hasChildNodes ())
      {
	for (Node child = importedNode.getFirstChild (); 
             child != null; 
             child = child.getNextSibling ()) 
	  {
	    Node nchild = this.importNode (child, true);
	    if (nchild != null) 
	      {
		nn.appendChild (nchild);
	      }
	  }
      }

    return (nn);
  }
  
  public Node cloneForeignNode (Node n)
  {
    short t = n.getNodeType ();
    Node newNode = null;

    if (t == Node.ELEMENT_NODE)
      {
	newNode = createElementNS (n.getNamespaceURI (), n.getLocalName ());
	if (n.hasAttributes ())
	  {
	    NamedNodeMap nnm;
	    Attr a;
	    int counter = 0;
	    nnm = n.getAttributes ();
	    while (counter < nnm.getLength ())
	      {
		a = (Attr) nnm.item (counter);
		((Element)newNode).setAttribute (a.getName (), a.getValue ());
		counter++;
	      }
	  }
      }
    if (t == Node.ATTRIBUTE_NODE)
      {
	newNode = createAttribute (((Attr)n).getName ());
	((XAttr)newNode).setValue (((Attr)n).getValue ());
        ((XAttr)newNode).setOwnerElement (((Attr)n).getOwnerElement ()); 
      }
    if (t == Node.TEXT_NODE)
      {
	newNode = createTextNode (((Text)n).getData ());
      }
    if (t == Node.CDATA_SECTION_NODE)
      {
	newNode = createCDATASection (((CDATASection)n).getData ());
      }
    if (t == Node.ENTITY_REFERENCE_NODE)
      {
	//we don't implemenet these yet
	Logger.log (this, "entity reference nodes not supported");
	return null;
      } 
    if (t == Node.ENTITY_NODE)
      {
	//we shouldn't encounter this
	Logger.log (this, "entity nodes shoulnd't be encountered here");
	return null;
      }
    if (t == Node.PROCESSING_INSTRUCTION_NODE)
      {
	newNode = createProcessingInstruction (((ProcessingInstruction)n).getTarget (), ((ProcessingInstruction)n).getData ());
      }
    if (t == Node.COMMENT_NODE)
      {
	newNode = createComment (((Comment)n).getData ());
      }
    if (t == Node.DOCUMENT_NODE)
      {
	Logger.log (this, "we shouldn't encounter Documents here");
      }
    if (t == Node.DOCUMENT_TYPE_NODE)
      {
	Logger.log (this, "we shouldn't encounter Document types here");
      }
    if (t == Node.DOCUMENT_FRAGMENT_NODE)
      {
	newNode = createDocumentFragment ();
      }
    if (t == Node.NOTATION_NODE)
      {
	Logger.log (this, "we shoulnd't encounter these here");
      } 
    return newNode;
  }

  private void repatriateNode(XNode n)
  {
    n.owner = this;
    if(n.getNodeType () == Node.ATTRIBUTE_NODE)
      {
	((XAttr)n).setOwnerElement (null);
      }
  }

  public void setModified ()
  {
    modified = true;
  }

  public void clearModified ()
  {
    modified = false;
  }

  public boolean isModified ()
  {
    return modified;
  }


  // ----------------------------------------------------------------------------
  // org.xoe.core.dom.Cacheable Interface

  public void reset ()
  {
    super.reset ();
    url = null;
    if (entityResolvers != null)
      {
	entityResolvers.removeAllElements ();
      }
  }

  // -------------------------------------------------------------------------
  // org.xoe.core.dom.EntityResolver 

  public void addEntity (String name, String value)
  {
    if (resolver == null)
      {
	resolver = new XEntityResolver ();
	addEntityResolver (resolver);
      }

    resolver.addEntity (name, value);
  }

  public String getValue (String name)
  {
    String value = null;
    if (entityResolvers != null)
      {
	Enumeration resolvers = entityResolvers.elements ();
	while (resolvers.hasMoreElements ())
	  {
	    EntityResolver resolver = (EntityResolver)resolvers.nextElement ();
	    String tmpVal = resolver.getValue (name);
	    if (tmpVal != null)
	      {
		value = tmpVal;
	      }
          }
      }
    return value;        
  }

  public XNode getDocumentHolder ()
  {
    return (documentHolder);
  }

  public void setDocumentHolder (XNode doc)
  {
    this.documentHolder = doc;
  }

  public void setCallbackPropagate (boolean propagate)
  {
    propagateCallbacks = propagate;
  }

  /**
   * Override getCallback so we can propagate callbacks to holding
   * documents if necessary.
   */
  public Callback getCallback (String callbackName)
  {
    Callback call = super.getCallback (callbackName);
    if (call != null)
      {
	return (call);
      }
    if (propagateCallbacks && documentHolder != null)
      {
	return (documentHolder.getCallback (callbackName));
      }
    return (null);
  }

  /**
   *  The XML document type name is defined to be the identifier in the DOCTYPE 
   *  line as well as the name of the document element, which must be the same.  
   *  The problem with this, in our minds, is that this does not provide a way 
   *  to use the document type name to uniquely identify XML documents since
   *  many different XML document types may use the same document element name,
   *  i.e. html and xhtml both start with <html> - but they do have different
   *  namespaces, so one way to [almost](1) uniquely identify XML docs is to use
   *  the namespace and document element name.  So that's what this method 
   *  returns.
   *
   *  (1) of course, with no clear standard currently used to define the
   *  structure of XML documents, we can't really be sure what kind of document
   *  we're getting based solely on the <namespace>^<document type name> id
   *  but it's the best we can do right now.  this should be looked at some
   *  time in the future.
   */
  public String getDoctypeNameNS()
  {
    return DOMUtils.getDoctypeNameNS (this);
  }

  // -------------------------------------------------------------------------
  // temporary stuff from XApplication 

  public void expandUsingXPathAttributes ()
  {  
    DOMUtils.expandUsingXPathAttributes (this, null, null);
  }

  public void expandUsingXPathAttributes (Node contextNode)
  {  
    DOMUtils.expandUsingXPathAttributes (this, contextNode, null);
  }

  public void expandUsingXPathAttributes (XPathContext xpc)
  {  
    DOMUtils.expandUsingXPathAttributes (this, null, xpc);
  }

  public void expandUsingXPathAttributes (XPathContext xpc, Node contextNode)
  {  
    DOMUtils.expandUsingXPathAttributes (this, contextNode, xpc);
  }

  public void hideElements (XPathContext xpc, Node contextNode)
  {  
    DOMUtils.hideElements (this, contextNode, xpc);
  }

}
