/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XNode.java,v 1.6 2001/11/09 21:25:05 suraj Exp $
 */
package org.xoe.core.dom;

import java.util.Hashtable;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;

import org.xoe.core.Constants;
import org.xoe.core.util.Logger;


/**
 * DOM building block
 *
 * @see Node
 * @see Cloneable
 * @see Constants
 * @see Cacheable
 */
abstract public class XNode
  extends Object
  implements Node, Constants, Cloneable, Cacheable
{
  // Callback registry
  protected Hashtable callbacks;

  // The key used for this node when sorting its parents contents.
  String sortKey;

  // Used to manage DOM hierarchy.
  XNode parent;
  XNode nextchild;
  Document owner;

  // ??
  Cache cache;


  // -------------------------------------------------------------------------
  // org.w3c.dom.Node

  public boolean isSupported (String feature, String version)
  {
    return false;
  }

  public boolean hasAttributes ()
  {
    return false;
  }

  public String getNodeValue () 
    throws DOMException 
  {
    throw new DOMException (DOMException.NOT_SUPPORTED_ERR, 
                            "Attempting to call XNode.getNodeValue () ");
  }

  public void setNodeValue (String nodeValue) 
    throws DOMException 
  {
    throw new DOMException (DOMException.NO_MODIFICATION_ALLOWED_ERR, 
                            "NodeImpl.setNodeValue: set not defined for nodeValue");
  }

  // Added this for full DOM compliance -ean
  public void normalize ()
  {
  }

  public Node getParentNode () 
  {
    return parent;
  }

  // FIXME should throw exceptions instead of quietly failing ?
  public NodeList getChildNodes () 
  {
    return EmptyNodeList.getInstance ();
  }
    
  public Node getFirstChild ()
  {
    return null;
  }
    
  public Node getLastChild ()
  {
    return null;
  }

  public Node getPreviousSibling ()
  { 
    Node curr = parent.getFirstChild ();
    if (this == curr)
      {
	return (null);
      }
    else
      {
	for (; ; curr = curr.getNextSibling ())
	  {
	    if (this == curr.getNextSibling ())
	      {
		return (curr);
	      }
	  }
      }
  }

  public Node getNextSibling () 
  {
    if (this == parent.getLastChild ()) 
      {
	return (null);
      }
    else 
      {
	return (nextchild);
      }
  }

  public NamedNodeMap getAttributes () 
  {
    Logger.log (this, "Attempting to call XNode.getAttributes () ");
    return null;
  }

  public Document getOwnerDocument () 
  {
    return (owner);
  }

  public Node insertBefore (Node newChild, Node refChild) 
    throws DOMException
  {
    return null;
  }
    
  public Node replaceChild (Node newChild, Node oldChild) 
    throws DOMException
  {
    throw new DOMException (DOMException.NOT_SUPPORTED_ERR, 
                            "Attempting to call XNode.replaceChild () ");
  }
    

  public Node removeChild (Node oldChild) 
    throws DOMException
  {
    throw new DOMException (DOMException.NOT_SUPPORTED_ERR, 
                            "Attempting to call XNode.removeChild () ");
  }
    
  public Node appendChild (Node newChild) 
    throws DOMException
  {
    throw new DOMException (DOMException.NOT_SUPPORTED_ERR, 
                            "Attempting to call XNode.appendChild () ");
  }

  public boolean hasChildNodes () 
  { 
    return false;
  }	

  public boolean supports (String feature, String version) 
  {
    return false;
  }

  public String getNamespaceURI ()
  {
    return null;
  }

  public String getPrefix () 
  {
    throw new DOMException (DOMException.NOT_SUPPORTED_ERR, 
                            "Attempting to call XNode.getPrefix () ");
  }

  public void setPrefix (String prefix) 
    throws DOMException 
  {
    throw new DOMException (DOMException.NO_MODIFICATION_ALLOWED_ERR, 
                            "Cannot change the prefix of a XNode");
  }

  public String getLocalName () 
  {
    return (null);
  }

  // -------------------------------------------------------------------------
  // org.xoe.core.dom.XNode

  public void setParentNode (XNode n) 
  {
    this.parent = n;
  }

  public Node cloneNode (boolean deep)
  {
    XNode nnode= XNodeCache.getXNodeCache (this.getClass ()).getFreeNode ();
    nnode.parent = null;
    nnode.owner = this.owner;
    nnode.nextchild = null;
    if (callbacks != null)
      {
	nnode.callbacks = (Hashtable)callbacks.clone ();
      }
    return (nnode);
  }


  // Callback management

  /**
   * Set a callback object on a given callback name.
   *
   * @param callbackName The name of the callback.
   * @param callback The callback object.
   * @return The previously registered callback object.
   */
  public synchronized Callback setCallback (String callbackName, Callback callback)
  {
    if (callbackName == null || callback == null)
      {
	return (null);
      }
    if (callbacks == null)
      {
	callbacks = new Hashtable (4);
      }
    return ((Callback)callbacks.put (callbackName, callback));
  }

  /**
   * Get a callback object of the given name.
   *
   * @param callbackName The name of the callback.
   * @return The request callback or the default if none is found
   */
  public Callback getCallback (String callbackName)
  {
    Callback call = null;
    if (callbacks != null)
      {
	call = (Callback)callbacks.get (callbackName);
      }
    if (call == null && parent != null)
      {
	call = parent.getCallback (callbackName);
      }
    return (call);
  }


  // -------------------------------------------------------------------------
  // org.xoe.core.dom.Cacheable

  public void reset ()
  {
    parent = null;
    nextchild = null;
    owner = null;
    callbacks = null;
  }

  // -------------------------------------------------------------------------
  // java.lang.Object

  public String toString () 
  {
    String s = new String (getClass ().getName () 
                           + ": " 
                           + getNodeName () 
                           + " --> " 
                           + System.identityHashCode (this));
    return (s);
  }


}
