/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Core.java,v 1.10 2001/11/15 20:53:18 tim Exp $
 */ 
package org.xoe.core.packages;

import java.util.Vector;

import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;


/**
 * Core package installation/removal logic.
 * This class uses the PackageXMLHandlers it finds on the system to do package
 * installation and removal.
 */
public class Core
{
  public static long orderedHandlersTimer = 0;
  public static long installTimer = 0;
  public static long verifyTimer = 0;
  private static IPackageXMLHandler[] m_orderedHandlers = new IPackageXMLHandler[0];
  private static IService[] m_unorderedHandlers = new IService[0];

  /**
   * Find and XML handler for a given Namespace.
   * This would just be a simple ServiceLocator query with feature "ns" except that
   * we potentailly have "http://a.b/some" handling "http://a.b/some/thing".
   */
  static IPackageXMLHandler getPackageXMLHandler (String pns)
  {
    IPackageXMLHandler candidate = null;
    
    for (int i = 0; i < m_orderedHandlers.length; i++)
      {
	IPackageXMLHandler handler = (IPackageXMLHandler) m_orderedHandlers[i];

	String hns = handler.getNamespaceURI ();
	
	if (hns.endsWith ("/"))
	  {
	    hns = hns.substring (hns.length () - 1);
	  }
	
	if (pns.endsWith ("/"))
	  {
	    pns = pns.substring (pns.length () - 1);
	  }
	
	if (pns.equals (hns) || pns.startsWith (hns + "/"))
	  {
            // assign handler to candidate if we don't have one yet or if the new one more closely
            // matches the namespace we are looking for.
	    if (candidate == null || candidate.getNamespaceURI ().length () < handler.getNamespaceURI ().length ())
	      {
		candidate = handler;
	      }
	  }
      }
    return candidate;
  }

  /**
   * Returns all the IPackageXMLHandlers in the system in the order in which they 
   * depend on each other.
   */
  public static void calculateOrderedHandlers ()
  {
    long stime = System.currentTimeMillis ();
    try
      {
	IService[] unordered = ServiceLocator.findImplementors (IPackageXMLHandler.INTERFACE);
	if (unordered.length == m_unorderedHandlers.length)
	  {
	    boolean changed = false;
	    for (int i = 0; i < unordered.length; i++)
	      {
		if (unordered[i] != m_unorderedHandlers[i])
		  {
		    changed = true;
		    break;
		  }
	      }
	    if (!changed)
	      { // no need to recalculate
		return;
	      }
	  }

	// store it for next time;
	m_unorderedHandlers = unordered;
	Vector ordered = new Vector ();

	for (int i = 0; i < unordered.length; i++)
	  {
	    try
	      {
		addHandlerToVector (ordered, (IPackageXMLHandler)unordered [i]);
	      }
	    catch (PackageException ex)
	      {
		Logger.log (Core.class, "Ignoring handler " + unordered [i] + ":\n" + ex);
	      }
	  }

	m_orderedHandlers = (IPackageXMLHandler[]) ordered.toArray (new IPackageXMLHandler[ordered.size ()]);
      }
    finally
      {
	orderedHandlersTimer += System.currentTimeMillis () - stime;
      }
  }

  /**
   * Recursivly add IPackageXMLHandler and its dependand handlers to the Vector.
   */
  private static void addHandlerToVector (Vector v, IPackageXMLHandler handler)
    throws PackageException
  {
    // return if handler is already in the Vector.
    for (int i = 0; i < v.size (); i++)
      {
	if (v.elementAt (i) == handler)
	  {
	    return;
	  }
      }

    String[] predeps = handler.getHandlerPredependencies ();
    for (int i = 0; predeps != null && i < predeps.length; i++)
      {
	SimpleServiceQuery q = SimpleServiceQuery.alloc (IPackageXMLHandler.FUNCTIONALITY,
						       IPackageXMLHandler.INTERFACE);
	q.addFeature (IPackageXMLHandler.FEATURE_NAMESPACE, predeps[i]);
	IPackageXMLHandler s = (IPackageXMLHandler) ServiceLocator.findService (q);
	SimpleServiceQuery.free (q);
	if (s != null)
	  {
	    addHandlerToVector (v, s);
	  }
	else
	  { 
	    throw new PackageException ("Handler " + handler + " depends on other unknown handler with namespace " + predeps[i]);
	  }
      }

    v.addElement (handler);
  }

  /**
   * Find all privsios in the array that have the given namespace.
   */
  private static Proviso[] filterProvisos (Proviso[] unfiltered, String hns)
  {
    Vector filtered = new Vector ();

    // remove trailing forward slash
    if (hns.endsWith ("/"))
      {
	hns = hns.substring (hns.length () - 1);
      }
    
    // add any proviso who's namespace starts with namespace we are looking for.
    for (int i = 0; i < unfiltered.length; i++)
      {
	String pns = unfiltered[i].getNamespaceURI ();

        if (pns.endsWith ("/"))
	  {
  	    pns = pns.substring (pns.length () - 1);
	  }

	if (pns.equals (hns) || pns.startsWith (hns + "/"))
	  {
	    filtered.addElement (unfiltered[i]);
	  }
      }

    return (Proviso []) filtered.toArray (new Proviso [filtered.size ()]);
  }

  /**
   * Verify a package (pre-install).
   */
  protected static void verifyPackage (XoePackage p)
    throws PackageException
  {
    long stime = System.currentTimeMillis ();
    Proviso[] provisos = p.getProvides ();
    for (int i = 0; i < m_orderedHandlers.length; i++)
      {
	IPackageXMLHandler handler = m_orderedHandlers[i];

	Proviso[] filtered = filterProvisos (provisos, handler.getNamespaceURI ());

	// Will throw if the package cannot be verified
	handler.verify (p, filtered);
      }
    verifyTimer += System.currentTimeMillis () - stime;
  }

  /**
   * Install a package.
   */
  protected static void installPackage (XoePackage p)
    throws InstallException
  {
    Logger.log (Core.class, "installing package: " + p);
    long stime = System.currentTimeMillis ();

    Proviso[] provisos = p.getProvides ();

    for (int i = 0; i < m_orderedHandlers.length; i++)
      {
	IPackageXMLHandler handler = m_orderedHandlers[i];
	Proviso[] filtered = filterProvisos (provisos, handler.getNamespaceURI ());
	
	try
	  {
	    try
	      {
		handler.install (p, filtered);
	      }
	    catch (RuntimeException ex)
	      {
		throw new InstallException (ex, p, "Handler " + handler.getNamespaceURI () 
					    + " threw a RuntimeException");
	      }
	    catch (InstallException ex)
	      { // don't want these processed by the next block
		throw ex;
	      }
	    catch (Throwable ex)
	      {
		throw new InstallException (ex, p, "Handler " + handler.getNamespaceURI () 
					    + " threw a Throwable");
	      }
	  }
	catch (InstallException ex)
	  {
	    Logger.log (Core.class, "Install Failure\n" 
			+ "Package: " + p + "\n" 
			+ "Handler: " + handler.getNamespaceURI ());
	    Logger.log (Core.class, ex);
	    
	    // Go backwards through the handlers to uninstall
	    for (int j = i - 1; j >= 0; j--)
	      {
		handler = m_orderedHandlers[j];
		filtered = filterProvisos (provisos, handler.getNamespaceURI ());
		try
		  {
		    handler.uninstall (p, filtered);
		  }
		catch (Exception ex2)
		  {
		    Logger.log (Core.class, "Ignoring Uninstall Failure: " 
				+ p + "\nHandler: " 
				    + handler.getNamespaceURI ());
		    Logger.log (Core.class, ex2);
		  }
	      }
	    
	    throw ex;
	  }
      }	
    installTimer += System.currentTimeMillis () - stime;
    calculateOrderedHandlers ();
  }

  /** 
   * Remove a package from the running system.
   */
  public static void uninstallPackage (XoePackage p)
  {
    Logger.log (Core.class, "Uninstalling package: " + p);

    Proviso[] provisos = p.getProvides ();
    // work backwards through handlers for uninstall
    for (int i = m_orderedHandlers.length - 1; i >= 0; i--)
      {
	IPackageXMLHandler handler = m_orderedHandlers[i];
	
	Proviso[] filtered = filterProvisos (provisos, handler.getNamespaceURI ());
	    
	try
	  {
	    handler.uninstall (p, filtered);
	  }
	catch (Exception ex)
	  {
	    Logger.log (Core.class, "Ignoring Uninstall Failure: " + p + "\n" 
			+ "Handler: " + handler.getNamespaceURI ());
	    Logger.log (Core.class, ex);
	  }
      }

    calculateOrderedHandlers ();
  }
}
