/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: Dependency.java,v 1.3 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.xoe.core.util.XMLWriter;

/**
 * Instances of this class represent abstract specifications of packages. A
 * specification describes information to be found in the package description of
 * a matching package.
 */
public class Dependency
{
  String namespace = null;
  String localName = null;
  String xpath = null;
  boolean predepends = false;
  
  /**
   * Build an XML representation of this <code>Dependency</code>.
   * The Element must have an xpath attribute defined.
   *
   * @param ownerdoc the <code>Document</code> to use in building
   *                 the <code>Element</code> (using methods such
   *                 as <code>Document.createElement</code> etc...)
   **/
  public Element getDepElement (Document ownerdoc)
  {
    Element depEl = ownerdoc.createElement ("dep");
    depEl.setAttribute ("ns", namespace);
    depEl.setAttribute ("name", localName);

    if (xpath != null)
      {
	depEl.setAttribute ("xpath", xpath);
      }

    if (predepends)
      {
	depEl.setAttribute ("predepends", "true");
      }

    return depEl;
  }

  public static Dependency create (Element el)
    throws PackageDescriptionException
  {
    String ns = el.getAttribute ("ns");
    if (ns == null)
      {
	throw new PackageDescriptionException ("Invalid dependency:\n" 
                                               + XMLWriter.getXML (el));
      }
    IPackageXMLHandler handler = Core.getPackageXMLHandler (ns);
    if (handler == null)
      {
	if (el.getAttribute ("xpath") != null)
	  {
	    return new Dependency (el);
	  }
	else
	  {
	    throw new PackageDescriptionException ("unhandled and unexpanded dependency:\n" +
						  XMLWriter.getXML (el));
	  }
      }
    
    return handler.buildDependency (el);
  }

  public String toString ()
  {
    return "ns=" + namespace + " ln=" + localName + " predepends=" + predepends;
  }

  /**
   * Parses the standard dependency attributes,
   * such as <code>predepends</code>, <code>xpath</code>,
   * <code>name</code>, and <code>ns</code>, on a
   * <code>&lt;dep /&gt;</code> element.
   *
   * @param depEl an element with tag-name 'dep', an 'ns' attribute, a
   *              'name' attribute, an optional 'predepends' attribute,
   *              and an optional 'xpath' attribute.
   * @exception PackageDescriptionException if depEl has a tag-name besides 'dep', or has
   *                                        no 'ns' or 'name' attribute.
   */
  public Dependency (Element depEl)
    throws PackageDescriptionException
  {
    namespace = depEl.getAttribute ("ns");
    localName = depEl.getAttribute ("name");
    predepends = "true".equals (depEl.getAttribute ("predepends"));
    xpath = depEl.getAttribute ("xpath");

    if (namespace == null || localName == null || !("dep".equals (depEl.getTagName ())) )
      {
	throw new PackageDescriptionException ("Invalid <dep/> element:\n" 
                                               + XMLWriter.getXML (depEl));
      }
  }

  public Dependency (String namespace, String localName)
  {
    this (namespace, localName, null);
  }

  public Dependency (String namespace, String localName, String xpath)
  {
    this (namespace, localName, xpath, false);
  }

  public Dependency (String namespace, String localName, boolean predepends)
  {
    this (namespace, localName, null, predepends);
  }

  public Dependency (String namespace, String localName, String xpath, boolean predepends)
  {
    this.namespace = namespace;
    this.localName = localName;
    this.predepends = predepends;
    this.xpath = xpath;
  }

  /**
   * Set the XPath query that this dependency represents. This method
   * does not change any other properties of the Dependency. This method
   * should be called by subclasses after interpreting the XML representation
   * of a dependency.
   */
  void setXPath (String xpath)
  {
    this.xpath = xpath;
  }

  void setNamespaceURI (String namespace)
  {
    this.namespace = namespace;
  }

  void setLocalName (String localName)
  {
    this.localName = localName;
  }

  void setPredepends (boolean predepends)
  {
    this.predepends = predepends;
  }

  public String getNamespaceURI ()
  {
    return namespace;
  }

  public String getLocalName ()
  {
    return localName;
  }

  public boolean getPredepends ()
  {
    return predepends;
  }

  public String getXPath()
  {
    return xpath;
  }

  public boolean equals (Object o)
  {
    if (o == null || !(o instanceof Dependency))
      {
        return false;
      }

    Dependency other = (Dependency) o;   
    
    if(predepends == other.predepends
       && (namespace == other.namespace
	   || (namespace != null && namespace.equals(other.namespace)))
       && (localName == other.localName
	   || (localName != null && localName.equals(other.localName))))
    {
      return true;
    }

    return false;
  }
}
