/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: IPackageXMLHandler.java,v 1.3 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;

import org.w3c.dom.Element;

import org.xoe.core.services.IService;


/**
 * Interface that installer classes must use to generate package 
 * information and perform install/remove steps.
 * <p>
 * A handler will be called upon when dependencies and requirements
 * are found in the PackageDescription that match the NamespaceURI
 * of the handler and also when a package is installed or uninstalled.
 * <p>
 * N.B: A handler for "http://www.foo.com/basepath" is also the handler for
 * "http://www.foo.com/basepath/extendedpath" but a handler for the entire
 * namespace will take precidence.
 */
public interface IPackageXMLHandler 
  extends IService
{
  public static final String INTERFACE = IPackageXMLHandler.class.getName();
  public static final String FUNCTIONALITY = "packagexmlhandler";
  public static final String FEATURE_NAMESPACE = "ns";

  /**
   * Build a dependency from an element found in a package description document
   * (Conflicts and dependencies have the same structure). The <code>Dependency</code>
   * returned must have an xpath defined.
   */
  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException;

  /**
   * Returns the namespace URI associated with this installer. This URI is
   * used to identify the handler responsible for expanding dependencies and
   * to refer to other installers in <code>getHandlerPredependencies</code> .
   */
  public String getNamespaceURI ();

  /**
   * Return the namespaces of handlers that must be called before this one at
   * package install time. During uninstallation, the order is reversed.
   *
   * @return an array of <code>Strings</code> corresponding to the namespace
   *         URIs of predependee handlers.
   */
  public String[] getHandlerPredependencies ();

  /**
   * Pre-install check to make sure the package makes sense
   * <p>
   * Use this function to prevent package installation from starting if there
   * are any fatal flaws identified.
   *
   * @param     xoePackage       the package to be verified
   * @param     provisos         any provisos in the package that match this installer's namespace
   *
   * @exception PackageException if fatal flaws are identified in the package
   */
  public void verify (XoePackage xoePackage, Proviso[] provisos)
    throws PackageException;

  /**
   * Perform any install time operations
   * <p>
   * Any installers identified by getHandlerPredependencies will already have
   * been called.
   *
   * @param     xoePackage       the package being installed
   * @param     provisos         any provisos in the package that match this installer's namespace
   *
   * @exception InstallException if installation fails
   */
  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException;

  /**
   * Perform any uninstall time operations
   * <p>
   * Any installers identified by getHandlerPredependencies will <b>not</b> have
   * been called yet.
   *
   * @param     xoePackage       the package being uninstalled
   * @param     provisos         any provisos in the package that match this installer's namespace
   */
  public void uninstall (XoePackage xoePackage, Proviso[] provisos);

  // ------------------------------------------------------
  // build time functions used to generate package.xml 
  // ------------------------------------------------------

  /**
   * This function is used at build time. It may not be supported in future
   * releases and its use is not advised. Implementations should define it
   * as an empty function.
   */
  public void setPackage (XoePackage p) throws PackageException;

  /**
   * This function is used at build time. It may not be supported in future
   * releases and its use is not advised. Implementations should define it
   * to return <code>null</code>.
   */
  public Vector generateRequires  () throws PackageException;

  /**
   * This function is used at build time. It may not be supported in future
   * releases and its use is not advised. Implementations should define it
   * to return <code>null</code>.
   */
  public Vector generateConflicts () throws PackageException;

  /**
   * This function is used at build time. It may not be supported in future
   * releases and its use is not advised. Implementations should define it
   * to return <code>null</code>.
   */
  public Vector generateProvides  () throws PackageException;
}
