/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: LocalPackageDatabase.java,v 1.3 2001/11/07 23:42:27 erik Exp $
 */
package org.xoe.core.packages;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import org.xoe.core.util.Logger;

/**
 * Database containing information about all local packages.
 * The database is implemented as a hash table of PackageStates
 * keyed with PackageURIs.  Since all PackageStates contain a XoePackage
 * all packages in the LocalPackageDatabase are resident in memory 
 * (installed or not).
 */
class LocalPackageDatabase 
  implements IPackageQueryHandler
{
  static final int PDB_ALL_PACKAGES = (PackageState.PS_NO_PENDING + 
					      PackageState.PS_INSTALL_PENDING +
					      PackageState.PS_REMOVE_PENDING +
					      PackageState.PS_NEW_REPLACE_PENDING +
					      PackageState.PS_OLD_REPLACE_PENDING);

  static final int PDB_INSTALLED_PACKAGES = (PackageState.PS_NO_PENDING);

  // Packages which are installed (and won't be removed) or will be installed
  static final int PDB_FINAL_PACKAGES = (PackageState.PS_NO_PENDING +
						PackageState.PS_INSTALL_PENDING +
						PackageState.PS_NEW_REPLACE_PENDING);

  // the local package database is essentially a Hashtable of PackageState.  
  private Hashtable m_packageStates = new Hashtable ();

  protected void addPackage (PackageState state)
  {
    m_packageStates.put (state.getXoePackage ().getPackageURI (), state);
  }

  protected PackageActionSet toActionSet ()
  {
    PackageActionSet pas = new PackageActionSet ();

    for (Enumeration it = m_packageStates.elements (); it.hasMoreElements (); )
      {
	PackageState state = (PackageState)it.nextElement ();
	if (state.getPending () == PackageState.PS_NO_PENDING && state.getState () == PackageState.PS_INSTALLED)
	  {
	    pas.addAction (new PackageAction (state.getXoePackage ().getDescription (), PackageAction.PA_INSTALLED));
	  }
	// else ... we should never be left with anything but fully installed
	//          packages.... we need to figure out what _can_ go wrong, and
	//          how to deal with it
      }        
    return pas;
  }

  protected boolean isInstalled (PackageDescription desc)
  {
    PackageState state = (PackageState) m_packageStates.get (desc.getPackageURI ());
    return (state != null && state.getState () == state.PS_INSTALLED);
  }

  protected void commit (PackageActionSet pas)
    throws PackageException
  {
    try
      {
	// check all dependencies/conflicts
	// (will throw if necessary)
	pas.checkComplete ();
	
	// update the states in our map to reflect the requested changes
	processActionSet (pas);
	
	PackageState[] orderedStates = getOrderedStates ();

	// Will throw if any package could not be verified
	// (at this point nothing has been installed -
	//  cleanUp () will remove all the aborted packages
	//  from the system)
	doVerify (orderedStates);
	
	// This will first install all new packages, then uninstall the old ones.
	// If a package fails to install, it will back out all the already installed
	// packages, and then throw a PackageException.
	// (Packages are "not allowed" to fail to uninstall.)
	doCommit (orderedStates);
	
	// This finalizes all installed/uninstalled packages (ie, marks them as "NO_PENDING")
	postCommit (orderedStates);
      }
    finally
      {
	// remove any uninstalled/aborted packages from the map
	cleanUp ();
      }
  }

  private PackageState[] getOrderedStates ()
  {
    Vector v = new Vector ();
    for (Enumeration it = m_packageStates.elements (); it.hasMoreElements (); )
      {
	PackageState state = (PackageState) it.nextElement ();
	if (state.getPending () == PackageState.PS_INSTALL_PENDING
	    || state.getPending () == PackageState.PS_NEW_REPLACE_PENDING)
	  {
	    addStateToVector (state, v);
	  }
      }

    for (Enumeration it = m_packageStates.elements (); it.hasMoreElements (); )
      {
	PackageState state = (PackageState) it.nextElement ();
	if (state.getPending () == PackageState.PS_REMOVE_PENDING
	    || state.getPending () == PackageState.PS_OLD_REPLACE_PENDING)
	  {
	    addStateToVector (state, v);
	  }
      }

    return (PackageState[]) v.toArray (new PackageState[v.size ()]);
  }

  private void addStateToVector (PackageState state, Vector v)
  {
    for (int i = 0; i < v.size (); i++)
      {
	if (v.elementAt (i) == state)
	  {
	    return;
	  }
      }

    if (state.getPending () == PackageState.PS_INSTALL_PENDING
     || state.getPending () == PackageState.PS_NEW_REPLACE_PENDING)
      {
	IPackageQuery[] predeps = state.getXoePackage ().getDescription ().getPredependencies ();

	for (int i = 0 ; predeps != null && i < predeps.length; i++)
	  {
	    Vector predepMatches = new Vector ();
	    performQuery (predeps[i], predepMatches, PDB_FINAL_PACKAGES);
	    for (int j = 0; j < predepMatches.size (); j++)
	      {
		PackageDescription desc = (PackageDescription)predepMatches.elementAt (j);
		PackageState pdState = (PackageState)m_packageStates.get (desc.getPackageURI());
		if (pdState.getPending () == PackageState.PS_INSTALL_PENDING
		    || pdState.getPending () == PackageState.PS_NEW_REPLACE_PENDING)
		  {
		    addStateToVector (pdState, v);
		  }
	      }
	  }
      }

    v.addElement (state);
  }

  /*  private void doDependencies ()
  {
    for (Enumeration it = m_packageStates.elements (); it.hasMoreElements (); )
      {
	XoePackage pkg = ((PackageState) it.nextElement ()).getXoePackage ();

	PackageDescription desc = pkg.getDescription ();
	
	IPackageQuery[] deps = desc.getDependencies ();

        if (deps != null)
          {
            for (int i = 0; i < deps.length; i++)
              {
                Vector matches = new Vector ();

                // search for packages that will be installed when we're done,
                // which match the dependency
                performQuery (deps[i], matches, PDB_FINAL_PACKAGES);
                for (int j = 0; j < matches.size (); j++)
                  {
                    desc = (PackageDescription) matches.elementAt (j);
                    XoePackage dpkg = PackageManager.retrievePackageLocally (desc);
                    pkg.getClassLoader ().addClassLoader (dpkg.getClassLoader ());
                  }
              }
          }
      }
  }
  */
  private void cleanUp ()
  {
    for (Enumeration it = m_packageStates.keys (); it.hasMoreElements (); )
      {
	Object key = it.nextElement ();
	PackageState state = (PackageState) m_packageStates.get (key);
	if (state.getState () != PackageState.PS_INSTALLED)
	  {
	    m_packageStates.remove (key);
	  }
	else
	  {
	    state.setPending (PackageState.PS_NO_PENDING);
	  }
      }    
  }

  private void doVerify (PackageState[] states)
    throws PackageException
  {
    for (int i = 0; i < states.length; i++)
      {
	states[i].verify ();
      }
  }

  private void doCommit (PackageState[] states)
    throws PackageException
  {
    for (int i = 0; i < states.length; i++)
      {
	try
	  {
	    states[i].doCommit ();
	  }
	catch (PackageException ex)
	  { // okay, back out everything before this
	    for (int j = i - 1; j >= 0; j--)
	      {
		states[j].backOut ();
	      }
	    throw ex;
	  }
      }
  }

  private void postCommit (PackageState[] states)
  {
    for (int i = 0; i < states.length; i++)
      {
	states[i].postCommit ();
      }
  }

  XoePackage getXoePackage (String pkgURI)
  {
    PackageState state = (PackageState) m_packageStates.get(pkgURI);
    return state == null ? null : state.getXoePackage();
  }

  PackageDescription[] getInstalledPackages ()
  {
    Vector v = new Vector ();

    for (Enumeration it = m_packageStates.elements (); it.hasMoreElements (); )
      {
	PackageState state = (PackageState)it.nextElement ();
	// REGISTERED is the state a package has after the installation is complete,
	// but while it still might be backed out (ie, if another package in the transaction
	//  fails to install)
	if (state.getState () == state.PS_INSTALLED || state.getState () == state.PS_REGISTERED)
	  {	   
	    v.addElement (state.getXoePackage ().getDescription ());
	  }
      }
    return (PackageDescription[]) getValues (v, new PackageDescription[v.size ()]);
  }

  public void performQuery (IPackageQuery in_q, Vector out_res)
  { // by default, look only at installed packages which are staying installed,
    // not pending to be installed, etc...
    performQuery (in_q, out_res, PDB_INSTALLED_PACKAGES);
  }

  public void performQuery (IPackageQuery in_q, Vector out_res, int type)
  {
    for (Enumeration it = m_packageStates.elements (); it.hasMoreElements (); )
      {
	((PackageState)it.nextElement ()).performQuery (in_q, out_res, type);
      }
  }

  private void processActionSet (PackageActionSet pas)
    throws PackageNotFoundException
  {
    Hashtable ht = pas.getActionMap ();
    PackageDescription desc;

    // store new package states in a vector until the entire
    // PackageActionSet has been processed.
    Vector newStates = new Vector ();
    
    for (Enumeration it = ht.elements (); it.hasMoreElements (); )
      {
	PackageAction action = (PackageAction)it.nextElement ();
	desc = action.getDescription ();
        // retrieve the actual XoePackage.
        XoePackage pkg = PackageManager.retrievePackage (desc);
	if (action.getAction () == PackageAction.PA_SELECTED_INSTALL)
	  {
	    // PackageState (XoePackage): pending=pending install, state=not installed
	    newStates.addElement (new PackageState (pkg));
	  }
	else if (action.getAction () == PackageAction.PA_SELECTED_REMOVE)
	  {
	    String uid = action.getDescription ().getPackageURI ();
	    PackageState state = (PackageState) m_packageStates.get (uid);
	    if (state != null)
	      {
		state.setPending (PackageState.PS_REMOVE_PENDING);
	      }
	  }
	else if (action.getAction () == PackageAction.PA_SELECTED_REPLACE)
	  {
	    String uid = action.getOriginal ().getPackageURI ();
	    PackageState state = (PackageState) m_packageStates.get (uid);
	    if (state != null)
	      {
		// PackageState (XoePackage doc, PackageState s)
		// pending=pending to replace s with doc, state=not installed
		// also marks s.pending as pending to be replaced by doc
		newStates.addElement (new PackageState (pkg, state));
	      }
	    else
	      { // is this the best way? turns the replace into an install
		newStates.addElement (new PackageState (pkg));
	      }
	  }
	// else it is marked installed already or not installed and not planned
	// for install.... don't need to do anything, but could validate this...
      }

    // if the entire PackageActionSet has been processed without any PackageNotFound
    // exceptions then we add the new states the hash table.
    for (int i = 0; i < newStates.size (); i++)
      {
        addPackage ((PackageState) newStates.elementAt (i));
      }
  }

  static Object[] getValues (Vector v, Object[] a)
  {
    int i = 0;
    for (Enumeration e = v.elements (); e.hasMoreElements (); i++)
      {
        a[i] = e.nextElement ();
      }
    return a;
  }
}
