/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: NativeInstaller.java,v 1.7 2001/11/14 01:16:21 samc Exp $
 */ 
package org.xoe.core.packages;

import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.net.FileNameMap;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.w3c.dom.Element;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;


/**
 * Installer for native libraries.  This includes JNI libs as well
 * as regular C libs and (RSN) kernel modules.
 */
public class NativeInstaller 
  extends FileIteratingInstaller
{
  private static final String helperURL = "stash:/NativeInstallerHelper.class";

  /**
   * The ExData for a native installer.  Simply an array of ths names
   * of the objects files that were installed in the system.
   */
  private static class NativeInstallerData
  {
    private File[] m_soFiles = null;

    private NativeInstallerData (File[] soFiles)
    {
      m_soFiles = soFiles;
    }

    private File[] getSharedObjects ()
    {
      return m_soFiles;
    }
  }

  static final String NAMESPACE = "http://www.xoe.org/installer/native";
  static final String ARCHITECTURE_NAMESPACE = "http://www.xoe.org/installer/native/architecture";

  private static final String LIBRARY_DIR = TEMP_DIR + "lib/";
  private static final FileNameMap map = new kaffe.net.DefaultFileNameMap ();
  private boolean archDepends = false;

  static
  {
    new File (LIBRARY_DIR).mkdirs ();
  }

  public String[] getHandlerPredependencies ()
  {
    return new String[] { ClassInstaller.NAMESPACE };
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    if (NAMESPACE.equals (depEl.getAttribute ("ns")) || ARCHITECTURE_NAMESPACE.equals (depEl.getAttribute ("ns")))
      {
	return new Dependency (depEl);
      }
    else
      {
	throw new PackageDescriptionException ("Unrecognized dependency type:\n" 
                                               + XMLWriter.getXML (depEl));
      }
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  /**
   * Add provisios for all object code elements in the ZipInputStream.
   * This happens at *build* time.
   */
  protected void processFile (ZipInputStream zis, ZipEntry entry)
  {
    String fileName = entry.getName ();
    if (fileIsSharedObject (fileName))
      {
	m_provides.addElement (new Proviso (NAMESPACE, fileName, null));
      }
  }

  /**
   * For now a shared object is anything with ending on ".so" with optional tailing
   * version numbers.
   * [FIXME] This should be a native method.
   */
  private static boolean fileIsSharedObject (String fileName)
  {
    int i = fileName.lastIndexOf (".");
    int j = fileName.length ();

    while (i != -1)
      {
        String field = fileName.substring (i + 1, j);
        if (field.equals ("so"))
          {
            return true;
          }
        else
          {
            try
              {
                Integer.parseInt (field);
              }
            catch (NumberFormatException ex)
              { // only numbers allowed after the .so
                // this isn't a shared-object
                return false;
              }
          }
        j = i;
        i = fileName.lastIndexOf (".", i - 1);
      }
    return false;
  }
 
  public void verify (XoePackage xoePackage, Proviso[] provisos)
  { // could check to make sure the architecture of the package
    // matches the system architecture, or that all native files
    // are actually built for this architecture, etc...
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    XoeClassLoader loader = ((ClassInstallerData) xoePackage.getExData (ClassInstaller.NAMESPACE)).getClassLoader ();
    INativeInstallerHelper helper = null;

    try
      {
        ContentElement ce = Stash.fetch (helperURL);
        if (ce == null)
          {
            throw new InstallException ("NativeInstaller requires: '" + helperURL + "'");
          }
        byte buf[] = new byte[ce.getContentLength ()];
        InputStream is = ce.getInputStream ();
        for (int res = 0, off = 0; res != -1; )
          {
            res = is.read (buf, off, buf.length - off);
            off += res;
          }

        Class helperclass = loader.addClass (buf, 0, buf.length);
        helper = (INativeInstallerHelper) helperclass.newInstance ();
      }
    catch (Exception ex)
      {
        Logger.log (this, ex);
        throw new InstallException (ex.toString ());
      }

    // first load all the shared libraries we depend on (gotta find'em)
    PackageDescription[] allPkgs = PackageManager.getInstalledPackages ();
    IPackageQuery[] deps = xoePackage.getDescription ().getDependencies ();

    for (int i = 0; i < allPkgs.length; i++)
      {
        // we are not interested in non-native packages
        if (allPkgs[i].getArchitecture () == null)
          {
            continue;
          }
        // try to find out if this package depends on the other one
        for (int j = 0; j < deps.length; j++)
          {
            Vector v = new Vector ();
            allPkgs[i].performQuery (deps[j], v);
            if (v.size () > 0)
              { // dep matched the package
                XoePackage other = PackageManager.retrievePackageLocally (allPkgs[i]);
                NativeInstallerData otherData = (NativeInstallerData) other.getExData (NAMESPACE);
                if (otherData != null)
                  {
                    File[] soFiles = otherData.getSharedObjects ();
                    for (int k = 0; k < soFiles.length; k++)
                      {
                        try 
                          {
                            helper.load (soFiles[k].toString ());
                          }
                        catch (UnsatisfiedLinkError ule)
                          {
                            throw new InstallException (xoePackage,
                                ("UnsatisfiedLinkError while installing " +
                                 "native library: " + soFiles[k]));
                          }
                      }
                  }
                break;
              }
          }
      }

    Vector soFiles = new Vector ();
    ZipInputStream zis = xoePackage.getZipInputStream ();
    
    try 
      {
	while (true)
	  {
	    ZipEntry entry = zis.getNextEntry ();
	    if (entry == null) break;
            String filename = entry.getName ();
	    
            if (fileIsSharedObject (filename))
	      {
                byte buf[] = new byte[(int) entry.getSize ()];
                for (int res = 0, off = 0; res != -1; )
                  {
                    res = zis.read (buf, off, buf.length - off);
                    off += res;
                  }

                /* make a temporary file. */
                /* [FIXME] must we do this? */
                File file = new File (filename);
                filename = LIBRARY_DIR + file.getName ();
                file = new File (filename);

                if (file.exists ())
                  {
                    Logger.log (NativeInstaller.class, "warning: file already exists: " + filename);
		    file.delete ();
                  }

                /* read the data from the ZipInputStream */
                /* write the library object to a file */
                FileOutputStream ofile = new FileOutputStream (filename);
                ofile.write (buf, 0, buf.length);  
                
                /* load the file */
                try 
                  {
		    helper.load (filename);
                  }
                catch (UnsatisfiedLinkError ule)
                  {
		    throw new InstallException (xoePackage,
					       ("UnsatisfiedLinkError while installing native " +
						"library: " + filename));
                  }

		soFiles.addElement (file);
	      }
	    zis.closeEntry ();
	  }
	zis.close ();
      }
    catch (IOException e)
      {
	throw new InstallException (e, xoePackage, ("An error occurred while performing " +
						    "shared library installation."));
      }

    xoePackage.setExData (NAMESPACE, new NativeInstallerData ( (File[])soFiles.toArray (new File[soFiles.size ()])));
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    NativeInstallerData data = (NativeInstallerData) xoePackage.getExData (NAMESPACE);
    
    File[] soFiles = data.getSharedObjects ();

    for (int i = 0; i < soFiles.length; i++)
	{
	  soFiles[i].delete ();
	}
    
    xoePackage.setExData (NAMESPACE, null);

    // sharedlibs are unloaded when classloaders are destroyed
  }

  // ------------------------------------------------------
  // build time functions used to generate package.xml 
  // ------------------------------------------------------
  
  public Vector generateRequires ()
  {
    String arch = m_package.getArchitecture ();
    if (arch != null)
      {
        StringTokenizer tokenizer = new StringTokenizer (arch, "-");
        while (tokenizer.hasMoreTokens ())
          {
            m_requires.addElement (new Dependency (ARCHITECTURE_NAMESPACE, tokenizer.nextToken ()));
          }
      }
    return m_requires;
  }

  public Vector generateProvides ()
  {
    if (m_package.getName ().equals ("architecture-native"))
      {
	String prop = System.getProperty ("xoe.hostarch");
        if (prop == null)
          {
            throw new Error ("xoe.hostarch property must be set when building native packages");
          }
	StringTokenizer tokenizer = new StringTokenizer (prop, "-");
	while (tokenizer.hasMoreTokens ())
	  {
	    String tok = tokenizer.nextToken ();
	    m_provides.addElement (new Proviso (ARCHITECTURE_NAMESPACE, tok));
	    // [FIXME] this is a hack
	    if (tok.equals ("i686"))
	      {
		m_provides.addElement (new Proviso (ARCHITECTURE_NAMESPACE, "i586"));
		m_provides.addElement (new Proviso (ARCHITECTURE_NAMESPACE, "i486"));
		m_provides.addElement (new Proviso (ARCHITECTURE_NAMESPACE, "i386"));
	      }
	    if (tok.equals ("i586"))
	      {
		m_provides.addElement (new Proviso (ARCHITECTURE_NAMESPACE, "i486"));
		m_provides.addElement (new Proviso (ARCHITECTURE_NAMESPACE, "i386"));
	      }
	    if (tok.equals ("i486"))
	      {
		m_provides.addElement (new Proviso (ARCHITECTURE_NAMESPACE, "i386"));
	      }
	  }
      }
    return m_provides;
  }
}
