/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageActionSet.java,v 1.2 2001/11/29 20:03:28 erik Exp $
 */
package org.xoe.core.packages;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import org.xoe.core.util.Logger;


public class PackageActionSet 
  implements IPackageQueryHandler
{
  private PackageDescription[] m_installs;
  private PackageDescription[] m_uninstalls;
  private PackageDescription[][] m_replaces;
  
  private Hashtable m_mapPackageActions = new Hashtable();

  // returns true if there are no conflicts or unmet dependencies
  public boolean isComplete ()
  {
    try
      {
	checkComplete ();
      }
    catch (PackageIncompleteException ex)
      {
	return false;
      }
    return true;
  }

  public void checkComplete ()
    throws PackageIncompleteException
  {
    PackageAction[] actions = (PackageAction[]) PackageUtils.getValues (m_mapPackageActions);

    for(int i = 0; i < actions.length; i++)
      {
	if(actions[i].getAction() == PackageAction.PA_SELECTED_INSTALL
	   || actions[i].getAction() == PackageAction.PA_SELECTED_REPLACE)
	  {
	    actions[i].checkCompleteInSet(this);
	  }
	else
	  {
	    actions[i].checkConflictsInSet(this);
	  }

	if(actions[i].getAction() == PackageAction.PA_SELECTED_REPLACE)
	  {
	    PackageDescription op = actions[i].getOriginal ();
	    PackageDescription np = actions[i].getDescription ();

	    Logger.log (this, "CHECK REPLACE: " + op + " -> " + np);
	    if (op.getName ().equals ("architecture-native") &&
		!op.getArchitecture ().equals (np.getArchitecture ()))
	      {
		Vector v = new Vector ();
		v.addElement (op);
		Logger.log (this, "DENIED");
		throw new PackageIncompleteException (np, v, null);
	      }
	  }
      }
  }

  public void performQuery (IPackageQuery in_q, Vector out_res)
  {
    PackageAction[] actions = (PackageAction[]) PackageUtils.getValues (m_mapPackageActions);

    for(int i = 0; i < actions.length; i++)
      {
	actions[i].performQuery(in_q, out_res);
      }    
  }

  public PackageActionSet (PackageActionSet orig, InstallContext add)
  {
    this(orig);

    PackageDescription[] descs = add.getPackages();

    for(int i = 0; i < descs.length; i++)
      {
	selectPackage(descs[i]);
      }
  }

  public PackageActionSet (PackageActionSet orig, UninstallContext remove)
  {
    this (orig);

    PackageDescription[] descs = remove.getPackages();

    for(int i = 0; i < descs.length; i++)
      {
	selectForRemoval(descs[i]);
      }
  }

  public PackageActionSet ()
  {
    m_mapPackageActions = new Hashtable();
  }

  private PackageActionSet (PackageActionSet toCopy)
  {
    try
      {
	m_mapPackageActions = new Hashtable(toCopy.m_mapPackageActions.size());

	for (Enumeration it = toCopy.m_mapPackageActions.elements(); it.hasMoreElements(); )
	  {
	    PackageAction action = (PackageAction)((PackageAction)it.nextElement()).clone();
	    m_mapPackageActions.put(action.getName(), action);
	  }
      }
    catch(CloneNotSupportedException ex)
      {
	// can't happen
      }
  }

  public Object clone ()
    throws CloneNotSupportedException
  {
    return new PackageActionSet (this);
  }

  public void selectForRemoval (PackageDescription desc)
  {
    m_installs = null;
    m_uninstalls = null;
    m_replaces = null;

    PackageAction action = (PackageAction)m_mapPackageActions.get(desc.getName());
    if (action == null)
      {
	Logger.log(this, "asked to remove package " + desc + " which was not installed");
	m_mapPackageActions.put(desc.getName(),
			      new PackageAction(desc, PackageAction.PA_SELECTED_REMOVE));
      }
    else if (action.getDescription().equals(desc))
      {
	if (action.getAction() == PackageAction.PA_INSTALLED)
	  {
	    m_mapPackageActions.put(desc.getName(),
				    new PackageAction(desc, PackageAction.PA_SELECTED_REMOVE));
	  }
	else if (action.getAction() == PackageAction.PA_SELECTED_INSTALL)
	  {
	    m_mapPackageActions.remove(desc.getName());
	  }
	else if (action.getAction() == PackageAction.PA_SELECTED_REPLACE )
	  { // leave the original package...
	    action.setAction(PackageAction.PA_INSTALLED);
	  }
	else if (action.getAction() == PackageAction.PA_SELECTED_REMOVE ||
		action.getAction() == PackageAction.PA_NOTINSTALLED)
	  {
	    // do nothing - this is what we want
	  }
      }
    else
      {
	// a package by that name is listed at a different version..
	// do nothing
      }

  }

  public void selectPackage (PackageDescription desc)
  {
    m_installs = null;
    m_uninstalls = null;
    m_replaces = null;

    PackageAction action = (PackageAction)m_mapPackageActions.get(desc.getName());
    if(action == null)
      {
	m_mapPackageActions.put(desc.getName(),
			      new PackageAction(desc, PackageAction.PA_SELECTED_INSTALL));
      }
    else if(action.getDescription().equals(desc))
      {
	if(action.getAction() == PackageAction.PA_INSTALLED ||
	   action.getAction() == PackageAction.PA_SELECTED_INSTALL ||
	   action.getAction() == PackageAction.PA_SELECTED_REPLACE )
	  {
	    // this package is already installed... do nothing
	  }
	else if(action.getAction() == PackageAction.PA_SELECTED_REMOVE)
	  {
	    action.setAction(PackageAction.PA_INSTALLED);
	  }
	else if(action.getAction() == PackageAction.PA_NOTINSTALLED)
	  {
	    action.setAction(PackageAction.PA_SELECTED_INSTALL);
	  }
      }
    else
      {
	// a package by that name is already listed at a different version..
	// either upgrade or downgrade....
	action.replaceWith(desc);
      }

  }

  protected void addAction (PackageAction action)
  {
    m_mapPackageActions.put(action.getDescription().getName(), action);
  }

  public PackageDescription[] getNewInstalls()
  {
    if(m_installs != null)
      {
	return m_installs;
      }
    Vector v = new Vector();
    PackageAction[] actions = (PackageAction[]) PackageUtils.getValues (m_mapPackageActions);

    for(int i = 0; i < actions.length; i++)
      {
	if(actions[i].getAction() == PackageAction.PA_SELECTED_INSTALL)
	  {
	    v.addElement(actions[i].getDescription());
	  }
      }
    m_installs = (PackageDescription[]) PackageUtils.getValues (v);
    if(m_installs == null)
      {
	m_installs = new PackageDescription[0];
      }
    return m_installs;
  }

  public PackageDescription[][] getReplacements()
  {
    if(m_replaces != null)
      {
	return m_replaces;
      }
    Vector v = new Vector();
    PackageAction[] actions = (PackageAction[]) PackageUtils.getValues (m_mapPackageActions);

    for(int i = 0; i < actions.length; i++)
      {
	if(actions[i].getAction() == PackageAction.PA_SELECTED_REPLACE)
	  {
	    PackageDescription[] set = new PackageDescription[2];
	    set[0] = actions[i].getOriginal();
	    set[1] = actions[i].getDescription();

	    v.addElement(set);
	  }
      }
    m_replaces = (PackageDescription[][]) PackageUtils.getValues(v);
    if(m_replaces == null)
      {
	m_replaces = new PackageDescription[0][];
      }
    return m_replaces;
  }

  public PackageDescription[] getUninstalls()
  {
    if(m_uninstalls != null)
      {
	return m_uninstalls;
      }
    Vector v = new Vector();
    PackageAction[] actions = (PackageAction[]) PackageUtils.getValues (m_mapPackageActions);

    for(int i = 0; i < actions.length; i++)
      {
	if(actions[i].getAction() == PackageAction.PA_SELECTED_REMOVE)
	  {
	    v.addElement(actions[i].getDescription());
	  }
      }
    m_uninstalls = (PackageDescription[]) PackageUtils.getValues (v);
    if(m_uninstalls == null)
      {
	m_uninstalls = new PackageDescription[0];
      }
    return m_uninstalls;
  }
  
  protected Hashtable getActionMap()
  {
    return m_mapPackageActions;
  }

  public String toString()
  {
    StringBuffer strb = new StringBuffer();
    PackageAction[] actions = (PackageAction[]) PackageUtils.getValues (m_mapPackageActions);

    if (actions.length > 0)
      {
	strb.append (actions[0].toString ());
	for (int i = 1 ; i < actions.length; i++)
	  {
	    strb.append (", " + actions[i].toString());
	  }
      }
    return strb.toString();
  }
}
