/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageDescription.java,v 1.7 2001/12/15 01:03:48 erik Exp $
 */
package org.xoe.core.packages;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;


/**
 * Encapsulation of an XML package description
 */
public class PackageDescription 
  implements IPackageQueryHandler
{
  // for debugging/profiling purposes
  private static long queryTime = 0;

  // The actual <package/> document
  private Document m_packageDoc = null;
  private XPathContext m_context = new XPathContext ();

  // IPackageQuery's for our dependencies and conflicts
  private IPackageQuery[] m_predependencies = null;
  private IPackageQuery[] m_dependencies = null;
  private IPackageQuery[] m_conflicts = null;

  // Basic information about the package
  private PackageVersion m_version = null;
  private String m_name = null;
  private String m_arch = null;
  private String m_uri = null;
  private Element m_infoTag = null;

  // Cache the results of PackageQueries performed on this package.
  // The hash maps XPath strings to a boolean flag indicating whether
  // or not the query was met by this package.
  private Hashtable m_cacheQueryResults = new Hashtable ();


  // Profiling
  // queryTime is the total time spent performing XPath queries
  // on this package since the last call to resetQueryTime
  public static void resetQueryTime ()
  {
    queryTime = 0;
  }

  // Profiling
  public static long getQueryTime ()
  {
    return queryTime;
  }

  /**
   * Builds a PackageDescription out of a DOM Document
   *
   * @exception PackageDescriptionException if the Document is not a valid package description
   */
  public PackageDescription (Document pkgDoc)
    throws PackageDescriptionException
  {
    try
      {
	m_packageDoc = pkgDoc;
	m_context.pushNode (m_packageDoc);
	m_infoTag = (Element) XPath.findFirstNode (m_packageDoc, "package/info");
	m_context.setFunctionProvider (PackageXPathFunctionProvider.singleton);

	// this will go away when all packages require an info tag
	if (m_infoTag == null)
	  {
	    m_infoTag = m_packageDoc.createElement ("info");
	    m_packageDoc.getDocumentElement ().appendChild (m_infoTag);
	  }

	// extract the required attributes.
	m_name = m_packageDoc.getDocumentElement ().getAttribute ("name");
	m_arch = m_packageDoc.getDocumentElement ().getAttribute ("architecture");
	String strVer = m_packageDoc.getDocumentElement ().getAttribute ("version");
	
	if (m_name == null || strVer == null)
	  {
	    throw new PackageDescriptionException ("package must have a name and a version");
	  }
	
	// create a PackageVersion from the string.
	try
	  {
	    m_version = new PackageVersion (strVer);
	  }
	catch (NumberFormatException ex)
	  {
	    throw new PackageDescriptionException ("invalid package version: " + strVer);
	  }
	
	readDependencies();
	readPredependencies();
	readConflicts();
      }
    catch (PackageDescriptionException ex)
      {
	ex.setPackageName (getPackageURI ());
	throw ex;
      }
  }

  /**
   * Reads an XML file from an InputStream and builds a PackageDescription out of it
   *
   * @exception PackageDescriptionException if the stream does not contain a valid
   *                                        package description
   */
  public PackageDescription (InputStream is)                                          
    throws IOException, PackageDescriptionException                                           
  {                                           
    this (DOMlet.create (is, "PackageDescription(InputStream)"));
  }

  /**
   * Builds an IPackageQuery object that will exclusively match all packages
   * that are upgrades to this one
   */
  public IPackageQuery getUpgradeQuery ()
  {
    // setMinVersion with false to mean "not inclusive"
    return (new PackageQuery ()).setName (getName ()).setMinVersion (getVersion (), false);
  }

  public void performQuery (IPackageQuery in_q, Vector out_res)
  {
    if (in_q == null)
      {
	return;
      }

    long stime = System.currentTimeMillis ();

    String strX = in_q.toXPath ();

    Boolean match = (Boolean) m_cacheQueryResults.get (strX);

    if (match == null)
      {
	NodeList ret = XPath.findNodes (m_context, strX);
	
	if (ret.getLength () > 0)
	  {
	    match = Boolean.TRUE;
	  }
	else
	  {
	    match = Boolean.FALSE;
	  }

	m_cacheQueryResults.put (strX, match);
      }

    if (match == Boolean.TRUE)
      {
	out_res.addElement(this);
      }

    long etime = System.currentTimeMillis();

    queryTime += (etime - stime);
  }

  /**
   * Returns a vector containing all packages that conflict with this one in the
   * set represented by the specified IPackageQueryHandler
   */
  public Vector getConflictsInSet (IPackageQueryHandler qHandler)
  {
    Vector confPkgs = new Vector ();
    
    if (m_conflicts != null)
      {
        for (int i = 0; i < m_conflicts.length; i++)
          {
            qHandler.performQuery (m_conflicts[i], confPkgs);
          }
      }

    return confPkgs;
  }

  /**
   * Returns a vector of IPackageQuery objects representing dependencies of this
   * package that were not met in the set of packages represented by the specified
   * IPackageQueryHandler
   */
  public Vector getUnmetDependenciesInSet (IPackageQueryHandler qHandler)
  {
    Vector unmetDeps = new Vector ();

    if (m_dependencies != null)
      for (int i = 0; i < m_dependencies.length; i++)
	{
	  Vector v = new Vector ();
	  qHandler.performQuery (m_dependencies[i], v);
	  if (v.size () == 0)
	    {
	      unmetDeps.addElement (m_dependencies[i]);
	    }
	}

    return unmetDeps;
  }

  /**
   * Throws an exception if any packages in the set conflict with this one, or if
   * this package has dependencies that are not met by the set. The set is
   * represented by the specified IPackageQueryHandler
   *
   * @exception PackageIncompleteException Thrown if this package has any conflicts
   *                                       or unmet dependencies in the set.
   */
  public void checkCompleteInSet (IPackageQueryHandler qHandler)
    throws PackageIncompleteException
  {
    Vector confPkgs = getConflictsInSet (qHandler);

    Vector unmetDeps = getUnmetDependenciesInSet (qHandler);

    if (unmetDeps.size() > 0 || confPkgs.size() > 0)
      {
	throw new PackageIncompleteException (this, unmetDeps, confPkgs);
      }
  }

  private void readConflicts ()
    throws PackageDescriptionException
  {
    Vector v = new Vector ();
    NodeList nl = XPath.findNodes (m_context, "package/conflicts/dep");
    for (int i = 0; i < nl.getLength (); i++)
      {
	v.addElement (new PackageQuery ((Element) nl.item (i)));
      }
    m_conflicts = (IPackageQuery[]) PackageUtils.getValues (v);

  }

  /**
   * Returns an array of IPackageQuery objects representing packages that conflict with
   * this one
   */
  public IPackageQuery[] getConflicts ()
  {
    return m_conflicts;
  }

  /**
   * Add an URL from which this package can be retrieved
   */
  public void setSuggestedURL (String url)                                          
  {
    Element urlEl = m_packageDoc.createElement ("package-url");                                          
    urlEl.appendChild(m_packageDoc.createTextNode (url));                                           
    m_infoTag.appendChild (urlEl);    
  }

  /**
   * Add a notation indicating the size of the XOE archive file for this package
   *
   * @param size the size, in bytes of the XOE archive file
   */
  public void setSize (long size)                                           
  {                                           
    NodeList nl = XPath.findNodes (m_context, "package/info/size/text()");
    
    if (nl.getLength () > 0)      
      {                                           
	((Text)nl.item (0)).setData (size + "B");
      }                                           
    else      
      {                                           
        Element sizeEl = m_packageDoc.createElement ("size");                                          
        sizeEl.appendChild (m_packageDoc.createTextNode (size + "B"));                                           
        m_infoTag.appendChild (sizeEl);    	
      }    
  }

  private void readPredependencies ()
    throws PackageDescriptionException
  {
    Vector v = new Vector ();
    NodeList nl = XPath.findNodes (m_context, "package/requires/dep[@predepends='true']");
    for (int i = 0; i < nl.getLength (); i++)
      {
	v.addElement (new PackageQuery ((Element) nl.item (i)));
      }
    m_predependencies = (IPackageQuery[]) PackageUtils.getValues (v);
    if (m_predependencies == null)
      {
	m_predependencies = new IPackageQuery[0];
      }
  }

  /**
   * Returns an array of package queries representing packages upon which this one
   * predepends
   */
  public IPackageQuery[] getPredependencies ()
  {
    return m_predependencies;
  }

  private void readDependencies ()
    throws PackageDescriptionException
  {
    Vector v = new Vector ();
    NodeList nl = XPath.findNodes (m_context, "package/requires/dep");
    for (int i = 0; i < nl.getLength (); i++)
      {
        v.addElement (new PackageQuery ((Element) nl.item (i)));
      }
    m_dependencies = (IPackageQuery[]) PackageUtils.getValues (v);
    if (m_dependencies == null)
      {
        m_dependencies = new IPackageQuery[0];
      }
  }

  /**
   * Returns an array of package queries representing packages upon which this one
   * depends
   */
  public IPackageQuery[] getDependencies ()
  {
    return m_dependencies;
  }

  /**
   * Gets the architecture the described package was built for
   */
  public String getArchitecture ()
  {
    return m_arch;
  }

  /**
   * Sets the architecture flag in the package. Should only be used at build time.
   */
  public void setArchitecture (String arch)
  {
    if (m_arch != null)
      {
        throw new Error ("setArchitecture can only be called once");
      }
    m_arch = arch;
  }

  /**
   * Gets the name listed in the info section of the description. The friendly
   * name is not guaranteed to be unique but has fewer restrictions on legal
   * characters and is usually more verbose. If no friendly name is provided,
   * returns the technical name.
   */
  public String getFriendlyName ()
  {
    if (m_infoTag == null)
      {
	return m_name;
      }

    String fname = XPath.findString (m_infoTag, "name");

    return (fname == null || fname.length() == 0) ? m_name : fname;
  }

  /**
   * Get the vendor name listed in the info section of the description.
   *
   * @return null if no vendor was specified
   */
  public String getVendor ()
  {
    return m_infoTag == null ? null : XPath.findString (m_infoTag, "vendor");
  }

  /**
   * Get the maintainer name listed in the info section of the description.
   *
   * @return null if no maintainer was specified
   */
  public String getMaintainerName ()
  {
    return m_infoTag == null ? null : XPath.findString (m_infoTag, "maintainer/name");
  }

  /**
   * Get the maintainer email listed in the info section of the description.
   *
   * @return null if no maintainer email was specified
   */
  public String getMaintainerEmail ()
  {
    return m_infoTag == null ? null : XPath.findString (m_infoTag, "maintainer/email");
  }

  /**
   * Get returns the year that the package was copyrighted
   *
   * @return null if no copyright year was specified
   */
  public String getCopyrightYear ()
  {
    return m_infoTag == null ? null : XPath.findString (m_infoTag, "copyright/year");
  }
  
  /**
   * Get the copyright holder name listed in the info section of the description.
   *
   * @return null if no copyright holder was specified
   */
  public String getCopyrightHolder ()
  {
    return m_infoTag == null ? null : XPath.findString (m_infoTag, "copyright/holder");
  }

  /**
   * Get the unique name of this package
   */
  public String getName ()
  {
    return m_name;
  }

  /**
   * Get the version of this package
   */
  public PackageVersion getVersion ()
  {
    return m_version;
  }

  /**
   * Get a DOM representation of this package description.
   */
  public Document getDocument ()
  {
    return m_packageDoc;
  }

  /**
   * Return the textual description of this package with no XHTML markup
   */
  public String getTextDescription ()
  {
    NodeList nl = XPath.findNodes (m_context, "package/info/description/text()");

    if (nl.getLength () > 0)
      {
	StringBuffer strb = new StringBuffer ();
	for (int i = 0; i < nl.getLength (); i++)
	  {
	    strb.append (((Text)nl.item (i)).getData () + " ");
	  }
	return strb.toString ();
      }

    return null;
  }

  /**
   * Return the textual description of this package. It may contain XHTML markup.
   */
  public Element getXHTMLDescription ()
  {
    return (Element) XPath.findFirstNode (m_context, "package/info/description");
  }
  
  /**
   * Returns the first URL listed from which the .xoe archive for this package can
   * be retrieved
   */
  public URL getSuggestedURL ()
  {
    Element elem = (Element) XPath.findFirstNode (m_context, "package/info/package-url");

    if (elem != null)
      {
        String surl = DOMUtils.getAssignedValue (elem);
        if (surl != null)
          {
            try
              {
                return new URL (surl);
              }
            catch (MalformedURLException ex)
              {
                // return null
              }
          }
      }
    return null;
  }

  /**
   * A more readable identifier of the package
   */
  public String toString ()
  {
    if (m_arch != null)
      {
        return m_name + " " + String.valueOf (m_version) + " (" + m_arch + ")";
      }
    else
      {
        return m_name + " " + String.valueOf(m_version);
      }
  }

  /**
   * Generate the filename for the package based on its name, version and arch.
   */
  public String getXoeFilename ()
  { 
    return getPackageURI () + ".xoe";
  }

  /**
   * m_arch is null for non-native packages.
   */
  public String getPackageURI ()
  {
    if (m_uri == null)
      {
        if (m_arch != null)
          {
            m_uri = m_name + "_" + m_arch + "_" + String.valueOf(m_version);
          }
        else
          {
            m_uri = m_name + "_" + String.valueOf(m_version);
          }
      }
    return m_uri;
  }

  /**
   * Compares objects by their unique identifier strings. 
   */
  public boolean equals (PackageDescription desc)
  {
    return getPackageURI ().equals (desc.getPackageURI ());
  }

  /**
   * Return the newer of two packages.
   */
  public static PackageDescription newest (PackageDescription p1, PackageDescription p2)
  {
    return p1.getVersion ().geq (p2.getVersion ()) ? p1 : p2;
  }

  /**
   * Return the newest out of a Vector of PackageDescriptions.
   */
  public static PackageDescription newest (Vector pkgs)
  {
    if (pkgs.size () == 0)
      {
	return null;
      }

    PackageDescription cand = (PackageDescription)pkgs.elementAt (0);

    for (int i = 1; i < pkgs.size (); i++)
      {
	cand = newest (cand, (PackageDescription)pkgs.elementAt (i));
      }

    return cand;
  }

  /** 
   * find out what the name of a package 'packageName', with version 'version',
   * would be (if it were compiled for 'arch'). Arch is ignored for non-native packages.
   */
  public static String getXoeFilename (String packageName, String arch, String version)
  { 
    return getPackageURI (packageName, arch, version) + ".xoe";
  }

  /**
   * Determines the package URI for a package with the specified name, architecture (null
   * if non-native), and version
   */
  public static String getPackageURI (String packageName, String arch, String version)
  {
    if (arch != null)
      {
        return packageName + "_" + arch + "_" + version;
      }
    else
      {
        return packageName + "_" + version;
      }
  }
}
