/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageException.java,v 1.3 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

import java.io.PrintStream;
import java.io.PrintWriter;

/**
 * Base class for all packaging exceptions. Can be used to wrap a Throwable
 * (such as an Error, RuntimeException, or other Exceptions) before throwing
 * from a packaging function. Can also be correlated to the package that 'caused'
 * the exception.
 */
public class PackageException 
  extends Exception
{
  private String m_strDetail = null;
  private Throwable m_origT = null;
  private String m_packageName = null;

  /**
   * Identify the package that 'caused' the exception. Can be called after the
   * exception is thrown since the thrower doesn't always know what package it's
   * dealing with. In these cases, the exception should be caught by the first
   * caller who _does_ know the package name, labeled, and optionally rethrown.
   */
  void setPackageName (String name)
  {
    m_packageName = name;
  }

  /**
   * Detail is in addition to the 'message' in Throwable base class.
   */
  void setDetail (String detail)
  {
    m_strDetail = detail;
  }

  /**
   * Detail is in addition to the 'message' in Throwable base class.
   */
  String getDetail ()
  {
    return m_strDetail;
  }

  /**
   * Constructs an exception with a brief message
   *
   * @param message the message passed to the Exception constructor
   */
  public PackageException (String message)
  {
    this (null, message);
  }

  /**
   * Wraps a Throwable in a PackageException
   *
   * @param t the original Throwable
   */
  public PackageException (Throwable t)
  {
    this (t, null);
  }

  /**
   * Wraps a Throwable in a PackageException and adds a short message
   *
   * @param t the original Throwable
   * @param message the message passed to the Exception constructor
   */
  public PackageException (Throwable t, String message)
  {
    this (t, message, null);
  }

  /**
   * Wraps a Throwable in a PackageException, adds a short message,
   * and also a more detailed description
   *
   * @param t the original Throwable
   * @param message the message passed to the Exception constructor
   * @param detail a more detailed description of the error
   */
  public PackageException (Throwable t, String message, String detail)
  {
    super (message);
    m_origT = t;
    m_strDetail = detail;
  }

  public void printStackTrace ()
  {
    printStackTrace(System.err);
  }

  public void printStackTrace (PrintStream s)
  {
    printStackTrace (new PrintWriter (s));
  }

  public void printStackTrace (PrintWriter pw)
  {
    if (m_packageName != null)
      {
	pw.println ("Exception processing package " + m_packageName);
      }
    super.printStackTrace (pw);
    if (m_origT != null)
      {
	pw.println ("Original exception/throwable:");
	m_origT.printStackTrace (pw);
      }
    pw.flush ();
  }

  public String toString ()
  {
    String ret = super.toString();
    
    if (m_packageName != null)
      {
	ret = "(" + m_packageName + "): " + ret;
      }

    if (m_strDetail != null)
      {
	ret += "\n" + m_strDetail;
      }

    return ret;
  }
}
