/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageManager.java,v 1.9 2001/11/15 20:53:18 tim Exp $
 */
package org.xoe.core.packages;

import java.io.IOException;
import java.lang.ref.SoftReference;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

import org.xoe.core.content.StashException;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;

/**
 * Pure singleton (all-static) class for managing packages both
 * on and off the system.
 */
public class PackageManager
{
  public static long batchInstallTimer = 0;
  // the local package database
  private static LocalPackageDatabase m_pdb = new LocalPackageDatabase ();
  // the packages servers to use when searching
  private static Vector m_vPackageServers = new Vector ();
  // Hashtable of SoftRefs to the actual XoePackages
  private static Hashtable m_xoePackageHash = new Hashtable ();

  // All PackageManager methods are static
  private PackageManager ()
  {
  }

  public static void addPackageServer (String type, String url)
  {
    // we know how to make SimplePackageQueryHandler without a factory
    if (type.equals ("SIMPLE"))
      {
        try
          {
            m_vPackageServers.addElement (new StashPackageQueryHandler (new URL (url)));
          }
        catch (StashException e)
          {
            Logger.log (PackageManager.class, e);
          }
        catch (MalformedURLException e)
          {
            Logger.log (PackageManager.class, e);
          }
        return;
      }

    // otherwise look for a factory
    SimpleServiceQuery q = SimpleServiceQuery.alloc (IPackageServerFactory.FUNCTIONALITY, IPackageServerFactory.INTERFACE);

    q.addFeature (IPackageServerFactory.TYPE_FEATURE, type);

    IPackageServerFactory factory = (IPackageServerFactory) ServiceLocator.findService (q);
    SimpleServiceQuery.free (q);

    if (factory == null)
      {
        // the default factory will look for PackageQueryHandlers at run time
	m_vPackageServers.addElement (new DefaultPackageServer (type, url));
	Logger.log(PackageManager.class, "added server with unknown type: {type='" + type + "', url='" + url + "'}");
      }
    else
      {
        // the factory is already registered in the system so just use that.
	m_vPackageServers.addElement (factory.getServer(type, url));
	Logger.log(PackageManager.class, "added server with known type: {type='" + type + "', url='" + url + "'}");
      }
  }

  public static PackageDescription getDescription (String name)
  {
    Vector v = new Vector ();
    performQuery ((new PackageQuery ()).setName (name), v);
    // newest returns null if there is nothing in the vector
    return PackageDescription.newest (v);
  }

  public static PackageDescription getDescription (String name, String version)
  {
    Vector v = new Vector ();
    performQuery ((new PackageQuery ()).setName (name).setVersion (version), v);
    // newest returns null if there is nothing in the vector
    return PackageDescription.newest (v);
  }
  
  public static void batchInstall (String[] names)
    throws PackageException
  {
    long stime = System.currentTimeMillis ();
    Vector v = InstallContext.getBatchContexts (names);
    if (v == null || v.size() == 0)
      {
	throw new InstallException ("failed to process batch install: " +
				    "not all packages were available.");
      }
    PackageActionSet pss = new PackageActionSet (m_pdb.toActionSet (), 
						 (InstallContext) v.elementAt (0));
    // checks to see if there are no internal conflicts
    if (!pss.isComplete ())
      {
	throw new InstallException ("failed to process batch install.");
      }

    batchInstallTimer += System.currentTimeMillis () - stime;
    commit (new InstallOption (((InstallContext) v.elementAt (0)).getPrimary (), pss));
  }

  public static SearchResults search (IPackageQuery in_q)
  {
    Vector queryResults = new Vector ();

    Vector installed = new Vector ();
    Vector installable = new Vector ();
    Vector uninstallable = new Vector ();

    // true is for "search remotely, even if something local matches the query"
    performQuery (in_q, queryResults, true);

    for (int i = 0; i < queryResults.size (); i++)
      {
	if (m_pdb.isInstalled ((PackageDescription) queryResults.elementAt (i)))
	  {
	    installed.addElement (queryResults.elementAt (i));
	  }
	else
	  {
	    Vector candidates = InstallContext.getCandidateContexts ((PackageDescription)queryResults.elementAt(i));

	    if (candidates.size () == 0)
	      {
		uninstallable.addElement (InstallContext.getFailure ());
	      }

	    for (int j = 0; j < candidates.size (); j++)
	      {
		PackageActionSet pas = new PackageActionSet (m_pdb.toActionSet (), 
							    (InstallContext) candidates.elementAt (j));
		
		// checks to see if there are no internal conflicts
		try
		  {
		    // throws
		    pas.checkComplete();
		    installable.addElement (new InstallOption
					    (((InstallContext) candidates.elementAt (j)).getPrimary (), pas));
		  }
		catch(PackageIncompleteException ex)
		  {
		    uninstallable.addElement (new UninstallablePackage ((PackageDescription)queryResults.elementAt(i), ex));
		  }
	      }
	  }
      }
    InstallOption[] installableA = (InstallOption[]) installable.toArray (new InstallOption[installable.size()]);
    UninstallablePackage[] uninstallableA = (UninstallablePackage[]) uninstallable.toArray (new UninstallablePackage[uninstallable.size()]);
    PackageDescription[] installedA = (PackageDescription[]) installed.toArray (new PackageDescription[installed.size()]);
    return new SearchResults (installableA, uninstallableA, installedA);
  }

  public static InstallOption getUninstallOption (PackageDescription pkg)
  {
    UninstallContext ctx = UninstallContext.getCandidateContext (pkg);

    //This line can potentially build PackageActionSets that contain conflicts
    PackageActionSet pas = new PackageActionSet (m_pdb.toActionSet (), ctx);

    return new InstallOption (pkg, pas);
  }

  public static PackageDescription[] getInstalledPackages ()
  {
    return m_pdb.getInstalledPackages ();
  }

  public static void commit (InstallOption opt)
    throws PackageException
  {
    // PackageManager.class could throw for 153 different reasons... can't find a .xoe file,
    // fail to install, etc..., etc...
    m_pdb.commit (opt.getActionSet());
    
    forgetPackages ();
    // might also launch up the "primary" app after install?
  }

  public static void installPackage (URL url) 
    throws IOException, PackageException
  {
    installPackage (new XoePackage (url));
  }

  public static void installPackage (XoePackage pkg)
    throws PackageException
  { 
    // this will be retrieved later if the package is actually installed
    m_xoePackageHash.put (pkg.getPackageURI (), new SoftReference (pkg));    

    PackageActionSet pas = m_pdb.toActionSet ();
    pas.selectPackage (pkg.getDescription ());

    // will throw if a package can't be retrieved, conflicts with the current
    // system, etc...
    m_pdb.commit (pas);
  }

  public static PackageActionSet getActionSet ()
  {
    return m_pdb.toActionSet ();
  }

  /**
   * First searches locally, then remotely.  
   * This is a utilty method -- use wisely.
   * This returns all matching packages, regardless of conflicts;
   * Therefore, these pacages may not be installable, due to conflicts
   */
  public static void performQuery (IPackageQuery in_q, Vector out_res)
  {
    performQuery (in_q, out_res, false);
  }

  public static void performQueryLocally (IPackageQuery in_q, Vector out_res)
  {
    m_pdb.performQuery (in_q, out_res);
  }

  public static void performQueryRemotely (IPackageQuery in_q, Vector out_res)
  {
    // don't need or want multiple references to the same pacakage
    // (ie, if it is listed on multiple servers...)
    // hash them by getPackageURI (a unique key of name+version)
    // this means that if the same description is found twice,
    // the second reference will replace the first in the hash map

    Hashtable thash = new Hashtable ();

    for (int i = 0; i < m_vPackageServers.size (); i++)
      {
	Vector tvec = new Vector ();
	((IPackageQueryHandler) m_vPackageServers.get (i)).performQuery (in_q, tvec);
	for (int j = 0; j < tvec.size (); j++)
	  {
	    thash.put (((PackageDescription) tvec.elementAt (j)).getPackageURI (),
	      	       ((PackageDescription) tvec.elementAt (j)));
	  }
      }

    for (Enumeration e = thash.elements (); e.hasMoreElements (); )
      {
	out_res.addElement (e.nextElement ());
      }
  }

  public static void performQuery (IPackageQuery in_q, Vector out_res, boolean all)
  {
    Vector tvec = new Vector ();

    m_pdb.performQuery (in_q, tvec,
		        all ? m_pdb.PDB_ALL_PACKAGES : m_pdb.PDB_INSTALLED_PACKAGES);
    if (!all && tvec.size () > 0)
      {// don't search remotely if we can find something already installed
	for (int i = 0; i < tvec.size (); i++)
	  {
	    out_res.addElement (tvec.elementAt (i));
	  }
	return;
      }

    // don't need or want multiple references to the same pacakage
    // (ie, if it is listed on multiple servers...)
    // hash them by getXoeFilename (a unique key of name+version)
    // this means that if the same description is found twice,
    // the second reference will replace the first in the hash map

    Hashtable thash = new Hashtable ();

    for (int j = 0; j < tvec.size (); j++)
      {
	thash.put (((PackageDescription) tvec.elementAt (j)).getPackageURI (),
		   ((PackageDescription) tvec.elementAt (j)));
      }

    for (int i = 0; i < m_vPackageServers.size (); i++)
      {
	tvec = new Vector ();
	((IPackageQueryHandler)m_vPackageServers.get (i)).performQuery (in_q, tvec);
	for(int j = 0; j < tvec.size(); j++)
	  {
	    thash.put (((PackageDescription) tvec.elementAt (j)).getPackageURI (),
		       ((PackageDescription) tvec.elementAt (j)));
	  }
      }

    for (Enumeration e = thash.elements (); e.hasMoreElements (); )
      {
	out_res.addElement (e.nextElement ());
      }
  }

  public static XoePackage retrievePackageLocally (PackageDescription desc)
  {
    XoePackage pkg = m_pdb.getXoePackage (desc.getPackageURI ());
    if (pkg != null)
      {
	return pkg;
      }

    SoftReference ref = (SoftReference) m_xoePackageHash.get (desc.getPackageURI());

    return ref == null ? null : (XoePackage) ref.get ();
  }
  
  public static void forgetPackages ()
  {
    Enumeration e = m_xoePackageHash.elements();
    
    while (e.hasMoreElements())
      {
	SoftReference ref = (SoftReference)e.nextElement();
	XoePackage pkg = (XoePackage)ref.get ();
	if (pkg != null)
	  {
	    pkg.forgetArchive ();
	  }
      }
    m_xoePackageHash = new Hashtable();
  }

  public static XoePackage retrievePackage (PackageDescription desc)
    throws PackageNotFoundException
  {
    Vector urls = new Vector ();

    XoePackage pkg = retrievePackageLocally (desc);
    
    if (pkg != null)
      {
	return pkg;
      }

    URL suggestion = desc.getSuggestedURL ();
    if (suggestion != null)
      {
	urls.addElement (suggestion);
      }

    /*
    for(int i = 0; i < m_vPAS.size(); i++)
      {
	try
	  {
	    urls.addElement(new URL((URL)m_vPAS.elementAt(i), desc.getPackageURI() ));
	  }
	catch(MalformedURLException ex)
	  { // do nothing...
	    Logger.log(PackageManager.class, ex);
	  }
      }
    */

    for (int i = 0; i < urls.size (); i++)
      {
	try 
	  {
	    pkg = new XoePackage (((URL) urls.elementAt (i)).openStream ());
	  }
	catch (IOException e)
	  {
	    continue;
	  }

	if (pkg != null)
	  {
	    m_xoePackageHash.put (pkg.getPackageURI (), new SoftReference (pkg));
	    return pkg;
	  }
      }
    
    throw new PackageNotFoundException (desc, urls);
  }
}
