/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageQuery.java,v 1.4 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.NoSuchElementException;
import java.util.Vector;
import java.util.StringTokenizer;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.DefaultEntityResolver;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;

/**
 * Class used to build most common types of package queries
 */
public class PackageQuery 
  implements IPackageQuery
{
  // A prebuilt XPath, or we already built our xpath...
  private String m_xpath = null;

  // Looking for keywords
  private Vector m_keywords = null;
  // whether kw's can also be found in the description or summary
  private boolean m_kweverywhere = false;
  // whether kw's are case sensitive (by default, not)
  private boolean m_kwcase = false;

  // whether we will only search for "task" packages
  // this flag will be set by the PkgMgr gui in "basic" mode
  // (these are user level packages, as opposed to small libraries that don't really
  //  make sense by themselves and might confuse the user)
  private boolean m_task = false;

  // Looking for a specific package
  private String m_pkgName = null;
  private String m_version = null;
  private String m_minVersion = null;
  private boolean m_minInclusive = true;
  private String m_maxVersion = null;
  private boolean m_maxInclusive = true;

  // Looking for a generic contentHandler, protocol handler, installer...
  private String m_functionality = null;
  private String m_interface = null;
  private Vector m_vFeatures = null;

  private Vector m_deps = new Vector ();

  public PackageQuery ()
  {
  }

  /**
   * Builds a package query out of a <code>&lt;dep /&gt;</code> element
   *
   * @exception PackageDescriptionElement if the element is not a valid
   *                                      dependency
   */
  public PackageQuery (Element elem)
    throws PackageDescriptionException
  {
    if (elem.getTagName ().equals ("dep"))
      {
	addDependency (Dependency.create (elem));
      }
    else
      {
	throw new PackageDescriptionException("Not a valid dependency:\n" + XMLWriter.getXML(elem));
      }
  }

  /**
   * Builds a <code>PackageQuery</code> out of an XPath query string.
   */
  public PackageQuery setXPath (String xpath)
  {
    if(xpath != null && xpath.length () == 0)
      {
	xpath = null;
      }

    m_xpath = xpath;
    return this;
  }

  /**
   * Add a single keyword to the query. Keywords are compared against
   * the value of the <code>keywords</code> field in a package description.
   * Use the <code>setKeywordsSearchEverywhere</code> method to expand the
   * list of fields searched. Matches are case-insensitive by default -
   * use <code>setKeywordsCaseSensitive</code> to change this setting.
   *
   * @param keyword A single keyword. Whitespace is considered part of the
   *                keyword.
   */
  public PackageQuery addKeyword (String keyword)
  {
    if(m_keywords == null)
      {
	m_keywords = new Vector();
      }
    m_keywords.addElement(DefaultEntityResolver.deEntify (keyword));
    m_xpath = null;
    return this;
  }

  /**
   * Add one or more keywords to the query. Keywords are compared against
   * the value of the <code>keywords</code> field in a package description.
   * Use the <code>setKeywordsSearchEverywhere</code> method to expand the
   * list of fields searched. Matches are case-insensitive by default -
   * use <code>setKeywordsCaseSensitive</code> to change this setting.
   *
   * @param keywords One or more keywords separated by whitespace. Phrases
   *                 with embedded whitespace can be surrounded by single or
   *                 double quotes. Phrases with embedded single quotes may
   *                 be surrounded by double quotes, etc...
   */
  public PackageQuery addKeywords (String keywords)
  {
    if (keywords == null || keywords.length() == 0)
      {
	return this;
      }

    StringTokenizer st = new StringTokenizer(keywords, " \t\n\r\f,\"'", true);
    try
      {
	do
	  {
	    String tok = st.nextToken();

	    if(tok.equals("\"") || tok.equals("'"))
	      {
		// will be the quoted-string
		// pass in the previous tok (either \" or \') to match the end quote
		// (ie, you can query "Mike's" or '"so-called"', and the embedded quotes
		// are preserved
		addKeyword(st.nextToken(tok));

		// will be the close-quote
		st.nextToken(" \t\n\r\f,\"'");
	      }
	    else if((" \t\n\r\f,").indexOf(tok.charAt(0)) > -1)
	      {// don't add the whitespace
	      }
	    else
	      {
		addKeyword (tok);
	      }
	  }while(true);
      }
    catch (NoSuchElementException ex)
      { // reached the end of the string
      }

    return this;
  }

  /**
   * Determines the fields in which a package query searches for keywords.
   * By default, only search the keywords field.
   *
   * @param kweverywhere If false, search only the keywords field. If true,
   *                     search description, summary, and name fields as well.
   *           
   */
  public PackageQuery setKeywordsSearchEverywhere (boolean kweverywhere)
  {
    m_kweverywhere = kweverywhere;
    m_xpath = null;
    return this;
  }

  /**
   * Determines whether keyword queries are case sensitive. By default,
   * they are case insensitive.
   *
   * @param kwcase whether or not keyword queries are case sensitive
   */
  public PackageQuery setKeywordsCaseSensitive (boolean kwcase)
  {
    m_kwcase = kwcase;
    m_xpath = null;
    return this;
  }

  /**
   * "Task" packages are high level packages that are more meaningful to the
   * user than others. Typically, in 'basic' searches, only "task" packages
   * should be displayed (taskmode=true). The default is false (show all
   * matching packages).
   *
   * @param task whether to limit searches to task packages
   */
  public PackageQuery setTaskMode (boolean task)
  {
    m_task = task;
    m_xpath = null;
    return this;
  }

  /**
   * Sets a substring to be found in the name of any matching package
   */
  public PackageQuery setName (String name)
  {
    if (name != null && name.length() == 0)
      {
	name = null;
      }
    m_pkgName = name;
    m_xpath = null;
    return this;
  }

  /**
   * Sets the exact version required of any matching package
   */
  public PackageQuery setVersion (PackageVersion version)
  {
    return setVersion (version.toString ());
  }

  /**
   * Sets the exact version required of any matching package
   */
  public PackageQuery setVersion (String version)
  {
    if(version != null && version.length () == 0)
      {
	version = null;
      }
    m_version = version;
    m_minVersion = null;
    m_maxVersion = null;
    m_xpath = null;
    return this;
  }

  /**
   * Sets the inclusive minimum version required of any matching package.
   * Use setMinVersion (PackageVersion, boolean) to set an exclusive minimum.
   */
  public PackageQuery setMinVersion (PackageVersion minVersion)
  {
    return setMinVersion (minVersion.toString ());
  }

  /**
   * Sets the inclusive minimum version required of any matching package.
   * Use setMinVersion (String, boolean) to set an exclusive minimum.
   */
  public PackageQuery setMinVersion (String minVersion)
  {
    // by default, inclusive
    return setMinVersion (minVersion, true);
  }

  /**
   * Sets the minimum version required of any matching package
   *
   * @param inclusive whether the minimum version is inclusive
   */
  public PackageQuery setMinVersion (PackageVersion minVersion, boolean inclusive)
  {
    return setMinVersion (minVersion.toString(), inclusive);
  }

  /**
   * Sets the minimum version required of any matching package
   *
   * @param inclusive whether the minimum version is inclusive
   */
  public PackageQuery setMinVersion (String minVersion, boolean inclusive)
  {
    if (minVersion != null && minVersion.length() == 0)
      {
	minVersion = null;
      }
    m_version = null;
    m_minVersion = minVersion;
    m_minInclusive = inclusive;
    m_xpath = null;
    return this;
  }

  /**
   * Sets the inclusive maximum version required of any matching package.
   * Use setMaxVersion (PackageVersion, boolean) to set an exclusive maximum.
   */
  public PackageQuery setMaxVersion (PackageVersion maxVersion)
  {
    return setMaxVersion (maxVersion.toString ());
  }

  /**
   * Sets the inclusive maximum version required of any matching package.
   * Use setMaxVersion (String, boolean) to set an exclusive maximum.
   */
  public PackageQuery setMaxVersion (String maxVersion)
  {
    // by default inclusive
    return setMaxVersion (maxVersion, true);
  }

  /**
   * Sets the maximum version required of any matching package
   *
   * @param inclusive whether the maximum version is inclusive
   */
  public PackageQuery setMaxVersion (PackageVersion maxVersion, boolean inclusive)
  {
    return setMaxVersion (maxVersion.toString ());
  }

  /**
   * Sets the maximum version required of any matching package
   *
   * @param inclusive whether the maximum version is inclusive
   */
  public PackageQuery setMaxVersion (String maxVersion, boolean inclusive)
  {
    if (maxVersion != null && maxVersion.length () == 0)
      {
	maxVersion = null;
      }
    m_version = null;
    m_maxVersion = maxVersion;
    m_maxInclusive = inclusive;
    m_xpath = null;
    return this;
  }

  /**
   * Requires any matching package to satisfy the specified dependency
   */
  public PackageQuery addDependency (Dependency c)
  {
    m_deps.addElement (c);
    m_xpath = null;
    return this;
  }

  /**
   * Remove a dependency from the list of required dependencies
   */
  public PackageQuery removeDependency (Dependency c)
  {
    m_deps.removeElement (c);
    m_xpath = null;
    return this;
  }

  /**
   * Helper function that adds a Dependency on a service with the
   * specified functionality. Multiple calls to <code>setFunctionality</code>,
   * <code>setInterface</code>, and <code>addFeature</code> do not add additional
   * dependencies but rather modify the same one.
   */
  public PackageQuery setFunctionality (String func)
  {
    if (func != null && func.length () == 0)
      {
	func = null;
      }
    m_functionality = func;
    m_xpath = null;
    return this;
  }

  /**
   * Helper function that adds a Dependency on a service with the
   * specified functionality. Multiple calls to <code>setFunctionality</code>,
   * <code>setInterface</code>, and <code>addFeature</code> do not add additional
   * dependencies but rather modify the same one.
   */
  public PackageQuery setInterface (String iface)
  {
    if (iface != null && iface.length () == 0)
      {
	iface = null;
      }
    m_interface = iface;
    m_xpath = null;
    return this;
  }

  /**
   * Helper function that adds a Dependency on a service with the
   * specified functionality. Multiple calls to <code>setFunctionality</code>,
   * <code>setInterface</code>, and <code>addFeature</code> do not add additional
   * dependencies but rather modify the same one.
   */
  public PackageQuery addFeature (String name, String value)
  {
    if (name == null || value == null)
      {
	return this;
      }

    if (m_vFeatures == null)
      {
	m_vFeatures = new Vector();
      }
    String[] el = {name, value};
    m_vFeatures.addElement(el);

    m_xpath = null;
    return this;
  }

  public String toXPath ()
  {
    if (m_xpath != null)
      {
	return m_xpath;
      }

    // Transform a single Service dependency into a Dependency object
    Dependency serviceDep = null;
    if (m_functionality != null || m_interface != null || m_vFeatures != null)
      {
	String[][] features = null;
	if (m_vFeatures != null)
	  {
	    features = (String[][])m_vFeatures.toArray (new String[m_vFeatures.size()][]);
	  }

	serviceDep = new ServiceDependency (m_functionality, m_interface, features);
	addDependency (serviceDep);
      }

    Dependency pkgDep = null;

    if (m_pkgName != null)
      {
	if (m_version != null)
	  {
	    pkgDep = new PackageDependency (m_pkgName, m_version);
	    addDependency (pkgDep);
	  }
	else
	  {
	    pkgDep = new PackageDependency (m_pkgName, m_minVersion, m_minInclusive, m_maxVersion, m_maxInclusive);
	    addDependency (pkgDep);
	  }
      }
    
    StringBuffer ret = new StringBuffer ("package");

    if (m_task != false)
      {
	ret.append ("[@task='true']");
      }

    // Process keywords
    if (m_keywords != null)
      {
	for (int i = 0; i < m_keywords.size (); i++)
	  {
	    ret.append ("[contains(");

	    if (!m_kwcase)
	      { // case insensitive
		ret.append ("translate(");
	      }

	    if (m_kweverywhere)
	      { // concatenate these four fields to search in
		ret.append ("concat(keywords, description, summary, @name)");
	      }
	    else
	      { // only search in keywords
		ret.append ("keywords");
	      }

	    if (!m_kwcase)
	      { // case insensitive
		ret.append (", 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', 'abcdefghijklmnopqrstuvwxyz')");
		ret.append (", '" + ((String)m_keywords.elementAt(i)).toLowerCase() + "')]");
	      }
	    else
	      { // case sensitive
		ret.append (", '" + m_keywords.elementAt(i) + "')]");
	      }
	  }
      }

    // Process dependencies
    if (m_deps.size () != 0)
      {
	ret.append ("[provides");
	for (int i = 0; i < m_deps.size (); i++)
	  {
	    Dependency dep = (Dependency) m_deps.elementAt (i);
	    ret.append("[proviso");
	    
            if (dep.getNamespaceURI () != null)
              {
                ret.append ("[@ns = '" + dep.getNamespaceURI () + "']");
              }
            if (dep.getLocalName () != null)
              {
                ret.append ("[@name = '" + dep.getLocalName () + "']");
              }
            if (dep.getXPath () != null)
              {
                ret.append ("[" + dep.getXPath () + "]");
              }

	    ret.append ("]");
	  }
	ret.append ("]");
      }
    // Make sure to recalc the "dynamic" dependencies next time we're called
    if (serviceDep != null)
      {
        removeDependency (serviceDep);
      }
    if (pkgDep != null)
      {
        removeDependency (pkgDep);
      }

    m_xpath = ret.toString();
    return m_xpath;
  }

  public String toString ()
  {
    return toXPath ();
  }

  // this is here for the benefit of serializing this object via soap
  public Vector getKeywords ()
  {
    return m_keywords;
  }

  public void setKeywords (Vector vec)
  {
    m_keywords = vec;
  }

  public String getFunctionality ()
  {
    return m_functionality;
  }

  public String getInterface ()
  {
    return m_interface;
  }

  public void setFeatures (Vector vec)
  {
    m_vFeatures = vec;
  }

  public Vector getFeatures ()
  {
    return m_vFeatures;
  }

  public void setPkgName (String name)
  {
    m_pkgName = name;
  }

  public String getPkgName ()
  {
    return m_pkgName;
  }

  public void setVer (String name)
  {
    m_version = name;
  }

  public String getVer ()
  {
    return m_version;
  }

  public void setMinVer (String name)
  {
    m_minVersion = name;
  }
  public String getMinVer()
  {
    return m_minVersion;
  }

  public void setMaxVer (String name)
  {
    m_maxVersion = name;
  }

  public String getMaxVer ()
  {
    return m_maxVersion;
  }
}
