/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: ResourceInstaller.java,v 1.8 2001/12/18 06:05:50 colin Exp $
 */
package org.xoe.core.packages;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.DefaultContentTypeResolver;
import org.xoe.core.content.Stash;
import org.xoe.core.content.StashException;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


public class ResourceInstaller
  extends FileIteratingInstaller
{
  public static final String NAMESPACE = "http://xoe.org/installer/resource";
  public static final String SHORTCUTS_FILENAME = "shortcuts.xml";
  private static final String LOCALIZEBASE = "localize/";
  private static final String MIMETYPES_FILENAME = "mimetypes.xml";
  private static final String RESOURCES_PATHPART = "resources/";

  public String[] getHandlerPredependencies ()
  {
    return new String[0] ;
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  public Dependency buildDependency ( Element depEl )
    throws PackageDescriptionException
  {
    throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML(depEl));
  }

  public void processFile (ZipInputStream zis, ZipEntry entry)
  {
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  { // Could check to see if there is enough space for all the files...
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    ResourceInstallerData data = null;

    String pkgDir = xoePackage.getPackageURI ();
    if (!pkgDir.endsWith (File.separator))
      {
        pkgDir += File.separator;
      }

    Hashtable contents = new Hashtable ();

    ZipInputStream zis = xoePackage.getZipInputStream ();
    URL pkgBase;

    try
      {
        // the stash url where this packages files will be located
        pkgBase = new URL (Stash.getPkgArchiveRoot (), pkgDir);
	
	data = new ResourceInstallerData (pkgBase);
	xoePackage.setExData (NAMESPACE, data);

        while (true)
          {
            ZipEntry entry = zis.getNextEntry ();
            if (entry == null)
              {
                break;
              }

	    String filename = entry.getName ();

	    // only files in 'resources' go into the stash
	    // (BUT remember to ignore the folder itself!)
	    if (filename.startsWith ( RESOURCES_PATHPART ) && (!filename.equals (RESOURCES_PATHPART)))
	      {
		// chop off 'resources/'
		filename = filename.substring (10);

		// the stash url of this particular file
		URL entryURL = new URL (pkgBase, filename);
		
		// get the content
		try
		  {
		    ContentElement content = Stash.getFactory ().createContentElement (entryURL, zis, (int)entry.getSize ());
		    content.setNodeName (filename);
		    // store it in the hash for further processing
		    contents.put (filename, content);
		  }
		catch (IOException ioe)
		  {
		    throw new InstallException (ioe, xoePackage,
						("An error occurred while installing the file " +
						 filename));
		  }
	      }
	    zis.closeEntry ();
          }
        zis.close ();

        // process the content elements from the xoe package
	processContents (contents, pkgBase.toString (), xoePackage);
      }
    catch (IOException ioe)
      {
	throw new InstallException (ioe, xoePackage,
				    "An error occurred while installing package resource files.");
      }

    // Install shortcuts
    try
      {
	ContentElement ce = Stash.fetch (new URL (pkgBase, SHORTCUTS_FILENAME));
	if (ce != null)
	  {
	    Document doc = DOMlet.create (ce.getInputStream (), SHORTCUTS_FILENAME);

	    NodeList list = XPath.findNodes (doc, "shortcuts/lnk");
	    int len = list.getLength ();
	    for (int i = 0; i < len; i++) 
	      {
		Element elem = (Element) list.item (i);
		try
		  {
		    installShortcut (data, elem.getAttribute ("file"),
				     elem.getAttribute ("shortcut"));
		  }
		catch (IOException ex)
		  {
		    Logger.log (this, "failed to install shortcut: " + elem.getAttribute ("shortcut"));
		    throw new InstallException (ex, xoePackage,
						"failed to install shortcut: \"" + elem.getAttribute ("shortcut") + "\"");
		  }
	      }
	  }
      }
    catch (IOException ex)
      {
	throw new InstallException (ex, xoePackage,
				    ("An error occurred while searching for " +
				     SHORTCUTS_FILENAME));
      }
  }

  private void processContents (Hashtable contents, String pkgBaseSrc, XoePackage pkg)
    throws IOException
  {
    Locale locale = Locale.getDefault ();
    String language = locale.getLanguage ();

    if (language != null) 
      {
	language = language.toLowerCase ();
      }
    String country = locale.getCountry ();
    if (country != null) 
      {
	country = country.toUpperCase ();
      }

    for (Enumeration k = contents.keys (); k.hasMoreElements(); ) 
      {
	String filename = (String)k.nextElement ();
	ContentElement content  = (ContentElement)contents.get (filename);
	if (filename.startsWith (LOCALIZEBASE))
          {
	    // Ignore classes, localized directory and content with no mimetype
	  }
	else 
          {
	    // ok, save the originally read content
	    String nonLocalizedSrc = content.getURL ().toString ();

	    // get the potentially localized version
	    content = getLocalizedContent (filename, contents, language, country);

	    // add the content element
	    Stash.put (content);

	    // check to see if we need to alias this content element
	    if (!nonLocalizedSrc.equals (content.getURL ().toString ()))
	      {
		// add the alias
		String alias = pkgBaseSrc + filename;
		Stash.addAlias (alias, content);
              }
	  }
      }
  }

  private ContentElement getLocalizedContent (String key, Hashtable contents, String lang, String country) 
  {
    String bkey = key;
    String ekey = "";
    int idx = bkey.lastIndexOf ('.');
    if (idx > -1) 
      {
	bkey = key.substring (0, idx);
	ekey = key.substring (idx);
      }

    Object content = null;
    if (lang != null && country != null) 
      {
	content = contents.get (LOCALIZEBASE + bkey + '_' + lang + '_' + country + ekey);
      }
    if (content == null && lang != null) 
      {
	content = contents.get (LOCALIZEBASE + bkey + '_' + lang + ekey);
      }
    if (content == null) 
      {
	content = contents.get (key);
      }
    return ((ContentElement)content);
  }

  private void installShortcut (ResourceInstallerData data, String file, String shortcut)
    throws IOException
  {
    URL surl = new URL(new URL(STASH_PROTOCOL + ":/home/default/apps/"), shortcut + ".lnk");
    ContentElement ce = Stash.getFactory().createContentElement(surl, ("<lnk src='" + new URL(data.getWorkingDirectory(), file) + "'/>").getBytes());

    Stash.put(ce);

    data.addShortcut(ce.getURL());
  }


  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    ResourceInstallerData data = (ResourceInstallerData)xoePackage.getExData (NAMESPACE);
    URL wd = data.getWorkingDirectory ();
    // delete the package files
    try
      {
	Stash.remove (wd);
      }
    catch (StashException ex)
      {
	Logger.log (this, "Ignoring failure while deleting package directory for " + xoePackage);
	Logger.log (this, ex);
      }
	

    // remove mimetype mappings added by this package
    DefaultContentTypeResolver dctr = DefaultContentTypeResolver.getInstance ();
    dctr.removeProvider (xoePackage);

    URL urls[] = data.getShortcuts();
    for(int i = 0; i < urls.length; i++)
      {
	try
	  {
	    ContentElement ce = Stash.fetch (urls[i]);
	    Document doc = DOMlet.create (ce.getInputStream (), urls[i]);
	    if (doc.getDocumentElement ().getAttribute ("src").startsWith (wd.toString()))
	      {
		Stash.remove ( urls[i] );
	      }
	  }
	catch (StashException ex)
	  {
	    Logger.log (this, "Ignoring failure while removing shortcut for " + xoePackage);
	    Logger.log (this, ex);
	  }
	catch (IOException ex)
	  {
	    Logger.log (this, "Ignoring failure while removing shortcut for " + xoePackage);
	    Logger.log (this, ex);
	  }
      }

    xoePackage.setExData (NAMESPACE, null);
  }

  private void installMapping (XoePackage xoePackage, String mimetype, String suffix)
  {
    DefaultContentTypeResolver dctr = DefaultContentTypeResolver.getInstance ();
    dctr.putMapping (suffix, mimetype, xoePackage);
  }

}
