/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: ServiceInstaller.java,v 1.9 2001/12/04 21:39:02 erik Exp $
 */
package org.xoe.core.packages;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.net.URL;
import java.util.Vector;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceInstanceElement;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


/**
 * Installer for services.  Since services are classes we extend
 * the ClassIteratingInstaller.
 */
public class ServiceInstaller 
  implements IPackageXMLHandler
{
  public static class ServiceInstallerData
  {
    private IService[] m_installedServices = null;
    
    private ServiceInstallerData (IService[] installedServices)
    {
      m_installedServices = installedServices;
    }

    public IService[] getInstalledServices ()
    {
      return m_installedServices;
    }
  }

  public static final String SERVICES_FILENAME = "services.xml";
  public static final String NAMESPACE = "http://www.xoe.org/installer/service";

  XoePackage m_package;

  private static Document m_doc = new XDocument ();

  public String[] getHandlerPredependencies ()
  {
    return new String[] { ClassInstaller.NAMESPACE,
			    ResourceInstaller.NAMESPACE,
			    NativeInstaller.NAMESPACE};
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    return new String[] { IPackageXMLHandler.FUNCTIONALITY };
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { IPackageXMLHandler.INTERFACE };
      }
    return null;
  }
  
  public String[] getFeatures (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { FEATURE_NAMESPACE, getNamespaceURI() };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    return new ServiceDependency (depEl);
  }

  /**
   * Helper function to build service provisios.
   */
  Proviso getProviso (String cname, String func, String[] ifaces, String[] features)
  {
    DocumentFragment frag = m_doc.createDocumentFragment ();

    Element funcE = m_doc.createElement ("functionality");
    funcE.appendChild (m_doc.createTextNode (func));
    frag.appendChild (funcE);

    Element classE = m_doc.createElement ("class");
    classE.appendChild (m_doc.createTextNode (cname));
    frag.appendChild (classE);

    if (ifaces != null)
      {
	for (int i = 0; i < ifaces.length; i++)
	  {
	    Element impl = m_doc.createElement ("implements");
	    Text t = m_doc.createTextNode (ifaces[i]);
	    impl.appendChild (t);
	    frag.appendChild (impl);
	  }
      }

    if (features != null)
      {
	for (int i = 0; i < features.length; i += 2)
	  {
	    Element feat = m_doc.createElement ("feature");
	    feat.setAttribute ("name", features[i]);
	    feat.setAttribute ("value", features[i + 1]);
	    frag.appendChild (feat);
	  }
      }

    return new Proviso (getNamespaceURI (), func, frag);
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  { // could maybe check to see if all referenced classes have
    // corresponding .class files...
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    Vector providers = new Vector();
    try
      {
	ContentElement ce = Stash.fetch (new URL (Stash.getWorkingDirectory (xoePackage), SERVICES_FILENAME));

	if (ce != null)
	  {
	    Document doc = DOMlet.create (ce.getInputStream (), SERVICES_FILENAME);
	    
	    NodeList list = XPath.findNodes (doc, "services/service");
	    
	    for (int i = 0; i < list.getLength (); i++) 
	      {
		Element e = (Element) list.item (i);
		providers.addElement (installService (xoePackage, e.getAttribute ("class"), e.getAttribute ("config")));
	      }
	  }
	xoePackage.setExData (NAMESPACE,
			      new ServiceInstallerData ( (IService[])providers.toArray (new IService[providers.size()])));
      }
    catch (Throwable ex)
      {
	// uninstall any successfully installed services
	for (int i = 0; i < providers.size (); i++)
	  {
	    try
	      {
		ServiceLocator.unregisterService ((IService)providers.elementAt (i));
	      }
	    catch (Throwable t)
	      {
		Logger.log (this, "WARNING: While aborting installation of " + xoePackage
			    + " failed to uninstall a service.");
		Logger.log (this, t);
	      }
	  }
	throw new InstallException (ex, xoePackage,
				    ("An error occurred while searching for " +
				     SERVICES_FILENAME));
      }
  }

  private IService installService (XoePackage xoePackage, String cname, String configpath)
    throws InstallException
  {
    try 
      {
	Class clazz = ((ClassInstallerData) xoePackage.getExData (ClassInstaller.NAMESPACE)).getClassLoader ().loadClass (cname);
        // [FIXME] if the pacakge dependencies are wrong this call can cause all sorts of strife.
        // There is a bug that can cause it to never return !!
	IService provider =  (IService) clazz.newInstance ();
	Document configdoc = null;

	if (configpath != null)
	  {
	    URL url = new URL (Stash.getWorkingDirectory (xoePackage), configpath);
	    ContentElement ce = Stash.fetch ( url );
	    if (ce == null)
	      {
		throw new InstallException (xoePackage,
					    ("Failed to find service config file:\n" +
					     "Class: " + cname + "\n" +
					     "Config: " + configpath ));
	      }

	    configdoc = DOMlet.create (ce.getInputStream (), url);
	  }

	provider.init (configdoc);

	// Register services as they are loaded
	ServiceInstanceElement sin = ServiceLocator.registerService (provider);
	sin.addInfo ("package", xoePackage.getPackageURI ());
	return provider;
      } 
    catch (InstallException ex)
      {
	throw ex;
      }
    catch (Exception ex)
      {
	throw new InstallException (ex, xoePackage,
				    ("An error occurred while registering a new service:\n" +
				     "Class: " + cname + "\n" +
				     "Config: " + (configpath == null ? "none" : configpath)));
      }
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    IService[] providers = ((ServiceInstallerData)xoePackage.getExData (NAMESPACE)).getInstalledServices ();
    for (int i = 0; i < providers.length; i++)
      {
	ServiceLocator.unregisterService (providers[i]);
      }
    xoePackage.setExData (NAMESPACE, null);
  }

  // ------------------------------------------------------
  // build time functions used to generate package.xml 
  // ------------------------------------------------------
  
  public void setPackage (XoePackage p)
  {
    m_package = p;
  }

  public Vector generateRequires ()
  {
    return null;
  }

  public Vector generateConflicts ()
  {
    return null;
  }

  public Vector generateProvides ()
    throws PackageException
  {
    Vector ret = new Vector();

    // [FIXME]
    // There are two distict ways we must do this depending on how the build process works.
    // If the build process actually installs the packages then we *much* use the ExData to
    // instanciate the IService class.
    // If the build process simply sets the appropriate CLASSPATH then we simple call forName ();


    if ("yes".equals (System.getProperty ("xoe.fastbuild")))
      {
        ZipInputStream zis = m_package.getZipInputStream ();
        ZipEntry entry;
        try
          {
            while ((entry = zis.getNextEntry ()) != null)
              {
                if (entry.getName ().equals ("resources/" + SERVICES_FILENAME))
                  {
                    byte buf[] = new byte[(int) entry.getSize ()];
                    zis.read (buf);
                    Document doc = DOMlet.create (new ByteArrayInputStream(buf), SERVICES_FILENAME);
                    NodeList services = XPath.findNodes (doc, "services/service/@class");
                    for (int i = 0; i < services.getLength (); i++)
                      {
                        String classname = services.item (i).getNodeValue ();
                        Class serviceClass = Class.forName (classname);
                        IService service = (IService) serviceClass.newInstance ();
                        String funcs[] = service.getFunctionalities ();
                        if (funcs != null)
                          {
                            for (int j = 0; j < funcs.length; j++)
                              {
                                String ifaces[] = service.getInterfaces (funcs[j]);
                                ret.addElement (getProviso (classname, funcs[j], ifaces, service.getFeatures (funcs[j])));
                              }
                          }
                      }
                    break;
                  }
              }
          }
        catch (Exception e)
          {
            throw new PackageException (e.toString ());
          }
      }
    else
      {
        ServiceInstallerData data = (ServiceInstallerData) m_package.getExData (NAMESPACE);
        if (data != null)
          {
            IService[] providers = data.getInstalledServices ();
            for(int i = 0; providers != null && i < providers.length; i++)
              {
                IService provider = providers[i];
                String funcs[] = provider.getFunctionalities ();
                if (funcs != null)
                  {
                    for (int j = 0; j < funcs.length; j++)
                      {
                        String ifaces[] = provider.getInterfaces (funcs[j]);
                        ret.addElement (getProviso (provider.getClass().toString(), funcs[j], ifaces, provider.getFeatures (funcs[j])));
                      }
                  }
              }
          }
      }
    return ret;
  }
}
