/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XoePackage.java,v 1.7 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import java.net.URL;
import java.util.Hashtable;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;

/**
 * Represents a XOE package in memory. Used to access the file archive during
 * installation and to store and retrieve references to runtime components
 * (classloaders, services, resources, etc...).
 */
public class XoePackage 
  implements IPackageQueryHandler
{
  private String m_filename = null;
  // binary data copy of the ZipInputStream
  private byte[] streamCopy = null;
  // the package meta-data
  private PackageDescription desc = null;
  // opaque data that each PackageXMLHandler can store with a package
  private Hashtable m_exData = new Hashtable ();

  public String toString ()
  {
    /**
     * Can't always rely on the PackageDescription because we may not
     * have one yet.
     */
    if (desc == null)
      {
	return "{XoePackage: " + m_filename + "}";
      }
    else
      {
        return desc.toString ();
      }
  }

  /**
   * Creates a XoePackage from a local file path
   *
   * @param file a local file path. Behavior is undefined if the path is
   *             relative.
   * @exception InvalidPackageException if the package cannot be parsed or the
   *                                    description is not understood
   * @exception IOException if an IO error occurs while accessing the file
   */
  public XoePackage (String file)
    throws InvalidPackageException, IOException
  {
    // pass file as the name
    this (new ZipInputStream (new FileInputStream (file)), file);
  }

  /**
   * Creates a XoePackage from an URL
   *
   * @param url The URL of a .xoe file
   * @exception InvalidPackageException if the package cannot be parsed or the
   *                                    description is not understood
   * @exception IOException if an IO error occurs while accessing the file
   */
  public XoePackage (URL url) throws InvalidPackageException, IOException
  {
    this (new ZipInputStream (url.openStream ()), url.toString());
  }

  /**
   * Creates a XoePackage from an InputStream
   *
   * @param stream An input stream from which to read a ZIP file
   * @exception InvalidPackageException if the package cannot be parsed or the
   *                                    description is not understood
   * @exception IOException if an IO error occurs while accessing the stream
   */
  public XoePackage (InputStream stream)
    throws InvalidPackageException, IOException
  {
    this (new ZipInputStream (stream));
  }

  /**
   * Creates a XoePackage from an InputStream and specifies the name of the
   * file the stream was created from. The filename is useful if an error
   * occurs while loading the package.
   *
   * @param stream An input stream from which to read a ZIP file
   * @param name A string identifying the resource from which the stream was created.
   * @exception InvalidPackageException if the package cannot be parsed or the
   *                                    description is not understood
   * @exception IOException if an IO error occurs while accessing the stream
   */
  public XoePackage (InputStream stream, String name)
    throws InvalidPackageException, IOException
  {
    this (new ZipInputStream (stream), name);
  }

  /**
   * Creates a XoePackage from a ZipInputStream
   *
   * @param stream A ZipInputStream from which to access the contents of a ZIP Archive
   * @exception InvalidPackageException if the package cannot be parsed or the
   *                                    description is not understood
   * @exception IOException if an IO error occurs while accessing the stream
   */
  public XoePackage (ZipInputStream stream) 
    throws InvalidPackageException
  {
    this (stream, null);
  }

  /**
   * Creates a XoePackage from a ZipInputStream and specifies the name of the
   * file the stream was created from. The filename is useful if an error
   * occurs while loading the package.
   *
   * @param stream A ZipInputStream from which to access the contents of a ZIP Archive
   * @param name A string identifying the resource from which the stream was created.
   * @exception InvalidPackageException if the package cannot be parsed or the
   *                                    description is not understood
   * @exception IOException if an IO error occurs while accessing the stream
   */
  public XoePackage (ZipInputStream stream, String name)
  {
    m_filename = name;
    try 
      {
	/* process the stream once to make a copy */
	ByteArrayOutputStream baos = new ByteArrayOutputStream ();
	byte buf[] = new byte[1024];

	while (true)
	  {
	    int len = stream.read (buf);
	    if (len == -1) break;
	    baos.write (buf, 0, len);
	  }

	baos.flush ();
	streamCopy = baos.toByteArray ();

	/* we need to process the stream again */
	stream = new ZipInputStream (new ByteArrayInputStream (streamCopy));

        ZipEntry ze;
	while ((ze = stream.getNextEntry ()) != null)
	  {
	    if ("package.xml".equals (ze.getName ()))
	      {
		byte data[] = new byte[(int)ze.getSize ()];
		stream.read (data, 0, data.length);
		desc = new PackageDescription (new ByteArrayInputStream (data));
		return;
	      }
	    stream.closeEntry ();
	  }
      } 
    catch (InvalidPackageException e)
      {
	throw e;
      }
    catch (Exception e)
      {
	Logger.log (this, e);
	throw new InvalidPackageException (e.toString ());
      }
    throw new InvalidPackageException ("no package.xml file found");
  }

  /**
   * Store data associated with the package. This is most useful to Installer Services
   * that may want to store data at package installation time that is needed at
   * uninstallation time. By publishing the key and value-type used, an Installer Service
   * can store information here for consumption by other software components while the
   * package is installed. If <code>value</code> is null any data already stored with the
   * key is removed.
   *
   * @param key the key to be used for lookup
   * @param value the data to store
   */
  public void setExData (Object key, Object value)
  {
    if (key == null)
      {
	return;
      }

    if (value == null)
      {
	m_exData.remove (key);
	return;
      }

    m_exData.put (key, value);
  }

  /**
   * Retrieve stored data associated with the specified key
   */
  public Object getExData (Object key)
  {
    return m_exData.get (key);
  }

  /**
   * Retrieve the package description that was found in the XOE archive
   */
  public PackageDescription getDescription ()
  {
    return desc;
  }

  /**
   * Get the architecture this package was built for
   */
  public String getArchitecture ()
  {
    return desc.getArchitecture ();
  }


  /**
   * Get the DOM representation of the package description
   */
  public Document getSpec ()
  {
    return desc.getDocument ();
  }

  /**
   * Removes all references to the original XOE archive. Used to free memory
   * after package installation is complete.
   */
  public void forgetArchive ()
  {
    streamCopy = null;
  }

  /**
   * Retrieves a ZipInputStream that can be used to access the contents of the
   * XOE archive
   */
  public ZipInputStream getZipInputStream ()
  {
    return new ZipInputStream (new ByteArrayInputStream (streamCopy));
  }

  /**
   * Same as getZipInputStream but cast to InputStream 
   */
  public InputStream getInputStream ()
  {
    return (InputStream) getZipInputStream ();
  }

  /**
   * Gets the package name listed in the package description
   */
  public String getName ()
  {
    return desc.getName ();
  }

  /**
   * Gets the package URI built from the name, version, and architecture
   * listed in the package description
   */
  public String getPackageURI ()
  {
    return desc.getPackageURI ();
  }

  /**
   * Gets the package version listed in the package description
   */
  public PackageVersion getVersion ()
  {
    return desc.getVersion ();
  }

  public void performQuery (IPackageQuery in_q, Vector out_res) 
  {
    desc.performQuery (in_q, out_res);
  }

  /**
   * Retrieves all the provisos listed in the package description
   */
  public Proviso[] getProvides ()
  {
    NodeList nl = XPath.findNodes (getSpec (), "package/provides/proviso");
    Proviso ret[] = new Proviso[nl.getLength ()];

    for (int i = 0; i < ret.length; i++)
      {
	Element item = (Element) nl.item (i);
	ret[i] = new Proviso (item);
      }

    return ret;
  }

  /**
   *  Given a class <code>c</code> return the xoe package which provided it.
   */
  public static XoePackage getPackage (Class c)
  {
    ClassLoader cl = c.getClassLoader ();
    if (cl instanceof XoeClassLoader)
      {
	return (((XoeClassLoader)cl).getPackage ());
      }
    return (null);
  }
}
