/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: ServiceInstanceElement.java,v 1.6 2001/11/05 02:50:35 colin Exp $
 */
package org.xoe.core.services;

import org.w3c.dom.Element;

import org.xoe.core.dom.XElement;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.xpath.XPath;


/**
 * DOM Element used to represent IServices inside the ServiceLocator.
 * Each ServiceInstanceElement contains a DOM representation of the 
 * IService that it represents.  
 * The only thing it adds to a simple BaseElement is a hash key for 
 * finding the actual IService.
 */
public class ServiceInstanceElement 
  extends XElement
{
  private Object m_providerKey = null;

  protected Object getProviderKey ()
  {
    return m_providerKey;
  }

  /** 
   * Build the DOM tree for this element based on information from the IService.
   */
  public void init (IService provider)
  {
    m_providerKey = ServiceLocator.storeProvider (provider);

    String[] funcs = provider.getFunctionalities ();
    
    // getFunctionalities could return null
    if (funcs == null)
      {
        addFunctionality (provider, "none");
      }
    else
      {
        for (int i = 0; i < funcs.length; i++)
          {
            addFunctionality (provider, funcs[i]);
          }
      }

    String[] info = provider.getInformation ();
    for (int i = 0; info != null && i < info.length; i += 2 )
      {
	addInfo (info[i], info[i+1]);
      }
  }

  /**
   * Add Interfaces and Features for the DOM tree for the given Functionality.
   */
  protected void addFunctionality (IService provider, String func)
  {
    String[] ifaces = provider.getInterfaces (func);
    for (int j = 0; ifaces != null && j < ifaces.length; j++)
      {
        addInterface (provider, func, ifaces[j]);
      }
    String[] features = provider.getFeatures (func);
    for (int j = 0; features != null && j < features.length; j += 2)
      {
        addFeature (func, features[j], features[j+1]);
      }
  }

  public IService getProvider ()
  {
    return ServiceLocator.retrieveProvider (m_providerKey);
  }

  protected void addFeature (String functionality, String name, String value)
  {
    Element func = (Element) XPath.findFirstNode(this, "xoeservices:functionality[@type='" + functionality + "']");
    if (func == null)
      {
        func = getOwnerDocument ().createElementNS (ServiceLocator.SERVICE_NAMESPACE, "functionality");
        func.setAttribute ("type", functionality);
        appendChild (func);
      }
    Element feature = getOwnerDocument ().createElementNS (ServiceLocator.SERVICE_NAMESPACE,"feature");
    feature.setAttribute ("name", name);
    feature.setAttribute ("value", value);
    func.appendChild (feature);
  }

  public void addInfo (String name, String value)
  {
    Element info = getOwnerDocument ().createElementNS (ServiceLocator.SERVICE_NAMESPACE, "info");
    DOMUtils.setTextContent (info, value);
    info.setAttribute ("name", name);
    appendChild (info);
  }

  protected void addInterface (Object provider, String functionality, String iface)
    throws ClassCastException
  {
    if(!checkImplements(provider.getClass(), iface))
      {
	throw new ClassCastException(provider.getClass() + " does not implement " + iface);
      }

    Element func = (Element) XPath.findFirstNode (this, "xoeservices:functionality[@type='" 
                                                         + functionality 
                                                         + "']");
    if(func == null)
      {
	func = getOwnerDocument ().createElementNS (ServiceLocator.SERVICE_NAMESPACE, 
                                                    "functionality");
	func.setAttribute ("type", functionality);
	appendChild (func);
      }

    Element impl = getOwnerDocument ().createElementNS (ServiceLocator.SERVICE_NAMESPACE, "implements");
    DOMUtils.setTextContent (impl, iface);
    func.appendChild (impl);
  }

  private boolean checkImplements(Class clazz, String iname)
  {
    if(clazz.getName().equals(iname))
      {
	return true;
      }

    Class[] ifaces = clazz.getInterfaces();
    boolean found = false;
    for(int i = 0; i < ifaces.length; i++)
      {
	found = checkImplements(ifaces[i], iname);
	if(found)
	  {
	    return true;
	  }
      }

    Class sup = clazz.getSuperclass();
    
    return (sup == null ? false : checkImplements(sup, iname));
  }
}
