/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: ServiceLocator.java,v 1.9 2001/11/15 22:47:31 tim Exp $
 */
package org.xoe.core.services;

import java.util.Hashtable;

import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

import org.xoe.core.Constants;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.packages.IPackageXMLHandler;
import org.xoe.core.packages.Core;


/**
 * Service locator is a searchable registry of sevices.  Services are
 * stored in a DOM tree and XPath is used to search for them.
 * The ServiceLocator returns newInstance of services it finds.
 */
public final class ServiceLocator
  implements Constants
{
  public static final String SERVICE_NAMESPACE = DEFAULT_NAMESPACE + "/services";

  private static final IService[] nullIService = new IService[0];

  // context for XPath searches 
  private static XPathContext m_context = new XPathContext ();

  // DOM tree representing services 
  private static Element m_registry;

  private static Hashtable m_providers = new Hashtable ();

  private static int nextKey = 0;

  // create the registry Document
  static
  {
    m_registry = DOMlet.create (("<serviceRegistry xmlns='" + SERVICE_NAMESPACE + "'/>").toCharArray (), "serviceRegistry" ).getDocumentElement ();
    m_context.pushNode (m_registry);
  }

  // can't make new ServiceLocators
  private ServiceLocator () {}

  static Object storeProvider (IService provider)
  {
    Integer key = new Integer (nextKey++);
    m_providers.put (key, provider);
    return key;
  }

  static IService retrieveProvider (Object key)
  {
    return (IService) m_providers.get (key);
  }

  public static Element getRegistry ()
  {
    return m_registry;
  }

  public static synchronized void unregisterService (IService provider)
  {
    NodeList nl = m_registry.getChildNodes ();
    for (int i = 0; i < nl.getLength (); i++ )
      {
        ServiceInstanceElement elem = (ServiceInstanceElement) nl.item (i);
	if (elem.getProvider () == provider)
	  {
	    m_registry.removeChild (elem);
	    m_providers.remove (elem.getProviderKey ());
	  }
      }
  }

  /**
   * Register a service with the service locator.  This involves adding the 
   * ServiceInstanceElement to the DOM tree.
   * [FIXME] make sure we don't add the same service more than once
   */
  public static synchronized ServiceInstanceElement registerService (IService provider)
  {
    ServiceInstanceElement service = (ServiceInstanceElement)((XDocument)m_registry.getOwnerDocument ()).createElementNS (SERVICE_NAMESPACE, "service", ServiceInstanceElement.class);

    service.init (provider);
    m_registry.appendChild (service);

    return service;
  }

  public static IService findService (IServiceQuery query)
  {
    return findService (query.toXPath ());
  }

  public static IService[] findServices (IServiceQuery query)
  {
    return findServices (query.toXPath ());
  }

  public static IService findImplementor (String iface)
  {
    SimpleServiceQuery q = SimpleServiceQuery.alloc (iface);
    IService r = findService (q);
    SimpleServiceQuery.free (q);
    return (r);
  }

  public static IService[] findImplementors (String iface)
  {
    SimpleServiceQuery q = SimpleServiceQuery.alloc (iface);
    IService[] r = findServices (q);
    SimpleServiceQuery.free (q);
    return (r);
  }

  public static IService findImplementor (String functionality, String iface)
  {
    SimpleServiceQuery q = SimpleServiceQuery.alloc (functionality, iface);
    IService r = findService (q);
    SimpleServiceQuery.free (q);
    return (r);
  }

  public static IService[] findImplementors (String functionality, String iface)
  {
    SimpleServiceQuery q = SimpleServiceQuery.alloc (functionality, iface);
    IService[] r = findServices (q);
    SimpleServiceQuery.free (q);
    return (r);
  }

  /**
   * Find all the services matching the XPath.  Never returns null.
   */
  public static synchronized IService[] findServices (String xpath)
  {
    NodeList list = XPath.findNodes (m_context, xpath);
    if (list.getLength () == 0)
      {
	return nullIService;
      }
    else
      {
	IService[] services = new IService [list.getLength ()];
	for (int i = 0; i < list.getLength (); i++)
          {
	    services[i] = ((ServiceInstanceElement) list.item (i)).getProvider ();
            if (services[i] instanceof ServiceFactory)
              {
		Logger.log (ServiceLocator.class, "** NOTE ** Currently don't support newInstances of services!!");
                // services[i] = ((ServiceFactory) services[i]).newInstance ();
              }
          }
	return services;
      }
  }

  /**
   * Find a single service.  Return null if no services found.
   */
  public static synchronized IService findService (String xpath)
  {
    IService[] ret = findServices (xpath);
    return ret.length > 0 ? ret[0] : null;
  }

  public static synchronized void logTree ()
  {
    Logger.log (ServiceLocator.class, XMLWriter.getXML (m_registry));
  }
}
