/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: DOMUtils.java,v 1.21 2001/11/26 22:26:08 samc Exp $
 */
package org.xoe.core.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;

import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.Constants;
import org.xoe.core.content.DefaultContentTypeResolver;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.ElementNamedNodeMapAdaptor;
import org.xoe.core.dom.Loadable;
import org.xoe.core.dom.XElement;
import org.xoe.core.dom.XLink;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathParser;


/**
 * Utilities for manipulating and extracting information from, DOM trees.
 */
public class DOMUtils
  implements Constants
{
  /**
   * the xpath used to find elements with the onLoad and xpath attributes
   */
  static final XPath REFRESH_XPATH = XPathParser.singleton.parse ("//*[@" 
                                                                   + NAME_ATTR 
                                                                   + " or @xoe:" 
                                                                   + XPATH_ATTR 
                                                                   + "]");

  static final XPath HIDE_IF_XPATH = XPathParser.singleton.parse ("//*[@xoe:" 
                                                                   + HIDE_IF_ATTR 
                                                                   + "]");
  static final String REPEAT_ATTR = "repeat";
  static final XPath  REPEAT_XPATH = XPathParser.singleton.parse ("//*[@xoe:" 
								 + REPEAT_ATTR
								 + "]");

  static final String TEXT_CONTENT_ATTR = "text-content";
  static final XPath  FILL_XPATH = XPathParser.singleton.parse ("//@xoe:" 
								 + TEXT_CONTENT_ATTR
								 + " | //@xoe-replace:*" );
  public static final String XOE_REPLACE_NAMESPACE = DEFAULT_NAMESPACE + "/replace";

  /**
   *  Return a document fragment containing clones of the child nodes of n
   */
  public static DocumentFragment cloneChildrenAsFragment (Node n)
  {
    DocumentFragment frag = n.getOwnerDocument ().createDocumentFragment ();
    NodeList nl = n.getChildNodes ();
    for (int i = 0; i < nl.getLength (); i++)
      {
	frag.appendChild ( nl.item (i).cloneNode (true) );
      }

    return frag;
  }

  /**
   * Assigns the string value to the node n.
   * @depricated use setTextContent
   */
  public static void setAssignedValue (Node n, String value)
  {
    setTextContent (n, value);
  }

  /**
   * Returns the string representing the assigned value of the node passed in
   * @depricated use getTextContent
   */
  public static String getAssignedValue (Node n)
  {
    return (getTextContent (n));
  }

  /**
   *  Implementation of the DOM3 textContent attribute of Node.
   */
  public static void setTextContent (Node n, String text)
  {
    // [FIXME] should we silently ignore null Nodes?
    if (n != null)
      {
        switch (n.getNodeType ()) 
          {
          case Node.ELEMENT_NODE:
	    removeChildNodes (n);
	    n.appendChild (n.getOwnerDocument().createTextNode (text));
            break;
          case Node.ATTRIBUTE_NODE:
            ((Attr)n).setValue (text);    
            break;
          case Node.TEXT_NODE:
            ((Text)n).setData (text);
            break;
          case Node.CDATA_SECTION_NODE:
            ((CDATASection)n).setData (text);
            break;
          case Node.COMMENT_NODE:
            ((CDATASection)n).setData (text);
            break;

          case Node.ENTITY_REFERENCE_NODE:
          case Node.ENTITY_NODE:
          case Node.PROCESSING_INSTRUCTION_NODE:
            Logger.log (DOMUtils.class, "setAssignedValue not implemented for nodetype: " + getNodeTypeName (n.getNodeType ()));
            break;

          case Node.DOCUMENT_NODE:
          case Node.DOCUMENT_TYPE_NODE:
          case Node.DOCUMENT_FRAGMENT_NODE:
          case Node.NOTATION_NODE:
          default:
            Logger.log (DOMUtils.class, "nothing to do in setAssignedValue for node type: " + getNodeTypeName (n.getNodeType ()));
          }
      }
  }

  /**
   *  Implementation of the DOM3 textContent attribute of Node.
   */
  public static String getTextContent (Node n)
  {
    if (n != null)
      {
	switch (n.getNodeType ()) 
	  {
	    case Node.ELEMENT_NODE:
	      Text pcdata = getPCDATA (n);
	      if (pcdata != null)
		{
		  return pcdata.getData ();
		}
	    case Node.ATTRIBUTE_NODE:
	      return (((Attr)n).getValue ());    
	    case Node.TEXT_NODE:
	      return (((Text)n).getData ());
	    case Node.CDATA_SECTION_NODE:
	      return (((CDATASection)n).getData ());
	    case Node.COMMENT_NODE:
	      return (((CDATASection)n).getData ());

	    case Node.ENTITY_REFERENCE_NODE:
	    case Node.ENTITY_NODE:
	    case Node.PROCESSING_INSTRUCTION_NODE:
	      Logger.log (DOMUtils.class, "getAssignedValue not implemented for nodetype: " + getNodeTypeName (n.getNodeType ()));
              break;

	    case Node.DOCUMENT_NODE:
	    case Node.DOCUMENT_TYPE_NODE:
	    case Node.DOCUMENT_FRAGMENT_NODE:
	    case Node.NOTATION_NODE:
	    default:
	      Logger.log (DOMUtils.class, "nothing to do in getAssignedValue for node type: " + getNodeTypeName (n.getNodeType ()));
	  }
      }

    return null;
  }

  public static String getDoctypeNameNS (Document doc)
  {
    Element e = doc.getDocumentElement ();

    // right now, some base documents don't have document elements, like .dtd
    // so, if the document element is null, use the file suffix
    if (e == null)
      {
        return DefaultContentTypeResolver.getSuffix (doc.getNodeName ());
      }
    return e.getNamespaceURI () + "^" + e.getLocalName ();
  }

  /**
   * Return all elemments in the document with the given ID string.
   */
  public static NodeList getElementsById (Document doc, String elementId) 
  {
    return XPath.findNodes (doc, "//*[@id='" + elementId + "']");
  }
 
  /**
   * Get the text data of the first child element with the given ID string.
   */
  public static Text getTextById (Document doc, String elementId) 
  {
    return getPCDATA (doc.getElementById (elementId));
  }

  /**
   */
  public static Text getPCDATA (Node n)
  {
    if (n != null)
      {
        switch (n.getNodeType ()) 
          {
          case Node.ELEMENT_NODE:
            // these nodes can have children so get the first one
            switch (n.getChildNodes ().getLength ()) 
              {
              case 0: 
                return ((Text) n.appendChild (n.getOwnerDocument ().createTextNode ("")));
              case 1: 
                if (n.getFirstChild ().getNodeType () == Node.TEXT_NODE)
                  {
                    return ((Text) n.getFirstChild ());
                  }
              }
            break;

          case Node.DOCUMENT_NODE:
          case Node.DOCUMENT_FRAGMENT_NODE:
          case Node.ENTITY_REFERENCE_NODE:
          case Node.DOCUMENT_TYPE_NODE:
          case Node.ATTRIBUTE_NODE:
          case Node.ENTITY_NODE:
          case Node.PROCESSING_INSTRUCTION_NODE:
          case Node.COMMENT_NODE:
          case Node.TEXT_NODE:
          case Node.CDATA_SECTION_NODE:
          case Node.NOTATION_NODE:
          default:
            // do nothing since these nodes can't have children
            break;
          }
      }

    return (null);
  }

  public static void removeChildNodes (Node n)
  {
    if (n != null)
      {
	switch (n.getNodeType ()) 
	  {
	    case Node.DOCUMENT_NODE:
	    case Node.DOCUMENT_FRAGMENT_NODE:
	    case Node.ENTITY_REFERENCE_NODE:
	    case Node.ELEMENT_NODE:
	    case Node.ATTRIBUTE_NODE:
	    case Node.ENTITY_NODE:
              // these nodes can have children so remove them
              Node c = n.getFirstChild ();
              while (c != null)
                {
                  n.removeChild (c);
                  c = n.getFirstChild ();
                }
              break;

	    case Node.DOCUMENT_TYPE_NODE:
	    case Node.PROCESSING_INSTRUCTION_NODE:
	    case Node.COMMENT_NODE:
	    case Node.TEXT_NODE:
	    case Node.CDATA_SECTION_NODE:
	    case Node.NOTATION_NODE:
	    default:
	      // do nothing since these nodes can't have children
	  }
      }
  }

  public static String getNodeTypeName (short nodeType)
  {
    switch (nodeType) 
      {
	case Node.DOCUMENT_NODE:
	  return ("DOCUMENT_NODE");
	case Node.DOCUMENT_FRAGMENT_NODE:
	  return ("DOCUMENT_FRAGMENT_NODE");
	case Node.ENTITY_REFERENCE_NODE:
	  return ("ENTITY_REFERENCE_NODE");
	case Node.ELEMENT_NODE:
	  return ("ELEMENT_NODE");
	case Node.ATTRIBUTE_NODE:
	  return ("ATTRIBUTE_NODE");
	case Node.ENTITY_NODE:
	  return ("ENTITY_NODE");
	case Node.DOCUMENT_TYPE_NODE:
	  return ("DOCUMENT_TYPE_NODE");
	case Node.PROCESSING_INSTRUCTION_NODE:
	  return ("PROCESSING_INSTRUCTION_NODE");
	case Node.COMMENT_NODE:
	  return ("COMMENT_NODE");
	case Node.TEXT_NODE:
	  return ("TEXT_NODE");
	case Node.CDATA_SECTION_NODE:
	  return ("CDATA_SECTION_NODE");
	case Node.NOTATION_NODE:
	  return ("NOTATION_NODE");
	default:
	  return ("UNKNOWN");
      }
  }

  public static String getBanner (Node n, String banner)
  {
    if (banner == null)
      {
	return getBanner (n);
      }

    boolean tokenIsAQuery = false;
    String currToken;
    StringTokenizer st = new StringTokenizer (banner, QUERY_DELIMITER, true);
    StringBuffer sb = new StringBuffer ();
    while (st.hasMoreTokens ())
      {
	currToken = st.nextToken ();
	
	// if the current token is a query delimiter, toggle the query flag
	if (QUERY_DELIMITER.equals (currToken))
	  {
	    tokenIsAQuery = (tokenIsAQuery) ? false : true;
	  }
	else
	  {
	    
	    // if the query flag is true, do the query and append it to the output buffer
	    if (tokenIsAQuery)
	      {
		sb.append (XPath.findString (n, currToken));
	      }
	    // otherwise, just append the string to the output buffer
	    else
	      {
		sb.append (currToken);
	      }
	  }
      }
    
    // if we didn't build anything, return the original banner
    if (sb.length () <= 0)
      {
	return getBanner (n);
      }
    return (sb.toString ());
  }

  public static String getBanner (Node n)
  {
    if (n.getNodeType () == Node.ELEMENT_NODE || n.getNodeType () == Node.ATTRIBUTE_NODE)
      { 
	return n.getLocalName ();
      } 
    else 
      {
	return n.getNodeName ();
      }
  }
  
  public static void sortChildren (Node n, String banner, String order) 
  {
    sortChildren (n, banner, (ASCENDING_SORT.equals (order)));
  }

  // helper class
  private static class SortNode 
    implements Comparable
  {
    Node node;
    String key;
    SortNode (Node node, String key)
      {
        this.node = node;
        this.key = key;
      }
    public int compareTo (Object o)
      {
        return key.compareTo (((SortNode) o).key);
      }
  }

  public static void sortChildren (Node n, String banner, boolean ascending) 
  {
    SortNode [] nodes = new SortNode [n.getChildNodes ().getLength ()];
    int i=0;
    for (Node child = n.getFirstChild (); child != null ; child = child.getNextSibling ())
      {
        nodes[i] = new SortNode (child, getBanner(child, banner));
        i++;
      }

    java.util.Arrays.sort (nodes);

    DOMUtils.removeChildNodes (n);
    if (ascending)
      {
	for (i = 0 ; i < nodes.length; i++)
	  {
	    n.appendChild (nodes[i].node);
	  }
      }
    else 
      {
	for (i = nodes.length-1 ; i >= 0 ; i--)
	  {
	    n.appendChild (nodes[i].node);
	  }
      }
  }

  /** 
   *  reads the xoe:sort-order and xoe:sort-banner attributes from element 
   *  <code>e</code> and uses them to sort the children of e.
   */
  public static void sortChildren (Element e) 
  {
    sortChildren (e,  
                  e.getAttributeNS (DEFAULT_NAMESPACE, SORT_BANNER_ATTR), 
                  e.getAttributeNS (DEFAULT_NAMESPACE, SORT_ORDER_ATTR));
  }

  public static void expandUsingXPathAttributes (Node node, Node contextNode, XPathContext xpathContext)
  {  
    hideElements (node, contextNode, xpathContext);
    NodeList nodes = XPath.findNodes (xpathContext, node, REFRESH_XPATH);
    int length = nodes.getLength ();
    for (int j=0; j<length; j++)
      {
	Node n = nodes.item (j);
	if (n instanceof Loadable) 
	  {
	    ((Loadable)n).onLoad (xpathContext, contextNode);
	  }
      }
  }

  public static void hideElements (Node node, Node contextNode, XPathContext xpathContext)
  {  
    // the logic here is that we only want to grab one element at a 
    // time since by potentially removing nodes we will change
    // the results of a findNodes call.  so once we process an 
    // element, we either remove it altogether or remove it's 
    // hide-if attr, either way it won't be returned by the next
    // call to findFirstNode.

    Element e = (Element)XPath.findFirstNode (node, HIDE_IF_XPATH);
    while (e != null)
      {
        String test = e.getAttributeNS (DEFAULT_NAMESPACE, "hide-if");

        // if the test evaluates to true, remove the element from the doc
        if (XPath.evalToBoolean (xpathContext, contextNode, test))
          {
            prune (e.getParentNode (), e);
          }
        else
          {
            // otherwise, remove the hide-if attribute so this element
            // won't show up when we look for the next element
	    e.removeAttributeNS (DEFAULT_NAMESPACE, "hide-if");
          }

        e = (Element)XPath.findFirstNode (node, HIDE_IF_XPATH);
      }
  }

  public static void prune (Node parent, Node n)
  {
    if (parent != null)
      {
	parent.removeChild (n);
	if (!parent.hasChildNodes ())
	  {
	    prune (parent.getParentNode (), parent);
	  }
      }
  }

  public static Node expandUsingRepeat (Node node,  Node contextNode, XPathContext xpc)
  {
    Element e = (Element)XPath.findFirstNode (node, REPEAT_XPATH);
    while (e != null)
      {
        String repeat = e.getAttributeNS (DEFAULT_NAMESPACE, "repeat");
	e.removeAttributeNS (DEFAULT_NAMESPACE, "repeat");
	Node parent = e.getParentNode ();
	if (parent == null)
	  { // error !!
	    return node; 
	  }
	NodeList nodes = XPath.findNodes (xpc, contextNode, repeat);
	for (int i=0; i < nodes.getLength (); i++)
	  {
	    Node clone = e.cloneNode (true);
	    Node context = nodes.item (i);
	    expandUsingRepeat(clone, context, xpc);
	    fillAttributesAndStringValues (clone, context, xpc);
	    parent.insertBefore (clone, e);
	  }
  	prune (parent, e);
        e = (Element)XPath.findFirstNode (node, REPEAT_XPATH);
      }
    return node;
  }


  public static Node fillAttributesAndStringValues (Node node, Node contextNode, XPathContext xpc)
  {
    NodeList nodes = XPath.findNodes (xpc, node, FILL_XPATH);
    int length = nodes.getLength ();
    for (int i=0; i<length; i++)
      {
	Attr attr = (Attr) nodes.item (i);
	Element owner = attr.getOwnerElement ();
	String value = XPath.findString (xpc, contextNode, attr.getNodeValue ());
	if (attr.getNamespaceURI () == XOE_REPLACE_NAMESPACE)
	  {
	    owner.setAttribute (attr.getLocalName (), value);
	  }
	else
	  {
	    DOMUtils.setTextContent (owner, value);
	  }
	owner.removeAttributeNode (attr);
      }
    return node;
  }

  /**
   * Populate the fields of the returned document with the name/value
   * pairs of the parameters passed with the request.
   */
  public static void populateFields (Document doc, Hashtable params)
  {
    Enumeration e = params.keys ();
    if (e != null)
      {
	while (e.hasMoreElements ())
	  {
	    String name = (String)e.nextElement ();
	    String value = (String) params.get (name);
	    NodeList list = XPath.findNodes (doc, "//*[@name='"+name+"']");
	    for (int i = list.getLength () - 1; i >= 0; i--)
	      {
		((Element)list.item (i)).setAttribute (VALUE_ATTR, value);
	      }
	  }
      }
  }

  // -----------------------------------------------------------------------------
  // static namespace lookup methods

  public static String lookupNSForPrefix (Node start, String prefix, Node stop) 
  {
    prefix = prefix == null ? null : prefix.intern ();
    if (prefix == XMLNS)
      {
	return XMLNS_NAMESPACE;
      }

    Node stopParent = stop == null ? null : stop.getParentNode ();
    for (Node n = start; n != null && n != stopParent; n = n.getParentNode ()) 
      {
	if (!(n instanceof XElement)) 
	  {
	    continue;
	  } 
	XElement bn = (XElement)n;
	ElementNamedNodeMapAdaptor attributes = (ElementNamedNodeMapAdaptor) bn.getAttributes ();
	if (attributes != null)
	  {
	    String ns = attributes.getPrefix (prefix);
	    if (ns != null) 
	      {
		return (ns);
	      }
	  }
       }
    // map the xoe: prefix to the default namespace, this will happen via a dtd evenually
    if (prefix == "xoe")
      {
	return DEFAULT_NAMESPACE;
      }	
    else if (prefix == "xlink")
      {
	return XLink.XLINK_NAMESPACE;
      }
    else if (prefix == "xoe-replace")
      {
	return XOE_REPLACE_NAMESPACE;
      }
    return null;

  }

  public static String lookupPrefixForNS (Node start, String namespaceURI, Node stop)
  {
    if (namespaceURI != null)
      {
	namespaceURI = namespaceURI.intern ();

	if (namespaceURI == XMLNS_NAMESPACE)
	  {
	    return XMLNS;
	  }
	if (namespaceURI == XML)
	  {
	    return XML_PREFIX_NAMESPACE;
	  }

	Node stopParent = (stop == null ? null : stop.getParentNode ());
	for (Node n = start; n != null && n != stopParent; n = n.getParentNode ()) 
	  {
	    if (! (n instanceof Element)) 
	      {
		continue;
	      } 
	    else if (! (n instanceof XElement))
	      {
		Logger.log (n, "Can only call lookupPrefixForNS on XElements");
	      }
	  
            ElementNamedNodeMapAdaptor attrs = (ElementNamedNodeMapAdaptor) n.getAttributes ();
	    Attr vattr = attrs.getItemByValue (namespaceURI);
	    if (vattr != null) 
	      {
		String localName = vattr.getName();
		String prefix = vattr.getPrefix();
		if (prefix == XMLNS && lookupNSForPrefix (start, localName, stop) == namespaceURI)
		  {
		    return localName;
		  }
		if (prefix == null && localName == XMLNS && lookupNSForPrefix(start, null, stop) == namespaceURI)
		  {
		    return "";  // Special return value. Not a real prefix, indicates a match with the "xmlns" attribute
		  }
	      }
	  }
      }
    return null;
  }
  public static void main (String[] args)
    throws IOException
  {
    if (args.length != 2)
      {
        System.err.println ("usage: " + DOMUtils.class.getName () + " <file to transform>  <context>");
        System.exit (1);
      }

    // create the document from the File or URL
    Document doc1 = null;
    try 
      {
        doc1 = DOMlet.create (new URL (args[0]));
      }
    catch (MalformedURLException e)
      {
        try
          {
            InputStream in = new FileInputStream (args[0]);
            doc1 = DOMlet.create (in, args[0]);
          }
        catch (FileNotFoundException e2)
          {
            System.err.println ("Invalid URL or filename: " + args[0]);
            System.exit (1);
          }
      }

    Document doc2 = null;
    try 
      {
        doc2 = DOMlet.create (new URL (args[1]));
      }
    catch (MalformedURLException e)
      {
        try
          {
            InputStream in = new FileInputStream (args[1]);
            doc2 = DOMlet.create (in, args[1]);
          }
        catch (FileNotFoundException e2)
          {
            System.err.println ("Invalid URL or filename: " + args[1]);
            System.exit (1);
          }
      }
    System.out.println ("Original document");
    System.out.println (XMLWriter.getXML (doc1));
    expandUsingRepeat (doc1, doc2, null);
    System.out.println ("After expandUsingRepeat");
    System.out.println (XMLWriter.getXML (doc1));
    
    System.exit (0);
  }

//   public void populateFields (XoeServlet s, Hashtable params)
//   {
//     populateFields (s.getDisplayedDocument (), params);
//   }
}
