/* 
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XMLWriter.java,v 1.14 2001/11/15 22:12:05 colin Exp $ 
 */
package org.xoe.core.util;

import java.io.Writer;
import java.io.StringWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.net.URL;
import java.net.MalformedURLException;

import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;

import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.DefaultEntityResolver;
import org.xoe.core.dom.ElementNamedNodeMapAdaptor;
import org.xoe.core.util.Logger;


/**
 * Utility class for generating XML from DOM structures.
 */
public class XMLWriter
{
  /**
   * XML constants
   */
  public static final String XMLNS = "xmlns";
  public static final String XMLNS_NAMESPACE = "http://www.w3.org/2000/xmlns/";
  public static final String XML = "xml";
  public static final String XML_PREFIX_NAMESPACE = "http://www.w3.org/XML/1998/namespace";

  public static final String DEFAULT_VERSION = "1.0";
  public static final String DEFAULT_ENCODING = "UTF-8";
  public static final String INDENT = "  ";

  protected boolean createNS;
  protected Writer  writer;
  protected Node    startNode;
  protected String  encoding;

  protected XMLWriter (Node startNode, Writer writer, boolean createNS, String encoding) 
  {
    this.startNode = startNode;
    this.writer    = writer;
    this.createNS  = createNS;
    this.encoding  = encoding;
  }

  /**
   * Write XML data from a given node and all it's children.
   * This is the only public method.  It calls the constructor.
   * @param node the Node to write
   * @param writer Writer to write to.
   * @param encoding the encoding used in the Writer.
   * @param prettify toggle pretty indentation.
   */
  public static void write (Node node, Writer writer, 
                            boolean createNS, String encoding, boolean prettify)
    throws IOException
  {
    XMLWriter xmlWriter = new XMLWriter (node, writer, createNS, encoding);
    if (prettify == false) 
      { 
        xmlWriter.writeNode (xmlWriter.startNode, null, -1);
      }
    else                   
      {
        xmlWriter.writeNode (xmlWriter.startNode, null, 1);
      }
  }

  /**
   * writeNode.
   * This is a recusrive (but not tail-recursive) function.
   * If somebody can think of a way do this with loops or tail
   * recusion then feel free :)
   */
  protected void writeNode (Node node, String defaultNS, int indent)
    throws IOException
  {
    // Lets see what kind of node we have
    switch (node.getNodeType ())
      {
      case Node.TEXT_NODE:
        writer.write (DefaultEntityResolver.deEntify (((Text)node).getData ()));
        break;
      case Node.CDATA_SECTION_NODE:
        writeIndent (indent);
        writer.write ("<![CDATA[" + ((CDATASection) node).getData () + "]]>");
        writeNewline (indent);
        break;
      case Node.COMMENT_NODE:
        writeIndent (indent);
        writer.write ("<!--" + ((Comment) node).getData () + "-->");
        writeNewline (indent);
        break;
      case Node.ELEMENT_NODE:
        writeElement ((Element) node, defaultNS, indent);
        break;
      case Node.DOCUMENT_NODE:
        writeDocument ((Document) node);
        break;
      case Node.DOCUMENT_TYPE_NODE:
        DocumentType docType = (DocumentType) node;
        writeIndent (indent);
        writer.write ("<!DOCTYPE " + docType.getName () + " " );
        if (docType.getPublicId () == null)
          {
            writer.write ("SYSTEM \"" + docType.getSystemId () + "\">");
          }
        else
          {
            writer.write ("PUBLIC \"" + docType.getPublicId () + "\" \"" + docType.getSystemId () + "\">");
          }
        writeNewline (indent);
        break;
      case Node.PROCESSING_INSTRUCTION_NODE:
	writeIndent (indent);
	writer.write ("<?" + ((ProcessingInstruction) node).getTarget () + node.getNodeValue () + "?>");
	writeNewline (indent);
	break;
      case Node.DOCUMENT_FRAGMENT_NODE:
	NodeList nl = node.getChildNodes ();
	if (indent != -1)
	  { // slight HACK
	    writeNewline (1);
	  }
	for (int i = 0; i < nl.getLength (); i++)
	  {
	    writeNode (nl.item (i), defaultNS, indent != -1 ? indent + 1 : indent);
	  }
	break;
      case Node.ATTRIBUTE_NODE:
        writeAttribute ((Attr) node);
	break;
      default:
        System.err.println ("FIXME don't know how to serialize this nodetype: " + DOMUtils.getNodeTypeName (node.getNodeType ()));
        break;
      }
  }

  /**
   * Write an entire XML document.
   * @param doc the Document to write.
   * @param version the XML version.
   * @param the encoding scheme to use.
   * @param prettify toggle pretty indentation.
   */
  protected void writeDocument (Document doc)
    throws IOException
  {
    // write XML version string
    // should only do this if the do has no version string already
    writer.write ("<?xml version=\"" + DEFAULT_VERSION + "\" encoding=\"" + encoding + "\"?>\n");

    // write all child nodes
    for (Node child = doc.getFirstChild ();
        child != null;
        child = child.getNextSibling ())
      {
        writeNode (child, null, 0);
      }
  }

  /**
   * Recursivly write an element and its child nodes.
   * This is a recusrive (but not tail-recursive) function.  If somebody
   * can think of a way do this with loops or tail recusion then feel
   * free :)
   */
  protected void writeElement (Element elem, String defaultNS, int indent)
    throws IOException
  {
    String newDefaultNS = defaultNS;
    String prefix = elem.getPrefix ();
    String ns     = elem.getNamespaceURI ();
    int childIndent = indent;
    if (childIndent != -1) { childIndent++; }

    // start the opening tag.
    writeIndent (indent);
    writer.write ("<");

    String newNamespaceDecl = null;
    String newPrefix = null;

    // get the attributes for this element
    NamedNodeMap attrs = elem.getAttributes ();

    // Perhaps we need to add a namespace declaration or even a new prefix declaration
    // if the element already has a default namespace declaration.  This is where DOM->XML
    // really sucks
    if (createNS == true && prefix == null)
      {
        // see if our namespace is different to the defaultNS remebering
        // that either one could be null.
        if ((defaultNS != null && !defaultNS.equals (ns)) 
            || (ns != null && !ns.equals (defaultNS))) 
          {
            // see if we already have a default xmlns attribute on this element
            newDefaultNS = ElementNamedNodeMapAdaptor.getPrefix (attrs, null);
            if (newDefaultNS != null)
              {
                if (!newDefaultNS.equals (ns))
                  {
                    // declare a new namespaceprefix.
                    newPrefix = elem.getLocalName () + "ns";
                    for (int i=0; DOMUtils.lookupNSForPrefix (elem, newPrefix, startNode) != null; i++)
                      {
                        newPrefix = elem.getLocalName () + "ns" + i;
                      }
                    if (ns == null)
                      {
                        newNamespaceDecl = " " + XMLNS + ":" + prefix + "=\"\"";
                      }
                    else
                      {
                        newNamespaceDecl = " " + XMLNS + ":" + prefix + "=\"" + ns + "\"";
                      }

                    // and use it
                    writer.write (newPrefix + ":");
                  }
              }
            else
              {
                // declare a new default namespace.
                if (ns == null)
                  {
                    newNamespaceDecl = " " + XMLNS + "=\"\"";
                  }
                else
                  {
                    newNamespaceDecl = " " + XMLNS + "=\"" + ns + "\"";
                  }
                newDefaultNS = ns;
              }
          }
      }

    writer.write (elem.getNodeName ());

    if (newNamespaceDecl != null)
      {
        writer.write (newNamespaceDecl);
      }

    // print any actual attributes.
    for (int i=0; i < attrs.getLength (); i++)
      {
        writeAttribute ((Attr) attrs.item (i));
      }

    // special case: we have no childen
    if (!elem.hasChildNodes ())
      {
        writer.write ("/>");
        writeNewline (indent);
        return;
      }

    // special case: we have a text child, so cancel indentation.
    if (childIndent != -1)
      {
        for (Node child = (Node) elem.getFirstChild ();
            child != null;
            child = (Node) child.getNextSibling ())
          {
            if (child.getNodeType () == Node.TEXT_NODE)
              {
                childIndent = -1;
                break;
              }
          }
      }

    // end the opening tag
    writer.write (">");
    writeNewline (childIndent);

    // write all child nodes
    for (Node child = (Node) elem.getFirstChild ();
        child != null;
        child = (Node) child.getNextSibling ())
      {
        writeNode (child, newDefaultNS, childIndent);
      }

    // write closing tag
    if (childIndent != -1) { writeIndent (indent); }
    writer.write ("</");
    if (newPrefix != null) { writer.write (newPrefix + ":"); }
    writer.write (elem.getNodeName () + ">");
    writeNewline (indent);
  }

  /**
   * Write out the appropriate amount of indentation.
   * Abstracted so that someone with more java knowledge can write this in
   * a better way.
   * @param in the level of indentation to write.
   */
  protected void writeIndent (int in) 
    throws IOException
  {
    for (int i = 0; i < in; i++) { writer.write (INDENT); }
  }

  /**
   * Write a newline if required.
   * @param in the level of indentation to write.
   */
  protected void writeNewline (int in) 
    throws IOException
  {
    if (in >= 0) { writer.write ("\n"); }
  }

  /**
   * Write an attribute including a prefix and a namespace declaration
   * if required.
   */
  protected void writeAttribute (Attr attr)               
    throws IOException
  {
    String prefix = attr.getPrefix ();
    String ns     = attr.getNamespaceURI ();
    Element elem  = (Element) attr.getParentNode ();

    writer.write (" ");

    if (createNS == true)
      {
        // see if we may need to declare a new prefix.
        // xmlns declarations do not require a prefix.
        if (ns != null && !ns.equals (XMLNS_NAMESPACE) && prefix == null)
          {
            prefix = DOMUtils.lookupPrefixForNS (elem, ns, startNode);
            if (prefix == null)
              {
                // pick a new prefix to declare
                prefix = attr.getLocalName () + "ns";
                for (int i=0; DOMUtils.lookupNSForPrefix (elem, prefix, startNode) != null; i++)
                  {
                    prefix = attr.getLocalName () + "ns" + i;
                  }
              } 
            writer.write (XMLNS + ":" + prefix + "=\"" + ns + "\" ");
          }
      }
    if (prefix != null)
      {
        writer.write (prefix + ":");
      }
    writer.write (attr.getLocalName () + "=\"" + DefaultEntityResolver.deEntify (attr.getValue ()) + "\"");
  }

  /** 
   * Test the XMLWriter class.
   */ 
  public static void main (String[] args)
    throws IOException
  {
    if (args.length != 1)
      {
        System.err.println ("usage: " + XMLWriter.class.getName () + " <url or filename>");
        System.exit (1);
      }

    // create the document from the File or URL
    Document doc = null;
    try 
      {
        doc = DOMlet.create (new URL (args[0]));
      }
    catch (MalformedURLException e)
      {
        try
          {
            InputStream in = new FileInputStream (args[0]);
            doc = DOMlet.create (in, args[0]);
          }
        catch (FileNotFoundException e2)
          {
            System.err.println ("Invalid URL or filename: " + args[0]);
            System.exit (1);
          }
      }

    boolean createNS = true;
    boolean prettify = true;
    System.out.println ("createNS: " + createNS);
    System.out.println ("prettify: " + prettify);
    FileOutputStream outFile;
    int c;

    System.out.println ("testing idempotency of parse->write cyle:");
    StringWriter sw = new StringWriter ();
    write ((Node) doc, sw, createNS, DEFAULT_ENCODING, prettify);
    String firstPass = sw.toString ();
    System.out.println ("  first pass --> firstpass.xml");
    outFile = new FileOutputStream ("firstpass.xml");
    outFile.write (firstPass.getBytes ());

    sw = new StringWriter ();
    doc = DOMlet.create (firstPass.toCharArray (), args[0]);
    write ((Node) doc, sw, createNS, DEFAULT_ENCODING, prettify);
    String secondPass = sw.toString ();
    System.out.println ("  second pass --> secondpass.xml");
    outFile = new FileOutputStream ("secondpass.xml");
    outFile.write (secondPass.getBytes ());
    if (firstPass.equals (secondPass))
      {
        System.out.println ("  first and second pass match: success!");
      }
    else
      {
        System.out.println ("  first and second pass differ: failure!");
        System.exit (1);
      }

    System.out.println ("testing namespace declaration creation:");
    System.out.println ("  Each child of the document element is in different namespace");
    System.out.println ("  All other tags are in the null namespace.");

    doc = new XDocument ();
    Element docElem = doc.createElement ("docelem");
    doc.appendChild (docElem);

    Element child = doc.createElementNS ("ns://test1", "tag");
    child.appendChild (doc.createElement ("tag2"));
    docElem.appendChild (doc.createElementNS ("ns://test1", "tag"));

    child = doc.createElementNS ("ns://test2", "tag");
    child.appendChild (doc.createElement ("tag2"));
    docElem.appendChild (child);

    child = doc.createElementNS ("ns://test3", "tag");
    child.appendChild (doc.createElement ("tag2"));
    docElem.appendChild (child);

    sw = new StringWriter ();
    write ((Node) doc, sw, createNS, DEFAULT_ENCODING, prettify);
    System.out.println (sw.toString ());

    System.exit (0);
  }

  /**
   * Argument free version getXML.
   * prettifer: on.
   * namespace creation: off.
   * encoding: default.
   */
  public static String getXML (Node node)
  {
    return getXML (node, false, true);
  }

  /** 
   * Utility function for writing XML to a String.
   * The resulting XML has the default encoding as the String is not yet encoded.
   * No encoding required as we are writing to a String.
   */
  public static String getXML (Node node, boolean createNS, boolean prettify)
  {
    StringWriter sw = new StringWriter ();
    try 
      {
	XMLWriter.write (node, sw, createNS, DEFAULT_ENCODING, prettify);
	sw.flush ();
      } 
    catch (IOException ioe) 
      { 
	Logger.log (XMLWriter.class, ioe);
	throw new Error (ioe.toString ()); 
      }
    return sw.toString ();
  } 

  // returns the text "null" for a null node
  public static String getXML_debug (Node node)
  {
    if (node == null)
      {
        return ("null node");
      }
    
    return getXML (node);
  } 
}
