/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XPathContext.java,v 1.5 2001/11/09 21:35:05 suraj Exp $
 */
package org.xoe.core.xpath;

import org.xoe.core.Constants;

import org.xoe.core.dom.Cacheable;
import org.xoe.core.dom.NodeConsumer;
import org.xoe.core.dom.XLink;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.DOMUtils;

import org.w3c.dom.Node;

import java.util.Hashtable;


public class XPathContext
  implements XPathConstants, NodeConsumer, Cacheable
{
  /** Calling consume(Node) increments consumeCount.
   * Used to count the number of nodes in a result node-set. */
  public static final int CONSUME_COUNT = 1;

  /** Calling consume(Node) sets consumeKind to true and returns false.
   * Used when converting a node-set to boolean. */
  public static final int CONSUME_BOOL = 2;

  /** Calling consume(Node) sets consumeObject to string value to true
   * and returns false (so one first node is processed).
   * Used to implement string(nodeset). */
  public static final int CONSUME_STRING = 3;

  /** Calling consume(Node) appends string value to StringBuffer in consumeObject.
   * Used to implement testsuite. */
  public static final int CONSUME_STRING_APPEND = 4;

  /** Calling consume(Node) save name of first node in consumeObject.
   * Used to implement name() function. */
  public static final int CONSUME_NAME = 5;

  /** Specifies what to do if this.consume(Node) is called. */
  public int consumeKind;

  public int consumeCount;
  public Object consumeObject;

  int resultType;
  public double numResult;
  public boolean boolResult;
  public Object objResult;

  int currentPosition;
  int currentSize = -1;

  Node[] path;
  int numParents;

  VariableProvider variableProvider = null;
  FunctionProvider functionProvider = null;

  /** Index in xpath of axis specifier that starts matchStep.
   * Used to evaluate position() and last(). */
  int stepPc;

  /** Index in xpath of most recent OP_PREDICATE.
   * Used to evaluate position() and last(). */
  int predPc;

  /** Array of counters used to keep track of position(). */
  int[] counters;


	/*
	 * Mapping from prefixes to namespaces
	 */
  Hashtable prefixes;
	
  public Node getCurrentNode ()
  {
	return path[numParents];
  }

  public void setCurrentNode (Node node)
  {
	path[numParents] = node;	
  }

  public Node getDocumentRoot () 
  {
	Node n = getCurrentNode ();
	while (n.getParentNode () != null)
      {
		n = n. getParentNode ();
      }

	int t = n.getNodeType ();
	if (t == Node.DOCUMENT_NODE 
        || t == Node.DOCUMENT_FRAGMENT_NODE)
	  {
		return n;
	  }
	else 
	  {
		return null;
	  }
  }

  public void pushNode (Node node)
  {
	if (path == null)
      {
		path = new Node[8];
      }

	numParents++;
	if (path.length == numParents) 
      {
		Node[] tmp = new Node[2 * numParents];
		System.arraycopy (path, 0, tmp, 0, numParents);
		path = tmp;
	  }

	path[numParents] = node;
  }

  public void popNode ()
  {
	numParents--;
  }

  public Node getParentNode (int repeat)
  {
	repeat = numParents - repeat;
	return repeat < 0 ? null : path[repeat];
  }

  public int getCurrentPosition ()
  {
	return currentPosition;
  }

  public int getCurrentSize ()
  {
	return currentSize;
  }

  public int getCounter (int counter)
  {
	if (counters == null)
      {
		counters = new int[counter > 10 ? counter : 10];
      }
	if (counter <= counters.length) 
      {
		int newLength = counters.length;
		if (counter >= newLength)
          {
			newLength = counter + 5;
          }
		int[] newCounters = new int[newLength];
		System.arraycopy (counters, 0, newCounters, 0, counters.length);
		counters = newCounters;
	  }
	return counters[counter];
  }

  public void setCounter (int counter, int value)
  {
	getCounter (counter);
	counters[counter] = value;
  }

  public void incrementCounter (int counter)
  {
	getCounter(counter);
	counters[counter]++;
  }


  /** Consume node.
   * If (resultType==BOOLEAN_RESULT), we only care *if* there are any nodes.
   */
  public boolean consume(Node node)
  {
	switch (consumeKind) 
      {
		case CONSUME_COUNT:
			consumeCount++;
			return true;
		case CONSUME_BOOL:
			consumeCount = 1;
			return false;
		case CONSUME_STRING_APPEND:
			if (consumeObject == null)
              {
				consumeObject = new StringBuffer (100);
              }
			XPath.valueOf ((StringBuffer) consumeObject, node);
			return true;
		case CONSUME_STRING:
			if (consumeObject == null)
              {
				consumeObject = XPath.valueOf(node);
              }
			return false;
		case CONSUME_NAME:
			if (consumeObject == null)
              {
				consumeObject = node.getNodeName();
              }
			return false;
		default:
			throw new Error("unknown consumeKind");
	  }
  }

  public void setFunctionProvider (FunctionProvider newProvider)
  {
	functionProvider = newProvider;
  }

  public XPathFunction getFunctionReference (String fname)
  {
	if (functionProvider != null) 
	  {
		return functionProvider.getFunction (fname);
	  }
	return null;
  }

  public void setVariableProvider (VariableProvider newProvider)
  {
	variableProvider = newProvider;
  }

  public void getVariableReference (Object qname, int resultType)
  {
	if (variableProvider != null) 
      {
		XPathVariable var = variableProvider.getVariable (qname);
		if (var != null) 
          {
			var.result( this, resultType );
          }
		else
          {
			XPath.result(this, 
						  "<unknown variable " + qname.toString () + '>', 
						  resultType);
          }
	  }
    else 
      {
		String value = System.getProperty (qname.toString());
		if (value == null)
          {
			value = "<unknown property " + qname + '>';
          }
		XPath.result (this, value, resultType);
	  }
  }

  public void reset ()
  {
	  resetAllbutPrefixes ();
	  resetPrefixDefinitions ();
  }

  public void resetAllbutPrefixes () 
  {
	currentPosition = 0;
	currentSize = -1;
	path = null;
	numParents = 0;
	variableProvider = null;
	functionProvider = null;
	stepPc = 0;
	predPc = 0;
	counters = null;
	resultType = 0;
	numResult = 0.0;
	boolResult = false;
	objResult = null;
	consumeKind = 0;
	consumeCount = 0;
	consumeObject = null;
  }


	public  String getNamespaceURI (String prefix)
	{
		if (prefix == null)
			{
				return null;
			}
		prefix = prefix.intern ();
		if (prefix == "xoe")
			{
				return Constants.DEFAULT_NAMESPACE;
			}
		else if (prefix == Constants.XMLNS)
			{
				return Constants.XMLNS_NAMESPACE;
			}
		else if (prefix == "xlink")
			{
				return XLink.XLINK_NAMESPACE;
			}
		else if (prefix == "xhtml")
			{
				return Constants.XHTML_NAMESPACE;
			}
		else if (prefix == "xoeservices")
			{
				return ServiceLocator.SERVICE_NAMESPACE ;
			}
		else if (prefix == "xoe-replace")
			{
				return DOMUtils.XOE_REPLACE_NAMESPACE;
			}
		else if (prefix != null && prefixes != null)
			{
				return (String) prefixes.get(prefix);
			}
		else
			{
				return null;
			}
	}

	public void definePrefix (String prefix, String namespace)
	{
		if (prefixes == null)
			{
				prefixes = new Hashtable ();
			}
		prefixes.put(prefix.intern (), namespace.intern ());
	}

	public void resetPrefixDefinitions ()
	{
		prefixes = null;
	}



}

// This is for people using the Emacs editor:
// Local Variables:
// c-file-style: "java"
// tab-width: 4
// indent-tabs-mode: t
// End:
