#!/bin/sh 
#
# The file is sourced from all the other build scripts.
# It sets up env vars and functions for common use.
#
# Copyright (c) 2001 Transvirtual Technologies, Inc.
# See the file COPYING for information on usage and redistribution.
#
# $tvt: tvt-include.sh,v 1.76 2001/12/06 00:20:40 jim Exp $

version="0.1.11"

MAKEDIR="`dirname $0`"
GNU_BUILD_ARCH=`uname -m`
MAKEDIR=`cd $MAKEDIR && pwd`
if [ -r $HOME/.tvtmakerc ]; then
  RCFILE="$HOME/.tvtmakerc"
else
  RCFILE="$MAKEDIR/tvtmakerc"
fi
CONFDIR="$MAKEDIR/config.d"
PATH="$PATH:$MAKEDIR"
BINARY_DOWNLOADS="http://xoe.org/ftp"
BUILD_ID_FILE=".build_id"

##
# List targets to stdout.  Given a argument generate TARGET_LIST rather than printing.
##
list_configs ()
{
  save_target=$TARGET
  for target in `ls $CONFDIR`; do
    unset AUTOBUILD
    if [ -r $CONFDIR/$target/tvt-make.config ]; then
      . $CONFDIR/$target/tvt-make.config
      if [ "$1" != "" ]; then
        if [ "$AUTOBUILD" = "yes" ]; then
          TARGET_LIST="$TARGET_LIST $target"
        fi
      else
        printf "  %-20s %s\n" "$target" "$TARGET_INFO"
      fi
    fi
  done
  if [ -n "$save_target" -a -f "$CONFDIR/$save_target/tvt-make.config" ]; then
    . $CONFDIR/$save_target/tvt-make.config
  fi
}

##
# List availavle CVS modules (MODULE_LIST)
##
list_modules ()
{
  for module in $MODULE_LIST; do
    echo "  $module"
  done
}

##
# Set up target specific variables.
##
set_target ()
{
  TARGET=$1
  if [ ! -f "$CONFDIR/$TARGET/tvt-make.config" ]; then
    echo "$canonical: error: invalid target ($TARGET)" >&2
    echo "Avaialble Targets:"
    list_configs
    exit 1
  else
    . $CONFDIR/$TARGET/tvt-make.config
    mkdir -p $TVTDEST/$TARGET/$ARCHIVE
    echo "   TARGET = $TARGET"
    echo "TOOLCHAIN = $TOOLCHAIN"
    export PATH="$TOOLCHAIN:$PATH"
    confopts="--host=$GNU_HOST_ARCH --build=$GNU_BUILD_ARCH"
  fi
  if [ -n "$TOOLCHAIN" -a -n "$TOOLPREFIX" ]; then
    if [ ! -x ${TOOLCHAIN}${TOOLPREFIX}gcc ]; then
      log "error: invalid toolchain prefix: ${TOOLCHAIN}${TOOLPREFIX}"
      exit 1
    fi
  fi
}

##
# Ask a yes/no question
##
ask ()
{
  if [ "$ANSWER" = "yes" ]; then return 0; fi
  if [ "$ANSWER" = "no"  ]; then return 1; fi
  while true; do
    echo -n "$1 [Y/n] "
    read answer
    case "$answer" in
      "yes"|"y"|"Y"|"Yes")
        return 0 ;;
      "no"|"n"|"N"|"No")
        return 1 ;;
      "")  # default action = yes
        return 0 ;;
    esac
  done
}

##
# Copy the given filename to the archive in TVTDEST.
# And update the symlink.  Filename given MUST be absolute.
# Keep the last N build only
##
archive ()
{
  filename=$1
  basefile=`basename $filename`
  if [ ! -f "$filename" ]; then
    log "error: could not find file to archive ($filename)" >&2
    exit 1
  fi
  log "archiving '$basefile'"
  (
    cd $TVTDEST
    if [ -n "$TARGET" ]; then
      cd $TARGET
    fi
    ln -sf $ARCHIVE/$basefile-$DATE-$TIME $basefile
    cp $filename $ARCHIVE/$basefile-$DATE-$TIME
    backlog=`ls -lt $ARCHIVE/$basefile* 2> /dev/null | wc -l`
    if [ "$backlog" -gt "$BACKLOG" ]; then
      num=$(($backlog - $BACKLOG))
      delete=`ls -t $ARCHIVE/$basefile* | tail -$num` 
      #echo "removing $num old copy(s) of $basefile from the archive"
      rm -f $delete
    fi
  )
}

##
# Print the current tvt-make environment
##
tvt_print_env ()
{
  echo "   script = $canonical"
  echo "  version = $version  "
  echo "     host = `hostname`"
  echo "     user = `whoami`  "
  echo "     DATE = $DATE     "
  echo "  TVTROOT = $TVTROOT  "
  echo "  TVTDEST = $TVTDEST  "
  echo "  CONFDIR = $CONFDIR  "
  echo "  ARCHIVE = $ARCHIVE  "
}

##
# bump the $BUILD_ID
##
build_inc ()
{
  # Create build ID file if none exists
  if [ ! -f "$BUILD_ID_FILE" ]; then
    echo "Build ID number is not set, restoring defaults."
    echo "Setting Build ID now."
    echo "Build id is set to 1001"
    echo 1001 > $BUILD_ID_FILE
  fi

  BUILD_ID=`cat $MAKEDIR/$BUILD_ID_FILE`
  echo "$(($BUILD_ID + 1))" > $MAKEDIR/$BUILD_ID_FILE
  BUILD_ID=`cat $MAKEDIR/$BUILD_ID_FILE`
  echo $BUILD_ID > $MAKEDIR/$BUILD_ID_FILE
  export BUILD_ID
}

##
# print a log message
##
log ()
{
  if [ -n "$1" ]; then
    echo ""
    echo "${INDENT}$canonical: $1"
  fi
}

##
# Log an error message and exit
##
die ()
{
  if [ -n "$1" ]; then 
    log "$1"
  fi

  exit 1
}

##
# configure
##
configure ()
{
  FILE=$1
  [   -z "$FILE" ] && FILE="Makefile"
  [ ! -f "$FILE" ] && CONFIGURE="yes"
  if [ "$CONFIGURE" = "yes" ]; then
    if [ -n "$TARGET" ]; then
      log "configuring '$module' for '$TARGET'"
      CPPFLAGS="-I$TVTDEST/$TARGET/include" LDFLAGS="-L$TVTDEST/$TARGET/lib" ./autogen.sh $confopts
      [ $? -ne 0 ] && die "error: failed to configure '$module' for '$TARGET'"
    else
      log "configuring '$module'"
      ./autogen.sh $confopts
      [ $? -ne 0 ] && die "error: failed to configure '$module'"
    fi
  fi
}

##
# clean a module
##
clean ()
{
  if [ "$CLEAN" = "yes" -a -f ./Makefile ]; then
    log "cleaning '$module'"
    make distclean
  fi
}

##
# build a module
##
build ()
{
  if [ -z "$TARGET" ]; then
    log "building '$module'"
  else
    log "building '$module' for '$TARGET'"
  fi
  if ! make $1; then
    log "error: failed to build '$module' for '$TARGET'"
    exit 1
  fi
}

install ()
{
  if ! make prefix=$TVTDEST/$TARGET install; then
    log "error: failed to perform 'make install' of '$module' for '$TARGET'"
    log "check that your Makefiles use the varibale 'prefix' for installation root"
    exit 1
  fi
}

##
# (Re)generate the rcfile
##
tvt_make_rc ()
{
  # could use "select" here but it only works in bash(1)
  echo "Where would you like to store the tvt source code?"
  UPONE=`(cd ..; pwd)`
  echo "a) $UPONE"
  echo "b) $HOME/projects"
  echo "c) /usr/local/src"
  echo "d) other..."
  
  DEFAULTCHOICE="$UPONE"
  if [ ! -z "$TVTROOT" ]; then DEFAULTCHOICE="$TVTROOT"; fi

  while true; do
    echo -n "Enter a letter: [press enter for $DEFAULTCHOICE] "
    read answer
    case $answer in
      a) TVTROOT="$UPONE" ;;
      b) TVTROOT="$HOME/projects" ;;
      c) TVTROOT="/usr/local/src" ;;
      d) echo -n "enter path: "
         read TVTROOT
         ;;
      "")TVTROOT="$DEFAULTCHOICE";;
      *) echo "invalid option: $answer"
         continue ;;
    esac
    echo "Setting \$TVTROOT to: $TVTROOT"
    break;
  done;
  echo ""
  echo "Where would you like to store resulting binaries:"
  echo "a) $TVTROOT/latest"
  echo "b) $HOME/latest"
  echo "c) other..."
  DEFAULTCHOICE="$TVTROOT/latest"
  if [ ! -z "$TVTDEST" ]; then DEFAULTCHOICE="$TVTDEST"; fi
   while true; do
    echo -n "Enter a letter: [press enter for $DEFAULTCHOICE] "
    read answer
    case $answer in
      a) TVTDEST="$TVTROOT/latest" ;;
      b) TVTDEST="$HOME/latest" ;;
      c) echo -n "enter path: "
         read TVTDEST
         ;;
      "") TVTDEST="$DEFAULTCHOICE" ;;
      *) echo "invalid option: $answer"
         continue ;;
    esac
    echo "Setting \$TVTDEST to: $TVTDEST"
    break;
  done;
  
  echo -n "What CVS tag would you like to use when checking out new modules? [HEAD] :"
  read CVSTAG
  cat > $RCFILE <<END
#!/bin/sh
# config options for tvt-make scripts
#

##
# The list of modules you want from TVT.
# These are the modules that the top level 'tvt-make' will build.
##
MODULE_LIST="calibrate tvt-make fgl ipl libfgl kaffe-all pocketlinux xs xoe"

##
# Name of the XOE jar file
##
XOEJAR=xoe.jar

##
# Subdir for storing archives
##
ARCHIVE=archive
BACKLOG=10

##
# String used to indent log messages during build
##
INDENT=">>> "

##
# Very import paths
##
[ -z "\$TVTROOT" ] && TVTROOT=$TVTROOT 
[ -z "\$TVTDEST" ] && TVTDEST=$TVTDEST

CVSTAG=$CVSTAG
CONFIGVERSION=$version

##
# Names of the direcories where the tvt modules
# reside.
##
       PLDIR=pocketlinux
      XOEDIR=xoe
    KAFFEDIR=kaffe-all
      IPLDIR=ipl
      FGLDIR=fgl
   LIBFGLDIR=libfgl
CALIBRATEDIR=calibrate
  LIBMPEGDIR=video/libmpeg3
   SERVERDIR=server
       XSDIR=xs
END
}

##
# Set DATE
##
if [ -z "$DATE" ]; then
  export DATE=`date +%Y%m%d`
  export TIME=`date +%H%M`
fi

##
# Trap signal at the top level
##
if [ -z "$TRAP" ]; then
  trap die SIGINT SIGKILL
  export TRAP=1
fi

##
# Create rcfile if none exists
##
if [ ! -f "$RCFILE" ]; then
  echo "tvt-make does not apear to be configured."
  echo "Performing configurations now."
  echo ""
  tvt_make_rc
fi

##
# Source the rcfile
##
. $RCFILE

if [ -z "$TVTROOT" -o -z "$TVTDEST" ]; then
  echo "error: tvt-make is not configured correctly"
  echo "please edit or remove $RCFILE"
  exit 1
fi

if [ "$CONFIGVERSION" != "$version" ]; then
  log "error: tvt-make is version $version but your rc file is for version $CONFIGVERSION"
  log "Re-configuring...."
  tvt_make_rc
  . $RCFILE
fi

export MAKEDIR INDENT TVTDEST TVTROOT

##
# Check for directories that we will need.
##
if [ ! -d "$TVTROOT" ]; then
  log "could not find TVTROOT ($TVTROOT)" >&2
  if ask "should I create '$TVTROOT'?"; then
    mkdir -p $TVTROOT
  else
    exit 1
  fi
fi

if [ ! -d "$TVTDEST" ]; then
  log "could not find TVTDEST directory ($TVTDEST)" >&2
  if ask "should I create '$TVTDEST'?"; then
    mkdir -p $TVTDEST
  else
    exit 1
  fi
fi

##
# Check for programs the we will need.
##
programs="autoconf aclocal wget cvs cp mv ln zip jikes"
for prog in $programs; do
  if [ ! -x "`which $prog`" ]; then
    log "error: you must have '$prog' in your '\$PATH' to use tvt-make" >&2
    exit 1
  fi
done

mkdir -p $TVTDEST/$ARCHIVE
