package org.xoe.test;

import com.pocketlinux.packages.InstallOption;
import com.pocketlinux.packages.PackageDescription;
import com.pocketlinux.packages.PackageException;
import com.pocketlinux.packages.PackageManager;
import com.pocketlinux.packages.PackageQuery;
import com.pocketlinux.packages.SearchResults;
import com.pocketlinux.util.Logger;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Vector;

public abstract class Test
{
  private TestSuite m_parent = null;
  private String m_status;
  private boolean m_executed;
  private boolean m_success;
  private long m_time;
  private Vector m_vPkgs = new Vector();
  // You must implement this functions.

  protected abstract void test()
    throws AssertionFailedException;

  void setParent(TestSuite parent)
  {
    m_parent = parent;
  }

  protected String getBanner()
  {
    return "Test: " + getName();
  }

  protected final void report(Throwable t)
  {
    StringWriter sw = new StringWriter ();
    PrintWriter pw = new PrintWriter (sw);
    t.printStackTrace (pw);
    report (sw.toString ());
  }

  protected final void report(String report)
  {
    if(report.length() == 0)
      {
	return;
      }

    if(m_parent != null)
      {
	m_parent.childReport(report);
      }
    else
      {
	System.out.println(report);
      }
  }

  public void reset()
  {
    m_status = "Not yet executed.";
    m_executed = false;
    m_time = -1;
    m_success = false;
  }

  public Test()
  {
    reset();
  }

  protected void ASSERT(boolean assertion, String message)
    throws AssertionFailedException
  {
    if (!assertion)
      {
	throw new AssertionFailedException(message);
      }
  }

  // "Success." or
  // "ASSERTION FAILED: 'This is the string passed to assert()' (SprocketTest.java:132)"
  public final String getStatus()
  {
    return m_status;
  }

  public final boolean getSuccess()
  {
    return m_success;
  }

  public final String getName()
  {
    return this.getClass().getName();
  }

  public final void execute()
  {
    report(getBanner());

    try
      {
	long stime = System.currentTimeMillis();

	test();

	long etime = System.currentTimeMillis();
	m_time = etime - stime;

	m_status = "Success. Time: " + m_time + "ms";
	m_success = true;
      }
    catch(AssertionFailedException ex)
      {
	m_status = ex.getMessage();
	m_success = false;
      }
    catch(RuntimeException ex)
      {
	StringWriter sw = new StringWriter ();
	PrintWriter pw = new PrintWriter (sw);
	pw.println("Unhandled runtime exception:");
	ex.printStackTrace (pw);
	m_status = sw.toString ();
	m_success = false;
      }
    catch(Error err)
      {
	StringWriter sw = new StringWriter ();
	PrintWriter pw = new PrintWriter (sw);
	pw.println("Unhandled runtime error:");
	err.printStackTrace (pw);
	m_status = sw.toString ();
	m_success = false;
      }
    m_executed = true;
    report(getStatus());

    while(m_vPkgs.size() > 0)
      {
	String name = (String)m_vPkgs.elementAt(0);
	try
	  {
	    uninstallPackage(name);
	  }
	catch(Exception ex)
	  {
	    Logger.log(this, "WARNING: failed to uninstall package '" + name + "' during test.");
	  }
      }
  }

  protected void installPackage(String name)
    throws AssertionFailedException
  {
    try
      {
	PackageQuery query = new PackageQuery ();
	query.setName (name);
	SearchResults res = PackageManager.getInstance().search (query);
	InstallOption[] options = res.getInstallablePackages ();

	ASSERT(options.length > 0, "Couldn't find package named: " + name);

	PackageManager.getInstance().commit (options[0]);

	// we'll try to uninstall this package later
	m_vPkgs.addElement(name);
      }
    catch(PackageException ex)
      {
	report(ex);
	ASSERT(false, "Failed to install package " + name);
      }
  }

  protected void uninstallPackage(String name)
    throws AssertionFailedException
  {
    try
      {
	// one way or the other, don't try to remove this one later (ie, even if it fails)
	m_vPkgs.removeElement(name);
	PackageQuery query = new PackageQuery ();
	query.setName (name);
	Vector v = new Vector();
	PackageManager.getInstance().performQueryLocally (query, v);
	ASSERT(v.size() > 0, "v.size() > 0");
	InstallOption uninstallOption = PackageManager.getInstance().getUninstallOption((PackageDescription)v.elementAt(0));
	PackageManager.getInstance ().commit (uninstallOption);
      }
    catch(PackageException ex)
      {
	report(ex);
	ASSERT(false, "Unexpected PackageException");
      }

  }

}
