/* -*- mode: C; mode: fold -*- */

/*
 * auth - wrapper program for PAM password access
 * 
 * NOTE: this requires PAM > 0.74-51 to propperly support changing
 * authentication tokens (pam_chauthtok otherwise segfaults deep inside
 * the conversation *after* it already manipulated the authentication token)
 * 
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * 10/29/01  Peter Mehlitz
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <unistd.h>

#include <security/pam_appl.h>
#include <security/pam_misc.h>


/******************************************************************************
 * defines and typedefs
 */                                                                     /*{{{*/

/*
 * make sure the following constants are kept in sync with the Auth.java in use
 */

/* request types */
#define REQT_AUTH      0
#define REQT_CHGAUTH   1

/* data types */
#define DT_STRING      0

/* return error values */
#define NO_ERROR        0
#define ERR_UNSPECIFIED 1
#define ERR_SERVICE     2
#define ERR_USER        3
#define ERR_BAD_PW      4
#define ERR_BAD_NEW_PW  5
#define ERR_CHAUTH      6
#define ERR_TIME        7

/*
 * internal constants
 */

/* length constants */
#define MAGIC_LEN      8
#define MAX_BLOCKS     3
#define MAX_DATA_LEN   4096
#define MAX_TIME       2      /* delta in sec */

/* PAM servic names (XX maps to /etc/pam.d/XX config file, or 'other',
 * if not existent
 */

#define SVC_AUTH_PW       "other"
#define SVC_AUTH_DEV      NULL /* unless you have one */
#define SVC_CHGAUTH_PW    "other"
#define SVC_CHGAUTH_DEV   NULL /* unless you have one */


#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

//typedef unsigned int  uint;
typedef unsigned char uchar;
typedef signed char   schar;

typedef struct _auth_data_t {
	uint          len;
	uchar         type;
	uchar*        data;
} auth_data_t;

typedef struct _auth_request_t  {
    uchar         magic[MAGIC_LEN];
    uchar         type;      // 0: authenticate, 1: set pw

	uchar         nBlocks;
	auth_data_t   *block;

	int           error;
} auth_request_t;
                                                                        /*}}}*/

/******************************************************************************
 * globals
 */                                                                     /*{{{*/

static auth_request_t    Request;
static int               FdIn=0, FdOut=1;

#ifdef DEBUG
FILE                     *Log;
#endif
                                                                        /*}}}*/

/******************************************************************************
 * functions
 */

/* simple Vigniere cypher */
uchar* decryptBlock (auth_request_t* req,                               /*{{{*/
					 auth_data_t* block, int stringize)
{
	int    i, j;
	uchar  *key = req->magic;
	int    m = sizeof(req->magic);
	uchar  *data = 0;
	int    len = (stringize) ? block->len + 1 : block->len;

	if ( (data = malloc(len)) ){
		for ( i=0,j=0; i<block->len; i++ ){
			data[i] = block->data[i] - key[j];
			j = ++j % m;
		}
		
		if ( stringize ){
			data[i] = 0;
		}
	}

	return data;
}                                                                       /*}}}*/

void erase ( uchar* data, int len )                                     /*{{{*/
{
	int i;
	char  *p;
	
	for ( i=0, p=data; i<len; i++, p++ ){
		*p = 0;
	}
}                                                                       /*}}}*/

#ifdef DEBUG
void dumpRequest ( auth_request_t* req )                                /*{{{*/
{
	int i, j, n;
	auth_data_t *block;
	uchar* buf;
	
	fprintf( Log, "magic:    ");
	for ( i=0; i<sizeof(req->magic); i++ ){
		fprintf( Log, "%x ", (req->magic[i] & 0xff));
	}
	fprintf( Log, "\n");
	
	fprintf( Log, "type:     %d\n", req->type);
	fprintf( Log, "nBlocks:  %d\n", req->nBlocks);
	
	for ( i=0; i<req->nBlocks; i++ ){
		block = &req->block[i];

		fprintf( Log, "---- block: %d\n", i);
		fprintf( Log, "       len:    %d\n", block->len);
		fprintf( Log, "       type:   %d\n", block->type);
		
		if ( (buf = decryptBlock( req, block, TRUE)) ){
			fprintf( Log, "       data:   %s\n", buf);
			free( buf);
		}
	}
	fflush( Log);
}
                                                                        /*}}}*/
#endif

int readBlock ( auth_data_t* block )                                    /*{{{*/
{
	// check length, skip if 0
	if ( read( FdIn, &(block->len), sizeof(block->len)) == sizeof(block->len) ){
		block->len = ntohl(block->len);
		
		if ( (block->len > 0) && (block->len < MAX_DATA_LEN) ){
			
			// get type
			if ( read( FdIn, &(block->type), sizeof(block->type)) != sizeof(block->type) ){
				return 0;
			}
			
			// now get data
			if ( !(block->data = (char*) malloc(block->len)) ){
				return 0;
			}
			if ( read( FdIn, block->data, block->len) == block->len ){
				return 1;
			}
		}
	}
	
	return 0;
}                                                                       /*}}}*/

int readRequest ( auth_request_t* req )                                 /*{{{*/
{
	int i;

	// get magic cookie
	if ( read( FdIn, &(req->magic), sizeof(req->magic)) != sizeof(req->magic) ){
		return 0;
	}

	// get request type
	if ( read( FdIn, &(req->type), sizeof(req->type)) != sizeof(req->type) ){
		return 0;
	}
	if ( (req->type != REQT_AUTH) && (req->type != REQT_CHGAUTH) ){
		return 0; /* unknown equest */
	}

	// get number of data blocks
	if ( read( FdIn, &(req->nBlocks), sizeof(req->nBlocks)) != sizeof(req->nBlocks) ){
		return 0;
	}

	// get data blocks themselves
	if ( (req->nBlocks > 0) && (req->nBlocks <= MAX_BLOCKS) ){
		if ( !(req->block = (auth_data_t*) calloc( req->nBlocks, sizeof(auth_data_t))) ){

			return 0;
		}
		for ( i=0; i< req->nBlocks; i++ ){
			if ( !readBlock( &(req->block[i])) ){
				return 0;
			}
		}
	}
	
	return 1;
}                                                                       /*}}}*/

/* PAM helper function - avoid using clib stuff in PAM conv */
int startsWithIgnoreCase ( char* a, char* b, int len )                  /*{{{*/
{
	for ( ; len--; a++, b++ ){
		if ( (*a | 0x20) != (*b | 0x20) ){
			return 0;
		}
	}
	return 1;
}                                                                       /*}}}*/

int deduceDataBlock ( char* prompt )                                    /*{{{*/
{
	int ret = -2; /* -2: fatal failure, -1: ignore, >= 0: data block index */

	/* That's the weak part - there are about ten gazillion possible
	 * complaints and conversation turns a PAM module might throw at us.
	 * We could pass this over to Java, but then we have full IPC
	 * needs (including GUI), making Auth.java much more complicated
	 */
	if ( startsWithIgnoreCase( prompt, "password:", 9) ||
		 startsWithIgnoreCase( prompt, "old password:", 13) ){
		if ( Request.nBlocks > 1 ){
			ret = 1;
		}
	}
	else if ( startsWithIgnoreCase( prompt, "new password:", 13) ||
			  startsWithIgnoreCase( prompt, "re-enter new password:", 22) ){
		if ( Request.nBlocks > 2 ){
			ret = 2;
		}
	}
	else if ( startsWithIgnoreCase( prompt, "Bad password:", 13) ){
		//Request.error = ERR_BAD_NEW_PW;
	}
	else {
		// no data, just empty return
		ret = -1;
	}

	return ret;
}                                                                       /*}}}*/

int pamConversation (int num_msg,                                       /*{{{*/
					 const struct pam_message **msg,
					 struct pam_response **resp,
					 void *appdata_ptr)
{
	int i, j;
	uchar *data = NULL;
	struct pam_response *reply = NULL;
	char  *prompt;
	int style;

	reply = (struct pam_response *) calloc(num_msg,
										   sizeof(struct pam_response));
	if ( !reply ){
		return PAM_CONV_ERR;
	}

	
	for ( j=0; j<num_msg; j++ ){
		prompt = (char*) (*msg)[j].msg;
		style = (*msg)[j].msg_style;
		
		if ( (i = deduceDataBlock( prompt)) >= 0 ){
			if ( (data = decryptBlock( &Request, &Request.block[i], TRUE)) ){

				reply[j].resp = data;
				reply[j].resp_retcode = 0;
#ifdef DEBUG	
				fprintf(Log,"PAM conv[%d]: <%s> (%d) -> <%s>\n",
						j, msg[j]->msg, msg[j]->msg_style, reply[j].resp);
#endif			
			}
			else {
			}
		}
		else {
		}
	}
	
	*resp = reply;
	
	return PAM_SUCCESS;
}                                                                       /*}}}*/

void runPam ( auth_request_t* req, char* service, char* user,           /*{{{*/
			  int (*pamAction)(pam_handle_t*,int) )
{
	int ret;
	pam_handle_t *pamh;
	
	struct pam_conv conv = {
		&pamConversation,
		NULL
	};

	if ( service ){
		if ( (ret = pam_start( service, user, &conv, &pamh) == PAM_SUCCESS)){
			if ( (ret = pamAction( pamh, 0)) == PAM_SUCCESS ){
				req->error = NO_ERROR;
			}
			else {
#ifdef DEBUG
				fprintf( Log, "pam action failed: %s\n", pam_strerror(pamh,ret));
#endif
				switch ( ret ){
				case PAM_USER_UNKNOWN: req->error = ERR_USER; break;
				case PAM_AUTH_ERR:     req->error = ERR_BAD_PW; break;
				case PAM_AUTHTOK_ERR:  req->error = ERR_BAD_NEW_PW; break;
				default:               req->error = ERR_UNSPECIFIED; break;
				}
			}
			
			ret = pam_end( pamh, ret);
		}
		else {
			req->error = ERR_USER;
		}
	}
	else {
		req->error = ERR_SERVICE;
	}
}                                                                       /*}}}*/
	
/*
 * This is the only function that actually interfaces to PAM. It can
 * be replaced with a simple own pw retrieve/change mechanism
 */
int processRequest ( auth_request_t* req )                              /*{{{*/
{
	uchar *user = 0;
	uchar *service = 0;

	/* all but explicit success is considered a failure */
	req->error = ERR_UNSPECIFIED;
	
	/* get the user block, which we always need */
	if ( (req->nBlocks < 1) ||
		(req->block[0].type != DT_STRING) ||
		!(user = decryptBlock( req, &req->block[0], TRUE)) ){
		return 0;
	}
	
	switch ( req->type ){
		
	case REQT_AUTH: /*-------------------------- user authentication request */
		/* now get the servive, which maps to the PAM config file to use
		 * (OTHER in most cases)
		 */
		if ( req->nBlocks == 2 ){
			if ( req->block[1].type == DT_STRING ){
				service = SVC_AUTH_PW;
			}
		}
		else {
			service = SVC_AUTH_DEV;
		}

		runPam( req, service, user, pam_authenticate);
				
		break;
		
	case REQT_CHGAUTH: /*--------------- authentication token change request */
		/* now get the servive, which maps to the PAM config file to use
		 * (OTHER in most cases)
		 */
		if ( req->nBlocks == 3 ){
			if ( (req->block[1].type == DT_STRING) &&
				 (req->block[2].type == DT_STRING) ){
				service = SVC_CHGAUTH_PW;
			}
		}
		else {
			service = SVC_CHGAUTH_DEV;
		}
		
		runPam( req, service, user, pam_chauthtok);
		
		break;		
	}
	
	if ( user ){
		free( user);
	}
	
	return (req->error == NO_ERROR);
}                                                                       /*}}}*/

/*
 * This is a simple way to authenticate the 'auth' program from
 * Java's perspective. It is just required in case we don't have
 * real file system permissions (e.g. WindowsCE). It could have all
 * sort of complexity, but keep in mind the Java counterpart can be
 * easily disassembled
 */
int writeReply ( auth_request_t* req )                                  /*{{{*/
{
	/* over-simplistic for now - we just return the cross sum plus the error
	 * return value
	 */
	
	int i;
	uchar  buf[2];
	schar a;
		
	buf[0] = req->error;
	
	for ( i=0, a=req->error; i<MAGIC_LEN; i++ ){
		a+= (schar) req->magic[i];
	}

	buf[1] = a;
	
	if ( !write( FdOut, buf, sizeof(buf)) ){
		return 0;
	}
	
	return 1;
}                                                                       /*}}}*/

int main ( int argc, char* argv[] )                                     /*{{{*/
{
	struct timeval tv0, tv1, tvd;
	long  td;

#ifdef DEBUG /*{{{*/
	Log = fopen( "log", "w+");
	setbuf( Log, 0);
#endif /*}}}*/	

	gettimeofday( &tv0, NULL); /* mark, to later-on check response time */
	
	if ( readRequest( &Request) ){
#ifdef DEBUG /*{{{*/
		dumpRequest(&Request);
#endif /*}}}*/

		processRequest( &Request);
	}

	gettimeofday( &tv1, NULL);  /* somebody spying on us? */
	timersub( &tv1, &tv0, &tvd);
	td = tvd.tv_usec;
	if ( tvd.tv_sec > MAX_TIME ){
		Request.error = ERR_TIME;
	}

	writeReply( &Request);

#ifdef DEBUG /*{{{*/
	  fprintf( Log, "error: %d\n", Request.error);
#endif /*}}}*/	
	
	return Request.error;
}                                                                       /*}}}*/
