/**
 * AuthProcess - IAuth implementation which uses a external process
 *               to auhenticate users
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 * @see IAuth
 */

package org.xoe.auth;

import java.io.OutputStream;
import java.io.InputStream;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;


public class AuthProcess implements IAuth {

	private int error;
	
	/**************************************************************************
	/* these constants have to be kept in sync with the auth program in use
	 */
	private static final String AuthProc     = "auth";
	private static final int KEY_LENGTH      = 8;
	private static final int REPLY_LENGTH    = 2;
	
	private static final int REQT_AUTH       = 0;
	private static final int REQT_CHGAUTH    = 1;
	private static final int DT_STRING       = 0;

	private static final int NO_ERROR        = 0;
	private static final int ERR_UNSPECIFIED = 1;
	private static final int ERR_SERVICE     = 2;
	private static final int ERR_USER        = 3;
	private static final int ERR_BAD_PW      = 4;
	private static final int ERR_BAD_NEW_PW  = 5;
	private static final int ERR_CHAUTH      = 6;
	private static final int ERR_TIME        = 7;
	private static final int ERR_PROG        = 8;
	
	
	/**************************************************************************
	 * private methods
	 */

	boolean writeBlock ( DataOutputStream dos, String data, byte[] key, int keyLen )
	{
		int i, j, n;
		byte[] d = data.getBytes();
		
		try {
			dos.writeInt( d.length);
			dos.writeByte( DT_STRING);

			// do a simple vignere encryption
			n = d.length;
			for ( i=0, j=0; i<n; i++ ){
				d[i] += key[j];
				j = ++j % keyLen;
			}

			dos.write( d, 0, d.length);
			dos.flush();
			
			return true;
		}
		catch ( Exception x ){
			x.printStackTrace();
		}
		
		return false;
	}

	// standard password authentication workhorse
	// (we avoid usng fields to add a little bit more safety)
	boolean authPw ( String user, String pw, String newPw )
	{
		int i;
		int ret;
		byte a;
		long magic;

		byte[] key = new byte[KEY_LENGTH];
		byte[] reply = new byte[REPLY_LENGTH];
	
		Process p;
		DataInputStream dis;
		DataOutputStream dos;

		try {
			error = ERR_PROG;
			
			// a trivial key for now - 2-shifted current time in 8 bytes
			magic = System.currentTimeMillis();
			for ( i=0; i<KEY_LENGTH; i++){
				key[i] = (byte) magic;
				magic >>= 2;
			}

			p = Runtime.getRuntime().exec( AuthProc, null);
			
			dos = new DataOutputStream( p.getOutputStream());
			dis = new DataInputStream( p.getInputStream());
			
			// write request to processes input stream
			dos.write( key, 0, key.length);
			
			if ( newPw == null ){
				dos.writeByte( REQT_AUTH);
				dos.writeByte( 2);          // number of blocks
			}
			else {
				dos.writeByte( REQT_CHGAUTH);
				dos.writeByte( 3);          // number of blocks
			}
			
			writeBlock( dos, user, key, key.length);
			writeBlock( dos, pw, key, key.length);
			if ( newPw != null ){
				writeBlock( dos, newPw, key, key.length);
			}
			
			dos.flush();

			// read reply
			if ( dis.read( reply, 0, reply.length) != reply.length ){
				return false;
			}
			else {
				// a trivial 2-byte answer for now:
				//   reply[0] : error value
				//   reply[1] : cross sum key + error value
				error = reply[0];
				for ( i=0, a= (byte)error; i<KEY_LENGTH; i++ ){
					a += key[i];
				}
				if ( a != reply[1] ){
					error = ERR_PROG; /* consistent with reply value? */
					return false;
				}
			}
						
			ret = p.waitFor();
			if ( ret != error ){ /* error value consistent with proc exit? */
				error = ERR_PROG;
				return false;
			}
			
			return (error == NO_ERROR);
		}
		catch ( Exception xx ){
			xx.printStackTrace();
		}
		
		return false;
	}

	
	/**************************************************************************
	 * public API
	 */
	
	public synchronized boolean authenticate ( String user, String pw )
	{
		return authPw( user, pw, null);
	}
	
	public synchronized boolean changeAuthentication ( String user, String pw, String newPw )
	{
		return authPw( user, pw, newPw);
	}

	public int getError ()
	{
		return error;
	}

	public String getErrorMessage ( int err )
	{
		switch ( err ){
		case NO_ERROR:         return "no error";
		case ERR_UNSPECIFIED:  return "unspecified error";
		case ERR_SERVICE:      return "no service";
		case ERR_USER:         return "unknown user";
		case ERR_BAD_PW:       return "illegal password";
		case ERR_BAD_NEW_PW:   return "bad new pssword";
		case ERR_CHAUTH:       return "change password error";
		case ERR_TIME:         return "timeout";
		case ERR_PROG:         return "program error";
		}
			
		return "unknown error";
	}
	
	/**************************************************************************
	 * test API
	 */

	public static void main ( String[] args )
	{
		IAuth auth;
		
		System.out.println( "AuthProcess test application");
		
		if ( !AuthFactory.registerDefault( "AuthProcess") ){
			System.out.println( "auth registration failed!");
			return;
		}
		
		auth = AuthFactory.getDefaultInstance();

		if ( args.length == 2 ){
			System.out.println( "performing password authentication");
			auth.authenticate( args[0], args[1]);
		}
		else if ( args.length == 3 ){
			System.out.println( "performing password change");
			auth.changeAuthentication( args[0], args[1], args[2]);
		}
		else {
			System.out.println( "usage: 'java AuthProcess <user> <pw> [<newPw>]");
			return;
		}
		
		System.out.print( "auth result: ");
		System.out.println( auth.getErrorMessage( auth.getError()));
	}
}
