/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: LinuxNetwork.java,v 1.7 2001/11/09 19:37:21 samc Exp $
 */
package org.xoe.device;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.StreamTokenizer;
import java.io.Reader;

import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Vector;

import org.xoe.core.device.INetwork;
import org.xoe.core.util.Logger;

/**
 * Java interface to native networking functions.
 */
public final class LinuxNetwork
  implements INetwork
{
  private static final String[] INTERFACES = { INetwork.class.getName () };

  // ---------------------------------------------------------------------
  // IService

  public String[] getInterfaces (String func) { return INTERFACES; }
  public String[] getFeatures (String func)   { return null;       }
  public String[] getFunctionalities ()       { return null;       }
  public String[] getInformation ()           { return null;       }
  public void init (org.w3c.dom.Document configDoc) {}

  // ---------------------------------------------------------------------
  // INetwork

  public String getHostname ()
    throws IOException
  {
    return getHostname0 ();
  }

  public void setHostname (String host)
    throws IOException
  {
    if (!setHostname0 (host))
      {
        throw new IOException ("error setting hostname");
      }
  }

  public String getDomainname ()
    throws IOException
  {
    return getDomainname0 ();
  }

  public void setDomainname (String domain)
    throws IOException
  {
    if (!setDomainname0 (domain))
      {
        throw new IOException ("error setting domainname");
      }
  }

  private native String  getHostname0   ()              throws IOException;
  private native boolean setHostname0   (String host)   throws IOException;
  private native String  getDomainname0 ()              throws IOException;
  private native boolean setDomainname0 (String domain) throws IOException;

  /**
   * Parse /etc/resolv.conf to find all the nameservers.
   */
  public String[] getNameservers ()
    throws IOException
  {
    FileInputStream fis = new FileInputStream ("/etc/resolv.conf");
    StreamTokenizer tok = new StreamTokenizer (new InputStreamReader (fis));
    tok.resetSyntax ();
    tok.eolIsSignificant (true);
    tok.wordChars ('a', 'z');
    tok.wordChars ('A', 'Z');
    tok.wordChars ('0', '9');
    tok.wordChars ('.', '.');
    tok.whitespaceChars (' ', ' ');
    Vector vector = new Vector ();
    while (tok.nextToken () != StreamTokenizer.TT_EOF)
      {
        // look for "nameserver" and expect an IP number to follow it.
        if (tok.ttype == StreamTokenizer.TT_WORD && tok.sval.equals ("nameserver"))
          {
            tok.nextToken ();
            if (tok.ttype != StreamTokenizer.TT_WORD)
              {
                throw new IOException ("error parsing /etc/resolv.conf: expected IP number: " + tok);
              }
            vector.addElement (tok.sval);
          }
      }
    fis.close ();
    return (String[]) vector.toArray (new String[vector.size ()]);
  }

  /**
   * Add a new nameserver to the *end* of the current list.
   */
  public void setNameservers (String[] servers)
    throws IOException
  {
    FileOutputStream fos = new FileOutputStream ("/etc/resolv.conf");
    // [TODO] we should probably to what pump does and all highter level
    // domains. e.g: xoe.org if the domainname is uk.xoe.org.
    String entry = "search " + getDomainname0 ();
    for (int i = 0; i < servers.length; i++)
      {
        entry = "nameserver " + servers[i] + "\n";
        fos.write (entry.getBytes ());
      }
    fos.close ();
  }

  /**
   * Parse '/proc/net/dev' to find all the network devices that
   * the kernel knows about.  This is slightly faster and less
   * prone the change that the output of 'ifconfig -a'.
   */
  public String[] getNetworkInterfaces ()
    throws IOException
  {
    FileInputStream fis = new FileInputStream (new File ("/proc/net/dev"));
    StreamTokenizer tok = new StreamTokenizer (new InputStreamReader (fis));
    Vector vector = new Vector ();
    tok.eolIsSignificant (true);

    // skip the first two lines.
    while (tok.nextToken () != StreamTokenizer.TT_EOF && tok.lineno () < 3);

    // take the first token of all other lines.
    boolean start = true;
    while (tok.nextToken () != StreamTokenizer.TT_EOF)
      {
        if (tok.ttype == StreamTokenizer.TT_EOL)
          {
            start = true;
          }
        else if (start)
          {
            start = false;
            if (tok.ttype != StreamTokenizer.TT_WORD)
              {
                throw new IOException ("error parsing /proc/net/dev: expected interface name");
              }
            vector.addElement (tok.sval);
          }
      }
    return (String[]) vector.toArray (new String[vector.size ()]);
  }

  /**
   * Helper to turn the reverse hex IP numbers (as found in the kernel)
   * into strings.
   */
  static String hexToIP (String hex)
    throws IOException
  {
    if (hex.length () != 8)
      {
        throw new IOException ("invalid reverse hex IP address: " + hex);
      }
    StringBuffer buf = new StringBuffer ();
    for (int end = 8; end > 0; end = end - 2)
      {
        buf.append (Integer.parseInt (hex.substring (end - 2, end), 16));
        if (end > 2) buf.append (".");
      }
    return buf.toString ();
  }

  /**
   * Get information about the given interface by parsing the output
   * of 'ifconfig <iface>' and the contents of '/proc/net/route'.
   * This could (should?) be a native call.
   */
  public NetworkInterface getSettings (String interfaceName)
    throws IOException
  {
    Logger.log (this, "getting interfaceName: " + interfaceName);
    NetworkInterface iface = new INetwork.NetworkInterface ();
    iface.name      = "";
    iface.transport = "";
    iface.address   = "";
    iface.netmask   = "";
    iface.broadcast = "";
    iface.gateway   = "";
    // run ifconfig
    Process proc;
    proc = Runtime.getRuntime ().exec ("/sbin/ifconfig " + interfaceName);
    try
      {
        proc.waitFor ();
      }
    catch (InterruptedException e)
      {
        throw new IOException (e.toString ());
      }
    if (proc.exitValue () != 0)
      {
        throw new IOException ("ifconfig returned " + proc.exitValue ());
      }

    // parse ifconfig output
    Reader reader = new BufferedReader (new InputStreamReader (proc.getInputStream ()));
    StreamTokenizer tok = new StreamTokenizer (reader);
    tok.nextToken ();
    if (tok.ttype != StreamTokenizer.TT_WORD)
      {
        throw new IOException ("expected interface name");
      }
    iface.name = tok.sval;
    iface.dhcp = checkPump (iface);
    while (tok.nextToken () != StreamTokenizer.TT_EOF)
      {
        if (tok.ttype == StreamTokenizer.TT_WORD)
          {
            switch (tok.lineno ())
              {
              case 1:
                if (tok.sval.equals ("encap"))
                  {
                    tok.nextToken ();
                    tok.nextToken ();
                    if (tok.ttype != StreamTokenizer.TT_WORD)
                      {
                        throw new IOException ("expected a word following encap: " + tok);
                      }
                    iface.transport = tok.sval;
                    while (tok.lineno () == 1 && tok.ttype != StreamTokenizer.TT_EOF)
                      {
                        tok.nextToken ();
                      }
                  }
                break;
              case 2:
                if (tok.sval.equals ("addr"))
                  {
                    tok.nextToken ();
                    tok.nextToken ();
                    if (tok.ttype == StreamTokenizer.TT_WORD)
                      {
                        iface.address = tok.sval;
                      }
                  }
                else if (tok.sval.equals ("Mask"))
                  {
                    tok.nextToken ();
                    tok.nextToken ();
                    if (tok.ttype == StreamTokenizer.TT_WORD)
                      {
                        iface.netmask = tok.sval;
                      }
                  }
                else if (tok.sval.equals ("Bcast"))
                  {
                    tok.nextToken ();
                    tok.nextToken ();
                    if (tok.ttype == StreamTokenizer.TT_WORD)
                      {
                        iface.broadcast = tok.sval;
                      }
                  }
                break;
              case 3:
                if (tok.sval.equals ("UP"))
                  {
                    iface.active = true;
                  }
              default:
                break;
              }
          }
      }

    // parse 'proc/net/route' for network and gateway information
    FileInputStream fis = new FileInputStream ("/proc/net/route");
    tok = new StreamTokenizer (new InputStreamReader (fis));
    tok.resetSyntax ();
    tok.eolIsSignificant (false);
    tok.wordChars ('a', 'z');
    tok.wordChars ('A', 'Z');
    tok.wordChars ('0', '9');
    tok.wordChars ('.', '.');
    tok.whitespaceChars (' ', ' ');
    tok.whitespaceChars ('\t', '\t');
    tok.whitespaceChars ('\n', '\n');
    tok.whitespaceChars ('\r', '\r');
    boolean start = true;
    while (tok.nextToken () != StreamTokenizer.TT_EOF)
      {
        if (tok.ttype == StreamTokenizer.TT_EOL)
          {
            start = true;
          }
        else if (start)
          {
            if (tok.ttype != StreamTokenizer.TT_WORD)
              {
                throw new IOException ("error parsing /proc/net/route: expected work and start of line: " + tok);
              }
            if (tok.sval.equals (iface.name))
              {
                tok.nextToken ();
                tok.nextToken ();
                iface.gateway = hexToIP (tok.sval);
              }
          }
      }
    return iface;
  }

  /**
   * Setup the network interface according to the given settings.
   * This is done by constucting a call the ifconfig.
   */
  public void applySettings (NetworkInterface iface)
    throws IOException
  {
    if (iface.name == null)
      {
        throw new IOException ("invalid Network Interface");
      }

    if (iface.dhcp)
      {
        runPump (iface);
      }
    else
      {
        runIfconfig (iface);
      }
  }

  private static boolean checkPump (NetworkInterface iface)
    throws IOException
  {
    String cmd = "/sbin/pump --status -i " + iface.name;
    Process proc = Runtime.getRuntime ().exec (cmd);
    try
      {
        proc.waitFor ();
      }
    catch (InterruptedException e)
      {
        throw new IOException (e.toString ());
      }
    if (proc.exitValue () == 0)
      {
        return true;
      }
    return false;
  }

  private static void runPump (NetworkInterface iface)
    throws IOException
  {
    String cmd = "/sbin/pump -i " + iface.name;
    Process proc = Runtime.getRuntime ().exec (cmd);
    try
      {
        proc.waitFor ();
      }
    catch (InterruptedException e)
      {
        throw new IOException (e.toString ());
      }
    if (proc.exitValue () != 0)
      {
        throw new IOException ("pump returned " + proc.exitValue ());
      }
  }

  private static void runIfconfig (NetworkInterface iface)
    throws IOException
  {
    StringBuffer buffer = new StringBuffer ();
    buffer.append ("/sbin/ifconfig " + iface.name);
    if (iface.address != null)
      {
        buffer.append (" " + iface.address);
      }
    if (iface.netmask != null)
      {
        buffer.append (" netmask " + iface.netmask);
      }
    if (iface.broadcast != null)
      {
        buffer.append (" broadcast " + iface.broadcast);
      }
    Logger.log (LinuxNetwork.class, buffer.toString ());
    Process proc = Runtime.getRuntime ().exec (buffer.toString ());
    try
      {
        proc.waitFor ();
      }
    catch (InterruptedException e)
      {
        throw new IOException (e.toString ());
      }
    if (proc.exitValue () != 0)
      {
        throw new IOException ("ifconfig returned " + proc.exitValue ());
      }
  }

  public static void main (String[] args)
    throws Exception
  {
    LinuxNetwork network = new LinuxNetwork ();

    String[] ifaces = network.getNetworkInterfaces ();
    System.err.println ("Network devices:");
    for (int i = 0; i < ifaces.length; i++)
      {
        System.err.println (i + ": " + ifaces[i]);
      }
    System.err.println ();

    String[] servers = network.getNameservers ();
    System.err.println ("Nameservers:");
    for (int i = 0; i < servers.length; i++)
      {
        System.err.println (i + ": " + servers[i]);
      }
    System.err.println ();

    for (int i = 0; i < ifaces.length; i++)
      {
        System.err.println (network.getSettings (ifaces[i]));
        System.err.println ();
      }

    System.err.println ("Setting Nameserver: 10.0.0.222");
    network.setNameservers (new String [] { "10.0.0.222" });
  }
}
