/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: Screen.java,v 1.17 2001/11/09 23:06:26 colin Exp $
 */
package org.xoe.device;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Window;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.MediaTracker;
import java.io.PrintWriter;
import java.net.URL;

import kaffe.awt.AWT;

import org.w3c.dom.Document;

import org.xoe.core.content.Stash;
import org.xoe.core.device.IScreen;
import org.xoe.core.device.IScreenRotate;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.ILogger;
import org.xoe.core.util.Logger;

/**
 * Base class for other screen classes for specific
 * display shapes and sizes.
 * This is a service and therefore you cannot expect anything
 * to work until you call register it with the ServiceLocator
 * (i.e. call init ())
 */
abstract public class Screen
  implements IScreen, IScreenRotate
{
  private Window window;
  private Container screen;
  private Image background;
  private String status = "";
  private ILogger logger;
  private PrintWriter out;

  protected Screen () {}

  protected Window getWindow ()
  {
    return new Window (null)
    {
      public void update (Graphics g)
      {
        paint (g);
      }
    };
  }

  // ---------------------------------------------------------------------
  // org.xoe.core.device.IScreen

  public Container getScreen ()
  {
    synchronized (this)
      {
	if (screen == null)
	  {
	    screen = new Container ()
	    {
	      private Graphics gra;
	      private Image img;

	      public void paint (Graphics og)
	      {
		Image i = img;
		Graphics g = gra;
		if (g == null || i == null)
		  {
		    Dimension d = getSize();
		    i = createImage (d.width, d.height);
		    g = i.getGraphics ();
		    img = i;
		    gra = g;
		  }
		long start = System.currentTimeMillis ();
	        super.paint (g);
		long epaint = System.currentTimeMillis ();
	        og.drawImage (i, 0, 0, null);
		Logger.log (this, "paint time " + (epaint-start) + "ms");
	      }

	      public void reshape (int x, int y, int w, int h)
	      {
		Dimension d = getSize ();
		if (d.width != w || d.height != h)
		  {
		    img = null;
		    gra = null;
		  }
		super.reshape (x, y, w, h);
	      }

	      public Component add(Component c)
	      {
		super.add (c, BorderLayout.CENTER);
		return c;
	      }
	    };

	    screen.setLayout (new BorderLayout ());
	    window.removeAll ();
	    window.add (screen, BorderLayout.CENTER);
	    background = null;
	    status = "";
	    if (logger != null)
	      {
		logger.unregister(out, ILogger.STATUS);
		logger = null;
		out = null;
	      }
	  }
      }
    return screen;
  }

  protected Image getBootingImage ()
  {
    // during the package build we get null from getWorkingDirectory
    // so ww catch the case. This should not happen, but it does
    URL url = Stash.getWorkingDirectory (this.getClass ());
    if (url != null)
      {
        String bkgImgSrc = url.toString ();
        if (!bkgImgSrc.endsWith ("/"))
          {
            bkgImgSrc += "/";
          }
        bkgImgSrc += "background.jpg";

        return Toolkit.getDefaultToolkit ().getImage (bkgImgSrc);
      }
    else
      {
        return null;
      }
  }

  protected Dimension getScreenSize ()
  {
    return Toolkit.getDefaultToolkit ().getScreenSize ();
  }

  // ---------------------------------------------------------------------
  // org.xoe.core.device.IScreenRotate

  public int getRotation ()
  {
    switch (AWT.extension.getScreenRotation ())
    {
      default:
      case 0:
	return 0;
      case 1:
	return 90;
      case 2:
	return 180;
      case 3:
	return 270;
    }
  }

  public void setRotation (int newrot)
  {
    if (newrot < 90)
      {
	newrot = 0;
      }
    else if (newrot < 180)
      {
	newrot = 1;
      }
    else if (newrot < 270)
      {
	newrot = 2;
      }
    else
     {
	newrot = 3;
     }

    int oldrot = AWT.extension.getScreenRotation ();
    AWT.extension.setScreenRotation (newrot);

    Dimension size = window.getSize ();
    window.invalidate ();
    if (((oldrot ^ newrot) & 1) == 1)
      {
	int v = size.height;
	size.height = size.width;
	size.width = v;
      }
    window.setBounds (0, 0, size.width, size.height);
  }

  public int[] getSupportedRotations ()
  {
    return new int[] { 0, 90, 180, 270 };
  }

  public int getDefaultRotation ()
  {
    return 90;
  }

  // ---------------------------------------------------------------------
  // org.xoe.core.services.IService

  public void init (Document configDoc)
  {
    window = getWindow ();
    setRotation (getDefaultRotation ());

    background = getBootingImage ();
    MediaTracker tracker = new MediaTracker (window);
    tracker.addImage (background, 0);
    try
      {
	tracker.waitForAll ();
      }
    catch (InterruptedException _)
      {
      }

    window.add (new Component ()
    {
      public void paint (Graphics g)
      {
        g.drawImage (background, 0, 0, this);
	FontMetrics fm = g.getFontMetrics ();
	int sw = fm.stringWidth (status);
	int sh = fm.getHeight ();
	Dimension d = getSize();
	int x = (d.width - sw) / 2;
	int y = d.height - sh;
	g.drawString (status, x, y);
      }
    }, BorderLayout.CENTER);

    window.setSize (getScreenSize ());
    window.setVisible (true);

    // Install status logger
    logger = (ILogger) ServiceLocator.findImplementor (ILogger.INTERFACE);
    if (logger != null)
      {
	out = new PrintWriter ((PrintWriter)null, true)
	{
	  public void write(char[] buf, int off, int len)
	  {
	      status = new String (buf, off, len);
	      window.repaint ();
	  }

	  public void println ()
	  {
	  }
	};
	logger.register (out, ILogger.STATUS);
      }
  }

  public String[] getFunctionalities ()
  {
    return new String[] { IScreen.FUNCTIONALITY, IScreenRotate.FUNCTIONALITY };
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals(IScreen.FUNCTIONALITY))
      {
	return new String[] { IScreen.INTERFACE };
      }
    else if (func.equals(IScreenRotate.FUNCTIONALITY))
      {
	return new String[] { IScreenRotate.INTERFACE };
      }
    return null;
  }

  public String[] getFeatures (String func)
  {
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }
}
