/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * $tvt: Canna.java,v 1.1 2001/11/29 01:44:45 samc Exp $ 
 */
package org.xoe.input.canna;

import org.xoe.core.util.Logger;
import java.io.UnsupportedEncodingException;

/**
 * Kana to kanji convertion.
 */
public class Canna
{
  String inputString;
  int context;
  int clauses;
  boolean initDone = false;

  /**
   * Create a new canna object for translating a String.
   */
  public Canna (String input) 
    throws InstantiationException
  {
    inputString = input;
    if (!initDone)
      {
        if (!init ())
          {
            throw new InstantiationException ("failed to initialize Canna libraries");
          }
        initDone = true;
      }
    if ((context = newContext ()) < 0)
      {
        throw new InstantiationException ("failed to initialize Canna libraries");
      }
    if (   !loadDict ("iroha")
        || !loadDict ("fuzokugo")
        || !loadDict ("hojomwd"))
      {
        throw new InstantiationException ("error loading canna dictionaries");
      }
    try
      {
        byte[] hira = romaToHira0 (input.getBytes ("EUCJP"));
        clauses = setWord (context, hira);
      }
    catch (UnsupportedEncodingException e)
      {
        throw new InstantiationException ("Canna required EUCJP support");
      }
  }

  /**
   * Close the canna context on finalize.
   */
  public void finalize () 
  {
    closeContext (context);
  }

  /**
   * Get the orignal input string.
   */ 
  public String getString ()
  {
    return inputString;
  }

  /**
   * Load a Canna dictionary.
   */
  public boolean loadDict (String name)
  {
    return loadDict (context, name);
  }

  /**
   * Find out how many clauses the current String has.
   */
  public int getClauseCount ()
  {
    return clauses;
  }

  public void setNextClause () 
  {
    nextClause (context);
  }

  public void setPreviousClause ()
  {
    previousClause (context);
  }

  /**
   * Get candidates.
   */
  public String[] getCandidates ()
  {
    try
      {
        int numCandidiates = 0;
        byte[] bytes = getCandidates0 (context);
        if (bytes.length > 0) numCandidiates=1;
        for (int i=0; i<bytes.length; i++)
          {
            if (bytes[i] == '\0')
              {
                numCandidiates++;
              }
          }

        System.out.println ("got " + numCandidiates);
        String rtn[] = new String[numCandidiates];
        int cstr = 0;
        int spos = 0;
        int i;
        for (i = 0; i < bytes.length; i++) 
          {
            if (bytes[i] == '\0') 
              {
                rtn[cstr] = new String (bytes, spos, i-spos, "EUCJP");
                cstr++;
                i++;
                spos = i;
              }
          }
        rtn[cstr] = new String (bytes, spos, i-spos, "EUCJP"); 
        return rtn;
      }
    catch (Exception e) 
      {
        Logger.log (Canna.class, e);
        return null;
      }
  }

  /**
   * Translate Romaju to Katakana.
   */
  static public String romaToKata (String input)
  {
    try 
      {
        byte[] eusChars = input.getBytes ("EUCJP");
        byte[] newChars = romaToKata0 (eusChars);
        return new String (newChars, "EUCJP");
      }
    catch (UnsupportedEncodingException e)
      {
        Logger.log (Canna.class, e);
        return input;
      }
  }

  /**
   * Translate Romaju to Hirigana.
   */
  static public String romaToHira (String input)
  {
    try 
      {
        byte[] eusChars = input.getBytes ("EUCJP");
        byte[] newChars = romaToHira0 (eusChars);
        return new String (newChars, "EUCJP");
      }
    catch (UnsupportedEncodingException e)
      {
        Logger.log (Canna.class, e);
        return input;
      }
  }

  /**
   * Translate Hirigana to Katakana.
   */
  static public String hiraToKata (String input)
  {
    try 
      {
        byte[] eusChars = input.getBytes ("EUCJP");
        byte[] newChars = hiraToKata0 (eusChars);
        return new String (newChars, "EUCJP");
      }
    catch (UnsupportedEncodingException e)
      {
        Logger.log (Canna.class, e);
        return input;
      }
  }

  /**
   * Canna native methods.
   */
  static private native boolean init ();
  static private native int     newContext ();
  static private native void    closeContext   (int context);
  static private native byte[]  getCandidates0 (int context);
  static private native boolean loadDict       (int context, String name);
  static private native int     setWord        (int context, byte[] string);
  static private native int     nextClause     (int context);
  static private native int     previousClause (int context);
  static private native byte[]  romaToKata0 (byte[] input);
  static private native byte[]  romaToHira0 (byte[] input);
  static private native byte[]  hiraToKata0 (byte[] input);

  /**
   * Main test function.
   * Read a romaji string argument and print out canna
   * translations in EUC-JP.
   */
  public static void main (String args[])
    throws Exception
  {
    if (args.length != 1)
      {
        System.err.println ("please pass in some romaji (ascii) text");
        System.exit (1);
      }

    System.loadLibrary ("canna");

    String hira = Canna.romaToHira (args[0]);
    String kata = Canna.romaToKata (args[0]);

    System.out.println ("Roma : " + args[0]);

    System.out.print ("roma->hira : ");
    System.out.write (hira.getBytes ("EUCJP"));
    System.out.println ("");

    System.out.print ("roma->kata : ");
    System.out.write (kata.getBytes ("EUCJP"));
    System.out.println ("");

    kata = Canna.hiraToKata (hira);
    System.out.print ("hira->kata : ");
    System.out.write (kata.getBytes ("EUCJP"));
    System.out.println ("");

    Canna canna = new Canna (args[0]);
    int clauses = canna.getClauseCount ();
    System.out.println ("We got "+clauses+" clauses in "+args[0]);
    for (int i = 0; i < clauses; i++ )
      {
        System.out.println ("clause : "+i);
        String strings[] = canna.getCandidates ();
        for (int j=0; j<strings.length; j++)
          {
            System.out.print ("cand "+j+" : ");
            System.out.write (strings[j].getBytes ("EUCJP"));
            System.out.println ();
          }
        canna.setNextClause ();
      }
  }
}
