#include "libsslio.h"
#include <errno.h>

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

extern int exDataIndex;

struct SslIONative
{
  BIO *bio;
  SSL *ssl;
};

struct CallbackInfo
{
  JNIEnv *env;
  jobject sslio;
};

static int xoe_ssl_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx)
{
  SSL* ssl;
  struct CallbackInfo *cb;
  jclass clazz;
  jmethodID cbMethID;
  jobject javachain;
  STACK_OF(X509) *cchain;
  int     err, depth;
  jstring errstring;

  if(0 == (ssl = X509_STORE_CTX_get_ex_data(x509_ctx, SSL_get_ex_data_X509_STORE_CTX_idx())))
    {
      throwSSLRuntimeException(cb->env, "X509_STORE_CTX_get_ex_data failed in xoe_ssl_verify_callback");
      return 0;
    }

  if(0 == (cb = SSL_get_ex_data(ssl, exDataIndex)))
    {
      throwSSLRuntimeException(cb->env, "SSL_get_ex_data failed in xoe_ssl_verify_callback");
      return 0;
    }

  err = X509_STORE_CTX_get_error(x509_ctx);
  depth = X509_STORE_CTX_get_error_depth(x509_ctx);

  if(NULL == (errstring = (*cb->env)->NewStringUTF(cb->env, X509_verify_cert_error_string(err))))
    { // already throws
      return 0;
    }
  
  if(0 == (clazz = (*cb->env)->GetObjectClass(cb->env, cb->sslio)))
    {
      throwSSLRuntimeException(cb->env, "GetObjectClass failed in xoe_ssl_verify_callback");
      return 0;
    }

  if(0 == (cbMethID = (*cb->env)->GetMethodID(cb->env, clazz, "verifyCallback",
					      "(Lcom/transvirtual/protocol/ssl/X509CertificateChain;IILjava/lang/String;Z)V")))
    {
      // this thing throws already
      return 0;
    }

  if(NULL == (cchain = X509_STORE_CTX_get1_chain(x509_ctx)))
    {
      throwSSLRuntimeException(cb->env,
			       "X509_STORE_CTX_get1_chain failed in xoe_ssl_verify_callback");
      return 0;
    }

  if(NULL == (javachain = (jobject)newX509CertificateChain(cb->env, cchain)))
    { // this throws already
      sk_X509_pop_free(cchain,X509_free);
      return 0;
    }

  // this function throws an exception if the app rejects the cert
  (*cb->env)->CallVoidMethod(cb->env, cb->sslio, cbMethID,
			     javachain, depth, 
			     err, errstring, (preverify_ok == 1) ? JNI_TRUE : JNI_FALSE );

  if((*cb->env)->ExceptionOccurred(cb->env))
    {
      return 0;
    }
  return 1;
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslIO_freeSsl0(JNIEnv* env, jobject this,
						   jobject internal)
{
  if(((struct SslIONative*)internal)->bio != NULL)
    {
      BIO_free_all(((struct SslIONative*)internal)->bio);
      ((struct SslIONative*)internal)->bio = 0;
    }
}

JNIEXPORT jobject JNICALL
Java_com_transvirtual_protocol_ssl_SslIO_getPeerCertificate0(JNIEnv* env, jobject this,
							     jobject internal)
{
  if(((struct SslIONative*)internal)->ssl != NULL)
    {
      return newX509Certificate (env, SSL_get_peer_certificate (((struct SslIONative*)internal)->ssl));
    }

  return NULL;
}

JNIEXPORT jobject JNICALL
Java_com_transvirtual_protocol_ssl_SslIO_initInstance0(JNIEnv* env, jobject this,
						       jobject ctx_obj)
{
  char *strErr = 0;
  struct SslIONative *internal;

  jclass clazz;
  jfieldID ctxIntID;
  SSL_CTX *ctx;

  if(0 == (internal = (struct SslIONative*) malloc (sizeof (struct SslIONative))))
    {
      strErr = "malloc failed";
      goto err;
    }

  if(0 == (clazz = (*env)->GetObjectClass(env, ctx_obj)))
    {
      strErr = "GetObjectClass(env, ctx_obj) failed";
      goto err;
    }

  if(0 == (ctxIntID = (*env)->GetFieldID(env, clazz, "m_internal", "Lkaffe/util/Ptr;")))
    {
      // this thing throws already
      goto cleanup;
    }

  if(0 == (ctx = (*env)->GetObjectField(env, ctx_obj, ctxIntID)))
    {
      strErr = "GetObjectField(env, ctx_obj, ctxIntID) failed";
      goto err;
    }

  if(0 == (internal->bio = BIO_new_ssl_connect(ctx)))
    {
      strErr = (char*)ERR_error_string(ERR_get_error(), NULL);
      goto err;
    }
  
  BIO_get_ssl(internal->bio, &internal->ssl);

  if(internal->ssl == 0)
    {
      strErr = (char*)ERR_error_string(ERR_get_error(), NULL);
      goto err;
    }

  // this is the default for BIO_new_ssl_connect
  SSL_set_connect_state(internal->ssl);
  SSL_set_mode(internal->ssl, SSL_MODE_AUTO_RETRY);
  SSL_set_verify(internal->ssl, SSL_VERIFY_PEER, xoe_ssl_verify_callback);


  if((*env)->ExceptionOccurred(env) != 0)
    {
      // some unexpected exception
      goto cleanup;
    }

  return (jobject)internal;

 err:;
  throwSSLRuntimeException(env, strErr);

 cleanup:;
  if(internal != 0)
    {
      if(internal->bio != 0)
	{
	  BIO_free_all(internal->bio);
	}
      free(internal);
    }

  return 0;
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslIO_connect0(JNIEnv* env, jobject this,
						  jobject internal, jstring addr, int port)
{
  struct CallbackInfo cb = {env, this};
  const char *s;
  int ret;

  s = (*env)->GetStringUTFChars(env, addr, 0);

  BIO_set_conn_hostname(((struct SslIONative*)internal)->bio, s);
  BIO_set_conn_int_port(((struct SslIONative*)internal)->bio, &port);
  BIO_set_nbio(((struct SslIONative*)internal)->bio, 0);
  BIO_set_nbio(SSL_get_rbio( ((struct SslIONative*)internal)->ssl ), 0);
  BIO_set_nbio(SSL_get_wbio( ((struct SslIONative*)internal)->ssl ), 0);
  SSL_set_ex_data(((struct SslIONative*)internal)->ssl, exDataIndex, &cb);
  ret = SSL_connect(((struct SslIONative*)internal)->ssl);

  if(ret != 1 && (*env)->ExceptionOccurred(env) == 0)
    {
      throwSSLException(env, (char*)ERR_error_string(SSL_get_error(((struct SslIONative*)internal)->ssl, ret), NULL));
    }

  (*env)->ReleaseStringUTFChars(env, addr, s);
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslIO_close0(JNIEnv* env, jobject this,
						jobject internal)
{
  struct CallbackInfo cb = {env, this};
  int ret;

  SSL_set_ex_data(((struct SslIONative*)internal)->ssl, exDataIndex, &cb);

  ret = SSL_shutdown(((struct SslIONative*)internal)->ssl);

  if(ret != 1 && (*env)->ExceptionOccurred(env) == 0)
    {
      throwSSLException(env, (char*)ERR_error_string(SSL_get_error(((struct SslIONative*)internal)->ssl, ret), NULL));
    }
}

/*
 * Read bytes from a file.
 */
JNIEXPORT jint JNICALL
Java_com_transvirtual_protocol_ssl_SslIO_read0(JNIEnv* env, jobject this,
					       jobject internal, jarray buffer,
					       jint offset, jint length)
{
  struct CallbackInfo cb = {env, this};
  char* buf;
  int s;
  
  buf = (*env)->GetByteArrayElements(env, buffer, 0);
  SSL_set_ex_data(((struct SslIONative*)internal)->ssl, exDataIndex, &cb);
  s = BIO_read(((struct SslIONative*)internal)->bio, buf+offset, length);
  (*env)->ReleaseByteArrayElements(env, buffer, buf, JNI_COMMIT);
  
  if(s < 0 && (*env)->ExceptionOccurred(env) == 0)
    {
      throwSSLException(env, (char*)ERR_error_string(ERR_get_error(), NULL));
    }

  return (s);
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslIO_write0(JNIEnv* env, jobject this,
						jobject internal, jarray buffer, 
						jint offset, jint length)
{
  struct CallbackInfo cb = {env, this};
  char* buf;
  int s;
  
  buf = (*env)->GetByteArrayElements(env, buffer, 0);
  SSL_set_ex_data(((struct SslIONative*)internal)->ssl, exDataIndex, &cb);
  s = BIO_write(((struct SslIONative*)internal)->bio, buf+offset, length);
  (*env)->ReleaseByteArrayElements(env, buffer, buf, JNI_ABORT);

  if(s < length && (*env)->ExceptionOccurred(env) == 0)
    {
      throwSSLException(env, (char*)ERR_error_string(ERR_get_error(), NULL));
    }
}
