package org.xoe.mime;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;

/**
  * Base Class for Multipart Mime support.  Contains all the parts of a mime attachment. 
  * (header, body, header parameter)
  *
  */
public class Entity
{  
  Hashtable headers = new Hashtable();
  Hashtable parameters = new Hashtable();

  BufferedInputStream body;
  static final String HARD_RETURN = "\r\n";
  String bodyString;


  /**
    * Sets the body of the mime part.
    *  
    * @param ibody body in the form of an InputStream
    */
  public void setBody (InputStream ibody)
  {
    this.body = new BufferedInputStream( ibody );
  }


  /**
    * Sets the body of the mime part.
    *  
    * @param sBody body in the form of a String
    */
  public void setBody (String sBody )
  {
    this.body = new BufferedInputStream( (InputStream)new ByteArrayInputStream( sBody.getBytes() ) ); 
  }

  /**
    * Retrieves the body of the mime part.
    *  
    * @return BufferedInputStream the body 
    */
  public BufferedInputStream getBody()
  {
    return this.body;
  }


  /**
    * Adds a header
    *  
    * @param name  header name
    * @param value value of the name 
    */
  public void addHeader (String name, String value)
  {
    //if this header has no value, convert it to an empty string    
    if( value != null )	  
      headers.put(name, value );
    else
      headers.put(name, "" );   
  }


  /**
    * Retrievs the parameters of a given header name
    *  
    * @param name the header name 
    *
    * @return a hash of all the parameters that match the given header name 
    */
  public Hashtable getHeaderParameters(String name)
  {
    return (Hashtable)parameters.get(name);
  }


  /**
    * Retrievs the value of a given header name
    *  
    * @param name the header name 
    *
    * @return the value of the given header name
    */
  public String getHeader(String name)
  {
    return (String)this.headers.get( name );
  }


  /**
    * Retrievs the value of a given parameter
    *  
    * @param headerName the header name 
    * @param paramName the name of the parameter
    *
    * @return the value of a given parameter
    */
  public String getHeaderParameterValue( String headerName, String paramName )
  {
    Hashtable paramValues = (Hashtable)this.parameters.get( headerName );
    return (String)paramValues.get( paramName );
  }


  /**
    * Adds a parameter
    *  
    * @param headerName the header name 
    * @param paramName the name of the parameter
    * @param paramValue the value of the parameter
    *
    */
  public void addParameter (String headerName, String paramName, String paramValue)
  {
    //if this parameter has no value, convert it to an empty string 
    if( paramValue == null )
      paramValue = "";

    if( this.parameters.get( headerName ) == null )
      {
	Hashtable paramHash = new Hashtable();
	paramHash.put( paramName, paramValue );
	this.parameters.put( headerName, paramHash );
      }
    else
      {
	Hashtable paramHash = (Hashtable)this.parameters.get( headerName );
	paramHash.put( paramName, paramValue );
      }
  }

  /**
    * Turns the entire entity to a correctly formatted string representation of this mime entity
    *
    * @return a formatted string
    */
  public String toString ()
  {
    try
      {
	return writeHeaders() + this.HARD_RETURN + writeBody ();
      }
    catch( IOException ioe )
      {
	ioe.printStackTrace();
      }
    return "";
  }

  /**
    * Gets the length of this mime entity as a string
    *
    * @return an integer representing the length of this mime entity
    * @exception IOException
    */
  public int getLength() throws IOException
  {
    return getHeaderLength() + getBodyLength();
  }

  /**
    * Gets the length of the header area as a string
    *
    * @return an integer representing the length of the header
    * @exception IOException
    */
  private int getHeaderLength() throws IOException
  {
    int leng = writeHeaders().length();    
    // add two for the "\r\n" that separates the header part from the body
    return leng + 2;
  }

  /**
    * Gets the length of the body as a string
    *
    * @return an integer representing the length of the body
    * @exception IOException
    */
  private int getBodyLength() throws IOException
  {
    return body.available();// + 4; 
    // add 4 bytes for double hard return after body is written.  
    //See writeBody() below
  }


  /**
    * Generates a string of just the header portion of this mime entity
    *
    * @return a string representation of the header portion
    * @exception IOException
    */
  public String writeHeaders() throws IOException
  {
    StringBuffer sb = new StringBuffer();
    for (Enumeration e = headers.keys(); e.hasMoreElements(); )
      {
	String headerName = (String) e.nextElement();
	String headerValue = (String) headers.get(headerName);
	sb.append( headerName + ": " + headerValue  );
	Hashtable paramHash = (Hashtable)this.parameters.get( headerName );
	if( paramHash != null )
	  {
	    Enumeration paramNames = paramHash.keys();
	    Enumeration paramValues = paramHash.elements();
	    while( paramNames.hasMoreElements() )
		sb.append( "; " + (String)paramNames.nextElement() + "=" + Entity.maybeQuote( (String)paramValues.nextElement() ) );	
	  } 
	sb.append( this.HARD_RETURN );
      }
    return sb.toString();
  }


  /**
    * Generates a string of just the body portion of this mime entity
    *
    * @return a string representation of the body portion
    * @exception IOException
    */
  private String writeBody () throws IOException
  {
    StringBuffer sb = new StringBuffer();
    if( body != null )
      {
	body.mark( body.available() + 1 );
	
	byte buf[] = new byte[ body.available() ];
	int bytes_read = body.read (buf, 0, body.available());
	if (bytes_read != -1)
	  {
	    sb.append ( new String( buf ) );
	    sb.append ( this.HARD_RETURN + this.HARD_RETURN );
	  }
	body.reset();
      }
    return sb.toString();
  }



  /**
    * Checks a given string if it needs to be quoted.  The string is checked (per RFC2045) and if 
    * it matches, it is quoted.
    *
    * @return a string representation of the header portion
    * @exception IOException
    */
  private static String maybeQuote (String str)
  {
    int strlen = str.length ();

    for (int i = 0; i < strlen; i++)
      {
	char ch = str.charAt (i);

	if (ch >= 0 && ch <= 31)
	  return "\"" + str + "\"";

	switch (ch)
	  {
	  case '(':
	  case ')':
	  case '<':
	  case '>':
	  case '@':
	  case ',':
	  case ';':
	  case ':':
	  case '\\':
	  case '"':
	  case '/':
	  case '[':
	  case ']':
	  case '?':
	  case '=':
	  case '\u007F':
	    return "\"" + str + "\"";
	  }
      }

    return str;
  }


  /**
    * It takes a BufferedReader of header info, and parses it into the hash
    *
    * @param buffReader the header in the form of a BufferedReader
    *
    * @return the BufferedReader passed into it, unchanged
    * @exception IOException
    */
  protected BufferedReader parseHeader( BufferedReader buffReader ) throws IOException
  {    
    String nextLine = buffReader.readLine();
    while( nextLine.indexOf( ':' ) != -1 )
      {
	System.err.println( "Multipart.parseInputStream(): parsing line [" + nextLine + "]" );

	String headerName = nextLine.substring( 0, nextLine.indexOf( ':' ) ).trim();
	String headerValue;
	if( nextLine.indexOf( ';' ) == -1 )
	  {
	    this.addHeader( headerName, nextLine.substring( nextLine.indexOf( ':' ) + 1 ).trim() );
	  }
	else
	  {
	    StringTokenizer st = new StringTokenizer( nextLine, ";" );
	    String firstToken = st.nextToken();
	    headerValue = firstToken.substring( firstToken.indexOf( ':' ) + 1 ).trim();
	    this.addHeader( headerName, headerValue );
	    
	    while( st.hasMoreTokens() )
	      {
		String nextToken = st.nextToken();
		if( nextToken.indexOf( '=' ) != -1 )
		  {
		    String paramName = nextToken.substring( 0, nextToken.indexOf( '=' ) ).trim();
		    String paramValue = nextToken.substring( nextToken.indexOf( '=' ) + 1 ).trim();   
		    this.addParameter( headerName, paramName, paramValue );
		  }
		else
		  {
		    this.addParameter( headerName, nextToken, null );
		  }
	      }
	  }    
	nextLine = buffReader.readLine();
      }
    return buffReader;
  }


}
