package org.xoe.mime;

import org.xoe.core.dom.XNode;
import org.xoe.core.xpath.XPath;
import org.xoe.core.dom.DOMlet;

import java.io.IOException;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.BufferedReader;
import java.util.Random;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Hashtable;



/**
 * Multipart Class 
 */
public class Multipart extends BodyPart
{
  static Random rand = new Random ();

  String subtype;
  String boundary;
  Vector bodyParts;

  public static final String HTTP_200_HEADER = "HTTP/1.0 200 OK";
  public static final String HTTP_500_HEADER = "HTTP/1.0 500";
  public static final String HTTP_SOAP_HEADER = "POST /soap/servlet/rpcrouter HTTP/1.0";


  /**
    * @param subtype the type added to the Header "Content-Type=multipart/YOUR_TYPE_HERE"
    * @param aBoundary the boundary string for this multipart
    *
    */
  public Multipart (String subtype, String aBoundary)
  {
    this.subtype = subtype;
    this.boundary = aBoundary;
    this.bodyParts = new Vector();

    addHeader ("Content-Type", "multipart/" + subtype);
    addParameter ("Content-Type", "boundary", boundary );
  }


  /**
    * @param InputStream an input stream that will be parsed into header and body parts
    *
    */
  public Multipart( InputStream inStream ) throws MultipartParsingException
  {
    //System.err.println( "creating Multipart object from InputStream..." );
    this.bodyParts = new Vector();
    parseInputStream( inStream );    
    //System.err.println( "finished creating Multipart object from InputStream..." );
  }

  /**
    * @param BodyPart add BodyPart object to this Multipart object
    *
    */
  public void addBodyPart (BodyPart part)
  {
    part.setParent (this);
    bodyParts.addElement (part);
  }

  /**
    * @return an integer represeting the number of BodyParts
    *
    */
  public int getBodyPartCount()
  {
    return this.bodyParts.size();
  }

  /**
    * @param int returns the BodyPart at this given position
    *
    * @return the BodyPart at the index
    *
    */
  public BodyPart getBodyPart (int index)
  {   
    if( bodyParts.size() <= index )
      return null;
    else
      return (BodyPart) bodyParts.elementAt (index);
  }



  /**
    * @param String returns the BodyPart represented by this String ID
    *
    * @return the BodyPart represented by the given String ID
    *
    */
  public BodyPart getBodyPart (String CID)
  {
    int bodyPartsLen = bodyParts.size ();

    for (int i = 0; i < bodyPartsLen; i++)
      {
	BodyPart bp = (BodyPart) bodyParts.elementAt (i);

	if (CID.equals (bp.getHeader ("Content-ID")))
	  {
	    return bp;
	  }
      }
    
    return null;
  }


  /**
    * Gets the boundary string
    * 
    * @return String the Boundary string
    *
    */
  public String getBoundary ()
  {
    return this.boundary;
  }

  /**
    * Sets the boundary string
    *
    * @param String the Boundary string
    *
    */
  public void setBoundary( String bound )
  {
    this.boundary = bound;
  }

  /**
   * Does the same thing as toString() except it skips the header in order 
   * to get an exact count of the body
   *
   * @return int exact count of the body
   */
  public int getPayloadLength() throws IOException
  {
    StringBuffer sb = new StringBuffer();

    int bpLength = bodyParts.size ();
    
    if (bpLength == 0)          // No body parts, so return
      {
	return 0;
      }
    
    String boundarySep = "--" + boundary;
    for (int i = 0; i < bpLength; i++)
      {
	sb.append ( boundarySep );
	sb.append ( this.HARD_RETURN );
	sb.append ((BodyPart)bodyParts.elementAt (i)).toString ();
      }
    sb.append ( this.HARD_RETURN );
    sb.append ( boundarySep + "--" );

    int counter = sb.toString().length();

    //System.err.println( "Multipart.getPayloadLength(): [" + counter + "]" );
    return counter;
  }

  /**
    * Generates the well-formed String representation of this mime Multipart
    *
    * @return String
    */
  public String toString ()
  {
    StringBuffer sb = new StringBuffer();
    sb.append ( super.toString() );
    
    int bpLength = bodyParts.size ();
    
    if (bpLength == 0)          // No body parts, so return
      {
	return sb.toString();
      }
    
    String boundarySep = "--" + boundary;
    for (int i = 0; i < bpLength; i++)
      {
	sb.append ( boundarySep );
	sb.append ( this.HARD_RETURN );
	sb.append ((BodyPart)bodyParts.elementAt (i)).toString ();
      }
    sb.append ( this.HARD_RETURN );
    sb.append ( boundarySep + "--" );
    
    return sb.toString();
    
  }

  /**
    * Randomly generates a boundary string
    *
    * @return String a boundary string
    */
  public static String createBoundary ()
  {
    StringBuffer sb = new StringBuffer (24);
    sb.append( "--=_Part_" ); //always begin a boundary with this so it gets quoted

    for (int i = 0; i < 15; i++) 
      {
	if ((i % 2) == 0)	// lowercase
	  {
	    sb.append ((byte) ((rand.nextInt () % 26) + 97));
	  }
	else			// uppercase
	  {
	    sb.append ((byte) ((rand.nextInt () % 26) + 65));
	  }
      }
    
    return sb.toString ();
  }


  /**
    * @param InputStream an input stream that will be parsed into Headers and BodyParts
    *
    */
  private void parseInputStream( InputStream inStream ) throws MultipartParsingException
  { 
    Reader reader = new InputStreamReader (inStream);
    BufferedReader buffReader = new BufferedReader( reader );

    try
      {
	String firstLine = buffReader.readLine();

	if( !firstLine.equals( HTTP_200_HEADER ) && !firstLine.equals( HTTP_SOAP_HEADER ) )
	  {
	    if( firstLine.startsWith( HTTP_500_HEADER ) )
	      throw new MultipartParsingException( "ERROR RESPONSE: [" + firstLine + "]" );
	    else
	      throw new MultipartParsingException( "Not a recognized header: [" + firstLine + "]" );
	  }
	else
	  {
	    parseHeader( buffReader );    	    
	  }
	  	

	//now try to read the body

	if( this.getHeader( "Content-Type" ).startsWith( "text/xml" ) )
	  {
	    int bodLength = new Integer( this.getHeader( "Content-Length" ) ).intValue();
	    char[] bod = new char[ bodLength ];
	    int bodRead = buffReader.read( bod, 0, bodLength );
	    //System.err.println( "Multipart.parseInputStream(): body read = [" + bodRead  + "]" );	   
	    while( bodRead < bodLength )
	      {
		//System.err.println( "Multipart.parseInputStream(): Whole body not read!!!" );
		bodRead += buffReader.read( bod, bodRead, bodLength - bodRead );
	      }
	    this.setBody( new String( bod ) );
	    //System.err.println( "Multipart.parseInputStream(): multipart = [" + this.toString() + "]" );
	  }	    
 	else if( this.getHeader( "Content-Type" ).startsWith( "multipart/related" ) )
	  {
	    Hashtable ctParameters = this.getHeaderParameters( "Content-Type" );
	    //check if the boundary parameter begins and ends with double-quotes
	    //if it does, strip them off
	    String bound = (String)ctParameters.get( "boundary" );
	    if( bound.startsWith( "\"" ) && bound.endsWith( "\"" ) )
	      bound = bound.substring( 1, bound.length() - 1 );
	    setBoundary( bound );		  
	    
	    //System.err.println( "BOUNDARY = [" + getBoundary() + "]" );	    
	    
	    int bodLength = new Integer( this.getHeader( "Content-Length" ) ).intValue();
	    //System.err.println( "Multipart.parseInputStream(): body length = [" + bodLength  + "]" );
	    char[] bod = new char[ bodLength ];	    
	    int bodRead = buffReader.read( bod, 0, bodLength );
	    //System.err.println( "Multipart.parseInputStream(): body read = [" + bodRead  + "]" );	    
	    while( bodRead < bodLength )
	      {
		//System.err.println( "Multipart.parseInputStream(): Whole body not read!!!" );
		bodRead += buffReader.read( bod, bodRead, bodLength - bodRead );
	      }

	    String wholeBody = new String( bod );
	    //System.err.println( "Multipart.parseInputStream(): whole body = [" + wholeBody  + "]" );
	    
	    int posFront = 0;
	    int posBack = 0;
	    while( posBack != -1 )
	      {
		posBack = wholeBody.indexOf( "--" + getBoundary(), posFront );
		//System.err.println( "Multipart.parseInputStream(): posFront = [" + posFront  + "]" );
		//System.err.println( "Multipart.parseInputStream(): posBack = [" + posBack  + "]" );
		  
		if( posBack != posFront && posBack != -1 )
		  {
		    String bodyString = wholeBody.substring( posFront, posBack );
		    //System.err.println( "Multipart.parseInputStream(): tokenized body = [" + bodyString  + "]" );
		    
		    //InputStream bodyStream = (InputStream)new ByteArrayInputStream( bodyString.getBytes() );
		    BodyPart bodyPart = new BodyPart( bodyString.trim() );
		    //bodyPart.parseHeader( new BufferedReader( (Reader)new InputStreamReader( bodyStream ) ) );		 
		    //bodyPart.setBody( bodyStream );
		    
		    this.addBodyPart( bodyPart );
		    
		    posFront = posBack;
		  }
		
		posFront += getBoundary().length();
		posFront += 2;
	      }
     	  }
	else
	  throw new MultipartParsingException( "Unknown Content-Type [" + this.getHeader( "Content-Type" ) + "]" );
	

	//inStream.close();
      }
    catch( IOException ioe )
      {	
	throw new MultipartParsingException( ioe.getMessage() );
      }
    catch( NumberFormatException nfe )
      {	
	throw new MultipartParsingException( nfe.getMessage() );
      }
  }


  /**
    * Transforms the Multipart to a DOMlet.  If there are no BodyParts, the default body is used.
    * If there are one or more BodyParts, only the first part is used.
    *
    * @return DOMlet 
    */
  public DOMlet toDOMlet()
  {    
    if( getBodyPartCount() == 0 )
      {
	// the 2nd param is printed out by LazyDOM on a parse error, pass in something appropriate
	DOMlet domlet = DOMlet.create( (InputStream)this.getBody(), "WholeResponse" ); 
	//Logger.log( this, "SoapMessage.send(): Entire resonse = [" + domletng() + "]" );          
	return domlet;  
      }
    else
      { //the first body part contains the soap xml response in a mulipart response
	DOMlet domlet = DOMlet.create( (InputStream)getBodyPart( 0 ).getBody(), "WholeResponse" ); 
	//Logger.log( this, "SoapMessage.send(): Entire resonse = [" + domletng() + "]" );       
	return domlet;
      }
  }

}
