#include <npapi.h>
#include <npupp.h>                                                            
#include <fgl.h>

#include <unistd.h>
#include <errno.h>
#include <signal.h>

#define FATAL(format, arg...)  fprintf( stderr, "FATAL:" format, ## arg); exit(1)
#define DBG(format, arg...)  fprintf( stderr, "DBG:" format, ## arg)

//extern int jthreadedFileDescriptor(int fd);

typedef ssize_t (*writefunc) (int,void*,size_t);

typedef struct _PPlugin {
  int   filedes[2];
  fgl_PClient_t *pClient;
  char  pipe_buf[4096];
  writefunc write;
  pid_t pid_dec;
} PPlugin_t;

/*****************************
 * convenience functions
 *****************************/
void player_updateStatus( NPP instance, char *msg)
{
  NPN_Status(instance, msg);
}

int player_open_pipes( PPlugin_t *plugin, char *argf) {
  int flags;

  if ( pipe( plugin->filedes) == -1 ) {
    FATAL( "failed to open pipes\n");
    return 0;
  }

  //set pipe handle async
  //plugin->filedes[1] = jthreadedFileDescriptor( plugin->filedes[1]);
  flags = fcntl( plugin->filedes[1], F_GETFL, 0);
  fcntl( plugin->filedes[1], F_SETFL, flags | O_NONBLOCK | O_ASYNC);
  sprintf( argf, "@%d", plugin->filedes[0] );

  DBG( "Plugin opened pipes: %d %d\n", plugin->filedes[0], plugin->filedes[1]);

  return 1;
}

void player_close_pipes( PPlugin_t *plugin) {
  int i;

  for ( i=0; i<2; i++ ) {
    if ( plugin->filedes[i] != -1 ){
      DBG( "Plugin closed pipe: %d\n", plugin->filedes[i]);
      close( plugin->filedes[i]);
      plugin->filedes[i] = -1;
    }
  }

}

void wait_state( fgl_PClient_t *pCli, u8 key, u16 value, u32 timeout) {
  u32 ut = timeout;
  do {
    if ( pCli->states[key] == value ) {
      return;
    }
    usleep( 10000);
    ut -= 10000;
  }while ( ut > 10000 );

  DBG( "wait state timed out: %d\n", timeout );
}

void player_suspend (NPP instance, PPlugin_t *plugin)
{
  fgl_PClientSetValue( plugin->pClient, PCS_SUSPEND, 1);
  wait_state( plugin->pClient, PCS_FLUSHING, 0, 100000);

  player_updateStatus( instance, "Suspended");

}

void player_resume ( NPP instance, PPlugin_t *plugin)
{
  fgl_PClientSetValue( plugin->pClient, PCS_SUSPEND, 0);
  player_updateStatus( instance, "Resumed");
}

void player_stop ( NPP instance, PPlugin_t *plugin)
{
  DBG( "killing decoder %s\n", DECODER);
  plugin->pClient->states[PCS_CONNECTED] = 0;
   
  if ( kill( plugin->pid_dec, SIGKILL) == -1 )
     {
	DBG( "failed to kill decoder reason: %d\n", errno);
     }
   
  DBG( "decoder done\n");

  player_updateStatus( instance, "Stopped");
}

int player_run_decoder( NPP instance, char *file)  {
  pid_t pid;
  PPlugin_t *plugin;
  char argf[255];

  plugin = (PPlugin_t*)instance->pdata;

  if ( file ) {
    strcpy( argf, file); 
  }
  else {
    player_open_pipes( plugin, argf);
  }

  plugin->pClient = fgl_PServerOpenConnection();
 
  DBG( "Plugin run_decoder: %s client: %p\n", DECODER, plugin->pClient);

  if ( (pid = fork() ) == 0 ) {
    /* not used by child */
    if ( file == 0 ) {
      close( plugin->filedes[1] );
    }

    if ( -1 == execlp( DECODER_PATHNAME, DECODER_ARGS( argf))) {
      FATAL( "failed to start decoder: %s (reason: %d)\n", DECODER, errno);
    }
    exit(0);
  }
   
  /* not used by parent */
  if ( file == 0 ) {
    close( plugin->filedes[0]);
  }

  plugin->pid_dec = pid;

  player_updateStatus( instance, "decoder started");
   
  return 0;
}


/**************************************************
 * public plugin interface
 *************************************************/

#ifdef __cplusplus
extern "C"
{
#endif
   
NPError
NPP_Destroy (NPP instance, NPSavedData **save)
{
  PPlugin_t *plugin = (PPlugin_t*)instance->pdata;  
  DBG( "NPP_Destroy: %p\n", plugin);

  free( plugin);
  instance->pdata = NULL;
   
  return NPERR_NO_ERROR;
}

NPError
NPP_DestroyStream (NPP instance, NPStream *stream, NPReason reason)
{
  PPlugin_t *plugin = (PPlugin_t*)instance->pdata;
  DBG( "NPP_DestroyStream: %p\n", plugin->pClient);

  player_close_pipes( plugin);

  if ( reason != NPRES_DONE ) {
    //disable rendering immediately...
    player_suspend( instance, plugin);

    //and wait to be stopped
    player_stop( instance, plugin);
  }


  return NPERR_NO_ERROR;
}

jref
NPP_GetJavaClass ()
{
  return NULL;
}

NPError
NPP_GetValue (void* instance, NPPVariable variable, void *value)
{
  NPError err = NPERR_NO_ERROR;
    
  switch (variable) {
  case NPPVpluginNameString:
    *((char **)value) = DECODER;
    break;
  case NPPVpluginDescriptionString:
    *((char **)value) = DECODER_DESCRIPTION;
    break;
  default:
    err = NPERR_GENERIC_ERROR;
  }
  return err;
}

NPError
NPP_Initialize ()
{
  return NPERR_NO_ERROR;
}

NPError
NPP_New (NPMIMEType pluginType, NPP instance, uint16 mode, int16 argc, char *argn[], char *argv[], NPSavedData *saved)
{
  
  PPlugin_t *plugin = (PPlugin_t*)calloc( 1, sizeof( PPlugin_t)); 
  
   plugin->write = (writefunc)saved->buf; 
   //plugin->write = write;
   instance->pdata = plugin;

  plugin->filedes[0] = -1;
  plugin->filedes[1] = -1;

  return NPERR_NO_ERROR;
}

NPError
NPP_NewStream (NPP instance, NPMIMEType type, NPStream *stream, NPBool seekable, uint16 *stype)
{ 
  SET_STYPE(stype);

  switch ( *stype) {
  case NP_SEEK:
    DBG( "NPP_NewStream: setting SEEK mode\n");
    //no supported yet
    return NPERR_GENERIC_ERROR; 
  case NP_NORMAL:
    DBG( "NPP_NewStream: setting NORMAL mode\n");
    player_run_decoder( instance, NULL);
    break;
  case NP_ASFILEONLY:
    DBG( "NPP_NewStream: setting ASFILEONLY mode\n");
    //player will be started on NPP_StreamAsFile
    break;
  }

  return NPERR_NO_ERROR;
}


void
NPP_Print (NPP instance, NPPrint *PrintInfo)
{
}

int16
NPP_HandleEvent (NPP instance, void *event)
{
   PPlugin_t *plugin = (PPlugin_t*)instance->pdata;
   NPEvent   *evt = (NPEvent*)event;
   u16 xoff, yoff;
   
   if ( plugin && plugin->pClient ) {
      xoff = (plugin->pClient->srv_bounds.width - plugin->pClient->src_width) / 2;
      yoff = (plugin->pClient->srv_bounds.height - plugin->pClient->src_height) / 2;
      fgl_PClientHandleEvent( plugin->pClient, evt->event, evt->p1-xoff, evt->p2-yoff); 
      return 1;
   }
   return 0;
}
   
NPError
NPP_SetValue (void *instance, NPNVariable variable, void *value)
{
  return NPERR_NO_ERROR;  
}

NPError
NPP_SetWindow (NPP instance, NPWindow *window)
{
  PPlugin_t *plugin = (PPlugin_t*)instance->pdata;
  DBG( "NPP_SetWindow: client %p %d,%d %dx%d\n", plugin->pClient,
       window->x, window->y, window->width, window->height);

  if ( plugin->pClient == NULL ) {
    DBG( "INVALID CLIENT\n");
    return NPERR_GENERIC_ERROR;
  }

  player_suspend( instance, plugin);

  fgl_PClientSetBounds( plugin->pClient, window->x, window->y,
			window->width, window->height);

  player_resume( instance, plugin);

  return NPERR_NO_ERROR;
}

void
NPP_Shutdown ()
{

}

void
NPP_StreamAsFile (NPP instance, NPStream *stream, const char *fname)
{
  player_run_decoder( instance, (char*)fname);
}

void
NPP_URLNotify (NPP instance, const char *url, NPReason reason, void *notifyData)
{
}

int32
NPP_Write (NPP instance, NPStream *stream, int32 offset, int32 len, void *buf)
{
  int wr;
  PPlugin_t *plugin = (PPlugin_t*)instance->pdata;

  if ( plugin->filedes[1] == -1 ) {
    return -1;
  }
  else {
    wr = (*plugin->write)( plugin->filedes[1], (char*)buf + offset, len);
    return wr;
  }
}

int32
NPP_WriteReady (NPP instance, NPStream *stream)
{
  PPlugin_t *plugin = (PPlugin_t*)instance->pdata;
  return sizeof( plugin->pipe_buf);
}

char*
NPP_GetMIMEDescription()
{
  return MIME_DESCRIPTION;
}

#ifdef __cplusplus
} //extern "C"
#endif
