/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: SoapDocument.java,v 1.13 2001/12/06 00:01:51 ben Exp $
 */ 
package org.xoe.soap;

import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XElement;
import org.xoe.core.content.ContentElement;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;

import java.util.Stack;
import java.util.Hashtable;
import java.net.*;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import org.xoe.core.util.XMLWriter;
import org.xoe.mime.Multipart;
import org.xoe.mime.BodyPart;
import org.xoe.mime.Util;



/**
  * This represents the XML portion of a Soap Message.
  *
  */
public class SoapDocument 
  extends XDocument
{
  private int counter = 0;
  private XElement currentMethodElement;
  private String methodElementEncodingStyle;
  private Hashtable mimeAttachments;
  private Stack mimeKeys;
  private URL url;
  private String methodID;

  private final XPathContext soapContext = new XPathContext ();


  /**
   * Constructor requires a URL for the Soap Server, because every Soap Document needs
   * to know of its destination.
   *
   * @param URL the URL of the Soap Server this document is destined for.
   *
   */
  public SoapDocument( URL aUrl )
  {
    soapContext.definePrefix (SoapConstants.ATTRIBUTE_PREFIX, SoapConstants.ENVELOPE_NS);

    this.url = aUrl;

    XElement env = (XElement)this.createElementNS( SoapConstants.ENVELOPE_NS, SoapConstants.ENVELOPE );      

    env.setAttribute( SoapConstants.XMLNS + SoapConstants.COLON + SoapConstants.XSD,		       
		      SoapConstants.XSD_URL );
    env.setAttribute( SoapConstants.XMLNS + SoapConstants.COLON + SoapConstants.XSI,	     
		      SoapConstants.XSI_URL );
    env.setAttribute( SoapConstants.XMLNS + SoapConstants.COLON + SoapConstants.ATTRIBUTE_PREFIX,
		      SoapConstants.ENVELOPE_NS );      

    this.appendChild( env );    
    XElement body = (XElement)this.createElementNS( SoapConstants.ENVELOPE_NS, SoapConstants.BODY );      
    env.appendChild( body );
  }


  /**
   * Utility method used to construct a Method Element to the Soap Document
   *
   * @param String the urn of the Soap Service to be called
   * @param String the name of the method being accessed
   * @param String the encoding style of the expected response
   * 
   */
  public XElement addMethodElement( String soapID, String methodName, String encodingStyle )
  {  
    this.methodID = soapID;
    XElement el = (XElement)this.createElementNS( soapID, methodName );        
    el.setAttribute( SoapConstants.ATTRIBUTE_PREFIX + SoapConstants.COLON + SoapConstants.ENCODING_STYLE , 
		     encodingStyle );   
    el.setAttribute( SoapConstants.XMLNS, soapID );   

    XElement body = (XElement) XPath.findFirstNode ( soapContext, this, SoapConstants.SOAP_BODY );
    body.appendChild (el);
  
    currentMethodElement = el;
    methodElementEncodingStyle = encodingStyle;
    return el;
  }


  /**
   * Utility method used to construct a Method Element to the Soap Document.  The added prefix parameter
   * is required for building Soap response messages.  See generateServerResponse in 
   * org.xoe.soap.server.SimpleSoapServerListener
   *
   * @param String the urn of the Soap Service to be called
   * @param String the name of the method being accessed
   * @param String the encoding style of the expected response
   * @param String the method prefix
   * 
   */
  public XElement addMethodElement (String soapID, String methodName, String encodingStyle, String prefix)
  {  
    this.methodID = soapID;
    XElement el = (XElement)this.createElementNS (soapID, prefix + ":" + methodName);
    el.setAttribute (SoapConstants.ATTRIBUTE_PREFIX + SoapConstants.COLON + SoapConstants.ENCODING_STYLE , 
		     encodingStyle);   
    el.setAttribute (SoapConstants.XMLNS + ":" + prefix, soapID);   

    XElement body = (XElement) XPath.findFirstNode (soapContext, this, SoapConstants.SOAP_BODY);
    body.appendChild (el);
  
    currentMethodElement = el;
    methodElementEncodingStyle = encodingStyle;
    return el;
  }

  /**
   * Accessor method for the element containing the SOAP method request.
   *
   * @return XElement the method element
   */
  public XElement getCurrentMethodElement()
  {
    return currentMethodElement;
  }


  /**
   * Accessor method for the encoding style of the method.
   *
   * @return String encoding style 
   */
  public String getMethodElementEncodingStyle()
  {
    return methodElementEncodingStyle;
  } 

  /** 
   *  Names of parameters in XOE Soap messages have the simplified
   *  format:  xxx1, xxx2, xxx3, where xxx = the value set in SoapConstants.PARAM_NAME
   *
   * @return String the generated param name
   */
  public String generateParamName()
  {
    this.counter++;
    return SoapConstants.PARAM_NAME + counter;
  }

  /**
   * Accessor method for all mime attachments associated with this document, in the form
   * of a java.util.Hashtable
   *
   * @return Hashtable a hash of Mime Attachments and their IDs 
   */
  public Hashtable getMimeAttachments()
  {
    if( this.mimeAttachments == null )
      this.mimeAttachments = new Hashtable();

    return this.mimeAttachments;
  }


  /**
   * Accessor method for all keys of the Mime Attachments associated with this document, in the
   * form of a java.util.Stack
   *
   * @return Stack mime keys
   */
  public Stack getMimeKeys()
  {
    if( this.mimeKeys == null )
      this.mimeKeys = new Stack();

    return this.mimeKeys;
  }

  /** 
   *  The URL of the Soap Server that this document is destined for.
   *
   */
  public URL getUrl()
  {
    return this.url;
  }

  /**
   * The unique id of the method on this document
   *
   * @return String method id
   */
  public String getMethodID()
  {
    return this.methodID;
  }

  /**
   * Constructs a well formed SOAP message, complete with HTTP header.
   *
   * @return String the Soap Document in a deliverable http-encoded String
   */
  public String toString()
  {
    System.err.println( "SoapDocument.toString(): called" );

    if( getMimeAttachments().isEmpty() )
      {
	//System.err.println( "SoapDocument.toString(): no mime attachments" );
	String docString =  XMLWriter.getXML (this.getDocumentElement());

	int mesgLength = docString.length();           
	String header = "";
	
	if (url != null)
	  {
	    header = 
	      "POST " + url.getFile() + " HTTP/1.0\r\n" +
	      "Host: " + url.getHost() + ":" + url.getPort() + "\r\n" +
	      //"Content-Type: text/xml; charset=utf-8\n" +
	      "Content-Type: text/xml\r\n" +
	      "Content-Length: " + mesgLength + "\r\n" +
	      "SOAPAction: \"\"\r\n\r\n";
	  }
	else
	  {
	    header = 
	      "HTTP/1.0 200 OK\r\n" +
	      "Content-Type: text/xml\r\n" +
	      "Content-Length: " + mesgLength + "\r\n\r\n";
	  }
	return header + docString;
      }
    else
      {
	System.err.println( "SoapDocument.toString(): with mime attachments" );
	try
	  {	   
	    String xmlPartID = Util.generateCID();	    
	    Multipart multipart = new Multipart( "related", Multipart.createBoundary() );
	    multipart.addParameter( "Content-Type", "type", "text/xml" );
	    multipart.addParameter( "Content-Type", "start", xmlPartID );
	    //add soap header stuff
	    if (url != null)
	      {
		multipart.addHeader( "Host", url.getHost() + ":" + url.getPort() );
	      }
	    multipart.addHeader( "SOAPAction", "\"\"" );
		
	    System.err.println( "SoapDocument.toString(): testing header [" + multipart.writeHeaders() + "]" );

	    BodyPart soapPart = new BodyPart();
	    String mesg = XMLWriter.getXML (this.getDocumentElement());
 
	    ByteArrayInputStream xmlStream = new ByteArrayInputStream( mesg.getBytes() );
 
	    soapPart.setBody( xmlStream );
	    soapPart.addHeader( "Content-Type", "text/xml" );
	    soapPart.addParameter( "Content-Type", "charset", "utf-8" );
	    soapPart.addHeader( "Content-Transfer-Encoding", "8bit" );
	    soapPart.addHeader( "Content-ID", "<" + xmlPartID + ">" );
	    soapPart.addHeader( "Content-Length", xmlStream.available()+"" );
	    multipart.addBodyPart( soapPart );
	    int bodyLengthSum = 0;

	    //System.err.println( "SoapDocument.toString(): number of Mime Keys = " + getMimeKeys().size() );	    
	    for( int i = 0; i < getMimeKeys().size(); i++ )
	      {
		String id = (String)getMimeKeys().elementAt( i );
		ContentElement contentEl = (ContentElement)getMimeAttachments().get( id );	   
		BodyPart bodyPart = new BodyPart();
		bodyPart.setBody( contentEl.getInputStream()  );

		//System.err.println( "SoapDocument.toString(): header Content-Type = [" + contentEl.getType().getMimetype() + "]" );
		//System.err.println( "SoapDocument.toString(): header Content-ID = [" + id + "]" );
		//System.err.println( "SoapDocument.toString(): header Content-Length = [" + contentEl.getInputStream().available() + "]" );

		bodyPart.addHeader( "Content-Type", contentEl.getType().getMimetype() );
		bodyPart.addHeader( "Content-Transfer-Encoding", "8bit" );
		bodyPart.addHeader( "Content-ID", "<" + id + ">" );
		bodyPart.addHeader( "Content-Length", contentEl.getInputStream().available()+"" );
		multipart.addBodyPart( bodyPart );	
		bodyLengthSum += bodyPart.getLength();
	      }

	    multipart.addHeader( "Content-Length", multipart.getPayloadLength() + ""  );
	    
	    String message = multipart.toString();
	    //System.err.println( "SoapDocument.toString(): multipart message = [" + message + "]" );
	    if (url != null)
	      {
		String soapHeader = "POST " + url.getFile() + " HTTP/1.0\r\n";
		return soapHeader + message;
	      }
	    else
	      {
		String soapHeader = "POST /soap/servlet/rpcrouter HTTP/1.0\r\n";
		return soapHeader + message;
	      }
	    
	  }
	catch( IOException ioe )
	  {
	    Logger.log( this, ioe.toString() );
	  }
	return null;    
      }
  }

}
