/*
 * Copyright  (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: SoapMessage.java,v 1.19 2001/12/06 00:01:51 ben Exp $
 */ 
package org.xoe.soap;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.io.*;
import java.net.*;
import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;

import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
//import org.xoe.core.util.DOMUtils;
//import org.xoe.core.util.XMLWriter;
import org.xoe.core.dom.XNode;
import org.xoe.core.dom.XElement;
import org.xoe.core.dom.DOMlet;
//import org.xoe.core.packages.PackageQuery;
import org.xoe.core.content.ContentElement;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPath;

import org.xoe.mime.Multipart;
import org.xoe.mime.BodyPart;
import org.xoe.mime.Util;
import org.xoe.mime.MultipartParsingException;

public class SoapMessage
{
  private final XPathContext soapContext = new XPathContext ();  
  private Multipart multipart;
  
  private String m_urn;
  private String m_methodName;
  private Stack m_parameters;
  private URL m_url;  
  private String m_encodingStyle;
  private int m_paramCounter;
  private SoapDocument m_baseDoc;
  private XElement m_methodEl;
  private boolean hasMimeContent = false;

  private Vector mimeAttachments;


  /**
    * This is the networking layer of the XOE SOAP system.  It contains all the 
    * information necessary to send a SOAP request to a SOAP server.
    *
    * This is the default constructor.
    *
    */
  public SoapMessage (String urn, String methodName, Stack parameters, URL url)
  {
    
    m_urn = urn;
    m_methodName = methodName;
    m_parameters = parameters;
    m_url = url;
    m_paramCounter = 0;
    m_encodingStyle = SoapConstants.STANDARD_ENCODING_STYLE_URL;    
  } 

  /**
    * This is the networking layer of the XOE SOAP system.  It contains all the 
    * information necessary to send a SOAP request to a SOAP server.
    *
    * This is the constructor with the optional encoding style, primarily used if the remote 
    * method is expected to return an xml document, rather than a primitive type.
    *
    */
  public SoapMessage (String urn, String methodName, Stack parameters, URL url, String encodingStyleURL)
  {
    
    m_urn = urn;
    m_methodName = methodName;
    m_parameters = parameters;
    m_url = url;
    m_encodingStyle = encodingStyleURL;
    m_paramCounter = 0;
  }


  /**
    * This is the networking layer of the XOE SOAP system.  It contains all the 
    * information necessary to send a SOAP request to a SOAP server.
    *
    * This is the constructor with the optional Vector object. It's designed so that 
    * the vector is sent empty and returns populated with whatever objects the remote
    * method places in it.
    *
    */
  public SoapMessage (String urn, String methodName, Stack parameters, URL url, Vector vec)
  {
    
    m_urn = urn;
    m_methodName = methodName;
    m_parameters = parameters;
    m_url = url;
    m_paramCounter = 0;
    m_encodingStyle = SoapConstants.STANDARD_ENCODING_STYLE_URL;  
    this.mimeAttachments = vec;
  }


  /**
    * This is the networking layer of the XOE SOAP system.  It contains all the 
    * information necessary to send a SOAP request to a SOAP server.
    *
    * This constructor is used to handle incoming soap messages.  This is primarily used
    * by the lightweight soap server.
    *
    */
  public SoapMessage (InputStream inStream)
  {
    try
      {
	this.multipart = new Multipart (inStream);
	if (multipart.getBodyPartCount () > 1) //skip the first attachment, because it contains the soap xml response vector
	  {
	    this.mimeAttachments = new Vector ();
	     
	    for (int i = 1; i < multipart.getBodyPartCount (); i++)  //nothing will happen if there are no body parts
	      {
		this.mimeAttachments.addElement (multipart.getBodyPart (i));
	      }
	  }
      }
    catch (MultipartParsingException mpe)
      {
	mpe.printStackTrace ();	
      } 
  }



  /**
    * This is the networking layer of the XOE SOAP system.  It contains all the 
    * information necessary to send a SOAP request to a SOAP server.
    *
    * This constructor is used to handle incoming soap messages with mime attachments.  
    * This is primarily used by the lightweight soap server.
    *
    */
  public SoapMessage (InputStream inStream, Vector vec)
  {
    this.mimeAttachments = vec;
    try
      {
	this.multipart = new Multipart (inStream);
	if (multipart.getBodyPartCount () > 1) //skip the first attachment, because it contains the soap xml response vector
	  {   
	    for (int i = 1; i < multipart.getBodyPartCount (); i++)  //nothing will happen if there are no body parts
	      {
		this.mimeAttachments.addElement (multipart.getBodyPart (i));
	      }
	  }
      }
    catch (MultipartParsingException mpe)
      {
	mpe.printStackTrace ();	
      } 
  }

  /**
    * This prepares the message for sending and then sends it.  It also extracts the body of the response and
    * returns it as an XNode.
    *
    * @return XNode the response from the remote method via the SOAP server
    */
  public XNode sendMessage ()
    throws SoapServiceException
  {
 
    //create static stuff	
    m_baseDoc = new SoapDocument (m_url);

    //create the dynamic info....
    m_methodEl = m_baseDoc.addMethodElement (m_urn, m_methodName, m_encodingStyle);

    SoapParameterFactory paramFactory = SoapParameterFactory.getInstance (); 

    while (!m_parameters.empty ()){
      Object obj = m_parameters.pop ();
      SoapParameter soapParameter = paramFactory.createSoapParameter (obj);
      soapParameter.generate (m_baseDoc);      
    }

    this.multipart = send (m_baseDoc);
 
    if (this.multipart.getBodyPartCount () > 1) //skip the first attachment, because it contains the soap xml response vector
      {
	if (this.mimeAttachments != null)
	  {
	    for (int i = 1; i < this.multipart.getBodyPartCount (); i++)  //nothing will happen if there are no body parts
	      {
		this.mimeAttachments.addElement (this.multipart.getBodyPart (i));
	      }
	  }
      }
    //System.err.println ("SoapMessage.sendMessage () multipart returned = [" + multipart.toString () + "]");
    DOMlet domlet = this.multipart.toDOMlet ();   
    soapContext.definePrefix  ("SOAP-ENV", "http://schemas.xmlsoap.org/soap/envelope/");
    XNode aNode =  (XNode) XPath.findFirstNode (soapContext, domlet, "//SOAP-ENV:Envelope/SOAP-ENV:Body");     
    return aNode;
  }
 
  /**
    * Accessor method for the Multipart of the SOAP message. 
    *
    */
  public Multipart getMultipart ()
  {
    return this.multipart;
  }

  /**
    * Accessor method for the Mime Attachments of the SOAP message. 
    *
    */
  public Vector getMimeAttachments ()
  {
    return this.mimeAttachments;
  }

  /**
    * Accessor method for a particular Mime Attachment of the SOAP message. 
    *
    */
  public BodyPart getMimeAttachmentByID (String id)
  {
    if (id.startsWith ("cid:"))
      {
	id = id.substring (4);
      }
    return this.multipart.getBodyPart ("<" + id + ">");
  }


  /**
    * This is where the networking job gets done.  This method blocks until it receives a response from the 
    * SOAP server.  Then the response is parsed as a new Multipart object and returned.
    *
    * @param SoapDocument the xml content of this message
    * @return Multipart the SOAP response object
    */
  private Multipart send (SoapDocument soapDocument)
    throws SoapServiceException
  {
    
    try {
      
      Logger.log (this, "Connecting: " + soapDocument.getUrl ().getHost () + ":" + soapDocument.getUrl ().getPort ());
      Socket sock = new Socket (soapDocument.getUrl ().getHost (), soapDocument.getUrl ().getPort ());
      
      OutputStream outStream = sock.getOutputStream ();
      InputStream  inStream = sock.getInputStream ();
      
      BufferedOutputStream bOutStream = new BufferedOutputStream (outStream);
      
      Logger.log (this, "SoapMessage.send (): conveting soapDocument to byte stream....");
      String soapString = soapDocument.toString ();
      Logger.log (this, "SoapMessage.send (): soapString length = " + soapString.length ());
      byte[] soapArray = soapString.getBytes ();
      bOutStream.write (soapArray);
      bOutStream.flush ();
      outStream.flush ();
      
      Logger.log (this, "reading soap response...");
          
      try
	{
	  return new Multipart (inStream);
	}
      catch (MultipartParsingException mpe)
	{
	  mpe.printStackTrace ();
	  return null;
	}      
      
    }
    catch  (IOException ioe) {
      ioe.printStackTrace ();
      throw new SoapServiceException (ioe.getMessage ());
      //return null;
    }
  }

  


}

