/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: SoapParameterFactory.java,v 1.3 2001/12/06 00:01:51 ben Exp $
 */ 
package org.xoe.soap;

import java.lang.InstantiationException;
import java.lang.IllegalAccessException;
import java.lang.ClassNotFoundException;

/**
 * Using a Singleton design pattern, the SoapParameterFactory is used to create
 * SoapParameter objects derived from their appropriate implementations.  The Factory
 * uses reflection to determine which implementation of the SoapParameter interface to use.
 *
 */
public class SoapParameterFactory
{

  static SoapParameterFactory instance;
  
  /**
   *  Singleton constructor
   *
   */
  private SoapParameterFactory()
  {
    instance = this;
  }

  /**
   * Public accessor for obtaining a handle to the factory.
   *
   * @return SoapParameterFactor this Factory
   */
  public static SoapParameterFactory getInstance()
  {
    if( instance == null )
      return new SoapParameterFactory();
    else
      return instance;
  }

  /**
   * This method does the reflection magic.  Take an object, look up its class name, 
   * add it to "SoapParameter", create an instance of the implementation that 
   * matches that name and return it.
   * 
   * @param Object any object
   * @return SoapParameter the implementation of the SoapParameter interface
   * @exception InstantiationException Raised if implementation could not be instantiated
   * @exception IllegalAccessException Raised if access to implementation is not allowed
   * @exception ClassNotFoundException Raised if implementation does not exist
   *
   */
  public SoapParameter createSoapParameter( Object obj )
  {

    Class cs = obj.getClass();
    String className = cs.getName();

    //hack to obtain the fundamental name of this object without the package info
    //i.e. String instead of java.lang.String
    className = className.substring( className.lastIndexOf( '.' ) + 1 );


    String fullClassName = "org.xoe.soap." + className + "SoapParameter";
    try
    {
	Class classDef = Class.forName( fullClassName );
	SoapParameter soapParameter = (SoapParameter)classDef.newInstance();
	soapParameter.setParameter( obj );

	return soapParameter;
    }	
    catch (InstantiationException e)
      {
	e.printStackTrace();
      } 
    catch (IllegalAccessException e)
      {
	e.printStackTrace();
      } 
    catch (ClassNotFoundException e)
      {
	e.printStackTrace();
      }
    return null;
	
  }
}
