/*
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Map cooridnated to char or special function keys.
 *
 * $tvt: VirtualKeymap.java,v 1.3 2001/12/08 00:45:33 damian Exp $
 */
package org.xoe.input.virtual;

import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.net.URL;
import java.net.MalformedURLException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;

/**
 * Creates VirtualKeymap to map keys on the VirtualKeyboard
 *
 */

public class VirtualKeymap 
{
  public static final int DEFAULT = 0;

  public static final int PEN = 1;
  public static final int MAP = 2;

  public static final int HIRAGANA = 3;
  public static final int KATAKANA = 4;
  public static final int KANJI    = 5;

  public static final int KEYMAP_OFFSET    = 0xf8ff;
  public static final int COMMAND_OFFSET   = 0xf800;
  public static final int COMMAND_SIZE     = 2;
  public static final int CHARACTER_OFFSET = 0xe000;

  public static final int SHIFT = 0;
  public static final int CLOSE = 1;

  public int type;
  public int mode;

  private Image image;
  private char[][] keymap;
  private String name;
  private String imageName;
  public String className;
  private int shiftKeyboard = -1;
  private String dir;
  private int width;
  private int height;
  private Toolkit toolkit;
  private URL docUrl;

  private static MediaTracker mediatracker = new MediaTracker (null);
  private static int nextID = 0;
  private int currentId;

  /**
   * Default constructor
   * @param map     map description
   * @param setId   associate keymap with id id
   *
   * @return: VirtualKeymap
   * @author: Damian Kowalewski
   */

  public VirtualKeymap (Element map, int setId)
  {
    String attr;

    Document doc = map.getOwnerDocument();
    URL docUrl = ((XDocument)doc).getURL();

    this.toolkit = Toolkit.getDefaultToolkit();
    this.currentId = setId;

    NodeList list = XPath.findNodes (doc,"/virtualkeyboard/map");
    String[] keyboards = new String[list.getLength()];
    for (int i = 0 ; i < list.getLength() ; i++)
      {
        keyboards[i] = ((Element)list.item(i)).getAttribute("name");
      }

    attr = map.getAttribute ("width");
    if ( attr != null )
      {
        this.width = Integer.parseInt (attr);
      }
    else
      {
        this.width = 240;
      }

    attr = map.getAttribute ("height");
    if ( attr != null )
      {
        this.height = Integer.parseInt (attr);
      }
    else
      {
        this.height = 95;
      }

    attr = map.getAttribute ("type");
    

    type = MAP;

    if (attr != null && attr.toUpperCase().equals ("PEN")) 
      {
        type = PEN;
      }

    attr = map.getAttribute ("mode");

    mode = DEFAULT;

    if (attr != null) 
      {
        attr = attr.toUpperCase ();
        if (attr.equals ("HIRAGANA")) { mode = HIRAGANA; }
        if (attr.equals ("KATAKANA")) { mode = KATAKANA; }
        if (attr.equals ("KANJI"))    { mode = KANJI;    }
      }

    className = map.getAttribute ("class");

    /* 
     * get the image for the keyboard.
     * [FIXME] this is path hardcoded.
     */
    this.name = map.getAttribute ("name");		
    this.imageName = map.getAttribute ("image");
    
    try
      {
        image = toolkit.getImage ( new URL (docUrl, "images/"+imageName));
      }
    catch (MalformedURLException e)
      {
        Logger.log (this, e);
      }

    mediatracker.addImage (image, currentId);
      
    String shiftName = map.getAttribute ("shift");
    if (shiftName != null) 
      {
        for (int i = 0; i < keyboards.length; i++) 
          {
            if (keyboards[i].equals (shiftName)) 
              {
                shiftKeyboard = i;
                break;
              }
          }
      }

    /* initialize the keymap */
    this.keymap = new char[this.width][this.height];

    /* create the keymap */
    for (Element n = (Element) map.getFirstChild (); 
        n != null ; n = (Element) n.getNextSibling ()) 
      {
        char value = 0;
        attr = n.getAttribute ("action");
        if (attr != null) 
          {
            String attrUpperCase = attr.toUpperCase ();
            if (attrUpperCase.equals ("SHIFT")) 
              {
                value = COMMAND_OFFSET - SHIFT;
              }
            else if (attrUpperCase.equals ("CLOSE")) 
              {
                value = COMMAND_OFFSET - CLOSE;
              }
            if (attrUpperCase.substring (0,5).equals ("GOTO:"))
              {
                String next = attr.substring(5);
                for (int i = 0; i < keyboards.length ; i++) 
                  {
                    if (keyboards[i].equals (next)) 
                      {
                        value = (char) (VirtualKeymap.KEYMAP_OFFSET - i);
                        break;
                      }
                  }
              } 
          }
        else 
          {
            attr = n.getAttribute ("char");
            if (attr != null) 
              {
                value = attr.charAt (0);
              }
          }

        // Get the coordinates of the key

        int x0 = Integer.parseInt (n.getAttribute ("x0"));
        int y0 = Integer.parseInt (n.getAttribute ("y0"));
        int x1 = Integer.parseInt (n.getAttribute ("x1"));
        int y1 = Integer.parseInt (n.getAttribute ("y1"));

        // set the key in the keymap
        for (int y = y0; y <= y1; y++) 
          {
            for (int x = x0; x <= x1; x++) 
              {
                keymap[x][y] = value;
              }
          }
      }						
  }

  /**
   * Get the name of the class associated with this keymap
   *
   * @return: String
   * @author: Damian Kowalewski
   */

  public String getClassName () 
  {
    return className;
  }

  /**
   * Maps the point to a character
   * @param x   the x coordinate
   * @param y   the y coordinate
   *
   * @return: char
   * @author: Damian Kowalewski
   */

  public char mapKey (int x, int y) 
  {
    if (x >= 0 && x < width && y >=0 && y < height) 
      {
        return keymap[x][y];
      }
    else { return 0; }
  }

  /**
   * Get the image associated with this keymap
   *
   * @return: java.awt.Image
   * @author: Damian Kowalewski
   */

  public Image getImage ()
  {
    /* we have to delay loading of the image until the theme
     * got loaded, otherwise we get a deadlock
     */
    if (image == null) 
      {
        try
          {
            image = toolkit.getImage ( new URL (docUrl, "images/"+imageName));
          }
        catch (MalformedURLException e)
          {
            Logger.log (this, e);
          }
      }
    return image;
  }

  /**
   * Returns the index of shift VirtualKeymap associated with this one
   * @return: int
   * @author: Damian Kowalewski
   */
   
  public int getShift () 
  {
    return shiftKeyboard;
  }

  /**
   * Get new free id to associate with the images.
   * @return: int
   * @author: Damian Kowalewski
   */

  public static synchronized int getNewId ()
  {
        return nextID++;
  }

  /**
   * Wait for all the images associated with an id to be loaded
   * @param id    id associated with te images
   * 
   * @author: Damian Kowalewski
   */

  public static void waitForImages (int id)
  {
    try
      {
        mediatracker.waitForID (setId);
      }
    catch (InterruptedException _)
      {
      }
  }

  /**
   * Get current id to associateid with the images.
   * @return: int
   * @author: Damian Kowalewski
   */

  public int getID ()
  {
    return this.currentId;
  }
}
