package org.xoe.display;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.xoe.core.Constants;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DefaultContentTypeResolver;
import org.xoe.core.content.Stash;
import org.xoe.core.packages.XoePackage;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.core.util.ProviderHashtable;
import org.xoe.services.IIconResolver;


public class IconManager
  implements Constants
{
  static final String DEFAULT_FILE_ICON = "file.gif";
  static final String DEFAULT_DIRECTORY_ICON = "dir.gif";

  private static DefaultContentTypeResolver m_instance = null;
  private static ProviderHashtable m_hash = new ProviderHashtable ();

  // this is the unique provider object used to represent user-specified icons
  private static Object m_userProvider = new Object ();
  
  static
  {
    // user options take precedence over those set by packages
    m_hash.setOverridingProvider (m_userProvider);
    setIcon (MIMETYPE_STASHDIR, getDefaultDirectoryIcon (), null);
  }

  public static URL getDefaultIcon ()
  {
    try
      {
	return (new URL (Stash.getWorkingDirectory (IconManager.class), DEFAULT_FILE_ICON));
      }
    catch (MalformedURLException ex)
      {// Don't ever expect to see this
	Logger.log (IconManager.class, ex);
	return (null);
      }
  }


  public static URL getDefaultDirectoryIcon ()
  {
    try
      {
	return (new URL (Stash.getWorkingDirectory (IconManager.class), DEFAULT_DIRECTORY_ICON));
      }
    catch (MalformedURLException ex)
      {// Don't ever expect to see this
	Logger.log (IconManager.class, ex);
	return (null);
      }
  }

  public static void unsetIcon (String suffixOrMimetype)
  {
    unsetIcon (suffixOrMimetype, m_userProvider);
  }

  public static void unsetIcon (String suffixOrMimetype, XoePackage pkg)
  {
    unsetIcon (suffixOrMimetype, (Object)pkg);
  }

  private static void unsetIcon (String suffixOrMimetype, Object provider)
  {
    // removes the '.' if there is one
    if(suffixOrMimetype.charAt(0) == '.')
      {
	suffixOrMimetype = suffixOrMimetype.substring(1);
      }

    m_hash.remove (suffixOrMimetype, provider);
  }

  public static void removeProvider(XoePackage pkg)
  {
    m_hash.removeProvider (pkg);
  }

  public static void setIcon (String suffixOrMimetype, URL icon)
  {
    setIcon (suffixOrMimetype, icon, m_userProvider);
  }

  public static void setIcon (String suffixOrMimetype, URL icon, XoePackage pkg)
  {
    setIcon (suffixOrMimetype, icon, (Object)pkg);
  }

  private static void setIcon (String suffixOrMimetype, URL icon, Object provider)
  {
    // removes the '.' if there is one
    if (suffixOrMimetype.charAt (0) == '.')
      {
	suffixOrMimetype = suffixOrMimetype.substring (1);
      }

    m_hash.put (suffixOrMimetype, icon, provider);
  }

  private static IService[] getCustomHandlers (ContentType type)
  {
    SimpleServiceQuery q = new SimpleServiceQuery (IIconResolver.FUNCTIONALITY,
						   IIconResolver.INTERFACE);

    q.addFeature (FEATURE_MIMETYPE, type.getMimetype ());
    return ServiceLocator.findServices (q);
  }

  public static URL getIconURL (ContentElement resource)
    throws IOException
  {
    ContentType type = resource.getType ();
    IService[] handlers = getCustomHandlers (type);
    URL iconURL = null;

    for (int i = 0; handlers != null && i < handlers.length; i++)
      {
	// 1) allow a custom handler to determine the icon
	//    (useful for text/vnd.tvt.lnk and text/xml)
	iconURL = ((IIconResolver)handlers[i]).getIconURL (resource);
	if (iconURL != null)
	  {
	    return (iconURL);
	  }
      }

    // 2) return the icon for this mimetype
    iconURL = (URL)m_hash.get (type.getMimetype ());
    if (iconURL != null)
      {
	return (iconURL);
      }

    // 3) return the icon for this suffix
    iconURL = (URL)m_hash.get (type.getSuffix());
    if (iconURL != null)
      {
	return (iconURL);
      }

    // 4) return default file icon
    return (getDefaultIcon ());
  }

  public static URL getIconURL (ContentType type)
  {
    IService[] handlers = getCustomHandlers (type);
    URL iconURL = null;

    for (int i = 0; handlers != null && i < handlers.length; i++)
      {
	// 1) allow a custom handler to determine the icon
	//    (useful for text/xml, which has additional info in the ContentType)
	iconURL = ((IIconResolver)handlers[i]).getIconURL (type);
	if (iconURL != null)
	  {
	    return (iconURL);
	  }
      }

    // 2) return the icon for this mimetype
    iconURL = (URL)m_hash.get (type.getMimetype ());
    if (iconURL != null)
      {
	return (iconURL);
      }

    // 3) return the icon for this suffix
    iconURL = (URL)m_hash.get (type.getSuffix ());
    if (iconURL != null)
      {
	return (iconURL);
      }

    // 4) return default file icon
    return (getDefaultIcon ());
  }
}
