package org.xoe.display.dom;

import java.awt.Component;
import java.awt.Image;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;

import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;


public class ImageFactory
{
  private static final int MAXLOADERS = 8;

  private static Hashtable cache = new Hashtable ();
  private static Stack pending = new Stack ();
  private static int inactive = MAXLOADERS;

  public static synchronized void getImage (String url, ImageHolder holder)
  {
    Object obj = cache.get (url);
    // Image in cache?  Deliver it now since it must be complete
    if (obj instanceof Image)
      {
	holder.imageDeliver (url, (Image)obj);
      }
    // Vector in cache?  A list of people waiting for the image.
    // Add myself so I'll get it later
    else if (obj instanceof Vector)
      {
	Vector vec = (Vector)obj;
	if (!vec.contains (holder))
	  {
	    vec.addElement (holder);
	  }
      }
    else if (obj instanceof ImageThread)
      {
	Vector vec = ((ImageThread)obj).vec;
	if (!vec.contains (holder))
	  {
	    vec.addElement (holder);
	  }
      }
    // Otherwise no ones getting this image yet - so we'd better schedule
    // it.
    else
      {
	Vector vec = new Vector ();
	vec.addElement (url);
	vec.addElement (holder);
	cache.put (url, vec);
	pending.addElement (vec);
	if (inactive > 0)
	  {
	    inactive--;
	    start ();
	  }
      }
  }

  public static synchronized void flush ()
  {
/* HACK HACK HACK
    cache.clear ();
*/
  }

  public static synchronized void stop ()
  {
    for (Enumeration e = cache.elements (); e.hasMoreElements (); )
      {
	Object obj = e.nextElement ();
	if (obj instanceof ImageThread)
	  {
	    ((Thread)obj).interrupt ();
	  }
      }
    cache.clear ();
    pending.removeAllElements ();
  }

  private static void start ()
  {
    ImageThread tid = new ImageThread ();
    tid.setPriority (Thread.MIN_PRIORITY);
    tid.start ();
  }

  private static Image fetchImage (String url)
    throws IOException
  {
    ContentElement content = ContentCache.fetch (url);
    if (content == null)
      {
	Logger.log (ImageFactory.class, "content not found: " + url);
	return (null);
      }
    SimpleServiceQuery query = new SimpleServiceQuery ("contenthandler", IImageBuilder.INTERFACE);
    query.addFeature ("mimetype", content.getType ().getMimetype ());
    IImageBuilder ib = (IImageBuilder)ServiceLocator.findService (query);
    if (ib == null)
      {
	Logger.log (ImageFactory.class, "image builder not found: " + content.getType ());
	return (null);
      }
    return (ib.getImage (content));
  }

  static class ImageThread
    extends Thread
  {
      Vector vec;

      public void run ()
      {
	for (;;)
	  {
	    String url = null;
	    try
	      {
		// Get some work to do ... if none terminate
		synchronized (ImageFactory.class)
		  {
		    if (pending.isEmpty ())
		      {
			inactive++;
			break;
		      }
		    vec = (Vector)pending.elementAt (0);
		    pending.removeElementAt (0);
		    url = (String)vec.elementAt (0);
		    cache.put (url, this);
	          }

		// Fetch the image
		Image img = fetchImage (url);
		// Someone interrupted me - throw IOException
		if (Thread.interrupted ())
		  {
		    throw new IOException ("interrupted " + url);
		  }

		// Put the fetched image into the cache and remove
		// ourself from inprogress
		synchronized (ImageFactory.class)
		  {
		    if (img == null)
		      {
		        cache.remove (url);
		      }
		    else if (cache.get (url) == this)
		      {
		        cache.put (url, img);
		      }
	          }

		// Dispatch image to anyone interested
		int l = vec.size ();
		for (int i = 1; i < l; i++)
		  {
		    ((ImageHolder)vec.elementAt (i)).imageDeliver (url, img);
	          }

		// Redraw if necessary
                if (redraw != null && !redraw.isValid())
                  {
                    Component r = redraw; // Because validating might change it!
                    r.validate();
                    r.repaint();
                  }

	      }
	    // If we get an exception (including getting interrupted) then
	    // remove the image from the cache
	    catch (IOException _)
	      {
		synchronized (ImageFactory.class)
		  {
		    if (url != null)
		      {
		        cache.remove (url);
		      }
		  }
	      }
	  }
      }
  }
}
