/*
 * Symbol management
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 * @author Peter Mehlitz
 * @author Philippe Laporte 
 * 
 * $tvt: symtab.c,v 1.1 2001/07/11 21:12:57 philippe Exp $
 */


#include <stdlib.h>
#include <string.h>

#include "symtab.h"
#include "xs_config.h"

#ifdef KAFFELIB  //if used as a kaffe jni lib

#include <jni.h>

#endif


/* very simple open hashing */

/***************************************************************************************************************
 * internal helpers
 */


static unsigned int hash ( char* sym )
{
	unsigned int h, g;

	for ( h=0; *sym; sym++ ){
		h = (h << 4) + *sym;
		if ( (g = h & 0xf0000000) ){
			h ^= (g >> 24);
			h = h ^ g;
		}
	}
	return h;
}

static inline int threshold ( int i ) {
	return (i * 80 / 100); 
}

static unsigned prime ( int n ) 
{
	if      ( n < 10 ) return 11;
	else if ( n < 15 ) return 17;
	else if ( n < 20 ) return 23;
	else if ( n < 30 ) return 31;	
	else if ( n < 45 ) return 47;
	else if ( n < 67 ) return 67;
	else if ( n < 100 ) return 101;
	else if ( n < 150 ) return 151;
	else if ( n < 200 ) return 199;
	else if ( n < 290 ) return 293; 
	else if ( n < 440 ) return 443;
        else if ( n < 670 ) return 673;
	else if ( n < 1010 ) return 1013; 
	else if ( n < 1530 ) return 1531;
	else if ( n < 2290 ) return 2297;
        else if ( n < 3440 ) return 3449;
	else if ( n < 5170 ) return 5179; 
	else if ( n < 7780 ) return 7789;
        else if ( n < 11680 ) return 11689;
        else if ( n < 17530 ) return 17539;
	else if ( n < 26300 ) return 26309; 
	else if ( n < 39490 ) return 39499;
	else if ( n < 59260 ) return 59263;
        else if ( n < 88890 ) return 88897;
	else if ( n < 133340 ) return 133349; 
	else{ 
	    printf("error: prime number overflow\n");
	    return 0;
	} 
}

static inline int nextIdx ( int i, unsigned hash, unsigned len )
{
	i += (16 - (hash & 0xf));
	return (i % len);
}

static inline int equals ( char* s1, char* s2 )
{
	for ( ; *s1 == *s2; s1++, s2++ ){
		if ( *s1 == 0 ) return 1;
	}
	return 0;
}

static void grow ( symtab_t* t )
{
      	int i,  j, k;
	int newLen = prime( t->len * 3 / 2);
	char** newSym = (char**) calloc( newLen, sizeof(char*));
	char* newType = (char*) calloc( newLen, sizeof(char*));
	AST** newNode = (AST**) calloc( newLen, sizeof(AST*));
	
	if(!newSym || !newType || !newNode){
	    fail("could not allocate memory in symtab grow");
	}
	
	
       // fprintf(stderr, "IN GROW\n");
	
	for ( j=0; j < t->len; j++ ) {
	       if ( t->sym[j] != NULL ) {
		        
			k = hash( t->sym[j]);
			i = k % newLen;
			
			do {
				if ( newSym[i] == NULL ){
					newSym[i] = t->sym[j];
					newType[i] = t->type[j];
					newNode[i] = t->node[j];
					break;
				}
		
				i = nextIdx( i, k, newLen);
			} while ( 1 );
		}
	}
			
	free( t->sym);
	free( t->type);
	free( t->node);
	
	t->thresh = threshold( newLen);
	t->sym = newSym;
	t->type = newType;
	t->node = newNode;
        t->len = newLen;
}


/***************************************************************************************************************
 * public APIs
 */

symtab_t* st_create ( int len, int isCopy )
{
	symtab_t *t = (symtab_t*) calloc( 1, sizeof( symtab_t));
	
	t->len    = prime(len * 3 / 2);
	t->thresh = threshold(t->len);
        t->sym    = (char**) calloc( t->len, sizeof(char*));
	t->type   = (char*) calloc(t->len, sizeof(char));
	t->node   = (AST**) calloc(t->len, sizeof(AST*));
	t->count  = 0;
	t->isCopy = isCopy;
	
	return t;
}

void st_reset ( symtab_t* t )
{
	int i;
	
	if ( t->count == 0 ){
		return; // nothing to do, it's empty anyway
	}
	
	for ( i=0; i < t->len; i++ ){
		if ( t->sym[i] ) {
			if ( t->isCopy ) {
				free( t->sym[i]);
			}
			t->sym[i]  = 0;
			t->type[i] = 0;
			t->node[i] = 0;
		}
	}
	
	t->count = 0;
}


void st_delete ( symtab_t* t )
{
	st_reset( t);
	free( t->sym);
	free( t->type);
	free( t->node);
}


void st_add ( symtab_t* t, char* sym, char type, AST* node)
{
       	unsigned k = hash( sym);
	unsigned i = k % t->len;
	
        if ( t->count + 1 > t->thresh ) {
	   grow( t);
	   i = k % t->len;
	}
//maybe I need to duplicate the nodes ???
	do {
		if ( t->sym[i] == 0 ) {
		        t->sym[i]  = t->isCopy ? StrDup(sym) : sym;
		        t->type[i] = type;
			t->node[i] = node;
		//	fprintf(stderr, "adding to %p, sym is %s, node is %p, at index %d\n", 
		//	                t, sym, node, i); 
			t->count++;
			return;
		}
		else if ( equals( sym, t->sym[i]) ){
			return; // already there
		}

		i = nextIdx( i, k, t->len);
		
	} while ( true );
	
	
}


int st_contains ( symtab_t* t, char* sym )
{
	unsigned k = hash( sym);
	unsigned i = k % t->len;
	
	do {
		if ( t->sym[i] == 0 ){ 
			return 0;
		}
	
		if ( equals( sym, t->sym[i]) ){
			return t->type[i];
		}
	
		i = nextIdx( i, k, t->len);
	} while ( true );
}


AST* st_contains_get_ast_node ( symtab_t* t, char* sym )
{
	unsigned k = hash( sym);
	unsigned i = k % t->len;
	
	do {
		if ( t->sym[i] == 0 ){
		      //  fprintf(stderr, "in %p, sym is %s, node is %p, at index %d\n", 
		//	                t, sym, 0, i); 
			return 0;
		}
	
		if ( equals( sym, t->sym[i]) ){
		      //  fprintf(stderr, "in %p, sym is %s, node is %p\n", t, sym, t->node[i]);   
			return t->node[i];
		}
	
		i = nextIdx( i, k, t->len);
	} while ( true );
}

/*
void st_remove ( symtab_t* t, char* sym )
{
	unsigned k = hash( sym);
	unsigned i = k % t->len;
	
	do {
		if ( t->sym[i] == 0 ){ 
			return;  // wasn't there
		}
	
		if ( equals( sym, t->sym[i]) ){
			if ( t->isCopy ) {
				free( t->sym[i]);
			}
			t->sym[i]  = 0;
			t->type[i] = 0;
		}
	
		i = nextIdx( i, k, t->len);
	} while ( 1 );
}
*/

