/*
 * Symbol management
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 * @author Peter Mehlitz
 * @author Philippe Laporte 
 * 
 * $tvt: symtab.c,v 1.3 2001/07/14 01:51:43 samc Exp $
 */

#include "symtab.h"
#include "xs_config.h"

#include <stdlib.h>
#include <string.h>

#ifdef KAFFELIB  //if used as a kaffe jni lib
#include <jni.h>
#endif


/* very simple open hashing */

/*************************************************************************
 * internal helpers
 */


static unsigned int hash ( char* sym )
{
	unsigned int h, g;

	for ( h=0; *sym; sym++ ){
		h = (h << 4) + *sym;
		if ( (g = h & 0xf0000000) ){
			h ^= (g >> 24);
			h = h ^ g;
		}
	}
	return h;
}

static inline int threshold ( int i ) {
	return (i * 80 / 100); 
}

//#include <stdio.h>
static unsigned  prime ( int n ) 
{	//fprintf(stderr, "In prime, n is %d\n", n);
	if ( n < 10 ) return 11;
	else if ( n < 15 ) return 17;
	else if ( n < 20 ) return 23;
	else if ( n < 30 ) return 31;	
	else if ( n < 45 ) return 47;
	else if ( n < 67 ) return 67;
	else if ( n < 100 ) return 101;
	else if ( n < 150 ) return 151;
	else return (n | 1); // cheat cheat cheat..
}

static inline int nextIdx ( int i, unsigned hash, unsigned len )
{
	i += (16 - (hash & 0xf));
	return (i % len);
}

static inline int equals ( char* s1, char* s2 )
{
	for ( ; *s1 == *s2; s1++, s2++ ){
		if ( *s1 == 0 ) return 1;
	}
	return 0;
}

static void grow ( symtab_t* t )
{
      	int i,  j, k;
	int newLen = prime( t->len * 3 / 2);
	char** newSym = (char**) calloc( newLen, sizeof( char*));
	char* newType = (char*) calloc( newLen, sizeof(char*));
	
        for ( j=0; j < t->len; j++ ) {
	       if ( t->sym[j] != 0 ) {
		        
			k = hash( t->sym[j]);
			i = k % newLen;
			
			do {
				if ( newSym[i] == 0 ){
					newSym[i] = t->sym[j];
					newType[i] = t->type[j];
					break;
				}
		
				i = nextIdx( i, k, newLen);
			} while ( 1 );
		}
	}
			
	free( t->sym);
	free( t->type);
	
	t->thresh = threshold( newLen);
	t->sym = newSym;
	t->type = newType;
        t->len = newLen;
}


/*************************************************************************
 * public APIs
 */

symtab_t* st_create ( int len, int isCopy )
{
	symtab_t *t = (symtab_t*) calloc( 1, sizeof( symtab_t));
	
	t->len    = prime(len * 3 / 2);
	t->thresh = threshold(t->len);
        t->sym    = (char**) calloc( t->len, sizeof(char*));
	t->type   = (char*) calloc(t->len, sizeof(char));
//	t->node   = (AST**) calloc(t->len, sizeof(AST*));
	t->count  = 0;
	t->isCopy = isCopy;
	
	return t;
}

#if 0

void st_reset ( symtab_t* t )
{
	int i;
	
	if ( t->count == 0 ){
		return; // nothing to do, it's empty anyway
	}
	
	for ( i=0; i < t->len; i++ ){
		if ( t->sym[i] ) {
			if ( t->isCopy ) {
				free( t->sym[i]);
			}
			t->sym[i]  = 0;
			t->type[i] = 0;
		}
	}
	
	t->count = 0;
}

void st_delete ( symtab_t* t )
{
	st_reset( t);
	free( t->sym);
	free( t->type);
}

#endif

void st_add ( symtab_t* t, char* sym, char type )
{
        
	
       	unsigned k = hash( sym);
	unsigned i = k % t->len;
	
        if ( t->count + 1 > t->thresh ) {
	   grow( t);
	   i = k % t->len;
	}

	do {
		if ( t->sym[i] == 0 ) {
		        t->sym[i]  = t->isCopy ? StrDup(sym) : sym;
		        t->type[i] = type;
			t->count++;
		        return;
		}
		else if ( equals( sym, t->sym[i]) ){
			return; // already there
		}

		i = nextIdx( i, k, t->len);
		
	} while ( true );
}

int st_contains ( symtab_t* t, char* sym )
{
	unsigned k = hash( sym);
	unsigned i = k % t->len;
	
	do {
		if ( t->sym[i] == 0 ){ 
			return 0;
		}
	
		if ( equals( sym, t->sym[i]) ){
			return t->type[i];
		}
	
		i = nextIdx( i, k, t->len);
	} while ( true );
}

/*
void st_remove ( symtab_t* t, char* sym )
{
	unsigned k = hash( sym);
	unsigned i = k % t->len;
	
	do {
		if ( t->sym[i] == 0 ){ 
			return;  // wasn't there
		}
	
		if ( equals( sym, t->sym[i]) ){
			if ( t->isCopy ) {
				free( t->sym[i]);
			}
			t->sym[i]  = 0;
			t->type[i] = 0;
		}
	
		i = nextIdx( i, k, t->len);
	} while ( 1 );
}
*/

/*************************************************************************
 * test stuff
 */

#ifdef TEST_SYMTAB

#include <stdio.h>
#include <string.h>

void st_dump( symtab_t* t )
{
	int i;
	
	printf( "len: %d, count: %d, thresh: %d\n", t->len, t->count, t->thresh);
	for ( i=0; i<t->count; i++ ){
		if ( t->sym[i] ){
			printf( "%3d: %8s (%u)\n", i, t->sym[i], hash(t->sym[i]));
		}
	}
}

int main ( int argc, char* argv[] )
{
	symtab_t *t = st_create( 10, 0);
	int i;

	char* a[] = { "one", "a", "gTriggerNum", "foo", "b", "stop", "i", "n",
	              "icon", "loc", "name", "zip", "isNice" };
		
	for ( i=0; i < (sizeof(a)/sizeof(a[0])); i++ ){
		st_add( t, a[i], T_PROP);
	}
	
	st_dump( t);
	
	st_contains( t, "one") ? puts( "has one") : puts( "no one");
	st_contains( t, "two") ? puts( "has two") : puts( "no two");
	
	st_remove( t, "one");
	
	return 0;
}

#endif
