/*
 * Code analysis utilities
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 * @author Peter Mehlitz
 * @author Philippe Laporte 
 * 
 * $tvt: util.c,v 1.10 2001/07/25 20:58:48 samc Exp $
 */

#include <stdio.h>
#include <string.h>

#include "xs_config.h"
#include "tokens.h"
#include "AParser.h" //small memory improvement, unused var
#include "util.h"

#include "ht.h"

/* 
 * hack in case JNIEXPORT is set to extern and we're already
 * inside the extern "C" {
 */
#define JNIEXPORT

/**************************************************************************
 * document type functions
 */

/*
 * return 1 if 'endTag' is supposed to be a valid 
 * end node of 'currentTag', 0 if not
 */
bool genericValidEndNode ( char* currentTag, char* endTag )
{
	return (strcasecmp( currentTag, endTag) == 0);
}

/*
 * return 0 if 'word' isn't a defined entity, "" if it expands to nil,
 * or otherwise its entity value
 */
char* genericExpand ( char* word )
{
	// 2DO - use a dynamically built table, here (PM)
	if ( !strcmp("test", word) ){
		return "MY TEST";
	}
	
	return 0;
}

DocType genericDocType = {
	0,     // we don't check parent-child relationship
	&genericValidEndNode,
	  
	&genericExpand
};

DocType *docType = &genericDocType;


int setDocType ( char* type )
{
	// a configurable registry of some sort would be probably better (PM)
	// (but the main point is to get it out of xs.g)
	if ( type ){
		if ( strcasecmp( type, "html") == 0 ){
			docType = &htmlDocType;
		}
	}
	else {
		docType = &genericDocType;
	}
}

/* generic utilities */

//need to rewrite since lib version never uses kaffe_malloc etc
extern "C"
char* StrDup(char* s)
{
  char* d;

  if(s == NULL)
    return NULL;

  d = (char*) calloc(strlen(s) + 1, sizeof(char));
  
  if(d != NULL)
    strcpy(d, s);
  
  return d;
}  
  


/****************************************************************************
 * operator method mapping
 */

char* binOp (ANTLRTokenType ttype)                                  /*{{{*/
{
    switch ( ttype ) {
	 case EQUAL:           return "eq";
	 case STRICT_EQUAL:    return "seq";
	 case STRICT_NOTEQUAL: return "snotEq";
	 case NOT_EQUAL:       return "notEq";
	 case LOG_NOT:         return "not";
	 case USHR_ASSIGN:     
	 case USHR:            return "ushr";
	 case SHR_ASSIGN:      
	 case SHR:             return "shr";
	 case GE:              return "ge";
	 case GT:              return "gt";
	 case SHL_ASSIGN:      
	 case SHL:             return "shl";
	 case LE:              return "le";
	 case TLT:             return "lt";
	 case MINUS_ASSIGN:   
	 case MINUS:           return "minus";
	 case PLUS_ASSIGN:    
	 case PLUS:            return "plus";
	 case MOD_ASSIGN:      
	 case MOD:             return "mod";
	 case MUL_ASSIGN:      
	 case MUL:             return "mul";
	 case DIV_ASSIGN:     
	 case DIV:             return "div";
	 case OR_ASSIGN:      
	 case LOG_OR:          return "logOr";
	 case OR:              return "or";
	 case AND_ASSIGN:      
	 case AND:             return "and";
	 case LOG_AND:         return "logAnd";
	 case XOR_ASSIGN:      
	 case XOR:             return "xor";
	
	 default:              return "";
    }  
} /*}}}*/


char* memberBinOp (ANTLRTokenType ttype)                               /*{{{*/
{
    switch ( ttype ) {
	 case USHR_ASSIGN:     return "ushrAssign";
	 case SHR_ASSIGN:      return "shrAssign";
	 case SHL_ASSIGN:      return "shlAssign";
	 case MINUS_ASSIGN:    return "minusAssign";
	 case PLUS_ASSIGN:     return "plusAssign";
	 case MOD_ASSIGN:      return "modAssign";
	 case MUL_ASSIGN:      return "mulAssign";
	 case DIV_ASSIGN:      return "divAssign";
	 case OR_ASSIGN:       return "logOrAssign";
	 case AND_ASSIGN:      return "andAssign";
	 case XOR_ASSIGN:      return "xorAssign";
	 
	 default:              return "";
    }  
} /*}}}*/

char* prefixOp ( ANTLRTokenType ttype )                                /*{{{*/
{
	switch ( ttype ){
	 case INC:            return "inc";
	 case DEC:            return "dec";
	 case TYPEOF:         return "xsGetType";
	 case VOID:           return "_void";
	 case NOT:            return "not";
	 case LOG_NOT:        return "logNot";
	 
	 default:             return "";
	}		
}
/*}}}*/

char* postfixOp ( ANTLRTokenType ttype )                               /*{{{*/
{
	switch ( ttype ){
	 case INC:            return "postInc";
	 case DEC:            return "postDec";
	 
	 default:             return "";
	}
}
/*}}}*/




/*****************************************************************************
 * parser auxiliary funcs
 */

void closeParen ()                                                      /*{{{*/
{
    for ( ;openParen; openParen-- ){
        out(")");
    }
    rawLiteral = false;
}


static void printVersion ()
{
	printf("%s v%s\n", PACKAGE, VERSION);
        printf("Copyright (c) 2001\nTransvirtual Technologies, Inc.  All rights reserved\n");
        printf("Compiled by  : %s\n", BUILD_USER);
        printf("Compiled on  : %s (%s)\n", BUILD_HOST,BUILD_ARCH);
        printf("Compile date : %s\n", BUILD_DATE);
#if defined(__GNUC__) & defined(__GNUC_MINOR__)
        printf("GCC Version  : %d.%d\n", __GNUC__, __GNUC_MINOR__);
#endif
        printf("Options      : %s\n", BUILD_OPTIONS);

}

static void printUsage ()
{
	printf("%s v%s\n", PACKAGE, VERSION);
	printf("Usage: xs [options] [<class_name> <input_file> <base_class> <factory_class> <imports>]\n");
	printf("Options:\n");
	printf("   -s   - script-only mode\n");
	printf("   -v   - print version information\n");
	printf("   -h   - display this message\n");
}

/*****************************************************************************
 * compiler driver
 */
 
#ifdef KAFFELIB


#include <setjmp.h>
#include <string.h>

//for exception handling
jmp_buf jmpbuffer;

JNIEnv* Genv;


inline void 
CheckException(){
	if(Genv->ExceptionOccurred())
	        longjmp(jmpbuffer, 1);
        
}



#define BUFFER_SIZE 4096

class DllExportPCCTS DLGJavaInputStream : public DLGInputStream {
 private:
        bool found_eof;
	
        jobject is;
	jclass  isclass;
	jmethodID read;
	JNIEnv* jni_env;
	
	jarray jbuf;    //Java data buffer
	
	jbyte buf[BUFFER_SIZE]; // data buffer
	
	jint buf_size;  //amount of data read
		
        short buf_idx; // index of next location to be read

	
 public:
        DLGJavaInputStream(JNIEnv* env, jobject istream) { 
	        is = istream; 
		
		isclass = env->GetObjectClass(istream); CheckException();
 
                read = env->GetMethodID(isclass, "read",  "([B)I"); CheckException();
		
		jbuf = env->NewByteArray((jsize)BUFFER_SIZE); CheckException();
		
		jni_env = env;
		
		found_eof = false;
		
		buf_size = 0;
		
		buf_idx = 0;  
        }
	
        	
	int nextChar() {
	        		
	        if ( found_eof ) 
		        return EOF;
                else {
		        if(buf_idx < buf_size){
			  return (int) (unsigned char) buf[buf_idx++];   
			}
			else{
			
                          buf_size = jni_env->CallIntMethod(is, read, jbuf); CheckException();
	
			  if (buf_size == EOF){ 
			    found_eof = true;
		            return EOF;
			  }	
				  
		          jni_env->GetByteArrayRegion(jbuf, 0, buf_size, buf); CheckException();			
					   
			  buf_idx = 0;
			  
			  return (int) (unsigned char) buf[buf_idx++]; 
		        }
		}
        }
};

#endif  


  
class XsLexer : public DLGLexer {
 
 public:
  
 
  XsLexer (DLGFileInput *input) : DLGLexer(input) {}
  
  XsLexer (DLGStringInput *input) : DLGLexer(input) {}

#ifdef KAFFELIB  
  
  XsLexer (DLGJavaInputStream *input) : DLGLexer(input) {}
  
#endif  
  
  
  void errstd(const char* msg){
     warn("%s, %s, ignored", msg, lextext()); // give line number ???
  }
  
      
  _ANTLRTokenPtr getToken() {
       
        if ( token_to_fill==NULL ) 
	  panic("NULL token_to_fill");

	ANTLRTokenType tt = nextTokenType();
	
        _ANTLRTokenPtr tk = token_to_fill->makeToken(tt, _lextext,_line);
	
        if(tt == SC && NoLineTerminatorHere)
	  NoLineTerminatorHere = false;
		
	if(tt == NEWLINE && NoLineTerminatorHere){
	  tk = token_to_fill->makeToken(SC, ";" ,_line);
	  NoLineTerminatorHere = false;
	}  
		
	return tk;
  }


};




int  nImp;
char* srcName = 0; // input file
char* imp[16];    // maximum 16 imports
    
    
inline void parseOptions(int argc, char* argv[]){

    int i;
       
    // parse options
    for (i=1; i<argc; i++){
        if ( !strcmp("-s",argv[i]) ){
            // script-only mode (testing purposes)
            scriptOnly = true;
        } else if ( !strcmp("-v",argv[i] ) ){
            printVersion();
            exit(0);
        } else if ( !strcmp("-h",argv[i] ) ){
            printUsage();
            exit(0);
	} else if ( !strcmp("-f",argv[i] ) ){
	    functionOnly = true;  //implies scriptOnly
        } 
	else break;
    }

    // class name
    if ( i < argc ){
        clsName = argv[i];
        i++;
    }

    // output file name (or stdout)
    if ( i < argc ) {
        if ( strcmp("-",argv[i]) != 0 ){
            srcName = argv[i];
	}
        i++;
    }

    // base class name
    if ( i < argc ){
        baseName = argv[i];
        i++;
    }

    // DOM tree factory class
    if ( !scriptOnly && (i < argc) ){
        factoryName = argv[i];
        i++;
    }

    // imports
    for ( nImp=0; (i < argc) && (nImp < (sizeof(imp)/sizeof(imp[0]))); i++, nImp++ ){
        imp[nImp] = argv[i];
    }
}



//need handles 
DLGInputStream* in;
XsLexer* scanner;
XsParser* parser;



int compile (  )
{
    int ret_signal;
    
    ANTLRTokenBuffer pipe(scanner);
    ANTLRToken tok;
    parser = new XsParser(&pipe);
    scanner->setToken(&tok);
    
    if ( scriptOnly ){
      scanner->mode( scanner->SCRIPT_MODE);
    }
    
    
    parser->init();
    
    //---- our own setup starts here
    initVars();
    initLabels();
    
    //---- set up global XsObject props/funcs
        
    addVar( "top", T_PROP);
    
    addVar( "Array", T_FUNC);
    addVar( "Date", T_FUNC);
    
    //addVar( "String", T_FUNC);
    addVar( "Number", T_FUNC);
    addVar( "Boolean", T_FUNC);
    addVar( "Object", T_FUNC);
    addVar( "Function", T_FUNC);
    
    bmain = create_sbuf( 1024);
    binit = create_sbuf( 256);
    btmp  = create_sbuf( 128);
    
    bfor = create_sbuf( 1024);
    blocals = create_sbuf( 256);
    
    
    tmp_stack = create_sbuf_stack(10);
    cur_stack = create_sbuf_stack(10);
    
    if ( !scriptOnly ){
        bdom  = create_sbuf( 1024);
    }
    
    //---- prep main buffer
    set_sbuf( bmain);
    out( "import kaffe.xs.*;"); ln();
    for ( int i=0; i<nImp; i++ ){
        out( "import %s;", imp[i]); ln();
    }
    
    lout( "public class %s extends %s\n{", clsName, baseName);
    
    if(!functionOnly){
      lout( "static %s Def;", clsName);

      //----  prep init buffer
      set_sbuf( binit);
      out( "public %s ()\n{", clsName);
      incInd();
      lout("Def = this;");
      if (scriptOnly)
        lout("reset();");
      decInd();
      lout("}");
      ln();
      lout("public static void main (String argv[])\n{");
      incInd();
      lout("new %s();", clsName);
      decInd();
      lout("}");
      ln();
      lout( "public void reset ()\n{");
      incInd();
      lout( "XsObject xThis = Def;");
      lout( "XsObject _def = Def;");
    }
    
    //---- compile
    if ( scriptOnly ){
        parser->script(&ret_signal);
    }
    else {
        parser->document(&ret_signal);
    }
    
   /* if ( scriptOnly ){
        parser->script();
    }
    else {
        parser->document();
    }
*/
    
    // TODO check ret_signal for exceptions 
     
     
    //---- finish init buffer
    if ( !scriptOnly ){
	lnb(binit); lnb(binit); outb( binit, "initDocument();");
    }
    
    if(!functionOnly){
      decIndb( binit); lnb(binit);
      outb( binit, "}");
    }
    
    //---- and merge it in
    lnb( bmain); lnb( bmain);
    mergeb( bmain, binit);

    if ( !scriptOnly ) {
        lnb( bmain); lnb( bmain);
        mergeb( bmain, bdom);
    }
    //this does not work yet
    lnb( bmain); outb( bmain, "}\n");
    write_sbuf( bmain, srcName);

    delete_sbuf( bmain);
    delete_sbuf( binit);
    delete_sbuf( btmp);
    delete_sbuf( bfor);
    delete_sbuf( blocals);
    
    if ( !scriptOnly ){
        delete_sbuf( bdom);
    }
    
    deleteVars();
    deleteLabels();
}




#ifndef KAFFELIB

int main ( int argc, char* argv[] )                                     
{
 parseOptions(argc, argv);
 
 DLGFileInput* fileinput = new DLGFileInput(stdin);  //avoid casting
 in = fileinput;
 scanner = new XsLexer(fileinput);
 
 compile();
}

#endif




// Java-JNI interface

#ifdef KAFFELIB

//for interp
jobject Gcontext;  // The Ctx object
jmethodID Gmethod_args; // The call method in Ctx.call
jmethodID Gmethod_no_args; 
jobject GxThis; // the this object


#define CheckLocalException() if(Genv->ExceptionOccurred()) {Genv = 0; return (jboolean) false;}


extern "C" {

JNIEXPORT jboolean JNICALL
Java_kaffe_xs_XsEngine_compile(JNIEnv* env, 
                                 jclass clazz, 
                                 jobject inputstream,
                                 jarray cmda)
{
 //save handle aside
 Genv = env;
 
 //reset global variables
 
 scriptOnly = false;    
 functionOnly = false;  
 enterScript = false;
 sconstDelim = 0;
 inEventScript = false;
 inAttribute = false;
 wasTextMode = false;
 openParen = 0;
 rawLiteral = false;    
 withLevel = 0;             
 statCall = false;  
 //isHtml = false; 
 handles_number = 0;  
 bcur = 0;
 levelLabels = 0;
 NoLineTerminatorHere = false;
 SrcInputStream = NULL; 
 
 dummy_declared = false;
 dummy_to_declare = false;
 
 memberLeftOp = false;

 //default values
 clsName     = "MyObject";
 baseName    = "XsObject";
 factoryName = "DocumentFactory";

 
 //read in the args: build native argv
 
 jsize cmdc = env->GetArrayLength(cmda); CheckLocalException();

 char** args = (char**) calloc(cmdc, sizeof(char*));

 jstring* strings = (jstring*) calloc(cmdc, sizeof(jstring)); // array of objects saved aside for later release
 
 
 for(jsize i = 0; i < cmdc; i++){
   strings[i] = env->GetObjectArrayElement(cmda, i); CheckLocalException();
   args[i] =  (char*) env->GetStringUTFChars(strings[i], NULL); CheckLocalException();
 }
 
 
 
 parseOptions(cmdc, args);    
  
 
 DLGJavaInputStream* isinput;
 
 //exception processing setup
 if(setjmp(jmpbuffer) != 0) 	
   goto GOT_EXCEPTION;

  	
 
 //setup the compiler
 isinput = new DLGJavaInputStream(env, inputstream);  //avoid casting
 in = isinput;
 scanner = new XsLexer(isinput);
 
 //compile
 compile();
 
 
 //done
 
 //release handle
 Genv = 0;
 
 //release strings    causes NullpointerException. If can't fix, do not use strings array
  //for(jsize i = 0; i < cmdc; i++){
  // env->ReleaseStringUTFChars(strings[i], args[i]); CheckLocalException();
  //}
 
 return (jboolean) true;  //success
 
 
 GOT_EXCEPTION:Genv = 0; 
 
               	      
               //release strings see above remark
               //for(jsize i = 0; i < cmdc; i++){
                // env->ReleaseStringUTFChars(strings[i], args[i]);CheckLocalException();
               //} 
  
               return (jboolean) false;  //failure
 
}


}//extern "C" 



static inline jobject 
interpretScript(char* script){

        int ret_signal;
	
	//reset global variables
	sconstDelim = 0;
        NoLineTerminatorHere = false;
 
	DLGStringInput* stringinput = new DLGStringInput(script);  //avoid casting
        in = stringinput;
        scanner = new XsLexer(stringinput);
	ANTLRTokenBuffer pipe(scanner);
        ANTLRToken tok;
        parser = new XsParser(&pipe);
        scanner->setToken(&tok);
        scanner->mode( scanner->SCRIPT_MODE);
        parser->init();
	
	return parser->call(&ret_signal);
}


#undef  CheckLocalException
#define CheckLocalException() if(Genv->ExceptionOccurred()) { Gcontext = 0; Genv = 0; Gmethod_args = 0; Gmethod_no_args = 0; GxThis = 0; return NULL;}


extern "C" {


JNIEXPORT jobject JNICALL
Java_kaffe_xs_XsEngine_interp(JNIEnv* env, 
                              jclass clazz, 
                              jobject xsobject,
                              jobject xthis, 
                              jstring inscript)
{
 //save handles aside
 Gcontext = xsobject;
 Genv = env;        
 GxThis = xthis;

 jobject ret; // value returned by call (expression)
 

 char* script = (char*) env->GetStringUTFChars(inscript, NULL); CheckLocalException();  
    	
 jclass xsobjectclass = env->GetObjectClass(xsobject); CheckLocalException();

 
 jmethodID mid_args = env->GetMethodID(xsobjectclass, "call",  "(Ljava/lang/Object;Lkaffe/xs/XsArray;)Lkaffe/xs/XsObject;"); CheckLocalException();

 jmethodID mid_no_args = env->GetMethodID(xsobjectclass, "call",  "(Ljava/lang/Object;)Lkaffe/xs/XsObject;"); CheckLocalException();


 //save handles aside
 Gmethod_args = mid_args;
 Gmethod_no_args = mid_no_args;

 //exception processing setup
 if(setjmp(jmpbuffer) != 0)
 	goto GOT_EXCEPTION;
 
 
 ret = interpretScript(script);
 
 
 env->ReleaseStringUTFChars(inscript, script); //not released if get exception, but only way since call resets exception status


 //lose the references
 Gcontext = 0;
 Genv = 0;        
 Gmethod_args = 0;
 Gmethod_no_args = 0;
 GxThis = 0;


 return ret;  //success
 

 GOT_EXCEPTION:
 
     //lose the references
     Gcontext = 0;
     Genv = 0;        
     Gmethod_args = 0;
     Gmethod_no_args = 0;
     GxThis = 0;

     return NULL;  //failure
}


} // extern "C"


#endif  // KAFFELIB


void warn ( char* fmt, ...)                                              
{
    va_list ap;
    
    va_start( ap, fmt);
                                         // was LT(1)->getLine()
    fprintf( stderr, "%s:%d: ", clsName, scanner->line());
    vfprintf( stderr, fmt, ap);
    fputc( '\n', stderr);
    va_end( ap);    
}



//System.out.println

//printf("System.out.println:\n");
   //jclass system = env->FindClass("java/lang/System"); CheckException(); 
   //jfieldID outid = env->GetStaticFieldID(system, "out","Ljava/io/PrintStream;"); CheckException();
   //jobject out =  env->GetStaticObjectField(system,outid);CheckException();
   //jclass printstream = env->FindClass("java/io/PrintStream");CheckException(); 
   //jmethodID println =  env->GetMethodID(printstream, "println","(Ljava/lang/Object;)V");CheckException();
   //env->CallObjectMethod(out, println, strings[i]); CheckException();
