/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Build.java,v 1.10 2001/11/09 00:57:14 samc Exp $
 */ 
package org.xoe.core.packages;

import org.xoe.core.Engine;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathParser;
import org.xoe.core.xpath.XPath;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.io.IOException;
import java.io.StringWriter;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * The java portion of the XOE package build system.
 */
class Build
{
  static final String PACKAGE_ARCHIVE = "stash:/packages";

  /**
   * Main function that gets called by the XOE package building
   * scripts.  Expects a single argument containing the filename
   * of package.  The output is a new package.xml file.
   */
  static public void main (String args[]) 
    throws Exception
  {
    if (args.length != 1)
      {
        System.err.println ("usage: " + Build.class.getName () + " <package-in-file>");
        System.exit (1);
      }

    boolean performInstall = true;
    if ("yes".equals (System.getProperty ("xoe.fastbuild")))
      {
        performInstall = false;
      }
    long timer = System.currentTimeMillis ();
    
    if (performInstall)
      {
        Engine.bootstrap ();
      }
    else
      {
        ServiceLocator.registerService (new ResourceInstaller ());
        ServiceLocator.registerService (new BaseInstaller ());
        ServiceLocator.registerService (new ClassInstaller ());
        ServiceLocator.registerService (new NativeInstaller ());
        ServiceLocator.registerService (new ServiceInstaller ());
        ServiceLocator.registerService (new BindingInstaller ());
        ServiceLocator.registerService (new MimeTypeInstaller ());
        Core.calculateOrderedHandlers ();
      }

    XoePackage     pkg    = new XoePackage (args[0]);
    StringWriter   sw     = new StringWriter ();
    if (pkg.getName ().endsWith ("-native"))
      {
        String arch = System.getProperty ("xoe.hostarch");
        if (arch == null)
          {
            System.err.println ("xoe.hostarch must be defined in order to build a native package");
            System.exit (1);
          }
        pkg.getDescription ().setArchitecture (arch);
      }

    // We only need to do this if we have explicit a dependancy 
    // whose Installer is not part of the core.
    // install required services
    if (performInstall)
      {
        installPackage (pkg);
      }

    Document doc = generatePackageXML (pkg);

    // log the time it took to do generatePackageXML ()
    timer = System.currentTimeMillis () - timer;
    Logger.log (Build.class, "package.xml generation took " + timer + " ms.");

    // write the XML to System.out
    System.out.println (XMLWriter.getXML (doc));
    System.exit (0);
  }

  /**
   * This will install all the packages required to fulfill the dependancies of the given
   * XoePackage.
   * Throws Exceptions when the cannot be met by the package manager or installtion
   * fails.
   */
  private static void installPackage (XoePackage pkg)
    throws Exception
  {
    PackageDescription desc = pkg.getDescription ();
    IPackageQuery[] deps = desc.getDependencies ();
    int len = deps.length;

    if (len == 0) { return; }

    // install each dependancy
    for (int i = 0; i < len; i++)
      {
        Logger.log (Build.class, "Looking for: " + deps[i].toXPath ());
        SearchResults res = PackageManager.search (deps[i]);
        if (res.getInstalledPackages ().length != 0)
          {
            Logger.log (Build.class, "Already installed: " + deps[i].toXPath ());
            continue;
          }
        InstallOption[] options = res.getInstallablePackages ();
        if (options.length > 0)
          {
            Logger.log (Build.class, "Installing: " + options[0]);
            PackageManager.commit (options[0]);
          }
        else
          {
            throw new Exception ("No installable packages for: " + deps[i].toXPath ());
          }
      }

    // don't install native package in case we are cross-compiling
    if (pkg.getDescription ().getArchitecture () == null)
      {
        PackageManager.installPackage (pkg);
      }
  }

  /**
   * Helper function to append vectors
   */
  private static void append (Vector v1, Vector v2)
  {
    if (v2 == null)
      {
	return;
      }

    for (int i = 0; i < v2.size (); i++)
      {
        v1.addElement (v2.elementAt (i));
      }
  }

  /**
   * Calculate info, provides, depends and conflicts for a package
   * and generate the "package.xml" Document.
   */
  private static Document generatePackageXML (XoePackage pkg)
    throws PackageException
  {
    Vector provides  = new Vector ();
    Vector depends   = new Vector ();
    Vector conflicts = new Vector ();

    // build vectors of all the <provides/>, <conflicts/>, and <requires/>
    // according to the installed PackageXMLHandler's.
    
    IService[] handlers = ServiceLocator.findImplementors (IPackageXMLHandler.INTERFACE);

    for (int i = 0; i < handlers.length; i++)
      {
        IPackageXMLHandler handler = (IPackageXMLHandler) handlers[i];
	handler.setPackage (pkg);
	append (provides , handler.generateProvides  ());
	append (depends  , handler.generateRequires  ());
	append (conflicts, handler.generateConflicts ());
      }

    // create the top level package.xml tag.
    Document doc = new XDocument ();

    Element packageE = doc.createElement ("package");
    doc.appendChild (packageE);

    packageE.setAttribute ("name", pkg.getName ());
    packageE.setAttribute ("version", pkg.getVersion ().toString ());

    String arch = pkg.getArchitecture ();
    if (arch != null)
      {
        packageE.setAttribute ("architecture", arch);
      }

    Element inputDoc = pkg.getDescription ().getDocument ().getDocumentElement ();

    if ("true".equals (inputDoc.getAttribute ("task")))
      {
	packageE.setAttribute ("task", "true");
      }

    // copy over the "info" section
    Element info = (Element) XPath.findFirstNode (inputDoc, "info");
    if (info != null)
      {
        packageE.appendChild (doc.importNode (info, true));
      }

    Element providesE = doc.createElement ("provides");
    for (int i = 0; i < provides.size (); i++)
      {
        Proviso proviso = (Proviso) provides.elementAt (i);

	Element provisoE = doc.createElement ("proviso");
	provisoE.setAttribute ("name", proviso.getLocalName ());
	provisoE.setAttribute ("ns", proviso.getNamespaceURI ());
	Node metadata = (Node) proviso.getMetadata ();
	if (metadata != null)
	  {
	    provisoE.appendChild (doc.importNode (metadata, true));
	  }
	providesE.appendChild (provisoE);
      }

    packageE.appendChild (providesE);

    XPathContext ctx = new XPathContext ();
    ctx.pushNode (doc);
    ctx.setFunctionProvider (PackageXPathFunctionProvider.singleton);
    
    Element requiresE = doc.createElement ("requires");
    packageE.appendChild (requiresE);
    for (int i = 0; i < depends.size (); i++)
      {
	Dependency dep = (Dependency) depends.elementAt (i);
	PackageQuery q = new PackageQuery ();
	q.addDependency (dep);
	NodeList nl = XPath.findNodes (ctx, q.toXPath ());
	if (nl.getLength () > 0)
	  { // don't add this dependency:
	    // it's satisfied by this package already!
	    Logger.log (Build.class, "Ignoring met dependency: " + q.toXPath ());
	    continue;
	  }
	requiresE.appendChild (dep.getDepElement (doc));
      }

    Element conflictsE = doc.createElement ("conflicts");
    packageE.appendChild (conflictsE);

    for (int i = 0; i < conflicts.size (); i++)
      {
	Dependency dep = (Dependency) conflicts.elementAt (i);
	conflictsE.appendChild (dep.getDepElement (doc));
      }

    return doc;
  }
}
