package java.awt;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;
import kaffe.awt.theme.ComponentDeco;

/**
 * class Button - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Button
  extends Component
  implements ShortcutConsumer
{
	final private static long serialVersionUID = -8774683716313001058L;
	String label;
	ActionListener aListener;
	String aCmd;
	FontMetrics fm;
	MenuShortcut shortcut;
	int state;
	static int PUSHED = 1;
	static int HILIGHTED = 2;
	static int named;

public Button () {
	this( "");
}

public Button ( String label) {
	cursor = Cursor.getPredefinedCursor( Cursor.HAND_CURSOR);
	flags |= IS_MOUSE_AWARE;
	setLabel( label);
}

public void addActionListener ( ActionListener a) {
	aListener = AWTEventMulticaster.add( aListener, a);
}

public void addNotify () {
	super.addNotify();

	if ( shortcut != null )
		ShortcutHandler.addShortcut( shortcut, getToplevel(), this);
		
	MenuShortcut s1 = new MenuShortcut( this, KeyEvent.VK_ENTER, 0);
	MenuShortcut s2 = new MenuShortcut( this, KeyEvent.VK_SPACE, 0);
	ShortcutHandler.addShortcut( s1, this, this);
	ShortcutHandler.addShortcut( s2, this, this);
}

void animate() {
	if ( Defaults.BtnAnimation ) {
		Graphics g = getGraphics();
	
		state |= PUSHED;
		paint( g);

		Toolkit.tlkSync();
		try { Thread.sleep( 100); }
		catch ( Exception _x) {}

		state &= ~PUSHED;
		paint( g);
		Toolkit.tlkSync();
	
		g.dispose();
	}
}

void checkMouseAware () {
	// we are always mouse aware if not disabled
	if ( (eventMask & AWTEvent.DISABLED_MASK) == 0 ){
		flags |= IS_MOUSE_AWARE;
	}
	else {
		flags &= ~IS_MOUSE_AWARE;
	}
}

void drawText ( Graphics g) {
	Color c1 = null, c2;
	int x, y;
	
	if ( label == null )
		return;
	
	x = (width - fm.stringWidth( label)) / 2;
	y = height - (height - fm.getHeight()) / 2 - fm.getDescent();

	// If button is disabled then the text as such.
	if (!isEnabled()) {
		c2 = bgClr.darker();
	}
	else if ( Defaults.BtnAnimation && ((state & PUSHED) > 0) ){
		x--; y--;
		c1 = Color.yellow;
		c2 = Color.red;
	}
	else {
		if ( (state & HILIGHTED) > 0 ) {
			c1 = Defaults.BtnTxtCarved ? bgClr.brighter() : null;
			c2 = (Defaults.BtnPointTxtClr != null) ? Defaults.BtnPointTxtClr : fgClr;
		}
		else {
			c1 = Defaults.BtnTxtCarved ? bgClr.brighter() : null;
			c2 = fgClr;
		}
	}

	if ( c1 != null ){
		g.setColor( c1);
		g.drawString( label, x+1, y+1);
	}
	
	g.setColor( c2);
	g.drawString( label, x, y);	
}

public String getActionCommand () {
	return ( (aCmd != null) ? aCmd : label);
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

int getDecoFlags() {
	int flags = 0;
	if ( (state & PUSHED) > 0 )
		flags |= ComponentDeco.CS_ACTIVE;
	if ( (state & HILIGHTED) > 0 )
		flags |= ComponentDeco.CS_CURSOR_OVER;
	return flags;
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "button" + (named++);
}

public String getLabel() {
	return label;
}

public void handleShortcut( MenuShortcut ms) {
	int mods = (ms != null) ? ms.mods : 0;

	if ( (state & PUSHED) == 0 )
		animate();
		
	if ( (aListener != null) ||
	     ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) ||
	     ((flags & IS_OLD_EVENT) != 0) ) {
		Toolkit.eventQueue.postEvent( ActionEvt.getEvent( this, ActionEvent.ACTION_PERFORMED,
		                                                  getActionCommand(), mods));
	}
}

public void paint ( Graphics g) {
	drawText( g);
}

protected String paramString() {
	return (super.paramString() + ",Label: " + label);
}

public Dimension preferredSize () {
	Dimension d = new Dimension( insets.left + insets.right, insets.top + insets.bottom);
	String ms = (label != null ) ? label : "xxx";
	FontMetrics mfm = (fm != null) ? fm : Defaults.WndFontMetrics;
	
	d.width += mfm.stringWidth( ms) * 4 / 3; 
	d.height += mfm.getHeight() * 3 / 2;

	return d;
}

void process ( ActionEvent e ) {
	if ( (aListener != null) || ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) ){
		processEvent( e);
	}

	if ( (flags & IS_OLD_EVENT) != 0 ) postEvent( Event.getEvent( e));
}

void process ( FocusEvent e ) {
	repaint();
	super.process( e);
}

protected void processActionEvent( ActionEvent e) {
	if ( aListener != null ) {
		aListener.actionPerformed( e);
	}
}

void processMouse ( MouseEvent e ) {
	switch ( e.id ) {
		case MouseEvent.MOUSE_ENTERED:
			state |= HILIGHTED;
			repaint();
			break;
		case MouseEvent.MOUSE_EXITED:
			state &= ~HILIGHTED;
			repaint();
			break;
		case MouseEvent.MOUSE_PRESSED:
			state |= PUSHED;
			if ( AWT.keyTgt != Button.this )
				requestFocus();
			else
				repaint();
			break;
		case MouseEvent.MOUSE_RELEASED:
			if ( contains( e.getX(), e.getY()))
				handleShortcut( null);
			state &= ~PUSHED;
			repaint();
			break;
	}

	super.processMouse( e);
}

void propagateFont( Font fnt) {
	super.propagateFont( fnt);
	fm = getFontMetrics( fnt);
	
	if ( (flags & IS_SHOWING) > 0 )
		repaint();
}

public void removeActionListener ( ActionListener a) {
	aListener = AWTEventMulticaster.remove( aListener, a);
}

public void removeNotify () {
	if ( shortcut != null )
		ShortcutHandler.removeFromOwner( getToplevel(), shortcut);
	ShortcutHandler.removeShortcuts( this);
	super.removeNotify();
}

public void setActionCommand ( String aCmd) {
	this.aCmd = aCmd;
}

public void setEnabled( boolean isEnabled) {
	super.setEnabled( isEnabled);
	repaint();
}

public void setLabel ( String label) {

	if ( label != null ) {
		int ti = label.indexOf( '~');

		if ( ti > -1){
			this.label = label.substring( 0, ti) + label.substring( ti+1);
			shortcut = new MenuShortcut( this, Character.toUpperCase( label.charAt(ti+1)), KeyEvent.ALT_MASK);
		}
		else {
			this.label = label;
		}
	}
	else {
		this.label = null;
	}

	if ( isShowing() ){
		repaint();
	}
}
}
